(* (** remove printing ~ *) *)

(** This module proves the Canonical Forms Lemmas, which allow us
    to retrieve the shape of an item given its type. *)

Set Implicit Arguments.

Require Import Coq.Program.Equality.
Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping
        RecordTypes InertContext RecordDefs PreciseTyping TightTyping InvertibleTyping
        GeneralToTight Subenvironments Weakening Narrowing Visibility
        Substitution HeapCorrespondence InertTightSubtyping EnvCorrespondence TFree ConfigVarkinds ReferenceTypes.

(** * Canonical forms lemmas *)

Section WithTypingMode.
Context (typing_mode: Mode).

(** * Functions under Inert Contexts *)
(** [inert G]            #<br>#
    [G ⊢ x∶ forall(T)U]       #<br>#
    [――――――――――――――-]    #<br>#
    [exists T', U',]          #<br>#
    [G(x) = forall(T')U']     #<br>#
    [G ⊢ T <: T']        #<br>#
    [forall fresh y, G, y: T ⊢ U'^y <: U^y] *)
Lemma var_typ_fun_to_binds: forall G e w T U,
    inert G -> sealed G -> ok e -> renv_corr G e ->
    G⁏e ⊢ w ∶ typ_fun T U ->
    (exists L T' U',
        ctx_binds w (typ_fun T' U') G /\
        G⁏e ⊢ T <: T' /\
        (forall x, x \notin L -> G § x / v_par ~ T ⁏e ⊢ (open x U') <: (open x U))).
Proof.
  introv Hin Hcl Hok Hcor Ht.
  forwards Htt: general_to_tight_typing Ht; fweauto.
  forwards Hinv: tight_to_invertible Htt; fweauto.
  forwards Hpf: invertible_to_precise_typ_fun Hinv; fweauto.
  destructc Hpf as T' U' V' L Htp Hs1 Hs2.
  exists L T' U'. splits.
  - apply* inert_precise_fun_inv.
  - apply~ tight_to_general.
  - assumption.
Qed.

(** * Objects under Inert Contexts *)
(** [inert G]                    #<br>#
    [G ⊢ x∶ {a: T}]              #<br>#
    [―――――――――――――――――――――――]    #<br>#
    [exists S, T', G(x) = mu(S)]       #<br>#
    [S^x = ... /\ {a: T'} /\ ...]  #<br>#
    [G ⊢ T' <: T]                *)
Lemma var_typ_rcd_to_binds: forall G e x a T U,
    inert G -> sealed G -> ok e -> renv_corr G e ->
    G⁏e ⊢ x ∶ typ_rcd (dec_fld a T U) ->
    (exists S S' T',
        ctx_binds x (typ_objref S S') G /\
        record_has (open x S) (dec_fld a T' T') /\
        G⁏e ⊢ T <: T' /\
        G⁏e ⊢ T' <: U).
Proof.
  introv Hin Hcl Hok Hcor Ht.
  destructc (typing_implies_bound Ht) as S BiG.
  lets Htt: (general_to_tight_typing Hin Hcor Ht).
  lets Hinv: (tight_to_invertible Hin Hcor Htt).
  forwards Hpf: invertible_to_precise_fld_dec Hinv; fweauto.
  destructc Hpf as S' T' U' Htp Hs1 Hs2.
  destruct (pf_dec_fld_inv Hin Htp).
  destruct (pf_inert_rcd_U Hin Htp) as [?U' [?S' Hr]]. subst.
  epose proof (precise_flow_record_has Hin Htp (rd_fld _ _)) as Hr'. apply pf_binds in Htp.
  exists U'0 S'0 S'. splits; try assumption; apply* tight_to_general.
Qed.

Lemma var_typ_rcd_met_to_binds: forall G e x m TC TZ TR,
    inert G -> sealed G -> ok e -> renv_corr G e ->
    G⁏e ⊢ x ∶ typ_rcd (dec_met m TC TZ TR) ->
    (exists L S S' TC' TZ' TR',
        ctx_binds x (typ_objref S S') G /\
        record_has (open x S) (dec_met m TC' TZ' TR') /\
        G⁏e ⊢ TZ <: TZ' /\
        (forall x, x \notin L -> G § x / v_par ~ TZ ⁏e ⊢ (open x TC) <: (open x TC')) /\
        (forall x, x \notin L -> forall y, y <> x -> y \notin L -> ctx_met G x TZ y TC ⁏e ⊢ (open_met x y TR') <: (open_met x y TR))
    ).
Proof.
  introv Hin Hcl Hok Hcor Ht.
  destruct (typing_implies_bound Ht) as [S BiG].
  lets Htt: (general_to_tight_typing Hin Hcor Ht).
  lets Hinv: (tight_to_invertible Hin Hcor Htt).
  forwards Hpf: invertible_to_precise_met_dec Hinv; fweauto.
  destruct Hpf as [L' [S' [TC' [TZ' [TR' [Htp [Hs1 [Hs2 [Hdp Hs3]]]]]]]]].
  destruct (pf_inert_rcd_U Hin Htp) as [?U' [?S' Hr]]. subst.
  lets Hr': (precise_flow_record_has Hin Htp (rd_met _ _ _ _)). apply pf_binds in Htp.
  exists L' U' S'0 TC' TZ' TR'. splits; try assumption; apply* tight_to_general.
Qed.

(** [inert G]                  #<br>#
    [G ⊢ v∶ mu(T)]             #<br>#
    [G ⊢ x∶ T^x]               #<br>#
    [T = ... /\ {a: U} /\ ...  ] #<br>#
    [――――――――――――――――――――――――] #<br>#
    [exists t, ds, v = nu(T)ds     ] #<br>#
    [ds^x = ... /\ {a = t} /\ ...] #<br>#
    [G ⊢ t∶ U] *)
(* For strongly bound variable of recursive type,
   there is object definitions on the heap, of that type opened *)
Lemma strong_mu_to_new: forall G e s x T S,
    inert G ->
    ty_item_s _ G e s x ->
    ctx_binds x (typ_objref T S) G ->
    exists ds,
      binds x (item_obj ds) s /\
      G §§ objctx_heap x T⁏e ⊢ ds ∶ open x T /\
      defs_varkinds G ds
.
Proof.
  introv Hi Hts Bi.
  inversions Hts; ctx_binds_eq.
  exists ds; split; auto.
Qed.

(** * Canonical Forms for Functions

    [inert G]            #<br>#
    [s: G]               #<br>#
    [G ⊢ x∶ forall(T)U]       #<br>#
    [――――――――――――――――――] #<br>#
    [s(x) = lambda(T')t] #<br>#
    [G ⊢ T <: T']        #<br>#
    [G, x: T ⊢ t∶ U]          *)
Lemma canonical_forms_fun_ref: forall G e s w y T U,
  inert G ->
  sealed G ->
  heap_correspond G e s ->
  binds w y e ->
  G⁏e ⊢ w ∶ typ_fun T U ->
  (exists L T' U' t,
    binds y (item_fun t) s /\
    (G⁏e ⊢ T <: T') /\
    all_tfree_k G t v_ref /\
    ctx_binds w (typ_fun T' U') G /\
    ctx_binds y (typ_fun T' U') G /\
    (forall x,
      x \notin L ->
      G § x / v_par ~ T ⁏e ⊢ open x t ∶ open x U) /\
    (forall x,
      x \notin L ->
      G § x / v_par ~ T ⁏e ⊢ open x U' <: open x U)
  ).
Proof.
  introv Hin Hcl Hst Hbi Hty.
  assert (ctx_ok G) as HokG by eauto.
  assert (ok e) as Hok by eauto.
  assert (renv_corr G e) as Hcor by eauto.
  destruct (var_typ_fun_to_binds Hin Hcl Hok Hcor Hty) as [L [S [T' [BiG [Hs1 Hs2]]]]].
  pose proof (corresponding_types_ref Hst Hbi BiG).
  destruct (renv_corr_binds_ref_to_loc HokG Hcor Hbi BiG) as [? [? [? Htcor]]].
  inversion Htcor; subst.
  inversions H; ctx_binds_eq.
  exists (L0 \u L \u ctx_dom G) S0 T0 t. splits; auto.
  intros.
  assert (Ht: G § x / v_par ~ S0 ⁏e ⊢ open x t ∶ open x T0) by auto.
  apply (narrow_ty_trm_subtyp Ht); auto 2.
  apply subctx_grow; auto 3.
Qed.

Lemma canonical_forms_fun: forall G e s w T U,
  inert G ->
  sealed G ->
  heap_correspond G e s ->
  ctx_k G w v_ref ->
  G⁏e ⊢ w ∶ typ_fun T U ->
  (exists y L T' U' t,
    binds w y e /\
    binds y (item_fun t) s /\
    ctx_binds w (typ_fun T' U') G /\
    ctx_binds y (typ_fun T' U') G /\
    (G⁏e ⊢ T <: T') /\
    all_tfree_k G t v_ref /\
    (forall x,
      x \notin L ->
      G § x / v_par ~ T ⁏e ⊢ open x t ∶ open x U) /\
    (forall x,
      x \notin L ->
      G § x / v_par ~ T ⁏e ⊢ open x U' <: open x U)
  ).
Proof.
  introv Hin Hcl Hst Hvk Hty.
  destruct (heap_correspond_exists_loc Hst Hvk) as [vLoc Hbi].
  exists vLoc.
  forwards Hcf: canonical_forms_fun_ref; fweauto.
  destructecs Hcf.
  exists L T' U' t.
  splits; auto.
Qed.

Lemma defs_has_fld_loc G ds a v:
  defs_varkinds G ds ->
  defs_has ds (def_fld a (avar_f v)) ->
  ctx_k G v v_loc.
Proof.
  introv Hvk Hhas.
  unfold defs_has in *.
  induction ds.
  - inversion Hhas.
  - unfold get_def in *. cases_if.
    + simplify_eq Hhas. intro. subst.
      inversions Hvk. inversions H3. auto.
    + inversions Hvk. auto.
Qed.

Lemma heap_met_varkinds G ds m t:
  defs_varkinds G ds ->
  defs_has ds (def_met m t) ->
  all_tfree_k G t v_ref.
Proof.
  introv Hvk Hhas.
  unfold defs_has in *.
  induction ds.
  - inversion Hhas.
  - unfold get_def in *. cases_if.
    + simplify_eq Hhas. intro. subst.
      inversions Hvk. inversions H3. auto.
    + inversions Hvk. auto.
Qed.


(** * Canonical Forms for Objects

    [inert G]            #<br>#
    [s: G]               #<br>#
    [G ⊢ x∶ {a:T}]       #<br>#
    [――――――――――――――――――] #<br>#
    [exists S, ds, t,] #<br>#
    [s(x) = nu(S)ds] #<br>#
    [ds^x = ... /\ {a = t} /\ ...] #<br>#
    [G ⊢ t∶ T] *)
Lemma canonical_forms_fld_ref: forall G e s vRef vLoc a S T,
  inert G ->
  sealed G ->
  heap_correspond G e s ->
  binds vRef vLoc e ->
  G⁏e ⊢ vRef ∶ (typ_rcd (dec_fld a S T)) ->
  (exists ds vVal,
    binds vLoc (item_obj ds) s /\
    defs_has ds (def_fld a (avar_f vVal)) /\
    G⁏e ⊢ vVal ∶ T /\
    G⁏e ⊢ S <: T /\
    ctx_k G vVal v_loc /\
    (exists Tg S1 S2,
    ctx_binds vRef (typ_objref Tg S1) G /\
    ctx_binds vLoc (typ_objref Tg S2) G /\
      exists Tf,
      record_has Tg /[-> vLoc] (dec_fld a Tf Tf) /\
      G⁏e ⊢ S <: Tf /\
      G⁏e ⊢ Tf <: T
    )
  ).
Proof.
  introv Hi Hcl Hst Hbi HtyRef.
  assert (ctx_ok G) as HokG by eauto.
  assert (ok e) as Hok by eauto.
  assert (renv_corr G e) as Hcor by eauto.
  assert ( G ⁏ e ⊢ vLoc ∶ typ_rcd (dec_fld a S T)) as HtyLoc.
  eapply ty_renv_binds_var; eauto.
  destructc (var_typ_rcd_to_binds Hi Hcl Hok Hcor HtyLoc) as ?S ?T' ?U HbiG Hrcd ?H ?H.
  destructc (var_typ_rcd_to_binds Hi Hcl Hok Hcor HtyRef) as ?S ?T' ?U HbiGRef ?Hrcd ?H ?H.
  pose proof (corresponding_types_ref Hst Hbi HbiGRef) as Hts.
  (*
  Note: this proof is kind of convoluted in how it applies var_typ_rcd_to_binds to both
  the location and reference separately.
  This code not used anymore:
  destruct (renv_corr_binds_ref_to_loc HokG Hcor Hbi HbiGRef) as [? [HbiGLoc [? Htcor]]].
  inversion Htcor; subst.*)
  destruct (strong_mu_to_new Hi Hts HbiG) as [?ds [?Bis [Htyds Hvkds]]].
  pose proof (record_has_ty_defs Htyds Hrcd) as [?d [? Htyd]].
  inversions Htyd.
  exists ds t. splits; eauto.
  -
  eapply defs_has_fld_loc; eauto.
  -
  assert (typ_corr (typ_objref S1 T'0) (typ_objref S0 T')) as Htcor.
  { eapply renv_corr_binds_ref_loc_corr; eauto. }
  assert (S0 = S1).
  { inversions Htcor; auto. }
  subst.
  exists S1 T'0 T'; splits; auto.
  exists U; splits; auto.
Qed.

Lemma canonical_forms_fld G e s vRef a S T:
  inert G ->
  sealed G ->
  heap_correspond G e s ->
  ctx_k G vRef v_ref ->
  G⁏e ⊢ vRef ∶ (typ_rcd (dec_fld a S T)) ->
  (exists vLoc ds vVal,
    binds vRef vLoc e /\
    binds vLoc (item_obj ds) s /\
    defs_has ds (def_fld a (avar_f vVal)) /\
    G⁏e ⊢ vVal ∶ T /\
    G⁏e ⊢ S <: T /\
    ctx_k G vVal v_loc /\
  
    (exists Tg S1 S2,
    ctx_binds vRef (typ_objref Tg S1) G /\
    ctx_binds vLoc (typ_objref Tg S2) G /\
      exists Tf,
      record_has Tg /[-> vLoc] (dec_fld a Tf Tf) /\
      G⁏e ⊢ S <: Tf /\
      G⁏e ⊢ Tf <: T
    )
  ).
Proof.
  introv Hin Hcl Hst Hvk Hty.
  destruct (heap_correspond_exists_loc Hst Hvk) as [vLoc Hbi].
  exists vLoc.
  destruct (canonical_forms_fld_ref Hin Hcl Hst Hbi Hty) as [s' [vVal Hrest]].
  exists s' vVal.
  auto.
Qed.

(* If a reference variable has a method type,
   then the corresponding location points to an object item,
   which contains definitions which contain a method definition,
   where the body only contains reference free-variables
   and the body has the expected type *)
Lemma canonical_forms_met_ref: forall G e s vRef vLoc m TC TZ TR,
  inert G ->
  sealed G ->
  heap_correspond G e s ->
  binds vRef vLoc e ->
  G⁏e ⊢ vRef ∶ (typ_rcd (dec_met m TC TZ TR)) ->
  (exists L ds t S S1 S2,
    ctx_binds vRef (typ_objref S S1) G /\
    ctx_binds vLoc (typ_objref S S2) G /\
    binds vLoc (item_obj ds) s /\
    defs_has ds (def_met m t) /\
    (exists TC0 TZ0 TR0,
      record_has S /[-> vLoc] (dec_met m TC0 TZ0 TR0) /\
      G ⁏ e ⊢ TZ <: TZ0 /\
      (forall x : var, x \notin L -> G §ᶻ x ~ TZ ⁏ e ⊢ TC /[-> x] <: TC0 /[-> x]) /\
      (forall x : var, x \notin L -> forall y : var, y <> x -> y \notin L -> ctx_met G x TZ y TC ⁏ e ⊢ TR0 /[# 1 -> y] /[-> x] <: TR /[# 1 -> y] /[-> x])
    ) /\
    all_tfree_k G t v_ref /\
    ( forall xz, xz \notin L ->
      forall xc, xc <> xz -> xc \notin L ->
      ctx_met_refined_heap (vis_hide typing_mode G) xz TZ xc TC vRef S ⁏e ⊢ open_met xz xc t ∶ open_met xz xc TR)
  ).
Proof.
  introv Hi Hcl Hst Hbi HtyRef.
  (* some properties following from heap correspondence *)
  assert (ctx_ok G) as HokG by eauto.
  assert (ok e) as Hok by eauto.
  assert (renv_corr G e) as Hcor by eauto.
  (* The type of the corresponding location *)
  assert ( G ⁏ e ⊢ vLoc ∶ typ_rcd (dec_met m TC TZ TR)) as HtyLoc.
  { eapply ty_renv_binds_var; eauto. }
  (*  Note: this proof is kind of convoluted in how it applies var_typ_rcd_met_to_binds to both
    the location and reference separately. *)
  destruct (var_typ_rcd_met_to_binds Hi Hcl Hok Hcor HtyLoc) as [?L [?S [?S1 [?TC [?TZ [?TR [HbiG [Hrcd [?H [?H ?H]]]]]]]]]].
  destruct (var_typ_rcd_met_to_binds Hi Hcl Hok Hcor HtyRef) as [?L [?S [?S1 [?TC [?TZ [?TR [HbiGRef [?Hrcd [?H [?H ?H]]]]]]]]]].
  (* The heap object type is the same for the reference and location *)
  assert (typ_corr (typ_objref S0 S2) (typ_objref S S1)).
  { eapply renv_corr_binds_ref_loc_corr; eauto. }
  assert (S0 = S).
  { inversions H5; auto. }
  subst.
  (* Extract the definition from the heap *)
  pose proof (corresponding_types_ref Hst Hbi HbiGRef) as Hts.
  destruct (strong_mu_to_new Hi Hts HbiG) as [?ds [?Bis [Htyds Hvkds]]].
  pose proof (record_has_ty_defs Htyds Hrcd) as [?d [? Htyd]].
  (* Type of the definition *)
  inversions Htyd.
  exists (L1 \u L \u ctx_dom G) ds t S S2 S1.
  splits; eauto.
  - exists TC0 TZ0 TR0.
    splits; eauto.
  - eapply heap_met_varkinds; eauto.
  - (* Type of the body *)
  intros.
  assert (ctx_obj_met _ G (objctx_heap vLoc S) xz TZ0 xc TC0 ⁏ e
      ⊢ open_met xz xc t ∶ open_met xz xc TR0) as Ht0 by auto.
  assert (ctx_obj_met typing_mode G (objctx_heap vRef S) xz TZ0 xc TC0 ⁏ e
      ⊢ open_met xz xc t ∶ open_met xz xc TR0) as Ht.
  { (* The body is typed with the location on the heap,
       but we will be opening it with a reference. *)
    unfold ctx_obj_met in *.
    eapply narrow_ty. eauto. apply Ht0.
    
    destruct (mode_visibility' typing_mode).
    - apply subctx_grow; eauto.
      apply (@subtyp_and _ (and_or_subtyp _)); auto.
      apply (@subtyp_and _ (and_or_subtyp _)); auto.
      apply subtyp_tyq.
      apply Indep.tyq_typ_open_indep_same.
      eapply (inert_bnd_indep Hi).  apply HbiG.
      auto.
    - apply subctx_grow; eauto.
      apply (@subtyp_and _ (and_or_subtyp _)); auto.
      apply (@subtyp_and _ (and_or_subtyp _)); auto.
      apply subtyp_tyq.
      apply Indep.tyq_typ_open_indep_same.
      eapply (inert_bnd_indep Hi).  apply HbiG.
      auto.
  }
  destruct (mode_visibility' typing_mode) eqn: Hm.
  *
    apply (narrow_ty_trm_subtyp Ht); auto 2.
    +
      unfold ctx_obj_met. rewrite Hm.
      apply subctx_grow; eauto.
      apply (@subtyp_and _ (and_or_subtyp _)); auto.
    +
      assert (ctx_met G xz TZ xc TC ⁏ e ⊢ open_met xz xc TR0 <: open_met xz xc TR).
      {subst.  apply H1; auto. }
      apply (narrow_subtyp H10); auto 2. constructor. auto. notin_ctx_obj_solve.
      apply subctx_grow; auto 3.
  *
    assert (ctx_met G xz TZ xc TC ⁏ e ⊢ open_met xz xc TR0 <: open_met xz xc TR).
    { subst. apply H1; auto. }
    unfold ctx_obj_met in *. rewrite Hm in *.
    ctx_flatten in Ht.
    apply (narrow_ty_trm_subtyp Ht); auto 2.
    +
      apply subctx_grow; eauto.
      apply subctx_grow; eauto.
      eapply hiding_hide; eauto.
      ctx_split_from xz.
      eapply hiding.
      ctx_flatten.
      apply (@subtyp_and _ (and_or_subtyp _)); auto.
    +

      ctx_split_from xz.
      eapply hiding.
      ctx_flatten.

      apply (narrow_subtyp H10); auto 2.
      constructor. auto. notin_ctx_obj_solve.
      apply subctx_grow; auto 3.
Qed.

Lemma canonical_forms_met: forall G e s vRef m TC TZ TR,
  inert G ->
  sealed G ->
  heap_correspond G e s ->
  ctx_k G vRef v_ref ->
  G⁏e ⊢ vRef ∶ (typ_rcd (dec_met m TC TZ TR)) ->
  (exists vLoc L ds t S S1 S2,
    ctx_binds vRef (typ_objref S S1) G /\
    ctx_binds vLoc (typ_objref S S2) G /\
    binds vRef vLoc e /\
    binds vLoc (item_obj ds) s /\
    defs_has ds (def_met m t) /\
    (exists TC0 TZ0 TR0,
      record_has S /[-> vLoc] (dec_met m TC0 TZ0 TR0) /\
      G ⁏ e ⊢ TZ <: TZ0 /\
      (forall x : var, x \notin L -> G §ᶻ x ~ TZ ⁏ e ⊢ TC /[-> x] <: TC0 /[-> x]) /\
      (forall x : var, x \notin L -> forall y : var, y <> x -> y \notin L -> ctx_met G x TZ y TC ⁏ e ⊢ TR0 /[# 1 -> y] /[-> x] <: TR /[# 1 -> y] /[-> x])
    ) /\
    all_tfree_k G t v_ref /\
    ( forall xz, xz \notin L ->
      forall xc, xc <> xz -> xc \notin L ->
      ctx_met_refined_heap (vis_hide typing_mode G) xz TZ xc TC vRef S ⁏e ⊢ open_met xz xc t ∶ open_met xz xc TR)
  ).
Proof.
  introv Hin Hcl Hst Hvk Hty.
  destruct (heap_correspond_exists_loc Hst Hvk) as [vLoc Hbi].
  exists vLoc.  
  pose proof (canonical_forms_met_ref Hin Hcl Hst Hbi Hty) as Hcf.
  destructecs Hcf.
  exists L ds t S S1 S2.
  splits; eauto.
Qed.

End WithTypingMode.
