(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax TypingMode.
Require Import TFree.
Require Import HeapCorrespondence InertContext EnvCorrespondence ConfigTyping ConfigNotations.

(** * Accessory premises of machine configuration typing *)

Section WithTypingMode.
Context {typing_mode: Mode}.

(* The parts from ty_config that are not actually typing. *)
Definition ty_config_accessories G e fs S t (T: typ) Lv :=
    inert G /\
    all_tfree_k_stack G fs v_ref /\
    all_tfree_k G t v_ref /\
    ty_nothing_fresh_in_config G T (t ; fs ; S ; e; Lv) /\
    has_null Lv /\
    sealed G /\
    config_closed (t; fs; S; e; Lv) /\
    closed T.
Definition heap_correspond_accessories G (e: renv) (s: heap) :=
  ctx_ok G /\
  ctx_weakens G /\
  ok s /\
  ok e /\
  ctx_kdom v_loc G = dom s /\
  ctx_kdom v_ref G = dom e /\
  renv_corr G e.

Lemma ty_config_to_accessories G e fs S t (T: typ) Lv:
  ty_config G (t; fs; S; e; Lv) T ->
  ty_config_accessories G e fs S t T Lv.
Proof.
  introv Hty.
  inversions Hty.
  unfold ty_config_accessories.
  splits; eauto.
Qed.
Lemma heap_correspond_to_accessories G e S:
  heap_correspond G e S ->
  heap_correspond_accessories G e S.
Proof.
  introv Hty.
  unfold heap_correspond in Hty.
  destructecs Hty.
  unfold heap_correspond_accessories.
  splits; eauto.
Qed.

Lemma ty_config_to_heap_accessories G t fs Sa e Lva T:
  ty_config G (t; fs ; Sa ; e ; Lva) T ->
  heap_correspond_accessories G e Sa.
Proof.
  introv Htyc.
  inversions Htyc.
  eapply heap_correspond_to_accessories; eauto.
Qed.


End WithTypingMode.