Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax TypingMode.
Require Import HeapValid ConfigTyping ConfigVarkinds EnvCorrespondence HeapCorrespondence OperationalSemantics HeapUpdate.

(** * Valid configurations. *)

(** A configuration is "valid" if variables bound in the heap and env are unique, and the env maps to variables on the heap *)
Inductive config_valid: config->Prop :=
  inductive_conf_c t s Sigma e Lv:
    ok Sigma ->
    ok e ->
    disjoint (dom e) (dom Sigma) ->
    (fv_env e) \c dom Sigma ->
    config_valid (t; s; Sigma; e; Lv).
Hint Constructors config_valid: core.

(** A configuration is "ok" if variables bound in the heap and env are unique *)
Definition conf_ok c: Prop :=
  match c with
  | (t1 ; s1 ; S1 ; e1 ; Lv1) =>
    fresh \{} (length (keys S1 ++ keys e1)%list) (keys S1 ++ keys e1)%list
  end.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma config_typed_valid G C T:
  ty_config G C T ->
  config_valid C.
Proof.
  intros.
  inversions H.
  inversions H1.
  destruct_all.
  assert(disjoint (dom e) (dom Sigma)). {
    rewrite <- H14.
    rewrite <- H13.
    eapply ctx_kdom_disjoint; eauto. discriminate.
  }
  assert(fv_env e \c dom Sigma). {
    intro. intro.
    rewrite <- H13.
    eapply renv_fv_ref; eauto.
  }
  constructor; eauto.
Qed.
Lemma config_hastyp_valid C:
  config_hastyp typing_mode C ->
  config_valid C.
Proof.
  introv Hht.
  inversions Hht.
  eapply config_typed_valid; eauto.
Qed.

Lemma heap_correspond_fields_in_dom G e S:
  heap_correspond G e S ->
  fields S \c dom S.
Proof.
  intros.
  hnf in H; destruct_all.
  rewrite <- H3.
  remember S as S0. (* S0 is the whole heap. *)
  rewrite HeqS0.
  replace (dom S0) with (dom S) in H6 by (rewrite HeqS0; eauto).

  rewrite <- concat_empty_r in HeqS0. remember empty as S1. clear HeqS1.
  reverts S1.
  induction S; intros; subst.
  - unfold fields, FieldsHeap, fields_heap. eapply subset_empty_l.
  - destruct a. unfold fields, FieldsHeap, fields_heap; fold fields_heap.
    forwards: (H6 v).
    rewrite dom_def; simpl. rewrite in_union, in_singleton; eauto.
    rewrite cons_to_push in *.

    eapply subset_union_l.
    eapply IHS; eauto.
    intros. eapply H6. simpl_dom; rewrite in_union; eauto. rewrite <- concat_assoc; eauto.

    inversions H7.
    { eapply binds_middle_eq_inv in H10; subst; eauto 1.
      unfold fields, FieldsItem, fields_item. eapply subset_empty_l. }
    { eapply binds_middle_eq_inv in H9; subst; eauto 1.
      unfold fields, FieldsItem, fields_item.
      eapply defs_varkinds_fields_loc in H12.
      eauto. }
Qed.

Lemma config_valid_renv_idempotent t s S e Lv:
  config_valid (t; s; S; e; Lv)->
  renv_idempotent e.
Proof.
  introv Hv.
  inversions Hv.
  unfold renv_idempotent.
  intro. intro.
  eapply disjoint_in_notin in H6; eauto.
  intro.
  eauto.
Qed.

Lemma config_valid_change t1 s1 S e Lv1 t2 s2 Lv2:
  config_valid (t1; s1; S; e; Lv1)->
  config_valid (t2; s2; S; e; Lv2).
Proof.
  introv Hv.
  inversions Hv.
  eauto.
Qed.

End WithTypingMode.
