(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode GeneralTyping RecordTypes InertContext PreciseTyping TightTyping InvertibleTyping FlatInvertibleTyping LayeredTypingMainCount Equivalence EnvCorrespondence.

Section WithTypingMode.
Context {typing_mode: Mode}.

(** * Sel-<: Premise

    [inert G]                    #<br>#
    [G ⊢## x∶ {A: S..U}]        #<br>#
    [――――――――――――――――――――――――――――]   #<br>#
    [exists T. G ⊢## x∶ {A: T..T}]   #<br>#
    [G ⊢# T <: U]               #<br>#
    [G ⊢# S <: T]                    *)
Lemma sel_premise: forall G e x A S U,
  inert G ->
  G⁏e ⊢## x ∶ typ_rcd (dec_typ A S U) ->
  exists TL TU V,
    G ⊢! x ∶ V ⪼ typ_rcd (dec_typ A TL TU) /\
    (forall y, G⁏e ⊢# open y TU <: open y U) /\
    (forall y, G⁏e ⊢# open y S <: open y TL).
Proof.
  introv HG Hinv.
  eapply invertible_to_precise_typ_dec; eauto.
Qed. 

(** * Sel-<: Replacement

    [inert G]              #<br>#
    [G ⊢# x∶ {A: S..U}]   #<br>#
    [――――――――――――――――――――――]   #<br>#
    [G ⊢# x.A <: U]       #<br>#
    [G ⊢# S <: x.A]            *)
Lemma sel_replacement: forall G e x A S U y,
    inert G ->
    renv_corr G e ->
    G⁏e ⊢# x ∶ typ_rcd (dec_typ A S U) ->
    (G⁏e ⊢# typ_sel (avar_f x) A (avar_f y) <: (open y U)) /\
    (G⁏e ⊢# (open y S) <: typ_sel (avar_f x) A (avar_f y)).
Proof.
  introv HinG Hcor Hty.
  forwards Hinv: tight_to_invertible HinG Hcor Hty.
  forwards Hsel: sel_premise e HinG Hinv.
  destructc Hsel as TL TU V Ht Hs1 Hs2.
  split; intros.
  - apply subtyp_sel1_t with(e:=e)(y:=y)(typing_mode:=typing_mode) in Ht. apply subtyp_trans_t with (T:=(open y TU)); auto.
  - apply subtyp_sel2_t with(e:=e)(y:=y)(typing_mode:=typing_mode) in Ht. apply subtyp_trans_t with (T:=(open y TL)); auto.
Qed.


(** * General to Tight [⊢ to ⊢#] *)
(** The following lemma says that in an inert environment, general typing
    ([ty_trm] [⊢]) can be reduced to tight typing ([ty_trm_t] [⊢#]).
    The proof is by mutual induction on the typing and subtyping judgements.

    [inert G]           #<br>#
    [G ⊢ t∶ T]          #<br>#
    [――――――――――――――]    #<br>#
    [G ⊢# t∶ T] #<br># #<br>#

    and                 #<br># #<br>#
    [inert G]           #<br>#
    [G ⊢ S <: U]        #<br>#
    [――――――――――――――――]  #<br>#
    [G ⊢# S <: U]         *)
Lemma general_to_tight: forall G0,
  inert G0 ->
  (forall G Go e x T,
     obj_typing G Go e x T ->
     renv_corr G e ->
     G = G0 ->
     Go = objctx_empty ->
     G⁏e ⊢# x ∶ T) /\
  (forall G Go e S U,
     subtyp _ G Go e S U ->
     renv_corr G e ->
     G = G0 ->
     Go = objctx_empty ->
     G⁏e ⊢# S <: U) /\
  (forall G Go e S U,
     ty_incap _ G Go e S U ->
     renv_corr G e ->
     G = G0 ->
     Go = objctx_empty ->
     G⁏e ⊢# typ_nocap <: U).
Proof.
  intros G0 HinG.
  apply ts_mutind; intros; subst;
    try match goal with
    | [ H : _ = (avar_f ?y) |- _ ] => inversion H; subst
    end;
    try solve [eapply sel_replacement; auto]; eauto 4.
    (* typ_bnd *)
    {
      eapply ty_rec_intro_t; intros.
      eauto. eapply o0. unfold mode_has_incap'. rewrite H2. auto. eapply H0; eauto. unfold mode_has_incap'. rewrite H2. auto. eapply H; eauto.
    }
    
    Unshelve. all: eapply typ_top.
Qed.

(** The general-to-tight lemma, formulated for term typing. *)
Lemma general_to_tight_typing: forall G e x T,
  inert G ->
  renv_corr G e ->
  G⁏e ⊢ x ∶ T ->
  G⁏e ⊢# x ∶ T.
Proof.
  intros. eapply general_to_tight; eauto.
Qed.

Lemma invertible_flat_to_tight G e U x:
    inert G ->
    renv_corr G e ->
    G⁏e ⊢##f x ∶ U ->
    G⁏e ⊢# x ∶ U.
Proof.
  introv HinG Hcor Hty.
  remember (trm_var (avar_f x)) as t eqn:Heq.
  
  induction Hty; try congruence.
  eapply precise_to_general in H.
  eapply general_to_tight. eauto.
  all: inversions Heq; eauto.
  (* bnd intro *)
  eapply ty_rec_intro_t. eauto.  eauto.
  intros. eapply H0.  unfold mode_has_incap'. rewrite H2. auto.
  intros. eapply general_to_tight; eauto. eapply H0.  unfold mode_has_incap'. rewrite H2. auto.
  eauto.
Qed.


Lemma invertible_to_tight G e U x:
    inert G ->
    renv_corr G e ->
    G⁏e ⊢## x ∶ U ->
    G⁏e ⊢# x ∶ U.
Proof.
  introv HinG Hcor Hty.
  unfold ty_inv in *.
  destruct (mode_ty_inv' typing_mode) eqn: Heq.
  - eapply invertible_flat_to_tight; eauto.
  - eapply ty_inv_main_tight; eauto with typing_mode_deps.
Qed.

End WithTypingMode.
