Set Implicit Arguments.

Require Import Coq.Program.Equality.
Require Import DotTactics LibExtra .
Require Import
        AbstractSyntax TypingMode GeneralTyping
        PreciseTyping EnvCorrespondence
        Substitution Weakening EWeakening Narrowing Visibility
        TFree ConfigVarkinds Deref HeapInstantiate Indep
        RecordTypes InertContext RecordDefs Equivalence
        TightTyping GeneralToTight InvertibleTyping Subenvironments ConfigNotations.

Lemma open_fun_inv : forall T U S y n,
    typ_fun T U = open_rec n y S ->
    exists (T' U' : typ), S = typ_fun T' U'.
Proof.
  introv Heq. destruct S; simpl in Heq; try congruence. eauto.
Qed.

Lemma open_dec_typ_inv : forall A L U S y n,
    dec_typ A L U = open_rec n y S ->
    exists (L' U' : typ), S = dec_typ A L' U'.
Proof.
  introv Heq. destruct S; simpl in Heq; try congruence.
  inversions Heq. eauto.
Qed.

Lemma open_dec_fld_inv : forall A L U S y n,
    dec_fld A L U = open_rec n y S ->
    exists (L' U' : typ), S = dec_fld A L' U'.
Proof.
  introv Heq. destruct S; simpl in Heq; try congruence.
  inversions Heq. eauto.
Qed.

Lemma open_dec_met_inv : forall m TC TZ TR S y n,
    dec_met m TC TZ TR = open_rec n y S ->
    exists (TC' TZ' TR' : typ), S = dec_met m TC' TZ' TR'.
Proof.
  introv Heq. destruct S; simpl in Heq; try congruence.
  inversions Heq. eauto.
Qed.

Lemma open_notin_subst_eq n w y S1 S2:
  w \notin fv_typ S1 \u fv_typ S2 ->
  (open_rec_typ n w S1) = (open_rec_typ n w S2) ->
  (open_rec_typ n y S1) = (open_rec_typ n y S2).
Proof.
  intros.
  assert (Heqsubst : ((open_rec_typ n w S1) /[ w -> y ]) =
                     ((open_rec_typ n w S2) /[ w -> y ])) by congruence.
  repeat(rewrite subst_open_commut_typ in Heqsubst).
  assert (Hfr1 : w \notin fv S1) by
      eauto using notin_union_r1, notin_union_r2.
  assert (Hfr2 : w \notin fv S2) by
      eauto using notin_union_r1, notin_union_r2.
  replace (S1 /[ w -> y]) with S1 in Heqsubst by
      (symmetry; eapply subst_fresh; auto).
  replace (S2 /[ w -> y]) with S2 in Heqsubst by
      (symmetry; eapply subst_fresh; auto).
  rewrite subst_eq_var in Heqsubst. auto.
Qed.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma open_def_tyq : forall G Go e (d : def) (D : dec) w y n,
    inert G -> renv_corr G e -> ctx_ok (G §§& Go) ->
    G §§ Go ⁏ e ⊢ d ∶d open_rec n w D ->
    e ⊢ w ≈ y ->
    w \notin fv D ->
    typ_indep_b n (typ_rcd D) ->
    G §§ Go ⁏ e ⊢ d ∶d open_rec n y D.
Proof.
  introv Hin Hcor Hok Htyd Htyv Hfr Hindep.
  remember (open_rec n w D) as Z.
  generalize dependent y. induction Htyd.
  - apply open_dec_typ_inv in HeqZ as HeqD. destruct HeqD as [S1 [S2 Heq]]. subst.
    simpl in HeqZ. inversions HeqZ. simpl. rename H into Htyq, H2 into Heqop.
    introv Htyv. simpl in Hfr.
    assert (Heq : (open_rec_typ (S n) y S1) = (open_rec_typ (S n) y S2)) by (eapply open_notin_subst_eq; eauto).
    rewrite <- Heq. eapply ty_def_typ. intros. eapply tyq_typ_trans. eapply Htyq. eauto 1. eapply tyq_typ_open.
    assert (Hindep1 : typ_indep_b (S n) S1).
    { inversions_tyq Hindep. auto. }
    eapply tyq_typ_open_rec_indep; eauto 2.
  - apply open_dec_fld_inv in HeqZ as HeqD. destruct HeqD as [S1 [S2 Heq]]. subst.
    simpl in HeqZ. inversions HeqZ. simpl. rename H into Htyq, H3 into Heqop.
    introv Htyv. simpl in Hfr.
    assert (Heq : (open_rec_typ n y S1) = (open_rec_typ n y S2)) by (eapply open_notin_subst_eq; eauto).
    rewrite <- Heq. constructor. eauto. eapply ty_var_sub; eauto.
    assert (Hindep1 : typ_indep_b n S1).
    { inversions_tyq Hindep. auto. }
    eapply subtyp_tyq. eapply tyq_typ_open_rec_indep; eauto 2.
  - apply open_dec_met_inv in HeqZ as HeqD.
    destruct HeqD as [TC' [TZ' [TR' Heq]]]. subst.
    simpl in HeqZ. inversions HeqZ. simpl. rename H into Hty.
    introv Htyv. simpl in Hfr. apply ty_def_met with (L := L \u ctx_dom (G §§& Go)).
    introv Hfrxz Hneq Hfrxc.
    assert (HfrxzL : xz \notin L) by auto.
    assert (HfrxcL : xc \notin L) by auto.
    specialize (Hty _ HfrxzL _ Hneq HfrxcL); clear HfrxzL HfrxcL.
    assert (Hsubctx :
              subctx e
                (ctx_obj_met typing_mode G Go xz (open_rec_typ n y TZ')
                         xc (open_rec_typ (S n) y TC'))
                (ctx_obj_met typing_mode G Go xz (open_rec_typ n w TZ')
                         xc (open_rec_typ (S n) w TC'))).
    { unfold ctx_obj_met.
      assert (Hindep' : typ_indep_b n TZ').
        { inversions_tyq Hindep. auto. }
        destruct Go; unfold ctx_obj in *;
        eapply subctx_grow;
        try eapply subctx_grow; try solve [destruct (mode_visibility' typing_mode); auto].
      - inversions_tyq Hindep.
        rename H3 into HindepTC, H5 into HindepTZ, H6 into HindepTR.
        apply subtyp_tyq. eapply tyq_typ_open.
        eapply tyq_typ_open_rec_indep; eauto 2.
      - iapply subtyp_and; auto.
        inversions_tyq Hindep.
        rename H3 into HindepTC, H5 into HindepTZ, H6 into HindepTR.
        apply subtyp_tyq. eapply tyq_typ_open.
        eapply tyq_typ_open_rec_indep; eauto 2.
      - iapply subtyp_and; auto.
        inversions_tyq Hindep.
        rename H3 into HindepTC, H5 into HindepTZ, H6 into HindepTR.
        apply subtyp_tyq. eapply tyq_typ_open.
        eapply tyq_typ_open_rec_indep; eauto 2.
      }
    pose proof (narrow_ty _ _ Hty Hsubctx) as Hty'.
    eapply ty_trm_sub; eauto 2. eapply subtyp_tyq; eauto 2.
    assert (Hindep' : typ_indep_b (S (S n)) TR').
    { inversions_tyq Hindep; auto. }
    assert (Htyq : e ⊢ (open_rec (S (S n)) w TR') ≈ (open_rec (S (S n)) y TR')).
    { eapply tyq_typ_open_rec_indep; eauto 2. }
    apply tyq_typ_open_rec with (n := 1) (y := xc) in Htyq.
    apply tyq_typ_open_rec with (n := 0) (y := xz) in Htyq. auto.
Qed.

Lemma open_typ_rcd_inv : forall T D y n,
    typ_rcd D = open_rec n y T ->
    exists (D' : dec), T = typ_rcd D'.
Proof.
  introv Heq. destruct T; simpl in Heq; try congruence.
  inversions Heq. eauto.
Qed.

Lemma open_and_inv : forall T U S y n,
    typ_and T U = open_rec n y S ->
    exists (T' U' : typ), S = typ_and T' U'.
Proof.
  introv Heq. destruct S; simpl in Heq; try congruence. eauto.
Qed.

Lemma open_defs_tyq : forall G Go e (ds : defs) (Ds : typ) w y n,
    inert G -> renv_corr G e -> ctx_ok (G §§& Go) ->
    G §§ Go ⁏ e ⊢ ds ∶ open_rec n w Ds ->
    e ⊢ w ≈ y ->
    w \notin fv Ds ->
    typ_indep_b n Ds ->
    G §§ Go ⁏ e ⊢ ds ∶ open_rec n y Ds.
Proof.
  induction ds; introv Hin Hcor Hok Htyd Htyv Hfr Hindep; inversions Htyd.
  - rename H5 into Hrcdeq.
    apply open_typ_rcd_inv in Hrcdeq as ?. destruct H. subst. simpl in Hrcdeq.
    inversions Hrcdeq. simpl. eauto using open_def_tyq.
  - apply open_and_inv in H1 as HeqD.
    destruct HeqD as [T' [U' Heq]]. subst. simpl in *.
    inversions H1. rename H2 into HeqZ.
    apply open_typ_rcd_inv in HeqZ as HeqD.
    destruct HeqD as [D' Heq]. subst. simpl in HeqZ. inversions HeqZ.
    simpl in *. apply ty_defs_cons; eauto 2.
    + eapply IHds; eauto 3 using notin_union_r2, notin_union_r1.
      inversions_tyq Hindep; auto.
    + eapply open_def_tyq; eauto 3 using notin_union_r2, notin_union_r1.
      inversions_tyq Hindep; auto.
Qed.

(** * Heap Correspondence *)

(** The operational semantics will be defined in terms of configurations
    consisting of a term [t], a stack [s], and a heap [Sigma].
    For typing a configuration, [heap_correspond] establishes a correspondence
    between the context [Gamma] and the heap [Sigma].

    We say that a context [Gamma] corresponds to a heap [Sigma] if
    - [G = {(xi mapsto Ti) | i = 1, ..., n}]
    - [s = {(xi mapsto vi) | i = 1, ..., n}]
    - If [vi] is a function then [G ⊢ vi∶ Ti]
    - If [vi = nu(T)ds] is an object then [Ti = mu(... /\ D /\ ...)] and
      [G ⊢ ds^xi :: (... /\ D /\ ...)^xi].
 *)

(** ** Strong Bindings for Locations *)


(** We first define a correspondence relation for a single location. This
    definition will later be lifted to entire contexts and heaps. *)

(** We say that a location [l] is strongly bound in the context [G], heap
    [s] and reference environment [e], denoted [s(l):G(l)] if there exists a
    [T] such that
    - if [s(x)] is a function [lambda(y: T)t] then [G ⊢ lambda(y∶ T)t ∶ T]
    - if [s(x)] is an object [nu(y: U)ds] for some [U], then [T = mu(U)] and
      [G ⊢ ds^x :: U^x] *)

Inductive ty_item_s : ctx -> renv -> heap -> var -> Prop :=
  (* The λ-abstraction case is no longer present in roDOT. But kept as possible
     room for an extension. *)
| ty_item_fun_s : forall G e s x S t L T,
    mode_has_fun typing_mode ->
    ctx_binds x (typ_fun S T) G ->
    binds x (item_fun t) s ->
    (forall x, x \notin L ->
      G § x / v_par ~ S ⁏e ⊢ open x t ∶ open x T) ->
    all_tfree_k G t v_ref -> (* all tfree variables in functions are references *)
    ty_item_s G e s x
| ty_item_obj_s : forall G e s x U S ds T,
    ctx_binds x T G ->
    binds x (item_obj ds) s ->
    T = typ_objref U S ->
    G §§ (objctx_heap x U) ⁏e ⊢ ds ∶ (open x U) ->
    defs_varkinds G ds -> (* all fields must contain locations, all tfree variables in methods are references *)
    ty_item_s G e s x.
Hint Constructors ty_item_s : core.

Local Hint Extern 1 (_ §# _) => ctx_simpl_dom; notin_solve : core.
Local Hint Extern 1 (ctx_ok (_ §& (_ / _ ~ _))) => constructor; auto : core.

(** A strongly bound location stays strongly bound in extended heaps and contexts *)
Lemma ty_item_s_push: forall G e s x y T v,
    ty_item_s G e s x ->
    ctx_ok G ->
    y §# G ->
    y \notin fv_env s ->
    ty_item_s (G § y / v_loc ~ T) e (s & y ~ v) x.
Proof.
  introv Hty Hok Hun Hnfv.
  assert (x \notin \{ y }).
  { apply notin_singleton.
    intros ?H. subst x.
    inversions Hty; eauto using ctx_binds_unbound_false. }
    destruct Hty.
  - (* λ-abstraction case *) eapply ty_item_fun_s; eauto.
    intros. assert (x0 \notin L) by eauto; specialize (H3 _ H5); clear H5.
    replace ((G § y / v_loc ~ T) § x0 / v_par ~ S) with
        (G §& (y / v_loc ~ T) §& (x0 / v_par ~ S)) by auto.
    apply weaken_ty; eauto.
  - (* Object case *) eapply ty_item_obj_s; eauto.
    replace (G § y / v_loc ~ T) with (G §& (y / v_loc ~ T)) by auto.
    eapply weaken_end; eauto.
    apply weaken_defs_varkinds; auto.
Qed.

Lemma ty_item_s_epush: forall G e s x w y T,
    ty_item_s G e s x ->
    ctx_ok G -> w §# G  ->
    ok e -> w # e ->
    ty_item_s (G § w / v_ref ~ T) (e & w ~ y) s x.
Proof.
  introv Htys HokG HunG Hoks Hfvs. inversions Htys.
  - (* ty_item_fun_s λ-abstraction case. *)
    assert (x <> w).
     intro; subst. eapply ctx_binds_unbound_false; eauto.
    eapply ty_item_fun_s; eauto.
    introv Hnin.
    specialize (H2 _ Hnin).
    rewrite <- (concat_empty_r e) in H2.
    eapply ewkn with (e2 := w ~ y) in H2; eauto.
    rewrite (concat_empty_r _) in H2.
    rewrite (ctx_concat_single G) in H2.
    rewrite (ctx_concat_double2 G).
    eapply weaken_ty; eauto.
  - (* ty_item_obj_s Object case. *)
    assert (x <> w).
     intro; subst. eapply ctx_binds_unbound_false; eauto.
    eapply ty_item_obj_s; eauto.
    rewrite <- (concat_empty_r _) in H2.
    eapply ewkn with (e2 := w ~ y) in H2; eauto.
    rewrite (concat_empty_r _) in H2.
    rewrite (ctx_concat_single G).
    eapply weaken_end ; eauto.
    apply weaken_defs_varkinds; auto.
Qed.

Lemma ty_item_s_Gpush: forall G e s x w vk T,
    ty_item_s G e s x ->
    ctx_ok G -> w §# G  ->
    ok e ->
    ty_item_s (G § w / vk ~ T) (e) s x.
Proof.
  introv Htys HokG HunG Hoke. inversions Htys.
  - (* ty_item_fun_s λ-abstraction case. *)
    assert (x <> w).
     intro; subst. eapply ctx_binds_unbound_false; eauto.
    eapply ty_item_fun_s; eauto.
    introv Hnin.
    specialize (H2 _ Hnin).
    rewrite <- (concat_empty_r e) in H2.
    rewrite (concat_empty_r _) in H2.
    rewrite (ctx_concat_single G) in H2.
    rewrite (ctx_concat_double2 G).
    eapply weaken_ty; eauto.
  - (* ty_item_obj_s Object case. *)
    assert (x <> w).
     intro; subst. eapply ctx_binds_unbound_false; eauto.
    eapply ty_item_obj_s; eauto.
    rewrite <- (concat_empty_r _) in H2.
    rewrite (concat_empty_r _) in H2.
    rewrite (ctx_concat_single G).
    eapply weaken_end ; eauto.
    apply weaken_defs_varkinds; auto.
Qed.

Lemma ebinm_ty_item_s G e e' x Sigma :
  ty_item_s G e Sigma x ->
  env_binds_more e e' ->
  ty_item_s G e' Sigma x.
Proof.
  introv Htyitm Hbm.
  inversions Htyitm; eauto using ebinm_ty_trm, ebinm_ty_defs.
Qed.

(** ** Heap Correspondence *)

(* Part of Figure 9 in "Reference mutability for DOT" *)
(* Part of Figure 14 in "Pure methods for roDOT" *)
(** [s:G] if [s] and [G] have the same domain and every location in [s] is
    strongly bound *)
Definition heap_correspond (G : ctx) (e: renv) (s : heap) : Prop :=
  ctx_ok G /\
  ctx_weakens G /\
  ok s /\
  ok e /\
  ctx_kdom v_loc G = dom s /\
  ctx_kdom v_ref G = dom e /\
  renv_corr G e /\
  (forall x,
      x \in dom s ->
      ty_item_s G e s x).
Hint Unfold heap_correspond : core.

(** ** Simple lemmas for Heap Correspondence *)

(** If [s: G], the variables in the domain of [s] are distinct. *)
Lemma heap_correspond_to_ok_G: forall s e G,
    heap_correspond G e s -> ctx_ok G.
Proof.
  intros. destruct H as [? [? ?]]. auto.
Qed.
Hint Resolve heap_correspond_to_ok_G : core.

Lemma heap_correspond_to_ok_s: forall s e G,
    heap_correspond G e s -> ok s.
Proof.
  intros. destruct H as [? [? [? ?]]]. auto.
Qed.
Hint Resolve heap_correspond_to_ok_s : core.

Lemma heap_correspond_to_ok_e: forall s e G,
    heap_correspond G e s -> ok e.
Proof.
  intros. destruct H as [? [? [? [? ?]]]]. auto.
Qed.
Hint Resolve heap_correspond_to_ok_e : core.

Lemma heap_correspond_to_ok_renv_corr: forall s e G,
    heap_correspond G e s -> renv_corr G e.
Proof.
  intros. destruct H as [? [? [? [? [? [? [? ?]]]]]]]. auto.
Qed.
Hint Resolve heap_correspond_to_ok_renv_corr : core.

Lemma heap_correspond_notin_dom_inv: forall G e s x,
    heap_correspond G e s ->
    x §# G ->
    x # e /\ x # s
.
Proof.
  introv Hwt. destruct Hwt as [? [? [? [? [?Hdoms [?Hdome ?]]]]]].
  unfold notin. rewrite <- Hdome. rewrite <- Hdoms.
  intros.
  split; apply ctx_notin_dom_kdom; auto.
Qed.

Lemma heap_correspond_in_dom_s: forall G e s x,
    heap_correspond G e s ->
    x \in dom s ->
    x \in ctx_kdom v_loc G.
Proof.
  introv Hwt. destruct Hwt as [? [? [? [? [?Hdom ?]]]]].
  rewrite Hdom.
  intros. auto.
Qed.

Lemma heap_correspond_env_fv_heap_dom G e S :
  heap_correspond G e S ->
  fv_env e \c dom S.
Proof.
  introv [HokG [HwkG [HokS [Hoke [Hlds [Hrde [Hcor Htys]]]]]]].
  intros y Hin. eapply fv_env_to_binds in Hin as [w Hbie]; eauto.
  eapply renv_corr_binds_loc, ctx_k_in_kdom in Hcor; eauto.
  rewrite Hlds in *. easy.
Qed.

(** ** Inductive lemmas for Heap Correspondence *)

Lemma heap_correspond_empty:
    heap_correspond ctx_empty empty empty.
Proof.
  repeat split; auto.
  - simpl_dom; ctx_simpl_kdom; auto. (* heap dom *)
  - simpl_dom; ctx_simpl_kdom; auto. (* env dom *)
  - introv B. simpl_dom.
    rewrite in_empty in B.
    destruct B.
Qed.
Hint Resolve heap_correspond_empty : core.

Lemma heap_correspond_push: forall G e s y T v,
    heap_correspond G e s ->
    y §# G ->
    y \notin fv_env s ->
    ty_item_s (G § y / v_loc ~ T) e (s & y ~ v) y ->
    heap_correspond (G § y / v_loc ~ T) e (s & y ~ v).
Proof.
  introv Hcor Hu Hnfv Htis. unfold heap_correspond in *.
  destruct (heap_correspond_notin_dom_inv Hcor Hu).
  destruct_all; splits; auto.
  (* The new renv_corr for roDOT is resolved automatically. *)
  - simpl_dom. ctx_simpl_kdom. rewrite <- H5; auto.
  - simpl_dom. ctx_simpl_kdom. rewrite ctx_kdom_push_n. rewrite <- H6; auto. simplify_eq.
  - intros. rewrite dom_push, in_union, in_singleton in H9.
    destruct_all; subst; auto using ty_item_s_push.
Qed.

Lemma heap_correspond_epush_nowfr : forall G G2 e s w y T1 T2,
    heap_correspond G e s ->
    w §# G ->
    w # e ->
    ctx_binds_k_after y v_loc T2 G G2 ->
    typ_corr T1 T2 ->
    heap_correspond (G § w / v_ref ~ T1) (e & w ~ y) s.
Proof.
  intros.
  unfold heap_correspond in *.
  destruct_all; splits; auto.
  - rewrite ctx_kdom_push. rewrite~ If_r. simplify_eq.
  - rewrite ctx_kdom_push. rewrite~ If_l. simpl_dom. rewrite H8. auto.
  - eapply renv_corr_ref; eauto.
  - intros.
    apply ty_item_s_epush; auto.
Qed.

Lemma heap_correspond_epush: forall G G2 e s w y T1 T2,
    heap_correspond G e s ->
    w §# G ->
    w # e ->
    w \notin fv_env s ->
    ctx_binds_k_after y v_loc T2 G G2 ->
    typ_corr T1 T2 ->
    heap_correspond (G § w / v_ref ~ T1) (e & w ~ y) s.
Proof.
  intros.
  unfold heap_correspond in *.
  destruct_all; splits; auto.
  - rewrite ctx_kdom_push. rewrite~ If_r. simplify_eq.
  - rewrite ctx_kdom_push. rewrite~ If_l. simpl_dom. rewrite H9. auto.
  - eapply renv_corr_ref; eauto.
  - intros.
    apply ty_item_s_epush; auto.
Qed.

Lemma heap_correspond_epermute G e e' Sigma :
  heap_correspond G e Sigma ->
  env_binds_more e  e' ->
  env_binds_more e' e  ->
  ok e' ->
  heap_correspond G e' Sigma.
Proof.
  introv [HokG [HwkG [HokS [Hoke [Hls [Hre [Hcorr Htyitm]]]]]]]
    Hbim Hbim' Hoke'.
  rewrite (proj1 (env_binds_more_permute Hbim Hbim')) in Hre.
  repeat(split); eauto using renv_corr_binds_more, ebinm_ty_item_s.
Qed.

Lemma renv_corr_push_obj G e w y T S:
  renv_corr G e ->
  w §# G ->
  y §# G ->
  w <> y ->
  typ_corr (typ_objref T S) (typ_objref T S) ->
  renv_corr
    (inert_G_for_obj G w y T S)
    (e & w ~ y).
Proof.
  intros.
  eapply renv_corr_ref.
  - eapply ctx_binds_k_e.
  - eapply ctx_binds_k_e.
  - auto.
  - apply renv_corr_push; auto.
Qed.
Lemma renv_corr_push_fun G e w y T S:
  renv_corr G e ->
  w §# G ->
  y §# G ->
  w <> y ->
  typ_corr (typ_fun T S) (typ_fun T S) ->
  renv_corr
    (inert_G_for_item G w y  (typ_fun T S) )
    (e & w ~ y).
Proof.
  intros.
  eapply renv_corr_ref.
  - eapply ctx_binds_k_e.
  - eapply ctx_binds_k_e.
  - auto.
  - apply renv_corr_push; auto.
Qed.


(* Helper lemma about visibility in extended context *)
Lemma ctx_vis_for_obj G w y T S x:
  ctx_vis G x ->
  ctx_vis (inert_G_for_obj G w y T S) x.
Proof.
  introv Hvis.
  destruct (classicT (x=w)).
  subst. eauto.
  constructor; eauto.
  destruct (classicT (x=y)).
  subst. eauto.
  constructor; eauto.
Qed.


Lemma ctx_ok_vis_hide G1 G2:
  ctx_ok (G1 §& G2) ->
  ctx_ok (G1 §& (vis_hide typing_mode G2)).
Proof.
  intros.
  destruct (mode_visibility' typing_mode);
  eauto.
  induction G2; ctx_flatten; eauto.
Qed.

Hint Resolve ctx_ok_vis_hide: core.

Lemma instantiation_met0z L G e w y z T S (t: trm) TZ TC TR L0:
  z \notin (L \u fv T \u fv S \u ctx_fv G \u dom e \u fv_env e \u \{ w } \u \{ y }) ->
  ok (e & w ~ y) ->
  ctx_ok ((inert_G_for_obj G w y T S § z / v_self ~ T /[-> z])) ->
  typ_indep_b 0 T ->
  (forall xz : var,
     xz \notin L0 ->
     forall xc : var,
     xc <> xz ->
     xc \notin L0 -> ctx_obj_met typing_mode (G § z / v_self ~ T /[-> z]) (objctx_heap z T) xz TZ xc TC ⁏ e ⊢ open_met xz xc t ∶ open_met xz xc TR) ->
  forall xz, xz \notin (L0 \u \{ z }\u \{ w}\u \{ y} \u (ctx_dom G)) ->
  forall xc, xc <> xz -> xc \notin (L0 \u \{ z }\u \{ w}\u \{ y} \u (ctx_dom G)) ->
  (inert_G_for_obj G w y T S § z / v_self ~ T /[-> z]) §& ctx_met_refined_heap (vis_hide typing_mode ctx_empty) xz TZ xc TC z T ⁏ e & w ~ y ⊢ t /[# 1 -> xc] /[-> xz] ∶ TR /[# 1 -> xc] /[-> xz].
Proof.
  introv Hn Hokw HokG Hind Hty.
  intros.

  destruct (mode_visibility' typing_mode) eqn: Hm.
    * (* basic *)

      ctx_split_between y z.
      eapply weaken_ty; auto.
      ctx_flatten.
      replace (e & w ~ y) with (e & w ~ y & empty) by
        (rewrite concat_empty_r; auto).
      eapply ewkn; auto.
      rewrite concat_empty_r by auto.
      unfold ctx_obj_met in Hty.
      rewrite Hm in Hty.
      eapply Hty; eauto. simpl. auto.
    *
      unfold ctx_obj_met in Hty.
      rewrite Hm in Hty.
      ctx_split_between y z.
      rewrite <- ctx_concat_hide.
      repeat rewrite <- ctx_concat_push.
      eapply weaken_ty; auto.
      ctx_flatten.
      replace (e & w ~ y) with (e & w ~ y & empty) by
        (rewrite concat_empty_r; auto).
      eapply ewkn; auto.
      rewrite concat_empty_r; auto. eauto. auto. ctx_flatten. auto.
Qed.

Lemma instantiation_met0_sub G w y T S xz TZ xc TC z e L L0:
 z \notin L \u fv T \u fv S \u ctx_fv G \u dom e \u fv_env e \u \{ w} \u \{ y} ->
ctx_ok (inert_G_for_obj G w y T S §ˢ z ~ T /[-> z]) ->
 typ_indep_b 0 T ->
xz \notin L0 \u \{ z} \u \{ w} \u \{ y} \u ctx_dom G ->
 xc <> xz ->
 xc \notin L0 \u \{ z} \u \{ w} \u \{ y} \u ctx_dom G ->

inert_G_for_obj G w y T S
§& (ctx_met_refined_heap
      match mode_visibility' typing_mode with
      | mode_visibility_all => ctx_empty
      | mode_visibility_hide_method => ctx_hide ctx_empty
      end xz TZ xc TC z T) /[§ z -> w] ⁏ e & w ~ y ⊢ w ∶ T /[-> z] /[ z -> w].
Proof.
introv Hnz HokG Hin Hnxz Hnx Hnxc.
    rewrite subst_open_commut.
    rewrite substi_eq_var.
    rewrite subst_fresh; auto.
    destruct (mode_visibility' typing_mode).
    * ctx_flatten. eauto 6. 
    * ctx_split_from xz.
      eapply hiding.
      ctx_split_from xz.
      eapply weaken_end; auto.
      eauto. eauto.
      constructor; eauto.
Qed.


Lemma instantiation_met0 L G e w y z T S (t: trm) TZ TC TR L0:
  z \notin (L \u fv T \u fv S \u ctx_fv G \u dom e \u fv_env e \u \{ w } \u \{ y }) ->
  ok (e & w ~ y) ->
  ctx_ok ((inert_G_for_obj G w y T S § z / v_self ~ T /[-> z])) ->
  typ_indep_b 0 T ->
  (forall xz : var,
     xz \notin L0 ->
     forall xc : var,
     xc <> xz ->
     xc \notin L0 -> ctx_obj_met typing_mode (G § z / v_self ~ T /[-> z]) (objctx_heap z T) xz TZ xc TC ⁏ e ⊢ open_met xz xc t ∶ open_met xz xc TR) ->
  forall xz, xz \notin (L0 \u \{ z }\u \{ w}\u \{ y} \u (ctx_dom G)) ->
  forall xc, xc <> xz -> xc \notin (L0 \u \{ z }\u \{ w}\u \{ y} \u (ctx_dom G)) ->
  (inert_G_for_obj G w y T S §& (subst_ctx z w (ctx_met_refined_heap (vis_hide typing_mode ctx_empty) xz TZ xc TC z T)) ⁏ e & w ~ y ⊢
        ((open_met xz xc t) /[ z -> w]) ∶ ((open_met xz xc TR) /[ z -> w]) ).
Proof.
  introv Hn Hokw HokG Hind Hty.
  intros.
  eapply subst_ty.
  + auto.
  + auto using (ewkn_ty_trm _).
  + eapply instantiation_met0z; eauto.
  + ctx_flatten.

    constructor. constructor.
    eapply ctx_ok_vis_hide; eauto.
    destruct (mode_visibility' typing_mode); ctx_flatten; notin_solve.
    destruct (mode_visibility' typing_mode); ctx_flatten; notin_solve.

  + sympls; eauto using ctx_dom.
  + sympls; eauto.

  + unfold renv_notin. notin_solve. auto.
    rewrite env_push_fv. notin_solve.
    unfold fv, VarFreeVar. auto.
  + eapply instantiation_met0_sub; eauto.
  + intros. destruct (mode_visibility' typing_mode). simpl. eauto.
    simpl. eauto. ctx_flatten in H3.
    inversions H3. false. eapply notin_same with (x:=z). notin_solve.
    inversions H10. false. eapply notin_same with (x:=z). notin_solve.
    inversions H11.
  Unshelve. all: eauto.
  all: try solve [eapply ewkn_ty_trm | eapply hiding_ty_var | eapply weaken_ty_var].
Qed.


Lemma instantiation_met L G e w y z T S m t TZ TC TR L0:
  z \notin (L \u fv T \u fv S \u ctx_fv G \u dom e \u fv_env e \u \{ w } \u \{ y }) ->
  ok (e & w ~ y) ->
  ctx_ok ((inert_G_for_obj G w y T S § z / v_self ~ T /[-> z])) ->
  typ_indep_b 0 T ->
  (forall xz : var,
     xz \notin L0 ->
     forall xc : var,
     xc <> xz ->
     xc \notin L0 -> ctx_obj_met typing_mode (G § z / v_self ~ open z T) (objctx_heap z T) xz TZ xc TC ⁏ e ⊢ open_met xz xc t ∶ open_met xz xc TR) ->
  inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y ⊢ def_met m (t /[ z -> w]) ∶d dec_met m (TC /[ z -> w]) (TZ /[ z -> w]) (TR /[ z -> w]).
Proof.
  introv Hn Hokw HokG Hind Hty.

  pose proof (instantiation_met0 _ _ Hn Hokw HokG Hind Hty) as Hty1.

  eapply ty_def_met.
  sympls.
  intros.
  ctx_split_from xz.

  specialize (Hty1 xz H xc H0 H1).

  sympls.

  repeat (
    rewrite subst_open_commut in Hty1;
    try rewrite substi_eq_var in Hty1;
    try rewrite substi_ne_var in Hty1 by auto
  ).
  eapply (narrow_ty); eauto.
  constructor; auto.

  +
    assert (ctx_ok G).
    inversions HokG. inversions H4. inversions H5. auto.

    destruct (mode_visibility' typing_mode).
    * constructor; eauto. constructor; auto 3. constructor; auto 3. inversions HokG. eauto.
    * eauto. constructor; eauto. constructor; auto 3.
  +
    destruct (mode_visibility' typing_mode); eauto 3. constructor; eauto.
  +
    iapply subtyp_and; auto.
    iapply subtyp_and; auto.

    replace (T /[ z -> w]) with T.
    apply subtyp_tyq.
    auto.
    rewrite subst_fresh; auto.

    replace (T /[ z -> w]) with T. auto.
    rewrite subst_fresh; auto.
Qed.

(* A definition typed in an object literal will be typed on the heap *)
Lemma instantiation_def L G e w y z T S D d:
  z \notin (L \u fv d \u fv T \u fv S \u ctx_fv G \u dom e \u fv_env e \u \{ w } \u \{ y }) ->
  ok (e & w ~ y) ->
  ctx_ok ((inert_G_for_obj G w y T S § z / v_self ~ T /[-> z])) ->
  typ_indep_b 0 T ->
  (G §§ objctx_lit z T ⁏ e ⊢ d /[-> z] ∶d D) ->
  inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y ⊢ (d /[-> z] /[ z -> w]) ∶d (D /[ z -> w]).
Proof.
  introv Hn Hokw HokG Hindep Htyd.
  inversions Htyd.
  * (* ty_def_typ *)
    sympls.
    eapply ty_def_typ.
    intros. assert (x \notin L0 \u \{z}) by apply H0.
    replace (e & w ~ y) with (e & w ~ y & empty) by (rewrite concat_empty_r; auto); eauto.
    apply ewkne; auto.
    rewrite concat_empty_r.
    rewrite <- subst_open_commut_ne by notin_solve.
    rewrite <- subst_open_commut_ne by notin_solve.
    apply subst_equiv; auto.
  * (* ty_def_fld *)
    sympls.
    eapply ty_def_fld.
    - (* visibility *)
      intro.
      unfold ctx_obj. cases_if; eauto.
      specialize (H0 H1).
      inversions H0.
      contradiction.
      eapply ctx_vis_for_obj; auto.
    - (* Field value type *)
      unfold ctx_obj.
      replace (t \[ z -> w]) with (t /[ z -> w]) by auto.
      rewrite <- ctx_concat_empty_r at 1.
      eapply subst_ty with (G1:=(inert_G_for_obj G w y T S)) (vk:=v_self) (x:=z) (e0:=(e& w ~ y)) (S0:=open z T) (G2:=ctx_empty).
      + eauto.
      + eauto using (ewkn_ty_var).
      + ctx_split_between y z.
        eapply weaken_ty; auto.
        ctx_flatten.
        replace (e & w ~ y) with (e & w ~ y & empty) by
          (rewrite concat_empty_r; auto).
        eapply ewkn.
        rewrite concat_empty_r; auto. auto.
        auto.
        constructor; auto.
        inversions HokG; auto.
        inversions HokG; inversions H3; auto.
      + ctx_flatten; auto.
      + ctx_simpl_fv. sympls. notin_solve.
      + auto.
      + unfold renv_notin.
        simpl_dom. notin_solve.
        rewrite env_push_fv.
        notin_solve.
        unfold fv, VarFreeVar. auto.
      +
        rewrite subst_open_commut.
        rewrite subst_fresh. rewrite substi_eq_var.
        apply ty_rec_elim; eauto. notin_solve.
      + intros. simpl. eauto.
  * (* ty_def_met *)
    eapply instantiation_met; eauto.
Qed.

Lemma instantiation_defs' L G e w y T S T0 (ds: defs) z:
  ok (e & w ~ y) ->
  ctx_ok (inert_G_for_obj G w y T S) ->
  typ_indep_b 0 T ->
  (G §§ objctx_lit z T ⁏ e ⊢ open z ds ∶ open z T0) ->
  z \notin     (L \u \{ w} \u \{ y} \u ctx_dom G \u ctx_fv G \u fv ds \u fv T \u fv S \u dom e
       \u fv_env e \u dom (e & w ~ y) \u fv_env (e & w ~ y)
       \u ctx_fv (inert_G_for_obj G w y T S) \u fv T0) ->
  (inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y ⊢ (open w ds) ∶ open w T0).
Proof.
  introv Hoke HokG Hin Hdefs Hn.

  assert (Hzds : z \notin (fv ds)). notin_solve.
  rewrite subst_intro with (x:=z) (y0:=w) by auto.
  rewrite subst_intro with (x:=z) (y0:=w) by auto.
  assert (z \notin L) as HzL by notin_solve.

  revert dependent T0.
  induction ds.
  + (* nil - not possible *)
    intros. inversions Hdefs.
  + intros. (* cons *)
    assert (Hzd : z \notin (fv d)). sympls. auto.
    inversions Hdefs.
    - (* one *)
      sympl.
      destruct ds.
      * apply ty_defs_one.
        eapply instantiation_def with (L:=L). notin_solve.
        auto.
        constructor; auto. auto. auto.
      * inversions H.
    - sympl.
      apply ty_defs_cons.
      destruct T0; inversions H1.
      eapply IHds; auto. sympls. notin_solve. sympls. notin_solve.
      eapply instantiation_def with (L:=L); auto.
      apply subst_defs_hasnt.
      unfold label_of, DefLabelOf in *.
      destruct d; sympls; auto.
Qed.

(* Lemma 5 in "Reference mutability for DOT" *)
(* A definition list typed in an object literal will be typed on the heap *)
Lemma instantiation_defs L G e w y T S T0 (ds: defs):
  ok (e & w ~ y) ->
  ctx_ok (inert_G_for_obj G w y T S) ->
  typ_indep_b 0 T ->
  (forall x : var, x \notin L -> G §§ objctx_lit x T ⁏ e ⊢ open x ds ∶ open x T0) ->
  (inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y ⊢ (open w ds) ∶ open w T0).
Proof.
  introv Hoke HokG Hin Hdefs.
  pick_fresh_gen
    (L \u \{ w} \u \{ y} \u ctx_dom G \u ctx_fv G \u fv ds \u fv T \u fv S \u dom e
       \u fv_env e \u dom (e & w ~ y) \u fv_env (e & w ~ y)
       \u ctx_fv (inert_G_for_obj G w y T S) \u fv T0) z.

  eapply instantiation_defs' with (z:=z); eauto.

Qed.


Lemma ctx_inert_for_obj_heap G w y T S e (ds: defs):
  inert G ->
  inert_typ (typ_and (typ_bnd T) S) v_loc ->
  y §# G ->
  y \notin ctx_fv G ->
  y \notin fv (typ_and (typ_bnd T) S) ->
  w \notin fv (typ_and (typ_bnd T) S) ->
  w §# G ->
  w <> y ->
w \notin ctx_fv G ->
  inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y ⊢ ds /[-> w] ∶ T /[-> w] ->
  inert (inert_G_for_obj G w y T S §§& objctx_heap y T).
Proof.
  introv HinG Hint HyuG HyuGfv HyfvTS HwfvTS.
  introv HwuG Hwy HwuGfv Htydsw.
  apply ty_defs_record_type in Htydsw as Hrcd.
  constructor; eauto.
  eapply inert_typ_loc_ref; eauto.
  { sympls. auto. }
Qed.

Lemma heap_correspond_push_obj: forall L G e s w y T S ds,
    heap_correspond G e s ->
    inert G ->
    inert_typ (typ_objref T S) v_loc ->
    typ_corr (typ_objref T S) (typ_objref T S) ->
    (forall x : var, x \notin L -> G §§ (objctx_lit x T) ⁏e ⊢ open x ds ∶ open x T) ->
    y §# G -> y \notin ctx_fv G ->
    y \notin fv ds ->
    y \notin fv_env s ->
    y \notin fv T ->
    y \notin fv S ->
    w §# G -> w \notin ctx_fv G ->
    w \notin fv ds ->
    w \notin fv_env s ->
    w \notin fv T ->
    w \notin fv S ->
    w <> y ->
    all_tfree_k_defs G ds v_ref ->
    heap_correspond (inert_G_for_obj G w y T S) (e & w ~ y)
                    (s & y ~ item_obj_dereffed e w y ds).
Proof.
  introv Hcor HinG Hint Htcor Hfvs HyuG HyuGfv.
  introv Hyds Hyfvs HyfvT HyfvS.
  introv HwuG HwuGfv Hwds Hwfvs HwfvT HwfvS Hwy Hvkds.
  unfold heap_correspond in *.
  (* Establish freshness of w and y in s and e *)
  destruct (heap_correspond_notin_dom_inv Hcor HyuG) as [Hyne Hyns].
  destruct (heap_correspond_notin_dom_inv Hcor HwuG) as [Hwne Hwns].
  destruct_all.
  (* Freshness in the contect *)
  assert (y \notin fv (typ_objref T S)) as HyfvTS. {
    unfold fv. unfold TypFreeVar. simpl. auto using notin_union_l.
  }
  assert (w \notin fv (typ_objref T S)) as HwfvTS. {
    unfold fv. unfold TypFreeVar. simpl. auto using notin_union_l.
  }
  (* Establish some properties of new context, heap, renv *)
  assert (ok (heap_with_obj_dereffed s e w y ds)) as Hoks2 by auto.
  assert (ctx_ok (inert_G_for_obj G w y T S)) as HokG2 by auto.
  assert (ctx_weakens (inert_G_for_obj G w y T S)) as HwkG2 by auto.
  (* binding of new variables in new context and heap *)
  assert (binds y (item_obj_dereffed e w y ds) (s & y ~ (item_obj_dereffed e w y ds))) as Hbisy by auto.
  assert (binds w y (e & w ~ y)) as Hbiew by auto.
  assert (ctx_binds y (typ_objref T S) (inert_G_for_obj G w y T S)) as HbiGy by auto.
  (* renv correspondence  *)
  assert (renv_corr (inert_G_for_obj G w y T S) (e & w ~ y)) as Hcor2.
    apply renv_corr_push_obj; auto.
  (* domain of renv e *)
  assert (ctx_kdom v_ref (inert_G_for_obj G w y T S) = dom (e & w ~ y)).
  { simpl_dom. rewrite ctx_kdom_push_e. rewrite ctx_kdom_push_n. rewrite H4. auto.
    simplify_eq. }
  (* typing of ds *)
  assert (inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y
          ⊢ (open w ds) ∶ open w T) as Htydsw.
          eapply instantiation_defs; eauto 6.
  (* Proving conditions of heap correspondence *)
  splits; auto.
  - (* domain of heap s *)
    simpl_dom. rewrite ctx_kdom_push_n. ctx_simpl_kdom. rewrite H3. auto.
    simplify_eq.
  - (* heap item typing *)
    intros x0 Hd.
    (* establish position of x0 in the new context *)
    assert (x0 \in ctx_dom (inert_G_for_obj G w y T S)) as HdG2.
      rewrite ctx_dom_push. rewrite in_union. right.
      rewrite ctx_dom_push. rewrite in_union.
      rewrite dom_push in Hd. rewrite in_union in Hd.
      destruct Hd; auto.
      right. eapply ctx_in_kdom_dom. apply heap_correspond_in_dom_s
                                       with (e:=e) (s:=s).
      unfold heap_correspond. splits; auto.
      auto.
    pose proof (ctx_dom_to_binds HokG2 HwkG2 HdG2) as [?T HbiG]; clear HdG2.
    pose proof (dom_to_binds Hd) as [?T Hbis].
    (* Cases for the new item and old items *)
    destruct (binds_push_inv Hbis) as [[? ?] | [? ?]]; subst.
    + (* The new item (x0 = y) *)
      apply (ty_item_obj_s (U := T) (S := S) HbiG Hbisy); auto.
      * eapply ctx_binds_functional; eauto.
      * apply ty_deref_defs; eauto 3.
        eapply ctx_inert_for_obj_heap; eauto. 
        eapply open_defs_tyq; eauto 3.
        -- constructor; eauto.
           eapply inert_typ_loc_ref; eauto.
           { simpl. auto. }
        -- inversions Hint; eauto.
      * eapply all_tfree_defs_deref; eauto.
        apply all_tfree_k_open; eauto.
        apply ctx_k_push_e.
    + (* An old item *)
      apply ty_item_s_epush; eauto.
      apply ty_item_s_push; eauto using binds_to_dom, binds_push_inv.
Qed.

(** ** Inversion lemmas for [s:G] *)

(** [s:G] implies every location is strongly typed *)

(** [s: G]              #<br>#
    [G(l) = T]          #<br>#
    [―――――――――――――]     #<br>#
    [s(l):G(l)]         *)
Lemma corresponding_types: forall G G1 e s x T,
    heap_correspond G e s ->
    ctx_binds_k_after x v_loc T G G1 ->
    ty_item_s G e s x.
Proof.
  intros. unfold heap_correspond in H.
  destruct_all. apply H7.
  rewrite <- H4.
  eauto using ctx_binds_kdom.
Qed.

Lemma corresponding_types_ref: forall G e s w y T,
    heap_correspond G e s ->
    binds w y e ->
    ctx_binds w T G ->
    ty_item_s G e s y.
Proof.
  introv Hhc Hbie HbiG. unfold heap_correspond in Hhc.
  pose proof Hhc as Hhc1.
  destruct Hhc1 as [Hok [? [? [? [? [Hdome [Hcor ?]]]]]]].

  destruct (renv_corr_binds_ref_to_loc Hok Hcor Hbie HbiG) as [TL [Hbiy [HvkL Htcor]]].
  destruct HvkL as [? [? HbikL]].
  eapply corresponding_types; eauto.
Qed.

(** [s: G]              #<br>#
    [s(l) = v]          #<br>#
    [―――――――――――――]     #<br>#
    [s(l):G(l)]         *)
Lemma corresponding_types_s: forall G e s x v,
    heap_correspond G e s ->
    binds x v s ->
    ty_item_s G e s x.
Proof.
  intros. unfold heap_correspond in H.
  destruct_all. apply H7.
  eauto using binds_to_dom.
Qed.

Lemma heap_correspond_exists_loc G e s w:
  heap_correspond G e s ->
  ctx_k G w v_ref ->
  exists y,
  binds w y e.
Proof.
  introv Hcor Hvk.
  destruct Hcor as [? [? [? [? [? [Hdom [Hecor ?]]]]]]].
  pose proof (ctx_k_in_kdom Hvk) as Hinkdom.
  rewrite Hdom in Hinkdom.
  apply dom_to_binds.
  assumption.
Qed.

Lemma heap_correspond_notin_dom: forall G e s x,
    heap_correspond G e s ->
    inert G ->
    x # s ->
    x # e ->
    x §# G.
Proof.
  introv Hcor Hin Hns Hne.
  destruct Hcor as [? [? [? [? [Hdoms [Hdome ?]]]]]].
  erewrite inert_dom.
  rewrite Hdoms, Hdome.
  auto.
  auto.
Qed.

Lemma heap_correspond_heap_binds_vis G e S y i:
  inert G ->
  heap_correspond G e S ->
  binds y i S ->
  ctx_vis G y.
Proof.
  introv Hin Hhc Hbi.
  destructc Hhc as ? ? ? Hoke Hkdom ? Hrest.

  eapply binds_to_dom in Hbi.
  rewrite <- Hkdom in Hbi.
  eapply ctx_in_kdom_dom in Hbi.
  eapply ctx_dom_to_binds in Hbi.
  destructc Hbi as t Ht.
  eapply inert_binds_vis; eauto.
  eauto.
  eauto.
Qed.

Lemma heap_correspond_renv_push_binds_vis G e S w1 y1 w2 y2:
  inert G ->
  heap_correspond G e S ->
  binds w1 y1 (e & w2 ~ y2) ->
  w1 <> w2 ->
  ctx_vis G w1.
Proof.
  introv Hin Hhc Hbi hn.
  destructc Hhc as ? ? ? Hoke ? Hkdom Hrest.

  eapply binds_to_dom in Hbi.
  rewrite dom_push in Hbi.

  rewrite in_union in Hbi. destruct Hbi. rewrite in_singleton in H3. contradiction. rename H3 into Hbiy.
  rewrite <- Hkdom in Hbiy.
  eapply ctx_in_kdom_dom in Hbiy. eapply ctx_dom_to_binds in Hbiy. destruct Hbiy.
  eapply inert_binds_vis; eauto. eauto. eauto.
Qed.

End WithTypingMode.

Hint Constructors ty_item_s : core.
Hint Resolve heap_correspond_to_ok_G : core.
Hint Resolve heap_correspond_to_ok_s : core.
Hint Resolve heap_correspond_to_ok_e : core.
Hint Resolve heap_correspond_to_ok_renv_corr : core.
Hint Resolve heap_correspond_empty : core.

Arguments heap_correspond {typing_mode}.
(* An inert context extended with a reference and a location for a new heap item *)
Notation inert_G_for_item G w y T := (G § y / v_loc ~ T § w / v_ref ~ T).
(* An inert context extended with a reference and a location for a new object *)
Notation inert_G_for_obj G w y T S := (inert_G_for_item G w y (typ_objref T S)).
(** S1 is a prefix of S2 *)
Notation heap_prefix S1 S2 := (exists S1', S2 = concat S1 S1').
