(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Program.Equality.
Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping.
Require Import RecordTypes InertContext PreciseTyping TightTyping Narrowing Env Equivalence
        EnvCorrespondence Subenvironments
.

Section WithTypingMode.
Context {typing_mode: Mode}.

(* NOTE:
  The main result of this file, inert_tight_subtyping, should hold in all typing modes,
  but this way of proving it does not work with mode_top_nm_subtype.
  It could maybe be proven using layered typing, but for now it does not seem to be worth it, because it is not used for safety proof.
*)
  
Context (Hm: mode_top_nm' typing_mode = mode_top_nm_none).

(** * Super Types of Objects *)
Inductive bnd_sup : ctx -> renv -> typ -> typ -> Prop :=
| bnd_sup_top : forall G e T,
    bnd_sup G e T typ_top
| bnd_sup_and : forall G e T T1 T2,
    bnd_sup G e T T1 ->
    bnd_sup G e T T2 ->
    bnd_sup G e T (typ_and T1 T2)
| bnd_sup_or1 : forall G e T T1 T2,
    bnd_sup G e T T1 ->
    bnd_sup G e T (typ_or T1 T2)
| bnd_sup_or2 : forall G e T T1 T2,
    bnd_sup G e T T2 ->
    bnd_sup G e T (typ_or T1 T2)
| bnd_sup_sel : forall G e T x A y U SL SU,
    G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A SL SU) ->
    bnd_sup G e T (open y SL) ->
    bnd_sup G e T (typ_sel (avar_f x) A (avar_f y))
| bnd_sup_rec : forall G e T,
    bnd_sup G e (typ_bnd T) (typ_bnd T)
| bnd_sup_eq : forall G e T T1 T2,
    bnd_sup G e T T1 ->
    typing_equiv e T1 T2 ->
    bnd_sup G e T T2
.
Local Hint Constructors bnd_sup : core.

Ltac unique_bind_bounds :=
  match goal with
  | [ Hin : inert ?G,
      H : ?G ⊢! ?x ∶ ?U ⪼ typ_rcd (dec_typ ?A ?S ?S),
      H' : ?G ⊢! ?x ∶ ?U' ⪼ typ_rcd (dec_typ ?A ?S' ?S') |- _ ] =>
    pose proof (x_bound_unique H H'); subst;
    pose proof (pf_inert_unique_tight_bounds Hin H H'); subst
  end.

Local Hint Extern 1 => unique_bind_bounds : core.

Ltac inversions_sup_tyq Htyq := (
    inversions Htyq;
    match goal with
    | [ Htybin : typ_binary _ |- _ ] => inversions Htybin; try congruence
    | [ Htybin : typ_andor _ |- _ ] => inversions Htybin; try congruence
    | [ Htynul : typ_nulary _ |- _ ] => inversions Htynul; try congruence
    | _ => idtac
    end).

Lemma bnd_sup_trans : forall G e S T U,
    bnd_sup G e S T ->
    bnd_sup G e T U ->
    bnd_sup G e S U.
Proof.
  introv Hbndsup. induction 1; eauto 3.
  specialize (IHbnd_sup1 Hbndsup). specialize (IHbnd_sup2 Hbndsup). eauto 3.
Qed.
Local Hint Resolve bnd_sup_trans : core.

Lemma bnd_sup_bot_inv : forall G e T,
    bnd_sup G e T typ_bot ->
    T = typ_bot.
Proof.
  introv Hbndsup; remember typ_bot as U; induction Hbndsup; subst;
    try congruence; eauto.
  inversions H; eauto.
  inversions H1; try congruence.
Qed.

Lemma bnd_sup_nocap_false : forall G e T,
    bnd_sup G e T typ_nocap ->
    False.
Proof.
  introv Hbndsup; remember typ_nocap as U; induction Hbndsup; subst;
    try congruence; eauto.
  inversions H; eauto.
  inversions H1; try congruence.
Qed.

Lemma bnd_sup_and_weak_tyq_l : forall G e S T U SU,
    bnd_sup G e T SU ->
    typing_equiv e SU (typ_and S U) ->
    bnd_sup G e T S.
Proof.
  induction 1; introv Htyeq;
    inversions_sup_tyq Htyeq; subst; eauto 2.
  - inversions H2. inversions H1. eauto 2.
  - inversions H1. assert (typing_equiv e (typ_and t11 t12) (typ_and S U)) by eauto 3.
    pose proof (tyq_typ_trans H0 H1). auto.
Qed.

Lemma bnd_sup_and_weak_tyq_r : forall G e S T U SU,
    bnd_sup G e T SU ->
    typing_equiv e SU (typ_and S U) ->
    bnd_sup G e T U.
Proof.
  induction 1; introv Htyeq;
    inversions_sup_tyq Htyeq; subst; eauto 2.
  - inversions H2. inversions H1. eauto 2.
  - inversions H1. assert (typing_equiv e (typ_and t11 t12) (typ_and S U)) by eauto 3.
    pose proof (tyq_typ_trans H0 H1). auto.
Qed.

Lemma bnd_sup_and_weak_l : forall G e S T U,
    bnd_sup G e T (typ_and S U) ->
    bnd_sup G e T S.
Proof. eauto 3 using bnd_sup_and_weak_tyq_l. Qed.

Lemma bnd_sup_and_weak_r : forall G e S T U,
    bnd_sup G e T (typ_and S U) ->
    bnd_sup G e T U.
Proof. eauto 3 using bnd_sup_and_weak_tyq_r. Qed.

Lemma bnd_sup_inv_or_tyq : forall G e S T U SU,
    bnd_sup G e T SU ->
    typing_equiv e SU (typ_or S U) ->
    bnd_sup G e T S \/ bnd_sup G e T U.
Proof.
  induction 1; introv Htyeq;
    inversions_sup_tyq Htyeq; subst; eauto 2.
  - inversions H1. inversions H0. eauto 3.
  - inversions H1. inversions H0. eauto 3.
  - inversions H1. assert (typing_equiv e (typ_or t11 t12) (typ_or S U)) by eauto 3.
    pose proof (tyq_typ_trans H0 H1). auto.
Qed.

Lemma bnd_sup_inv_or : forall G e S T U,
    bnd_sup G e T (typ_or S U) ->
    bnd_sup G e T S \/ bnd_sup G e T U.
Proof. eauto 3 using bnd_sup_inv_or_tyq. Qed.

Lemma bnd_sup_no_rcd_sup : forall G e S TR,
    bnd_sup G e S TR ->
    (exists T, typing_equiv e TR (typ_rcd T)) ->
    False.
Proof.
  introv Hbndsup.
  induction Hbndsup; introv [?T ?Htyq]; inversions_sup_tyq Htyq; inversions_tyq H; eauto.
Qed.

Lemma bnd_sup_no_bnd_fun : forall G e TS TP,
    bnd_sup G e TS TP ->
    (exists S, typing_equiv e TS (typ_bnd S)) ->
    (exists A B, typing_equiv e TP (typ_fun A B)) ->
    False.
Proof.
  introv Hbndsup.
  induction Hbndsup; introv [?S ?HtyqS] [?A [?B ?HtyqP]];
    inversions_sup_tyq HtyqS;
    inversions_sup_tyq HtyqP.
    eauto.
Qed.

Lemma pf_dec_typ_subtyp_t : forall G x T A S U G2 e2,
    inert G ->
    G ⊢! x ∶ T ⪼ typ_rcd (dec_typ A S U) ->
    G2⁏e2 ⊢# S <: U .
Proof.
  introv Hi Pf.
  destruct (pf_dec_typ_inv Hi Pf);
  subst; auto.
Qed.

Lemma bnd_sup_unsel_inert : forall G e x U A y TL TU S AR,
    bnd_sup G e (typ_bnd S) AR ->
    inert G -> renv_corr G e ->
    G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
    typing_equiv e AR (typ_sel (avar_f x) A (avar_f y)) ->
    bnd_sup G e (typ_bnd S) (open y TU).
Proof.
  (* introv Hin Hpf Hbndsup Htyq. *)
  introv Hbndsup.
  dependent induction Hbndsup;
    introv Hin Hcor Hpf Htyq; inversions_sup_tyq Htyq;
      specialize (IHHbndsup Hm _ eq_refl Hin Hcor Hpf); eauto 3.
  - rename x0 into x', U0 into U', H2 into Htyqv, H into Hpf'.
    apply bnd_sup_eq with (T1 := (open y SL)) (T2 := (open y TU)); auto; clear IHHbndsup.
    pose proof (pf_tyq_two_upper_tyq Hin Hcor Htyqv Hpf' Hpf) as Htyq.
    destruct (pf_dec_typ_inv Hin Hpf'); subst; eauto 2.
    eapply tyq_typ_open; eauto.
    apply bnd_sup_bot_inv in Hbndsup. congruence.
  - assert (U = U0) by eauto 3 using x_bound_unique; subst.
    pose proof (pf_inert_unique_tight_bounds Hin H Hpf) as [? ?]; subst.
    clear H.
    pose proof (pf_dec_typ_inv Hin Hpf) as [? | Hcontra]; subst; eauto 3.
    apply bnd_sup_bot_inv in Hbndsup. congruence.
  - apply IHHbndsup.
    assert (typing_equiv e
                    (typ_sel (avar_f v1) (typ_label_normal l) (avar_f y))
                    (typ_sel (avar_f x) (typ_label_normal l) (avar_f y))) by eauto 3.
    eauto 3 using tyq_typ_trans.
Qed.

Lemma tight_bnd_sup: forall G e S T U,
    inert G -> renv_corr G e ->
    G⁏e ⊢# T <: U ->
    bnd_sup G e (typ_bnd S) T ->
    bnd_sup G e (typ_bnd S) U.
Proof.
  introv Hi Hcor H.
  induction H; introv Hbinsup;
    match goal with
    | [ H : bnd_sup _ _ _ _ |- _ ] =>
      try solve [false; eapply bnd_sup_nocap_false; eauto 1 | inversions H; auto 3]
    end; eauto 4;
    repeat match goal with
      | [ Hin : inert G, Hp : inert G -> _ |- _ ] => specialize (Hp Hin)
        end.
  - apply bnd_sup_bot_inv in Hbinsup. congruence.
  - eauto 2 using bnd_sup_and_weak_l.
  - eauto 2 using bnd_sup_and_weak_r.
  - destruct (bnd_sup_inv_or Hbinsup); eauto 3.
  - exfalso; eauto 3 using bnd_sup_no_rcd_sup.
  - exfalso; eauto 3 using bnd_sup_no_rcd_sup.
  - eauto 3 using bnd_sup_unsel_inert.
  - exfalso. eapply bnd_sup_no_bnd_fun; eauto 3.
  - exfalso; eauto 3 using bnd_sup_no_rcd_sup.
  - inversions Hbinsup.
    + false.
      eapply bnd_sup_no_rcd_sup. apply H4. eexists. auto.
    + eapply bnd_sup_and_weak_tyq_r in H0. 2: apply H.
      false.
      eapply bnd_sup_no_rcd_sup. apply H0. eexists. auto.
  - inversions Hbinsup.
    inversions H5; auto.
    destruct (bnd_sup_inv_or_tyq H H0).
    apply bnd_sup_or1. eapply bnd_sup_and; auto.
    apply bnd_sup_or2. eapply bnd_sup_and; auto.
    symmetry in H0.
    destruct (tyq_typ_andor_inv T1 (typ_or T2 T3) typ_andor_and H0) as [t1 [t2 [Ht [Htyq1 Htyq2]]]]; subst.
    inversions H.
    +
      destruct (tyq_typ_andor_inv T2 T3 typ_andor_or Htyq2) as [t21 [t22 [Ht2 [Htyq21 Htyq22]]]]; subst.
      symmetry in Htyq1.
      symmetry in Htyq21.
      symmetry in Htyq22.
      destruct (bnd_sup_inv_or H7).
      * apply bnd_sup_or1.
        eapply bnd_sup_and.
        apply (bnd_sup_eq H6 Htyq1).
        apply (bnd_sup_eq H Htyq21).
      * apply bnd_sup_or2.
        eapply bnd_sup_and.
        apply (bnd_sup_eq H6 Htyq1).
        apply (bnd_sup_eq H Htyq22).
    +
      pose proof (bnd_sup_and_weak_tyq_r H1 H2).
      pose proof (bnd_sup_and_weak_tyq_l H1 H2).
      symmetry in Htyq2.
      symmetry in Htyq1.
      destruct (bnd_sup_inv_or_tyq H Htyq2).
      * apply bnd_sup_or1.
        eapply bnd_sup_and.
        apply (bnd_sup_eq H3 Htyq1).
        auto.
      * apply bnd_sup_or2.
        eapply bnd_sup_and.
        apply (bnd_sup_eq H3 Htyq1).
        auto.
  -  typing_mode_contradiction typing_mode.
Qed.

Local Hint Resolve tight_bnd_sup : core.

(** * Super Types of Functions *)
Inductive fun_sup : ctx -> renv -> typ -> typ -> Prop :=
| fun_sup_top : forall G e T,
    fun_sup G e T typ_top
| fun_sup_and : forall G e T T1 T2,
    fun_sup G e T T1 ->
    fun_sup G e T T2 ->
    fun_sup G e T (typ_and T1 T2)
| fun_sup_or1 : forall G e T T1 T2,
    fun_sup G e T T1 ->
    fun_sup G e T (typ_or T1 T2)
| fun_sup_or2 : forall G e T T1 T2,
    fun_sup G e T T2 ->
    fun_sup G e T (typ_or T1 T2)
| fun_sup_sel : forall G e T x A y U S,
    G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A S S) ->
    fun_sup G e T (open y S) ->
    fun_sup G e T (typ_sel (avar_f x) A (avar_f y))
| fun_sup_fun : forall L G e S1 T1 S2 T2,
    G⁏e ⊢# S2 <: S1 ->
    (forall x, x \notin L ->
       G § x / v_par ~ S2 ⁏e ⊢ open x T1 <: open x T2) ->
    fun_sup G e (typ_fun S1 T1) (typ_fun S2 T2)
| fun_sup_eq : forall G e T T1 T2,
    fun_sup G e T T1 ->
    typing_equiv e T1 T2 ->
    fun_sup G e T T2
.
Local Hint Constructors fun_sup : core.

Lemma fun_sup_refl: forall G e S T,
    fun_sup G e (typ_fun S T) (typ_fun S T).
Proof.
  intros. apply_fresh fun_sup_fun as z; auto.
Qed.
Local Hint Resolve fun_sup_refl : core.

Lemma fun_sup_bot : forall G e T,
    not (fun_sup G e T typ_bot).
Proof.
  unfold not. introv Hsup. dependent induction Hsup; try eauto 3.
  inversions H; try auto 3.
  inversions H1; try congruence.
Qed.

Lemma fun_sup_nocap_false : forall G e T,
    fun_sup G e T typ_nocap ->
    False.
Proof.
  introv Hbndsup; remember typ_nocap as U; induction Hbndsup; subst;
    try congruence; eauto.
  inversions H; eauto.
  inversions H1; try congruence.
Qed.

Lemma fun_sup_and_weak_tyq_l : forall G e S T U SU,
    fun_sup G e T SU ->
    typing_equiv e SU (typ_and S U) ->
    fun_sup G e T S.
Proof.
  induction 1; introv Htyeq;
    inversions_sup_tyq Htyeq; subst; eauto 2.
  - inversions H2. inversions H1. eauto 2.
  - inversions H1. assert (typing_equiv e (typ_and t11 t12) (typ_and S U)) by eauto 3.
    pose proof (tyq_typ_trans H0 H1). auto.
Qed.

Lemma fun_sup_and_weak_tyq_r : forall G e S T U SU,
    fun_sup G e T SU ->
    typing_equiv e SU (typ_and S U) ->
    fun_sup G e T U.
Proof.
  induction 1; introv Htyeq;
    inversions_sup_tyq Htyeq; subst; eauto 2.
  - inversions H2. inversions H1. eauto 2.
  - inversions H1. assert (typing_equiv e (typ_and t11 t12) (typ_and S U)) by eauto 3.
    pose proof (tyq_typ_trans H0 H1). auto.
Qed.

Lemma fun_sup_or_inv : forall G e T tor t1 t2,
    fun_sup G e T tor -> typing_equiv e tor (typ_or t1 t2) ->
    fun_sup G e T t1 \/ fun_sup G e T t2.
Proof.
  induction 1; introv Htyeq;
    inversions_sup_tyq Htyeq; subst; eauto 2.
  - inversions H1; inversions H0. left. eapply fun_sup_eq; eauto 2.
  - inversions H1; inversions H0. right. eapply fun_sup_eq; eauto 2.
  - inversions H1.
    assert (typing_equiv e T1 (typ_or t1 t2)) by eauto 3 using tyq_typ_trans. auto.
Qed.

Lemma fun_sup_fun_rcd_no : forall G e S T Trcd R,
    fun_sup G e (typ_fun S T) Trcd ->
    typing_equiv e Trcd (typ_rcd R) ->
    False.
Proof.
  induction 1; introv Htyeq;
    inversions_sup_tyq Htyeq; subst; eauto 2; inversions_tyq H0.
  - apply IHfun_sup. eapply tyq_typ_fld; eapply tyq_typ_trans; eauto 3.
  - apply IHfun_sup. eapply tyq_typ_typ; intros; eapply tyq_typ_trans; eauto 3.
  - apply IHfun_sup. eapply tyq_typ_met; intros; eapply tyq_typ_trans; eauto 3.
Qed.

Lemma fun_sup_fun_bnd_no : forall G e S T Tbnd R,
    fun_sup G e (typ_fun S T) Tbnd ->
    typing_equiv e Tbnd (typ_bnd R) ->
    False.
Proof.
  induction 1; introv Htyeq;
    inversions_sup_tyq Htyeq; subst; eauto 2.
    apply IHfun_sup. eapply tyq_typ_trans; eauto 3.
Qed.

Lemma fun_sup_fun_inv : forall G e U P S T,
    fun_sup G e U P ->
    ok e -> inert G ->
    typing_equiv e P (typ_fun S T) ->
    exists S' T' L,
      (U = typ_fun S' T') /\
      (G ⁏ e ⊢# S <: S')  /\
      (forall x : var, x \notin L ->
                  G § x / v_par ~ S ⁏ e ⊢ open x T' <: open x T).
Proof.
  induction 1; introv Hok Hin Htyeq;
    inversions_sup_tyq Htyeq; subst; eauto 2.
  -
    exists S1 T1 (L \u (ctx_dom G)); repeat split; auto.
    + apply subtyp_trans_t with (T := S2); auto. apply tyq_typ_sym in H5. auto.
    + introv Hfr. assert (HfrL : x \notin L) by eauto. specialize (H0 _ HfrL).
      assert (Hst : G § x / v_par ~ S ⁏ e ⊢ open x T1 <: open x T2).
      { eapply narrow_ty; eauto 2.
        constructor; eauto 3.
        eapply subtyp_tyq.  auto using tyq_typ_sym. }
      assert (Hst' : G § x / v_par ~ S ⁏ e ⊢ open x T2 <: open x T).
      { eapply subtyp_tyq.  eauto using tyq_typ_sym. }
      eapply subtyp_trans; eauto.
  - 
    assert (typing_equiv e T1 (typ_fun S T)) by eauto using tyq_typ_trans.
    specialize (IHfun_sup Hok Hin H1) as [S' [T' [L' [Heq [HstT Hst]]]]].
    exists S' T' L'; eauto 3.
Qed.

Lemma fun_sup_trans : forall G e T1 T2 T3,
    ok e -> inert G ->
    fun_sup G e T1 T2 ->
    fun_sup G e T2 T3 ->
    fun_sup G e T1 T3.
Proof.
  introv Hok Hin Hsup. induction 1; eauto 4.
  eapply fun_sup_fun_inv in Hsup; eauto 3.
  destruct Hsup as [S' [T' [L' [? [HstT Hst]]]]]; subst.
  eapply fun_sup_fun with (L := L \u L' \u ctx_dom G); eauto 2.
  introv Hfr.
  assert (HfrL : x \notin L) by auto. assert (HfrL' : x \notin L') by auto.
  specialize (H0 _ HfrL). specialize (Hst _ HfrL').
  assert (Hst2: G § x / v_par ~ S2 ⁏ e ⊢ open x T' <: open x T0).
  { eapply narrow_ty; eauto 3.
    constructor; eauto 3.
    pose proof tight_to_general as [_ Htg]. auto. }
  eauto.
Qed.

Lemma fun_sup_unsel_inert : forall G e x U A y TL TU S T AR,
    fun_sup G e (typ_fun S T) AR ->
    inert G -> renv_corr G e ->
    G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
    typing_equiv e AR (typ_sel (avar_f x) A (avar_f y)) ->
    fun_sup G e (typ_fun S T) (open y TU).
Proof.
  introv Hbndsup.
  dependent induction Hbndsup;
    introv Hin Hcor Hpf Htyq; inversions_sup_tyq Htyq;
    specialize (IHHbndsup Hm _ _ eq_refl Hin Hcor Hpf); eauto 3.
  - rename x0 into x', U0 into U', H2 into Htyqv, H into Hpf'.
    apply fun_sup_eq with (T1 := (open y S0)) (T2 := (open y TU)); auto; clear IHHbndsup.
    pose proof (pf_tyq_two_upper_tyq Hin Hcor Htyqv Hpf' Hpf) as Htyq.
     eauto.
  - assert (U = U0) by eauto 3 using x_bound_unique; subst.
    pose proof (pf_inert_unique_tight_bounds Hin H Hpf) as [? ?]; subst.
    clear H.
    pose proof (pf_dec_typ_inv Hin Hpf) as [? | Hcontra]; subst; eauto 3.
  - apply IHHbndsup.
    assert (typing_equiv e
                    (typ_sel (avar_f v1) (typ_label_normal l) (avar_f y))
                    (typ_sel (avar_f x) (typ_label_normal l) (avar_f y))) by eauto 3.
    eauto 3 using tyq_typ_trans.
Qed.

Lemma tight_fun_sup: forall G e S T U1 U2,
    ok e -> inert G -> renv_corr G e ->
    G⁏e ⊢# U1 <: U2 ->
    fun_sup G e (typ_fun S T) U1 ->
    fun_sup G e (typ_fun S T) U2.
Proof.
  introv He Hi Hc Ht.
  induction Ht; intros; auto 3;
    try solve [ eapply fun_sup_and_weak_tyq_l; eauto 3
              | eapply fun_sup_and_weak_tyq_r; eauto 3
              | exfalso; eapply fun_sup_fun_rcd_no; eauto 2
              | (* nocap *)
                false; eapply fun_sup_nocap_false; eauto 1
              | eauto 3 ].
  - apply fun_sup_bot in H; tauto.
  - eapply fun_sup_or_inv in H; eauto 2. destruct_all; auto.
  - replace TU with TL in H; eauto 3.
    eapply pf_dec_typ_inv in H; destruct_all; subst; auto.
    apply fun_sup_bot in H0. tauto.
  - eauto 3 using fun_sup_unsel_inert.
  - assert (fun_sup G e (typ_fun S1 T1) (typ_fun S2 T2)) by
        (eapply fun_sup_fun; eauto 3).
    eauto 2 using fun_sup_trans.
  - inversions H.
    + false.
      eapply fun_sup_fun_rcd_no. apply H5. eauto.
    + eapply fun_sup_and_weak_tyq_r in H0. 2: apply H1.
      false.
      eapply fun_sup_fun_rcd_no. apply H0. eauto.
  - inversions H.
    inversions H6; auto.
    destruct (fun_sup_or_inv H H0).
    apply fun_sup_or1. eapply fun_sup_and; auto.
    apply fun_sup_or2. eapply fun_sup_and; auto.
    symmetry in H1.
    destruct (tyq_typ_andor_inv T1 (typ_or T2 T3) typ_andor_and H1) as [t1 [t2 [Ht [Htyq1 Htyq2]]]]; subst.
    inversions H0.
    +
      destruct (tyq_typ_andor_inv T2 T3 typ_andor_or Htyq2) as [t21 [t22 [Ht2 [Htyq21 Htyq22]]]]; subst.
      symmetry in Htyq1.
      symmetry in Htyq21.
      symmetry in Htyq22.
      destruct (fun_sup_or_inv H7 (tyq_typ_refl _ _)).
      * apply fun_sup_or1; eauto.
      * apply fun_sup_or2; eauto.
    +
      pose proof (fun_sup_and_weak_tyq_r H H2).
      pose proof (fun_sup_and_weak_tyq_l H H2).
      symmetry in Htyq2.
      symmetry in Htyq1.
      destruct (fun_sup_or_inv H0 Htyq2).
      * apply fun_sup_or1; eauto.
      * apply fun_sup_or2; eauto.
  -  typing_mode_contradiction typing_mode.
Qed.
Local Hint Resolve tight_fun_sup : core.

Lemma fun_sup_top_any : forall G e T U,
    fun_sup G e typ_top T ->
    fun_sup G e U       T.
Proof.
  introv Hsup. generalize dependent U. dependent induction Hsup; try eauto 3.
  intros. constructor; auto 3.
Qed.

Lemma inert_tight_fun_typing : forall G e S t U,
    inert G ->
    G⁏e ⊢ lit_fun S t ∶ U ->
    exists T, fun_sup G e (typ_fun S T) U.
Proof.
  introv Hi Hty.
  inversions Hty; eauto.
Qed.

Lemma inert_tight_fun_typing_precise : forall G e S t U,
    inert G ->
    G⁏e ⊢ lit_fun S t ∶ U ->
    exists T L,
      (forall x, x \notin L -> G § x / v_par ~ S ⁏e ⊢ open x t ∶ open x T) /\
      fun_sup G e (typ_fun S T) U.
Proof.
  introv Hi Hty. inversion Hty; eauto.
Qed.

(** * Tight SubTyping Contradictions *)
Lemma inert_tight_subtyping: forall G e,
  inert G -> ok e -> renv_corr G e ->
  (forall T S U, ~ G⁏e ⊢# (typ_bnd T) <: (typ_fun S U)) /\
  (forall S U T, ~ G⁏e ⊢# (typ_fun S U) <: (typ_bnd T)).
Proof.
  introv Hi He Hc; splits; introv H.
  - assert (bnd_sup G e (typ_bnd T) (typ_bnd T)) by auto.
    pose proof (tight_bnd_sup Hi Hc H H0).
    eapply bnd_sup_no_bnd_fun; eauto 3.
  - assert (fun_sup G e (typ_fun S U) (typ_fun S U)) by auto.
    pose proof (tight_fun_sup He Hi Hc H H0).
    eapply fun_sup_fun_bnd_no; eauto 3.
Qed.

End WithTypingMode.
