Set Implicit Arguments.

Require Import Coq.Program.Equality.
Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode.
Require Import Equivalence GeneralTyping Indep.
Require Import PreciseTyping RecordTypes InertContext EnvCorrespondence.
Require Import TightTyping InertTightSubtyping.
Require Import Orctx Logctx OrctxLogctx Logpath.

Require Import LayeredCommon LayeredSubtypingAtomic LayeredCapabilityBasic.

(** * N and M supertypes on the union closure layer. *)

Section WithTypingMode.
Context {typing_mode: Mode}.

(** [G ⊢b N◁ T]   #<br>#
    [―――――――――――(LNC-Basic)]   #<br>#
    [G ⊢c N◁ T] #<br>#
    #<br>#
    [―――――――――――(LNC-Top)]   #<br>#
    [G ⊢c N◁ ⊤] #<br>#
    #<br>#
    [G ⊢c N◁ T]   #<br>#
    [―――――――――――(LNC-Or1)]   #<br>#
    [G ⊢c N◁ (T ∨ U)] #<br>#
    #<br>#
    [G ⊢c N◁ U]   #<br>#
    [―――――――――――(LNC-Or2)]   #<br>#
    [G ⊢c N◁ (T ∨ U)]
    *)
(* Part of Figure 22 in "Pure methods for roDOT" *)
Definition ty_has_M_closure G e := ty_or_closure (ty_has_M G e).

(** [G ⊢b M◁ T]   #<br>#
    [―――――――――――(LMC-Basic)]   #<br>#
    [G ⊢c M◁ T] #<br>#
    #<br>#
    [―――――――――――(LMC-Top)]   #<br>#
    [G ⊢c M◁ ⊤] #<br>#
    #<br>#
    [G ⊢c M◁ T]   #<br>#
    [―――――――――――(LMC-Or1)]   #<br>#
    [G ⊢c M◁ (T ∨ U)] #<br>#
    #<br>#
    [G ⊢c M◁ U]   #<br>#
    [―――――――――――(LMC-Or2)]   #<br>#
    [G ⊢c M◁ (T ∨ U)]
    *)
(* Part of Figure 22 in "Pure methods for roDOT" *)
Definition ty_has_N_closure G e := ty_or_closure (ty_has_N G e).
Definition ty_has_top_closure G e := ty_or_closure (ty_has_top_atomic G e).

(** ** Closure properties *)

Lemma ty_has_M_closure_tyq G e U V:
  ty_has_M_closure G e U ->
  typing_equiv e U V ->
  ty_has_M_closure G e V.
Proof.
  introv Hh Htyq.
  unfold ty_has_M_closure in *.
  reverts V.
  induction Hh; intros.
  -
  eapply ty_or_closure_atomic.
  eapply ty_has_M_tyq; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_l; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_r; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_top; eauto.
Qed.

Lemma ty_has_N_closure_tyq G e U V:
  ty_has_N_closure G e U ->
  typing_equiv e U V ->
  ty_has_N_closure G e V.
Proof.
  introv Hh Htyq.
  unfold ty_has_N_closure in *.
  reverts V.
  induction Hh; intros.
  -
  eapply ty_or_closure_atomic.
  eapply ty_has_N_tyq; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_l; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_r; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_top; eauto.
Qed.

Lemma ty_has_top_closure_tyq G e U V:
  ty_has_top_closure G e U ->
  typing_equiv e U V ->
  ty_has_top_closure G e V.
Proof.
  introv Hh Htyq.
  unfold ty_has_top_closure in *.
  reverts V.
  induction Hh; intros.
  -
  eapply ty_or_closure_atomic.
  eapply ty_has_top_atomic_tyq; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_l; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_r; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_top; eauto.
Qed.



Lemma ty_has_M_closure_lc_subtyp G e oc U V:
  inert G ->
  ty_has_M_closure G e (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_has_M_closure G e (logctx_typ oc V).
Proof.
  introv Hi Hh Hsu.
  remember (logctx_typ oc U).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_or_closure_atomic.
    eapply ty_has_M_lc_subtyp; eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    inversions Hsu.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    inversions Hsu.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    inversions Hsu.
Qed.
Lemma ty_has_N_closure_lc_subtyp G e oc U V:
  inert G ->
  ty_has_N_closure G e (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_has_N_closure G e (logctx_typ oc V).
Proof.
  introv Hi Hh Hsu.
  remember (logctx_typ oc U).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_or_closure_atomic.
    eapply ty_has_N_lc_subtyp; eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    inversions Hsu.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    inversions Hsu.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    inversions Hsu.
Qed.
Lemma ty_has_top_closure_lc_subtyp G e oc U V:
  inert G ->
  ty_has_top_closure G e (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_has_top_closure G e (logctx_typ oc V).
Proof.
  introv Hi Hh Hsu.
  remember (logctx_typ oc U).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_or_closure_atomic.
    eapply ty_has_top_atomic_lc_subtyp; eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    inversions Hsu.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    inversions Hsu.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    inversions Hsu.
Qed.


Lemma ty_has_N_closure_lc_rev_subtyp G e oc U V:
  inert G ->
  ty_has_N_closure G e (logctx_typ oc U) ->
  subtyp_atomic G e V U ->
  ty_has_N_closure G e (logctx_typ oc V) \/ ty_has_top_atomic G e U.
Proof.
  introv Hi Hh Hsu.
  remember (logctx_typ oc U).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - forwards Hhas: ty_has_N_lc_rev_subtyp H; fweauto.
    destructecs Hhas.
    + left. eapply ty_or_closure_atomic; eauto.
    + right. eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    *
    inversions Hsu.
    *
    forwards Hhas: IHHh; fweauto.
    destructecs Hhas.
    + left. eapply ty_or_closure_l. eauto.
    + eauto.
    *
    left. eapply ty_or_closure_l; eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    *
    inversions Hsu.
    * left. eapply ty_or_closure_r; eauto.
    *
    forwards Hhas: IHHh; fweauto.
    destructecs Hhas.
    + left. eapply ty_or_closure_r. eauto.
    + eauto.

  -
    destruct oc; simpls; inversions Heqt.
    inversions Hsu.
Qed.



(* INVERSION AND REPLACEMENT *)


Lemma ty_has_M_closure_lc_bot_replace G e oc U:
  ty_has_M_closure G e (logctx_typ oc typ_bot) ->
  ty_has_M_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc typ_bot).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_M_lc_bot_false in H. contradiction.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.
Lemma ty_has_N_closure_lc_bot_replace G e oc U:
  ty_has_N_closure G e (logctx_typ oc typ_bot) ->
  ty_has_N_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc typ_bot).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_N_lc_bot_false in H. contradiction.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.
Lemma ty_has_M_closure_lc_sel_replace G e oc U x A y:
  ty_has_M_closure G e (logctx_typ oc (typ_sel x A y)) ->
  ty_has_M_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_sel x A y)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_M_lc_sel_false in H. contradiction.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.
Lemma ty_has_N_closure_lc_sel_replace G e oc U x A y:
  ty_has_N_closure G e (logctx_typ oc (typ_sel x A y)) ->
  ty_has_N_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_sel x A y)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_N_lc_sel_false in H. contradiction.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.

Lemma ty_has_M_closure_lc_bnd_replace G e oc U T:
  ty_has_M_closure G e (logctx_typ oc (typ_bnd T)) ->
  ty_has_M_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_bnd T)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_M_lc_bnd_false in H. contradiction.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.
Lemma ty_has_N_closure_lc_bnd_replace G e oc U T:
  ty_has_N_closure G e (logctx_typ oc (typ_bnd T)) ->
  ty_has_N_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_bnd T)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_N_lc_bnd_false in H. contradiction.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.

Lemma ty_has_M_closure_lc_and_replace G e oc U T1 T2:
  ty_has_M_closure G e (logctx_typ oc (typ_and T1 T2)) ->
  ty_has_M_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_and T1 T2)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_M_lc_and_false in H. contradiction.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.
Lemma ty_has_N_closure_lc_and_replace G e oc U T1 T2:
  ty_has_N_closure G e (logctx_typ oc (typ_and T1 T2)) ->
  ty_has_N_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_and T1 T2)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_N_lc_and_false in H. contradiction.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.
Lemma ty_has_top_closure_lc_and_replace G e oc U T1 T2:
  ty_has_top_closure G e (logctx_typ oc (typ_and T1 T2)) ->
  ty_has_top_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_and T1 T2)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_top_atomic_lc_and_false in H. contradiction.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.


Lemma ty_has_M_closure_oc_and_replace G e T oc U S:
  ty_has_M_closure G e (orctx_typ oc (typ_and T U)) ->
  ty_has_M_closure G e (orctx_typ oc S).
Proof.
  introv Hty.
  rewrite ! orctx_logctx_typ  in *.
  eapply ty_has_M_closure_lc_and_replace; eauto.
Qed.
Lemma ty_has_N_closure_oc_and_inv G e T oc U S:
  ty_has_N_closure G e (orctx_typ oc (typ_and T U)) ->
  ty_has_N_closure G e (orctx_typ oc S).
Proof.
  introv Hty.
  rewrite ! orctx_logctx_typ  in *.
  eapply ty_has_N_closure_lc_and_replace; eauto.
Qed.
Lemma ty_has_top_closure_oc_and_inv G e T oc U S:
  ty_has_top_closure G e (orctx_typ oc (typ_and T U)) ->
  ty_has_top_closure G e (orctx_typ oc S).
Proof.
  introv Hty.
  rewrite ! orctx_logctx_typ  in *.
  eapply ty_has_top_closure_lc_and_replace; eauto.
Qed.


Lemma ty_has_N_closure_replace G e oc U T:
  ty_has_N_closure G e (logctx_typ oc U) ->
  ty_has_N G e T ->
  ty_has_N_closure G e (logctx_typ oc T).
Proof.
  introv Hh HhN.
  remember (logctx_typ oc U).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_or_closure_atomic. eapply ty_has_N_lc_N_replace; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_atomic; auto.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_atomic; auto.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    eapply ty_or_closure_atomic; auto.
Qed.


Lemma ty_has_N_closure_lc_M_replace G e oc T:
  inert G ->
  renv_corr G e ->
  ty_has_N_closure G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_has_N_closure G e (logctx_typ oc T).
Proof.
  introv Hi Hcor Hh.
  remember (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_or_closure_atomic. eapply ty_has_N_lc_M_replace; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.

Lemma ty_has_M_closure_lc_nocap_replace G e oc T:
  ty_has_M_closure G e (logctx_typ oc typ_nocap) ->
  ty_has_M_closure G e (logctx_typ oc T).
Proof.
  introv Hh.
  remember (logctx_typ oc typ_nocap).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_or_closure_atomic. eapply ty_has_M_lc_nocap_replace; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.


Lemma ty_has_N_closure_lc_has_top_replace G e oc U T:
  inert G ->
  renv_corr G e ->
  ty_has_N_closure G e (logctx_typ oc U) ->
  ty_has_top_atomic G e T ->
  ty_has_N_closure G e (logctx_typ oc T).
Proof.
  introv Hi Hcor Hh.
  remember (logctx_typ oc U).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_or_closure_atomic. eapply ty_has_N_lc_has_top_replace; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_atomic. eauto.
    eapply ty_or_closure_l; eauto. eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto. 
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_atomic. eauto.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.  eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    eapply ty_or_closure_atomic. eauto.
Qed.

Lemma ty_has_M_closure_lc_has_top_replace G e oc U T:
  inert G ->
  renv_corr G e ->
  ty_has_M_closure G e (logctx_typ oc U) ->
  ty_has_top_atomic G e T ->
  ty_has_M_closure G e (logctx_typ oc T).
Proof.
  introv Hi Hcor Hh.
  remember (logctx_typ oc U).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_or_closure_atomic. eapply ty_has_M_lc_has_top_replace; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_atomic. eapply ty_has_M_has_top; eauto.
    eapply ty_or_closure_l; eauto. eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto. 
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_atomic.  eapply ty_has_M_has_top; eauto.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.  eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    eapply ty_or_closure_atomic.  eapply ty_has_M_has_top; eauto.
Qed.

Lemma ty_has_N_atomic_lc_top_replace G e oc T:
  ty_has_N_atomic G e (logctx_typ oc T) ->
  ty_has_N_closure G e (logctx_typ oc typ_top).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh; eapply ty_or_closure_top.
Qed.
Lemma ty_has_top_atomic_lc_top_replace_N G e oc T:
  ty_has_top_atomic G e (logctx_typ oc T) ->
  ty_has_N_closure G e (logctx_typ oc typ_top).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh; eapply ty_or_closure_top.
Qed.
Lemma ty_has_top_atomic_lc_top_replace G e oc T:
  ty_has_top_atomic G e (logctx_typ oc T) ->
  ty_has_top_closure G e (logctx_typ oc typ_top).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh; eapply ty_or_closure_top.
Qed.
Lemma ty_has_N_atomic_lc_top_replace_top G e oc T:
  ty_has_N_atomic G e (logctx_typ oc T) ->
  ty_has_top_closure G e (logctx_typ oc typ_top).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh; eapply ty_or_closure_top.
Qed.
Lemma ty_has_N_lc_top_replace G e oc T:
  ty_has_N G e (logctx_typ oc T) ->
  ty_has_N_closure G e (logctx_typ oc typ_top).
Proof.
  introv Hh.
  inversions Hh; eauto using ty_has_N_atomic_lc_top_replace, ty_has_top_atomic_lc_top_replace_N.
Qed.
Lemma ty_has_M_lc_top_replace G e oc T:
  ty_has_M G e (logctx_typ oc T) ->
  ty_has_M_closure G e (logctx_typ oc typ_top).
Proof.
  introv Hh.
  destruct oc; simpls; inversions Hh; eapply ty_or_closure_top.
Qed.
Lemma ty_has_N_closure_lc_top_replace G e oc T:
  ty_has_N_closure G e (logctx_typ oc T) ->
  ty_has_N_closure G e (logctx_typ oc typ_top).
Proof.
  introv Hh.
  remember (logctx_typ oc T).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_N_lc_top_replace; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_top.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_top.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    eapply ty_or_closure_top.
Qed.
Lemma ty_has_M_closure_lc_top_replace G e oc T:
  ty_has_M_closure G e (logctx_typ oc T) ->
  ty_has_M_closure G e (logctx_typ oc typ_top).
Proof.
  introv Hh.
  remember (logctx_typ oc T).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_M_lc_top_replace; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_top.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_top.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    eapply ty_or_closure_top.
Qed.

Lemma ty_has_M_closure_lc_NM_replace G e oc U T1 T2:
  ty_has_M_closure G e (logctx_typ oc U) ->
  (mode_has_top_nm_complement typing_mode -> forall x0 : var, G ⁏ e ⊢# T1 /[-> x0] <: typ_bot) ->
  ty_has_M_closure G e (logctx_typ oc (typ_NM T1 T2)).
Proof.
  introv Hh Hbot.
  remember (logctx_typ oc U).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - destruct oc; simpls; inversions H.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_atomic; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_atomic; eauto.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_atomic; eauto.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_atomic; eauto.
Qed.

Lemma ty_has_N_closure_lc_NM_replace G e oc U T1 T2:
  ty_has_N_closure G e (logctx_typ oc U) ->
  ty_has_N_closure G e (logctx_typ oc (typ_NM T1 T2)).
Proof.
  introv Hh.
  remember (logctx_typ oc U).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - destruct oc; simpls.
    eapply ty_or_closure_l; eauto.
    eapply ty_or_closure_atomic; eauto.
    all: false; solve [eapply ty_has_N_lc_or_false with (oc0:=logctx_nil); eauto | eapply ty_has_N_lc_and_false with (oc0:=logctx_nil); eauto].
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l; eauto.
    eapply ty_or_closure_atomic; eauto.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_l; eauto.
    eapply ty_or_closure_atomic; eauto.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    eapply ty_or_closure_l; eauto.
    eapply ty_or_closure_atomic; eauto.
Qed.

Lemma ty_has_M_closure_lc_or_diag G e oc U:
  ty_has_M_closure G e (logctx_typ oc (typ_or U U)) ->
  ty_has_M_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_or U U)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - destruct oc; simpls; inversions H.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls; eauto.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.

Lemma ty_has_M_closure_lc_or_inv G e oc T U:
  ty_has_M_closure G e (logctx_typ oc (typ_or T U)) ->
  ty_has_M_closure G e (logctx_typ oc T) \/  ty_has_M_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_or T U)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - destruct oc; simpls; inversions H.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    forwards HHh: IHHh. eauto. destructecs HHh.
    left.
    eapply ty_or_closure_l. eauto.
    right.
    eapply ty_or_closure_l. eauto.
    left.
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    left.
    eapply ty_or_closure_r. eauto.
    forwards HHh: IHHh. eauto. destructecs HHh.
    left.
    eapply ty_or_closure_r. eauto.
    right.
    eapply ty_or_closure_r. eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.

Lemma ty_has_N_closure_lc_or_diag G e oc U:
  ty_has_N_closure G e (logctx_typ oc (typ_or U U)) ->
  ty_has_N_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_or U U)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - false; eapply ty_has_N_lc_or_false; eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls; eauto.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.

Lemma ty_has_N_closure_lc_or_inv G e oc T U:
  ty_has_N_closure G e (logctx_typ oc (typ_or T U)) ->
  ty_has_N_closure G e (logctx_typ oc T) \/  ty_has_N_closure G e (logctx_typ oc U).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_or T U)).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - forwards: ty_has_N_lc_inv H.  subst; simpls. inversions H; inversions H0.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    forwards HHh: IHHh. eauto. destructecs HHh.
    left.
    eapply ty_or_closure_l. eauto.
    right.
    eapply ty_or_closure_l. eauto.
    left.
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt; simpls; eauto.
    left.
    eapply ty_or_closure_r. eauto.
    forwards HHh: IHHh. eauto. destructecs HHh.
    left.
    eapply ty_or_closure_r. eauto.
    right.
    eapply ty_or_closure_r. eauto.
  -
    destruct oc; simpls; inversions Heqt.
Qed.

Lemma ty_has_N_closure_lc_typ_normal_replace G e oc A SL SU TL TU:
  ty_has_N_closure G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_normal A) SL SU))) ->
  ty_has_N_closure G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_normal A) TL TU))).
Proof.
  introv Hh.
  dependent induction Hh.
  - eapply ty_or_closure_atomic.
    eapply ty_has_N_lc_typ_normal_replace; eauto.
  - destruct oc; simpls; inversions x.
    forwards: IHHh; eauto.
    eapply ty_or_closure_l; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; simpls; inversions x.
    eapply ty_or_closure_r; eauto.
    forwards: IHHh; eauto.
    eapply ty_or_closure_r; eauto.
  - destruct oc; simpls; inversions x.
Qed.

Lemma ty_has_N_closure_lc_typ_replace G e oc A SL SU:
  ty_has_N_closure G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  (forall  TL TU, ty_has_N_closure G e (logctx_typ oc (typ_rcd (dec_typ A TL TU)))) \/ ty_has_top_atomic G e (typ_rcd (dec_typ A SL SU)).
Proof.
  introv Hh.
  dependent induction Hh.
  - forwards: ty_has_N_lc_typ_replace H; fweauto.
    destructecs H0.
    left; intros; eapply ty_or_closure_atomic; eauto.
    eauto.
  - destruct oc; simpls; inversions x.
    forwards: IHHh; eauto.
    destructecs H.
    left. intros.
    eapply ty_or_closure_l; eauto. eapply H.
    eauto.
    left. intros.
    eapply ty_or_closure_l; eauto.
  - destruct oc; simpls; inversions x.
    +
    left. intros.
    eapply ty_or_closure_r; eauto.
    +
    forwards: IHHh; fweauto.
    destructecs H.
    left. intros.
    eapply ty_or_closure_r; eauto. eapply H.
    eauto.
  - destruct oc; simpls; inversions x.
Qed.

Lemma ty_has_M_closure_lc_typ_replace G e oc A SL SU TL TU:
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  (mode_has_top_nm_complement typing_mode -> forall x : var, G⁏e ⊢# TL /[-> x] <: typ_bot) ->
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A TL TU))).
Proof.
  introv Hh Hsu.
  dependent induction Hh.
  - eapply ty_or_closure_atomic.
    eapply ty_has_M_lc_typ_replace; eauto.
  - destruct oc; simpls; inversions x.
    forwards: IHHh; eauto.
    eapply ty_or_closure_l; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; simpls; inversions x.
    eapply ty_or_closure_r; eauto.
    forwards: IHHh; eauto.
    eapply ty_or_closure_r; eauto.
  - destruct oc; simpls; inversions x.
Qed.

Lemma ty_has_M_closure_lc_typ_replace' G e oc A SL SU TL TU:
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A TL TU))) \/ (mode_has_top_nm_complement typing_mode -> forall x : var, G⁏e ⊢# SL /[-> x] <: typ_bot) .
Proof.
  introv Hh.
  dependent induction Hh.
  - forwards: ty_has_M_lc_typ_replace' SL SU TL TU; fweauto.
    destructecs H0; eauto 2.
    left.
    eapply ty_or_closure_atomic; eauto.
    
  - destruct oc; simpls; inversions x.
    forwards: IHHh; fweauto.
    destructecs H.
    left.
    eapply ty_or_closure_l; eauto.
    eauto.
    left.
    eapply ty_or_closure_l; eauto.
  - destruct oc; simpls; inversions x.
    left. eapply ty_or_closure_r; eauto.
    forwards: IHHh; fweauto.
    destructecs H.
    left.
    eapply ty_or_closure_r; eauto.
    eauto.
  - destruct oc; simpls; inversions x.
Qed.

Lemma ty_has_M_closure_lc_typ_dist G e oc A SL SU TL TU:
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A TL TU))) ->
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))).
Proof.
  introv HhS HhT.
  forwards HrS: ty_has_M_closure_lc_typ_replace' (typ_or TL SL) (typ_and TU SU) HhS.
  destructecs HrS.
  auto.
  forwards HrT: ty_has_M_closure_lc_typ_replace' (typ_or TL SL) (typ_and TU SU) HhT.
  destructecs HrT.
  auto.
  eapply ty_has_M_closure_lc_typ_replace. eapply HhS.
  intros.
  eapply subtyp_or2_t; eauto.
Qed.


Lemma ty_has_M_closure_lc_typ_pfor_replace G e x oc A SL SU TL TU:
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  inert G ->
  x \in ctx_dom G ->
  ty_pf_or_none G x A TL TU ->
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A TL TU))).
Proof.
  introv Hh Hi Hdom Hpf.
  dependent induction Hh.
  - eapply ty_or_closure_atomic.
    eapply ty_has_M_lc_typ_pfor_replace; eauto.
  - destruct oc; simpls; inversions x.
    forwards: IHHh; eauto.
    eapply ty_or_closure_l; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; simpls; inversions x.
    eapply ty_or_closure_r; eauto.
    forwards: IHHh; eauto.
    eapply ty_or_closure_r; eauto.
  - destruct oc; simpls; inversions x.
Qed.

Lemma ty_has_N_oc_orl G e oc U T:
  ty_has_N G e (orctx_typ oc U) ->
  ty_has_N_closure G e (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  reverts T U.
  induction oc; intros; simpls.
  - eapply ty_or_closure_l. apply ty_or_closure_atomic. eauto.
  - false. eapply ty_has_N_lc_or_false with (oc0:=logctx_nil) in Hty. eauto.
  - false. eapply ty_has_N_lc_or_false with (oc0:=logctx_nil) in Hty. eauto.
Qed.

Lemma ty_has_N_closure_oc_orl G e oc U T:
  ty_has_N_closure G e (orctx_typ oc U) ->
  ty_has_N_closure G e (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  remember (orctx_typ oc U).
  reverts oc U.
  induction Hty; intros; subst.
  - eapply ty_has_N_oc_orl; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_l. eauto.
    + eapply ty_or_closure_l. eapply IHHty. eauto.
    + eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r. eapply IHHty. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_top.
Qed.

Lemma ty_has_N_closure_oclc_orl G e oc2 oc U T:
  ty_has_N_closure G e (orctx_typ oc2 (logctx_typ oc U)) ->
  ty_has_N_closure G e (orctx_typ oc2 (logctx_typ oc (typ_or U T))).
Proof.
  introv Hty.
  reverts oc2.
  induction oc; intros; simpls.
  - eapply ty_has_N_closure_oc_orl; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_has_N_closure_oc_and_inv in Hty; eauto.
  - eapply ty_has_N_closure_oc_and_inv in Hty; eauto.
Qed.
Lemma ty_has_N_closure_lc_orl G e oc U T:
  ty_has_N_closure G e (logctx_typ oc U) ->
  ty_has_N_closure G e (logctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  eapply ty_has_N_closure_oclc_orl with (oc2:=orctx_nil); eauto.
Qed.

Lemma ty_has_N_oc_orr G e oc U T:
  ty_has_N G e (orctx_typ oc T) ->
  ty_has_N_closure G e (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  reverts T U.
  induction oc; intros; simpls.
  - eapply ty_or_closure_r. apply ty_or_closure_atomic. eauto.
  - false. eapply ty_has_N_lc_or_false with (oc0:=logctx_nil) in Hty. eauto.
  - false. eapply ty_has_N_lc_or_false with (oc0:=logctx_nil) in Hty. eauto.
Qed.


Lemma ty_has_N_closure_oc_orr G e oc U T:
  ty_has_N_closure G e (orctx_typ oc T) ->
  ty_has_N_closure G e (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  remember (orctx_typ oc T).
  reverts oc T.
  induction Hty; intros; subst.
  - eapply ty_has_N_oc_orr; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_l. eauto.
    + eapply ty_or_closure_l. eapply IHHty. eauto.
    + eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r. eapply IHHty. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_top.
Qed.

Lemma ty_has_N_closure_oclc_orr G e oc2 oc U T:
  ty_has_N_closure G e (orctx_typ oc2 (logctx_typ oc T)) ->
  ty_has_N_closure G e (orctx_typ oc2 (logctx_typ oc (typ_or U T))).
Proof.
  introv Hty.
  reverts oc2.
  induction oc; intros; simpls.
  - eapply ty_has_N_closure_oc_orr; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_has_N_closure_oc_and_inv in Hty; eauto.
  - eapply ty_has_N_closure_oc_and_inv in Hty; eauto.
Qed.
Lemma ty_has_N_closure_lc_orr G e oc U T:
  ty_has_N_closure G e (logctx_typ oc T) ->
  ty_has_N_closure G e (logctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  eapply ty_has_N_closure_oclc_orr with (oc2:=orctx_nil); eauto.
Qed.


Lemma ty_has_M_oc_orl G e oc U T:
  ty_has_M G e (orctx_typ oc U) ->
  ty_has_M_closure G e (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  remember (orctx_typ oc U).
  reverts oc U.
  induction Hty; intros; subst; destruct oc; simpls; inversions Heqt.
  - eapply ty_or_closure_l. eapply ty_or_closure_atomic. eauto.
Qed.

Lemma ty_has_M_closure_oc_orl G e oc U T:
  ty_has_M_closure G e (orctx_typ oc U) ->
  ty_has_M_closure G e (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  remember (orctx_typ oc U).
  reverts oc U.
  induction Hty; intros; subst.
  - eapply ty_has_M_oc_orl; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_l. eauto.
    + eapply ty_or_closure_l. eapply IHHty. eauto.
    + eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r. eapply IHHty. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_top.
Qed.

Lemma ty_has_M_oc_orr G e oc U T:
  ty_has_M G e (orctx_typ oc T) ->
  ty_has_M_closure G e (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  remember (orctx_typ oc T).
  reverts oc T.
  induction Hty; intros; subst; destruct oc; simpls; inversions Heqt.
  - eapply ty_or_closure_r. eapply ty_or_closure_atomic. eauto.
Qed.

Lemma ty_has_M_closure_oc_orr G e oc U T:
  ty_has_M_closure G e (orctx_typ oc T) ->
  ty_has_M_closure G e (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  remember (orctx_typ oc T).
  reverts oc T.
  induction Hty; intros; subst.
  - eapply ty_has_M_oc_orr; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_l. eauto.
    + eapply ty_or_closure_l. eapply IHHty. eauto.
    + eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r. eapply IHHty. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_top.
Qed.

Lemma ty_has_M_closure_oclc_orl G e oc2 oc U T:
  ty_has_M_closure G e (orctx_typ oc2 (logctx_typ oc U)) ->
  ty_has_M_closure G e (orctx_typ oc2 (logctx_typ oc (typ_or U T))).
Proof.
  introv Hty.
  reverts oc2.
  induction oc; intros; simpls.
  - eapply ty_has_M_closure_oc_orl; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_has_M_closure_oc_and_replace in Hty; eauto.
  - eapply ty_has_M_closure_oc_and_replace in Hty; eauto.
Qed.
Lemma ty_has_M_closure_lc_orl G e oc U T:
  ty_has_M_closure G e (logctx_typ oc U) ->
  ty_has_M_closure G e (logctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  eapply ty_has_M_closure_oclc_orl with (oc2:=orctx_nil); eauto.
Qed.

Lemma ty_has_M_closure_oclc_orr G e oc2 oc U T:
  ty_has_M_closure G e (orctx_typ oc2 (logctx_typ oc T)) ->
  ty_has_M_closure G e (orctx_typ oc2 (logctx_typ oc (typ_or U T))).
Proof.
  introv Hty.
  reverts oc2.
  induction oc; intros; simpls.
  - eapply ty_has_M_closure_oc_orr; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_has_M_closure_oc_and_replace in Hty; eauto.
  - eapply ty_has_M_closure_oc_and_replace in Hty; eauto.
Qed.
Lemma ty_has_M_closure_lc_orr G e oc U T:
  ty_has_M_closure G e (logctx_typ oc T) ->
  ty_has_M_closure G e (logctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  eapply ty_has_M_closure_oclc_orr with (oc2:=orctx_nil); eauto.
Qed.



Lemma ty_has_N_closure_or_inv G e oc T S:
  ty_has_N_closure G e (logctx_typ oc (typ_or T S)) ->
  ty_has_N_closure G e (logctx_typ oc T) \/ ty_has_N_closure G e (logctx_typ oc S).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_or T S)).
  reverts oc T S.
  induction Hh; intros; subst.
  - false; eapply ty_has_N_lc_or_false; eauto.
  - destruct oc; simpls; inversions Heqt.
    + auto.
    + forwards: IHHh; eauto. destruct_all. left. eapply ty_or_closure_l. auto. right. eapply ty_or_closure_l. auto. 
    + left. eapply ty_or_closure_l. auto.
  - destruct oc; simpls; inversions Heqt.
    + auto.
    + right. eapply ty_or_closure_r. auto.
    + forwards: IHHh; eauto. destruct_all. left. eapply ty_or_closure_r. auto. right. eapply ty_or_closure_r. auto.
  - destruct oc; simpl in Heqt; inversions Heqt. 
Qed.


Lemma ty_has_M_closure_or_inv G e oc T S:
  ty_has_M_closure G e (logctx_typ oc (typ_or T S)) ->
  ty_has_M_closure G e (logctx_typ oc T) \/ ty_has_M_closure G e (logctx_typ oc S).
Proof.
  introv Hh.
  remember (logctx_typ oc (typ_or T S)).
  reverts oc T S.
  induction Hh; intros; subst.
  - false; eapply ty_has_M_lc_or_false; eauto.
  - destruct oc; simpls; inversions Heqt.
    + auto.
    + forwards: IHHh; eauto. destruct_all. left. eapply ty_or_closure_l. auto. right. eapply ty_or_closure_l. auto. 
    + left. eapply ty_or_closure_l. auto.
  - destruct oc; simpls; inversions Heqt.
    + auto.
    + right. eapply ty_or_closure_r. auto.
    + forwards: IHHh; eauto. destruct_all. left. eapply ty_or_closure_r. auto. right. eapply ty_or_closure_r. auto.
  - destruct oc; simpl in Heqt; inversions Heqt. 
Qed.

Lemma ty_has_N_closure_lc_toplike_replace G e oc U T:
  ty_has_N_closure G e (logctx_typ oc U) ->
  typ_logic_toplike G e T ->
  ty_has_N_closure G e (logctx_typ oc T).
Proof.
  introv Hh Ht.
  inversions Ht.
  eapply ty_has_N_closure_lc_top_replace; eauto.
  eapply ty_has_N_closure_lc_NM_replace; eauto.
  eapply ty_has_N_closure_lc_has_top_replace; eauto.
Qed.


Lemma ty_has_N_closure_lc_typ_dist G e oc A SL SU TL TU:
  inert G ->
  renv_corr G e ->
  ty_has_N_closure G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_has_N_closure G e (logctx_typ oc (typ_rcd (dec_typ A TL TU))) ->
  ty_has_N_closure G e (logctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))).
Proof.
  introv Hi Hcor HhS HhT.
  forwards HrS: ty_has_N_closure_lc_typ_replace HhS.
  destructecs HrS. auto.
  forwards HrT: ty_has_N_closure_lc_typ_replace HhT.
  destructecs HrT. auto.

  eapply ty_has_N_closure_lc_toplike_replace.
  eapply HhS.
  inversions HrS.
  inversions HrT.
  eapply typ_logic_toplike_has_top; auto 1.
  intros. eapply subtyp_or2_t; eauto.
  intros. eapply subtyp_and2_t; eauto.
Qed.

Lemma ty_has_N_lc_nocap_replace G e oc T:
  ty_has_N G e (logctx_typ oc T) ->
  ty_has_N_closure G e (logctx_typ oc typ_nocap).
Proof.
  introv Hh.
  eapply ty_or_closure_atomic.
  eapply ty_has_N_lc_N_replace; eauto.
Qed.
Lemma ty_has_N_closure_lc_nocap_replace G e oc T:
  ty_has_N_closure G e (logctx_typ oc T) ->
  ty_has_N_closure G e (logctx_typ oc typ_nocap).
Proof.
  introv Hh.
  remember (logctx_typ oc T).
  reverts oc.
  induction Hh; intros; subst; simpls.
  - eapply ty_has_N_lc_nocap_replace; eauto.
  - destruct oc; inversions Heqt; simpls.
    eapply ty_or_closure_atomic; eauto.
    eapply ty_or_closure_l.
    eapply IHHh; eauto.
    eapply ty_or_closure_l; eauto.
  - destruct oc; inversions Heqt; simpls.
     eapply ty_or_closure_atomic; eauto.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_r.
    eapply IHHh; eauto.
  -
    destruct oc; simpls; inversions Heqt.
     eapply ty_or_closure_atomic; eauto.
Qed.


Lemma ty_has_N_closure_closure_replace G e oc U T:
  ty_has_N_closure G e (logctx_typ oc U) ->
  ty_has_N_closure G e T ->
  ty_has_N_closure G e (logctx_typ oc T).
Proof.
  introv Hh HhN.

  induction HhN.
  - eapply ty_has_N_closure_replace; eauto.
  - eapply ty_has_N_closure_lc_orl; eauto.
  - eapply ty_has_N_closure_lc_orr; eauto.
  - eapply ty_has_N_closure_lc_top_replace; eauto.
Qed.

Lemma ty_has_M_closure_lc_cases G e oc T0:
  ty_has_M_closure G e (logctx_typ oc T0) ->
  ty_has_M_closure G e T0 \/ (forall T1, ty_has_M_closure G e (logctx_typ oc T1)).
Proof.
  introv Hty.
  remember (logctx_typ oc T0).
  reverts oc T0.
  induction Hty; intros; subst.
  - forwards: ty_has_M_lc_inv H. subst. left. eapply ty_or_closure_atomic; eauto.
  - destruct oc; simpls; inversions Heqt.
    + left. eapply ty_or_closure_l. eauto.
    + forwards: IHHty; eauto. destruct H; eauto. right. intros. eapply ty_or_closure_l. eapply H.
    + right. intros. eapply ty_or_closure_l. eauto.
  -     destruct oc; simpls; inversions Heqt.
    + left. eapply ty_or_closure_r. eauto.

    + right. intros. eapply ty_or_closure_r. eauto.
    + forwards: IHHty; eauto. destruct H; eauto. right. intros. eapply ty_or_closure_r. eapply H.
  -  destruct oc; simpls; inversions Heqt.
   left. eapply ty_or_closure_top; eauto.
Qed.


Lemma ty_has_N_closure_tight G e T:
  ty_has_N_closure G e T ->
  G⁏e ⊢# typ_nocap <: T.
Proof.
  introv Hh.
  induction Hh; eauto 2 using ty_has_N_tight.
Qed.
Lemma ty_has_M_closure_tight G e T:
  ty_has_M_closure G e T ->
  G⁏e ⊢# (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_top typ_bot)) <: T.
Proof.
  introv Hh.
  induction Hh; eauto 2 using ty_has_M_tight.
Qed.
Lemma ty_has_M_closure_tight_bot G e T:
  mode_has_top_nm_complement typing_mode ->
  ty_has_M_closure G e T ->
  G⁏e ⊢# (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) <: T.
Proof.
  introv Hm Hh.
  induction Hh; eauto 2 using ty_has_M_tight.
  inversions H.
  eauto.
Qed.


Lemma ty_has_N_closure_atomic_basic G e T:
  ty_has_N_closure G e T ->
  typ_atomic T ->
  ty_has_N G e T.
Proof.
  introv Hty Ha.
  induction Hty; eauto; inversions Ha.
Qed.


Lemma ty_has_N_or_closure_lc_cases  G e oc T:
  ty_has_N_closure G e (logctx_typ oc T) ->
  ty_has_N_closure G e T \/ (forall U, ty_has_N_closure G e (logctx_typ oc U)).
Proof.
  introv Hh.
  remember (logctx_typ oc T).
  reverts T oc.
  induction Hh; intros; subst.
  - forwards: ty_has_N_lc_inv H; subst; simpls. left. eapply ty_or_closure_atomic; eauto.
  - destruct oc; simpls; subst.
    +
    forwards: IHHh T logctx_nil; simpls. auto. destruct_all. left. eapply ty_or_closure_l; eauto. eauto.
    +
    inversions Heqt.
    forwards: IHHh T0 oc; simpls. auto. destruct_all. left. eauto. right. intros. eapply ty_or_closure_l; eauto. eapply H.
    +
    inversions Heqt.
    forwards: IHHh T1 logctx_nil; simpls. auto. destruct_all.  right. intros. eapply ty_or_closure_l; eauto. right. intros. eapply ty_or_closure_r; eauto. eapply H.
    + inversion Heqt.
    + inversion Heqt.
  - destruct oc; simpls; subst.
    +
    forwards: IHHh T logctx_nil; simpls. auto. destruct_all. left. eapply ty_or_closure_r; eauto. eauto.
    +
    inversions Heqt.
    forwards: IHHh T1 logctx_nil; simpls. auto. destruct_all.  right. intros. eapply ty_or_closure_r; eauto. right. intros. eapply ty_or_closure_r; eauto.

    +
    inversions Heqt.
    forwards: IHHh T0 oc; simpls. auto. destruct_all. left. eauto. right. intros. eapply ty_or_closure_r; eauto. eapply H.
    + inversion Heqt.
    + inversion Heqt.
  - destruct oc; simpls; subst; try discriminate. left. eapply ty_or_closure_top; eauto.
Qed.

Lemma ty_has_N_or_closure_oc_cases  G e oc T:
  ty_has_N_closure G e (orctx_typ oc T) ->
  ty_has_N_closure G e T \/ (forall U, ty_has_N_closure G e (orctx_typ oc U)).
Proof.
  introv Hh.
  forwards: ty_has_N_or_closure_lc_cases (orctx_logctx oc) T.
  rewrite <- orctx_logctx_typ. auto.
eapply Hh.
  destruct H. auto.
  right. intro.
  rewrite  orctx_logctx_typ. auto.
Qed.


Lemma ty_has_M_or_closure_lc_cases G e  oc T:
  ty_has_M_closure G e (logctx_typ oc T) ->
  ty_has_M_closure G e T \/ (forall U, ty_has_M_closure G e (logctx_typ oc U)).
Proof.
  introv Hh.
  remember (logctx_typ oc T).
  reverts T oc.
  induction Hh; intros; subst.
  - forwards: ty_has_M_lc_inv H; subst; simpls. left. eapply ty_or_closure_atomic; eauto.
  - destruct oc; simpls; subst.
    +
    forwards: IHHh T logctx_nil; simpls. auto. destruct_all. left. eapply ty_or_closure_l; eauto. eauto.
    +
    inversions Heqt.
    forwards: IHHh T0 oc; simpls. auto. destruct_all. left. eauto. right. intros. eapply ty_or_closure_l; eauto. eapply H.
    +
    inversions Heqt.
    forwards: IHHh T1 logctx_nil; simpls. auto. destruct_all.  right. intros. eapply ty_or_closure_l; eauto. right. intros. eapply ty_or_closure_r; eauto. eapply H.
    + inversion Heqt.
    + inversion Heqt.
  - destruct oc; simpls; subst.
    +
    forwards: IHHh T logctx_nil; simpls. auto. destruct_all. left. eapply ty_or_closure_r; eauto. eauto.
    +
    inversions Heqt.
    forwards: IHHh T1 logctx_nil; simpls. auto. destruct_all.  right. intros. eapply ty_or_closure_r; eauto. right. intros. eapply ty_or_closure_r; eauto.

    +
    inversions Heqt.
    forwards: IHHh T0 oc; simpls. auto. destruct_all. left. eauto. right. intros. eapply ty_or_closure_r; eauto. eapply H.
    + inversion Heqt.
    + inversion Heqt.
  - destruct oc; simpls; subst; try discriminate. left. eapply ty_or_closure_top; eauto.
Qed.

Lemma ty_has_M_closure_has_N_top G e T:
  ty_has_M_closure G e T ->
  ty_has_N G e T ->
  ty_has_top_atomic G e T.
Proof.
  introv HhM HhN.
  inversions HhN.
  - inversions H; inversions HhM; inversions H.
  - eauto.
Qed.


End WithTypingMode.

