Set Implicit Arguments.

Require Import Coq.Program.Equality.
Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode.
Require Import Equivalence GeneralTyping Indep.
Require Import PreciseTyping RecordTypes InertContext EnvCorrespondence.
Require Import TightTyping InertTightSubtyping TightTypingHelpers.
Require Import Orctx Logctx OrctxLogctx Logpath.

Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure.

Section WithTypingMode.
Context {typing_mode: Mode}.

(* Logic Has-N
  Unions and intersection of types.
*)

Inductive ty_has_N_logic (G: ctx) (e: renv): typ -> Prop :=
| ty_has_N_logic_or T:
    ty_has_N_closure G e T ->
    ty_has_N_logic G e T
| ty_has_N_logic_oc_and oc T U:
    ty_has_N_logic G e (orctx_typ oc T) ->
    ty_has_N_logic G e (orctx_typ oc U) ->
    ty_has_N_logic G e (orctx_typ oc (typ_and T U))
.

(* CLOSURE PROPERTIES *)
Lemma ty_has_N_logic_tyq G e U V:
  ty_has_N_logic G e U ->
  typing_equiv e U V ->
  ty_has_N_logic G e V.
Proof.
  introv Hty Htyq.
  reverts V.
  induction Hty; intros.
  - eapply ty_has_N_logic_or; eauto.
    eapply ty_has_N_closure_tyq; eauto.
  - eapply orctx_tyq_case in Htyq; destruct_all; subst.
    inversions_tyq H1.
    eapply ty_has_N_logic_oc_and.
    eapply IHHty1; eauto using orctx_tyq_typ.
    eapply IHHty2; eauto using orctx_tyq_typ.
Qed.

(* FORWARD DERIVATION *)

Lemma ty_has_N_logic_and G e T U:
    ty_has_N_logic G e T ->
    ty_has_N_logic G e U ->
    ty_has_N_logic G e (typ_and T U).
Proof.
  introv HtyT HtyU.
  forwards: ty_has_N_logic_oc_and orctx_nil T U; eauto.
Qed.

(* INVERSION *)

Lemma ty_has_N_logic_oc_and_invl  G e oc T U:
  ty_has_N_logic G e (orctx_typ oc (typ_and T U)) ->
  ty_has_N_logic G e (orctx_typ oc T).
Proof.
  introv Hty.
  remember (orctx_typ oc (typ_and T U)).
  reverts T U oc.
  induction Hty; intros; subst.
  -
    eapply ty_has_N_closure_oc_and_inv in H.
    eapply ty_has_N_logic_or; eauto.
  -
  eapply orctx_intersect_and_and_cases in Heqt.
  destruct_all; subst. eauto.
  forwards ocST1 Hrest: (H1 T0).
  destructc Hrest as HocST1 HocS1T1 HocS2T1 Hf.
      forwards Hty01: IHHty1; eauto.
      forwards Hty02: IHHty2; eauto.
      rewrite <- HocST1.
      rewrite <- HocS1T1 in Hty01.
      rewrite <- HocS2T1 in Hty02.
      eapply ty_has_N_logic_oc_and; eauto.
Qed.
Lemma ty_has_N_logic_oc_and_invr G e  oc T U:
  ty_has_N_logic  G e (orctx_typ oc (typ_and T U)) ->
  ty_has_N_logic  G e (orctx_typ oc U).
Proof.
  introv Hty.
  remember (orctx_typ oc (typ_and T U)).
  reverts T U oc.
  induction Hty; intros; subst.
  -
    eapply ty_has_N_closure_oc_and_inv in H.
    eapply ty_has_N_logic_or; eauto.
  -
  eapply orctx_intersect_and_and_cases in Heqt.
  destruct_all; subst. eauto.
  forwards ocST1 Hrest: (H1 U0).
  destructc Hrest as HocST1 HocS1T1 HocS2T1 Hf.
      forwards Hty01: IHHty1; eauto.
      forwards Hty02: IHHty2; eauto.
      rewrite <- HocST1.
      rewrite <- HocS1T1 in Hty01.
      rewrite <- HocS2T1 in Hty02.
      eapply ty_has_N_logic_oc_and; eauto.
Qed.
Lemma ty_has_N_logic_oclc_and_invl G e oc oc2 T U:
  ty_has_N_logic G e (orctx_typ oc (logctx_typ oc2 (typ_and T U))) ->
  ty_has_N_logic G e (orctx_typ oc (logctx_typ oc2 T)).
Proof.
  introv Hty.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_has_N_logic_oc_and_invl; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_has_N_logic_oc_and.
    forwards: ty_has_N_logic_oc_and_invl Hty. eauto.
    forwards: ty_has_N_logic_oc_and_invr Hty. eauto.
  - eapply ty_has_N_logic_oc_and.
    forwards: ty_has_N_logic_oc_and_invl Hty. eauto.
    forwards: ty_has_N_logic_oc_and_invr Hty. eauto.
Qed.

Lemma ty_has_N_logic_oclc_and_invr G e oc oc2 T U:
  ty_has_N_logic G e (orctx_typ oc (logctx_typ oc2 (typ_and T U))) ->
  ty_has_N_logic G e (orctx_typ oc (logctx_typ oc2 U)).
Proof.
  introv Hty.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_has_N_logic_oc_and_invr; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_has_N_logic_oc_and.
    forwards: ty_has_N_logic_oc_and_invl Hty. eauto.
    forwards: ty_has_N_logic_oc_and_invr Hty. eauto.
  - eapply ty_has_N_logic_oc_and.
    forwards: ty_has_N_logic_oc_and_invl Hty. eauto.
    forwards: ty_has_N_logic_oc_and_invr Hty. eauto.
Qed.
Lemma ty_has_N_logic_lc_and_invl G e oc2 T U:
  ty_has_N_logic G e ( (logctx_typ oc2 (typ_and T U))) ->
  ty_has_N_logic G e ( (logctx_typ oc2 T)).
Proof.
  introv Hty.
  applys ty_has_N_logic_oclc_and_invl orctx_nil; eauto.
Qed.
Lemma ty_has_N_logic_lc_and_invr G e oc2 T U:
  ty_has_N_logic G e ( (logctx_typ oc2 (typ_and T U))) ->
  ty_has_N_logic G e ( (logctx_typ oc2 U)).
Proof.
  introv Hty.
  applys ty_has_N_logic_oclc_and_invr orctx_nil; eauto.
Qed.

Lemma ty_has_N_logic_oclc_and  G e oc oc2 T U:
    ty_has_N_logic G e  (orctx_typ oc (logctx_typ oc2 T)) ->
    ty_has_N_logic G e  (orctx_typ oc (logctx_typ oc2 U)) ->
    ty_has_N_logic G e  (orctx_typ oc (logctx_typ oc2 (typ_and T U))).
Proof.
  introv HtyT HtyU.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_has_N_logic_oc_and; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_has_N_logic_oc_and.
    forwards: ty_has_N_logic_oc_and_invl HtyT.
    forwards: ty_has_N_logic_oc_and_invl HtyU.
    eauto.
    forwards: ty_has_N_logic_oc_and_invr HtyT.
    eauto.
  -  eapply ty_has_N_logic_oc_and.
    forwards: ty_has_N_logic_oc_and_invl HtyT.
    eauto.
    forwards: ty_has_N_logic_oc_and_invr HtyT.
    forwards: ty_has_N_logic_oc_and_invr HtyU.
    eauto.
Qed.

Lemma ty_has_N_logic_lc_and G e  oc2 T U:
    ty_has_N_logic G e (logctx_typ oc2 T) ->
    ty_has_N_logic G e (logctx_typ oc2 U) ->
    ty_has_N_logic G e (logctx_typ oc2 (typ_and T U)).
Proof.
  introv HtyT HtyU.
  forwards: ty_has_N_logic_oclc_and orctx_nil T U; simpls; eauto.
Qed.


Lemma ty_has_N_logic_toplike G e T:
  typ_logic_toplike G e T ->
  ty_has_N_logic G e T.
Proof.
  introv Htl.
  inversions Htl.
  - (* typ_top *)
    eapply ty_has_N_logic_or; eauto.
    eapply ty_or_closure_top.
  - (* typ_NM *)
    eapply ty_has_N_logic_or; eauto.
    eapply ty_or_closure_l.
    eapply ty_or_closure_atomic.
    eauto.
  -
    eapply ty_has_N_logic_or.
    eapply ty_or_closure_atomic.
    eapply ty_has_N_has_top; eauto.
Qed.

Lemma ty_has_N_logic_nocap G e:
  ty_has_N_logic G e typ_nocap.
Proof.
  eapply ty_has_N_logic_or.
  eapply ty_or_closure_atomic.
  eauto.
Qed.

Lemma ty_has_N_logic_orl G e  T U:
  ty_has_N_logic G e T ->
  ty_has_N_logic G e (typ_or T U).
Proof.
  introv Hty.
  induction Hty.
  - eapply ty_has_N_logic_or; eauto.
    eapply ty_or_closure_l. eauto.
  - rewrite orctx_fold_l in *.
    eapply ty_has_N_logic_oc_and; eauto.
Qed.
Lemma ty_has_N_logic_orr  G e T U:
  ty_has_N_logic G e U ->
  ty_has_N_logic G e (typ_or T U).
Proof.
  introv Hty.
  induction Hty.
  - eapply ty_has_N_logic_or; eauto.
    eapply ty_or_closure_r. eauto.
  - rewrite orctx_fold_r in *.
    eapply ty_has_N_logic_oc_and; eauto.
Qed.
Lemma ty_has_N_logic_oc G e oc U:
  ty_has_N_logic G e U ->
  ty_has_N_logic G e (orctx_typ oc U).
Proof.
  introv Hty.
  induction oc; simpls; eauto using ty_has_N_logic_orl, ty_has_N_logic_orr.
Qed.


Lemma ty_has_N_logic_oc_orr G e oc T U:
  ty_has_N_logic G e (orctx_typ oc U) ->
  ty_has_N_logic G e (orctx_typ oc (typ_or T U)).
Proof.
  introv Hty.
  remember (orctx_typ oc (U)).
  reverts T U oc.
  induction Hty; intros; subst.
  -
    eapply ty_has_N_closure_oc_orr in H.
    eapply ty_has_N_logic_or; eauto.
  -
    symmetry in Heqt.
    pose proof Heqt as Heqt0.
    eapply orctx_intersect_cases_r in Heqt. 2: intros; discriminate.
    destruct_all; subst.
    *
      rewrite orctx_fold_r.
      rewrite <- ! orctx_extend_typ.
      eapply ty_has_N_logic_oc_and.
      +
        rewrite ! orctx_extend_typ; simpls.
        eapply IHHty1.
        rewrite ! orctx_extend_typ; eauto.
      +
        rewrite ! orctx_extend_typ; simpls.
        eapply IHHty2.
        rewrite ! orctx_extend_typ; eauto.
    *
        forwards: orctx_separate_cross_r Heqt0 H (typ_or T0 U0).
        destruct_all.
        forwards HT: H1 T.
        forwards HU: H1 U.
        destruct_all.
        rewrite H0.
        eapply ty_has_N_logic_oc_and.
        rewrite <- H7. eauto.
        rewrite <- H5. eauto.
Qed.

Lemma ty_has_N_logic_oclc_orr G e  oc2 oc T U:
  ty_has_N_logic G e (orctx_typ oc2 (logctx_typ oc U)) ->
  ty_has_N_logic G e (orctx_typ oc2 (logctx_typ oc (typ_or T U))).
Proof.
  introv Hty.
  reverts oc2.
  induction oc; intros; simpls.
  - eapply ty_has_N_logic_oc_orr; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  -
    forwards: ty_has_N_logic_oc_and_invl Hty. 
    forwards: ty_has_N_logic_oc_and_invr Hty.
    forwards: IHoc; eauto.
    eapply ty_has_N_logic_oc_and; eauto.
  -
    forwards: ty_has_N_logic_oc_and_invl Hty. 
    forwards: ty_has_N_logic_oc_and_invr Hty.
    forwards: IHoc; eauto.
    eapply ty_has_N_logic_oc_and; eauto.
Qed.
Lemma ty_has_N_logic_lc_orr G e oc T U:
  ty_has_N_logic G e (logctx_typ oc U) ->
  ty_has_N_logic G e (logctx_typ oc (typ_or T U)).
Proof.
  introv Hty.
  applys ty_has_N_logic_oclc_orr orctx_nil; eauto.
Qed.

Lemma ty_has_N_logic_oc_orl G e oc T U:
  ty_has_N_logic G e (orctx_typ oc T) ->
  ty_has_N_logic G e (orctx_typ oc (typ_or T U)).
Proof.
  introv Hty.
  remember (orctx_typ oc T).
  reverts T U oc.
  induction Hty; intros; subst.
  -
    eapply ty_has_N_closure_oc_orl in H.
    eapply ty_has_N_logic_or; eauto.
  -
    symmetry in Heqt.
    pose proof Heqt as Heqt0.
    eapply orctx_intersect_cases_r in Heqt. 2: intros; discriminate.
    destruct_all; subst.
    *
      rewrite orctx_fold_l.
      rewrite <- ! orctx_extend_typ.
      eapply ty_has_N_logic_oc_and.
      +
        rewrite ! orctx_extend_typ; simpls.
        eapply IHHty1.
        rewrite ! orctx_extend_typ; eauto.
      +
        rewrite ! orctx_extend_typ; simpls.
        eapply IHHty2.
        rewrite ! orctx_extend_typ; eauto.
    *
        forwards: orctx_separate_cross_r Heqt0 H (typ_or T0 U0).
        destruct_all.
        forwards HT: H1 T.
        forwards HU: H1 U.
        destruct_all.
        rewrite H0.
        eapply ty_has_N_logic_oc_and.
        rewrite <- H7. eauto.
        rewrite <- H5. eauto.
Qed.

Lemma ty_has_N_logic_oclc_orl  G e oc2 oc T U:
  ty_has_N_logic  G e (orctx_typ oc2 (logctx_typ oc T)) ->
  ty_has_N_logic  G e (orctx_typ oc2 (logctx_typ oc (typ_or T U))).
Proof.
  introv Hty.
  reverts oc2.
  induction oc; intros; simpls.
  - eapply ty_has_N_logic_oc_orl; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  -
    forwards: ty_has_N_logic_oc_and_invl Hty. 
    forwards: ty_has_N_logic_oc_and_invr Hty.
    forwards: IHoc; eauto.
    eapply ty_has_N_logic_oc_and; eauto.
  -
    forwards: ty_has_N_logic_oc_and_invl Hty. 
    forwards: ty_has_N_logic_oc_and_invr Hty.
    forwards: IHoc; eauto.
    eapply ty_has_N_logic_oc_and; eauto.
Qed.
Lemma ty_has_N_logic_lc_orl  G e oc T U:
  ty_has_N_logic G e (logctx_typ oc T) ->
  ty_has_N_logic G e (logctx_typ oc (typ_or T U)).
Proof.
  introv Hty.
  applys ty_has_N_logic_oclc_orl orctx_nil; eauto.
Qed.


Lemma ty_has_N_logic_lift_or G e oc V S P:
    (forall oc2,
      ty_has_N_closure G e (logctx_typ oc2 V) ->
      ty_has_N_closure G e (logctx_typ oc2 S) \/ P
    ) ->
    ty_inv_logic_liftable G e V S ->
    ty_has_N_logic G e (logctx_typ oc V) ->
    ty_has_N_logic G e (logctx_typ oc S) \/ P.
Proof.
  introv Hbas Hlift Hty.
  remember (logctx_typ oc V). reverts oc.
  induction Hty; intros; subst.
  - forwards Hl: Hbas H. destructecs Hl. left. eapply ty_has_N_logic_or; eauto. eauto.
  - (* AND case *)
    rewrite orctx_logctx_typ in Heqt.
    lets Heqt0: Heqt.
    (* And happened under oc, bot happened under oc0. Try cases of intersecting oc and oc0. *)
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + (* oc = oc0 - impossible if V is not and *)
      inversions Hlift.
      * (* not and *)
        false.
        forwards: H orctx_nil; simpls; eauto.
      * (* toplike *)
        left.
        rewrite <- orctx_logctx_typ.
        eapply ty_has_N_logic_oc.
        eapply ty_has_N_logic_toplike; eauto.
    + (* oc is prefix of oc0, meaning that the V happens in one of the and branches *)
      (* destruct the next step in oc0, which tells us which branch to go to *)
      destruct x; simpls; inversions H0.
      * (* Must be a proper prefix *)
        false.
      * (* left branch *)
        forwards: IHHty1 (logctx_extend (orctx_logctx oc) x). eauto. 
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls; auto.
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls.
        rewrite logctx_extend_typ, <- orctx_logctx_typ in H.
        destructecs H.
        left. eapply ty_has_N_logic_oc_and; eauto.
        eauto.
      * (* right branch *)
        forwards: IHHty2 (logctx_extend (orctx_logctx oc) x). eauto. 
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls; auto.
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls.
        rewrite logctx_extend_typ, <- orctx_logctx_typ in H.
        destructecs H.
        left. eapply ty_has_N_logic_oc_and; eauto.
        eauto.
    + (* oc0 is a prefix of oc - impossible if V is not and *)
      eapply orctx_logctx_extend in H; destruct_all; subst.
      inversions Hlift.
      * (* not and *)
        false.
        forwards: H x1; simpls; rewrite <- orctx_logctx_typ in H0; eauto.
      * (* toplike *)
        left.
        rewrite <- orctx_logctx_typ.
        eapply ty_has_N_logic_oc.
        eapply ty_has_N_logic_toplike; eauto.
    + (* The paths diverge *)
      (* Put oc0 on the LHS *)
      symmetry in Heqt0.
      eapply logctx_separate_symm in H.
      (* Find variant of oc, where V is replaced by S *)
      forwards: logctx_separate_cross_r Heqt0 H S.
      destruct_all.
      (* Find variant of oc0, where (typ_and T U) is replaced by T *)
      forwards: H1 T.
      (* Find variant of oc0, where (typ_and T U) is replaced by U *)
      forwards: H1 U.
      destruct_all.
      rewrite <- orctx_logctx_typ in *.
      (* Recurse on the types where (typ_and T U) is replaced by T and by U individually *)
      forwards: IHHty1 x1; eauto.
      forwards: IHHty2 x0; eauto.
      (* Rewrite the types into a form that uses the variant of oc where V is replaced by S, therefore exposing the intersection of T and U *)
      rewrite  H7 in H9.
      rewrite  H5 in H10.
      rewrite H0.
      (* Because oc does not have intersection, then its version with V replaced by S also does not *)
      forwards Hocx0: logctx_follows_isor logctx_isor_orctx H2.
      hnf in Hocx0. destructecs Hocx0; subst.
      rewrite <- orctx_logctx_typ in *; simpls.
      destructecs_all; try auto 2.
      (* Now it is in the form to apply the rule *)
      left. eapply ty_has_N_logic_oc_and; eauto.
Qed.

Lemma ty_has_N_logic_lift G e oc V S:
    (forall oc2,
      ty_has_N_closure G e (logctx_typ oc2 V) ->
      ty_has_N_closure G e (logctx_typ oc2 S)
    ) ->
    ty_inv_logic_liftable G e V S ->
    ty_has_N_logic G e (logctx_typ oc V) ->
    ty_has_N_logic G e (logctx_typ oc S).
Proof.
  introv Hbas Hlift Hty.
  forwards: ty_has_N_logic_lift_or False Hlift Hty; fweauto.
  destructecs H; easy.
Qed.


Lemma ty_has_N_logic_lc_typ_normal_replace G e oc A SL SU TL TU:
  ty_has_N_logic G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_normal A) SL SU))) ->
  ty_has_N_logic G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_normal A) TL TU))).
Proof.
  introv Hty.
  eapply ty_has_N_logic_lift.
  intros.
  eapply ty_has_N_closure_lc_typ_normal_replace; eauto.
  eapply ty_inv_logic_liftable_atomic; eauto.
  eauto.
Qed.

Lemma ty_has_N_logic_lc_typ_replace G e oc A SL SU:
  ty_has_N_logic G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  (forall TL TU, ty_has_N_logic G e (logctx_typ oc (typ_rcd (dec_typ A TL TU)))) \/ ty_has_top_atomic G e (typ_rcd (dec_typ A SL SU)).
Proof.
  introv Hty.
  remember (logctx_typ oc (typ_rcd (dec_typ A SL SU))). reverts oc SL SU.
  induction Hty; intros; subst.
  - forwards Hl: ty_has_N_closure_lc_typ_replace H. destructecs Hl. left. intros. eapply ty_has_N_logic_or; eauto. eauto.
  - (* AND case *)
    rewrite orctx_logctx_typ in Heqt.
    lets Heqt0: Heqt.
    (* And happened under oc, bot happened under oc0. Try cases of intersecting oc and oc0. *)
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + (* oc = oc0 - impossible if V is not and *)
      inversions H0.
    + (* oc is prefix of oc0, meaning that the V happens in one of the and branches *)
      (* destruct the next step in oc0, which tells us which branch to go to *)
      destruct x; simpls; inversions H0.
      * (* left branch *)
        forwards: IHHty1 (logctx_extend (orctx_logctx oc) x). 
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls; auto.
        destructecs H.
        left. intros.
        specialize (H TL TU).
        rewrite logctx_extend_typ, <- orctx_logctx_typ in H.
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls.
        eapply ty_has_N_logic_oc_and; eauto.
        eauto.
      * (* right branch *)
        forwards: IHHty2 (logctx_extend (orctx_logctx oc) x).
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls; auto.
        destructecs H.
        left. intros.
        specialize (H TL TU).
        rewrite logctx_extend_typ, <- orctx_logctx_typ; simpls.
        rewrite logctx_extend_typ, <- orctx_logctx_typ in H.
        eapply ty_has_N_logic_oc_and; eauto.
        eauto.
    + (* oc0 is a prefix of oc - impossible if V is not and *)
      destruct x; simpls; inversions H0.
    + (* The paths diverge *)
      (* Find variant of oc, where V is replaced by S *)
      forwards HcrT: logctx_separate_cross_r Heqt0 H T.
      forwards HcrU: logctx_separate_cross_r Heqt0 H U.
      destructecs_all.
      rewrite <- orctx_logctx_typ in *.
      forwards HHty1: IHHty1 HcrT0; eauto.
      forwards HHty2: IHHty2 HcrU0; eauto.
      destructecs_all.
      * left. intros.
      specialize (HHty1 TL TU).
      specialize (HHty2 TL TU).
      specialize (HcrT1 (typ_rcd (dec_typ A TL TU))).
      specialize (HcrU1 (typ_rcd (dec_typ A TL TU))).
      destructecs_all.
      forwards Heq: logctx_follow_eq_inv ocST0 ocST1. eapply logctx_follow_trans. eapply HcrT1. eapply logctx_follow_symm. eauto.
      eapply eq_trans. eapply HcrT2. symmetry. eapply HcrU2. destructecs Heq. subst. clear Heq0.

      rewrite <- HcrT3 in HHty1.
      rewrite <- HcrU3 in HHty2.
      rewrite <- HcrT2.

      forwards Hocx0: logctx_follows_isor.  eapply logctx_isor_orctx. eapply logctx_follow_symm. eauto. 
      hnf in Hocx0. destructecs Hocx0; subst.
      rewrite <- orctx_logctx_typ in *; simpls.
      eapply ty_has_N_logic_oc_and; eauto.
      * eauto.
      * eauto.
      * eauto.
Qed.

Lemma ty_has_N_logic_lc_top_replace G e oc T U:
  ty_has_N_logic G e (logctx_typ oc T) ->
  typ_logic_toplike G e U ->
  ty_has_N_logic G e (logctx_typ oc U).
Proof.
  introv Hh Ht.
  eapply ty_has_N_logic_lift.
  intros.
  eapply ty_has_N_closure_lc_toplike_replace; eauto.
  eapply ty_inv_logic_liftable_toplike; eauto.
  eauto.
Qed.

Lemma ty_has_N_logic_lc_nocap_N_replace G e oc T:
  ty_has_N_logic G e (logctx_typ oc typ_nocap) ->
  ty_has_N G e T ->
  ty_has_N_logic G e (logctx_typ oc T).
Proof.
  introv Hh Ht.
  eapply ty_has_N_logic_lift.
  intros.
  eapply ty_has_N_closure_replace with (U:=typ_nocap); eauto.
  eapply ty_inv_logic_liftable_not_and; intros; destruct oc0; simpls; discriminate; eauto.
  eauto.
Qed.
Lemma ty_has_N_logic_lc_nocap_closure_replace G e oc T:
  ty_has_N_logic G e (logctx_typ oc typ_nocap) ->
  ty_has_N_closure G e T ->
  ty_has_N_logic G e (logctx_typ oc T).
Proof.
  introv Hh Ht.
  eapply ty_has_N_logic_lift.
  intros.
  eapply ty_has_N_closure_closure_replace with (U:=typ_nocap); eauto.
  eapply ty_inv_logic_liftable_not_and; intros; destruct oc0; simpls; discriminate; eauto.
  eauto.
Qed.



Lemma ty_has_N_logic_lc_M_replace G e oc T:
  inert G ->
  renv_corr G e ->
  ty_has_N_logic G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_has_N_logic G e (logctx_typ oc T).
Proof.
  introv Hi Hcor Hh.
  eapply ty_has_N_logic_lift.
  intros.
  eapply ty_has_N_closure_lc_M_replace; eauto.
  eapply ty_inv_logic_liftable_not_and; intros; destruct oc0; simpls; discriminate; eauto.
  eauto.
Qed.


Lemma ty_has_N_logic_lc_sel_replace G e oc U x A y:
  ty_has_N_logic G e (logctx_typ oc (typ_sel x A y)) ->
  ty_has_N_logic G e (logctx_typ oc U).
Proof.
  introv Ht.
  eapply ty_has_N_logic_lift.
  intros.
  eapply ty_has_N_closure_lc_sel_replace; eauto.
  eapply ty_inv_logic_liftable_not_and; intros; destruct oc0; simpls; discriminate; eauto.
  eauto.
Qed.
Lemma ty_has_N_logic_lc_bnd_replace G e oc U T:
  ty_has_N_logic G e (logctx_typ oc (typ_bnd T)) ->
  ty_has_N_logic G e (logctx_typ oc U).
Proof.
  introv Ht.
  eapply ty_has_N_logic_lift.
  intros.
  eapply ty_has_N_closure_lc_bnd_replace; eauto.
  eapply ty_inv_logic_liftable_not_and; intros; destruct oc0; simpls; discriminate; eauto.
  eauto.
Qed.
Lemma ty_has_N_logic_lc_bot_replace G e oc U:
  ty_has_N_logic G e (logctx_typ oc typ_bot) ->
  ty_has_N_logic G e (logctx_typ oc U).
Proof.
  introv Ht.
  eapply ty_has_N_logic_lift.
  intros.
  eapply ty_has_N_closure_lc_bot_replace; eauto.
  eapply ty_inv_logic_liftable_bot; eauto.
  eauto.
Qed.
Lemma ty_has_N_logic_lc_subtyp G e oc U V:
  inert G ->
  ty_has_N_logic G e (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_has_N_logic G e (logctx_typ oc V).
Proof.
  introv Hi Hh Ht.
  eapply ty_has_N_logic_lift.
  intros.
  eapply ty_has_N_closure_lc_subtyp; eauto.
  eapply ty_inv_logic_liftable_atomic; eauto.
  eapply subtyp_atomic_atomic_l; eauto.
  eauto.
Qed.


Lemma ty_has_N_logic_lc_rev_subtyp G e oc U V:
  inert G ->
  ty_has_N_logic G e (logctx_typ oc U) ->
  subtyp_atomic G e V U ->
  ty_has_N_logic G e (logctx_typ oc V) \/ ty_has_top_atomic G e U.
Proof.
  introv Hi Hh Ht.
  eapply ty_has_N_logic_lift_or.
  intros.
  eapply ty_has_N_closure_lc_rev_subtyp; eauto.
  eapply ty_inv_logic_liftable_atomic; eauto.
  eapply subtyp_atomic_atomic_r; eauto.
  eauto.
Qed.


(* NOTE: this also follows from ty_has_N_logic_lc_dist *)
Lemma ty_has_N_logic_oc_dist G e oc T1 T2 T3:
  ty_has_N_logic G e (orctx_typ oc (typ_and T1 (typ_or T2 T3))) ->
  ty_has_N_logic G e (orctx_typ oc (typ_or (typ_and T1 T2) (typ_and T1 T3))).
Proof.
  introv Hty.
  remember (orctx_typ oc
           (typ_and T1 (typ_or T2 T3))).
  reverts oc T1 T2 T3.
  induction Hty; intros; subst.
  - eapply ty_has_N_logic_or.
    rewrite orctx_logctx_typ in *.
    eapply ty_has_N_closure_lc_and_replace; eauto.
  -
    lets: Heqt.
    forwards: orctx_intersect_cases_nonnil Heqt.
    destruct_all; subst.
    + inversions H1.

      assert (ty_has_N_logic G e (orctx_typ oc0 (typ_or (typ_and T1 T2) T3))).
      *
      rewrite orctx_extend_l.
      eapply ty_has_N_logic_oc_and; rewrite orctx_extend_typ; simpls.
      eapply ty_has_N_logic_oc_orl; auto.
      auto.
      *
      rewrite orctx_extend_r.
      eapply ty_has_N_logic_oc_and; rewrite orctx_extend_typ; simpls.
      eapply ty_has_N_logic_oc_orr; auto.
      auto.

    + destruct x; simpls; try contradiction.
      inversions H1.
      inversions H1.
    + destruct x; simpls; try contradiction.
      inversions H1.
      inversions H1.
    +
      symmetry in H.
      eapply orctx_separate_symm in H0.
      forwards: orctx_separate_cross_r H H0 (typ_or (typ_and T1 T2) (typ_and T1 T3)).
      destruct_all.
      forwards: H2 T; destruct_all.
      forwards: H2 U; destruct_all.

      forwards: IHHty1; eauto.
      forwards: IHHty2; eauto.

      rewrite H8 in H11.
      rewrite H5 in H10.
      rewrite H1.

      eapply ty_has_N_logic_oc_and; eauto.
Qed.


Lemma ty_has_N_logic_lc_dist G e oc T1 T2 T3:
  ty_has_N_logic G e (logctx_typ oc (typ_and T1 (typ_or T2 T3))) ->
  ty_has_N_logic G e (logctx_typ oc (typ_or (typ_and T1 T2) (typ_and T1 T3))).
Proof.
  introv Hty.
  remember (logctx_typ oc
           (typ_and T1 (typ_or T2 T3))).
  reverts oc T1 T2 T3.
  induction Hty; intros; subst.
  - eapply ty_has_N_logic_or.
    eapply ty_has_N_closure_lc_and_replace; eauto.
  -
    lets: Heqt.
    rewrite orctx_logctx_typ in *.
    forwards: logctx_intersect_cases_nonnil Heqt.
    destruct_all; subst.
    + inversions H1.

      assert (ty_has_N_logic G e (logctx_typ (orctx_logctx oc)  (typ_or (typ_and T1 T2) T3))).
      rewrite <- orctx_logctx_typ in *.
      rewrite orctx_extend_l.

      eapply ty_has_N_logic_oc_and; rewrite orctx_extend_typ; simpls.
      eapply ty_has_N_logic_oc_orl; auto.
      auto.

      rewrite <- orctx_logctx_typ in *.
      rewrite orctx_extend_r.
      eapply ty_has_N_logic_oc_and; rewrite orctx_extend_typ; simpls.
      eapply ty_has_N_logic_oc_orr; auto.
      auto.

    + destruct x; simpls; try contradiction.
      inversions H1.
      inversions H1.

      inversions H1.
      rewrite logctx_extend_typ in *. simpls.
      rewrite <- orctx_logctx_typ in *. simpls.
      eapply ty_has_N_logic_oc_and; eauto.
      rewrite orctx_logctx_typ in *.
      rewrite <- logctx_extend_typ in *.
      eapply IHHty1; auto.

      inversions H1.
      rewrite logctx_extend_typ in *. simpls.
      rewrite <- orctx_logctx_typ in *. simpls.
      eapply ty_has_N_logic_oc_and; eauto.
      rewrite orctx_logctx_typ in *.
      rewrite <- logctx_extend_typ in *.
      eapply IHHty2; auto.


    + destruct x; simpls; try contradiction.
      inversions H1.
      inversions H1.

      inversions H1.
      eapply orctx_logctx_extend_and_l_false in H0. contradiction.
      inversions H1.
      eapply orctx_logctx_extend_and_r_false in H0. contradiction.
    +
      symmetry in H.
      eapply logctx_separate_symm in H0.
      forwards: logctx_separate_cross_r H H0 (typ_or (typ_and T1 T2) (typ_and T1 T3)).
      destruct_all.
      forwards: H2 T; destruct_all.
      forwards: H2 U; destruct_all.

      forwards: IHHty1; eauto.
      forwards: IHHty2; eauto.

      rewrite H8 in H11.
      rewrite H5 in H10.
      rewrite H1.
      forwards: logctx_follows_isor (orctx_logctx oc); eauto.

      hnf in H12; destruct_all; subst.
      rewrite <- orctx_logctx_typ in *.
      eapply ty_has_N_logic_oc_and; eauto.
Qed.

Lemma ty_has_N_logic_oc_and_dist G e oc T1 T2 T3:
  ty_has_N_logic G e (orctx_typ oc T1) ->
  ty_has_N_logic G e (orctx_typ oc (typ_or T2 T3)) ->
  ty_has_N_logic G e (orctx_typ oc (typ_or (typ_and T1 T2) (typ_and T1 T3))).
Proof.
  introv H1 H23.
  eapply ty_has_N_logic_oc_dist.
  eapply ty_has_N_logic_oc_and; eauto.
Qed.
Lemma ty_has_N_logic_lc_and_dist G e oc T1 T2 T3:
  ty_has_N_logic G e (logctx_typ oc T1) ->
  ty_has_N_logic G e (logctx_typ oc (typ_or T2 T3)) ->
  ty_has_N_logic G e (logctx_typ oc (typ_or (typ_and T1 T2) (typ_and T1 T3))).
Proof.
  introv H1 H23.
  eapply ty_has_N_logic_lc_dist.
  eapply ty_has_N_logic_lc_and; eauto.
Qed.

Lemma ty_has_N_logic_oc_replace_nocap G e oc T:
  ty_has_N_logic G e (orctx_typ oc T) ->
  ty_has_N_logic G e (orctx_typ oc typ_nocap).
Proof.
  introv Hh.
  remember (orctx_typ oc T).
  reverts oc T.
  induction Hh; intros; subst.
  - rewrite orctx_logctx_typ in *. eapply ty_has_N_logic_or.  eapply ty_has_N_closure_lc_nocap_replace; eauto.
  -
  pose proof Heqt as Heqt'.
  eapply orctx_intersect_cases_nonnil in Heqt.
  destruct_all; subst.
  + forwards: IHHh2 oc0 U; auto.
  + destruct x; simpls; try discriminate; contradiction.
  +
    forwards: IHHh2 oc0 (orctx_typ x U) . rewrite orctx_extend_typ; auto.
    forwards: IHHh1 oc0 (orctx_typ x T) . rewrite orctx_extend_typ; auto.
    destruct_all; eauto.
  +
    forwards: orctx_separate_cross_l Heqt' typ_nocap. eauto.
    destruct_all.

    forwards Hty01: IHHh1; eauto.
    forwards Hty02: IHHh2; eauto.
    destruct_all; eauto.

    forwards: H1 T.
    forwards: H1 U.
    destructecs_all.
    rewrite <- H0.
    eapply ty_has_N_logic_oc_and; eauto.
    rewrite H8. eauto. 
    rewrite H6. eauto.
Qed.

Lemma ty_has_N_logic_oclc_replace_nocap G e oc oc2 T:
  ty_has_N_logic G e (orctx_typ oc (logctx_typ oc2 T)) ->
  ty_has_N_logic G e (orctx_typ oc (logctx_typ oc2 typ_nocap)).
Proof.
  introv Hty.
  reverts oc.
  induction oc2; intros; simpls.
  - eapply ty_has_N_logic_oc_replace_nocap; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  -
    forwards: ty_has_N_logic_oc_and_invl Hty. 
    forwards: ty_has_N_logic_oc_and_invr Hty.
    forwards: IHoc2; eauto.
    eapply ty_has_N_logic_oc_and; eauto.
  -
    forwards: ty_has_N_logic_oc_and_invl Hty. 
    forwards: ty_has_N_logic_oc_and_invr Hty.
    forwards: IHoc2; eauto.
    eapply ty_has_N_logic_oc_and; eauto.
Qed.

Lemma ty_has_N_logic_lc_replace_nocap G e oc T:
  ty_has_N_logic G e (logctx_typ oc T) ->
  ty_has_N_logic G e (logctx_typ oc typ_nocap).
Proof.
  introv Hh.
  eapply ty_has_N_logic_oclc_replace_nocap with (oc:=orctx_nil). simpl. eauto.
Qed.



Lemma ty_has_N_logic_oc_cases G e  oc T:
  ty_has_N_logic G e (orctx_typ oc T) ->
  ty_has_N_logic G e T \/ (forall U, ty_has_N_logic G e (orctx_typ oc U)).
Proof.
  introv Hh.
  remember (orctx_typ oc T).
  reverts T oc.
  induction Hh; intros; subst.
  - forwards: ty_has_N_or_closure_oc_cases H. destruct_all. left. eapply ty_has_N_logic_or. eauto. right. intros. eapply ty_has_N_logic_or. eauto. 
  -
  pose proof Heqt as Heqt'.
  eapply orctx_intersect_cases_nonnil in Heqt.
  destruct_all; subst.
  + forwards: IHHh2 U oc0; auto.
    forwards: IHHh1 T oc0; auto.
    destruct_all; eauto.
    left. eapply ty_has_N_logic_and; eauto.
  + destruct x; simpls; try discriminate; contradiction.
  +
    forwards: IHHh2 (orctx_typ x U) oc0. rewrite orctx_extend_typ; auto.
    forwards: IHHh1 (orctx_typ x T) oc0. rewrite orctx_extend_typ; auto.
    destruct_all; eauto.
    left.
    eapply ty_has_N_logic_oc_and; eauto.
  +
    forwards: orctx_separate_cross_r Heqt' T. eauto.
    forwards: orctx_separate_cross_r Heqt' U. eauto.
    destruct_all.

    forwards Hty01: IHHh1; eauto.
    forwards Hty02: IHHh2; eauto.
    destruct_all; eauto.
    right. intro.

    forwards: H4 U0.
    forwards: H2 U0.
    destruct_all.

    forwards: orctx_follow_eq_inv x1 x2. eauto using orctx_follow_trans, orctx_follow_symm. rewrite H9. eauto. destruct_all; subst. cleanup.
    rewrite <- H9.
    eapply ty_has_N_logic_oc_and; eauto.
    rewrite H12. eauto. 
    rewrite H10. eauto.
Qed.

Lemma ty_has_N_logic_oclc_cases G e oc oc2 T:
  ty_has_N_logic G e (orctx_typ oc (logctx_typ oc2 T)) ->
  ty_has_N_logic G e T \/ (forall U, ty_has_N_logic G e (orctx_typ oc (logctx_typ oc2 U))).
Proof.
  introv Hty.
  reverts oc.
  induction oc2; intros; subst; simpls.
  - eapply ty_has_N_logic_oc_cases; eauto.
  - rewrite orctx_extend_l in *. forwards: IHoc2; eauto.
    destruct_all. auto.
    right; intros. specialize (H U).
    rewrite orctx_extend_typ in H; simpls. auto.
  - rewrite orctx_extend_r in *. forwards: IHoc2; eauto.
    destruct_all. auto.
    right; intros. specialize (H U).
    rewrite orctx_extend_typ in H; simpls. auto.
  -
    forwards: ty_has_N_logic_oc_and_invr Hty.
    forwards: ty_has_N_logic_oc_and_invl Hty.
    forwards: IHoc2; eauto.
    destruct_all.
    auto.
    right. intros. eapply  ty_has_N_logic_oc_and; eauto.
  -
    forwards: ty_has_N_logic_oc_and_invr Hty.
    forwards: ty_has_N_logic_oc_and_invl Hty.
    forwards: IHoc2; eauto.
    destruct_all.
    auto.
    right. intros. eapply  ty_has_N_logic_oc_and; eauto.
Qed.

Lemma ty_has_N_logic_lc_cases G e oc T:
  ty_has_N_logic G e (logctx_typ oc T) ->
  ty_has_N_logic G e T \/ (forall U, ty_has_N_logic G e (logctx_typ oc U)).
Proof.
  introv Hty.
  eapply ty_has_N_logic_oclc_cases with (oc:=orctx_nil); eauto.
Qed.


(* NOTE: this will not hold if nocap is subtype of {M:bot..top} *)
Lemma ty_has_N_or_closure_oc_or_inv G e oc T U:
  ty_has_N_closure G e (orctx_typ oc (typ_or T U)) ->
  ty_has_N_closure G e (orctx_typ oc T) \/ ty_has_N_closure G e (orctx_typ oc U).
Proof.
  introv Hh.
  remember (orctx_typ oc (typ_or T U)).
  reverts T U oc.
  induction Hh; intros; subst.
  - forwards: ty_has_N_oc_inv H; subst; simpls.
    false. eapply ty_has_N_lc_or_false with (oc:=logctx_nil) in H.
    eauto.
  - destruct oc; simpls; subst.
    +
    inversions Heqt. left. eauto.
    +
    inversions Heqt.
    forwards: IHHh T0 oc; simpls. auto. destruct_all. left. eapply ty_or_closure_l; eauto. right. eapply ty_or_closure_l; eauto.
    +
    inversions Heqt.

    left. eapply ty_or_closure_l; eauto.
  - destruct oc; simpls; subst.
    +
    inversions Heqt. right. eauto.
    +
    inversions Heqt.
    left. eapply ty_or_closure_r; eauto.

    +
    inversions Heqt.
    forwards: IHHh T0 oc; simpls. auto. destruct_all. left.
    eapply ty_or_closure_r; eauto.

    right. eapply ty_or_closure_r; eauto.
  - destruct oc; simpls; subst; inversions Heqt.
Qed.

Lemma ty_has_N_logic_lc_nocap_logic_replace G e oc T:
  ty_has_N_logic G e (logctx_typ oc typ_nocap) ->
  ty_has_N_logic G e T ->
  ty_has_N_logic G e (logctx_typ oc T).
Proof.
  introv Hh Ht.
  induction Ht.
  - eapply  ty_has_N_logic_lc_nocap_closure_replace; eauto.
  - rewrite orctx_logctx_typ, <- logctx_extend_typ in *.  eapply ty_has_N_logic_lc_and; eauto.
Qed.

Lemma ty_has_N_logic_atomic_closure G e T:
  ty_has_N_logic G e T ->
  typ_atomic T ->
  ty_has_N_closure G e T.
Proof.
  introv Hty Ha.
  induction Hty; eauto; inversions Ha.
  destruct oc; simpls; inversions H0.
  destruct oc; simpls; inversions H0.
Qed.

Lemma ty_has_N_logic_tight G e T:
  ty_has_N_logic G e T ->
  G⁏e ⊢# typ_nocap <: T.
Proof.
  introv Hh.
  induction Hh; eauto 2 using ty_has_N_closure_tight.
  eapply subtyp_oc_and_tight; eauto.
Qed.



End WithTypingMode.