Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening.

Require Import Orctx Logctx OrctxLogctx Logpath.
Require Import LayeredCommon LayeredSubtypingAtomic LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic.
Require Import WrapAction.

Section WithTypingMode.
Context {typing_mode: Mode}.



Definition ty_has_N_and_closure G e := ty_and_closure (ty_has_N_closure G e).

Inductive ty_has_N_logic_linear (G: ctx) (e: renv): typ -> Prop :=
| ty_has_N_logic_linear_and T:
    ty_has_N_and_closure G e T ->
    ty_has_N_logic_linear G e T
| ty_has_N_logic_linear_l oc S T U:
  logctx_andor oc ->
  ty_has_N_logic_linear G e (logctx_typ oc (typ_and (typ_or T S) (typ_or U S))) ->
  ty_has_N_logic_linear G e (logctx_typ oc (typ_or (typ_and T U) S))
| ty_has_N_logic_linear_r oc S T U:
  logctx_andor oc ->
  ty_has_N_logic_linear G e (logctx_typ oc (typ_and (typ_or S T) (typ_or S U))) ->
  ty_has_N_logic_linear G e (logctx_typ oc (typ_or S (typ_and T U)))
.

Lemma ty_has_N_and_closure_logic G e T:
  ty_has_N_and_closure G e T ->
  ty_has_N_logic G e T.
Proof.
  introv Hty.
  induction Hty.
  - eapply ty_has_N_logic_or; eauto.
  - eapply ty_has_N_logic_oc_and with (oc := orctx_nil); simpl; auto.
Qed.

Lemma ty_has_N_logic_linear_logic G e T:
  ty_has_N_logic_linear G e T ->
  ty_has_N_logic G e T.
Proof.
  introv Hty.
  induction Hty.
  - eapply ty_has_N_and_closure_logic; eauto.
  -
    forwards: ty_has_N_logic_lc_and_invr IHHty.
    forwards: ty_has_N_logic_lc_and_invl IHHty.
    rewrite logctx_extend_orl in *.
    eapply ty_has_N_logic_lc_and; eauto.
  -
    forwards: ty_has_N_logic_lc_and_invr IHHty.
    forwards: ty_has_N_logic_lc_and_invl IHHty.
    rewrite logctx_extend_orr in *.
    eapply ty_has_N_logic_lc_and; eauto.
Qed.

Lemma ty_has_N_logic_linear_and_closure G e T U:
  ty_has_N_and_closure G e T ->
  ty_has_N_logic_linear G e U ->
  ty_has_N_logic_linear G e (typ_and T U).
Proof.
  introv HtyT HtyU.
  induction HtyU.
  - eapply ty_has_N_logic_linear_and.
    eapply ty_and_closure_and; auto.
  - rewrite logctx_fold_andr in *.
    eapply ty_has_N_logic_linear_l; auto.
  - rewrite logctx_fold_andr in *.
    eapply ty_has_N_logic_linear_r; auto.
Qed.

Lemma ty_has_N_logic_linear_and_nil G e T U:
  ty_has_N_logic_linear G e T ->
  ty_has_N_logic_linear G e U ->
  ty_has_N_logic_linear G e (typ_and T U).
Proof.
  introv HtyT HtyU.
  induction HtyT.
  - eapply ty_has_N_logic_linear_and_closure; eauto.
  - rewrite logctx_fold_andl in *.
   eapply ty_has_N_logic_linear_l; auto.
  - rewrite logctx_fold_andl in *.
   eapply ty_has_N_logic_linear_r; auto.
Qed.


Lemma ty_has_N_logic_linear_and_orctx G e oc T U:
  ty_has_N_logic_linear G e (orctx_typ oc T) ->
  ty_has_N_logic_linear G e (orctx_typ oc U) ->
  ty_has_N_logic_linear G e (orctx_typ oc (typ_and T U)).
Proof.
  introv HtyT HtyU.
  reverts T U.
  induction oc using orctx_reverse_ind; intros; simpls.
  - eapply ty_has_N_logic_linear_and_nil; eauto.
  - rewrite orctx_extend_typ in *; simpls.
    forwards: IHoc HtyT HtyU.

    rewrite orctx_logctx_typ in *.
    eapply ty_has_N_logic_linear_l; eauto.
  - rewrite orctx_extend_typ in *; simpls.
    forwards: IHoc HtyT HtyU.

    rewrite orctx_logctx_typ in *.
    eapply ty_has_N_logic_linear_r; eauto.
Qed.


Lemma ty_has_N_logic_logic_linear G e T:
  ty_has_N_logic G e T ->
  ty_has_N_logic_linear G e T.
Proof.
  introv Hty.
  induction Hty.
  - eapply ty_has_N_logic_linear_and.
    eapply ty_and_closure_atomic. auto.
  -
    reverts T U.
    induction oc; intros; simpls.
    + eapply ty_has_N_logic_linear_and_nil; eauto.
    + rewrite orctx_fold_l in *.
      eapply ty_has_N_logic_linear_and_orctx; eauto.
    + rewrite orctx_fold_r in *.
      eapply ty_has_N_logic_linear_and_orctx; eauto.
Qed.

Inductive wrap_action_has_N_logic: wrap_action -> Prop :=
| wrap_action_has_N_logic_l S T U oc:
    logctx_andor oc ->
    wrap_action_has_N_logic (wrap_action_c (logctx_path oc) (typ_and (typ_or T S) (typ_or U S))  (typ_or (typ_and T U) S))
| wrap_action_has_N_logic_r S T U oc:
    logctx_andor oc ->
    wrap_action_has_N_logic (wrap_action_c (logctx_path oc) (typ_and (typ_or S T) (typ_or S U))  (typ_or S (typ_and T U)))
.
Hint Constructors wrap_action_has_N_logic: core. 

Lemma ty_has_N_logic_plan G e T:
  ty_has_N_logic_linear G e T ->
  exists wp S,
    ty_has_N_and_closure G e S /\
    wrap_plan_utyp S wp T /\
    List.Forall (wrap_action_has_N_logic) wp.
Proof.
  introv Hty.
  induction Hty.
  - exists (nil: wrap_plan) T; splits; eauto. (* constructor.*)
  - destruct_all.
    exists (cons (wrap_action_c (logctx_path oc) (typ_and (typ_or T S) (typ_or U S))  (typ_or (typ_and T U) S) ) x) x0.
    splits; eauto. econstructor; eauto. constructor.
  - destruct_all.
    exists (cons (wrap_action_c (logctx_path oc) (typ_and (typ_or S T) (typ_or S U))  (typ_or S (typ_and T U)) ) x) x0.
    splits; eauto. econstructor; eauto. constructor.
Qed.

Lemma ty_has_N_plan_logic G e T wp S:
  ty_has_N_logic_linear G e S ->
  wrap_plan_utyp S wp T ->
  List.Forall (wrap_action_has_N_logic) wp ->
  ty_has_N_logic_linear G e T.
Proof.
  introv Hhas Hu Ha.
  induction Hu.
  - eauto.
  - inversions Ha.
    forwards: IHHu; eauto.
    inversions H.
    inversions H2.
    + eapply ty_has_N_logic_linear_l; eauto.
      eapply logctx_path_eq_follow in H.
      eapply logctx_follow_andor; eauto.
    + eapply ty_has_N_logic_linear_r; eauto.
      eapply logctx_path_eq_follow in H.
      eapply logctx_follow_andor; eauto.
Qed.

(*
This lemma would not be true because we can do and (or A B) (or A C)  -> (or A (B and C)) then do or
Lemma wrap_action_has_N_logic_leaf wa:
  wrap_action_has_N_logic wa ->
  wrap_action_leaf_for (logpath_andor) wa.
*)

End WithTypingMode.