Set Implicit Arguments.

Require Import Coq.Program.Equality.
Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode.
Require Import Equivalence GeneralTyping Indep.
Require Import PreciseTyping RecordTypes InertContext EnvCorrespondence.
Require Import TightTyping InertTightSubtyping TightTypingHelpers.
Require Import Orctx Logctx OrctxLogctx Logpath.

Require Import LayeredCommon LayeredSubtypingAtomic LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic LayeredCapabilityLogicOrdered.
Require Import WrapAction WrapActionMain.

Section WithTypingMode.
Context {typing_mode: Mode}.

(*
  Derivation of typ_nocap <: T on the main layer.
  Unlike ty_inv_main, there is no rule to replace (open x T) by (typ_bnd T),
  for that situation, there is ty_has_N_main_lc_nocap_main_replace
*)

Inductive ty_has_N_main: ctx -> renv -> typ -> Prop :=
| ty_has_N_main_logic G e T:
    ty_has_N_logic G e T ->
    ty_has_N_main G e T
| ty_has_N_main_name G e oc S1 S2 U y A z:
    ty_has_N_main G e (logctx_typ oc (open z S1)) ->
    G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
    ty_has_N_main G e (logctx_typ oc (typ_sel (avar_f y) A (avar_f z)))
| ty_has_N_main_bnd G e oc T:
    ty_has_N_main G e (logctx_typ oc typ_nocap) ->
    ty_has_N_main G e (logctx_typ oc (typ_bnd T))
.

(* WRAP ACTIONS *)
(* Expressing path through the derivation above, using a list of actions that operate on the type *)

Lemma ty_has_N_main_plan G e T:
  inert G ->
  ty_has_N_main G e T ->
  exists wp S,
    ty_has_N_logic G e S /\
    wrap_plan_utyp S wp T /\
    List.Forall (wrap_action_has_N_main G e) wp.
Proof.
  introv Hi Hty.
  induction Hty.
  - exists (nil: wrap_plan) T; splits; eauto. (* constructor.*)
  - specialize (IHHty Hi). destruct_all.
    exists (cons (wrap_action_c (logctx_path oc) (open z S1) (typ_sel (avar_f y) A (avar_f z)) ) x) x0.
    splits; eauto. econstructor; eauto. constructor.
  - specialize (IHHty Hi). destruct_all.
    exists (cons (wrap_action_c (logctx_path oc) (typ_nocap) (typ_bnd T)) x) x0.
    splits; eauto. econstructor; eauto. constructor.
Qed.

Lemma ty_has_N_plan_logic_main G e T wp S:
  ty_has_N_logic G e S ->
  wrap_plan_utyp S wp T ->
  List.Forall (wrap_action_has_N_main G e) wp ->
  ty_has_N_main G e T.
Proof.
  introv Hhas Hu Ha.
  induction Hu.
  - eapply ty_has_N_main_logic; eauto.
  - inversions Ha.
    forwards: IHHu; eauto.
    inversions H.
    inversions H2.
    + eapply ty_has_N_main_name; eauto.
    + eapply ty_has_N_main_bnd; eauto.
Qed.
Lemma ty_has_N_plan_main G e T wp S:
  ty_has_N_main G e S ->
  wrap_plan_utyp S wp T ->
  List.Forall (wrap_action_has_N_main G e) wp ->
  ty_has_N_main G e T.
Proof.
  introv Hhas Hu Ha.
  induction Hu.
  - eauto.
  - inversions Ha.
    forwards: IHHu; eauto.
    inversions H.
    inversions H2.
    + eapply ty_has_N_main_name; eauto.
    + eapply ty_has_N_main_bnd; eauto.
Qed.
Lemma ty_has_N_main_plan_segment G e oc T:
  inert G ->
  ty_has_N_main G e (logctx_typ oc T) ->
  exists wpsep wpex oc2 U,
    ty_has_N_logic G e (logctx_typ oc2 U) /\
    wrap_plan_utyp (logctx_typ oc2 U) wpsep (logctx_typ oc U) /\
    wrap_plan_utyp U wpex T /\
    wrap_plan_separate (logctx_path oc) wpsep /\
    logctx_path oc2 = logctx_path oc /\
    List.Forall (wrap_action_has_N_main G e) wpsep /\
    List.Forall (wrap_action_has_N_main G e) (wrap_plan_under (logctx_path oc) wpex).
Proof.
  introv Hi Hty.
  forwards: ty_has_N_main_plan Hi Hty.
  destructc H as wp0 S Hhas Hplan Hf.
  forwards: wrap_plan_segment (logctx_path oc); eauto.
  eapply logctx_path_typ.

  eapply List.Forall_impl. 2: eauto. intros. eapply wrap_action_has_N_main_leaf; eauto.
  destructc H as wpex wpsep wpsepex ns Hrest.
  destructc Hrest as Hwpsepex Hre Hnd Hl Htyp Hsep.
  rewrite Hwpsepex in Htyp.
  eapply wrap_plan_utyp_app_inv in Htyp.
  destructc Htyp as US HSUS HUST.
  forwards: wrap_plan_utyp_logpath_ctx_rev_preserve HUST. eapply logctx_path_ctx.
  forwards: logpath_ctx_typ_exists H; eauto.
  destructc H0 as U HU.
  forwards: logpath_ctx_typ HU H; subst.

  forwards: list_forall_reorder Hre Hf.
  forwards: list_forall_app_invr H0.
  forwards: list_forall_app_invl H0.

  forwards: wrap_plan_utyp_logpath_typ_rev_preserve HSUS Hsep. eapply logctx_path_typ.
  forwards: logpath_typ_ctx_exists H3; eauto.
  destructc H4 as ocSU HocSU.
  forwards: logpath_ctx_typ H3 HocSU; subst.
  forwards: logpath_ctx_path HocSU.

  exists wpsep wpex ocSU U; splits; eauto.
  eapply wrap_plan_utyp_under_case; eauto.
Qed.

(* CLOSURE PROPERTIES *)


Lemma ty_has_N_main_tyq G e U V:
  inert G ->
  renv_corr G e ->
  ty_has_N_main G e U ->
  typing_equiv e U V ->
  ty_has_N_main G e V.
Proof.
  introv Hin Hcor Hty Htyq.
  reverts V.
  induction Hty; intros.
  - eapply ty_has_N_main_logic; eauto. eapply ty_has_N_logic_tyq; eauto.
  - eapply logctx_tyq_case in Htyq; destruct_all; subst.
    apply tyq_typ_sel_inv in H2.
    destruct H2 as [[v21 [v22 [tnl [Htnl [HU [? Htyq2]]]]]] | HT2]; subst.
    + inversions Htnl.
      destruct (renv_corr_pf_normal_typ_tyq Hin Hcor Htyq2 H).
      destruct H0 as [T21 [T22 [Hpf [Htyq11 Htyq12]]]].
      eapply ty_has_N_main_name.
      apply IHHty. auto. auto. apply logctx_tyq_typ. eauto. eapply Htyq11. eauto.
    + eapply ty_has_N_main_name. eapply IHHty; eauto. apply logctx_tyq_typ. eauto. 2: eauto. eauto.
  - eapply logctx_tyq_case in Htyq; destruct_all; subst.

    inversions_tyq H1.

    eapply ty_has_N_main_bnd.

    eapply IHHty; eauto. apply logctx_tyq_typ. eauto.
    eauto.
Qed.

(* LIFTING PROPERTIES FROM LOGIC LAYER *)
Lemma ty_has_N_main_lift_or G e oc S T P:
    inert G ->
    (forall oc2,
      ty_has_N_logic G e (logctx_typ oc2 T) ->
      ty_has_N_logic G e (logctx_typ oc2 S) \/ P
    ) ->
    ty_inv_main_liftable T S ->
    ty_has_N_main G e (logctx_typ oc T) ->
    ty_has_N_main G e (logctx_typ oc S) \/ P.
Proof.
  introv Hin Hlog Hlcnd Hty.
  remember (logctx_typ oc T). (* reverts T.*)
  reverts oc.
  induction Hty; intros; subst.
  - forwards Hl: Hlog H.
    destructecs Hl. left. eapply ty_has_N_main_logic. eauto. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs_all; subst.
    + (* Equal *)
      inversions Hlcnd.
      *
        false.
        destruct_all.
        eapply (H0 logctx_nil); simpls; eauto.
    +
      destruct ocST; simpl in Heqt; inversions Heqt.
      inversions Hlcnd.
      false.
      destructecs_all.
      eapply (H1 logctx_nil); simpls; eauto.
    +
      false.
      inversions Hlcnd.
      destruct_all.
      eapply (H0 ocTS); simpls; eauto.
    + (* separate *)
      forwards Hcross: logctx_separate_cross_r Heqt0 Heqt (open z S1).
      destructecs_all.

      forwards HHty: IHHty. eauto. apply Hlog. eauto.
      specialize (Hcross1 S).
      destructecs_all.
      left.
      rewrite <- Hcross3 in HHty.
      rewrite <- Hcross2.
      eapply ty_has_N_main_name; eauto.
      eauto.
  -
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + (* Equal *)
      inversions Hlcnd.
      *
        false.
        destruct_all.
        eapply (H0 logctx_nil); simpls; eauto.
    +
      destruct x; simpl in H0; inversions H0.
      inversions Hlcnd.
      false.
      destruct_all.
      eapply (H1 logctx_nil); simpls; eauto.
    +
      false.
      inversions Hlcnd.
      destruct_all.
      eapply (H0 x); simpls; eauto.
    + (* separate *)
      forwards Hcross: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destructecs_all.

      forwards HHty: IHHty. eauto. apply Hlog. eauto.
      specialize (Hcross1 S).
      destructecs_all.
      left.
      rewrite <- Hcross3 in HHty.
      rewrite <- Hcross2.
      eapply ty_has_N_main_bnd; eauto.
      eauto.
Qed.
Lemma ty_has_N_main_lift G e oc S T:
    inert G ->
    (forall oc2,
      ty_has_N_logic G e (logctx_typ oc2 T) ->
      ty_has_N_logic G e (logctx_typ oc2 S)
    ) ->
    ty_inv_main_liftable T S ->
    ty_has_N_main G e (logctx_typ oc T) ->
    ty_has_N_main G e (logctx_typ oc S).
Proof.
  introv Hin Hlog Hlcnd Hty.
  forwards: ty_has_N_main_lift_or False Hlcnd Hty; fweauto.
  destructecs H; easy.
Qed.

(* FORWARD DERIVATIONS *)
Lemma ty_has_N_main_lc_top_replace G e oc T:
  inert G ->
  ty_has_N_main G e (logctx_typ oc T) ->
  ty_has_N_main G e (logctx_typ oc typ_top).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc T). (* reverts T.*)
  reverts oc T.
  induction Hty; intros; subst.
  - eapply ty_has_N_main_logic. eapply ty_has_N_logic_lc_top_replace; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + eauto.
    + destruct x; simpls; inversions H1.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 typ_top.
      destruct_all.
      rewrite <- H4.
      eapply ty_has_N_main_name. 2: eauto.
      forwards: IHHty. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + eauto.
    + destruct x; simpls; inversions H0.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H typ_nocap.
      destruct_all.
      forwards: H1 typ_top.
      destruct_all.
      rewrite <- H3.
      eapply ty_has_N_main_bnd.
      forwards: IHHty. eauto. eapply H0. rewrite <- H4 in H6. eauto.
Qed.
Lemma ty_has_N_main_replace_nocap G e oc T:
  inert G ->
  ty_has_N_main G e (logctx_typ oc T) ->
  ty_has_N_main G e (logctx_typ oc typ_nocap).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc T). (* reverts T.*)
  reverts oc T.
  induction Hty; intros; subst.
  - eapply ty_has_N_main_logic. eapply ty_has_N_logic_lc_replace_nocap; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + eauto.
    + destruct x; simpls; inversions H1.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 typ_nocap.
      destruct_all.
      rewrite <- H4.
      eapply ty_has_N_main_name. 2: eauto.
      forwards: IHHty. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + eauto.
    + destruct x; simpls; inversions H0.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H typ_nocap.
      destruct_all.
      forwards: H1 typ_nocap.
      destruct_all.
      rewrite <- H3.
      eapply ty_has_N_main_bnd.
      forwards: IHHty. eauto. eapply H0. rewrite <- H4 in H6. eauto.
Qed.
Lemma ty_has_N_main_NM_replace G e oc T T1 T2:
  mode_has_top_nm_subtype typing_mode ->
  inert G ->
  (mode_has_top_nm_complement typing_mode -> forall x0 : var, G ⁏ e ⊢# T1 /[-> x0] <: typ_bot) ->
  ty_has_N_main G e (logctx_typ oc T) ->
  ty_has_N_main G e (logctx_typ oc (typ_NM T1 T2)).
Proof.
  introv Hm Hin Hbot Hty.
  remember (logctx_typ oc T). (* reverts T.*)
  reverts oc T.
  induction Hty; intros; subst.
  - eapply ty_has_N_main_logic. eapply ty_has_N_logic_lc_top_replace; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + eauto.
    + destruct x; simpls; inversions H1.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (typ_NM T1 T2).
      destruct_all.
      rewrite <- H4.
      eapply ty_has_N_main_name. 2: eauto.
      forwards: IHHty Hbot. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + eauto.
    + destruct x; simpls; inversions H0.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H typ_nocap.
      destruct_all.
      forwards: H1 (typ_NM T1 T2).
      destruct_all.
      rewrite <- H3.
      eapply ty_has_N_main_bnd.
      forwards: IHHty Hbot. eauto. eapply H0. rewrite <- H4 in H6. eauto.
Qed.

Lemma ty_has_N_main_lc_bot_replace G e oc T:
  inert G ->
  ty_has_N_main G e (logctx_typ oc typ_bot) ->
  ty_has_N_main G e (logctx_typ oc T).
Proof.
  intros Hin Hty.
  forwards: ty_has_N_main_lift Hin.
  intros. eapply ty_has_N_logic_lc_bot_replace. eauto. 
  eauto.
  eauto using ty_inv_main_liftable_bot.
  eauto.
  eauto.
Qed.

Lemma ty_has_N_main_lc_nocap_replace G e oc T:
  inert G ->
  ty_has_N G e T ->
  ty_has_N_main G e (logctx_typ oc typ_nocap) ->
  ty_has_N_main G e (logctx_typ oc T).
Proof.
  intros Hin Hhas Hty.
  forwards: ty_has_N_main_lift Hin.
  intros. eapply ty_has_N_logic_lc_nocap_logic_replace. eauto.
  eapply ty_has_N_logic_or. eapply ty_or_closure_atomic. eauto.
  eauto using ty_inv_main_liftable_nocap.
  eauto.
  eauto.
Qed.

Lemma ty_has_N_main_lc_nocap_logic_replace G e oc T:
  inert G ->
  ty_has_N_logic G e T ->
  ty_has_N_main G e (logctx_typ oc typ_nocap) ->
  ty_has_N_main G e (logctx_typ oc T).
Proof.
  intros Hin Hhas Hty.
  forwards: ty_has_N_main_lift Hin.
  intros. eapply ty_has_N_logic_lc_nocap_logic_replace. eauto. eauto. 
  eauto using ty_inv_main_liftable_nocap.
  eauto.
  eauto.
Qed.
Lemma ty_has_N_main_lc_nocap_main_replace G e oc T:
  inert G ->
  ty_has_N_main G e T ->
  ty_has_N_main G e (logctx_typ oc typ_nocap) ->
  ty_has_N_main G e (logctx_typ oc T).
Proof.
  intros Hin Hhas Hty.
  induction Hhas.
  - eapply ty_has_N_main_lc_nocap_logic_replace; eauto.
  - rewrite  <- logctx_extend_typ in *. eapply ty_has_N_main_name; eauto.
  - rewrite  <- logctx_extend_typ in *. eapply ty_has_N_main_bnd; eauto.

Qed.

Lemma ty_has_N_main_lc_N_main_replace G e oc T U:
  inert G ->
  ty_has_N_main G e T ->
  ty_has_N_main G e (logctx_typ oc U) ->
  ty_has_N_main G e (logctx_typ oc T).
Proof.
  intros Hin Hhas Hty.
  eapply ty_has_N_main_lc_nocap_main_replace.
  eauto. eauto.
  eapply ty_has_N_main_replace_nocap; eauto.
Qed.


Lemma ty_has_N_main_lc_M_replace G e oc T:
  inert G ->
  renv_corr G e ->
  ty_has_N_main G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_has_N_main G e (logctx_typ oc T).
Proof.
  intros Hin Hcor Hhas.
  forwards: ty_has_N_main_lift Hin.
  intros. eapply ty_has_N_logic_lc_M_replace. eauto.  eauto.
  eauto.
  eapply ty_inv_main_liftable_atomic. eauto.
  eauto.
  eauto.
Qed.

Lemma ty_has_N_logic_main_action_rev G e S1 S2 wa :
  ty_has_N_logic G e S2 ->
  wrap_action_utyp S1 wa S2 ->
  wrap_action_has_N_main G e wa ->
  ty_has_N_logic G e S1.
Proof.
  introv Hhas Hut Hm.
  destruct wa.
  inversions Hm; inversions Hut; simpls.
  -
    eapply ty_has_N_logic_lc_sel_replace; eauto.
  -
    eapply ty_has_N_logic_lc_bnd_replace; eauto.
Qed.


Lemma ty_has_N_logic_main_plan_rev G e S1 S2 wa :
  ty_has_N_logic G e S2 ->
  wrap_plan_utyp S1 wa S2 ->
  List.Forall (wrap_action_has_N_main G e) wa ->
  ty_has_N_logic G e S1.
Proof.
  introv Hhas Hut Hm.
  reverts S1 S2.
  induction wa; intros.
  - inversion Hut. auto.
  - inversions Hut. inversions Hm.
    forwards: ty_has_N_logic_main_action_rev; eauto. 
Qed.

Lemma ty_has_N_main_lc_and G e oc T U:
  inert G ->
  ty_has_N_main G e (logctx_typ oc T) ->
  ty_has_N_main G e (logctx_typ oc U) ->
  ty_has_N_main G e (logctx_typ oc (typ_and T U)).
Proof.
  introv Hi HtyT HtyU.
  (* Break up the paths to T and U *)
  eapply ty_has_N_main_plan_segment in HtyT; auto 1.
  eapply ty_has_N_main_plan_segment in HtyU; auto 1.
  destructecs_all.
  rename U1 into ST, wpsep0 into wpocT.
  rename oc0 into ocT, wpex0 into wpST.
  rename U0 into SU, wpsep into wpocU.
  rename oc2 into ocU, wpex into wpSU.

  forwards HlfU: List.Forall_impl (@wrap_action_has_N_main_leaf typing_mode G e) wpocU; eauto.
  forwards HlfT: List.Forall_impl (@wrap_action_has_N_main_leaf typing_mode G e) wpocT; eauto.
  
  (* Find a common prefix *)
  forwards Hsep: wrap_plan_separate_common  (wrap_action_has_N_main G e) ocT HtyT1 HtyU1; eauto 1.
  eapply wrap_action_has_N_main_determined.
  eapply wrap_action_has_N_main_closed.

  rewrite HtyT4; auto.
  rewrite HtyU4; auto.


  eapply logctx_path_eq_follow; eauto.
  eapply logctx_path_eq_follow; eauto.
  destructc Hsep as ocTU waTUU waTUT nU nT Hrest.
  destructc Hrest as HutSU HutST HreU HreT.
    
  (* Layer below *)
  forwards HtySU: ty_has_N_logic_main_plan_rev HtyU0 HutSU.
  eapply list_forall_reorder; eauto.


  forwards HtyST: ty_has_N_logic_main_plan_rev HtyT0 HutST.
  eapply list_forall_reorder; eauto.

  forwards HhSTU: ty_has_N_logic_lc_and HtyST HtySU.

  (* Connect the plan through intermediate types *)

  forwards HwpU: wrap_plan_separate_replace (typ_and ST SU) HtyU1 ; eauto.
  rewrite HtyU4; eauto.
  forwards HwpTU: HutST (typ_and ST SU).

  assert (wrap_plan_utyp (logctx_typ oc (typ_and ST SU)) (wrap_plan_under (logctx_path oc) (wrap_plan_under (logpath_andl logpath_nil) wpST)) (logctx_typ oc (typ_and T SU))) as HwpuTSU. {
    eapply wrap_plan_utyp_under.
    replace (typ_and ST SU) with (logctx_typ (logctx_andl logctx_nil SU) ST).
    replace (typ_and T SU) with (logctx_typ (logctx_andl logctx_nil SU) T).
    eapply wrap_plan_utyp_under. auto.
    simpls; auto.
    simpls; auto.
  }
  assert (wrap_plan_utyp (logctx_typ oc (typ_and T SU)) (wrap_plan_under (logctx_path oc) (wrap_plan_under (logpath_andr logpath_nil) wpSU)) (logctx_typ oc (typ_and T U))) as HwpuTU. {
    eapply wrap_plan_utyp_under.
    replace (typ_and T SU) with (logctx_typ (logctx_andr T logctx_nil) SU).
    replace (typ_and T U) with (logctx_typ (logctx_andr T logctx_nil) U).
    eapply wrap_plan_utyp_under. auto.
    simpls; auto.
    simpls; auto.
  }

  (* Apply the plans to typing *)

  eapply ty_has_N_plan_main; eauto.
  eapply ty_has_N_plan_main; eauto.
  eapply ty_has_N_plan_main; eauto.
  eapply ty_has_N_plan_logic_main; eauto.
  eapply list_forall_reorder; eauto.
  
  (* Consistency *)
  eapply list_forall_map.
  eapply list_forall_map.
  eapply list_forall_map_inv.
  eapply HtyT.
  intros; eapply wrap_action_has_N_main_closed; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.

  eapply list_forall_map.
  eapply list_forall_map.
  eapply list_forall_map_inv.
  eapply HtyU.
  intros; eapply wrap_action_has_N_main_closed; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.
Qed.

(* INVERSIONS *)
Lemma ty_has_N_main_lc_and_invr G e oc T U:
  inert G ->
  ty_has_N_main G e (logctx_typ oc (typ_and T U)) ->
  ty_has_N_main G e (logctx_typ oc U).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc (typ_and T U)).
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_has_N_main_logic. eapply ty_has_N_logic_lc_and_invr; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + discriminate.
    + destruct x; simpls; discriminate.
    + destruct x; simpls; inversions H1. 
      * eapply IHHty with (T:=(logctx_typ x (open z S1))); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
      * rewrite <- logctx_extend_typ.
        eapply ty_has_N_main_name; eauto.
        rewrite logctx_extend_typ.
        eapply IHHty with (T:=T0); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (U0).
      destruct_all.
      rewrite <- H4.
      eapply ty_has_N_main_name. 2: eauto.
      forwards: IHHty. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + discriminate.
    + destruct x; simpls; discriminate.
    + destruct x; simpls; inversions H0. 
      * eapply IHHty with (T:=(logctx_typ x (typ_nocap))); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
      * rewrite <- logctx_extend_typ.
        eapply ty_has_N_main_bnd; eauto.
        rewrite logctx_extend_typ.
        eapply IHHty with (T:=T1); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destruct_all.
      forwards: H1 (U).
      destruct_all.
      rewrite <- H3.
      eapply ty_has_N_main_bnd.  eauto.  eauto.
      forwards: IHHty. eauto. eapply H0. rewrite <- H4 in H6. eauto.
Qed.

Lemma ty_has_N_main_lc_and_invl G e oc T U:
  inert G ->
  ty_has_N_main G e (logctx_typ oc (typ_and T U)) ->
  ty_has_N_main G e (logctx_typ oc T).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc (typ_and T U)).
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_has_N_main_logic. eapply ty_has_N_logic_lc_and_invl; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + discriminate.
    + destruct x; simpls; discriminate.
    + destruct x; simpls; inversions H1. 

      * rewrite <- logctx_extend_typ.
        eapply ty_has_N_main_name; eauto.
        rewrite logctx_extend_typ.
        eapply IHHty with (U:=T0); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
      * eapply IHHty with (U:=(logctx_typ x (open z S1))); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (T).
      destruct_all.
      rewrite <- H4.
      eapply ty_has_N_main_name. 2: eauto.
      forwards: IHHty. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + discriminate.
    + destruct x; simpls; discriminate.
    + destruct x; simpls; inversions H0. 

      * rewrite <- logctx_extend_typ.
        eapply ty_has_N_main_bnd; eauto.
        rewrite logctx_extend_typ.
        eapply IHHty with (U:=T1); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
      * eapply IHHty with (U:=(logctx_typ x (typ_nocap))); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destruct_all.
      forwards: H1 (T0).
      destruct_all.
      rewrite <- H3.
      eapply ty_has_N_main_bnd.  eauto.  eauto.
      forwards: IHHty. eauto. eapply H0. rewrite <- H4 in H6. eauto.
Qed.
Lemma ty_has_N_main_lc_orl G e oc T U:
  inert G ->
  ty_has_N_main G e (logctx_typ oc T) ->
  ty_has_N_main G e (logctx_typ oc (typ_or T U)).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc T). (* reverts T.*)
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_has_N_main_logic. eapply ty_has_N_logic_lc_orl; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + rewrite logctx_extend_orl. eapply ty_has_N_main_name. 2: eauto.
      rewrite <- logctx_extend_orl. eauto.
    + destruct x; simpls; inversions H1.
      rewrite logctx_extend_nil_l in *.

      rewrite logctx_extend_orl. eapply ty_has_N_main_name. 2: eauto.
      rewrite <- logctx_extend_orl. eauto.
    +
      rewrite logctx_extend_typ in *.
      rewrite logctx_extend_orl.
      rewrite <- logctx_extend_typ.
      eapply ty_has_N_main_name. 2: eauto.
      rewrite logctx_extend_typ.
      rewrite logctx_extend_typ.
      simpls.
      eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (typ_or T U0).
      destruct_all.
      rewrite <- H4.
      eapply ty_has_N_main_name. 2: eauto.
      forwards: IHHty. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + rewrite logctx_extend_orl. eapply ty_has_N_main_bnd. eauto. eauto.
      rewrite <- logctx_extend_orl. eauto.
    + destruct x; simpls; inversions H0.
      rewrite logctx_extend_nil_l in *.

      rewrite logctx_extend_orl. eapply ty_has_N_main_bnd. eauto. eauto.
      rewrite <- logctx_extend_orl. eauto.
    +
      rewrite logctx_extend_typ in *.
      rewrite logctx_extend_orl.
      rewrite <- logctx_extend_typ.
      eapply ty_has_N_main_bnd.
      rewrite logctx_extend_typ.
      rewrite logctx_extend_typ.
      simpls.
      eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H typ_nocap.
      destruct_all.
      forwards: H1 (typ_or T0 U).
      destruct_all.
      rewrite <- H3.
      eapply ty_has_N_main_bnd. eauto. eauto.
      forwards: IHHty. eauto. eapply H0. rewrite <- H4 in H6. eauto.
Qed.

Lemma ty_has_N_main_lc_orr G e oc T U:
  inert G ->
  ty_has_N_main G e (logctx_typ oc U) ->
  ty_has_N_main G e (logctx_typ oc (typ_or T U)).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc U). (* reverts T.*)
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_has_N_main_logic. eapply ty_has_N_logic_lc_orr; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + rewrite logctx_extend_orr. eapply ty_has_N_main_name. 2: eauto.
      rewrite <- logctx_extend_orr. eauto.
    + destruct x; simpls; inversions H1.
      rewrite logctx_extend_nil_l in *.

      rewrite logctx_extend_orr. eapply ty_has_N_main_name. 2: eauto.
      rewrite <- logctx_extend_orr. eauto.
    +
      rewrite logctx_extend_typ in *.
      rewrite logctx_extend_orr.
      rewrite <- logctx_extend_typ.
      eapply ty_has_N_main_name. 2: eauto.
      rewrite logctx_extend_typ.
      rewrite logctx_extend_typ.
      simpls.
      eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (typ_or T U0).
      destruct_all.
      rewrite <- H4.
      eapply ty_has_N_main_name. 2: eauto.
      forwards: IHHty. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + rewrite logctx_extend_orr. eapply ty_has_N_main_bnd. eauto. eauto.
      rewrite <- logctx_extend_orr. eauto.
    + destruct x; simpls; inversions H0.
      rewrite logctx_extend_nil_l in *.

      rewrite logctx_extend_orr. eapply ty_has_N_main_bnd. eauto. eauto.
      rewrite <- logctx_extend_orr. eauto.
    +
      rewrite logctx_extend_typ in *.
      rewrite logctx_extend_orr.
      rewrite <- logctx_extend_typ.
      eapply ty_has_N_main_bnd.
      rewrite logctx_extend_typ.
      rewrite logctx_extend_typ.
      simpls.
      eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H typ_nocap.
      destruct_all.
      forwards: H1 (typ_or T0 U).
      destruct_all.
      rewrite <- H3.
      eapply ty_has_N_main_bnd. eauto. eauto.
      forwards: IHHty. eauto. eapply H0. rewrite <- H4 in H6. eauto.
Qed.
Lemma ty_has_N_main_lc_subtyp G e oc T U:
  inert G ->
  subtyp_atomic G e T U ->
  ty_has_N_main G e (logctx_typ oc T) ->
  ty_has_N_main G e (logctx_typ oc U).
Proof.
  introv Hi Hsub Hty.
  forwards: ty_has_N_main_lift Hi.
  intros. eapply ty_has_N_logic_lc_subtyp. eauto.
  eauto.
  eauto.
  eapply ty_inv_main_liftable_not_bndsel; intros; eapply subtyp_atomic_atomic_l in Hsub; destruct oc0; simpls; inversions Hsub; discriminate.
  eauto.
  eauto.
Qed.

Lemma ty_has_N_main_lc_typ_normal_replace G e oc A SL SU TL TU:
  inert G ->
  ty_has_N_main G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_normal A) SL SU))) ->
  ty_has_N_main G e (logctx_typ oc (typ_rcd (dec_typ (typ_label_normal A) TL TU))).
Proof.
  introv Hi Hty.

  forwards: ty_has_N_main_lift Hi.
  intros. eapply ty_has_N_logic_lc_typ_normal_replace. eauto.
  eapply ty_inv_main_liftable_not_bndsel; intros; destruct oc0; simpls; discriminate.
  eauto.
  eauto.
Qed.


Lemma ty_has_N_main_lc_typ_replace G e oc A SL SU:
  inert G ->
  ty_has_N_main G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  (forall  TL TU, ty_has_N_main G e (logctx_typ oc (typ_rcd (dec_typ A TL TU)))) \/ ty_has_top_atomic G e (typ_rcd (dec_typ A SL SU)).
Proof.
  introv Hi Hty.
  remember (logctx_typ oc (typ_rcd (dec_typ A SL SU))). (* reverts T.*)
  reverts oc.
  induction Hty; intros; subst.
  - forwards Hl: ty_has_N_logic_lc_typ_replace H.
    destructecs Hl. left. intros. eapply ty_has_N_main_logic. eauto. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs_all; subst.
    + inversions Heqt.
    +
      destruct ocST; simpl in Heqt; inversions Heqt.
    +
      destruct ocTS; simpl in Heqt; inversions Heqt.
    + (* separate *)
      forwards Hcross: logctx_separate_cross_r Heqt0 Heqt (open z S1).
      destructecs_all.

      forwards HHty: IHHty. eauto. eauto.
      destructecs HHty.
      left. intros.
      specialize (Hcross1 (typ_rcd (dec_typ A TL TU))).
      destructecs_all.
      specialize (HHty TL TU).
      rewrite <- Hcross3 in HHty.
      rewrite <- Hcross2.
      eapply ty_has_N_main_name; eauto.
      eauto.
  -
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs_all; subst.
    + inversions Heqt.
    +
      destruct ocST; simpl in Heqt; inversions Heqt.
    +
      destruct ocTS; simpl in Heqt; inversions Heqt.
    + (* separate *)
      forwards Hcross: logctx_separate_cross_r Heqt0 Heqt (typ_nocap).
      destructecs_all.

      forwards HHty: IHHty. eauto. eauto.
      destructecs HHty.
      left. intros.
      specialize (Hcross1 (typ_rcd (dec_typ A TL TU))).
      destructecs_all.
      specialize (HHty TL TU).
      rewrite <- Hcross3 in HHty.
      rewrite <- Hcross2.
      eapply ty_has_N_main_bnd; eauto.
      eauto.
Qed.




Lemma ty_has_N_main_lc_dist G e oc T U1 U2:
  inert G ->
  ty_has_N_main G e (logctx_typ oc (typ_and T (typ_or U1 U2))) ->
  ty_has_N_main G e (logctx_typ oc (typ_or (typ_and T U1) (typ_and T U2))).
Proof.
  introv Hi Hty.
  forwards HtyT: ty_has_N_main_lc_and_invr G; eauto.
  forwards HtyU: ty_has_N_main_lc_and_invl G; eauto.


  eapply ty_has_N_main_plan_segment in HtyT; auto 1.
  eapply ty_has_N_main_plan_segment in HtyU; auto 1.
  destruct_all.
  rename x6 into ST, x4 into wpST.
  rename x5 into ocT, x3 into wpocT.
  rename x2 into SU, x0 into wpSU.
  rename x1 into ocU, x into wpocU.

  forwards HlfU: List.Forall_impl (@wrap_action_has_N_main_leaf typing_mode G e) wpocU; eauto.
  forwards HlfT: List.Forall_impl (@wrap_action_has_N_main_leaf typing_mode G e) wpocT; eauto.

  forwards: wrap_plan_andor_segment logpath_or H8.
  eapply wrap_plan_has_N_main_leaf.

  eapply list_forall_map_inv. eapply H12.
  intros. eapply wrap_action_has_N_main_closed. eauto.

  destruct_all.
  rename x into wpSU1, x0 into wpSU2, x1 into wpSU12, x2 into nSU.
  subst.
  forwards: wrap_plan_andor_split H15.
  destructc H13 as ST1 ST2 HST HST1 HST2; subst. simpls.

  
  (* Find a common prefix *)
  forwards: wrap_plan_separate_common (wrap_action_has_N_main G e) H1 H7; eauto 1.
  eapply wrap_action_has_N_main_determined.
  eapply wrap_action_has_N_main_closed.

  rewrite H4; auto.
  rewrite H10; auto.

  eapply logctx_path_eq_follow; eauto.
  eapply logctx_path_eq_follow; eauto.
  destructc H13 as ocTU waTUU waTUT nU nT Hrest.
  destructc Hrest as HutSU HutST HreU HreT.
    
  (* Layer below *)
  forwards: ty_has_N_logic_main_plan_rev H0 HutSU.
  eapply list_forall_reorder; eauto.

  forwards: ty_has_N_logic_main_plan_rev H HutST.
  eapply list_forall_reorder; eauto.

  forwards: ty_has_N_logic_lc_and_dist H17 H13.

  (* Connect the plan through intermediate types *)

  forwards: wrap_plan_separate_replace (typ_or (typ_and SU ST1) (typ_and SU ST2)) H1 ; eauto.
  rewrite H4; eauto.
  forwards: HutST (typ_or (typ_and SU ST1) (typ_and SU ST2)).

  assert (wrap_plan_utyp (logctx_typ oc (typ_or (typ_and SU ST1) (typ_and SU ST2))) (wrap_plan_under (logctx_path oc) (wrap_plan_under (logpath_orl (logpath_andl logpath_nil)) wpSU)) (logctx_typ oc (typ_or (typ_and T ST1) (typ_and SU ST2)))). {
    eapply wrap_plan_utyp_under.
    replace (typ_or (typ_and SU ST1) (typ_and SU ST2)) with (logctx_typ (logctx_orl (logctx_andl logctx_nil ST1) (typ_and SU ST2)) SU).
    replace (typ_or (typ_and T ST1) (typ_and SU ST2)) with (logctx_typ (logctx_orl (logctx_andl logctx_nil ST1) (typ_and SU ST2)) T).
    eapply wrap_plan_utyp_under. auto.
    simpls; auto.
    simpls; auto.
  }
  assert (wrap_plan_utyp (logctx_typ oc (typ_or (typ_and T ST1) (typ_and SU ST2))) (wrap_plan_under (logctx_path oc) (wrap_plan_under (logpath_orr (logpath_andl logpath_nil)) wpSU)) (logctx_typ oc (typ_or (typ_and T ST1) (typ_and T ST2)))). {
    eapply wrap_plan_utyp_under.
    replace (typ_or (typ_and T ST1) (typ_and SU ST2)) with (logctx_typ (logctx_orr (typ_and T ST1) (logctx_andl logctx_nil ST2)) SU).
    replace (typ_or (typ_and T ST1) (typ_and T ST2)) with (logctx_typ (logctx_orr (typ_and T ST1) (logctx_andl logctx_nil ST2)) T).

    eapply wrap_plan_utyp_under. auto.
    simpls; auto.
    simpls; auto.
  }

  assert (wrap_plan_utyp (logctx_typ oc (typ_or (typ_and T ST1) (typ_and T ST2))) (wrap_plan_under (logctx_path oc) (wrap_plan_under (logpath_orl (logpath_andr logpath_nil)) wpSU1)) (logctx_typ oc (typ_or (typ_and T U1) (typ_and T ST2)))). {
    eapply wrap_plan_utyp_under.
    replace (typ_or (typ_and T ST1) (typ_and T ST2)) with (logctx_typ (logctx_orl (logctx_andr T logctx_nil) (typ_and T ST2)) ST1).
    replace (typ_or (typ_and T U1) (typ_and T ST2)) with (logctx_typ (logctx_orl (logctx_andr T logctx_nil) (typ_and T ST2)) U1).
    eapply wrap_plan_utyp_under. auto.
    simpls; auto.
    simpls; auto.
  }

  assert (wrap_plan_utyp (logctx_typ oc (typ_or (typ_and T U1) (typ_and T ST2))) (wrap_plan_under (logctx_path oc) (wrap_plan_under (logpath_orr (logpath_andr logpath_nil)) wpSU2)) (logctx_typ oc (typ_or (typ_and T U1) (typ_and T U2)))). {
    eapply wrap_plan_utyp_under.
    replace (typ_or (typ_and T U1) (typ_and T ST2)) with (logctx_typ (logctx_orr (typ_and T U1) (logctx_andr T logctx_nil)) ST2).
    replace (typ_or (typ_and T U1) (typ_and T U2)) with (logctx_typ (logctx_orr (typ_and T U1) (logctx_andr T logctx_nil)) U2).

    eapply wrap_plan_utyp_under. auto.
    simpls; auto.
    simpls; auto.
  }


  (* Apply the plans to typing *)

  eapply ty_has_N_plan_main. 2: eauto.
  eapply ty_has_N_plan_main. 2: eauto.
  eapply ty_has_N_plan_main. 2: eauto.
  eapply ty_has_N_plan_main. 2: eauto.
  eapply ty_has_N_plan_main. 2: eauto.
  eapply ty_has_N_plan_logic_main. 2: eauto. eauto.
 
  (* Consistency *)
  eapply list_forall_reorder; eauto.
  eauto.
  
  eapply list_forall_map.
  eapply list_forall_map.
  eapply list_forall_map_inv.
  eapply H6.
  intros; eapply wrap_action_has_N_main_closed; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.

  eapply list_forall_map.
  eapply list_forall_map.
  eapply list_forall_map_inv.
  eapply H6.
  intros; eapply wrap_action_has_N_main_closed; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.

  eapply list_forall_map.
  eapply list_forall_map.
  eapply list_forall_map_inv.
  eapply list_forall_app_invl.
  eapply list_forall_reorder.
  eapply H14.
  eapply list_forall_map_inv.
  eauto.
  intros; eapply wrap_action_has_N_main_closed; eauto.
  intros; eapply wrap_action_has_N_main_closed; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.

  eapply list_forall_map.
  eapply list_forall_map.
  eapply list_forall_map_inv.
  eapply list_forall_app_invr.
  eapply list_forall_reorder.
  eapply H14.
  eapply list_forall_map_inv.
  eauto.
  intros; eapply wrap_action_has_N_main_closed; eauto.
  intros; eapply wrap_action_has_N_main_closed; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.
  intros; eapply wrap_action_has_N_main_closed_inv; eauto.
Qed.

Lemma ty_has_N_main_lc_sel_open G e oc U TL TU x0 A y :
  inert G ->
  G ⊢! x0 ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
  ty_has_N_main G e (logctx_typ oc (typ_sel (avar_f x0) A (avar_f y))) ->
  ty_has_N_main G e (logctx_typ oc (open y TU)).
Proof.
  introv Hi Hpf Hty.

  remember (logctx_typ oc (typ_sel (avar_f x0) A
              (avar_f y))).
  reverts oc.
  induction Hty; intros; subst.
  - eapply ty_has_N_main_logic.
    eapply ty_has_N_logic_lc_sel_replace; eauto.
  - 
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + inversions H1.
      lets Hu: (x_bound_unique Hpf H). subst.
      pose proof (pf_inert_unique_tight_bounds Hi Hpf H) as Hu. destruct Hu. subst.
      destruct (pf_dec_typ_inv Hi H) as [He|Hb].
      * subst. assumption.
      * subst. simpls.
        eapply ty_has_N_main_lc_bot_replace; eauto.
    + destruct x; simpls; inversions H1.
      contradiction.
    + destruct x; simpls; inversions H1.
      contradiction.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (open y TU).
      destruct_all.
      rewrite <- H4.
      eapply ty_has_N_main_name. 2: eauto.
      forwards: IHHty. eauto. eauto. eapply H1. rewrite <- H5 in H7. eauto.
 - 
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + inversions H0.
    + destruct x; simpls; inversions H0.
    + destruct x; simpls; inversions H0.
    +
      forwards: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destruct_all.
      forwards: H1 (open y TU).
      destruct_all.
      rewrite <- H3.
      eapply ty_has_N_main_bnd.
      forwards: IHHty. eauto. eauto. eapply H0. eauto. rewrite H4. eauto.
Qed.

Lemma ty_has_N_main_lc_sel_open1 G e oc U TL TU x0 A y :
  inert G ->
  G ⊢! x0 ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
  ty_has_N_main G e (logctx_typ oc (typ_sel (avar_f x0) A (avar_f y))) ->
  ty_has_N_main G e (logctx_typ oc (open y TL)).
Proof.
  introv Hi Hpf Hty.

  remember (logctx_typ oc (typ_sel (avar_f x0) A
              (avar_f y))).
  reverts oc.
  induction Hty; intros; subst.
  - eapply ty_has_N_main_logic.
    eapply ty_has_N_logic_lc_sel_replace; eauto.
  - 
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + inversions H1.
      lets Hu: (x_bound_unique Hpf H). subst.
      pose proof (pf_inert_unique_tight_bounds Hi Hpf H) as Hu. destruct Hu. subst.
      destruct (pf_dec_typ_inv Hi H) as [He|Hb].
      * subst. assumption.
      * subst. simpls.
        eapply ty_has_N_main_lc_bot_replace; eauto.
    + destruct x; simpls; inversions H1.
      contradiction.
    + destruct x; simpls; inversions H1.
      contradiction.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (open y TL).
      destruct_all.
      rewrite <- H4.
      eapply ty_has_N_main_name. 2: eauto.
      forwards: IHHty. eauto. eauto. eapply H1. rewrite <- H5 in H7. eauto.
 - 
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + inversions H0.
    + destruct x; simpls; inversions H0.
    + destruct x; simpls; inversions H0.
    +
      forwards: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destruct_all.
      forwards: H1 (open y TL).
      destruct_all.
      rewrite <- H3.
      eapply ty_has_N_main_bnd.
      forwards: IHHty. eauto. eauto. eapply H0. eauto. rewrite H4. eauto.
Qed.

Lemma ty_has_N_main_lc_or_diag G e oc T:
  inert G ->
  ty_has_N_main G e (logctx_typ oc (typ_or T T)) ->
  ty_has_N_main G e (logctx_typ oc T).
Proof.
  introv Hi Hh.
  (* Extract the plan of changes *)
  eapply ty_has_N_main_plan_segment in Hh; auto 1.
  destructecs_all.
  rename oc2 into ocU, Hh0 into HhU, Hh1 into Hwpusep, Hh2 into Hwpuex, Hh3 into Hsep, Hh4 into Heqp, Hh into Hwasep.
  assert (List.Forall (wrap_action_has_N_main G e) wpex) as Hwaex. {
    eapply list_forall_map_inv. eauto.
    intros; eapply wrap_action_has_N_main_closed; eauto.
  }

  eapply wrap_plan_andor_segment with (TB:=logpath_or) in Hwpuex.
  2: {
    eapply List.Forall_impl. 2: eapply list_forall_map_inv. 2: eapply Hwasep. intros; eapply wrap_action_has_N_main_leaf; eauto.
    intros; eapply wrap_action_has_N_main_closed; eauto.
  }
  destructecs_all; subst; simpls.
  rename ns into nsex, wpex0 into wpexl, wpsep0 into wpexr.

  assert (List.Forall (wrap_action_has_N_main G e) wpexl) as Hwpexl. {

    eapply list_forall_map_inv.
    eapply list_forall_app_invl.
    eapply list_forall_reorder. eauto. eauto.
    intros; eapply wrap_action_has_N_main_closed; eauto.
  }
  assert (List.Forall (wrap_action_has_N_main G e) wpexr) as Hwpexr. {

    eapply list_forall_map_inv.
    eapply list_forall_app_invr.
    eapply list_forall_reorder. eauto. eauto.
    intros; eapply wrap_action_has_N_main_closed; eauto.
  }

  eapply wrap_plan_andor_split with  (TB:=logpath_or) in Hwpuex2.
  destructecs_all; subst.
  rename Tl0 into Ul, Tr0 into Ur.
  forwards Hwpscs: wrap_plan_separate_common_same (wrap_action_has_N_main G e) Hwpuex2 Hwpuex3.
  eapply wrap_action_has_N_main_determined. eapply wrap_action_has_N_main_closed.
  eapply wrap_plan_has_N_main_leaf; eauto. eapply wrap_plan_has_N_main_leaf; eauto. eauto. eauto.
  destructecs_all.
  rename TU into V, waTUU into wpVUl, waTUT into wpVUr, nU into nsVUr, nT into nsVUl.
  simpls.

  assert (List.Forall (wrap_action_has_N_main G e) wpVUl) as HwpVUl. {

    eapply list_forall_reorder. eauto.
    eapply list_forall_map_inv.
    eapply list_forall_app_invr.
    eapply list_forall_reorder. eauto. eauto.
    intros; eapply wrap_action_has_N_main_closed; eauto.
  }
  assert (ty_has_N_logic G e (logctx_typ ocU (typ_or V V))) as HtyV. {
    forwards Hrevl: ty_has_N_logic_main_plan_rev G e (logctx_typ (logctx_extend ocU (logctx_orl logctx_nil Ur)) V) (logctx_typ (logctx_extend ocU (logctx_orl logctx_nil Ur)) Ul).
    rewrite logctx_extend_typ. simpls.  eauto.
    eapply wrap_plan_utyp_under. eauto.
    eapply list_forall_map. eauto.
    intros; eapply wrap_action_has_N_main_closed_inv; eauto.

    forwards Hrevr: ty_has_N_logic_main_plan_rev G e (logctx_typ (logctx_extend ocU (logctx_orr V logctx_nil)) V) (logctx_typ (logctx_extend ocU (logctx_orr V logctx_nil)) Ur).
    rewrite logctx_extend_typ.  rewrite logctx_extend_typ in Hrevl. simpls. auto.
    eapply wrap_plan_utyp_under. eauto.
    eapply list_forall_map.
    eapply list_forall_reorder. eauto. eauto.
    intros; eapply wrap_action_has_N_main_closed_inv; eauto.

    rewrite logctx_extend_typ in Hrevr. simpls. auto.

  }
  eapply ty_has_N_logic_lc_or_diag in HtyV.
  assert (ty_has_N_main G e (logctx_typ ocU T)) as HtyT. {
    eapply ty_has_N_plan_logic_main. eauto. 
    eapply wrap_plan_utyp_under.
    eapply wrap_plan_utyp_app. eapply Hwpscs0. eauto.

    eapply list_forall_map.
    eapply list_forall_app. eauto.
    eauto.
    intros; eapply wrap_action_has_N_main_closed_inv; eauto.
  }

  eapply ty_has_N_plan_main. eauto. 
  eapply wrap_plan_separate_replace. eauto. eapply Hwpusep. rewrite Heqp. eauto. eauto.
Qed.


Lemma ty_has_N_main_lc_cases G e oc T:
  ty_has_N_main G e (logctx_typ oc T) ->
  ty_has_N_main G e T \/ (forall U, ty_has_N_main G e (logctx_typ oc U)).
Proof.
  introv Hty.
  remember (logctx_typ oc (T)).
  reverts oc T.
  induction Hty; intros; subst.
  - forwards: ty_has_N_logic_lc_cases; eauto. destruct_all. left. eapply ty_has_N_main_logic. eauto. right. intros. eapply ty_has_N_main_logic; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destructecs Heqt; subst.
    + forwards HHty: IHHty; eauto. destructecs HHty. left. eapply ty_has_N_main_name with (oc:=logctx_nil); eauto. eauto.
    + destruct ocST; simpls. false. all: discriminate.
    + forwards: IHHty oc0 (logctx_typ ocTS (open z S1)); simpls; eauto. rewrite logctx_extend_typ. auto.
      destruct_all. left. eapply ty_has_N_main_name with (oc:=ocTS); eauto. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 Heqt (open z S1).
      destruct_all.
      forwards: IHHty; eauto.
      destruct_all.
      auto.
      right. intro.

      forwards HocST1: H1 (U0).
      destructecs HocST1.
 
      rewrite <- HocST0.
      eapply ty_has_N_main_name. 2: eauto. rewrite HocST2. eauto.
  - (* Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destructecs Heqt; subst.
    + forwards: IHHty; eauto. destruct_all. left. eapply ty_has_N_main_bnd with (oc:=logctx_nil); eauto. eauto.
    + destruct ocST; simpls. false. all: discriminate.
    + forwards: IHHty oc0 (logctx_typ ocTS (typ_nocap)); simpls; eauto. rewrite logctx_extend_typ. auto.
      destruct_all. left. eapply ty_has_N_main_bnd with (oc:=ocTS); eauto. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (typ_nocap).
      destructecs Hcr.
      forwards HHty: IHHty; eauto.
      destructecs HHty.
      auto.
      right. intro.

      forwards HocST1: Hcr1 (U).
      destructecs HocST1.

      rewrite <- HocST0.
      eapply ty_has_N_main_bnd. rewrite HocST2. eauto.
Qed.

Lemma ty_has_N_main_atomic_logic G e  T:
  ty_has_N_main G e T ->
  typ_atomic T ->
  ty_has_N_logic G e T.
Proof.
  introv Hty Ha.
  induction Hty; eauto.
  all:
  forwards: logctx_atomic_inv Ha; destruct_all; subst; simpls; inversions Ha.
Qed.

Lemma ty_has_N_atomic_atomic G e T:
  ty_has_N_main G e T ->
  typ_atomic T ->
  ty_has_N G e T.
Proof.
  introv Hty Ha.
  eapply ty_has_N_closure_atomic_basic.
  eapply ty_has_N_logic_atomic_closure.
  eapply ty_has_N_main_atomic_logic.
  all: eauto.
Qed.

Lemma ty_has_N_main_typ_and G e A S1 S2 T1 T2:
  inert G ->
  ty_has_N_main G e (typ_rcd (dec_typ A S1 T1)) ->
  ty_has_N_main G e (typ_rcd (dec_typ A S2 T2)) ->
  ty_has_N_main G e (typ_rcd (dec_typ A (typ_or S1 S2) (typ_and T1 T2))).
Proof.
  introv Hin Hc1 Hc2.

  forwards Hh1: ty_has_N_atomic_atomic Hc1; fweauto.
  forwards Hh2: ty_has_N_atomic_atomic Hc2; fweauto.

  inversions Hh1.
  -
  inversions H.
  eapply ty_has_N_main_lc_typ_normal_replace with (oc:=logctx_nil); eauto.
  -
  inversions Hh2.
  +
  inversions H0.
  eapply ty_has_N_main_lc_typ_normal_replace with (oc:=logctx_nil); eauto.
  +
  inversions H.
  inversions H0.
  eapply ty_has_N_main_lc_subtyp  with (oc:=logctx_nil); eauto.
  constructor.
  intros. eapply subtyp_or2_t. eauto. eapply subtyp_trans_t. eapply H3. eauto.
  intros. eapply subtyp_and2_t. eauto. eapply subtyp_trans_t. 2: eapply H7. eauto.
Qed.


Lemma ty_has_N_main_lc_typ_and G e oc A S1 S2 T1 T2:
  inert G ->
  ty_has_N_main G e (logctx_typ oc (typ_and (typ_rcd (dec_typ A S1 T1)) (typ_rcd (dec_typ A S2 T2)))) ->
  ty_has_N_main G e (logctx_typ oc (typ_rcd (dec_typ A (typ_or S1 S2) (typ_and T1 T2)))).
Proof.
  introv Hin Hh.
  forwards Hh1: ty_has_N_main_lc_and_invl Hh; fweauto.
  forwards Hh2: ty_has_N_main_lc_and_invr Hh; fweauto.

  (* Destruct cases and solve simple ones *)
  forwards Hc1: ty_has_N_main_lc_cases Hh1.
  forwards Hc2: ty_has_N_main_lc_cases Hh2.
  destructecs_all; eauto 1.

  forwards Hha: ty_has_N_main_typ_and Hc1 Hc2; fweauto.
  eapply ty_has_N_main_lc_N_main_replace; eauto.
Qed.


Lemma ty_has_N_main_lc_closure_tight G e oc T U:
  inert G ->
  renv_corr G e ->
  ty_has_N_main G e (logctx_typ oc T) ->
  G⁏e ⊢# T <: U ->
  ty_has_N_main G e (logctx_typ oc U).
Proof.
  introv Hin Hcor Hty Hsub.
  reverts oc.
  dependent induction Hsub; intros; eauto.
  - eapply ty_has_N_main_lc_top_replace; eauto.
  - eapply ty_has_N_main_lc_bot_replace; eauto.
  - eapply ty_has_N_main_lc_and_invl; eauto.
  - eapply ty_has_N_main_lc_and_invr; eauto.
  - eapply ty_has_N_main_lc_and; eauto.
  - eapply ty_has_N_main_lc_orl; eauto.
  - eapply ty_has_N_main_lc_orr; eauto.
  -
    lets HTU: Hty.
    rewrite logctx_extend_orl in HTU.
    eapply IHHsub1 in HTU; eauto.

    rewrite <- logctx_extend_orl in HTU.

    lets HTS: HTU.
    rewrite logctx_extend_orr in HTS.
    eapply IHHsub2 in HTS; eauto.
    rewrite <- logctx_extend_orr in HTS.
    eapply ty_has_N_main_lc_or_diag; eauto.
  - eapply ty_has_N_main_lc_subtyp; eauto. eapply subtyp_atomic_fld; eauto.
  - eapply ty_has_N_main_lc_subtyp; eauto. eapply subtyp_atomic_typ; eauto.
  - eapply ty_has_N_main_name; eauto.
  - eapply ty_has_N_main_lc_sel_open; eauto.
  - eapply ty_has_N_main_lc_subtyp; eauto. eapply subtyp_atomic_fun; eauto.
  - eapply ty_has_N_main_lc_subtyp; eauto. eapply subtyp_atomic_met; eauto.
  - eapply ty_has_N_main_tyq; eauto. eapply logctx_tyq_same; eauto. 
  - eapply ty_has_N_main_lc_typ_and; eauto.
  - eapply ty_has_N_main_lc_dist; eauto.
  - eapply ty_has_N_main_bnd; eauto.
  - eapply ty_has_N_main_lc_nocap_replace; eauto.
  - eapply ty_has_N_main_lc_nocap_replace; eauto.
  - eapply ty_has_N_main_lc_nocap_replace; eauto.
  - eapply ty_has_N_main_lc_nocap_replace; eauto.
  - eapply ty_has_N_main_NM_replace.
    + destruct (mode_has_top_nm_subtype' typing_mode) eqn: Hm.
      auto.
      typing_mode_contradiction typing_mode.
    + eauto.
    + intros. typing_mode_contradiction typing_mode.
    + eauto.
  - eapply ty_has_N_main_lc_orl. eauto.
    eapply ty_has_N_main_replace_nocap; eauto.
  - eapply ty_has_N_main_lc_and_invr in Hty; rweauto. 
    eapply ty_has_N_main_lc_M_replace in Hty; eauto.
  - eapply ty_has_N_main_lc_N_main_replace; eauto.
    eapply ty_has_N_main_logic. eapply ty_has_N_logic_or. eapply ty_or_closure_atomic. eapply ty_has_N_has_top. eauto.
Qed.


Lemma subtyp_N_has_N_main G e T:
  inert G ->
  renv_corr G e ->
  G⁏e ⊢# typ_nocap <: T ->
  ty_has_N_main G e T.
Proof.
  introv Hi Hcor Hsub.
  applys ty_has_N_main_lc_closure_tight logctx_nil; eauto.
  simpls.
  eapply ty_has_N_main_logic.
  eapply ty_has_N_logic_or.
  eapply ty_or_closure_atomic.
  eauto.
Qed.

Lemma ty_has_N_main_tight G e T:
  ty_has_N_main G e  T ->
  G⁏e ⊢# typ_nocap <: T.
Proof.
  introv Hh.
  induction Hh; eauto 2 using ty_has_N_logic_tight.
  - eapply subtyp_trans_t. 2: eapply subtyp_lc_tight. eauto. eauto.
  - eapply subtyp_trans_t. 2: eapply subtyp_lc_tight. eauto. eauto.
Qed.




End WithTypingMode.
