Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx.


Section WithTypingMode.
Context {typing_mode: Mode}.

(** * Helper definitions and lemmas for layered typing *)

Lemma logctx_tyq_same e oc T U: 
  typing_equiv e T U ->
  typing_equiv e (logctx_typ oc T) (logctx_typ oc U).
Proof.
  induction oc; simpls; eauto.
Qed.
Lemma logctx_tyq_inv e oc T U: 
  typing_equiv e (logctx_typ oc T) (logctx_typ oc U) ->
  typing_equiv e T U.
Proof.
  introv Htyq.
  induction oc; simpls; eauto.
  - inversions_tyq Htyq. inversions H; eauto.
  - inversions_tyq Htyq. inversions H; eauto.
  - inversions_tyq Htyq. inversions H; eauto.
  - inversions_tyq Htyq. inversions H; eauto.
Qed.

Inductive logctx_tyq: renv -> logctx -> logctx -> Prop :=
| logctx_tyq_nil e:
    logctx_tyq e logctx_nil logctx_nil
| logctx_tyq_orl e oc1 oc2 T1 T2:
    typing_equiv e T1 T2 ->
    logctx_tyq e oc1 oc2 ->
    logctx_tyq e (logctx_orl oc1 T1) (logctx_orl oc2 T2)
| logctx_tyq_orr e oc1 oc2 T1 T2:
    typing_equiv e T1 T2 ->
    logctx_tyq e oc1 oc2 ->
    logctx_tyq e (logctx_orr T1 oc1) (logctx_orr T2 oc2)
| logctx_tyq_andl e oc1 oc2 T1 T2:
    typing_equiv e T1 T2 ->
    logctx_tyq e oc1 oc2 ->
    logctx_tyq e (logctx_andl oc1 T1) (logctx_andl oc2 T2)
| logctx_tyq_andr e oc1 oc2 T1 T2:
    typing_equiv e T1 T2 ->
    logctx_tyq e oc1 oc2 ->
    logctx_tyq e (logctx_andr T1 oc1) (logctx_andr T2 oc2)
.
Hint Constructors logctx_tyq: core.

Lemma logctx_tyq_refl e oc:
  logctx_tyq e oc oc.
Proof.
  induction oc; simpls; eauto.
Qed.

Lemma logctx_top_inv oc T:
  typ_top = (logctx_typ oc T) ->
  oc = logctx_nil /\ T = typ_top.
Proof.
  introv Ha.
  destruct oc.
  - auto.
  - false; simpls; inversions Ha.
  - false.
  - false.
  - false.
Qed.
Lemma logctx_tyq_top_inv e oc T:
  typing_equiv e typ_top (logctx_typ oc T) ->
  oc = logctx_nil /\ T = typ_top.
Proof.
  introv Htyq.
  destruct oc; simpls; inversions_tyq Htyq; eauto; discriminate.
Qed.

Lemma logctx_tyq_typ e oc1 oc2 T U: 
  logctx_tyq e oc1 oc2 ->
  typing_equiv e T U ->
  typing_equiv e (logctx_typ oc1 T) (logctx_typ oc2 U).
Proof.
  introv Hoc Htyq.
  induction Hoc; simpls; eauto.
Qed.

Inductive orctx_tyq: renv -> orctx -> orctx -> Prop :=
| orctx_tyq_nil e:
    orctx_tyq e orctx_nil orctx_nil
| orctx_tyq_orl e oc1 oc2 T1 T2:
    typing_equiv e T1 T2 ->
    orctx_tyq e oc1 oc2 ->
    orctx_tyq e (orctx_l oc1 T1) (orctx_l oc2 T2)
| orctx_tyq_orr e oc1 oc2 T1 T2:
    typing_equiv e T1 T2 ->
    orctx_tyq e oc1 oc2 ->
    orctx_tyq e (orctx_r T1 oc1) (orctx_r T2 oc2)
.
Hint Constructors orctx_tyq: core.

Lemma orctx_tyq_typ e oc1 oc2 T U: 
  orctx_tyq e oc1 oc2 ->
  typing_equiv e T U ->
  typing_equiv e (orctx_typ oc1 T) (orctx_typ oc2 U).
Proof.
  introv Hoc Htyq.
  induction Hoc; simpls; eauto.
Qed.


Lemma orctx_tyq_case e oc1 T V:
  typing_equiv e (orctx_typ oc1 T) V ->
  exists oc2 U,
    V = (orctx_typ oc2 U) /\
    orctx_tyq e oc1 oc2 /\
    typing_equiv e T U.
Proof.
  introv Htyq.
  reverts V.
  induction oc1; intros; simpls.
  - exists orctx_nil V; simpls; eauto.
  - inversions_tyq Htyq.
    forwards: IHoc1; eauto.
    destruct_all; subst.
    exists (orctx_l x t22); simpls; eauto.
  - inversions_tyq Htyq.
    forwards: IHoc1; eauto.
    destruct_all; subst.
    exists (orctx_r t21 x); simpls; eauto.
Qed.
Lemma logctx_tyq_case e oc1 T V:
  typing_equiv e (logctx_typ oc1 T) V ->
  exists oc2 U,
    V = (logctx_typ oc2 U) /\
    logctx_tyq e oc1 oc2 /\
    typing_equiv e T U.
Proof.
  introv Htyq.
  reverts V.
  induction oc1; intros; simpls.
  - exists logctx_nil V; simpls; eauto.
  - inversions_tyq Htyq.
    forwards: IHoc1; eauto.
    destruct_all; subst.
    exists (logctx_orl x t22); simpls; eauto.
  - inversions_tyq Htyq.
    forwards: IHoc1; eauto.
    destruct_all; subst.
    exists (logctx_orr t21 x); simpls; eauto.
  - inversions_tyq Htyq.
    forwards: IHoc1; eauto.
    destruct_all; subst.
    exists (logctx_andl x t22); simpls; eauto.
  - inversions_tyq Htyq.
    forwards: IHoc1; eauto.
    destruct_all; subst.
    exists (logctx_andr t21 x); simpls; eauto.
Qed.


(* Closes a type property under intersection *)
Inductive ty_or_closure (P: typ -> Prop): typ -> Prop :=
| ty_or_closure_atomic (T: typ):
    P T ->
    ty_or_closure P T
| ty_or_closure_l T U:
    ty_or_closure P T ->
    ty_or_closure P (typ_or T U)
| ty_or_closure_r T U:
    ty_or_closure P T ->
    ty_or_closure P (typ_or U T)
| ty_or_closure_top:
    ty_or_closure P typ_top
.

(* Identify atomic types *)
Inductive typ_atomic: typ -> Prop :=
| typ_atomic_rcd D:
    typ_atomic (typ_rcd D)
| typ_atomic_fun S T:
    typ_atomic (typ_fun S T)
.
Hint Constructors typ_atomic: core.

Lemma typ_atomic_tyq e U V:
  typ_atomic U ->
  typing_equiv e U V ->
  typ_atomic V.
Proof.
  introv Ht Htyq.
  induction Ht; inversions_tyq Htyq; eauto.
Qed.

Lemma logctx_atomic_inv oc T:
  typ_atomic (logctx_typ oc T) ->
  oc = logctx_nil /\ typ_atomic T.
Proof.
  introv Ha.
  destruct oc.
  - auto.
  - false; simpls; inversions Ha.
  - false. simpls. inversions Ha.
  - false. simpls. inversions Ha.
  - false. simpls. inversions Ha.
Qed.

Lemma orctx_atomic_inv oc T:
  typ_atomic (orctx_typ oc T) ->
  oc = orctx_nil /\ typ_atomic T.
Proof.
  introv Ha.
  destruct oc.
  - auto.
  - false; simpls; inversions Ha.
  - false. simpls. inversions Ha.
Qed.

(* Closes a type property under intersection *)
Inductive ty_and_closure (P: typ -> Prop): typ -> Prop :=
| ty_and_closure_atomic (T: typ):
    P T ->
    ty_and_closure P T
| ty_and_closure_and T U:
    ty_and_closure P T ->
    ty_and_closure P U ->
    ty_and_closure P (typ_and T U)
.

End WithTypingMode.

Notation ty_pf_not_typ G x A := (forall U TL TU, G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU) -> False) (only parsing).
Definition ty_pf_or_none G x A TL TU := ((exists U, G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU)) \/ ty_pf_not_typ G x A).

Notation typ_NM T1 T2 := (typ_or typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) T1 T2))) (only parsing).
Notation typ_NM_canon := (typ_NM typ_top typ_bot) (only parsing).
(* Repeated hints *)
Hint Constructors typ_atomic: core.

Section WithTypingMode.
Context {typing_mode: Mode}.

Lemma ty_pf_typ_decide G x A:
  inert G ->
  (exists U TL TU, 
    G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU)) \/ (ty_pf_not_typ G x A).
Proof.
  introv Hin.
  destruct (classicT (exists U TL TU, G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU))).
  eauto.
  right.
  intros.
  eapply n; eauto.
Qed.

Lemma ty_pf_mutability_bound G e x T Tl Tu:
  inert G ->
  G ⊢! x ∶ T ⪼ typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) Tl Tu) ->
  mode_has_mutability typing_mode ->
  forall x0 : var, G ⁏ e ⊢# Tl /[-> x0] <: typ_bot.
Proof.
  introv Hi Hpf Hm.
  forwards Hbi: pf_binds Hpf; fweauto.
  forwards Hpfm: pf_mutability typing_mode Hi; fweauto.
  destructecs Hpfm.
  forwards Hpfb: pf_inert_unique_tight_bounds Hpf Hpfm; fweauto.
  destructecs Hpfb; subst.
  eauto.
Qed.


Lemma ty_pf_or_none_mutability_bound G e x TL TU:
  inert G ->
  x \in ctx_dom G ->
  ty_pf_or_none G x (typ_label_refcap typ_refcap_mutability) TL TU ->
  mode_has_top_nm_complement typing_mode ->
  forall x0 : var, G ⁏ e ⊢# TL /[-> x0] <: typ_bot.
Proof.
  introv Hi Hdom Hpf Hm.
  forwards Hbi: ctx_dom_to_binds Hdom; fweauto.
  destructecs Hbi.
  forwards Hpfm: pf_mutability typing_mode Hi.
  destruct (mode_has_mutability' typing_mode) eqn: Hmm. auto. typing_mode_contradiction typing_mode.
  eauto.
  destructecs Hpfm.
  inversions Hpf.
  -
  destructecs H.
  forwards Hpfb: pf_inert_unique_tight_bounds H Hpfm; fweauto.
  destructecs Hpfb; subst.
  eauto.
  -
  false; eauto.
Qed.

(* ------------- Logic typing ------------------ *)

(* Restricted version of (T :> Top) *)
Inductive typ_logic_toplike (G: ctx) (e: renv): typ -> Prop :=
| typ_logic_toplike_top: 
    typ_logic_toplike G e typ_top
| typ_logic_toplike_NM T1 T2:
    mode_has_top_nm_subtype typing_mode ->
    (mode_has_top_nm_complement typing_mode -> forall x0 : var, G ⁏ e ⊢# T1 /[-> x0] <: typ_bot) ->
    typ_logic_toplike G e (typ_NM T1 T2)
| typ_logic_toplike_has_top TL1 TU1: 
  inert G ->
  renv_corr G e ->
  mode_has_mut_top typing_mode ->
  (forall x : var, G⁏e ⊢# TL1/[-> x] <: typ_bot) ->
  (forall x : var, G⁏e ⊢# typ_top <: TU1/[-> x]) ->
  typ_logic_toplike G e (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) TL1 TU1))
.

(* Which types can be used with ty_inv_logic_lift *)
Inductive ty_inv_logic_liftable (G: ctx) (e: renv): forall (V S: typ), Prop :=
| ty_inv_logic_liftable_not_and V S:
  (forall oc T U, V <> orctx_typ oc (typ_and T U)) ->
  ty_inv_logic_liftable G e V S
| ty_inv_logic_liftable_toplike V S:
  typ_logic_toplike G e S ->
  ty_inv_logic_liftable G e V S
.


(* Which types can be used with ty_inv_main_lift *)
Inductive ty_inv_main_liftable: forall (T S: typ), Prop :=
| ty_inv_main_liftable_not_bndsel T S:
  (forall oc x A y, T <> logctx_typ oc (typ_sel x A y)) ->
  (forall oc U, T <> logctx_typ oc (typ_bnd U)) ->
  ty_inv_main_liftable T S
.

End WithTypingMode.

Hint Constructors typ_logic_toplike: core.
Hint Constructors ty_inv_logic_liftable: core.
Hint Constructors ty_inv_main_liftable: core.

Section WithTypingMode.
Context {typing_mode: Mode}.

Lemma ty_inv_logic_liftable_bot G e S:
  ty_inv_logic_liftable G e typ_bot S.
Proof.
  eapply ty_inv_logic_liftable_not_and.
  intros; destruct oc; simpls; discriminate.
Qed.
Lemma ty_inv_logic_liftable_nocap G e S:
  ty_inv_logic_liftable G e typ_nocap S.
Proof.
  eapply ty_inv_logic_liftable_not_and.
  intros; destruct oc; simpls; discriminate.
Qed.
Lemma ty_inv_logic_liftable_atomic G e T S:
  typ_atomic T ->
  ty_inv_logic_liftable G e T S.
Proof.
  introv Hat.
  eapply ty_inv_logic_liftable_not_and.
  intros; destruct oc; simpls; inversions Hat; discriminate.
Qed.
Lemma ty_inv_main_liftable_bot S:
  ty_inv_main_liftable typ_bot S .
Proof.
  eapply ty_inv_main_liftable_not_bndsel;
  intros; destruct oc; simpls; discriminate.
Qed.
Lemma ty_inv_main_liftable_nocap S:
  ty_inv_main_liftable typ_nocap S.
Proof.
  eapply ty_inv_main_liftable_not_bndsel;
  intros; destruct oc; simpls; discriminate.
Qed.
Lemma ty_inv_main_liftable_atomic T S:
  typ_atomic T ->
  ty_inv_main_liftable T S.
Proof.
  introv Ht.
  eapply ty_inv_main_liftable_not_bndsel;
  intros; destruct oc; inversions Ht; simpls; discriminate.
Qed.


End WithTypingMode.
