Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath LogpathOrder.

Require Import TypeInduction TypLogicInduction.
Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic LayeredCapabilityMain.
Require Import LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure LayeredTypingLogic LayeredTypingMain LayeredTypingLogicOrdered
               LayeredSubtypingMainTree LayeredSubtypingMainOrd LayeredSubtypingMainPrep.

Require Import Lia.

(* This file contains "hard" properties about subtyping on the main layer, involving the various alternative definitions. *)

Section WithTypingMode.
Context {typing_mode: Mode}.

Section WithContext.
Context (G: ctx) (e: renv) (x: var).
Context (Hin: inert G).


(* Analog of ty_inv_logic_ord_logic *)

Lemma ty_has_N_main_sub_ord_main n ocl U T:
  ty_has_N_main G e U ->
  ty_sub_main_ord G e x n ocl U T ->
  ty_has_N_main G e T.
Proof.
  introv Hty Hh.
  reverts T U.
  induction n using nat_lt_induction; intros.
  induction Hh.
  - eauto.
  - eapply ty_has_N_main_name; eauto.
    eapply IHHh.
    + intros. eapply H. 2: eauto.
    2: eapply ty_sub_main_ord_le. 2: eauto. 2: eauto.  eapply Nat.lt_lt_succ_r; eauto.
    +
    eauto.
  - eapply ty_has_N_main_bnd. eapply ty_has_N_main_replace_nocap. eauto.
    eapply IHHh.
    + intros. eapply H. 3: eauto. 2: eauto.
      eapply Nat.lt_lt_succ_r; eauto. eapply Plus.lt_plus_trans. eauto.
    + eauto.
  - eapply ty_has_N_main_bnd; eauto.  eapply IHHh. intros. eapply H. 2: eauto. 2: eauto.
    2: eapply ty_sub_main_ord_le. 2: eauto. 2: eauto.  eapply Nat.lt_lt_succ_r; eauto.
    eauto.
Qed.

End WithContext.


Lemma ty_main_tree_sub_ord_cap G e x tr T U:
  inert G ->
  renv_corr G e ->
  ty_main_tree_sub G e x tr T U ->
  ty_main_tree_bndfolds tr = 0 ->
  ty_sub_main_ord_cap G e x (ty_main_tree_folds tr) logctx_nil T U.
Proof.
  introv Hin Hcor Htr Hf.
  remember (ty_main_tree_folds tr).
  reverts tr T U.

  induction n using nat_lt_0S_induction; intros; subst.
  - forwards Heq: ty_main_tree_sub_folds_0_inv; eauto.
    subst. eapply ty_sub_main_ord_cap_refl.
  -
    induction Htr; intros; simpls.
    + eapply Nat.lt_irrefl in H. contradiction.
    +
      eapply Plus.plus_is_O in Hf. destructecs Hf. 
      forwards Hadd: Nat.add_pos_cases H. destructecs Hadd.
      * forwards HH0: H0 trr  Tr.  2: auto.
        2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. eauto.
        forwards: IHHtr1 .
        intros; subst.  3: eauto.  eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r. auto.
        eapply ty_sub_main_ord_cap_le.
        2: eapply ty_sub_main_ord_cap_or; eauto. constructor.
      * forwards HH0: H0 trl  Tl. 2: auto.
        2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. eauto.

        forwards: IHHtr2 .
        intros; subst.  3: eauto.  eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l. auto.
        eapply ty_sub_main_ord_cap_le.
        2: eapply ty_sub_main_ord_cap_or; eauto. constructor.
    +
      eapply Plus.plus_is_O in Hf. destructecs Hf. 
      forwards Hadd: Nat.add_pos_cases H. destructecs Hadd.
      * forwards HH0: H0 trr (*logctx_extend oc (logctx_orr Tl logctx_nil)*)  Tr. 2: auto.
        2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. eauto.
          forwards: IHHtr1 .
        intros; subst.  3: eauto.  eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r. auto.
        eapply ty_sub_main_ord_cap_le.
        2: eapply ty_sub_main_ord_cap_and; eauto. constructor.
      * forwards HH0: H0 trl (*logctx_extend oc (logctx_orr Tl logctx_nil)*)  Tl. 2: auto.
        2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. eauto.
          forwards: IHHtr2 .
        intros; subst.  3: eauto.  eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l. auto.
        eapply ty_sub_main_ord_cap_le.
        2: eapply ty_sub_main_ord_cap_and; eauto. constructor.
    + inversions Hf.

    +
      replace   (typ_bnd T) with (logctx_typ logctx_nil   (typ_bnd T)).
      2: simpl; auto.
      eapply ty_sub_main_ord_cap_bnd; simpls; eauto. constructor. constructor.
    +
      replace  (typ_sel (avar_f y) A (avar_f z)) with (logctx_typ logctx_nil  (typ_sel (avar_f y) A (avar_f z))).
      2: simpl; auto.
      eapply ty_sub_main_ord_cap_sel; simpls; eauto. constructor. 
Qed.



Lemma ty_main_tree_sub_ord G e x tr T U:
  inert G ->
  renv_corr G e ->
  ty_main_tree_sub G e x tr T U ->
  ty_sub_main_ord G e x (ty_main_tree_folds tr) logctx_nil T U.
Proof.
  introv Hin Hcor Htr.
  remember (ty_main_tree_folds tr).
  reverts tr T U.

  induction n using nat_lt_0S_induction; intros; subst.
  - forwards Heq: ty_main_tree_sub_folds_0_inv; eauto.
    subst. eapply ty_sub_main_ord_refl.
  -
    induction Htr; intros; simpls.
    + eapply Nat.lt_irrefl in H. contradiction.
    + forwards Hadd: Nat.add_pos_cases H. destructecs Hadd.
      * forwards HH0: H0 trr (*logctx_extend oc (logctx_orr Tl logctx_nil)*)  Tr.
        2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. eauto.

        forwards: IHHtr1 .
        intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r.
        eapply ty_sub_main_ord_le.
        2: eapply ty_sub_main_ord_or; eauto. constructor.
    * forwards HH0: H0 trl (*logctx_extend oc (logctx_orr Tl logctx_nil)*)  Tl.
        2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. eauto.

        forwards: IHHtr2 .
        intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l.
        eapply ty_sub_main_ord_le.
        2: eapply ty_sub_main_ord_or; eauto. constructor.
    +
      forwards Hadd: Nat.add_pos_cases H. destructecs Hadd.
      * forwards HH0: H0 trr (*logctx_extend oc (logctx_orr Tl logctx_nil)*)  Tr.
        2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. eauto.

        forwards: IHHtr1 .
        intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r.
        eapply ty_sub_main_ord_le.
        2: eapply ty_sub_main_ord_and; eauto. constructor.
      * forwards HH0: H0 trl (*logctx_extend oc (logctx_orr Tl logctx_nil)*)  Tl.
        2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. eauto.

        forwards: IHHtr2 .
        intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l.
        eapply ty_sub_main_ord_le.
        2: eapply ty_sub_main_ord_and; eauto. constructor.
    +
      destruct (mode_has_incap' typing_mode ) eqn: Hm.
      **
        forwards: H1. eauto. destructecs H3; subst. 

        replace   (typ_bnd U) with (logctx_typ logctx_nil   (typ_bnd U)).
        2: simpl; auto.
        destruct (ty_main_tree_folds tr) eqn: He.
        *
          eapply ty_sub_main_ord_bnd; simpls; eauto. constructor. (* intros; splits. 2: eapply ty_main_tree_sub_ord_cap; eauto. eauto. eauto. constructor. *)
        *
          eapply ty_sub_main_ord_bnd; simpls; eauto. 
          eapply IHHtr; intros. eapply ty_sub_main_ord_le; eauto. constructor. eapply H0. 2: eauto. 2: eauto. eapply Nat.lt_lt_succ_r. rewrite <- Nat.add_succ_l. eapply Nat.lt_lt_add_r. eauto.  eapply Nat.lt_0_succ.
          intros. splits; eauto.
          applys H0. 2: eauto. eapply Nat.lt_lt_succ_r. rewrite <- Nat.add_succ_r.   eapply Nat.lt_lt_add_l. eapply Nat.lt_succ_diag_r. auto. constructor.
      **
        replace   (typ_bnd U) with (logctx_typ logctx_nil   (typ_bnd U)).
        2: simpl; auto.
        destruct (ty_main_tree_folds tr) eqn: He.
        *
          eapply ty_sub_main_ord_bnd; simpls; eauto. intros. rewrite Hm in H3; discriminate. constructor.
        *
          eapply ty_sub_main_ord_bnd; simpls; eauto. 
          eapply IHHtr; intros. eapply ty_sub_main_ord_le; eauto. constructor. eapply H0. 2: eauto. 2: eauto. eapply Nat.lt_lt_succ_r. rewrite <- Nat.add_succ_l. eapply Nat.lt_lt_add_r. eauto.  eapply Nat.lt_0_succ.
          intros. rewrite Hm in H3; discriminate. constructor.
    +
      replace   (typ_bnd T) with (logctx_typ logctx_nil   (typ_bnd T)).
      2: simpl; auto.
      eapply ty_sub_main_ord_nocap_bnd; simpls; eauto. constructor. constructor.
    +
      replace  (typ_sel (avar_f y) A (avar_f z)) with (logctx_typ logctx_nil  (typ_sel (avar_f y) A (avar_f z))).
      2: simpl; auto.
      eapply ty_sub_main_ord_sel; simpls; eauto. constructor. 
  Unshelve. all: eauto.
Qed.


Lemma ty_sub_main_ord_cap_tree G e x n oc T U:
  inert G ->
  renv_corr G e ->
  ty_sub_main_ord_cap G e x n oc T U ->
  exists tr,
    ty_main_tree_sub G e x tr T U /\
    (ty_main_tree_folds tr) = n /\
    (ty_main_tree_bndfolds tr) = 0.
Proof.
  introv Hin Hcor Hty.

  reverts T U oc.
  induction n using nat_lt_induction; intros; subst.
  inversions Hty.
  -
    eexists. splits.
    eapply ty_main_tree_sub_refl. eapply ty_main_tree_refl_folds. eapply ty_main_tree_refl_bndfolds.
  -
    forwards IHHty: H. 2: eapply H0. eapply Nat.lt_succ_diag_r.
    destructecs IHHty; subst.

    forwards Hoc: ty_main_tree_sub_lc_inv  IHHty0.
    destructecs Hoc; subst.

    forwards HtrS1: ty_main_tree_sub_refl (open z S1).
    forwards: ty_main_tree_sub_lc.
    eapply ty_main_tree_sub_sel; eauto.
    forwards Htrt: ty_main_tree_sub_trans IHHty0.
    eauto.
    eexists. splits. eauto.
    erewrite ty_main_tree_trans_folds_eq by eauto.
    rewrite ty_main_tree_lc_folds.
    simpl. rewrite ty_main_tree_refl_folds. rewrite Nat.add_1_r. auto.
    erewrite ty_main_tree_trans_bndfolds_eq by eauto.
    rewrite IHHty.
    rewrite ty_main_tree_lc_bndfolds.
    simpl. rewrite ty_main_tree_refl_bndfolds. auto.


  -
    (* Main predecessor *)
    forwards IHHty: H. 2: eapply H0. eapply Nat.lt_succ_diag_r.
    destructecs IHHty; subst.

    forwards Hoc: ty_main_tree_sub_lc_inv  IHHty0.
    destructecs Hoc; subst.

    forwards HtrS1: ty_main_tree_sub_refl G e x (open x T0).
    forwards: ty_main_tree_sub_lc.
    eapply ty_main_tree_sub_nocap_bnd; eauto.
    forwards Htrt: ty_main_tree_sub_trans IHHty0. eauto.

    eexists. splits. eauto.
    erewrite ty_main_tree_trans_folds_eq by eauto.
    rewrite ty_main_tree_lc_folds.
    simpl.
    rewrite Nat.add_1_r. auto.

    erewrite ty_main_tree_trans_bndfolds_eq by eauto.
    rewrite IHHty.
    rewrite ty_main_tree_lc_bndfolds.
    simpl. auto.
Qed.

Lemma ty_sub_main_ord_tree G e x n oc T U:
  inert G ->
  renv_corr G e ->
  ty_sub_main_ord G e x n oc T U ->
  exists tr,
    ty_main_tree_sub G e x tr T U /\
    (ty_main_tree_folds tr) <= n.
Proof.
  introv Hin Hcor Hty.

  reverts T U oc.
  induction n using nat_lt_induction; intros; subst.
  inversions Hty.
  -
    eexists. split.
    eapply ty_main_tree_sub_refl. eapply Nat.eq_le_incl. eapply ty_main_tree_refl_folds.
  -
    forwards IHHty: H. 2: eapply H0. eapply Nat.lt_succ_diag_r.
    destructecs IHHty; subst.

    forwards Hoc: ty_main_tree_sub_lc_inv  IHHty0.
    destructecs Hoc; subst.

    forwards HtrS1: ty_main_tree_sub_refl (open z S1).
    forwards: ty_main_tree_sub_lc.
    eapply ty_main_tree_sub_sel; eauto.
    forwards Htrt: ty_main_tree_sub_trans IHHty0.
    eauto.
    eexists. split. eauto.

    erewrite ty_main_tree_trans_folds_eq by eauto.
    rewrite ty_main_tree_lc_folds.
    simpl. rewrite ty_main_tree_refl_folds. rewrite Nat.add_1_r. lia.
  -
    (* Main predecessor *)
    forwards IHHty: H. 2: eapply H0. rewrite <- Nat.add_succ_l. eapply Nat.lt_lt_add_r. eapply Nat.lt_succ_diag_r.
    destructecs IHHty; subst.

    destruct (mode_has_incap' typing_mode ) eqn: Hm.
    **

      (* Secondary predecessor *)
      forwards HH: H1. auto.
      destructecs HH; subst.

      forwards IHHh: H HH. rewrite <- Nat.add_succ_r. eapply Nat.lt_lt_add_l. eapply Nat.lt_succ_diag_r.
      destructecs IHHh; subst.

      forwards Hoc: ty_main_tree_sub_lc_inv  IHHty0.
      destructecs Hoc; subst.

      forwards HtrS1: ty_main_tree_sub_refl (open x T0).
      forwards: ty_main_tree_sub_lc.
      eapply ty_main_tree_sub_bnd; eauto. 

      forwards Htrt: ty_main_tree_sub_trans IHHty0. eauto.
      eexists. split. eauto.
      erewrite ty_main_tree_trans_folds_eq by eauto.
      rewrite ty_main_tree_lc_folds.
      simpl. rewrite ty_main_tree_refl_folds.
      rewrite Nat.add_succ_r. lia.
    **
      forwards Hoc: ty_main_tree_sub_lc_inv  IHHty0.
      destructecs Hoc; subst.

      forwards HtrS1: ty_main_tree_sub_refl (open x T0).
      forwards: ty_main_tree_sub_lc.
      eapply ty_main_tree_sub_bnd with (trn:= (ty_main_tree_refl (open x T0))); eauto.
      intros . rewrite Hm in H4; discriminate.

      forwards Htrt: ty_main_tree_sub_trans IHHty0. eauto.
      eexists. split. eauto.
      erewrite ty_main_tree_trans_folds_eq by eauto.
      rewrite ty_main_tree_lc_folds.
      simpl. rewrite ty_main_tree_refl_folds.
      rewrite Nat.add_succ_r. lia.
  -
    (* Main predecessor *)
    forwards IHHty: H. 2: eapply H0. eapply Nat.lt_succ_diag_r.
    destructecs IHHty; subst.

    forwards Hoc: ty_main_tree_sub_lc_inv  IHHty0.
    destructecs Hoc; subst.

    forwards HtrS1: ty_main_tree_sub_refl G e x (open x T0).
    forwards: ty_main_tree_sub_lc.
    eapply ty_main_tree_sub_nocap_bnd; eauto.
    forwards Htrt: ty_main_tree_sub_trans IHHty0. eauto.

    eexists. split. eauto.
    erewrite ty_main_tree_trans_folds_eq by eauto.
    rewrite ty_main_tree_lc_folds.
    simpl.
    rewrite Nat.add_1_r. lia.
  Unshelve.
    eauto.
Qed.

End WithTypingMode.
