Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath LogpathOrder.

Require Import TypeInduction TypLogicInduction.
Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic LayeredCapabilityMain.
Require Import LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure LayeredTypingLogic LayeredTypingMain.

Section WithTypingMode.
Context {typing_mode: Mode}.

(* Ordered derivation on the main layer *)
Inductive ty_sub_main_ord_cap (G: ctx) (e: renv) (x: var): nat -> logctx -> typ -> typ -> Prop :=
| ty_sub_main_ord_cap_refl ocl T:
    ty_sub_main_ord_cap G e x 0 ocl T T
(* Replace lower bound by selection *)
| ty_sub_main_ord_cap_sel n ocl T ocp S1 S2 U y A z:
    ty_sub_main_ord_cap G e x n ocp T (logctx_typ ocp (open z S1)) ->
    G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
    logctx_le ocl ocp ->
    ty_sub_main_ord_cap G e x (S n) ocl T (logctx_typ ocp (typ_sel (avar_f y) A (avar_f z)))
| ty_sub_main_ord_cap_bnd n ocl ocp T U:
    ty_sub_main_ord_cap G e x n ocp U (logctx_typ ocp typ_nocap) ->
    logctx_le ocl ocp ->
    ty_sub_main_ord_cap G e x (S n) ocl U (logctx_typ ocp (typ_bnd T))
.

Inductive ty_sub_main_ord (G: ctx) (e: renv) (x: var): nat -> logctx -> typ -> typ -> Prop :=
| ty_sub_main_ord_refl ocl T:
    ty_sub_main_ord G e x 0 ocl T T
(* Replace lower bound by selection *)
| ty_sub_main_ord_sel n ocl T ocp S1 S2 U y A z:
    ty_sub_main_ord G e x n ocp T (logctx_typ ocp (open z S1)) ->
    G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
    logctx_le ocl ocp ->
    ty_sub_main_ord G e x (S n) ocl T (logctx_typ ocp (typ_sel (avar_f y) A (avar_f z)))
| ty_sub_main_ord_bnd n1 n2 ocl ocp T Tn U:
    ty_sub_main_ord G e x n1 ocl U (logctx_typ ocp (open x T)) ->
    ((mode_has_incap typing_mode) -> (ty_has_N_logic G e Tn /\ ty_sub_main_ord G e x n2 logctx_nil Tn (open x T))) ->
    typ_indep_b 0 T ->
    logctx_le ocl ocp ->
    ty_sub_main_ord G e x (S (n1 + n2)) ocl U (logctx_typ ocp (typ_bnd T))
| ty_sub_main_ord_nocap_bnd n ocl ocp T U:
    ty_sub_main_ord G e x n ocp U (logctx_typ ocp typ_nocap) ->
    logctx_le ocl ocp ->
    ty_sub_main_ord G e x (S n) ocl U (logctx_typ ocp (typ_bnd T))
.

Section WithContext.
Context (G: ctx) (e: renv) (x: var).
Context (Hin: inert G).

Lemma ty_sub_main_ord_cap_0_inv ocl T U:
  ty_sub_main_ord_cap G e x 0 ocl T U ->
  T = U.
Proof.
  introv Ht.
  remember 0.
  induction Ht; auto; discriminate.
Qed.

(* Analog of ty_inv_logic_ord_le *)
Lemma ty_sub_main_ord_cap_le n op1 op2 U T:
  logctx_le op2 op1 ->
  ty_sub_main_ord_cap G e x n op1 U T ->
  ty_sub_main_ord_cap G e x n op2 U T.
Proof.
  introv Hle Hh.
  induction Hh.
  - eapply ty_sub_main_ord_cap_refl; eauto.
  - eapply ty_sub_main_ord_cap_sel; eauto.
    unfold orctx_le in *. eapply logctx_le_trans; eauto.
  - eapply ty_sub_main_ord_cap_bnd; eauto.
    unfold orctx_le in *. eapply logctx_le_trans; eauto.
Qed.


Lemma ty_sub_main_ord_cap_lc n ocl oc U T:
  ty_sub_main_ord_cap G e x n ocl U T ->
  ty_sub_main_ord_cap G e x n (logctx_extend oc ocl) (logctx_typ oc U) (logctx_typ oc T).
Proof.
  introv Htr.
  reverts oc.
  induction Htr; intros.
  - eapply ty_sub_main_ord_cap_refl; eauto.
  - rewrite <- logctx_extend_typ in *.
    eapply ty_sub_main_ord_cap_sel; eauto.
    rewrite logctx_extend_typ. eauto.
  - rewrite <- logctx_extend_typ in *.
    eapply ty_sub_main_ord_cap_bnd; eauto.
    rewrite logctx_extend_typ. eauto.
Qed.



Lemma ty_sub_main_ord_cap_or nr Tr Ur nl ocl ocr Tl Ul:
  ty_sub_main_ord_cap G e x nl ocl Tl Ul ->
  ty_sub_main_ord_cap G e x nr ocr Tr Ur ->
  ty_sub_main_ord_cap G e x (nl + nr) (logctx_orl ocl Ur) (typ_or Tl Tr) (typ_or Ul Ur).
Proof.
  introv Htrl Htrr.
  induction Htrl; subst.
  - simpl.
    forwards: ty_sub_main_ord_cap_lc (logctx_orr T logctx_nil) Htrr .
    eapply ty_sub_main_ord_cap_le; eauto.
    simpl in H. constructor.
  - rewrite logctx_fold_orl in *.
    rewrite Nat.add_succ_l.
    eapply ty_sub_main_ord_cap_sel.
    2: eauto. auto. hnf; simpl. constructor. auto.
  - rewrite logctx_fold_orl in *.
    rewrite Nat.add_succ_l.
    eapply ty_sub_main_ord_cap_bnd.
    auto. hnf; simpl. constructor. auto.
Qed.

Lemma ty_sub_main_ord_cap_and nr Tr Ur nl ocl ocr Tl Ul:
  ty_sub_main_ord_cap G e x nl ocl Tl Ul ->
  ty_sub_main_ord_cap G e x nr ocr Tr Ur ->
  ty_sub_main_ord_cap G e x (nl + nr) (logctx_andl ocl Ur) (typ_and Tl Tr) (typ_and Ul Ur).
Proof.
  introv Htrl Htrr.
  induction Htrl; subst.
  - simpl.
    forwards: ty_sub_main_ord_cap_lc (logctx_andr T logctx_nil) Htrr .
    eapply ty_sub_main_ord_cap_le; eauto.
    simpl in H. constructor.
  - rewrite logctx_fold_andl in *.
    rewrite Nat.add_succ_l.
    eapply ty_sub_main_ord_cap_sel.
    2: eauto. auto. hnf; simpl. constructor. auto.
  - rewrite logctx_fold_andl in *.
    rewrite Nat.add_succ_l.
    eapply ty_sub_main_ord_cap_bnd.
    auto. hnf; simpl. constructor. auto.
Qed.



Lemma ty_sub_main_ord_0_inv ocl T U:
  ty_sub_main_ord G e x 0 ocl T U ->
  T = U.
Proof.
  introv Ht.
  remember 0.
  induction Ht; auto; discriminate.
Qed.

(* Analog of ty_inv_logic_ord_le *)
Lemma ty_sub_main_ord_le n op1 op2 U T:
  logctx_le op2 op1 ->
  ty_sub_main_ord G e x n op1 U T ->
  ty_sub_main_ord G e x n op2 U T.
Proof.
  introv Hle Hh.
  induction Hh.
  - eapply ty_sub_main_ord_refl; eauto.
  - eapply ty_sub_main_ord_sel; eauto.
    unfold orctx_le in *. eapply logctx_le_trans; eauto.
  - eapply ty_sub_main_ord_bnd; eauto.
    unfold orctx_le in *. eapply logctx_le_trans; eauto.
  - eapply ty_sub_main_ord_nocap_bnd; eauto.
    unfold orctx_le in *. eapply logctx_le_trans; eauto.
Qed.


Lemma ty_sub_main_ord_lc n ocl oc U T:
  ty_sub_main_ord G e x n ocl U T ->
  ty_sub_main_ord G e x n (logctx_extend oc ocl) (logctx_typ oc U) (logctx_typ oc T).
Proof.
  introv Htr.
  reverts oc.
  induction Htr; intros.
  - eapply ty_sub_main_ord_refl; eauto.
  - rewrite <- logctx_extend_typ in *.
    eapply ty_sub_main_ord_sel; eauto.
    rewrite logctx_extend_typ. eauto.
  - rewrite <- logctx_extend_typ in *.
    eapply ty_sub_main_ord_bnd; eauto.
    rewrite logctx_extend_typ. eauto.
  - rewrite <- logctx_extend_typ in *.
    eapply ty_sub_main_ord_nocap_bnd; eauto.
    rewrite logctx_extend_typ. eauto.
Qed.



Lemma ty_sub_main_ord_or nr Tr Ur nl ocl ocr Tl Ul:
  ty_sub_main_ord G e x nl ocl Tl Ul ->
  ty_sub_main_ord G e x nr ocr Tr Ur ->
  ty_sub_main_ord G e x (nl + nr) (logctx_orl ocl Ur) (typ_or Tl Tr) (typ_or Ul Ur).
Proof.
  introv Htrl Htrr.
  induction Htrl; subst.
  - simpl.
    forwards: ty_sub_main_ord_lc (logctx_orr T logctx_nil) Htrr .
    eapply ty_sub_main_ord_le; eauto.
    simpl in H. constructor.
  - rewrite logctx_fold_orl in *.
    rewrite Nat.add_succ_l.
    eapply ty_sub_main_ord_sel.
    2: eauto. auto. hnf; simpl. constructor. auto.
  - rewrite logctx_fold_orl in *.
    rewrite Nat.add_succ_l.
    rewrite <- Nat.add_assoc.
    rewrite <- (Nat.add_comm nr).
    rewrite Nat.add_assoc.

    eapply ty_sub_main_ord_bnd. eauto. eauto. eauto.
    hnf; simpl. constructor. auto.
  - rewrite logctx_fold_orl in *.
    rewrite Nat.add_succ_l.
    eapply ty_sub_main_ord_nocap_bnd.
    auto. hnf; simpl. constructor. auto.
Qed.

Lemma ty_sub_main_ord_and nr Tr Ur nl ocl ocr Tl Ul:
  ty_sub_main_ord G e x nl ocl Tl Ul ->
  ty_sub_main_ord G e x nr ocr Tr Ur ->
  ty_sub_main_ord G e x (nl + nr) (logctx_andl ocl Ur) (typ_and Tl Tr) (typ_and Ul Ur).
Proof.
  introv Htrl Htrr.
  induction Htrl; subst.
  - simpl.
    forwards: ty_sub_main_ord_lc (logctx_andr T logctx_nil) Htrr .
    eapply ty_sub_main_ord_le; eauto.
    simpl in H. constructor.
  - rewrite logctx_fold_andl in *.
    rewrite Nat.add_succ_l.
    eapply ty_sub_main_ord_sel.
    2: eauto. auto. hnf; simpl. constructor. auto.
  - rewrite logctx_fold_andl in *.
    rewrite Nat.add_succ_l.
    rewrite <- Nat.add_assoc.
    rewrite <- (Nat.add_comm nr).
    rewrite Nat.add_assoc.

    eapply ty_sub_main_ord_bnd. eauto. eauto. eauto.
    hnf; simpl. constructor. auto.
  - rewrite logctx_fold_andl in *.
    rewrite Nat.add_succ_l.
    eapply ty_sub_main_ord_nocap_bnd.
    auto. hnf; simpl. constructor. auto.
Qed.

Lemma ty_sub_main_ord_cap_ord n ocl T U:
  ty_sub_main_ord_cap G e x n ocl T U ->
  ty_sub_main_ord G e x n ocl T U.
Proof.
  introv Hs.
  induction Hs.
  - eapply ty_sub_main_ord_refl.
  - eapply ty_sub_main_ord_sel; eauto.
  - eapply ty_sub_main_ord_nocap_bnd; eauto.
Qed.

End WithContext.

Lemma ty_has_N_main_main_ord_cap G e x n T U:
  inert G ->
  renv_corr G e ->
  ty_sub_main_ord_cap G e x n logctx_nil T U ->
  ty_has_N_main G e T ->
  ty_has_N_main G e U.
Proof.
  introv Hin Hcor Hty Hh.
  induction Hty.
  - auto.
  - eapply ty_has_N_main_name; eauto.
  - eapply ty_has_N_main_bnd; eauto.
Qed.

Lemma ty_has_N_main_main_ord G e x n T U:
  inert G ->
  renv_corr G e ->
  ty_sub_main_ord G e x n logctx_nil T U ->
  ty_has_N_main G e T ->
  ty_has_N_main G e U.
Proof.
  introv Hin Hcor Hty Hh.
  induction Hty.
  - auto.
  - eapply ty_has_N_main_name; eauto.
  - eapply ty_has_N_main_bnd.
    eapply ty_has_N_main_replace_nocap; eauto.
  - eapply ty_has_N_main_bnd; eauto.
Qed.

End WithTypingMode.