Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath LogpathOrder OrctxLogctx.

Require Import TypeInduction TypLogicInduction TypeBndSel.

Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic.
Require Import LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure LayeredTypingLogic LayeredTypingMain.


Section WithTypingMode.
Context {typing_mode: Mode}.

(* Derivation at the last typing layer according to a tree structure *)

(* The tree structure that guides the derivation. *)
Inductive ty_main_tree: Set :=
| ty_main_tree_atomic (T: typ): ty_main_tree
| ty_main_tree_or (trl trr: ty_main_tree): ty_main_tree
| ty_main_tree_and (trl trr: ty_main_tree): ty_main_tree
| ty_main_tree_bnd (tr trn: ty_main_tree): ty_main_tree
| ty_main_tree_nocap_bnd (T: typ): ty_main_tree
| ty_main_tree_sel (tr: ty_main_tree) (y: var) (A: typ_label) (z: var): ty_main_tree
.

(* Subtyping guided by ty_main_tree *)
Inductive ty_main_tree_sub (G: ctx) (e: renv) (x: var): ty_main_tree -> typ -> typ -> Prop :=
| ty_main_tree_sub_atomic T:
  typ_notlog T ->
  ty_main_tree_sub G e x (ty_main_tree_atomic T) T T

| ty_main_tree_sub_or trl trr Tl Tr Ul Ur:
  ty_main_tree_sub G e x trl Tl Ul ->
  ty_main_tree_sub G e x trr Tr Ur ->
  ty_main_tree_sub G e x (ty_main_tree_or trl trr) (typ_or Tl Tr) (typ_or Ul Ur)

| ty_main_tree_sub_and trl trr Tl Tr Ul Ur:
  ty_main_tree_sub G e x trl Tl Ul ->
  ty_main_tree_sub G e x trr Tr Ur ->
  ty_main_tree_sub G e x (ty_main_tree_and trl trr) (typ_and Tl Tr) (typ_and Ul Ur)

| ty_main_tree_sub_bnd tr T U trn Un:
  ty_main_tree_sub G e x tr T (open x U) ->
  ((mode_has_incap typing_mode) ->
    (ty_has_N_logic G e Un /\ ty_main_tree_sub G e x trn Un (open x U) ) ) ->
  typ_indep_b 0 U ->
  ty_main_tree_sub G e x (ty_main_tree_bnd tr trn) T (typ_bnd U)

| ty_main_tree_sub_nocap_bnd T:
  ty_main_tree_sub G e x (ty_main_tree_nocap_bnd T) typ_nocap (typ_bnd T)

| ty_main_tree_sub_sel tr T y A z U S1 S2:
  ty_main_tree_sub G e x tr T (open z S1) ->
  G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
  ty_main_tree_sub G e x (ty_main_tree_sel tr y A z) T  (typ_sel (avar_f y) A (avar_f z))
.
Hint Constructors ty_main_tree_sub: core.

(* Version where bnd proof uses the same tree *)
Inductive ty_main_tree_linear_sub (G: ctx) (e: renv) (x: var): ty_main_tree -> typ -> typ -> Prop :=
| ty_main_tree_linear_sub_atomic T:
  typ_notlog T ->
  ty_main_tree_linear_sub G e x (ty_main_tree_atomic T) T T

| ty_main_tree_linear_sub_or trl trr Tl Tr Ul Ur:
  ty_main_tree_linear_sub G e x trl Tl Ul ->
  ty_main_tree_linear_sub G e x trr Tr Ur ->
  ty_main_tree_linear_sub G e x (ty_main_tree_or trl trr) (typ_or Tl Tr) (typ_or Ul Ur)

| ty_main_tree_linear_sub_and trl trr Tl Tr Ul Ur:
  ty_main_tree_linear_sub G e x trl Tl Ul ->
  ty_main_tree_linear_sub G e x trr Tr Ur ->
  ty_main_tree_linear_sub G e x (ty_main_tree_and trl trr) (typ_and Tl Tr) (typ_and Ul Ur)

| ty_main_tree_linear_sub_bnd tr T U:
  ty_main_tree_linear_sub G e x tr T (open x U) ->
  ((mode_has_incap typing_mode) -> ty_has_N_logic G e T) ->
  typ_indep_b 0 U ->
  ty_main_tree_linear_sub G e x (ty_main_tree_bnd tr (ty_main_tree_atomic typ_top)) T (typ_bnd U)

| ty_main_tree_linear_sub_nocap_bnd T:
  ty_main_tree_linear_sub G e x (ty_main_tree_nocap_bnd T) typ_nocap (typ_bnd T)

| ty_main_tree_linear_sub_sel tr T y A z U S1 S2:
  ty_main_tree_linear_sub G e x tr T (open z S1) ->
  G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
  ty_main_tree_linear_sub G e x (ty_main_tree_sel tr y A z) T  (typ_sel (avar_f y) A (avar_f z))
.
Hint Constructors ty_main_tree_linear_sub: core.

(* Number of non-trivial steps in the derivation. *)
Fixpoint ty_main_tree_folds (tr: ty_main_tree): nat :=
match tr with
| ty_main_tree_atomic T => 0
| ty_main_tree_or trl trr => (ty_main_tree_folds trl) + (ty_main_tree_folds trr)
| ty_main_tree_and trl trr => (ty_main_tree_folds trl) + (ty_main_tree_folds trr)
| ty_main_tree_bnd tr trn => S ((ty_main_tree_folds tr) + (ty_main_tree_folds trn))
| ty_main_tree_nocap_bnd T => 1
| ty_main_tree_sel tr y A z => S (ty_main_tree_folds tr)
end.

(* Number of ty_main_tree_bnd steps in the derivation. *)
Fixpoint ty_main_tree_bndfolds (tr: ty_main_tree): nat :=
match tr with
| ty_main_tree_atomic T => 0
| ty_main_tree_or trl trr => (ty_main_tree_bndfolds trl) + (ty_main_tree_bndfolds trr)
| ty_main_tree_and trl trr => (ty_main_tree_bndfolds trl) + (ty_main_tree_bndfolds trr)
| ty_main_tree_bnd tr trn => S (ty_main_tree_bndfolds tr)
| ty_main_tree_nocap_bnd T => 0
| ty_main_tree_sel tr y A z => (ty_main_tree_bndfolds tr)
end.


(* Finds a subtree at a specific path. When the path cannot continue, returns the current tree *)
Fixpoint ty_main_tree_under_lc (tr: ty_main_tree) oc {struct oc}: ty_main_tree :=
match oc, tr with
| logpath_nil, _ => tr
| logpath_orl ocp, ty_main_tree_or trl trr => ty_main_tree_under_lc trl ocp
| logpath_orl ocp, _ => tr
| logpath_orr ocp, ty_main_tree_or trl trr => ty_main_tree_under_lc trr ocp
| logpath_orr ocp, _ => tr
| logpath_andl ocp, ty_main_tree_and trl trr => ty_main_tree_under_lc trl ocp
| logpath_andl ocp, _ => tr
| logpath_andr ocp, ty_main_tree_and trl trr => ty_main_tree_under_lc trr ocp
| logpath_andr ocp, _ => tr
end.
Fixpoint ty_main_tree_set_under_lc (tr tr2: ty_main_tree) oc {struct oc}: ty_main_tree :=
match oc, tr with
| logpath_nil, _ => tr2
| logpath_orl ocp, ty_main_tree_or trl trr => ty_main_tree_or (ty_main_tree_set_under_lc trl tr2 ocp) trr
| logpath_orl ocp, _ => tr2
| logpath_orr ocp, ty_main_tree_or trl trr => ty_main_tree_or trl (ty_main_tree_set_under_lc trr tr2 ocp)
| logpath_orr ocp, _ => tr2
| logpath_andl ocp, ty_main_tree_and trl trr => ty_main_tree_and (ty_main_tree_set_under_lc trl tr2 ocp) trr
| logpath_andl ocp, _ => tr2
| logpath_andr ocp, ty_main_tree_and trl trr => ty_main_tree_and trl (ty_main_tree_set_under_lc trr tr2 ocp)
| logpath_andr ocp, _ => tr2
end.

Fixpoint ty_main_tree_refl T: ty_main_tree :=
match T with
| typ_and Tl Tr => ty_main_tree_and (ty_main_tree_refl Tl) (ty_main_tree_refl Tr)
| typ_or Tl Tr => ty_main_tree_or (ty_main_tree_refl Tl) (ty_main_tree_refl Tr)
| _ => (ty_main_tree_atomic T)
end.


Fixpoint ty_main_tree_extend_lc (tr: ty_main_tree) oc {struct oc}: ty_main_tree :=
match oc with
| logctx_nil => tr
| logctx_orl ocp Tr => ty_main_tree_or (ty_main_tree_extend_lc tr ocp) (ty_main_tree_refl Tr)
| logctx_orr Tl ocp => ty_main_tree_or (ty_main_tree_refl Tl) (ty_main_tree_extend_lc tr ocp) 
| logctx_andl ocp Tr => ty_main_tree_and (ty_main_tree_extend_lc tr ocp) (ty_main_tree_refl Tr)
| logctx_andr Tl ocp => ty_main_tree_and (ty_main_tree_refl Tl) (ty_main_tree_extend_lc tr ocp) 
end.

Fixpoint ty_main_tree_trans (tr1 tr2: ty_main_tree) {struct tr2} := 
  match tr2, tr1 with
  | ty_main_tree_atomic tr2l, _ => tr1
  | ty_main_tree_or tr2l tr2r, ty_main_tree_or tr1l tr1r => ty_main_tree_or (ty_main_tree_trans tr1l tr2l) (ty_main_tree_trans tr1r tr2r)
  | ty_main_tree_and tr2l tr2r, ty_main_tree_and tr1l tr1r => ty_main_tree_and (ty_main_tree_trans tr1l tr2l) (ty_main_tree_trans tr1r tr2r)
  | ty_main_tree_bnd tr2l T2, _ => ty_main_tree_bnd (ty_main_tree_trans tr1 tr2l) T2
  | ty_main_tree_sel tr2l y2 A2 z2, _ => ty_main_tree_sel (ty_main_tree_trans tr1 tr2l) y2 A2 z2
  | ty_main_tree_nocap_bnd T2, _ => ty_main_tree_nocap_bnd T2
  | _,_ => tr1 (* Misaligned case *)
  end.



Section WithContext.
Context (G: ctx) (e: renv) (x: var).
Context (Hin: inert G).



Lemma ty_main_tree_sub_or_inv tr T Ul Ur:
  ty_main_tree_sub G e x tr T (typ_or Ul Ur) ->
  exists trr trl Tr Tl,
      T = (typ_or Tl Tr) /\ 
      tr = (ty_main_tree_or trl trr) /\
      ty_main_tree_sub G e x trl Tl Ul /\
      ty_main_tree_sub G e x trr Tr Ur.
Proof.
  introv Htr.
  inversions Htr.
  - hnf in H; destructc H as Ho Ha. false Ho; eauto.
  - eauto 8.
Qed.
Lemma ty_main_tree_sub_and_inv tr T Ul Ur:
  ty_main_tree_sub G e x tr T (typ_and Ul Ur) ->
  exists trr trl Tr Tl,
      T = (typ_and Tl Tr) /\ 
      tr = (ty_main_tree_and trl trr) /\
      ty_main_tree_sub G e x trl Tl Ul /\
      ty_main_tree_sub G e x trr Tr Ur.
Proof.
  introv Htr.
  inversions Htr.
  - hnf in H; destructc H as Ho Ha. false Ha; eauto.
  - eauto 8.
Qed.
Lemma ty_main_tree_sub_nologbndsel_inv tr T U:
  typ_notlog U ->
  typ_notbndsel U ->
  ty_main_tree_sub G e x tr T U ->
  T = U /\ tr = ty_main_tree_atomic U.
Proof.
  introv Hnl Hnb Htr.
  hnf in Hnl, Hnb; destruct_all.
  inversions Htr.
  - eauto.
  - false H1; eauto.
  - false H2; eauto.
  - false H0 (logctx_nil); simpl; eauto.
  - false H0 (logctx_nil); simpl; eauto.
  - false H (logctx_nil); simpl; eauto.
Qed.


(* Any path in the output is also in the input *)

Lemma ty_main_tree_sub_lc_inv tr oc U T:
  ty_main_tree_sub G e x tr U (logctx_typ oc T) ->
  exists ocU Uoc,
    U = logctx_typ ocU Uoc /\
    ty_main_tree_sub G e x (ty_main_tree_under_lc tr (logctx_path oc)) Uoc T /\
    logctx_follow oc ocU.
Proof.
  introv Htr.
  reverts tr U.
  induction oc; intros; simpls.
  - exists logctx_nil U; simpls; eauto.
  - eapply ty_main_tree_sub_or_inv in Htr; destructecs Htr; subst.
    forwards Hoc: IHoc; eauto.
    destructecs Hoc; subst.
    exists (logctx_orl ocU Tr) Uoc; splits; simpls; eauto. 
  - eapply ty_main_tree_sub_or_inv in Htr; destructecs Htr; subst.
    forwards Hoc: IHoc; eauto.
    destructecs Hoc; subst.
    exists (logctx_orr Tl ocU) Uoc; splits; simpls; eauto.
  - eapply ty_main_tree_sub_and_inv in Htr; destructecs Htr; subst.
    forwards Hoc: IHoc; eauto.
    destructecs Hoc; subst.
    exists (logctx_andl ocU Tr) Uoc; splits; simpls; eauto. 
  - eapply ty_main_tree_sub_and_inv in Htr; destructecs Htr; subst.
    forwards Hoc: IHoc; eauto.
    destructecs Hoc; subst.
    exists (logctx_andr Tl ocU) Uoc; splits; simpls; eauto.
Qed.

Lemma ty_main_tree_sub_oc_inv tr oc U T:
  ty_main_tree_sub G e x tr  U (orctx_typ oc T) ->
  exists ocU Uoc,
    U = orctx_typ ocU Uoc /\
    ty_main_tree_sub  G e x (ty_main_tree_under_lc tr (orctx_path oc)) Uoc T /\
    orctx_follow oc ocU.
Proof.
  introv Hty.
  rewrite orctx_logctx_typ in Hty.
  eapply ty_main_tree_sub_lc_inv in Hty.
  destruct_all.
  forwards: logctx_follows_isor (orctx_logctx oc); eauto. 
  destruct H2; subst.
  exists x2 x1.
  splits.
  rewrite <- orctx_logctx_typ in *. auto.
  auto.
  eapply logctx_follow_orctx; eauto.
Qed.

Lemma ty_main_tree_sub_lc_follow_replace tr ocT ocU U T:
  logctx_follow ocT ocU ->
  ty_main_tree_sub G e x tr (logctx_typ ocT T) (logctx_typ ocU U) ->
  forall tr2 (T2 U2: typ),
  ty_main_tree_sub G e x tr2 T2 U2 ->
  ty_main_tree_sub G e x (ty_main_tree_set_under_lc tr tr2 (logctx_path ocT)) (logctx_typ ocT T2) (logctx_typ ocU U2).

Proof.
  introv Hf Htyoc Hty.
  reverts tr T U ocT.
  induction ocU; intros; inversions Hf; simpls.
  - auto.
  - eapply ty_main_tree_sub_or_inv in Htyoc; destructecs Htyoc; subst.
    inversions Htyoc0. eauto.
  - eapply ty_main_tree_sub_or_inv in Htyoc; destructecs Htyoc; subst.
    inversions Htyoc0. eauto.
  - eapply ty_main_tree_sub_and_inv in Htyoc; destructecs Htyoc; subst.
    inversions Htyoc0. eauto.
  - eapply ty_main_tree_sub_and_inv in Htyoc; destructecs Htyoc; subst.
    inversions Htyoc0. eauto.
Qed.

Lemma ty_main_tree_sub_refl T:
  (ty_main_tree_sub G e x (ty_main_tree_refl T) T T).
Proof.
  induction T using typ_logic_ind; simpls; eauto.
  destruct T; simpls; eauto.
  - false (proj22 H); eauto.
  - false (proj21 H); eauto.
Qed.


Lemma ty_main_tree_refl_folds T:
  (ty_main_tree_folds (ty_main_tree_refl T)  = 0).
Proof.
  induction T using typ_logic_ind; simpls; eauto.
  +
  destruct T; simpls; auto.
  - false (proj22 H); eauto.
  - false (proj21 H); eauto.
  + rewrite IHT1, IHT2. auto.
  + rewrite IHT1, IHT2. auto.
Qed.
Lemma ty_main_tree_refl_bndfolds T:
  (ty_main_tree_bndfolds (ty_main_tree_refl T)  = 0).
Proof.
  induction T using typ_logic_ind; simpls; eauto.
  +
  destruct T; simpls; auto.
  - false (proj22 H); eauto.
  - false (proj21 H); eauto.
  + rewrite IHT1, IHT2. auto.
  + rewrite IHT1, IHT2. auto.
Qed.


Lemma ty_main_tree_sub_lc tr oc T U:
  ty_main_tree_sub G e x tr T U ->
  ty_main_tree_sub G e x (ty_main_tree_extend_lc tr oc) (logctx_typ oc T) (logctx_typ oc U).
Proof.
  introv Htr.
  induction oc; intros; simpls.
  - auto.
  - eapply ty_main_tree_sub_or; eauto using ty_main_tree_sub_refl.
  - eapply ty_main_tree_sub_or; eauto using ty_main_tree_sub_refl.
  - eapply ty_main_tree_sub_and; eauto using ty_main_tree_sub_refl.
  - eapply ty_main_tree_sub_and; eauto using ty_main_tree_sub_refl.
Qed.

Lemma ty_main_tree_lc_folds tr oc:
  (ty_main_tree_folds (ty_main_tree_extend_lc tr oc)) = (ty_main_tree_folds tr).
Proof.
  induction oc; simpls; rewrite ? ty_main_tree_refl_folds, ? IHoc; auto.
Qed.
Lemma ty_main_tree_lc_bndfolds tr oc:
  (ty_main_tree_bndfolds (ty_main_tree_extend_lc tr oc)) = (ty_main_tree_bndfolds tr).
Proof.
  induction oc; simpls; rewrite ? ty_main_tree_refl_bndfolds, ? IHoc; auto.
Qed.

Lemma ty_main_tree_sub_oc_follow_replace tr ocT ocU U T:
  orctx_follow ocT ocU ->
  ty_main_tree_sub G e x tr (orctx_typ ocT T) (orctx_typ ocU U) ->
  forall tr2 (T2 U2: typ),
  ty_main_tree_sub G e x tr2 T2 U2 ->
  ty_main_tree_sub G e x  (ty_main_tree_set_under_lc tr tr2 (orctx_path ocT)) (orctx_typ ocT T2) (orctx_typ ocU U2).
Proof.
  introv Hf Htyoc Hty.
  eapply orctx_follow_logctx in Hf.
  rewrite ! orctx_logctx_typ in Htyoc.
  forwards: ty_main_tree_sub_lc_follow_replace; eauto.
  rewrite <- ! orctx_logctx_typ in H.
  eauto.
Qed.

Lemma ty_main_tree_sub_trans tr1 tr2 T1 T2 T3:
  ty_main_tree_sub G e x  tr1 T1 T2 ->
  ty_main_tree_sub G e x  tr2 T2 T3 ->
  ty_main_tree_sub G e x  (ty_main_tree_trans tr1 tr2) T1 T3.
Proof.
  introv Htr1 Htr2.
  reverts tr1 T1.
  induction Htr2; intros; simpls.
  - eauto.
  - eapply ty_main_tree_sub_or_inv in Htr1; destruct_all; subst.
    eauto.
  - eapply ty_main_tree_sub_and_inv in Htr1; destruct_all; subst.
    eauto.
  - inversions Htr1; forwards: IHHtr2; eauto.
  - inversions Htr1. eauto.
  - inversions Htr1; forwards: IHHtr2; eauto.
Qed.

Lemma ty_main_tree_trans_folds tr1 tr2:
  ((ty_main_tree_folds (ty_main_tree_trans tr1 tr2)) <= ty_main_tree_folds tr1 + ty_main_tree_folds tr2).
Proof.
  reverts tr1.
  induction tr2; intros; simpls.
  - rewrite Nat.add_0_r. auto.
  - destruct tr1; simpls.
    + eapply Nat.le_0_l.
    + rewrite Nat.add_shuffle1.
      eapply Nat.add_le_mono; auto.
    + eapply Plus.le_plus_trans. auto.
    + apply -> Nat.succ_le_mono. eapply Plus.le_plus_trans. auto.
    + apply -> Nat.succ_le_mono. eapply Nat.le_0_l.
    + apply -> Nat.succ_le_mono. eapply Plus.le_plus_trans. auto.
  - destruct tr1; simpls.
    + eapply Nat.le_0_l.
    + eapply Plus.le_plus_trans. auto.
    + rewrite Nat.add_shuffle1.
      eapply Nat.add_le_mono; auto.
    + apply -> Nat.succ_le_mono. eapply Plus.le_plus_trans. auto.
    + apply -> Nat.succ_le_mono. eapply Nat.le_0_l.
    + apply -> Nat.succ_le_mono. eapply Plus.le_plus_trans. auto.
  - rewrite Nat.add_succ_r.
    apply -> Nat.succ_le_mono.
    rewrite Nat.add_assoc.
    apply -> Nat.add_le_mono_r.
    auto.
  - rewrite Nat.add_succ_r.
    apply -> Nat.succ_le_mono.
    eapply Nat.le_0_l.
  - rewrite Nat.add_succ_r.
    apply -> Nat.succ_le_mono.
    auto.
Qed.

(* Version of the above, but additionally claim equiality about the complexity of the tree *)
Lemma ty_main_tree_trans_folds_eq tr1 tr2 T1 T2 T3:
  ty_main_tree_sub G e x tr1 T1 T2 ->
  ty_main_tree_sub G e x tr2 T2 T3 ->
  ((ty_main_tree_folds (ty_main_tree_trans tr1 tr2)) = ty_main_tree_folds tr1 + ty_main_tree_folds tr2).
Proof.
  introv Htr1 Htr2.
  reverts tr1 T1.
  induction Htr2; intros.
  - eauto.
  - eapply ty_main_tree_sub_or_inv in Htr1; destruct_all; subst.
    forwards HHtr2_1: IHHtr2_1; eauto.
    forwards HHtr2_2: IHHtr2_2; eauto.
    simpls. rewrite  HHtr2_1,  HHtr2_2. 
    rewrite Nat.add_assoc.
    rewrite Nat.add_assoc.
    rewrite <- (Nat.add_assoc (ty_main_tree_folds x1 )).
    rewrite <- (Nat.add_assoc (ty_main_tree_folds x1 ) (ty_main_tree_folds x0)).
    rewrite (Nat.add_comm (ty_main_tree_folds trl )). auto.
  - eapply ty_main_tree_sub_and_inv in Htr1; destruct_all; subst.
    forwards  HHtr2_1: IHHtr2_1; eauto.
    forwards HHtr2_2: IHHtr2_2; eauto.

    simpls. rewrite  HHtr2_1,  HHtr2_2. 
    rewrite Nat.add_assoc.
    rewrite Nat.add_assoc.
    rewrite <- (Nat.add_assoc (ty_main_tree_folds x1 )).
    rewrite <- (Nat.add_assoc (ty_main_tree_folds x1 ) (ty_main_tree_folds x0)).
    rewrite (Nat.add_comm (ty_main_tree_folds trl )). auto.
  -
    forwards HHtr2: IHHtr2; eauto.

    simpls. rewrite HHtr2.
    rewrite Nat.add_succ_r.
    rewrite Nat.add_assoc. auto.
  - 
    inversions Htr1.
    eauto.
  - forwards HHtr2: IHHtr2; eauto.
    simpls. rewrite HHtr2.
    rewrite Nat.add_succ_r.
    auto.
Qed.


Lemma ty_main_tree_trans_bndfolds tr1 tr2:
  ((ty_main_tree_bndfolds (ty_main_tree_trans tr1 tr2)) <= ty_main_tree_bndfolds tr1 + ty_main_tree_bndfolds tr2).
Proof.
  reverts tr1.
  induction tr2; intros; simpls.
  - rewrite Nat.add_0_r. auto.
  - destruct tr1; simpls.
    + eapply Nat.le_0_l.
    + rewrite Nat.add_shuffle1.
      eapply Nat.add_le_mono; auto.
    + eapply Plus.le_plus_trans. auto.
    + apply -> Nat.succ_le_mono. eapply Plus.le_plus_trans. auto.
    +  eapply Nat.le_0_l.
    +  eapply Plus.le_plus_trans. auto.
  - destruct tr1; simpls.
    + eapply Nat.le_0_l.
    + eapply Plus.le_plus_trans. auto.
    + rewrite Nat.add_shuffle1.
      eapply Nat.add_le_mono; auto.
    + apply -> Nat.succ_le_mono. eapply Plus.le_plus_trans. auto.
    +  eapply Nat.le_0_l.
    + eapply Plus.le_plus_trans. auto.
  - rewrite Nat.add_succ_r.
    apply -> Nat.succ_le_mono.
    auto.
  - 
    eapply Nat.le_0_l.
  - 
    auto.
Qed.

(* Version of the above, but additionally claim equiality about the complexity of the tree *)
Lemma ty_main_tree_trans_bndfolds_eq tr1 tr2 T1 T2 T3:
  ty_main_tree_sub G e x tr1 T1 T2 ->
  ty_main_tree_sub G e x tr2 T2 T3 ->
  ((ty_main_tree_bndfolds (ty_main_tree_trans tr1 tr2)) = ty_main_tree_bndfolds tr1 + ty_main_tree_bndfolds tr2).
Proof.
  introv Htr1 Htr2.
  reverts tr1 T1.
  induction Htr2; intros.
  - eauto.
  - eapply ty_main_tree_sub_or_inv in Htr1; destruct_all; subst.
    forwards HHtr2_1: IHHtr2_1; eauto.
    forwards HHtr2_2: IHHtr2_2; eauto.
    simpls. rewrite  HHtr2_1,  HHtr2_2. 
    rewrite Nat.add_assoc.
    rewrite Nat.add_assoc.
    rewrite <- (Nat.add_assoc (ty_main_tree_bndfolds x1 )).
    rewrite <- (Nat.add_assoc (ty_main_tree_bndfolds x1 ) (ty_main_tree_bndfolds x0)).
    rewrite (Nat.add_comm (ty_main_tree_bndfolds trl )). auto.
  - eapply ty_main_tree_sub_and_inv in Htr1; destruct_all; subst.
    forwards  HHtr2_1: IHHtr2_1; eauto.
    forwards HHtr2_2: IHHtr2_2; eauto.

    simpls. rewrite  HHtr2_1,  HHtr2_2. 
    rewrite Nat.add_assoc.
    rewrite Nat.add_assoc.
    rewrite <- (Nat.add_assoc (ty_main_tree_bndfolds x1 )).
    rewrite <- (Nat.add_assoc (ty_main_tree_bndfolds x1 ) (ty_main_tree_bndfolds x0)).
    rewrite (Nat.add_comm (ty_main_tree_bndfolds trl )). auto.
  -
    forwards HHtr2: IHHtr2; eauto.

    simpls. rewrite HHtr2.
    rewrite Nat.add_succ_r.
     auto.
  - 
    inversions Htr1.
    eauto.
  - forwards HHtr2: IHHtr2; eauto.
Qed.



Lemma ty_main_tree_set_under_lc_folds tr tr2 op:
  (ty_main_tree_folds (ty_main_tree_set_under_lc tr tr2 op)) + (ty_main_tree_folds (ty_main_tree_under_lc tr op)) = (ty_main_tree_folds tr2) + (ty_main_tree_folds tr).
Proof.
  reverts tr tr2.
  induction op; intros; simpls; auto; destruct tr; simpls; eauto.
  - rewrite ! Nat.add_assoc. rewrite Nat.add_shuffle0. rewrite IHop. auto.
  - rewrite ! Nat.add_assoc. rewrite <- Nat.add_assoc. rewrite IHop. rewrite Nat.add_shuffle0. rewrite Nat.add_comm. auto.
  - rewrite ! Nat.add_assoc. rewrite Nat.add_shuffle0. rewrite IHop. auto.
  - rewrite ! Nat.add_assoc. rewrite <- Nat.add_assoc. rewrite IHop. rewrite Nat.add_shuffle0. rewrite Nat.add_comm. auto.
Qed.

Lemma ty_main_tree_set_under_lc_bndfolds tr tr2 op:
  (ty_main_tree_bndfolds (ty_main_tree_set_under_lc tr tr2 op)) + (ty_main_tree_bndfolds (ty_main_tree_under_lc tr op)) = (ty_main_tree_bndfolds tr2) + (ty_main_tree_bndfolds tr).
Proof.
  reverts tr tr2.
  induction op; intros; simpls; auto; destruct tr; simpls; eauto.
  - rewrite ! Nat.add_assoc. rewrite Nat.add_shuffle0. rewrite IHop. auto.
  - rewrite ! Nat.add_assoc. rewrite <- Nat.add_assoc. rewrite IHop. rewrite Nat.add_shuffle0. rewrite Nat.add_comm. auto.
  - rewrite ! Nat.add_assoc. rewrite Nat.add_shuffle0. rewrite IHop. auto.
  - rewrite ! Nat.add_assoc. rewrite <- Nat.add_assoc. rewrite IHop. rewrite Nat.add_shuffle0. rewrite Nat.add_comm. auto.
Qed.

Lemma ty_main_tree_set_under_lc_bndfolds_le tr tr2 op:
  (ty_main_tree_bndfolds (ty_main_tree_set_under_lc tr tr2 op) ) <= (ty_main_tree_bndfolds tr) + (ty_main_tree_bndfolds tr2).
Proof.
  apply <- Nat.add_le_mono_r.
  rewrite ty_main_tree_set_under_lc_bndfolds.
  rewrite Nat.add_comm.
  rewrite <- Nat.add_assoc.
  apply Nat.add_le_mono_l.
  apply Nat.le_add_r.
Qed.

Lemma ty_main_tree_sub_folds_0_inv tr T U:
  ty_main_tree_sub G e x tr T U ->
  ty_main_tree_folds tr = 0 ->
  U = T.
Proof.
  introv Htr Hf.
  induction Htr; simpls.
  - auto.
  - eapply Plus.plus_is_O in Hf. destructecs Hf.
    rewrite IHHtr1, IHHtr2; auto.
  - eapply Plus.plus_is_O in Hf. destructecs Hf.
    rewrite IHHtr1, IHHtr2; auto.
  - discriminate.
  - discriminate.
  - discriminate.
Qed.

Lemma ty_has_N_atomic_main_tree tr T U:
  ty_main_tree_sub G e x tr T U ->
  ty_has_N_atomic G e U ->
  ty_has_N_atomic G e T.
Proof.
  introv Hty Hh.
  induction Hh; inversions Hty; eauto.
Qed.
Lemma ty_has_top_atomic_main_tree tr T U:
  ty_main_tree_sub G e x tr T U ->
  ty_has_top_atomic G e U ->
  ty_has_top_atomic G e T.
Proof.
  introv Hty Hh.
  induction Hh; inversions Hty; eauto.
Qed.



Lemma ty_has_N_main_tree tr T U:
  ty_main_tree_sub G e x tr T U ->
  ty_has_N G e U ->
  ty_has_N G e T.
Proof.
  introv Hty Hh.
  inversions Hh.
  eapply ty_has_N_has_atomic. eapply ty_has_N_atomic_main_tree; eauto.
  eapply ty_has_N_has_top. eapply ty_has_top_atomic_main_tree; eauto.
Qed.

Lemma ty_has_N_closure_main_tree tr T U:
  ty_main_tree_sub G e x tr T U ->
  ty_has_N_closure G e U ->
  ty_has_N_closure G e T.
Proof.
  introv Hty Hh.
  reverts T tr.
  induction Hh; intros.
  - eapply ty_or_closure_atomic. eapply ty_has_N_main_tree; eauto. 
  - inversions Hty.
    + hnf in H; destruct_all. false H; eauto.
    + eapply ty_or_closure_l. eapply IHHh; eauto.
  - inversions Hty.
    + hnf in H; destruct_all. false H; eauto.
    + eapply ty_or_closure_r. eapply IHHh; eauto.
  - inversions Hty.
    eapply ty_or_closure_top; eauto.
Qed.

Lemma ty_has_N_logic_main_tree tr T U:
  ty_main_tree_sub G e x tr T U ->
  ty_has_N_logic G e U ->
  ty_has_N_logic G e T.
Proof.
  introv Hty Hh.
  reverts T tr.
  induction Hh; intros.
  - eapply ty_has_N_logic_or. eapply ty_has_N_closure_main_tree; eauto.
  - pose proof Hty as Hty'. 
    eapply ty_main_tree_sub_oc_inv in Hty; destruct_all; subst.
    eapply ty_main_tree_sub_and_inv in H0; destruct_all; subst.
    eapply orctx_follow_symm in H1.
    forwards: ty_main_tree_sub_oc_follow_replace Hty' H2; eauto.
    forwards: ty_main_tree_sub_oc_follow_replace Hty' H3; eauto. destruct_all.
    eapply ty_has_N_logic_oc_and; eauto.
Qed.


Lemma ty_main_tree_set_sel_folds tr op y A z:
  ty_main_tree_folds (ty_main_tree_set_under_lc tr (ty_main_tree_sel (ty_main_tree_under_lc tr op) y A z) op) = 
  S (ty_main_tree_folds tr).
Proof.
  reverts tr.
  induction op; intros; destruct tr; simpl; auto.
  -
  rewrite <- Nat.add_succ_l.
  apply <- Nat.add_cancel_r. auto.
  -
  rewrite <- Nat.add_succ_r.
  apply <- Nat.add_cancel_l. auto.
  -
  rewrite <- Nat.add_succ_l.
  apply <- Nat.add_cancel_r. auto.
  -
  rewrite <- Nat.add_succ_r.
  apply <- Nat.add_cancel_l. auto.
Qed.
Lemma ty_main_tree_set_sel_bndfolds tr op y A z:
  ty_main_tree_bndfolds (ty_main_tree_set_under_lc tr (ty_main_tree_sel (ty_main_tree_under_lc tr op) y A z) op) = 
  (ty_main_tree_bndfolds tr).
Proof.
  reverts tr.
  induction op; intros; destruct tr; simpl; auto.
Qed.

Lemma ty_main_tree_set_bnd_folds tr op trn:
  ty_main_tree_folds (ty_main_tree_set_under_lc tr (ty_main_tree_bnd (ty_main_tree_under_lc tr op) trn) op) = 
  S (ty_main_tree_folds tr + ty_main_tree_folds trn).
Proof.
  reverts tr.
  induction op; intros; destruct tr; simpl; auto.
  -
  rewrite <- Nat.add_succ_l.
  rewrite <- Nat.add_succ_l.
  rewrite <- Nat.add_assoc.
  rewrite <- (Nat.add_comm (ty_main_tree_folds trn)).
  rewrite Nat.add_assoc.
  apply <- Nat.add_cancel_r.
  rewrite Nat.add_succ_l. auto.
  -
  rewrite <- Nat.add_succ_r.
  rewrite <- Nat.add_assoc.
  apply <- Nat.add_cancel_l.
  rewrite Nat.add_succ_r. auto.
  -
  rewrite <- Nat.add_succ_l.
  rewrite <- Nat.add_succ_l.
  rewrite <- Nat.add_assoc.
  rewrite <- (Nat.add_comm (ty_main_tree_folds trn)).
  rewrite Nat.add_assoc.
  apply <- Nat.add_cancel_r.
  rewrite Nat.add_succ_l. auto.
  -
  rewrite <- Nat.add_succ_r.
  rewrite <- Nat.add_assoc.
  apply <- Nat.add_cancel_l.
  rewrite Nat.add_succ_r. auto.
Qed.


Lemma ty_main_tree_set_bnd_bndfolds tr op trn:
  ty_main_tree_bndfolds (ty_main_tree_set_under_lc tr (ty_main_tree_bnd (ty_main_tree_under_lc tr op) trn) op) = 
  S (ty_main_tree_bndfolds tr).
Proof.
  reverts tr.
  induction op; intros; destruct tr; simpl; auto.
  -
  rewrite <- Nat.add_succ_l.
  apply <- Nat.add_cancel_r.
  auto.
  -
  rewrite <- Nat.add_succ_r.
  apply <- Nat.add_cancel_l.
  auto.
  -
  rewrite <- Nat.add_succ_l.
  apply <- Nat.add_cancel_r.
  auto.
  -
  rewrite <- Nat.add_succ_r.
  apply <- Nat.add_cancel_l.
  auto.
Qed.

Lemma ty_main_tree_under_lc_bndfolds tr op: 
  ty_main_tree_bndfolds tr = 0 ->
  ty_main_tree_bndfolds (ty_main_tree_under_lc tr op) = 0.
Proof.
  introv Hf.
  reverts tr.
  induction op; intros; subst; simpls; destruct tr; eauto.
  all: inversions Hf; rewrite H0; eapply IHop; apply -> Nat.eq_add_0 in H0; destruct H0; auto.
Qed.
Lemma ty_main_tree_set_under_lc_bndfolds_0 tr trn op: 
  ty_main_tree_bndfolds tr = 0 ->
  ty_main_tree_bndfolds trn = 0 ->
  ty_main_tree_bndfolds (ty_main_tree_set_under_lc tr trn op) = 0.
Proof.
  introv Hf.
  reverts tr.
  induction op; intros; subst; simpls; destruct tr; eauto.
  all: inversions Hf; rewrite H1; simpls; rewrite IHop; apply -> Nat.eq_add_0 in H1; destruct H1; rewrite ? Nat.add_0_r; auto.
Qed.



Lemma ty_main_tree_linear_sub_or_inv tr T Ul Ur:
  ty_main_tree_linear_sub G e x tr T (typ_or Ul Ur) ->
  exists trr trl Tr Tl,
      T = (typ_or Tl Tr) /\ 
      tr = (ty_main_tree_or trl trr) /\
      ty_main_tree_linear_sub G e x trl Tl Ul /\
      ty_main_tree_linear_sub G e x trr Tr Ur.
Proof.
  introv Htr.
  inversions Htr.
  - hnf in H; destructc H as Ho Ha. false Ho; eauto.
  - eauto 8.
Qed.
Lemma ty_main_tree_linear_sub_and_inv tr T Ul Ur:
  ty_main_tree_linear_sub G e x tr T (typ_and Ul Ur) ->
  exists trr trl Tr Tl,
      T = (typ_and Tl Tr) /\ 
      tr = (ty_main_tree_and trl trr) /\
      ty_main_tree_linear_sub G e x trl Tl Ul /\
      ty_main_tree_linear_sub G e x trr Tr Ur.
Proof.
  introv Htr.
  inversions Htr.
  - hnf in H; destructc H as Ho Ha. false Ha; eauto.
  - eauto 8.
Qed.
Lemma ty_main_tree_linear_sub_nologbndsel_inv tr T U:
  typ_notlog U ->
  typ_notbndsel U ->
  ty_main_tree_linear_sub G e x tr T U ->
  T = U /\ tr = ty_main_tree_atomic U.
Proof.
  introv Hnl Hnb Htr.
  hnf in Hnl, Hnb; destruct_all.
  inversions Htr.
  - eauto.
  - false H1; eauto.
  - false H2; eauto.
  - false H0 (logctx_nil); simpl; eauto.
  - false H0 (logctx_nil); simpl; eauto.
  - false H (logctx_nil); simpl; eauto.
Qed.


Lemma ty_main_tree_linear_sub_lc_inv tr oc U T:
  ty_main_tree_linear_sub G e x tr U (logctx_typ oc T) ->
  exists ocU Uoc,
    U = logctx_typ ocU Uoc /\
    ty_main_tree_linear_sub G e x (ty_main_tree_under_lc tr (logctx_path oc)) Uoc T /\
    logctx_follow oc ocU.
Proof.
  introv Htr.
  reverts tr U.
  induction oc; intros; simpls.
  - exists logctx_nil U; simpls; eauto.
  - eapply ty_main_tree_linear_sub_or_inv in Htr; destructecs Htr; subst.
    forwards Hoc: IHoc; eauto.
    destructecs Hoc; subst.
    exists (logctx_orl ocU Tr) Uoc; splits; simpls; eauto. 
  - eapply ty_main_tree_linear_sub_or_inv in Htr; destructecs Htr; subst.
    forwards Hoc: IHoc; eauto.
    destructecs Hoc; subst.
    exists (logctx_orr Tl ocU) Uoc; splits; simpls; eauto.
  - eapply ty_main_tree_linear_sub_and_inv in Htr; destructecs Htr; subst.
    forwards Hoc: IHoc; eauto.
    destructecs Hoc; subst.
    exists (logctx_andl ocU Tr) Uoc; splits; simpls; eauto. 
  - eapply ty_main_tree_linear_sub_and_inv in Htr; destructecs Htr; subst.
    forwards Hoc: IHoc; eauto.
    destructecs Hoc; subst.
    exists (logctx_andr Tl ocU) Uoc; splits; simpls; eauto.
Qed.

Lemma ty_main_tree_linear_sub_oc_inv tr oc U T:
  ty_main_tree_linear_sub G e x tr  U (orctx_typ oc T) ->
  exists ocU Uoc,
    U = orctx_typ ocU Uoc /\
    ty_main_tree_linear_sub G e x  (ty_main_tree_under_lc tr (orctx_path oc)) Uoc T /\
    orctx_follow oc ocU.
Proof.
  introv Hty.
  rewrite orctx_logctx_typ in Hty.
  eapply ty_main_tree_linear_sub_lc_inv in Hty.
  destruct_all.
  forwards: logctx_follows_isor (orctx_logctx oc); eauto. 
  destruct H2; subst.
  exists x2 x1.
  splits.
  rewrite <- orctx_logctx_typ in *. auto.
  auto.
  eapply logctx_follow_orctx; eauto.
Qed.

Lemma ty_main_tree_linear_sub_lc_follow_replace tr ocT ocU U T:
  logctx_follow ocT ocU ->
  ty_main_tree_linear_sub G e x tr (logctx_typ ocT T) (logctx_typ ocU U) ->
  forall tr2 (T2 U2: typ),
  ty_main_tree_linear_sub G e x tr2 T2 U2 ->
  ty_main_tree_linear_sub G e x (ty_main_tree_set_under_lc tr tr2 (logctx_path ocT)) (logctx_typ ocT T2) (logctx_typ ocU U2).

Proof.
  introv Hf Htyoc Hty.
  reverts tr T U ocT.
  induction ocU; intros; inversions Hf; simpls.
  - auto.
  - eapply ty_main_tree_linear_sub_or_inv in Htyoc; destructecs Htyoc; subst.
    inversions Htyoc0. eauto.
  - eapply ty_main_tree_linear_sub_or_inv in Htyoc; destructecs Htyoc; subst.
    inversions Htyoc0. eauto.
  - eapply ty_main_tree_linear_sub_and_inv in Htyoc; destructecs Htyoc; subst.
    inversions Htyoc0. eauto.
  - eapply ty_main_tree_linear_sub_and_inv in Htyoc; destructecs Htyoc; subst.
    inversions Htyoc0. eauto.
Qed.

Lemma ty_main_tree_linear_sub_refl T:
  ty_main_tree_linear_sub G e x (ty_main_tree_refl T) T T.
Proof.
  induction T using typ_logic_ind; simpls; eauto.
  destruct T; simpls; eauto.
  - false (proj22 H); eauto.
  - false (proj21 H); eauto.
Qed.


Lemma ty_main_tree_linear_sub_folds_0_inv tr T U:
  ty_main_tree_linear_sub G e x tr T U ->
  ty_main_tree_folds tr = 0 ->
  U = T.
Proof.
  introv Htr Hf.
  induction Htr; simpls.
  - auto.
  - eapply Plus.plus_is_O in Hf. destructecs Hf.
    rewrite IHHtr1, IHHtr2; auto.
  - eapply Plus.plus_is_O in Hf. destructecs Hf.
    rewrite IHHtr1, IHHtr2; auto.
  - discriminate.
  - discriminate.
  - discriminate.
Qed.

Lemma ty_has_N_atomic_main_tree_linear tr T U:
  ty_main_tree_linear_sub G e x tr T U ->
  ty_has_N_atomic G e U ->
  ty_has_N_atomic G e T.
Proof.
  introv Hty Hh.
  induction Hh; inversions Hty; eauto.
Qed.

Lemma ty_has_top_atomic_main_tree_linear tr T U:
  ty_main_tree_linear_sub G e x tr T U ->
  ty_has_top_atomic G e U ->
  ty_has_top_atomic G e T.
Proof.
  introv Hty Hh.
  induction Hh; inversions Hty; eauto.
Qed.

Lemma ty_has_N_main_tree_linear tr T U:
  ty_main_tree_linear_sub G e x tr T U ->
  ty_has_N G e U ->
  ty_has_N G e T.
Proof.
  introv Hty Hh.
  inversions Hh.
  eapply ty_has_N_has_atomic. eapply ty_has_N_atomic_main_tree_linear; eauto.
  eapply ty_has_N_has_top. eapply ty_has_top_atomic_main_tree_linear; eauto.
Qed.

Lemma ty_has_N_closure_main_tree_linear tr T U:
  ty_main_tree_linear_sub G e x tr T U ->
  ty_has_N_closure G e U ->
  ty_has_N_closure G e T.
Proof.
  introv Hty Hh.
  reverts T tr.
  induction Hh; intros.
  - eapply ty_or_closure_atomic. eapply ty_has_N_main_tree_linear; eauto. 
  - inversions Hty.
    + hnf in H; destruct_all. false H; eauto.
    + eapply ty_or_closure_l. eapply IHHh; eauto.
  - inversions Hty.
    + hnf in H; destruct_all. false H; eauto.
    + eapply ty_or_closure_r. eapply IHHh; eauto.
  - inversions Hty.
    eapply ty_or_closure_top; eauto.
Qed.


Lemma ty_main_tree_linear_sub_oc_follow_replace tr ocT ocU U T:
  orctx_follow ocT ocU ->
  ty_main_tree_linear_sub G e x tr (orctx_typ ocT T) (orctx_typ ocU U) ->
  forall tr2 (T2 U2: typ),
  ty_main_tree_linear_sub G e x tr2 T2 U2 ->
  ty_main_tree_linear_sub G e x  (ty_main_tree_set_under_lc tr tr2 (orctx_path ocT)) (orctx_typ ocT T2) (orctx_typ ocU U2).
Proof.
  introv Hf Htyoc Hty.
  eapply orctx_follow_logctx in Hf.
  rewrite ! orctx_logctx_typ in Htyoc.
  forwards: ty_main_tree_linear_sub_lc_follow_replace; eauto.
  rewrite <- ! orctx_logctx_typ in H.
  eauto.
Qed.

Lemma ty_has_N_logic_main_tree_linear tr T U:
  ty_main_tree_linear_sub G e x tr T U ->
  ty_has_N_logic G e U ->
  ty_has_N_logic G e T.
Proof.
  introv Hty Hh.
  reverts T tr.
  induction Hh; intros.
  - eapply ty_has_N_logic_or. eapply ty_has_N_closure_main_tree_linear; eauto.
  - pose proof Hty as Hty'. 
    eapply ty_main_tree_linear_sub_oc_inv in Hty; destruct_all; subst.
    eapply ty_main_tree_linear_sub_and_inv in H0; destruct_all; subst.
    eapply orctx_follow_symm in H1.
    forwards: ty_main_tree_linear_sub_oc_follow_replace Hty' H2; eauto.
    forwards: ty_main_tree_linear_sub_oc_follow_replace Hty' H3; eauto. destruct_all.
    eapply ty_has_N_logic_oc_and; eauto.
Qed.
Lemma ty_main_tree_linear_sub_trans tr1 tr2 T1 T2 T3:
  ty_main_tree_linear_sub G e x tr1 T1 T2 ->
  ty_main_tree_linear_sub G e x tr2 T2 T3 ->
  ty_main_tree_linear_sub G e x (ty_main_tree_trans tr1 tr2) T1 T3.
Proof.
  introv Htr1 Htr2.
  reverts tr1 T1.
  induction Htr2; intros; simpls.
  - eauto.
  - eapply ty_main_tree_linear_sub_or_inv in Htr1; destruct_all; subst.
    eauto.
  - eapply ty_main_tree_linear_sub_and_inv in Htr1; destruct_all; subst.
    eauto.
  - forwards: IHHtr2; eauto. econstructor. eauto. intros. eapply ty_has_N_logic_main_tree_linear. eapply Htr1. eauto. eauto.
  - inversions Htr1. eauto.
  - inversions Htr1; forwards: IHHtr2; eauto.
Qed.

Lemma ty_main_tree_linear_sub_lc tr oc T U:
  ty_main_tree_linear_sub G e x tr T U ->
  ty_main_tree_linear_sub G e x (ty_main_tree_extend_lc tr oc) (logctx_typ oc T) (logctx_typ oc U).
Proof.
  introv Htr.
  induction oc; intros; simpls.
  - auto.
  - eapply ty_main_tree_linear_sub_or; eauto using ty_main_tree_linear_sub_refl.
  - eapply ty_main_tree_linear_sub_or; eauto using ty_main_tree_linear_sub_refl.
  - eapply ty_main_tree_linear_sub_and; eauto using ty_main_tree_linear_sub_refl.
  - eapply ty_main_tree_linear_sub_and; eauto using ty_main_tree_linear_sub_refl.
Qed.

Lemma ty_main_tree_linear_trans_folds_eq tr1 tr2 T1 T2 T3:
  ty_main_tree_linear_sub G e x tr1 T1 T2 ->
  ty_main_tree_linear_sub G e x tr2 T2 T3 ->
  ((ty_main_tree_folds (ty_main_tree_trans tr1 tr2)) = ty_main_tree_folds tr1 + ty_main_tree_folds tr2).
Proof.
  introv Htr1 Htr2.
  reverts tr1 T1.
  induction Htr2; intros.
  - eauto.
  - eapply ty_main_tree_linear_sub_or_inv in Htr1; destruct_all; subst.
    forwards HHtr2_1: IHHtr2_1; eauto.
    forwards HHtr2_2: IHHtr2_2; eauto.
    simpls. rewrite  HHtr2_1,  HHtr2_2. 
    rewrite Nat.add_assoc.
    rewrite Nat.add_assoc.
    rewrite <- (Nat.add_assoc (ty_main_tree_folds x1 )).
    rewrite <- (Nat.add_assoc (ty_main_tree_folds x1 ) (ty_main_tree_folds x0)).
    rewrite (Nat.add_comm (ty_main_tree_folds trl )). auto.
  - eapply ty_main_tree_linear_sub_and_inv in Htr1; destruct_all; subst.
    forwards  HHtr2_1: IHHtr2_1; eauto.
    forwards HHtr2_2: IHHtr2_2; eauto.

    simpls. rewrite  HHtr2_1,  HHtr2_2. 
    rewrite Nat.add_assoc.
    rewrite Nat.add_assoc.
    rewrite <- (Nat.add_assoc (ty_main_tree_folds x1 )).
    rewrite <- (Nat.add_assoc (ty_main_tree_folds x1 ) (ty_main_tree_folds x0)).
    rewrite (Nat.add_comm (ty_main_tree_folds trl )). auto.
  -
    forwards HHtr2: IHHtr2; eauto.

    simpls. rewrite HHtr2.
    rewrite Nat.add_succ_r.
    rewrite Nat.add_assoc. auto.
  - 
    inversions Htr1.
    eauto.
  - forwards HHtr2: IHHtr2; eauto.
    simpls. rewrite HHtr2.
    rewrite Nat.add_succ_r.
    auto.
Qed.

Lemma ty_main_tree_under_lc_extend tr oc ocp:
  ty_main_tree_under_lc (ty_main_tree_extend_lc tr oc) (logpath_extend (logctx_path oc) ocp) = 
  ty_main_tree_under_lc tr ocp.
Proof.  
  reverts tr ocp.
  induction oc; intros; simpls; auto.
Qed.


End WithContext.
End WithTypingMode.

(* Repeat hints outside of section *)
Hint Constructors ty_main_tree_sub: core.
Hint Constructors ty_main_tree_linear_sub: core.
