Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath LogpathOrder.

Require Import TypeInduction TypLogicInduction.
Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic LayeredCapabilityMain.
Require Import LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure LayeredTypingLogic LayeredTypingMain.
Require Import LayeredSubtypingMainTree.

Section WithTypingMode.
Context {typing_mode: Mode}.

Inductive ty_sub_main_tree_ord (G: ctx) (e: renv) (x: var): nat -> typ -> typ -> Prop :=
| ty_sub_main_tree_ord_refl T:
    ty_sub_main_tree_ord G e x 0 T T
(* Replace lower bound by selection *)
| ty_sub_main_tree_ord_sel n tr T ocp S1 S2 U y A z:
    ty_main_tree_linear_sub G e x tr T (logctx_typ ocp (open z S1)) ->
    G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
    n =  (S (ty_main_tree_folds tr)) ->
    ty_sub_main_tree_ord G e x n T (logctx_typ ocp (typ_sel (avar_f y) A (avar_f z)))
| ty_sub_main_tree_ord_bnd n tr ocp ocl T U:
    ty_main_tree_linear_sub G e x tr (logctx_typ ocl U) (logctx_typ ocp (open x T)) ->
    ((mode_has_incap typing_mode) -> (ty_has_N_logic G e U /\ ty_main_tree_linear_sub G e x (ty_main_tree_under_lc tr (logctx_path ocp)) U (open x T))) ->
    logctx_follow ocl ocp ->
    typ_indep_b 0 T ->
    n = (S (ty_main_tree_folds tr)) ->
    ty_sub_main_tree_ord G e x n (logctx_typ ocl U) (logctx_typ ocp (typ_bnd T))
| ty_sub_main_tree_ord_nocap_bnd n tr ocp T U:
    ty_main_tree_linear_sub G e x tr U (logctx_typ ocp typ_nocap) ->
    n = (S (ty_main_tree_folds tr)) ->
    ty_sub_main_tree_ord G e x n U (logctx_typ ocp (typ_bnd T))
.


Section WithContext.
Context (G: ctx) (e: renv) (x: var).
Context (Hin: inert G).

Lemma ty_sub_main_tree_ord_lc n oc U T:
  ty_sub_main_tree_ord G e x n U T ->
  ty_sub_main_tree_ord G e x n (logctx_typ oc U) (logctx_typ oc T).
Proof.
  introv Htr.
  reverts oc.
  induction Htr; intros; subst.
  - eapply ty_sub_main_tree_ord_refl; eauto.
  - rewrite <- logctx_extend_typ in *.
    applys ty_sub_main_tree_ord_sel (logctx_extend oc ocp).
    + rewrite logctx_extend_typ in *. eapply ty_main_tree_linear_sub_lc. eauto.
    + eauto.
    + rewrite ty_main_tree_lc_folds. eauto.
  - rewrite <- ! logctx_extend_typ in *.
    applys ty_sub_main_tree_ord_bnd (logctx_extend oc ocp) (logctx_extend oc ocl).
    + rewrite ! logctx_extend_typ in *. eapply ty_main_tree_linear_sub_lc. eauto.
    + intros.
      forwards: H0; eauto. destructecs H4. split. auto. rewrite logctx_path_extend. rewrite ty_main_tree_under_lc_extend. auto.
    + eapply logctx_follow_extend. auto.
    + auto.
    + rewrite ty_main_tree_lc_folds.
      eauto.
  - rewrite <- logctx_extend_typ in *.
    applys ty_sub_main_tree_ord_nocap_bnd (logctx_extend oc ocp).
    + rewrite ! logctx_extend_typ in *. eapply ty_main_tree_linear_sub_lc. eauto.
    + rewrite ty_main_tree_lc_folds.
      eauto.
Qed.

End WithContext.

Lemma ty_sub_main_tree_ord_tree_linear G e x n T U:
  inert G ->
  ty_sub_main_tree_ord G e x n T U ->
  exists tr,
    ty_main_tree_linear_sub G e x tr T U /\
    (ty_main_tree_folds tr) = n.
Proof.
  introv Hin Hty. 

  reverts T U (* oc *).
  induction n using nat_lt_induction; intros; subst.
  inversions Hty.
  -
    eexists. split.
    eapply ty_main_tree_linear_sub_refl. eapply ty_main_tree_refl_folds.
  -
    forwards HtrS1: ty_main_tree_linear_sub_refl (open z S1).
    forwards: ty_main_tree_linear_sub_lc.
    eapply ty_main_tree_linear_sub_sel; eauto.
    forwards Htrt: ty_main_tree_linear_sub_trans H0.
    eauto.
    eexists. split. eauto.
    erewrite ty_main_tree_linear_trans_folds_eq by eauto.
    rewrite ty_main_tree_lc_folds.
    simpl. rewrite ty_main_tree_refl_folds. rewrite Nat.add_1_r. auto.
  -
    (* Main predecessor *)

    forwards HtrS1: ty_main_tree_linear_sub_refl G e x (open x T0).
    forwards: ty_main_tree_linear_sub_lc_follow_replace.
    eauto. eauto.

    {
      destruct (mode_has_incap' typing_mode ) eqn: Hm.
      *

        forwards HH: H1. auto.
        destructecs HH; subst.
        eapply ty_main_tree_linear_sub_bnd. eapply HH. eauto. eauto.
      *
        eapply ty_main_tree_linear_sub_bnd. 
        forwards Hoc: ty_main_tree_linear_sub_lc_inv  H0.
        destructecs Hoc.
        forwards: logctx_follow_eq_inv Hoc0. eapply logctx_follow_trans; eauto. destruct H4; subst.
        eauto. intro. rewrite Hm in H4; discriminate. auto.
    }
    eexists. split. eapply H4.
    erewrite logctx_follow_path_eq with (oc1 := ocl) by eauto.
    rewrite ty_main_tree_set_bnd_folds.
    simpl. auto.
  -
    (* Main predecessor *)
    forwards HtrS1: ty_main_tree_linear_sub_refl G e x (open x T0).
    forwards: ty_main_tree_linear_sub_lc.
    eapply ty_main_tree_linear_sub_nocap_bnd; eauto.
    forwards Htrt: ty_main_tree_linear_sub_trans H0. eauto.

    eexists. split. eauto.
    erewrite ty_main_tree_linear_trans_folds_eq by eauto.
    rewrite ty_main_tree_lc_folds.
    simpl.
    rewrite Nat.add_1_r. auto.
Qed.


Section WithContext.
Context (G: ctx) (e: renv) (x: var).
Context (Hin: inert G).


Lemma ty_sub_main_tree_ord_or nr Tr Ur nl Tl Ul:
  ty_sub_main_tree_ord G e x nl Tl Ul ->
  ty_sub_main_tree_ord G e x nr Tr Ur ->
  ty_sub_main_tree_ord G e x (nl + nr) (typ_or Tl Tr) (typ_or Ul Ur).
Proof.
  introv Htrl Htrr.
  reverts Tr Ur nr.
  induction Htrl; intros; subst.
  - simpl.
    forwards: ty_sub_main_tree_ord_lc (logctx_orr T logctx_nil) Htrr .
    simpl in H. auto.
  - 
    rewrite logctx_fold_orl in *.
    forwards Hxxx: ty_sub_main_tree_ord_tree_linear Htrr. eauto. auto.
    destructecs Hxxx.
    applys ty_sub_main_tree_ord_sel (typ_or T Tr) (logctx_orl ocp Ur).
    2: eauto.
    simpl.
    constructor. 2: eauto. eauto.
    subst. simpls. eauto.
  - 
    rewrite ! logctx_fold_orl in *.
    forwards Hxxx: ty_sub_main_tree_ord_tree_linear Htrr. eauto. auto.
    destructecs Hxxx.
    applys ty_sub_main_tree_ord_bnd (logctx_orl ocp Ur) (logctx_orl ocl Tr).
    simpl.
    constructor. 2: eauto. eauto.
    eauto.
    eauto.
    eauto.
    simpl.
    subst. simpls. eauto.
  - 
    rewrite logctx_fold_orl in *.
    forwards: ty_sub_main_tree_ord_tree_linear Htrr. eauto. auto.
    destructecs H0.
    applys ty_sub_main_tree_ord_nocap_bnd  (logctx_orl ocp Ur) T  (typ_or U Tr) .

    simpl.
    constructor. 2: eauto. eauto.
    subst. simpls. eauto.

Qed.

Lemma ty_sub_main_tree_ord_and nr Tr Ur nl Tl Ul:
  ty_sub_main_tree_ord G e x nl Tl Ul ->
  ty_sub_main_tree_ord G e x nr Tr Ur ->
  ty_sub_main_tree_ord G e x (nl + nr) (typ_and Tl Tr) (typ_and Ul Ur).
Proof.
  introv Htrl Htrr.
  reverts Tr Ur nr.
  induction Htrl; intros; subst.
  - simpl.
    forwards: ty_sub_main_tree_ord_lc (logctx_andr T logctx_nil) Htrr .
    simpl in H. auto.
  - 
    rewrite logctx_fold_andl in *.
    forwards Hxxx: ty_sub_main_tree_ord_tree_linear Htrr. eauto. auto.
    destructecs Hxxx.
    applys ty_sub_main_tree_ord_sel (typ_and T Tr) (logctx_andl ocp Ur).
    2: eauto.
    simpl.
    constructor. 2: eauto. eauto.
    subst. simpls. eauto.
  - 
    rewrite ! logctx_fold_andl in *.
    forwards Hxxx: ty_sub_main_tree_ord_tree_linear Htrr. eauto. auto.
    destructecs Hxxx.
    applys ty_sub_main_tree_ord_bnd (logctx_andl ocp Ur) (logctx_andl ocl Tr).
    simpl.
    constructor. 2: eauto. eauto.
    eauto.
    eauto.
    eauto.
    simpl.
    subst. simpls. eauto.
  - 
    rewrite logctx_fold_andl in *.
    forwards: ty_sub_main_tree_ord_tree_linear Htrr. eauto. auto.
    destructecs H0.
    applys ty_sub_main_tree_ord_nocap_bnd (logctx_andl ocp Ur) T  (typ_and U Tr) .

    simpl.
    constructor. 2: eauto. eauto.
    subst. simpls. eauto.

Qed.



End WithContext.

Lemma ty_main_linear_sub_tree_ord G e x tr T U:
  inert G ->
  ty_main_tree_linear_sub G e x tr T U ->
  ty_sub_main_tree_ord G e x (ty_main_tree_folds tr) T U.
Proof.
  introv Hin Htr.
  remember (ty_main_tree_folds tr).
  reverts tr T U.

  induction n using nat_lt_0S_induction; intros; subst.
  - forwards Heq: ty_main_tree_linear_sub_folds_0_inv; eauto.
    subst. eapply ty_sub_main_tree_ord_refl.
  -
    induction Htr; intros; simpls.
    + eapply Nat.lt_irrefl in H. contradiction.
    + forwards Hadd: Nat.add_pos_cases H. destructecs Hadd.
      * forwards HH0: H0 trr  Tr.
        2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. eauto.
        forwards: IHHtr1 .
        intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r.
        eapply ty_sub_main_tree_ord_or; eauto.
      * forwards HH0: H0 trl  Tl.
        2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. eauto.
        forwards: IHHtr2 .
        intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l.
        eapply ty_sub_main_tree_ord_or; eauto.
    +
      forwards Hadd: Nat.add_pos_cases H. destructecs Hadd.
      * forwards HH0: H0 trr  Tr.
        2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. eauto.
        forwards: IHHtr1 .
        intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r.
        eapply ty_sub_main_tree_ord_and; eauto.
      * forwards HH0: H0 trl  Tl.
        2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. eauto.
        forwards: IHHtr2 .
        intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l.
        eapply ty_sub_main_tree_ord_and; eauto.
    +
      destruct (mode_has_incap' typing_mode ) eqn: Hm.
      *
        forwards: H1. eauto. 
        replace   (typ_bnd U) with (logctx_typ logctx_nil   (typ_bnd U)).
        2: simpl; auto.
        replace T with (logctx_typ logctx_nil T).
        applys ty_sub_main_tree_ord_bnd. simpls; eauto. intros; splits. eauto. simpls. eauto. eauto. eauto. auto. simpl; auto.
      *
        replace   (typ_bnd U) with (logctx_typ logctx_nil   (typ_bnd U)).
        2: simpl; auto.

        replace T with (logctx_typ logctx_nil T).
        applys ty_sub_main_tree_ord_bnd. simpls; eauto. intros. rewrite Hm in H3; discriminate. eauto. auto. simpl; auto. auto.
    +
      replace   (typ_bnd T) with (logctx_typ logctx_nil   (typ_bnd T)).
      2: simpl; auto.
      replace typ_nocap with (logctx_typ logctx_nil typ_nocap).

      eapply ty_sub_main_tree_ord_nocap_bnd; simpls; eauto. constructor. constructor; intros; discriminate.
      simpl. auto. simpl. auto. 
    +
      replace  (typ_sel (avar_f y) A (avar_f z)) with (logctx_typ logctx_nil  (typ_sel (avar_f y) A (avar_f z))).
      2: simpl; auto.
      eapply ty_sub_main_tree_ord_sel; simpls; eauto.
Qed.



End WithTypingMode.