Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath.

Require Import LayeredCommon LayeredSubtypingAtomic LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityMain LayeredTypingAtomic LayeredTypingBasic.
Require Import LayeredCapabilityLogic.

Section WithTypingMode.
Context {typing_mode: Mode}.
(** * Layered typing on the union layer *)

(* Figure 17 in "Pure methods for roDOT" *)
(** [G ⊢b x : T]   #<br>#
    [―――――――――――(LTC-Basic)]   #<br>#
    [G ⊢c x : T] #<br>#
    #<br>#
    [―――――――――――(LTC-Top)]   #<br>#
    [G ⊢c x : ⊤] #<br>#
    #<br>#
    [G ⊢c x : T]   #<br>#
    [―――――――――――(LTC-Or1)]   #<br>#
    [G ⊢c x : (T ∨ U)] #<br>#
    #<br>#
    [G ⊢c x : U]   #<br>#
    [―――――――――――(LTC-Or2)]   #<br>#
    [G ⊢c x : (T ∨ U)]
    *)
Definition ty_inv_basic_closure G e x:= ty_or_closure (ty_inv_basic G e x).

Lemma ty_inv_basic_closure_atomic_basic G e x T:
  ty_inv_basic_closure G e x T ->
  typ_atomic T ->
  ty_inv_basic G e x T.
Proof.
  introv Hty Ha.
  unfold ty_inv_basic_closure in *.
  induction Hty; eauto.
  inversions Ha.
  inversions Ha.
  inversions Ha.
Qed.

Lemma ty_inv_basic_closure_tyq G e x U V:
  inert G ->
  ty_inv_basic_closure G e x U ->
  typing_equiv e U V ->
  ty_inv_basic_closure G e x V.
Proof.
  introv Hin Hty Htyq.
  unfold ty_inv_basic_closure in *.
  reverts V.
  induction Hty; intros.
  -
  eapply ty_or_closure_atomic.
  eapply ty_inv_basic_tyq; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_l; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_r; eauto.
  - inversions_tyq Htyq.
    eapply ty_or_closure_top; eauto.
Qed.



Lemma ty_inv_basic_closure_oc_and_inv G e x oc T U S: 
   ty_inv_basic_closure G e x (orctx_typ oc (typ_and T U)) ->
   ty_inv_basic_closure G e x (orctx_typ oc S).
Proof.
  introv Hty.
  remember ( (orctx_typ oc (typ_and T U))).
  reverts T U S oc.
  induction Hty; intros; subst; simpls.
  - eapply ty_or_closure_atomic.
    eapply ty_inv_basic_oc_and_replace. eauto.
  - destruct oc; simpls; inversions Heqt.
    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
Qed.


Lemma ty_has_NM_closure_basic G e x U:
  mode_has_top_nm_subtype typing_mode ->
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_has_N_closure G e U ->
  ty_has_M_closure G e U ->
  ty_inv_basic_closure G e x U.
Proof.
  introv Hm Hin Hcor Hdom HhN HhM.
  induction HhM.
  - inversions H; inversions HhN; inversions H. inversions H1.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_atomic.
    eapply ty_inv_atomic_has_top_atomic; eauto.
  - inversions HhN.
    inversions H; inversions H0.
    eapply ty_or_closure_l.
    eapply IHHhM. eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_MN; eauto.
  - inversions HhN.
    inversions H; inversions H0.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_or_closure_r.
    eapply IHHhM. eauto.
  - eapply ty_or_closure_top.
Qed.

Lemma ty_inv_basic_oc_orl G e x oc U T:
  ty_inv_basic G e x (orctx_typ oc U) ->
  ty_inv_basic_closure G e x (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  remember (orctx_typ oc U).
  reverts oc U.
  induction Hty; intros; subst.
  - forwards Hta: ty_inv_atomic_atomic H.
    forwards: orctx_atomic_inv Hta.
    destruct H0; subst; simpls.
    eapply ty_or_closure_l; eauto.
    eapply ty_or_closure_atomic; eapply ty_inv_basic_atomic; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_atomic. eapply ty_inv_basic_NM; eauto.
    + eapply ty_or_closure_atomic. eapply ty_inv_basic_NM; eauto.
      eapply ty_has_N_closure_oc_orl; eauto.
    + eapply ty_or_closure_atomic. eapply ty_inv_basic_NM; eauto.
      eapply ty_has_M_closure_oc_orl; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_atomic. eapply ty_inv_basic_MN; eauto.
    + eapply ty_or_closure_atomic. eapply ty_inv_basic_MN; eauto.
      eapply ty_has_M_closure_oc_orl; eauto.
    + eapply ty_or_closure_atomic. eapply ty_inv_basic_MN; eauto.
      eapply ty_has_N_closure_oc_orl; eauto.
Qed.

Lemma ty_inv_basic_closure_oc_orl G e x oc U T:
  ty_inv_basic_closure G e x (orctx_typ oc U) ->
  ty_inv_basic_closure G e x (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  remember (orctx_typ oc U).
  reverts oc U.
  induction Hty; intros; subst.
  - eapply ty_inv_basic_oc_orl; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_l. eauto.
    + eapply ty_or_closure_l. eapply IHHty; eauto.
    + eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r.  eapply IHHty; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_l. eapply ty_or_closure_top.
Qed.
Lemma ty_inv_basic_closure_oclc_orl G e x oc2 oc U T:
  ty_inv_basic_closure G e x (orctx_typ oc2 (logctx_typ oc U)) ->
  ty_inv_basic_closure G e x (orctx_typ oc2 (logctx_typ oc (typ_or U T))).
Proof.
  introv Hty.
  reverts oc2.
  induction oc; intros; simpls.
  - eapply ty_inv_basic_closure_oc_orl; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_inv_basic_closure_oc_and_inv in Hty; eauto.
  - eapply ty_inv_basic_closure_oc_and_inv in Hty; eauto.
Qed.
Lemma ty_inv_basic_closure_lc_orl G e x oc U T:
  ty_inv_basic_closure G e x (logctx_typ oc U) ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  eapply ty_inv_basic_closure_oclc_orl with (oc2:=orctx_nil); eauto.
Qed.

Lemma ty_inv_basic_oc_orr G e x oc U T:
  ty_inv_basic G e x (orctx_typ oc T) ->
  ty_inv_basic_closure G e x (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  remember (orctx_typ oc T).
  reverts oc T.
  induction Hty; intros; subst.
  - forwards Hta: ty_inv_atomic_atomic H.
    forwards: orctx_atomic_inv Hta.
    destruct H0; subst; simpls.
    eapply ty_or_closure_r; eauto.
    eapply ty_or_closure_atomic; eapply ty_inv_basic_atomic; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_atomic. eapply ty_inv_basic_NM; eauto.
    + eapply ty_or_closure_atomic. eapply ty_inv_basic_NM; eauto.
      eapply ty_has_N_closure_oc_orr; eauto.
    + eapply ty_or_closure_atomic. eapply ty_inv_basic_NM; eauto.
      eapply ty_has_M_closure_oc_orr; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_atomic. eapply ty_inv_basic_MN; eauto.
    + eapply ty_or_closure_atomic. eapply ty_inv_basic_MN; eauto.
      eapply ty_has_M_closure_oc_orr; eauto.
    + eapply ty_or_closure_atomic. eapply ty_inv_basic_MN; eauto.
      eapply ty_has_N_closure_oc_orr; eauto.
Qed.

Lemma ty_inv_basic_closure_oc_orr G e x oc U T:
  ty_inv_basic_closure G e x (orctx_typ oc T) ->
  ty_inv_basic_closure G e x (orctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  remember (orctx_typ oc T).
  reverts oc T.
  induction Hty; intros; subst.
  - eapply ty_inv_basic_oc_orr; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_l. eauto.
    + eapply ty_or_closure_l. eapply IHHty; eauto.
    + eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r. eauto.
    + eapply ty_or_closure_r.  eapply IHHty; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_r. eapply ty_or_closure_top.
Qed.
Lemma ty_inv_basic_closure_oclc_orr G e x oc2 oc U T:
  ty_inv_basic_closure G e x (orctx_typ oc2 (logctx_typ oc T)) ->
  ty_inv_basic_closure G e x (orctx_typ oc2 (logctx_typ oc (typ_or U T))).
Proof.
  introv Hty.
  reverts oc2.
  induction oc; intros; simpls.
  - eapply ty_inv_basic_closure_oc_orr; eauto.
  - rewrite orctx_extend_l in *. eauto.
  - rewrite orctx_extend_r in *. eauto.
  - eapply ty_inv_basic_closure_oc_and_inv in Hty; eauto.
  - eapply ty_inv_basic_closure_oc_and_inv in Hty; eauto.
Qed.
Lemma ty_inv_basic_closure_lc_orr G e x oc U T:
  ty_inv_basic_closure G e x (logctx_typ oc T) ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_or U T)).
Proof.
  introv Hty.
  eapply ty_inv_basic_closure_oclc_orr with (oc2:=orctx_nil); eauto.
Qed.

Lemma ty_inv_basic_lc_or_diag G e x oc U:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic G e x (logctx_typ oc (typ_or U U)) ->
  ty_inv_basic_closure G e x (logctx_typ oc U).
Proof.
  introv Hin Hcor Hdom Hty.
  inversions Hty.
  - eapply ty_inv_atomic_atomic in H.
    destruct oc; simpls; inversions H.
  - destruct oc; simpls; inversions H.
    eapply ty_has_NM_closure_basic; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_N_closure_lc_or_diag; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_M_closure_lc_or_diag; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_has_NM_closure_basic; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_M_closure_lc_or_diag; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_N_closure_lc_or_diag; eauto.
Qed.
Lemma ty_inv_basic_closure_lc_or_diag G e x oc U:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_or U U)) ->
  ty_inv_basic_closure G e x (logctx_typ oc U).
Proof.
  introv Hin Hcor Hdom Hty.
  remember ( (logctx_typ oc (typ_or U U))).
  reverts U oc.
  induction Hty; intros; subst; simpls.
  - eapply ty_inv_basic_lc_or_diag; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eauto.
    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eauto.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
Qed.

Lemma ty_inv_basic_closure_lc_subtyp G e x oc U V:
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc U) ->
  subtyp_atomic G e U V ->
  ty_inv_basic_closure G e x (logctx_typ oc V).
Proof.
  introv Hin Hty Hsu.
  remember  (logctx_typ oc U). reverts oc U.
  induction Hty; intros; subst.
  - eapply ty_or_closure_atomic.
    eapply ty_inv_basic_lc_subtyp; eauto.
  - destruct oc; simpls; inversions Heqt.
    + inversions Hsu.
    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl. eauto. 
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    + inversions Hsu.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl. eauto.
  -
    destruct oc; simpls; inversions Heqt.
    inversions Hsu.
Qed.

Lemma ty_inv_basic_closure_lc_bot_replace G e x oc U:
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc typ_bot) ->
  ty_inv_basic_closure G e x (logctx_typ oc U).
Proof.
  introv Hin Hty.
  remember  (logctx_typ oc typ_bot). reverts oc.
  induction Hty; intros; subst.
  - eapply ty_or_closure_atomic.
    eapply ty_inv_basic_lc_bot_replace; eauto.
  - destruct oc; simpls; inversions Heqt.

    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
Qed.
Lemma ty_inv_basic_closure_lc_sel_replace G e x oc U z A y:
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_sel z A y)) ->
  ty_inv_basic_closure G e x (logctx_typ oc U).
Proof.
  introv Hin Hty.
  remember  (logctx_typ oc  (typ_sel z A y)). reverts oc.
  induction Hty; intros; subst.
  - eapply ty_or_closure_atomic.
    eapply ty_inv_basic_lc_sel_replace; eauto.
  - destruct oc; simpls; inversions Heqt.

    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
Qed.
Lemma ty_inv_basic_closure_lc_bnd_replace G e x oc U T:
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_bnd T)) ->
  ty_inv_basic_closure G e x (logctx_typ oc U).
Proof.
  introv Hin Hty.
  remember  (logctx_typ oc  (typ_bnd T)). reverts oc.
  induction Hty; intros; subst.
  - eapply ty_or_closure_atomic.
    eapply ty_inv_basic_lc_bnd_replace; eauto.
  - destruct oc; simpls; inversions Heqt.

    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
Qed.
Lemma ty_inv_basic_closure_lc_nocap_replace G e x oc U:
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc typ_nocap) ->
  ty_has_N G e U ->
  ty_inv_basic_closure G e x (logctx_typ oc U).
Proof.
  introv Hin Hty HhN.
  remember  (logctx_typ oc typ_nocap). reverts oc.
  induction Hty; intros; subst.
  - eapply ty_or_closure_atomic.
    eapply ty_inv_basic_lc_nocap_replace; eauto.
  - destruct oc; simpls; inversions Heqt.

    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
Qed.


Lemma ty_inv_basic_closure_lc_M_replace G e x oc U:
  mode_top_nm' typing_mode = mode_top_nm_complement ->
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_has_M G e U ->
  ty_inv_basic_closure G e x (logctx_typ oc U).
Proof.
  introv Hm Hin Hty HhN.
  remember  (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))). reverts oc.
  induction Hty; intros; subst.
  - eapply ty_or_closure_atomic.
    eapply ty_inv_basic_lc_M_replace; eauto.
  - destruct oc; simpls; inversions Heqt.

    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
Qed.



Lemma ty_inv_basic_closure_lc_and_replace G e x oc U T S:
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_and T S)) ->
  ty_inv_basic_closure G e x (logctx_typ oc U).
Proof.
  introv Hin Hty.
  remember  (logctx_typ oc  (typ_and T S)). reverts oc.
  induction Hty; intros; subst.
  - eapply ty_or_closure_atomic.
    eapply ty_inv_basic_lc_and_replace; eauto.
  - destruct oc; simpls; inversions Heqt.

    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
Qed.
Lemma ty_inv_basic_lc_top_replace G e x oc U:
  inert G ->
  ty_inv_basic G e x (logctx_typ oc U) ->
  ty_inv_basic_closure G e x (logctx_typ oc typ_top).
Proof.
  introv Hin Hty.
  inversions Hty.
  - eapply ty_inv_atomic_atomic in H. destruct oc; simpls; inversions H; eapply ty_or_closure_top; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_or_closure_top; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_N_closure_lc_top_replace; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_NM; eauto.
    eapply ty_has_M_closure_lc_top_replace; eauto.
  - destruct oc; simpls; inversions H.
    eapply ty_or_closure_top; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_M_closure_lc_top_replace; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_MN; eauto.
    eapply ty_has_N_closure_lc_top_replace; eauto.
Qed.
Lemma ty_inv_basic_closure_lc_top_replace G e x oc U:
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc U) ->
  ty_inv_basic_closure G e x (logctx_typ oc typ_top).
Proof.
  introv Hin Hty.
  remember  (logctx_typ oc U). reverts oc.
  induction Hty; intros; subst.
  - 
    eapply ty_inv_basic_lc_top_replace; eauto.
  - destruct oc; simpls; inversions Heqt.
    +  eapply ty_or_closure_top; eauto. 
    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    +
    eapply ty_or_closure_top; eauto.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
    eapply ty_or_closure_top; eauto.
Qed.

Lemma ty_inv_basic_closure_lc_M_closure_replace G e x oc U:
  mode_has_top_nm_complement typing_mode ->
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_has_M_closure G e U ->
  ty_inv_basic_closure G e x (logctx_typ oc U).
Proof.
  introv Hm Hin Hty HhN.
  induction HhN.
  - eapply ty_inv_basic_closure_lc_M_replace; eauto.
  - eapply ty_inv_basic_closure_lc_orl; eauto.
  - eapply ty_inv_basic_closure_lc_orr; eauto.
  - eapply ty_inv_basic_closure_lc_top_replace; eauto.
Qed.

Lemma ty_inv_basic_closure_NM_canon G e x T1 T2:
  mode_has_top_nm_subtype typing_mode ->
  (mode_has_top_nm_complement typing_mode -> forall x0 : var, G ⁏ e ⊢# T1 /[-> x0] <: typ_bot) ->
  ty_inv_basic_closure G e x (typ_or typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) T1 T2))).
Proof.
  introv Hm Hbot.
  eapply ty_or_closure_atomic.
  eapply ty_inv_basic_NM; eauto.
  eapply ty_or_closure_atomic. eauto.
  eapply ty_or_closure_atomic. eauto.
Qed.

Lemma ty_inv_basic_closure_lc_NM_replace G e x oc U T1 T2:
  mode_has_top_nm_subtype typing_mode ->
  inert G ->
  (mode_has_top_nm_complement typing_mode -> forall x0 : var, G ⁏ e ⊢# T1 /[-> x0] <: typ_bot) ->
  ty_inv_basic_closure G e x (logctx_typ oc U) ->
  ty_inv_basic_closure G e x (logctx_typ oc  (typ_NM T1 T2)).
Proof.
  introv Hm Hin Hbot Hty.
  remember  (logctx_typ oc U). reverts oc.
  induction Hty; intros; subst.
  - 
    eapply ty_or_closure_atomic; eauto.
    eapply ty_inv_basic_lc_NM_replace; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_inv_basic_closure_NM_canon; eauto.
    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_inv_basic_closure_NM_canon; eauto.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
    eapply ty_inv_basic_closure_NM_canon; eauto.
Qed.

Lemma ty_inv_basic_closure_lc_has_top_replace G e x oc U T:
  inert G ->
renv_corr G e ->
x \in ctx_dom G ->
  ty_inv_basic_closure G e x (logctx_typ oc U) ->
  ty_has_top_atomic G e T ->
  ty_inv_basic_closure G e x (logctx_typ oc T).
Proof.
  introv Hin Hcor Hdom Hty Htop.
  remember  (logctx_typ oc U). reverts oc.
  induction Hty; intros; subst.
  - 
    eapply ty_or_closure_atomic; eauto.
    eapply ty_inv_basic_lc_has_top_replace; eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_atomic. eapply ty_inv_basic_atomic. eapply ty_inv_atomic_has_top_atomic; eauto.
    +
    eapply ty_or_closure_l.
    eapply IHHty. eapply eq_refl.
    +
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions Heqt.
    + eapply ty_or_closure_atomic. eapply ty_inv_basic_atomic. eapply ty_inv_atomic_has_top_atomic; eauto.
    +
    eapply ty_or_closure_r.  eauto.
    +
    eapply ty_or_closure_r. 
    eapply IHHty. eapply eq_refl.
  -
    destruct oc; simpls; inversions Heqt.
    eapply ty_or_closure_atomic. eapply ty_inv_basic_atomic. eapply ty_inv_atomic_has_top_atomic; eauto.
Qed.

Lemma ty_inv_basic_closure_lc_typ_pfor_replace G e x oc A SL SU TL TU:
  inert G ->
  x \in ctx_dom G ->
  ty_pf_or_none G x A TL TU ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU)))  \/ ty_has_top_atomic G e (typ_rcd (dec_typ A SL SU)).
Proof.
  introv Hin Hdom Hpf Hty.
  dependent induction Hty; intros; subst.
  - forwards: ty_inv_basic_lc_typ_pfor_replace; fweauto.
    destructecs H0.
    left. eapply ty_or_closure_atomic; eauto.
    eauto.
  - destruct oc; simpls; inversions x.
    forwards: IHHty; eauto. destructecs H.
    left. eapply ty_or_closure_l. eauto.
    eauto.
    left.
    eapply ty_or_closure_l. eauto.
  - destruct oc; simpls; inversions x.
    left. eapply ty_or_closure_r. eauto.
    forwards: IHHty; eauto. destructecs H.
    left.
    eapply ty_or_closure_r; eauto.
    eauto.
  - destruct oc; simpls; inversions x.
Qed.

Lemma ty_inv_basic_lc_or_inv G e x oc T S:
  inert G ->
  ty_inv_basic G e x (logctx_typ oc (typ_or T S)) -> 
  ty_inv_basic G e x (logctx_typ oc T) \/
  ty_inv_basic G e x (logctx_typ oc S) \/ 
  (ty_has_N_closure G e (logctx_typ oc T) /\ ty_has_M_closure G e (logctx_typ oc S) /\ mode_has_top_nm_subtype typing_mode ) \/
  (ty_has_N_closure G e (logctx_typ oc S) /\ ty_has_M_closure G e (logctx_typ oc T) /\ mode_has_top_nm_subtype typing_mode ).
Proof.
  introv Hi Hty.
  remember (logctx_typ oc (typ_or T S)).
  reverts oc T S.
  induction Hty; intros; subst.
  -
  false; eapply ty_inv_atomic_lc_or_false; eauto.
  -
    destruct oc; simpls; inversions Heqt; eauto 6.
    +
      eapply ty_has_N_closure_or_inv in H0; destruct_all.
      *
        left.
        eapply ty_inv_basic_NM; eauto.
      *
        right; left.
        eapply ty_inv_basic_NM; eauto.
    +
      eapply ty_has_M_closure_or_inv in H1; destruct_all.
      *
        left.
        eapply ty_inv_basic_NM; eauto.
      *
        right; left.
        eapply ty_inv_basic_NM; eauto.
  -
    destruct oc; simpls; inversions Heqt; eauto 6.
    +
      eapply ty_has_M_closure_or_inv in H0; destruct_all.
      *
        left.
        eapply ty_inv_basic_MN; eauto.
      *
        right; left.
        eapply ty_inv_basic_MN; eauto.
    +
      eapply ty_has_N_closure_or_inv in H1; destruct_all.
      *
        left.
        eapply ty_inv_basic_MN; eauto.
      *
        right; left.
        eapply ty_inv_basic_MN; eauto.
Qed.


Lemma ty_inv_basic_closure_lc_or_inv G e x oc T S:
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_or T S)) ->
  ty_inv_basic_closure G e x (logctx_typ oc T) \/
  ty_inv_basic_closure G e x (logctx_typ oc S) \/ 
  (ty_has_N_closure G e (logctx_typ oc T) /\ ty_has_M_closure G e (logctx_typ oc S) /\ mode_has_top_nm_subtype typing_mode ) \/
  (ty_has_N_closure G e (logctx_typ oc S) /\ ty_has_M_closure G e (logctx_typ oc T) /\ mode_has_top_nm_subtype typing_mode ).
Proof.
  introv Hi Hh.
  remember (logctx_typ oc (typ_or T S)).
  reverts oc T S.
  induction Hh; intros; subst.
  - eapply ty_inv_basic_lc_or_inv in H; eauto; destruct_all; eauto 6;
    eapply ty_or_closure_atomic in H; eauto.
  - destruct oc; simpls; inversions Heqt.
    + auto.
    + forwards: IHHh; eauto. destruct_all.
      left. eapply ty_or_closure_l. auto.
      right. left. eapply ty_or_closure_l. auto. 
      right. right. left. splits; auto 1; eapply ty_or_closure_l; eauto.
      right. right. right. splits; auto 1; eapply ty_or_closure_l; eauto.
    + left. eapply ty_or_closure_l. auto.
  - destruct oc; simpls; inversions Heqt.
    + auto.
    + left. eapply ty_or_closure_r. auto.
    + forwards: IHHh; eauto. destruct_all.
     left. eapply ty_or_closure_r. auto.
      right. left. eapply ty_or_closure_r. auto. 
      right. right. left. splits; auto 1; eapply ty_or_closure_r; eauto.
      right. right. right. splits; auto 1; eapply ty_or_closure_r; eauto.
  - destruct oc; simpl in Heqt; inversions Heqt. 
Qed.

Lemma ty_has_N_also_M_basic_closure G e x  T:
  mode_has_top_nm_subtype typing_mode ->
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_has_N_closure G e T ->
  ty_has_M_closure G e T ->
  ty_inv_basic_closure G e x T.
Proof.
  introv Hm Hi Hcor Hdom HhN HhM.
  induction HhN.
  - forwards: ty_has_M_closure_has_N_top; fweauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_atomic.
    eapply ty_inv_atomic_has_top_atomic; eauto.
  - inversions HhM. inversions H. eapply ty_or_closure_l. eapply IHHhN. eapply H0.
    eapply ty_or_closure_atomic. eapply ty_inv_basic_NM; eauto.
  - inversions HhM. inversions H. 
    eapply ty_or_closure_atomic. eapply ty_inv_basic_MN; eauto.
    eapply ty_or_closure_r. 
    eapply IHHhN. eapply H0.
  - eapply ty_or_closure_top.
Qed.

Lemma ty_has_N_also_M_lc_or_basic_closure G e x oc U V:
  mode_has_top_nm_subtype typing_mode ->
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_has_N_closure G e (logctx_typ oc U) ->
  ty_has_M_closure G e (logctx_typ oc V) ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_or U V)).
Proof.
  introv Hm Hi Hcor Hdom HhN HhM. reverts U V.
  induction oc; intros; simpls.
  - eapply ty_or_closure_atomic. eapply ty_inv_basic_NM; eauto.
  - eapply ty_has_M_closure_or_inv with (oc0:=logctx_nil) in HhM.
    eapply ty_has_N_closure_or_inv with (oc0:=logctx_nil) in HhN.
    simpls. destruct_all.
    + eapply ty_or_closure_l. eapply IHoc; eauto.
    + eapply ty_or_closure_atomic.
       eapply ty_inv_basic_MN; eauto. eapply ty_has_M_closure_lc_orr; eauto.
    + eapply ty_or_closure_atomic.
       eapply ty_inv_basic_NM; eauto. eapply ty_has_N_closure_lc_orl; eauto.
    + eapply ty_or_closure_r. eapply ty_has_N_also_M_basic_closure; eauto.
  - eapply ty_has_M_closure_or_inv with (oc0:=logctx_nil) in HhM.
    eapply ty_has_N_closure_or_inv with (oc0:=logctx_nil) in HhN.
    simpls. destruct_all.
    + eapply ty_or_closure_l. eapply ty_has_N_also_M_basic_closure; eauto.

    + eapply ty_or_closure_atomic.
       eapply ty_inv_basic_MN; eauto. eapply ty_has_N_closure_lc_orl; eauto.
    + eapply ty_or_closure_atomic.
       eapply ty_inv_basic_NM; eauto. eapply ty_has_M_closure_lc_orr; eauto.
    + eapply ty_or_closure_r. eapply IHoc; eauto.
  - eapply ty_has_N_closure_lc_and_replace with (oc0:=logctx_nil) (U0:=typ_bot) in HhN . simpls. inversions HhN. false. eapply ty_has_N_lc_bot_false with (oc0:=logctx_nil); eauto.
  - eapply ty_has_N_closure_lc_and_replace with (oc0:=logctx_nil) (U0:=typ_bot) in HhN . simpls. inversions HhN. false. eapply ty_has_N_lc_bot_false with (oc0:=logctx_nil); eauto.
Qed.
Lemma ty_has_M_also_N_lc_or_basic_closure G e x oc U V:
  mode_has_top_nm_subtype typing_mode ->
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_has_M_closure G e (logctx_typ oc U) ->
  ty_has_N_closure G e (logctx_typ oc V) ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_or U V)).
Proof.
  introv Hm Hi Hcor Hdom HhM HhN. reverts U V.
  induction oc; intros; simpls.
  - eapply ty_or_closure_atomic. eapply ty_inv_basic_MN; eauto.
  - eapply ty_has_M_closure_or_inv with (oc0:=logctx_nil) in HhM.
    eapply ty_has_N_closure_or_inv with (oc0:=logctx_nil) in HhN.
    simpls. destruct_all.
    + eapply ty_or_closure_l. eapply IHoc; eauto.
    + eapply ty_or_closure_atomic.
       eapply ty_inv_basic_NM; eauto. eapply ty_has_N_closure_lc_orr; eauto.
    + eapply ty_or_closure_atomic.
       eapply ty_inv_basic_MN; eauto. eapply ty_has_M_closure_lc_orl; eauto.
    + eapply ty_or_closure_r. eapply ty_has_N_also_M_basic_closure; eauto.
  - eapply ty_has_M_closure_or_inv with (oc0:=logctx_nil) in HhM.
    eapply ty_has_N_closure_or_inv with (oc0:=logctx_nil) in HhN.
    simpls. destruct_all.
    + eapply ty_or_closure_l. eapply ty_has_N_also_M_basic_closure; eauto.

    + eapply ty_or_closure_atomic.
       eapply ty_inv_basic_NM; eauto. eapply ty_has_M_closure_lc_orl; eauto.
    + eapply ty_or_closure_atomic.
       eapply ty_inv_basic_MN; eauto. eapply ty_has_N_closure_lc_orr; eauto.
    + eapply ty_or_closure_r. eapply IHoc; eauto.
  - eapply ty_has_N_closure_lc_and_replace with (oc0:=logctx_nil) (U0:=typ_bot) in HhN . simpls. inversions HhN. false. eapply ty_has_N_lc_bot_false with (oc0:=logctx_nil); eauto.
  - eapply ty_has_N_closure_lc_and_replace with (oc0:=logctx_nil) (U0:=typ_bot) in HhN . simpls. inversions HhN. false. eapply ty_has_N_lc_bot_false with (oc0:=logctx_nil); eauto.
Qed.


Lemma ty_inv_basic_closure_lc_typ_pf_inv G e x oc U A SL SU TL TU:
  inert G ->
  G ⊢! x ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ((forall SL2 SU2,       ((forall y : var, G ⁏ e ⊢# open y SL <: typ_bot) -> (forall y : var, G ⁏ e ⊢# open y SL2 <: typ_bot)) -> ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A SL2 SU2)))) \/
  (forall y : var, G ⁏ e ⊢# open y SL <: open y TL) /\ (forall y : var, G ⁏ e ⊢# open y TU <: open y SU)).
Proof.
  introv Hin Hpf Hty.
  dependent induction Hty; intros; subst.
  -
    forwards: ty_inv_basic_lc_typ_pf_inv; eauto.
    destruct_all.
    + left. intros. eapply ty_or_closure_atomic; eauto.
    + right; eauto.
  -
    destruct oc; simpls; inversions x.
    *
    forwards: IHHty; eauto.
    destruct_all.
    + left. intros. eapply ty_or_closure_l. eapply H. eauto.
    + right; eauto.
    *
      left. intros. eapply ty_or_closure_l. eapply Hty.
  -
    destruct oc; simpls; inversions x.
    *
      left. intros. eapply ty_or_closure_r. eapply Hty.
    *
    forwards: IHHty; eauto.
    destruct_all.
    + left. intros. eapply ty_or_closure_r. eapply H. eauto.
    + right; eauto.
  -
    destruct oc; simpls; inversions x.
Qed.



Lemma ty_inv_basic_basic_closure_N_cases G e x oc T:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic G e x (logctx_typ oc T) ->
  ty_has_N_closure G e T ->
  ty_inv_basic_closure G e x T \/
  ty_inv_basic_closure G e x (logctx_typ oc (typ_nocap)).
Proof.
  introv Hi Hcor Hdom Hty Hh.

  remember ( (logctx_typ oc T)).
  reverts oc T.
  induction Hty; intros; subst.
  - left. eapply ty_or_closure_atomic. eapply ty_inv_basic_atomic. eapply ty_inv_atomic_lc_inv; eauto.
  -
    destruct oc; simpls; inversions Heqt.
    +
      left. eapply ty_or_closure_atomic. eapply ty_inv_basic_NM; eauto.
    +
      right. eapply ty_or_closure_atomic.  eapply ty_inv_basic_NM; eauto. eapply ty_has_N_closure_lc_nocap_replace; eauto.
    + forwards: ty_has_M_closure_lc_cases; eauto.
      destruct H2.
      * left. eapply ty_has_N_also_M_basic_closure; eauto.
      * right. eapply ty_or_closure_atomic. eapply ty_inv_basic_NM; eauto. 
  -
    destruct oc; simpls; inversions Heqt.
    +
      left. eapply ty_or_closure_atomic. eapply ty_inv_basic_MN; eauto.

    + forwards: ty_has_M_closure_lc_cases; eauto.
      destruct H2.
      * left. eapply ty_has_N_also_M_basic_closure; eauto.
      * right. eapply ty_or_closure_atomic. eapply ty_inv_basic_MN; eauto.
    + 
      right. eapply ty_or_closure_atomic.  eapply ty_inv_basic_MN; eauto. eapply ty_has_N_closure_lc_nocap_replace; eauto. 
Qed.


Lemma ty_inv_basic_closure_N_cases G e x oc T:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic_closure G e x (logctx_typ oc T) ->
  ty_has_N_closure G e T ->
  ty_inv_basic_closure G e x T \/
  ty_inv_basic_closure G e x (logctx_typ oc (typ_nocap)).
Proof.
  introv Hi Hcor Hdom Hty Hh.

  remember ( (logctx_typ oc T)).
  reverts oc T.
  induction Hty; intros; subst.
  - (* basic typing *)
    eapply ty_inv_basic_basic_closure_N_cases; eauto.
  - (* types with left *)
    destruct oc; simpls; inversions Heqt.
    + left. eapply ty_or_closure_l; eauto. 
    + forwards: IHHty; eauto. destruct H. eauto. right. eapply ty_or_closure_l. eauto.
    +  right. eapply ty_or_closure_l. eauto.
  - (* types with right *)
    destruct oc; simpls; inversions Heqt.
    + left. eapply ty_or_closure_r; eauto.
    +  right. eapply ty_or_closure_r. eauto.
    + forwards: IHHty; eauto. destruct H. eauto. right. eapply ty_or_closure_r. eauto.
  - destruct oc; simpls; inversions Heqt. left. eapply ty_or_closure_top; eauto.
Qed.

Lemma ty_inv_basic_closure_lc_nocap_closure_replace G e x oc U:
  inert G ->
  ty_inv_basic_closure G e x (logctx_typ oc typ_nocap) ->
  ty_has_N_closure G e U ->
  ty_inv_basic_closure G e x (logctx_typ oc U).
Proof.
  introv Hin Hty HhN.
  induction HhN.
  - eapply ty_inv_basic_closure_lc_nocap_replace; eauto.
  - eapply ty_inv_basic_closure_lc_orl; eauto.
  - eapply ty_inv_basic_closure_lc_orr; eauto.
  - eapply ty_inv_basic_closure_lc_top_replace; eauto.
Qed.


Lemma ty_inv_basic_closure_has_NM_cases G e w T:
  inert G ->
  ty_inv_basic_closure G e w T ->
  ty_has_N_closure G e T \/ ty_has_M_closure G e T.
Proof.
  introv Hi Hty.
  induction Hty.
  - eapply ty_inv_basic_has_NM_cases; eauto.
  - destructecs IHHty.
    left. eapply ty_or_closure_l; eauto.
    right. eapply ty_or_closure_l; eauto.
  - destructecs IHHty.
    left. eapply ty_or_closure_r; eauto.
    right. eapply ty_or_closure_r; eauto.
  - left. eapply ty_or_closure_top.
Qed.


Lemma ty_inv_basic_closure_has_NM_replace G e w oc T U:
  inert G ->
  ty_inv_basic_closure G e w (logctx_typ oc T) ->
  ty_has_N_closure G e T \/ ty_has_M_closure G e T \/  ty_inv_basic_closure G e w (logctx_typ oc U).
Proof.
  introv Hi Hty.
  remember ( (logctx_typ oc T)).
  reverts oc T.
  induction Hty; intros; subst.
  - forwards Hc: ty_inv_basic_has_NM_replace; fweauto.
    destructecs Hc; auto.
    right; right. eapply ty_or_closure_atomic; eauto.
  - (* types with left *)
    destruct oc; simpls; inversions Heqt.
    + forwards: IHHty logctx_nil T; fweauto. destructecs H0. left. eapply ty_or_closure_l. eauto. right. left.  eapply ty_or_closure_l. eauto. simpls. auto.
    + forwards: IHHty; fweauto. destructecs H. eauto. eauto. right. right. eapply ty_or_closure_l. eauto.
    + right. right. eapply ty_or_closure_l. eauto.
  - (* types with right *)
    destruct oc; simpls; inversions Heqt.
    + forwards: IHHty logctx_nil T; fweauto. destructecs H0. left. eapply ty_or_closure_r. eauto. right. left.  eapply ty_or_closure_r. eauto. simpls. auto.
    + right. right. eapply ty_or_closure_r. eauto.
    + forwards: IHHty; fweauto. destructecs H. eauto. eauto. right. right. eapply ty_or_closure_r. eauto.
  - destruct oc; simpls; inversions Heqt.
    left. eapply ty_or_closure_top.
Qed.


Lemma ty_inv_basic_N_also_M_replace G e w oc U:
  inert G ->
  renv_corr G e ->
  w \in ctx_dom G ->
  ty_inv_basic G e w (logctx_typ oc typ_nocap) ->
  ty_inv_basic G e w (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_inv_basic_closure G e w (logctx_typ oc U).
Proof.
  introv Hin Hcor Hdom HtyN HtyM.
  induction oc; simpls.
  - (* nil *)
    inversions HtyN. apply ty_inv_atomic_atomic in H. inversions H.
  - (* orl *)
    inversions HtyN.
    + (* basic N - impossible *)
      apply ty_inv_atomic_atomic in H. inversions H.
    + (* N M *)
      inversions HtyM.
      * (* basic M - impossible *) apply ty_inv_atomic_atomic in H. inversions H.
      * forwards: ty_has_N_closure_lc_M_replace U H8; fweauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM; eauto.
      * eapply ty_or_closure_r. eapply ty_has_NM_closure_basic; eauto.
    + (* M N *)
      inversions HtyM.
      * apply ty_inv_atomic_atomic in H. inversions H.
      * eapply ty_or_closure_r. eapply ty_has_NM_closure_basic; eauto.
      * forwards: ty_has_M_closure_lc_nocap_replace U H5.
        auto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN; eauto.
  - (* orr *)
    inversions HtyN.
    + (* basic N - impossible *)
      apply ty_inv_atomic_atomic in H. inversions H.
    + (* N M *)
      inversions HtyM.
      * (* basic M - impossible *) apply ty_inv_atomic_atomic in H. inversions H.
      * forwards: ty_has_M_closure_lc_nocap_replace U H6.
        auto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM; eauto.
      * eapply ty_or_closure_l. eapply ty_has_NM_closure_basic; eauto.
    + (* M N *)
      inversions HtyM.
      * apply ty_inv_atomic_atomic in H. inversions H.
      * eapply ty_or_closure_l. eapply ty_has_NM_closure_basic; eauto.
      * forwards: ty_has_N_closure_lc_M_replace U H9; fweauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN; eauto.
  - (* andl *) inversions HtyN. apply ty_inv_atomic_atomic in H. inversions H.
  - (* andr *) inversions HtyN. apply ty_inv_atomic_atomic in H. inversions H.
Qed.

Lemma ty_inv_basic_or_assoc G e w T1 T2 T3:
  ty_inv_basic G e w (typ_or T1 (typ_or T2 T3)) ->
  ty_inv_basic_closure G e w (typ_or (typ_or T1 T2) T3).
Proof.
  introv Hty.
  inversions Hty.
  - apply ty_inv_atomic_atomic in H. inversions H.
  - forwards: ty_has_M_closure_or_inv logctx_nil H6.
    destructecs H; simpls.
    + apply ty_or_closure_l. apply ty_or_closure_atomic.
      apply ty_inv_basic_NM; eauto. 
    + apply ty_or_closure_atomic.
      apply ty_inv_basic_NM; eauto.
      apply ty_or_closure_l; auto.
  - forwards: ty_has_N_closure_or_inv logctx_nil H6.
    destructecs H; simpls.
    + apply ty_or_closure_l. apply ty_or_closure_atomic.
      apply ty_inv_basic_MN; eauto. 
    + apply ty_or_closure_atomic.
      apply ty_inv_basic_MN; eauto.
      apply ty_or_closure_l; auto.
Qed.

Lemma ty_inv_basic_or_prepend G e w T1 T2 T3:
  ty_inv_basic G e w  (typ_or T2 T3) ->
  ty_inv_basic_closure G e w (typ_or (typ_or T1 T2) T3).
Proof.
  introv Hty.
  inversions Hty.
  - apply ty_inv_atomic_atomic in H. inversions H.
  - (* NM *)
      apply ty_or_closure_atomic.
      apply ty_inv_basic_NM; eauto.
      apply ty_or_closure_r; auto.
  - (* MN *)
    apply ty_or_closure_atomic.
    apply ty_inv_basic_MN; eauto.
    apply ty_or_closure_r; auto.
Qed.

Lemma ty_inv_basic_closure_or_assoc G e w T1 T2 T3:
  ty_inv_basic_closure G e w (typ_or T1 (typ_or T2 T3)) ->
  ty_inv_basic_closure G e w (typ_or (typ_or T1 T2) T3).
Proof.
  introv Hty.
  inversions Hty.
  - (* basic *)
    apply ty_inv_basic_or_assoc; eauto.
  - (* l -> T1 *)
    apply ty_or_closure_l. apply ty_or_closure_l. auto.
  - inversions H0.
    + (* basic *)
    apply ty_inv_basic_or_prepend; eauto.
    + (* l -> T2 *)
      apply ty_or_closure_l; auto.
      apply ty_or_closure_r; auto.
    + (* r -> T3 *)
     apply ty_or_closure_r; auto.
Qed.

Lemma ty_inv_basic_closure_or_comm G e w T1 T2:
  inert G ->
  ty_inv_basic_closure G e w (typ_or T1 T2) ->
  ty_inv_basic_closure G e w (typ_or T2 T1).
Proof.
  introv Hin Hty.
  forwards Htyc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin Hty.
  simpls.
  destructecs Htyc.
      apply ty_or_closure_r; auto.
      apply ty_or_closure_l; auto.
      apply ty_or_closure_atomic; apply ty_inv_basic_MN; auto.
      apply ty_or_closure_atomic; apply ty_inv_basic_NM; auto.
Qed.

Lemma ty_inv_basic_lc_pull_or G e w oc T U:
  inert G ->
  ty_inv_basic G e w (logctx_typ oc T) ->
  ty_inv_basic_closure G e w (typ_or (logctx_typ oc U) T).
Proof.
  introv Hin Hty.
  induction oc; simpls.
  +    apply ty_or_closure_r.    apply ty_or_closure_atomic. auto.
  + forwards: ty_inv_basic_lc_or_inv logctx_nil Hin Hty.
    simpls. destructecs H.
    * forwards: IHoc H.
      apply ty_inv_basic_closure_or_assoc.
      eapply ty_inv_basic_closure_or_comm. auto.
      apply ty_inv_basic_closure_or_assoc.
      apply ty_or_closure_r.
      eapply ty_inv_basic_closure_or_comm. auto.
      auto.
    * apply ty_or_closure_l.    apply ty_or_closure_r. apply ty_or_closure_atomic. auto.
    * forwards: ty_has_N_or_closure_lc_cases H0.
      destructecs H2.
      - apply ty_or_closure_atomic. apply ty_inv_basic_MN. auto. apply ty_or_closure_r. auto. auto.
      - apply ty_or_closure_l.  apply ty_or_closure_atomic. apply ty_inv_basic_NM. auto. apply H2. auto.
    * forwards: ty_has_M_or_closure_lc_cases H1.
      destructecs H2.
      - apply ty_or_closure_atomic. apply ty_inv_basic_NM. auto. apply ty_or_closure_r. auto. auto.
      - apply ty_or_closure_l.  apply ty_or_closure_atomic. apply ty_inv_basic_MN. auto. apply H2. auto.
  + forwards: ty_inv_basic_lc_or_inv logctx_nil Hin Hty.
    simpls. destructecs H.
    * apply ty_or_closure_l.    apply ty_or_closure_l. apply ty_or_closure_atomic. auto.
    * forwards: IHoc H.
      apply ty_inv_basic_closure_or_assoc.
      apply ty_or_closure_r. auto.
    * forwards: ty_has_M_or_closure_lc_cases H1.
      destructecs H2.
      - apply ty_or_closure_atomic. apply ty_inv_basic_NM. auto. apply ty_or_closure_l. auto. auto.
      - apply ty_or_closure_l.  apply ty_or_closure_atomic. apply ty_inv_basic_NM. auto. auto. apply H2.
    * forwards: ty_has_N_or_closure_lc_cases H0.
      destructecs H2.
      - apply ty_or_closure_atomic. apply ty_inv_basic_MN. auto. apply ty_or_closure_l. auto. auto.
      - apply ty_or_closure_l.  apply ty_or_closure_atomic. apply ty_inv_basic_MN. auto. auto. apply H2.
  + inversions Hty. apply ty_inv_atomic_atomic in H. inversions H.
  + inversions Hty. apply ty_inv_atomic_atomic in H. inversions H.
Qed.


Lemma ty_inv_basic_closure_lc_pull_or G e w oc T U:
  inert G ->
  ty_inv_basic_closure G e w (logctx_typ oc T) ->
  ty_inv_basic_closure G e w (typ_or (logctx_typ oc U) T).
Proof.
  introv Hin Hty.
  remember  (logctx_typ oc T).
  reverts oc T.
  induction Hty; intros; subst; simpls.
  - eapply ty_inv_basic_lc_pull_or; eauto.
  - destruct oc; simpls; inversions Heqt.
    + apply ty_or_closure_r.  apply ty_or_closure_l. auto.
    + forwards: IHHty. eauto.

      apply ty_inv_basic_closure_or_assoc.
      eapply ty_inv_basic_closure_or_comm. auto.
      apply ty_inv_basic_closure_or_assoc.
      apply ty_or_closure_r.
      eapply ty_inv_basic_closure_or_comm. auto. auto.
    + 
    apply ty_or_closure_l.
    apply ty_or_closure_l. auto.

  - destruct oc; simpls; inversions Heqt.
    + apply ty_or_closure_r.  apply ty_or_closure_r. auto.
    +  apply ty_or_closure_l.  apply ty_or_closure_r. auto.
    + forwards: IHHty. eauto.
      apply ty_inv_basic_closure_or_assoc.
      apply ty_or_closure_r. auto.
  - destruct oc; simpls; inversions Heqt.
    apply ty_or_closure_r.
    apply ty_or_closure_top.
Qed.


Lemma ty_inv_basic_closure_N_also_M_replace G e w oc U:
  mode_top_nm' typing_mode = mode_top_nm_complement ->
  inert G ->
  renv_corr G e ->
  w \in ctx_dom G ->
  ty_inv_basic_closure G e w (logctx_typ oc typ_nocap) ->
  ty_inv_basic_closure G e w (logctx_typ oc (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) ->
  ty_inv_basic_closure G e w (logctx_typ oc U).
Proof.
  introv Hm Hin Hcor Hdom HtyN HtyM.
  induction oc; simpls.
  - (* nil *)
    inversions HtyN. inversions H. apply ty_inv_atomic_atomic in H0. inversions H0.
  - (* orl *)
    forwards HtyNc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyN.
    forwards HtyMc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyM.
    simpls.
    destructecs HtyNc.
    +
      destructecs HtyMc.
      * apply ty_or_closure_l. eapply IHoc; eauto.
      * apply ty_or_closure_r. eauto.
      * forwards: ty_has_N_closure_lc_M_replace HtyMc0; fweauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM. auto. eauto. auto. 
      * eapply ty_inv_basic_closure_lc_pull_or. auto.
        eapply ty_inv_basic_closure_lc_nocap_closure_replace; eauto.
    + apply ty_or_closure_r. eauto.
    +
      destructecs HtyMc.
      * 
        eapply ty_inv_basic_closure_lc_pull_or. eauto.
        eapply ty_inv_basic_closure_lc_M_closure_replace; eauto.
      * apply ty_or_closure_r. eauto.
      * forwards: ty_has_N_closure_lc_M_replace HtyMc0; fweauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM. auto. eauto. auto. 
      * apply ty_or_closure_r.
        eapply ty_has_N_also_M_basic_closure; eauto.
    +
      destructecs HtyMc.
      * eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN. auto.
        forwards: ty_has_M_closure_lc_nocap_replace HtyNc1; fweauto.
        eauto. auto.
      * apply ty_or_closure_r. eauto.
      * forwards: ty_has_N_closure_lc_M_replace HtyMc0; fweauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM. auto. eauto. auto. 
      * eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN. auto.
        forwards: ty_has_M_closure_lc_nocap_replace HtyNc1; fweauto.
        eauto. auto.
  - (* orr *)
    forwards HtyNc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyN.
    forwards HtyMc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyM.
    simpls.
    destructecs HtyNc.
    + apply ty_or_closure_l. eauto.
    +
      destructecs HtyMc.
      * apply ty_or_closure_l. eauto.
      * apply ty_or_closure_r. eapply IHoc; eauto.
      * eapply ty_inv_basic_closure_or_comm; fweauto.
        eapply ty_inv_basic_closure_lc_pull_or; fweauto.
        eapply ty_inv_basic_closure_lc_nocap_closure_replace; eauto.
      *
        forwards: ty_has_N_closure_lc_M_replace HtyMc0; fweauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN. auto. eauto. eauto.
    +
        forwards: ty_has_M_closure_lc_nocap_replace HtyNc1; fweauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM. auto. eauto. eauto.
    +
      destructecs HtyMc.
      * apply ty_or_closure_l. eauto.
      *
        eapply ty_inv_basic_closure_or_comm; fweauto.
        eapply ty_inv_basic_closure_lc_pull_or; fweauto.
        eapply ty_inv_basic_closure_lc_M_closure_replace; eauto.
      * apply ty_or_closure_l.
        eapply ty_has_N_also_M_basic_closure; eauto.
      * forwards: ty_has_N_closure_lc_M_replace HtyMc0; fweauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN. auto. eauto. eauto.
  - eapply ty_inv_basic_closure_lc_and_replace with (oc:=logctx_nil); eauto.
  - eapply ty_inv_basic_closure_lc_and_replace with (oc:=logctx_nil); eauto.
Qed.


Lemma ty_inv_basic_lc_typ_also_dist G e x oc A SL SU TL TU:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU))) ->
  ty_inv_basic G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))).
Proof.
  introv Hin Hcor Hdom HtyT HtyU.

  remember (typ_rcd (dec_typ A TL TU)).
  reverts TL TU A HtyU.
  dependent induction HtyT; intros; subst.
  - forwards Ha: ty_inv_atomic_atomic H. forwards: logctx_atomic_inv Ha. destructecs H0; subst; simpls.
    inversions HtyU.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_atomic.
    eapply ty_inv_atomic_lc_typ_also_dist with (oc:=logctx_nil); eauto.
  - destruct oc; simpls; inversions x.
    forwards Hhc: ty_has_N_closure_lc_typ_replace; fweauto.
    destructecs Hhc.
    +
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_NM; eauto.
    +
    rewrite logctx_fold_orl.
    rewrite logctx_fold_orl in HtyU.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_lc_subtyp. eauto. eauto. eapply ty_has_top_atomic_subtyp_distl; eauto.
    +
    inversions HtyU.
    * false. eapply ty_inv_atomic_lc_or_false with (oc0:=logctx_nil) in H2. auto.
    * eapply ty_or_closure_atomic. eapply ty_inv_basic_NM. eauto. eauto. eapply ty_has_M_closure_lc_typ_dist; eauto.
    * eapply ty_or_closure_l. eapply ty_has_N_also_M_basic_closure; eauto.
  - destruct oc; simpls; inversions x.
    +
    inversions HtyU.
    * false. eapply ty_inv_atomic_lc_or_false with (oc0:=logctx_nil) in H2. auto.
    * eapply ty_or_closure_r. eapply ty_has_N_also_M_basic_closure; eauto.
    * eapply ty_or_closure_atomic. eapply ty_inv_basic_MN. eauto. eapply ty_has_M_closure_lc_typ_dist; eauto. eauto.
    +
    forwards Hhc: ty_has_N_closure_lc_typ_replace; fweauto.
    destructecs Hhc.
    ++
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_MN; eauto.
    ++
    rewrite logctx_fold_orr.
    rewrite logctx_fold_orr in HtyU.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_lc_subtyp. eauto. eauto. eapply ty_has_top_atomic_subtyp_distl; eauto.
Qed.

Lemma ty_inv_atomic_mutability_has_M G e x TL TU:
  mode_has_mutability typing_mode ->
  inert G ->
  x \in ctx_dom G ->
  ty_inv_atomic G e x (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) TL TU)) ->
  ty_has_M G e (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) TL TU)).
Proof.
  introv Hm Hi Hdom Hty.
  remember (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) TL TU)).
  reverts TL TU.
  induction Hty; intros; subst.
  -
  inversions_tyq H1.
  eapply ctx_dom_to_binds in Hdom; rweauto. destructecs Hdom.
  forwards Hpf: pf_mutability; fweauto.
  destructecs Hpf.
  forwards: x_bound_unique H Hpf; subst.
  forwards Hu: pf_inert_unique_tight_bounds H Hpf; fweauto.
  destructecs Hu; subst.
  constructor. intros. eapply subtyp_tyq_t. symmetry. eauto.
  -
  inversions H.
  forwards: IHHty; fweauto.
  eapply ty_has_M_lc_subtyp with (oc:=logctx_nil). eauto. eapply H. constructor; eauto.
Qed.

Lemma ty_has_M_typ_dist G e A TL TU SL SU:
ty_has_M G e (typ_rcd (dec_typ A SL SU)) ->
ty_has_M G e (typ_rcd (dec_typ A TL TU)) ->
ty_has_M G e (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))).
Proof.
  introv HhS HhT.
  inversions HhS.
  inversions HhT.
  constructor. intros. eapply subtyp_or2_t; eauto.
Qed.

Lemma ty_has_top_atomic_typ_dist G e A TL TU SL SU:
ty_has_top_atomic G e (typ_rcd (dec_typ A SL SU)) ->
ty_has_top_atomic G e (typ_rcd (dec_typ A TL TU)) ->
ty_has_top_atomic G e (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))).
Proof.
  introv HhS HhT.
  inversions HhS;
  inversions HhT.
  constructor; intros. eauto. eapply subtyp_or2_t; eauto. eapply subtyp_and2_t; eauto.
Qed.

Lemma ty_has_N_atomic_typ_dist G e A TL TU SL SU:
ty_has_N_atomic G e (typ_rcd (dec_typ A SL SU)) ->
ty_has_N_atomic G e (typ_rcd (dec_typ A TL TU)) ->
ty_has_N_atomic G e (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))).
Proof.
  introv HhS HhT.
  inversions HhS;
  inversions HhT.
  constructor.
Qed.

Lemma ty_has_N_typ_dist G e A TL TU SL SU:
ty_has_N G e (typ_rcd (dec_typ A SL SU)) ->
ty_has_N G e (typ_rcd (dec_typ A TL TU)) ->
ty_has_N G e (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))).
Proof.
  introv HhS HhT.
  inversions HhS;
  inversions HhT.
  - eapply ty_has_N_has_atomic. eapply ty_has_N_atomic_typ_dist; eauto.
  - inversions H; inversions H0.
  - inversions H; inversions H0.
  - eapply ty_has_N_has_top. eapply ty_has_top_atomic_typ_dist; eauto.
Qed.

Lemma ty_inv_basic_has_M_typ_dist G e x A TL TU SL SU:
  mode_has_mutability typing_mode ->
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic G e x (typ_rcd (dec_typ A TL TU)) ->
  ty_has_M G e (typ_rcd (dec_typ A SL SU)) ->
  ty_has_M G e (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))).
Proof.
  introv Hm Hin Hcor Hdom HtyT HtyS.
  lets HtyS': HtyS.
  inversions HtyS.
  inversions HtyT.
  forwards: ty_inv_atomic_mutability_has_M; fweauto.
  eapply ty_has_M_typ_dist; eauto.
Qed.

Lemma ty_inv_basic_has_N_typ_dist G e x A TL TU SL SU:
  mode_has_mutability typing_mode ->
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic G e x (typ_rcd (dec_typ A TL TU)) ->
  ty_has_N G e (typ_rcd (dec_typ A SL SU)) ->
  ty_has_N G e (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))) \/ ty_inv_basic G e x (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))).
Proof.
  introv Hm Hin Hcor Hdom HtyT HtyS.
  lets HtyS': HtyS.
  inversions HtyS.
  - left. eapply ty_has_N_has_atomic. eapply ty_has_N_atomic_lc_typ_replace with (oc:=logctx_nil). eauto.
  - right.
    inversions HtyT.
    inversions H.
    eapply ty_inv_basic_atomic.
    eapply ty_inv_atomic_subtyp. eauto. constructor; intros.
    eapply subtyp_or2_t; eauto.
    eapply subtyp_and2_t. eauto. eapply subtyp_trans_t with (T:=typ_top). eauto. eauto.  
Qed.

Lemma ty_inv_basic_closure_lc_typ_also_has_M_dist G e x oc A SL SU TL TU:
  mode_has_mutability typing_mode ->
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU))) ->
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  (ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))) \/
  ty_has_M_closure G e (logctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))))).
Proof.
  introv Hm Hin Hcor Hdom HtyT HtyS.

  induction oc; simpls.
  - (* nil *)
    inversions HtyT. inversions HtyS.
    right.
    eapply ty_or_closure_atomic.  
    eapply ty_inv_basic_has_M_typ_dist; eauto.
  - (* orl *)
    forwards HtyTc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyT.
    forwards HtySc: ty_has_M_closure_lc_or_inv logctx_nil HtyS.
    simpls.
    destructecs HtySc.
    + destructecs HtyTc.
      *
        forwards Hoc: IHoc; fweauto.
        destructecs Hoc.
        ** left. eapply ty_or_closure_l. eauto.
        ** right. eapply ty_or_closure_l. eauto.
      * left. eapply ty_or_closure_r. eauto.
      * right. eapply ty_or_closure_r. eauto.
      * right. eapply ty_or_closure_l. eapply ty_has_M_closure_lc_typ_dist; eauto.
    + right. eapply ty_or_closure_r. eauto.
  - (* orr *)
    forwards HtyTc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyT.
    forwards HtySc: ty_has_M_closure_lc_or_inv logctx_nil HtyS.
    simpls.
    destructecs HtySc.
    + right. eapply ty_or_closure_l. eauto.
    + destructecs HtyTc.
      * left. eapply ty_or_closure_l. eauto.
      *
        forwards Hoc: IHoc; fweauto.
        destructecs Hoc.
        ** left. eapply ty_or_closure_r. eauto.
        ** right. eapply ty_or_closure_r. eauto.
      * right. eapply ty_or_closure_r. eapply ty_has_M_closure_lc_typ_dist; eauto.
      * right. eapply ty_or_closure_l. eauto.
  - left. eapply ty_inv_basic_closure_lc_and_replace with (oc:=logctx_nil); eauto.
  - left. eapply ty_inv_basic_closure_lc_and_replace with (oc:=logctx_nil); eauto.
Qed.

Lemma ty_inv_basic_closure_lc_typ_also_has_N_dist G e x oc A SL SU TL TU:
  mode_has_mutability typing_mode ->
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU))) ->
  ty_has_N_closure G e (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  (ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))) \/
  ty_has_N_closure G e (logctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))))).
Proof.
  introv Hm Hin Hcor Hdom HtyT HtyS.

  induction oc; simpls.
  - (* nil *)
    inversions HtyT. inversions HtyS.
    forwards: ty_inv_basic_has_N_typ_dist; fweauto.
    destructecs H1.
    right.
    eapply ty_or_closure_atomic. eauto.
    left.
    eapply ty_or_closure_atomic. eauto.
  - (* orl *)
    forwards HtyTc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyT.
    forwards HtySc: ty_has_N_closure_lc_or_inv logctx_nil HtyS.
    simpls.
    destructecs HtySc.
    + destructecs HtyTc.
      *
        forwards Hoc: IHoc; fweauto.
        destructecs Hoc.
        ** left. eapply ty_or_closure_l. eauto.
        ** right. eapply ty_or_closure_l. eauto.
      * left. eapply ty_or_closure_r. eauto.
      * right. eapply ty_or_closure_l. eapply ty_has_N_closure_lc_typ_dist; eauto.
      * right. eapply ty_or_closure_r. eauto.
    + right. eapply ty_or_closure_r. eauto.
  - (* orr *)
    forwards HtyTc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyT.
    forwards HtySc: ty_has_N_closure_lc_or_inv logctx_nil HtyS.
    simpls.
    destructecs HtySc.
    + right. eapply ty_or_closure_l. eauto.
    + destructecs HtyTc.
      * left. eapply ty_or_closure_l. eauto.
      *
        forwards Hoc: IHoc; fweauto.
        destructecs Hoc.
        ** left. eapply ty_or_closure_r. eauto.
        ** right. eapply ty_or_closure_r. eauto.
      * right. eapply ty_or_closure_l. eauto.
      * right. eapply ty_or_closure_r. eapply ty_has_N_closure_lc_typ_dist; eauto.
  - left. eapply ty_inv_basic_closure_lc_and_replace with (oc:=logctx_nil); eauto.
  - left. eapply ty_inv_basic_closure_lc_and_replace with (oc:=logctx_nil); eauto.
Qed.


Lemma subtyp_atomic_typ_dist_swap G e A SL SU TL TU:
  subtyp_atomic G e (typ_rcd (dec_typ A (typ_or SL TL) (typ_and SU TU))) (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU))).
Proof.
  constructor; intros.
  eapply subtyp_or2_t. 
  eapply subtyp_or12_t.
  eapply subtyp_or11_t.
  eapply subtyp_and2_t. 
  eapply subtyp_and12_t.
  eapply subtyp_and11_t.
Qed.

Lemma ty_inv_basic_closure_lc_typ_also_dist G e x oc A SL SU TL TU:
  inert G ->
  renv_corr G e ->
  x \in ctx_dom G ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU))) ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_basic_closure G e x (logctx_typ oc (typ_rcd (dec_typ A (typ_or TL SL) (typ_and TU SU)))).
Proof.
  introv Hin Hcor Hdom HtyT HtyS.

  induction oc; simpls.
  - (* nil *)
    inversions HtyT. inversions HtyS. eapply ty_inv_basic_lc_typ_also_dist with (oc:=logctx_nil); eauto.
  - (* orl *)
    forwards HtyTc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyT.
    forwards HtySc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyS.
    simpls.
    destructecs HtyTc.
    +
      destructecs HtySc.
      * apply ty_or_closure_l. eapply IHoc; eauto.
      * apply ty_or_closure_r. eauto.
      * forwards: ty_has_N_closure_lc_typ_replace; fweauto.
        destruct H.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM. auto. eauto. auto.
        rewrite logctx_fold_orl.
        rewrite logctx_fold_orl in HtyT.
        eapply ty_inv_basic_closure_lc_subtyp. eauto. eapply HtyT.
      eapply ty_has_top_atomic_subtyp_distr; eauto.
      * 
        forwards: ty_inv_basic_closure_lc_typ_also_has_M_dist; fweauto. 
        destruct (mode_has_mutability' typing_mode) eqn: Hmm. auto. typing_mode_contradiction typing_mode.
        destructecs H.
        apply ty_or_closure_l. eauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN; auto.
    + apply ty_or_closure_r. eauto.
    +
      destructecs HtySc.
      *
        forwards: ty_inv_basic_closure_lc_typ_also_has_N_dist; fweauto.
        destruct (mode_has_mutability' typing_mode) eqn: Hmm. auto. typing_mode_contradiction typing_mode.
        destructecs H.
        apply ty_or_closure_l.
        eapply ty_inv_basic_closure_lc_subtyp; fweauto. eapply subtyp_atomic_typ_dist_swap.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM. eauto. eapply ty_has_N_closure_lc_subtyp; fweauto. eapply subtyp_atomic_typ_dist_swap.
        eauto.
      * apply ty_or_closure_r. eauto.
      * forwards: ty_has_N_closure_lc_typ_dist HtySc0 HtyTc0; fweauto. 
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM; auto.
      * apply ty_or_closure_r.
        eapply ty_has_N_also_M_basic_closure; eauto.
    +
      destructecs HtySc.
      *
        forwards: ty_inv_basic_closure_lc_typ_also_has_M_dist; fweauto.
        destruct (mode_has_mutability' typing_mode) eqn: Hmm. auto. typing_mode_contradiction typing_mode.
      destructecs H.
        apply ty_or_closure_l.
        eapply ty_inv_basic_closure_lc_subtyp; fweauto. eapply subtyp_atomic_typ_dist_swap.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN; auto.
        eapply ty_has_M_closure_lc_subtyp; fweauto. eapply subtyp_atomic_typ_dist_swap.
      * apply ty_or_closure_r. eauto.
      *

        eapply ty_or_closure_r.
  eapply ty_has_N_also_M_basic_closure; eauto.
      *
        forwards: ty_has_M_closure_lc_typ_dist HtySc1 HtyTc1; fweauto. 
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN; auto.
  - (* orr *)
    forwards HtyNc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyT.
    forwards HtyMc: ty_inv_basic_closure_lc_or_inv logctx_nil Hin HtyS.
    simpls.
    destructecs HtyNc.
    + apply ty_or_closure_l. eauto.
    +
      destructecs HtyMc.
      * apply ty_or_closure_l. eauto.
      * apply ty_or_closure_r. eapply IHoc; eauto.
      * 
        forwards: ty_inv_basic_closure_lc_typ_also_has_M_dist; fweauto.
        destruct (mode_has_mutability' typing_mode) eqn: Hmm. auto. typing_mode_contradiction typing_mode.
        destructecs H.
        apply ty_or_closure_r. eauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM; eauto.
      *
        forwards: ty_inv_basic_closure_lc_typ_also_has_N_dist; fweauto.
        destruct (mode_has_mutability' typing_mode) eqn: Hmm. auto. typing_mode_contradiction typing_mode.
        destructecs H.
        apply ty_or_closure_r. eauto.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN; eauto.
    + destructecs HtyMc.
      * apply ty_or_closure_l. eauto.
      * 
        forwards: ty_inv_basic_closure_lc_typ_also_has_M_dist; fweauto.
        destruct (mode_has_mutability' typing_mode) eqn: Hmm. auto. typing_mode_contradiction typing_mode.
        destructecs H.
        apply ty_or_closure_r. eapply ty_inv_basic_closure_lc_subtyp; fweauto. eapply subtyp_atomic_typ_dist_swap.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM; eauto.
        eapply ty_has_M_closure_lc_subtyp; fweauto. eapply subtyp_atomic_typ_dist_swap.
      *
        forwards: ty_has_M_closure_lc_typ_dist HtyMc1 HtyNc1; fweauto. 
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_NM; auto.
      * 
        eapply ty_or_closure_l.
  eapply ty_has_N_also_M_basic_closure; eauto.
    +
      destructecs HtyMc.
      * apply ty_or_closure_l. eauto.
      *         forwards: ty_inv_basic_closure_lc_typ_also_has_N_dist; fweauto.
        destruct (mode_has_mutability' typing_mode) eqn: Hmm. auto. typing_mode_contradiction typing_mode.
        destructecs H.
        apply ty_or_closure_r.  eapply ty_inv_basic_closure_lc_subtyp; fweauto. eapply subtyp_atomic_typ_dist_swap.
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN; eauto.
        eapply ty_has_N_closure_lc_subtyp; fweauto. eapply subtyp_atomic_typ_dist_swap.
      * apply ty_or_closure_l.
  eapply ty_has_N_also_M_basic_closure; eauto.
      * 
        forwards: ty_has_N_closure_lc_typ_dist HtyMc0 HtyNc0; fweauto. 
        eapply ty_or_closure_atomic.
        eapply ty_inv_basic_MN. auto. eauto. eauto.
  - eapply ty_inv_basic_closure_lc_and_replace with (oc:=logctx_nil); eauto.
  - eapply ty_inv_basic_closure_lc_and_replace with (oc:=logctx_nil); eauto.
Qed.

Lemma ty_inv_basic_closure_ro_ref G e w y T:
  mode_has_top_nm_subtype typing_mode ->
  inert G ->
  renv_corr G e ->
  binds w y e ->
  ty_has_N_closure G e T ->
  ty_inv_basic_closure G e y T ->
  ty_inv_basic_closure G e w T.
Proof.
  introv Hm Hin Hcor Hbi Hh Hty.
  induction Hty.
  - eapply ty_or_closure_atomic.
    eapply ty_inv_basic_ro_ref; eauto.
    eapply ty_has_N_main_logic. eapply ty_has_N_logic_or; eauto.
  -
    forwards Ho: ty_has_N_closure_or_inv logctx_nil Hh; simpls.
    forwards Hc: ty_inv_basic_closure_has_NM_cases Hty. eauto.
    destruct Hc; destruct Ho.
    1,2,3: eapply ty_or_closure_l; eapply IHHty; auto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_MN; eauto.
  -
    forwards Ho: ty_has_N_closure_or_inv logctx_nil Hh; simpls.
    forwards Hc: ty_inv_basic_closure_has_NM_cases Hty. eauto.
    destruct Hc; destruct Ho.
    1,2,4: eapply ty_or_closure_r; eapply IHHty; auto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_NM; eauto.
  -
    eapply ty_or_closure_top.
Qed.


Lemma ty_inv_basic_closure_tight G e x T:
    inert G ->
    renv_corr G e ->
    x \in ctx_dom G ->
    ty_inv_basic_closure G e x T ->
    G⁏e ⊢# x ∶ T.
Proof.
  introv Hi Hcor Hdom Hty.
  induction Hty; eauto 2 using ty_inv_basic_tight, ty_tight_dom_top.
Qed.


End WithTypingMode.