Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening TypeInduction.
Require Import Orctx Logctx Logpath.
Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityMain LayeredCapabilityBasic LayeredCapabilityLogic.
Require Import LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure.
Require Import LayeredTypingLogic LayeredTypingLogicOrdered WrapAction.

Section WithTypingMode.
Context {typing_mode: Mode}.

(** * Main layer typing *)
(* Figure 19 in "Pure methods for roDOT" *)
Inductive ty_inv_main: ctx -> renv -> var -> typ -> Prop :=
(** [G ⊢l x : T]   #<br>#
    [―――――――――――(LTM-Logic)]   #<br>#
    [G ⊢m x : T]   *)
| ty_inv_main_logic G e x T:
    ty_inv_logic G e x T ->
    ty_inv_main G e x T
(** [G ⊢m x : C[T1^z]]   #<br>#
    [G ⊢! y : {B:T1..T2}}]  #<br>#
    [――――――――――――――(LTM-Sel)]   #<br>#
    [G ⊢m x : C[y.B(z)]   *)
| ty_inv_main_sel G e x oc S1 S2 U y A z:
    ty_inv_main G e x (logctx_typ oc (open z S1)) ->
    G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
    ty_inv_main G e x (logctx_typ oc (typ_sel (avar_f y) A (avar_f z)))
(** [G ⊢m x : C[T^z]]   #<br>#
    [T indep]  #<br>#
    [G ⊢# N <: T^x]   #<br>#
    [――――――――――――――(LTM-Rec)]   #<br>#
    [G ⊢m x : C[μ(T)]   *)
| ty_inv_main_bnd G e x oc T:
    typ_indep_b 0 T ->
    ((mode_has_incap typing_mode) -> (G⁏e ⊢# typ_nocap <: (open x T))) ->
    ty_inv_main G e x (logctx_typ oc (open x T)) ->
    ty_inv_main G e x (logctx_typ oc (typ_bnd T))
(** [G ⊢m x : C[N]]   #<br>#
    [――――――――――――――(LTM-N)]   #<br>#
    [G ⊢m x : C[μ(T)]]   *)
| ty_inv_main_nocap_bnd G e x oc T:
    ty_inv_main G e x (logctx_typ oc typ_nocap) ->
    ty_inv_main G e x (logctx_typ oc (typ_bnd T))
.

Inductive wrap_action_ty_inv_main: ctx -> renv -> var -> wrap_action -> Prop :=
|  wrap_action_ty_inv_main_name G e x y U z A S1 S2 oc:
    inert G ->
    G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
    wrap_action_ty_inv_main G e x (wrap_action_c oc (open z S1) (typ_sel (avar_f y) A (avar_f z)))
| wrap_action_ty_inv_main_bnd G e x oc T:
    typ_indep_b 0 T ->
    ((mode_has_incap typing_mode) -> (G⁏e ⊢# typ_nocap <: (open x T))) ->
    wrap_action_ty_inv_main G e x (wrap_action_c oc (open x T) (typ_bnd T))
| wrap_action_ty_inv_main_nocap_bnd G e x oc T:
    wrap_action_ty_inv_main G e x (wrap_action_c oc typ_nocap (typ_bnd T))
.
Hint Constructors wrap_action_ty_inv_main: core.

Lemma ty_inv_main_plan G e x T:
  inert G ->
  ty_inv_main G e x T ->
  exists wp S,
    ty_inv_logic G e x S /\
    wrap_plan_utyp S wp T /\
    List.Forall (wrap_action_ty_inv_main G e x) wp.
Proof.
  introv Hi Hty.
  induction Hty.
  - exists (nil: wrap_plan) T; splits; eauto. (* constructor. *)
  - specialize (IHHty Hi). destruct_all.
    exists (cons (wrap_action_c (logctx_path oc) (open z S1) (typ_sel (avar_f y) A (avar_f z)) ) x0) x1.
    splits; eauto. econstructor; eauto. constructor.
  - specialize (IHHty Hi). destruct_all.
    exists (cons (wrap_action_c (logctx_path oc) (open x T) (typ_bnd T)) x0) x1.
    splits; eauto. econstructor; eauto. constructor.
  - specialize (IHHty Hi). destruct_all.
    exists (cons (wrap_action_c (logctx_path oc) (typ_nocap) (typ_bnd T)) x0) x1.
    splits; eauto. econstructor; eauto. constructor.
Qed.
Lemma ty_inv_plan_main G e x T wp S:
  ty_inv_main G e x S ->
  wrap_plan_utyp S wp T ->
  List.Forall (wrap_action_ty_inv_main G e x) wp ->
  ty_inv_main G e x T.
Proof.
  introv Hhas Hu Ha.
  induction Hu.
  - eauto.
  - inversions Ha.
    forwards: IHHu; eauto.
    inversions H.
    inversions H2.
    + eapply ty_inv_main_sel; eauto.
    + eapply ty_inv_main_bnd; eauto.
    + eapply ty_inv_main_nocap_bnd; eauto.
Qed.

Lemma ty_inv_main_dom G e x T:
  ty_inv_main G e x T ->
  x \in ctx_dom G.
Proof.
  introv Hty.
  induction Hty; eauto.
  eapply ty_inv_logic_dom; eauto.
Qed.

Lemma ty_inv_main_atomic_logic G e x T:
  ty_inv_main G e x T ->
  typ_atomic T ->
  ty_inv_logic G e x T.
Proof.
  introv Hty Ha.
  induction Hty; eauto.
  all:
  forwards: logctx_atomic_inv Ha; destruct_all; subst; simpls; inversions Ha.
Qed.

Lemma ty_inv_main_atomic_atomic G e x T:
  ty_inv_main G e x T ->
  typ_atomic T ->
  ty_inv_atomic G e x T.
Proof.
  introv Hty Ha.
  eapply ty_inv_basic_atomic_atomic.
  eapply ty_inv_basic_closure_atomic_basic.
  eapply ty_inv_logic_atomic_basic_closure.
  eapply ty_inv_main_atomic_logic.
  all: eauto.
Qed.



Lemma ty_inv_main_tyq G e x U V:
  inert G ->
  renv_corr G e ->
  ty_inv_main G e x U ->
  typing_equiv e U V ->
  ty_inv_main G e x V.
Proof.
  introv Hin Hcor Hty Htyq.
  reverts V.
  induction Hty; intros.
  - eapply ty_inv_main_logic; eauto. eapply ty_inv_logic_tyq; eauto.
  - eapply logctx_tyq_case in Htyq; destruct_all; subst.
    apply tyq_typ_sel_inv in H2.
    destruct H2 as [[v21 [v22 [tnl [Htnl [HU [? Htyq2]]]]]] | HT2]; subst.
    + inversions Htnl.
      destruct (renv_corr_pf_normal_typ_tyq Hin Hcor Htyq2 H).
      destruct H0 as [T21 [T22 [Hpf [Htyq11 Htyq12]]]].
      eapply ty_inv_main_sel.
      apply IHHty. auto. auto. apply logctx_tyq_typ. eauto. eapply Htyq11. eauto.
    + eapply ty_inv_main_sel. eapply IHHty; eauto. apply logctx_tyq_typ. eauto. 2: eauto. eauto.
  - eapply logctx_tyq_case in Htyq; destruct_all; subst.
    apply tyq_typ_bnd_inv in H3; auto.
    destruct H3 as [t21 [HT2 Htyq21]]; subst.

    eapply ty_inv_main_bnd. 
    +
      eapply typ_indep_tyq_open_all_indep; eauto.
    +
      intro Hm. specialize (H0 Hm).
      specialize (Htyq21 x).
      eapply subtyp_trans_t. eauto.
      eauto.
(*      eapply ty_icp_opened_tyq; eauto. *)
    +
      eapply IHHty; eauto.
      eapply logctx_tyq_typ; eauto.
  - 
    eapply logctx_tyq_case in Htyq; destruct_all; subst.
    apply tyq_typ_bnd_inv in H1; auto.
    destruct H1 as [t21 [HT2 Htyq21]]; subst.

    eapply ty_inv_main_nocap_bnd.
    +
      eapply IHHty; eauto.
      eapply logctx_tyq_typ; eauto.
Qed.

Lemma ty_inv_main_lift_or G e x oc S T P:
    inert G ->
    (forall oc2,
      ty_inv_logic G e x (logctx_typ oc2 T) ->
      ty_inv_logic G e x (logctx_typ oc2 S) \/ P
    ) ->
    ty_inv_main_liftable T S ->
    ty_inv_main G e x (logctx_typ oc T) ->
    ty_inv_main G e x (logctx_typ oc S) \/ P.
Proof.
  introv Hin Hlog Hlcnd Hty.
  remember (logctx_typ oc T). (* reverts T.*)
  reverts oc.
  induction Hty; intros; subst.
  - forwards: Hlog. eauto. destruct H0. left. eapply ty_inv_main_logic. eauto. auto. 
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + (* Equal *)
      inversions Hlcnd.
      false (H0 logctx_nil); simpls; eauto.
    +
      destruct x0; simpl in H1; inversions H1.
      inversions Hlcnd.
      false.
      destruct_all.
      eapply (H1 logctx_nil); simpls; eauto.
    +
      false.
      inversions Hlcnd.
      destruct_all.
      eapply (H0 x0); simpls; eauto.
    + (* separate *)
      forwards Hcr: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destructecs Hcr.

      forwards HHty: IHHty. eauto. 2: apply Hcr0. eauto.
      specialize (Hcr1 S).
      destructecs Hcr1.
      rewrite <- Hcr3 in HHty.
      rewrite <- Hcr2.
      destruct HHty.
      left.
      eapply ty_inv_main_sel; eauto.
      auto.
  - (* Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + (* Equal *)
      false.
      inversions Hlcnd.
      destruct_all.
      eapply (H2 logctx_nil); simpls; eauto.
    +
      destruct ocST; simpl in Heqt; inversions Heqt.
      false.
      inversions Hlcnd.
      destruct_all.
      eapply (H3 logctx_nil); simpls; eauto.
    +
      false.
      inversions Hlcnd.
      destruct_all.
      eapply (H2 ocTS); simpls; eauto.
    + (* separate *)
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open x T0).
      destructecs Hcr.

      forwards HHty: IHHty. eauto. eauto. apply Hcr0.
      specialize (Hcr1 S).
      destructecs Hcr1.
      rewrite <- Hcr3 in HHty.
      rewrite <- Hcr2.
      destruct HHty.
      left.
      eapply ty_inv_main_bnd; eauto.
      auto.
  - (* Nocap-Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + (* Equal *)
      false.
      inversions Hlcnd.
      destruct_all.
      eapply (H0 logctx_nil); simpls; eauto.
    +
      destruct x0; simpl in H0; inversions H0.
      false.
      inversions Hlcnd.
      destruct_all.
      eapply (H1 logctx_nil); simpls; eauto.
    +
      false.
      inversions Hlcnd.
      destruct_all.
      eapply (H0 x0); simpls; eauto.
    + (* separate *)
      forwards: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destruct_all.

      forwards HHty: IHHty. eauto. eauto. apply H0.
      destructecs HHty.
      specialize (H1 S).
      destruct_all.
      rewrite <- H3 in HHty.
      rewrite <- H1.
      left.
      eapply ty_inv_main_nocap_bnd; eauto.
      eauto.
Qed.
Lemma ty_inv_main_lift G e x oc S T:
    inert G ->
    (forall oc2,
      ty_inv_logic G e x (logctx_typ oc2 T) ->
      ty_inv_logic G e x (logctx_typ oc2 S)
    ) ->
    ty_inv_main_liftable T S ->
    ty_inv_main G e x (logctx_typ oc T) ->
    ty_inv_main G e x (logctx_typ oc S).
Proof.
  introv Hin Hlog Hlcnd Hty.
  forwards: ty_inv_main_lift_or False Hin Hlcnd Hty.
  intros. left. eapply Hlog; eauto.
  destruct H; easy.
Qed.

Lemma ty_inv_main_lc_bot_replace G e x oc S:
    inert G ->
    ty_inv_main G e x (logctx_typ oc typ_bot) ->
    ty_inv_main G e x (logctx_typ oc S).
Proof.
  intros Hin Hty.
  forwards: ty_inv_main_lift Hin.
  intros. eapply ty_inv_logic_lc_bot_replace. eauto. 
  eauto.
  eauto using ty_inv_main_liftable_bot.
  eauto.
  eauto.
Qed.

Lemma ty_inv_main_lc_nocap_replace G e x oc S:
    inert G ->
    ty_has_N G e S ->
    ty_inv_main G e x (logctx_typ oc typ_nocap) ->
    ty_inv_main G e x (logctx_typ oc S).
Proof.
  intros Hin HhN Hty.
  forwards: ty_inv_main_lift Hin.
  intros. eapply ty_inv_logic_lc_nocap_replace. eauto. 
  eauto.
  eauto.
  eauto using ty_inv_main_liftable_nocap.
  eauto.
  eauto.
Qed.


Lemma ty_inv_main_lc_nocap_logic_replace G e x oc S:
    inert G ->
    ty_has_N_logic G e S ->
    ty_inv_main G e x (logctx_typ oc typ_nocap) ->
    ty_inv_main G e x (logctx_typ oc S).
Proof.
  intros Hin HhN Hty.
  forwards: ty_inv_main_lift Hin.
  intros. eapply ty_inv_logic_lc_nocap_logic_replace. eauto. 
  eauto.
  eauto.
  eauto using ty_inv_main_liftable_nocap.
  eauto.
  eauto.
Qed.

(* Lemma 36 in "Pure methods for roDOT" *)
Lemma ty_inv_main_lc_nocap_main_replace G e x oc S:
    inert G ->
    ty_has_N_main G e S ->
    ty_inv_main G e x (logctx_typ oc typ_nocap) ->
    ty_inv_main G e x (logctx_typ oc S).
Proof.
  intros Hin HhN Hty.
  induction HhN; eauto 2 using ty_inv_main_lc_nocap_logic_replace.
  - forwards: IHHhN; eauto.
    rewrite <- logctx_extend_typ in *.
    eapply ty_inv_main_sel; eauto.
  - forwards: IHHhN; eauto.
    rewrite <- logctx_extend_typ in *.
    eapply ty_inv_main_nocap_bnd; eauto.
Qed.

Lemma ty_inv_main_lc_typ_pfor_replace G e x oc A SL SU TL TU:
  inert G ->
  ty_pf_or_none G x A TL TU ->
  ty_inv_main G e x (logctx_typ oc (typ_rcd (dec_typ A SL SU))) ->
  ty_inv_main G e x (logctx_typ oc (typ_rcd (dec_typ A TL TU)))  \/ ty_has_top_atomic G e (typ_rcd (dec_typ A SL SU)).
Proof.
  intros Hin Hpf Hty.
  forwards: ty_inv_main_lift_or Hin.
  intros. eapply ty_inv_logic_lc_typ_pfor_replace. eauto. 
  eauto.
  eauto.
  eauto using ty_inv_main_liftable_atomic.
  eauto.
  eauto.
Qed.

Lemma invertible_main_bot_false G e x:
    inert G ->
    ty_inv_main G e x typ_bot ->
    False.
Proof.
  introv Hi Hinv.
  inversion Hinv; subst.
  - inversions H. inversions H1. inversions H.
    eapply ty_inv_atomic_lc_bot_false with (oc := logctx_nil); simpls; eauto.
    destruct oc; simpls; discriminate.
  - destruct oc; simpls; discriminate.
  - destruct oc; simpls; discriminate.
  - destruct oc; simpls; discriminate.
Qed.

Lemma invertible_main_nocap_false G e x:
    inert G ->
    ty_inv_main G e x typ_nocap ->
    False.
Proof.
  introv Hi Hinv.
  inversion Hinv; subst.
  - inversions H. inversions H1. inversions H.
    eapply ty_inv_atomic_lc_nocap_false with (oc := logctx_nil); simpls; eauto.
    destruct oc; simpls; discriminate.
  - destruct oc; simpls; discriminate.
  - destruct oc; simpls; discriminate.
  - destruct oc; simpls; discriminate.
Qed.

(* Part of Lemma 34 in "Pure methods for roDOT" *)
Lemma ty_inv_main_lc_and_invr G e x oc T U:
  inert G ->
  ty_inv_main G e x (logctx_typ oc (typ_and T U)) ->
  ty_inv_main G e x (logctx_typ oc U).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc (typ_and T U)).
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic. eapply ty_inv_logic_lc_and_invr; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + discriminate.
    + destruct ocST; simpls; discriminate.
    + destruct ocTS; simpls; inversions Heqt. 
      * eapply IHHty with (T:=(logctx_typ ocTS (open z S1))); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
      * rewrite <- logctx_extend_typ.
        eapply ty_inv_main_sel; eauto.
        rewrite logctx_extend_typ.
        eapply IHHty with (T:=T0); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open z S1).
      destructecs Hcr.
      forwards HcrU0: Hcr1 U0.
      destructecs HcrU0.
      rewrite <- HcrU1.
      eapply ty_inv_main_sel. 2: eauto.
      forwards HHty: IHHty. eauto. eapply Hcr0. rewrite <- HcrU2 in HHty. eauto.
  - (* Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + discriminate.
    + destruct ocST; simpls; discriminate.
    + destruct ocTS; simpls; inversions Heqt. 
      * eapply IHHty with (T0:=(logctx_typ ocTS (open x T))); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
      * rewrite <- logctx_extend_typ.
        eapply ty_inv_main_bnd; eauto.
        rewrite logctx_extend_typ.
        eapply IHHty with (T0:=T1); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open x T).
      destructecs Hcr.
      forwards HcrU: Hcr1 (U).
      destructecs HcrU.
      rewrite <- HcrU0.
      eapply ty_inv_main_bnd.  eauto.  eauto.
      forwards HHty: IHHty. eauto. eapply Hcr0. rewrite <- HcrU1 in HHty. eauto.
  - (* Nocap-Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + discriminate.
    + destruct ocST; simpls; discriminate.
    + destruct ocTS; simpls; inversions Heqt. 
      * eapply IHHty with (T:=(logctx_typ ocTS (typ_nocap))); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
      * rewrite <- logctx_extend_typ.
        eapply ty_inv_main_nocap_bnd; eauto.
        rewrite logctx_extend_typ.
        eapply IHHty with (T:=T1); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (typ_nocap).
      destructecs Hcr.
      forwards HcrU: Hcr1 (U).
      destructecs HcrU.
      rewrite <- HcrU0.
      eapply ty_inv_main_nocap_bnd.  eauto.  eauto.
      forwards HHty: IHHty. eauto. eapply Hcr0. rewrite <- HcrU1 in HHty. eauto.
Qed.

(* Part of Lemma 34 in "Pure methods for roDOT" *)
Lemma ty_inv_main_lc_and_invl G e x oc T U:
  inert G ->
  ty_inv_main G e x (logctx_typ oc (typ_and T U)) ->
  ty_inv_main G e x (logctx_typ oc T).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc (typ_and T U)).
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic. eapply ty_inv_logic_lc_and_invl; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + discriminate.
    + destruct x0; simpls; discriminate.
    + destruct x0; simpls; inversions H1. 

      * rewrite <- logctx_extend_typ.
        eapply ty_inv_main_sel; eauto.
        rewrite logctx_extend_typ.
        eapply IHHty with (U:=T0); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
      * eapply IHHty with (U:=(logctx_typ x0 (open z S1))); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (T).
      destruct_all.
      rewrite <- H4.
      eapply ty_inv_main_sel. 2: eauto.
      forwards: IHHty. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + discriminate.
    + destruct x0; simpls; discriminate.
    + destruct x0; simpls; inversions H2. 
      * rewrite <- logctx_extend_typ.
        eapply ty_inv_main_bnd; eauto.
        rewrite logctx_extend_typ.
        eapply IHHty with (U:=T1); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
      * eapply IHHty with (U:=(logctx_typ x0 (open x T))); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H1 (open x T).
      destruct_all.
      forwards: H3 (T0).
      destruct_all.
      rewrite <- H5.
      eapply ty_inv_main_bnd.  eauto.  eauto.
      forwards: IHHty. eauto. eapply H2. rewrite <- H6 in H8. eauto.
  - (* Nocap-Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + discriminate.
    + destruct x0; simpls; discriminate.
    + destruct x0; simpls; inversions H0. 

      * rewrite <- logctx_extend_typ.
        eapply ty_inv_main_nocap_bnd; eauto.
        rewrite logctx_extend_typ.
        eapply IHHty with (U:=T1); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
      * eapply IHHty with (U:=(logctx_typ x0 (typ_nocap))); eauto.
        rewrite logctx_extend_typ. simpls. eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destruct_all.
      forwards: H1 (T0).
      destruct_all.
      rewrite <- H3.
      eapply ty_inv_main_nocap_bnd.  eauto.  eauto.
      forwards: IHHty. eauto. eapply H0. rewrite <- H4 in H6. eauto.
Qed.

Lemma ty_inv_main_lc_orl G e x oc T U:
  inert G ->
  ty_inv_main G e x (logctx_typ oc T) ->
  ty_inv_main G e x (logctx_typ oc (typ_or T U)).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc T). (* reverts T.*)
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic. eapply ty_inv_logic_lc_orl; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + rewrite logctx_extend_orl. eapply ty_inv_main_sel. 2: eauto.
      rewrite <- logctx_extend_orl. eauto.
    + destruct x0; simpls; inversions H1.
      rewrite logctx_extend_nil_l in *.

      rewrite logctx_extend_orl. eapply ty_inv_main_sel. 2: eauto.
      rewrite <- logctx_extend_orl. eauto.
    +
      rewrite logctx_extend_typ in *.
      rewrite logctx_extend_orl.
      rewrite <- logctx_extend_typ.
      eapply ty_inv_main_sel. 2: eauto.
      rewrite logctx_extend_typ.
      rewrite logctx_extend_typ.
      simpls.
      eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (typ_or T U0).
      destruct_all.
      rewrite <- H4.
      eapply ty_inv_main_sel. 2: eauto.
      forwards: IHHty. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + rewrite logctx_extend_orl. eapply ty_inv_main_bnd. eauto. eauto.
      rewrite <- logctx_extend_orl. eauto.
    + destruct x0; simpls; inversions H2.
      rewrite logctx_extend_nil_l in *.

      rewrite logctx_extend_orl. eapply ty_inv_main_bnd. eauto. eauto.
      rewrite <- logctx_extend_orl. eauto.
    +
      rewrite logctx_extend_typ in *.
      rewrite logctx_extend_orl.
      rewrite <- logctx_extend_typ.
      eapply ty_inv_main_bnd. 2: eauto. eauto.
      rewrite logctx_extend_typ.
      rewrite logctx_extend_typ.
      simpls.
      eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H1 (open x T).
      destruct_all.
      forwards: H3 (typ_or T0 U).
      destruct_all.
      rewrite <- H5.
      eapply ty_inv_main_bnd. eauto. eauto.
      forwards: IHHty. eauto. eapply H2. rewrite <- H6 in H8. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + rewrite logctx_extend_orl. eapply ty_inv_main_nocap_bnd.
      rewrite <- logctx_extend_orl. eauto.
    + destruct x0; simpls; inversions H0.
      rewrite logctx_extend_nil_l in *.

      rewrite logctx_extend_orl. eapply ty_inv_main_nocap_bnd.
      rewrite <- logctx_extend_orl. eauto.
    +
      rewrite logctx_extend_typ in *.
      rewrite logctx_extend_orl.
      rewrite <- logctx_extend_typ.
      eapply ty_inv_main_nocap_bnd.
      rewrite logctx_extend_typ.
      rewrite logctx_extend_typ.
      simpls.
      eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H typ_nocap.
      destruct_all.
      forwards: H1 (typ_or T0 U).
      destruct_all.
      rewrite <- H3.
      eapply ty_inv_main_nocap_bnd.
      forwards: IHHty. eauto. eapply H0. rewrite <- H4 in H6. eauto.

Qed.

Lemma ty_inv_main_lc_orr G e x oc T U:
  inert G ->
  ty_inv_main G e x (logctx_typ oc U) ->
  ty_inv_main G e x (logctx_typ oc (typ_or T U)).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc U). (* reverts T.*)
  reverts oc T U.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic. eapply ty_inv_logic_lc_orr; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst.
    + rewrite logctx_extend_orr. eapply ty_inv_main_sel. 2: eauto.
      rewrite <- logctx_extend_orr. eauto.
    + destruct x0; simpls; inversions H1.
      rewrite logctx_extend_nil_l in *.

      rewrite logctx_extend_orr. eapply ty_inv_main_sel. 2: eauto.
      rewrite <- logctx_extend_orr. eauto.
    +
      rewrite logctx_extend_typ in *.
      rewrite logctx_extend_orr.
      rewrite <- logctx_extend_typ.
      eapply ty_inv_main_sel. 2: eauto.
      rewrite logctx_extend_typ.
      rewrite logctx_extend_typ.
      simpls.
      eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (typ_or T U0).
      destruct_all.
      rewrite <- H4.
      eapply ty_inv_main_sel. 2: eauto.
      forwards: IHHty. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst. 
    + rewrite logctx_extend_orr. eapply ty_inv_main_bnd. eauto. eauto.
      rewrite <- logctx_extend_orr. eauto.
    + destruct x0; simpls; inversions H2.
      rewrite logctx_extend_nil_l in *.

      rewrite logctx_extend_orr. eapply ty_inv_main_bnd. eauto. eauto.
      rewrite <- logctx_extend_orr. eauto.
    +
      rewrite logctx_extend_typ in *.
      rewrite logctx_extend_orr.
      rewrite <- logctx_extend_typ.
      eapply ty_inv_main_bnd. 2: eauto. eauto.
      rewrite logctx_extend_typ.
      rewrite logctx_extend_typ.
      simpls.
      eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H1 (open x T).
      destruct_all.
      forwards: H3 (typ_or T0 U).
      destruct_all.
      rewrite <- H5.
      eapply ty_inv_main_bnd. eauto. eauto.
      forwards: IHHty. eauto. eapply H2. rewrite <- H6 in H8. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destruct_all; subst. 
    + rewrite logctx_extend_orr. eapply ty_inv_main_nocap_bnd.
      rewrite <- logctx_extend_orr. eauto.
    + destruct x0; simpls; inversions H0.
      rewrite logctx_extend_nil_l in *.

      rewrite logctx_extend_orr. eapply ty_inv_main_nocap_bnd.
      rewrite <- logctx_extend_orr. eauto.
    +
      rewrite logctx_extend_typ in *.
      rewrite logctx_extend_orr.
      rewrite <- logctx_extend_typ.
      eapply ty_inv_main_nocap_bnd.
      rewrite logctx_extend_typ.
      rewrite logctx_extend_typ.
      simpls.
      eauto.
    +
      forwards: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destruct_all.
      forwards: H1 (typ_or T0 U).
      destruct_all.
      rewrite <- H3.
      eapply ty_inv_main_nocap_bnd.
      forwards: IHHty. eauto. eapply H0. rewrite <- H4 in H6. eauto.
Qed.

Lemma ty_inv_main_lc_top_replace G e x oc T:
  inert G ->
  ty_inv_main G e x (logctx_typ oc T) ->
  ty_inv_main G e x (logctx_typ oc typ_top).
Proof.
  introv Hin Hty.
  remember (logctx_typ oc T). (* reverts T.*)
  reverts oc T.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic. eapply ty_inv_logic_lc_top_replace; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + eauto.
    + destruct ocST; simpls; inversions Heqt.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open z S1).
      destructecs Hcr.
      forwards Hcrt: Hcr1 typ_top.
      destructecs Hcrt.
      rewrite <- Hcrt0.
      eapply ty_inv_main_sel. 2: eauto.
      forwards HHty: IHHty. eauto. eapply Hcr0. rewrite <- Hcrt1 in HHty. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + eauto.
    + destruct ocST; simpls; inversions Heqt.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open x T).
      destructecs Hcr.
      forwards Hcrt: Hcr1 typ_top.
      destructecs Hcrt.
      rewrite <- Hcrt0.
      eapply ty_inv_main_bnd. 2: eauto. eauto.
      forwards HHty: IHHty. eauto. eapply Hcr0. rewrite <- Hcrt1 in HHty. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + eauto.
    + destruct ocST; simpls; inversions Heqt.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt typ_nocap.
      destructecs Hcr.
      forwards Hcrt: Hcr1 typ_top.
      destructecs Hcrt.
      rewrite <- Hcrt0.
      eapply ty_inv_main_nocap_bnd.
      forwards HHty: IHHty. eauto. eapply Hcr0. rewrite <- Hcrt1 in HHty. eauto.
Qed.

Lemma ty_inv_main_lc_NM_replace G e x oc T T1 T2:
    mode_has_top_nm_subtype typing_mode ->
  inert G ->
    (mode_has_top_nm_complement typing_mode -> forall x0 : var, G ⁏ e ⊢# T1 /[-> x0] <: typ_bot) ->
  ty_inv_main G e x (logctx_typ oc T) ->
  ty_inv_main G e x (logctx_typ oc (typ_NM T1 T2)).
Proof.
  introv Hm Hin Hbot Hty.
  remember (logctx_typ oc T). (* reverts T.*)
  reverts oc T.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic. eapply ty_inv_logic_lc_NM_replace; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + eauto.
    + destruct ocST; simpls; inversions Heqt.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open z S1).
      destructecs Hcr.
      forwards Hcrt: Hcr1 (typ_NM T1 T2).
      destructecs Hcrt.
      rewrite <- Hcrt0.
      eapply ty_inv_main_sel. 2: eauto.
      forwards HHty: IHHty Hbot. eauto. eapply Hcr0. rewrite <- Hcrt1 in HHty. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + eauto.
    + destruct ocST; simpls; inversions Heqt.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open x T).
      destructecs Hcr.
      forwards Hcrt: Hcr1 (typ_NM T1 T2).
      destructecs Hcrt.
      rewrite <- Hcrt0.
      eapply ty_inv_main_bnd. 2: eauto. eauto.
      forwards HHty: IHHty Hbot. eauto. eapply Hcr0. rewrite <- Hcrt1 in HHty. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + eauto.
    + destruct ocST; simpls; inversions Heqt.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt typ_nocap.
      destructecs Hcr.
      forwards Hcrt: Hcr1 (typ_NM T1 T2).
      destructecs Hcrt.
      rewrite <- Hcrt0.
      eapply ty_inv_main_nocap_bnd.
      forwards HHty: IHHty Hbot. eauto. eapply Hcr0. rewrite <- Hcrt1 in HHty. eauto.
Qed.

Lemma ty_inv_main_lc_has_top_replace G e x oc T U:
    inert G ->
 renv_corr G e ->
  ty_inv_main G e x (logctx_typ oc T) ->
  ty_has_top_atomic G e U ->
  ty_inv_main G e x (logctx_typ oc U).
Proof.
  introv Hin Hcor Hty Htop.
  remember (logctx_typ oc T). (* reverts T.*)
  reverts oc T.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic. eapply ty_inv_logic_lc_has_top_replace; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + eauto.
    + destruct ocST; simpls; inversions Heqt.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open z S1).
      destructecs Hcr.
      forwards Hcrt: Hcr1 U.
      destructecs Hcrt.
      rewrite <- Hcrt0.
      eapply ty_inv_main_sel. 2: eauto.
      forwards HHty: IHHty; fweauto. rewrite <- Hcrt1 in HHty. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + eauto.
    + destruct ocST; simpls; inversions Heqt.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open x T).
      destructecs Hcr.
      forwards Hcrt: Hcr1 U.
      destructecs Hcrt.
      rewrite <- Hcrt0.
      eapply ty_inv_main_bnd. 2: eauto. eauto.
      forwards HHty: IHHty; fweauto. rewrite <- Hcrt1 in HHty. eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + eauto.
    + destruct ocST; simpls; inversions Heqt.
      rewrite logctx_extend_nil_l in *.
      eauto.
    +
      rewrite logctx_extend_typ in *. eauto.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt typ_nocap.
      destructecs Hcr.
      forwards Hcrt: Hcr1 U.
      destructecs Hcrt.
      rewrite <- Hcrt0.
      eapply ty_inv_main_nocap_bnd.
      forwards HHty: IHHty; fweauto. rewrite <- Hcrt1 in HHty. eauto.
Qed.



Lemma ty_inv_main_lc_subtyp G e x oc T U:
  inert G ->
  ty_inv_main G e x (logctx_typ oc T) ->
  subtyp_atomic G e T U ->
  ty_inv_main G e x (logctx_typ oc U).
Proof.
  intros Hin Hty Hsu.
  forwards: ty_inv_main_lift Hin.
  intros. eapply ty_inv_logic_lc_subtyp. eauto.
  eauto.
  eauto.
  eapply ty_inv_main_liftable_not_bndsel; intros; eapply subtyp_atomic_atomic_l in Hsu; destruct oc0; simpls; inversions Hsu; discriminate.
  eauto.
  eauto.
Qed.

(*
NOTE: Stating the lemma under a logctx (see below) is not possible, because we can have
x : M | rec (s: s.A)

where rec(s: s.A) >: N,
but we cannot say y.A >: for any y

Lemma ty_inv_main_bnd_open G e x oc T:
  inert G ->
  ty_inv_main G e x (logctx_typ oc (typ_bnd T)) ->
  ty_inv_main G e x (logctx_typ oc (open x T)).
*)

Lemma ty_inv_main_bnd_open G e x T:
  inert G ->
  ty_inv_main G e x (typ_bnd T) ->
  ty_inv_main G e x (open x T).
Proof.
  introv Hi Hty.
  remember (typ_bnd T).
  inversions Hty.
  - forwards: ty_inv_logic_lc_bnd_replace G logctx_nil (open x T); eauto. simpls.
    eapply ty_inv_main_logic; eauto.
  - destruct oc; simpls; inversions H4.
  - destruct oc; simpls; inversions H5.
    eauto.
  - (* nocap_bnd cannot be used at top level, because no variable has typ_nocap type *)
    destruct oc; simpls; inversions H3.
    false. eapply invertible_main_nocap_false; eauto.
Qed.

Lemma ty_inv_main_lc_sel_open G e x oc U TL TU x0 A y :
  inert G ->
  G ⊢! x0 ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
  ty_inv_main G e x (logctx_typ oc (typ_sel (avar_f x0) A (avar_f y))) ->
  ty_inv_main G e x (logctx_typ oc (open y TU)).
Proof.
  introv Hi Hpf Hty.

  remember (logctx_typ oc (typ_sel (avar_f x0) A
              (avar_f y))).
  reverts oc.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic.
    eapply ty_inv_logic_lc_sel_replace; eauto.
  - 
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destructecs Heqt; subst.
    + inversions Heqt.
      lets Hu: (x_bound_unique Hpf H). subst.
      pose proof (pf_inert_unique_tight_bounds Hi Hpf H) as Hu. destruct Hu. subst.
      destruct (pf_dec_typ_inv Hi H) as [He|Hb].
      * subst. assumption.
      * subst. simpls. eapply ty_inv_main_lc_bot_replace; eauto.
    + destruct ocST; simpls; inversions Heqt2.
      contradiction.
    + destruct ocTS; simpls; inversions Heqt2.
      contradiction.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open z S1).
      destructecs Hcr.
      forwards Hcrt: Hcr1 (open y TU).
      destructecs Hcrt.
      rewrite <- Hcrt0.
      eapply ty_inv_main_sel. 2: eauto.
      forwards HHty: IHHty. eauto. eauto. eapply Hcr0. rewrite <- Hcrt1 in HHty. eauto.
 - 
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + inversions H2.
    + destruct x1; simpls; inversions H2.
    + destruct x1; simpls; inversions H2.
    +
      forwards: logctx_separate_cross_r Heqt0 H1 (open x T).
      destruct_all.
      forwards: H3 (open y TU).
      destruct_all.
      rewrite <- H5.
      eapply ty_inv_main_bnd. 2: eauto.
      forwards: IHHty. eauto. eauto. eapply H2. eauto. rewrite H6. eauto.
 - 
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + inversions H0.
    + destruct x1; simpls; inversions H0.
    + destruct x1; simpls; inversions H0.
    +
      forwards: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destruct_all.
      forwards: H1 (open y TU).
      destruct_all.
      rewrite <- H3.
      eapply ty_inv_main_nocap_bnd.
      forwards: IHHty. eauto. eauto. eapply H0. rewrite H4. eauto.
Qed.
Lemma ty_inv_main_lc_sel_open_l G e x oc U TL TU x0 A y :
  inert G ->
  G ⊢! x0 ∶ U ⪼ typ_rcd (dec_typ A TL TU) ->
  ty_inv_main G e x (logctx_typ oc (typ_sel (avar_f x0) A (avar_f y))) ->
  ty_inv_main G e x (logctx_typ oc (open y TL)).
Proof.
  introv Hi Hpf Hty.

  remember (logctx_typ oc (typ_sel (avar_f x0) A
              (avar_f y))).
  reverts oc.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic.
    eapply ty_inv_logic_lc_sel_replace; eauto.
  - 
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + inversions H1.
      lets Hu: (x_bound_unique Hpf H). subst.
      pose proof (pf_inert_unique_tight_bounds Hi Hpf H) as Hu. destruct Hu. subst.
      destruct (pf_dec_typ_inv Hi H) as [He|Hb].
      * subst. assumption.
      * subst. simpls. eapply ty_inv_main_lc_bot_replace; eauto.
    + destruct x1; simpls; inversions H1.
      contradiction.
    + destruct x1; simpls; inversions H1.
      contradiction.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 (open y TL).
      destruct_all.
      rewrite <- H4.
      eapply ty_inv_main_sel. 2: eauto.
      forwards: IHHty. eauto. eauto. eapply H1. rewrite <- H5 in H7. eauto.
 - 
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + inversions H2.
    + destruct x1; simpls; inversions H2.
    + destruct x1; simpls; inversions H2.
    +
      forwards: logctx_separate_cross_r Heqt0 H1 (open x T).
      destruct_all.
      forwards: H3 (open y TL).
      destruct_all.
      rewrite <- H5.
      eapply ty_inv_main_bnd. 2: eauto.
      forwards: IHHty. eauto. eauto. eapply H2. eauto. rewrite H6. eauto.
 - 
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt.
    destruct_all; subst.
    + inversions H0.
    + destruct x1; simpls; inversions H0.
    + destruct x1; simpls; inversions H0.
    +
      forwards: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destruct_all.
      forwards: H1 (open y TL).
      destruct_all.
      rewrite <- H3.
      eapply ty_inv_main_nocap_bnd.
      forwards: IHHty. eauto. eauto. eapply H0. rewrite H4. eauto.
Qed.

Lemma ty_inv_main_lc_logic_and G e x oc T U:
  inert G ->
  ty_inv_logic G e x (logctx_typ oc T) ->
  ty_inv_main G e x (logctx_typ oc U) ->
  ty_inv_main G e x (logctx_typ oc (typ_and T U)).
Proof.
  introv Hi HtyT HtyU.
  remember (logctx_typ oc U).
  reverts oc U.
  induction HtyU; intros; subst.
  -
    eapply ty_inv_main_logic.
    eapply ty_inv_logic_oclc_and with (oc0 := orctx_nil); simpls; eauto.
  -
    lets Heqt1: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt1.
    destruct_all; subst.
    * (* same *) forwards: IHHtyU; eauto.
      rewrite logctx_extend_andr in *.
      eapply ty_inv_main_sel; eauto.
    * destruct x0; simpls; inversions H1. contradiction.
    *
      forwards: IHHtyU; [> eauto..|]. rewrite logctx_extend_typ. eauto.
      rewrite logctx_extend_andr in *.
      rewrite <- logctx_extend_typ in *.
      eapply ty_inv_main_sel; eauto.
    * (* separate *)
      forwards: logctx_separate_cross_r Heqt H0 (open z S1).
      destruct_all.
      forwards: H2 T. destruct_all.
      forwards: H2 (typ_and T U0). destruct_all.

      forwards: IHHtyU Hi.

      2: eapply H1. rewrite <- H5. rewrite H1 in HtyU.
      eapply ty_inv_logic_lc_sel_replace; eauto.
      rewrite H4. eauto. rewrite <- H8 in H10.
      rewrite <- H7.
      eapply ty_inv_main_sel; eauto.
  - (* bnd U *)
    lets Heqt1: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt1.
    destruct_all; subst.
    * forwards: IHHtyU; eauto.
      rewrite logctx_extend_andr in *.
      eapply ty_inv_main_bnd; eauto.
    * destruct x0; simpls; inversions H2. contradiction.
    *

      forwards: IHHtyU; [> eauto..|]. rewrite logctx_extend_typ. eauto.
      rewrite logctx_extend_andr in *.
      rewrite <- logctx_extend_typ in *.
      eapply ty_inv_main_bnd; eauto.
    *
      forwards: logctx_separate_cross_r Heqt H1 (open x T0).
      destruct_all.
      forwards: H3 T. destruct_all.
      forwards: H3 (typ_and T U). destruct_all.

      symmetry in H5.
      forwards: IHHtyU Hi.
      
      2: eapply H2. rewrite <- H6. rewrite H5 in HtyT.
      eapply ty_inv_logic_lc_bnd_replace; eauto.
      rewrite <- H8. rewrite <- H9 in H11.
      eapply ty_inv_main_bnd; eauto.
  - (* nocap_bnd U *)

    lets Heqt1: Heqt.
    eapply logctx_intersect_cases_nonnil in Heqt1.
    destruct_all; subst.
    * forwards: IHHtyU; eauto.
      rewrite logctx_extend_andr in *.
      eapply ty_inv_main_nocap_bnd; eauto.
    * destruct x0; simpls; inversions H0. contradiction.
    *

      forwards: IHHtyU; [> eauto..|]. rewrite logctx_extend_typ. eauto.
      rewrite logctx_extend_andr in *.
      rewrite <- logctx_extend_typ in *.
      eapply ty_inv_main_nocap_bnd; eauto.
    *
      forwards: logctx_separate_cross_r Heqt H typ_nocap.
      destruct_all.
      forwards: H1 T. destruct_all.
      forwards: H1 (typ_and T U). destruct_all.

      symmetry in H3.
      forwards: IHHtyU Hi.
      
      2: eapply H0. rewrite <- H4. rewrite H3 in HtyT.
      eapply ty_inv_logic_lc_bnd_replace; eauto.
      rewrite <- H6. rewrite <- H7 in H9.
      eapply ty_inv_main_nocap_bnd; eauto.
Qed.


Lemma precise_to_invertible_atomic G e x T U:
  inert G ->
  G ⊢! x ∶ T ⪼ U ->
  typ_atomic U ->
  ty_inv_main G e x U.
Proof.
  introv Hi Hpf Ht.
    eapply ty_inv_main_logic.
    eapply ty_inv_logic_or. eapply ctx_binds_dom. eapply pf_binds; eauto.
    eapply ty_or_closure_atomic.
    eapply ty_inv_basic_atomic.
    eapply ty_inv_atomic_precise; eauto.
Qed.



Lemma ty_inv_atomic_main G e x T:
  ty_inv_atomic G e  x T ->
  ty_inv_main G e x T.
Proof.
  introv Hty.
  eapply ty_inv_main_logic.
  eapply ty_inv_logic_or.
  eapply ty_inv_atomic_dom; eauto.
  eapply ty_or_closure_atomic.
  eapply ty_inv_basic_atomic.
  eauto.
Qed.




End WithTypingMode.