Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping TightTypingHelpers Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath LogpathOrder.

Require Import TypeInduction TypLogicInduction.
Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic LayeredCapabilityMain.
Require Import LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure LayeredTypingLogic LayeredTypingMain LayeredTypingLogicOrdered
               LayeredSubtypingMainTree LayeredSubtypingMainTreeOrd LayeredSubtypingMainOrd LayeredSubtypingMainPrep LayeredSubtypingMain LayeredTypingMainHard.

Require Import Lia.

Section WithTypingMode.
Context {typing_mode: Mode}.

Section WithContext.
Context (G: ctx) (e: renv) (x: var).
Context (Hin: inert G) (Hcor: renv_corr G e).

(* Version of ty_inv_main which requires a specific number of steps be performed on the main layer *)
Definition ty_inv_main_count n T :=  x \in ctx_dom G /\ exists U tr, ty_inv_logic G e x U /\ ty_main_tree_linear_sub G e x tr U T /\ (ty_main_tree_folds tr = n).
(* Version of ty_has_N_main which requires a specific number of steps be performed on the main layer *)
Definition ty_has_N_main_count n T := exists U tr, ty_has_N_logic G e U /\  ty_main_tree_linear_sub G e x tr U T /\ (ty_main_tree_folds tr = n).

Lemma ty_inv_main_count_main n T:
  ty_inv_main_count n T ->
  ty_inv_main G e x T.
Proof.
  introv Hty.
  hnf in Hty; destructecs Hty.
  eapply ty_inv_main_tree_linear_sub_main; eauto.
  eapply ty_inv_main_logic; eauto.
Qed.


Lemma ty_inv_main_main_count T:
  ty_inv_main G e x T ->
  exists n,
  ty_inv_main_count n T.
Proof.
  introv Hty.
  forwards Htr: ty_inv_main_inv_tree_linear_sub; eauto.
  destructecs Htr.
  eexists. hnf; splits. eapply ty_inv_main_dom; eauto.
  eexists; eexists; splits; eauto.
Qed.



Lemma ty_main_tree_under_lc_folds_le tr op:
  (ty_main_tree_folds (ty_main_tree_under_lc tr op)) <= ty_main_tree_folds tr.
Proof.
  reverts tr.
  induction op; intros; destruct tr; simpls; auto.
  all: eapply Nat.le_trans; eauto; solve [eapply Nat.le_add_r| eapply Plus.le_plus_r]. 
Qed.



Lemma ty_inv_main_lc_N_cases' n1 n2 oc T:
  mode_has_incap typing_mode ->
  ty_inv_main_count n1 (logctx_typ oc T) ->
  ty_has_N_main_count n2 T ->
  (exists n3, ty_inv_main_count n3 T /\ n3 <= n1 + n2) \/
  (exists n3, ty_inv_main_count n3 (logctx_typ oc (typ_nocap)) /\ n3 <= n1 + n2).
Proof.
  introv Hm Hty Hh.
  (* Trace to the previous layer *)
  hnf in Hty.
  destructecs Hty.
  hnf in Hh.
  destructecs Hh.
  subst.
  forwards HT: ty_main_tree_linear_sub_lc_inv Hty2.
  destructecs HT; subst.
  (* Find a common path through this layer *)
  forwards HT3: ty_main_tree_linear_common HT1 Hh1. eauto. eauto.
  destructecs HT3.
  (* Solve the previous layer *)
  forwards Htyr: ty_inv_logic_main_prep.
  eauto. eauto. eauto.
  eapply ty_main_prep_lc. eauto.

  forwards Hhl: ty_has_N_logic_main_prep; eauto.

  forwards: ty_inv_logic_lc_N_cases G e; eauto.
  (* Apply the path through this layer *)
  eapply logctx_follow_symm in HT.

  destructecs H.


  - left. eexists. split;hnf. splits. auto. eauto.
    eapply Nat.le_trans. eauto. apply -> Nat.add_le_mono_r. eapply ty_main_tree_under_lc_folds_le.
  - forwards Hn: ty_main_tree_linear_sub_refl G e x typ_nocap.
    right. eexists. split. hnf. splits. eauto. exists (logctx_typ ocU typ_nocap). eexists. splits. eauto.
    eapply ty_main_tree_linear_sub_lc_follow_replace; eauto. eauto.
    apply <- Nat.add_le_mono_r.
    rewrite  ty_main_tree_set_under_lc_folds. rewrite Nat.add_comm.
    rewrite <- Nat.add_assoc.
    rewrite ty_main_tree_refl_folds.
    lia.
Qed.

Lemma ty_inv_main_lc_N_cases'' tr oc ocU Uoc T:
  x \in ctx_dom G -> 
  ty_inv_logic G e x (logctx_typ ocU Uoc) ->
  ty_main_tree_linear_sub G e x tr (logctx_typ ocU Uoc) (logctx_typ oc T) ->
  ty_has_N_logic G e Uoc -> 
  logctx_follow ocU oc ->
  (exists n3, ty_inv_main_count n3 T /\ n3 <= (ty_main_tree_folds tr)) \/
  (exists n3, ty_inv_main_count n3 (logctx_typ oc (typ_nocap)) /\ n3 <= (ty_main_tree_folds tr)).
Proof.
  introv Hi Hty Htr Hh Hf.

  (* Trace to the previous layer *)
  forwards HT: ty_main_tree_linear_sub_lc_inv Htr.
  destructecs HT; subst.
  forwards Heq: logctx_follow_eq_inv HT0. eapply logctx_follow_trans;eauto.
  destructecs Heq; subst.
  (* Solve the previous layer *)
  forwards: ty_inv_logic_lc_N_cases; eauto.
  (* Apply the path through this layer *)
  eapply logctx_follow_symm in HT.
  destructecs H.
  - left. eexists. split;hnf. splits. auto.
    eexists. eexists. splits. eauto. eauto. eauto.
    eapply Nat.le_trans. eauto. eapply ty_main_tree_under_lc_folds_le.
  - forwards Hn: ty_main_tree_linear_sub_refl G e x typ_nocap.
    right. eexists. split. hnf. splits. eauto. exists (logctx_typ ocU0 typ_nocap). eexists. splits. eauto.
    eapply ty_main_tree_linear_sub_lc_follow_replace; eauto. eauto.
    apply <- Nat.add_le_mono_r.
    rewrite  ty_main_tree_set_under_lc_folds. rewrite Nat.add_comm.
    apply -> Nat.add_le_mono_l. 
    rewrite ty_main_tree_refl_folds.
    lia.
Qed.




End WithContext.

Lemma ty_inv_main_count_0 G e x T:
    ty_inv_main_count G e  x 0 T ->
    ty_inv_logic G e x T.
Proof.
  introv HtyT.
  hnf in HtyT.
  destructecs HtyT.
  forwards Heq: ty_main_tree_linear_sub_folds_0_inv HtyT2 HtyT.
  subst.
  auto.
Qed.

Lemma ty_inv_main_count_in_dom G e x n T:
  ty_inv_main_count G e  x n T ->
  x \in ctx_dom G.
Proof.
  introv HtyT.
  hnf in HtyT.
  destructecs HtyT.
  auto.
Qed.

Lemma ty_inv_main_count_logic_tree_sub G e x tr U T :
  x \in ctx_dom G  ->
  ty_inv_logic G e x  U ->
  ty_main_tree_linear_sub G e x tr U T ->
  ty_inv_main_count G e x  (ty_main_tree_folds tr) T.
Proof.
  introv Hi Hty Htr.
  hnf.
  split; eauto.
Qed.
Lemma ty_has_N_main_count_logic_tree_sub G e x tr U T :
  ty_has_N_logic G e U ->
  ty_main_tree_linear_sub G e x tr U T ->
  ty_has_N_main_count G e x  (ty_main_tree_folds tr) T.
Proof.
  introv Hty Htr.
  hnf.
  eauto.
Qed.



Lemma ty_inv_main_count_tight G e x n T:
    mode_incap' typing_mode = mode_incap_nocap ->
    inert G ->
    renv_corr G e ->
    ty_inv_main_count G e  x n T ->
    G⁏e ⊢# x ∶ T.
Proof.
  introv Hm Hi Hcor HtyT.

  assert (mode_has_incap typing_mode) as Hmi. {
    unfold mode_has_incap'; rewrite Hm; auto.
  }

  reverts T.
  (* Induction on the complexity of the derivation *)
  induction n using nat_lt_0S_induction; intros; subst.
  --
    forwards: ty_inv_main_count_0. eauto.
    eapply ty_inv_logic_tight; eauto.
  --
    remember (typ_logic_count T) as m.
    reverts T.
    (* Induction on the structure of the type *)
    induction m using nat_lt_0S_induction; intros; subst.
    ++ (* Top-level case *)
      symmetry in Heqm.
      apply <- typ_logic_count_notlog in Heqm.

      hnf in HtyT.
      destructecs HtyT.

      forwards Hord: ty_main_linear_sub_tree_ord. eauto. eauto.
      inversions Hord.
      - eapply ty_inv_logic_tight; eauto.
      - (* Sel *)
        forwards: H0. 2: eapply ty_inv_main_count_logic_tree_sub; eauto. rewrite H3. apply Nat.lt_succ_diag_r.
        eapply ty_sub_t. eauto.
        eapply subtyp_tight_lc. eauto.
      - (* Bnd *)
        (* This is the top-level case, so we can use a typing rule *)
        forwards Hnoc: typ_notlog_logctx_nil Heqm. destructecs Hnoc; subst; simpls.
        forwards Htyc: ty_inv_main_count_logic_tree_sub  HtyT1 H1. eauto.
        forwards: H0 Htyc. rewrite H5. apply Nat.lt_succ_diag_r.
        eapply ty_rec_intro_t. auto. intro Hmc. rewrite Hm in Hmc; discriminate. 2: auto. intros.
        forwards HlT0: H2. auto. destructecs HlT0; subst.
        eapply ty_has_N_main_tight.

        eapply ty_has_N_main_tree_linear_sub_main; eauto.
        eapply ty_has_N_main_logic; auto.
      -
        forwards Htyc: ty_inv_main_count_logic_tree_sub  HtyT1 H1. eauto.
        forwards: H0 Htyc. rewrite H2. apply Nat.lt_succ_diag_r.
        eapply ty_sub_t. eauto.
        eapply subtyp_tight_lc. eauto.
    ++

      hnf in HtyT.
      destructecs HtyT.

      forwards Hord: ty_main_linear_sub_tree_ord. eauto. eauto.
      inversions Hord.
      - eapply ty_inv_logic_tight; eauto.
      - (* Sel *)
        forwards Htyc: ty_inv_main_count_logic_tree_sub  HtyT1 H3. eauto.
        forwards: H0 Htyc. rewrite H5. apply Nat.lt_succ_diag_r. 
        eapply ty_sub_t. eauto.
        eapply subtyp_tight_lc. eauto.
      - (* Bnd *)
        forwards Hl: H4. eauto.
        destructecs Hl.

        forwards Hn: ty_inv_main_lc_N_cases''.
        eauto.
        eauto. eauto. eapply HtyT1. eauto.
        eauto. eauto.
       destructecs Hn.
        + (* Go towards the top-level *)

        assert (G ⁏ e ⊢# x ∶ (open x T0)). {
          eapply H0. rewrite H7. 2: eauto. eapply Lt.le_lt_n_Sm. eauto.
        } 


          eapply ty_tight_lc_replace. 
          eapply ty_rec_intro_t. auto. intro Hmc. rewrite Hm in Hmc; discriminate.  intros.
          eapply ty_has_N_main_tight.
          eapply ty_has_N_main_tree_linear_sub_main; eauto.
          eapply ty_has_N_main_logic; eauto.
          eauto.

          eapply H0.
          2: hnf; split; auto; eexists; eexists; splits. 3: eauto. 3: eauto. rewrite H7. eapply Nat.lt_succ_diag_r.
          eauto.

        +
      
          forwards: H0 Hn0. rewrite H7. eapply Lt.le_lt_n_Sm. eauto.
          eapply ty_sub_t. eauto.
          eapply subtyp_tight_lc. eauto.
      -
        forwards Htyc: ty_inv_main_count_logic_tree_sub  HtyT1 H3. eauto.
        forwards: H0 Htyc. rewrite H4. apply Nat.lt_succ_diag_r.
        eapply ty_sub_t. eauto.
        eapply subtyp_tight_lc. eauto.
Qed.
(* Lemma 37 in "Pure methods for roDOT" *)
Lemma ty_inv_main_tight G e x T:
  mode_incap' typing_mode = mode_incap_nocap ->
  inert G ->
  renv_corr G e ->
  ty_inv_main G e x T ->
  G⁏e ⊢# x ∶ T.
Proof.
  introv Hm Hi Hcor Hty.
  forwards: ty_inv_main_main_count.
  eauto.
  eauto.
  eauto.
  destructecs H.
  eapply ty_inv_main_count_tight; eauto.
Qed.

End WithTypingMode.
