Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath LogpathOrder.

Require Import TypeInduction TypLogicInduction.
Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityLogic LayeredCapabilityMain.
Require Import LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure LayeredTypingLogic LayeredTypingMain LayeredTypingLogicOrdered
               LayeredSubtypingMainTree LayeredSubtypingMainOrd LayeredSubtypingMainTreeOrd LayeredSubtypingMainPrep LayeredSubtypingMain.

Require Import Lia.

(* This file contains "hard" properties of ty_inv_main, making use of the various alternative definitions. *)

Section WithTypingMode.
Context {typing_mode: Mode}.

Section WithContext.
Context (G: ctx) (e: renv) (x: var).
Context (Hin: inert G) (Hcor: renv_corr G e).



(* Analog of ty_inv_logic_ord_logic *)
Lemma ty_inv_main_sub_ord_main n ocl U T:
  ty_inv_main G e x U ->
  ty_sub_main_ord G e x n ocl U T ->
  ty_inv_main G e x T.
Proof.
  introv Hty Hh.
  induction Hh.
  - eauto.
  - eapply ty_inv_main_sel; eauto.
  - eapply ty_inv_main_bnd. eauto. intros.
    forwards HH: H. eauto. destructecs HH.
    eapply ty_has_N_main_tight.
    eapply ty_has_N_main_sub_ord_main with (x0:=x).
    eauto. 2: eauto. eapply ty_has_N_main_logic.  eauto. eauto.
  - eapply ty_inv_main_nocap_bnd; eauto.
Qed.

Lemma ty_main_tree_sub_ord_cap  tr T U:
  ty_main_tree_sub G e x tr T U ->
  (ty_main_tree_bndfolds tr) = 0 ->
  ty_sub_main_ord_cap G e x (ty_main_tree_folds tr) logctx_nil T U.
Proof.
  introv Htr Hbf.
  remember (ty_main_tree_folds tr).
  reverts tr T U.

  induction n using nat_lt_0S_induction; intros; subst.
  - forwards Heq: ty_main_tree_sub_folds_0_inv; eauto.
    subst. eapply ty_sub_main_ord_cap_refl.
  -
  induction Htr; intros; simpls.
  + eapply Nat.lt_irrefl in H. contradiction.
  +
    apply -> Nat.eq_add_0 in Hbf. destruct Hbf.
    forwards Hadd: Nat.add_pos_cases H. destructecs Hadd.
    * forwards HH0: H0 trr   Tr.
      2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. eauto. eauto. eauto.

      forwards: IHHtr1 (*logctx_extend oc (logctx_orl logctx_nil Ur)*).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r. eauto. eauto.

      eapply ty_sub_main_ord_cap_le.
      2: eapply ty_sub_main_ord_cap_or; eauto. constructor.
   * forwards HH0: H0 trl   Tl.
      2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. eauto. lia. eauto.
      rewrite ! Nat.add_0_r in HH0.

      forwards: IHHtr2 (*logctx_extend oc (logctx_orl logctx_nil Ur)*).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l.  eauto.  eauto.
      eapply ty_sub_main_ord_cap_le.
      2: eapply ty_sub_main_ord_cap_or; eauto. constructor.
  +
    apply -> Nat.eq_add_0 in Hbf. destruct Hbf.
    forwards Hadd: Nat.add_pos_cases H. destructecs Hadd.
    * forwards HH0: H0 trr   Tr.
      2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. eauto. eauto. eauto.

      forwards: IHHtr1 (*logctx_extend oc (logctx_orl logctx_nil Ur)*).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r. eauto. eauto.
      eapply ty_sub_main_ord_cap_le.
      2: eapply ty_sub_main_ord_cap_and; eauto. constructor.
    * forwards HH0: H0 trl   Tl.
      2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. eauto. lia. eauto.
      rewrite ! Nat.add_0_r in HH0.
      forwards: IHHtr2 (*logctx_extend oc (logctx_orl logctx_nil Ur)*).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l. eauto. eauto.
      eapply ty_sub_main_ord_cap_le.
      2: eapply ty_sub_main_ord_cap_and; eauto. constructor.
  +
    inversion Hbf.
  +
    replace   (typ_bnd T) with (logctx_typ logctx_nil   (typ_bnd T)).
    2: simpl; auto.
    eapply ty_sub_main_ord_cap_bnd; simpls; eauto. constructor. constructor.
  +
    replace  (typ_sel (avar_f y) A (avar_f z)) with (logctx_typ logctx_nil  (typ_sel (avar_f y) A (avar_f z))).
    2: simpl; auto.
    eapply ty_sub_main_ord_cap_sel; simpls; eauto. constructor. 
Qed.

Lemma ty_main_sub_ord_tree_cap n oc T U:
  ty_sub_main_ord_cap G e x n oc T U ->
  exists tr,
    ty_main_tree_sub G e x tr T U /\
    (ty_main_tree_folds tr) = n /\
    (ty_main_tree_bndfolds tr) = 0.
Proof.
  introv Hty.

  reverts T U oc.
  induction n using nat_lt_induction; intros; subst.
  inversions Hty.
  -
  eexists. splits.
  eapply ty_main_tree_sub_refl. eapply ty_main_tree_refl_folds.  eapply ty_main_tree_refl_bndfolds. 
  -
  forwards IHHty: H. 2: eapply H0. eapply Nat.lt_succ_diag_r.
  destructecs IHHty; subst.

  forwards Hoc: ty_main_tree_sub_lc_inv  IHHty0.
  destructecs Hoc; subst.

  forwards HtrS1: ty_main_tree_sub_refl (open z S1).
  forwards: ty_main_tree_sub_lc.
  eapply ty_main_tree_sub_sel; eauto.
  forwards Htrt: ty_main_tree_sub_trans IHHty0.
  eauto.
  eexists. splits. eauto.
  erewrite ty_main_tree_trans_folds_eq by eauto.
  rewrite ty_main_tree_lc_folds.
  simpl. rewrite ty_main_tree_refl_folds. rewrite Nat.add_1_r. auto.
  erewrite ty_main_tree_trans_bndfolds_eq by eauto.
  rewrite ty_main_tree_lc_bndfolds. rewrite IHHty. simpl. rewrite ty_main_tree_refl_bndfolds. auto.

  -
  (* Main predecessor *)
  forwards IHHty: H. 2: eapply H0. eapply Nat.lt_succ_diag_r.
  destructecs IHHty; subst.

  forwards Hoc: ty_main_tree_sub_lc_inv  IHHty0.
  destructecs Hoc; subst.

  forwards HtrS1: ty_main_tree_sub_refl G e x (open x T0).
  forwards: ty_main_tree_sub_lc.
  eapply ty_main_tree_sub_nocap_bnd; eauto.
  forwards Htrt: ty_main_tree_sub_trans IHHty0.
  eauto.
  eexists. splits. eauto.
  erewrite ty_main_tree_trans_folds_eq by eauto.

  rewrite ty_main_tree_lc_folds.
  simpl.
  rewrite Nat.add_1_r. auto.
  erewrite ty_main_tree_trans_bndfolds_eq by eauto.
  rewrite ty_main_tree_lc_bndfolds. rewrite IHHty. simpl. auto.
Qed.

Lemma ty_has_N_main_tree_linear_sub_main tr T U:
  ty_main_tree_linear_sub G e x tr T U ->
  ty_has_N_main G e T ->
  ty_has_N_main G e U.
Proof.
  introv Hty Hh.
  forwards Hord: ty_main_linear_sub_tree_ord; eauto.
  remember  (ty_main_tree_folds tr).
  reverts tr T U.
  induction n using nat_lt_induction; intros; subst.
  inversions Hord.
  - eauto.
  - forwards: H tr0; eauto. rewrite H2. lia.
    eapply ty_main_linear_sub_tree_ord; eauto.
    eapply ty_has_N_main_name; eauto.
  - forwards: H tr0; eauto. rewrite H4. lia.
    eapply ty_main_linear_sub_tree_ord; eauto.
    eapply ty_has_N_main_bnd; eauto.
    eapply ty_has_N_main_replace_nocap; eauto.
  - forwards: H tr0; eauto. rewrite H1. lia.
    eapply ty_main_linear_sub_tree_ord; eauto.
    eapply ty_has_N_main_bnd; eauto.
Qed.

Lemma ty_has_N_main_tree_sub_main tr T U:
  ty_main_tree_sub G e x tr T U ->
  ty_has_N_main G e T ->
  ty_has_N_main G e U.
Proof.
  introv Hty Hh.

  forwards Hord: ty_main_tree_sub_ord; eauto.
  eapply ty_has_N_main_main_ord; eauto.
Qed.

End WithContext.

(* For unknown reason, if this and the previous lemma are in the same section, then Qed of this lemma fails. *)
Section WithContext.
Context (G: ctx) (e: renv) (x: var).
Context (Hin: inert G) (Hcor: renv_corr G e).


Lemma ty_has_N_main_tree_sub_main_inv tr T U:
  mode_has_incap typing_mode -> (* This probably only works with incap, because without it, anything can be wrapped in typ_bnd which is incap*)
  ty_main_tree_sub G e x tr T U ->
  ty_has_N_main G e U ->
  ty_has_N_main G e T.
Proof.
  introv Hm Hty Hh.
  remember (ty_main_tree_folds tr).
  reverts tr T U.
  induction n using nat_lt_induction; intros; subst.
  reverts tr T. 
  induction Hh; intros.
    - eapply ty_has_N_main_logic; eauto. eapply ty_has_N_logic_main_tree; eauto.
  - pose proof Hty as Hty'.
    eapply ty_main_tree_sub_lc_inv in Hty; destructecs Hty; subst.
    eapply logctx_follow_symm in Hty.
    inversions Hty1.
    + (* notlog *)
      forwards HtrS1: ty_main_tree_sub_refl (open z S1).  
      eapply ty_has_N_main_name; eauto.
      eapply IHHh.
      eauto.       eauto.
      2: eapply ty_main_tree_sub_lc_follow_replace; eauto.
      intros; subst; eauto. eapply H0; eauto.
      eapply Nat.lt_le_trans. eauto.
      eapply Plus.plus_le_reg_l. 
      rewrite Nat.add_comm. rewrite ty_main_tree_set_under_lc_folds.
      eapply Plus.plus_le_compat_r.
      rewrite ty_main_tree_refl_folds. 
      eapply Nat.le_0_l.
    + (* Sel fold *)

      lets Hu: (x_bound_unique H H7). subst.
      pose proof (pf_inert_unique_tight_bounds Hin H H7) as Hu. destruct Hu. subst.

      forwards Htroc2: ty_main_tree_sub_lc_follow_replace Hty'; eauto.

      forwards: IHHh (ty_main_tree_set_under_lc tr tr0 (logctx_path ocU)); intros; subst; eauto. eapply  H0. 2: eauto. 2: eauto. 2: eauto.
      eapply Nat.lt_le_trans. eauto. eapply Nat.le_trans. 2: eauto.
      eapply Plus.plus_le_reg_l. 
      rewrite Nat.add_comm. rewrite ty_main_tree_set_under_lc_folds.
      eapply Plus.plus_le_compat_r.
      forwards: logctx_follow_path_eq; eauto.
      rewrite H3.
      rewrite <- H1. simpls. eapply Nat.le_succ_diag_r.
  - pose proof Hty as Hty'.
    eapply ty_main_tree_sub_lc_inv in Hty; destructecs Hty; subst.
    eapply logctx_follow_symm in Hty.
    inversions Hty1.
    + (* notlog *)
      forwards Htrn: ty_main_tree_sub_refl G e x (typ_nocap).  
      eapply ty_has_N_main_bnd; eauto.
      eapply IHHh.
      eauto.       eauto.
      2: eapply ty_main_tree_sub_lc_follow_replace; eauto.
      intros; subst; eauto. eapply H; eauto.
      eapply Nat.lt_le_trans. eauto.
      eapply Plus.plus_le_reg_l. 
      rewrite Nat.add_comm. rewrite ty_main_tree_set_under_lc_folds.
      eapply Plus.plus_le_compat_r.
      rewrite ty_main_tree_refl_folds. 
      eapply Nat.le_0_l.
    + (* Bnd *)


      eapply H.
      3: eapply ty_main_tree_sub_lc_follow_replace; eauto.
      3: eapply ty_has_N_main_lc_nocap_main_replace. 3: eauto. 3: eapply subtyp_N_has_N_main. 3: eauto. 3: eauto. 3: {
        forwards Hn: H3. eauto. destructecs Hn.
        eapply ty_has_N_main_tight.
        eapply  ty_has_N_main_tree_sub_main; eauto.
        eapply  ty_has_N_main_logic; auto.
      }
     

      2: eauto.

      eapply Plus.plus_lt_reg_l. 
      rewrite Nat.add_comm. rewrite ty_main_tree_set_under_lc_folds.
      eapply Plus.plus_lt_compat_r.
      forwards Hop: logctx_follow_path_eq; eauto. 
      rewrite Hop. rewrite <- H0. simpls. rewrite <- Nat.add_succ_l. eapply Nat.lt_lt_add_r.  eapply Nat.lt_succ_diag_r.  eauto.
    +
      forwards: IHHh; intros; subst.
      4: eapply ty_main_tree_sub_lc_follow_replace; eauto.
      4: eapply  (ty_main_tree_sub_atomic). 4: hnf; split; intros; discriminate. eauto. eauto. eapply H; eauto.
      eapply Nat.lt_le_trans. eauto.
      eapply Plus.plus_le_reg_l. 
      rewrite Nat.add_comm. rewrite ty_main_tree_set_under_lc_folds.
      eapply Plus.plus_le_compat_r.
      forwards Hop: logctx_follow_path_eq; eauto.
      rewrite Hop. rewrite <- H1. simpls. auto. auto.
Qed.


Lemma ty_has_N_main_tree_sub_main_tree_logic tr T U:
  mode_has_incap typing_mode ->
  ty_main_tree_sub G e x tr T U ->
  ty_has_N_main G e U ->
  exists T2 tr2,
  ty_main_prep G e T T2 /\
  ty_main_tree_sub G e x tr2 T2 U /\
  ty_has_N_logic G e T2.
Proof.
  introv Hm Hty Hh.
  remember (ty_main_tree_folds tr).
  reverts tr T U.
  induction n using nat_lt_induction; intros; subst.
  reverts tr T. 
  induction Hh; intros.
  - exists T0 tr. splits. constructor. eauto.
    eapply ty_has_N_logic_main_tree; eauto. (* auto. *)
  - pose proof Hty as Hty'.
    eapply ty_main_tree_sub_lc_inv in Hty; destructecs Hty; subst.
    eapply logctx_follow_symm in Hty.
    inversions Hty1.
    + (* notlog *)
      forwards HtrS1: ty_main_tree_sub_refl G e x (open z S1).  

      forward_block HHh: IHHh.
      { forward.

        eauto.       eauto.
        2: eapply ty_main_tree_sub_lc_follow_replace; eauto.
        intros; subst; eauto. eapply H0; eauto.
        eapply Nat.lt_le_trans. eauto.
        eapply Plus.plus_le_reg_l. 
        rewrite Nat.add_comm. rewrite ty_main_tree_set_under_lc_folds.
        eapply Plus.plus_le_compat_r.
        rewrite ty_main_tree_refl_folds. 
        eapply Nat.le_0_l.
      }
      destructecs HHh.
        
      forwards Hhoc: ty_main_tree_sub_lc_inv HHh1.
      destructecs Hhoc; subst.

      forwards Hioc: ty_main_prep_lc_inv HHh0. destructecs Hioc.
      forwards Heq: logctx_follow_eq_inv Hioc0. eauto using logctx_follow_trans, logctx_follow_symm.
      destructecs Heq; subst.

      {
        exists (logctx_typ oc2 T2). eexists.
        eapply logctx_follow_symm in Hhoc.
        splits. 3: eauto. 2: eapply ty_main_tree_sub_lc_follow_replace. 2: eauto. 2: eauto.
        eapply ty_main_prep_sel. 
        eapply ty_main_prep_lc_replace.
        eapply ty_main_prep_refl.
        eauto.
        eauto.
        eapply ty_main_tree_sub_sel; eauto. 
      }
    + (* Sel fold *)

      lets Hu: (x_bound_unique H H7). subst.
      pose proof (pf_inert_unique_tight_bounds Hin H H7) as Hu. destruct Hu. subst.

      forwards Htroc2: ty_main_tree_sub_lc_follow_replace Hty'; eauto.

      forward_block HHh: IHHh (ty_main_tree_set_under_lc tr tr0 (logctx_path ocU)).

      { forward; intros; subst; eauto.
        eapply  H0. 2: eauto. 2: eauto. 2: eauto.
        eapply Nat.lt_le_trans. eauto. eapply Nat.le_trans. 2: eauto.
        eapply Plus.plus_le_reg_l. 
        rewrite Nat.add_comm. rewrite ty_main_tree_set_under_lc_folds.
        eapply Plus.plus_le_compat_r.
        forwards: logctx_follow_path_eq; eauto.
        rewrite H3.
        rewrite <- H1. simpls. eapply Nat.le_succ_diag_r.
      }
      destructecs HHh.
      forwards HiUoc: ty_main_prep_lc_inv HHh0.
      destructecs HiUoc; subst.
      forwards HtrT0: ty_main_tree_sub_lc_inv HHh1.
      destructecs HtrT0; subst.
      forwards Heq: logctx_follow_eq_inv HtrT1. eauto using logctx_follow_trans, logctx_follow_symm.
      destructecs Heq; subst.
      {
        exists (logctx_typ ocU0 Uoc0). eexists. splits. eauto. 
        eapply ty_main_tree_sub_lc_follow_replace; eauto. eauto using logctx_follow_trans, logctx_follow_symm. eauto. 
      }
  - pose proof Hty as Hty'.
    eapply ty_main_tree_sub_lc_inv in Hty; destructecs Hty; subst.
    eapply logctx_follow_symm in Hty.
    inversions Hty1.
    + (* notlog *)
      forwards Htrn: ty_main_tree_sub_refl G e x (typ_nocap).  
      forward_block HHh: IHHh.
      { forward. 
        eauto.       eauto.
        2: eapply ty_main_tree_sub_lc_follow_replace; eauto.
        intros; subst; eauto. eapply H; eauto.
        eapply Nat.lt_le_trans. eauto.
        eapply Plus.plus_le_reg_l. 
        rewrite Nat.add_comm. rewrite ty_main_tree_set_under_lc_folds.
        eapply Plus.plus_le_compat_r.
        rewrite ty_main_tree_refl_folds. 
        eapply Nat.le_0_l.
      }
      destructecs HHh.
      forwards Hioc: ty_main_prep_lc_inv HHh0.
      destructecs Hioc; subst.
      forwards Hhoc: ty_main_tree_sub_lc_inv HHh1.
      destructecs Hhoc; subst.
      forwards Heq: logctx_follow_eq_inv Hhoc0. eauto using logctx_follow_trans, logctx_follow_symm.
      destructecs Heq; subst.
      {
        exists  (logctx_typ ocU0 Uoc). eexists. splits. 3: eauto.
        eapply ty_main_prep_bnd.
        eapply ty_main_prep_lc_replace. eapply ty_main_prep_refl.
        eauto using logctx_follow_trans, logctx_follow_symm.
        eauto.
        eapply ty_main_tree_sub_lc_follow_replace; eauto. eauto using logctx_follow_trans, logctx_follow_symm.
        inversions Hhoc1.
        eapply ty_main_tree_sub_nocap_bnd.
      }
    + (* Bnd *)

      forward_block HH: H.
      { forward.
        3: eapply ty_main_tree_sub_lc_follow_replace; eauto.
        3: eapply ty_has_N_main_lc_nocap_main_replace.
        3: eauto.
        3: eapply subtyp_N_has_N_main.
        3: eauto.
        3: eauto.
        3: {
          forwards Hln: H3. eauto. destructecs Hln.
          eapply ty_has_N_main_tight. eapply ty_has_N_main_tree_sub_main; eauto. eapply ty_has_N_main_logic. eauto.
        }
        2: eauto.

        eapply Plus.plus_lt_reg_l. 
        rewrite Nat.add_comm. rewrite ty_main_tree_set_under_lc_folds.
        eapply Plus.plus_lt_compat_r.
       
        forwards Hop: logctx_follow_path_eq; eauto.
        rewrite Hop. rewrite <- H0. simpls.  rewrite <- Nat.add_succ_l. eapply Nat.lt_lt_add_r.  eapply Nat.lt_succ_diag_r.  eauto.
      }
      destructecs HH.
      forwards HH0lc: ty_main_prep_lc_inv HH0.
      destructecs HH0lc; subst.
      forwards HH1lc: ty_main_tree_sub_lc_inv HH1.
      destructecs HH1lc; subst.
      forwards Heq: logctx_follow_eq_inv HH1lc0.
      eauto using logctx_follow_trans, logctx_follow_symm.
      destructecs Heq; subst.
      {
        exists (logctx_typ ocU0 Uoc0). eexists.
        splits.
        3: eauto.
        eauto.
        eapply ty_main_tree_sub_lc_follow_replace; eauto. eauto using logctx_follow_trans, logctx_follow_symm.
      }
    +
      forward_block HHh: IHHh.
      { forward; intros; subst.
      4: eapply ty_main_tree_sub_lc_follow_replace; eauto.
      4: eapply  (ty_main_tree_sub_atomic). 4: hnf; split; intros; discriminate. eauto. eauto. eapply H; eauto.
      eapply Nat.lt_le_trans. eauto.
      eapply Plus.plus_le_reg_l. 
      rewrite Nat.add_comm. rewrite ty_main_tree_set_under_lc_folds.
      eapply Plus.plus_le_compat_r.
      forwards Hop: logctx_follow_path_eq; eauto.
      rewrite Hop. rewrite <- H1. simpls. auto.
      }
      destructecs HHh.
      forwards HH0lc: ty_main_prep_lc_inv HHh0.
      destructecs HH0lc; subst.
      forwards HH1lc: ty_main_tree_sub_lc_inv HHh1.
      destructecs HH1lc; subst.
      forwards Heq: logctx_follow_eq_inv HH1lc0.
      eauto using logctx_follow_trans, logctx_follow_symm.
      destructecs Heq; subst.
      {
        exists (logctx_typ ocU0 Uoc). eexists.
        splits.
        3: eauto.
        eauto.
        eapply ty_main_tree_sub_lc_follow_replace; eauto. eauto using logctx_follow_trans, logctx_follow_symm.
        inversions HH1lc1.
        eapply ty_main_tree_sub_nocap_bnd; eauto.
      }
Qed. 

Lemma ty_has_N_main_main_tree_logic T:
  mode_has_incap typing_mode ->
  ty_has_N_main G e T ->
  exists T2 tr2,
  ty_main_prep G e T T2 /\
  ty_main_tree_sub G e x tr2 T2 T /\
  ty_has_N_logic G e T2.
Proof.
  intros.
  forwards HtrT: ty_main_tree_sub_refl T.
  forwards: ty_has_N_main_tree_sub_main_tree_logic; eauto.
Qed.



Lemma ty_sub_main_ord_cap_lc_sel n U1 oc y A z S1 S2 U:
  ty_sub_main_ord_cap G e x n logctx_nil U1 (logctx_typ oc (open z S1)) ->
  G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
  ty_sub_main_ord_cap G e x (S n) logctx_nil U1 (logctx_typ oc (typ_sel (avar_f y) A (avar_f z))).
Proof.
  introv Hsub Hpf.
  forwards Htr: ty_sub_main_ord_cap_tree Hsub; eauto.

  destructecs Htr.
  forwards Htro: ty_main_tree_sub_lc_inv Htr0.
  destructecs Htro; subst.
  forwards Hsel: ty_main_tree_sub_sel Htro1 Hpf.
  forwards Hoc: ty_main_tree_sub_lc_follow_replace Htr0 Hsel. eapply logctx_follow_symm; eauto.
  forwards Hord: ty_main_tree_sub_ord_cap Hoc; eauto.
  simpl. erewrite logctx_follow_path_eq by eauto. rewrite ty_main_tree_set_sel_bndfolds. auto.
  erewrite logctx_follow_path_eq in Hord by eauto. 
  rewrite ty_main_tree_set_sel_folds in Hord.
  auto.
Qed.

Lemma ty_sub_main_ord_cap_lc_bnd n U1 oc T:
  ty_sub_main_ord_cap G e x n logctx_nil U1 (logctx_typ oc typ_nocap) ->
  exists m,
  ty_sub_main_ord_cap G e x m logctx_nil U1 (logctx_typ oc (typ_bnd T)) /\ m <= (S n).
Proof.
  introv Hsub.
  forwards Htr: ty_sub_main_ord_cap_tree Hsub; eauto.

  destructecs Htr.
  forwards Htro: ty_main_tree_sub_lc_inv Htr0.
  destructecs Htro; subst.
  forwards Hsel: ty_main_tree_sub_nocap_bnd G e x T.
  forwards HUoc: ty_main_tree_sub_trans Htro1 Hsel.
  forwards Hoc: ty_main_tree_sub_lc_follow_replace Htr0 HUoc. eapply logctx_follow_symm; eauto.
  forwards Hord: ty_main_tree_sub_ord_cap Hoc; eauto.
  simpl.  erewrite logctx_follow_path_eq by eauto.
  eapply ty_main_tree_set_under_lc_bndfolds_0. auto. simpls. auto.

  eexists; splits.
  eauto.

  apply <- Nat.add_le_mono_r.

rewrite ty_main_tree_set_under_lc_folds.
rewrite Nat.add_comm.
rewrite Nat.add_succ_l.
rewrite <- Nat.add_succ_r.
apply -> Nat.add_le_mono_l.
 simpls. apply le_n_S. apply Nat.le_0_l.
Qed.


Lemma ty_main_tree_common tr1 tr2 T1 T2 U:
  mode_has_incap typing_mode -> (* Required, because without it, we cannot merge derivations of typ_bnd U from open U and from typ_nocap *)
  ty_main_tree_sub G e x tr1 T1 U ->
  ty_main_tree_sub G e x tr2 T2 U ->
  exists tr3 T3,
    ty_main_prep G e T1 T3 /\
    ty_main_prep G e T2 T3 /\
    ty_main_tree_sub G e x tr3 T3 U.
Proof.
  introv Hm Htr1 Htr2.
  reverts tr1 T1.
  induction Htr2; intros.
  - hnf in H. destructecs H.
    inversions Htr1.
    + eexists. exists T. splits; eauto using ty_main_prep_refl.
    + false H0; eauto.
    + false H; eauto.
    + eexists. exists T1. splits. eapply ty_main_prep_refl. 2: eapply ty_main_tree_sub_bnd; eauto. eapply ty_main_prep_bnd with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl.
    + eexists. exists typ_nocap. splits. eapply ty_main_prep_refl. 2: eapply ty_main_tree_sub_nocap_bnd; eauto. eapply ty_main_prep_bnd with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl.
    + eexists. exists T1. splits. eapply ty_main_prep_refl. 2: eapply ty_main_tree_sub_sel; eauto. eapply ty_main_prep_sel with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl.
  - eapply ty_main_tree_sub_or_inv in Htr1; destruct_all; subst.
    forwards: IHHtr2_1; eauto.
    forwards: IHHtr2_2; eauto.
    destruct_all.
    eauto 8.
  - eapply ty_main_tree_sub_and_inv in Htr1; destruct_all; subst.
    forwards: IHHtr2_1; eauto.
    forwards: IHHtr2_2; eauto.
    destruct_all.
    eauto 8.
  - (* U is (typ_bnd U) derived from (open x U) in tr2 *)
    inversions Htr1.
    + (* U is the same in tr1 *)
      eexists. exists T. splits. 3: eapply ty_main_tree_sub_bnd; eauto. eapply ty_main_prep_bnd with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl.  eapply ty_main_prep_refl.  
    + (* both trees agree on (open x U) *)
      forwards: IHHtr2 H2; eauto.
      destruct_all.
      eauto 7.
    + (* in tr1, (typ_bnd U) is derived from typ_nocap, while in tr2 from (open x U) *)
   
        forwards: H. auto.
        destructecs H1.
        forwards: subtyp_N_has_N_main.
        eauto.
        eauto.
        eapply ty_has_N_main_tight.


        eapply ty_has_N_main_tree_sub_main. eauto. eauto. eauto. eauto. eapply ty_has_N_main_logic; eauto.

        forwards Heel: ty_has_N_main_tree_sub_main_tree_logic Htr2; [>eauto..|].
        destructecs Heel.
        exists (ty_main_tree_bnd tr2 trn) T2.
        splits; eauto.
        eapply ty_main_prep_nocap with (oc := logctx_nil).
        simpls. eapply ty_main_prep_refl. eauto.
  - (* U is (typ_bnd U) derived from typ_nocap in tr2 *)
    inversions Htr1.
    + (* U is the same in tr1 *)
      eexists. exists (typ_nocap). splits. 3: eapply ty_main_tree_sub_nocap_bnd; eauto. eapply ty_main_prep_bnd with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl.  eapply ty_main_prep_refl.  
    + 
        forwards: H1. auto.
        destructecs H.
        forwards: subtyp_N_has_N_main.
        eauto.
        eauto.
        eapply ty_has_N_main_tight.
        eapply ty_has_N_main_tree_sub_main. eauto. eauto. eauto. eauto. eapply ty_has_N_main_logic; eauto.



        forwards Heel: ty_has_N_main_tree_sub_main_tree_logic H0; [>eauto..|].
        destructecs Heel.
        exists (ty_main_tree_bnd tr2 trn) T2.
        splits; eauto.
        eapply ty_main_prep_nocap with (oc := logctx_nil).
        simpls. eapply ty_main_prep_refl. eauto.
    + eexists. exists typ_nocap. splits; eauto using ty_main_prep_refl.
  - (* Sel fold *)
    inversions Htr1.
    + (* U is the same in tr1 *)
      eexists. exists T. splits.   3: eapply ty_main_tree_sub_sel; eauto.   eapply ty_main_prep_sel with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl. eapply ty_main_prep_refl.  
    + (* Sel fold *)
      lets Hu: (x_bound_unique H H6). subst.
      pose proof (pf_inert_unique_tight_bounds Hin H H6) as Hu. destruct Hu. subst.
   

      forwards: IHHtr2 T1; eauto.
      destructecs H0; eauto. eexists. exists T3. splits; eauto.
Qed.

Lemma ty_has_N_logic_main_prep T U:
  ty_has_N_logic G e T ->
  ty_main_prep G e T U ->
  ty_has_N_logic G e U.
Proof.
  introv Hty Hi.
  induction Hi; eauto.
  -
    forwards: IHHi; eauto.
    eapply ty_has_N_logic_lc_nocap_logic_replace; eauto.
  -
    forwards: IHHi; eauto.
    eapply ty_has_N_logic_lc_sel_replace; eauto.
  -
    forwards: IHHi; eauto.
    eapply ty_has_N_logic_lc_bnd_replace; eauto.
Qed.


Lemma ty_inv_logic_main_prep T U:
  ty_inv_logic G e x T ->
  ty_main_prep G e T U ->
  ty_inv_logic G e x U.
Proof.
  introv  Hty Hi.
  induction Hi; eauto.
  -
    forwards: IHHi; eauto.
    eapply ty_inv_logic_lc_nocap_logic_replace; eauto.
  -
    forwards: IHHi; eauto.
    eapply ty_inv_logic_lc_sel_replace; eauto.
  -
    forwards: IHHi; eauto.
    eapply ty_inv_logic_lc_bnd_replace; eauto.
Qed.

(* Finds a common derivation of U from two types T1 and T2 on the main layer.
   The common derivation will go from T3 to U on the main layer,
   and the paths from T1 to T3 and from T2 to T3 are moved to the logic layer.

   The derivations are in the form of a tree, so they can be compared according to the structure of the type *)
Lemma ty_main_tree_linear_common tr1 tr2 T1 T2 U:
  mode_has_incap typing_mode -> (* Required, because without it, we cannot merge derivations of typ_bnd U from open U and from typ_nocap *)
  ty_main_tree_linear_sub G e x tr1 T1 U ->
  ty_main_tree_linear_sub G e x tr2 T2 U ->
  exists tr3 T3,
    ty_main_prep G e T1 T3 /\
    ty_main_prep G e T2 T3 /\
    ty_main_tree_linear_sub G e x tr3 T3 U /\
    (ty_main_tree_folds tr3 <= (ty_main_tree_folds tr1) + (ty_main_tree_folds tr2)).
Proof.
  introv Hm Htr1 Htr2.
  reverts tr1 T1.
  (* Induction on the derivation tree T2 -> U *)
  induction Htr2; intros.
  - (* atomic - T2 = U *)
    hnf in H. destructecs H.
    (* Invert the second derivation, and move the inverse of that action to the previous layer, then use the second derivation on the main layer *)
    inversions Htr1.
    + (* atomic *) eexists. exists T. splits; eauto using ty_main_prep_refl.
    + (* or *) false H0; eauto.
    + (* and *) false H; eauto.
    + (* bnd *)
       eexists. exists T1. splits.
       1: eapply ty_main_prep_refl.
       2: eapply ty_main_tree_linear_sub_bnd; eauto.
       eapply ty_main_prep_bnd with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl.
       eapply Nat.le_add_r.
    + eexists. exists typ_nocap. splits. eapply ty_main_prep_refl. 2: eapply ty_main_tree_linear_sub_nocap_bnd; eauto. eapply ty_main_prep_bnd with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl.
       eapply Nat.le_add_r.
    + eexists. exists T1. splits. eapply ty_main_prep_refl. 2: eapply ty_main_tree_linear_sub_sel; eauto. eapply ty_main_prep_sel with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl.
      eapply Nat.le_add_r.

  - eapply ty_main_tree_linear_sub_or_inv in Htr1.
    destructecs Htr1; subst.
    forwards HHtr2_1: IHHtr2_1; eauto.
    forwards HHtr2_2: IHHtr2_2; eauto.
    destructecs HHtr2_1.
    destructecs HHtr2_2.
    exists (ty_main_tree_or tr3 tr0) (typ_or T3 T0); splits; eauto.
    simpls. rewrite Nat.add_shuffle1. 
    eapply Nat.add_le_mono; eauto.
  - eapply ty_main_tree_linear_sub_and_inv in Htr1.
    destructecs Htr1; subst.
    forwards HHtr2_1: IHHtr2_1; eauto.
    forwards HHtr2_2: IHHtr2_2; eauto.
    destructecs HHtr2_1.
    destructecs HHtr2_2.
    exists (ty_main_tree_and tr3 tr0) (typ_and T3 T0); splits; eauto.
    simpls. rewrite Nat.add_shuffle1. 
    eapply Nat.add_le_mono; eauto.
  - (* U is (typ_bnd U) derived from (open x U) in tr2 *)
    inversions Htr1.
    + (* U is the same in tr1 *)
      eexists. exists T. splits. 3: eapply ty_main_tree_linear_sub_bnd; eauto. eapply ty_main_prep_bnd with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl.  eapply ty_main_prep_refl. rewrite Nat.add_comm.   eapply Nat.le_add_r.
    + (* both trees agree on (open x U) *)
      forwards HHtr2: IHHtr2 H2; eauto.
      destructecs HHtr2.
      eexists. eexists. splits; eauto. eapply ty_main_tree_linear_sub_bnd; eauto.
      intro. eapply ty_has_N_logic_main_prep; eauto.
      simpls. eapply le_n_S. rewrite <- Nat.add_succ_r.  rewrite Nat.add_shuffle1.
      eapply Nat.add_le_mono; eauto.    eapply Nat.le_add_r.

    + (* in tr1, (typ_bnd U) is derived from typ_nocap, while in tr2 from (open x U) *)


      forwards HH: H. auto.


      eexists; eexists; splits.
      3: eapply ty_main_tree_linear_sub_bnd.
      eapply ty_main_prep_nocap with (oc := logctx_nil).
      simpls. eapply ty_main_prep_refl. eauto.
      simpls. constructor. simpls. eauto. intro. simpls. eauto. eauto.
      simpls.
      eapply Nat.le_succ_diag_r.



  - (* U is (typ_bnd U) derived from typ_nocap in tr2 *)
    inversions Htr1.
    + (* U is the same in tr1 *)
      eexists. exists (typ_nocap). splits. 3: eapply ty_main_tree_linear_sub_nocap_bnd; eauto. eapply ty_main_prep_bnd with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl.  eapply ty_main_prep_refl.  
    simpls. auto.
    + 

        forwards: H1. auto.


        eexists; eexists; splits.
        3: eapply ty_main_tree_linear_sub_bnd.
        eapply ty_main_prep_refl.
        eapply ty_main_prep_nocap with (oc := logctx_nil).
        simpls. eapply ty_main_prep_refl. eauto.
        eauto. eauto. eauto.
        simpls. apply -> Nat.succ_le_mono.  rewrite Nat.add_succ_r. rewrite ! Nat.add_0_r.  eapply Nat.le_succ_diag_r.

    + eexists. exists typ_nocap. splits; eauto using ty_main_prep_refl.
  - (* Sel fold *)
    inversions Htr1.
    + (* U is the same in tr1 *)
      eexists. exists T. splits.   3: eapply ty_main_tree_linear_sub_sel; eauto.   eapply ty_main_prep_sel with (oc:=logctx_nil). simpls. eapply ty_main_prep_refl. eapply ty_main_prep_refl.
      simpls.  eapply Nat.le_refl.
    + (* Sel fold *)
      lets Hu: (x_bound_unique H H6). subst.
      pose proof (pf_inert_unique_tight_bounds Hin H H6) as Hu. destruct Hu. subst.
   

      forwards: IHHtr2 T1; eauto.
      destructecs H0; eauto. eexists. exists T3. splits; eauto.
      simpls.
      apply -> Nat.succ_le_mono. 
      rewrite Nat.add_succ_r.
      eapply Nat.le_le_succ_r. eauto.
Qed.


Lemma ty_main_tree_tree_linear tr1 T U:
  ty_main_tree_sub G e x tr1 T U ->
  exists tr2 T2, 
  ty_main_prep G e T T2 /\
  ty_main_tree_linear_sub G e x tr2 T2 U.
Proof.
  introv Htr.
  remember (ty_main_tree_folds tr1) as n.
  reverts tr1 T U.
  induction n using nat_lt_induction; intros; subst.
  induction Htr.
  - (* refl *)
    eexists. exists T;splits. constructor. eapply ty_main_tree_linear_sub_refl.
  - (* or *)

    destruct (Nat.eq_0_gt_0_cases (ty_main_tree_folds trl));
    destruct (Nat.eq_0_gt_0_cases (ty_main_tree_folds trr)).
    +
    forwards: ty_main_tree_sub_folds_0_inv Htr1 H0.
    forwards: ty_main_tree_sub_folds_0_inv Htr2 H1.
    subst.
    eexists. exists (typ_or Tl Tr); splits. constructor. eapply ty_main_tree_linear_sub_refl.
    +
    forwards: IHHtr2. intros.
    eapply H. 2: eauto. simpls. lia. eauto.
    destructecs H2.

    forwards Hl: H Htr1. 2: eauto. simpls. lia.  destructecs Hl.
    eexists. exists (typ_or T0 T2); splits. eapply ty_main_prep_or; eauto. constructor; eauto.
    +
    forwards: IHHtr1. intros.
    eapply H. 2: eauto. simpls. lia. eauto.
    destructecs H2.

    forwards Hl: H Htr2. 2: eauto. simpls. lia.  destructecs Hl.
    eexists. exists (typ_or T2 T0); splits. eapply ty_main_prep_or; eauto. constructor; eauto.


    +

    forwards Hl: H Htr1. 2: eauto. simpls. lia.  destructecs Hl.
    forwards Hr: H Htr2. 2: eauto. simpls. lia.  destructecs Hr.
    eexists. exists (typ_or T2 T0); splits. eapply ty_main_prep_or; eauto. constructor; eauto.

  - (* and *)

    destruct (Nat.eq_0_gt_0_cases (ty_main_tree_folds trl));
    destruct (Nat.eq_0_gt_0_cases (ty_main_tree_folds trr)).
    +
    forwards: ty_main_tree_sub_folds_0_inv Htr1 H0.
    forwards: ty_main_tree_sub_folds_0_inv Htr2 H1.
    subst.
    eexists. exists (typ_and Tl Tr); splits. constructor. eapply ty_main_tree_linear_sub_refl.
    +
    forwards: IHHtr2. intros.
    eapply H. 2: eauto. simpls. lia. eauto.
    destructecs H2.

    forwards Hl: H Htr1. 2: eauto. simpls. lia.  destructecs Hl.
    eexists. exists (typ_and T0 T2); splits. eapply ty_main_prep_and; eauto. constructor; eauto.
    +
    forwards: IHHtr1. intros.
    eapply H. 2: eauto. simpls. lia. eauto.
    destructecs H2.

    forwards Hl: H Htr2. 2: eauto. simpls. lia.  destructecs Hl.
    eexists. exists (typ_and T2 T0); splits. eapply ty_main_prep_and; eauto. constructor; eauto.


    +

    forwards Hl: H Htr1. 2: eauto. simpls. lia.  destructecs Hl.
    forwards Hr: H Htr2. 2: eauto. simpls. lia.  destructecs Hr.
    eexists. exists (typ_and T2 T0); splits. eapply ty_main_prep_and; eauto. constructor; eauto.

  - (* bnd *)

    destruct ( mode_has_incap' typing_mode ) eqn: Hm.
    *
      forwards HH: H0; eauto.
      destructecs HH.

      forwards HtrT: H Htr. 2: eauto. simpls. lia.
      forwards HtrUn: H HH. 2: eauto. simpls. lia.
      destructecs HtrT.
      destructecs HtrUn.
      forwards Htrc: ty_main_tree_linear_common HtrT HtrUn. eauto.
      destructecs Htrc.    
      eexists.    eexists. splits.
      eapply ty_main_prep_trans; eauto.
      eapply ty_main_tree_linear_sub_bnd. eapply Htrc2.
      intros. eapply ty_has_N_logic_main_prep; eauto. eapply ty_main_prep_trans; eauto. eauto.
    *
      forwards HtrT: H Htr. 2: eauto. simpls. lia.
      destructecs HtrT.

      eexists.    eexists. splits. eauto.
      eapply ty_main_tree_linear_sub_bnd. eapply HtrT.
      intros. rewrite Hm in H2. discriminate. eauto.
  - (* bnd *)
    eexists.    eexists. splits.
    2: eapply ty_main_tree_linear_sub_nocap_bnd.
    constructor.
  - (* bnd *)
    forwards HtrT: H Htr. 2: eauto. simpls. lia.
    destructecs HtrT.
    eexists.    eexists. splits.
    eauto.
    eapply ty_main_tree_linear_sub_sel. eapply HtrT. eauto.
Qed.




Lemma ty_inv_main_lc_tree_sub_main oc tr T U:
  ty_inv_main G e x (logctx_typ oc T) ->
  ty_main_tree_sub G e x tr T U ->
  ty_inv_main G e x (logctx_typ oc U).
Proof.
  introv Hty Htr.
  remember (ty_main_tree_folds tr).
  reverts oc tr T U.
  induction n using nat_lt_0S_induction; intros; subst.
  - forwards: ty_main_tree_sub_folds_0_inv; eauto. subst. eauto.
  -
  reverts oc. 
  induction Htr; intros; simpls.
  + eauto.
  + forwards: Nat.add_pos_cases H. destructecs H1.
    * forwards HH0: H0 (logctx_extend oc (logctx_orr Tl logctx_nil)) trr Tr.
      2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. rewrite logctx_extend_typ. simpl. auto. eauto.
      rewrite logctx_extend_typ in HH0. simpl in HH0.
      forwards: IHHtr1 (logctx_extend oc (logctx_orl logctx_nil Ur)).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r. eauto.
      rewrite logctx_extend_typ. simpl. eauto.
      rewrite logctx_extend_typ in H2. simpl in H2. eauto.
    * forwards HH0: H0 (logctx_extend oc (logctx_orl logctx_nil Tr)) trl Tl.
      2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. rewrite logctx_extend_typ. simpl. auto. eauto.
      rewrite logctx_extend_typ in HH0. simpl in HH0.
      forwards: IHHtr2 (logctx_extend oc (logctx_orr Ul logctx_nil)).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l. eauto.
      rewrite logctx_extend_typ. simpl. eauto.
      rewrite logctx_extend_typ in H2. simpl in H2. eauto.
  + forwards: Nat.add_pos_cases H. destructecs H1.
    * forwards HH0: H0 (logctx_extend oc (logctx_andr Tl logctx_nil)) trr Tr.
      2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. rewrite logctx_extend_typ. simpl. auto. eauto.
      rewrite logctx_extend_typ in HH0. simpl in HH0.
      forwards: IHHtr1 (logctx_extend oc (logctx_andl logctx_nil Ur)).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r. eauto.
      rewrite logctx_extend_typ. simpl. eauto.
      rewrite logctx_extend_typ in H2. simpl in H2. eauto.
    * forwards HH0: H0 (logctx_extend oc (logctx_andl logctx_nil Tr)) trl Tl.
      2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. rewrite logctx_extend_typ. simpl. auto. eauto.
      rewrite logctx_extend_typ in HH0. simpl in HH0.
      forwards: IHHtr2 (logctx_extend oc (logctx_andr Ul logctx_nil)).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l. eauto.
      rewrite logctx_extend_typ. simpl. eauto.
      rewrite logctx_extend_typ in H2. simpl in H2. eauto.
  + forwards: H0 tr. 2: eauto. rewrite <- Nat.add_succ_l. eapply Nat.lt_lt_add_r.  eapply Nat.lt_succ_diag_r. eauto. eauto. eapply ty_inv_main_bnd. eauto.
    intros. forwards: H1. eauto. destructecs H5. eapply ty_has_N_main_tight. eapply ty_has_N_main_tree_sub_main; eauto. eapply ty_has_N_main_logic; eauto. eauto.
  + eapply ty_inv_main_nocap_bnd; eauto.
  + forwards: H0 tr. 2: eauto. eapply Nat.lt_succ_diag_r. eauto. eauto. eapply ty_inv_main_sel; eauto.
Qed.


Lemma ty_inv_main_lc_tree_linear_sub_main oc tr T U:
  ty_inv_main G e x (logctx_typ oc T) ->
  ty_main_tree_linear_sub G e x tr T U ->
  ty_inv_main G e x (logctx_typ oc U).
Proof.
  introv Hty Htr.
  remember (ty_main_tree_folds tr).
  reverts oc tr T U.
  induction n using nat_lt_0S_induction; intros; subst.
  - forwards: ty_main_tree_linear_sub_folds_0_inv; eauto. subst. eauto.
  -
  reverts oc. 
  induction Htr; intros; simpls.
  + eauto.
  + forwards: Nat.add_pos_cases H. destructecs H1.
    * forwards HH0: H0 (logctx_extend oc (logctx_orr Tl logctx_nil)) trr Tr.
      2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. rewrite logctx_extend_typ. simpl. auto. eauto.
      rewrite logctx_extend_typ in HH0. simpl in HH0.
      forwards: IHHtr1 (logctx_extend oc (logctx_orl logctx_nil Ur)).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r. eauto.
      rewrite logctx_extend_typ. simpl. eauto.
      rewrite logctx_extend_typ in H2. simpl in H2. eauto.
    * forwards HH0: H0 (logctx_extend oc (logctx_orl logctx_nil Tr)) trl Tl.
      2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. rewrite logctx_extend_typ. simpl. auto. eauto.
      rewrite logctx_extend_typ in HH0. simpl in HH0.
      forwards: IHHtr2 (logctx_extend oc (logctx_orr Ul logctx_nil)).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l. eauto.
      rewrite logctx_extend_typ. simpl. eauto.
      rewrite logctx_extend_typ in H2. simpl in H2. eauto.
  + forwards: Nat.add_pos_cases H. destructecs H1.
    * forwards HH0: H0 (logctx_extend oc (logctx_andr Tl logctx_nil)) trr Tr.
      2: auto. rewrite <- (Nat.add_0_l (ty_main_tree_folds trr)) at 1. apply -> PeanoNat.Nat.add_lt_mono_r. auto. rewrite logctx_extend_typ. simpl. auto. eauto.
      rewrite logctx_extend_typ in HH0. simpl in HH0.
      forwards: IHHtr1 (logctx_extend oc (logctx_andl logctx_nil Ur)).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_r. eauto.
      rewrite logctx_extend_typ. simpl. eauto.
      rewrite logctx_extend_typ in H2. simpl in H2. eauto.
    * forwards HH0: H0 (logctx_extend oc (logctx_andl logctx_nil Tr)) trl Tl.
      2: auto. rewrite <- (Nat.add_0_r (ty_main_tree_folds trl)) at 1. apply -> PeanoNat.Nat.add_lt_mono_l. auto. rewrite logctx_extend_typ. simpl. auto. eauto.
      rewrite logctx_extend_typ in HH0. simpl in HH0.
      forwards: IHHtr2 (logctx_extend oc (logctx_andr Ul logctx_nil)).
      intros; subst. eapply H0. 2: eauto. 2: eauto. 2: eauto. eauto using Nat.lt_lt_add_l. eauto.
      rewrite logctx_extend_typ. simpl. eauto.
      rewrite logctx_extend_typ in H2. simpl in H2. eauto.
  + forwards: H0 tr. 2: eauto. rewrite <- Nat.add_succ_l. eapply Nat.lt_lt_add_r.  eapply Nat.lt_succ_diag_r. eauto. eauto. eapply ty_inv_main_bnd. eauto.
    intros. forwards: H1. eauto. eapply ty_has_N_main_tight. eapply ty_has_N_main_tree_linear_sub_main; eauto. eapply ty_has_N_main_logic; eauto. eauto.
  + eapply ty_inv_main_nocap_bnd; eauto.
  + forwards: H0 tr. 2: eauto. eapply Nat.lt_succ_diag_r. eauto. eauto. eapply ty_inv_main_sel; eauto.
Qed.

Lemma ty_inv_main_tree_sub_main tr T U:
  ty_inv_main G e x T ->
  ty_main_tree_sub G e x tr T U ->
  ty_inv_main G e x U.
Proof.
  introv Hty Htr.
  eapply ty_inv_main_lc_tree_sub_main with (oc:=logctx_nil); simpls; eauto.
Qed.

Lemma ty_inv_main_tree_linear_sub_main tr T U:
  ty_inv_main G e x T ->
  ty_main_tree_linear_sub G e x tr T U ->
  ty_inv_main G e x U.
Proof.
  introv Hty Htr.
  eapply ty_inv_main_lc_tree_linear_sub_main with (oc:=logctx_nil); simpls; eauto.
Qed.

End WithContext.

(* For unknown reason, if this and the previous lemma are in the same section, then Qed of this lemma fails. *)
Section WithContext.
Context (G: ctx) (e: renv) (x: var).
Context (Hin: inert G) (Hcor: renv_corr G e).



Lemma ty_inv_main_inv_tree_sub T:
  ty_inv_main G e x T ->
  exists tr U,
  ty_main_tree_sub G e x tr U T /\
  ty_inv_logic G e x U.
Proof.
  introv Hty.
  induction Hty.
  - forwards: ty_main_tree_sub_refl G e x T. eauto.
  - (* Sel fold *)
    forwards: IHHty; eauto. destructecs H0.
    forwards: ty_main_tree_sub_lc_inv H1. destructecs H2; subst.
    eapply logctx_follow_symm in H2.
    eexists. exists (logctx_typ ocU Uoc). splits.
    eapply ty_main_tree_sub_lc_follow_replace.
    eauto. eauto. eapply ty_main_tree_sub_sel; eauto. eauto.
  - (* bnd *)
    forwards: IHHty; eauto. destructecs H1.
    forwards: ty_main_tree_sub_lc_inv H2. destructecs H3; subst.
    eapply logctx_follow_symm in H3.

    destruct (mode_has_incap' typing_mode) eqn: Hm.
    + 
    forwards HsuT: H0. auto.
    forwards HtrT: ty_main_tree_sub_refl (open x T).
    forwards HhT: ty_has_N_main_tree_sub_main_tree_logic. eauto. eauto. eauto. eapply HtrT.
    eapply subtyp_N_has_N_main. auto. auto. auto.
    destructecs HhT.

    eexists. exists (logctx_typ ocU Uoc). splits.
    eapply ty_main_tree_sub_lc_follow_replace.
    eauto. eauto. eapply ty_main_tree_sub_bnd; eauto.
    eauto. 
    +
    eexists. exists (logctx_typ ocU Uoc). splits.
    eapply ty_main_tree_sub_lc_follow_replace.
    eauto. eauto. eapply ty_main_tree_sub_bnd; eauto. 
    intro. splits; eauto. rewrite Hm in H4. discriminate. (*  rewrite Hm in H4. discriminate. *)
    eauto.
  - (* nocap bnd *)
    forwards: IHHty; eauto. destructecs H.
    forwards: ty_main_tree_sub_lc_inv H0. destructecs H1; subst.
    eapply logctx_follow_symm in H1.
    forwards: ty_main_tree_sub_nologbndsel_inv H3. hnf; split; intros; discriminate. hnf; split; intros; destruct oc0; simpls; discriminate.
    destructecs H2; subst.
    eexists. exists (logctx_typ ocU typ_nocap). splits.
    eapply ty_main_tree_sub_lc_follow_replace.
    eauto. eauto. eapply ty_main_tree_sub_nocap_bnd; eauto. eauto.
Qed.

Lemma ty_inv_main_inv_tree_linear_sub T:
  ty_inv_main G e x T ->
  exists tr U,
    ty_main_tree_linear_sub G e x tr U T /\
    ty_inv_logic G e x U.
Proof.
  introv Hty.
  forwards Htr: ty_inv_main_inv_tree_sub; eauto.
  destructecs Htr.
  forwards Htrx: ty_main_tree_tree_linear; eauto.
  destructecs Htrx.
  exists tr2 T2.
  splits; eauto.
  eapply ty_inv_logic_main_prep; eauto.

Qed.


(* OR DIAG *)

(* Lemma 35 in "Pure methods for roDOT" *)
Lemma ty_inv_main_lc_or_diag oc T:
  mode_has_incap typing_mode ->
  inert G ->
  ty_inv_main G e x (logctx_typ oc (typ_or T T)) ->
  ty_inv_main G e x (logctx_typ oc T).
Proof.
  introv Hm Hi Hty.
  (* Trace the union to the previous layer *)
  forwards Hee: ty_inv_main_inv_tree_sub; eauto.
  destructecs Hee.
  forwards Ho: ty_main_tree_sub_lc_inv Hee0.
  destructecs Ho; subst.
  forwards HoT: ty_main_tree_sub_or_inv Ho1.
  destructecs HoT; subst.
  (* Find a common path through this layer *)
  forwards HT3: ty_main_tree_common HoT HoT2; fweauto.
  destructecs HT3.
  (* Solve the previous layer *)
  forwards Htyr: ty_inv_logic_main_prep (logctx_typ (logctx_extend ocU (logctx_orr Tl logctx_nil)) Tr).
  eauto.
  rewrite logctx_extend_typ; simpls. eauto.
  eapply ty_main_prep_lc. eauto.
  rewrite logctx_extend_typ in Htyr; simpl in Htyr.

  forwards Htyl: ty_inv_logic_main_prep (logctx_typ (logctx_extend ocU (logctx_orl logctx_nil T3)) Tl).
  eauto.
  rewrite logctx_extend_typ; simpls. eauto.
  eapply ty_main_prep_lc. eauto.
  rewrite logctx_extend_typ in Htyl; simpl in Htyl.

  eapply ty_inv_logic_lc_or_diag in Htyl. 2: eauto. 2: eauto. 2: eapply ty_inv_main_dom; eauto.
  (* Apply the path through this layer *)

  eapply logctx_follow_symm in Ho.
  forwards: ty_main_tree_sub_lc_follow_replace Hee0 HT3. auto.
  eapply ty_inv_main_tree_sub_main. eauto. eauto. 2: eapply H.
  eapply ty_inv_main_logic. eauto.

Qed.

(* Lemma 38 in "Pure methods for roDOT" *)
Lemma ty_inv_main_lc_N_cases oc T:
  mode_has_incap typing_mode ->
  ty_inv_main G e x (logctx_typ oc T) ->
  ty_has_N_main G e T ->
  ty_inv_main G e x T \/
  ty_inv_main G e x (logctx_typ oc (typ_nocap)).
Proof.
  introv Hm Hty Hh.
  (* Trace to the previous layer *)
  forwards Htyee: ty_inv_main_inv_tree_sub; eauto.
  destructecs Htyee.
  forwards Hhee: ty_has_N_main_main_tree_logic x Hh; eauto.
  destructecs Hhee.
  forwards HT: ty_main_tree_sub_lc_inv Htyee0.
  destructecs HT; subst.
  (* Find a common path through this layer *)
  forwards HT3: ty_main_tree_common HT1 Hhee1; fweauto.
  destructecs HT3.
  (* Solve the previous layer *)
  forwards Htyr: ty_inv_logic_main_prep.
  eauto. eauto.
  eapply ty_main_prep_lc. eauto.

  forwards Hhl: ty_has_N_logic_main_prep Hhee; eauto.

  forwards: ty_inv_logic_lc_N_cases; eauto.  eapply ty_inv_main_dom; eauto.
  (* Apply the path through this layer *)
  eapply logctx_follow_symm in HT.

  destructecs H.
  - left. eapply ty_inv_main_tree_sub_main. eauto. eauto. eapply ty_inv_main_logic. eauto. eauto.
  - forwards Hn: ty_main_tree_sub_refl G e x typ_nocap.
    right. eapply ty_inv_main_tree_sub_main. eauto. eauto. eapply ty_inv_main_logic. eauto.
    eapply ty_main_tree_sub_lc_follow_replace; eauto.
Qed.

(* Lemma 33 in "Pure methods for roDOT" *)
Lemma ty_inv_main_lc_and oc T U:
  mode_has_incap typing_mode ->
  ty_inv_main G e x (logctx_typ oc T) ->
  ty_inv_main G e x (logctx_typ oc U) ->
  ty_inv_main G e x (logctx_typ oc (typ_and T U)).
Proof.
  introv Hm HtyT HtyU.
  (* Trace to the previous layer *)
  forwards HeeT: ty_inv_main_inv_tree_sub HtyT; eauto.
  destructecs HeeT.
  forwards HeeU: ty_inv_main_inv_tree_sub HtyU; eauto.
  destructecs HeeU.
  forwards HT: ty_main_tree_sub_lc_inv HeeT0.
  destructecs HT; subst.
  forwards HU: ty_main_tree_sub_lc_inv HeeU0.
  destructecs HU; subst.

  (* Common type *)

  forwards Htop: ty_main_tree_sub_refl G e x typ_top.
(*  destructecs Htop. *)

  forwards HtopT: ty_main_tree_sub_lc_follow_replace HeeT0 Htop.
  eauto using logctx_follow_trans, logctx_follow_symm.

  forwards HtopU: ty_main_tree_sub_lc_follow_replace HeeU0 Htop.
  eauto using logctx_follow_trans, logctx_follow_symm.  

  (* Find a common path through this layer *)
  forwards HT3: ty_main_tree_common HtopT HtopU; fweauto.
  destructecs HT3.

  forwards HeT3: ty_main_tree_sub_lc_inv HT3.
  destructecs HeT3; subst.

  (* Specialize to source types *)
  forwards HiT: ty_main_prep_lc_replace Uoc Uoc HT0. eapply ty_main_prep_refl.
  eauto using logctx_follow_trans, logctx_follow_symm.  

  forwards HiU: ty_main_prep_lc_replace Uoc0 Uoc0 HT2. eapply ty_main_prep_refl.
  eauto using logctx_follow_trans, logctx_follow_symm.  

  (* Solve the previous layer *)
  forwards HtylT: ty_inv_logic_main_prep HeeT HiT; eauto.
  forwards HtylU: ty_inv_logic_main_prep HeeU HiU; eauto.

  forwards: ty_inv_logic_lc_and HtylT HtylU.

  (* Apply the path through this layer *)
  forwards: ty_main_tree_sub_lc_follow_replace HT3.
  eauto using logctx_follow_trans, logctx_follow_symm.
  eapply ty_main_tree_sub_and. eapply HT1. eapply HU1.

  eapply ty_inv_main_tree_sub_main. eauto. eauto.  2: eapply H0.
  eapply ty_inv_main_logic. eauto.
Qed.

End WithContext.


Lemma typ_notlog_tyq e T1 T2:
  typ_notlog T1 ->
  typing_equiv e T1 T2 ->
  typ_notlog T2.
Proof.
  introv Hn Htyq.
  destruct T1; inversions Hn; inversions_tyq Htyq.
  all: hnf;split; intros; try discriminate.
  false H0; eauto.
  false H; eauto.
Qed.

Lemma ty_main_tree_linear_sub_tyq G e w y tr1 T1 U1 U2:
  inert G ->
  renv_corr G e ->
  binds w y e ->
  typing_equiv e U1 U2 ->
  ty_main_tree_linear_sub G e y tr1 T1 U1 ->
  exists T2 tr2,
  typing_equiv e T1 T2 /\
  ty_main_tree_linear_sub G e w tr2 T2 U2.
Proof.
  introv Hin Hcor Hbi Htyq Htr.
  reverts U2.
  induction Htr; intros.
  - forwards: typ_notlog_tyq H Htyq. exists U2; eauto.
  - inversions_tyq Htyq.
    forwards HHtr1: IHHtr1. eauto.
    forwards HHtr2: IHHtr2. eauto.
    destructecs HHtr1. destructecs HHtr2.
    exists (typ_or T2 T0); eauto.
  -  inversions_tyq Htyq.
    forwards HHtr1: IHHtr1. eauto.
    forwards HHtr2: IHHtr2. eauto.
    destructecs HHtr1. destructecs HHtr2.
    exists (typ_and T2 T0); eauto.
  - inversions_tyq Htyq.
    forwards HHtr1: IHHtr (open w t2).
    forwards: tyq_typ_open_indep_same e y w; eauto.
    eapply tyq_typ_trans; eauto.
    destructecs HHtr1.

    destruct (mode_has_incap' typing_mode ) eqn: Hm.
    *

    forwards HH: H. auto.

    exists T2. eexists. splits. eauto. eapply ty_main_tree_linear_sub_bnd. eauto.
    intros. eapply ty_has_N_logic_tyq; eauto.
    pick_fresh z.
    eapply typ_indep_tyq_open_indep with (e:=e) (xs:=z); eauto.
    *
    exists T2. eexists. splits. eauto. eapply ty_main_tree_linear_sub_bnd. eauto.
    intros. rewrite Hm in H1. discriminate.
    pick_fresh z.
    eapply typ_indep_tyq_open_indep with (e:=e) (xs:=z); eauto.


    - inversions_tyq Htyq.
    eauto.
    -
 inversions_tyq Htyq.
  + forwards Hpf: renv_corr_pf_normal_typ_tyq; eauto.
    destructecs Hpf.

    forwards HHtr: IHHtr. eapply Hpf1. destructecs HHtr.
    exists T2. eauto.
  +
    forwards HHtr: IHHtr. eauto. destructecs HHtr.
    exists T2. eauto.
Qed.

Lemma ty_inv_main_ro_ref G e w y T:
  mode_has_incap typing_mode ->
    mode_has_top_nm_subtype typing_mode ->
  inert G ->
  renv_corr G e ->
  binds w y e ->
  ty_has_N_main G e T ->
  ty_inv_main G e y T ->
  ty_inv_main G e w T.
Proof.
  introv Hm' Hm Hin Hcor Hbi Hh Hty.

  (* Trace to the previous layer *)
  forwards Htyee: ty_inv_main_inv_tree_sub; eauto.
  destructecs Htyee.
  forwards Hhee: ty_has_N_main_main_tree_logic y Hh; eauto.
  destructecs Hhee.
  (* Find a common path through this layer *)
  forwards HT3: ty_main_tree_common Htyee0 Hhee1; eauto.
  destructecs HT3.
  (* Use the path for the reference variable *)
  forwards Htrx: ty_main_tree_tree_linear HT3; eauto.
  destructecs Htrx.
  forwards Htrq: ty_main_tree_linear_sub_tyq Htrx; eauto.
  destructecs Htrq.
  (* Solve the previous layer *)
  forwards Hhr: ty_has_N_logic_main_prep HT1; eauto.
  forwards Htyr: ty_inv_logic_main_prep HT0; eauto.
  forwards Htyr2: ty_inv_logic_main_prep Htrx0; eauto.
  forwards Hhr2: ty_has_N_logic_main_prep Htrx0; eauto.
  forwards Htyl: ty_inv_logic_ro_ref Hhr2 Htyr2; eauto.
  forwards Htyl2: ty_inv_logic_tyq Htyl Htrq0; eauto.
  (* Use the path on the main layer *)
  eapply ty_inv_main_tree_linear_sub_main; eauto.
  eapply ty_inv_main_logic; eauto.
Qed.

(* Lemma 28 in "Pure methods for roDOT" *)
Lemma invertible_main_to_precise_typ_dec: forall G e x a S T,
  ty_inv_main G e x ( typ_rcd (dec_typ a S T)) ->
  exists S' T' U,
    G ⊢! x ∶ U ⪼ typ_rcd (dec_typ a S' T') /\
    (forall y, G⁏e ⊢# open y T' <: open y  T) /\
    (forall y, G⁏e ⊢# open y S <: open y  S').
Proof.
  introv Hinv.
  eapply ty_inv_main_atomic_atomic in Hinv; eauto.
  dependent induction Hinv.
  - symmetry in H1.
    apply tyq_typ_dec_typ_inv in H1.
    destruct H1 as [t21 [t22 [Ht [Htyq1 Htyq2]]]]; subst.
    exists t21 t22 T0.
    split; auto; split; intros.
    specialize (Htyq1 y).
    specialize (Htyq2 y).
    apply tyq_typ_sym in Htyq2; eauto.
    eauto.
  - inversions H.
    specialize (IHHinv _ _ _ eq_refl).
    destructc IHHinv as V V' V'' Hx Hs1 Hs2. auto. auto.
    exists V V' V''.
    splits; auto; intros.
    eapply subtyp_trans_t; try eapply Hs1; eauto.
    eapply subtyp_trans_t; try eapply Hs2; eauto.
Qed.

(* Lemma 30 in "Pure methods for roDOT" *)
Lemma invertible_main_to_precise_met_dec: forall G e x m TC TZ TR,
  inert G -> ok e ->
  ty_inv_main G e x (typ_rcd (dec_met m TC TZ TR)) ->
  exists L S TC' TZ' TR',
    G ⊢! x ∶ S ⪼ typ_rcd (dec_met m TC' TZ' TR') /\
    G⁏e ⊢# TZ <: TZ' /\
    (forall z, z \notin L -> G § z / v_par ~ TZ ⁏e ⊢ (open z TC) <: (open z TC')) /\
    (mode_has_sef typing_mode -> typ_indep_b 0 TC -> typ_indep_b 0 TC') /\
    (forall z, z \notin L -> forall y, y <> z -> y \notin L -> ctx_met G z TZ y TC ⁏e ⊢ (open_met z y TR') <: (open_met z y TR)).
Proof.
  introv Hi He Hinv.
  eapply ty_inv_main_atomic_atomic in Hinv; eauto.
  dependent induction Hinv.

  - symmetry in H1.
    apply tyq_typ_dec_met_inv in H1.
    destruct H1 as [t21 [t22 [t23 [Ht [Htyq1 [Htyq2 Htyq3]]]]]]; subst.
    exists (ctx_dom G) T t21 t22 t23.
    splits; eauto using typ_indep_tyq_open_all_indep.
    intros.
    specialize (Htyq1 z).
    specialize (Htyq3 z y).
    symmetry in Htyq3.
    eauto.
  - inversions H.
    specialize (IHHinv _ _ _ _ Hi He eq_refl).
    destruct IHHinv
      as [L' [S' [TC3 [TZ3 [TR3 [Hx [Hs1 [Hs2 [Hidp Hs3]]]]]]]]].
    exists (ctx_dom G \u L \u L') S' TC3 TZ3 TR3.
    splits; auto.
    + eapply subtyp_trans_t; eassumption.
    + intros z' Fr.
      assert (Hok: ctx_ok (G § z' / v_par ~ TZ)) by eauto.
      apply tight_to_general in H7; auto.
      assert (Hnarrow: G § z' / v_par ~ TZ ⁏e ⊢ open z' TC <: open z' TC1).
      { eapply narrow; auto using subctx_last. }
      eauto using narrow_subtyping.
    + intros z' Frz y' Nyz Fry.
      assert (Hok: ctx_ok (ctx_met G z' TZ y' TC)) by eauto using ctx_unbound_push.
      apply tight_to_general in H7; auto.
      assert (Hnarrow: (ctx_met G z' TZ y' TC) ⁏e ⊢ open_met z' y' TR3 <: open_met z' y' TR1).
      { apply narrow with (G0:= (ctx_met G z' TZ1 y' TC1)); auto using subctx_last2. }
      eauto.
Qed.

Lemma invertible_main_to_precise_typ_fun: forall G e x S T,
  inert G -> ok e ->
  ty_inv_main G e x (typ_fun S T) ->
  exists S' T' U L,
    G ⊢! x ∶ U ⪼ typ_fun S' T' /\
    G⁏e ⊢# S <: S' /\
    (forall y,
        y \notin L ->
            G § y / v_par ~ S ⁏e ⊢ open y T' <: open y T).
Proof.
  introv HG HE Hinv.
  eapply ty_inv_main_atomic_atomic in Hinv; eauto.
  dependent induction Hinv.
  - destructc (tyq_typ_fun_invr H1) as t21 t22 Ht Htyq1 Htyq2; subst.
    exists t21 t22 T0 (ctx_dom G).
    apply tyq_typ_sym in Htyq1.
    split; auto.
    split; auto.
    intros.
    specialize (Htyq2 y ).
    symmetry in Htyq2.
    eauto.
  - inversions H.
    specialize (IHHinv _ _ HG HE eq_refl).
    destruct IHHinv as [S' [T' [V' [L' [Hpt [HSsub HTsub]]]]]].
    exists S' T' V' (ctx_dom G \u L \u L').
    split; auto.
    assert (Hsub2 : G⁏e ⊢# typ_fun S1 T0 <: typ_fun S T).
    { apply subtyp_fun_t with (L:=L); assumption. }
    split.
    + eapply subtyp_trans_t; eauto.
    + intros y Fr.
      assert (Hok: ctx_ok (G § y / v_par ~ S)) by eauto.
      apply tight_to_general in H6; auto.
      assert (Hnarrow: G § y / v_par ~ S ⁏e ⊢ open y T' <: open y T0).
      { eapply narrow; auto using subctx_last. }
      eauto.
Qed.

Lemma ty_inv_main_lc_typ_and G e x oc A S1 T1 S2 T2:
  inert G ->
  renv_corr G e ->
  ty_inv_main G e x (logctx_typ oc (typ_and (typ_rcd (dec_typ A S1 T1)) (typ_rcd (dec_typ A S2 T2)))) ->
  ty_inv_main G e x (logctx_typ oc (typ_rcd (dec_typ A (typ_or S1 S2) (typ_and T1 T2)))).
Proof.
  introv Hin Hcor Hty.
  remember (logctx_typ oc (typ_and (typ_rcd (dec_typ A S1 T1)) (typ_rcd (dec_typ A S2 T2)))) .
  reverts oc.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic. eapply ty_inv_logic_lc_typ_dist; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + discriminate.
    + destruct ocST; simpls; discriminate.
    + destruct ocTS; simpls; inversions Heqt.
      destruct ocTS; simpls; inversions H1.
      destruct ocTS; simpls; inversions H2.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open z S0).
      destructecs Hcr.
      forwards HcrU0: Hcr1 (typ_rcd (dec_typ A (typ_or S1 S2) (typ_and T1 T2))).
      destructecs HcrU0.
      rewrite <- HcrU1.
      eapply ty_inv_main_sel. 2: eauto.
      forwards HHty: IHHty; fweauto. rewrite <- HcrU2 in HHty. eauto.
  - (* Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + discriminate.
    + destruct ocST; simpls; discriminate.
    + destruct ocTS; simpls; inversions Heqt. 
      destruct ocTS; simpls; inversions H2.
      destruct ocTS; simpls; inversions H3.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open x T).
      destructecs Hcr.
      forwards HcrU: Hcr1  (typ_rcd (dec_typ A (typ_or S1 S2) (typ_and T1 T2))).
      destructecs HcrU.
      rewrite <- HcrU0.
      eapply ty_inv_main_bnd.  eauto.  eauto.
      forwards HHty: IHHty; fweauto. rewrite <- HcrU1 in HHty. eauto.
  - (* Nocap-Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + discriminate.
    + destruct ocST; simpls; discriminate.
    + destruct ocTS; simpls; inversions Heqt.
      destruct ocTS; simpls; inversions H0.
      destruct ocTS; simpls; inversions H1.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (typ_nocap).
      destructecs Hcr.
      forwards HcrU: Hcr1 (typ_rcd (dec_typ A (typ_or S1 S2) (typ_and T1 T2))).
      destructecs HcrU.
      rewrite <- HcrU0.
      eapply ty_inv_main_nocap_bnd.  eauto.  eauto.
      forwards HHty: IHHty; fweauto. rewrite <- HcrU1 in HHty. eauto.
Qed.



Lemma ty_inv_main_lc_dist G e x oc T1 T2 T3:
  inert G ->
  renv_corr G e ->
  ty_inv_main G e x (logctx_typ oc (typ_and T1 (typ_or T2 T3))) ->
  ty_inv_main G e x (logctx_typ oc (typ_or (typ_and T1 T2) (typ_and T1 T3))).
Proof.
  introv Hi Hcor Hty.

  remember (logctx_typ oc (typ_and T1 (typ_or T2 T3))) as T.
  reverts oc T1 T2 T3.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic.
    eapply ty_inv_logic_lc_dist; eauto.
  - (* Sel *)
    lets Heqt0: HeqT.
    eapply logctx_intersect_cases in HeqT.
    destruct_all; subst.
    + discriminate.
    + destruct x0; simpls; discriminate.
    + destruct x0; simpls; inversions H1. 
      * rewrite logctx_extend_orl.
        rewrite logctx_extend_andl with (Tr:=T2).
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_sel; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite <- logctx_extend_orl.
        rewrite logctx_extend_orr.
        rewrite logctx_extend_andl with (Tr:=T3).
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_sel; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        simpls.
        eapply IHHty; eauto.
        rewrite logctx_extend_typ.
        simpls. auto.
      * destruct x0; simpls; inversions H3.
        **
        rewrite logctx_extend_orl.
        rewrite logctx_extend_andr.
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_sel; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        simpls.
        eapply IHHty; eauto.
        rewrite logctx_extend_typ.
        simpls. auto.
        **
        rewrite logctx_extend_orr.
        rewrite logctx_extend_andr.
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_sel; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        simpls.
        eapply IHHty; eauto.
        rewrite logctx_extend_typ.
        simpls. auto.
    +
      forwards: logctx_separate_cross_r Heqt0 H0 (open z S1).
      destruct_all.
      forwards: H2 ( (typ_or (typ_and T1 T2) (typ_and T1 T3))).
      destruct_all.
      rewrite <- H4.
      eapply ty_inv_main_sel. 2: eauto.
      forwards: IHHty. eauto. eauto. eapply H1. rewrite <- H5 in H7. eauto.
  - (* Bnd *)
    lets Heqt0: HeqT.
    eapply logctx_intersect_cases in HeqT.
    destruct_all; subst.
    + discriminate.
    + destruct x0; simpls; discriminate.
    + destruct x0; simpls; inversions H2.

      * rewrite logctx_extend_orl.
        rewrite logctx_extend_andl with (Tr:=T2).
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_bnd; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite <- logctx_extend_orl.
        rewrite logctx_extend_orr.
        rewrite logctx_extend_andl with (Tr:=T3).
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_bnd; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        simpls.
        eapply IHHty; eauto.
        rewrite logctx_extend_typ.
        simpls. auto.
      * destruct x0; simpls; inversions H4.
        **
        rewrite logctx_extend_orl.
        rewrite logctx_extend_andr.
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_bnd; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        simpls.
        eapply IHHty; eauto.
        rewrite logctx_extend_typ.
        simpls. auto.
        **
        rewrite logctx_extend_orr.
        rewrite logctx_extend_andr.
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_bnd; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        simpls.
        eapply IHHty; eauto.
        rewrite logctx_extend_typ.
        simpls. auto.
    +
      forwards: logctx_separate_cross_r Heqt0 H1 (open x T).
      destruct_all.
      forwards: H3 ( (typ_or (typ_and T1 T2) (typ_and T1 T3))).
      destruct_all.
      rewrite <- H5.
      eapply ty_inv_main_bnd.  eauto.  eauto.
      forwards: IHHty. eauto. eauto. eapply H2. rewrite <- H6 in H8. eauto.

  - (* Nocap-Bnd *)
    lets Heqt0: HeqT.
    eapply logctx_intersect_cases in HeqT.
    destruct_all; subst.
    + discriminate.
    + destruct x0; simpls; discriminate.
    + destruct x0; simpls; inversions H0.

      * rewrite logctx_extend_orl.
        rewrite logctx_extend_andl with (Tr:=T2).
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_nocap_bnd; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite <- logctx_extend_orl.
        rewrite logctx_extend_orr.
        rewrite logctx_extend_andl with (Tr:=T3).
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_nocap_bnd; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        simpls.
        eapply IHHty; eauto.
        rewrite logctx_extend_typ.
        simpls. auto.
      * destruct x0; simpls; inversions H2.
        **
        rewrite logctx_extend_orl.
        rewrite logctx_extend_andr.
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_nocap_bnd; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        simpls.
        eapply IHHty; eauto.
        rewrite logctx_extend_typ.
        simpls. auto.
        **
        rewrite logctx_extend_orr.
        rewrite logctx_extend_andr.
        rewrite <- logctx_extend_typ.
        eapply ty_inv_main_nocap_bnd; eauto.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        rewrite logctx_extend_typ.
        simpls.
        eapply IHHty; eauto.
        rewrite logctx_extend_typ.
        simpls. auto.
    +
      forwards: logctx_separate_cross_r Heqt0 H (typ_nocap).
      destruct_all.
      forwards: H1 ( (typ_or (typ_and T1 T2) (typ_and T1 T3))).
      destruct_all.
      rewrite <- H3.
      eapply ty_inv_main_nocap_bnd.  eauto.  eauto.
      forwards: IHHty. eauto. eauto. eapply H0. rewrite <- H4 in H6. eauto.
Qed.

Lemma ty_inv_main_lc_NM_bot_replace G e x oc:
  mode_top_nm' typing_mode = mode_top_nm_complement ->
  inert G ->
  renv_corr G e ->
  ty_inv_main G e x (logctx_typ oc (typ_and typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)))) ->
  ty_inv_main G e x (logctx_typ oc typ_bot).
Proof.
  introv Hm Hin Hcor Hty.
  remember (logctx_typ oc (typ_and typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)))).
  reverts oc.
  induction Hty; intros; subst.
  - eapply ty_inv_main_logic. eapply ty_inv_logic_lc_MN_bot_replace; eauto.
  - (* Sel *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + discriminate.
    + destruct ocST; simpls; discriminate.
    + destruct ocTS; simpls; inversions Heqt.
      destruct ocTS; simpls; inversions H1.
      destruct ocTS; simpls; inversions H2.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open z S1).
      destructecs Hcr.
      forwards HcrU0: Hcr1 typ_bot.
      destructecs HcrU0.
      rewrite <- HcrU1.
      eapply ty_inv_main_sel. 2: eauto.
      forwards HHty: IHHty; fweauto. rewrite <- HcrU2 in HHty. eauto.
  - (* Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + discriminate.
    + destruct ocST; simpls; discriminate.
    + destruct ocTS; simpls; inversions Heqt. 
      destruct ocTS; simpls; inversions H2.
      destruct ocTS; simpls; inversions H3.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (open x T).
      destructecs Hcr.
      forwards HcrU: Hcr1 typ_bot.
      destructecs HcrU.
      rewrite <- HcrU0.
      eapply ty_inv_main_bnd.  eauto.  eauto.
      forwards HHty: IHHty; fweauto. rewrite <- HcrU1 in HHty. eauto.
  - (* Nocap-Bnd *)
    lets Heqt0: Heqt.
    eapply logctx_intersect_cases in Heqt.
    destructecs Heqt; subst.
    + discriminate.
    + destruct ocST; simpls; discriminate.
    + destruct ocTS; simpls; inversions Heqt.
      destruct ocTS; simpls; inversions H0.
      destruct ocTS; simpls; inversions H1.
    +
      forwards Hcr: logctx_separate_cross_r Heqt0 Heqt (typ_nocap).
      destructecs Hcr.
      forwards HcrU: Hcr1 typ_bot.
      destructecs HcrU.
      rewrite <- HcrU0.
      eapply ty_inv_main_nocap_bnd.  eauto.  eauto.
      forwards HHty: IHHty; fweauto. rewrite <- HcrU1 in HHty. eauto.
Qed.


(* Lemma 32 in "Pure methods for roDOT" *)
Lemma invertible_main_typing_closure_tight G e x oc T U:
  mode_has_incap typing_mode ->
  inert G ->
  renv_corr G e ->
  ty_inv_main G e x (logctx_typ oc T) ->
  G⁏e ⊢# T <: U ->
  ty_inv_main G e x (logctx_typ oc U).
Proof.
  introv Hm Hi Hcor HT Hsub. reverts oc.
  dependent induction Hsub; intros; eauto.
  - eapply ty_inv_main_lc_top_replace; eauto.
  - eapply ty_inv_main_lc_bot_replace; eauto.
  - eapply ty_inv_main_lc_and_invl; eauto.
  - eapply ty_inv_main_lc_and_invr; eauto.
  - eapply ty_inv_main_lc_and; eauto.
  - eapply ty_inv_main_lc_orl; eauto.
  - eapply ty_inv_main_lc_orr; eauto.
  -
    lets HTU: HT.
    rewrite logctx_extend_orl in HTU.
    eapply IHHsub1 in HTU; eauto.

    rewrite <- logctx_extend_orl in HTU.

    lets HTS: HTU.
    rewrite logctx_extend_orr in HTS.
    eapply IHHsub2 in HTS; eauto.
    rewrite <- logctx_extend_orr in HTS.
    eapply ty_inv_main_lc_or_diag; eauto.
  - eapply ty_inv_main_lc_subtyp; eauto. eapply subtyp_atomic_fld; eauto.
  - eapply ty_inv_main_lc_subtyp; eauto. eapply subtyp_atomic_typ; eauto.
  - eapply ty_inv_main_sel; eauto.
  - eapply ty_inv_main_lc_sel_open; eauto.
  - eapply ty_inv_main_lc_subtyp; eauto. eapply subtyp_atomic_fun; eauto.
  - eapply ty_inv_main_lc_subtyp; eauto. eapply subtyp_atomic_met; eauto.
  - eapply ty_inv_main_tyq; eauto. eapply logctx_tyq_same; eauto.
  - eapply ty_inv_main_lc_typ_and; eauto.
  - eapply ty_inv_main_lc_dist; eauto.
  - eapply ty_inv_main_nocap_bnd; eauto.
  - eapply ty_inv_main_lc_nocap_replace; eauto.
  - eapply ty_inv_main_lc_nocap_replace; eauto.
  - eapply ty_inv_main_lc_nocap_replace; eauto.
  - eapply ty_inv_main_lc_nocap_replace; eauto.
  - eapply ty_inv_main_lc_NM_replace.  destruct (mode_has_top_nm_subtype' typing_mode) eqn: Hms. eauto. typing_mode_contradiction typing_mode. eauto. intros.  typing_mode_contradiction typing_mode. eauto.
  - eapply ty_inv_main_lc_NM_replace.  destruct (mode_has_top_nm_subtype' typing_mode) eqn: Hms. eauto. typing_mode_contradiction typing_mode. eauto. intros.  sympl.  eauto. eauto.
  - eapply ty_inv_main_lc_NM_bot_replace; eauto.
  - eapply ty_inv_main_lc_has_top_replace; eauto. 
Qed.


Lemma precise_to_invertible_main G e x T U:
  mode_has_incap typing_mode ->
  inert G ->
  renv_corr G e ->
  G ⊢! x ∶ T ⪼ U ->
  ty_inv_main G e x U.
Proof.
  introv Hm Hi Hcor Hpf.

  forwards: pf_inert_T Hi Hpf.
  unfold inert_type in *; destruct_all.
  inversions H.
  - (* Fun *)
    forwards: precise_flow_fun_inv Hpf; subst.
    eapply precise_to_invertible_atomic; eauto.
  - (* Obj *)
    assert (typ_opening_ind U) as Ti by apply typ_opening_ind_all.
    induction Ti; intros.
    + eapply pf_binds in Hpf.
      eapply ty_inv_main_logic.
      eapply ty_inv_logic_or.
      eapply ctx_binds_dom; eauto.
      eapply ty_or_closure_top.
    + eapply pf_bot_false in Hpf; eauto; contradiction.
    + eapply pf_bnd_same_or_rcd in Hpf; eauto; destruct_all;
      try discriminate; inversions H4; inversions H5.
    + eapply precise_to_invertible_atomic; eauto.
    + eapply precise_to_invertible_atomic; eauto.
    + eapply precise_to_invertible_atomic; eauto.
    + forwards: IHTi1. eauto.
      forwards: IHTi2. eauto.
      forwards: ty_inv_main_lc_and logctx_nil H H3;
      eauto.
    + eapply pf_or_false in Hpf; eauto; contradiction.
    + eapply pf_psel_false in Hpf; eauto; contradiction.
    +
      assert (T0 = T). {
        forwards: pf_inert_bnd_U; eauto.
        destruct_all; subst.
        inversions H4. auto.
      }
      subst.
      forwards: open_record_type x H0.

      forwards: ty_inv_main_bnd G e x logctx_nil T.
      ++ eauto.
      ++
      intros.
      eapply record_type_subtyp_nocap; eauto.
      ++
      simpls. eauto.
      ++
      simpls. eauto.
    + eapply pf_nocap_false in Hpf; eauto; contradiction.
Qed.
(* Lemma 31 in "Pure methods for roDOT" *)
Lemma tight_to_invertible_main G e U x:
    mode_incap' typing_mode = mode_incap_nocap -> 
    inert G ->
    renv_corr G e ->
    G⁏e ⊢# x ∶ U ->
    ty_inv_main G e x U.
Proof.
  introv Hm Hgd Hcor Hty.
  
  assert (mode_has_incap typing_mode) as Hmi. {
    unfold mode_has_incap'; rewrite Hm; auto.
  }

  induction Hty; try congruence; eauto.
  - eapply precise_to_invertible_main; eauto.
  - forwards: IHHty; eauto.
    forwards: ty_inv_main_bnd logctx_nil. eauto. intros. eapply H1. eauto. eauto. eauto.
  - forwards: ty_inv_main_bnd_open; eauto.
  - forwards: ty_inv_main_lc_and logctx_nil T U; eauto.
  - forwards: invertible_main_typing_closure_tight logctx_nil T U; eauto.
  - destruct (tight_typing_implies_bound H0).
    forwards Hmut: (pf_mutability H Hgd H1).
    destructc Hmut  as T1 ?H.
    forwards: precise_to_invertible_main e; eauto.
    forwards: ty_inv_main_lc_subtyp logctx_nil; eauto.
    constructor; eauto.
Qed.

End WithTypingMode.