Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath LogpathOrder.

Require Import TypeInduction TypLogicInduction TypeBndSel.

Require Import LayeredCommon LayeredSubtypingAtomic.
Require Import LayeredCapabilityBasic LayeredCapabilityBasicClosure LayeredCapabilityMain.
Require Import LayeredTypingAtomic LayeredTypingBasic LayeredTypingBasicClosure LayeredTypingLogic LayeredTypingMain.

Section WithTypingMode.
Context {typing_mode: Mode}.

Section WithContext.
Context (G: ctx) (e: renv) (x: var).
Context (Hin: inert G).


(* Analog of ty_inv_logic_ord
  First argument is maximum number of steps.
  Second argument is leftmost allowed prefix
 *)
Inductive ty_inv_main_ord: nat -> logctx -> typ -> Prop :=
| ty_inv_main_ord_logic n oc T:
    ty_inv_logic G e x T ->
    typ_notbnd T ->
    ty_inv_main_ord n oc T
(* Replace lower bound by selection *)
| ty_inv_main_ord_sel n1 n2 oc ocp S1 S2 U y A z:
    ty_inv_main_ord n1 ocp (logctx_typ ocp (open z S1)) ->
    G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
    logctx_le oc ocp ->
    n1 < n2 ->
    ty_inv_main_ord n2 oc (logctx_typ ocp (typ_sel (avar_f y) A (avar_f z)))
| ty_inv_main_ord_bnd n1 n2 oc ocp T:
    ty_inv_main_ord n1 ocp (logctx_typ ocp (open x T)) ->
    ((mode_has_incap typing_mode) -> (G⁏e ⊢# typ_nocap <: (open x T))) ->
    typ_indep_b 0 T ->
    logctx_le oc ocp ->
    n1 < n2 ->
    ty_inv_main_ord n2 oc (logctx_typ ocp (typ_bnd T))
| ty_inv_main_ord_nocap_bnd n1 n2 oc ocp T:
    ty_inv_main_ord n1 ocp (logctx_typ ocp typ_nocap) ->
    logctx_le oc ocp ->
    n1 < n2 ->
    ty_inv_main_ord n2 oc (logctx_typ ocp (typ_bnd T))
.


(* Analog of ty_inv_logic_ord_logic *)
Lemma ty_inv_main_ord_main n op T:
  ty_inv_main_ord n op T ->
  ty_inv_main G e x T.
Proof.
  introv Hh.
  induction Hh.
  - eapply ty_inv_main_logic; eauto.
  - eapply ty_inv_main_sel; eauto.
  - eapply ty_inv_main_bnd; eauto.
  - eapply ty_inv_main_nocap_bnd; eauto.
Qed.


(* Analog of ty_inv_logic_ord_le *)
Lemma ty_inv_main_ord_le n op1 op2 T:
  logctx_le op2 op1 ->
  ty_inv_main_ord n op1 T ->
  ty_inv_main_ord n op2 T.
Proof.
  introv Hle Hh.
  induction Hh.
  - eapply ty_inv_main_ord_logic; eauto.
  - eapply ty_inv_main_ord_sel; eauto.
    unfold orctx_le in *. eapply logpath_le_trans; eauto.
  - eapply ty_inv_main_ord_bnd; eauto.
    unfold orctx_le in *. eapply logpath_le_trans; eauto.
  - eapply ty_inv_main_ord_nocap_bnd; eauto.
    unfold orctx_le in *. eapply logpath_le_trans; eauto.
Qed.

Lemma ty_inv_main_ord_nge n1 n2 op T:
  n1 <= n2 ->
  ty_inv_main_ord n1 op T ->
  ty_inv_main_ord n2 op T.
Proof.
  introv Hle Hh.
  induction Hh.
  - eapply ty_inv_main_ord_logic; eauto.
  - eapply ty_inv_main_ord_sel; eauto using Nat.lt_le_trans.
  - eapply ty_inv_main_ord_bnd; eauto using Nat.lt_le_trans.
  - eapply ty_inv_main_ord_nocap_bnd; eauto using Nat.lt_le_trans.
Qed.

(* Analog of ty_inv_logic_ord_has_closure *)
Lemma ty_inv_main_ord_leftmost_logic T:
    ty_inv_logic G e x T ->
    exists n,
    ty_inv_main_ord n (typ_bnd_leftmost_logctx T) T.
Proof.
  introv Hh.

  induction T using typ_bnd_simple_ind; intros.
  - exists 0. eapply ty_inv_main_ord_logic; eauto.
  -
    forwards: IHT.
    eapply ty_inv_logic_lc_bnd_replace; eauto.
    destruct H0.
    exists (S x0).

    eapply ty_inv_main_ord_nocap_bnd.
    eapply ty_inv_main_ord_le.
    eapply typ_bnd_leftmost_logctx_notbnd_le; eauto.
    eapply H0.
    eapply typ_bnd_leftmost_logctx_bnd_le.
    eauto.
Qed.

(* Analog of ty_inv_logic_ord_orctx_notorand_le *)
Lemma ty_inv_main_ord_logctx_notbnd_le n oc T:
  ty_inv_main_ord n oc (logctx_typ oc T) ->
  logctx_notbnd_le oc.
Proof.
  introv Hh.
  remember (logctx_typ oc T).
  reverts T.
  induction Hh; intros; subst.
  - eapply logctx_notbnd_le_notbnd; eauto.
  -
    forwards: IHHh; eauto.
    eapply logctx_notbnd_le_le; eauto.
  -
    forwards: IHHh; eauto.
    eapply logctx_notbnd_le_le; eauto.
  -
    forwards: IHHh; eauto.
    eapply logctx_notbnd_le_le; eauto.
Qed.



Lemma ty_inv_logic_ord_sel_lc_le n oc ocp y A z U S1 S2 Tl Tr:
    logctx_typ ocp Tl = logctx_typ oc Tr ->
    ty_inv_main_ord n oc (logctx_typ ocp (open z S1)) ->
    G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
    logctx_le oc ocp ->
    ty_inv_main_ord (S n) oc (logctx_typ ocp (typ_sel (avar_f y) A (avar_f z))).
Proof.
  introv He HhT HhU Hle.
  rename ocp into ocTU.
  rename oc into ocL.
  remember (logctx_typ ocTU (open z S1)).
  reverts Tl Tr ocTU. (* S1 S2. *)

  induction HhT; intros; subst.
  -  eapply ty_inv_main_ord_sel; eauto.
     eapply ty_inv_main_ord_logic; eauto.
  - (* sel *)
    edestruct (logctx_le_cases ocTU ocp). symmetry; eauto.
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_inv_main_ord_sel; eauto.
      rewrite <- Heqt.
      eapply ty_inv_main_ord_sel; eauto.
    + (* ocp left of ocp0  -> need induction *)

      symmetry in Heqt.
      forwards: logctx_intersect_cases_nonnil Heqt.
      destruct_all; subst.
      * (* ocp = ocp0 -> solve without inductions same as the case above *)
      eapply ty_inv_main_ord_sel; eauto.
      rewrite H4.
      eapply ty_inv_main_ord_sel; eauto.
      *
      eapply logctx_le_extend_l_inv in H2. contradiction.
      * destruct x0; simpls; inversions H4. contradiction.
      * (* actually need induction *)
        forwards: logctx_separate_cross_r Heqt H3 (typ_sel (avar_f y) A (avar_f z)).
        destruct_all.
        forwards: H5 (open z0 S0).
        destruct_all.
        forwards: IHHhT.
        2: symmetry; eauto.
        eauto.
        eapply logctx_follow_symm in H9.
        eauto using logctx_le_trans, logctx_follow_le.

        rewrite H8 in H10.
        rewrite H4.
        eapply ty_inv_main_ord_sel; eauto.
        eapply ty_inv_main_ord_nge.
        2: eapply ty_inv_main_ord_le.
        2: eapply logctx_follow_le; eapply logctx_follow_symm; eauto.
        2: eauto.
        eapply H1.
        eauto using logctx_le_trans, logctx_follow_le.
  - (* bnd *)
    edestruct (logctx_le_cases ocTU ocp). symmetry; eauto.
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_inv_main_ord_sel; eauto.
      rewrite <- Heqt.
      eapply ty_inv_main_ord_bnd; eauto.
    + (* ocp left of ocp0  -> need induction *)

      symmetry in Heqt.
      forwards: logctx_intersect_cases_nonnil Heqt.
      destruct_all; subst.
      * (* ocp = ocp0 -> solve without inductions same as the case above *)
      eapply ty_inv_main_ord_sel; eauto.
      rewrite H5.
      eapply ty_inv_main_ord_bnd; eauto.
      *
      eapply logctx_le_extend_l_inv in H3. contradiction.
      * destruct x0; simpls; inversions H5. contradiction.
      * (* actually need induction *)
        forwards: logctx_separate_cross_r Heqt H4 (typ_sel (avar_f y) A (avar_f z)).
        destruct_all.
        forwards: H6 (open x T).
        destruct_all.
        forwards: IHHhT.
        2: symmetry; eauto.
        eauto.
        eapply logctx_follow_symm in H10.
        eauto using logctx_le_trans, logctx_follow_le.

        rewrite H9 in H11.
        rewrite H5.
        eapply ty_inv_main_ord_bnd; eauto.
        eapply ty_inv_main_ord_nge.
        2: eapply ty_inv_main_ord_le.
        2: eapply logctx_follow_le; eapply logctx_follow_symm; eauto.
        2: eauto.
        eapply H2.
        eauto using logctx_le_trans, logctx_follow_le.
  - (* nocap_bnd *)
    edestruct (logctx_le_cases ocTU ocp). symmetry; eauto.
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_inv_main_ord_sel; eauto.
      rewrite <- Heqt.
      eapply ty_inv_main_ord_nocap_bnd; eauto.
    + (* ocp left of ocp0  -> need induction *)

      symmetry in Heqt.
      forwards: logctx_intersect_cases_nonnil Heqt.
      destruct_all; subst.
      * (* ocp = ocp0 -> solve without inductions same as the case above *)
      eapply ty_inv_main_ord_sel; eauto.
      rewrite H3.
      eapply ty_inv_main_ord_nocap_bnd; eauto.
      *
      eapply logctx_le_extend_l_inv in H1. contradiction.
      * destruct x0; simpls; inversions H3. contradiction.
      * (* actually need induction *)
        forwards: logctx_separate_cross_r Heqt H2 (typ_sel (avar_f y) A (avar_f z)).
        destruct_all.
        forwards: H4 typ_nocap.
        destruct_all.
        forwards: IHHhT.
        2: symmetry; eauto.
        eauto.
        eapply logctx_follow_symm in H8.
        eauto using logctx_le_trans, logctx_follow_le.

        rewrite H7 in H9.
        rewrite H3.
        eapply ty_inv_main_ord_nocap_bnd; eauto.
        eapply ty_inv_main_ord_nge.
        2: eapply ty_inv_main_ord_le.
        2: eapply logctx_follow_le; eapply logctx_follow_symm; eauto.
        2: eauto.
        eapply H0.
        eauto using logctx_le_trans, logctx_follow_le.
Qed.

Lemma ty_inv_main_ord_sel_lc n oc ocp y A z U S1 S2 Tl Tr:
    logctx_typ ocp Tl = logctx_typ oc Tr ->
    ty_inv_main_ord n oc (logctx_typ ocp (open z S1)) ->
    G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
    ty_inv_main_ord (S n) (logctx_leb ocp oc) (logctx_typ ocp (typ_sel (avar_f y) A (avar_f z))).
Proof.
  introv He HhT HhU.

  destruct (logctx_le_cases ocp oc Tl Tr He).
  - (* ocp <= oc    easy case *)
  rewrite logctx_leb_l in * by auto.
  eapply ty_inv_main_ord_sel.
  eapply ty_inv_main_ord_le.
  eauto.
  eapply HhT.
  eauto.
  eapply logctx_le_refl.
  eauto.
  - (* oc <= ocp *)
  eapply ty_inv_main_ord_le.
  eapply logctx_leb_le_comm.
  eauto.
  rewrite logctx_leb_l in * by auto.
  eapply ty_inv_logic_ord_sel_lc_le; eauto.
Qed.



Lemma ty_inv_logic_ord_bnd_lc_le n oc ocp U Tl Tr:
    logctx_typ ocp Tl = logctx_typ oc Tr ->
    ty_inv_main_ord n oc (logctx_typ ocp (open x U)) ->
    (mode_has_incap typing_mode -> G ⁏ e ⊢# typ_nocap <: open x U) ->
    typ_indep_b 0 U ->
    logctx_le oc ocp ->
    ty_inv_main_ord (S n) oc (logctx_typ ocp (typ_bnd U)).
Proof.
  introv He HhT HsU HiU Hle.
  rename ocp into ocTU.
  rename oc into ocL.
  remember (logctx_typ ocTU (open x U)).
  reverts Tl Tr ocTU. (* S1 S2. *)

  induction HhT; intros; subst.
  -  eapply ty_inv_main_ord_bnd; eauto.
     eapply ty_inv_main_ord_logic; eauto.
  - (* sel *)
    edestruct (logctx_le_cases ocTU ocp). symmetry; eauto.
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_inv_main_ord_bnd; eauto.
      rewrite <- Heqt.
      eapply ty_inv_main_ord_sel; eauto.
    + (* ocp left of ocp0  -> need induction *)

      symmetry in Heqt.
      forwards: logctx_intersect_cases_nonnil Heqt.
      destruct_all; subst.
      * (* ocp = ocp0 -> solve without inductions same as the case above *)
      eapply ty_inv_main_ord_bnd; eauto.
      rewrite H4.
      eapply ty_inv_main_ord_sel; eauto.
      *
      eapply logctx_le_extend_l_inv in H2. contradiction.
      * destruct x0; simpls; inversions H4. contradiction.
      * (* actually need induction *)
        forwards: logctx_separate_cross_r Heqt H3 (typ_bnd U).
        destruct_all.
        forwards: H5 (open z S1).
        destruct_all.
        forwards: IHHhT.
        2: symmetry; eauto.
        eauto.
        eapply logctx_follow_symm in H9.
        eauto using logctx_le_trans, logctx_follow_le.

        rewrite H8 in H10.
        rewrite H4.
        eapply ty_inv_main_ord_sel; eauto.
        eapply ty_inv_main_ord_nge.
        2: eapply ty_inv_main_ord_le.
        2: eapply logctx_follow_le; eapply logctx_follow_symm; eauto.
        2: eauto.
        eapply H1.
        eauto using logctx_le_trans, logctx_follow_le.
  - (* bnd *)
    edestruct (logctx_le_cases ocTU ocp). symmetry; eauto.
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_inv_main_ord_bnd; eauto.
      rewrite <- Heqt.
      eapply ty_inv_main_ord_bnd; eauto.
    + (* ocp left of ocp0  -> need induction *)

      symmetry in Heqt.
      forwards: logctx_intersect_cases_nonnil Heqt.
      destruct_all; subst.
      * (* ocp = ocp0 -> solve without inductions same as the case above *)
      eapply ty_inv_main_ord_bnd; eauto.
      rewrite H5.
      eapply ty_inv_main_ord_bnd; eauto.
      *
      eapply logctx_le_extend_l_inv in H3. contradiction.
      * destruct x0; simpls; inversions H5. contradiction.
      * (* actually need induction *)
        forwards: logctx_separate_cross_r Heqt H4 (typ_bnd U).
        destruct_all.
        forwards: H6 (open x T).
        destruct_all.
        forwards: IHHhT.
        2: symmetry; eauto.
        eauto.
        eapply logctx_follow_symm in H10.
        eauto using logctx_le_trans, logctx_follow_le.

        rewrite H9 in H11.
        rewrite H5.
        eapply ty_inv_main_ord_bnd; eauto.
        eapply ty_inv_main_ord_nge.
        2: eapply ty_inv_main_ord_le.
        2: eapply logctx_follow_le; eapply logctx_follow_symm; eauto.
        2: eauto.
        eapply H2.
        eauto using logctx_le_trans, logctx_follow_le.
  - (* nocap_bnd *)
    edestruct (logctx_le_cases ocTU ocp). symmetry; eauto.
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_inv_main_ord_bnd; eauto.
      rewrite <- Heqt.
      eapply ty_inv_main_ord_nocap_bnd; eauto.
    + (* ocp left of ocp0  -> need induction *)

      symmetry in Heqt.
      forwards: logctx_intersect_cases_nonnil Heqt.
      destruct_all; subst.
      * (* ocp = ocp0 -> solve without inductions same as the case above *)
      eapply ty_inv_main_ord_bnd; eauto.
      rewrite H3.
      eapply ty_inv_main_ord_nocap_bnd; eauto.
      *
      eapply logctx_le_extend_l_inv in H1. contradiction.
      * destruct x0; simpls; inversions H3. contradiction.
      * (* actually need induction *)
        forwards: logctx_separate_cross_r Heqt H2 (typ_bnd U).
        destruct_all.
        forwards: H4 typ_nocap.
        destruct_all.
        forwards: IHHhT.
        2: symmetry; eauto.
        eauto.
        eapply logctx_follow_symm in H8.
        eauto using logctx_le_trans, logctx_follow_le.

        rewrite H7 in H9.
        rewrite H3.
        eapply ty_inv_main_ord_nocap_bnd; eauto.
        eapply ty_inv_main_ord_nge.
        2: eapply ty_inv_main_ord_le.
        2: eapply logctx_follow_le; eapply logctx_follow_symm; eauto.
        2: eauto.
        eapply H0.
        eauto using logctx_le_trans, logctx_follow_le.
Qed.

Lemma ty_inv_main_ord_bnd_lc n oc ocp U Tl Tr:
    logctx_typ ocp Tl = logctx_typ oc Tr ->
    ty_inv_main_ord n oc (logctx_typ ocp (open x U)) ->
    (mode_has_incap typing_mode -> G ⁏ e ⊢# typ_nocap <: open x U) ->
    typ_indep_b 0 U ->
    ty_inv_main_ord (S n) (logctx_leb ocp oc) (logctx_typ ocp (typ_bnd U)).
Proof.
  introv He HhT HhU HiU.

  destruct (logctx_le_cases ocp oc Tl Tr He).
  - (* ocp <= oc    easy case *)
  rewrite logctx_leb_l in * by auto.
  eapply ty_inv_main_ord_bnd.
  eapply ty_inv_main_ord_le.
  eauto.
  eapply HhT.
  eauto.
  eauto.
  eapply logctx_le_refl.
  eauto.
  - (* oc <= ocp *)
  eapply ty_inv_main_ord_le.
  eapply logctx_leb_le_comm.
  eauto.
  rewrite logctx_leb_l in * by auto.
  eapply ty_inv_logic_ord_bnd_lc_le; eauto.
Qed.

Lemma ty_inv_logic_ord_nocap_bnd_lc_le n oc ocp U Tl Tr:
    logctx_typ ocp Tl = logctx_typ oc Tr ->
    ty_inv_main_ord n oc (logctx_typ ocp typ_nocap) ->
    logctx_le oc ocp ->
    ty_inv_main_ord (S n) oc (logctx_typ ocp (typ_bnd U)).
Proof.
  introv He HhT Hle.
  rename ocp into ocTU.
  rename oc into ocL.
  remember (logctx_typ ocTU typ_nocap).
  reverts Tl Tr ocTU. (* S1 S2. *)

  induction HhT; intros; subst.
  -  eapply ty_inv_main_ord_nocap_bnd; eauto.
     eapply ty_inv_main_ord_logic; eauto.
  - (* sel *)
    edestruct (logctx_le_cases ocTU ocp). symmetry; eauto.
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_inv_main_ord_nocap_bnd; eauto.
      rewrite <- Heqt.
      eapply ty_inv_main_ord_sel; eauto.
    + (* ocp left of ocp0  -> need induction *)

      symmetry in Heqt.
      forwards: logctx_intersect_cases_nonnil Heqt.
      destruct_all; subst.
      * (* ocp = ocp0 -> solve without inductions same as the case above *)
      eapply ty_inv_main_ord_nocap_bnd; eauto.
      rewrite H4.
      eapply ty_inv_main_ord_sel; eauto.
      *
      eapply logctx_le_extend_l_inv in H2. contradiction.
      * destruct x0; simpls; inversions H4.
      * (* actually need induction *)
        forwards: logctx_separate_cross_r Heqt H3 (typ_bnd U).
        destruct_all.
        forwards: H5 (open z S1).
        destruct_all.
        forwards: IHHhT.
        2: symmetry; eauto.
        eauto.
        eapply logctx_follow_symm in H9.
        eauto using logctx_le_trans, logctx_follow_le.

        rewrite H8 in H10.
        rewrite H4.
        eapply ty_inv_main_ord_sel; eauto.
        eapply ty_inv_main_ord_nge.
        2: eapply ty_inv_main_ord_le.
        2: eapply logctx_follow_le; eapply logctx_follow_symm; eauto.
        2: eauto.
        eapply H1.
        eauto using logctx_le_trans, logctx_follow_le.
  - (* bnd *)
    edestruct (logctx_le_cases ocTU ocp). symmetry; eauto.
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_inv_main_ord_nocap_bnd; eauto.
      rewrite <- Heqt.
      eapply ty_inv_main_ord_bnd; eauto.
    + (* ocp left of ocp0  -> need induction *)

      symmetry in Heqt.
      forwards: logctx_intersect_cases_nonnil Heqt.
      destruct_all; subst.
      * (* ocp = ocp0 -> solve without inductions same as the case above *)
      eapply ty_inv_main_ord_nocap_bnd; eauto.
      rewrite H5.
      eapply ty_inv_main_ord_bnd; eauto.
      *
      eapply logctx_le_extend_l_inv in H3. contradiction.
      * destruct x0; simpls; inversions H5.
      * (* actually need induction *)
        forwards: logctx_separate_cross_r Heqt H4 (typ_bnd U).
        destruct_all.
        forwards: H6 (open x T).
        destruct_all.
        forwards: IHHhT.
        2: symmetry; eauto.
        eauto.
        eapply logctx_follow_symm in H10.
        eauto using logctx_le_trans, logctx_follow_le.

        rewrite H9 in H11.
        rewrite H5.
        eapply ty_inv_main_ord_bnd; eauto.
        eapply ty_inv_main_ord_nge.
        2: eapply ty_inv_main_ord_le.
        2: eapply logctx_follow_le; eapply logctx_follow_symm; eauto.
        2: eauto.
        eapply H2.
        eauto using logctx_le_trans, logctx_follow_le.
  - (* nocap_bnd *)
    edestruct (logctx_le_cases ocTU ocp). symmetry; eauto.
    + (* ocpTU left of ocp -> stop induction *)
      eapply ty_inv_main_ord_nocap_bnd; eauto.
      rewrite <- Heqt.
      eapply ty_inv_main_ord_nocap_bnd; eauto.
    + (* ocp left of ocp0  -> need induction *)

      symmetry in Heqt.
      forwards: logctx_intersect_cases_nonnil Heqt.
      destruct_all; subst.
      * (* ocp = ocp0 -> solve without inductions same as the case above *)
        discriminate.
      *
      eapply logctx_le_extend_l_inv in H1. contradiction.
      * destruct x0; simpls; inversions H3.
      * (* actually need induction *)
        forwards: logctx_separate_cross_r Heqt H2 (typ_bnd U).
        destruct_all.
        forwards: H4 typ_nocap.
        destruct_all.
        forwards: IHHhT.
        2: symmetry; eauto.
        eauto.
        eapply logctx_follow_symm in H8.
        eauto using logctx_le_trans, logctx_follow_le.

        rewrite H7 in H9.
        rewrite H3.
        eapply ty_inv_main_ord_nocap_bnd; eauto.
        eapply ty_inv_main_ord_nge.
        2: eapply ty_inv_main_ord_le.
        2: eapply logctx_follow_le; eapply logctx_follow_symm; eauto.
        2: eauto.
        eapply H0.
        eauto using logctx_le_trans, logctx_follow_le.
Qed.

Lemma ty_inv_main_ord_nocap_bnd_lc n oc ocp U Tl Tr:
    logctx_typ ocp Tl = logctx_typ oc Tr ->
    ty_inv_main_ord n oc (logctx_typ ocp typ_nocap) ->
    ty_inv_main_ord (S n) (logctx_leb ocp oc) (logctx_typ ocp (typ_bnd U)).
Proof.
  introv He HhT.

  destruct (logctx_le_cases ocp oc Tl Tr He).
  - (* ocp <= oc    easy case *)
  rewrite logctx_leb_l in * by auto.
  eapply ty_inv_main_ord_nocap_bnd.
  eapply ty_inv_main_ord_le.
  eauto.
  eapply HhT.
  eapply logctx_le_refl.
  eauto.
  - (* oc <= ocp *)
  eapply ty_inv_main_ord_le.
  eapply logctx_leb_le_comm.
  eauto.
  rewrite logctx_leb_l in * by auto.
  eapply ty_inv_logic_ord_nocap_bnd_lc_le; eauto.
Qed.

End WithContext.
End WithTypingMode.