Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping.
Require Import PreciseTyping RecordTypes InertContext
        TightTyping Subenvironments Narrowing
        InertTightSubtyping EnvCorrespondence Indep Weakening Orctx Logctx Logpath LogpathOrder.

Require Import TypeInduction TypLogicInduction.

Definition typ_notbndsel T := 
  (forall oc x A y, T <> logctx_typ oc (typ_sel x A y)) /\
  (forall oc U, T <> logctx_typ oc (typ_bnd U)).

Lemma typ_notbndsel_and_invl T1 T2:
  typ_notbndsel (typ_and T1 T2) ->
  typ_notbndsel T1.
Proof.
  introv Hn.
  unfold typ_notbndsel in *.
  destruct_all; split; intros; intro; subst.
  - forwards: H (logctx_andl oc T2); simpls.
    false H1; eauto.
  - forwards: H0 (logctx_andl oc T2); simpls.
    false H1; eauto.
Qed.
Lemma typ_notbndsel_and_invr T1 T2:
  typ_notbndsel (typ_and T1 T2) ->
  typ_notbndsel T2.
Proof.
  introv Hn.
  unfold typ_notbndsel in *.
  destruct_all; split; intros; intro; subst.
  - forwards: H (logctx_andr T1 oc); simpls.
    false H1; eauto.
  - forwards: H0 (logctx_andr T1 oc); simpls.
    false H1; eauto.
Qed.
Lemma typ_notbndsel_or_invl T1 T2:
  typ_notbndsel (typ_or T1 T2) ->
  typ_notbndsel T1.
Proof.
  introv Hn.
  unfold typ_notbndsel in *.
  destruct_all; split; intros; intro; subst.
  - forwards: H (logctx_orl oc T2); simpls.
    false H1; eauto.
  - forwards: H0 (logctx_orl oc T2); simpls.
    false H1; eauto.
Qed.
Lemma typ_notbndsel_or_invr T1 T2:
  typ_notbndsel (typ_or T1 T2) ->
  typ_notbndsel T2.
Proof.
  introv Hn.
  unfold typ_notbndsel in *.
  destruct_all; split; intros; intro; subst.
  - forwards: H (logctx_orr T1 oc); simpls.
    false H1; eauto.
  - forwards: H0 (logctx_orr T1 oc); simpls.
    false H1; eauto.
Qed.

Lemma typ_notbndsel_sel_false x A y:
  typ_notbndsel (typ_sel x A y) ->
  False.
Proof.
  introv Hn.
  unfold typ_notbndsel in *.
  destruct_all.
  false (H logctx_nil); simpls; eauto.
Qed.
Lemma typ_notbndsel_bnd_false T:
  typ_notbndsel (typ_bnd T) ->
  False.
Proof.
  introv Hn.
  unfold typ_notbndsel in *.
  destruct_all.
  false (H0 logctx_nil); simpls; eauto.
Qed.

Lemma typ_notbndsel_lc_sel_false oc x A y:
  typ_notbndsel (logctx_typ oc (typ_sel x A y)) ->
  False.
Proof.
  introv Hn.
  induction oc; simpls;
  eauto using typ_notbndsel_sel_false,
    typ_notbndsel_or_invl, typ_notbndsel_or_invr,
    typ_notbndsel_and_invl, typ_notbndsel_and_invr.
Qed.
Lemma typ_notbndsel_lc_bnd_false oc T:
  typ_notbndsel (logctx_typ oc (typ_bnd T)) ->
  False.
Proof.
  introv Hn.
  induction oc; simpls;
  eauto using typ_notbndsel_bnd_false,
    typ_notbndsel_or_invl, typ_notbndsel_or_invr,
    typ_notbndsel_and_invl, typ_notbndsel_and_invr.
Qed.



(* Count of and constructors *)
Fixpoint typ_bndsel_count (T: typ): nat :=
  match T with
  | typ_and T1 T2 => (typ_bndsel_count T2) + (typ_bndsel_count T1)
  | typ_or T1 T2 => (typ_bndsel_count T2) + (typ_bndsel_count T1)
  | typ_sel x T1 y => 1
  | typ_bnd T1 => 1 + (typ_bndsel_count T1)
  | _ => 0
  end.

(* Analog of typ_orand_count_notorand *) 
Lemma typ_bndsel_count_notbndsel T:
  typ_notbndsel T <->
  typ_bndsel_count T = 0.
Proof.
  split.
  - introv Hn.
    pose proof (typ_opening_ind_all T).
    induction H; simpls; auto.
    +
    forwards: IHtyp_opening_ind1; eauto using typ_notbndsel_and_invl.
    forwards: IHtyp_opening_ind2; eauto using typ_notbndsel_and_invr.
    rewrite H1, H2. auto.
    +
    forwards: IHtyp_opening_ind1; eauto using typ_notbndsel_or_invl.
    forwards: IHtyp_opening_ind2; eauto using typ_notbndsel_or_invr.
    rewrite H1, H2. auto.
    + false typ_notbndsel_sel_false; eauto.
    + false typ_notbndsel_bnd_false; eauto.
  - introv He.
    pose proof (typ_opening_ind_all T).
    induction H; simpls; unfold typ_notbndsel; split; intros; try discriminate; destruct oc; simpls; try discriminate.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind1; eauto. destruct H3.
      eapply H3; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind2; eauto. destruct H3.
      eapply H3; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind1; eauto. destruct H3.
      eapply H4; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind2; eauto. destruct H3.
      eapply H4; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind1; eauto. destruct H3.
      eapply H3; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind2; eauto. destruct H3.
      eapply H3; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind1; eauto. destruct H3.
      eapply H4; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind2; eauto. destruct H3.
      eapply H4; eauto.
Qed.

(* Analog of typ_orand_count_orand *)
Lemma typ_bndsel_count_sel oc x A y:
  typ_bndsel_count (logctx_typ oc (typ_sel x A y)) <> 0.
Proof.
  intro.
  apply  <- typ_bndsel_count_notbndsel in H.
  false typ_notbndsel_lc_sel_false; eauto.
Qed.
(* Analog of typ_orand_count_orand *)
Lemma typ_bndsel_count_bnd oc T:
  typ_bndsel_count (logctx_typ oc (typ_bnd T)) <> 0.
Proof.
  intro.
  apply  <- typ_bndsel_count_notbndsel in H.
  false typ_notbndsel_lc_bnd_false; eauto.
Qed.

Inductive logctx_notbndsel_le: logctx->Prop :=
| logctx_notbndsel_le_nil:
  logctx_notbndsel_le logctx_nil
| logctx_notbndsel_le_orl oc T:
  logctx_notbndsel_le oc ->
  logctx_notbndsel_le (logctx_orl oc T)
| logctx_notbndsel_le_r oc T:
  typ_notbndsel T ->
  logctx_notbndsel_le oc ->
  logctx_notbndsel_le (logctx_orr T oc)
| logctx_notbndsel_le_andl oc T:
  logctx_notbndsel_le oc ->
  logctx_notbndsel_le (logctx_andl oc T)
| logctx_notbndsel_le_andr oc T:
  typ_notbndsel T ->
  logctx_notbndsel_le oc ->
  logctx_notbndsel_le (logctx_andr T oc).
Hint Constructors logctx_notbndsel_le: core.


Fixpoint typ_bndsel_index T: nat :=
  match T with
  | typ_bnd T => 1
  | typ_sel x A y => 1
  | typ_or T1 T2
  | typ_and T1 T2  =>
      let nL := (typ_bndsel_index T1) in
      match nL with
      (* Nothing in left -> go to right *)
      | 0 => 1 + (typ_bndsel_index T2) 
      (* In left -> go to left and add right *)
      | _ => 1 + nL + (typ_bndsel_count T2)
      end
  | _ => 0
  end.

Fixpoint typ_bndsel_leftmost_logctx T: logctx :=
  match T with
  | typ_or T1 T2 =>
      let nL := (typ_bndsel_count T1) in
      match nL with
      (* Nothing in left -> go to right *)
      | 0 => logctx_orr T1 (typ_bndsel_leftmost_logctx T2)
      (* In left -> go to left *)
      | _ => logctx_orl (typ_bndsel_leftmost_logctx T1) T2
      end
  | typ_and T1 T2 =>
      let nL := (typ_bndsel_count T1) in
      match nL with
      (* Nothing in left -> go to right *)
      | 0 => logctx_andr T1 (typ_bndsel_leftmost_logctx T2)
      (* In left -> go to left *)
      | _ => logctx_andl (typ_bndsel_leftmost_logctx T1) T2
      end
  | _ => logctx_nil
  end.

(* Analog of typ_logic_ind *)
Lemma typ_logic_bndsel_simple_ind (P: typ->Prop):
  (forall T, typ_notlog T -> typ_notbndsel T -> P T) ->
  (forall T U, P T -> P U -> P (typ_or T U)) ->
  (forall T U, P T -> P U -> P (typ_and T U)) ->
  (forall x A y, P (typ_sel x A y)) ->
  (forall T, P (typ_bnd T)) ->
  (forall T, P T).
Proof.
  introv Hnl Ho Ha.
  intros.
  induction T; eauto; eapply Hnl; hnf; split; intros; try destruct oc; simpls; discriminate.
Qed.


Lemma typ_bndsel_leftmost_logctx_notlog T:
  typ_notlog T ->
  typ_notbndsel T ->
  typ_bndsel_leftmost_logctx T = logctx_nil.
Proof.
  introv Hn.
  hnf in Hn. destruct Hn.
  destruct T; simpls; eauto.
  false H0; eauto.
  false H; eauto.
Qed.


Lemma typ_bndsel_leftmost_logctx_orl_inv oc T U:
  logctx_orl oc T = typ_bndsel_leftmost_logctx U ->
  exists U1,
    U = (typ_or U1 T) /\ oc = (typ_bndsel_leftmost_logctx U1) /\ typ_bndsel_count U1 <> 0.
Proof.
  introv Heq.
  induction U using typ_logic_bndsel_simple_ind.
  - eapply typ_bndsel_leftmost_logctx_notlog in H. rewrite H in Heq. discriminate. auto.
  - exists U1. simpls. destruct (typ_bndsel_count U1). discriminate.
    inversions Heq. eauto.
  - exists U1. simpls. destruct (typ_bndsel_count U1). discriminate.
    inversions Heq.
  - simpls. discriminate.
  - simpls. discriminate.
Qed.
Lemma typ_bndsel_leftmost_logctx_orr_inv oc T U:
  logctx_orr T oc = typ_bndsel_leftmost_logctx U ->
  exists U2,
    U = (typ_or T U2) /\ oc = (typ_bndsel_leftmost_logctx U2) /\ typ_notbndsel T.
Proof.
  introv Heq.
  induction U using typ_logic_bndsel_simple_ind.
  - eapply typ_bndsel_leftmost_logctx_notlog in H. rewrite H in Heq. discriminate. auto.
  - exists U2. simpls. remember (typ_bndsel_count U1). destruct n.
    inversions Heq.  symmetry in Heqn. apply <- typ_bndsel_count_notbndsel in Heqn. eauto.
    discriminate.
  - exists U2. simpls. destruct (typ_bndsel_count U1). discriminate.
    inversions Heq.
  - simpls. discriminate.
  - simpls. discriminate.
Qed.
Lemma typ_bndsel_leftmost_logctx_andl_inv oc T U:
  logctx_andl oc T = typ_bndsel_leftmost_logctx U ->
  exists U1,
    U = (typ_and U1 T) /\ oc = (typ_bndsel_leftmost_logctx U1) /\ typ_bndsel_count U1 <> 0.
Proof.
  introv Heq.
  induction U using typ_logic_bndsel_simple_ind.
  - eapply typ_bndsel_leftmost_logctx_notlog in H. rewrite H in Heq. discriminate. auto.
  - exists U1. simpls. destruct (typ_bndsel_count U1). discriminate.
    inversions Heq.
  - exists U1. simpls. destruct (typ_bndsel_count U1). discriminate.
    inversions Heq. eauto.
  - simpls. discriminate.
  - simpls. discriminate.
Qed.
Lemma typ_bndsel_leftmost_logctx_andr_inv oc T U:
  logctx_andr T oc = typ_bndsel_leftmost_logctx U ->
  exists U2,
    U = (typ_and T U2) /\ oc = (typ_bndsel_leftmost_logctx U2) /\ typ_notbndsel T.
Proof.
  introv Heq.
  induction U using typ_logic_bndsel_simple_ind.
  - eapply typ_bndsel_leftmost_logctx_notlog in H. rewrite H in Heq. discriminate. auto.
  - exists U2. simpls. destruct (typ_bndsel_count U1). discriminate.
    inversions Heq.

  - exists U2. simpls. remember (typ_bndsel_count U1). destruct n.
    inversions Heq.  symmetry in Heqn. apply <- typ_bndsel_count_notbndsel in Heqn. eauto.
    discriminate.
  - simpls. discriminate.
  - simpls. discriminate.
Qed.

(* Analog to typ_orand_leftmost_orctx_notorand *)
Lemma typ_bndsel_leftmost_logctx_notbndsel T:
  logctx_notbndsel_le (typ_bndsel_leftmost_logctx T).
Proof.
  remember (typ_bndsel_leftmost_logctx T) as oc.
  reverts T.
  induction oc; constructor.
  - eapply typ_bndsel_leftmost_logctx_orl_inv in Heqoc. destruct_all. eapply IHoc; eauto.
  - eapply typ_bndsel_leftmost_logctx_orr_inv in Heqoc. destruct_all. eauto.
  - eapply typ_bndsel_leftmost_logctx_orr_inv in Heqoc. destruct_all. eauto.
  - eapply typ_bndsel_leftmost_logctx_andl_inv in Heqoc. destruct_all. eapply IHoc; eauto.
  - eapply typ_bndsel_leftmost_logctx_andr_inv in Heqoc. destruct_all. eauto.
  - eapply typ_bndsel_leftmost_logctx_andr_inv in Heqoc. destruct_all. eauto.
Qed.
Hint Resolve typ_bndsel_leftmost_logctx_notbndsel: core.


(* Analog to typ_orand_leftmost_orctx_orand_typ *)
Lemma typ_bndsel_leftmost_logctx_bndsel_typ T:
  typ_bndsel_count T = 0 \/
  (exists U, T = (logctx_typ (typ_bndsel_leftmost_logctx T) (typ_bnd U))) \/
  (exists x A y, T = (logctx_typ (typ_bndsel_leftmost_logctx T) (typ_sel x A y))).
Proof.
  remember (typ_bndsel_leftmost_logctx T) as oc.
  reverts T.
  induction oc; intros; simpls.
  - destruct T; simpls; eauto 6.
    destruct (typ_bndsel_count T1); discriminate.
    destruct (typ_bndsel_count T1); discriminate.
  - eapply typ_bndsel_leftmost_logctx_orl_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto; destruct_all; subst.
    + contradiction.
    + right. left. exists x0. fequal; auto.
    + right. right. exists x0 x1 x2. fequal; auto. 
  - eapply typ_bndsel_leftmost_logctx_orr_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto. destruct_all.
    + apply -> typ_bndsel_count_notbndsel in H1.
      left. simpl. rewrite H1, H. auto.
    + right. left. exists x0. fequal; auto.
    + right. right. exists x0 x1 x2. fequal; auto.
  - eapply typ_bndsel_leftmost_logctx_andl_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto; destruct_all; subst.
    + contradiction.
    + right. left. exists x0. fequal; auto.
    + right. right. exists x0 x1 x2. fequal; auto. 
  - eapply typ_bndsel_leftmost_logctx_andr_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto. destruct_all.
    + apply -> typ_bndsel_count_notbndsel in H1.
      left. simpl. rewrite H1, H. auto.
    + right. left. exists x0. fequal; auto.
    + right. right. exists x0 x1 x2. fequal; auto.
Qed.

(* Analog of typ_orand_ind *)
Lemma typ_bndsel_simple_ind (P: typ->Prop):
  (forall T, typ_notbndsel T -> P T) ->
(*  (forall oc T U TC, typ_andor TC -> logctx_notbndsel_le oc -> P (logctx_typ oc T) -> P (logctx_typ oc U) -> P (logctx_typ oc (TC T U))) ->*)
  (forall oc x A y, logctx_notbndsel_le oc ->  P (logctx_typ oc (typ_sel x A y))) ->
  (forall oc T, logctx_notbndsel_le oc -> P (logctx_typ oc (typ_bnd T))) ->
  (forall T, P T).
Proof.
  introv Hn Hos Hob.
  intros.
  remember (typ_bndsel_count T) as n.
  reverts T.
  induction n using nat_lt_induction; destruct n; intros.
  - (* No and-under ors *)
    symmetry in Heqn.
    eapply Hn. apply <- typ_bndsel_count_notbndsel. auto.
  - (* There is an and-under-or *)
    rename H into IHn.
    forwards: typ_bndsel_leftmost_logctx_bndsel_typ T.
    destruct H.
    +
      rewrite <- Heqn in H; discriminate.
    + destruct H.
    ++
      destructc H as U1 HU.
      rewrite HU.
      eapply Hob.
      eauto.
    ++
      destructc H as x A y HU.
      rewrite HU.
      eapply Hos.
      eauto.
Qed.



(* Analog of orctx_notorand_eq_and_le*)
Lemma logctx_notbndsel_eq_sel_le oc x A y S ocp:
  logctx_notbndsel_le oc ->
  logctx_typ ocp (typ_sel x A y) = logctx_typ oc S ->
  logctx_le oc ocp.
Proof.
  introv Hn He.
  reverts ocp.
  induction Hn; intros; simpls.
  - eapply logctx_le_nil.
  - destruct ocp; simpls; inversions He.
    + eapply logctx_le_orl; eauto.
    + constructor.
  - destruct ocp; simpls; inversions He.
    + false.
      false typ_notbndsel_lc_sel_false; eauto.
    + eapply logctx_le_orr; eauto.
  - destruct ocp; simpls; inversions He.
    + eapply logctx_le_andl; eauto.
    + constructor.
  - destruct ocp; simpls; inversions He.
    + false.
      false typ_notbndsel_lc_sel_false; eauto.
    + eapply logctx_le_andr; eauto.
Qed.
Lemma logctx_notbndsel_eq_bnd_le oc T S ocp:
  logctx_notbndsel_le oc ->
  logctx_typ ocp (typ_bnd T) = logctx_typ oc S ->
  logctx_le oc ocp.
Proof.
  introv Hn He.
  reverts ocp.
  induction Hn; intros; simpls.
  - eapply logctx_le_nil.
  - destruct ocp; simpls; inversions He.
    + eapply logctx_le_orl; eauto.
    + constructor.
  - destruct ocp; simpls; inversions He.
    + false.
      false typ_notbndsel_lc_bnd_false; eauto.
    + eapply logctx_le_orr; eauto.
  - destruct ocp; simpls; inversions He.
    + eapply logctx_le_andl; eauto.
    + constructor.
  - destruct ocp; simpls; inversions He.
    + false.
      false typ_notbndsel_lc_bnd_false; eauto.
    + eapply logctx_le_andr; eauto.
Qed.

(* Analog of orctx_notorand_le_notorand *)
Lemma logctx_notbndsel_le_notbndsel ocp U:
  typ_notbndsel (logctx_typ ocp U) ->
  logctx_notbndsel_le ocp.
Proof.
  introv Hn.
  induction ocp; eauto; simpls;
  constructor; eauto using typ_notbndsel_or_invl, typ_notbndsel_or_invr, typ_notbndsel_and_invl, typ_notbndsel_and_invr.
Qed.

(* Analog of orctx_notorand_le_le *)
Lemma logctx_notbndsel_le_le ocL ocR TL TR:
  logctx_le ocL ocR ->
  logctx_typ ocL TL = logctx_typ ocR TR ->
  logctx_notbndsel_le ocR ->
  logctx_notbndsel_le ocL.
Proof.
  introv Hle He Hn.
  unfold logctx_le in *.
  reverts ocL TL TR.
  induction Hn; intros.
  - eapply logctx_le_nil_inv in Hle. subst. eauto.
  - destruct ocL; simpls; inversions Hle; eauto; inversions He; eauto.
  - destruct ocL; simpls.
    + inversions Hle; eauto.
    + inversions He. econstructor. eapply logctx_notbndsel_le_notbndsel; eauto.
    + inversions He. inversions Hle. eauto.
    + inversions He.
    + inversions He.
  - destruct ocL; simpls; inversions Hle; eauto; inversions He; eauto.
  - destruct ocL; simpls.
    + inversions Hle; eauto.
    + inversions He.
    + inversions He.
    + inversions He. econstructor. eapply logctx_notbndsel_le_notbndsel; eauto.
    + inversions He. inversions Hle. eauto.
Qed.

(* Analog of typ_orand_leftmost_orctx_le *)
Lemma typ_bndsel_leftmost_logctx_sel_le oc x A y:
  logctx_le (typ_bndsel_leftmost_logctx (logctx_typ oc (typ_sel x A y))) oc.
Proof.
  induction oc; simpls.
  - unfold logctx_le, logctx_path in *. simpls. constructor.
  - forwards: typ_bndsel_count_sel oc x A y.
    cases_if.
    eapply logctx_le_orl; eauto.
  -
    cases_if.
    eapply logctx_le_orr; eauto.
    constructor.
  - forwards: typ_bndsel_count_sel oc x A y.
    cases_if.
    eapply logctx_le_andl; eauto.
  -
    cases_if.
    eapply logctx_le_andr; eauto.
    constructor.
Qed.
(* Analog of typ_orand_leftmost_orctx_le *)
Lemma typ_bndsel_leftmost_logctx_bnd_le oc T:
  logctx_le (typ_bndsel_leftmost_logctx (logctx_typ oc (typ_bnd T))) oc.
Proof.
  induction oc; simpls.
  - unfold logctx_le, logctx_path in *. simpls. constructor.
  - forwards: typ_bndsel_count_bnd oc T.
    cases_if.
    eapply logctx_le_orl; eauto.
  -
    cases_if.
    eapply logctx_le_orr; eauto.
    constructor.
  - forwards: typ_bndsel_count_bnd oc T.
    cases_if.
    eapply logctx_le_andl; eauto.
  -
    cases_if.
    eapply logctx_le_andr; eauto.
    constructor.
Qed.

(* Analog of typ_orand_leftmost_orctx_notorand_le *)
Lemma typ_bndsel_leftmost_logctx_notbndsel_le oc T:
  logctx_notbndsel_le oc ->
  logctx_le oc (typ_bndsel_leftmost_logctx (logctx_typ oc T)).
Proof.
  introv Hn.

  remember (typ_bndsel_leftmost_logctx (logctx_typ oc T)) as o.
  reverts oc T.
  induction o; intros.
  - destruct oc; simpls.
    + constructor.
    + destruct (typ_bndsel_count (logctx_typ oc T)); discriminate.
    + destruct (typ_bndsel_count T0); discriminate.
    + destruct (typ_bndsel_count (logctx_typ oc T)); discriminate.
    + destruct (typ_bndsel_count T0); discriminate.
  -
    eapply typ_bndsel_leftmost_logctx_orl_inv in Heqo; destruct_all; subst.
    destruct oc; simpls; subst.
    + constructor.
    + inversions H. inversions Hn.
      eapply logctx_le_orl; eauto.
    + inversions H. inversions Hn.
      apply typ_bndsel_count_notbndsel in H2. contradiction.
    + inversions H.
    + inversions H.
  -
    eapply typ_bndsel_leftmost_logctx_orr_inv in Heqo; destruct_all; subst.
    destruct oc; simpls; subst.
    + constructor.
    + constructor. 
    + inversions H. inversions Hn.
      eapply logctx_le_orr; eauto.
    + inversions H.
    + inversions H.
  -
    eapply typ_bndsel_leftmost_logctx_andl_inv in Heqo; destruct_all; subst.
    destruct oc; simpls; subst.
    + constructor.
    + inversions H.
    + inversions H.
    + inversions H. inversions Hn.
      eapply logctx_le_andl; eauto.
    + inversions H. inversions Hn.
      apply typ_bndsel_count_notbndsel in H2. contradiction.

  -
    eapply typ_bndsel_leftmost_logctx_andr_inv in Heqo; destruct_all; subst.
    destruct oc; simpls; subst.
    + constructor.
    + inversions H.
    + inversions H.
    + constructor. 
    + inversions H. inversions Hn.
      eapply logctx_le_andr; eauto.

Qed.

(* Version of the above but just for bnd *)

Definition typ_notbnd T := 
  (forall oc U, T <> logctx_typ oc (typ_bnd U)).

Lemma typ_notbnd_and_invl T1 T2:
  typ_notbnd (typ_and T1 T2) ->
  typ_notbnd T1.
Proof.
  introv Hn.
  unfold typ_notbnd in *.
  intros; intro; subst.
  - forwards: Hn (logctx_andl oc T2); simpls.
    false H; eauto.
Qed.
Lemma typ_notbnd_and_invr T1 T2:
  typ_notbnd (typ_and T1 T2) ->
  typ_notbnd T2.
Proof.
  introv Hn.
  unfold typ_notbnd in *.
  intros; intro; subst.
  - forwards: Hn (logctx_andr T1 oc); simpls.
    false H; eauto.
Qed.
Lemma typ_notbnd_or_invl T1 T2:
  typ_notbnd (typ_or T1 T2) ->
  typ_notbnd T1.
Proof.
  introv Hn.
  unfold typ_notbnd in *.
  intros; intro; subst.
  - forwards: Hn (logctx_orl oc T2); simpls.
    false H; eauto.
Qed.
Lemma typ_notbnd_or_invr T1 T2:
  typ_notbnd (typ_or T1 T2) ->
  typ_notbnd T2.
Proof.
  introv Hn.
  unfold typ_notbnd in *.
  intros; intro; subst.
  - forwards: Hn (logctx_orr T1 oc); simpls.
    false H; eauto.
Qed.

Lemma typ_notbnd_bnd_false T:
  typ_notbnd (typ_bnd T) ->
  False.
Proof.
  introv Hn.
  unfold typ_notbnd in *.
  false (Hn logctx_nil); simpls; eauto.
Qed.

Lemma typ_notbnd_lc_bnd_false oc T:
  typ_notbnd (logctx_typ oc (typ_bnd T)) ->
  False.
Proof.
  introv Hn.
  induction oc; simpls;
  eauto using typ_notbnd_bnd_false,
    typ_notbnd_or_invl, typ_notbnd_or_invr,
    typ_notbnd_and_invl, typ_notbnd_and_invr.
Qed.



(* Count of and constructors *)
Fixpoint typ_bnd_count (T: typ): nat :=
  match T with
  | typ_and T1 T2 => (typ_bnd_count T2) + (typ_bnd_count T1)
  | typ_or T1 T2 => (typ_bnd_count T2) + (typ_bnd_count T1)
  | typ_bnd T1 => 1 + (typ_bnd_count T1)
  | _ => 0
  end.

(* Analog of typ_orand_count_notorand *) 
Lemma typ_bnd_count_notbnd T:
  typ_notbnd T <->
  typ_bnd_count T = 0.
Proof.
  split.
  - introv Hn.
    pose proof (typ_opening_ind_all T).
    induction H; simpls; auto.
    +
    forwards: IHtyp_opening_ind1; eauto using typ_notbnd_and_invl.
    forwards: IHtyp_opening_ind2; eauto using typ_notbnd_and_invr.
    rewrite H1, H2. auto.
    +
    forwards: IHtyp_opening_ind1; eauto using typ_notbnd_or_invl.
    forwards: IHtyp_opening_ind2; eauto using typ_notbnd_or_invr.
    rewrite H1, H2. auto.
    + false typ_notbnd_bnd_false; eauto.
  - introv He.
    pose proof (typ_opening_ind_all T).
    induction H; simpls; unfold typ_notbnd; intros; try discriminate; destruct oc; simpls; try discriminate.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind1; eauto.
      eapply H3; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind2; eauto.
      eapply H3; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind1; eauto.
      eapply H3; eauto.
    + eapply Plus.plus_is_O in He; destruct He.
      intro He. inversions He.
      forwards: IHtyp_opening_ind2; eauto.
      eapply H3; eauto.
Qed.

(* Analog of typ_orand_count_orand *)
Lemma typ_bnd_count_bnd oc T:
  typ_bnd_count (logctx_typ oc (typ_bnd T)) <> 0.
Proof.
  intro.
  apply  <- typ_bnd_count_notbnd in H.
  false typ_notbnd_lc_bnd_false; eauto.
Qed.

Inductive logctx_notbnd_le: logctx->Prop :=
| logctx_notbnd_le_nil:
  logctx_notbnd_le logctx_nil
| logctx_notbnd_le_orl oc T:
  logctx_notbnd_le oc ->
  logctx_notbnd_le (logctx_orl oc T)
| logctx_notbnd_le_r oc T:
  typ_notbnd T ->
  logctx_notbnd_le oc ->
  logctx_notbnd_le (logctx_orr T oc)
| logctx_notbnd_le_andl oc T:
  logctx_notbnd_le oc ->
  logctx_notbnd_le (logctx_andl oc T)
| logctx_notbnd_le_andr oc T:
  typ_notbnd T ->
  logctx_notbnd_le oc ->
  logctx_notbnd_le (logctx_andr T oc).
Hint Constructors logctx_notbnd_le: core.


Fixpoint typ_bnd_index T: nat :=
  match T with
  | typ_bnd T => 1
  | typ_or T1 T2
  | typ_and T1 T2  =>
      let nL := (typ_bnd_index T1) in
      match nL with
      (* Nothing in left -> go to right *)
      | 0 => 1 + (typ_bnd_index T2) 
      (* In left -> go to left and add right *)
      | _ => 1 + nL + (typ_bnd_count T2)
      end
  | _ => 0
  end.

Fixpoint typ_bnd_leftmost_logctx T: logctx :=
  match T with
  | typ_or T1 T2 =>
      let nL := (typ_bnd_count T1) in
      match nL with
      (* Nothing in left -> go to right *)
      | 0 => logctx_orr T1 (typ_bnd_leftmost_logctx T2)
      (* In left -> go to left *)
      | _ => logctx_orl (typ_bnd_leftmost_logctx T1) T2
      end
  | typ_and T1 T2 =>
      let nL := (typ_bnd_count T1) in
      match nL with
      (* Nothing in left -> go to right *)
      | 0 => logctx_andr T1 (typ_bnd_leftmost_logctx T2)
      (* In left -> go to left *)
      | _ => logctx_andl (typ_bnd_leftmost_logctx T1) T2
      end
  | _ => logctx_nil
  end.

(* Analog of typ_logic_and_count_and_l, typ_logic_and_count_and_r *)
Lemma typ_bnd_count_bnd_inv oc T:
  (typ_bnd_count (logctx_typ oc typ_nocap)) < typ_bnd_count (logctx_typ oc (typ_bnd T)).
Proof.
  induction oc; simpls.
  - eapply Nat.lt_0_succ.
  - eapply Plus.plus_lt_compat_l. eauto.
  - eapply Plus.plus_lt_compat_r. eauto.
  - eapply Plus.plus_lt_compat_l. eauto.
  - eapply Plus.plus_lt_compat_r. eauto.
Qed.

(* Analog of typ_logic_ind *)
Lemma typ_logic_bnd_simple_ind (P: typ->Prop):
  (forall T, typ_notlog T -> typ_notbnd T -> P T) ->
  (forall T U, P T -> P U -> P (typ_or T U)) ->
  (forall T U, P T -> P U -> P (typ_and T U)) ->
  (forall T, P (typ_bnd T)) ->
  (forall T, P T).
Proof.
  introv Hnl Ho Ha.
  intros.
  induction T; eauto; eapply Hnl; hnf; try split; intros; try destruct oc; simpls; try discriminate.
Qed.


Lemma typ_bnd_leftmost_logctx_notlog T:
  typ_notlog T ->
  typ_notbnd T ->
  typ_bnd_leftmost_logctx T = logctx_nil.
Proof.
  introv Hn.
  hnf in Hn. destruct Hn.
  destruct T; simpls; eauto.
  false H0; eauto.
  false H; eauto.
Qed.


Lemma typ_bnd_leftmost_logctx_orl_inv oc T U:
  logctx_orl oc T = typ_bnd_leftmost_logctx U ->
  exists U1,
    U = (typ_or U1 T) /\ oc = (typ_bnd_leftmost_logctx U1) /\ typ_bnd_count U1 <> 0.
Proof.
  introv Heq.
  induction U using typ_logic_bnd_simple_ind.
  - eapply typ_bnd_leftmost_logctx_notlog in H. rewrite H in Heq. discriminate. auto.
  - exists U1. simpls. destruct (typ_bnd_count U1). discriminate.
    inversions Heq. eauto.
  - exists U1. simpls. destruct (typ_bnd_count U1). discriminate.
    inversions Heq.
  - simpls. discriminate.
Qed.
Lemma typ_bnd_leftmost_logctx_orr_inv oc T U:
  logctx_orr T oc = typ_bnd_leftmost_logctx U ->
  exists U2,
    U = (typ_or T U2) /\ oc = (typ_bnd_leftmost_logctx U2) /\ typ_notbnd T.
Proof.
  introv Heq.
  induction U using typ_logic_bnd_simple_ind.
  - eapply typ_bnd_leftmost_logctx_notlog in H. rewrite H in Heq. discriminate. auto.
  - exists U2. simpls. remember (typ_bnd_count U1). destruct n.
    inversions Heq.  symmetry in Heqn. apply <- typ_bnd_count_notbnd in Heqn. eauto.
    discriminate.
  - exists U2. simpls. destruct (typ_bnd_count U1). discriminate.
    inversions Heq.
  - simpls. discriminate.
Qed.
Lemma typ_bnd_leftmost_logctx_andl_inv oc T U:
  logctx_andl oc T = typ_bnd_leftmost_logctx U ->
  exists U1,
    U = (typ_and U1 T) /\ oc = (typ_bnd_leftmost_logctx U1) /\ typ_bnd_count U1 <> 0.
Proof.
  introv Heq.
  induction U using typ_logic_bnd_simple_ind.
  - eapply typ_bnd_leftmost_logctx_notlog in H. rewrite H in Heq. discriminate. auto.
  - exists U1. simpls. destruct (typ_bnd_count U1). discriminate.
    inversions Heq.
  - exists U1. simpls. destruct (typ_bnd_count U1). discriminate.
    inversions Heq. eauto.
  - simpls. discriminate.
Qed.
Lemma typ_bnd_leftmost_logctx_andr_inv oc T U:
  logctx_andr T oc = typ_bnd_leftmost_logctx U ->
  exists U2,
    U = (typ_and T U2) /\ oc = (typ_bnd_leftmost_logctx U2) /\ typ_notbnd T.
Proof.
  introv Heq.
  induction U using typ_logic_bnd_simple_ind.
  - eapply typ_bnd_leftmost_logctx_notlog in H. rewrite H in Heq. discriminate. auto.
  - exists U2. simpls. destruct (typ_bnd_count U1). discriminate.
    inversions Heq.

  - exists U2. simpls. remember (typ_bnd_count U1). destruct n.
    inversions Heq.  symmetry in Heqn. apply <- typ_bnd_count_notbnd in Heqn. eauto.
    discriminate.
  - simpls. discriminate.
Qed.

(* Analog to typ_orand_leftmost_orctx_notorand *)
Lemma typ_bnd_leftmost_logctx_notbnd T:
  logctx_notbnd_le (typ_bnd_leftmost_logctx T).
Proof.
  remember (typ_bnd_leftmost_logctx T) as oc.
  reverts T.
  induction oc; constructor.
  - eapply typ_bnd_leftmost_logctx_orl_inv in Heqoc. destruct_all. eapply IHoc; eauto.
  - eapply typ_bnd_leftmost_logctx_orr_inv in Heqoc. destruct_all. eauto.
  - eapply typ_bnd_leftmost_logctx_orr_inv in Heqoc. destruct_all. eauto.
  - eapply typ_bnd_leftmost_logctx_andl_inv in Heqoc. destruct_all. eapply IHoc; eauto.
  - eapply typ_bnd_leftmost_logctx_andr_inv in Heqoc. destruct_all. eauto.
  - eapply typ_bnd_leftmost_logctx_andr_inv in Heqoc. destruct_all. eauto.
Qed.
Hint Resolve typ_bnd_leftmost_logctx_notbnd: core.


(* Analog to typ_orand_leftmost_orctx_orand_typ *)
Lemma typ_bnd_leftmost_logctx_bnd_typ T:
  typ_bnd_count T = 0 \/
  (exists U, T = (logctx_typ (typ_bnd_leftmost_logctx T) (typ_bnd U))).
Proof.
  remember (typ_bnd_leftmost_logctx T) as oc.
  reverts T.
  induction oc; intros; simpls.
  - destruct T; simpls; eauto 6.
    destruct (typ_bnd_count T1); discriminate.
    destruct (typ_bnd_count T1); discriminate.
  - eapply typ_bnd_leftmost_logctx_orl_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto; destruct_all; subst.
    + contradiction.
    + right. exists x0. fequal; auto.

  - eapply typ_bnd_leftmost_logctx_orr_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto. destruct_all.
    + apply -> typ_bnd_count_notbnd in H1.
      left. simpl. rewrite H1, H. auto.
    + right. exists x0. fequal; auto.
  - eapply typ_bnd_leftmost_logctx_andl_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto; destruct_all; subst.
    + contradiction.
    + right. exists x0. fequal; auto.
  - eapply typ_bnd_leftmost_logctx_andr_inv in Heqoc.
    destruct_all; subst.
    forwards: IHoc; eauto. destruct_all.
    + apply -> typ_bnd_count_notbnd in H1.
      left. simpl. rewrite H1, H. auto.
    + right. exists x0. fequal; auto.
Qed.

(* Analog of typ_orand_ind *)
Lemma typ_bnd_simple_ind (P: typ->Prop):
  (forall T, typ_notbnd T -> P T) ->
  (forall oc T, logctx_notbnd_le oc -> P (logctx_typ oc typ_nocap) -> P (logctx_typ oc (typ_bnd T))) ->
  (forall T, P T).
Proof.
  introv Hn Hob.
  intros.
  remember (typ_bnd_count T) as n.
  reverts T.
  induction n using nat_lt_induction; destruct n; intros.
  - (* No and-under ors *)
    symmetry in Heqn.
    eapply Hn. apply <- typ_bnd_count_notbnd. auto.
  - (* There is an and-under-or *)
    rename H into IHn.
    forwards: typ_bnd_leftmost_logctx_bnd_typ T.
    destruct H.
    +
      rewrite <- Heqn in H; discriminate.
    + destructc H as U Ht.
  
      rewrite Ht.
      eapply Hob.
      eauto.

      eapply IHn.
      2: eapply eq_refl.
      rewrite Heqn.
      rewrite Ht at 2.
      eapply typ_bnd_count_bnd_inv.
Qed.



(* Analog of orctx_notorand_eq_and_le*)

Lemma logctx_notbnd_eq_bnd_le oc T S ocp:
  logctx_notbnd_le oc ->
  logctx_typ ocp (typ_bnd T) = logctx_typ oc S ->
  logctx_le oc ocp.
Proof.
  introv Hn He.
  reverts ocp.
  induction Hn; intros; simpls.
  - eapply logctx_le_nil.
  - destruct ocp; simpls; inversions He.
    + eapply logctx_le_orl; eauto.
    + constructor.
  - destruct ocp; simpls; inversions He.
    + false.
      false typ_notbnd_lc_bnd_false; eauto.
    + eapply logctx_le_orr; eauto.
  - destruct ocp; simpls; inversions He.
    + eapply logctx_le_andl; eauto.
    + constructor.
  - destruct ocp; simpls; inversions He.
    + false.
      false typ_notbnd_lc_bnd_false; eauto.
    + eapply logctx_le_andr; eauto.
Qed.

(* Analog of orctx_notorand_le_notorand *)
Lemma logctx_notbnd_le_notbnd ocp U:
  typ_notbnd (logctx_typ ocp U) ->
  logctx_notbnd_le ocp.
Proof.
  introv Hn.
  induction ocp; eauto; simpls;
  constructor; eauto using typ_notbnd_or_invl, typ_notbnd_or_invr, typ_notbnd_and_invl, typ_notbnd_and_invr.
Qed.

(* Analog of orctx_notorand_le_le *)
Lemma logctx_notbnd_le_le ocL ocR TL TR:
  logctx_le ocL ocR ->
  logctx_typ ocL TL = logctx_typ ocR TR ->
  logctx_notbnd_le ocR ->
  logctx_notbnd_le ocL.
Proof.
  introv Hle He Hn.
  unfold logctx_le in *.
  reverts ocL TL TR.
  induction Hn; intros.
  - eapply logctx_le_nil_inv in Hle. subst. eauto.
  - destruct ocL; simpls; inversions Hle; eauto; inversions He; eauto.
  - destruct ocL; simpls.
    + inversions Hle; eauto.
    + inversions He. econstructor. eapply logctx_notbnd_le_notbnd; eauto.
    + inversions He. inversions Hle. eauto.
    + inversions He.
    + inversions He.
  - destruct ocL; simpls; inversions Hle; eauto; inversions He; eauto.
  - destruct ocL; simpls.
    + inversions Hle; eauto.
    + inversions He.
    + inversions He.
    + inversions He. econstructor. eapply logctx_notbnd_le_notbnd; eauto.
    + inversions He. inversions Hle. eauto.
Qed.

(* Analog of typ_orand_leftmost_orctx_le *)
Lemma typ_bnd_leftmost_logctx_bnd_le oc T:
  logctx_le (typ_bnd_leftmost_logctx (logctx_typ oc (typ_bnd T))) oc.
Proof.
  induction oc; simpls.
  - unfold logctx_le, logctx_path in *. simpls. constructor.
  - forwards: typ_bnd_count_bnd oc T.
    cases_if.
    eapply logctx_le_orl; eauto.
  -
    cases_if.
    eapply logctx_le_orr; eauto.
    constructor.
  - forwards: typ_bnd_count_bnd oc T.
    cases_if.
    eapply logctx_le_andl; eauto.
  -
    cases_if.
    eapply logctx_le_andr; eauto.
    constructor.
Qed.

(* Analog of typ_orand_leftmost_orctx_notorand_le *)
Lemma typ_bnd_leftmost_logctx_notbnd_le oc T:
  logctx_notbnd_le oc ->
  logctx_le oc (typ_bnd_leftmost_logctx (logctx_typ oc T)).
Proof.
  introv Hn.

  remember (typ_bnd_leftmost_logctx (logctx_typ oc T)) as o.
  reverts oc T.
  induction o; intros.
  - destruct oc; simpls.
    + constructor.
    + destruct (typ_bnd_count (logctx_typ oc T)); discriminate.
    + destruct (typ_bnd_count T0); discriminate.
    + destruct (typ_bnd_count (logctx_typ oc T)); discriminate.
    + destruct (typ_bnd_count T0); discriminate.
  -
    eapply typ_bnd_leftmost_logctx_orl_inv in Heqo; destruct_all; subst.
    destruct oc; simpls; subst.
    + constructor.
    + inversions H. inversions Hn.
      eapply logctx_le_orl; eauto.
    + inversions H. inversions Hn.
      apply typ_bnd_count_notbnd in H2. contradiction.
    + inversions H.
    + inversions H.
  -
    eapply typ_bnd_leftmost_logctx_orr_inv in Heqo; destruct_all; subst.
    destruct oc; simpls; subst.
    + constructor.
    + constructor. 
    + inversions H. inversions Hn.
      eapply logctx_le_orr; eauto.
    + inversions H.
    + inversions H.
  -
    eapply typ_bnd_leftmost_logctx_andl_inv in Heqo; destruct_all; subst.
    destruct oc; simpls; subst.
    + constructor.
    + inversions H.
    + inversions H.
    + inversions H. inversions Hn.
      eapply logctx_le_andl; eauto.
    + inversions H. inversions Hn.
      apply typ_bnd_count_notbnd in H2. contradiction.

  -
    eapply typ_bnd_leftmost_logctx_andr_inv in Heqo; destruct_all; subst.
    destruct oc; simpls; subst.
    + constructor.
    + inversions H.
    + inversions H.
    + constructor. 
    + inversions H. inversions Hn.
      eapply logctx_le_andr; eauto.

Qed.