Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode.
Require Import PreciseTyping InertContext Logctx Logpath.
Require Import WrapAction.

Section WithTypingMode.
Context {typing_mode: Mode}.

Inductive wrap_action_has_N_main: ctx -> renv -> wrap_action -> Prop :=
|  wrap_action_has_N_main_name G e y U z A S1 S2 oc:
    inert G ->
    G ⊢! y ∶ U ⪼ typ_rcd (dec_typ A S1 S2) ->
    wrap_action_has_N_main G e (wrap_action_c oc (open z S1) (typ_sel (avar_f y) A (avar_f z)))
| wrap_action_has_N_main_bnd G e oc T:
    wrap_action_has_N_main G e (wrap_action_c oc typ_nocap (typ_bnd T))
.
Hint Constructors wrap_action_has_N_main: core. 

Lemma wrap_action_has_N_main_leaf G e wa:
  wrap_action_has_N_main G e wa ->
  wrap_action_leaf wa.
Proof.
  introv Hhas.
  inversions Hhas.
  - constructor; hnf; intros; discriminate.
  - constructor; hnf; intros; discriminate.
Qed.
Lemma wrap_action_has_N_main_context_free G e oc1 oc2 T1 T2:
  wrap_action_has_N_main G e (wrap_action_c oc1 T1 T2) ->
  wrap_action_has_N_main G e (wrap_action_c oc2 T1 T2).
Proof.
  introv Hhas.
  inversions Hhas; eauto.
Qed.



Lemma wrap_action_has_N_main_injective G e wa:
  wrap_action_has_N_main G e wa ->
  wrap_action_injective wa.
Proof.
  introv Hhas Hut1 Hut2.
  inversions Hhas.
  - inversions Hut1. inversions Hut2.
    rename H3 into Hocp.
    eapply logctx_path_eq_follow in Hocp.
    forwards: logctx_follow_eq_inv Hocp; eauto.
    destruct_all; subst. auto.
  - inversions Hut1. inversions Hut2.
    rename H1 into Hocp.
    eapply logctx_path_eq_follow in Hocp.
    forwards: logctx_follow_eq_inv Hocp; eauto.
    destruct_all; subst. auto.
Qed.

Lemma wrap_action_has_N_main_direct_determined G e wa wb T1 T2 U:
  wrap_action_has_N_main G e wa ->
  wrap_action_has_N_main G e wb ->
  wrap_action_direct wa ->
  wrap_action_direct wb ->
  wrap_action_utyp T1 wa U ->
  wrap_action_utyp T2 wb U ->
  wa = wb /\ T1 = T2.
Proof.
  introv Hhas1 Hhas2 Hd1 Hd2 Hut1 Hut2.
  inversions Hhas1.
  - inversions Hut1. inversions Hut2.
    inversions Hd1. inversions Hd2.
    destruct oc0; inversions H2.
    destruct oc; inversions H3.
    simpls; subst.
    inversions Hhas2.

    lets Hu: (x_bound_unique H0 H9). subst.
    pose proof (pf_inert_unique_tight_bounds H H0 H9) as Hu. destruct Hu. subst.
    auto.
  - inversions Hut1. inversions Hut2.
    inversions Hd1. inversions Hd2.
    destruct oc0; inversions H0.
    destruct oc; inversions H1.
    simpls; subst.
    inversions Hhas2.
    auto.
Qed.

Instance wrap_action_has_N_main_determined G e:
  WrapActionDetermined (wrap_action_has_N_main G e).
Proof.
  hnf; intros.
  eapply wrap_action_has_N_main_direct_determined; eauto.
Qed.
Instance wrap_action_has_N_main_closed G e:
  WrapActionClosed (wrap_action_has_N_main G e).
Proof.
  hnf; intros.
  inversions H; destruct wa; simpls.
  - inversions H0. eauto.
  - inversions H3. eauto.
Qed.
Instance wrap_action_has_N_main_closed_inv G e:
  WrapActionClosedInv (wrap_action_has_N_main G e).
Proof.
  hnf; intros.
  inversions H; simpls.
  - eauto.
  - eauto.
Qed.
Lemma wrap_plan_has_N_main_leaf G e xs:
  List.Forall (wrap_action_has_N_main G e) xs ->
  wrap_plan_leaf xs.
Proof.
  introv Hf.
  eapply List.Forall_impl.  eapply wrap_action_has_N_main_leaf; eauto. eauto.
Qed.



End WithTypingMode.

Hint Constructors wrap_action_has_N_main: core. 
