Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypeInduction .

(** * Types that do not select members form a given variable *)

Inductive nosel : typ -> var -> Prop :=

| nosel_top : forall (x : var), nosel typ_top x

| nosel_bot : forall (x : var), nosel typ_bot x

| nosel_nocap : forall (x : var), nosel typ_nocap x

| nosel_bin : forall (x : var) (typ_bin : typ -> typ -> typ) (T1 T2 : typ),
    typ_binary typ_bin ->
    nosel T1 x ->
    nosel T2 x ->
    nosel (typ_bin T1 T2) x

| nosel_sel : forall (x : var) (y z: avar) (B : typ_label),
    y <> (avar_f x) ->
    z <> (avar_f x) ->
    nosel (typ_sel y B z) x

| nosel_rec : forall (x : var) (T : typ),
    nosel (typ_bnd T) x

| nosel_met : forall (x : var) (T1 T2 T3 : typ) (m : met_label),
    nosel T1 x ->
    nosel T2 x ->
    nosel T3 x ->
    nosel (typ_rcd (dec_met m T1 T2 T3)) x
.
Hint Constructors nosel : core.


(* 5.112 NoselFV *)
Lemma notin_fv_nosel T y:
  y \notin fv T ->
  nosel T y.
Proof.
  einduction T using typ_mut; intros; eauto.
  - revert H. eapply IHt. (* Fixes property for induction on declarations *)
  - rewrite fv_typ_eq in *; unfold fv_typ in *; eauto.
  - rewrite fv_typ_eq in *; unfold fv_typ in *; eauto.
  - rewrite fv_typ_eq in *; unfold fv_typ in *.
    rewrite notin_union in H. destructc H as Hna Hna0.
    constructor.
    replace (fv a) with (AvarFreeVar a) in Hna; eauto.
    unfold AvarFreeVar in Hna.
    intro; subst. apply Hna. rewrite in_singleton; auto.
    replace (fv a0) with (AvarFreeVar a0) in Hna0; eauto.
    unfold AvarFreeVar in Hna0.
    intro; subst. apply Hna0. rewrite in_singleton; auto.
  - rewrite fv_typ_eq in *; unfold fv_typ in *; eauto.
  - intro. rewrite fv_typ_eq in *.
    unfold fv_typ in *.
    apply (nosel_bin (typ_binary_rcd_typ t1)); eauto.
  - intro. rewrite fv_typ_eq in *.
    unfold fv_typ in *.
    apply (nosel_bin (typ_binary_rcd_fld f)); eauto.
  - intro. rewrite fv_typ_eq in *.
    unfold fv_typ in *; eauto.
Qed.

Lemma nosel_open n x y T:
  x <> y ->
  nosel T y ->
  nosel (open_rec n x T) y.
Proof.
  introv Hn Hns.
  reverts n.
  induction Hns; intros; sympls; eauto.
  - inversions H; sympls; eauto.
    eapply (nosel_bin (typ_binary_rcd_typ A)); eauto.
    eapply (nosel_bin (typ_binary_rcd_fld A)); eauto.
  - eapply nosel_sel.
    intro. symmetry in H1.
    destruct (avar_f_open_cases n x y H1); try destruct H2; subst; contradiction.
    intro. symmetry in H1.
    destruct (avar_f_open_cases n x z H1); try destruct H2; subst; contradiction.
Qed.
Hint Resolve nosel_open: core.

Lemma nosel_open_inv n x y T:
  nosel (open_rec n x T) y ->
  nosel T y.
Proof.
  introv Hns.
  remember (open_rec n x T).
  reverts n x T.
  induction Hns; intros; sympls; eauto; destruct T; try discriminate; eauto.

  - inversions H; sympls; inversions Heqt; eauto; destruct d; inversions H0.
    eapply (nosel_bin (typ_binary_rcd_typ t)); eauto.
    eapply (nosel_bin (typ_binary_rcd_fld f)); eauto.
  - inversions_tyq Heqt; eauto.
  - inversions_tyq Heqt; eauto.
  - inversions_tyq Heqt; eauto.
  - inversions_tyq Heqt; eauto.
  - inversions_tyq Heqt; eauto; constructor; intro; subst; sympls; contradiction.
  - destruct T0; inversions_tyq Heqt; eauto.
  - destruct T0; inversions_tyq Heqt; eauto.
  - destruct T0; inversions_tyq Heqt; eauto.

  - destruct T0; inversions_tyq Heqt; eauto.
  - destruct T0; inversions_tyq Heqt; eauto.
  - destruct T0; inversions_tyq Heqt; eauto.
  - destruct T0; inversions_tyq Heqt; eauto.
  - destruct T0; inversions_tyq Heqt; eauto.
  - destruct T0; inversions_tyq Heqt; eauto.
  - destruct d; inversions_tyq Heqt; eauto.
Qed.

Lemma nosel_close n x y T:
  nosel T y ->
  nosel (close_rec n x T) y.
Proof.
  introv Hns.
  reverts n.
  induction Hns; intros; autorewrite with simpl_close; eauto.
  - inversions H; autorewrite with simpl_close; eauto.
    eauto.
    eauto.
    eauto.
    eapply (nosel_bin (typ_binary_rcd_typ A)); eauto.
    eapply (nosel_bin (typ_binary_rcd_fld A)); eauto.
  - autorewrite with simpl_close.
    constructor.
    destruct y.
    autorewrite with simpl_close; eauto. rewrite avar_close_f. cases_if. discriminate. auto.
    destruct z.
    autorewrite with simpl_close; eauto. rewrite avar_close_f. cases_if. discriminate. auto.
Qed.
Hint Resolve nosel_close: core.