Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import ConfigNotations.

(** * Lemmas about collecting free variables in machine configurations. *)

Lemma null_p (G:ctx) (U:typ) (t: trm) fs Sigma1 e Lv:
   (exists null, null \in Lv) ->
   (exists null, null \in fv_config (t ; fs ; Sigma1 ; e ; Lv)).
Proof.
  introv Hnull.
  destructc Hnull as null Hnull.
  exists null.
  union_in_solver.
Qed.

Lemma ty_nothing_fresh_in_config_p (G:ctx) (U:typ) (t: trm) fs Sigma1 e Lv t2 fs2 Sigma2 e2:
   ty_nothing_fresh_in_config G U (t ; fs ; Sigma1 ; e ; Lv) ->
   ty_nothing_fresh_in_config G U (t2; fs2 ; Sigma2 ; e2; fv_config (t ; fs ; Sigma1 ; e ; Lv)).
Proof.
  introv Hfr.
  destructc Hfr as [? ?] ?.
  unfold nothing_fresh_in_config in *.
  repeat split;
  union_sub_solver.
Qed.

Lemma ty_nothing_fresh_in_config_push (G:ctx) w y vk T (U:typ) (t: trm) fs Sigma1 e Lv t2 fs2 Sigma2:
   nothing_fresh_in_config T (t ; fs ; Sigma1 ; e ; Lv) ->
   ty_nothing_fresh_in_config G U (t ; fs ; Sigma1 ; e ; Lv) ->
   ty_nothing_fresh_in_config (G § w /vk ~ T) U (t2; fs2 ; Sigma2 ; e & (w ~ y); fv_config (t ; fs ; Sigma1 ; e ; Lv)).
Proof.
  introv HfrT Hfr.
  destructc Hfr as [? ?] ?.
  unfold nothing_fresh_in_config in *.
  repeat split;  ctx_simpl_dom; simpl_dom; change (@fv ctx CtxFreeVar) with (@ctx_fv); ctx_simpl_fv;
  union_sub_solver.
Qed.

Lemma ty_nothing_fresh_in_config_push2 (G:ctx) w y T (U:typ) (t: trm) fs Sigma1 e Lv t2 fs2 Sigma2:
   nothing_fresh_in_config T (t ; fs ; Sigma1 ; e ; Lv) ->
   ty_nothing_fresh_in_config G U (t ; fs ; Sigma1 ; e ; Lv) ->
   ty_nothing_fresh_in_config (inert_G_for_item G w y T) U (t2; fs2 ; Sigma2 ; e & (w ~ y); fv_config (t ; fs ; Sigma1 ; e ; Lv)).
Proof.
  introv HfrT Hfr.
  unfold nothing_fresh_in_config in *.
  destructc Hfr as [? ?] ?.
  repeat split;  ctx_simpl_dom; simpl_dom; change (@fv ctx CtxFreeVar) with (@ctx_fv); ctx_simpl_fv; rewrite ? env_push_fv;
  change (@fv var _ ?a) with \{a};
  union_sub_solver.
Qed.

Lemma nothing_fresh_in_config_Lv_subset (A: Set) {AF: FreeVar A} (T: A) t fs Si e Lv1 Lv2:
   nothing_fresh_in_config T (t ; fs ; Si ; e ; Lv1) ->
   Lv1 \c Lv2 ->
   nothing_fresh_in_config T (t ; fs ; Si ; e ; Lv2).
Proof.
  introv Hnfr Hsu.
  unfold nothing_fresh_in_config in *.
  eapply subset_trans. eapply Hnfr.
  union_sub_solver.
Qed.
Lemma ty_nothing_fresh_in_config_Lv_subset (A: Set) {AF: FreeVar A}  G (T: A) t fs Si e Lv1 Lv2:
   ty_nothing_fresh_in_config G T (t ; fs ; Si ; e ; Lv1) ->
   Lv1 \c Lv2 ->
   ty_nothing_fresh_in_config G T (t ; fs ; Si ; e ; Lv2).
Proof.
  introv Hnfr Hsu.
  destructecs Hnfr.
  destructecs Hnfr0.
  splits. split.
  eapply subset_trans. eapply Hnfr1.
  union_sub_solver.
  eapply nothing_fresh_in_config_Lv_subset; eauto.
  eapply nothing_fresh_in_config_Lv_subset; eauto.
Qed.


Lemma ty_nothing_fresh_in_config_push2_Lv_T (G:ctx) w y T2 (U:typ) (t: trm) fs Sigma1 e Lv t2 fs2 Sigma2 Lv2:
   ty_nothing_fresh_in_config G U (t ; fs ; Sigma1 ; e ; Lv) ->
   fv_config (t ; fs ; Sigma1 ; e ; Lv) \c Lv2  ->
   fv T2 \c Lv2 ->
   ty_nothing_fresh_in_config (inert_G_for_item G w y T2) U (t2; fs2 ; Sigma2 ; e & (w ~ y); Lv2).
Proof.
  introv Hfr Hsuc HsuT.

  unfold nothing_fresh_in_config in *.
  destructc Hfr as [? ?] ?.
  repeat split;  ctx_simpl_dom; simpl_dom; change (@fv ctx CtxFreeVar) with (@ctx_fv); ctx_simpl_fv; rewrite ? env_push_fv.
  change (@fv var _ ?a) with \{a}.
  union_sub_solver_destruct_ls.
  union_sub_solver.
  union_sub_solver.
  eapply subset_trans. eapply H.
  eapply subset_trans. eapply Hsuc.
  union_sub_solver.
  union_sub_solver_destruct_ls.
  union_sub_solver.
  union_sub_solver.
  eapply subset_trans. eapply H0.
  eapply subset_trans. eapply Hsuc.
  union_sub_solver.
  union_sub_solver_destruct_ls.
  eapply subset_trans. eapply H1.
  eapply subset_trans. eapply Hsuc.
  union_sub_solver.
Qed.
