Set Implicit Arguments.

Require Import DotTactics LibExtra AbstractSyntax.
Require Import GeneralTyping TypingMode Defs RecordTypes Indep InertContext
  Substitution EWeakening Weakening .

Section WithTypingMode.
Context (typing_mode: Mode).

(** * Lemmas About Records and Record Types *)

(** [G ⊢ ds :: U]                          #<br>#
    [U] is a record type with labels [ls]  #<br>#
    [ds] are definitions with label [ls']  #<br>#
    [l \notin ls']                          #<br>#
    [―――――――――――――――――――――――――――――――――――]  #<br>#
    [l \notin ls] *)
Lemma hasnt_notin : forall G Go e (ds : defs) ls l U,
    G §§ Go ⁏e ⊢ ds ∶ U ->
    record_typ U ls ->
    defs_hasnt ds l ->
    l \notin ls.
Proof.

  Local Ltac inversion_def_typ :=
    match goal with
    | [ H: _§§_⁏_ ⊢ _ ∶d _ |- _ ] => inversions H
    end.

  introv Hds Hrec Hhasnt.
  inversions Hhasnt. gen ds. induction Hrec; intros; inversions Hds.
  - inversion_def_typ; simpl in *; case_if; apply* notin_singleton.
  - apply notin_union; split; simpl in *.
    + apply* IHHrec. case_if*.
    + inversion_def_typ; case_if; apply* notin_singleton.
Qed.

(** The type of definitions is a record type. *)
Lemma ty_defs_record_type : forall G Go e (ds :defs) T,
    G §§ Go ⁏e ⊢ ds ∶ T ->
    record_type T.
Proof.
 intros. induction H; destruct D;
    repeat match goal with
        | [ H: record_type _ |- _ ] =>
          destruct H
        | [ Hd: _ §§ _ ⁏_ ⊢ _ ∶d dec_typ _ _ _ |- _ ] =>
          inversions Hd
        | [ Hd: _§§ _ ⁏_ ⊢ _ ∶d dec_fld _ _ _ |- _ ] =>
          inversions Hd
        | [ Hd: _§§ _ ⁏_ ⊢ _ ∶d dec_met _ _ _ _ |- _ ] =>
          inversions Hd
    end.
    all: match goal with
    | [ ls: fset label,
        t: fld_label |- _ ] =>
      exists (ls \u \{ label_fld t })
    | [ ls: fset label,
        t: typ_normal_label |- _ ] =>
      exists (ls \u \{ label_typ (typ_label_normal t) })
    | [ ls: fset label,
        t: met_label |- _ ] =>
      exists (ls \u \{ label_met t })
    | [ t: fld_label |- _ ] =>
      exists \{ label_fld t }
    | [ t: typ_normal_label |- _ ] =>
      exists \{ label_typ (typ_label_normal t) }
    | [ t: met_label |- _ ] =>
      exists \{ label_met t }
    end.
    all: constructor*.
    all: try constructor.
    all: apply (hasnt_notin H); eauto.
Qed.

(** [G ⊢ ds :: ... /\ D /\ ...]       #<br>#
    [―――――――――――――――――――――――]       #<br>#
    [exists d, ds = ... /\ d /\ ...]       #<br>#
    [G ⊢ d: D]                      *)
Lemma record_has_ty_defs: forall G Go e T ds D,
  G §§ Go ⁏e ⊢ ds ∶ T ->
  record_has T D ->
  exists d, defs_has ds d /\ G §§ Go ⁏e ⊢ d ∶d D.
Proof.
  introv Hdefs Hhas. induction Hdefs.
  - inversion Hhas; subst. exists d. split.
    + unfold defs_has. simpl. cases_if; reflexivity.
    + assumption.
  - inversion Hhas; subst.
    + destruct (IHHdefs H5) as [d' [H3 H4]].
      exists d'. split.
      * unfold defs_has. simpl. cases_if. apply H3.
      * assumption.
    + exists d. split.
      * unfold defs_has. simpl. rewrite If_l; reflexivity.
      * inversions* H5.
Qed.

Lemma record_has_ty_defs_has G Go e d a R T1 T2:
  record_has R (dec_fld a T1 T2) ->
  G §§ Go ⁏ e ⊢ d ∶ R ->
  exists x,
  defs_has d (def_fld a x).
Proof.
  introv HhasR Hty.
  induction Hty.
  - (* one *)
    inversions HhasR. inversions H.
    exists (avar_f t). compute. cases_if. auto.
  - (* cons *)
    inversions HhasR.
    + (* tail *)
      forwards x Hhas: IHHty; auto.
      exists x.
      unfold defs_has, get_def.
      cases_if.
      * unfold defs_has, get_def in Hhas. auto.
    + (* head *)
      inversions H5. inversions H0.
      exists (avar_f t).
      compute. cases_if. auto.
Qed.

(* If literal defs have a field with a value
   and a record type has the same field,
   and the defs have that record type,
   then the value of the field has the type of the field.
 *)
Lemma defs_has_record_has_ty G Go e d a x R T1 T2:
  defs_has d (def_fld a x) ->
  record_has R (dec_fld a T1 T2) ->
  G §§ Go ⁏ e ⊢ d ∶ R ->
  G §§ Go ⁏ e ⊢ (def_fld a x) ∶d (dec_fld a T1 T2).
Proof.
  introv Hhasd HhasR Hty.
  revert dependent R.
  induction d.
  - (* Cannot happen in empty defs *)
    intros.
    inversions Hhasd.
  - intros.
    sympls.
    inversions Hty.
    + (* one *)
      destruct d0.
      * (* typ - mismatch *)
        compute in Hhasd. cases_if.
      * (* fld *)
        (* Match field def *)
        compute in Hhasd. cases_if. inversions C. inversions Hhasd.
        (* Match field decl *)
        inversions HhasR.
        auto.
      * (* met - mismatch *)
        compute in Hhasd. cases_if.
   + (* cons *)
     eapply defs_has_cons_inv in Hhasd. destruct Hhasd; subst.
     * (* head in defs *)
       inversions HhasR.
       ** (* tail in type - contradiction *)
         destructc (record_has_ty_defs_has H3 H1) as x0 Hhas.
         eapply (defs_has_hasnt_neq _ Hhas) in H7.
         compute in H7. contradiction.
       ** (* head in type *)
         inversions H3. auto.
     * (* tail in defs *)
       inversions HhasR.
       **
         eapply IHd; eauto.
       **
         inversions H5; inversions H4.
         eapply (defs_has_hasnt_neq _ H) in H7.
         compute in H7. contradiction.
Qed.

(* If literal defs have a field with a value
   and a record type has the same field,
   and the defs have that record type,
   then the value of the field has the type of the field.
 *)
Lemma defs_has_record_has_ty_open F1 e1 d a x y1 R T L0:
  defs_has d (def_fld a x) ->
  record_has (open y1 R) (dec_fld a T T) ->
  (forall x0 : var, x0 \notin L0 -> F1 §§ objctx_lit x0 R ⁏ e1 ⊢ open x0 d ∶ open x0 R) ->
  exists U1 U2,
  (forall x0 : var, x0 \notin L0 -> F1 §§ objctx_lit x0 R ⁏ e1 ⊢ def_fld a (open x0 x) ∶d open x0 (dec_fld a U1 U2)) /\
  (open y1 U1 = T) /\ (open y1 U2 = T).
Proof.
  introv Hhasd HhasR Hty.

  destructc (record_open_has_fld y1 R HhasR) as U1 U2 HhasRc ? ?.
  exists U1 U2.
  splits; try assumption.
  intros x0 Hx0.
  specialize (Hty x0 Hx0).

  pose proof (record_has_open x0 HhasRc) as HhasRo.
  simpl in HhasRo.
  pose proof (defs_has_open x0 Hhasd) as Hhasdo.
  simpl in Hhasdo.

  pose proof (defs_has_record_has_ty Hhasdo HhasRo Hty).
  simpl. auto.
Qed.


Lemma defs_has_record_has_ty_open_var F1 e1 d a x y1 R T L0:
  defs_has d (def_fld a (avar_f x)) ->
  record_has (open y1 R) (dec_fld a T T) ->
  (forall x0 : var, x0 \notin L0 -> F1 §§ objctx_lit x0 R ⁏ e1 ⊢ open x0 d ∶ open x0 R) ->
  exists U, (forall x0 : var, x0 \notin L0 -> F1 § x0 / v_self ~ open x0 R ⁏ e1 ⊢ x ∶ open x0 U) /\ (open y1 U = T).
Proof.
  introv Hhasd HhasR Hty.
  forwards U1 U2 Hty0 HU1 HU2: defs_has_record_has_ty_open; eauto.
  exists U1.
  split; try assumption.
  intros x0 Hx0.
  specialize (Hty0 x0 Hx0).
  inversions Hty0.
  unfold ctx_obj in H7.
  auto.
Qed.

(* If literal defs have a field with a value
   and a record type has the same field,
   and the defs have that record type,
   then the value of the field has the type of the field.
 *)
Lemma defs_has_record_has_ty_obj_fld F1 e1 d a x y1 R S T L0:
  defs_has d (def_fld a (avar_f x)) ->
  record_has (open y1 R) (dec_fld a T T) ->
  (forall x0 : var, x0 \notin L0 -> F1 §§ objctx_lit x0 R ⁏ e1 ⊢ open x0 d ∶ open x0 R) ->
  inert (F1 § y1 / v_loc ~ (typ_objref R S)) ->
  ok e1 ->
  F1 § y1 / v_loc ~ (typ_objref R S) ⁏ e1 ⊢ x ∶ T.
Proof.
  introv Hhasd HhasR Hty Hin Hoke.
  forwards U Hty0 HU: defs_has_record_has_ty_open_var; eauto.

  pick_fresh y.
  rewrite <- HU.

  forwards: (Hty0 y). eauto.

  assert (typ_indep_b 0 R). inversions Hin. inversions H6. auto.

  assert (F1 § y1 / v_loc ~ (typ_objref R S)  ⁏ e1 ⊢ y1 ∶ open y1 R).
  eapply ty_rec_elim.
  eapply ty_var_sub. eapply ty_var_var. eauto.
  eauto. auto.

  rewrite subst_intro with (x0:=y) (y0:=y1) by auto.
  rewrite subst_intro with (x0:=y) (y0:=y1) in H1 by auto.
  replace x with (x /[ y -> y1]).
  eapply @subst_end.
  eapply subst_ty_var.
  eapply ewkn_ty_var.
  6: apply H1.
  ctx_split_between y1 y.
  eapply weaken_ty.
  ctx_flatten. eapply Hty0. eauto.
  eauto. eauto.
  eauto. ctx_simpl_fv. rewrite notin_union. split; eauto.
  replace (fv (typ_and (typ_bnd R) S)) with ((fv R) \u (fv S)).
  rewrite notin_union. eauto.
  compute. eauto.
  eauto.
  eauto.
  eauto.
  rewrite subst_ne_var; eauto.
Qed.

Lemma inert_obj_typ G e T0 ds T S vk:
  G ⁏ e ⊢ lit_obj T0 ds ∶ T ->
  ctx_ok G ->

  refcap_inert_type S vk ->
  inert_typ (typ_objref T0 S) vk.
Proof.
  introv Htyl Hok Hrc.
  inversions Htyl.
  apply inert_typ_obj; auto.
  pick_fresh x.

  eapply record_type_open_rec.
  2: eapply ty_defs_record_type.
  2: apply H2.
  assert (x \notin fv T0) as Hn by notin_solve.
  apply Hn.
  notin_solve.
Qed.


End WithTypingMode.