(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax GeneralTyping TypingMode PreciseTyping
        EWeakening Indep RecordTypes InertContext Equivalence SubstCommon
        TightTyping InvertibleTyping GeneralToTight TypeApproximation
        RestrictedTightSubtyping EnvCorrespondence Substitution
        UnboundSubstitution  HeapCorrespondence TFree ConfigVarkinds
        TFreeTyping.

(** * Elimination of references *)

Definition env_elim {A} (e : env A) x a (e' : env A) :=
  exists e1 e2, e = e1 & x ~ a & e2 /\ e' = e1 & e2.

Lemma env_elim_binds_exists {A} x (a : A) e :
  ok e ->
  binds x a e ->
  exists e', env_elim e x a e' /\ ok e'.
Proof.
  reverts e.
  apply (env_ind (fun e => ok e -> binds x a e ->
                        exists e', env_elim e x a e' /\ ok e')).
  - introv _ Hc. false. eauto using binds_empty_inv.
  - intros e x' a' IH Hok Hbi.
    apply binds_push_inv in Hbi.
    destruct Hbi as [[? ?] | [Hneq Hbi]]; subst.
    + exists e. repeat(split); eauto. hnf.
      exists e (@empty A). repeat(rewrite concat_empty_r). eauto.
    + apply ok_push_inv_ok in Hok as Hok'. specialize (IH Hok' Hbi).
      destruct IH as [e0' [Helim Hoke0']].
      exists (e0' & x' ~ a'). split.
      all: destruct Helim as [e1 [e2 [? ?]]]; subst.
      * hnf. exists e1 (e2 & x' ~ a'). repeat(rewrite concat_assoc). easy.
      * constructor; eauto. apply ok_push_inv in Hok.
        destruct Hok as [Hok Hx'fr].
        rewrite dom_concat, dom_push in Hx'fr.
        repeat(rewrite notin_union in Hx'fr).
        repeat(rewrite dom_concat). repeat(rewrite notin_union). tauto.
Qed.

Lemma env_elim_ok_notin {A} (e e' : env A) w y :
  ok e -> env_elim e w y e' -> w # e'.
Proof.
  introv Hok [e1 [e2 [? ?]]]; subst.
  eapply ok_middle_inv_r in Hok as Hokr.
  eapply ok_middle_inv_l in Hok as Hokl.
  rewrite dom_concat, notin_union. easy.
Qed.

Lemma env_elim_binds {A} (e e' : env A) x a :
  ok e -> env_elim e x a e' -> binds x a e.
Proof.
  introv Hok Helim. hnf in Helim. destruct_all; subst.
  rename x0 into e0, x1 into e1. eapply binds_concat_left_ok; eauto.
Qed.

Lemma env_elim_binds_more {A} (e : env A) x a e' :
  env_elim e x a e' -> ok e ->
  forall y b, binds y b e' -> binds y b e.
Proof.
  introv Helim Hoke Hbi. destruct Helim as [e1 [e2 [? ?]]]; subst.
  apply binds_concat_inv in Hbi.
  destruct Hbi as [Hbi2 | [Hnin Hbi1]]; eauto.
  apply ok_concat_inv_l, ok_push_inv in Hoke.
  eapply binds_concat_left; eauto.
  eapply binds_concat_left; eauto. rewrite dom_single, notin_singleton.
  hnf. intros. subst. destruct Hoke as [_ Hnin1].
  apply binds_to_dom in Hbi1. eauto.
Qed.

Lemma env_elim_correct {A} (e : env A) x a e' :
  env_elim e x a e' ->
  forall y b , y <> x -> binds y b e -> binds y b e'.
Proof.
  introv Hbi. destruct Hbi as [e1 [e2 [? ?]]]; subst.
  introv Hneq Hbi. induction e2.
  - rewrite <- empty_def, concat_empty_r in *.
    apply binds_push_inv in Hbi. destruct_all; subst; eauto. congruence.
  - destruct a0. rewrite cons_to_push in *.
    rewrite concat_assoc in *.
    apply binds_push_inv in Hbi. destruct_all; subst; eauto.
Qed.

Lemma env_elim_permute {A} (e : env A) x a e' :
  ok e ->
  env_elim e x a e' ->
  exists e0, (env_binds_more e (e0 & x ~ a)
         /\ ok (e0 & x ~ a)
         /\ env_elim e x a e0
         /\ env_binds_more e0 e'
         /\ env_binds_more e' e0).
Proof.
  introv Hok Helim. destruct Helim as [e1 [e2 [Heq Heq']]]. subst.
  exists (e1 & e2).
  repeat(split); eauto 3 using ok_permute_push, binds_permute_push.
  eexists. repeat(split); eauto 3.
Qed.

Lemma env_elim_push_permute {A} (e : env A) x a e' :
  env_elim e x a e' -> ok e ->
  forall y b, binds y b (e' & x ~ a) <-> binds y b e.
Proof.
  introv Helim Hoke. destruct Helim as [e1 [e2 [? ?]]]; subst.
  split; introv Hbi.
  - apply binds_push_inv in Hbi as [[? ?] | [Hneq Hbi]]; subst;
      eauto using binds_middle_eq, ok_middle_inv_r.
    apply binds_concat_inv in Hbi as [Hbie2 | [Hnbi Hbie1]]; eauto.
  - apply binds_middle_inv in Hbi.
    destruct Hbi as [Hbi | [[Hnin [? ?]] | [Hnin [Hne Hbi]]]]; subst; eauto.
    eapply binds_push_neq; eauto.
    rewrite <- concat_assoc in Hoke. apply ok_concat_inv_r in Hoke.
    apply binds_to_dom in Hbi.
    rewrite <- concat_empty_l, concat_assoc in Hoke.
    eapply ok_middle_inv_r in Hoke. hnf. intros; subst. easy.
Qed.

Lemma env_elim_ok {A} (e : env A) x a e' :
  ok e ->
  env_elim e x a e' ->
  ok e'.
Proof.
  introv Hok Helim. destruct Helim as [e1 [e2 [Heq Heq']]]. subst.
  apply ok_concat_inv_r in Hok as Hok2.
  apply ok_concat_inv_l in Hok as Hok1.
  inversions Hok1; eauto.
Qed.

(** Context G' is G but with references w replaced by y when
    - G  = G₁(w:Tw)G₂
    - G' = G₁G₂[w=y]
    If G is an ok context and G corresponds to the environment (w ~ y)
    then G' is G but with references w eliminated by y because in this
    case we actually have
    - G  = G₁₁(y:Ty)G₁₂(w:Tw)G₂ with y,w ∉ dom(G₁₁),dom(G₁₂),dom(G₂)
    - G' = G₁₁(y:Ty)G₁₂G₂[w=y] *)
Definition ref_elim_ctx (G : ctx) (w y : var) (G' : ctx) :=
  exists vkw Tw G1 G2, (G = G1 § w / vkw ~ Tw §& G2)
                  /\ (G' = G1 §& (subst_ctx w y G2)).

Lemma ref_elim_ctx_fv_subset G G' w y :
  ref_elim_ctx G w y G' ->
  (fv G') \c (fv G \u \{y}).
Proof.
  introv [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  unfold fv, CtxFreeVar. rewrite ctx_fv_concat, ctx_fv_concat, ctx_fv_push.
  introv Hin. repeat(rewrite in_union in *). destruct_all; eauto.
  apply ctx_fv_subst_var_subset in H.
  repeat(rewrite in_union in *). destruct_all; eauto.
Qed.

Lemma ref_elim_ctx_ok_correct G w y G' :
  ctx_ok G ->
  ref_elim_ctx G w y G' ->
  w §# G'.
Proof.
  introv Hok [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  apply ctx_ok_middle_inv in Hok.
  rewrite ctx_dom_concat, ctx_dom_map, notin_union. easy.
Qed.

Lemma ref_elim_ctx_ok G G' w y :
  ctx_ok G ->
  ref_elim_ctx G w y G' ->
  ctx_ok G'.
Proof.
  introv HokG Helim. hnf in Helim.
  destruct Helim as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  apply ctx_ok_middle_inv in HokG.
  destruct_all; eauto using ctx_ok_concat_map.
Qed.

Lemma ref_elim_ctx_weakens G G' w y :
  ctx_weakens G ->
  ref_elim_ctx G w y G' ->
  ctx_weakens G'.
Proof.
  introv HwkG Helim. hnf in Helim.
  destruct Helim as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  apply ctx_weakens_concat_right in HwkG as Hwk2.
  apply ctx_weakens_concat_left in HwkG as Hwk1.
  inversions Hwk1. eauto using ctx_weakens_concat, ctx_weakens_map.
Qed.

Lemma ref_elim_ctx_sealed G G' w y :
  sealed G ->
  ref_elim_ctx G w y G' ->
  sealed G'.
Proof.
  introv Hs Helim. destruct Helim as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  rewrite ctx_sealed_concat in Hs. destruct Hs as [HS1 HS2].
  apply ctx_closed_push_G in HS1. rewrite ctx_sealed_concat.
  split; eauto using ctx_sealed_subst.
Qed.

Lemma ref_elim_ctx_vk_same G G' w y x vk :
  ctx_ok G -> ctx_weakens G ->
  ctx_k G x vk ->
  ref_elim_ctx G w y G' ->
  x <> w ->
  ctx_k G' x vk.
Proof.
  introv Hok Hwk Hvk Helim Hne. hnf in Helim.
  destruct Helim as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  apply ctx_k_in_kdom in Hvk; eauto.
  rewrite ctx_kdom_concat in Hvk; simpl in Hvk.
  assert (Hvkx : x \in ctx_kdom vk G1 \u ctx_kdom vk G2).
  { destruct (classicT (vk = vkw)); repeat(rewrite in_union in Hvk);
      repeat(rewrite in_singleton in Hvk); rewrite in_union; tauto. }
  clear Hvk; rename Hvkx into Hvk. rewrite in_union in Hvk.
  eapply ctx_ok_concat_inv in Hok.
  destruct Hok as [Hok1 [Hok2 Hdis]]. inversions Hok1. rename H1 into Hok1.
  rename H4 into Hfrw. apply ctx_weakens_concat_right in Hwk as Hwk2.
  apply ctx_weakens_concat_left in Hwk as Hwk1. inversions Hwk1.
  rename H0 into Hwk1.
  destruct Hvk as [Hvk | Hvk]; eapply ctx_in_kdom_k in Hvk; eauto.
  - destruct Hvk as [t0 [G0 Hbk]]. exists t0 G0.
    eapply ctx_binds_k_after_concat; eauto using ctx_weakens_map.
    eapply ctx_ok_concat_map. eapply ctx_ok_concat; eauto.
    simpl in Hdis. hnf in Hdis. rewrite inter_union_l in Hdis.
    hnf. eapply fset_extens; eauto using subset_empty_l.
    rewrite <- Hdis. eauto using subset_union_weak_r.
  - destruct Hvk as [t0 [G0 Hbk]].
    exists (t0 /[ w -> y]) (G1 §& (subst_ctx w y G0)).
    eauto using ctx_binds_k_after_ctx_concat_map.
Qed.

Section WithTypingMode.
Context (typing_mode : Mode).

Lemma ref_elim_ctx_inert_typ_loc_stable G G' Ty0 Ty0' w y y0 :
  ctx_ok G ->
  ref_elim_ctx G w y G' ->
  ctx_binds y0 Ty0 G ->
  ctx_binds y0 Ty0' G' ->
  inert_typ Ty0 v_loc ->
  inert_typ Ty0' v_loc.
Proof.
  introv Hok Helim Hbi0 Hbi0' Hin0. hnf in Helim.
  destruct Helim as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  eapply ctx_binds_concat_inv in Hbi0. destruct Hbi0 as [Hbi0 | Hbi0].
  inversions Hbi0.
  - false. eapply ctx_ok_middle_inv in Hok. destruct Hok as [_ [Hfrw1 Hfrw2]].
    apply ctx_binds_dom in Hbi0'.
    rewrite ctx_dom_concat, ctx_dom_map, in_union in Hbi0'.
    destruct_all; eauto.
  - rename H3 into Hne, H6 into Hbi0.
    apply ctx_ok_concat_inv in Hok.
    apply ctx_binds_concat_inv in Hbi0'. destruct Hbi0' as [Hbi0' | Hbi0'].
    2: { false. apply ctx_binds_dom in Hbi0; apply ctx_binds_dom in Hbi0'.
         rewrite ctx_dom_map in Hbi0'. destruct Hok as [_ [_ Hdis]].
         simpl in Hdis. hnf in Hdis. rewrite inter_union_l in Hdis.
         erewrite <- in_empty, <- Hdis, in_union. repeat(rewrite in_inter).
         eauto. }
    destruct Hok as [Hok _]; inversions Hok. clear H4; rename H1 into Hok.
    rewrite <- (ctx_binds_functional Hbi0 Hbi0'). easy.
  - eapply ctx_binds_map with (f := subst_var w y) in Hbi0.
    eapply ctx_binds_concat_inv in Hbi0'. destruct Hbi0' as [Hbi0' | Hbi0'].
    1: { false. apply ctx_binds_dom in Hbi0; apply ctx_binds_dom in Hbi0'.
         rewrite ctx_dom_map in Hbi0. apply ctx_ok_concat_inv in Hok.
         destruct Hok as [_ [_ Hdis]].
         simpl in Hdis. hnf in Hdis. rewrite inter_union_l in Hdis.
         erewrite <- in_empty, <- Hdis, in_union. repeat(rewrite in_inter).
         eauto. }
    rewrite <- (ctx_binds_functional Hbi0 Hbi0').
    eauto using inert_type_subst.
Qed.

Lemma ref_elim_ctx_inert_typ_loc G G' w y y0 w0 Ty0 e :
  inert G -> renv_corr G e -> binds w0 y0 e -> binds w y e ->
  ref_elim_ctx G w y G' ->
  ctx_binds y0 Ty0 G' ->
  inert_typ Ty0 v_loc.
Proof.
  introv HinG Hcor Hbie0 Hbie Helim HbiG0.
  eapply renv_corr_binds_to_ref_loc with (G:=G) in Hbie0 as Hbk; eauto.
  destruct Hbk as [TL [TR [_ [_ [HbyG [HbkyG Htycorr]]]]]].
  eapply inert_binds_inert in HbyG as HinTL; eauto.
  eapply inert_type_corr_loc in HinTL; eauto.
  eapply ref_elim_ctx_inert_typ_loc_stable
    with (Ty0:=TL) (y0:=y0) (G':=G') (G:=G); eauto.
Qed.

End WithTypingMode.

Section WithSefMode.

Context (typing_mode : Mode).
Context (Hms : mode_has_sef typing_mode).
Context (Hmm : mode_has_mutability typing_mode).

Lemma ref_elim_ctx_inert G Gw Gy vkw vky Tw Ty G' w y :
  ctx_binds_k_after w vkw Tw G Gw ->
  ctx_binds_k_after y vky Ty Gw Gy ->
  inert G ->
  ref_elim_ctx G w y G' ->
  inert G'.
Proof.
  introv Hbkw Hbky HinG Helim. hnf in Helim.
  destruct Helim as [vkw' [Tw' [G1 [G2 [? ?]]]]]; subst.

  assert (Hfry : (y §# G2) /\ (w §# G2)).
  { apply ctx_binds_k_after_initial_segment in Hbkw.
    destruct Hbkw as [G2' Heq]. rewrite Heq in HinG.
    apply ctx_binds_k_after_initial_segment in Hbky.
    destruct Hbky as [Gw' Heq']. subst.
    apply ctx_weakens_concat_inv in Heq;
      eauto using inert_concat_invert_r, inert_weakens.
    destruct Heq as [[G' [Heq ?]] | [G' [Heq ?]]]; subst.
    - apply ctx_push_concat_inv in Heq as Heq'.
      destruct Heq' as [[? _] | [G2' ?]]; subst.
      + rewrite ctx_concat_empty_r in *.
        inversions Heq. rewrite ctx_concat_empty_l in *.
        rewrite <- ctx_concat_push, <- ctx_concat_assoc in *.
        hnf. split. all: introv Hin.
        1: erewrite <- (in_empty y).
        2: erewrite <- (in_empty w).
        2: rewrite ctx_concat_assoc in HinG.
        all: apply inert_ok, ctx_ok_concat_inv in HinG;
        destruct HinG as [_ [_ Hd]].
        all: erewrite <- Hd; repeat(rewrite ctx_dom_push);
          repeat(rewrite ctx_dom_concat); repeat(rewrite ctx_dom_push).
        all: repeat(rewrite in_inter).
        all: repeat(rewrite in_union). all: repeat(rewrite in_singleton).
        all: try tauto.
      + false. clear Heq. apply inert_ok in HinG.
        apply ctx_ok_concat_inv in HinG. destruct HinG as [_ [_ Hd]].
        rewrite <- (in_empty w), <- Hd, ctx_dom_concat.
        repeat(rewrite ctx_dom_push). repeat(rewrite ctx_dom_concat).
        repeat(rewrite ctx_dom_push). repeat(rewrite in_inter).
        repeat(rewrite in_union). repeat(rewrite in_singleton). tauto.
    - apply ctx_push_concat_inv in Heq as Heq'.
      destruct Heq' as [[? _] | [G2'' Heq']]; subst.
      + rewrite ctx_concat_empty_r in *.
        inversions Heq. rewrite ctx_concat_empty_l in *.
        rewrite <- ctx_concat_push, <- ctx_concat_assoc in *.
        hnf. split. all: introv Hin.
        1: erewrite <- (in_empty y).
        2: erewrite <- (in_empty w).
        2: rewrite ctx_concat_assoc in HinG.
        all: apply inert_ok, ctx_ok_concat_inv in HinG;
        destruct HinG as [_ [_ Hd]].
        all: erewrite <- Hd; repeat(rewrite ctx_dom_push);
          repeat(rewrite ctx_dom_concat); repeat(rewrite ctx_dom_push).
        all: repeat(rewrite in_inter).
        all: repeat(rewrite in_union). all: repeat(rewrite in_singleton).
        all: try tauto.
      + rewrite ctx_concat_push in Heq. inversions Heq. rename H0 into Heq.
        apply eq_sym, ctx_weakens_concat_inv in Heq.
        2: { apply inert_weakens, ctx_weakens_concat_left in HinG.
             inversions HinG. eauto using ctx_weakens_concat_right. }
        destruct Heq as [[G' [Heq ?]] | [G' [Heq ?]]]; subst.
        * apply ctx_push_concat_inv in Heq as Heq'.
          destruct Heq' as [[? _] | [G3 ?]]; subst.
          { rewrite ctx_concat_empty_r in Heq. inversions Heq.
            rewrite ctx_concat_empty_l in *. false.
            rewrite <- ctx_concat_push, <- ctx_concat_assoc in HinG.
            apply inert_ok, ctx_ok_concat_inv in HinG.
            destruct HinG as [_ [_ Hd]].
            rewrite <- (in_empty y), <- Hd.
            repeat(rewrite ctx_dom_push). repeat(rewrite ctx_dom_concat).
            repeat(rewrite ctx_dom_push). repeat(rewrite in_inter).
            repeat(rewrite in_union). repeat(rewrite in_singleton). tauto. }
          { false. rewrite ctx_concat_push in Heq. inversions Heq.
            rewrite <- ctx_concat_push, <- ctx_concat_assoc in HinG.
            apply inert_concat_invert_r in HinG.
            rewrite <- ctx_concat_push, <- ctx_concat_assoc in HinG.
            apply inert_ok, ctx_ok_concat_inv in HinG.
            destruct HinG as [_ [_ Hd]].
            rewrite <- (in_empty w), <- Hd.
            repeat(rewrite ctx_dom_push). repeat(rewrite ctx_dom_concat).
            repeat(rewrite ctx_dom_push). repeat(rewrite in_inter).
            repeat(rewrite in_union). repeat(rewrite in_singleton). tauto. }
        * apply ctx_push_concat_inv in Heq as Heq'.
          destruct Heq' as [[? _] | [G3 ?]]; subst.
          { rewrite ctx_concat_empty_r in Heq. inversions Heq.
            rewrite ctx_concat_empty_l in *. false.
            rewrite <- ctx_concat_push, <- ctx_concat_assoc in HinG.
            apply inert_ok, ctx_ok_concat_inv in HinG.
            destruct HinG as [_ [_ Hd]].
            rewrite <- (in_empty w), <- Hd.
            repeat(rewrite ctx_dom_push). repeat(rewrite ctx_dom_concat).
            repeat(rewrite ctx_dom_push). repeat(rewrite in_inter).
            repeat(rewrite in_union). repeat(rewrite in_singleton). tauto. }
          { false. rewrite ctx_concat_push in Heq. inversions Heq.
            apply inert_concat_invert_l in HinG.
            rewrite <- ctx_concat_push in HinG.
            apply inert_ok, ctx_ok_concat_inv in HinG.
            destruct HinG as [_ [_ Hd]].
            rewrite <- (in_empty w), <- Hd.
            repeat(rewrite ctx_dom_push). repeat(rewrite ctx_dom_concat).
            repeat(rewrite ctx_dom_push). repeat(rewrite in_inter).
            repeat(rewrite in_union). repeat(rewrite in_singleton). tauto. }
  }

  induction G2; simpl in *;
    try solve[inversions HinG; auto].
  inversions HinG.
  rename H3 into HinG, H4 into Hinvk, H5 into Hint, H6 into Hwd,
    H7 into Hwf, H8 into Hwt.

  econstructor; eauto using inert_type_subst.
  - eapply IHG2; eauto.
    + inversions Hbkw; eauto. false.
      rewrite ctx_dom_concat, ctx_dom_push in Hwd.
      repeat(rewrite notin_union in Hwd). rewrite notin_singleton in Hwd.
      tauto.
    + repeat(rewrite notin_union in Hfry). tauto.
  - rewrite ctx_fv_concat, notin_union in *.
    rewrite ctx_fv_push in *. rewrite notin_union in *.
    split; try tauto. eapply notin_subst_ctx_fv; try tauto.
    all: repeat(rewrite notin_singleton in Hfry). all: apply neq_sym; tauto.
  - eapply notin_fv_subst_var; auto.
    repeat(rewrite notin_union, notin_singleton in Hfry).
    apply neq_sym. tauto.
Qed.

Lemma renv_corr_ref_elim_ctx_exists G e w y :
  binds w y e -> renv_corr G e -> ctx_ok G ->
  exists G', ref_elim_ctx G w y G'.
Proof.
  introv Hbie Hcor HokG.
  eapply renv_corr_binds_to_binds_after in Hcor; eauto.
  destruct Hcor as [TL [TR [G2 [G3 [Hbk1 [Hbk2 _]]]]]].
  eapply ctx_binds_k_after_initial_segment in Hbk1
      as [G4 ?]; subst.
  eapply ctx_binds_k_after_initial_segment in Hbk2
      as [G5 ?]; subst.
  repeat(eexists).
Qed.

Lemma ref_elim_ctx_renv_corr G G' e w y :
  fresh_telescope G -> ok e ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  renv_corr G e ->
  exists G11 G12 G2 Ty Tw,
    ((G = G11 § y / v_loc ~ Ty §& G12 § w / v_ref ~ Tw §& G2)
     /\ (y \notin ctx_fv G11)
     /\ (G' = G11 § y / v_loc ~ Ty §& G12 §& (subst_ctx w y G2))).
Proof.
  introv HftG Hoke Helim Hbi Hcor.
  assert (HokG : ctx_ok G) by eauto using fresh_telescope_ok.
  assert (HwkG : ctx_weakens G) by eauto using fresh_telescope_weakens.
  hnf in Helim. apply (renv_corr_binds_to_binds_after HokG Hcor) in Hbi.
  destruct Hbi as [Ty [Tw [G2 [G3 [Hbkw [Hbky Htycor]]]]]].
  apply ctx_binds_k_after_initial_segment in Hbkw.
  destruct Hbkw as [G3' ?]; subst.
  apply ctx_binds_k_after_initial_segment in Hbky.
  destruct Hbky as [G4' ?]; subst.
  destruct Helim as [vkw [Tw0 [G1' [G2' [Heq Heq']]]]]; subst.
  apply eq_sym in Heq. apply ctx_weakens_concat_inv in Heq;
    eauto using ctx_weakens_concat_right.
  destruct Heq as [[G' [Heq1 Heq3]] | [G' [Heq3 Heq2]]]; subst.
  - apply ctx_push_concat_inv in Heq1 as Heq1'.
    destruct Heq1' as [[? Heq1'] | [G'' ?]]; subst.
    + rewrite ctx_concat_empty_r in *.
      rewrite ctx_concat_empty_l in *.
      inversions Heq1'. repeat(exists); repeat(split); eauto.
      rewrite <- ctx_concat_push, <- ctx_concat_assoc in HftG.
      apply fresh_telescope_concat_invert_l in HftG. inversions HftG; auto.
    + simpl in Heq1. inversions Heq1.
      false. rewrite ctx_concat_assoc in HokG.
      apply ctx_ok_concat_l in HokG. rewrite ctx_concat_push in HokG.
      inversions HokG. clear H1. rename H4 into Hfrw.
      rewrite ctx_dom_concat, ctx_dom_push, notin_union,
        notin_union, notin_singleton in Hfrw. tauto.
  - apply ctx_push_concat_inv in Heq3 as Heq3'.
    destruct Heq3' as [[? Heq3'] | [G'' ?]]; subst.
    + rewrite ctx_concat_empty_r in *.
      rewrite ctx_concat_empty_l in *. inversions Heq3'.
      repeat(exists); repeat(split); eauto.
      rewrite <- ctx_concat_push, <- ctx_concat_assoc in HftG.
      apply fresh_telescope_concat_invert_l in HftG. inversions HftG; auto.
    + simpl in Heq3. inversions Heq3. rename H0 into Heq3.
      apply eq_sym in Heq3. eapply ctx_weakens_concat_inv in Heq3.
      2: { apply ctx_weakens_concat_left in HwkG.
           rewrite <- ctx_concat_push in HwkG.
           apply ctx_weakens_concat_right in HwkG.
           inversions HwkG; eauto. }
      destruct Heq3 as [[G0 [Heq3 ?]] | [G0 [Heq3 ?]]]; subst.
      * apply ctx_push_concat_inv in Heq3 as Heq3'.
        destruct Heq3' as [[? Heq3'] | [G0' ?]]; subst.
        { inversions Heq3'.
          rewrite ctx_concat_empty_l in *.
          rewrite ctx_concat_empty_r in *.
          false. apply ctx_ok_concat_l in HokG.
          inversions HokG. clear H1. rename H4 into Hfry.
          rewrite ctx_dom_concat, ctx_dom_push, notin_union,
            notin_union, notin_singleton in Hfry. tauto. }
        { simpl in Heq3. inversions Heq3. false.
          apply ctx_ok_concat_l in HokG.
          inversions HokG. clear H1. rename H4 into Hfrw.
          rewrite ctx_dom_concat, ctx_dom_push, notin_union,
            notin_union, notin_singleton in Hfrw.
          rewrite ctx_dom_concat, ctx_dom_push, notin_union,
            notin_union, notin_singleton in Hfrw.
          tauto. }
      * apply ctx_push_concat_inv in Heq3 as Heq3'.
        destruct Heq3' as [[? Heq3'] | [G0' ?]]; subst.
        { inversions Heq3'.
          rewrite ctx_concat_empty_l in *.
          rewrite ctx_concat_empty_r in *.
          false. apply ctx_ok_concat_l in HokG.
          inversions HokG. clear H1. rename H4 into Hfrw.
          rewrite ctx_dom_concat, ctx_dom_push, notin_union,
            notin_union, notin_singleton in Hfrw. tauto. }
        { simpl in Heq3. inversions Heq3. false.
          apply ctx_ok_concat_l in HokG.
          inversions HokG. clear H1. rename H4 into Hfrw.
          rewrite ctx_dom_concat, ctx_dom_push, notin_union,
            notin_union, notin_singleton in Hfrw.
          rewrite ctx_dom_concat, ctx_dom_push, notin_union,
            notin_union, notin_singleton in Hfrw.
          tauto. }
Qed.

Lemma ref_elim_ctx_notin_fv G G' e w y x :
  fresh_telescope G -> ok e ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  renv_corr G e ->
  x \notin ctx_fv G ->
  x <> w -> x <> y ->
  x \notin ctx_fv G'.
Proof.
  introv Hfrt Hoke Helim Hbiwy Hcorr Hnin Hnew Hney.
  eapply ref_elim_ctx_renv_corr in Helim; eauto.
  destruct Helim as [G11 [G12 [G2 [Ty [Tw [? [Hfry ?]]]]]]]; subst.
  repeat(repeat(rewrite ctx_fv_concat in *); repeat(simpl in *)).
  repeat(rewrite notin_union in *). repeat(split); try tauto.
  eapply notin_subst_ctx_fv; eauto; try tauto.
Qed.

Lemma ref_elim_typ_corr G G' e w y :
  fresh_telescope G -> ok e ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  renv_corr G e ->
  exists Tw Ty, ctx_binds w Tw G /\ ctx_binds y Ty G' /\ ctx_binds y Ty G
           /\ typ_corr Tw Ty.
Proof.
  intros HftG Hoke Helim Hbie HcorGe.
  eapply ref_elim_ctx_renv_corr in Helim; eauto.
  destruct Helim as [G11 [G12 [G2 [Ty [Tw [? [Hfry ?]]]]]]].

  assert (HbiwG : ctx_binds w Tw G).
  { subst. apply ctx_binds_concat_left;
      eauto using fresh_telescope_weakens, ctx_weakens_concat_left,
      ctx_weakens_concat_right.
    rewrite <- ctx_concat_push, <- ctx_concat_assoc in HftG.
    apply fresh_telescope_concat_invert_r in HftG.
    rewrite ctx_concat_single, <- ctx_concat_assoc in HftG.
    apply fresh_telescope_concat_invert_r, fresh_telescope_ok,
      ctx_ok_concat_inv in HftG.
    destruct HftG as [_ [_ Hdis]]. simpl in Hdis.
    rewrite union_empty_r in Hdis. hnf. intros. hnf in Hdis.
    erewrite <- in_empty, <- Hdis, in_inter, in_singleton. eauto. }

  assert (HbiyG : ctx_binds y Ty G).
  { subst. do 2 (rewrite ctx_concat_single).
    repeat (apply ctx_binds_concat_left;
            eauto using fresh_telescope_weakens, ctx_weakens_concat_left,
              ctx_weakens_concat_right, ctx_weakens_map).
    rewrite <- ctx_concat_push in HftG.
    + apply fresh_telescope_concat_invert_l,
        fresh_telescope_concat_invert_r, fresh_telescope_weakens
        in HftG. inversions HftG. easy.
    + apply fresh_telescope_concat_invert_l in HftG. inversions HftG.
      clear H4 H5 H6. rename H3 into HftG.
      rewrite ctx_concat_single, <- ctx_concat_assoc in HftG.
      apply fresh_telescope_concat_invert_r,
        fresh_telescope_ok, ctx_ok_concat_inv in HftG.
      destruct HftG as [_ [_ Hdis]]. simpl in Hdis.
      rewrite union_empty_r in Hdis. hnf. intros. hnf in Hdis.
      erewrite <- in_empty, <- Hdis, in_inter, in_singleton. eauto.
    + simpl. rewrite union_empty_r, notin_singleton.
      apply fresh_telescope_concat_invert_l, fresh_telescope_ok in HftG.
      inversions HftG. clear H1. rename H4 into Hfrw.
      rewrite ctx_dom_concat in Hfrw. simpl in Hfrw. hnf. intros. subst.
      repeat(rewrite notin_union in Hfrw). rewrite notin_singleton in *.
      tauto.
    + do 2 (rewrite ctx_concat_single in HftG).
      repeat(rewrite <- ctx_concat_assoc in HftG).
      apply fresh_telescope_concat_invert_r,
        fresh_telescope_ok, ctx_ok_concat_inv in HftG.
      destruct HftG as [_ [_ Hdis]]. simpl in Hdis.
      repeat(rewrite ctx_dom_concat in Hdis).
      simpl in Hdis. repeat(rewrite union_empty_r in Hdis).
      hnf. intros. hnf in Hdis.
      erewrite <- in_empty, <- Hdis, in_inter, in_singleton.
      repeat(erewrite in_union). eauto. }

  exists Tw Ty. repeat(split); eauto using renv_corr_binds_ref_loc_corr,
    fresh_telescope_ok.

  subst. repeat (apply ctx_binds_concat_left;
                 eauto using fresh_telescope_weakens, ctx_weakens_concat_left,
                   ctx_weakens_concat_right, ctx_weakens_map).
  rewrite <- ctx_concat_push in HftG.
  + apply fresh_telescope_concat_invert_l,
      fresh_telescope_concat_invert_r, fresh_telescope_weakens
      in HftG. inversions HftG. easy.
  + apply fresh_telescope_concat_invert_l in HftG. inversions HftG.
    clear H4 H5 H6. rename H3 into HftG.
    rewrite ctx_concat_single, <- ctx_concat_assoc in HftG.
    apply fresh_telescope_concat_invert_r,
      fresh_telescope_ok, ctx_ok_concat_inv in HftG.
    destruct HftG as [_ [_ Hdis]]. simpl in Hdis.
    rewrite union_empty_r in Hdis. hnf. intros. hnf in Hdis.
    erewrite <- in_empty, <- Hdis, in_inter, in_singleton. eauto.
  + do 2 (rewrite ctx_concat_single in HftG).
    repeat(rewrite <- ctx_concat_assoc in HftG).
    apply fresh_telescope_concat_invert_r,
      fresh_telescope_ok, ctx_ok_concat_inv in HftG.
    destruct HftG as [_ [_ Hdis]]. simpl in Hdis.
    repeat(rewrite ctx_dom_concat in Hdis).
    simpl in Hdis. repeat(rewrite union_empty_r in Hdis).
    hnf. intros. hnf in Hdis. rewrite ctx_dom_map in *.
    erewrite <- in_empty, <- Hdis, in_inter, in_singleton.
    repeat(erewrite in_union). eauto.

Qed.

Lemma ref_elim_typ_end G Go G' e w y (t : trm) T :
  fresh_telescope G ->
  ctx_ok (G §§& Go) -> ok (e & w ~ y) ->
  renv_corr G (e & w ~ y) ->
  G §§ Go ⁏ (e & w ~ y) ⊢ t ∶ T ->
  ref_elim_ctx G w y G' ->
  G' §§ (subst_objctx w y Go) ⁏ e ⊢ (t /[ w -> y]) ∶ (T /[ w -> y]).
Proof.
  introv HftG HokG Hoke Hcorr Hty Helim.
  apply ref_elim_ctx_renv_corr with (e := e & w ~ y) in Helim;
    eauto using ctx_ok_obj_inv_l.
  destruct Helim as [G11 [G12 [G2 [Ty [Tw [HeqG [Hyfr HeqG']]]]]]]. subst.
  apply ok_push_inv in Hoke. destruct Hoke as [Hoke Hfrwe].

  assert (Hwk : ctx_weakens (G12 §& subst_ctx w y G2)).
  { apply fresh_telescope_concat_invert_r in HftG as HftG2.
    apply fresh_telescope_concat_invert_l in HftG as HftG1.
    inversions HftG1. apply fresh_telescope_concat_invert_r in H3.
    apply ctx_weakens_concat;
      eauto using fresh_telescope_weakens, ctx_weakens_map. }
  eapply subst_ty_trm; eauto.

  - (* Have G₁₁(y:Ty)G₁₂(w:Tw)G₂ ⊢ t : T so G₁₁(y:Ty)G₁₂(w:Tw)G₂ must
       be well-formed. And so w can't be free in the initial telescope
       before it thanks to the freshness condition of inert contexts. *)
    apply fresh_telescope_concat_invert_l in HftG. inversions HftG; auto.

  - apply renv_notin_dom_fv; auto.
    hnf. introv Hwin. apply (fv_env_to_binds Hoke) in Hwin.
    destruct Hwin as [w0 Hbie].
    eapply binds_push_neq with (x2 := w) (v2 := y) in Hbie.
    2: { hnf. intros; subst. eauto using binds_to_dom. }
    apply ctx_ok_obj_inv_l in HokG.
    pose proof (renv_corr_binds_loc HokG Hcorr Hbie) as Hbkloc.
    assert (ctx_k (((G11 § y / v_loc ~ Ty) §& G12 § w / v_ref ~ Tw) §& G2) w
                  v_ref) as Hbkref.
    { hnf. exists Tw ((G11 § y / v_loc ~ Ty) §& G12).
      eauto using fresh_telescope_weakens, ctx_weakens_concat_right,
        ctx_binds_k_after_concat. }
    pose proof (ctx_k_eq Hbkloc Hbkref) as Hcontra. congruence.

  - rewrite subst_fresh.

    + (* Ty is a subtype of Tw by renv_corr *)
      assert (Htycor : typ_corr Tw Ty).
      { inversions Hcorr.
        - false. eauto using empty_push_inv.
        - apply eq_push_inv in H. destruct H as [? [? ?]]; subst.
          replace Ty with Ty0. replace Tw with Tw0. auto.
          + apply ctx_binds_k_after_binds in H0.
            assert (ctx_binds w Tw
                              (((G11 § y / v_loc ~ Ty)
                                  §& G12 § w / v_ref ~ Tw) §& G2)).
            { eapply ctx_binds_concat_left;
                eauto 3 using fresh_telescope_weakens,
                ctx_weakens_concat_right.
              apply ctx_ok_obj_inv_l, ctx_ok_middle_inv in HokG.
              tauto. }
            eauto using ctx_binds_functional.
          + pose proof (ctx_binds_k_after_trans H1 H0
                          (fresh_telescope_ok HftG))
              as Hbiy.
            apply ctx_binds_k_after_binds in Hbiy.
            assert (ctx_binds y Ty
                              (((G11 § y / v_loc ~ Ty)
                                  §& G12 § w / v_ref ~ Tw) §& G2)).
            { rewrite <- ctx_concat_push, <- ctx_concat_assoc in HftG.
              rewrite <- ctx_concat_push, <- ctx_concat_assoc.
              eapply ctx_binds_concat_left;
                eauto 3 using fresh_telescope_weakens,
                ctx_weakens_concat_right.
              apply fresh_telescope_ok, ctx_ok_middle_inv in HftG. tauto. }
            eauto using ctx_binds_functional. }

      assert (Hoksubst : ctx_ok ((G11 § y / v_loc ~ Ty)
                                   §& G12 §& subst_ctx w y G2)).
      { apply fresh_telescope_ok, ctx_ok_middle_inv in HftG.
        clear HokG. destruct HftG as [HokG [Hwnin1 Hwnin2]].
        rewrite <- ctx_concat_assoc in HokG.
        apply ctx_ok_concat_inv in HokG.
        destruct HokG as [Hok11 [Hok2 HD1]].
        apply ctx_ok_concat_inv in Hok2.
        destruct Hok2 as [Hok12 [Hok2 HD2]].
        apply ctx_ok_concat. apply ctx_ok_concat; eauto.
        - simpl. hnf. hnf in HD1. simpl in HD1.
          rewrite ctx_dom_concat, inter_union_r in HD1.
          apply fset_extens; eauto using subset_empty_l.
          rewrite <- HD1. auto using subset_union_weak_l.
        - rewrite <- ctx_concat_empty_l.
          apply ctx_ok_concat_map.
          rewrite ctx_concat_empty_l. auto.
        - rewrite ctx_dom_map. rewrite ctx_dom_concat in HD1.
          hnf. hnf in HD1. rewrite inter_union_r in HD1.
          apply fset_extens; auto using subset_empty_l.
          rewrite ctx_dom_concat. rewrite inter_union_l.
          hnf in HD2. rewrite HD2, union_empty_r.
          rewrite <- HD1. auto using subset_union_weak_r. }

      eapply typ_corr_subtyp with (e:=e) in Htycor; eauto.
      eapply ty_var_sub; eauto.
      eapply ty_var_var. rewrite <- ctx_concat_assoc in Hoksubst.
      rewrite <- ctx_concat_assoc, <- ctx_concat_empty_r.
      apply ctx_binds_weaken; simpl; eauto.

    + (* Have G₁₁(y:Ty)G₁₂(w:Tw)G₂ ⊢ t : T so G₁₁(y:Ty)G₁₂(w:Tw)G₂ must
       be well-formed. And so w can't be free in Tw again by
       freshness condition of inert contexts. *)
      apply fresh_telescope_concat_invert_l in HftG. inversions HftG; auto.

  - (* Because renv_corr and ctx_binds_k_after, the segment between y
       and w that is G₁₂ must consist only of pushes. *)
    intros _ Hvis. apply ctx_vis_objctx_concat. rewrite <- ctx_concat_assoc.
    eauto using ctx_vis_weakens.
Qed.

(* This is the same proof as ref_elim_typ_end so make the
   proof use syntax classes *)
Lemma ref_elim_typ_var_end G Go G' e w y (x : var) T :
  fresh_telescope G ->
  ctx_ok (G §§& Go) -> ok (e & w ~ y) ->
  renv_corr G (e & w ~ y) ->
  G §§ Go ⁏ (e & w ~ y) ⊢ x ∶ T ->
  ref_elim_ctx G w y G' ->
  G' §§ (subst_objctx w y Go) ⁏ e ⊢ (x /[ w -> y]) ∶ (T /[ w -> y]).
Proof.
  introv HftG HokG Hoke Hcorr Hty Helim.
  apply ref_elim_ctx_renv_corr with (e := e & w ~ y) in Helim;
    eauto using ctx_ok_obj_inv_l.
  destruct Helim as [G11 [G12 [G2 [Ty [Tw [HeqG [Hyfr HeqG']]]]]]]. subst.
  apply ok_push_inv in Hoke. destruct Hoke as [Hoke Hfrwe].

  assert (Hwk : ctx_weakens (G12 §& subst_ctx w y G2)).
  { apply fresh_telescope_concat_invert_r in HftG as HftG2.
    apply fresh_telescope_concat_invert_l in HftG as HftG1.
    inversions HftG1. apply fresh_telescope_concat_invert_r in H3.
    apply ctx_weakens_concat;
      eauto using fresh_telescope_weakens, ctx_weakens_map. }
  eapply subst_ty_var; eauto.

  - (* Have G₁₁(y:Ty)G₁₂(w:Tw)G₂ ⊢ t : T so G₁₁(y:Ty)G₁₂(w:Tw)G₂ must
       be well-formed. And so w can't be free in the initial telescope
       before it thanks to the freshness condition of inert contexts. *)
    apply fresh_telescope_concat_invert_l in HftG. inversions HftG; auto.

  - apply renv_notin_dom_fv; auto.
    hnf. introv Hwin. apply (fv_env_to_binds Hoke) in Hwin.
    destruct Hwin as [w0 Hbie].
    eapply binds_push_neq with (x2 := w) (v2 := y) in Hbie.
    2: { hnf. intros; subst. eauto using binds_to_dom. }
    apply ctx_ok_obj_inv_l in HokG.
    pose proof (renv_corr_binds_loc HokG Hcorr Hbie) as Hbkloc.
    assert (ctx_k (((G11 § y / v_loc ~ Ty) §& G12 § w / v_ref ~ Tw) §& G2) w
                  v_ref) as Hbkref.
    { hnf. exists Tw ((G11 § y / v_loc ~ Ty) §& G12).
      eauto using fresh_telescope_weakens, ctx_weakens_concat_right,
        ctx_binds_k_after_concat. }
    pose proof (ctx_k_eq Hbkloc Hbkref) as Hcontra. congruence.

  - rewrite subst_fresh.

    + (* Ty is a subtype of Tw by renv_corr *)
      assert (Htycor : typ_corr Tw Ty).
      { inversions Hcorr.
        - false. eauto using empty_push_inv.
        - apply eq_push_inv in H. destruct H as [? [? ?]]; subst.
          replace Ty with Ty0. replace Tw with Tw0. auto.
          + apply ctx_binds_k_after_binds in H0.
            assert (ctx_binds w Tw
                              (((G11 § y / v_loc ~ Ty)
                                  §& G12 § w / v_ref ~ Tw) §& G2)).
            { eapply ctx_binds_concat_left;
                eauto 3 using fresh_telescope_weakens,
                ctx_weakens_concat_right.
              apply ctx_ok_obj_inv_l, ctx_ok_middle_inv in HokG.
              tauto. }
            eauto using ctx_binds_functional.
          + pose proof (ctx_binds_k_after_trans H1 H0
                          (fresh_telescope_ok HftG))
              as Hbiy.
            apply ctx_binds_k_after_binds in Hbiy.
            assert (ctx_binds y Ty
                              (((G11 § y / v_loc ~ Ty)
                                  §& G12 § w / v_ref ~ Tw) §& G2)).
            { rewrite <- ctx_concat_push, <- ctx_concat_assoc in HftG.
              rewrite <- ctx_concat_push, <- ctx_concat_assoc.
              eapply ctx_binds_concat_left;
                eauto 3 using fresh_telescope_weakens,
                ctx_weakens_concat_right.
              apply fresh_telescope_ok, ctx_ok_middle_inv in HftG. tauto. }
            eauto using ctx_binds_functional. }

      assert (Hoksubst : ctx_ok ((G11 § y / v_loc ~ Ty)
                                   §& G12 §& subst_ctx w y G2)).
      { apply fresh_telescope_ok, ctx_ok_middle_inv in HftG.
        clear HokG. destruct HftG as [HokG [Hwnin1 Hwnin2]].
        rewrite <- ctx_concat_assoc in HokG.
        apply ctx_ok_concat_inv in HokG.
        destruct HokG as [Hok11 [Hok2 HD1]].
        apply ctx_ok_concat_inv in Hok2.
        destruct Hok2 as [Hok12 [Hok2 HD2]].
        apply ctx_ok_concat. apply ctx_ok_concat; eauto.
        - simpl. hnf. hnf in HD1. simpl in HD1.
          rewrite ctx_dom_concat, inter_union_r in HD1.
          apply fset_extens; eauto using subset_empty_l.
          rewrite <- HD1. auto using subset_union_weak_l.
        - rewrite <- ctx_concat_empty_l.
          apply ctx_ok_concat_map.
          rewrite ctx_concat_empty_l. auto.
        - rewrite ctx_dom_map. rewrite ctx_dom_concat in HD1.
          hnf. hnf in HD1. rewrite inter_union_r in HD1.
          apply fset_extens; auto using subset_empty_l.
          rewrite ctx_dom_concat. rewrite inter_union_l.
          hnf in HD2. rewrite HD2, union_empty_r.
          rewrite <- HD1. auto using subset_union_weak_r. }

      eapply typ_corr_subtyp with (e:=e) in Htycor; eauto.
      eapply ty_var_sub; eauto.
      eapply ty_var_var. rewrite <- ctx_concat_assoc in Hoksubst.
      rewrite <- ctx_concat_assoc, <- ctx_concat_empty_r.
      apply ctx_binds_weaken; simpl; eauto.

    + (* Have G₁₁(y:Ty)G₁₂(w:Tw)G₂ ⊢ t : T so G₁₁(y:Ty)G₁₂(w:Tw)G₂ must
       be well-formed. And so w can't be free in Tw again by
       freshness condition of inert contexts. *)
      apply fresh_telescope_concat_invert_l in HftG. inversions HftG; auto.

  - (* Because renv_corr and ctx_binds_k_after, the segment between y
       and w that is G₁₂ must consist only of pushes. *)
    intros _ Hvis. apply ctx_vis_objctx_concat. rewrite <- ctx_concat_assoc.
    eauto using ctx_vis_weakens.
Qed.

Lemma ref_elim_typ_trm G Go G' e e' w y (t : trm) T :
  fresh_telescope G ->
  ctx_ok (G §§& Go) -> ok e ->
  renv_corr G e ->
  G §§ Go ⁏ e ⊢ t ∶ T ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  env_elim e w y e' ->
  G' §§ (subst_objctx w y Go) ⁏ e' ⊢ (t /[ w -> y]) ∶ (T /[ w -> y]).
Proof.
  introv HftG HokG Hoke Hcorr Hty HelimG Hbie Helime.
  pose proof (env_elim_permute Hoke Helime) as Hp.
  destruct Hp as [e0 [Hbm0 [Hok [Helime0 [Hbm0' Hbm'0]]]]].
  eapply ebinm_ty_trm in Hty; eauto.
  eapply ref_elim_typ_end with (G':=G') in Hty; auto.
  eapply ebinm_ty_trm in Hty; eauto.
  eapply renv_corr_binds_more; try eassumption;
    eauto using fresh_telescope_weakens, fresh_telescope_ok.
  introv Hbi0. apply binds_push_inv in Hbi0.
  destruct Hbi0 as [[? ?] | [Hneq Hbi0]]; subst; eauto.
  specialize (Hbm0' _ _ Hbi0).
  eauto using env_elim_binds_more.
Qed.

Lemma ref_elim_typ_end_hide G1 G2 Go G1' e w y (t : trm) T :
  fresh_telescope G1 -> ctx_weakens G2 ->
  ctx_ok (G1 §& G2 §§& Go) -> ok (e & w ~ y) ->
  renv_corr G1 (e & w ~ y) ->
  (ctx_hide G1) §& G2 §§ Go ⁏ (e & w ~ y) ⊢ t ∶ T ->
  ref_elim_ctx G1 w y G1' ->
  (ctx_hide G1') §& (subst_ctx w y G2) §§ (subst_objctx w y Go) ⁏ e
                 ⊢ (t /[ w -> y]) ∶ (T /[ w -> y]).
Proof.
  introv HftG HwkG2 HokG Hoke Hcorr Hty Helim.
  apply ref_elim_ctx_renv_corr with (e := e & w ~ y) in Helim;
    eauto using ctx_ok_obj_inv_l.
  destruct Helim as [G11 [G12 [G13 [Ty [Tw [HeqG [Hyfr HeqG']]]]]]]. subst.
  apply ok_push_inv in Hoke. destruct Hoke as [Hoke Hfrwe].

  assert (Hwk : ctx_weakens (G12 §& subst_ctx w y G13)).
  { apply fresh_telescope_concat_invert_r in HftG as HftG3.
    apply fresh_telescope_concat_invert_l in HftG as HftG12.
    inversions HftG12. apply fresh_telescope_concat_invert_r in H3.
    apply ctx_weakens_concat;
      eauto using fresh_telescope_weakens, ctx_weakens_map. }

  replace (ctx_hide (((G11 § y / v_loc ~ Ty) §& G12 § w / v_ref ~ Tw) §& G13))
    with (((G11 § y / v_loc ~ Ty) §& G12) § w / v_ref ~ Tw §& ctx_hide G13)
    in Hty by eauto.
  rewrite <- ctx_concat_assoc in Hty.
  replace (ctx_hide ((G11 § y / v_loc ~ Ty) §& G12 §& subst_ctx w y G13))
    with (((G11 § y / v_loc ~ Ty) §& G12) §& subst_ctx w y (ctx_hide G13))
    by eauto.
  rewrite <- ctx_concat_assoc, <- ctx_map_concat.
  eapply subst_ty_trm; eauto.

  - eapply ctx_ok_obj_disjoint_converse.
    + rewrite ctx_concat_assoc. simpl.
      apply ctx_ok_obj_inv_l, ctx_ok_concat_inv in HokG.
      destruct_all; eauto using ctx_ok_concat.
    + eapply ctx_ok_obj_disjoint in HokG.
      repeat(rewrite ctx_dom_concat in *). rewrite ctx_dom_hide.
             rewrite union_assoc. eauto.

  - (* Have G₁₁(y:Ty)G₁₂(w:Tw)G₂ ⊢ t : T so G₁₁(y:Ty)G₁₂(w:Tw)G₂ must *)
    (* be well-formed. And so w can't be free in the initial telescope *)
    (* before it thanks to the freshness condition of inert contexts. *)
    apply fresh_telescope_concat_invert_l in HftG. inversions HftG; auto.

  - apply renv_notin_dom_fv; auto.
    hnf. introv Hwin. apply (fv_env_to_binds Hoke) in Hwin.
    destruct Hwin as [w0 Hbie].
    eapply binds_push_neq with (x2 := w) (v2 := y) in Hbie.
    2: { hnf. intros; subst. eauto using binds_to_dom. }
    apply ctx_ok_obj_inv_l in HokG.
    pose proof (renv_corr_binds_loc
                  (fresh_telescope_ok HftG) Hcorr Hbie) as Hbkloc.
    assert (ctx_k (((G11 § y / v_loc ~ Ty) §& G12 § w / v_ref ~ Tw) §& G13) w
                  v_ref) as Hbkref.
    { hnf. exists Tw ((G11 § y / v_loc ~ Ty) §& G12).
      eauto using fresh_telescope_weakens, ctx_weakens_concat_right,
        ctx_binds_k_after_concat, fresh_telescope_ok. }
    pose proof (ctx_k_eq Hbkloc Hbkref) as Hcontra. congruence.

  - rewrite subst_fresh.

    + (* Ty is a subtype of Tw by renv_corr *)
      assert (Htycor : typ_corr Tw Ty).
      { inversions Hcorr.
        - false. eauto using empty_push_inv.
        - apply eq_push_inv in H. destruct H as [? [? ?]]; subst.
          replace Ty with Ty0. replace Tw with Tw0. auto.
          + apply ctx_binds_k_after_binds in H0.
            assert (ctx_binds w Tw
                              (((G11 § y / v_loc ~ Ty)
                                  §& G12 § w / v_ref ~ Tw) §& G13)).
            { eapply ctx_binds_concat_left;
                eauto 3 using fresh_telescope_weakens,
                ctx_weakens_concat_right.
              apply fresh_telescope_ok, ctx_ok_middle_inv in HftG. tauto. }
            eauto using ctx_binds_functional.
          + pose proof (ctx_binds_k_after_trans H1 H0
                          (fresh_telescope_ok HftG))
              as Hbiy.
            apply ctx_binds_k_after_binds in Hbiy.
            assert (ctx_binds y Ty
                              (((G11 § y / v_loc ~ Ty)
                                  §& G12 § w / v_ref ~ Tw) §& G13)).
            { rewrite <- ctx_concat_push, <- ctx_concat_assoc in HftG.
              rewrite <- ctx_concat_push, <- ctx_concat_assoc.
              eapply ctx_binds_concat_left;
                eauto 3 using fresh_telescope_weakens,
                ctx_weakens_concat_right.
              apply fresh_telescope_ok, ctx_ok_middle_inv in HftG. tauto. }
            eauto using ctx_binds_functional. }

      assert (Hoksubst : ctx_ok
                           ((G11 § y / v_loc ~ Ty)
                              §& G12 §& subst_ctx w y (ctx_hide G13 §& G2))).
      { apply ctx_ok_obj_inv_l in HokG.
        rewrite <- ctx_concat_assoc in HokG.
        apply ctx_ok_middle_inv in HokG.
        destruct HokG as [HokG [Hwnin1 Hwnin2]].
        rewrite <- ctx_concat_assoc in HokG.
        apply ctx_ok_concat_inv in HokG.
        destruct HokG as [Hok11 [Hok2 HD1]].
        apply ctx_ok_concat_inv in Hok2.
        destruct Hok2 as [Hok12 [Hok2 HD2]].
        apply ctx_ok_concat. apply ctx_ok_concat; eauto.
        - simpl. hnf. hnf in HD1. simpl in HD1.
          rewrite ctx_dom_concat, inter_union_r in HD1.
          apply fset_extens; eauto using subset_empty_l.
          rewrite <- HD1. auto using subset_union_weak_l.
        - rewrite <- ctx_concat_empty_l. apply ctx_ok_concat_map.
          rewrite ctx_concat_empty_l.
          apply ctx_ok_concat_inv in Hok2. destruct_all.
          apply ctx_ok_concat; eauto.
        - rewrite ctx_dom_map. rewrite ctx_dom_concat in HD1.
          hnf. hnf in HD1. rewrite inter_union_r in HD1.
          apply fset_extens; auto using subset_empty_l.
          rewrite ctx_dom_concat. rewrite inter_union_l.
          hnf in HD2. rewrite ctx_dom_concat, ctx_dom_hide.
          rewrite ctx_dom_concat in *.
          rewrite HD2, union_empty_r.
          rewrite <- HD1. auto using subset_union_weak_r. }

      eapply typ_corr_subtyp with
        (e:=e)
        (G := (G11 § y / v_loc ~ Ty)
                §& G12 §& subst_ctx w y (ctx_hide G13 §& G2))
        in Htycor; eauto.
      eapply ty_var_sub; eauto.
      eapply ty_var_var. rewrite <- ctx_concat_assoc in Hoksubst.
      rewrite ctx_map_concat, ctx_concat_assoc, <- ctx_concat_empty_r.
      apply ctx_binds_weaken; eauto using ctx_weakens_map.
      * rewrite ctx_map_hide. simpl. apply ctx_binds_hide.
        rewrite <- ctx_concat_assoc. eapply ctx_binds_concat_left; eauto.
        eapply ctx_ok_middle_inv in Hoksubst. destruct_all.
        rewrite ctx_dom_concat, ctx_dom_map in *.
        rewrite ctx_dom_concat in *. repeat(rewrite notin_union in *). tauto.
      * rewrite <- ctx_concat_assoc, <- ctx_map_concat, <- ctx_concat_assoc.
        auto.

    + (* Have G₁₁(y:Ty)G₁₂(w:Tw)G₂ ⊢ t : T so G₁₁(y:Ty)G₁₂(w:Tw)G₂ must *)
      (* be well-formed. And so w can't be free in Tw again by *)
      (* freshness condition of inert contexts. *)
      apply fresh_telescope_concat_invert_l in HftG. inversions HftG; auto.

  - intros _ Hvis. false. apply ctx_vis_objctx_inv in Hvis.
    destruct Hvis as [Hvis | Hwin].
    + rewrite ctx_concat_assoc in Hvis.
      apply ctx_vis_concat_inv in Hvis. destruct Hvis as [Hvis | Hvis].
      * simpl in *. eauto using ctx_vis_hide_inv.
      * eapply ctx_ok_obj_inv_l, ctx_ok_concat_inv in HokG.
        destruct HokG as [_ [_ Hdis]]. apply ctx_vis_in_dom in Hvis.
        hnf in Hdis. rewrite ctx_dom_concat, ctx_dom_push in Hdis.
        erewrite <- in_empty, <- Hdis. rewrite in_inter.
        repeat(rewrite in_union). rewrite in_singleton. tauto.
    + apply ctx_ok_obj_disjoint in HokG. hnf in HokG.
      erewrite <- in_empty with (x:=w), <- HokG, in_inter.
      repeat(rewrite ctx_dom_concat).
      rewrite ctx_dom_push. repeat(rewrite in_union). rewrite in_singleton.
      tauto.
Qed.

Lemma ref_elim_typ_trm_met_heap G G' e e' w y (t : trm) T xz TZ xc TC x U :
  fresh_telescope G -> ok e ->
  renv_corr G e ->
  ctx_ok (ctx_met_refined_heap (ctx_hide G) xz TZ xc TC x U) ->
  ctx_met_refined_heap (ctx_hide G) xz TZ xc TC x U ⁏ e ⊢ t ∶ T ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  env_elim e w y e' ->
  ctx_met_refined_heap
    (ctx_hide G')
    xz (TZ /[ w -> y]) xc (TC /[ w -> y])
    (x /[ w -> y]) (U /[ w -> y]) ⁏ e'
    ⊢ (t /[ w -> y]) ∶ (T /[ w -> y]).
Proof.
  introv HftG Hoke Hcorr Hokmet Hty HeG Hbi Helime.
  pose proof (env_elim_permute Hoke Helime) as Hp.
  destruct Hp as [e0 [Hbm0 [Hok [Helime0 [Hbm0' Hbm'0]]]]].
  eapply ebinm_ty_trm in Hty; eauto.
  rewrite ctx_met_refined_heap_concat in *.
  assert (HokGmet : ctx_ok (ctx_met_refined_heap G xz TZ xc TC x U)).
  { apply ctx_ok_concat_inv in Hokmet.
    destruct Hokmet as [HokG [Hokmet Hdis]].
    rewrite ctx_met_refined_heap_concat. rewrite ctx_dom_hide in *.
    apply ctx_ok_concat; eauto. }
  eapply ref_elim_typ_end_hide with (G1':=G') in Hty;
    try solve[simpl in *; auto].
  - assert (Hwin : w \in ctx_dom G).
    { destruct (renv_corr_binds_vk (fresh_telescope_ok HftG) Hcorr Hbi)
      as [Hwin _]. eauto using ctx_k_in_dom. }
    apply ctx_ok_concat_inv in Hokmet.
    destruct Hokmet as [HokG [Hokmet Hdis]].
    apply ctx_ok_hide_inv in HokG. rewrite ctx_dom_hide in *.
    repeat(rewrite ctx_dom_push in *). hnf in Hdis.
    assert (Hneqzw : xz <> w).
    { hnf. intros; subst.
      erewrite <- in_empty, <- Hdis, in_inter. repeat(rewrite in_union).
      repeat(rewrite in_singleton). eauto. }
    assert (Hneqzc : xc <> w).
    { hnf. intros; subst. hnf in Hdis.
      erewrite <- in_empty, <- Hdis, in_inter. repeat(rewrite in_union).
      repeat(rewrite in_singleton). eauto. }
    assert (subst_ctx w y (ctx_met_refined_heap ctx_empty xz TZ xc TC x U)
            = ctx_met_refined_heap
                ctx_empty
                xz (TZ /[ w -> y])
                xc (TC /[ w -> y])
                (x /[ w -> y]) (U /[ w -> y])) as Heqsubst.
    { pose proof (subst_ctx_met_refined_heap0) as Hsubst.
      specialize (Hsubst ctx_empty w y ctx_empty xz TZ xc TC x U
                         Hneqzw Hneqzc). eauto. }
    rewrite <- Heqsubst; simpl in *; clear Heqsubst.
    eapply ebinm_ty_trm in Hty; eauto.
  - eapply renv_corr_binds_more;
      eauto using fresh_telescope_weakens, fresh_telescope_ok.
    introv Hbi0. apply binds_push_inv in Hbi0.
    destruct Hbi0 as [[? ?] | [Hneq Hbi0]]; subst; eauto.
    specialize (Hbm0' _ _ Hbi0).
    eauto using env_elim_binds_more.
Qed.

Lemma ref_elim_tyq_typ w y e e' T U :
  ok e -> w \notin fv_env e -> w <> y ->
  env_elim e w y e' ->
  e  ⊢ T ≈ U ->
  e' ⊢ (T /[ w -> y]) ≈ (U /[ w -> y]).
Proof.
  introv Hoke Hwnin Hneq Helim Htyq.
  hnf in Helim. destruct Helim as [e1 [e2 [? ?]]]; subst.
  eapply ebinm_tyq_typ with (e' := e1 & e2 & w ~ y) in Htyq; eauto.
  - eapply subst_tyq_typ
      with (S:=typ_bot) (G1:=ctx_empty) (vk:=v_ref) (Go:=objctx_empty)
           (G2 := ctx_hide (y / v_loc ~ typ_bot));
      simpl in *; eauto.
    + apply ok_middle_inv_l in Hoke as Hoke1.
      apply ok_middle_inv_r in Hoke as Hoke2.
      hnf. rewrite env_concat_fv, dom_concat.
      hnf in Hwnin. repeat(rewrite fv_env_concat in Hwnin).
      repeat (rewrite in_union in *). introv Hcontra.
      destruct_all; hnf in *; try tauto.
    + intros _ Hcontra. false. eauto using ctx_vis_hide_inv.
  - introv Hbi. apply binds_middle_inv in Hbi.
    destruct Hbi as [Hbi | [[Hfrw0 [? ?]] | [Hfrw0 [Hneq0 Hbi]]]];
      subst; eauto. eapply binds_push_neq; eauto using binds_concat_right.
    hnf. intros. subst. apply binds_to_dom in Hbi.
    apply ok_middle_inv_r in Hoke. eauto.
Qed.

Lemma ref_elim_ctx_vis G w y G' x :
  ctx_weakens G ->
  ref_elim_ctx G w y G' ->
  x <> w ->
  ctx_vis G  x ->
  ctx_vis G' x.
Proof.
  introv Hwk Helim Hneq Hvis.
  hnf in Helim. destruct Helim as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  apply ctx_vis_concat_inv in Hvis.
  destruct Hvis as [Hvis | Hvis];
    eauto using ctx_vis_concat_map, ctx_vis_weakens_left, ctx_vis_map.
  inversions Hvis; try congruence.
  eauto using ctx_weakens_concat_right, ctx_weakens_map, ctx_vis_weakens.
Qed.

Lemma ref_elim_typ_loc G G' e w y Ty :
  fresh_telescope G -> ok e ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  renv_corr G e ->
  ctx_binds y Ty G ->
  ctx_binds y Ty G'.
Proof.
  introv HftG Hoke Helim Hbie Hcor Hbiy.
  eapply ref_elim_typ_corr in Helim; eauto.
  destruct Helim as [_ [Ty' [_ [HbiyG' [HbiyG _]]]]].
  apply (ctx_binds_functional Hbiy) in HbiyG. subst. easy.
Qed.

Lemma ref_elim_precise_eq G G' e w y T TMBnd U :
  fresh_telescope G -> (* sealed G -> *)
  ctx_ok G -> ok e ->
  renv_corr G e ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  U <> typ_objref T (typ_mut_marker TMBnd) ->
  U <> typ_mut_marker TMBnd ->
  G  ⊢! w ∶ (typ_objref T (typ_mut_marker TMBnd)) ⪼ U ->
  G' ⊢! y ∶ (typ_objref T (typ_mut_marker typ_bot)) ⪼ (U /[ w -> y]).
Proof.
  introv Hft HokG Hoke Hcor Helim Hbi HneqU1 HneqU2 Hpf.
  apply pf_binds in Hpf as Hbiw.
  eapply fresh_telescope_binds_fv in Hbiw as Hfrw; eauto. simpl in Hfrw.
  rewrite union_empty_l, notin_union in Hfrw.

  eapply open_proj_pf_subst; simpl; eauto using pf_dec_typ_path; try easy.

  eapply ref_elim_typ_corr in Helim; eauto.
  destruct Helim as [Tw [Ty [Hbiw' [Hbiy [_ Htycor]]]]].
  apply (ctx_binds_functional Hbiw) in Hbiw'; subst.
  inversions Htycor; simpl in *. 
  destruct ( mode_refcap_corr' typing_mode) eqn: Hmc.
  typing_mode_contradiction typing_mode.
  destruct H2 as [Tbnd [? ?]]; subst.
  inversions H. easy.
Qed.

Lemma ref_elim_precise_eq_rcd G G' e w y T A TMBnd TU TL :
  fresh_telescope G -> (* sealed G -> *)
  ctx_ok G -> ok e ->
  renv_corr G e ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  G  ⊢! w ∶ (typ_objref T (typ_mut_marker TMBnd)) ⪼
    (typ_rcd (dec_typ A TL TU)) ->
  (typ_rcd (dec_typ A TL TU) = typ_mut_marker TMBnd)
    \/ (G' ⊢! y ∶ (typ_objref T (typ_mut_marker typ_bot)) ⪼
         (typ_rcd (dec_typ A TL TU)) /[ w -> y]).
Proof.
  introv Hft HokG Hoke Hcor Helim Hbi Hpf.
  remember (typ_rcd (dec_typ A TL TU)) as U.
  assert (U <> typ_objref T (typ_mut_marker TMBnd)) as HneqU by (subst; easy).
  destruct (classicT (U = typ_mut_marker TMBnd)); eauto.
  eapply ref_elim_precise_eq in Hpf; eauto.
Qed.

Lemma ref_elim_typ_var G Go G' e e' w y (x : var) T :
  fresh_telescope G ->
  ctx_ok (G §§& Go) -> ok e ->
  renv_corr G e ->
  G §§ Go ⁏ e ⊢ x ∶ T ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  env_elim e w y e' ->
  G' §§ (subst_objctx w y Go) ⁏ e' ⊢ (x /[ w -> y]) ∶ (T /[ w -> y]).
Proof.
  introv HinG HokG Hoke Hcorr Hty HelimG Hbie Helime.
  pose proof (env_elim_permute Hoke Helime) as Hp.
  destruct Hp as [e0 [Hbm0 [Hok [Helime0 [Hbm0' Hbm'0]]]]].
  eapply ebinm_ty_var in Hty; eauto.
  eapply ref_elim_typ_var_end with (G':=G') in Hty; auto.
  eapply ebinm_ty_var in Hty; eauto.
  eapply renv_corr_binds_more; try eassumption;
    eauto using fresh_telescope_weakens, fresh_telescope_ok.
  introv Hbi0. apply binds_push_inv in Hbi0.
  destruct Hbi0 as [[? ?] | [Hneq Hbi0]]; subst; eauto.
  specialize (Hbm0' _ _ Hbi0).
  eauto using env_elim_binds_more.
Qed.

Lemma ref_elim_subtyp G G' e e' w y S U :
  fresh_telescope G -> (* -> sealed G -> *) ok e ->
  renv_corr G e ->
  G ⁏ e ⊢ S <: U ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  env_elim e w y e' ->
  G' ⁏ e' ⊢ (S /[ w -> y]) <: (U /[ w -> y]).
Proof.
  introv HftG Hoke Hcorr Hty HelimG Hbie Helime.
  reverts G'.

  induction Hty; simpl; try eauto; try discriminate; intros.

  - eapply subtyp_typ.
    all: intros; pick_fresh z.
    all: specialize (H0 z); specialize (H2 z); autospecialize_prop.
    all: repeat(rewrite subst_open_commut in H0).
    all: repeat(rewrite subst_open_commut in H2).
    all: destruct (classicT (z = w)); subst;
      try solve[repeat(rewrite notin_union in Fr);
                repeat(rewrite notin_singleton in Fr); tauto].

    all: specialize (H0 _ HelimG); specialize (H2 _ HelimG).

    all: apply usubst_end1 with (x0:=z) (y0:=x) in H0;
      eauto 3 using ref_elim_ctx_ok, fresh_telescope_ok.
    all: apply usubst_end1 with (x0:=z) (y0:=x) in H2;
      eauto 3 using ref_elim_ctx_ok, fresh_telescope_ok.
    all: repeat(rewrite subst_open_commut in H0).
    all: repeat(rewrite subst_open_commut in H2).
    all: destruct (classicT (z = z)) as [_ | ?]; try easy.

    all: repeat(rewrite notin_union in Fr);
      repeat(rewrite notin_singleton in Fr).
    all: do 2 (rewrite subst_fresh with (x0:=z) (y0:=x) in H0);
      do 2 (rewrite subst_fresh with (x0:=z) (y0:=x) in H2);
      try solve[assumption].

    all: hnf; intros Hin.
    all: epose proof (fv_subst_var_subset _ w y Hin) as Hin';
      clear Hin; rename Hin' into Hin.
    all: rewrite in_union, in_singleton in Hin.
    all: destruct Hin; subst; try tauto.
    all: match goal with
         | [H : ?z \in fv ?T |- False] => assert (z \notin fv T) by tauto; easy
         end.

  - rename y0 into z. rewrite subst_open_commut.
    destruct (classicT (z = w)); subst.
    all: destruct (classicT (x = w)); subst.

    all: eapply subtyp_sel2.
    all: eapply ref_elim_typ_var in H; eauto.
    all: simpl in H.
    all: try (rewrite subst_eq_var in H); try (rewrite subst_ne_var in H);
      eauto 3 using fresh_telescope_ok.

  - rename y0 into z. rewrite subst_open_commut.
    destruct (classicT (z = w)); subst.
    all: destruct (classicT (x = w)); subst.

    all: eapply subtyp_sel1.
    all: eapply ref_elim_typ_var in H; eauto.
    all: simpl in H.
    all: try (rewrite subst_eq_var in H); try (rewrite subst_ne_var in H);
      eauto 3 using fresh_telescope_ok.

  - info_typing_mode_subst.

  - autospecialize_prop. eapply subtyp_met; eauto 3.

    + clear H H0 H2 H3 H4. rename H1 into IH.
      intro_fresh x Hfrx. repeat(rewrite union_empty_l in Hfrx).
      repeat(rewrite union_empty_r in Hfrx).
      assert (HfrxL : x \notin L) by eauto; specialize (IH _ HfrxL);
        clear HfrxL.
      assert (HftGx : fresh_telescope (G § x / v_par ~ TZ2)) by
        (constructor; eauto); specialize (IH HftGx); clear HftGx.
      specialize (IH Hoke).
      assert (HfrxG : x §# G) by eauto;
        assert (HcorrGx : renv_corr (G § x / v_par ~ TZ2) e) by
        eauto using renv_corr_push;
        specialize (IH HcorrGx); clear HfrxG HcorrGx.
      autospecialize_prop.

      repeat(rewrite subst_open_commut in IH).
      replace (x \[ w -> y]) with x in IH
          by (destruct (classicT (x = w)); eauto; subst; notin_false).
      eapply IH; clear IH.

      hnf. hnf in HelimG.
      destruct HelimG as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
      rewrite <- ctx_concat_push. repeat(eexists).

    + clear H H0 H1 H3 H4. rename H2 into IH.
      intros Hm Hindep. simpl in IH. autospecialize_prop.
      rewrite typ_indep_b_subst in Hindep. autospecialize_prop.
      rewrite <- typ_indep_b_subst in IH. eauto.

    + repeat(rewrite union_empty_l). repeat(rewrite union_empty_r).
      clear H H0 H1 H2 H3. rename H4 into IH.

      introv Hfrx Hneq Hfry0.
      assert (HfrxL : x \notin L) by eauto;
        assert (Hfry0L : y0 \notin L) by eauto;
        specialize (IH _ HfrxL _ Hneq Hfry0L);
        clear HfrxL Hfry0L.

      assert (HftGm : fresh_telescope (ctx_met G x TZ2 y0 TC2)).
      { constructor; eauto; simpl in *; notin_solve.
        eapply notin_fv_open_bnd; eauto. simpl.
        replace (fv_typ TC2) with (fv TC2) by eauto. notin_solve. }
      assert (HcorrGm : renv_corr (ctx_met G x TZ2 y0 TC2) e)
        by (eauto using renv_corr_push).
      autospecialize_prop.

      repeat(rewrite subst_open_commut in IH).
      replace (x \[ w -> y]) with x in IH
          by (destruct (classicT (x = w)); eauto; subst; notin_false).
      replace (y0 \[ w -> y]) with y0 in IH
          by (destruct (classicT (y0 = w)); eauto; subst; notin_false).
      eapply IH; clear IH.

      hnf. hnf in HelimG.
      destruct HelimG as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
      repeat(rewrite <- ctx_concat_push). repeat(eexists).

      simpl. rewrite subst_open_commut.
      replace (x \[ w -> y]) with x
        by (destruct (classicT (x = w)); eauto; subst; notin_false). easy.

  - eapply renv_corr_binds_to_ref_loc in Hcorr as Hvk;
      eauto using fresh_telescope_ok.
    destruct Hvk as [_ [_ [_ [Hvkw [_ [Hvky _]]]]]].

    eapply ref_elim_tyq_typ in H; eauto.

    + hnf. introv Hin. apply fv_env_to_binds in Hin; eauto.
      destruct Hin as [w0 Hbiw].
      eapply renv_corr_binds_to_ref_loc in Hbiw as Hvk;
        try eassumption; eauto using fresh_telescope_ok.
      destruct Hvk as [_ [_ [_ [_ [_ [Hvkw' _]]]]]].
      apply ctx_k_eq with (vk1 := v_ref) in Hvkw'; eauto. discriminate.

    + hnf; intros; subst. apply ctx_k_eq with (vk1 := v_loc) in Hvkw; eauto.
      discriminate.

Qed.

Lemma ref_elim_typ_def G G' e e' w y (d : def) T x U :
  fresh_telescope (G §§& (objctx_heap x U)) -> ok e ->
  renv_corr G e ->
  G §§ (objctx_heap x U) ⁏ e ⊢ d ∶d T ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  env_elim e w y e' ->
  G' §§ (subst_objctx w y (objctx_heap x U)) ⁏ e'
     ⊢ (d /[ w -> y]) ∶d (T /[ w -> y]).
Proof.
  introv HftG Hoke Hcorr Hty HelimG Hbie Helime.
  assert (Hwyneq : w <> y).
  { hnf. intros; subst.
    eapply renv_corr_binds_vk with (G:=G) in Hbie as [Hck1 Hck2];
      eauto using fresh_telescope_ok.
    pose proof (ctx_k_eq Hck1 Hck2). discriminate. }
  inversions Hty.
  - (* ty_def_typ *)
    simpl. eapply ty_def_typ with (L := L \u \{w}). introv Hfrx.
    replace (subst_typ w y U0) with (U0 /[w->y]) by eauto.
    rewrite <- ! subst_open_commut_ne; eauto.
    eapply ref_elim_tyq_typ; eauto.
    hnf. introv Hwin. apply fv_env_to_binds in Hwin; eauto.
    destruct Hwin as [w0 Hbie0].
    eapply renv_corr_binds_vk with (G:=G) in Hbie0 as [_ Hbie0];
      eauto using fresh_telescope_ok.
    eapply renv_corr_binds_vk with (G:=G) in Hbie as [Hbie _];
      eauto using fresh_telescope_ok.
    pose proof (ctx_k_eq Hbie Hbie0). discriminate.
  - (* ty_def_fld *)
    rename H into Hvis, H0 into Htyp, t into v. simpl. eapply ty_def_fld.
    + introv Hm. specialize (Hvis Hm).
      destruct (classicT (v = w)) as [Heq | Hneq]; subst; simpl in *;
        eauto using ref_elim_ctx_vis.
      (* If v = w then the goal becomes y being visible in
         G' §§& subst_objctx w y Go and we have Hvis : ctx_vis (G §§& Go) y.
         Since HelimG : ref_elim_ctx G w y G' at most w becomes invisible,
         so the result follows.
         Otherwise the goal is to show x remains visible. Once again
         result follows from HelimG : ref_elim_ctx G w y G'. *)
      all: eapply ref_elim_ctx_vis with (G:=G) (w:=w) (y:=y);
        eauto using fresh_telescope_weakens, fresh_telescope_ok.
      eapply renv_corr_binds_vk with (G:=G) in Hbie as [_ Hbi];
        eauto using fresh_telescope_weakens, fresh_telescope_ok.
      eapply ctx_k_in_dom, ctx_dom_to_binds in Hbi;
        eauto using fresh_telescope_weakens, fresh_telescope_ok.
      destruct Hbi as [T Hbi]. eapply fresh_telescope_binds_vis; eauto.
    + (* Need ref_elim_typ_trm for variables *)
      pose proof (ref_elim_typ_var
                    (w:=w) (y:=y) (e:=e) (e':=e') (T:=T0)
                    (G' := G' §§& objctx_heap (x /[ w -> y]) (U /[ w -> y]))
                    (Go:=objctx_empty) (x:=v) HftG) as Htypelim.
      simpl in *. apply Htypelim; eauto using fresh_telescope_ok.
  - (* ty_def_met *)
    rename H into Hty. simpl.
    eapply ty_def_met with (L := ctx_dom (ctx_hide G) \u L \u \{w}).
    introv Hfrxz Hneq Hfrxc.
    simpl in *.
    assert (HokG : ctx_ok (ctx_met_refined_heap (ctx_hide G)
                                                xz TZ xc TC x U)).
    { eapply ctx_met_refined_heap_ok;
        eauto using fresh_telescope_weakens, fresh_telescope_ok. }
    assert (Htymet : ctx_met_refined_heap
                       (ctx_hide G) xz TZ xc TC x U ⁏ e
                       ⊢ open_met xz xc t ∶ open_met xz xc TR)
      by (info_typing_mode_subst).
    pose proof (ref_elim_typ_trm_met_heap
                  TC x U
                  HftG Hoke Hcorr HokG Htymet HelimG Hbie Helime)
      as Htyelim.
    assert (Heqmet : open_met xz xc (subst_trm w y t) =
                       ((open_met xz xc t) /[ w -> y])).
    { rewrite ! subst_open_commut_ne; eauto. }
    assert (Heqtyp : open_met xz xc (subst_typ w y TR) =
                       ((open_met xz xc TR) /[ w -> y])).
    { rewrite ! subst_open_commut_ne; eauto. }
    rewrite Heqmet, Heqtyp.
    replace (subst_typ w y TZ) with (TZ /[ w -> y]) by auto.
    replace (subst_typ w y TC) with (TC /[ w -> y]) by auto.
    info_typing_mode_subst.
Qed.

Lemma ref_elim_typ_defs G G' e e' w y (ds : defs) T x U :
  fresh_telescope (G §§& (objctx_heap x U)) -> ok e ->
  renv_corr G e ->
  G §§ (objctx_heap x U) ⁏ e ⊢ ds ∶ T ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  env_elim e w y e' ->
  G' §§ (subst_objctx w y (objctx_heap x U)) ⁏ e'
     ⊢ (ds /[ w -> y]) ∶ (T /[ w -> y]).
Proof.
  introv HftG Hoke Hcorr Hty HelimG Hbie Helime.
  remember (objctx_heap x U) as Go. induction Hty; subst.
  - simpl. apply ty_defs_one. eapply ref_elim_typ_def; eauto.
  - simpl. apply ty_defs_cons.
    + eapply IHHty; eauto.
    + eapply ref_elim_typ_def; eauto.
    + eapply subst_defs_hasnt. erewrite <- subst_label_of_def. auto.
Qed.

Lemma ref_elim_ctx_k_neq G G' w y x vk :
  ctx_ok G -> ctx_weakens G ->
  ref_elim_ctx G w y G' ->
  ctx_k G x vk ->
  x <> w ->
  ctx_k G' x vk.
Proof.
  introv HokG HwkG Helim Hctxk Hneq.
  apply ref_elim_ctx_ok with (G:=G) (G':=G') (w:=w) (y:=y)
    in HokG as HokG'; eauto.
  hnf in Helim.
  destruct Helim as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  hnf in Hctxk. destruct Hctxk as [T [G0 Hctxk]].
  apply ctx_binds_k_after_concat_inv in Hctxk.
  destruct Hctxk as [Hctxk | [G02 [? Hctxk]]]; subst.
  - inversions Hctxk; try easy. hnf. exists T G0.
    apply ctx_binds_k_after_concat;
      eauto using ctx_weakens_map, ctx_weakens_concat_right.
  - hnf. exists (T /[w->y]) (G1 §& subst_ctx w y G02).
    eauto using ctx_binds_k_after_concat_map.
Qed.

Lemma ref_elim_def_varkinds G G' e d T w y v U :
  ctx_ok G -> ctx_weakens G ->
  ref_elim_ctx G w y G' ->
  renv_corr G e ->
  binds w y e ->
  G §§ (objctx_heap v U) ⁏ e ⊢ d ∶d T ->
  def_varkinds G d ->
  def_varkinds G' (d /[ w -> y]).
Proof.
  introv HokG HwkG Helim Hcor Hbye Hty Hvk.
  inversions Hty; simpl in *; eauto; inversions Hvk.
  - apply def_varkinds_fld. rename t into x.
    destruct (classicT (x = w)) as [? | Hneq]; subst.
    + destruct (renv_corr_binds_vk HokG Hcor Hbye) as [Hwvk _].
      false.
      apply ctx_k_ne_false with (vk1:=v_loc) (vk2:=v_ref) (v:=w) (G:=G);
        easy.
    + eapply ref_elim_ctx_k_neq with (G:=G) (w:=w) (y:=y); auto.
  - apply def_varkinds_met.
    rename H into Hty. hnf. introv Htf0.
    pick_fresh xz. rename Fr into Hfrxz.
    pick_fresh xc. rename Fr into Hfrxc.
    assert (Hneq : xc <> xz) by eauto.
    assert (FrxcL : xc \notin L) by eauto.
    assert (FrxzL : xz \notin L) by eauto.
    specialize (Hty _ FrxzL _ Hneq FrxcL).
    assert (mode_visibility' typing_mode = mode_visibility_hide_method) as
      Hmv by (info_typing_mode_subst).
    rewrite Hmv in Hty.
    pose proof (tfree_met_params Hmv _ _ _ _ Hty) as Htf. false.
    (* v₀ is tfree in t[w=y]. Then either:
       1. v₀ is tfree in t; or
       2. w is tfree in t and v₀ = y.
       In the first case, by tfree_open, v₀ is tfree in open_met xc xz t
       which means xc = v₀ or xz = v₀, but either cases give a contradiction.
       In the second case, w is tfree in t so once again w is tfree in
       open_met xc xz t so xc = w or xz = w, another contradiction. *)
    apply tfree_sbust_inv_trm in Htf0.
    assert (xc \in \{xc}) as Hixc by eauto using in_singleton_self.
    assert (xz \in \{xz}) as Hixz by eauto using in_singleton_self.
    destruct Htf0 as [Htf0 | [Htfw ?]]; subst.
    + specialize (Htf v0).
      eapply tfree_open_met in Htf0. apply Htf in Htf0.
      destruct Htf0 as [? | ?]; subst.
      * apply Hfrxc. repeat(rewrite in_union). tauto.
      * apply Hfrxz. repeat(rewrite in_union). tauto.
    + specialize (Htf w).
      eapply tfree_open_met in Htfw. apply Htf in Htfw.
      destruct Htfw as [? | ?]; subst.
      * apply Hfrxc. repeat(rewrite in_union). tauto.
      * apply Hfrxz. repeat(rewrite in_union). tauto.
Qed.

Lemma ref_elim_defs_varkinds G G' v U e ds T w y :
  ctx_ok G -> ctx_weakens G -> renv_corr G e -> binds w y e ->
  ref_elim_ctx G w y G' ->
  G §§ (objctx_heap v U) ⁏ e ⊢ ds ∶ T ->
  defs_varkinds G ds ->
  defs_varkinds G' (ds /[ w -> y]).
Proof.
  introv HokG HwkG Hcorr Hbi Helim Hty Hvk. reverts y w G' e. reverts T v U.
  induction Hvk; intros; eauto.
  autospecialize_prop. inversions Hty.
  all: simpl in *; constructor; eauto using ref_elim_def_varkinds.
Qed.

Lemma ref_elim_ctx_neq_binds G w y G' x T :
  fresh_telescope G ->
  ref_elim_ctx G w y G' ->
  ctx_binds x T G ->
  x <> w ->
  ctx_binds x (T /[w->y]) G'.
Proof.
  introv Hft Helim Hbi Hneq. hnf in Helim.
  destruct Helim as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  apply ctx_binds_middle_inv in Hbi.
  destruct Hbi as [Hbi2 | [[Hnin [? ?]] | [Hnin [_ Hbi1]]]]; subst;
    try congruence.
  + eauto using ctx_binds_map.
  + apply ctx_binds_concat_left;
      eauto using ctx_weakens_concat_right, ctx_weakens_map,
      fresh_telescope_weakens.
    apply fresh_telescope_concat_invert_l in Hft.
    inversions Hft.
    assert (Hwnin : w \notin fv T).
    { hnf. intros. apply ctx_binds_fv_subset in Hbi1. eauto. }
    rewrite subst_fresh; eauto.
Qed.

Lemma ref_elim_ctx_binds_not_more G w y x G' :
  ref_elim_ctx G w y G' ->
  (x \in ctx_dom G') ->
  x \in ctx_dom G.
Proof.
  introv [vkw [Tw [G1 [G2 [? ?]]]]] Hin; subst.
  rewrite ctx_dom_concat, ctx_dom_map, in_union in Hin.
  rewrite ctx_dom_concat; simpl; repeat(rewrite in_union). tauto.
Qed.

Lemma ref_elim_ty_item G w y e Sigma x e' G' :
  ok e -> renv_corr G e ->
  ctx_kdom v_loc G = dom Sigma ->
  fresh_telescope G ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  env_elim e w y e' ->
  ty_item_s _ G e Sigma x ->
  ty_item_s _ G' e' (subst_heap w y Sigma) x.
Proof.
  introv Hoke Hcorr HSloc Hft HeG Hbiwy Hee Htys.
  inversion Htys; subst.
  - simpl in H. info_typing_mode_subst.
    (* Problem: we need everything tfree in t[w=y] be a reference.
       But w can appear in tfree in t and then y appears tfree in t[w=y]
       and y is not a reference *)
  - assert (Hneq : x <> w).
      { hnf. intros. subst. apply binds_to_dom in H0. rewrite <- HSloc in H0.
        destruct (renv_corr_binds_vk (fresh_telescope_ok Hft) Hcorr Hbiwy)
          as [Hcontra _]. apply ctx_k_in_kdom in Hcontra.
        assert (Hvkneq : (v_loc <> v_ref)) by congruence.
        pose proof
             (ctx_kdom_disjoint
                (fresh_telescope_ok Hft) (fresh_telescope_weakens Hft)
                Hvkneq)
          as Hdis.
        hnf in Hdis. erewrite <- in_empty with (x := w), <- Hdis, in_inter.
        tauto. }
      eapply ty_item_obj_s
        with (T := (typ_and (typ_bnd U) S) /[ w -> y]) (ds := ds /[ w -> y])
             (U := U /[ w -> y]) (S := S /[ w -> y]); eauto.
    + eapply ref_elim_ctx_neq_binds; try eassumption.
    + exact (binds_map (subst_var w y) H0).
    + replace (objctx_heap x (U /[ w -> y]))
        with (subst_objctx w y (objctx_heap x U)); eauto.
      2: { simpl. rewrite subst_ne_var; eauto. }
      pose proof (subst_open_commut_typ U w y x 0) as HsU.
      rewrite subst_ne_var in HsU; eauto. rewrite <- HsU; clear HsU.
      eapply ref_elim_typ_defs; eauto.
    + apply fresh_telescope_ok in Hft as Hok.
      apply fresh_telescope_weakens in Hft as Hwk.
      eapply ref_elim_defs_varkinds; eauto.
Qed.

Lemma ref_elim_ctx_neq_kdom G G' w y vk :
  ctx_ok G -> ctx_weakens G ->
  ref_elim_ctx G w y G' ->
  ~ (ctx_k G w vk) ->
  ctx_kdom vk G = ctx_kdom vk G'.
Proof.
  introv Hok Hwk Helim Hnvk. hnf in Helim.
  destruct Helim as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  repeat(rewrite ctx_kdom_concat).
  assert (Hneqvk : vk <> vkw).
  { hnf. intros. subst. apply Hnvk. hnf. exists Tw G1.
    eapply ctx_binds_k_after_concat; eauto using ctx_weakens_concat_right. }
  rewrite ctx_kdom_push_n; eauto. rewrite ctx_kdom_map. auto.
Qed.

Lemma ref_elim_ctx_env_kdom_eq G G' e e' w y vk :
  ctx_ok G -> ok e -> ctx_weakens G ->
  ref_elim_ctx G w y G' ->
  env_elim e w y e' ->
  ctx_kdom vk G  = dom e ->
  ctx_kdom vk G' = dom e'.
Proof.
  introv HokG Hoke HwkG HeG Hee Hkdomeq.
  hnf in HeG. destruct HeG as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  hnf in Hee. destruct Hee as [e1 [e2 [? ?]]]; subst.
  rewrite ctx_kdom_concat, ctx_kdom_map, dom_concat in *.
  rewrite dom_push in *.
  assert (HwninG0 : w \notin ctx_dom G1 \u ctx_dom G2).
  { apply ctx_ok_concat_inv in HokG.
    destruct HokG as [Hok1 [Hok2 Hdis]].
    inversions Hok1. rewrite notin_union; split; eauto. hnf.
    introv Hin. erewrite <- in_empty, <- Hdis, in_inter.
    split; eauto. simpl. rewrite in_union, in_singleton; eauto. }
  assert (HwninG : forall k, w \notin ctx_kdom k G1 \u ctx_kdom k G2).
  { intros k. hnf. intros Hin. apply HwninG0.
    rewrite <- ctx_kdom_concat in Hin. rewrite <- ctx_dom_concat.
    eapply ctx_kdom_subset; eauto.
    - apply ctx_ok_concat_inv in HokG.
      destruct HokG as [Hok1 [Hok2 Hdis]].
      inversions Hok1. eapply ctx_ok_concat; eauto.
      hnf. eapply fset_extens; eauto using subset_empty_l.
      rewrite <- Hdis. rewrite ctx_dom_push, inter_union_l.
      eauto using subset_union_weak_r.
    - apply ctx_weakens_concat_left in HwkG as Hwk1. inversions Hwk1.
      apply ctx_weakens_concat_right in HwkG as Hwk2.
      eauto using ctx_weakens_concat. }
  assert (Hwnine : w \notin dom e1 \u dom e2).
  { apply ok_middle_inv in Hoke. rewrite notin_union. auto. }
  destruct (vk_dec vk vkw) as [? | Hneq]; subst.
  - rewrite ctx_kdom_push_e in *; eauto.
    repeat(rewrite <- union_assoc in *).
    eapply union_disjoint_inv; eauto; eapply disjoint_singleton; eauto.
  - false. rewrite ctx_kdom_push_n in *; eauto. specialize (HwninG vk).
    rewrite Hkdomeq in HwninG.
    repeat(rewrite notin_union in *). rewrite notin_singleton in *. tauto.
Qed.

Lemma renv_corr_push_subst G1 w y vkw Tw G2 e :
  fresh_telescope (G1 § w / vkw ~ Tw §& G2) ->
  ok e -> w # e -> w \notin fv_env e ->
  renv_corr (G1 § w / vkw ~ Tw §& G2) e ->
  renv_corr (G1 §& subst_ctx w y G2) e.
Proof.
  reverts w y vkw Tw. reverts G1 G2. reverts e.
  apply (env_ind (fun e => forall G1 G2 w y vkw Tw,
                      fresh_telescope ((G1 § w / vkw ~ Tw) §& G2) ->
                      ok e -> w # e -> w \notin fv_env e ->
                      renv_corr ((G1 § w / vkw ~ Tw) §& G2) e ->
                      renv_corr (G1 §& subst_ctx w y G2) e));
    try solve [intros; eauto].
  intros e w0 y0 IH. introv HftG Hoke Hwd Hwf Hcor.
  inversions Hcor.
  1: { false. eauto using empty_push_inv. }
  apply eq_push_inv in H. destruct_all; subst.
  rename H0 into Hbkw0, H1 into Hbky0, H2 into Htycor, H4 into Hcor.
  apply ctx_binds_k_after_initial_segment in Hbky0.
  destruct Hbky0 as [G2' ?]; subst.
  apply ctx_binds_k_after_initial_segment in Hbkw0.
  destruct Hbkw0 as [G5 Heq].
  apply eq_sym in Heq.
  apply ctx_weakens_concat_inv in Heq as Heq';
    eauto using ctx_weakens_concat_right.
  destruct Heq' as [[G' [Heq' ?]] | [G' [Heq' ?]]]; subst.
  - apply ctx_push_concat_inv in Heq' as Heq''.
    destruct Heq'' as [[? Heq''] | [G6 ?]]; subst; simpl in *.
    + inversions Heq''. false.
      rewrite dom_push, notin_union, notin_singleton in Hwd. tauto.
    + inversions Heq'. clear Heq. rename H0 into Heq.
      apply ctx_weakens_concat_inv in Heq as Heq'.
      2: { eapply fresh_telescope_weakens,
        ctx_weakens_concat_right, ctx_weakens_concat_left in HftG.
           inversions HftG; eauto. }
      destruct Heq' as [[G' [Heq' ?]] | [G' [Heq' ?]]]; subst.
      * apply ctx_push_concat_inv in Heq' as Heq''.
        destruct Heq'' as [[? Heq''] | [G6 ?]]; subst; simpl in *.
        { rewrite ctx_concat_empty_l in *. inversions Heq'.
          false. rewrite fv_env_push, notin_union in Hwf.
          unfold fv, VarFreeVar in Hwf. rewrite notin_singleton in *.
          tauto. }
        { inversions Heq'.
          eapply renv_corr_ref with
            (Tw:=Tw0/[w->y]) (Ty:=Ty/[w->y])
            (G2 := G1 §& subst_ctx w y ((G6 § y0 / v_loc ~ Ty) §& G2'))
            (G3 := G1 §& subst_ctx w y G6);
            eauto using typ_corr_subst.
          - repeat(rewrite subst_ctx_concat). simpl.
            repeat(rewrite subst_ctx_concat). simpl.
            rewrite ctx_concat_assoc.
            eapply ctx_binds_k_after_concat; simpl; eauto.
            + repeat(rewrite <- ctx_subst_push).
              rewrite <- ctx_concat_push.
              repeat(rewrite <- subst_ctx_concat).
              repeat(rewrite <- ctx_subst_push).
              rewrite <- ctx_concat_assoc.
              repeat(rewrite <- subst_ctx_concat).
              apply ctx_ok_concat_map.
              apply fresh_telescope_ok, ctx_ok_concat_inv in HftG.
              destruct HftG as [Hok1 [Hok6 Hd1]].
              inversions Hok1. clear H4. rename H1 into Hok1.
              apply ctx_ok_concat; eauto.
              hnf. apply fset_extens; eauto using subset_empty_l.
              introv Hin. rewrite <- Hd1.
              rewrite ctx_dom_push, inter_union_l, in_union. tauto.
            + eauto using
                fresh_telescope_concat_invert_r,
                ctx_weakens_map, fresh_telescope_weakens.
          - repeat(rewrite subst_ctx_concat). simpl.
            rewrite ctx_concat_assoc.
            eapply ctx_binds_k_after_concat; simpl; eauto.
            + rewrite <- ctx_concat_push.
              repeat(rewrite <- ctx_subst_push).
              rewrite <- ctx_concat_assoc.
              repeat(rewrite <- subst_ctx_concat).
              apply ctx_ok_concat_map.
              rewrite ctx_concat_assoc in HftG.
              apply fresh_telescope_concat_invert_l in HftG.
              rewrite ctx_concat_push in HftG.
              inversions HftG. clear H4 H5 H6. rename H3 into HftG.
              apply fresh_telescope_ok, ctx_ok_concat_inv in HftG.
              destruct HftG as [Hok1 [Hok6 Hd1]].
              inversions Hok1. clear H4. rename H1 into Hok1.
              apply ctx_ok_concat; eauto.
              hnf. apply fset_extens; eauto using subset_empty_l.
              introv Hin. rewrite <- Hd1.
              rewrite ctx_dom_push, inter_union_l, in_union. tauto.
            + apply fresh_telescope_concat_invert_r,
                fresh_telescope_concat_invert_l in HftG.
              inversions HftG. clear H4 H5 H6. rename H3 into HftG.
              eauto using
                    fresh_telescope_concat_invert_r, ctx_weakens_map,
                fresh_telescope_weakens.
          - eapply IH; eauto. rewrite fv_env_push, notin_union in Hwf.
            tauto. }
      * apply ctx_push_concat_inv in Heq' as Heq''.
        destruct Heq'' as [[? Heq''] | [G7 ?]]; subst; simpl in *.
        { rewrite ctx_concat_empty_l in *. inversions Heq'.
          false. rewrite fv_env_push, notin_union in Hwf.
          unfold fv, VarFreeVar in Hwf. rewrite notin_singleton in *.
          tauto. }
        { inversions Heq'.
          eapply renv_corr_ref with
            (Tw:=Tw0/[w->y]) (Ty:=Ty)
            (G2 := (G4 § y0 / v_loc ~ Ty) §& G7 §& subst_ctx w y G6)
            (G3 := G4).
          - repeat(rewrite subst_ctx_concat). simpl.
            rewrite ctx_concat_assoc.
            eapply ctx_binds_k_after_concat; simpl; eauto.
            + rewrite <- ctx_concat_push.
              repeat(rewrite <- ctx_subst_push).
              rewrite <- ctx_concat_assoc.
              repeat(rewrite <- subst_ctx_concat).
              apply ctx_ok_concat_map.
              apply fresh_telescope_ok, ctx_ok_concat_inv in HftG.
              destruct HftG as [Hok4 [Hok6 Hd4]].
              inversions Hok4. clear H4. rename H1 into Hok1.
              apply ctx_ok_concat; eauto.
              hnf. apply fset_extens; eauto using subset_empty_l.
              introv Hin. rewrite <- Hd4.
              rewrite ctx_dom_push, inter_union_l, in_union. tauto.
            + eauto using
                    fresh_telescope_concat_invert_r,
                ctx_weakens_map, fresh_telescope_weakens.
          - rewrite <- ctx_concat_assoc.
            eapply ctx_binds_k_after_concat; simpl; eauto.
            + rewrite ctx_concat_assoc in *. apply ctx_ok_concat_map.
              apply fresh_telescope_concat_invert_l in HftG.
              rewrite ctx_concat_push in HftG.
              inversions HftG. clear H4 H5 H6. rename H3 into HftG.
              apply fresh_telescope_ok, ctx_ok_concat_inv in HftG.
              destruct HftG as [Hok4 [Hok6 Hd]].
              inversions Hok4. clear H4. rename H1 into Hok4.
              apply ctx_ok_concat; eauto.
              hnf. apply fset_extens; eauto using subset_empty_l.
              introv Hin. rewrite <- Hd.
              rewrite ctx_dom_push, inter_union_l, in_union. tauto.
            + rewrite ctx_concat_assoc in *.
              apply fresh_telescope_concat_invert_l in HftG.
              rewrite ctx_concat_push in HftG.
              inversions HftG. clear H4 H5 H6. rename H3 into HftG.
              rewrite <- ctx_concat_push, <- ctx_concat_assoc in HftG.
              apply fresh_telescope_concat_invert_r in HftG.
              apply fresh_telescope_concat_invert_r in HftG as HftG6.
              apply fresh_telescope_concat_invert_l in HftG.
              inversions HftG. clear H4 H5 H6. rename H3 into HinG7.
              eauto using fresh_telescope_weakens, ctx_weakens_concat,
                ctx_weakens_map.
          - inversions Htycor; simpl in *.
            destruct ( mode_refcap_corr' typing_mode) eqn: Hmc.
            typing_mode_contradiction typing_mode. 2: typing_mode_contradiction typing_mode.
            destruct H as [Tbnd [? ?]]; subst. simpl in *.
            replace (subst_typ w y T) with (T /[ w -> y]) by eauto.
            rewrite subst_fresh.
            + constructor; simpl; rewrite Hmc; eauto.
            + rewrite ctx_concat_assoc in HftG.
              do 2 (apply fresh_telescope_concat_invert_l in HftG).
              inversions HftG. clear Heq H3 H4 H6.
              rename H5 into Hfrw. rewrite ctx_fv_concat in *.
              simpl in Hfrw. repeat(rewrite notin_union in Hfrw). tauto.
          - eapply IH; eauto. rewrite fv_env_push, notin_union in Hwf.
            tauto. }
  - apply ctx_push_concat_inv in Heq' as Heq''.
    destruct Heq'' as [[? Heq''] | [G5 ?]]; subst; simpl in *.
    + inversions Heq''. false.
      rewrite dom_push, notin_union, notin_singleton in Hwd. tauto.
    + inversions Heq'.
      eapply renv_corr_ref with
        (Tw:=Tw0) (Ty:=Ty)
        (G2 := (G4 § y0 / v_loc ~ Ty) §& G2')
        (G3 := G4); eauto.
      * rewrite <- ctx_concat_assoc.
        eapply ctx_binds_k_after_concat; simpl; eauto.
        { rewrite ctx_concat_assoc.
          apply fresh_telescope_ok, ctx_ok_concat_inv in HftG.
          destruct HftG as [HokG4 [HokG2 Hd]].
          inversions HokG4. clear H4. rename H1 into HokG4.
          eapply ctx_ok_concat_map, ctx_ok_concat; eauto.
          hnf. apply fset_extens; eauto using subset_empty_l.
          introv Hin. rewrite <- Hd.
          rewrite ctx_dom_push, inter_union_l, in_union. tauto. }
        { rewrite <- ctx_concat_push, <- ctx_concat_assoc in HftG.
          apply fresh_telescope_concat_invert_r in HftG.
          apply fresh_telescope_concat_invert_r in HftG as HftG2.
          apply fresh_telescope_concat_invert_l in HftG.
          inversions HftG. clear H4 H5 H6. rename H3 into HftG5.
          eauto using ctx_weakens_concat, ctx_weakens_map,
            fresh_telescope_weakens. }
      * eapply ctx_binds_k_after_concat; simpl; eauto.
        { apply fresh_telescope_concat_invert_l in HftG.
          inversions HftG. clear H4 H5 H6. rename H3 into HftG.
          apply fresh_telescope_concat_invert_l in HftG.
          inversions HftG. clear H4 H5 H6. rename H3 into HftG.
          eauto using fresh_telescope_ok. }
        { apply fresh_telescope_concat_invert_l in HftG.
          inversions HftG. clear H4 H5 H6. rename H3 into HftG.
          apply fresh_telescope_concat_invert_l in HftG.
          inversions HftG. clear H4 H5 H6. rename H3 into HftG.
          apply fresh_telescope_concat_invert_r in HftG.
          eauto using fresh_telescope_weakens. }
      * eapply IH; eauto. rewrite fv_env_push, notin_union in Hwf. tauto.
  - eauto using fresh_telescope_concat_invert_r, fresh_telescope_weakens.
Qed.

Lemma ref_elim_renv_corr G G' e e' w y :
  ok e -> fresh_telescope G -> w \notin fv_env e ->
  renv_corr G e ->
  ref_elim_ctx G w y G' ->
  env_elim e w y e' ->
  renv_corr G' e'.
Proof.
  introv Hoke HftG Hwnin Hcor HeG Hee.
  hnf in HeG. destruct HeG as [vkw [Tw [G1 [G2 [? ?]]]]]; subst.
  hnf in Hee. destruct Hee as [e1 [e2 [? ?]]]; subst.
  reverts w y vkw Tw. reverts G1 G2 e1.
  reverts e2.
  apply (env_ind (fun e2 => forall (G1 G2 : ctx) (e1 : env var) (w y : var),
                    ok (e1 & w ~ y & e2) ->
                    w \notin fv_env (e1 & w ~ y & e2) ->
                    forall (vkw : varkind) (Tw : typ),
                      fresh_telescope ((G1 § w / vkw ~ Tw) §& G2) ->
                      renv_corr ((G1 § w / vkw ~ Tw) §& G2)
                                (e1 & w ~ y & e2) ->
                      renv_corr (G1 §& subst_ctx w y G2) (e1 & e2))).
  - introv Hoke Hwnin HftG Hcor.
    rewrite concat_empty_r in *.
    apply renv_corr_epush_inv in Hcor; eauto using fresh_telescope_ok.
    apply ok_push_inv in Hoke; destruct_all.
    eapply renv_corr_push_subst; eauto.
    rewrite fv_env_push, notin_union in Hwnin. tauto.
  - intros e2 w0 y0 IH. introv Hoke Hwnin HftG Hcor.
    repeat(rewrite concat_assoc in *).
    apply ok_push_inv in Hoke. destruct Hoke as [Hoke Hnin].
    apply renv_corr_epush_inv in Hcor as Hcor0; eauto.
    inversions Hcor.
    1: { false. eauto using empty_push_inv. }
    rename H0 into Hbkw1, H1 into Hbky1, H2 into Htycor, H4 into Hecor,
      Ty into Ty0.
    rename H into Heqenv. apply eq_push_inv in Heqenv.
    destruct Heqenv as [? [? ?]]; subst.
    assert (Hneqwy0 : y0 <> w).
    { hnf. intros. subst. rewrite fv_env_push, notin_union in Hwnin.
      unfold fv, VarFreeVar in Hwnin. rewrite notin_singleton in Hwnin.
      tauto. }
    (* G₁(w:Tw)G₂ = G₃(w₀:Tw₀)...  and G₃ = G₄(y₀:Ty₀)...
       so G₁(w:Tw)G₂ = G₄(y₀:Ty₀)...(w₀:Tw₀)...
       Then either w can be either in between y₀,w₀ or at the start or
       at the end:
       1. w...y₀...w₀ ─ in which case G₂ = ...(y₀:Ty₀)...(w₀:Tw₀)...
          Then, G₁(G₂[w=y]) = G₁...(y₀:Ty₀[w=y])...(w₀:Tw₀[w=y])...
       2. y₀...w...w₀ ─ in which case G₁ = ...(y₀:Ty₀)... and
          G₂ = ...(w₀:Tw₀)....
          Then, G₁(G₂[w=y]) = ...(y₀:Ty₀)...(w₀:Tw₀[w=y])...
       3. y₀...w₀...w ─ in which case G₁ = ...(y₀:Ty₀)...(w₀:Tw₀)...
          Then, G₁(G₂[w=y]) = ...(y₀:Ty₀)...(w₀:Tw₀)...
     *)
    apply ctx_binds_k_after_initial_segment in Hbky1.
    destruct Hbky1 as [G3' Heq3]. subst.
    apply ctx_binds_k_after_concat_inv in Hbkw1.
    destruct Hbkw1 as [Hbk | [G21 [Heq Hbk]]].
    + (* G₁(w:Tw) = G₄(y₀:Ty₀)G₃'(w₀:Tw₀)G₅ *)
      apply ctx_binds_k_after_initial_segment in Hbk.
      destruct Hbk as [G5 Heq]. apply ctx_push_concat_inv in Heq as Heq'.
      destruct Heq' as [[? _] | [G5' ?]]; subst.
      * inversions Heq. false. repeat(rewrite dom_concat in *).
        repeat(rewrite dom_single in *). repeat(rewrite notin_union in *).
        repeat(rewrite notin_singleton in *). tauto.
      * simpl in Heq. inversions Heq.
        eapply renv_corr_ref
          with (Tw:=Tw0) (Ty:=Ty0)
               (G2 := (G4 § y0 / v_loc ~ Ty0) §& G3') (G3 := G4); auto.
        { rewrite <- ctx_concat_assoc.
          rewrite <- ctx_concat_push, <- ctx_concat_assoc in HftG.

          eapply ctx_binds_k_after_concat; eauto.
          - apply fresh_telescope_ok, ctx_ok_concat_inv in HftG.
            destruct HftG as [Hok4 [Hok5 Hdis1]].
            apply ctx_ok_concat_inv in Hok5.
            destruct Hok5 as [Hok5 [Hok2 Hdis2]].
            apply ctx_ok_concat; eauto.
            + inversions Hok5. apply ctx_ok_concat_map; eauto.
              apply ctx_ok_concat; eauto.
              apply fset_extens; eauto using subset_empty_l.
              hnf. introv Hin.
              rewrite <- Hdis2, ctx_dom_push, inter_union_l, in_union. tauto.
            + apply fset_extens; eauto using subset_empty_l.
              hnf. introv Hin.
              rewrite <- Hdis1, ctx_dom_push, inter_union_l, in_union.
              repeat(rewrite ctx_dom_concat in *).
              repeat(rewrite ctx_dom_push in *).
              rewrite ctx_dom_concat, ctx_dom_push, ctx_dom_map in Hin.
              repeat(rewrite in_inter, in_union in *).
              repeat(rewrite in_union in *). tauto.
          - apply fresh_telescope_concat_invert_r,
              fresh_telescope_weakens in HftG.
            apply ctx_weakens_concat_left in HftG as HftG5.
            apply ctx_weakens_concat_right in HftG as HftG2.
            inversions HftG5.
            eauto using ctx_weakens_concat, ctx_weakens_map. }
        { eapply ctx_binds_k_after_concat; eauto.
          - apply fresh_telescope_concat_invert_l in HftG.
            rewrite <- ctx_concat_push in HftG.
            apply fresh_telescope_concat_invert_l in HftG.
            inversions HftG. eauto using fresh_telescope_ok.
          - apply fresh_telescope_concat_invert_l in HftG.
            rewrite <- ctx_concat_push in HftG.
            apply fresh_telescope_concat_invert_l in HftG.
            rewrite <- ctx_concat_push in HftG.
            apply fresh_telescope_concat_invert_r in HftG.
            inversions HftG; eauto using fresh_telescope_weakens. }
        { (* Use Hecor *)
          eapply IH; eauto.
          rewrite fv_env_push, fv_env_concat, fv_env_push in Hwnin.
          rewrite fv_env_concat, fv_env_push.
          repeat(rewrite notin_union in *). tauto. }
    + apply ctx_binds_k_after_initial_segment in Hbk.
      (* G₂ = G₂₁(w₀:Tw₀)G₂₂ and G₄(y₀:Ty₀)G₃' = G₁(w:Tw)G₂₁ *)
      destruct Hbk as [G22 ?]; subst.
      apply ctx_weakens_concat_inv in Heq as Heq'.
      destruct Heq' as [[G5 [Heq' ?]] | [G5 [Heq' ?]]]; subst.
      * apply ctx_push_concat_inv in Heq' as Heq''.
        destruct Heq'' as [[? Heq''] | [G6 ?]]; subst.
        { inversions Heq''. simpl in *. tauto. }
        { simpl in *. inversions Heq'.
          eapply renv_corr_ref
            with (Tw:=Tw0/[w->y]) (Ty:=Ty0/[w->y])
                 (G2 := (G1 §& (subst_ctx w y
                                          ((G6 § y0 / v_loc ~ Ty0) §& G3'))))
                 (G3 := (G1 §& (subst_ctx w y G6)));
            eauto using typ_corr_subst.
          - repeat(rewrite subst_ctx_concat). simpl.
            repeat(rewrite subst_ctx_concat). simpl.
            rewrite ctx_concat_assoc.
            eapply ctx_binds_k_after_concat; simpl; eauto.
            + apply fresh_telescope_ok in HftG.
              apply ctx_ok_concat_map with (f := subst_var w y) in HftG.
              repeat(rewrite subst_ctx_concat in HftG). simpl in HftG.
              repeat(rewrite subst_ctx_concat in HftG). simpl in HftG.
              rewrite <- ctx_concat_push, <- ctx_concat_assoc.
              apply ctx_ok_concat_inv in HftG.
              destruct HftG as [Hok1 [Hok6 Hd1]].
              inversions Hok1. rename H4 into Hfrw, H1 into Hok1.
              apply ctx_ok_concat; eauto.
              apply fset_extens; eauto using subset_empty_l.
              introv Hin. rewrite <- Hd1. rewrite ctx_dom_push.
              rewrite inter_union_l, in_union. tauto.
            + eauto using fresh_telescope_weakens, ctx_weakens_map,
                fresh_telescope_concat_invert_r.
          - repeat(rewrite subst_ctx_concat). simpl.
            rewrite ctx_concat_assoc.
            eapply ctx_binds_k_after_concat; simpl; eauto.
            + rewrite <- ctx_concat_push, <- ctx_subst_push,
                <- ctx_concat_assoc, <- subst_ctx_concat.
              apply ctx_ok_concat_map.
              rewrite ctx_concat_assoc in HftG.
              apply fresh_telescope_concat_invert_l in HftG.
              rewrite ctx_concat_push in HftG. inversions HftG.
              clear H4 H5 H6. rename H3 into HftG.
              apply fresh_telescope_ok, ctx_ok_concat_inv in HftG.
              destruct HftG as [Hok1 [Hok6 Hd1]].
              inversions Hok1. clear H4. rename H1 into Hok1.
              apply ctx_ok_concat; eauto.
              apply fset_extens; eauto using subset_empty_l.
              introv Hin. rewrite <- Hd1. rewrite ctx_dom_push.
              rewrite inter_union_l, in_union. tauto.
            + apply ctx_weakens_map.
              apply fresh_telescope_concat_invert_r,
                fresh_telescope_concat_invert_l in HftG.
              inversions HftG. clear H4 H5 H6. rename H3 into HftG.
              apply fresh_telescope_concat_invert_r in HftG;
                eauto using fresh_telescope_weakens.
          - eapply IH; eauto.
            rewrite fv_env_push, fv_env_concat, fv_env_push in Hwnin.
            rewrite fv_env_concat, fv_env_push.
            repeat(rewrite notin_union in *). tauto. }
      * apply ctx_push_concat_inv in Heq' as Heq''.
        destruct Heq'' as [[? Heq''] | [G6 ?]]; subst.
        { inversions Heq''. simpl in *. tauto. }
        { simpl in *. inversions Heq'.
          eapply renv_corr_ref
            with (Tw:=Tw0/[w->y]) (Ty:=Ty0)
                 (G2 := (G4 § y0 / v_loc ~ Ty0) §& G6 §& subst_ctx w y G21)
                 (G3 := G4).
          - repeat(rewrite subst_ctx_concat). simpl.
            rewrite ctx_concat_assoc.
            eapply ctx_binds_k_after_concat; simpl; eauto.
            + apply fresh_telescope_ok, ctx_ok_concat_inv in HftG.
              destruct HftG as [Hok4 [Hok2 Hd1]].
              apply  ctx_ok_concat_inv in Hok2.
              destruct Hok2 as [Hok21 [Hok22 Hd2]].
              eapply ctx_ok_concat_map. eapply ctx_ok_concat; eauto.
              * inversions Hok21. rename H1 into Hok21, H4 into Hfrw0.
                constructor; eauto.
                { eapply ctx_ok_concat_map. eapply ctx_ok_concat; eauto.
                  hnf. apply fset_extens, subset_empty_l. introv Hin.
                  rewrite <- Hd1.
                  repeat(rewrite ctx_dom_concat in *).
                  repeat(rewrite ctx_dom_push in *).
                  repeat(rewrite in_inter in *).
                  repeat(rewrite in_union in *).
                  repeat(rewrite ctx_dom_concat in *).
                  repeat(rewrite ctx_dom_push in *).
                  repeat(rewrite in_inter in *).
                  repeat(rewrite in_union in *). tauto. }
                { hnf. introv Hw0in.
                  repeat(rewrite ctx_dom_concat in *).
                  repeat(rewrite ctx_dom_push in *).
                  repeat(rewrite ctx_dom_concat in Hd1).
                  repeat(rewrite ctx_dom_push in Hd1).
                  repeat(rewrite ctx_dom_map in *).
                  erewrite <- in_empty with (x:=w0), <- Hd1.
                  repeat(rewrite in_inter in *).
                  repeat(rewrite in_union in *).
                  destruct Hw0in as [Hw0in | Hw0in].
                  - repeat(rewrite in_singleton in *). tauto.
                  - false. }
              * rewrite <- ctx_concat_push, <- ctx_subst_push,
                  ctx_dom_concat, ctx_dom_map.
                apply disjoint_union_l; eauto.
                hnf. apply fset_extens; eauto using subset_empty_l.
                introv Hin. rewrite <- Hd1.
                repeat(rewrite ctx_dom_concat in *).
                repeat(rewrite ctx_dom_push in *).
                repeat(rewrite ctx_dom_concat in *).
                repeat(rewrite ctx_dom_push in *).
                repeat(rewrite in_inter in *).
                repeat(rewrite in_union in *). tauto.
            + eauto using ctx_weakens_map,
                fresh_telescope_concat_invert_r,
                fresh_telescope_weakens.
          - rewrite <- ctx_concat_assoc.
            eapply ctx_binds_k_after_concat; eauto.
            + apply fresh_telescope_ok in HftG.
              apply ctx_ok_concat_inv in HftG.
              destruct HftG as [Hok4 [Hok2 Hdis]].
              inversions Hok4. rename H4 into Hfrw, H1 into Hok4.
              rewrite ctx_concat_assoc.
              apply ctx_ok_concat_l in Hok2. inversions Hok2.
              rename H1 into Hok2, H4 into Hfrw0.
              eapply ctx_ok_concat_map, ctx_ok_concat; eauto.
              hnf. apply fset_extens; eauto using subset_empty_l.
              introv Hin. rewrite <- Hdis.
              repeat(rewrite ctx_dom_concat in *).
              repeat(rewrite ctx_dom_push in *).
              repeat(rewrite ctx_dom_concat in *).
              repeat(rewrite ctx_dom_push in *).
              repeat(rewrite in_inter in *).
              repeat(rewrite in_union in *). tauto.
            + rewrite ctx_concat_assoc in HftG.
              apply fresh_telescope_weakens, ctx_weakens_concat_left in HftG.
              apply ctx_weakens_concat_left in HftG as Hwk6.
              inversions Hwk6. rename H0 into Hwk6.
              apply ctx_weakens_concat_right in Hwk6.
              apply ctx_weakens_concat_right in HftG.
              inversions HftG. rename H0 into Hwk1.
              eauto using ctx_weakens_concat, ctx_weakens_map.
          - inversions Htycor; simpl in *; try congruence.
            destruct ( mode_refcap_corr' typing_mode) eqn: Hmc.
            typing_mode_contradiction typing_mode. 2: typing_mode_contradiction typing_mode.
            destruct H as [Tbnd [? ?]]; subst. simpl in *.
            replace (subst_typ w y T) with (T /[ w -> y]) by eauto.
            rewrite subst_fresh.
            + constructor; simpl; rewrite Hmc; eauto.
            + rewrite ctx_concat_assoc in HftG.
              do 2 (apply fresh_telescope_concat_invert_l in HftG).
              inversions HftG. clear Heq H3 H4 H6.
              rename H5 into Hfrw. rewrite ctx_fv_concat in *.
              simpl in Hfrw. repeat(rewrite notin_union in Hfrw). tauto.
          - eapply IH; eauto.
            rewrite fv_env_push, fv_env_concat, fv_env_push in Hwnin.
            rewrite fv_env_concat, fv_env_push.
            repeat(rewrite notin_union in *). tauto. }
      * eapply fresh_telescope_weakens,
          ctx_weakens_concat_right, ctx_weakens_concat_left in HftG.
        inversions HftG; eauto.
    + eauto using fresh_telescope_ok.
Qed.

(* G,e ~ Σ and G' is G with references w ~ y eliminated and Σ' is the Σ with
   w ~ y eliminated then G',e' ~ Σ' where the e' is e with w ~ y eliminated *)
Lemma ref_elim_heap G e Sigma e' G' w y :
  fresh_telescope G -> heap_correspond G e Sigma ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  env_elim e w y e' ->
  heap_correspond G' e' (subst_heap w y Sigma).
Proof.
  introv Hft Hhc HeG Hbi Hee.
  destruct Hhc as [HokG [HwkG [HokS [Hoke [Hdl [Hdr [Hcorr Htys]]]]]]].
  repeat(split);
    eauto using ref_elim_ctx_ok, ref_elim_ctx_weakens, ok_map, env_elim_ok,
    ref_elim_ctx_env_kdom_eq.
  - (* Use HeG and Hdl *)
    unfold subst_heap. rewrite dom_map, <- Hdl.
    erewrite <- ref_elim_ctx_neq_kdom; eauto. hnf. introv Hloc.
    eapply renv_corr_binds_vk in Hbi; eauto. destruct Hbi as [Hbi _].
    eapply (ctx_k_ne_false _ Hloc Hbi). Unshelve. discriminate.
  - eapply ref_elim_renv_corr; eauto. hnf. introv Hwin.
    apply fv_env_to_binds in Hwin; eauto.
    destruct Hwin as [w0 Hbi0].
    apply (renv_corr_binds_vk HokG Hcorr) in Hbi; eauto.
    apply (renv_corr_binds_vk HokG Hcorr) in Hbi0; eauto.
    destruct_all. assert (v_ref <> v_loc) by congruence.
    eauto using ctx_k_ne_false.
  - (* Use ref_elim_ty_item *)
    introv Hxdom. unfold subst_heap in Hxdom. rewrite dom_map in Hxdom.
    eapply ref_elim_ty_item with (e:=e) (G:=G); eauto.
Qed.

Lemma ref_elim_heap_push_back G e Sigma e' G' w y Tw Ty :
  fresh_telescope G -> heap_correspond G e Sigma ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  env_elim e w y e' ->
  ctx_binds y Ty G ->
  typ_corr Tw Ty ->
  heap_correspond (G' § w / v_ref ~ Tw) (e' & w ~ y) (subst_heap w y Sigma).
Proof.
  introv Hft Hhc HeG Hbi Hee HbiGy Htycorr.
  assert (Hhc0 : heap_correspond G e Sigma) by eauto.
  destruct Hhc0 as
    [HokG [HwkG [HokS [Hoke [Hdl [Hdr [Hcorr Htys]]]]]]].
  eapply ref_elim_heap with (e':=e') in Hhc; eauto.
  destruct Hhc as
    [HokG' [HwkG' [HokS' [Hoke' [Hdl' [Hdr' [Hcorr' Htys']]]]]]].

  repeat(split);
    eauto using ref_elim_ctx_ok_correct, fresh_telescope_ok, env_elim_ok,
    env_elim_ok_notin.

  - simpl. destruct (classicT (v_loc = v_ref)) as [? | _]; try discriminate.
    easy.

  - simpl. destruct (classicT (v_ref = v_ref)) as [_ | ?]; try tauto.
    rewrite dom_concat, dom_single, Hdr', union_comm. easy.

  - eapply renv_corr_push with (x:=w) (vk:=v_ref) (t:=Tw) in Hcorr';
      eauto using ref_elim_ctx_ok_correct.
    eapply ref_elim_ctx_renv_corr with (G:=G) in Hbi; eauto.
    destruct Hbi as [G11 [G12 [G2 [Ty0 [Tw0 [HeqG [Hfry HeqG']]]]]]].
    eapply renv_corr_ref with (G3:=G11); eauto. rewrite HeqG'.
    replace Ty0 with Ty in *.
    2: { rewrite HeqG, <- ctx_concat_push, <- ctx_concat_assoc in HbiGy.
         apply ctx_binds_middle_eq_inv in HbiGy; try easy.
         rewrite ctx_concat_assoc, ctx_concat_push, <- HeqG. easy. }
    rewrite <- ctx_concat_assoc. eapply ctx_binds_k_after_concat; eauto.
    + rewrite ctx_concat_assoc, <- HeqG'; easy.
    + rewrite HeqG' in HwkG'. rewrite <- ctx_concat_assoc in HwkG'.
      eauto using ctx_weakens_concat_right.

  - introv Hxdom. unfold subst_heap in Hxdom.
    specialize (Htys' _ Hxdom).
    eapply ty_item_s_epush; eauto 3 using ref_elim_ctx_ok_correct,
      env_elim_ok_notin.

Qed.

Lemma ref_elim_heap_ctx_push G e Sigma e' G' w y Tw Ty :
  fresh_telescope G -> heap_correspond G e Sigma ->
  ref_elim_ctx G w y G' ->
  binds w y e ->
  env_elim e w y e' ->
  ctx_binds y Ty G ->
  typ_corr Tw Ty ->
  heap_correspond (G' § w / v_ref ~ Tw) e (subst_heap w y Sigma).
Proof.
  introv Hft Hhc HeG Hbi Hee HbiGy Htycorr.
  assert (Hoke : ok e) by eauto.
  eapply ref_elim_heap_push_back in Hhc; eauto.
  pose proof (env_elim_push_permute Hee Hoke) as Hpe.
  eapply heap_correspond_epermute; eauto.
  all: intros; eapply Hpe; easy.
Qed.

End WithSefMode.
