Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping
        InertTightSubtyping TypeInduction
        GeneralToTight TightTyping InvertibleTyping FlatInvertibleTyping
        RecordTypes InertContext PreciseTyping
        Substitution Weakening
         Narrowing
        EnvCorrespondence EWeakening
        TFree Deref Incap.
Require Import Equivalence.
Require Import Indep.

Require Import LayeredCapabilityMain LayeredTypingMainHard.

Hint Resolve general_to_tight_typing : core.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma ty_icp_diag_reopen G e x v2 T:
  mode_has_incap typing_mode ->
  inert G ->
  ok e ->
  renv_corr G e -> 
  typ_indep_b 0 T ->
  e ⊢ x ≈ v2 ->
  G ⁏ e ⊢ open x T ∶icp open x T ->
  G ⁏ e ⊢ open v2 T ∶icp open v2 T.
Proof.
  introv Hm Hin Hoke Hcor Hind Htyq Hty.
  destruct (mode_incap' typing_mode) eqn: Hmi.
  - unfold  mode_has_incap' in Hm. rewrite Hmi in Hm. discriminate.
  - eapply ty_icp_diag_reopen_mode_structural; eauto.
  - eapply ty_incap_nocap. auto.
    eapply tight_to_general.
    eapply ty_has_N_main_tight.
    eapply ty_has_N_main_tyq with (U:= open x T). auto. auto.
    eapply subtyp_N_has_N_main. auto. auto.
    eapply (proj32 (general_to_tight Hin)).
    eapply ty_incap_diag_nocap_sup. eauto. eauto. eauto. eauto.
    eapply tyq_typ_open_indep_same; eauto.
    auto.
    auto.
Qed.

(* Corresponding variables and typing *)

Lemma ty_renv_binds_var G e1 v1 v2 T:
  inert G ->
  binds v1 v2 e1 ->
  ok e1 ->
  renv_corr G e1 ->
  typing G e1 v1 T ->
  typing G e1 v2 T.
Proof.
  introv Hin Hbi Hoke1 Hcor1 Hty1.
  assert (ctx_ok G) as HokG by eauto.
  destruct_renv_corr_binds_to_ref_loc HokG Hcor1 Hbi  TL TR.
  induction Hty1; eauto 4.
  - pose proof (ctx_binds_functional H HbiR); subst.
    eapply ty_var_sub.
    apply ty_var_var; eauto.
    apply typ_corr_subtyp; eauto.
  - apply ty_rec_intro; auto.
    eapply ty_var_sub.
    apply IHHty1; auto 1.
    apply subtyp_tyq; auto.

    (* In rodot, the read-only condition in recursive types is a bit ugly here *)
    
    intro. specialize (H1 H2).
    eapply ty_icp_diag_reopen with (x:=x); eauto.
  - eapply ty_var_sub.
    apply ty_rec_elim; eauto.
    apply subtyp_tyq; auto.
Qed.

(* This holds in basic modes where the type of reference is a subtype of type of location *)
Lemma ty_renv_binds_rev_var G e1 w y T:
  mode_refcap_corr' typing_mode = mode_refcap_corr_same ->
  inert G ->
  binds w y e1 ->
  ok e1 ->
  renv_corr G e1 ->
  typing G e1 y T ->
  typing G e1 w T.
Proof.
  introv Hm Hin Hbi Hoke1 Hcor1 Hty1.
  assert (ctx_ok G) as HokG by eauto.
  destruct_renv_corr_binds_to_ref_loc HokG Hcor1 Hbi  TL TR.
  induction Hty1; eauto 4.
  - pose proof (ctx_binds_functional H HbiL); subst.
    eapply ty_var_sub.
    apply ty_var_var; eauto.
    apply typ_corr_rev_subtyp; eauto.
  - apply ty_rec_intro; auto.
    eapply ty_var_sub. apply IHHty1; auto.
    apply subtyp_tyq; auto.
    (* We don't need to prove icp in modes without mutability *)
    intros Hmi.
    eapply ty_icp_diag_reopen with (x:=x); eauto.
  - eapply ty_var_sub.
    apply ty_rec_elim; eauto.
    apply subtyp_tyq; auto.
Qed.

Lemma ty_renv_binds_rev_sametype_var G e1 v1 v2 T U:
  inert G ->
  ctx_binds v1 U G ->
  ctx_binds v2 U G ->
  binds v1 v2 e1 ->
  ok e1 ->
  renv_corr G e1 ->
  typing G e1 v2 T ->
  typing G e1 v1 T.
Proof.
  introv Hin Hv1U Hv2U Hbi Hoke1 Hcor1 Hty1.
  assert (ctx_ok G) as HokG by eauto.
  destruct_renv_corr_binds_to_ref_loc HokG Hcor1 Hbi  TL TR.
  induction Hty1; eauto 4.
  - ctx_binds_eq.
    ctx_binds_eq.
    ctx_binds_eq.
    eapply ty_var_sub.
    apply ty_var_var; eauto.
    apply typ_corr_subtyp; eauto.
  - apply ty_rec_intro; auto.
    eapply ty_var_sub.
    apply IHHty1; auto 1.
    apply subtyp_tyq; auto.

    (* In rodot, the read-only condition in recursive types is a bit ugly here *)
    intro.
    eapply ty_icp_diag_reopen with (x:=x); eauto.

  - eapply ty_var_sub.
    apply ty_rec_elim; eauto.
    apply subtyp_tyq; auto.
Qed.

(* Equivalents of record types are incapacitated *)
Lemma record_tyq_incap G e T1 T2:
  record_type T1 ->
  e ⊢ T1 ≈ T2 ->
  G ⁏ e ⊢ T2 ∶icp T2.
Proof.
  intros.
  inversions H.
  revert dependent T2.
  induction H1; intros; subst.
  destruct D.
  * inversions_tyq H1. inversions H; eauto.
  * inversions_tyq H1. eauto.
  * inversions_tyq H1. eauto.
  * inversions_tyq H3. constructor.
    eapply IHrecord_typ; eauto.
    destruct D; inversions_tyq H7; try inversions H; eauto.
Qed.

(* Open forms of record types are incapacitated *)
Lemma record_open_tyq_incap G e T1 T2 w:
  record_type T1 ->
  e ⊢ (open w T1) ≈ T2 ->
  G ⁏ e ⊢ T2 ∶icp T2.
Proof.
  intros.
  inversions H.
  revert dependent T2.
  induction H1; intros; subst.
  destruct D.
  * inversions_tyq H1. inversions H; eauto.
  * inversions_tyq H1. eauto.
  * inversions_tyq H1. eauto.
  * inversions_tyq H3. constructor.
    eapply IHrecord_typ; eauto.
    destruct D; inversions_tyq H7; try inversions H; eauto.
Qed.

(* If opening a type with location is icapacitated, then so is opening with a reference *)
Lemma incap_refl_open_ref G e w y T:
  inert G ->
  ok e ->
  renv_corr G e ->
  binds w y e ->
  Indep.typ_indep_b 0 T ->
   G ⁏ e ⊢ open y T ∶icp open y T ->
   G ⁏ e ⊢ open w T ∶icp open w T.
Proof.
  introv Hin Hoke Hcor Hbi HidT Hicp.

  destruct (mode_has_incap' typing_mode) eqn: Hm.
  +
  eapply ty_icp_diag_reopen with (x:=y); eauto.
  +

  induction T; eauto.
  all: inversions_tyq HidT; simpls; inversions Hicp; eauto.
  1,2,3,5,7,8: unfold mode_has_incap' in  Hm; rewrite H in Hm; discriminate.
  all: destruct a; simpls.
  - remember (typ_sel (avar_f x) t (avar_f y0)).
    induction H8; try congruence; eauto.
    unfold mode_has_incap' in  Hm; rewrite H0 in Hm; discriminate.
  -
    inversions H. destruct a0. assert (0 <> n). intro. subst. contradiction. unfold open, AvarOpenable in H3. cases_if. unfold open, AvarOpenable in H3. inversions H3.
    replace (open w (avar_f v0)) with (avar_f v0). eauto. unfold open, AvarOpenable. eauto.
  - cases_if; inversions H.
    remember (typ_sel (avar_f y) (typ_label_normal l) (avar_f y0)).
    induction H7; subst; try congruence; simpl; eauto.
    unfold mode_has_incap' in  Hm; rewrite H in Hm; discriminate.
  - inversions H; eauto.

destruct a0. assert (0 <> n). intro. subst. contradiction.  unfold open, AvarOpenable in H3. cases_if. unfold open, AvarOpenable in H3. inversions H3.
    replace (open w (avar_f v0)) with (avar_f v0). eauto. unfold open, AvarOpenable. eauto.
Qed.

(* Incapacitated declaration.
   This helps extract common parts from proofs of reference_refl_incap_type and reference_incap_type.
   It is very similar to record_dec, but allows arbitrary bounds for field and type members.
*)
Inductive incap_dec : dec -> Prop :=
| incap_dec_typ : forall A S T, incap_dec (dec_typ (typ_label_normal A) S T)
| incap_dec_fld : forall a S T, incap_dec (dec_fld a S T)
| incap_dec_met : forall m TC TZ TR, incap_dec (dec_met m TC TZ TR)
.
Hint Constructors incap_dec: core.

Inductive incap_atomic_type : typ -> Prop :=
| incap_atomic_dec D: incap_dec D -> incap_atomic_type (typ_rcd D)
| incap_atomic_fun S T: incap_atomic_type (typ_fun S T)
.
Hint Constructors incap_atomic_type: core.

(* If a location has a declaration type,
   then a corresponding reference also has that type. *)
Lemma reference_incap_atomic_type G (e: renv) y2 T w2:
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  incap_atomic_type T ->
  G ⁏ e ⊢ y2 ∶ T ->
  binds w2 y2 e ->
  G ⁏ e ⊢ w2 ∶ T.
Proof.
  introv Hin Hcl Hoke HcorG HaT Hty Hbi.

  assert (G ⁏ e ⊢## y2 ∶ T) as Htyinv by eauto using tight_to_invertible.
  inversions HaT.
  inversions H.
  - (* typ *)
    eapply invertible_to_precise_typ_dec in Htyinv.
    destructc Htyinv as S' T' U Hpf Hst1 Hst2.
    eapply renv_corr_pf_normal_typ_tyq in Hpf.
    destructc Hpf as ? ? ? Hpfw ? ?.
    4: symmetry; apply tyq_var_binds; eauto.
    eapply ty_var_sub.
    eapply precise_to_general; eauto.
    eapply subtyp_typ.
    ++
      intros. specialize (Hst2 x2).
      eapply subtyp_trans.
      eapply tight_to_general; eauto.
      eapply subtyp_tyq; eauto using tyq_typ_open.
    ++
      intros. specialize (Hst1 x2).
      eapply subtyp_trans.
      eapply subtyp_tyq.
      symmetry; eapply H0.
      eapply tight_to_general; eauto.
    ++ eauto.
    ++ eauto.
  - (* fld *)
    eapply invertible_to_precise_fld_dec in Htyinv.
    destructc Htyinv as ? ? ? Hpf ? ?.
    eapply renv_corr_pf_tyq with (v2:=w2) in Hpf; eauto.
    destructc Hpf as ? ? Hpfw ? .
    eapply ty_var_sub.
    eapply precise_to_general.
    eauto.
    eapply Hpfw.
    inversions_tyq H1.
    eapply subtyp_fld.
    +
      eapply subtyp_trans.
      eapply tight_to_general; eauto.
      eauto.
    + intros.
      eapply subtyp_trans.
      eapply subtyp_tyq.
      symmetry; eauto.
      eapply tight_to_general; eauto.
    +
      right.
      eapply pf_dec_fld_inv in Hpf; subst.
      eexists.
      constructor.
      eapply rd_fld.
      eauto.
      eauto.
  - eapply invertible_to_precise_met_dec in Htyinv ; [ | eauto | eauto].
    destructc Htyinv as ? ? ? ? ? Hpf H.
    destructc H as ? ? ?.
    eapply renv_corr_pf_tyq with (v2 := w2) in Hpf; eauto.
    destructc Hpf as ? ? Hpfw ?.

    eapply ty_var_sub.
    eapply precise_to_general.
    eauto.
    eapply Hpfw.
    inversions_tyq H2.
    eapply subtyp_met.
    +
      eapply subtyp_trans.
      eapply tight_to_general; eauto.
      eauto.
    +
      intros.
      eapply subtyp_trans.
      eapply H0; eauto.
      eapply subtyp_tyq.
      eauto.
    + (* Caused by modified method subtyping for sef *)
      destructecs H1.
      intros.
      eapply typ_indep_tyq_open_all_indep.
      eapply H2; auto.
      intros. eapply H9.
    +
      intros.
      eapply subtyp_trans.
      eapply subtyp_tyq.
      symmetry. eauto.
      eapply H1; eauto.
    +
      right. eexists. constructor. constructor. eauto.
  -
    eapply invertible_to_precise_typ_fun in Htyinv; eauto.
    destructc Htyinv as ? ? ? ? Hpf ? ?.
    eapply renv_corr_pf_tyq in Hpf.
    destructc Hpf as ? ? ? ?.
    eapply ty_var_sub.
    eapply precise_to_general.
    eauto.

    eapply H1.
    inversions_tyq H2.
    eapply subtyp_fun.

    pose proof H1.
    eapply pf_binds in H2.
    eapply pf_inert_fun_U in H1; subst.
    eapply inert_binds_inert in H2.
    inversions H2.
    inversions H1.
    eapply H5. eauto. eauto.
    eapply subtyp_trans.
    eapply tight_to_general; eauto.
    eauto.
    intros.
    specialize (H0 x4 H2).
    eapply subtyp_trans.
    eapply subtyp_tyq.
    symmetry.
    eauto. eauto. eauto. eauto. eauto.
    left; eauto.
Qed.

(* If a location has a type that is incapacitated,
   then a corresponding reference also has that type. *)
Lemma reference_refl_incap_type_structural G (e: renv) y2 T w2:
  mode_incap' typing_mode = mode_incap_structural ->
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  G ⁏ e ⊢ y2 ∶ T ->
  binds w2 y2 e ->
  G ⁏ e ⊢ T ∶icp T ->
  G ⁏ e ⊢ w2 ∶ T.
Proof.
  introv Hmode Hin Hcl Hoke  HcorG  Hty Hbi Hicp.

  assert (ctx_ok G) as Hok by eauto.
  destruct_renv_corr_binds_to_ref_loc Hok HcorG Hbi TL TR.

  (* Induction using typ_icp_ind *)
  assert (typ_icp_ind T) as Ti by apply typ_icp_ind_all.
  induction Ti; inversions Hicp.
  all: try match goal with Hmi: mode_incap' typing_mode = mode_incap_nocap |- _ => rewrite Hmi in Hmode; discriminate end.
  - (* typ_top *)
    eapply ty_var_sub.
    eapply ty_var_var; eauto.
    eauto.
  - (* dec_typ *)
     eauto using reference_incap_atomic_type.
  - (* dec_met *)
     eauto using reference_incap_atomic_type.
  - (* dec_fld *)
     eauto using reference_incap_atomic_type.
  - (* typ_and *)
    eapply ty_var_and.
    eapply IHTi1; eauto.
    eapply IHTi2; eauto.
  - assert (G ⁏ e ⊢##f y2 ∶ (typ_or T1 T2)) as H1 by eauto using tight_to_invertible_flat.
    inversions H1.
    + inversions_tyq H0.
      eapply pf_or_false in H; eauto.
      contradiction.
    +
      assert (G ⁏ e ⊢ w2 ∶ T1).
      eapply IHTi1; eauto.
      eapply tight_to_general; [| eauto | eauto].
      eapply invertible_flat_to_tight; eauto.
      eauto.
    +
      assert (G ⁏ e ⊢ w2 ∶ T2).
       eapply IHTi2; eauto.
       eapply tight_to_general; [| eauto | eauto].
       eapply invertible_flat_to_tight;
       eauto. eauto.
  - remember (typ_sel (avar_f x0) L (avar_f y0)).
    induction H6; inversions Heqt.
    eauto.
    rewrite H in Hmode; discriminate.
  - assert (G ⁏ e ⊢##f y2 ∶ ( typ_bnd T)) as H1 by eauto using tight_to_invertible_flat.
    inversions H1.
    + inversions_tyq H3.
      pose proof H2.
      eapply pf_binds in H2.
      eapply pf_inert_bnd_U in H1.
      destructc H1 as ? ?; subst.
      ctx_binds_eq.
      clear H1.
      inversions_tyq Hcor.
      eapply ty_rec_intro.
      * eapply ty_var_sub.
        eapply ty_rec_elim.
        eapply ty_var_sub; eauto.
        apply subtyp_and11.
        eauto.
        eapply subtyp_tyq.
        eauto.
      * eapply typ_indep_tyq_open_all_indep.
        2: apply (H6 ).
        eauto.
      * intros.
        eapply inert_binds_inert in HbiR.
        inversion HbiR.
        inversions H2.
        eapply record_open_tyq_incap; eauto.
        eauto.
      * eauto.
    + forwards: H4.
      unfold mode_has_incap'. rewrite Hmode. auto.
      eapply ty_rec_intro; eauto.
      assert (G ⁏ e ⊢ w2 ∶ open y2 T ) by eauto.
      eapply ty_var_sub. eauto. eauto.
      intro.
      eapply incap_refl_open_ref; eauto.
  - eauto using reference_incap_atomic_type.
Qed.

Lemma reference_incap_type_structural G (e: renv) y2 T w2 Tstrip:
  mode_incap' typing_mode = mode_incap_structural ->
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  G ⁏ e ⊢ y2 ∶ T ->
  binds w2 y2 e ->
  G ⁏ e ⊢ T ∶icp Tstrip ->
  G ⁏ e ⊢ w2 ∶ Tstrip.
Proof.
  introv Hmode Hin Hcl Hoke HcorG Hty Hbi Hicp.

  assert (ctx_ok G) as Hok by eauto.
  destruct_renv_corr_binds_to_ref_loc Hok HcorG Hbi TL TR.

  induction Hicp.
  all: try match goal with Hmi: mode_incap' typing_mode = mode_incap_nocap |- _ => rewrite Hmi in Hmode; discriminate end.
  - eapply ty_var_sub.
    eapply ty_var_var; eauto.
    eauto.
  - assert (G ⁏ e ⊢##f y2 ∶ typ_bot) as H by eauto using tight_to_invertible_flat.
    inversions H.
    inversions_tyq H1.
    eauto.
    false.
    eapply pf_bot_false; eauto.
  - eauto using reference_incap_atomic_type.
  - eapply ty_var_sub.
    eapply ty_var_var.
    eauto.
    eauto.
  - eauto using reference_incap_atomic_type.
  - eauto using reference_incap_atomic_type.
  - assert ( G ⁏ e ⊢ y2 ∶ (open y T) ) by eauto.
    eapply IHHicp; eauto.
  - assert (G ⁏ e ⊢##f y2 ∶ (typ_bnd T)) as H by eauto using tight_to_invertible_flat.
    inversions H.
    + inversions_tyq H1.
      pose proof H0.
      eapply pf_binds in H.
      eapply pf_inert_bnd_U in H0.
      destructc H0 as ? ?; subst.
      ctx_binds_eq.
      clear H0.
      inversions_tyq Hcor.
      eapply ty_rec_intro.
      * eapply ty_var_sub.
        eapply ty_rec_elim.
        eapply ty_var_sub.
        eauto.
        apply subtyp_and11.
        eauto.
        eapply subtyp_tyq.
        eauto.
      *
        eapply typ_indep_tyq_open_all_indep.
        2: apply H4.
        eauto.
      *
        intros.
        eapply inert_binds_inert in HbiR.
        inversion HbiR.
        inversions H0.
        eapply record_open_tyq_incap; eauto.
        eauto.
      * eauto.
    + forwards: H2.
      unfold mode_has_incap'. rewrite Hmode. auto.
      eapply reference_refl_incap_type_structural; eauto.

  - eapply ty_var_and.
    eapply IHHicp; eauto.
    eapply IHHicp0; eauto.
  - assert (G ⁏ e ⊢##f y2 ∶ ( typ_or S1 T1)) as H1 by eauto using tight_to_invertible_flat.
    inversions H1.
    +
      inversions_tyq H3.
      eapply pf_or_false in H2; eauto.
      contradiction.
    +
      assert (G ⁏ e ⊢ w2 ∶ S2).
      eapply IHHicp; eauto.
      eapply tight_to_general.
      eapply invertible_flat_to_tight; eauto.
      eauto.
      eauto.
      eauto.
    +
      assert (G ⁏ e ⊢ w2 ∶ T2).
      eapply IHHicp0; eauto.
      eapply tight_to_general.
      eapply invertible_flat_to_tight; eauto.
      eauto.
      eauto.
      eauto.
  - eauto using reference_incap_atomic_type.
Qed.

(**
[G ⁏ e ⊢ y ∶ T]
[G ⁏ e ⊢ N <: T]
[-------------]
[G ⁏ e ⊢ w ∶ T]
*)
(* Lemma 9 in "Pure methods for roDOT" *)
Lemma reference_nocap_type G (e: renv) y2 T w2:
  mode_has_incap typing_mode ->
  mode_has_top_nm_subtype typing_mode ->
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  G ⁏ e ⊢ y2 ∶ T ->
  binds w2 y2 e ->
  G ⁏ e ⊢ typ_nocap <: T ->
  G ⁏ e ⊢ w2 ∶ T.
Proof.
  introv Hmode Hms Hin Hcl Hoke HcorG Hty Hbi Hicp.
  destruct (mode_incap' typing_mode) eqn: Hmi;
  destruct (mode_ty_inv' typing_mode) eqn: Hmt;
  try typing_mode_contradiction typing_mode.
  eapply tight_to_general. eapply invertible_to_tight. eauto. eauto.
  unfold ty_inv. rewrite Hmt.
  eapply ty_inv_main_ro_ref. eauto. eauto. eauto. eauto.  eauto.
  eapply subtyp_N_has_N_main. eauto. eauto. eapply (proj32 general_to_tight Hin). eauto.  eauto. eauto. eauto.
  eapply tight_to_invertible_main. eauto. eauto. eauto. eauto.  eauto.   eauto. 
Qed.

(* Lemma 3 in "Reference mutability for DOT" *)
Lemma reference_incap_type G (e: renv) y2 T w2 Tstrip:
  mode_has_incap typing_mode ->
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  G ⁏ e ⊢ y2 ∶ T ->
  binds w2 y2 e ->
  G ⁏ e ⊢ T ∶icp Tstrip ->
  G ⁏ e ⊢ w2 ∶ Tstrip.
Proof.
  introv Hmode Hin Hcl Hoke HcorG Hty Hbi Hicp.
  destruct (mode_incap' typing_mode) eqn: Hmi.
  - unfold mode_has_incap' in Hmode. rewrite Hmi in Hmode; discriminate.
  - eapply reference_incap_type_structural; eauto.
  - destruct (mode_refcap_corr' typing_mode) eqn: Hmc.
      + (* No capabilities *)
        eapply ty_renv_binds_rev_var; eauto.
        eapply ty_var_sub. eauto. eapply ty_incap_subtyp. eauto.
      + (* Lot of feature dependencies *)
        destruct (mode_ty_inv' typing_mode) eqn: Hmt;
        destruct (mode_has_top_nm_subtype' typing_mode) eqn: Hmns;
        try typing_mode_contradiction typing_mode.
        eapply reference_nocap_type; eauto.
        eapply ty_var_sub. eauto. eapply ty_incap_subtyp. eauto.  
        eapply ty_incap_diag_nocap_sup. eauto.
Qed.
   
End WithTypingMode.
