(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra.
Require Import AbstractSyntax TypingMode GeneralTyping PreciseTyping Equivalence
        EWeakening Indep RecordTypes InertContext
        TightTyping TypeApproximation Equivalence EnvCorrespondence UnboundSubstitution Nosel.
Require Import GeneralToTight.
Require Import Coq.Program.Equality.

Section WithTypingMode.
Context {typing_mode: Mode}.

Reserved Notation "F '⁏' e ⊢ atg ⦇ p ⦈ T1 '<:' T2"
         (at level 40, e at level 39, T1 at level 59, T2 at level 59,
          p at level 59, atg at level 59).

Inductive subtyp_restricted : ctx -> renv ->
                              approx_target -> polarity ->
                              typ -> typ -> Prop :=

| subtyp_top_res : forall (F : ctx) (e : renv)
                (atg : approx_target) (δ : polarity)
                (T : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ T <: typ_top

| subtyp_bot_res : forall (F : ctx) (e : renv)
                (atg : approx_target) (δ : polarity)
                (T : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ typ_bot <: T

| subtyp_refl_res : forall (F : ctx) (e : renv)
                 (atg : approx_target) (δ : polarity)
                 (T : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ T <: T

| subtyp_tyq_res : forall (F : ctx) (e : renv)
               (atg : approx_target) (δ : polarity)
               (T1 T2 : typ),
    approx_target_subtyping atg ->
    e ⊢ T1 ≈ T2 ->
    F⁏e ⊢atg⦇δ⦈ T1 <: T2

| subtyp_trans_res : forall (F : ctx) (e : renv)
                  (atg : approx_target) (δ : polarity)
                  (T1 T2 T3 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ T1 <: T2 ->
    F⁏e ⊢atg⦇δ⦈ T2 <: T3 ->
    F⁏e ⊢atg⦇δ⦈ T1 <: T3

| subtyp_or11_res : forall (F : ctx) (e : renv)
                (atg : approx_target) (δ : polarity)
                (T1 T2 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ T1 <: typ_or T1 T2

| subtyp_or12_res : forall (F : ctx) (e : renv)
                (atg : approx_target) (δ : polarity)
                (T1 T2 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ T2 <: typ_or T1 T2

| subtyp_or2_res : forall (F : ctx) (e : renv)
               (atg : approx_target) (δ : polarity)
               (T1 T2 T3 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ T1 <: T2 ->
    F⁏e ⊢atg⦇δ⦈ T3 <: T2 ->
    F⁏e ⊢atg⦇δ⦈ typ_or T1 T3 <: T2

| subtyp_and11_res : forall (F : ctx) (e : renv)
                 (atg : approx_target) (δ : polarity)
                 (T1 T2 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ typ_and T1 T2 <: T1

| subtyp_and12_res : forall (F : ctx) (e : renv)
                 (atg : approx_target) (δ : polarity)
                 (T1 T2 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ typ_and T1 T2 <: T2

| subtyp_and2_res : forall (F : ctx) (e : renv)
                (atg : approx_target) (δ : polarity)
                (T1 T2 T3 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ T1 <: T2 ->
    F⁏e ⊢atg⦇δ⦈ T1 <: T3 ->
    F⁏e ⊢atg⦇δ⦈ T1 <: typ_and T2 T3

| subtyp_nocap_bnd_res : forall (F : ctx) (e : renv)
                  (atg : approx_target) (δ : polarity)
                  (T : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ typ_nocap <: typ_bnd T

| subtyp_nocap_fld_res : forall (F : ctx) (e : renv)
                  (atg : approx_target) (δ : polarity)
                  (a : fld_label) (T1 T2 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ typ_nocap <: typ_rcd (dec_fld a T1 T2)

| subtyp_nocap_typ_res : forall (F : ctx) (e : renv)
                  (atg : approx_target) (δ : polarity)
                  A (T1 T2 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈ typ_nocap <: typ_rcd (dec_typ (typ_label_normal A) T1 T2)

| subtyp_nocap_fun_res : forall (F : ctx) (e : env var)
                           (atg : approx_target) (δ : polarity)
                           (TZ TR : typ),
    approx_target_subtyping atg ->
    F ⁏ e ⊢atg⦇δ⦈ (match mode_has_fun' typing_mode with
                   | true => typ_nocap
                   | false => typ_bot
                   end) <: typ_fun TZ TR

| subtyp_fld_res : forall (F : ctx) (e : renv)
                (atg : approx_target) (δ : polarity)
                (a : fld_label) (T1 T1' T2 T2' : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇─δ⦈ T1' <: T1 ->
    F⁏e ⊢atg⦇δ⦈ T2 <: T2' ->
    F⁏e ⊢atg⦇δ⦈ typ_rcd (dec_fld a T1 T2) <: typ_rcd (dec_fld a T1' T2')

| subtyp_typ_res : forall (F : ctx) (e : renv)
                (atg : approx_target) (δ : polarity)
                (A : typ_label) (T1 T1' T2 T2' : typ),
    approx_target_subtyping atg ->
    (forall x, F⁏e ⊢atg⦇─δ⦈ (open x T1') <: (open x T1)) ->
    (forall x, F⁏e ⊢atg⦇δ⦈ (open x T2) <: (open x T2')) ->
    F⁏e ⊢atg⦇δ⦈ typ_rcd (dec_typ A T1 T2) <: typ_rcd (dec_typ A T1' T2')

| subtyp_typ_and_res : forall (F : ctx) (e : renv)
                   (atg : approx_target) (δ : polarity)
                   (A : typ_label) (T1 T2 T3 T4 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈
      typ_and (typ_rcd (dec_typ A T1 T2)) (typ_rcd (dec_typ A T3 T4)) <:
      typ_rcd (dec_typ A (typ_or T1 T3) (typ_and T2 T4))

| subtyp_dist_res : forall (F : ctx) (e : renv)
                 (atg : approx_target) (δ : polarity)
                 (T1 T2 T3 : typ),
    approx_target_subtyping atg ->
    F⁏e ⊢atg⦇δ⦈
      typ_and T1 (typ_or T2 T3) <:
      typ_or (typ_and T1 T2) (typ_and T1 T3)

| subtyp_sel_u_res_s : forall (F : ctx) (e : renv)
                (v v2: var) (B : typ_label) (T1 T2 T : typ),
    F ⊢! v ∶ T ⪼ typ_rcd (dec_typ B T1 T2) ->
    F⁏e ⊢ˢ⦇⊕⦈ typ_sel (avar_f v) B (avar_f v2) <: (open v2 T2)

| subtyp_sel_l_res_s : forall (F : ctx) (e : renv)
                (v v2: var) (B : typ_label) (T1 T2 T : typ),
    F ⊢! v ∶ T ⪼ typ_rcd (dec_typ B T1 T2) ->
    F⁏e ⊢ˢ⦇⊝⦈ (open v2 T1) <: typ_sel (avar_f v) B (avar_f v2)

| subtyp_fun_res_s: forall L G e (δ : polarity) S1 T1 S2 T2,
    mode_has_fun typing_mode ->
    G⁏e ⊢# S2 <: S1 ->
    (forall x, x \notin L ->
       G § x / v_par ~ S2 ⁏e ⊢ open x T1 <: open x T2) ->
    G⁏e ⊢ˢ⦇δ⦈ typ_fun S1 T1 <: typ_fun S2 T2

| subtyp_nocap_res_nm_boundless: forall G e atg (δ : polarity) T1 T2,
    approx_target_subtyping atg ->
    mode_has_top_nm_subtype_boundless typing_mode ->
    G⁏e ⊢atg⦇δ⦈ typ_top <: (typ_or typ_nocap
                                 (typ_rcd (dec_typ (typ_label_refcap
                                                      typ_refcap_mutability)
                                                   T1 T2)))

| subtyp_nocap_res_nm_complement: forall G e atg (δ : polarity),
    approx_target_subtyping atg ->
    mode_has_top_nm_complement typing_mode ->
    G⁏e ⊢atg⦇δ⦈ typ_top <: (typ_or typ_nocap
                                 (typ_rcd (dec_typ (typ_label_refcap
                                                      typ_refcap_mutability)
                                                   typ_bot typ_bot)))

| subtyp_nocap_mut_bot_res G e atg (δ : polarity):
    approx_target_subtyping atg ->
    mode_has_top_nm_complement typing_mode ->
    G⁏e ⊢atg⦇δ⦈  (typ_and typ_nocap (typ_mut_marker typ_bot) ) <: typ_bot

| subtyp_mut_top_res G e atg (δ : polarity):
    approx_target_subtyping atg ->
    mode_has_mut_top typing_mode ->
    G⁏e ⊢atg⦇δ⦈  typ_top <: (typ_mut_marker typ_top)


| subtyp_met_res_s : forall (F : ctx) (e : renv)
               (δ : polarity)
               (m : met_label) (T1 T2 T3 T4 T5 T6 : typ) (L : fset var),
    F⁏e ⊢# T3 <: T1 ->
    (forall (z : var),
        z \notin L -> (* z is the parameter *)
        subtyping (F § z / v_par ~ T3) objctx_empty e
               (open z T6) (open z T5)) ->
    (* If typ_indep_b 0 T6 then typ_indep_b 0 T5 in modes with sef *)
    (mode_has_sef typing_mode -> typ_indep_b 0 T6 -> typ_indep_b 0 T5) ->
    (forall (z : var),
        z \notin L ->
        forall (r : var),
          r <> z -> r \notin L -> (* r is the receiver *)
          subtyping (ctx_met F z T3 r T6) objctx_empty e
                 (open_met z r T2) (open_met z r T4)) ->
    F⁏e ⊢ˢ⦇δ⦈ typ_rcd (dec_met m T5 T1 T2) <: typ_rcd (dec_met m T6 T3 T4)

| subtyp_nocap_met_res_s : forall (F : ctx) (e : renv)
                 (δ : polarity)
                 (m : met_label) (T1 T2 T3 : typ),
    F⁏e ⊢ˢ⦇δ⦈ typ_nocap <: typ_rcd (dec_met m T1 T2 T3)

| subtyp_sel_u_res_m : forall (F : ctx) (e : renv)
                (v v2: var) (B : typ_label) (T1 T2 T3 T : typ),
    F ⊢! v ∶ T ⪼ typ_rcd (dec_typ B T1 T2) ->
    (open v2 T2) ↝ᵐ⦇⊕⦈ T3 ->
    F⁏e ⊢ᵐ⦇⊕⦈ typ_sel (avar_f v) B (avar_f v2) <: T3

| subtyp_sel_l_res_m : forall (F : ctx) (e : renv)
                (v v2: var) (B : typ_label) (T1 T2 T3 T : typ),
    F ⊢! v ∶ T ⪼ typ_rcd (dec_typ B T1 T2) ->
    (open v2 T1) ↝ᵐ⦇⊝⦈ T3 ->
    F⁏e ⊢ᵐ⦇⊝⦈ T3 <: typ_sel (avar_f v) B (avar_f v2)

where "F '⁏' e ⊢ atg ⦇ p ⦈ T1 '<:' T2" := (subtyp_restricted F e atg p T1 T2).

Hint Constructors subtyp_restricted : core.
Remove Hints subtyp_trans_res : core.
Hint Resolve subtyp_trans_res | 10 : core.

(* Used in shortening_subtyp_bot_loc *)
Lemma approx_target_subtyping_s: approx_target_subtyping ˢ.
Proof.
  auto.
Qed.

(* Subtyping in the direction of polarity: T1 <: T2 for ⊕, T2 <: T1 for ⊝ *)
Notation "G '⁏' e ⊢ atg ⦇ δ ⦈ T1 '<:>' T2" := (polarity_apply (subtyp_restricted G e atg δ) δ T1 T2)
         (at level 40, e at level 39, T1 at level 59, T2 at level 59,
          δ at level 59, atg at level 59).

Notation "G '⁏' e ⊢ ⦇ δ ⦈ T1 '<:>' T2" := (polarity_apply (subtyping G objctx_empty e) δ T1 T2)
         (at level 40, e at level 39, T1 at level 59, T2 at level 59,
          δ at level 59).

Lemma subtyp_res_polar_inv G e atg δ T1 T2:
  G ⁏ e ⊢ atg ⦇ δ ⦈ T1 <:> T2 ->
  exists T3 T4,
  G ⁏ e ⊢ atg ⦇ δ ⦈ T3 <: T4 /\
  match δ with
  | ⊕ => T3 = T1 /\ T4 = T2
  | ⊝ => T3 = T2 /\ T4 = T1
  end.
Proof.
  intros.
  destruct δ;
  eauto.
Qed.

Definition general_subtyping_restricted (atg : approx_target) (δ : polarity): (ctx -> renv -> typ -> typ -> Prop) := fun G e => (subtyp_restricted G e atg δ).
Hint Unfold general_subtyping_restricted: core.

Instance and_or_subtyp_restricted (atg : approx_target) (δ : polarity) (atgSM: atg = ˢ \/ atg = ᵐ ): GeneralizedSubtyping.AndOrSubtyping (general_subtyping_restricted atg δ) := {
  and11 G e T1 T2 := @subtyp_and11_res G e atg δ T1 T2 atgSM;
  and12 G e T1 T2 := @subtyp_and12_res G e atg δ T1 T2 atgSM;
  and2 G e T1 T2 T3 := @subtyp_and2_res G e atg δ T1 T2 T3 atgSM;
  or11 G e T1 T2 := @subtyp_or11_res G e atg δ T1 T2 atgSM;
  or12 G e T1 T2 := @subtyp_or12_res G e atg δ T1 T2 atgSM;
  or2 G e T1 T2 T3 := @subtyp_or2_res G e atg δ T1 T2 T3 atgSM;
  trans G e T1 T2 T3 := @subtyp_trans_res G e atg δ T1 T2 T3 atgSM;
}.

(* Lemma 5.88 AndSubS and Lemma 5.89 OrSubS *)
Instance and_or_subtyp_restricted_s (δ : polarity): GeneralizedSubtyping.AndOrSubtyping (general_subtyping_restricted ˢ δ).
Proof.
  eapply and_or_subtyp_restricted; eauto.
Qed.
Hint Resolve and_or_subtyp_restricted_s: core.

(* Lemma 5.90 AndSubM and Lemma 5.91 OrSubM *)
Instance and_or_subtyp_restricted_m (δ : polarity): GeneralizedSubtyping.AndOrSubtyping (general_subtyping_restricted ᵐ δ).
Proof.
  eapply and_or_subtyp_restricted; eauto.
Qed.
Hint Resolve and_or_subtyp_restricted_m: core.

(* Lemma 5.92 SRedSub *)
(* Type approximation ˢ produces a supertype in ⊕ and subtype in ⊝ *)
Lemma typ_approx_s_sub δ G e T1 T2:
  sealed G ->
  G ⊢ T1 ↝ˢ⦇δ⦈ T2 ->
  G⁏e ⊢ˢ⦇δ⦈ T1 <:> T2.
Proof.
  introv Hcl Hap.
  dependent induction Hap.
  all: try destruct δ. all: eauto.
  - destruct H; subst.
    + eapply (@subtyp_and _ (and_or_subtyp_restricted_s _)); eauto.
    + eapply (@subtyp_or _ (and_or_subtyp_restricted_s _)); eauto.
  - destruct H; subst.
    + eapply (@subtyp_and _ (and_or_subtyp_restricted_s _)); eauto.
    + eapply (@subtyp_or _ (and_or_subtyp_restricted_s _)); eauto.
  - destruct H; try discriminate.
 Qed.

Lemma pf_dec_typ_subset_fv_l F v T B T1 T2:
  F ⊢! v ∶ T ⪼ typ_rcd (dec_typ B T1 T2) ->
  fv T1 \c (ctx_fv F \u ctx_dom F).
Proof.
  introv Hpf.
  intro. intro.
  destruct (fv_decide x (ctx_fv F \u ctx_dom F)). auto.
  eapply pf_dec_typ_notin_fv_l with (y:=x) in Hpf. contradiction. rewrite notin_union in n. destruct n. auto.
rewrite notin_union in n. destruct n. auto.
Qed.
Lemma pf_dec_typ_subset_fv_u F v T B T1 T2:
  F ⊢! v ∶ T ⪼ typ_rcd (dec_typ B T1 T2) ->
  fv T2 \c (ctx_fv F \u ctx_dom F).
Proof.
  introv Hpf.
  intro. intro.
  destruct (fv_decide x (ctx_fv F \u ctx_dom F)). auto.
  eapply pf_dec_typ_notin_fv_u with (y:=x) in Hpf. contradiction. rewrite notin_union in n. destruct n. auto.
rewrite notin_union in n. destruct n. auto.
Qed.

Lemma typ_approx_subst_closed0 δ T2 T3 x y:
  (forall z, open z T2 = T2) ->
  (forall z, open z T3 = T3) ->
  T2 ↝ ᵐ ⦇ δ ⦈ T3 ->
  (T2 /[ x -> y]) ↝ ᵐ ⦇ δ ⦈ (T3 /[ x -> y]).
Proof.
  introv Ho2 Ho3 Hap.
  rewrite <- (Ho2  x).
  rewrite  <-  open_close.
  rewrite <- subst_intro.

  rewrite <- (Ho3  x).
  rewrite  <- (open_close 0 x).
  rewrite <- subst_intro.
  eapply typ_approx_reopen with (x0:=x). simpls. notin_solve.  eapply notin_fv_close.  eapply notin_fv_close.
  rewrite open_close.

  rewrite open_close.
  rewrite Ho2.
  rewrite Ho3. auto.
  eapply notin_fv_close.  eapply notin_fv_close.
Qed.

Lemma typ_approx_subst_vars_closed0 δ T2 T3 x y:
  length(x) = length y ->
  (forall z, open z T2 = T2) ->
  (forall z, open z T3 = T3) ->
  T2 ↝ ᵐ ⦇ δ ⦈ T3 ->
  (T2 /[ x --> y]) ↝ ᵐ ⦇ δ ⦈ (T3 /[ x --> y]).
Proof.
  introv Hl Ho2 Ho3 Hap.

  reverts y.
  induction x; intros; simpls; auto; destruct y; auto.
  eapply typ_approx_subst_closed0; intros. eapply closed0_subst_vars; eauto.
  eapply closed0_subst_vars; eauto.
  eapply IHx. simpls. inversions Hl. auto.
Qed.


Lemma typ_approx_subst_pvars_closed0 δ T2 T3 x y:
  length(x) = length y ->
  (forall z, open z T2 = T2) ->
  (forall z, open z T3 = T3) ->
  T2 ↝ ᵐ ⦇ δ ⦈ T3 ->
  (T2 //[ x --> y]) ↝ ᵐ ⦇ δ ⦈ (T3 //[ x --> y]).
Proof.
  introv Hl Ho2 Ho3 Hap.

  lets vs0 Hf0 Heq0: (subst_pvars_unfold_L T2 x y (fv T2 \u fv T3)).

(* edestruct (subst_pvars_unfold_L T2 ) as [zs [Hfr Hsubst]].
  rewrite Hsubst.*) rewrite Heq0.

  edestruct (subst_pvars_unfold ( (T3))) as [zs3 [Hfr3 Hsubst3]].
  rewrite Hsubst3.

  rewrite subst_vars_fresh_middle_change with (zs:=zs3) (ws:=vs0); eauto.
  eapply typ_approx_subst_vars_closed0. eapply fresh_length in Hf0. auto.
  eapply closed0_subst_vars; eauto.
  eapply closed0_subst_vars; eauto.

  eapply typ_approx_subst_vars_closed0. eapply fresh_length in Hf0. rewrite Hf0. auto.
  auto.
  auto.
  auto.
Qed.



Lemma typ_approx_open_subst_pvars δ  v2 T2  T3 x y:
  length y = length x ->
  T2 /[-> v2] ↝ ᵐ ⦇ δ ⦈ T3 ->
  T2 /[-> v2] //[ x --> y] ↝ ᵐ ⦇ δ ⦈ (T3 //[ x --> y]).
Proof.
  introv Hl Hap.
  eapply typ_approx_subst_pvars_closed0; eauto.
  intro. rewrite open_twice. auto.
  intro.
  eapply typ_approx_closed_n in Hap; eauto.
  rewrite open_twice. auto.
Qed.


Lemma subtyp_res_usubst_pvars atg δ G e T1 T2 xs ys:
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  length ys = length xs ->
  fresh (fv_renv e \u ctx_dom G \u ctx_fv G) (length xs) xs ->
  G⁏e ⊢atg⦇δ⦈ T1 <: T2 ->
  G⁏e ⊢atg⦇δ⦈ (T1 //[xs --> ys]) <: (T2 //[xs --> ys]).
Proof.
  introv Hin HslG Hoke Hcor Hl Hn Hsub.
  reverts xs ys.
  induction Hsub; intros; sympls.
  all: autorewrite with simpl_subst_pvars; eauto 4.
  - (* tyq *)
    eapply usubst_pvars_tyq_typ in H0; eauto.
  - pose proof subtyp_nocap_fun_res.
    destruct (mode_has_fun' typing_mode) eqn:Htm.
    all: rewrite typ_subst_pvars_nulary; eauto.
  - (* typ *)

    pick_fresh_with z (from_list xs \u from_list ys).
    assert (z \notin (from_list xs)) as Hnzxs by eauto.
    assert (z \notin (from_list ys)) as Hnzys by eauto.
    assert (z \notin (fv F)) as HnzG2. { unfold fv, CtxFreeVar.
      notin_solve.
    }

    eapply subtyp_typ_res; intros. eauto.

    rewrite subst_intro with (x0:=z).
    rewrite subst_intro with (x0:=z) (X:=(T1 //[ xs --> ys])).
    repeat rewrite <- (subst_pvars_open_fresh_commut ) by eauto.
    rewrite <- subst_pvars_push by eauto.
    rewrite <- subst_pvars_push by eauto.

    eapply H1; simpls; eauto.
    eapply notin_fv_subst_pvars; eauto. eapply notin_fv_subst_pvars; eauto.

    rewrite subst_intro with (x0:=z).
    rewrite subst_intro with (x0:=z) (X:=(T2' //[ xs --> ys])).
    repeat rewrite <- (subst_pvars_open_fresh_commut ) by eauto.
    rewrite <- subst_pvars_push by eauto.
    rewrite <- subst_pvars_push by eauto.

    eapply H3; simpls; eauto.
    eapply notin_fv_subst_pvars; eauto. eapply notin_fv_subst_pvars; eauto.
  - (* sel_u *)
   
    pose proof H. eapply pf_binds in H0. eapply ctx_binds_dom in H0.
    erewrite subst_pvars_fresh_ctx_dom; eauto.
    rewrite subst_pvars_open_commut.

    replace (T2 //[xs --> ys]) with T2.

    eapply subtyp_sel_u_res_s. eauto. symmetry.
    eapply subst_pvars_fresh. eapply fresh_subset.
    eapply pf_dec_typ_subset_fv_u; eauto. fresh_solve.
  - (* sel_l *)

    pose proof H. eapply pf_binds in H0. eapply ctx_binds_dom in H0.
    erewrite subst_pvars_fresh_ctx_dom; eauto.
    rewrite subst_pvars_open_commut.

    replace (T1 //[xs --> ys]) with T1.

    eapply subtyp_sel_l_res_s. eauto. symmetry.
    eapply subst_pvars_fresh. eapply fresh_subset.
    eapply pf_dec_typ_subset_fv_l; eauto. fresh_solve.
  - (* fun *)
    eapply subtyp_fun_res_s. auto.
    eapply tight_to_general in H0.
    eapply general_to_tight; eauto.
    eapply usubst_end; eauto. eauto. eauto.

    intro_fresh_with x Hx (from_list xs \u from_list ys).

    rewrite <- ! (subst_pvars_open_fresh_commut ) by eauto.
    replace G with (G §& (subst_ctx_pvars xs ys ctx_empty)).
    rewrite subst_ctx_pvars_push.
    eapply usubstitution; ctx_flatten; eauto.
    simpls. auto.
  - rename F into G.
    eapply subtyp_met_res_s. auto.
    eapply tight_to_general in H.
    eapply general_to_tight; eauto. sympls.
    eapply usubst_end; eauto.  eauto. eauto.

    intro_fresh_with x Hx (from_list xs \u from_list ys).
    intros.
    rewrite <- ! (subst_pvars_open_fresh_commut ) by eauto.
    replace G with (G §& (subst_ctx_pvars xs ys ctx_empty)).
    rewrite  subst_ctx_pvars_push.
    eapply usubstitution.
    ctx_flatten. eauto. ctx_flatten. eauto. eauto. ctx_simpl_dom. eauto.
    simpls. auto.

    (* Caused by modified method subtyping for sef *)
    {
    intros.
        eapply subst_pvars_indep_b. auto. eapply H1. auto.
    eapply subst_pvars_indep_b_inv. symmetry. eauto. eauto.

    }

    intro_fresh_with z Hz (from_list xs \u from_list ys).
    intros.
    rewrite <- ! (subst_pvars_open_fresh_commut ) by eauto.
    replace G with (G §& (subst_ctx_pvars xs ys ctx_empty)).
    rewrite  subst_ctx_pvars_push.
    rewrite  subst_ctx_pvars_push.

    eapply usubstitution.
    ctx_flatten. eauto. ctx_flatten. eauto. eauto. ctx_simpl_dom. eauto.
    simpls. auto.
  - (* sel_u *)

    pose proof H. eapply pf_binds in H1. eapply ctx_binds_dom in H1.
    erewrite subst_pvars_fresh_ctx_dom; eauto.
    eapply subtyp_sel_u_res_m.
    eauto.
    replace T2 with (T2 //[xs --> ys]).
    rewrite <- subst_pvars_open_commut.

    eapply typ_approx_open_subst_pvars; eauto.
    eapply subst_pvars_fresh. eapply fresh_subset.
    eapply pf_dec_typ_subset_fv_u; eauto. eauto.
 -
    pose proof H. eapply pf_binds in H1. eapply ctx_binds_dom in H1.
    erewrite subst_pvars_fresh_ctx_dom; eauto.
    eapply subtyp_sel_l_res_m.
    eauto.
    replace T1 with (T1 //[xs --> ys]).
    rewrite <- subst_pvars_open_commut.

    eapply typ_approx_open_subst_pvars; eauto.
    eapply subst_pvars_fresh. eapply fresh_subset.
    eapply pf_dec_typ_subset_fv_l; eauto. eauto.
Qed.

Lemma subtyp_res_usubst atg δ G e T1 T2 x y:
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  x \notin (fv_renv e \u ctx_dom G \u ctx_fv G) ->
  G⁏e ⊢atg⦇δ⦈ T1 <: T2 ->
  G⁏e ⊢atg⦇δ⦈ (T1 /[x -> y]) <: (T2 /[x -> y]).
Proof.
  intros.
  rewrite <- subst_pvars_one.
  rewrite <- subst_pvars_one.
  eapply subtyp_res_usubst_pvars; eauto.
Qed.

Lemma typ_approx_reopen_subst δ atg F x y T1 T2:
  sealed F ->
  inert F ->
  x \notin ctx_dom F \u ctx_fv F \u fv T1 ->
  F ⊢ open x T1 ↝ atg ⦇ δ ⦈ T2 ->
  open y (close x T2) = (T2 /[ x -> y ]).
Proof.
  introv HclG HinG Hn Hap.
  rewrite <- subst_open_close.

  pose proof Hap. eapply (typ_approx_closed_n 0) with (x0:=x) in H; eauto.
  rewrite H. auto. rewrite open_twice. auto.
Qed.
Lemma typ_approx_bot_inv atg d G T:
  G ⊢  typ_bot ↝ atg ⦇ d ⦈ T ->
  T = typ_bot.
Proof.
  introv Hap.
  inversions Hap. auto.
  destructecs_all; subst; inversions H.
  auto.
Qed.


Lemma typ_approx_bot_open_inv atg d G T:
  (forall x : var, G ⊢  typ_bot /[-> x] ↝ atg ⦇ d ⦈ T /[-> x] )->
  T = typ_bot.
Proof.
  introv Hap.
  pick_fresh y.
  specialize (Hap y).
  eapply typ_approx_bot_inv in Hap.
  destruct T; sympl; try discriminate. auto.
Qed.

Lemma typ_approx_top_inv atg d G T:
  G ⊢  typ_top ↝ atg ⦇ d ⦈ T ->
  T = typ_top.
Proof.
  introv Hap.
  inversions Hap. auto.
  destructecs_all; subst; inversions H.
  auto.
Qed.


Lemma typ_approx_top_open_inv atg d G T:
  (forall x : var, G ⊢  typ_top /[-> x] ↝ atg ⦇ d ⦈ T /[-> x] )->
  T = typ_top.
Proof.
  introv Hap.
  pick_fresh y.
  specialize (Hap y).
  eapply typ_approx_top_inv in Hap.
  destruct T; sympl; try discriminate. auto.
Qed.


Lemma typ_approx_nocap_inv atg d G T:
  G ⊢  typ_nocap ↝ atg ⦇ d ⦈ T ->
  T = typ_nocap.
Proof.
  introv Hap.
  inversions Hap. auto.
  destructecs_all; subst; inversions H.
  auto.
Qed.

(* Lemma 5.98 SRedSubCom *)
Lemma typ_approx_s_subcom δ G e T1 T2 T3 T4:
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  G⁏e ⊢ˢ⦇δ⦈ T1 <: T2 ->
  G ⊢ T1 ↝ˢ⦇δ⦈ T3 ->
  G ⊢ T2 ↝ˢ⦇δ⦈ T4 ->
  exists T5 T6,
  G ⊢ T3 ↝ˢ⦇δ⦈ T5 /\
  G ⊢ T4 ↝ˢ⦇δ⦈ T6 /\
  G⁏e ⊢ˢ⦇δ⦈ T5 <: T6.
Proof.
  introv Hin Hcl Hoke Hcor Hsub12 Hap1 Hap2.
  revert dependent T4.
  revert dependent T3.
  dependent induction Hsub12; intros; eauto.
  - inversions_typ Hap2. exists T3. eauto 8.
  - inversions_typ Hap1. exists typ_bot. eauto.
  - destruct (typ_approx_s_com Hin Hcl Hap1 Hap2) as [T5 [Hap35 Hap45]].
    exists T5. eauto.
  - (*tyq*)
    destruct δ.
    + destruct (typ_approx_tyq Hin Hcl Hcor H0 Hap1) as [T5 [Hap25 Htyq35]].
      destruct (typ_approx_s_com Hin Hcl Hap2 Hap25) as [T6 [Hap35 Hap45]].
      pose proof (typ_approx_s_sub e Hcl Hap45). simpl in H1.
      eauto 7.
    + destruct (typ_approx_tyq Hin Hcl Hcor (tyq_typ_sym H0) Hap2) as [T5 [Hap25 Htyq45]].
      destruct (typ_approx_s_com Hin Hcl Hap1 Hap25) as [T6 [Hap36 Hap56]].
      pose proof (typ_approx_s_sub e Hcl Hap56). simpl in H1.
      symmetry in Htyq45.
      eauto 7.
  - (* trans *)
    autospecialize_prop.
    destruct δ.
    +
      destruct (IHHsub12_1 T0 Hap1 T2) as [T5 [T6 [? [? ?]]]]. eauto.
      destruct (IHHsub12_2 T6 H1 T4 Hap2) as [T7 [T8 [? [? ?]]]].
      pose proof (typ_approx_s_sub e Hcl H3). simpl in H6.
      eauto 7.
    + edestruct IHHsub12_2 as [T5 [T6 [? [? ?]]]]. eapply typ_approx_refl. eauto.
      edestruct IHHsub12_1 as [T7 [T8 [? [? ?]]]]. eauto. eauto.
      pose proof (typ_approx_s_sub e Hcl H4). simpl in H6.
      eauto 7.
 - inversions_typ Hap2. eauto 8.
   destruct (typ_approx_s_com Hin Hcl Hap1 H2) as [T6 [? ?]].
   eauto 8.
 - inversions_typ Hap2. eauto 8.
   destruct (typ_approx_s_com Hin Hcl Hap1 H4) as [T6 [? ?]].
   eauto 8.
 - epose proof (typ_approx_s_inv_pos _ _ _ Hap1 ) as Hinv. (* Shelves *)
   destruct Hinv as [T01 [T02 [? [? ?]]]]; subst.
   autospecialize_prop.
   + edestruct IHHsub12_1 as [T5 [T6 [? [Hap46 ?]]]].
     apply H1.
     apply Hap2.
     edestruct IHHsub12_2 as [T7 [T8 [? [Hap48 ?]]]].
     apply H2.
     apply Hap2.
     destruct (typ_approx_s_com Hin Hcl Hap46 Hap48) as [T9 [Hap69 Hap89]].
     pose proof (typ_approx_s_sub e Hcl Hap69) as Hsub69.
     pose proof (typ_approx_s_sub e Hcl Hap89) as Hsub89.
     pose proof (typ_approx_s_sub e Hcl Hap46) as Hsub46.
     pose proof (typ_approx_s_sub e Hcl Hap48) as Hsub48.
     destruct δ.
     * exists (typ_or T5 T7) T9.
       repeat split; eauto.
       eapply typ_approx_s_trans; eauto.
     * exists (typ_or T5 T7) T4.
       repeat split; eauto.
   Unshelve. eauto.
 - inversions_typ Hap1. eauto 8.
   destruct (typ_approx_s_com Hin Hcl Hap2 H2) as [T6 [? ?]].
   exists (typ_and T6 T2') T6. repeat split; eauto.
 - inversions_typ Hap1. eauto 8.
   destruct (typ_approx_s_com Hin Hcl Hap2 H4) as [T6 [? ?]].
   exists (typ_and T1' T6) T6. repeat split; eauto.
 - epose proof (typ_approx_s_inv_pos _ _ _ Hap2) as Hinv. (* Shelves *)
   destruct Hinv as [T02 [T03 [? [? ?]]]]; subst.
   autospecialize_prop.
   + edestruct IHHsub12_1 as [T5 [T6 [Hap5 [? ?]]]].
     apply Hap1.
     apply H1.
     edestruct IHHsub12_2 as [T7 [T8 [Hap7 [? ?]]]].
     apply Hap1.
     apply H2.
     destruct (typ_approx_s_com Hin Hcl Hap5 Hap7) as [T9 [Hap59 Hap79]].
     pose proof (typ_approx_s_sub e Hcl Hap59) as Hsub59.
     pose proof (typ_approx_s_sub e Hcl Hap79) as Hsub79.
     pose proof (typ_approx_s_sub e Hcl Hap5) as Hsub5.
     pose proof (typ_approx_s_sub e Hcl Hap7) as Hsub7.
     destruct δ.
     * exists T0 (typ_and T6 T8).
       repeat split; eauto.
     * exists T9 (typ_and T6 T8).
       repeat split; eauto.
       eapply typ_approx_s_trans; eauto.
   Unshelve. eauto.
 - inversions_typ Hap1. inversions_typ Hap2. eauto 6.
 - inversions_typ Hap1. inversions_typ Hap2. eauto 6. eauto 6.
 - inversions_typ Hap1. inversions_typ Hap2. eauto 6. eauto 6.
 - (* fun *)
   pose proof subtyp_nocap_fun_res.
   destruct (mode_has_fun' typing_mode) eqn:Htm.
   + inversions Hap1. inversions Hap2. eauto 6.
     destruct H2; subst; try discriminate.
     destruct H7; subst; try discriminate.
     destruct H2; subst; try discriminate.
     destruct H1; subst; try discriminate.
   + inversions Hap1. inversions Hap2. eauto 6.
     destruct H2; subst; try discriminate.
     destruct H7; subst; try discriminate.
     destruct H2; subst; try discriminate.
     destruct H1; subst; try discriminate.
 - (* fld *)
   destruct (typ_approx_s_inv_fld Hap1) as [T01 [T02 [? [? ?]]]]; subst.
   destruct (typ_approx_s_inv_fld Hap2) as [T01' [T02' [? [? ?]]]]; subst.
   autospecialize_prop. destruct δ.
   + edestruct IHHsub12_1 as [T5 [T6 [Hap5 [? ?]]]]; eauto.
     edestruct IHHsub12_2 as [T7 [T8 [Hap7 [? ?]]]]; eauto.
     exists (typ_rcd (dec_fld a T6 T7)).
     exists (typ_rcd (dec_fld a T5 T8)).
     repeat split; eauto.
   + edestruct IHHsub12_1 as [T5 [T6 [Hap5 [? ?]]]]; eauto.
     edestruct IHHsub12_2 as [T7 [T8 [Hap7 [? ?]]]]; eauto.
     exists (typ_rcd (dec_fld a T6 T7)).
     exists (typ_rcd (dec_fld a T5 T8)).
     repeat split; eauto.
 - destruct (typ_approx_s_inv_typ Hap1) as [T01 [T02 [? [? ?]]]]; subst.
   destruct (typ_approx_s_inv_typ Hap2) as [T01' [T02' [? [? ?]]]]; subst.

    pick_fresh y.
    forwards T5 T6 Hap5 ? ?: (H1 y); eauto.
    forwards T7 T8 Hap7 ? ?: (H3 y); eauto.
    exists (typ_rcd (dec_typ A (close y T6) (close y T7))).
    exists (typ_rcd (dec_typ A (close y T5) (close y T8))).
    splits.
    * constructor.
      intros. eapply typ_approx_reopen_closed'; eauto using ctx_sealed_empty.
      intros. eapply typ_approx_reopen_closed'; eauto using ctx_sealed_empty.
    * constructor.
      intros. eapply typ_approx_reopen_closed'; eauto using ctx_sealed_empty.
      intros. eapply typ_approx_reopen_closed'; eauto using ctx_sealed_empty.
    * eapply subtyp_typ_res. eauto. intro.
      erewrite typ_approx_reopen_subst; eauto.
      erewrite typ_approx_reopen_subst; eauto.
      eapply subtyp_res_usubst; eauto.
      notin_solve.
      notin_solve.
 intro.
      erewrite typ_approx_reopen_subst; eauto.
      erewrite typ_approx_reopen_subst; eauto.
      eapply subtyp_res_usubst; eauto.
      notin_solve.
      notin_solve.

 - epose proof (typ_approx_s_inv_pos _ _ _ Hap1) as Hinv. (* Shelves *)
   destruct Hinv as [T12 [T34 [? [? ?]]]]; subst.
   destruct (typ_approx_s_inv_typ H1) as [T01 [T02 [? [? ?]]]]; subst.
   destruct (typ_approx_s_inv_typ H2) as [T03 [T04 [? [? ?]]]]; subst.
   destruct (typ_approx_s_inv_typ Hap2) as [T013 [T024 [? [? ?]]]]; subst.

    pick_fresh z.

   epose proof (typ_approx_s_inv_pos _ _ _ (H7 z)) as Hinv. (* Shelves *)
   destruct Hinv as [T001 [T003 [? [? ?]]]]; subst.
   epose proof (typ_approx_s_inv_pos _ _ _ (H8 z)) as Hinv. (* Shelves *)
   destruct Hinv as [T002 [T004 [? [? ?]]]]; subst.

   autospecialize.

   typ_approx_s_com_for_type (open z T1) T1.
   typ_approx_s_com_for_type (open z T2) T2.
   typ_approx_s_com_for_type (open z T3) T3.
   typ_approx_s_com_for_type (open z T4) T4.

   destruct δ.
   exists (typ_and (typ_rcd (dec_typ A (close z T0) (close z T5))) (typ_rcd (dec_typ A (close z T6) (close z T7)))).
   exists (typ_rcd (dec_typ A (typ_or (close z T0) (close z T6)) (typ_and (close z T5) (close z T7)))).
   repeat split; eauto. constructor; eauto;
      constructor; intros; eapply typ_approx_reopen_closed'; eauto.
      constructor; intros.
      destruct T013; simpl; inversions H0; sympls.
      constructor; eauto; eapply typ_approx_reopen_closed'; eauto.
      destruct T024; simpl; inversions H11; sympls.
      constructor; eauto; eapply typ_approx_reopen_closed'; eauto.

   exists (typ_and (typ_rcd (dec_typ A (close z T0) (close z T5))) (typ_rcd (dec_typ A (close z T6) (close z T7)))).
   exists (typ_rcd (dec_typ A (typ_or (close z T0) (close z T6)) (typ_and (close z T5) (close z T7)))).
   repeat split; eauto. constructor; eauto;
      constructor; intros; eapply typ_approx_reopen_closed'; eauto.
      constructor; intros.
      destruct T013; simpl; inversions H0; sympls.
      constructor; eauto; eapply typ_approx_reopen_closed'; eauto.
      destruct T024; simpl; inversions H11; sympls.
      constructor; eauto; eapply typ_approx_reopen_closed'; eauto.
   Unshelve. all: auto.
 - (* dist *)
   epose proof (typ_approx_s_inv_pos _ _ _ Hap1) as Hinv. (* Shelves *)
   destruct Hinv as [T01 [T023 [? [? ?]]]]; subst.
   epose proof (typ_approx_s_inv_pos _ _ _ H2) as Hinv. (* Shelves *)
   destruct Hinv as [T02 [T03 [? [? ?]]]]; subst.
   epose proof (typ_approx_s_inv_pos _ _ _ Hap2) as Hinv. (* Shelves *)
   destruct Hinv as [T012 [T013 [? [? ?]]]]; subst.
   epose proof (typ_approx_s_inv_pos _ _ _ H5) as Hinv. (* Shelves *)
   destruct Hinv as [T201 [T102 [? [? ?]]]]; subst.
   epose proof (typ_approx_s_inv_pos _ _ _ H6) as Hinv. (* Shelves *)
   destruct Hinv as [T301 [T103 [? [? ?]]]]; subst.

   destruct (typ_approx_s_com Hin Hcl H1 H7) as [T021 [? ?]].
   destruct (typ_approx_s_com Hin Hcl H1 H9) as [T031 [? ?]].
   destruct (typ_approx_s_com Hin Hcl H0 H12) as [T0321 [? ?]].

   typ_approx_s_com_for T2.
   typ_approx_s_com_for T3.

   exists (typ_and T0321 (typ_or T0 T4)).
   exists (typ_or (typ_and T0321 T0) (typ_and T0321 T4)).
   repeat split.
   + constructor; eauto using typ_approx_s_trans.
   + constructor; eauto using typ_approx_s_trans.
   + eauto.
   Unshelve. all: auto.
 - inversions_typ Hap1. eauto 8.
   destruct (pf_inert_unique_tight_bounds Hin H H3); subst.
   typ_approx_s_com_for_type (open v2 T5) T5. exists T1, T1. eauto.
 - inversions_typ Hap2. eauto 8.
   destruct (pf_inert_unique_tight_bounds Hin H H3); subst.
   typ_approx_s_com_for_type (open v2 T0) T0. exists T1, T1. eauto.
 - inversions_typ Hap1. inversions_typ Hap2. eauto 8.
 - (* subtyp_nocap_nm_s *)
   inversions Hap1. inversions Hap2. eauto 8.
   all: repeat(match goal with
               | [ H : _ = _ \/ _ |- _ ] => destruct H; subst; try discriminate
               end).
   inversions H1. inversions H3.
   all: repeat(match goal with
               | [ H : _ = _ \/ _ |- _ ] => destruct H; subst; try discriminate
               end).
   inversions H5.
   all: repeat(match goal with
               | [ H : _ = _ \/ _ |- _ ] => destruct H; subst; try discriminate
               end).
   { (* Reflexivity case *)
     exists (typ_top), (typ_or typ_nocap
                          (typ_rcd (dec_typ (typ_label_refcap
                                               typ_refcap_mutability)
                                            T1 T2))). eauto. }
   { (* typ_approx_typ case *)
     exists (typ_top), (typ_or typ_nocap
                          (typ_rcd (dec_typ (typ_label_refcap
                                               typ_refcap_mutability)
                                            T1' T2'0))). eauto. }

 - (* subtyp_nocap_nm_s *)
   inversions Hap1. inversions Hap2. eauto 8.
   all: repeat(match goal with
               | [ H : _ = _ \/ _ |- _ ] => destruct H; subst; try discriminate
               end).
   inversions H1. inversions H3.
   all: repeat(match goal with
               | [ H : _ = _ \/ _ |- _ ] => destruct H; subst; try discriminate
               end).
   inversions H5.
   all: repeat(match goal with
               | [ H : _ = _ \/ _ |- _ ] => destruct H; subst; try discriminate
               end).
   { (* Reflexivity case *)
     exists (typ_top), (typ_or typ_nocap
                          (typ_rcd (dec_typ (typ_label_refcap
                                               typ_refcap_mutability)
                                            typ_bot typ_bot))). eauto. }
   { (* typ_approx_typ case *)

     eapply typ_approx_bot_open_inv in H9.
     eapply typ_approx_bot_open_inv in H10.
     subst.
     exists (typ_top), (typ_or typ_nocap
                          (typ_rcd (dec_typ (typ_label_refcap
                                               typ_refcap_mutability)
                                             typ_bot typ_bot))).
    splits; eauto. }
 -
exists (typ_and typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))  ) (typ_bot).
     eapply typ_approx_bot_inv in Hap2. subst. inversions_typ Hap1.  splits; eauto.

     eapply typ_approx_nocap_inv in H3. subst. inversions_typ H5. eauto.
  eapply typ_approx_bot_open_inv in H9.
     eapply typ_approx_bot_open_inv in H10. subst. eauto.
 - eapply typ_approx_top_inv in Hap1. subst.
  exists typ_top (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_top) ).
 inversions_typ Hap2. eauto.

     eapply typ_approx_bot_open_inv in H8.
     eapply typ_approx_top_open_inv in H9.
  subst; eauto.
 - inversions_typ Hap1. inversions_typ Hap2. eauto 8.
 - inversions_typ Hap1. inversions_typ Hap2. eauto 8.
Qed.

(* Lemma 5.99 ToSRed *)
Lemma typ_approx_tight_to_s δ G e T1 T2:
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  G⁏e ⊢# T1 <: T2 ->
  exists T3 T4,
  G ⊢ T1 ↝ˢ⦇δ⦈ T3 /\
  G ⊢ T2 ↝ˢ⦇δ⦈ T4 /\
  G⁏e ⊢ˢ⦇δ⦈ T3 <: T4.
Proof.
  introv Hin Hsl Hoke Hcor Hsub12.
  revert δ.
  dependent induction Hsub12; intros; eauto.
  - exists T, typ_top. eauto.
  - exists typ_bot, T. eauto.
  - exists T, T. eauto.
  - (* trans *)
    rename S into T1 (*1*). rename T into T2 (*5*). rename U into T3 (*2*).
    destruct (IHHsub12_1 Hin Hsl Hoke Hcor δ) as [T1s (*7*) [T21s (*8*) [Hap1 [Hap21 Hsub21]]]].
    destruct (IHHsub12_2 Hin Hsl Hoke Hcor δ) as [T23s (*9*) [T3s (*10*) [Hap23 [Hap3 Hsub23]]]].
    destruct (typ_approx_s_com Hin Hsl Hap21 Hap23) as [T2s (*11*) [Hap12 Hap32]].
    destruct δ.
    + destruct (typ_approx_s_subcom Hin Hsl Hoke Hcor Hsub23 Hap32 (typ_approx_refl G T3s ⊕)) as [T2ss [T3ss [Hap2s [Hap3s Hsub23s]]]].
      exists T1s, T3ss.
      repeat split; eauto using typ_approx_s_trans.
      pose proof (typ_approx_s_sub e Hsl Hap12) as Hsub12. simpl in Hsub12.
      pose proof (typ_approx_s_sub e Hsl Hap2s) as Hsub2s. simpl in Hsub2s.
      eapply subtyp_trans_res; eauto.
    + destruct (typ_approx_s_subcom Hin Hsl Hoke Hcor Hsub21 (typ_approx_refl G T1s ⊝) Hap12) as [T1ss [T2ss [Hap1s [Hap2s Hsub21s]]]].
      exists T1ss, T3s.
      repeat split; eauto using typ_approx_s_trans.
      pose proof (typ_approx_s_sub e Hsl Hap32) as Hsub32. simpl in Hsub32.
      pose proof (typ_approx_s_sub e Hsl Hap2s) as Hsub2s. simpl in Hsub2s.
      eapply subtyp_trans_res; eauto.
  - (* and11 *)
    exists (typ_and T U), T. eauto.
  - (* and12 *)
    exists (typ_and T U), U. eauto.
  - (* and2 *)
    rename S into T1 (*1*). rename T into T2 (*5*). rename U into T3 (*6*).
    autospecialize.
    destruct (IHHsub12_1) as [T12s (*7*) [T2s (*8*) [Hap12 [Hap2 Hsub21]]]].
    destruct (IHHsub12_2) as [T13s (*9*) [T3s (*10*) [Hap13 [Hap3 Hsub31]]]].
    destruct δ.
    + exists T1 (typ_and T2s T3s).
      repeat split; eauto.
      pose proof (typ_approx_s_sub e Hsl Hap12) as Hsub12. simpl in Hsub12.
      pose proof (typ_approx_s_sub e Hsl Hap13) as Hsub13. simpl in Hsub13.
      eapply subtyp_and2_res; eauto.
    + destruct (typ_approx_s_com Hin Hsl Hap12 Hap13) as [T1s [Hap21 Hap31]].
      pose proof (typ_approx_s_sub e Hsl Hap21) as Hsub12. simpl in Hsub12.
      pose proof (typ_approx_s_sub e Hsl Hap31) as Hsub13. simpl in Hsub13.
      exists T1s (typ_and T2s T3s).
      repeat split; eauto using typ_approx_s_trans.
      eapply subtyp_and2_res; eauto.
  - (* or11 *)
    exists T, (typ_or T U). eauto.
  - (* or12 *)
    exists U, (typ_or T U). eauto.
  - (* or2 *)
    rename S into T2. rename T into T1. rename U into T3.
    destruct (IHHsub12_1 Hin Hsl Hoke Hcor δ) as [T2s [T12s [Hap2 [Hap12 Hsub12]]]].
    destruct (IHHsub12_2 Hin Hsl Hoke Hcor δ) as [T3s [T13s [Hap3 [Hap13 Hsub13]]]].
    destruct δ.
    + destruct (typ_approx_s_com Hin Hsl Hap12 Hap13) as [T1s [Hap21 Hap31]].
      pose proof (typ_approx_s_sub e Hsl Hap21) as Hsub21. simpl in Hsub21.
      pose proof (typ_approx_s_sub e Hsl Hap31) as Hsub31. simpl in Hsub31.
      exists (typ_or T2s T3s) T1s.
      repeat split; eauto using typ_approx_s_trans.
      eapply subtyp_or2_res; eauto.
    + exists (typ_or T2s T3s) T1.
      repeat split; eauto.
      pose proof (typ_approx_s_sub e Hsl Hap12) as Hsub21. simpl in Hsub21.
      pose proof (typ_approx_s_sub e Hsl Hap13) as Hsub31. simpl in Hsub31.
      eapply subtyp_or2_res; eauto.
  - (* fld *)
    destruct (IHHsub12_1 Hin Hsl Hoke Hcor (─δ)) as [T1s [T2s [HapT1 [HapT2 HsubT]]]].
    destruct (IHHsub12_2 Hin Hsl Hoke Hcor δ) as [U1s [U2s [HapU1 [HapU2 HsubU]]]].
    exists (typ_rcd (dec_fld a T2s U1s) ), (typ_rcd (dec_fld a T1s U2s) ).
    repeat split;
    eauto.
  - (* typ *)
    pick_fresh z.
    destruct (H0 z Hin Hsl Hoke  Hcor (─δ)) as [T1s [T2s [HapT1 [HapT2 HsubT]]]].
    destruct (H2 z Hin Hsl Hoke Hcor δ) as [U1s [U2s [HapU1 [HapU2 HsubU]]]].
    exists (typ_rcd (dec_typ A (close z T2s) (close z U1s)) ), (typ_rcd (dec_typ A (close z T1s) (close z U2s)) ).
    repeat split;
    eauto.
      eapply subtyp_typ_res; eauto.
      intros.
      erewrite typ_approx_reopen_subst; eauto. erewrite typ_approx_reopen_subst; eauto.
      eapply subtyp_res_usubst. eauto. eauto. eauto. eauto. notin_solve. eauto. notin_solve. notin_solve.
      intros.
      erewrite typ_approx_reopen_subst; eauto.
      erewrite typ_approx_reopen_subst; eauto.
      eapply subtyp_res_usubst; eauto. notin_solve. notin_solve.
  - destruct δ.
    + exists (open y TL), (open y TU). repeat split; eauto.
      destruct (pf_dec_typ_inv Hin H); subst; eauto.
    + exists (open y TL), (open y TL). repeat split; eauto.
  - destruct δ.
    + exists (open y TU), (open y TU). repeat split; eauto.
    + exists (open y TL), (open y TU). repeat split; eauto.
      destruct (pf_dec_typ_inv Hin H); subst; eauto.
  - exists (typ_fun S1 T1), (typ_fun S2 T2). eauto.
  - exists (typ_rcd (dec_met m TC1 TZ1 TR1)), (typ_rcd (dec_met m TC2 TZ2 TR2)). eauto.
  - exists T1, T2. eauto.
  - exists (typ_and (typ_rcd (dec_typ A S1 T1)) (typ_rcd (dec_typ A S2 T2))), (typ_rcd (dec_typ A (typ_or S1 S2) (typ_and T1 T2))). eauto.
  - exists (typ_and T1 (typ_or T2 T3) ), ( typ_or (typ_and T1 T2) (typ_and T1 T3)).  eauto.
  - exists (typ_nocap), (typ_bnd T). eauto.
  - exists (typ_nocap), (typ_rcd (dec_typ (typ_label_normal A) T1 T2)). eauto.
  - exists (typ_nocap), (typ_rcd (dec_fld a T1 T2)). eauto.
  - pose proof subtyp_nocap_fun_res.
    destruct (mode_has_fun' typing_mode); try discriminate.
    exists (typ_nocap), (typ_fun TZ TR). splits; eauto.
  - exists (typ_top),
      (typ_or typ_nocap
              (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability)
                                T1 T2))).
    repeat(split); eauto 5.
  - exists (typ_top),
      (typ_or typ_nocap
              (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability)
                                 typ_bot typ_bot))).
    repeat(split); eauto 5.
  - exists ( typ_and typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) ) typ_bot.
    repeat(split); eauto 5.
  - exists typ_top ( (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_top)) ) .
    repeat(split); eauto 5.
    (* Because of the new ⊤ <: (N | M{T₁..T₂}) rule. Need to have the
       corresponding restricted tight subtyping rule. Namely,
       mode_top_nm' typing_mode = mode_top_nm_subtype ->
       G ⁏ e ⊢ ˢ ⦇ δ ⦈ typ_top <:
       typ_or typ_nocap
              (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability)
                                T1 T2))
     *)
Qed.

(* Lemma 5.105 ERedCom *)
Lemma typ_approx_e_subcom G e δ T1 T2 T3 y:
  ctx_ok G ->
  T1 ↝ᵉ⦇δ⦈ T2 ->
  T1 ↝ᵉ⦇δ⦈ T3 ->
  nosel T2 y ->
  nosel T3 y ->
  exists T4,
  G⁏e ⊢⦇δ⦈ T2 <:> T4 /\
  G⁏e ⊢⦇δ⦈ T3 <:> T4 /\
  nosel T4 y /\
  T1 ↝ᵉ⦇δ⦈ T4.
Proof.
  (* Note: The proof is almost exactly the same as typ_approx_e_com,
     except in sel cases, the types are different *)
  introv Hok  Hap12 Hap13 Hns2 Hns3.
  revert dependent T3.
  dependent induction Hap12; intros.
  - destruct H; subst.
    +
      inversions_tyq Hns2.
      inversions_typ Hap13; eauto.
      inversions_tyq Hns3.
      autospecialize.
      destruct (IHHap12_1) as [T5 [H15 [H105 [Hns5 Hap5]]]].
      destruct (IHHap12_2) as [T6 [H16 [H106 [Hns6 Hap6]]]].
      exists (typ_and T5 T6).
      repeat split; eauto; destruct δ; eapply (@subtyp_and _ (and_or_subtyp _)); eauto.
    +
      inversions_tyq Hns2.
      inversions_typ Hap13; eauto.
      inversions_tyq Hns3.
      autospecialize.
      destruct (IHHap12_1) as [T5 [H15 [H105 [Hns5 Hap5]]]].
      destruct (IHHap12_2) as [T6 [H16 [H106 [Hns6 Hap6]]]].
      exists (typ_or T5 T6).
      repeat split; eauto; destruct δ; eapply (@subtyp_or _ (and_or_subtyp _)); eauto.
  - inversions_tyq Hns2.
    inversions_typ Hap13; eauto.
    inversions_tyq Hns3.
    autospecialize.
    destruct δ.
    +
      destruct (IHHap12_1) as [T5 [H15 [H105 [Hns5 Hap5]]]].
      destruct (IHHap12_2) as [T6 [H16 [H106 [Hns6 Hap6]]]].
      all: exists (typ_rcd (dec_fld a T5 T6)).
      repeat split; eauto. apply (nosel_bin (typ_binary_rcd_fld a)); eauto.
    +
      destruct (IHHap12_1) as [T5 [H15 [H105 [Hns5 Hap5]]]].
      destruct (IHHap12_2) as [T6 [H16 [H106 [Hns6 Hap6]]]].
      all: exists (typ_rcd (dec_fld a T5 T6)).
      repeat split; eauto. apply (nosel_bin (typ_binary_rcd_fld a)); eauto.
  - inversions_tyq Hns2.
    inversions_typ Hap13; eauto.
    inversions_tyq Hns3.
    destruct δ.
    +
      pick_fresh z.
      clear H H1.
      forwards T5 H15 H105 Hns5 Hap5: H0; eauto.
      eapply (nosel_open) with (x:=z); eauto.
      eapply (nosel_open) with (x:=z); eauto.
      forwards T6 H16 H106 Hns6 Hap6: H2; eauto.
      eapply (nosel_open) with (x:=z); eauto.
      eapply (nosel_open) with (x:=z); eauto.
      exists (typ_rcd (dec_typ B (close z T5) (close z T6))).

      repeat split; eauto.
      constructor; intro.
      erewrite typ_approx_reopen_subst; eauto.
      rewrite (subst_intro) with (x0:=z) by eauto.
      eapply usubst_end1; eauto.
      all: eauto.

      erewrite typ_approx_reopen_subst; eauto.
      rewrite (subst_intro) with (x0:=z).
      eapply usubst_end1; eauto.
      all: eauto.

      specialize (H12 z).
      specialize (H13 z).
      eapply subtyp_typ; intros.
      erewrite typ_approx_reopen_subst; eauto. rewrite (subst_intro) with (x0:=z).
      eapply usubst_end1; eauto.
      all: eauto.
      erewrite typ_approx_reopen_subst; eauto. rewrite (subst_intro) with (x0:=z).
      eapply usubst_end1 ; eauto.
      all: eauto 6.
      apply (nosel_bin (typ_binary_rcd_typ B)); eauto.
    +
      pick_fresh z.
      clear H H1.
      forwards T5 H15 H105 Hns5 Hap5: H0; eauto.
      eapply (nosel_open) with (x:=z); eauto.
      eapply (nosel_open) with (x:=z); eauto.
      forwards T6 H16 H106 Hns6 Hap6: H2; eauto.
      eapply (nosel_open) with (x:=z); eauto.
      eapply (nosel_open) with (x:=z); eauto.
      exists (typ_rcd (dec_typ B (close z T5) (close z T6))).

      repeat split; eauto.
      constructor; intro.
      erewrite typ_approx_reopen_subst; eauto. rewrite (subst_intro) with (x0:=z).
      eapply usubst_end1; eauto.
      all: eauto.


      erewrite typ_approx_reopen_subst; eauto. rewrite (subst_intro) with (x0:=z).
      eapply usubst_end1; eauto.
      all: eauto.

      specialize (H12 z).
      specialize (H13 z).
  eapply subtyp_typ; intros.
 erewrite typ_approx_reopen_subst; eauto. rewrite (subst_intro) with (x0:=z).
     eapply usubst_end1; eauto.
      all: eauto.

      erewrite typ_approx_reopen_subst; eauto. rewrite (subst_intro) with (x0:=z).
      eapply usubst_end1; eauto.
      all: eauto.
    apply (nosel_bin (typ_binary_rcd_typ B)); eauto.
    eauto 6.

  - inversions_typ Hap13; eauto.  exists typ_top. repeat split; eauto; destruct δ; eauto.
  - inversions_typ Hap13; eauto.  eexists. repeat split; eauto; destruct δ; eauto.
  - inversions_typ Hap13; eauto.  exists typ_nocap. repeat split; eauto; destruct δ; eauto.
  - inversions_typ Hap13; eauto. destruct δ.
    + exists typ_top. repeat split; eauto. constructor; eauto.
    + exists typ_nocap. repeat split; eauto. constructor; eauto.
  - pose proof subtyp_nocap_fun_res.
    destruct (mode_has_fun' typing_mode) eqn:Htm.
    (* Because of update to typ_approx_me_fun *)
    + eexists. repeat(split); eauto. destruct δ; eauto.
      * inversions Hap13. inversions H2; subst; try discriminate.
        destruct (mode_has_fun' typing_mode); try discriminate; eauto.
      * inversions Hap13. inversions H2; subst; try discriminate.
        destruct (mode_has_fun' typing_mode); try discriminate; eauto.
      * destruct δ; eauto.
    + eexists. repeat(split); eauto. destruct δ; eauto.
      * inversions Hap13. inversions H2; subst; try discriminate.
        destruct (mode_has_fun' typing_mode); try discriminate; eauto.
      * inversions Hap13. inversions H2; subst; try discriminate.
        destruct (mode_has_fun' typing_mode); try discriminate; eauto.
      * destruct δ; eauto.
  - inversions_typ Hap13; eauto; destruct δ; exists (typ_sel v B v2); eauto.
  - inversions_typ Hap13; eauto.
    destruct δ; exists (typ_bnd T); eauto.
  - inversions_typ Hap13; eauto.
    + destruct δ; exists (typ_sel v B v2); eauto.
    + destruct δ.
      ++ exists (typ_bot). eauto.
      ++ exists (typ_top). eauto.
Qed.

(* Lemma 5.107 TRedCom *)
Lemma typ_approx_m_subcom δ G e T1 T2 T3:
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  G⁏e ⊢ˢ⦇δ⦈ T1 <:> T2 ->
  T1 ↝ᵐ⦇δ⦈ T3 ->
  exists T4,
  G⁏e ⊢ᵐ⦇δ⦈ T3 <:> T4 /\
  T2 ↝ᵐ⦇δ⦈ T4.
Proof.
  introv Hin Hsl Hoke Hcor Hsub12.
  revert T3.
  (* We need both directions in the induction so we cannot destruct δ, but the subtyping must be in normal form *)
  rename T1 into T1p. rename T2 into T2p. rename Hsub12 into Hsub12p.
  destruct (subtyp_res_polar_inv G e ˢ δ T1p T2p Hsub12p) as [T1 [T2 [Hsub12 Hp]]].

  assert ( match δ with
  | ⊕ =>
     (forall T3 : typ, T1 ↝ ᵐ ⦇ ⊕ ⦈ T3 -> exists T4, G ⁏ e ⊢ ᵐ ⦇ ⊕ ⦈ T3 <: T4 /\ T2 ↝ ᵐ ⦇ ⊕ ⦈ T4)
  | ⊝ =>
     (forall T3 : typ, T2 ↝ ᵐ ⦇ ⊝ ⦈ T3 -> exists T4, G ⁏ e ⊢ ᵐ ⦇ ⊝ ⦈ T4 <: T3 /\ T1 ↝ ᵐ ⦇ ⊝ ⦈ T4)
  end).
  clear dependent T1p.
  clear dependent T2p.

  dependent induction Hsub12; try destruct δ; intros.
  - exists typ_top. split; eauto.
  - inversions_typ H0.
    destruct (typ_approx_m_exists ⊝ T) as [T4 ?].
    eauto.
  - inversions_typ H0.
    destruct (typ_approx_m_exists ⊕ T) as [T4 ?].
    eauto.
  - exists typ_bot. split; eauto.
  - eauto.
  - eauto.
  - destruct (typ_approx_tyq Hin Hsl Hcor H0 (typ_approx_m_ctx F H1)) as [T4 [? ?]].
    exists T4. split; eauto using typ_approx_m_ctx.
  - symmetry in H0.
    destruct (typ_approx_tyq Hin Hsl Hcor H0 (typ_approx_m_ctx F H1)) as [T4 [? ?]].
    symmetry in H3.
    exists T4. split; eauto using typ_approx_m_ctx.
  - (* trans *)
    autospecialize_prop.
    destruct (IHHsub12_1 T0) as [T4 [? ?]]. auto.
    destruct (IHHsub12_2 T4) as [T5 [? ?]]. auto.
    eauto.
  - (* trans *)
    autospecialize_prop.
    destruct (IHHsub12_2 T0) as [T4 [? ?]]. auto.
    destruct (IHHsub12_1 T4) as [T5 [? ?]]. auto.
    eauto.
  - destruct (typ_approx_m_exists ⊕ T2) as [T4 ?].
    exists (typ_or T3 T4).
    split; eauto.
  - inversions_typ H0. eauto.
  - destruct (typ_approx_m_exists ⊕ T1) as [T4 ?].
    exists (typ_or T4 T3).
    split; eauto.
  - inversions_typ H0. eauto.
  - (* or2 *)
    autospecialize_prop.
    inversions_typ H0.
    destruct (IHHsub12_1 T1') as [T4 [? ?]]. auto.
    destruct (IHHsub12_2 T2') as [T5 [? ?]]. auto.
    assert (T4 = T5) by (eapply typ_approx_m_eq; eauto).
    subst.
    exists T5.
    split; eauto.
  - (* or2 *)
    autospecialize_prop.
    destruct (IHHsub12_1 T0) as [T4 [? ?]]. auto.
    destruct (IHHsub12_2 T0) as [T5 [? ?]]. auto.
    exists (typ_or T4 T5).
    split; eauto.
  - inversions_typ H0. eauto.
  - destruct (typ_approx_m_exists ⊝ T2) as [T4 ?].
    exists (typ_and T3 T4).
    split; eauto.
  - inversions_typ H0. eauto.
  - destruct (typ_approx_m_exists ⊝ T1) as [T4 ?].
    exists (typ_and T4 T3).
    split; eauto.
  - (* and2 *)
    autospecialize_prop.
    destruct (IHHsub12_1 T0) as [T4 [? ?]]. auto.
    destruct (IHHsub12_2 T0) as [T5 [? ?]]. auto.
    exists (typ_and T4 T5).
    split; eauto.
  - (* and2 *)
    autospecialize_prop.
    inversions_typ H0.
    destruct (IHHsub12_1 T1') as [T4 [? ?]]. auto.
    destruct (IHHsub12_2 T2') as [T5 [? ?]]. auto.
    assert (T4 = T5) by (eapply typ_approx_m_eq; eauto).
    subst.
    exists T5.
    split; eauto.
  - inversions_typ H0. exists (typ_bnd T). eauto.
  - inversions_typ H0. exists (typ_nocap). eauto.
  - inversions_typ H0.
    destruct (typ_approx_m_exists ⊝ T1) as [T4 ?].
    destruct (typ_approx_m_exists ⊕ T2) as [T5 ?].
    exists (typ_rcd (dec_fld a T4 T5)). split; eauto.
  - inversions_typ H0. exists (typ_nocap). split; eauto.
  - inversions_typ H0.
    pick_fresh z.
    destruct (typ_approx_m_exists ⊝ (open z T1)) as [T4 ?].
    destruct (typ_approx_m_exists ⊕ (open z T2)) as [T5 ?].
    exists (typ_rcd (dec_typ (typ_label_normal A) (close z T4) (close z T5))).
    split; eauto 8.
  - inversions_typ H0. exists (typ_nocap). split; eauto.
  - (* Because of update to typ_approx_me_fun *)
    pose proof subtyp_nocap_fun_res as Hnfr.
    pose proof typ_approx_me_fun as Hamf.
    destruct (mode_has_fun' typing_mode) eqn:Htm.
    + inversions H0. destruct H2; subst; try discriminate.
      destruct (typ_approx_m_exists ⊕ (typ_fun TZ TR)).
      exists typ_top. split; eauto. apply Hamf with (δ := ⊕). auto.
    + inversions H0. destruct H2; subst; try discriminate.
      destruct (typ_approx_m_exists ⊕ (typ_fun TZ TR)).
      exists typ_top. split; eauto. apply Hamf with (δ := ⊕). auto.
  - (* Because of update to typ_approx_me_fun *)
    pose proof subtyp_nocap_fun_res as Hnfr.
    pose proof typ_approx_me_fun as Hamf.
    destruct (mode_has_fun' typing_mode) eqn:Htm.
    + inversions H0. destruct H2; subst; try discriminate.
      destruct (typ_approx_m_exists ⊕ (typ_fun TZ TR)).
      destruct (mode_has_fun' typing_mode); try discriminate.
      exists typ_nocap; split; eauto.
    + destruct (typ_approx_m_exists ⊕ (typ_fun TZ TR)).
      destruct (mode_has_fun' typing_mode); try discriminate.
      exists typ_bot; split; eauto.
  - inversions_typ H0.
    autospecialize_prop.
    destruct (IHHsub12_1 T1'0) as [T4 [? ?]]. auto.
    destruct (IHHsub12_2 T2'0) as [T5 [? ?]]. auto.
    exists (typ_rcd (dec_fld a T4 T5)). split; eauto.
  - inversions_typ H0. autospecialize_prop.
    destruct (IHHsub12_1 T1'0) as [T4 [? ?]]. auto.
    destruct (IHHsub12_2 T2'0) as [T5 [? ?]]. auto.
    exists (typ_rcd (dec_fld a T4 T5)). split; eauto.
  - inversions_typ H4.
    pick_fresh z.
    forwards T4 ? ?: (H1); try apply H12; eauto.
    forwards T5 ? ?: (H3); try apply H12; eauto.
    exists (typ_rcd (dec_typ A (close z T4) (close z T5))).
    split; eauto. eapply subtyp_typ_res; auto.
    intro. specialize (H12 x). specialize (H13 x).
     erewrite typ_approx_reopen_subst; eauto using typ_approx_m_ctx.
    rewrite subst_intro with (x0:=z).
    eapply subtyp_res_usubst; eauto. notin_solve. notin_solve.
    intro. specialize (H12 x). specialize (H13 x).
     erewrite typ_approx_reopen_subst; eauto using typ_approx_m_ctx.
    rewrite subst_intro with (x0:=z).
    eapply subtyp_res_usubst; eauto. notin_solve. notin_solve.
    eauto 6.
  - inversions_typ H4. pick_fresh z.
    forwards [T4 [? ?]]: (H1); eauto.
    forwards [T5 [? ?]]: (H3); eauto.
    exists (typ_rcd (dec_typ A (close z T4) (close z T5))).
    split; eauto.
     eapply subtyp_typ_res; eauto.
    intro. specialize (H12 x); specialize (H13 x).
     erewrite typ_approx_reopen_subst; eauto using typ_approx_m_ctx.
    rewrite subst_intro with (x0:=z).
    eapply subtyp_res_usubst; eauto. notin_solve. notin_solve.
    intro. specialize (H12 x); specialize (H13 x).
     erewrite typ_approx_reopen_subst; eauto using typ_approx_m_ctx.
    rewrite subst_intro with (x0:=z).
    eapply subtyp_res_usubst; eauto. notin_solve. notin_solve.
    eauto 6.
  - (* typ_and *)
    inversions_typ H0.
    inversions_typ H3.
    inversions_typ H5.
    exists (typ_rcd (dec_typ A (typ_or T1'0 T1') (typ_and  T2'0 T2'1))).
    split; eauto 6.
    constructor; eauto; intros.
    specialize (H8 x).
    specialize (H9 x).
    specialize (H10 x).
    specialize (H11 x).
    all: constructor; eauto.
  - inversions_typ H0.
    pick_fresh z.
    specialize (H8 z).
    specialize (H9 z).
    destruct T1'; sympls; inversions_typ H8.
    destruct T2'; sympls; inversions_typ H9.
    exists (typ_and (typ_rcd (dec_typ A T1'1 T2'1)) (typ_rcd (dec_typ A T1'2 T2'2))).
    split; eauto 6.
    constructor; eauto; constructor; eauto; intros.
    all: eapply typ_approx_reopen; eauto; sympls; notin_solve.

  - inversions_typ H0.
    inversions_typ H5.
    eexists; split; eauto 7.
  - inversions_typ H0.
    inversions_typ H3.
    inversions_typ H5.
    assert (T1' = T1'0) by (eapply typ_approx_m_eq; eauto).
    subst.
    eexists; split; eauto.
  - inversions_typ H0.
    destruct (typ_approx_m_exists ⊕ (open v2 T2)) as [T4 ?].
    exists T4; eauto.
  - inversions_typ H0.
    destruct (typ_approx_m_exists ⊝ (open v2 T1)) as [T4 ?].
    exists T4; eauto.
  - (* Because of update to typ_approx_me_fun *)
    pose proof typ_approx_me_fun.
    destruct (mode_has_fun' typing_mode); try discriminate.
    exists typ_top. split; try constructor; eauto.
    eapply H3 with (δ := ⊕). auto.
  - (* Because of update to typ_approx_me_fun *)
    pose proof typ_approx_me_fun.
    destruct (mode_has_fun' typing_mode) eqn:Htm; try discriminate.
    exists typ_nocap. split.
    + inversions H2; auto. destruct H5; subst; try discriminate.
      destruct (mode_has_fun' typing_mode); try discriminate.
      auto.
    + eapply H3 with (δ := ⊝). auto.
  - (* Because of subtyp_nocap_nm_s *)
    inversions H1. destruct H3; subst; try discriminate.
    (* Goal. (N | M{T₁..T₂}) ↝ᵐ (N | M{T₁..T₂})
       Solution. Add reflexivity for (N | M{T₁..T₂}) in ↝ᵐ
       (typ_approx) since (N | M{T₁..T₂}) is an extreme with the new
       ⊤ <: (N | M{T₁..T₂}) rule. *)
    
    forwards HT1: typ_approx_m_exists T1.
    destructc HT1 as T1' HT1'.
    forwards HT2: typ_approx_m_exists T2.
    destructc HT2 as T2' HT2'.

    exists (typ_or typ_nocap
              (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability)
                                T1' T2'))).
    split; eauto.

    eapply typ_approx_andor; eauto.
    eapply typ_approx_typ; intros.
    eapply typ_approx_open; eauto.
    eapply typ_approx_open; eauto.

  - exists typ_top. split; try solve[ constructor; eauto ].
    (* Goal. G⁏e ⊢ᵐ⦇⊝⦈ ⊤ <: T₃ where (N | M{T₁..T₂}) ↝ᵐ T₃.
       Solution. Basically add this rule to restricted subtyping? *)
    (* Inverting approximation of RHS *)
    inversions H1. destruct H3; subst; inversions H2.
    (* We have N ↝ᵐ⦇⊝⦈ T₁' approximating N down to ⊥ in (♠) means T₁' = ⊥
       (◆) *)
    inversions H4. destruct H2; subst; inversions H1.
    inversions H6. destruct H3; subst; inversions H2.

    auto.
  - eapply typ_approx_top_inv in H1. subst.
    exists (typ_or typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))). splits;eauto. eapply typ_approx_andor; eauto. eapply typ_approx_typ; eauto.
  - exists typ_top. inversions_typ H1. eapply typ_approx_nocap_inv in H4. subst. inversions_typ H6.
    eapply typ_approx_bot_open_inv in H9.
    eapply typ_approx_bot_open_inv in H10. subst.
    splits; eauto.
  - exists typ_bot. inversions_typ H1. eapply typ_approx_nocap_inv in H4. subst. inversions_typ H6.
    eapply typ_approx_bot_open_inv in H9.
    eapply typ_approx_bot_open_inv in H10. subst.
    splits; eauto.
  - eapply typ_approx_bot_inv in H1. subst.
    exists (typ_and typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) ). splits; eauto. eapply typ_approx_andor; eauto. eapply typ_approx_typ; eauto.
  -  eapply typ_approx_top_inv in H1. subst.
    exists ((typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_top))). splits;eauto. eapply typ_approx_typ; eauto.
  - exists typ_top. splits; eauto. 
inversions_typ H1. 
    eapply typ_approx_bot_open_inv in H9.
    eapply typ_approx_top_open_inv in H10. subst. eauto.
  - exists typ_top. split; try constructor; eauto.
  - exists typ_nocap. split; try constructor; eauto.
    inversions H3; auto. destruct H5; subst; try discriminate.
  - exists typ_top. split; try constructor; eauto.
  - exists typ_nocap. split; try constructor; eauto.
    inversions H; auto. destruct H1; subst; try discriminate.
  - (* Now destruct δ and prove the lemma *)
  intros.
  destruct δ; destruct Hp; subst; eauto.
Qed.

(* Lemma 5.108 ToTRed *)
Lemma typ_approx_s_to_m δ G e T1 T2:
  inert G ->
  sealed G ->
ok e->
  renv_corr G e ->
  G⁏e ⊢ˢ⦇δ⦈ T1 <: T2 ->
  exists T3 T4,
  T1 ↝ᵐ⦇δ⦈ T3 /\
  T2 ↝ᵐ⦇δ⦈ T4 /\
  G⁏e ⊢ᵐ⦇δ⦈ T3 <: T4.
Proof.
  introv Hin Hsl Hoke Hcor Hsub12.
  destruct δ.
  -
    destruct (typ_approx_m_exists ⊕ T1) as [T3 Hap13].
    epose proof (typ_approx_m_subcom T2 Hin Hsl  Hoke Hcor ?[C] Hap13).
    destruct H as [T4 [H34 Hap24]].
    exists T3, T4. eauto.
    Unshelve. eauto.
  -
    destruct (typ_approx_m_exists ⊝ T2) as [T4 Hap24].
    epose proof (typ_approx_m_subcom T1 Hin Hsl Hoke Hcor ?[C] Hap24).
    destruct H as [T3 [H34 Hap13]].
    exists T3, T4. eauto.
    Unshelve. eauto.
Qed.

(* Lemma 5.114 TRedCut *)
Lemma typ_approx_m_cut δ G e T1 T2 y:
  ctx_ok G ->
  nosel T1 y ->
  T1 ↝ᵐ⦇δ⦈ T2 ->
  nosel T2 y /\
  G⁏e ⊢⦇δ⦈ T1 <:> T2 /\
  T2 ↝ᵉ⦇δ⦈ T2.
Proof.
  introv Hok Hns Hap12.
  dependent induction Hap12; intros; try solve [try destruct δ; eauto].
  - (* and, or *)
    destruct H; subst.
    + inversions_tyq Hns.
      autospecialize. autodestruct.
      repeat split; eauto.
      destruct δ; eapply (@subtyp_and _ (and_or_subtyp _)); eauto.
    + inversions_tyq Hns.
      autospecialize. autodestruct.
      repeat split; eauto.
      destruct δ; eapply (@subtyp_or _ (and_or_subtyp _)); eauto.
  - inversions_tyq Hns.
    destruct δ; autospecialize; autodestruct; repeat split; eauto.
    all: apply (nosel_bin (typ_binary_rcd_fld a)); eauto.
  - inversions_tyq Hns.

    pick_fresh z.
    specialize (H z).
    specialize (H0 z).
    specialize (H1 z).
    specialize (H2 z).

    forwards Hns1 Hsub1 Hap1: H0; eauto using nosel_open.
    forwards Hns2 Hsub2 Hap2: H2; eauto using nosel_open.

    eapply nosel_open_inv in Hns1.
    eapply nosel_open_inv in Hns2.

    destruct δ; splits.

    all: try apply (nosel_bin (typ_binary_rcd_typ B)); eauto.

    eapply subtyp_typ; intro.
    rewrite subst_intro with (X:=T1') (x0:=z) by notin_solve.
    rewrite subst_intro with(X:=T1)  (x0:=z) by notin_solve.
    eapply usubst_end1; eauto.


    rewrite subst_intro with (X:=T2') (x0:=z) by notin_solve.
    rewrite subst_intro with(X:=T2)  (x0:=z) by notin_solve.
    eapply usubst_end1; eauto.

   constructor; intros; eapply typ_approx_reopen; eauto; sympls; notin_solve.

  eapply subtyp_typ; intro.
    rewrite subst_intro with (X:=T1') (x0:=z) by notin_solve.
    rewrite subst_intro with(X:=T1)  (x0:=z) by notin_solve.
    eapply usubst_end1; eauto.

    rewrite subst_intro with (X:=T2') (x0:=z) by notin_solve.
    rewrite subst_intro with(X:=T2)  (x0:=z) by notin_solve.
    eapply usubst_end1; eauto.

    constructor; intros; eapply typ_approx_reopen; eauto; sympls; notin_solve.
  - (* Because of update to typ_approx_me_fun *)
    pose proof subtyp_nocap_fun_res.
    destruct (mode_has_fun' typing_mode) eqn:Htm.
    all: repeat split; destruct δ; auto.
Qed.

Lemma nosel_or y T1 T2:
  nosel T1 y ->
  nosel T2 y ->
  nosel (typ_or T1 T2) y.
Proof.
  eapply nosel_bin with (typ_bin:=typ_or).
  eauto.
Qed.
Lemma nosel_and y T1 T2:
  nosel T1 y ->
  nosel T2 y ->
  nosel (typ_and T1 T2) y.
Proof.
  eapply nosel_bin with (typ_bin:=typ_and).
  eauto.
Qed.
Lemma nosel_typ y A T1 T2:
  nosel T1 y ->
  nosel T2 y ->
  nosel (typ_rcd (dec_typ A T1 T2)) y.
Proof.
  eapply nosel_bin with (typ_bin:=(typ_binary_rcd_typ_ctor A)).
  eauto.
Qed.
Hint Resolve nosel_or nosel_and nosel_typ : core.


(* Lemma 5.115 TSubCut *)
Lemma typ_approx_e_cut δ G e T1 T2 T3 y vk T:
  inert (G § y / vk ~ T) ->
  sealed (G § y / vk ~ T) ->
  renv_corr (G § y / vk ~ T) e ->
  y \notin ctx_fv G ->
  y # e ->
  renv_notin y e ->
  (G § y / vk ~ T)⁏e ⊢ᵐ⦇δ⦈ T1 <:> T2 ->
  T1 ↝ᵉ⦇δ⦈ T3 ->
  nosel T3 y ->
  exists T4,
  G⁏e ⊢⦇δ⦈ T3 <:> T4 /\
  T2 ↝ᵉ⦇δ⦈ T4 /\
  nosel T4 y.
Proof.
  (* This proof uses  typ_approx_reopen_closed'  in eauto *)

  introv Hin Hsl Hcor Hn Hned Hnef.
  introv Hsub12 Hap13 Hns3.
  assert (inert G) as HinG by (inversion Hin; auto).
  assert (sealed G) as HslG by (eapply ctx_closed_push_G; eauto).
  assert (ctx_ok G) as HokG by eauto.
  assert (renv_corr G e) as HcorG.
  eapply renv_corr_push_inv; eauto.

  (* Same trick as in typ_approx_m_subcom *)
  rename T1 into T1p. rename T2 into T2p. rename Hsub12 into Hsub12p.
  destruct (subtyp_res_polar_inv (G § y / vk ~ T) e ᵐ δ T1p T2p Hsub12p) as [T1 [T2 [Hsub12 Hp]]].
  replace T1p with match δ with
  | ⊕ => T1
  | ⊝ => T2
  end in Hap13.

  assert ( match δ with
  | ⊕ =>
     (exists T4, G ⁏ e ⊢ T3 <: T4 /\ T2 ↝ ᵉ ⦇ ⊕ ⦈ T4 /\ nosel T4 y)
  | ⊝ =>
     (exists T4, G ⁏ e ⊢ T4 <: T3 /\ T1 ↝ ᵉ ⦇ ⊝ ⦈ T4 /\ nosel T4 y)
  end).
  clear dependent T1p.
  clear dependent T2p.

  revert dependent T3.

  dependent induction Hsub12; intros; try destruct δ; eauto;
    try solve [eexists; repeat split; eauto; auto]. (* The second auto is for some reason required *)
  - inversions_typ Hap13.
    destruct (typ_approx_e_exists ⊝ T0 y) as [T4 [? ?]].
    eexists; repeat split; eauto; auto.
  - inversions_typ Hap13.
    destruct (typ_approx_e_exists ⊕ T0 y) as [T4 [? ?]].
    eexists; repeat split; eauto; auto.
  - (* tyq *)
    destruct (typ_approx_tyq_e HinG HslG HcorG Hnef H0 (typ_approx_e_ctx G Hap13)) as [T4 [? [? ?]]]; eauto.
    exists T4. repeat split; eauto using typ_approx_e_ctx.
  - symmetry in H0.
    destruct (typ_approx_tyq_e HinG HslG HcorG Hnef H0 (typ_approx_e_ctx G Hap13)) as [T4 [? [? ?]]]; eauto.
    symmetry in H2.
    exists T4. repeat split; eauto using typ_approx_e_ctx.
  - (* trans *)
    edestruct (IHHsub12_1) as [T4 [? [? ?]]]; eauto.
    edestruct (IHHsub12_2) as [T5 [? [? ?]]]; eauto.
    exists T5. repeat split; eauto.
  - (* trans *)
    edestruct (IHHsub12_2) as [T4 [? [? ?]]]; eauto.
    edestruct (IHHsub12_1) as [T5 [? [? ?]]]; eauto.
    exists T5. repeat split; eauto.
  - (* or 11 *)
    destruct (typ_approx_e_exists ⊕ T2 y) as [T4 [? ?]]. exists (typ_or T3 T4). repeat split; eauto.
  - (* or 11 *)
    inversions_typ Hap13. inversions_tyq Hns3. exists T1'. repeat split; eauto.
  - (* or 12 *)
    destruct (typ_approx_e_exists ⊕ T1 y) as [T4 [? ?]]. exists (typ_or T4 T3). repeat split; eauto.
  - (* or 12 *)
    inversions_typ Hap13. inversions_tyq Hns3. exists T2'. repeat split; eauto.
  - (* or2 *)
    inversions_typ Hap13. inversions_tyq Hns3.
    edestruct IHHsub12_1 as [T4 [? [? ?]]]; eauto.
    edestruct IHHsub12_2 as [T5 [? [? ?]]]; eauto.
    destruct (typ_approx_e_subcom  e HokG H1 H8 H6 H9) as [T6 [? [? [? ?]]]].
    exists (T6).
    repeat split; eauto.
  - (* or2 *)
    edestruct IHHsub12_1 as [T4 [? [? ?]]]; eauto.
    edestruct IHHsub12_2 as [T5 [? [? ?]]]; eauto.
    exists (typ_or T4 T5).
    repeat split; eauto.
  - (* and 11 *)
    inversions_typ Hap13. inversions_tyq Hns3. exists T1'. repeat split; eauto.
  - (* and 11 *)
    destruct (typ_approx_e_exists ⊝ T2 y) as [T4 [? ?]]. exists (typ_and T3 T4). repeat split; eauto.
  - (* and 12 *)
    inversions_typ Hap13. inversions_tyq Hns3. exists T2'. repeat split; eauto.
  - (* and 12 *)
    destruct (typ_approx_e_exists ⊝ T1 y) as [T4 [? ?]]. exists (typ_and T4 T3). repeat split; eauto.
  - (* and2 *)
    edestruct IHHsub12_1 as [T4 [? [? ?]]]; eauto.
    edestruct IHHsub12_2 as [T5 [? [? ?]]]; eauto.
    exists (typ_and T4 T5).
    repeat split; eauto.
  - (* and2 *)
    inversions_typ Hap13. inversions_tyq Hns3.
    edestruct IHHsub12_1 as [T4 [? [? ?]]]; eauto.
    edestruct IHHsub12_2 as [T5 [? [? ?]]]; eauto.
    destruct (typ_approx_e_subcom e HokG H1 H8 H6 H9) as [T6 [? [? [? ?]]]].
    exists (T6).
    repeat split; eauto.
  - inversions_typ Hap13.
    exists (typ_bnd T0).
    repeat split; eauto.
  - inversions_typ Hap13.
    exists (typ_nocap).
    repeat split; eauto.
  - inversions_typ Hap13.

    destruct (typ_approx_e_exists ⊝ T1 y) as [T3 [? ?]].
    destruct (typ_approx_e_exists ⊕ T2 y) as [T4 [? ?]].

    exists (typ_rcd (dec_fld a T3 T4)).
    repeat split; eauto. apply (nosel_bin (typ_binary_rcd_fld a)); eauto.
  - exists typ_nocap. repeat split; eauto.
    inversions Hap13. destruct H1; subst; try discriminate.
    apply subtyp_nocap_fld.
  - inversions_typ Hap13.
    pick_fresh z.
    destruct (typ_approx_e_exists ⊝ (open z T1) y) as [T3 [? ?]].
    destruct (typ_approx_e_exists ⊕ (open z T2) y) as [T4 [? ?]].

    exists (typ_rcd (dec_typ (typ_label_normal A) (close z T3) (close z T4))).
    repeat split; eauto.
    constructor; eauto.
  - (* Because of update to typ_approx_me_fun *)
    exists typ_nocap. repeat split; eauto.
    inversions Hap13. destruct H1; subst; try discriminate. eauto.
  - (* Because of update to typ_approx_me_fun *)
    pose proof typ_approx_me_fun as Hmefun.
    destruct (mode_has_fun' typing_mode) eqn:Htm.
    + exists typ_top. repeat split; eauto.
      inversions Hap13. destruct H1; subst; try discriminate. auto.
      eapply Hmefun with (δ := ⊕). auto.
    + exists typ_top. repeat split; eauto.
      inversions Hap13. destruct H1; subst; try discriminate. auto.
      eapply Hmefun with (δ := ⊕). auto.
  - (* Because of update to typ_approx_me_fun *)
    pose proof typ_approx_me_fun as Hmefun.
    destruct (mode_has_fun' typing_mode) eqn:Htm.
    + eexists. repeat split; eauto.
      inversions Hap13. destruct H1; subst; try discriminate.
      destruct (mode_has_fun' typing_mode); eauto. discriminate.
    + eexists. repeat split; eauto.
      inversions Hap13. destruct H1; subst; try discriminate.
      destruct (mode_has_fun' typing_mode); eauto.
  - (* fld *)
    inversions_typ Hap13. inversions_tyq Hns3.
    edestruct IHHsub12_1 as [T4 [? [? ?]]]; eauto.
    edestruct IHHsub12_2 as [T5 [? [? ?]]]; eauto.
    exists (typ_rcd (dec_fld a T4 T5)).
    repeat split; eauto.
    apply (nosel_bin (typ_binary_rcd_fld a)); eauto.
  - (* fld *)
    inversions_typ Hap13. inversions_tyq Hns3.
    edestruct IHHsub12_1 as [T4 [? [? ?]]]; eauto.
    edestruct IHHsub12_2 as [T5 [? [? ?]]]; eauto.
    exists (typ_rcd (dec_fld a T4 T5)).
    repeat split; eauto.
    apply (nosel_bin (typ_binary_rcd_fld a)); eauto.
  - (* typ *)
    inversions_typ Hap13. inversions_tyq Hns3.
    pick_fresh z.
    edestruct (H1 z) as [T4 [? [? ?]]]; eauto.
    edestruct (H3 z) as [T5 [? [? ?]]]; eauto.
    exists (typ_rcd (dec_typ A (close z T4) (close z T5))).
    repeat split; eauto.
    eapply subtyp_typ; intro.

      erewrite typ_approx_reopen_subst; eauto.
      rewrite (subst_intro) with (x0:=z).
      eapply usubst_end1; eauto.

      all: eauto. 2: { eapply typ_approx_e_ctx in H5. apply H5. } notin_solve.

      erewrite typ_approx_reopen_subst; eauto.
      rewrite (subst_intro) with (x0:=z).
      eapply usubst_end1; eauto.

      all: eauto. 2: { eapply typ_approx_e_ctx in H10. apply H10. } notin_solve.
    eauto 7.
  - (* typ *)

    inversions_typ Hap13. inversions_tyq Hns3.
    pick_fresh z.
    edestruct (H1 z) as [T4 [? [? ?]]]; eauto.
    edestruct (H3 z) as [T5 [? [? ?]]]; eauto.
    exists (typ_rcd (dec_typ A (close z T4) (close z T5))).
    repeat split; eauto.
    eapply subtyp_typ; intro.

      erewrite typ_approx_reopen_subst; eauto.
      rewrite (subst_intro) with (x0:=z).
      eapply usubst_end1; eauto.

      all: eauto. 2: { eapply typ_approx_e_ctx in H5. apply H5. } notin_solve.

      erewrite typ_approx_reopen_subst; eauto.
      rewrite (subst_intro) with (x0:=z).
      eapply usubst_end1; eauto.

      all: eauto. 2: { eapply typ_approx_e_ctx in H10. apply H10. } notin_solve.
    eauto 7.
  - inversions_typ Hap13. inversions_tyq Hns3.
    inversions_typ H2. inversions_tyq H3.
    inversions_typ H4. inversions_tyq H5.
    exists (typ_rcd (dec_typ A (typ_or T1'0 T1') (typ_and T2'0 T2'1))).
    repeat split; eauto 6.
    constructor; intro; sympls; constructor; eauto.
  - inversions_typ Hap13. inversions_tyq Hns3.
    pick_fresh z. specialize (H7 z). specialize (H8 z).

  destruct T1'; sympls; inversions_typ H7.
  destruct T2'; sympls; inversions_typ H8.

   inversions_tyq H2.
    inversions_tyq H3.
    exists (typ_and (typ_rcd (dec_typ A T1'1 T2'1)) (typ_rcd (dec_typ A T1'2 T2'2))).
    repeat split; eauto 6.

constructor; eauto; constructor; intro; eapply typ_approx_reopen; eauto; sympls; notin_solve.

  - inversions_typ Hap13. inversions_tyq Hns3.
    inversions_typ H4. inversions_tyq H5.
    exists (typ_or (typ_and T1' T1'0) (typ_and T1' T2'0)).
    repeat split; eauto 6. constructor; eauto.
  - inversions_typ Hap13. inversions_tyq Hns3.
    inversions_typ H2. inversions_tyq H3.
    inversions_typ H4. inversions_tyq H5.
    destruct (typ_approx_e_subcom e HokG H6 H3 H2 H4) as [T6 [? [? [? ?]]]].
    exists (typ_and T6 (typ_or T2'0 T2'1)).
    repeat split; eauto.
    assert (G ⁏ e ⊢ typ_and T6 (typ_or T2'0 T2'1) <: typ_or (typ_and T6 T2'0) (typ_and T6 T2'1)) by eauto.
    eapply subtyp_trans; eauto.
    eapply (@subtyp_or _ (and_or_subtyp _)); eauto;
    eapply (@subtyp_and _ (and_or_subtyp _)); eauto.

  - (* Because of subtyp_nocap_res_nm *)
    inversions Hap13. destruct H2; subst; try discriminate.
    forwards HT1: typ_approx_e_exists ⊝ T1.
    destructc HT1 as T1' HT1' HT1ns.
    forwards HT2: typ_approx_e_exists ⊕ T2.
    destructc HT2 as T2' HT2' HT2ns.

    exists (typ_or typ_nocap
              (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability)
                                T1' T2'))).
    repeat(split); eauto.

     eapply typ_approx_andor; eauto.
      eapply typ_approx_typ; intros.
      eapply typ_approx_open; eauto.
      eapply typ_approx_open; eauto.

  - (* Because of subtyp_nocap_res_nm *)
    inversions Hap13; match goal with
                      | [ H : _ = _ \/ _ |- _ ] => destruct H; subst;
                                                try discriminate
                      end. inversions H1.
    inversions H3; match goal with
                   | [ H : _ = _ \/ _ |- _ ] => destruct H; subst;
                                             try discriminate
                   end.
    inversions H5; match goal with
                   | [ H : _ = _ \/ _ |- _ ] => destruct H; subst;
                                             try discriminate
                   end.
    exists typ_top. repeat(split). all: eauto.
  - inversions_typ Hap13.
    exists (typ_or typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) ). splits; eauto.
    constructor; eauto.
  - inversions_typ Hap13.
    inversions_typ H3.
    inversions_typ H5.
    eapply typ_approx_bot_open_inv in H10.
    eapply typ_approx_bot_open_inv in H11.
    subst.
    exists typ_top; splits; eauto.
  - inversions_typ Hap13.
inversions_typ H3.
    inversions_typ H5.
    eapply typ_approx_bot_open_inv in H10.
    eapply typ_approx_bot_open_inv in H11.
    subst.
    exists typ_bot. splits; eauto.
  - eapply typ_approx_bot_inv in Hap13. subst.
    exists (typ_and typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) ). splits; eauto.
    eapply typ_approx_andor; eauto. eapply typ_approx_typ; eauto.
  - eapply typ_approx_top_inv in Hap13. subst.
  exists ( typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_top) ). splits; eauto.
eapply typ_approx_typ; eauto.
  - inversions_typ Hap13.
    eapply typ_approx_bot_open_inv in H8.
    eapply typ_approx_top_open_inv in H9.
    subst.
    exists typ_top. splits; eauto.

  - (* sel_u *)
    inversions_typ Hap13. inversions_tyq Hns3.
    + eapply pf_push_inv in H. exists T3.
      edestruct typ_approx_m_cut as [? [? ?]]. eapply HokG.
      assert (y \notin fv (typ_rcd ((dec_typ B T1 T2)))).
      {
        rewrite <- (ctx_concat_empty_r G) in H.
        eapply pf_concat_notin_fv; eauto.
        assert (ctx_ok (G § y / vk ~ T)) as HokG1 by eauto.
        inversion HokG1; auto.
      }
      assert (y \notin fv (open v2 T2)).
      {
        eapply notin_fv_open. intro. subst. contradiction.
        rewrite fv_typ_eq in *. unfold fv_typ in *. rewrite notin_union in H1. destruct H1. auto.
      }
      eapply notin_fv_nosel; eauto.
      eauto.
      repeat split; eauto.
      assert (G ⁏ e ⊢ (open v2 T2) <: T3). { apply H2. }
      assert (G ⁏ e ⊢ typ_sel (avar_f v) B (avar_f v2) <: (open v2 T2)).
      { eapply subtyp_sel1. eapply precise_to_general; eauto. }
      eauto.
      intro; subst; auto.
    +
      destruct (typ_approx_e_exists ⊕ T3 y) as [T4 [? ?]].
      exists T4. eauto.
  - (* sel_l *)
    inversions_typ Hap13. inversions_tyq Hns3.
    + eapply pf_push_inv in H. exists T3.
      edestruct typ_approx_m_cut as [? [? ?]]. eapply HokG.
      assert (y \notin fv (typ_rcd ((dec_typ B T1 T2)))).
      {
        rewrite <- (ctx_concat_empty_r G) in H.
        eapply pf_concat_notin_fv; eauto.
        assert (ctx_ok (G § y / vk ~ T)) as HokG1 by eauto.
        inversion HokG1; auto.
      }
      assert (y \notin fv (open v2 T1)).
      {
        eapply notin_fv_open. intro. subst. contradiction.
        rewrite fv_typ_eq in *. unfold fv_typ in *. auto.
      }
      eapply notin_fv_nosel; eauto.
      eauto.
      repeat split; eauto.
      assert (G ⁏ e ⊢ T3 <: (open v2 T1)). { apply H2. }
      assert (G ⁏ e ⊢ (open v2 T1) <: typ_sel (avar_f v) B (avar_f v2)).
      { eapply subtyp_sel2. eapply precise_to_general; eauto. }
      eauto.
      intro; subst; auto.
    +
      destruct (typ_approx_e_exists ⊝ T3 y) as [T4 [? ?]].
      exists T4. eauto.
  - destruct δ; destruct Hp; subst; eauto.
  - destruct δ; destruct Hp; eauto.
Qed.

(* Lemma 5.116 FromSSub *)
Lemma subtyp_restricted_to_general δ G e T1 T2:
  inert G ->
  ok e ->
  G⁏e ⊢ˢ⦇δ⦈ T1 <: T2 ->
  G⁏e ⊢ T1 <: T2.
Proof.
  introv Hin Hoke Hsub12.
  dependent induction Hsub12; intros; eauto.
  - destruct (mode_has_fun' typing_mode) eqn:Htm; eauto.
  - assert (F ⁏ e ⊢# typ_sel (avar_f v) B (avar_f v2) <: (open v2 T2)) by eauto.
    eapply tight_to_general; eauto.
  - assert (F ⁏ e ⊢# (open v2 T1) <: typ_sel (avar_f v) B (avar_f v2)) by eauto.
    eapply tight_to_general; eauto.
  - eapply tight_to_general in H0; eauto.
  - eapply tight_to_general in H; eauto.
Qed.

End WithTypingMode.
