Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax GeneralTyping TypingMode PreciseTyping
        EWeakening Indep RecordTypes InertContext Equivalence SubstCommon
        TightTyping InvertibleTyping GeneralToTight TypeApproximation
        RestrictedTightSubtyping EnvCorrespondence Substitution
        UnboundSubstitution  HeapCorrespondence TFree ConfigVarkinds
        TFreeTyping Narrowing Subenvironments.

(** * Read-only typing context *)
Inductive typ_refcap_ro: typ -> typ -> Prop :=
| typ_refcap_ro_top:
    typ_refcap_ro typ_top typ_top
| typ_refcap_ro_and S1 S2 T1 T2:
    typ_refcap_ro S1 S2 ->
    typ_refcap_ro T1 T2 ->
    typ_refcap_ro (typ_and S1 T1) (typ_and S2 T2)
| typ_refcap_ro_mut S T:
  typ_refcap_ro (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability)
                                  S T))
                (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability)
                                  typ_bot typ_top))
| typ_refcap_ro_other l S T:
    (l <> (typ_label_refcap typ_refcap_mutability) ->
    typ_refcap_ro (typ_rcd (dec_typ l S T))
                  (typ_rcd (dec_typ l S T)))
.

Hint Constructors typ_refcap_ro: core.

(* Read-only version of an inert type *)
Inductive typ_ro: typ -> typ -> Prop :=
| typ_ro_bnd T:
  typ_ro (typ_bnd T) (typ_bnd T)
| typ_ro_fun S T:
  typ_ro (typ_fun S T) (typ_fun S T)
| typ_ro_obj T Tc Tcr:
  typ_refcap_ro Tc Tcr ->
  typ_ro (typ_and (typ_bnd T) Tc) (typ_and (typ_bnd T) Tcr)
.

Hint Constructors typ_ro: core.

(* Read-only version of an inert context *)
Inductive ctx_ro: ctx -> ctx -> Prop :=
| ctx_ro_empty:
  ctx_ro ctx_empty ctx_empty
| ctx_ro_push_loc G Gr x T:
  ctx_ro G Gr ->
  ctx_ro (G § x / v_loc ~ T) (Gr § x / v_loc ~ T)
| ctx_ro_push_ref G Gr x T Tr:
  ctx_ro G Gr ->
  typ_ro T Tr ->
  ctx_ro (G § x / v_ref ~ T) (Gr § x / v_ref ~ Tr)
.

Hint Constructors ctx_ro: core.

(* LEMMAS *)

Lemma typ_refcap_ro_fv T Tr : typ_refcap_ro T Tr -> fv Tr \c fv T.
Proof.
  introv Hro. induction Hro; simpl;
    eauto using subset_refl, subset_union_2.
  rewrite union_empty_l. apply subset_empty_l.
Qed.

Lemma typ_ro_fv T Tr : typ_ro T Tr -> fv Tr \c fv T.
Proof.
  introv Hro. inversions Hro;
    simpl; eauto using typ_refcap_ro_fv, subset_refl, subset_union_2.
Qed.

Lemma typ_refcap_ro_sealed Tc Tcr :
  typ_refcap_ro Tc Tcr -> sealed Tc -> sealed Tcr.
Proof.
  induction 1; try eauto.

  - introv Hs.
    assert (HsS : sealed S1).
    { intro x. specialize (Hs x). simpl in Hs. inversions Hs.
      rewrite H2 at 1. auto. }
    assert (HsT : sealed T1).
    { intro x. specialize (Hs x). simpl in Hs. inversions Hs.
      rewrite H3 at 1. auto. }
    autospecialize_prop. intros x. simpl.
    specialize (IHtyp_refcap_ro1 x). specialize (IHtyp_refcap_ro2 x).
    rewrite IHtyp_refcap_ro1 at 1. rewrite IHtyp_refcap_ro2 at 1. auto.

  - introv _. hnf. intros x. simpl. auto.

Qed.

Lemma typ_ro_sealed T Tr : typ_ro T Tr -> sealed T -> sealed Tr.
Proof.
  introv Hro Hs. destruct Hro; try easy.
  assert (sealed Tc).
  { intros x. specialize (Hs x). simpl in Hs. inversions Hs.
    rewrite H2 at 1. auto. }
  apply typ_refcap_ro_sealed in H; eauto. intros x.
  specialize (Hs x). specialize (H x).
  simpl in *. inversions Hs. rewrite H at 1. rewrite H2 at 1. auto.
Qed.

Lemma ctx_ro_fv G Gr : ctx_ro G Gr -> ctx_fv Gr \c ctx_fv G.
Proof.
  introv Hro. induction Hro;
    simpl; eauto using typ_ro_fv, subset_refl, subset_union_2.
Qed.

Section WithTypingMode.
Context {typing_mode: Mode}.

Hint Constructors RecordTypes.refcap_typ RecordTypes.refcap_dec refcap_inert_mutability inert: core.

Lemma typ_refcap_ro_exists T:
  refcap_inert_type T v_ref ->
  exists Tr,
    typ_refcap_ro T Tr /\ refcap_inert_type Tr v_ref .
Proof.
  introv Hin.
  unfold refcap_inert_type in *.
  destruct ( mode_refcaps' typing_mode).
  - induction Hin; subst; eauto.
    + inversions H.
      exists ((typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_top))).
      destruct lt.
      split.
      eapply typ_refcap_ro_mut.
      simpls; eauto.
    + inversions H.
      destruct lt.
      forwards Tr HTTr HTr: IHHin; eauto.
  - dependent induction Hin; eauto.
Qed.

Lemma typ_ro_exists T:
  inert_typ T v_ref ->
  exists Tr,
    typ_ro T Tr /\ inert_typ Tr v_ref .
Proof.
  introv Hin.
  inversions Hin.
  - eauto.
  - forwards Tr HSTr HTr: typ_refcap_ro_exists; eauto.
Qed.

End WithTypingMode.

Section WithRodotMode.

Context (typing_mode : Mode).
Context (Hmm : mode_has_mutability typing_mode).

Lemma typ_ro_exists_bnd T:
  inert_typ T v_ref ->
  exists S,
    typ_ro T (typ_and (typ_bnd S)
                (typ_mut_marker typ_top)).
Proof.
  introv Hin. inversions Hin. simpl in H.
  try congruence; info_typing_mode_contradiction typing_mode.
  forwards Tr HSTr HTr: typ_refcap_ro_exists; eauto.
  hnf in HTr. hnf in H0. info_typing_mode_contradiction typing_mode.
  all: try (inversions HTr). all: try (inversions HSTr). all: eauto.
Qed.

End WithRodotMode.

Section WithTypingMode.
Context {typing_mode: Mode}.

Lemma ctx_ro_exists G:
  inert G ->
  exists Gr,
    ctx_ro G Gr /\ inert Gr /\ ctx_dom Gr = ctx_dom G /\ ctx_fv Gr \c ctx_fv G.
Proof.
  introv Hin.
  induction Hin.
  - eauto using subset_refl.
  - destructc IHHin as Gr HGGr HGr Hdom Hfv.
    assert ( x §# Gr) by (rewrite Hdom; eauto).
    inversions H.
    + eexists. splits; eauto using subset_union_2, subset_refl.
      * constructor; auto. hnf. intros; eauto.
      * ctx_simpl_dom. rewrite Hdom; eauto.
    + forwards Tr HTTr HTr: typ_ro_exists; eauto.
      eexists. splits; eauto. ctx_simpl_dom.
      * constructor; auto.
        { hnf. intros; eauto. }
        { apply typ_ro_fv in HTTr. hnf; intros; eauto. }
      * simpl. rewrite Hdom; eauto.
      * simpl. eauto using subset_union_2, typ_ro_fv.
Qed.

Lemma record_has_ro T D Tr:
  record_has T D ->
  typ_refcap_ro T Tr ->
  record_has Tr D \/ (
    record_has Tr (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_top) /\
    exists SL SU,
      D = (dec_typ (typ_label_refcap typ_refcap_mutability) SL SU)
  ).
Proof.
  introv Hrh Hro.
  reverts Tr.
  induction Hrh; intros.
  - inversions Hro;
    eauto.
  -
    inversions Hro.
    forwards: IHHrh; eauto. destruct_all; subst; eauto.
    right. split; eauto.
  -
    inversions Hro.
    forwards: IHHrh; eauto. destruct_all; subst; eauto.
    right. split; eauto.
Qed.


Lemma ctx_ro_binds G Gr T x:
  inert G ->
  ctx_ro G Gr ->
  ctx_binds x T G ->
  ( (* References have ro types *)
    ctx_k G x v_ref /\
    exists Tr,
      ctx_binds x Tr Gr /\
      typ_ro T Tr
  ) \/ (
    ctx_k G x v_loc /\
    ctx_binds x T Gr
  ).
Proof.
  introv Hin Hro Hbi.
  reverts Gr.
  induction Hbi; intros; inversions Hin.
  - inversions Hro.
    + right. splits; eauto.
      eapply ctx_k_push_e.
    + left. splits; eauto.
      eapply ctx_k_push_e.
  - inversions Hro;
    forwards: IHHbi; destruct_all; eauto using ctx_k_push_n;
    left; split; eauto using ctx_k_push_n.
Qed.

Lemma pf_ro_fun G Gr x U S T:
  inert G ->
  ctx_ro G Gr ->
  G ⊢! x ∶ U ⪼ typ_fun S T ->
  Gr ⊢! x ∶ U ⪼ typ_fun S T.
Proof.
  introv Hin Hro Hty.
  forwards: pf_inert_fun_U Hty; eauto; subst.
  forwards: inert_precise_fun_inv; eauto.
  forwards: ctx_ro_binds; eauto.
  destruct_all.
  - inversions H2. eapply pf_bind. eauto.
  - eauto.
Qed.

Lemma pf_ro_bnd G Gr x U T:
  inert G ->
  ctx_ro G Gr ->
  G ⊢! x ∶ U ⪼ typ_bnd T ->
  exists Ur,
    Gr ⊢! x ∶ Ur ⪼ typ_bnd T.
Proof.
  introv Hin Hro Hty.
  forwards: pf_inert_bnd_U Hty; eauto.
  destruct_all;  subst.
  forwards: pf_binds; eauto.
  forwards: ctx_ro_binds; eauto.
  destruct_all.
  - inversions H2.
    exists (typ_and (typ_bnd T) Tcr). eauto.
  - eauto.
Qed.

Lemma pf_record_has : forall x G U T D,
    inert G ->
    G ⊢! x ∶ U ⪼ T ->
    record_has T D ->
    G ⊢! x ∶ U ⪼ (typ_rcd D).
Proof.
  introv Hi Hpf Hhas.
  induction Hhas; eauto.
Qed.

Lemma pf_obj_record_has : forall x G T S  D,
    inert G ->
    ctx_binds x (typ_objref T S) G ->
    record_has T D ->
    G ⊢! x ∶ typ_objref T S ⪼ (typ_rcd (open x D)).
Proof.
  introv Hi Hbi Hhas.
  eapply pf_record_has; eauto.
  eapply record_has_open; eauto.
Qed.
Lemma pf_refcaps_record_has : forall x G T S  D,
    inert G ->
    ctx_binds x (typ_objref T S) G ->
    record_has S D ->
    G ⊢! x ∶ typ_objref T S ⪼ (typ_rcd D).
Proof.
  introv Hi Hbi Hhas.
  eapply pf_record_has. eauto. 2: eauto.
  eauto.
Qed.

Lemma record_has_open_case x T D:
  record_has (open x T) D ->
  exists Dx, record_has T Dx /\ D = (open x Dx).
Proof.
  introv Hhas.
  remember (open x T) as Tx. reverts T.
  induction Hhas; intros; subst.
  - destruct T; simpls; inversions HeqTx.
    eauto.
  - destruct T0; simpls; inversions HeqTx.
    forwards: IHHhas; eauto. destruct_all; eauto.
  - destruct T0; simpls; inversions HeqTx.
    forwards: IHHhas; eauto. destruct_all; eauto.
Qed.

Lemma pf_ro_rcd G Gr x U D:
  inert G ->
  inert Gr ->
  ctx_ro G Gr ->
  G ⊢! x ∶ U ⪼ typ_rcd D ->
  exists Ur, (
    Gr ⊢! x ∶ Ur ⪼ typ_rcd D \/
    Gr ⊢! x ∶ Ur ⪼ typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_top) /\ (
    exists SL SU,
      D = (dec_typ (typ_label_refcap typ_refcap_mutability) SL SU)
    )
  ).
Proof.
  introv Hin Hir Hro Hty.

  forwards: pf_inert_rcd_U Hty; eauto.
  destruct_all;  subst.
  forwards: pf_binds; eauto.
  forwards: pf_record_sub Hty; eauto; destruct_all.
  -
    forwards: record_has_open_case H0. destruct_all; subst.
    forwards: ctx_ro_binds Hro; eauto.

    destruct_all.
    + inversions H5.
      exists (typ_and (typ_bnd x0) Tcr).
      left.
      eapply pf_obj_record_has; eauto.
    +
      exists (typ_and (typ_bnd x0) x1).
      left.
      eapply pf_obj_record_has; eauto.
  -
    forwards: ctx_ro_binds Hro; eauto.
    destruct_all.
    + inversions H4.
      exists  (typ_and (typ_bnd x0) Tcr).
      forwards: record_has_ro H0 H8.
      destruct_all.
      * left. eapply pf_refcaps_record_has; eauto.
      * right. subst. splits; eauto.
        eapply pf_refcaps_record_has; eauto.
    +
      exists (typ_and (typ_bnd x0) x1).
      left.
      eapply pf_refcaps_record_has; eauto.
Qed.

Lemma refcap_typ_empty_ro_same Tc Tcr :
  typ_refcap_ro Tc Tcr -> refcap_typ Tc \{} -> Tc = Tcr.
Proof.
  induction 1; eauto; introv Hrf.
  - inversions Hrf. inversions H5. simpl in H3. false.
    erewrite <- in_empty, <- H3, in_union. eauto using in_singleton_self.
  - inversions Hrf. false. simpl in H0.
    erewrite <- in_empty, <- H0. eauto using in_singleton_self.
Qed.

Lemma inert_typ_ro_corr Tr T :
  inert_typ T v_loc -> typ_ro T Tr -> typ_corr Tr T.
Proof.
  introv Hin Hro. inversions Hin.

  - inversions Hro. eapply typ_corr_fun. auto.

  - rename H into Hrcd, H0 into Hrf, H1 into Hindep.
    inversions Hro. rename H2 into Hrfro. eapply typ_corr_obj.
    hnf in Hrf. destruct (mode_refcap_corr' typing_mode) eqn:Htm.
    + destruct (mode_refcaps' typing_mode) eqn:Hrfm.
      2: { destruct typing_mode; simpl in *; try congruence. }
      eauto using refcap_typ_empty_ro_same.
    + destruct (mode_refcaps' typing_mode) eqn:Hrfm.
      1: { destruct typing_mode; simpl in *; try congruence. }
      inversions Hrf. inversions Hrfro; eauto.

Qed.

End WithTypingMode.
