Set Implicit Arguments.

Require Import Coq.Program.Equality.

Require Import DotTactics LibExtra.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping
        Weakening EWeakening Visibility.
Require Import  Indep ConfigTyping Closedness Sealing.

Section WithTypingMode.
Context {typing_mode: Mode}.

(** * Lemmas about sealed and closed elements and typing *)

Lemma seal_ty_lit_fun G e t1a T U x:
  closed (lit_fun T t1a) ->
  sealed G ->
  G ⁏ e ⊢ lit_fun T t1a ∶ typ_fun T U ->
  G ⁏ e ⊢ lit_fun T t1a ∶ typ_fun T (seal_rec 1 x U).
Proof.
  introv Hcl HslG Hty.
  eapply sealing_obj with (Go:=objctx_empty) (x0:=x) in Hty.
  rewrite (closed_sealed 0 x Hcl) in Hty.
  assert (closed T).
  eapply closed_lit_fun_invl; eauto.
  sympl in Hty. sympl_seal_in_all.
  rewrite (closed_sealed 0 x H) in Hty.
  rewrite HslG in Hty.
  auto.
Qed.

Lemma seal_ty_lit_fun_cont G e L T U (t2a: trm) Tt y:
  sealed G ->
  (forall x : var, closed t2a /[-> x] ) ->
  (closed T) ->
  (forall x : var, x \notin L -> G §ˣ x ~ typ_fun T U ⁏ e ⊢ t2a /[-> x] ∶ Tt) ->
  (forall x : var, x \notin L -> G §ˣ x ~ typ_fun T (seal_rec 1 y U) ⁏ e ⊢ t2a /[-> x] ∶ Tt  /[# 0 ->> y]).
Proof.
  introv HslG Hclt HclT Hty Hnx.
  specialize (Hty x Hnx).
  eapply sealing_obj with (Go:=objctx_empty) (x0:=y) in Hty.
  sympl in Hty. sympl_seal_in_all.
  rewrite HslG in Hty.
  rewrite (closed_sealed 0 y HclT) in Hty.
  rewrite (closed_sealed 0 y (Hclt x)) in Hty.
  auto.
Qed.

Lemma seal_ty_stack G e T (fs: stack) S y:
  ty_stack G e  T fs S ->
  closed fs ->
  ty_stack (G /[# 0 ->> y]) e  (T /[# 0 ->> y]) fs (S /[# 0 ->> y]).
Proof.
  introv Htys Hclfs.
  induction Htys.
  - (* subtyp *)
    eapply sealing_obj with (x:=y) in H.
    sympl in H.
    eauto.
  - 
    
    forwards: IHHtys.
    eapply stack_closed_fs; eauto.
    econstructor. eauto.
    intros.
    forwards: stack_closed_t x; eauto.
    forwards: H x. eauto.
    eapply sealing_obj with (x0:=y) in H3.
    rewrite (closed_sealed 0 y H2) in H3.
    eauto.
Unshelve.
  eapply sealing_subtyp.
Qed.


Lemma seal_ty_stack_closed G e T (fs: stack) S y:
  sealed G ->
  closed S ->
  closed fs ->
  ty_stack G e  T fs S ->
  ty_stack G e  (T /[# 0 ->> y]) fs S.
Proof.
  intros HslG HclS Hclfs Htys.
  forwards: seal_ty_stack Htys.
  auto.
  
    rewrite HslG in H.
    rewrite (closed_sealed 0 y HclS)  in H.
 
    eauto.
Qed.

End WithTypingMode.