(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax GeneralTyping TypingMode PreciseTyping
        EWeakening Indep RecordTypes InertContext Equivalence
        TightTyping InvertibleTyping GeneralToTight TypeApproximation RestrictedTightSubtyping EnvCorrespondence Substitution UnboundSubstitution  Nosel.

(** * Shortening lemmas *)

Notation typ_flag l := (typ_rcd (dec_typ l typ_bot typ_bot)) (only parsing).
Notation fld_typ_flag lf lt := (typ_rcd (dec_fld lf typ_bot (typ_rcd (dec_typ lt typ_bot typ_bot)))) (only parsing).

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma renv_corr_ctx_remove_end_notin G x vk T e :
  x # e ->
  renv_corr (G § x / vk ~ T) e ->
  renv_corr G e.
Proof.
  introv Hxnin Hcor.
  remember (G § x / vk ~ T) as G0.
  reverts G x T vk.
  induction Hcor; intros; subst; eauto.
  assert (Hxnin' : x # e) by eauto.
  specialize (IHHcor _ _ Hxnin' _ _ eq_refl).
  eapply renv_corr_ref; eauto.
  inversions H; eauto. false.
  rewrite dom_concat, dom_single, notin_union, notin_singleton in Hxnin.
  tauto.
Qed.

Lemma renv_corr_push_inv G w2 T2 e y2:
  w2 # e ->
  inert (G § w2 / v_ref ~ T2) ->
  renv_corr (G § w2 / v_ref ~ T2) (e & w2 ~ y2) ->
  y2 <> w2 /\
  exists Tw Ty G2 G3,
    renv_corr G2 e /\
    ctx_binds_k_after w2 v_ref Tw (G § w2 / v_ref ~ T2) G2 /\
    ctx_binds_k_after y2 v_loc Ty G2 G3 /\
    typ_corr Tw Ty.
Proof.
  introv Hfrw2 Hin Hcor.
  inversions Hcor. eapply empty_concat_inv in H1. destruct H1. eapply empty_single_inv in H0. contradiction.
  eapply eq_push_inv in H.
  destructc H as ? ? ?; subst.

  split.
  -
    eapply ctx_binds_k_after_unbound_r in H0.
    intro; subst.
    eapply ctx_binds_k_after_binds in H1.
    eapply ctx_binds_dom in H1. contradiction.
    eauto.
  - exists T2 Ty G G3. repeat(split).
    + eapply renv_corr_ctx_remove_end_notin in H4; auto.
    + eapply ctx_binds_k_e.
    + apply ctx_binds_k_after_initial_segment in H1.
      destruct H1 as [?G ?]; subst.
      apply ctx_binds_k_after_initial_segment in H0.
      destruct H0 as [?G ?]; subst.
      apply ctx_push_concat_inv in H as ?H.
      destruct_all; subst.
      * inversions H1. eapply ctx_binds_k_after_concat; eauto.
        { inversions Hin; eauto. }
        { rewrite <- ctx_concat_push in Hin.
          apply inert_concat_invert_r, inert_weakens in Hin.
          inversions Hin; eauto. }
      * inversions H. rewrite <- ctx_concat_push, <- ctx_concat_assoc.
        eapply ctx_binds_k_after_concat; eauto.
        { inversions Hin. rewrite ctx_concat_assoc. eauto using inert_ok. }
        { apply inert_weakens in Hin.
          inversions Hin.
          rewrite <- ctx_concat_push, <- ctx_concat_assoc in H0.
          eauto using ctx_weakens_concat_right. }
    + apply ctx_binds_k_after_binds in H0.
      inversions H0; eauto. tauto.
Qed.

Lemma inert_push_corr_vis G w T e y:
  w # e ->
  inert (G § w / v_ref ~ T) ->
  renv_corr (G § w / v_ref ~ T) (e & w ~ y) ->
  ctx_vis G y.
Proof.
  introv Hfrw Hin Hcor.
  destructc (renv_corr_push_inv Hfrw Hin Hcor) as Hn Tw Ty G2 G3 Hcor2 H.
  destructc H as Hbiw Hbiy Htcor.
  eapply ctx_binds_k_after_binds in Hbiy.
  inversions Hin.
  eapply ctx_vis_push_n_inv.
  eapply inert_binds_vis; eauto.
  eapply ctx_binds_k_after_initial_binds; eauto.
  eauto.
Qed.
(* Hint Resolve inert_push_corr_vis: core. *)

Lemma shortening_corr_ty G e w y T:
  inert (G § w / v_ref ~ T) ->
  w \notin ctx_fv G ->
  ok e ->
  renv_notin w e ->
  y <> w ->
  renv_corr (G § w / v_ref ~ T) (e & w ~ y) ->
  G ⁏ e ⊢ y ∶ (T /[ w -> y]).
Proof.
  introv Hin HnG Hoke Hne Hnyw Hcorr.
  assert (ctx_ok (G § w / v_ref ~ T)) as HokG by eauto.
  assert (binds w y (e & w ~ y)) as Hbi by auto.
  destruct_renv_corr_binds_to_ref_loc HokG Hcorr Hbi TL TR.

  inversions HbiR; try contradiction.
  apply ctx_binds_push_inv_n in HbiL.

  assert (w \notin fv TL).
  eapply ctx_binds_notin_fv. apply HbiL. auto.

  assert (G ⁏ e ⊢ y ∶ (TL /[ w -> y])).
  rewrite subst_fresh.
  eauto.
  eauto.

  assert (ctx_ok (G § w / v_ref ~ TL)). inversion HokG. auto.
  assert (ctx_weakens (G § w / v_ref ~ TL)). inversion Hin. eauto.
  assert (G § w / v_ref ~ TL ⁏ e ⊢ TL <: T). eapply typ_corr_subtyp; eauto.
  assert (G ⁏ e ⊢ (TL /[ w -> y]) <: (T /[ w -> y])).
  assert (w # e).
  { hnf in Hne. rewrite in_union in Hne. tauto. }
  eapply subst_end with (property := subtyp _); eauto.

  - intros. eapply inert_push_corr_vis; eassumption.
  - eapply ty_var_sub; eauto.
  - eauto.
Qed.

Lemma shortening_typ G e w y T v S:
  inert (G § w / v_ref ~ T) ->
  w \notin ctx_fv G ->
  ok e ->
  renv_notin w e ->
  y <> w ->
  v <> w ->
  w \notin fv S ->
  renv_corr (G § w / v_ref ~ T) (e & w ~ y)->
  (G § w / v_ref ~ T)⁏(e & w ~ y) ⊢ v ∶ S ->
  G⁏ e ⊢ v ∶ S.
Proof.
  introv Hin HnG Hoke Hne Hnyw Hnvw HnS Hcorr Hty.
  replace v with (v /[ w -> y]).
  replace S with (S /[ w -> y]).
  rewrite <- (ctx_concat_empty_r G).
  rewrite <- (ctx_map_empty (subst_var w y)).
  eapply subst_ty_epush with (AT:= ty_var _); eauto.

  autorewrite with ctx_flatten.
  eapply shortening_corr_ty; eauto.

  sympls; eauto.
  - intros; eapply inert_push_corr_vis; try eassumption.
    hnf in Hne. rewrite in_union in Hne. tauto.
  - apply subst_fresh; auto.
  - unfold subst_var, VarSubstVar.
    rewrite substi_ne_var; eauto.
Qed.

(* Lemma 15 (StnMRef) in "Reference mutability for DOT" *)
Lemma shortening_typ_ref G e w y T v l:
  inert (G § w / v_ref ~ T) ->
  w \notin ctx_fv G ->
  ok e ->
  renv_notin w e ->
  y <> w ->
  v <> w ->
  renv_corr (G § w / v_ref ~ T) (e & w ~ y)->
  (G § w / v_ref ~ T)⁏(e & w ~ y) ⊢ v ∶ typ_flag l ->
  G⁏ e ⊢ v ∶ typ_flag l.
Proof.
  introv Hin HnG Hoke Hne Hnyw Hnvw Hcorr Hty.
  assert (w \notin fv (typ_flag l)) by (simpl; notin_solve).
  eapply shortening_typ with (y := y); eauto.
Qed.

(* Helper *)
Lemma shortening_subtyp_typ_helper G e y T T1 B:
  (exists l, B = typ_flag l) \/ (B = typ_bot) ->
  inert (G § y / v_loc ~ T) ->
  sealed (G § y / v_loc ~ T) ->
  y \notin ctx_fv G ->
  ok e ->
  y # e ->
  renv_notin y e ->
  renv_corr (G § y / v_loc ~ T) e ->
  nosel T1 y ->
  (G § y / v_loc ~ T)⁏e ⊢# T1 <: B ->
  G⁏e ⊢ T1 <: B.
Proof with (match goal with
            | [HBeq : (exists l, ?B = typ_flag l) \/ ?B = typ_bot |- _] =>
              destruct HBeq as [[?l ?] | ?]; subst;
              [ eapply typ_approx_typ_bot_inv; eauto
              | eapply typ_approx_bot_inv; eauto]
            | _ => idtac
            end; subst).

  introv HBeq Hin HsG Hn Hoke Hned Hnef Hcorr Hnosel Hsubb.
  assert (ctx_ok (G § y / v_loc ~ T)) as HokG by eauto.
  assert (ctx_ok (G )) as HokG0 by eauto.

  (* To s subtyping *)
  destruct (typ_approx_tight_to_s ⊕ Hin HsG Hoke Hcorr Hsubb) as
      [T3 [T4 [Hsap [Hsbap Hssub]]]].
  assert (T4 = B)...
  pose proof (typ_approx_s_sub e HsG Hsap); simpl in H.
  pose proof (subtyp_trans_res approx_target_subtyping_s H Hssub).

  (* To m subtyping *)
  destruct (typ_approx_s_to_m Hin HsG Hoke Hcorr H0) as [T5 [T6 [Hmap [Hmbap Hmsub]]]].
  assert (T6 = B)...

  (* To e *)
  destruct (typ_approx_m_cut e HokG0 Hnosel Hmap) as [Henosel [Hesub Heap]].
  epose proof (typ_approx_e_cut B Hin HsG Hcorr Hn Hned Hnef
                                ?[C] Heap Henosel).
  destruct H1 as [T4 [Hbsub [Hbap Hbnosel]]].
  assert (T4 = B)...
  eapply subtyp_trans; eauto.

 (* Solve existential ?[C]. Is there a better way? *)
  Unshelve. eauto.

Qed.

Corollary shortening_subtyp_bot_loc G e y T T1:
  inert (G § y / v_loc ~ T) ->
  sealed (G § y / v_loc ~ T) ->
  y \notin ctx_fv G ->
  ok e ->
  y # e ->
  renv_notin y e ->
  renv_corr (G § y / v_loc ~ T) e ->
  nosel T1 y ->
  (G § y / v_loc ~ T)⁏e ⊢# T1 <: typ_bot ->
  G⁏e ⊢ T1 <: typ_bot.
Proof.
  eauto using shortening_subtyp_typ_helper.
Qed.

Lemma shortening_subtyp_bot_ref G e w y T T1:
  inert (G § w / v_ref ~ T) ->
  w \notin ctx_fv G ->
  ok e ->
  renv_notin w e ->
  y <> w ->
  renv_corr (G § w / v_ref ~ T) (e & w ~ y)->
  w \notin fv T1 ->
  (G § w / v_ref ~ T)⁏ (e & w ~ y) ⊢# T1 <: typ_bot ->
  G⁏e ⊢ T1 <: typ_bot.
Proof.
  introv Hin Hn Hoke Hne Hwy Hcor Hns Hsub.

  replace T1 with (T1 /[ w -> y]).
  replace typ_bot with (typ_bot /[ w -> y]).
  rewrite <- (ctx_concat_empty_r G).
  rewrite <- (ctx_map_empty (subst_var w y)).
  eapply substitution_epush with (property:= subtyp _); eauto.
  autorewrite with ctx_flatten. eapply tight_to_general. eauto. unfold renv_notin in Hne. rewrite notin_union in Hne. destruct Hne. eauto. eauto.
  eapply shortening_corr_ty; eauto.
  sympls; eauto.
  - intros. eapply inert_push_corr_vis; try eassumption.
    hnf in Hne. rewrite in_union in Hne. tauto.
  - eapply subst_fresh; auto.
  - eapply subst_fresh; auto.
Qed.

(* Lemma 13 (StnMLoc) in "Reference mutability for DOT" *)
Lemma shortening_typ_loc G e y T v l:
  inert (G § y / v_loc ~ T) ->
  sealed (G § y / v_loc ~ T) ->
  y \notin ctx_fv G ->
  ok e ->
  y # e ->
  renv_notin y e ->
  v <> y ->
  renv_corr (G § y / v_loc ~ T) e->
  (G § y / v_loc ~ T)⁏e ⊢ v ∶ typ_flag l ->
  G⁏ e ⊢ v ∶ typ_flag l.
Proof.
  introv Hin HclG Hn Hoke Hned Hnef Hvy Hcor Hty.
  set (S := typ_flag l).
  assert (inert G) as HinG by (inversion Hin; eauto).
  assert (ctx_ok (G § y / v_loc ~ T)) as HokG by eauto.
  assert ((G § y / v_loc ~ T)⁏e ⊢## v ∶ S ) as Htyi. {
  eauto using general_to_tight_typing, tight_to_invertible.
  }

  forwards Hpf: invertible_to_precise_typ_dec Htyi.
  destruct Hpf as [TL [TU [TG [? [? ?]]]]].
  pose proof H as HbiG.
  apply pf_push_inv in HbiG; auto.


  assert (forall x, G ⁏ e ⊢ (open x TU) <: (open x typ_bot)). {
    pick_fresh z.
    assert(G ⁏ e ⊢ (open z TU) <: (open z typ_bot)).
    specialize (H0 z). sympls.
    eapply shortening_subtyp_bot_loc. eauto. eauto. eauto. eauto.  eauto.  eauto. eauto.
    apply notin_fv_nosel. eapply notin_fv_open. notin_solve.
    eapply pf_dec_typ_notin_fv_u. 3: apply HbiG. inversions Hin. auto. eauto. auto.

    intros.
    replace (open x TU) with ((open z TU) /[z -> x]).
    replace (typ_bot) with (typ_bot /[z -> x]).
    eapply usubst_end1; eauto.

    simpls. auto.
    rewrite <- subst_intro; eauto.
  }

  assert (y \notin fv (typ_rcd (dec_typ l TL TU))).
  ctx_split_from y in H.
  eapply pf_concat_notin_fv; eauto. ctx_simpl_dom. notin_solve. inversion HokG; auto.
  rewrite fv_typ_eq in *. unfold fv_typ in *. auto.
  eapply precise_to_general with (e:=e) in HbiG; eauto.
  eapply ty_var_sub; eauto.
  constructor; auto.
Qed.

(* Lemma 16 (StnMFRef) in "Reference mutability for DOT" *)
Lemma shortening_fld_typ_ref G e w y T v lf lt:
  inert (G § w / v_ref ~ T) ->
  sealed (G § w / v_ref ~ T) ->
  w \notin ctx_fv G ->
  ok e ->
  renv_notin w e ->
  y <> w ->
  v <> w ->
  renv_corr (G § w / v_ref ~ T) (e & w ~ y)->
  (G § w / v_ref ~ T)⁏(e & w ~ y) ⊢ v ∶ fld_typ_flag lf lt ->
  G⁏ e ⊢ v ∶ fld_typ_flag lf lt.
Proof.
  introv Hin Hn Hoke Hnef Hnyw Hnwv Hcor Hty.
  assert (w \notin fv (fld_typ_flag lf lt)) by (simpl; notin_solve).
  eapply shortening_typ with (y := y); eauto.
Qed.

(* Lemma 17 (StnSub) in "Reference mutability for DOT" *)
Corollary shortening_subtyp_typ_loc G e y T T1 l:
  inert (G § y / v_loc ~ T) ->
  sealed (G § y / v_loc ~ T) ->
  y \notin ctx_fv G ->
  ok e ->
  y # e ->
  renv_notin y e ->
  renv_corr (G § y / v_loc ~ T) e ->
  nosel T1 y ->
  (G § y / v_loc ~ T)⁏e ⊢# T1 <: typ_flag l ->
  G⁏e ⊢ T1 <: typ_flag l.
Proof.
  eauto using shortening_subtyp_typ_helper.
Qed.

(* Lemma 14 (StnMFLoc) in "Reference mutability for DOT" *)
Lemma shortening_fld_typ_loc G e y T v lf lt:
  inert (G § y / v_loc ~ T) ->
  sealed (G § y / v_loc ~ T) ->
  y \notin ctx_fv G ->
  ok e ->
  y # e ->
  renv_notin y e ->
  v <> y ->
  renv_corr (G § y / v_loc ~ T) e->
  (G § y / v_loc ~ T)⁏e ⊢ v ∶ fld_typ_flag lf lt ->
  G⁏ e ⊢ v ∶ fld_typ_flag lf lt.
Proof.
  introv Hin HclG Hn Hoke Hned Hnef Hvy Hcor Hty.
  set (S := typ_rcd (dec_fld lf typ_bot (typ_rcd (dec_typ lt typ_bot typ_bot)))).
  assert (inert G) as HinG by (inversion Hin; eauto).
  assert (ctx_ok (G § y / v_loc ~ T)) as HokG by eauto.
  assert ((G § y / v_loc ~ T)⁏e ⊢## v ∶ S ) as Htyi. {
    eauto using general_to_tight_typing, tight_to_invertible.
  }

  forwards Hpf: invertible_to_precise_fld_dec Htyi.
  destruct Hpf as [TL [TU [TG [? [? ?]]]]].

  pose proof H as HbiG.
  apply pf_push_inv in HbiG; auto.

  assert (G ⁏ e ⊢ TU <: typ_flag lt).
  eapply shortening_subtyp_typ_loc; eauto.
  apply notin_fv_nosel.

  assert (y \notin fv (typ_rcd (dec_fld lf TL TU))).
  ctx_split_from y in H.
  eapply pf_concat_notin_fv; eauto. ctx_simpl_dom. notin_solve. inversion HokG; auto.
  rewrite fv_typ_eq in *. unfold fv_typ in *. auto.
  eapply precise_to_general with (e:=e) in HbiG; eauto.
  eapply ty_var_sub; eauto.
  constructor; auto.
Qed.

End WithTypingMode.
