(** remove printing ~ *)

(** This module contains lemmas related to tight typing [G ⊢# t: T] *)

Set Implicit Arguments.

Require Import LibExtra.
Require Import AbstractSyntax TypingMode GeneralTyping Equivalence.
Require Import TightTyping.
Require Import Orctx Logctx.

(** * Lemmas about tight typing. *)

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma subtyp_or_tight G e S1 S2 T1 T2:
  G ⁏ e ⊢# T1 <:  S1->
  G ⁏ e ⊢# T2 <:  S2->
  G ⁏ e ⊢# (typ_or T1 T2) <:  (typ_or S1 S2).
Proof.
  intros; eauto.
Qed.
Lemma subtyp_and_tight G e S1 S2 T1 T2:
  G ⁏ e ⊢# T1 <:  S1->
  G ⁏ e ⊢# T2 <:  S2->
  G ⁏ e ⊢# (typ_and T1 T2) <:  (typ_and S1 S2).
Proof.
  intros; eauto.
Qed.
Lemma subtyp_or_swap_t G e T1 T2:
  G⁏e ⊢# (typ_or T1 T2) <: (typ_or T2 T1).
Proof.
  eauto.
Qed.
Lemma subtyp_and_swap_t G e T1 T2:
  G⁏e ⊢# (typ_and T1 T2) <: (typ_and T2 T1).
Proof.
  eauto.
Qed.
Lemma subtyp_distl_tight G e T1 T2 T3:
  G⁏e ⊢# typ_or (typ_and T1 T2) (typ_and T1 T3) <: typ_and T1 (typ_or T2 T3).
Proof.  
  eapply subtyp_or2_t; eapply subtyp_and_tight; eauto.
Qed.
Lemma subtyp_dist_and_or_l_tight G e T1 T2 T3:
  G⁏e ⊢# (typ_and (typ_or T1 T2) (typ_or T1 T3)) <: typ_or T1 (typ_and T2 T3).
Proof.
  eapply subtyp_trans_t.
  eapply subtyp_dist_t.
  eapply subtyp_trans_t.
  eapply subtyp_or_tight.
  eapply subtyp_and12_t.
  eapply subtyp_trans_t.
  simple apply subtyp_and_swap_t.
  simple apply subtyp_dist_t.
  simple apply subtyp_or2_t.
  simple apply subtyp_or11_t.
  simple apply subtyp_or_tight.
  simple apply subtyp_and12_t.
  sapply subtyp_and2_t.
  simple apply subtyp_and12_t.
  simple apply subtyp_and11_t.
Qed.
Lemma subtyp_dist_and_or_r_tight G e T1 T2 T3:
  G⁏e ⊢# (typ_and (typ_or T1 T3) (typ_or T2 T3)) <: typ_or (typ_and T1 T2) T3.
Proof.
  eapply subtyp_trans_t.
  eapply subtyp_and_tight.
  eapply subtyp_or_swap_t.
  eapply subtyp_or_swap_t.
  eapply subtyp_trans_t.
  eapply subtyp_dist_and_or_l_tight.
  eapply subtyp_or_swap_t.
Qed.

Lemma subtyp_oc_swap_or_l_tight G e oc U T:
  G ⁏ e ⊢# (orctx_typ oc (typ_or U T) ) <: typ_or (orctx_typ oc U) T.
Proof.
  reverts T U. induction oc; intros; simpls.
  - auto.
  - eapply subtyp_trans_t. eapply subtyp_or_tight. eapply IHoc. eapply subtyp_refl_t. eauto.
  - eapply subtyp_trans_t. eapply subtyp_or_tight.  eapply subtyp_refl_t.  eapply IHoc. eauto.
Qed.

Lemma subtyp_oc_swap_or_r_tight G e oc U T:
  G ⁏ e ⊢# (orctx_typ oc (typ_or U T) ) <: typ_or U  (orctx_typ oc  T).
Proof.
  reverts T U. induction oc; intros; simpls.
  - auto.
  - eapply subtyp_trans_t. eapply subtyp_or_tight. eapply IHoc. eapply subtyp_refl_t. eauto.
  - eapply subtyp_trans_t. eapply subtyp_or_tight.  eapply subtyp_refl_t.  eapply IHoc. eauto.
Qed.
Lemma subtyp_oc_or_swap_l_tight G e oc U T:
  G ⁏ e ⊢# typ_or (orctx_typ oc U) T <: (orctx_typ oc (typ_or U T) ).
Proof.
  reverts T U. induction oc; intros; simpls.
  - auto.
  - eapply subtyp_trans_t. 2: eapply subtyp_or_tight. 2: eapply IHoc. 2: eapply subtyp_refl_t. eauto.
  - eapply subtyp_trans_t. 2: eapply subtyp_or_tight. 2:  eapply subtyp_refl_t. 2:  eapply IHoc. eauto.
Qed.

Lemma subtyp_oc_or_swap_r_tight G e oc U T:
  G ⁏ e ⊢# typ_or U (orctx_typ oc T) <: (orctx_typ oc (typ_or U T) ).
Proof.
  reverts T U. induction oc; intros; simpls.
  - auto.
  - eapply subtyp_trans_t. 2:  eapply subtyp_or_tight. 2:  eapply IHoc.  2:  eapply subtyp_refl_t. eauto.
  - eapply subtyp_trans_t. 2:  eapply subtyp_or_tight.  2:  eapply subtyp_refl_t.  2:   eapply IHoc. eauto.
Qed.

Lemma subtyp_oc_tight G e oc U T:
  G ⁏ e ⊢# T <: U ->
  G ⁏ e ⊢# (orctx_typ oc T) <: (orctx_typ oc U ).
Proof.
  reverts T U. induction oc; intros; simpls.
  - auto.
  - eapply subtyp_or_tight; eauto.
  - eapply subtyp_or_tight; eauto.
Qed.




Lemma subtyp_oc_and_tight G e oc T U1 U2:
  G ⁏ e ⊢# T <: orctx_typ oc U1 ->
  G ⁏ e ⊢# T <: orctx_typ oc U2 ->
  G ⁏ e ⊢# T <: orctx_typ oc (typ_and U1 U2).
Proof.
  introv Hty1 Hty2.
  reverts T U1 U2.
  induction oc; intros; simpls.
  - auto.
  - forwards: IHoc T0 (typ_or U1 T) (typ_or U2 T).
    eapply subtyp_trans_t. 2: eapply subtyp_oc_or_swap_l_tight. eauto. 
    eapply subtyp_trans_t. 2: eapply subtyp_oc_or_swap_l_tight. eauto. 
    eapply subtyp_trans_t. 2: eapply subtyp_oc_swap_or_l_tight.
    eapply subtyp_trans_t. 2: eapply subtyp_oc_tight. eauto.
    eapply subtyp_dist_and_or_r_tight.
  - forwards: IHoc T0 (typ_or T U1) (typ_or T U2).
    eapply subtyp_trans_t. 2: eapply subtyp_oc_or_swap_r_tight. eauto. 
    eapply subtyp_trans_t. 2: eapply subtyp_oc_or_swap_r_tight. eauto. 
    eapply subtyp_trans_t. 2: eapply subtyp_oc_swap_or_r_tight.
    eapply subtyp_trans_t. 2: eapply subtyp_oc_tight. eauto.
    eapply subtyp_dist_and_or_l_tight.
Qed.

Lemma subtyp_lc_tight G e oc U T:
  G ⁏ e ⊢# T <: U ->
  G ⁏ e ⊢# (logctx_typ oc  T) <: (logctx_typ oc U ).
Proof.
  reverts T U. induction oc; intros; simpls.
  - auto.
  - eapply subtyp_or_tight; eauto.
  - eapply subtyp_or_tight; eauto.
  - eapply subtyp_and_tight; eauto.
  - eapply subtyp_and_tight; eauto.
Qed.


Lemma ty_tight_oc_dist G e oc T U:
  G ⁏ e ⊢# (typ_and (orctx_typ oc T) (orctx_typ oc U)) <: (orctx_typ oc (typ_and T U)).
Proof.
  reverts T U.
  induction oc; intros; simpls.
  - eapply subtyp_refl_t.
  - eapply subtyp_trans_t. eapply subtyp_dist_and_or_r_tight. eauto.
  - eapply subtyp_trans_t. eapply subtyp_dist_and_or_l_tight. eauto.
Qed.
Lemma subtyp_tight_lc G e oc T U:
  G ⁏ e ⊢# T <: U ->
  G ⁏ e ⊢# (logctx_typ oc T) <: (logctx_typ oc U).
Proof.
  reverts T U.
  induction oc; intros; simpls; eauto using subtyp_or_tight,subtyp_and_tight,subtyp_refl_t with nocore. (* Using nocore to speed it up *)
Qed.
Lemma ty_tight_lc_dist G e oc T U:
  G ⁏ e ⊢# (typ_and (logctx_typ oc T) (logctx_typ oc U)) <: (logctx_typ oc (typ_and T U)).
Proof.
  reverts T U.
  induction oc; intros; simpls.
  - eapply subtyp_refl_t.
  - eapply subtyp_trans_t. eapply subtyp_dist_and_or_r_tight. eauto.
  - eapply subtyp_trans_t. eapply subtyp_dist_and_or_l_tight. eauto.
  - eapply subtyp_and2_t.
    eapply subtyp_trans_t. 2: eapply IHoc. eapply subtyp_and_tight. eauto. eauto.
    eapply subtyp_trans_t. eapply subtyp_and11_t. eapply subtyp_and12_t.
  - eapply subtyp_and2_t.
    eapply subtyp_trans_t. eapply subtyp_and11_t. eapply subtyp_and11_t.
    eapply subtyp_trans_t. 2: eapply IHoc. eapply subtyp_and_tight. eauto. eauto.
Qed.

Lemma ty_tight_lc_and G e oc x T U:
  G⁏e ⊢# x ∶ (logctx_typ oc U) ->
  G⁏e ⊢# x ∶ (logctx_typ oc T) ->
  G⁏e ⊢# x ∶ (logctx_typ oc (typ_and T U)).
Proof.
  introv HtyT HtyU.
  reverts T U.
  induction oc; intros.
  -  simpls. eauto.
  - 
    forwards: ty_and_intro_t HtyU HtyT.
    eapply ty_sub_t. eapply H.
    eapply ty_tight_lc_dist.
  - 
    forwards: ty_and_intro_t HtyU HtyT.
    eapply ty_sub_t. eapply H.
    eapply ty_tight_lc_dist.
  - 
    forwards: ty_and_intro_t HtyU HtyT.
    eapply ty_sub_t. eapply H.
    eapply ty_tight_lc_dist.
  - 
    forwards: ty_and_intro_t HtyU HtyT.
    eapply ty_sub_t. eapply H.
    eapply ty_tight_lc_dist.
Qed.

Lemma ty_tight_lc_replace G e oc x T U:
  G⁏e ⊢# x ∶ T ->
  G⁏e ⊢# x ∶ (logctx_typ oc U) ->
  G⁏e ⊢# x ∶ (logctx_typ oc T).
Proof.
  introv HtyT HtyU.
  reverts T U.
  induction oc using logctx_reverse_ind; intros; simpls.
  - eauto.
  - rewrite logctx_extend_typ in *; simpls.
    forwards: IHoc T0; eauto.
  - rewrite logctx_extend_typ in *; simpls.
    forwards: IHoc T0; eauto.
  - rewrite logctx_extend_typ in *; simpls.
    forwards: IHoc T0. eauto. eauto.
    forwards: ty_tight_lc_and HtyU H.
    eapply ty_sub_t.
    eapply H0.
    eapply subtyp_tight_lc. eapply subtyp_and_tight; eauto.
  - rewrite logctx_extend_typ in *; simpls.
    forwards: IHoc T0. eauto. eauto.
    forwards: ty_tight_lc_and H HtyU.
    eapply ty_sub_t.
    eapply H0.
    eapply subtyp_tight_lc. eapply subtyp_and_tight; eauto.
Qed.
Lemma subtyp_top_mut_t G e TL1 TU1 :
  mode_has_mut_top typing_mode ->
  (forall x : var, G ⁏ e ⊢# TL1 /[-> x] <: typ_bot) -> 
  (forall x : var, G ⁏ e ⊢# typ_top <: TU1 /[-> x]) ->
  G ⁏ e ⊢# typ_top <: typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) TL1 TU1).
Proof.
  introv Hm Hsl Hsu.
  eapply subtyp_trans_t.
  -
  eapply subtyp_mut_top_t. auto.
  - eapply subtyp_typ_t; eauto.
Qed.
(* Used to automatically solve goals from ty_has_M. *)
Lemma subtyp_bot_tyq_t G e TL T:
  (forall x : var, G ⁏ e ⊢# TL /[-> x] <: typ_bot) ->
  (forall x : var, e ⊢ TL /[-> x] ≈ T /[-> x]) ->
  (forall x : var, G ⁏ e ⊢# T /[-> x] <: typ_bot).
Proof.
  intros.
  eapply subtyp_trans_t.
  eapply subtyp_tyq_t.
  symmetry. eapply H0. eauto.
Qed.
Lemma subtyp_top_tyq_t G e TL T:
  (forall x : var, G ⁏ e ⊢# typ_top <: TL /[-> x]) ->
  (forall x : var, e ⊢ TL /[-> x] ≈ T /[-> x]) ->
  (forall x : var, G ⁏ e ⊢# typ_top  <: T /[-> x]).
Proof.
  intros.
  eapply subtyp_trans_t.
  eapply H.
  eapply subtyp_tyq_t. eauto.
Qed.



End WithTypingMode.