(** remove printing ~ *)
(** remove printing mu *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Equivalence Indep.
Require Export TypingMode TypingContext.

Notation EnvProperty A B := (forall (e:renv) (v: A) (T: B), Prop).
Notation TypingProperty A B := (forall (G:ctx) (Go: objctx) (e:renv) (v: A) (T: B), Prop).

(** * Typing Rules *)

Class DecTyping (A : Set) := obj_dec_typing: ctx -> objctx -> renv -> A -> dec -> Prop.
Notation "G '§§' Go '⁏' e ⊢ t '∶d' T" := (obj_dec_typing G Go e t T) (at level 40, Go at level 39, e at level 39, t at level 59).
Class Typing (A : Set) := obj_typing: ctx -> objctx -> renv -> A -> typ -> Prop.
Notation "G '§§' Go '⁏' e ⊢ t '∶' T" := (obj_typing G Go e t T) (at level 40, Go at level 39, e at level 39, t at level 59).

Notation typing G e a t := (obj_typing G objctx_empty e a t).
Notation "G '⁏' e ⊢ t '∶' T" := (typing G e t T) (at level 40, e at level 39, t at level 59).
Notation dec_typing G e a d := (obj_dec_typing G objctx_empty e a d).
Notation "G '⁏' e ⊢ t '∶d' T" := (dec_typing G e t T) (at level 40, e at level 39, t at level 59).

Class Subtyping := subtyping: ctx -> objctx -> renv -> typ -> typ -> Prop.
Notation "G '⁏' e '⊢' T '<:' U" := (subtyping G objctx_empty e T U) (at level 40, e at level 39, T at level 59).
Hint Unfold Subtyping: core.

(** The 'ro' relation, which removes references capabilities *)
Class ReferenceIncapacitate := obj_ref_incap :
                                 ctx -> objctx -> renv -> typ -> typ -> Prop.
Notation "G '§§' Go '⁏' e  ⊢ T1 '∶icp' T2" := (obj_ref_incap G Go e T1 T2)
  (at level 40, Go at level 39, e at level 39, T1 at level 59, T2 at level 59).
Notation ref_incap G e T1 T2 := (obj_ref_incap G objctx_empty e T1 T2).
Notation "G '⁏' e ⊢ T1 '∶icp' T2" :=
  (ref_incap G e T1 T2)
    (at level 40, e at level 39, T1 at level 59, T2 at level 59).

(** The 'mu' relation, which extracts the bound of a capability *)
Class CapabilityBound (C : typ_refcap_label) :=
  obj_cap_bnd :
     ctx -> objctx -> renv -> typ -> typ -> Prop.
Notation "G '§§' Go '⁏' e  ⊢ T1 '∶cpb' ⟦ C ⟧  T2" := (@obj_cap_bnd C _ G Go e T1 T2)
  (at level 40, Go at level 39, e at level 39, T1 at level 59, T2 at level 59).
Notation cap_bnd C G e T1 T2 := (@obj_cap_bnd C _ G objctx_empty e T1 T2).
Notation "G '⁏' e ⊢ T1 '∶cpb' ⟦ C ⟧ T2" := (cap_bnd C G e T1 T2)
    (at level 40, e at level 39, T1 at level 59, T2 at level 59).

(* Tcap can use avar_b 0 to refer to the receiver of the type declaration *)
Notation typ_cap_adapt C x Tcap := (
  typ_cap_marker C (typ_or Tcap
    (typ_sel (avar_f x) (typ_label_refcap C) (avar_b 0))
  )).

Section WithTypingMode.
Context (typing_mode: Mode).

(* Part of Figure 5 in "Reference mutability for DOT" *)
(* Figure 9 in "Pure methods for roDOT" *)
(** ** Term typing [G ⊢ t: T] *)
Inductive ty_var : Typing var :=

(** [G(x) = T]  #<br>#
    [――――――――(VT-Var)]  #<br>#
    [G ⊢ x: T]  *)
| ty_var_var : forall G e x T,
    ctx_binds x T G ->
    G⁏e ⊢ x ∶ T

(** [G ⊢ x: T^x]   #<br>#
    [――――――――――――(VT-RecI)] #<br>#
    [G ⊢ x: μ(T)]     *)
| ty_rec_intro : forall G e x T,
    G⁏e ⊢ x ∶ open x T ->
    typ_indep_b 0 T ->
    (* NOTE: This is stated as it was originally for rodot, using icp. With mode_incap_nocap, it is equivalent to typ_nocap <: (open x T). See ty_incap_diag_nocap_sup and ty_incap_nocap. *)
    ((mode_has_incap typing_mode) -> (G⁏e ⊢ (open x T) ∶icp (open x T))) ->
    G⁏e ⊢ x ∶ typ_bnd T

(** [G ⊢ x: μ(T)] #<br>#
    [――――――――――――(VT-RecE)] #<br>#
    [G ⊢ x: T^x]   *)
| ty_rec_elim : forall G e x T,
    G⁏e ⊢ x ∶ typ_bnd T ->
    typ_indep_b 0 T ->
    G⁏e ⊢ x ∶ open x T

(** [G ⊢ x: T]     #<br>#
    [G ⊢ x: U]     #<br>#
    [――――――――――――(VT-AndI)] #<br>#
    [G ⊢ x: T ∧ U]     *)
| ty_var_and : forall G e x T U,
    G⁏e ⊢ x ∶ T ->
    G⁏e ⊢ x ∶ U ->
    G⁏e ⊢ x ∶ typ_and T U

(** [G ⊢ x: T]   #<br>#
    [G ⊢ T <: U] #<br>#
    [――――――――――(VT-Sub)] #<br>#
    [G ⊢ x: U]   *)
| ty_var_sub : forall G e x T U,
    G⁏e ⊢ x ∶ T ->
    G⁏e ⊢ T <: U ->
    G⁏e ⊢ x ∶ U

(** [G ⊢ x: T]   #<br>#
    [――――――――――(VT-MutTop)] #<br>#
    [G ⊢ x: {M:⊥..⊤}}]   *)
| ty_var_mut_top : forall G e x T,
    mode_has_mutability typing_mode ->
    G⁏e ⊢ x ∶ T ->
    G⁏e ⊢ x ∶ (typ_mut_marker typ_top)

(* Figure 4 in "Reference mutability for DOT" *)
(* Figure 10 in "Pure methods for roDOT" *)
(** ** Subtyping [G ⊢ T <: U] *)
with subtyp : Subtyping :=

(** [――――――――――(ST-Top)] #<br>#
    [G ⊢ T <: ⊤] *)
| subtyp_top: forall G e T,
    G⁏e ⊢ T <: typ_top

(** [――――――――――(ST-Bot)] #<br>#
    [G ⊢ ⊥ <: T] *)
| subtyp_bot: forall G e T,
    G⁏e ⊢ typ_bot <: T

(** [――――――――――(ST-Refl)] #<br>#
    [G ⊢ T <: T] *)
| subtyp_refl: forall G e T,
    G⁏e ⊢ T <: T

(** [G ⊢ S <: T]     #<br>#
    [G ⊢ T <: U]     #<br>#
    [――――――――――(ST-Trans)]     #<br>#
    [G ⊢ S <: U]         *)
| subtyp_trans: forall G e S T U,
    G⁏e ⊢ S <: T ->
    G⁏e ⊢ T <: U ->
    G⁏e ⊢ S <: U

(** [――――――――――(ST-And1)] #<br>#
    [G ⊢ T ∧ U <: T] *)
| subtyp_and11: forall G e T U,
    G⁏e ⊢ typ_and T U <: T

(** [――――――――――(ST-And2)] #<br>#
    [G ⊢ T ∧ U <: U] *)
| subtyp_and12: forall G e T U,
    G⁏e ⊢ typ_and T U <: U

(** [G ⊢ S <: T]       #<br>#
    [G ⊢ S <: U]       #<br>#
    [――――――――――――――(ST-And)]   #<br>#
    [G ⊢ S <: T ∧ U]          *)
| subtyp_and2: forall G e S T U,
    G⁏e ⊢ S <: T ->
    G⁏e ⊢ S <: U ->
    G⁏e ⊢ S <: typ_and T U

(** [――――――――――(ST-Or1)] #<br>#
    [G ⊢ T <: T ∨ U] *)
| subtyp_or11: forall G e T U,
    G⁏e ⊢ T <: typ_or T U

(** [――――――――――(ST-Or2)] #<br>#
    [G ⊢ U <: T ∨ U] *)
| subtyp_or12: forall G e T U,
    G⁏e ⊢ U <: typ_or T U

(** [G ⊢ S <: T]       #<br>#
    [G ⊢ U <: T]       #<br>#
    [――――――――――――――(ST-Or)]   #<br>#
    [G ⊢ S ∨ U <: T]          *)
| subtyp_or2: forall G e S T U,
    G⁏e ⊢ S <: T ->
    G⁏e ⊢ U <: T ->
    G⁏e ⊢ typ_or S U <: T


(** [G ⊢ T1 <: T2]           #<br>#
    [G ⊢ U1 <: U2]           #<br>#
    [――――――――――――――――――――(ST-Fld)] #<br>#
    [G ⊢ {a: T2..U1} <: {a: T1..U2}] *)
| subtyp_fld: forall G e a T1 T2 U1 U2,
    G⁏e ⊢ T1 <: T2 ->
    G⁏e ⊢ U1 <: U2 ->
    G⁏e ⊢ typ_rcd (dec_fld a T2 U1) <: typ_rcd (dec_fld a T1 U2)

(** [G ⊢ S2^x <: S1^x]                   #<br>#
    [G ⊢ T1^x <: T2^x]                   #<br>#
    [x fresh]                     #<br>#
    [――――――――――――――――――――――――――――――(ST-Typ)] #<br>#
    [G ⊢ {B: S1..T1} <: {B: S2..T2}]     *)
| subtyp_typ: forall G e A S1 T1 S2 T2,
    (forall x, G⁏e ⊢ (open x S2) <: (open x S1)) ->
    (forall x, G⁏e ⊢ (open x T1) <: (open x T2)) ->
    G⁏e ⊢ typ_rcd (dec_typ A S1 T1) <: typ_rcd (dec_typ A S2 T2)

(** [G ⊢ x: {A: S..T}] #<br>#
    [――――――――――――――――(ST-SelL)] #<br>#
    [G ⊢ S^r <: x.A(r)]     *)
| subtyp_sel2: forall G e x A y S T,
    G⁏e ⊢ x ∶ typ_rcd (dec_typ A S T) ->
    G⁏e ⊢ (open y S) <: typ_sel (avar_f x) A (avar_f y)

(** [G ⊢ x: {A: S..T}] #<br>#
    [――――――――――――――――(ST-SelU)] #<br>#
    [G ⊢ x.A(r) <: T^r]     *)
| subtyp_sel1: forall G e x A y S T,
    G⁏e ⊢ x ∶ typ_rcd (dec_typ A S T) ->
    G⁏e ⊢ typ_sel (avar_f x) A (avar_f y) <: (open y T)

(** [G ⊢ S2 <: S1]                #<br>#
    [G, x: S2 ⊢ T1^x <: T2^x]     #<br>#
    [x fresh]                     #<br>#
    [―――――――――――――――――――――――(ST-Fun)]     #<br>#
    [G ⊢ ∀(S1)T1 <: ∀(S2)T2]
This rule is disabled in rodot and rodot_sef typing modes. *)
| subtyp_fun: forall L G e S1 T1 S2 T2,
    mode_has_fun typing_mode ->
    G⁏e ⊢ S2 <: S1 ->
    (forall x, x \notin L ->
       G § x / v_par ~ S2 ⁏e ⊢ open x T1 <: open x T2) ->
    G⁏e ⊢ typ_fun S1 T1 <: typ_fun S2 T2

(** [G ⊢ TZ2 <: TZ1]                #<br>#
    [G, x: TZ2 ⊢ TC2^x <: TC1^x]     #<br>#
    [G, x: TZ2, y: TC2^x ⊢ TR1^x^y <: TR2^x^y]     #<br>#
    [x,y fresh]                     #<br>#
    [―――――――――――――――――――――――(ST-Met)]     #<br>#
    [G ⊢ {m(TC1, TZ1): TR1} <: {m(TC2, TZ2): TR2}}]      *)
| subtyp_met: forall L G e m TC1 TZ1 TR1 TC2 TZ2 TR2,
    G⁏e ⊢ TZ2 <: TZ1 ->
    (forall x, x \notin L -> (* x is the parameter *)
    G § x / v_par ~ TZ2 ⁏e ⊢ open x TC2 <: open x TC1) ->
    (* If typ_indep_b 0 TC2 then typ_indep_b 0 TC1 in modes with sef *)
    (mode_has_sef typing_mode -> typ_indep_b 0 TC2 -> typ_indep_b 0 TC1) ->
    (forall x, x \notin L ->
     forall y, y <> x -> y \notin L -> (* y is the receiver *)
    (ctx_met G x TZ2 y TC2) ⁏e ⊢ open_met x y TR1 <: open_met x y TR2) ->
    G⁏e ⊢ typ_rcd (dec_met m TC1 TZ1 TR1) <: typ_rcd (dec_met m TC2 TZ2 TR2)


(** [e ⊢ T1 ≈ T2] #<br>#
    [――――――――――――――――(ST-Eq)] #<br>#
    [G⁏e ⊢ T1 <: T2)]     *)
| subtyp_tyq: forall G e T1 T2,
    e ⊢ T1 ≈ T2 ->
    G⁏e ⊢ T1 <: T2

(** [――――――――――(ST-TypAnd)] #<br>#
    [G ⊢ {A: S1..T1} ∧ {A: S2..T2} <: {A: S1 ∨ S2 .. T1 ∧ T2}] *)
| subtyp_typ_and: forall G e A S1 T1 S2 T2,
    G⁏e ⊢
     typ_and
       (typ_rcd (dec_typ A S1 T1))
       (typ_rcd (dec_typ A S2 T2))
     <: typ_rcd (dec_typ A (typ_or S1 S2) (typ_and T1 T2))

(** [――――――――――(ST-Dist)] #<br>#
    [G ⊢ T1 ∧ (T2 ∨ T3) <: (T1 ∧ T2) ∨ (T1 ∧ T3)] *)
| subtyp_dist: forall G e T1 T2 T3,
    G⁏e ⊢
      typ_and T1 (typ_or T2 T3)
      <: typ_or (typ_and T1 T2) (typ_and T1 T3)

(** [――――――――――(ST-N-Rec)] #<br>#
    [G ⊢ N < μ(T)] *)
| subtyp_nocap_bnd G e T:
    G⁏e ⊢ typ_nocap <: typ_bnd T

(** [――――――――――(ST-N-Typ)] #<br>#
    [G ⊢ N <: {A: T1..T2}] *)
| subtyp_nocap_typ G e A T1 T2:
    G⁏e ⊢ typ_nocap <: typ_rcd (dec_typ (typ_label_normal A) T1 T2)

(** [――――――――――(ST-N-Fld)] #<br>#
    [G ⊢ N <: {a: T1..T2}] *)
| subtyp_nocap_fld G e a T1 T2:
    G⁏e ⊢ typ_nocap <: typ_rcd (dec_fld a T1 T2)

(** [――――――――――(ST-N-Met)] #<br>#
    [G ⊢ N <: {m(TC,TZ):TR}] *)
| subtyp_nocap_met G e a TC TZ TR:
    G⁏e ⊢ typ_nocap <: typ_rcd (dec_met a TC TZ TR)

(** [――――――――――(ST-N-Fun)] #<br>#
    [G ⊢ N <: ∀(TZ)TR]  #<br>#
    This rule is disabled in rodot_sef typing mode.  *)
| subtyp_nocap_fun G e TZ TR:
    mode_has_fun typing_mode ->
    G⁏e ⊢ typ_nocap <: typ_fun TZ TR

(* Part of Figure 4 in "Pure methods for roDOT" *)
(** [――――――――――(ST-NM)] #<br>#
    [G ⊢ ⊤ <: N ∨ {M: T1..T2}] *)
| subtyp_nocap_mut_top_boundless G e T1 T2:
    mode_has_top_nm_subtype_boundless typing_mode ->
    G⁏e ⊢ typ_top <: (typ_or typ_nocap (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) T1 T2)))

(** [G ⊢ ⊤ <: N ∨ {M: ⊥..⊥}]  #<br>#
This rule is disabled in rodot and rodot_sef typing modes. *)
| subtyp_nocap_mut_top_complement G e:
    mode_has_top_nm_complement typing_mode ->
    G⁏e ⊢ typ_top <: (typ_or typ_nocap (typ_mut_marker typ_bot) )

(** [G ⊢ N ∧ {M: ⊥..⊥} <: ⊥]  #<br>#
This rule is disabled in rodot and rodot_sef typing modes. *)
| subtyp_nocap_mut_bot G e:
    mode_has_top_nm_complement typing_mode ->
    G⁏e ⊢ (typ_and typ_nocap (typ_mut_marker typ_bot)) <: typ_bot

(** [G ⊢ ⊤ <: {M: ⊥..⊤}]  #<br>#
This rule is disabled in rodot and rodot_sef typing modes. *)
| subtyp_mut_top G e:
    mode_has_mut_top typing_mode ->
    G⁏e ⊢ typ_top <: (typ_mut_marker typ_top)

(* Part of Figure 7 in "Reference mutability for DOT" *)
(* Part of Figure 11 in "Pure methods for roDOT" *)
(** ** Removing reference capabilities [G;e ⊢ T1 :icp T2] *)
with ty_incap : ReferenceIncapacitate :=

(** [G ⊢ ⊤ ro ⊤] *)
| ty_incap_top G e:
    G⁏e ⊢ typ_top ∶icp typ_top

(** [G ⊢ ⊥ ro N] *)
| ty_incap_bot G e:
    G⁏e ⊢ typ_bot ∶icp typ_nocap

(** [G ⊢ {A: T1..T2} ro {A: T1..T2}] *)
| ty_incap_typ G e A S T:
    G⁏e ⊢ typ_rcd (dec_typ (typ_label_normal A) S T)
     ∶icp typ_rcd (dec_typ (typ_label_normal A) S T)

(** [G ⊢ {M: T1..T2} ro ⊤] *)
| ty_incap_cap G e M S T:
    G⁏e ⊢ typ_rcd (dec_typ (typ_label_refcap M) S T)
     ∶icp typ_top

(** [G ⊢ {m(TC, TZ): TR} ro ⊤] *)
| ty_incap_met G e m TC TZ TR:
    G⁏e ⊢ typ_rcd (dec_met m TC TZ TR) ∶icp typ_rcd (dec_met m TC TZ TR)

(** [G ⊢ {a: T1..T2} ro {a: T1..T2}] *)
| ty_incap_fld G e a T1 T2:
    G⁏e ⊢ typ_rcd (dec_fld a T1 T2) ∶icp typ_rcd (dec_fld a T1 T2)

(** 
    [G ⊢ x: {B: S..T}] #<br>#
    [G ⊢ T^y ro T'] #<br>#
    [――――――――――――――] #<br>#
    [G ⊢ x.B(y) ro T'] *)
| ty_incap_sel G e x B y S T T':
    G⁏e ⊢ x ∶ typ_rcd (dec_typ B S T) ->
    G⁏e ⊢ (open y T) ∶icp T' ->
    G⁏e ⊢ typ_sel (avar_f x) B (avar_f y) ∶icp T'

(** [G ⊢ μ(T) ro μ(T)] *)
| ty_incap_rec G e T:
    G⁏e ⊢ typ_bnd T ∶icp typ_bnd T

(** [G ⊢ S1 ro S2]       #<br>#
    [G ⊢ T1 ro T2]       #<br>#
    [――――――――――――――]   #<br>#
    [G ⊢ S1 ∧ T1 ro S2 ∧ T2] *)
| ty_incap_and G e S1 T1 S2 T2:
    G⁏e ⊢ S1 ∶icp S2 ->
    G⁏e ⊢ T1 ∶icp T2 ->
    G⁏e ⊢ typ_and S1 T1 ∶icp typ_and S2 T2

(** [G ⊢ S1 ro S2]       #<br>#
    [G ⊢ T1 ro T2]       #<br>#
    [――――――――――――――]   #<br>#
    [G ⊢ S1 ∨ T1 ro S2 ∨ T2] *)
| ty_incap_or G e S1 T1 S2 T2:
    G⁏e ⊢ S1 ∶icp S2 ->
    G⁏e ⊢ T1 ∶icp T2 ->
    G⁏e ⊢ typ_or S1 T1 ∶icp typ_or S2 T2

(** [G ⊢ ∀(S)T ro ∀(S)T]  #<br>#
This rule is disabled in rodot and rodot_sef typing modes. *)
| ty_incap_fun G e S T:
    mode_has_fun typing_mode ->
    G⁏e ⊢ typ_fun S T ∶icp typ_fun S T

(** [G ⊢ N <: T ]      #<br>#
    [――――――――――――――]   #<br>#
    [G ⊢ T ro T] *)
(* Part of Lemma 12 in "Pure methods for roDOT" *)
(* Part of Figure 4 in "Pure methods for roDOT" *)
| ty_incap_nocap G e T:
    mode_incap' typing_mode = mode_incap_nocap ->
    G⁏e ⊢ typ_nocap <: T ->
    G⁏e ⊢ T ∶icp T
.

Global Existing Instances ty_var subtyp ty_incap.

(* Part of Figure 7 in "Reference mutability for DOT" *)
(* Part of Figure 11 in "Pure methods for roDOT" *)
(* The result can refer to the receiver by avar_b 0 *)
Inductive ty_capbnd (C : typ_refcap_label) : CapabilityBound C :=
| ty_capbnd_top G e:
    G⁏e ⊢ typ_top ∶cpb⟦C⟧ typ_top

(** [G ⊢ ⊥ mu ⊥] *)
| ty_capbnd_bot G e:
    G⁏e ⊢ typ_bot ∶cpb⟦C⟧ typ_bot

(** [G ⊢ {A: T1..T2} mu ⊤] *)
| ty_capbnd_typ G e A S T:
    G⁏e ⊢ typ_rcd (dec_typ (typ_label_normal A) S T) ∶cpb⟦C⟧ typ_top

(** [G ⊢ {M: S..T} mu S] *)
| ty_capbnd_cap_sel G e S T:
    (* Here, T is not opened *)
    G⁏e ⊢ typ_rcd (dec_typ (typ_label_refcap C) S T) ∶cpb⟦C⟧ T

(** [G ⊢ {A: S..T} mu ⊤] *)
| ty_capbnd_cap_ignore G e C' S T:
    C' <> C ->
    G⁏e ⊢ typ_rcd (dec_typ (typ_label_refcap C') S T) ∶cpb⟦C⟧ typ_top

(** [G ⊢ {m(TC, TZ): TR} mu ⊤] *)
| ty_capbnd_met G e m TC TZ TR:
    G⁏e ⊢ typ_rcd (dec_met m TC TZ TR) ∶cpb⟦C⟧ typ_top

(** [G ⊢ {a: T1..T2} mu ⊤] *)
| ty_capbnd_fld G e a T1 T2:
    G⁏e ⊢ typ_rcd (dec_fld a T1 T2) ∶cpb⟦C⟧ typ_top
(** 
    [G ⊢ x: {B: S..T}] #<br>#
    [G ⊢ T^y mu T'] #<br>#
    [――――――――――――――] #<br>#
    [G ⊢ x.B(y) mu T'] *)
| ty_capbnd_sel G e x B y S T T':
    G⁏e ⊢ x ∶ typ_rcd (dec_typ B S T) ->
    G⁏e ⊢ (open y T) ∶cpb⟦C⟧ T' ->
    G⁏e ⊢ typ_sel (avar_f x) B (avar_f y) ∶cpb⟦C⟧ T'

(** [G ⊢ μ(T) mu ⊤] *)
| ty_capbnd_rec G e T:
    G⁏e ⊢ typ_bnd T ∶cpb⟦C⟧ typ_top

(** [G ⊢ S1 mu S2]       #<br>#
    [G ⊢ T1 mu T2]       #<br>#
    [――――――――――――――]   #<br>#
    [G ⊢ S1 ∧ T1 mu S2 ∧ T2] *)
| ty_capbnd_and G e S1 T1 S2 T2:
    G⁏e ⊢ S1 ∶cpb⟦C⟧ S2 ->
    G⁏e ⊢ T1 ∶cpb⟦C⟧ T2 ->
    G⁏e ⊢ typ_and S1 T1 ∶cpb⟦C⟧ typ_and S2 T2

(** [G ⊢ S1 mu S2]       #<br>#
    [G ⊢ T1 mu T2]       #<br>#
    [――――――――――――――]   #<br>#
    [G ⊢ S1 ∨ T1 mu S2 ∨ T2] *)
| ty_capbnd_or G e S1 T1 S2 T2:
    G⁏e ⊢ S1 ∶cpb⟦C⟧ S2 ->
    G⁏e ⊢ T1 ∶cpb⟦C⟧ T2 ->
    G⁏e ⊢ typ_or S1 T1 ∶cpb⟦C⟧ typ_or S2 T2

(** [G ⊢ ∀(S)T mu ⊤] #<br>#
This rule is disabled in rodot and rodot_sef typing modes. *)
| ty_capbnd_all G e S T:
    G⁏e ⊢ typ_fun S T ∶cpb⟦C⟧ typ_top

.

Global Existing Instances ty_capbnd.

(** ** Viewpoint adaptation *)

(** Viewpoint adapts a single reference capability in a type *)
Inductive viewpoint_adapt_cap: typ_refcap_label -> var -> (TypingProperty typ typ) :=
| viewpoint_adapt_cap_strip_combine C x G Go e T Tstrip Tcap:
    G§§Go⁏e ⊢ T ∶icp Tstrip ->
    G§§Go⁏e ⊢ T ∶cpb⟦C⟧ Tcap ->
    viewpoint_adapt_cap C x G Go e T (typ_and Tstrip (typ_cap_adapt C x Tcap)).

(** Viewpoint adapts the type of a field to a source reference *)
Definition viewpoint_adapt x G Go e Tf Tr :=
  match (mode_viewpoint_adaptation' typing_mode) with
  (* In basic mode, the type of the field is unchanged *)
  | mode_viewpoint_adaptation_same => Tf = Tr
  (* In rodot, the mutability is adapted *)
  | mode_viewpoint_adapt_mutability => viewpoint_adapt_cap typ_refcap_mutability x G Go e Tf Tr
  end.

End WithTypingMode.

Notation "G ⁏ e ⊢ x ▷⟦ C ⟧ T1 ⟶ T2" :=
  (viewpoint_adapt_cap _ C x G objctx_empty e T1 T2)
    (at level 40, e at level 39, x at level 59, T1 at level 59, T2 at level 59).

Notation "G ⁏ e ⊢ x ▷ T1 ⟶ T2" :=
  (viewpoint_adapt _ x G objctx_empty e T1 T2)
    (at level 40, e at level 39, x at level 59, T1 at level 59, T2 at level 59).

(* Notation for types of object literals, including the initial reference capabilities. *)
Notation typ_objlit typing_mode T := (
  match (mode_objlit_refcaps' typing_mode) with
  (* In basic mode, object reference has the same type as the literal,
     (intersected with top to make it syntactically similar to rodot, which may not be necessary) *)
  | mode_objlit_refcaps_none => typ_and (typ_bnd T) typ_top
  (* In rodot mode, object reference has a read-write type *)
  | mode_objlit_refcaps_mutable => typ_and (typ_bnd T) (typ_mut_marker typ_bot)
  end) (only parsing).


(* Is a variable visible?
   In basic modes, all variables are visible.
   In rodot, some variables are hidden *)
Notation vis_var typing_mode G x := (
  (mode_has_visibility typing_mode) -> (ctx_vis G x)
  ) (only parsing).

(* Is a variable writeable? *)
Notation mut_var_writeable typing_mode G e x := (
  (mode_has_mutability typing_mode) -> (G⁏e ⊢ x ∶ (typ_mut_marker typ_bot) )
  ) (only parsing).

Section WithTypingMode.
Context (typing_mode: Mode).

(* Part of Figure 5 in "Reference mutability for DOT" *)
(* Figure 8 in "Pure methods for roDOT" *)
(** ** Term typing [G ⊢ t: T] *)
Inductive ty_trm : Typing trm :=
(** [G ⊢ x: T]  #<br>#
    [G vis x]#<br>#
    [――――――――(TT-Var)]  #<br>#
    [G ⊢ v x: T]  *)
| ty_trm_var : forall G e x T,
    vis_var typing_mode G x ->
    G⁏e ⊢ x ∶ T ->
    G⁏e ⊢ trm_var (avar_f x) ∶ T
(** [G ⊢ x: ∀(S)T] #<br>#
    [G ⊢ z: S]     #<br>#
    [G vis x,z]#<br>#
    [――――――――――――(TT-Apply)] #<br>#
    [G ⊢ x z: T^z] #<br>#
    This rule is disabled in rodot and rodot_sef typing modes. *)
| ty_trm_apply : forall G e x z S T,
    mode_has_fun typing_mode ->
    vis_var typing_mode G x ->
    vis_var typing_mode G z ->
    G⁏e ⊢ x ∶ typ_fun S T ->
    G⁏e ⊢ z ∶ S ->
    G⁏e ⊢ trm_apply (avar_f x) (avar_f z) ∶ open z T

(** [G ⊢ x: {a: S..T}] #<br>#
    [G vis x]#<br>#
    [G ⊢ T ro T1] #<br>#
    [G ⊢ T mu T2] #<br>#
    [U = T1 ∧ {M: ⊥..(x.M ∨ T2)}] #<br>#
    [―――――――――――――(TT-Read)] #<br>#
    [G ⊢ x.a: U]        *)
| ty_trm_read : forall G e x a S T U,
    vis_var typing_mode G x ->
    G⁏e ⊢ x ∶ typ_rcd (dec_fld a S T) ->
    G⁏e ⊢ x ▷ T ⟶ U ->
    G⁏e ⊢ trm_read (avar_f x) a ∶ U

(** [G ⊢ x: {a: S..T}] #<br>#
    [G ⊢ y: S] #<br>#
    [G ⊢ x: {M: ⊥..⊥}] #<br>#
    [G vis x,y]#<br>#
    [―――――――――――――(TT-Write)] #<br>#
    [G ⊢ x.a = y: T]        *)
| ty_trm_write : forall G e x y a S T,
    vis_var typing_mode G x ->
    vis_var typing_mode G y ->
    mut_var_writeable typing_mode G e x ->
    G⁏e ⊢ x ∶ typ_rcd (dec_fld a S T) ->
    G⁏e ⊢ y ∶ S ->
    G⁏e ⊢ trm_write (avar_f x) a (avar_f y) ∶ T

(** [G ⊢ x: {m(TC,TZ): TR}] #<br>#
    [G ⊢ y: TZ] #<br>#
    [G ⊢ x: TC^y] #<br>#
    [G vis x,y]#<br>#
    [―――――――――――――(TT-Call)] #<br>#
    [G ⊢ x.m(y): TR^x^y]        *)
| ty_trm_call : forall G e x y m TC TZ TR,
    vis_var typing_mode G x ->
    vis_var typing_mode G y ->
    G⁏e ⊢ x ∶ typ_rcd (dec_met m TC TZ TR) -> (* x is the receiver *)
    G⁏e ⊢ y ∶ TZ -> (* y is the argument *)
    (mode_has_sef typing_mode -> typ_indep_b 0 TC) ->
    G⁏e ⊢ x ∶ (open y TC) ->
    G⁏e ⊢ trm_call (avar_f x) m (avar_f y) ∶ open_met y x TR

(** [G ⊢ t: T]          #<br>#
    [G, x: T ⊢ u^x: U]  #<br>#
    [x fresh]           #<br>#
    [―――――――――――――――――(TT-Let)] #<br>#
    [G ⊢ let t in u: U]     *)
| ty_trm_let : forall L G e t u T U,
    G⁏e ⊢ t ∶ T ->
    (forall x, x \notin L ->
      G § x / v_let ~ T ⁏e ⊢ open x u ∶ U) ->
    G⁏e ⊢ trm_let t u ∶ U

(** [G ⊢ l: T]          #<br>#
    [G, x: T ⊢ u^x: U]  #<br>#
    [x fresh]           #<br>#
    [―――――――――――――――――(TT-LetLit)] #<br>#
    [G ⊢ let l in u: U]     *)
| ty_trm_lit : forall L G e l u T U,
    G⁏e ⊢ l ∶ T ->
    (forall x, x \notin L ->
      G § x / v_let ~ T ⁏e ⊢ open x u ∶ U) ->
    G⁏e ⊢ trm_lit l u ∶ U

(** [G ⊢ t: T]   #<br>#
    [G ⊢ T <: U] #<br>#
    [――――――――――(TT-Sub)] #<br>#
    [G ⊢ t: U]   *)
| ty_trm_sub : forall G e (t : trm) T U,
    G⁏e ⊢ t ∶ T ->
    G⁏e ⊢ T <: U ->
    G⁏e ⊢ t ∶ U

with ty_lit : Typing lit :=
(** [G, x: T ⊢ t^x: U^x]     #<br>#
    [x fresh]                #<br>#
    [――――――――――――――――――――――] #<br>#
    [G ⊢ lambda(T)t: ∀(T)U] #<br>#
    This rule is disabled in rodot_sef typing mode. *)
| ty_lit_fun : forall L G e T t U,
    mode_has_fun typing_mode ->
    (forall x, x \notin L ->
      G § x / v_par ~ T ⁏e ⊢ open x t ∶ open x U) ->
    G⁏e ⊢ lit_fun T t ∶ typ_fun T U

(** [G, x: T ⊢ ds^x: T^x]     #<br>#
    [x fresh]                #<br>#
    [T indep]                #<br>#
    [――――――――――――――――――――――] #<br>#
    [G ⊢ ν(T)ds: T] #<br>#
The self-variable x is passed through [objctx_lit] to allow special handling in rule [ty_def_met].
*)
| ty_lit_obj : forall L G e T ds,
    (forall x, x \notin L -> G §§ (objctx_lit x T) ⁏e ⊢ (open x ds) ∶ (open x T)) ->
    typ_indep_b 0 T ->
    G⁏e ⊢ lit_obj T ds ∶ typ_objlit typing_mode T

(* Part of Figure 6 in "Reference mutability for DOT" *)
(* Part of Figure 12 in "Pure methods for roDOT" *)
(** ** Single-definition typing [G ⊢ d: D] *)
with ty_def : DecTyping def :=
(**  [―――――――――――――――――――(DT-Typ)] #<br>#
     [G ⊢ {A = T}: {A: T..T}]   *)
| ty_def_typ : forall L G Go e A T U,
    (forall x, x \notin L -> e ⊢ (open x T) ≈ (open x U)) ->
    G §§ Go⁏e ⊢ def_typ (typ_label_normal A) T ∶d dec_typ (typ_label_normal A) U U

(** [G ⊢ t: T]            #<br>#
    [―――――――――――――――――――(DT-Fld)] #<br>#
    [G ⊢ {a = t}: {a: T}] *)
| ty_def_fld : forall G Go e a t T,
    vis_var typing_mode (G §§& Go) t ->
    G §§& Go⁏e ⊢ t ∶ T ->
    G §§ Go⁏e ⊢ def_fld a (avar_f t) ∶d dec_fld a T T

(** [G, x:T^x, !, xz:TZ, xc: TC^z ∧ T^x ∧ T^xc ⊢ t^xz^xc: TR^xz^xc]            #<br>#
    [―――――――――――――――――――(DT-Met)] #<br>#
    [G, x:T^x ⊢ {m(xr, xz) = t}: {m(r: TC, z: TZ): TR}] *)
(* Warning: the receiver comes before the parameter in declarations and call,
   but after the parameter in the typing context and when opening the term and the result type.
  *)
| ty_def_met : forall L G Go e m t TC TZ TR,
    (forall xz, xz \notin L ->
    forall xc, xc <> xz -> xc \notin L ->
    ctx_obj_met typing_mode G Go xz TZ xc TC ⁏e ⊢ open_met xz xc t ∶ open_met xz xc TR) ->
    G §§ Go⁏e ⊢ def_met m t ∶d dec_met m TC TZ TR

(* Part of Figure 6 in "Reference mutability for DOT" *)
(* Part of Figure 12 in "Pure methods for roDOT" *)
(** ** Multiple-definition typing [G ⊢ ds :: T] *)
with ty_defs: Typing defs :=
(** [G ⊢ d: D]              #<br>#
    [―――――――――――――――――――――] #<br>#
    [G ⊢ d ++ defs_nil : D] *)
| ty_defs_one : forall G Go e d D,
    G §§ Go⁏e ⊢ d ∶d D ->
    G §§ Go⁏e ⊢ defs_cons defs_nil d ∶ typ_rcd D

(** [G ⊢ ds :: T]         #<br>#
    [G ⊢ d: D]            #<br>#
    [d \notin ds]         #<br>#
    [―――――――――――――――――――(DT-And)] #<br>#
    [G ⊢ ds ++ d : T /\ D] *)
| ty_defs_cons : forall G Go e ds d T D,
    G §§ Go⁏e ⊢ ds ∶ T ->
    G §§ Go⁏e ⊢ d ∶d D ->
    defs_hasnt ds (label_of d) ->
    G §§ Go⁏e ⊢ defs_cons ds d ∶ typ_and T (typ_rcd D)
.

Global Existing Instances ty_trm ty_def | 0.
Global Existing Instances ty_lit ty_defs | 1.

(** ** Mutual Induction Principles *)

Scheme ts_ty_var_mut := Induction for ty_var Sort Prop
with   ts_subtyp     := Induction for subtyp Sort Prop
with   ts_ty_incap   := Induction for ty_incap Sort Prop.
Combined Scheme ts_mutind from ts_ty_var_mut, ts_subtyp, ts_ty_incap.

Scheme rules_trm_mut    := Induction for ty_trm Sort Prop
with   rules_lit_mut    := Induction for ty_lit Sort Prop
with   rules_def_mut    := Induction for ty_def Sort Prop
with   rules_defs_mut   := Induction for ty_defs Sort Prop.
Combined Scheme rules_mutind from rules_trm_mut, rules_lit_mut,
rules_def_mut, rules_defs_mut.

End WithTypingMode.

Hint Constructors ty_var ty_trm ty_lit ty_def ty_defs subtyp
     ty_incap ty_capbnd : core.
Remove Hints subtyp_trans ty_var_sub ty_trm_sub : core.
Hint Resolve subtyp_trans ty_var_sub ty_trm_sub | 10 : core.
Hint Mode subtyping + + + + + + : core.

(** ** Tactics *)

(* Tactic used in Narrowing and Substitution to handle cases which add
   variables to the typing context *)
(* Note: application of typing rules with @ to _ is required, otherwise apply_fresh fails *)
Ltac fresh_constructor_extern :=
  match goal with
  | [ |- _⁏_ ⊢ lit_obj _ _ ∶ typ_objlit _ _ ] =>
    let z := fresh "z" in
    apply_fresh (@ty_lit_obj _) as z
  | [ |- _⁏_ ⊢ lit_fun _ _ ∶ typ_fun _ _ ] =>
    let z := fresh "z" in
    apply_fresh (@ty_lit_fun _) as z
  | [ |- _⁏_ ⊢ trm_let _ _ ∶ _ ] =>
    let z := fresh "z" in
    apply_fresh (@ty_trm_let _) as z
  | [ |- _⁏_ ⊢ typ_fun _ _ <: typ_fun _ _ ] =>
    let z := fresh "z" in
    apply_fresh (@subtyp_fun _) as z
  | [ |- _⁏_ ⊢ trm_lit _ _ ∶ _ ] =>
    let z := fresh "z" in
    apply_fresh (@ty_trm_lit _) as z
  | [ |- _⁏_ ⊢ typ_rcd (dec_met _ _ _ _) <: typ_rcd (dec_met _ _ _ _) ] =>
    let z := fresh "z" in
    apply_fresh (@subtyp_met _) as z
  end.

Hint Extern 4 (_⁏_ ⊢ _ ∶ _) => fresh_constructor_extern : core.

Ltac fresh_constructor :=
  fresh_constructor_extern; eauto.

(* For proving the case of field typing, do cases on the object context *)
Ltac objctx_destruct :=
  match goal with
  | H: (_ §§& ?Go) ⁏_ ⊢ _ ∶ _ |- _ §§ ?Go ⁏_ ⊢ _ ∶d _ => unfold ctx_obj in H; destruct Go
  end.

Section WithTypingMode.
Context (typing_mode: Mode).
(** * Lemmas *)

(** If a variable can be typed in an environment,
    then it is bound in that environment. *)
Lemma typing_implies_bound: forall G e (x: var) T,
  G⁏e ⊢ x ∶ T ->
  exists S, ctx_binds x S G.
Proof.
  introv Ht. remember (x).
  induction Ht; try congruence;
    match goal with
    | [ H : _ = _ |- _ ] => inversions H
    end; eauto.
Qed.

Lemma typing_implies_dom G e (x: var) T:
  G⁏e ⊢ x ∶ T ->
  x \in ctx_dom G.
Proof.
  introv Ht.
  destruct (typing_implies_bound Ht).
  eapply ctx_binds_dom.
  eauto.
Qed.

Lemma ctx_ty_ne x0 G1 G2 e z S:
  x0 §# G1 ->
  x0 §# G2 ->
  G1 §& G2 ⁏ e ⊢ z ∶ S ->
  x0 <> z.
Proof.
  introv Hn1 Hn2 Hty.
  assert (x0 §# (G1 §& G2)). eauto. apply typing_implies_bound in Hty. destruct Hty.
  intro; subst. eapply ctx_binds_unbound_false. 2: apply H. apply H0.
Qed.


Lemma var_typing_implies_avar_f: forall G e a T,
    G⁏e ⊢ (trm_var a) ∶ T ->
    exists x, a = avar_f x.
Proof.
  intros G e a T H.
  remember (trm_var a) as t eqn:Heqt.
  induction H; inversion Heqt; eauto.
Qed.

(** Some simple lemmas for automation *)
Lemma typing_ctx_assoc A (P: Typing A) G1 G2 G3 e (t : A) T:
   G1 §& (G2 §& G3) ⁏e ⊢ t ∶ T ->
   G1 §& G2 §& G3 ⁏e ⊢ t ∶ T.
Proof.
  intros; rewrite <- ctx_concat_assoc; auto.
Qed.

Lemma typing_ctx_assoc2 A (P: Typing A) G1 G2 v1 vk T1 e (t : A) T:
   G1 §& (G2 § v1 / vk ~ T1) ⁏e ⊢ t ∶ T ->
   G1 §& G2 § v1 / vk ~ T1 ⁏e ⊢ t ∶ T.
Proof.
  intros; rewrite <- ctx_concat_push; auto.
Qed.


Lemma dec_typing_ctx_assoc A (P: DecTyping A) G1 G2 G3 e (t : A) T:
   G1 §& (G2 §& G3) ⁏e ⊢ t ∶d T ->
   G1 §& G2 §& G3 ⁏e ⊢ t ∶d T.
Proof.
  intros; rewrite <- ctx_concat_assoc; auto.
Qed.

Lemma dec_typing_ctx_assoc2 A (P: DecTyping A) G1 G2 v1 vk T1 e (t : A) T:
   G1 §& (G2 § v1 / vk ~ T1) ⁏e ⊢ t ∶d T ->
   G1 §& G2 § v1 / vk ~ T1 ⁏e ⊢ t ∶d T.
Proof.
  intros; rewrite <- ctx_concat_push; auto.
Qed.


Lemma subtyp_ctx_assoc (ST: Subtyping) G1 G2 G3 e T U:
   G1 §& (G2 §& G3) ⁏e ⊢ T <: U ->
   G1 §& G2 §& G3 ⁏e ⊢ T <: U.
Proof.
  intros; rewrite <- ctx_concat_assoc; auto.
Qed.


Lemma subtyp_ctx_assoc2 (ST: Subtyping) G1 G2 v1 vk T1 e T U:
   G1 §& (G2 § v1 / vk ~ T1) ⁏e ⊢ T <: U ->
   G1 §& G2 § v1 / vk ~ T1 ⁏e ⊢ T <: U.
Proof.
  intros; rewrite <- ctx_concat_push; auto.
Qed.

(* Inverts typing of variable term *)
Lemma ty_trm_var_inv G e x T:
  G ⁏ e ⊢ trm_var (avar_f x) ∶ T ->
  G ⁏ e ⊢ x ∶ T.
Proof.
  introv Hty.
  remember (trm_var (avar_f x)) as t.
  induction Hty; inversions Heqt; eauto.
Qed.

(* Inverts typing of object literal *)
Lemma ty_trm_lit_obj_inv F1 e1 R d t T0:
  F1 ⁏ e1 ⊢ trm_lit (lit_obj R d) t ∶ T0 ->
  exists L0,
    (forall x0 : var, x0 \notin L0 -> F1 §§ objctx_lit x0 R ⁏ e1 ⊢ open x0 d ∶ open x0 R).
Proof.
  intros. remember (trm_lit (lit_obj R d) t). induction H; inversions Heqt0.
  - inversions H. exists L0. eauto.
  - eapply IHty_trm. auto.
Qed.

(* Part of Lemma 12 in "Pure methods for roDOT" *)
Lemma ty_incap_diag_nocap_sup G e S T:
  G⁏e ⊢ S ∶icp T ->
  G⁏e ⊢ typ_nocap <: T.
Proof.
  introv Hsu.
  induction Hsu; eauto.
Qed.

End WithTypingMode.

Hint Resolve typing_ctx_assoc | 7 : core.
Hint Resolve typing_ctx_assoc2 | 7 : core.
Hint Resolve dec_typing_ctx_assoc | 7 : core.
Hint Resolve dec_typing_ctx_assoc2 | 7 : core.
Hint Resolve subtyp_ctx_assoc | 7 : core.
Hint Resolve subtyp_ctx_assoc2 | 7 : core.

Notation SubtypingPropery := (ctx -> renv -> typ -> typ -> Prop) (only parsing).

(* Properties of typing common to GeneralTyping and TightTyping *)
Module GeneralizedSubtyping.
Class AndOrSubtyping (ST: SubtypingPropery) := {
  and11 G e T U: ST G e (typ_and T U) T;
  and12 G e T U: ST G e (typ_and T U) U;
  and2 G e S T U:
    ST G e S T ->
    ST G e S U ->
    ST G e S (typ_and T U);
  or11 G e T U: ST G e T (typ_or T U);
  or12 G e T U: ST G e U (typ_or T U);
  or2 G e S T U:
    ST G e S T ->
    ST G e U T ->
    ST G e (typ_or S U) T;
  trans G e S T U:
    ST G e S T ->
    ST G e T U ->
    ST G e S U;
}.

End GeneralizedSubtyping.

Section GeneralizedSubtypingLemmata.
Context {ST} {AOS: GeneralizedSubtyping.AndOrSubtyping ST}.

(* For some reason, `eapply subtyp_and` fails at unification.
  Instead, use `iapply subtyp_and`. Also seems to work: `applys subtyp_and`.
*)

Lemma subtyp_or G e T11 T12 T21 T22:
  ST G e T11 T21 ->
  ST G e T12 T22 ->
  ST G e (typ_or T11 T12) (typ_or T21 T22).
Proof.
  intros.
  apply GeneralizedSubtyping.or2.
  all: eapply GeneralizedSubtyping.trans; eauto.
  - apply GeneralizedSubtyping.or11.
  - apply GeneralizedSubtyping.or12.
Qed.

Lemma subtyp_and G e T11 T12 T21 T22:
  ST G e T11 T21 ->
  ST G e T12 T22 ->
  ST G e (typ_and T11 T12) (typ_and T21 T22).
Proof.
  intros.
  apply GeneralizedSubtyping.and2;
  eapply GeneralizedSubtyping.trans.
  2: eauto. apply GeneralizedSubtyping.and11.
  2: eauto. apply GeneralizedSubtyping.and12.
Qed.

End GeneralizedSubtypingLemmata.

Section WithTypingMode.
Context (typing_mode: Mode).

Global Instance and_or_subtyp: GeneralizedSubtyping.AndOrSubtyping (fun G => (subtyping)  G objctx_empty) := {
  and11 := subtyp_and11 typing_mode;
  and12 := subtyp_and12 typing_mode;
  and2 := @subtyp_and2 typing_mode;
  or11 := subtyp_or11 typing_mode;
  or12 := subtyp_or12 typing_mode;
  or2 := @subtyp_or2 typing_mode;
  trans := @subtyp_trans typing_mode;
}.

End WithTypingMode.
