(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Weakening EWeakening Substitution Narrowing Subenvironments.

Section WithTypingMode.
Context (typing_mode: Mode).

(* In rodot, the rhs of icp cannot be selection *)
(* It could be possible to have a better icp where this is allowed. *)
Lemma ty_incap_sel_false G e T v B v2:
  mode_incap' typing_mode = mode_incap_structural ->
  G ⁏ e ⊢ T ∶icp typ_sel v B v2 -> False.
Proof.
  introv Hm.
  (* State it a bit differently for the induction *)
  assert (
    forall T2,
    G ⁏ e ⊢ T ∶icp T2 -> (T2 <> typ_sel v B v2)
  ). {
    intros. induction H; simplify_eq; eauto.
    (* The nocap rule is not avaiable in structural mode *)
    rewrite Hm in H. discriminate.
  }
  intro.
  specialize (H _ H0).
  contradiction.
Qed.

Lemma ty_icp_diag_reopen_mode_structural G e x v2 T:
  mode_incap' typing_mode = mode_incap_structural ->
  G ⁏ e ⊢ open x T ∶icp open x T ->
  G ⁏ e ⊢ open v2 T ∶icp open v2 T.
Proof.
  introv Hm Hicp.
    induction T; sympls; inversions Hicp; sympls; eauto.
    + destruct d; sympls; inversions H0; eauto.
    + destruct d; sympls; inversions H0; eauto.
    + destruct d; sympls; inversions H0; eauto.
    + rewrite Hm in H; discriminate.
    + rewrite Hm in H; discriminate.
    + rewrite Hm in H; discriminate.
    + apply ty_incap_sel_false in H6. 2: eauto.
      contradiction.
    + rewrite Hm in H; discriminate.
    + rewrite Hm in H; discriminate.
Qed.

Lemma ty_icp_opened_tyq G e T1 x T T2:
  mode_incap' typing_mode = mode_incap_structural ->
  T1 = open x T ->
  G ⁏e ⊢ T1 ∶icp T1 ->
  e ⊢ T1 ≈ T2 ->
  G ⁏e ⊢ T2 ∶icp T2.
Proof.
  introv Hm HT1 Hicp1 Htyq.
  reverts T T2;
  induction T1; intros; subst; sympls; eauto;  try solve [inversions_tyq Htyq; eauto].
  destruct d; inversions_tyq Htyq; sympls; eauto. inversions Hicp1. eauto.
  rewrite Hm in H; discriminate.
  inversions Hicp1.
  destruct T; try discriminate; sympls.
  inversions HT1.

  inversions_tyq Htyq.   constructor.
  eapply IHT1_1; eauto.
  eapply IHT1_2; eauto.
  rewrite Hm in H; discriminate.

  inversions Hicp1.
  destruct T; try discriminate; sympls.
  inversions HT1.
  inversions_tyq Htyq.
  constructor.
  eapply IHT1_1; eauto.
  eapply IHT1_2; eauto.
  rewrite Hm in H; discriminate.

  eapply ty_incap_sel_false in Hicp1; eauto. contradiction.
  
  inversions_tyq Htyq. 
  destruct T; sympls; try discriminate.
   
  inversions HT1. inversions Hicp1. eauto.
  rewrite Hm in H; discriminate.
Qed.

Lemma ty_incap_subtyp G (e: renv) T U:
  G ⁏ e ⊢ T ∶icp U ->
  G ⁏ e ⊢ T <: U.
Proof.
  introv Hicp.
  induction Hicp; eauto.
  - eapply subtyp_trans. 2: eapply IHHicp; eauto. eapply subtyp_sel1. eauto.
  - eapply (@subtyp_and _ (and_or_subtyp _)); eauto.
Qed.

(* Simpler variant of ty_trm_read that works in every mode without case analysis *)
Lemma ty_trm_read_incap G e x a S T U Tc:
    vis_var typing_mode G x ->
    G⁏e ⊢ x ∶ typ_rcd (dec_fld a S T) ->
    G⁏e ⊢ T ∶icp U ->
    G ⁏ e ⊢ T ∶cpb ⟦ typ_refcap_mutability ⟧ Tc ->
    G⁏e ⊢ trm_read (avar_f x) a ∶ U.
Proof.
  introv Hvi Htyx Hicp Hcpb.
  destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hmva.
  - 
    eapply ty_trm_read.

    3: {hnf. cases_if. eauto. }
    eauto.
    simple eapply ty_var_sub.
    eauto.
    eapply subtyp_fld. eapply subtyp_refl. eapply ty_incap_subtyp.
    auto.
  - 
    eapply ty_trm_sub.
    eapply ty_trm_read.

    3: {hnf. cases_if. econstructor. eapply Hicp. eauto. }
    eauto.
    eauto.
    eapply subtyp_and11.
Qed.

(* Simpler variant of ty_trm_lit with ty_lit_obj that works in every mode without case analysis.
  The two parts are put together because lit typing does not have subsumption. *)
Lemma ty_trm_lit_obj_incap L G e T U u ds:
    ctx_ok G ->
    ctx_dom G \c L ->
    (forall x, x \notin L -> G §§ (objctx_lit x T) ⁏e ⊢ (open x ds) ∶ (open x T)) ->
    typ_indep_b 0 T ->
    (forall x, x \notin L -> G § x / v_let ~ (typ_bnd T) ⁏e ⊢ open x u ∶ U) ->
    G⁏e ⊢ trm_lit (lit_obj T ds) u ∶ U.
Proof.
  introv HinG HnxG Htyds HinT Htyu.

  eapply ty_trm_lit with (L:=L).
  eauto.
  intros.
  
  assert (forall vk T, ctx_ok (G § x / vk ~ T)). {
    constructor. auto. intro. apply H. apply HnxG. auto.
  }
  eapply narrow; eauto.
  cases_if; eauto.
Qed.



End WithTypingMode.
