Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses AbstractSyntax.

Require Import Vars Types Context Env Equivalence .
Require Import Coq.Program.Equality.

(* Type Independence *)

(* A type T is independent on variable v,
   when it contains no selections of special type members on
   variable v. *)


Inductive typ_indep: avar -> typ -> Prop :=
| typ_indep_nulary v T:
  typ_nulary T -> typ_indep v T

| typ_indep_and v t1 t2:
    typ_indep v t1->
    typ_indep v t2->
    typ_indep v (typ_and t1 t2)
| typ_indep_or v t1 t2:
    typ_indep v t1->
    typ_indep v t2->
    typ_indep v (typ_or t1 t2)
| typ_indep_fld v a t1 t2:
    typ_indep v t1->
    typ_indep v t2->
    typ_indep v (typ_rcd (dec_fld a t1 t2))
| typ_indep_typ_f v a t1 t2:
    typ_indep (avar_f v) t1->
    typ_indep (avar_f v) t2->
    typ_indep (avar_f v) (typ_rcd (dec_typ a t1 t2))
| typ_indep_typ_b n a t1 t2:
    typ_indep (avar_b (S n)) t1->
    typ_indep (avar_b (S n)) t2->
    typ_indep (avar_b n) (typ_rcd (dec_typ a t1 t2))
| typ_indep_fun_f v t1 t2:
    typ_indep (avar_f v) t1 ->
    typ_indep (avar_f v) t2 ->
    typ_indep (avar_f v) (typ_fun t1 t2)
| typ_indep_fun_b n t1 t2:
    typ_indep (avar_b n) t1 ->
    typ_indep (avar_b (S n)) t2 ->
    typ_indep (avar_b n) (typ_fun t1 t2)
| typ_indep_sel(v v1 v2: avar)(l: typ_label):
    v <> v1 ->
    v <> v2 ->
    typ_indep v (typ_sel v1 l v2)
| typ_indep_sel_normal(v v1 v2: avar)(l: typ_normal_label):
    v <> v2 ->
    typ_indep v (typ_sel v1 (typ_label_normal l) v2)
| typ_indep_bnd_f v t:
    typ_indep (avar_f v) t ->
    typ_indep (avar_f v) (typ_bnd t)
| typ_indep_bnd_b n t:
    typ_indep (avar_b (S n)) t ->
    typ_indep (avar_b n) (typ_bnd t)
| typ_indep_met_f v t1 t2 t3 l:
    typ_indep (avar_f v) t1 ->
    typ_indep (avar_f v) t2 ->
    typ_indep (avar_f v) t3 ->
    typ_indep (avar_f v) (typ_rcd (dec_met l t1 t2 t3))
| typ_indep_met_b n t1 t2 t3 l:
    typ_indep (avar_b (S n)) t1 ->
    typ_indep (avar_b n) t2 ->
    typ_indep (avar_b (S (S n))) t3 ->
    typ_indep (avar_b n) (typ_rcd (dec_met l t1 t2 t3))
.
Hint Constructors typ_indep : core.

Notation typ_indep_f v T := (typ_indep (avar_f v) T).
Notation typ_indep_b n T := (typ_indep (avar_b n) T).

(* INDEPENDENCE PROPERTIES *)

Lemma typ_indep_bnd_b_inv n t:
  typ_indep (avar_b n) (typ_bnd t) ->
  typ_indep (avar_b (S n)) t.
Proof.
  intros. inversions H.
  - inversions H0.
  - assumption.
Qed.

Lemma typ_indep_open n m x T:
  typ_indep_b n T ->
  typ_indep_b n (open_rec m x T).
Proof.
  introv Hin.
  revert dependent m.
  dependent induction Hin; subst; intros.
  inversions H; simpls; eauto.
  all: simpls; eauto.
  - eapply typ_indep_sel.
    destruct v1; simpls; eauto.
    cases_if; eauto. discriminate.
    destruct v2; simpls; eauto.
    cases_if; eauto. discriminate.
  -
    eapply typ_indep_sel_normal.
    destruct v2; simpls; eauto.
    cases_if; eauto. discriminate.
Qed.

Lemma typ_indep_seal n m x T:
  typ_indep_b n T ->
  typ_indep_b n (seal_rec m x T).
Proof.
  introv Hin.
  revert dependent m.
  dependent induction Hin; subst; intros.
  inversions H; simpls; eauto.
  all: simpls; eauto.
  - eapply typ_indep_sel.
    destruct v1; simpls; eauto.
    cases_if; eauto. discriminate.
    destruct v2; simpls; eauto.
    cases_if; eauto. discriminate.
  -
    eapply typ_indep_sel_normal.
    destruct v2; simpls; eauto.
    cases_if; eauto. discriminate.
Qed.

Lemma sel_rec_avar_b_lt m x n:
  n < m ->
  seal_rec m x (avar_b n) = (avar_b n).
Proof.
  introv Hle.
simpls.
cases_if. eapply Lt.le_not_lt in C. contradiction.
auto.
Qed.


Lemma typ_indep_seal_inv n m x T:
  n < m ->
  typ_indep_b n (seal_rec m x T) ->
  typ_indep_b n T.
Proof.
  introv Hle Hin.
  remember (seal_rec m x T).
  remember (avar_b n).
  reverts T m n.
  induction Hin; intros; subst; sympls; try inversions Heqa.
  - destruct T0; sympls; inversions H; eauto.
  - destruct T; sympls; inversions Heqt. eauto.
  - destruct T; sympls; inversions Heqt. eauto.
  - destruct T; sympls; inversions Heqt. destruct d; sympls; inversions H0. eauto.
  - destruct T; sympls; inversions Heqt. destruct d; sympls; inversions H0. eauto.
    eapply typ_indep_typ_b. eapply IHHin1. eauto. 2: eauto. apply -> Nat.succ_lt_mono; eauto.
     eapply IHHin2.  eauto. 2: eauto. apply -> Nat.succ_lt_mono; eauto.

  - destruct T; sympls; inversions Heqt.
    eapply typ_indep_fun_b;  eauto.
    eapply IHHin2.  eauto. 2: eauto. apply -> Nat.succ_lt_mono; eauto.
  -  destruct T; sympls; inversions Heqt.
    eapply typ_indep_sel; intro; intros; subst. rewrite sel_rec_avar_b_lt in H by auto. contradiction.
    rewrite sel_rec_avar_b_lt in H0 by auto. contradiction.
  -
    destruct T; sympls; inversions Heqt.
    eapply typ_indep_sel_normal; intro; intros; subst. rewrite sel_rec_avar_b_lt in H by auto. contradiction.
  -
    destruct T; sympls; inversions Heqt.
    eapply typ_indep_bnd_b. eapply IHHin. 3: eauto. eauto. apply -> Nat.succ_lt_mono; eauto.
  -
    destruct T; sympls; inversions Heqt.
    destruct d; sympls; inversions H0.
    eapply typ_indep_met_b.
    eapply IHHin1. 3: eauto. eauto. apply -> Nat.succ_lt_mono; eauto.
    eapply IHHin2. 3: eauto. eauto. auto.
    eapply IHHin3. 3: eauto. eauto. apply -> Nat.succ_lt_mono; eauto.  apply -> Nat.succ_lt_mono; eauto.
Qed.





(* Opening with equivalent variables at indep places preserves equivalence *)

Lemma tyq_typ_open_rec_indep : forall e t1 t2 n y1 y2,
  typ_indep (avar_b n) t1 ->
  typ_indep (avar_b n) t2 -> (* this one is not required, see typ_indep_tyq_indep *)
  typing_equiv e y1 y2 ->
  (forall x, typing_equiv e (open_rec n x t1) (open_rec n x t2)) ->
  typing_equiv e (open_rec n y1 t1) (open_rec n y2 t2).
Proof.
  introv Hindep1 Hindep2 Htyqvar Htyqtyp.
  pick_fresh x.
  specialize (Htyqtyp x).
  remember (open_rec n x t1).
  remember (open_rec n x t2).
  reverts n t1 t2.
  induction Htyqtyp; intros; subst; sympls.
  - inversions H; destruct t1; simpls; try congruence; destruct t2; simpls; try congruence; eauto.
  - inversions H; destruct t1; sympls; try congruence; destruct t2; sympls; try congruence.
    all: inversions Heqt0.
    all: inversions Heqt.
    + inversions_tyq Hindep1.
      inversions_tyq Hindep2.
      eapply tyq_typ_andor. eauto.
      eapply IHHtyqtyp; eauto.
      eapply IHHtyqtyp0; eauto.
    + inversions_tyq Hindep1.
      inversions_tyq Hindep2.
      eapply tyq_typ_andor. eauto.
      eapply IHHtyqtyp; eauto.
      eapply IHHtyqtyp0; eauto.
  - destruct t1; sympls; try congruence; destruct t2; sympls; try congruence.
    inversions_tyq Hindep1.
    inversions_tyq Hindep2.
    inversions Heqt0; inversions Heqt.
    eapply tyq_typ_fun.
    eapply IHHtyqtyp; eauto.
    intros.

    pick_fresh z.
    eapply tyq_typ_reopen with (x:=z).
    notin_solve; eapply notin_fv_open; eauto.

    rewrite (open_commut z).
    rewrite (open_commut z).
    eapply H1; eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z); eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z); eauto.
    notin_solve; eapply notin_fv_open; eauto.
    discriminate.
    discriminate.
  - destruct t1; sympls; try congruence; destruct t2; sympls; try congruence.
    inversions Heqt. inversions Heqt0.
    destruct d; sympls; try congruence; destruct d0; sympls; try congruence.
    inversions H2. inversions H3.
    inversions_tyq Hindep1.
    inversions_tyq Hindep2.
    eauto.
  - destruct t1; sympls; try congruence; destruct t2; sympls; try congruence.
    inversions Heqt. inversions Heqt0.
    destruct d; sympls; try congruence; destruct d0; sympls; try congruence.
    inversions H4. inversions H5.
    inversions_tyq Hindep1.
    inversions_tyq Hindep2.
    eapply tyq_typ_typ.

    intros.
    pick_fresh z.
    eapply tyq_typ_reopen with (x:=z).
    notin_solve; eapply notin_fv_open; eauto.

    rewrite (open_commut z) by discriminate.
    rewrite (open_commut z) by discriminate.

    eapply H0; eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z); eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z); eauto.
    notin_solve; eapply notin_fv_open; eauto.

    intros.
    pick_fresh z.
    eapply tyq_typ_reopen with (x:=z).
    notin_solve; eapply notin_fv_open; eauto.

    rewrite (open_commut z) by discriminate.
    rewrite (open_commut z) by discriminate.

    eapply H2; eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z); eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z); eauto.
    notin_solve; eapply notin_fv_open; eauto.
  - (* sel normal *)
    destruct t1; sympls; try congruence; destruct t2; sympls; try congruence.
    inversions Heqt. inversions Heqt0.
    (* The receiver is equivalent *)
    forwards w1 w2 Hw1 Hw2 Htyqw: tyq_var_reopen_tyq_rec. 2: eapply H. 2: eapply Htyqvar. 2: eauto. 2: eauto.
    notin_solve.
    rewrite <- Hw1.
    rewrite <- Hw2.
    (* The argument cannot be affected by opening because of indep *)
    assert((open_rec n y1 a0) = (open_rec n y2 a2)) as Heqy. {
      destruct (open_rec_avar_cases n a0 a2 H4). notin_solve.
      + destructc H0 as q Ha0 Ha2. subst. sympls. auto.
      + destructc H0 as q Ha0 Ha2. subst. sympls. cases_if. inversions_tyq Hindep1; contradiction. auto.
    }
    rewrite Heqy.
    sympls.
    eapply tyq_typ_sel_normal; eauto.
  - (* sel general *)
    destruct t1; sympls; try congruence; destruct t2; sympls; try congruence.
    inversions Heqt. inversions Heqt0.

    (* argument *)
    assert((open_rec n y1 a0) = (open_rec n y2 a2)) as Heqy. {
      destruct (open_rec_avar_cases n a0 a2 H2). notin_solve.
      + destructc H as q Ha0 Ha2. subst. sympls. auto.
      + destructc H as q Ha0 Ha2. subst. sympls. cases_if. inversions_tyq Hindep1; contradiction. auto.
    }
    rewrite Heqy.
    (* receiver *)
    assert((open_rec n y1 a) = (open_rec n y2 a1) \/ (exists l, t0 = typ_label_normal l) /\ (open_rec n y1 a) = (avar_f y1) /\ (open_rec n y2 a1) = (avar_f y2)) as Heqy0. {
      destruct (open_rec_avar_cases n a a1 H0). notin_solve.
      + destructc H as q Ha Ha1. subst. sympls. auto.
      + destructc H as q Ha Ha1. subst. sympls. cases_if. inversions_tyq Hindep1. contradiction.  inversions_tyq Hindep2. contradiction.  right; eauto. auto.
    }
    destruct Heqy0.
    + (* same receiver *)
      rewrite H.
      eapply tyq_typ_sel.
    + (* normal label *)
      destruct_all; subst.
      rewrite H1.
      rewrite H3.
      eapply tyq_typ_sel_normal. auto.
  - (* bnd *)
    destruct t0; sympls; try congruence; destruct t3; sympls; try congruence.
    inversions Heqt. inversions Heqt0.
    inversions_tyq Hindep1.
    inversions_tyq Hindep2.
    eapply tyq_typ_bnd.
    intros.
    pick_fresh z.
    eapply tyq_typ_reopen with (x:=z).
    notin_solve; eapply notin_fv_open; eauto.

    rewrite (open_commut z) by discriminate.
    rewrite (open_commut z) by discriminate.

    eapply H0; eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z); eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z); eauto.
    notin_solve; eapply notin_fv_open; eauto.
  - (* met *)
    destruct t1; sympls; try congruence; destruct t2; sympls; try congruence.
    inversions Heqt. inversions Heqt0.
    destruct d; sympls; try congruence; destruct d0; sympls; try congruence.
    inversions H5. inversions H6.
    inversions_tyq Hindep1.
    inversions_tyq Hindep2.
    eapply tyq_typ_met.
    +
    intros.
    pick_fresh z.
    eapply tyq_typ_reopen with (x:=z).
    notin_solve; eapply notin_fv_open; eauto.

    rewrite (open_commut z) by discriminate.
    rewrite (open_commut z) by discriminate.

    eapply H0; eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z); eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z); eauto.
    notin_solve; eapply notin_fv_open; eauto.
   + eauto.
   +    intros.
    pick_fresh z1. pick_fresh z2.
    rewrite <- (open_vars_single 1).
    rewrite open_vars_push.
    rewrite <- (open_vars_single 1).
    rewrite open_vars_push.
    eapply tyq_typ_reopen_vars with (xs:=(z1::z2::nil)%list).
    fresh_solve; auto; eapply notin_fv_open; eauto.

    rewrite <- open_vars_push.
    rewrite <- open_vars_push.
    rewrite <- open_vars_push.
    rewrite <- open_vars_push. sympls.

    rewrite (open_commut z2) by discriminate.
    rewrite (open_commut z1) by discriminate.
    rewrite (open_commut z2) by discriminate.
    rewrite (open_commut z1 y2) by discriminate.
    eapply (H3 z1 z2).
    eauto.
    eapply typ_indep_open; eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z2) by discriminate.
    rewrite (open_commut z1) by discriminate. eauto.
    eapply typ_indep_open; eauto.
    eapply typ_indep_open; eauto.
    rewrite (open_commut z2) by discriminate.
    rewrite (open_commut z1) by discriminate. eauto.
    notin_solve; eapply notin_fv_open; eauto; eapply notin_fv_open; eauto.
Qed.
Hint Resolve tyq_typ_open_rec_indep : core.

Lemma tyq_typ_open_indep_same: forall e t y1 y2,
  typ_indep_b 0 t ->
  typing_equiv e y1 y2 ->
  typing_equiv e (open y1 t) (open y2 t).
Proof.
  intros.
  apply tyq_typ_open_rec_indep; auto.
Qed.

(* Equivalence preserves independence *)


Lemma typ_indep_tyq_open_vars_indep e t1 t2 n xs:
  typ_indep (avar_b n) t1 ->
  fresh ((fv t1) \u (fv t2) \u fv_renv e) (length xs) xs ->
  typing_equiv e (open_rec_vars 0 xs t1) (open_rec_vars 0 xs t2) ->
  typ_indep (avar_b n) t2.
Proof.
  introv Hindep Hfr Htyq.
  remember (open_rec_vars 0 xs t1).
  remember (open_rec_vars 0 xs t2).
  reverts t1 t2 n .
  reverts xs.
  induction Htyq; intros; subst; sympls; eauto.
  - inversions H; inversions_tyq H1; rewrite Heqt0 in H0; invert_open; eauto.
  - inversions H; invert_open; invert_open.
    + inversions_tyq Hindep. eapply typ_indep_and. eapply (IHHtyq xs); eauto; sympls; eauto. eapply (IHHtyq0 xs); eauto; sympls; eauto.
    + inversions_tyq Hindep. eapply typ_indep_or. eapply (IHHtyq xs); eauto; sympls; eauto. eapply (IHHtyq0 xs); eauto; sympls; eauto.
  - invert_open; invert_open. inversions_tyq Hindep; sympls. eapply typ_indep_fun_b; eauto.
    pick_fresh_with z (from_list xs).
    eapply (H1 z (z::xs)%list); eauto.
    1, 3: rewrite open_vars_push; auto.
    eapply fresh_push. fresh_solve. notin_solve.
  - invert_open; invert_open. inversions_tyq Hindep. sympls. eauto.
  - (* typ *)
    invert_open; invert_open. inversions_tyq Hindep. sympls.
    eapply typ_indep_typ_b.
    pick_fresh_with z (from_list xs).
    eapply (H0 z (z::xs)%list).
    4: eapply H6.
    1, 3: rewrite open_vars_push; auto.
    eapply fresh_push. fresh_solve. notin_solve.
    pick_fresh_with z (from_list xs).
    eapply (H2 z (z::xs)%list).
    4: eapply H8.
    1, 3: rewrite open_vars_push; auto.
    eapply fresh_push. fresh_solve. notin_solve.
  - (* sel normal *)
    invert_open; invert_open.
    eapply typ_indep_sel_normal.
    intro. subst. clear H2 H1.
    destruct (open_vars_avar_cases 0 xs (avar_b n) x2 H3); subst; sympls.
    fresh_solve.
    + destruct_all; subst. discriminate H0.
    + destruct_all; subst. inversions H0.
      inversions_tyq Hindep; contradiction.

  - invert_open; invert_open.
    assert(avar_b n <> x0). {
    intro. subst.
      destruct (open_vars_avar_cases 0 xs (avar_b n) x2 H1); subst; sympls.
      fresh_solve.
      + destruct_all; subst. discriminate H.
      + destruct_all; subst. inversions H.
        inversions_tyq Hindep; contradiction.
    }
    assert(avar_b n <> x \/ (exists ln, l= typ_label_normal ln)). {
      destruct (open_vars_avar_cases 0 xs x x1 H0); subst; sympls.
      fresh_solve.
      + destruct_all; subst. left. discriminate.
      + destruct_all; subst. inversions_tyq Hindep. auto.
        right; exists l0; auto.
    }
    destruct H2. eauto. destruct H2; subst; eauto.
  - (* bnd *)
    invert_open; invert_open.
    inversions_tyq Hindep. sympls.
    eapply typ_indep_bnd_b.
    pick_fresh_with z (from_list xs).
    eapply (H0 z (z::xs)%list).
    4: eapply H3.
    1, 3: rewrite open_vars_push; auto.
    eapply fresh_push. fresh_solve. notin_solve.
  - (* met *)
    invert_open; invert_open.
    inversions_tyq Hindep. sympls.
    eapply typ_indep_met_b.
    +
      pick_fresh_with z (from_list xs).
      eapply (H0 z (z::xs)%list).
      4: eapply H8.
      1, 3: rewrite open_vars_push; auto.
      eapply fresh_push. fresh_solve. notin_solve.
    + eauto.
    + pick_fresh_with z2 (from_list xs). pick_fresh_with z1 (from_list xs).
      eapply (H3 z1 z2 (z1::z2::xs)%list).
      4: eapply H11.
      rewrite open_vars_push; rewrite open_vars_push; auto.
      2: rewrite open_vars_push; rewrite open_vars_push; auto.
      eapply fresh_push. eapply fresh_push. fresh_solve. notin_solve. rewrite from_list_cons. notin_solve.
Qed.

Lemma typ_indep_tyq_open_indep e t1 t2 n xs:
  typ_indep (avar_b n) t1 ->
  xs \notin ((fv t1) \u (fv t2) \u fv_renv e)->
  typing_equiv e (open xs t1) (open xs t2) ->
  typ_indep (avar_b n) t2.
Proof.
  introv Hindep Hn Htyq.
  rewrite <- open_vars_single in Htyq.
  rewrite <- open_vars_single in Htyq.
  eapply typ_indep_tyq_open_vars_indep; eauto.
  fresh_solve; eauto.
Qed.
Lemma typ_indep_tyq_open_all_indep e t1 t2 n:
  typ_indep (avar_b n) t1 ->
  (forall xs, typing_equiv e (open xs t1) (open xs t2)) ->
  typ_indep (avar_b n) t2.
Proof.
  introv Hindep Htyq.
  pick_fresh z.
  specialize (Htyq z).
  eapply typ_indep_tyq_open_indep; eauto. notin_solve.
Qed.
Lemma typ_indep_tyq_indep e t1 t2 n:
  typ_indep (avar_b n) t1 ->
  typing_equiv e t1 t2 ->
  typ_indep (avar_b n) t2.
Proof.
  introv Hindep Htyq.
  rewrite <- (open_vars_nil n t1) in Htyq.
  rewrite <- (open_vars_nil n t2) in Htyq.
  eapply typ_indep_tyq_open_vars_indep with (xs:=nil); eauto.
Qed.

Lemma tyq_typ_open_indep_var e t1 t2 y1 y2:
  typ_indep_b 0 t1 ->
  (forall x, typing_equiv e (open x t1) (open x t2)) ->
  typing_equiv e y1 y2 ->
  typing_equiv e (open y1 t1) (open y2 t2).
Proof.
  introv Hin Htyqt Htyqy.
  eapply tyq_typ_open_rec_indep; eauto.
  pick_fresh z. specialize (Htyqt z).
  forwards: typ_indep_tyq_open_indep; eauto. notin_solve.
Qed.

Lemma subst_indep_b v T x y:
  typ_indep_b v T ->
  typ_indep_b v (subst_var x y T).
Proof.
  introv H.
  assert (exists v1, (avar_b v) = (avar_b v1)).
  exists v; auto.
  induction H.
  all: sympl in *.
  all: eauto.
  - destruct H; sympl; auto.
  - destruct H0; subst.
    inversions H0.
    assert (exists v1, (avar_b (S x0)) = (avar_b v1)).
    exists (S x0); auto. eauto.
  - destruct H0; subst.
    destruct (subst_cases2_avar x y v1) as [[? ?]|[? ?]].
    all: destruct (subst_cases2_avar x y v2) as [[? ?]|[? ?]].
    all: rewrite H2; rewrite H4.
    all: apply typ_indep_sel.
    all: try simplify_eq; auto.
  - destruct H0; subst.
    destruct (subst_cases2_avar x y v1) as [[? ?]|[? ?]].
    all: destruct (subst_cases2_avar x y v2) as [[? ?]|[? ?]].
    all: rewrite H1; rewrite H3.
    all: apply typ_indep_sel_normal.
    all: try simplify_eq; auto.
  - apply typ_indep_met_b;
    eauto.
Qed.
Hint Resolve subst_indep_b: core.

Lemma subst_vars_indep_b v T x y:
  length x = length y ->
  typ_indep_b v T ->
  typ_indep_b v (subst_vars x y T).
Proof.
  introv Hl Hindep.
  reverts y.
  induction x; intros; destruct y; simpls; try discriminate; eauto.
Qed.
Hint Resolve subst_vars_indep_b: core.



Lemma subst_pvars_indep_b v T x y:
  length x = length y ->
  typ_indep_b v T ->
  typ_indep_b v (subst_pvars x y T).
Proof.
  introv Hl Hindep.

  edestruct (subst_pvars_unfold (T)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst.
  eapply subst_vars_indep_b. eapply fresh_length in Hfr. auto.
  eapply subst_vars_indep_b. eapply fresh_length in Hfr. rewrite Hfr. auto.
  auto.
Qed.
Hint Resolve subst_pvars_indep_b: core.

Lemma typ_indep_b_subst T n x y :
  typ_indep_b n (T /[ x -> y]) <-> typ_indep_b n T.
Proof.
  split.

  { remember (T /[ x -> y]) as S. remember (avar_b n) as v.
    introv Hindep. reverts T n.
    induction Hindep; intros; subst; eauto.

    all: try solve [ destruct T; simpl in *; try discriminate; eauto;
                     inversions HeqS; eauto
                   | destruct T; simpl in *; try discriminate; eauto;
                     inversions HeqS;
                     destruct d; simpl in *;  try discriminate; eauto;
                     inversions H0; eauto ].

    - inversions H. all: destruct T0; simpl in *; try discriminate; eauto.
    - destruct T; simpl in *; try discriminate; eauto; inversions HeqS.
      eapply typ_indep_sel.
      all: destruct a; simpl in *; eauto; try discriminate.
      all: destruct a0; simpl in *; eauto; try discriminate.
    - destruct T; simpl in *; try discriminate; eauto; inversions HeqS.
      eapply typ_indep_sel_normal.
      destruct a0; simpl in *; eauto; try discriminate. }

  { remember (avar_b n) as v. introv Hindep. reverts n.
    induction Hindep; simpl in *; intros; subst; eauto.

    - eapply typ_indep_sel.
      all: destruct v1; simpl in *; eauto; try discriminate.
      all: destruct v2; simpl in *; eauto; try discriminate.
    - eapply typ_indep_sel_normal.
      destruct v2; simpl in *; eauto; try discriminate. }

Qed.

Lemma subst_vars_indep_b_inv v T x y:
  length x = length y ->
  typ_indep_b v (subst_vars x y T) ->
  typ_indep_b v T.
Proof.
  introv Hl Hindep.
  reverts y.
  induction x; intros; destruct y; simpls; try discriminate.
  eauto.
  inversions Hl.
  eapply IHx. eauto. apply -> typ_indep_b_subst; eauto.
Qed.
Lemma subst_pvars_indep_b_inv v T x y:
  length x = length y ->
  typ_indep_b v (subst_pvars x y T) ->
  typ_indep_b v T.
Proof.
  introv Hl Hindep.

  edestruct (subst_pvars_unfold (T)) as [?zs [?Hfr ?Hsubst]].
  rewrite Hsubst in Hindep.
  eapply subst_vars_indep_b_inv. eapply fresh_length in Hfr. rewrite Hl. symmetry. eapply Hfr.
  eapply subst_vars_indep_b_inv. eapply fresh_length in Hfr. rewrite Hfr. auto.
  eapply Hindep.
Qed.