(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Program.Equality.

Require Import DotTactics LibExtra.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping
        Weakening EWeakening Visibility.
Require Import  Substitution.

Section WithTypingMode.
Context {typing_mode: Mode}.

(** * Renaming  *)
Section Renaming.

Context {A: Set} {AO: Openable A} {AF: FreeVar A} {AT: Typing A} {AS: SubstVar A}
  {AX: AbstractSyntax AO AF AS}
  {AI: SubstIntro AX} {AW: Substitution AS subst_var AT}{AWE: EWeakening (obj_typing)}.


Lemma renaming_open_ty G n z vk T U e (ds : A) x:
    ctx_ok G ->
    z §# G ->
    ok e ->
    renv_notin z e ->
    z \notin (ctx_fv G \u fv ds \u fv T \u fv U) ->
    G § z / vk ~ open z U ⁏e ⊢ open_rec n z ds ∶ open_rec n z T ->
    G⁏e ⊢ x ∶ open x U ->
    (mode_has_visibility typing_mode -> ctx_vis G x) ->
    G⁏e ⊢ open_rec n x ds ∶ open_rec n x T.
Proof.
  introv Hok Hoke Hnz Hnz'' Hnz' Hz Hx.
  rewrite subst_intro with (x0:=z) (y:=x) by auto.
  rewrite subst_intro with (x0:=z) (y:=x) by auto.
  eapply subst_end; auto. eapply Hz. rewrite <- subst_intro. all: auto.
Qed.

(** Renaming the name of the opening variable for definition typing.  #<br>#

    [ok G]                   #<br>#
    [z] fresh                #<br>#
    [G, z: T^z ⊢ ds^z ∶ T^z] #<br>#
    [G ⊢ x∶ T^x]             #<br>#
    [――――――――――――――――――――――] #<br>#
    [G ⊢ ds^x ∶ T^x]         *)
Lemma renaming_def G n z vk T e (ds : A) x:
    ctx_ok G ->
    z §# G ->
    ok e ->
    renv_notin z e ->
    z \notin (ctx_fv G \u fv ds \u fv T) ->
    G § z / vk ~ open z T ⁏e ⊢ open_rec n z ds ∶ open_rec n z T ->
    G⁏e ⊢ x ∶ open x T ->
    (mode_has_visibility typing_mode -> ctx_vis G x) ->
    G⁏e ⊢ open_rec n x ds ∶ open_rec n x T.
Proof.
  intros.
  eapply renaming_open_ty; eauto.
  notin_solve.
Qed.

(** Renaming the name of the opening variable for term typing. #<br>#
    [ok G]                   #<br>#
    [z] fresh                #<br>#
    [G, z: U ⊢ t^z ∶ T^z]    #<br>#
    [G ⊢ x∶ U]               #<br>#
    [――――――――――――――――――――――] #<br>#
    [G ⊢ t^x ∶ T^x]         *)
Lemma renaming_ty G n z vk e T U (t : A) x:
    ctx_ok G ->
    z §# G ->
    ok e ->
    renv_notin z e ->
    z \notin (ctx_fv G \u fv U \u fv T \u fv t) ->
    G § z / vk ~ U ⁏e ⊢ open_rec n z t ∶ open_rec n z T ->
    G⁏e ⊢ x ∶ U ->
    (mode_has_visibility typing_mode -> ctx_vis G x) ->
    G⁏e ⊢ open_rec n x t ∶ open_rec n x T.
Proof.
  introv Hok Hoke Hnz Hnz'' Hnz' Hz Hx.
  rewrite subst_intro with (x0:=z) (y:=x).
  rewrite subst_intro with (x0:=z) (y:=x).
  eapply subst_end; auto. eapply Hz. rewrite subst_fresh. all: auto.
Qed.

Lemma renaming_ty2 G n z1 vk1 z2 vk2 e T U1 U2 (t : A) x:
    ctx_ok G ->
    z1 §# G ->
    z2 §# G ->
    ok e ->
    renv_notin z1 e ->
    z1 \notin (ctx_fv G \u fv U1 \u fv U2 \u fv T \u fv t) ->
    z1 <> z2 ->
    G § z1 / vk1 ~ U1 § z2 / vk2 ~ (open z1 U2) ⁏e ⊢ open_rec n z1 t ∶ open_rec n z1 T ->
    G⁏e ⊢ x ∶ U1 ->
    (mode_has_visibility typing_mode ->
     ctx_vis ((G § z1 / vk1 ~ U1) §& (z2 / vk2 ~ open z1 U2)) z1 ->
     ctx_vis (G §& subst_ctx z1 x (z2 / vk2 ~ open z1 U2)) x) ->
    G § z2 / vk2 ~ (open x U2) ⁏e ⊢ open_rec n x t ∶ open_rec n x T.
Proof.
  introv Hok Hnz1 Hnz2 Hoke Hnz'' Hnz' Hnz12 Hz Hx Hvis.
  rewrite subst_intro with (x0:=z1) (y:=x); auto.
  rewrite subst_intro with (x0:=z1) (y:=x); auto.
  rewrite subst_intro with (x0:=z1) (y:=x); auto.
  rewrite ctx_concat_single.
  rewrite <- (ctx_subst_empty z1 x).
  rewrite <- ctx_subst_push.
  eapply subst_ty; auto.
  rewrite <- ctx_concat_single.
  eapply Hz.
  rewrite <- ctx_concat_single.
  apply ctx_ok_push; auto.
  eapply weaken_end; auto.
  rewrite subst_fresh; auto.
  apply ctx_weakens_map; auto.
  apply ctx_ok_concat_map.
  rewrite <- ctx_concat_single; auto.
  apply Hvis.
Qed.

(** Renaming the name of the opening variable for term typing. #<br>#
    [ok G]                   #<br>#
    [z] fresh                #<br>#
    [G, z: U ⊢ t^z ∶ T^z]    #<br>#
    [――――――――――――――――――――――] #<br>#
    [G ⊢ t^x ∶ T^x]         *)
Lemma renaming_fresh : forall L G e T (t : A) U x vk,
    ctx_ok G ->
    ok e ->
    (forall z : var, z \notin L -> renv_notin z e -> G § z / vk ~ T ⁏e ⊢ open z t ∶ U) ->
    G⁏e ⊢ x ∶ T ->
    (mode_has_visibility typing_mode -> ctx_vis G x) ->
    G⁏e ⊢ open x t ∶ U.
Proof.
  introv Hok Hoke Hu Hx Hvis.
  pick_fresh_with y (fv t).
  rewrite subst_intro with (x0:=y) (y0:=x) by auto.
  rewrite <- subst_fresh with (x0:=y) (y0:=x) by auto.
  eapply subst_end; auto.
  rewrite subst_fresh; auto.
Qed.

Lemma renaming_push {AWW: Weakening AT}: forall L G e T (t : A) U x vk,
    ctx_ok (G § x / vk ~ T) ->
    ok e ->
    (forall z : var, z \notin L -> renv_notin z e -> G § z / vk ~ T ⁏e ⊢ open z t ∶ U) ->
    G § x / vk ~ T ⁏e ⊢ open x t ∶ U.
Proof.
  introv Hok Hoke Hu. eapply renaming_fresh with (L:=L); eauto.
  intros.
  set (Hu z H).
  rewrite <- ctx_concat_empty_r with (G1 := (G § x / vk ~ T)) .
  rewrite <- ctx_concat_push.
  rewrite <- ctx_concat_empty_r with (G1 := G) .
  rewrite <- ctx_concat_push.
  apply weaken_ty; eauto.
Qed.

End Renaming.

End WithTypingMode.
