Set Implicit Arguments.

Require Import Coq.Program.Equality.

Require Import DotTactics LibExtra.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping
        Weakening EWeakening Visibility.
Require Import  Indep TypingContext.

Section WithTypingMode.
Context {typing_mode: Mode}.


(* N parameters is because of capbnd, where we want to allow index 0 on right side *)
Class Sealing (A B: Set) (AO: Sealable A) (BO: Sealable B) (property: TypingProperty A B) (n: nat) :=
  sealing_obj G Go e (v: A) T x:
    property G Go e v T ->
    property (seal x G) (seal x Go) e (seal x v) (seal_rec n x T).


Definition SealingInd (A B: Set) (AO: Sealable A) (BO: Sealable B) (property: TypingProperty A B) (n: nat) :=
  (forall G Go e (v: A) T,
    property G Go e v T ->
      forall x,
      property (seal x G) (seal x Go) e (seal x v) (seal_rec n x T)).

Definition SealingEInd (A: Set) (AO: Sealable A) (t: TypingEquiv A) :=
  SealingInd AO AO (fun G Go e (a b: A) => typing_equiv e a b) 0.
Definition SealingTInd (A: Set) (AO: Sealable A) (t: Typing A) :=
  SealingInd AO TypSealable obj_typing 0.
Definition SealingDInd (A: Set) (AO: Sealable A) (t: DecTyping A) :=
  SealingInd AO DecSealable obj_dec_typing 0.
Definition SealingIncapInd (t: ReferenceIncapacitate) :=
  SealingInd TypSealable TypSealable obj_ref_incap 0.
Definition SealingCapbndInd (C : typ_refcap_label) (_ : CapabilityBound C)  :=
  SealingInd TypSealable TypSealable obj_cap_bnd 1.


Definition SealingIndN (A B: Set) (AO: Sealable A) (BO: Sealable B) (property: TypingProperty A B) :=
  (forall G Go e (v: A) T,
    property G Go e v T ->
      forall n x,
      property (seal x G) (seal x Go) e (seal_rec n x v) (seal_rec n x T)).

(*
Definition SealingEIndN (A: Set) (AO: Sealable A) (t: TypingEquiv A) :=
  SealingIndN AO AO (fun G Go e (a b: A) => typing_equiv e a b).*)

Lemma sealing_tyq_var: SealingEInd _ tyq_var.
Proof.
  unfold SealingEInd, SealingInd. auto.
Qed.


Lemma sealing_tyq_typ_ind: SealingEInd _ tyq_typ.
Proof.
  unfold SealingEInd, SealingInd.
  intros.
  induction H; intros; sympl; auto.
  - inversion H; intros; subst; simpl; auto.
  - eapply tyq_typ_fun; eauto. intros. repeat rewrite open_seal_down; auto.
  - eapply tyq_typ_typ; intros; repeat rewrite open_seal_down; auto.
  - eapply tyq_typ_bnd; intros; repeat rewrite open_seal_down; auto.
  - eapply tyq_typ_met; intros; repeat rewrite open_seal_down; auto. sympl.
    rewrite (open_seal_down 1).
    rewrite (open_seal_down 1).
    rewrite open_seal_down.
    rewrite (open_seal_down 0 x). (* If not applied, picks up wrong place, confusing seal and open *)
    auto.
Qed.

Instance sealing_tyq_typ: Sealing TypSealable TypSealable (fun G Go e (a b: typ) => typing_equiv e a b) 0.
Proof.
  unfold Sealing. intros.
  eapply sealing_tyq_typ_ind; eauto.
Qed.



Lemma seal_var (x y: var):
  seal x y = y.
Proof.
  unfold seal, VarSealable. auto.
Qed.
Hint Rewrite seal_var: seal_simpl.

Lemma sealing_ts_ind:
  SealingTInd _ (ty_var _) /\
  SealingInd _ _ subtyping 0 /\
  SealingIncapInd (ty_incap _).
Proof.
  introv.
  apply ts_mutind.
  all: intros.
  all: sympls.
  all: eauto 4.
  all: fold_typ_seal_rec.
  all: autorewrite with seal_simpl in *.
  - eauto using ctx_seal_rec_binds.
  - (* Rec intro *)
    eapply ty_rec_intro; eauto using typ_indep_seal.
    all: intros; rewrite open_seal_down; eauto.
  - (* Rec elim *)
    rewrite <- open_seal_down.
    eapply ty_rec_elim; eauto using typ_indep_seal.
  - eapply subtyp_typ;
    intros;
    repeat rewrite open_seal_down; eauto.
  - rewrite <- open_seal_down.
    eapply subtyp_sel2.
    eauto.
  - rewrite <- open_seal_down.
    eapply subtyp_sel1.
    eauto.
  - eapply subtyp_fun; eauto.
    intros.
    repeat rewrite open_seal_down.
    eauto.
  - eapply subtyp_met; eauto.
    intros.
    repeat rewrite open_seal_down.
    eauto.
    + (* Caused by modified method subtyping for sef *)
      intros.
      eapply typ_indep_seal.
      eapply t. auto.
      eapply typ_indep_seal_inv; eauto.
    + intros.
      repeat rewrite (open_seal_down (1)).
      repeat rewrite (open_seal_down (0) x x0).
      eauto.
  - eapply subtyp_tyq.
    pose proof sealing_tyq_typ_ind.
    hnf in H.
    eapply (H ctx_empty objctx_empty).
    eauto.
  - eapply ty_incap_sel; eauto.
    fold_typ_seal_rec.
    rewrite open_seal_down.
    eauto.
Qed.

Instance sealing_ty_var: Sealing VarSealable TypSealable (ty_var _) 0.
Proof.
  instance_by_ind Sealing sealing_ts_ind.
Qed.
Instance sealing_ty_incap: Sealing TypSealable TypSealable (ty_incap _) 0.
Proof.
  instance_by_ind Sealing sealing_ts_ind.
Qed.
Instance sealing_subtyp: Sealing TypSealable TypSealable (subtyp _) 0.
Proof.
  instance_by_ind Sealing sealing_ts_ind.
Qed.
Lemma sealing_capbnd_ind C:
  SealingCapbndInd (ty_capbnd _ C).
Proof.
  hnf. intros.

  induction H.
  8: {
    eapply ty_capbnd_sel; eauto.
    2: rewrite <- open_seal_down in IHty_capbnd.
    2: eapply IHty_capbnd.
    eapply sealing_obj with (x1:=x) in H. simpls.
    rewrite seal_var in H.
    eapply H.
  }
  all: sympls; eauto.
Qed.


Instance sealing_ty_capbnd C: Sealing TypSealable TypSealable (ty_capbnd _ C) 1.
Proof.
  instance_by_ind Sealing sealing_capbnd_ind.
Qed.
Instance sealing_ty_viewpoint_adapt_cap C x: Sealing TypSealable TypSealable (viewpoint_adapt_cap _ C x) 0.
Proof.
  hnf. intros.
  inversions H.
  simpls. cases_if. eapply Nat.nle_succ_0 in C0. contradiction.
  constructor; eapply sealing_obj; eauto.
Qed.
Instance sealing_ty_viewpoint_adapt x: Sealing TypSealable TypSealable (viewpoint_adapt _ x) 0.
Proof.
  hnf. intros.
  unfold viewpoint_adapt in *.
  destruct (mode_viewpoint_adaptation' typing_mode).
  - subst; eauto.
  - eapply sealing_obj; eauto.
Qed.

Lemma seal_typ_objlit T v:
  seal v (typ_objlit typing_mode T) = (typ_objlit typing_mode (seal_rec 1 v T)).
Proof.
  destruct typing_mode;
  simpls; auto.
Qed.
Lemma seal_ctx_obj v G Go:
  seal v (G §§& Go) = (seal v G) §§& (seal v Go).
Proof.
  unfold ctx_obj; destruct Go;
  simpls; auto. rewrite open_seal_down. auto.
Qed.

Lemma seal_ctx_obj_met x G Go xz TZ xc TC:
  seal x (ctx_obj_met typing_mode G Go xz TZ xc TC) = (ctx_obj_met typing_mode (seal x G) (seal x Go) xz (seal x TZ) xc (seal_rec 1 x TC)).
Proof.
  unfold ctx_obj_met; destruct Go; destruct typing_mode;
  simpls; auto; repeat rewrite open_seal_down; auto.
Qed.

Lemma seal_defs_hasnt ds d x:
  defs_hasnt ds (label_of d) ->
  defs_hasnt (seal x ds) (label_of (seal x d)).
Proof.
  introv Hhn.
  unfold defs_hasnt in *.
  induction ds.
  - destruct d; simpls; auto.
  - destruct d; simpls; cases_if; cases_if; destruct d0; simpls; solve [discriminate | eauto | contradiction].
Qed.

Lemma sealing_rules_ind:
  SealingTInd _ (ty_trm _) /\
  SealingTInd _ (ty_lit _) /\
  SealingDInd _ (ty_def _) /\
  SealingTInd _ (ty_defs _).
Proof.
  introv.
  apply rules_mutind.
  all: intros.
  all: simpls; eauto.
  - (* var *)
    eapply sealing_obj with (x1:=x0) in o.
    simpls.
    eapply ty_trm_var; eauto.
  - (* apply *)
    rewrite <- open_seal_down.
    eapply sealing_obj with (x1:=x0) in o.
    eapply sealing_obj with (x1:=x0) in o0.
    simpls.
    eapply ty_trm_apply with (S:=(seal x0 S)); eauto.
  - eapply sealing_obj with (x1:=x0) in o.
    eapply @sealing_obj with (x:=x0) in v.
    2: eapply sealing_ty_viewpoint_adapt.
    simpls.
    eauto.
  - eapply sealing_obj with (x1:=x0) in o0.
    eapply sealing_obj with (x1:=x0) in o1.
    simpls.
    eapply ty_trm_write; eauto.
    intro.
    eapply sealing_obj with (x1:=x0) in o; eauto.
  -
    rewrite <- open_seal_down.
    rewrite <- open_seal_down.
    eapply sealing_obj with (x1:=x0) in o.
    eapply sealing_obj with (x1:=x0) in o0.
    eapply sealing_obj with (x1:=x0) in o1.
    simpls.
    rewrite <- open_seal_down in o1.
    eapply ty_trm_call; eauto 6 using typ_indep_seal.
  - (* let *)
    eapply ty_trm_let. eauto.
    intros.
    specialize (H0 x0 H1 x).
    rewrite <- open_seal_down in H0.
    eauto.
  - eapply ty_trm_lit; eauto.
    intros.
    specialize (H0 x0 H1 x).
    rewrite <- open_seal_down in H0.
    eauto.
  - eapply sealing_obj with (x0:=x) in s.
    eauto.
  - eapply ty_lit_fun; eauto.
    intros.
    specialize (H x0 H0 x).
    rewrite <- open_seal_down in H.
    rewrite <- open_seal_down in H.
    eauto.
  - (* lit_obj *)
    rewrite seal_typ_objlit.
    simpls; eapply ty_lit_obj; eauto using typ_indep_seal.
    intros.
    specialize (H x0 H0 x).
    rewrite <- open_seal_down in H.
    rewrite <- open_seal_down in H.
    unfold ctx_obj.
    eauto.
  - (* dec_typ *)
    eapply ty_def_typ.
    intros.
    specialize (t x0 H).
    fold_typ_seal_rec.
    rewrite open_seal_down.
    rewrite open_seal_down.
    pose proof (@sealing_obj _ _ _ _ _ _ sealing_tyq_typ). eapply H0; eauto.
  - (* fld *)
    eapply ty_def_fld; eauto.
    rewrite <- seal_ctx_obj. eauto.
    eapply sealing_obj with (x0:=x) in o.
    simpls.
    rewrite <- seal_ctx_obj.
    eauto.
  - eapply ty_def_met; eauto.
    intros.
    specialize (H xz H0 xc H1 H2 x).
    rewrite <- open_seal_down in H.
    rewrite <- open_seal_down in H.
    rewrite <- open_seal_down in H.
    rewrite <- open_seal_down in H.
    rewrite <- seal_ctx_obj_met.
    eauto.
  - eapply ty_defs_cons; eauto using seal_defs_hasnt.
Qed.

Global Instance sealing_ty_trm: Sealing TrmSealable TypSealable (ty_trm _) 0.
Proof.
  instance_by_ind Sealing sealing_rules_ind.
Qed.
Global Instance sealing_ty_lit: Sealing LitSealable TypSealable (ty_lit _) 0.
Proof.
  instance_by_ind Sealing sealing_rules_ind.
Qed.
Global Instance sealing_ty_def: Sealing DefSealable DecSealable (ty_def _) 0.
Proof.
  instance_by_ind Sealing sealing_rules_ind.
Qed.
Global Instance sealing_ty_defs: Sealing DefsSealable TypSealable (ty_defs _) 0.
Proof.
  instance_by_ind Sealing sealing_rules_ind.
Qed.


Lemma fv_seal_subset (T: typ) n x:
  fv (seal_rec n x T) \c fv T \u \{x}.
Proof.
  forwards Hc: fv_seal_cases x T n.
  destructecs Hc; rewrite Hc; union_sub_solver.
Qed.
Lemma fv_seal_notin (T: typ) n x y: 
  y <> x ->
  y \notin fv T ->
  y \notin fv (seal_rec n x T).
Proof.
  introv Hnx HnT.
  eapply notin_subset.
  eapply fv_seal_subset.
  rewrite notin_union; auto.
Qed.



Lemma ctx_fv_seal_subset (T: ctx) n x:
  ctx_fv (seal_rec n x T) \c ctx_fv T \u \{x}.
Proof.
  induction T; sympl; try union_sub_solver.
  union_sub_solver_destruct_ls.
  
  eapply subset_trans.
  eapply fv_seal_subset.
  union_sub_solver.
  eapply subset_trans.
  eapply IHT.
  union_sub_solver.
Qed.
Lemma ctx_fv_seal_notin (T: ctx) n x y: 
  y <> x ->
  y \notin ctx_fv T ->
  y \notin ctx_fv (seal_rec n x T).
Proof.
  introv Hnx HnT.
  eapply notin_subset.
  eapply ctx_fv_seal_subset.
  rewrite notin_union; auto.
Qed.




End WithTypingMode.
