(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Program.Equality.

Require Import DotTactics LibExtra.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping
        Weakening EWeakening Visibility TypingContext.
Require Import  Indep.

(** * Tactics and lemmas for variable substitution. *)

Ltac unfold_fvar :=
  unfold subst_fvar; rewrite~ If_r.

Ltac subst_open_fresh :=
  repeat match goal with
    | [ |- context [ (?O ?n ?z (?S ?x ?y ?T)) ] ] =>
        (replace (O n z (S x y T)) with (O n (subst_fvar x y z) (S x y T))
          by unfold_fvar);
        let H := fresh in
        pose proof (subst_open_commut x y z T) as H;
        unfold subst_fvar; rewrite <- H; clear H
         end;
  try unfold_fvar.


Ltac fold_subst_ctx :=
  try rewrite ? subst_singles_rev_to_list by auto;
  try rewrite ? ctx_subst_push;
  rewrite <- ? ctx_concat_assoc, <- ? ctx_map_concat, ? ctx_concat_assoc.

Ltac unfold_subst_ctx :=
  rewrite ctx_map_concat, <- subst_single.

Ltac subst_solver :=
    fresh_constructor;
    subst_open_fresh;
    match goal with
    | [ H: forall z, z \notin ?L -> forall G, _
        |- context [ ?G1 §& ctx_map (?S1 ?x ?y) ?G2 § ?z / ?vk ~ ?S2 ?x ?y ?T] ] =>
      assert (G1 §& ctx_map (S1 x y) G2 § z / vk ~ S2 x y T
              = G1 §& subst_ctx x y (G2 § z / vk ~ T))
        as B by (unfold_subst_ctx; rewrite ctx_concat_assoc; auto);
      rewrite B; apply H; rewrite ? ctx_concat_assoc; auto 2;
      rewrite ctx_map_concat; rewrite ? ctx_concat_assoc;
      auto using weaken_ty_trm, weaken_ty_var
    end.

Ltac subst_solver2 :=
  fresh_constructor;
  subst_open_fresh;
  match goal with
    | [ H: forall z, z \notin ?L -> forall G, _
        |- context [ ?G1 §& ctx_map (?S1 ?x ?y) ?G2 § ?z / ?vk ~ ?S2 ?x ?y ?T] ] =>
      assert (G1 §& ctx_map (S1 x y) G2 § z / vk ~ S2 x y T
              = G1 §& subst_ctx x y (G2 § z / vk ~ T))
        as B  by (rewrite ctx_subst_push; rewrite ctx_concat_push; auto);
      rewrite B;
      eapply H;  rewrite ? ctx_concat_push; eauto 2;
      rewrite ctx_map_push; rewrite ctx_concat_push
  end.

Ltac subst_open_commut_solver :=
  sympls;
  try rewrite subst_open_commut;
  auto; eauto 4.

Ltac fold_subst :=
  repeat match goal with
    | [ |- context [ (avar_f (If ?x = ?y then ?z else ?x)) ] ] =>
        asserts_rewrite ( (avar_f (If x = y then z else x))
                         = subst_var y z ((avar_f x))); auto
    | [ |- context [ trm_var (avar_f (If ?x = ?y then ?z else ?x)) ] ] =>
        asserts_rewrite (trm_var (avar_f (If x = y then z else x))
                         = subst_trm y z (trm_var (avar_f x))); auto
    | [ |- context [ open (If ?x = ?y then ?z else ?x) (subst_var ?y ?z ?T) ] ] =>
        asserts_rewrite (open (If x = y then z else x) (subst_var y z T)
                     = subst_var y z (open x T)); auto  end.

Ltac simpl_length_s :=
  repeat match goal with
  | [ |- context[ length_s (_ /[ _ -> _]) ] ] =>
    rewrite length_s_subst
  | [ H: context[ length_s (_ /[ _ -> _]) ] |- _ ] =>
    rewrite length_s_subst in H
  end.


Lemma subst_ctx_met G1 x y G2 x1 TZ x2 TC:
  x1 <> x ->
  x2 <> x ->
  ctx_met (G1 §& subst_ctx x y G2) x1 (TZ /[ x -> y]) x2 (TC /[ x -> y]) =
  G1 §& subst_ctx x y (ctx_met G2 x1 TZ x2 TC).
Proof.
  introv Hnx1 Hnx2.
  rewrite ctx_subst_push.
  rewrite ctx_concat_push.
  rewrite ctx_subst_push.
  rewrite ctx_concat_push.
  repeat rewrite subst_open_commut.
  rewrite substi_ne_var; auto.
Qed.

Lemma subst_ctx_met_refined G1 x y G2 x1 TZ x2 TC TC2:
  x1 <> x ->
  x2 <> x ->
  ctx_met_refined (G1 §& subst_ctx x y G2) x1 (TZ /[ x -> y]) x2 (TC /[ x -> y]) (TC2 /[ x -> y]) =
  G1 §& subst_ctx x y (ctx_met_refined G2 x1 TZ x2 TC TC2).
Proof.
  introv Hnx1 Hnx2.
  rewrite ctx_subst_push.
  rewrite ctx_concat_push.
  rewrite ctx_subst_push.
  rewrite ctx_concat_push.
  sympl.
  repeat rewrite subst_open_commut.
  rewrite substi_ne_var; auto.
Qed.

Lemma subst_ctx_met_refined_heap G1 x y G2 x1 TZ x2 TC y2 TC2:
  x1 <> x ->
  x2 <> x ->
  y2 <> x ->
  ctx_met_refined_heap (G1 §& subst_ctx x y G2) x1 (TZ /[ x -> y]) x2 (TC /[ x -> y]) y2 (TC2 /[ x -> y]) =
  G1 §& subst_ctx x y (ctx_met_refined_heap G2 x1 TZ x2 TC y2 TC2).
Proof.
  introv Hnx1 Hnx2 Hny2.
  rewrite ctx_subst_push.
  rewrite ctx_concat_push.
  rewrite ctx_subst_push.
  rewrite ctx_concat_push.
  sympl.
  repeat rewrite subst_open_commut.
  rewrite substi_ne_var; auto.
  rewrite substi_ne_var; auto.
  rewrite substi_ne_var; auto.
Qed.

Lemma subst_ctx_met_refined_heap0 G1 x y G2 x1 TZ x2 TC y2 TC2:
  x1 <> x ->
  x2 <> x ->
  ctx_met_refined_heap (G1 §& subst_ctx x y G2) x1 (TZ /[ x -> y]) x2 (TC /[ x -> y]) (y2 /[ x -> y]) (TC2 /[ x -> y]) =
  G1 §& subst_ctx x y (ctx_met_refined_heap G2 x1 TZ x2 TC y2 TC2).
Proof.
  introv Hnx1 Hnx2.
  rewrite ctx_subst_push.
  rewrite ctx_concat_push.
  rewrite ctx_subst_push.
  rewrite ctx_concat_push.
  sympl.
  repeat rewrite subst_open_commut.
  rewrite substi_ne_var; auto.
  rewrite (substi_ne_var y Hnx2); auto.
Qed.

Lemma subst_met A (AS: SubstVar A) (AO: Openable A) (AF: FreeVar A)  (AAS: AbstractSyntax AO AF AS)  (ASO: SubstOpenCommut AAS)  x y (x1 x2: var) (TR: A):
  x1 <> x ->
  x2 <> x ->
  (open_met x1 x2 (TR /[x -> y])) = ((open_met x1 x2 TR) /[x -> y]).
Proof.
  introv Hnx1 Hnx2.
  repeat rewrite subst_open_commut.
  repeat rewrite substi_ne_var; auto.
Qed.

Lemma substitution_neq_visible G1 G2 v vk S x y:
  x <> v ->
  ctx_vis ((G1 § v / vk ~ S) §& G2) x ->
  ctx_vis (G1 §& subst_ctx v y G2) x.
Proof.
  introv Hneq Hvis. remember ((G1 § v / vk ~ S) §& G2) as G.
  generalize dependent G1. generalize dependent G2.
  induction Hvis; intros; subst.
  - apply ctx_push_concat_inv in HeqG.
    destruct HeqG as [[? ?] | [? ?]]; subst; simpl; eauto.
    inversions H0; try tauto.
  - apply ctx_push_concat_inv in HeqG as ?.
    destruct H0 as [[? ?] | [? ?]]; subst; simpl in *; eauto.
    + inversions H1; simpl; eauto.
    + inversions HeqG. specialize (IHHvis Hneq).
      simpl in IHHvis. eauto.
Qed.



Lemma ctx_vis_objctx_concat G Go x:
  ctx_vis (G) x ->
  ctx_vis (G §§& Go) x.
Proof.
  intros Hvis.
  unfold ctx_obj.
  destruct Go; eauto.
  destruct (classicT (x= x0)); subst;
  eauto.
Qed.
Lemma subst_vars_label_of_def: forall (xs ys : list var) (d : def), label_of d = label_of (d /[ xs --> ys]).
Proof.
  intros.
  reverts ys d.
  induction xs; intros; simpls; auto; destruct ys; auto.
  rewrite <- subst_label_of_def.
  rewrite <- IHxs. auto.
Qed.
Lemma subst_pvars_label_of_def: forall (xs ys : list var) (d : def), label_of d = label_of (d //[ xs --> ys]).
Proof.
  intros.
  edestruct (subst_pvars_unfold (d)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  rewrite <- (subst_vars_label_of_def  zs ys); auto.
  rewrite <- (subst_vars_label_of_def  xs zs d); auto.
Qed.

Lemma subst_vars_defs_hasnt: forall (xs ys : list var) (l : label) (ds : defs), defs_hasnt ds l -> defs_hasnt (ds /[ xs --> ys]) l.
Proof.
  intros.
  reverts ys ds.
  induction xs; intros; simpls; auto; destruct ys; auto.
  eapply subst_defs_hasnt; eauto.
Qed.
Lemma subst_pvars_defs_hasnt: forall (xs ys : list var) (l : label) (ds : defs), defs_hasnt ds l -> defs_hasnt (ds //[ xs --> ys]) l.
Proof.
  intros.
  edestruct (subst_pvars_unfold (ds)) as [zs [Hfr Hsubst]].
  rewrite Hsubst.
  eapply (subst_vars_defs_hasnt  zs ys); auto.
  eapply (subst_vars_defs_hasnt  xs zs); auto.
Qed.


Lemma ctx_obj_subst_pvars G1 G2 Go xs ys:
  fresh (ctx_dom (G2 §§& Go)) (length xs) xs ->
  (G1 §& subst_ctx_pvars xs ys G2 §§& subst_objctx_pvars xs ys Go) =  (G1 §& subst_ctx_pvars xs ys (G2 §§&  Go)).
Proof.
  introv Hfr.
  rewrite ctx_obj_assoc.
  eapply f_equal.
  unfold ctx_obj.
  destruct Go; simpls; eauto.
  rewrite (subst_pvars_open_fresh_commut).
  eauto.
  fresh_solve.
Qed.

Lemma ctx_obj_met_subst_pvars typing_mode G2 Go xz TZ xc TC xs ys :
  fresh (ctx_dom (ctx_obj_met typing_mode G2 Go xz TZ xc TC)) (length xs) xs ->
  (subst_ctx_pvars xs ys (ctx_obj_met typing_mode G2 Go xz TZ xc TC)) = ctx_obj_met typing_mode (subst_ctx_pvars xs ys G2) (subst_objctx_pvars xs ys Go) xz (TZ //[ xs --> ys]) xc (TC //[ xs --> ys]).
Proof.
  introv Hfr.
  destruct Go; destruct typing_mode; unfold ctx_obj_met.
  all: simpls; eauto.
  all: repeat rewrite typ_subst_pvars_andor with (T:=typ_and) by eauto.
  all: repeat rewrite (subst_pvars_open_fresh_commut ) by eauto. all: eauto.
  all: rewrite subst_pvars_open_commut; eauto.
  all: repeat rewrite (subst_pvars_open_fresh_commut ) by eauto. all: eauto.
Qed.
Lemma typ_subst_objlit typing_mode x y T:
  (typ_objlit typing_mode T) /[ x -> y] = (typ_objlit typing_mode (T /[ x -> y])).
Proof.
    destruct (mode_objlit_refcaps' typing_mode);
    simpl; eauto.
Qed.
Lemma typ_subst_pvars_objlit typing_mode x y T:
  (typ_objlit typing_mode T) //[ x --> y] = (typ_objlit typing_mode (T //[ x --> y])).
Proof.
    destruct (mode_objlit_refcaps' typing_mode);
    autorewrite with simpl_subst_pvars; eauto.
Qed.

