(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Program.Equality.

Require Import DotTactics LibExtra.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping
        Weakening EWeakening Visibility.
Require Import  Indep SubstCommon UnboundSubstitution.

Section WithTypingMode.
Context {typing_mode: Mode}.

(** * Substitution Lemma *)
Class Substitution (A B: Set) (su: SubstVar A) (sub: SubstVar B)
    (property: TypingProperty A B) :=
  substitution_obj (y: var) (S: typ) G1 x vk G2 Go e (a: A) (b: B):
      property (G1 § x / vk ~ S §& G2) Go (e & x ~ y) a b ->
      ctx_ok (G1 § x / vk ~ S §& G2 §§& Go) ->
      x \notin ctx_fv G1 ->
      renv_notin x e ->
      G1 §& G2 /[§ x->y] ⁏e ⊢ y ∶ S /[x -> y] ->
      (* y must be visible in the rodot case, like in SubT from the tech report *)
      (mode_has_visibility typing_mode ->
       ctx_vis (G1 § x / vk ~ S §& G2 §§& Go) x ->
       ctx_vis (G1 §& G2 /[§x->y] §§& (subst_objctx x y Go)) y) ->
      property (G1 §& G2 /[§x->y]) (subst_objctx x y Go) e a/[x->y] b/[x->y] .
Hint Unfold Substitution: core.



Section SubstitutionLemmata.
Context (A B: Set) (AS: SubstVar A) (BS: SubstVar B)
  (property: TypingProperty A B)(AW: Substitution AS BS property){AEW: EWeakening property}.

Lemma substitution_epush (y: var) (S: typ) G1 x vk G2 e (a: A) (b: B):
      property (G1 § x / vk ~ S §& G2) objctx_empty (e & x ~ y) a b ->
      ctx_ok (G1 § x / vk ~ S §& G2) ->
      x \notin ctx_fv G1 ->
      renv_notin x e ->
      G1 §& G2 /[§ x -> y] ⁏e ⊢ y ∶ S /[x -> y] ->
      (mode_has_visibility typing_mode ->
       ctx_vis (G1 § x / vk ~ S §& G2) x ->
       ctx_vis (G1 §& G2 /[§ x -> y]) y) ->
      property (G1 §& G2 /[§ x -> y]) objctx_empty  e (a/[ x -> y]) (b/[ x -> y]).
Proof.
  intros.
  replace objctx_empty with (subst_objctx x y objctx_empty).
  eapply substitution_obj; eauto.
  simpl. auto.
Qed.

Lemma substitution (y: var) (S: typ) G1 x vk G2 e (a: A) (b: B):
      property (G1 § x / vk ~ S §& G2) objctx_empty e a b ->
      ctx_ok (G1 § x / vk ~ S §& G2) ->
      x \notin ctx_fv G1 ->
      ok e ->
      renv_notin x e ->
      G1 §& G2 /[§ x -> y] ⁏e ⊢ y ∶ S /[x -> y] ->
      (mode_has_visibility typing_mode ->
        ctx_vis (G1 § x / vk ~ S §& G2) x ->
        ctx_vis (G1 §& G2 /[§ x -> y]) y
      ) ->
      property (G1 §& G2 /[§ x -> y]) objctx_empty  e (a/[ x -> y]) (b/[ x -> y]).
Proof.
  intros.
  eapply ewkn_push in H.
  eapply substitution_epush; eauto.
  auto. unfold renv_notin in H3. auto.
Qed.


Lemma subst_end:
  forall y S G x vk e (t : A) (T: B),
    property (G § x / vk ~ S) objctx_empty e t T ->
    ctx_ok (G § x / vk ~ S) ->
    x \notin ctx_fv G ->
    ok e ->
    renv_notin x e ->
    G⁏e ⊢ y ∶ S /[ x -> y] ->
    (mode_has_visibility typing_mode -> ctx_vis G y) ->
    property G objctx_empty e (t /[ x -> y]) (T /[ x -> y]).
Proof.
  intros.
  apply substitution with (G1:=G) (G2:=ctx_empty) (x:=x) (vk := vk) (y:=y) in H;
  try rewrite map_empty in *; try rewrite concat_empty_r in *; auto.
Qed.
End SubstitutionLemmata.


Section SubstitutionTyLemmata.
Context (A: Set)(AS: SubstVar A)(AT: Typing A)
  {AW: Substitution AS subst_var AT} {AEW: EWeakening (obj_typing)}.

Lemma subst_ty_epush(y: var) (S: typ) G1 x vk G2 e (a: A) (b: typ):
  typing (G1 § x / vk ~ S §& G2) (e & x ~ y) a b ->
  ctx_ok (G1 § x / vk ~ S §& G2) ->
  x \notin ctx_fv G1 ->
  renv_notin x e ->
  G1 §& G2 /[§ x -> y] ⁏e ⊢ y ∶ S /[ x -> y] ->
  (mode_has_visibility typing_mode ->
   ctx_vis (G1 § x / vk ~ S §& G2) x ->
   ctx_vis (G1 §& G2 /[§ x -> y]) y) ->
  typing (G1 §& G2 /[§ x -> y]) e (a /[ x -> y]) (b /[ x -> y]) .
Proof.
  intros.
  eapply substitution_epush; eauto.
Qed.
(** [G1, x: S, G2 ⊢ t∶ T]            #<br>#
    [ok(G1, x: S, G2)]               #<br>#
    [x \notin fv(G1)]                 #<br>#
    [G1, G2[y/x] ⊢ y∶ S[y/x]]       #<br>#
    [―――――――――――――――――――――――――――――]  #<br>#
    [G1, G2[y/x] ⊢ t[y/x]∶ T[y/x]] #<br>#  #<br>#
*)
Lemma subst_ty(y: var) (S: typ) G1 x vk G2 e (a: A) (b: typ):
  typing (G1 § x / vk ~ S §& G2) e a b ->
  ctx_ok (G1 § x / vk ~ S §& G2) ->
  x \notin ctx_fv G1 ->
  ok e ->
  renv_notin x e ->
  G1 §& G2 /[§ x -> y] ⁏e ⊢ y ∶ S /[x -> y] ->
  (mode_has_visibility typing_mode ->
   ctx_vis (G1 § x / vk ~ S §& G2) x ->
   ctx_vis (G1 §& G2 /[§ x -> y]) y) ->
  typing (G1 §& G2 /[§ x -> y]) e (a /[x -> y]) (b /[x -> y]).
Proof.
  intros.
  eapply substitution; eauto.
Qed.
Hint Resolve subst_ty: core.

End SubstitutionTyLemmata.

Definition SubstitutionInd (y: var) (S: typ) (A B: Set) (su: SubstVar A) (sub: SubstVar B) (property: TypingProperty A B) :=
  (forall G Go e (v: A) T,
    property G Go e v T ->
    forall G1 G2 e1 x vk,
      G = G1 § x / vk ~ S §& G2 ->
      e = (e1 & x ~ y) ->
      ctx_ok (G1 § x / vk ~ S §& G2 §§& Go) ->
      x \notin ctx_fv G1 ->
      renv_notin x e1 ->
      G1 §& G2 /[§ x -> y] ⁏ e1 ⊢ y ∶ S /[x -> y] ->
      (mode_has_visibility typing_mode ->
       ctx_vis (G1 § x / vk ~ S §& G2 §§& Go) x ->
       ctx_vis (G1 §& G2 /[§ x -> y] §§& (subst_objctx x y Go)) y) ->
      property (G1 §& G2 /[§ x -> y]) (subst_objctx x y Go) e1 (subst_var x y v) (subst_var x y T)) .
Definition SubstitutionEInd (y: var) (S: typ) (A: Set) (t: TypingEquiv A) (su: SubstVar A) :=
  SubstitutionInd y S (su) (su) (fun G Go e (a b: A) => typing_equiv e a b).
Definition SubstitutionTInd (y: var) (S: typ) (A: Set) (t: Typing A) (su: SubstVar A) :=
  SubstitutionInd y S (su) (TypSubstVar) (obj_typing).
Definition SubstitutionIncapInd (y: var) (S: typ) (_: ReferenceIncapacitate) :=
  SubstitutionInd y S TypSubstVar TypSubstVar obj_ref_incap.
Definition SubstitutionCapbndInd (y: var) (S: typ)
           (C : typ_refcap_label) (_ : CapabilityBound C) :=
  SubstitutionInd y S TypSubstVar TypSubstVar obj_cap_bnd.
Definition SubstitutionDInd (y: var) (S: typ) (A: Set) (t: DecTyping A) (su: SubstVar A) :=
  SubstitutionInd y S (su) (subst_dec) (obj_dec_typing).

Local Hint Extern 1 (_ §# _) => ctx_simpl_dom; notin_solve : core.

Lemma subst_tyq_var y S:
  SubstitutionEInd y S tyq_var VarSubstVar.
Proof.
  unfold SubstitutionEInd, SubstitutionInd.
  introv Heq HG He Hok HnG Hne Hty; subst. clear Hok.

  unfold typing_equiv in Heq. (* unfolding seems to be needed for dependent induction *)
  dependent induction Heq.
  - auto.
  -
    destruct (subst_cases2 x y v1) as [[Hv1x Hv1y] | [Hv1x Hv1v1]]; subst.
    + destruct (subst_cases2 x y v2) as [[Hv2x Hv2y] | [Hv2x Hv2v2]]; subst.
      * auto.
      * rewrite Hv1y, Hv2v2.
        pose proof (binds_push_eq_inv H). subst. auto.
    +
      pose proof (binds_push_neq_inv H Hv1x).
      destruct (subst_cases2 x y v2) as [[Hv2x Hv2y] | [Hv2x Hv2v2]]; subst.
      * false. eapply renv_binds_notin_false. apply H0. auto.
      * rewrite Hv2v2, Hv1v1. auto.

  - symmetry. eapply IHHeq; eauto.
  - transitivity (v2 /[ x -> y]); eauto.
Qed.

Lemma subst_tyq_typ y S:
  SubstitutionEInd y S tyq_typ TypSubstVar.
Proof.
  unfold SubstitutionEInd, SubstitutionInd.
  intros.
  induction H.
  6: (* sel *) {
    apply tyq_typ_sel_normal.
    pose proof (subst_tyq_var S H G1 G2 vk H0 H1 H2 H3 H4 H5 H6). sympls. auto.
  }
  all: subst; sympls; eauto.
  - (* andor *)
    inversion H; subst; sympl; auto; apply (tyq_typ_binary H); auto.
  - (* fun *)
    eapply (tyq_typ_fun); eauto; intros. pick_fresh z. forwards: (H8 z); eauto.
    repeat rewrite (subst_open_commut_ne ) in H0 by notin_solve.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_var; eauto.
  - (* typ*)
    eapply (tyq_typ_typ); eauto.

    intros. pick_fresh z. forwards: (H9 z); eauto. forwards: (H7 z); eauto.
    repeat rewrite (subst_open_commut_ne ) in H0 by notin_solve.
    repeat rewrite (subst_open_commut_ne ) in H1 by notin_solve.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_var; eauto.

    intros. pick_fresh z. forwards: (H9 z); eauto. forwards: (H7 z); eauto.
    repeat rewrite (subst_open_commut_ne ) in H0 by notin_solve.
    repeat rewrite (subst_open_commut_ne ) in H1 by notin_solve.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_var; eauto.

  - apply tyq_typ_bnd; eauto; intros. pick_fresh z. forwards: (H7 z); eauto.
    repeat rewrite (subst_open_commut_ne ) in H0 by notin_solve.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_var; eauto.
  - apply tyq_typ_met; auto.
    intros. pick_fresh z. forwards: (H7 z); eauto.
    repeat rewrite (subst_open_commut_ne ) in H0 by notin_solve.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_var; eauto.
    intros. pick_fresh z1. pick_fresh z2. forwards: (H10 z1 z2); eauto.
    repeat rewrite (subst_open_commut_ne ) in H0 by notin_solve.

    repeat rewrite <- (open_vars_single 1) in *.
    repeat rewrite open_vars_push in *.
    eapply tyq_typ_reopen_vars; eauto.
    fresh_solve; auto; eapply notin_fv_subst_var; eauto.
Qed.

Lemma typ_binary_subst : forall (T : typ -> typ -> typ) (t1 t2 : typ) (x y : var),
    typ_binary T ->
    (T t1 t2) /[ x -> y] = T (t1 /[ x -> y]) (t2 /[ x -> y]).
Proof. inversion 1; subst; auto. Qed.

Lemma subst_tyq_fresh : forall e v1 v2 x y,
    renv_notin x e ->
    typing_equiv e v1 v2 ->
    typing_equiv e (v1 \[ x -> y]) (v2 \[ x -> y]).
Proof.
  introv Hxfresh Htyq.
  induction Htyq; eauto.
  unfold renv_notin in Hxfresh.
  assert (x \notin fv_env e) by auto. assert (x \notin dom e) by auto.
  apply binds_to_dom in H as ?.
  apply renv_bind_in_fv in H as ?.
  cases_if; cases_if; auto.
Qed.

Lemma notin_subst_ctx_fv x y z G:
    z \notin ctx_fv G ->
    z <> x ->
    z <> y ->
    z \notin ctx_fv (subst_ctx x y G).
Proof.
  introv HnG Hnx Hny.
  induction G;
  sympls; eauto.
  rewrite notin_union in HnG; destruct HnG.
  destruct (fv_subst_cases x y t).
  destruct H1. rewrite H1. rewrite notin_union; split; eauto.
  destruct H1. destruct H2. rewrite H3. repeat rewrite notin_union. splits; eauto; splits; eauto. intro. rewrite in_remove in H4. destruct H4. contradiction.

Qed.

Lemma subst_equiv e T1 T2 x y:
    renv_notin x e ->
    e ⊢ T1 ≈ T2 ->
    e ⊢ (T1 /[ x -> y]) ≈ (T2 /[ x -> y]).
Proof.
  introv Hxfresh Hequiv.
  induction Hequiv;
    repeat match goal with
          | [ H : renv_notin x e -> _ |- _] =>
            specialize (H Hxfresh)
          end;
    try solve[ simpl; econstructor;
              eauto using subst_tyq_fresh ].
  - inversions H; simpl; constructor; auto.
  - inversions H; simpl; constructor; auto.
  - sympls. eapply tyq_typ_fun; auto. intros.
    pick_fresh z. forwards: (H1 z); eauto.
    repeat rewrite (subst_open_commut_ne ) in H2 by notin_solve.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_var; eauto.

- (* typ*)
subst; sympls.
    eapply (tyq_typ_typ); eauto.

    intros. pick_fresh z. forwards: (H0 z); eauto. forwards: (H2 z); eauto.
    repeat rewrite (subst_open_commut_ne ) in H3 by notin_solve.
    repeat rewrite (subst_open_commut_ne ) in H4 by notin_solve.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_var; eauto.

    intros. pick_fresh z. forwards: (H0 z); eauto. forwards: (H2 z); eauto.
    repeat rewrite (subst_open_commut_ne ) in H3 by notin_solve.
    repeat rewrite (subst_open_commut_ne ) in H4 by notin_solve.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_var; eauto.
 - subst; sympls.  apply tyq_typ_bnd; eauto; intros. pick_fresh z. forwards: (H0 z); eauto.
    repeat rewrite (subst_open_commut_ne ) in H1 by notin_solve.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_var; eauto.
  - subst; sympls.  apply tyq_typ_met; auto.
    intros. pick_fresh z. forwards: (H0 z); eauto.
    repeat rewrite (subst_open_commut_ne ) in H4 by notin_solve.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_var; eauto.
    intros. pick_fresh z1. pick_fresh z2. forwards: (H3 z1 z2); eauto.
    repeat rewrite (subst_open_commut_ne ) in H4 by notin_solve.

    repeat rewrite <- (open_vars_single 1) in *.
    repeat rewrite open_vars_push in *.
    eapply tyq_typ_reopen_vars; eauto.
    fresh_solve; auto; eapply notin_fv_subst_var; eauto.
Qed.

(** The proof is by mutual induction on variable typing, and subtyping. *)
Lemma subst_ts_ind y S:
  SubstitutionTInd y S (ty_var _) VarSubstVar /\
  SubstitutionInd y S subst_typ subst_typ subtyping /\
  SubstitutionIncapInd y S (ty_incap _).
Proof.
  introv.
  apply ts_mutind.
  all: intros.
  all: subst.
  all: sympl in *.
  all: try rewrite subst_open_commut.
  all: eauto 4.

  - Case "ty_var".
    unfold subst_var.
    unfold VarSubstVar.
    cases_if.
    + apply ctx_binds_middle_eq_inv in c; subst; assumption.
    + apply (subst_fresh_ctx y) in H2.
      rewrite <- H2, <- ctx_map_concat.
      eauto using ctx_binds_map, ctx_binds_subst, ctx_binds_middle_concat.

  - Case "ty_rec_intro".
    apply ty_rec_intro; auto.
    + unfold subst_var at 2 3.
      unfold VarSubstVar. rewrite <- subst_open_commut; eauto.
    + intros. rewrite <- subst_open_commut_typ. eauto.
  - eapply subtyp_typ; intros.
    pick_fresh z.
    assert (G1 §& G2 /[§ x -> y ] ⁏ e1 ⊢ S2 /[-> z] /[ x -> y] <: S1 /[-> z] /[ x -> y]). {
      eapply H; eauto.
    }
    assert (G1 §& G2 /[§ x -> y ] ⁏ e1 ⊢ S2 /[ x -> y] /[-> z] <: S1 /[ x -> y] /[-> z]). {
      rewrite subst_open_commut_ne in H1 by eauto.
      rewrite subst_open_commut_ne in H1 by eauto. eauto.
    }
    rewrite subst_intro with (x1:=z).
    rewrite subst_intro with (x1:=z) (X:= S1 /[ x -> y]).
    eapply usubst_end1; eauto.
  eapply ctx_ok_concat_map. eapply ctx_ok_middle_inv; eauto. ctx_simpl_fv. notin_solve. eapply notin_subst_ctx_fv; notin_solve.
    eapply notin_fv_subst_var; eauto.
    eapply notin_fv_subst_var; eauto.

  pick_fresh z.
    assert (G1 §& G2 /[§ x -> y ] ⁏ e1 ⊢ T1 /[-> z] /[ x -> y] <: T2 /[-> z] /[ x -> y]).
    eapply H0; eauto.
    assert (G1 §& G2 /[§ x -> y ] ⁏ e1 ⊢ T1 /[ x -> y] /[-> z] <: T2 /[ x -> y] /[-> z]).
    rewrite subst_open_commut_ne in H1 by eauto.
    rewrite subst_open_commut_ne in H1 by eauto. eauto.
    rewrite subst_intro with (x1:=z).
    rewrite subst_intro with (x1:=z) (X:=(T2 /[ x -> y])).
    eapply usubst_end1; eauto.
  eapply ctx_ok_concat_map. eapply ctx_ok_middle_inv; eauto. ctx_simpl_fv. notin_solve. eapply notin_subst_ctx_fv; notin_solve.
    eapply notin_fv_subst_var; eauto.
    eapply notin_fv_subst_var; eauto.

  - rewrite subst_open_commut_typ.
    eapply subtyp_sel2. eapply H; eauto.
  - rewrite subst_open_commut_typ.
    eapply subtyp_sel1. eapply H; eauto.
  - (* Function subtyping *)
    subst_solver2.
    apply weaken_push; auto.
    subst_open_fresh.
    apply ctx_ok_concat_map.
    eapply ctx_ok_push_concat; eauto.
    destruct (fv_decide y \{z}).
    + rewrite in_singleton in m; subst; eauto.
    + apply notin_singleton_r in n. intros.
      assert (z <> x) by eauto. inversions H2; eauto; tauto.

  - (* Method subtyping *)
    eapply subtyp_met with
        (L := ctx_dom ((G1 § x / vk ~ S) §& G2) \u L \u \{x});
      eauto.

    { introv Hx0fr; clear s0 s1 H H1.
      assert (Hx0frL : x0 \notin L) by auto;
        specialize (H0 _ Hx0frL); clear Hx0frL.
      specialize (H0 G1 (G2 § x0 / v_par ~ TZ2) e1 x vk).
      replace ((G1 § x / vk ~ S) §& G2 § x0 / v_par ~ TZ2) with
          ((G1 § x / vk ~ S) §& (G2 § x0 / v_par ~ TZ2)) in H0 by auto.
      specialize (H0 eq_refl).
      assert (ctx_ok ((G1 § x / vk ~ S) §& (G2 § x0 / v_par ~ TZ2))) as Hok.
      { replace ((G1 § x / vk ~ S) §& (G2 § x0 / v_par ~ TZ2))
          with (((G1 § x / vk ~ S) §& G2) § x0 / v_par ~ TZ2) by auto.
        constructor; auto. }
      specialize (H0 (eq_refl _) Hok H5 H6). simpl in H0.
      assert (G1 §& subst_ctx x y G2 § x0 / v_par ~ (TZ2 /[ x -> y]) ⁏ e1 ⊢ y
                 ∶ (S /[ x -> y])).
      {  apply weaken_push;
           eauto using ctx_ok_push_concat,
           ctx_ok_concat_map. }
      specialize (H0 H); clear H.
      repeat (rewrite subst_open_commut_typ in H0).
      replace (x0 /[ x -> y]) with x0 in H0; eauto.
      - eapply H0. destruct (fv_decide y \{x0}).
        + rewrite in_singleton in m0; subst; eauto.
        + apply notin_singleton_r in n. intros.
          assert (x0 <> x) by eauto. inversions H1; eauto; tauto.
      - unfold subst_var. unfold VarSubstVar. cases_if; auto. exfalso.
        assert (x <> x) by auto. tauto. }

    (* Caused by modified method subtyping for sef *)
    { introv Hms HinC2.
    apply <- typ_indep_b_subst. eapply t. auto.
    apply -> typ_indep_b_subst. eauto.
    }

    { introv Hx0fr Hy0neqx0 Hy0fr. clear s1 H0 s0 H.
      assert (Hx0frL : x0 \notin L) by auto;
        assert (Hy0frL : y0 \notin L) by auto;
        specialize (H1 _ Hx0frL _ Hy0neqx0 Hy0frL); clear Hx0frL Hy0frL.
      specialize (H1 G1
                     ((G2 § x0 / v_par ~ TZ2) § y0 / v_recv ~ open x0 TC2)
                     e1 x vk).
      assert
        (((G1 § x / vk ~ S) §& G2 § x0 / v_par ~ TZ2)
           § y0 / v_recv ~ open x0 TC2 =
         (G1 § x / vk ~ S) §& ((G2 § x0 / v_par ~ TZ2)
                                 § y0 / v_recv ~ open x0 TC2))
        as Heq by auto; specialize (H1 Heq); clear Heq.
      assert (Hok :
                ctx_ok
                  ((G1 § x / vk ~ S)
                     §& ((G2 § x0 / v_par ~ TZ2)
                           § y0 / v_recv ~ open x0 TC2))).
      { replace ((G1 § x / vk ~ S)
                   §& ((G2 § x0 / v_par ~ TZ2) § y0 / v_recv ~ open x0 TC2))
          with  (((G1 § x / vk ~ S) §& G2 § x0 / v_par ~ TZ2)
                   § y0 / v_recv ~ open x0 TC2) by auto.
        constructor; eauto. }
      specialize (H1 (eq_refl _) Hok H5 H6).
      simpl in H1.
      assert ((G1 §& subst_ctx x y G2 § x0 / v_par ~ (TZ2 /[ x -> y]))
                § y0 / v_recv ~ (open x0 TC2) /[ x -> y]
                ⁏ e1 ⊢ y ∶ S /[ x -> y]).
      { apply weaken_push;
          eauto using ctx_ok_push_concat, ctx_ok_concat_map.
        apply weaken_push;
          eauto using ctx_ok_push_concat, ctx_ok_concat_map. }
      specialize (H1 H); clear H.
      assert (Heqx : x0 = x0 /[ x -> y]).
      { unfold subst_var, VarSubstVar; cases_if; auto. exfalso.
        assert (x <> x) by eauto. tauto. }
      assert (Heqy : y0 = y0 /[ x -> y]).
      { unfold subst_var, VarSubstVar; cases_if; auto. exfalso.
        assert (x <> x) by eauto. tauto. }
      replace (TC2 /[ x -> y] /[-> x0]) with
          (TC2 /[-> x0] /[ x -> y]) by
          (rewrite (subst_open_commut_typ TC2 x y x0 0);
           rewrite <- Heqx; auto).
      replace ((open_rec 1 y0 (TR1 /[ x -> y]))/[-> x0]) with
          ( ((open_rec 1 y0 TR1) /[-> x0]) /[ x -> y])
      by (rewrite (subst_open_commut_typ _ x y x0 0); rewrite <- Heqx;
          rewrite (subst_open_commut_typ _ x y y0 1); rewrite <- Heqy; auto).
      replace ((open_rec 1 y0 (TR2 /[ x -> y]))/[-> x0]) with
          (((open_rec 1 y0 TR2)/[-> x0]) /[ x -> y])
      by (rewrite (subst_open_commut_typ _ x y x0 0); rewrite <- Heqx;
          rewrite (subst_open_commut_typ _ x y y0 1); rewrite <- Heqy; auto).
      apply H1.
      destruct (fv_decide y \{y0}).
      + rewrite in_singleton in m0; subst. eauto.
      + apply notin_singleton_r in n. intros. eapply ctx_vis_n; eauto.
        destruct (fv_decide y \{x0}).
        * rewrite in_singleton in m0; subst. eauto.
        * apply notin_singleton_r in n0. intros.
          assert (x <> y0) by eauto. inversions H0; eauto; try tauto.
          assert (x <> x0) by eauto. inversions H14; eauto; try tauto. }

  - eapply subst_tyq_typ in t. eauto.
  - eapply ty_incap_sel; eauto. rewrite <- subst_open_commut. eauto.
Qed.

Global Instance subst_ty_var: Substitution subst_var subst_var (ty_var _).
Proof.
  instance_by_ind Substitution subst_ts_ind.
Qed.
Hint Resolve subst_ty_var: core.

Global Instance subst_sub: Substitution subst_var subst_var (subtyp _).
Proof.
  instance_by_ind Substitution subst_ts_ind.
Qed.
Hint Resolve subst_sub: core.

Global Instance subst_ty_incap: Substitution subst_var subst_var (ty_incap _).
Proof.
  instance_by_ind Substitution subst_ts_ind.
Qed.
Hint Resolve subst_ty_incap: core.


Lemma subst_ty_capbnd_ind y S:
  forall C, SubstitutionCapbndInd y S (ty_capbnd _ C).
Proof.
  introv. hnf. introv.
  induction 1; intros; subst;
    sympl in *; try rewrite subst_open_commut;
      eauto 4.
  apply ty_capbnd_sel with (T := T /[x0 -> y])
                           (S := S0 /[x0 -> y]); eauto.
  replace objctx_empty with (subst_objctx x0 y objctx_empty); eauto.
  replace (typ_rcd (dec_typ B (S0 /[ x0 -> y]) (T /[ x0 -> y])))
    with ((typ_rcd (dec_typ B S0 T)) /[x0 -> y]); eauto.
  eapply subst_ty_var; eauto.
  rewrite <- subst_open_commut; eauto.
Qed.

Global Instance subst_ty_capbnd:
  forall C, Substitution subst_var subst_var (ty_capbnd _ C).
Proof. intros. hnf. intros. eapply subst_ty_capbnd_ind; eauto. Qed.
Hint Resolve subst_ty_capbnd: core.


Lemma subst_ty_viewpoint_adapt
      (y: var) (S: typ) G1 x vk G2 Go e (T1 T2: typ) v :
  viewpoint_adapt _ v
                  (G1 § x / vk ~ S §& G2) Go
                  (e & x ~ y) T1 T2 ->
  ctx_ok (G1 § x / vk ~ S §& G2 §§& Go) ->
  x \notin ctx_fv G1 ->
  renv_notin x e ->
  G1 §& G2 /[§ x -> y] ⁏e ⊢ y ∶ S /[x -> y] ->
  (mode_has_visibility typing_mode ->
   ctx_vis (G1 § x / vk ~ S §& G2 §§& Go) x ->
   ctx_vis (G1 §& G2 /[§ x -> y] §§& (subst_objctx x y Go)) y) ->
  viewpoint_adapt _
                  (v \[ x -> y])
                  (G1 §& subst_ctx x y G2) (subst_objctx x y Go)
                  e (T1 /[ x -> y]) (T2 /[ x -> y]).
Proof.
  intros. hnf. hnf in H. hnf in H3.
  pose proof subst_ty_incap as Hicp. hnf in Hicp.
  pose proof subst_ty_capbnd as Hcbp. unfold Substitution in Hcbp.
  destruct (mode_viewpoint_adaptation' typing_mode); subst; eauto.
  inversions H. simpl. constructor.
  - eapply Hicp; eauto.
  - eapply Hcbp; eauto.
Qed.

(* Restate subst_ty for use in subst_rules_ind *)
Lemma subst_ty_var_u (y: var) (S: typ) G1 x vk G2 e (a: var) (b: typ):
  typing (G1 § x / vk ~ S §& G2) (e & x ~ y) a b ->
  ctx_ok (G1 § x / vk ~ S §& G2) ->
  x \notin ctx_fv G1 ->
  renv_notin x e ->
  G1 §& G2 /[§ x -> y] ⁏ e ⊢ y ∶ S /[ x -> y] ->
  (mode_has_visibility typing_mode ->
   ctx_vis (G1 § x / vk ~ S §& G2) x ->
   ctx_vis (G1 §& G2 /[§ x -> y]) y) ->
  typing (G1 §& G2 /[§ x -> y]) e (a \[ x -> y ]) (b /[ x -> y]) .
Proof.
  intros.
  replace ( a \[ x -> y ] ) with (a /[x -> y]).
  eapply subst_ty_epush; eauto.
  unfold subst_var, VarSubstVar. auto.
Qed.
Hint Resolve subst_ty_var_u: core.

Lemma ctx_obj_met_subst_var  G2 Go xz TZ xc TC xs ys :
  xs \notin  (ctx_dom (ctx_obj_met typing_mode G2 Go xz TZ xc TC)) ->
  (subst_ctx xs ys (ctx_obj_met typing_mode G2 Go xz TZ xc TC))
  = ctx_obj_met
      typing_mode (subst_ctx xs ys G2) (subst_objctx xs ys Go)
      xz (TZ /[ xs -> ys]) xc (TC /[ xs -> ys]).
Proof.
  intros Hnin. induction Go;
  try solve [ simpl in *; destruct (mode_visibility' typing_mode);
              repeat(simpl in *; rewrite subst_open_commut_typ, subst_ne_var;
                     eauto) ].
  simpl in *; destruct (mode_visibility' typing_mode).
  - destruct (var_decide y xs);
      subst; try (rewrite subst_eq_var); try (rewrite subst_ne_var; eauto).
    + repeat(rewrite subst_open_commut_typ).
      repeat(rewrite subst_eq_var); repeat(rewrite subst_ne_var; eauto).
    + repeat(simpl in *; rewrite subst_open_commut_typ, subst_ne_var;
             eauto).
  - destruct (var_decide y xs);
      subst; try (rewrite subst_eq_var); try (rewrite subst_ne_var; eauto).
    + repeat(rewrite subst_open_commut_typ).
      repeat(rewrite subst_eq_var); repeat(rewrite subst_ne_var; eauto).
    + repeat(simpl in *; rewrite subst_open_commut_typ, subst_ne_var;
             eauto).
Qed.

Lemma ctx_obj_met_weakens_var G1 G2 Go xz TZ xc TC e (x : var) T :
  G1 §& G2 ⁏ e ⊢ x ∶ T ->
  ctx_ok (G1 §& G2
             §& ctx_obj_met typing_mode ctx_empty Go xz TZ xc TC) ->
  G1 §& G2
     §& ctx_obj_met typing_mode ctx_empty Go xz TZ xc TC
     ⁏ e ⊢ x ∶ T.
Proof.
  intros Hty. induction Go; simpl;
    destruct (mode_visibility' typing_mode); simpl; intros Hok.
  5,3: eapply ctx_met_refined_heap_weakens; eauto.
  1: eapply ctx_met_weakens; eauto.
  - rewrite ctx_met_concat. apply hiding_ty_var. simpl.
    eapply ctx_met_weakens; eauto.
    repeat(inversions Hok;
           match goal with [H : ctx_ok _ |- _] => rename H into Hok end).
    eauto.
  - repeat(inversions Hok;
           match goal with [H : ctx_ok _ |- _] => rename H into Hok end).
    eapply weaken_push; eauto 3.
  - rewrite ctx_met_refined_heap_concat. apply hiding_ty_var. simpl.
    repeat(inversions Hok;
           match goal with [H : ctx_ok _ |- _] => rename H into Hok end).
    eapply ctx_met_refined_heap_weakens; eauto.
    eapply weaken_push; eauto 3.
  - rewrite ctx_met_refined_heap_concat. apply hiding_ty_var. simpl.
    repeat(inversions Hok;
           match goal with [H : ctx_ok _ |- _] => rename H into Hok end).
    eapply ctx_met_refined_heap_weakens; eauto.
Qed.

Lemma ctx_obj_met_weakens_var' G1 Go xz TZ xc TC e (x : var) T :
  G1 §§& Go ⁏ e ⊢ x ∶ T ->
  ctx_ok (G1
             §& ctx_obj_met typing_mode ctx_empty Go xz TZ xc TC) ->
  G1
     §& ctx_obj_met typing_mode ctx_empty Go xz TZ xc TC
     ⁏ e ⊢ x ∶ T.
Proof.
  intros Hty. induction Go; simpl;
    destruct (mode_visibility' typing_mode); simpl; intros Hok.
  5,3: eapply ctx_met_refined_heap_weakens; eauto.
  1: eapply ctx_met_weakens; eauto.
  - rewrite ctx_met_concat. apply hiding_ty_var. simpl.
    eapply ctx_met_weakens; eauto.
    repeat(inversions Hok;
           match goal with [H : ctx_ok _ |- _] => rename H into Hok end).
    eauto.
  - repeat(inversions Hok;
           match goal with [H : ctx_ok _ |- _] => rename H into Hok end).
    unfold ctx_obj in Hty.

    eapply weaken_push; eauto 3.
    eapply weaken_push; eauto 3.
    eapply hiding_hide; eauto.
    eauto.
  - rewrite ctx_met_refined_heap_concat. apply hiding_ty_var. simpl.
    repeat(inversions Hok;
           match goal with [H : ctx_ok _ |- _] => rename H into Hok end).
    eapply ctx_met_refined_heap_weakens; eauto.
Qed.

Lemma subst_rules_ind y S:
  SubstitutionTInd y S (ty_trm _) TrmSubstVar /\
  SubstitutionTInd y S (ty_lit _) LitSubstVar /\
  SubstitutionDInd y S (ty_def _) DefSubstVar /\
  SubstitutionTInd y S (ty_defs _) DefsSubstVar.
Proof.
  introv.
  (* Induction cases *)
  apply rules_mutind; intros.
  all: subst.
  (* Pushes substitution down *)
  all: sympl in *.
  (* In cases where variables are typed, specializes substitution lemma for those variables.
     Also solves ty_trm_var case. *)
  try repeat match goal with
    | H: (?G1 § ?x0 / ?vk ~ ?S) §& ?G2 ⁏_ ⊢ ?x ∶ ?T |- _ =>  apply subst_ty_var_u with (y:=y) in H ; eauto
  end.

  all: subst_open_commut_solver.
  all: try subst_solver.
  all: try subst_solver2.
  all: try eapply weaken_push.
  all: eauto using ctx_ok_concat_map, ctx_ok_push_concat, subst_ty.

  (* Deal with visibility hypotheses *)
  all: try
         solve[ match goal with
                  [ |- mode_has_visibility _ ->
                      ctx_vis (?C1 § ?z / _ ~ _) ?x ->
                      ctx_vis (?C2 § ?z / _ ~ _) ?y ] =>
                  assert (x <> z) by eauto; assert (y <> z) by eauto;
                  introv ?Hhv ?Hvis; inversions Hvis; try tauto; eauto;
                  idtac
                end ].

  - Case "ty_trm_var". apply ty_trm_var; eauto.
    (* Requires hypothesis that y is also visible *)
    cases_if; eauto using substitution_neq_visible.

  - Case "ty_trm_apply".
    apply ty_trm_apply with (S := (S0 /[ x0 -> y])); eauto.
    all: cases_if; eauto using substitution_neq_visible.

  - Case "ty_trm_read".
    eapply ty_trm_read with (S := (S0 /[ x0 -> y])) (T := (T /[ x0 -> y]));
      eauto.
    + cases_if; eauto using substitution_neq_visible.
    + hnf. hnf in o. hnf in v.
      pose proof subst_ty_incap as Hicp. hnf in Hicp.
      pose proof subst_ty_capbnd as Hcpb. unfold Substitution in Hcpb.
      destruct (mode_viewpoint_adaptation' typing_mode); subst; eauto.
      inversions v. simpl. constructor.
      * replace objctx_empty with (subst_objctx x0 y objctx_empty) by eauto.
        eapply Hicp; eauto.
      * replace objctx_empty with (subst_objctx x0 y objctx_empty) by eauto.
        eapply Hcpb; eauto.

  - Case "ty_trm_write".
    eapply ty_trm_write with (S := (S0 /[ x0 -> y])) (T := (T /[ x0 -> y]));
      eauto.
    all: cases_if; eauto using substitution_neq_visible.

  - Case "ty_trm_call".
    rewrite (subst_open_commut_typ TR x0 y x _).
    eapply ty_trm_call with (TC := TC /[ x0 -> y]); eauto 3.

    1: unfold subst_var at 2; unfold VarSubstVar at 1.
    1,2: cases_if; eauto using substitution_neq_visible.

    all: unfold subst_var at 2; unfold VarSubstVar at 1.
    all: eauto.

    pose proof (subst_open_commut_typ TC x0 y y0 0).
    unfold subst_var at 2 in H. unfold VarSubstVar at 1 in H.
    rewrite <- H. eauto.

  - Case "ty_trm_sub". eapply ty_trm_sub.
    eapply H; eauto. eapply subst_ty_epush; eauto.

  - Case "ty_lit_obj".

    rewrite typ_subst_objlit.
    apply_fresh (@ty_lit_obj _) as z; eauto.
    assert (z <> x) as Hx1 by auto.

    rewrite <- (substi_ne_var y Hx1) at 2 3.
    rewrite <- subst_open_commut.
    rewrite <- subst_open_commut.
    eapply H; eauto. intros. inversions H1; eauto; try tauto.

  - Case "ty_def_typ".
    eapply ty_def_typ.
    intros.
    assert (x0 \notin (L \u \{x})) by apply H.
    specialize (t x0).
    pose proof (subst_tyq_typ). specialize (H6 y S). hnf in H6.

    rewrite <- subst_open_commut_ne.
    rewrite <- subst_open_commut_ne.
    eapply H6; eauto. notin_solve. notin_solve.


  - Case "ty_def_fld". destruct Go; unfold subst_objctx; sympls.

    + apply ty_def_fld; simpl; eauto.
      cases_if; eauto using substitution_neq_visible.

    + apply ty_def_fld.
      { unfold ctx_obj. intros. specialize (c H).
        cases_if; eauto using substitution_neq_visible.
        - destruct (fv_decide x0 \{y}); subst; eauto.
          rewrite in_singleton in m. subst.
          + inversions c; eauto using substitution_neq_visible.
          + inversions c; eauto using substitution_neq_visible. }

      unfold ctx_obj.
      assert (T0 /[ x -> y] /[-> x0] = T0 /[-> x0] /[ x -> y]).
      { rewrite subst_open_commut. rewrite substi_ne_var. auto.
      intro; subst. inversions H1. apply H9.
      rewrite ctx_dom_concat. rewrite in_union. left. unfold ctx_dom.
      rewrite in_union. left. rewrite in_singleton. auto. }
      rewrite H.
      rewrite <- ctx_concat_push.
      rewrite <- ctx_subst_push.
      eapply subst_ty_var_u; eauto.
      2: { simpl. destruct (fv_decide x0 \{y}).
           - rewrite in_singleton in m. subst. eauto.
           - apply notin_singleton_r in n. intros.
             eapply ctx_vis_n; eauto.
             specialize (H5 H0 H6). inversions H5; try tauto; eauto. }
      rewrite ctx_subst_push.
      rewrite ctx_concat_push.
      apply weaken_push; eauto. inversions H1. auto.
      apply ctx_ok_concat_map.
      eapply ctx_ok_push_concat; eauto.
      inversions H1. ctx_simpl_dom. notin_solve.
      rewrite ctx_dom_map. auto.

    + apply ty_def_fld. all: unfold ctx_obj; eauto.
      cases_if; eauto using substitution_neq_visible.

  - Case "ty_def_met".
    apply_fresh_base (@ty_def_met _) gather_vars xz.
    intros_fresh xc; intros.
    assert (xc <> x) as Hxc by auto.
    assert (xz <> x) as Hxz by auto.

    rename Frxc into Hxcxz, H0 into Frxc.
    assert (HxzL : xz \notin L) by eauto.
    assert (HxcL : xc \notin L) by eauto.

    rewrite ctx_obj_met_assoc in *.
    rewrite <- ctx_obj_met_subst_var in *.
    rewrite <- 4 subst_open_commut_ne in * by assumption.

    assert
      (x §# ctx_obj_met typing_mode ctx_empty Go xz TZ xc TC) as Hfrx.
    { rewrite ctx_obj_met_dom; notin_solve.
      rewrite objctx_dom_ctx_dom.
      apply ctx_ok_obj_disjoint in H2. hnf in H2.
      rewrite ctx_dom_concat, ctx_dom_push in H2.
      hnf. intros Hin. rewrite <- in_empty with (x := x), <- H2.
      repeat (rewrite in_inter); repeat (rewrite in_union);
        repeat (rewrite in_singleton).
      simpl in Hin; rewrite union_empty_l in Hin. tauto. }

    eapply H; eauto.

    + rewrite ctx_obj_met_assoc; auto.

    + rewrite <- (ctx_concat_empty_r G2), ctx_obj_met_assoc,
        ctx_map_concat, ctx_concat_assoc, ctx_obj_met_subst_var; simpl;
        try solve [ eapply ctx_obj_met_weakens_var; eauto ]; eauto.
      eapply ctx_obj_met_weakens_var; eauto.
      replace ctx_empty with (subst_ctx x y ctx_empty) by eauto.
      rewrite <- ctx_obj_met_subst_var; eauto 2.
      eapply ctx_ok_concat_map, ctx_ok_concat.
      * eapply ctx_ok_obj_inv_l, ctx_ok_push_concat in H2.
        eapply ctx_ok_concat_map; eauto.
      * eapply ctx_obj_met_ok_empty; eauto.
      * rewrite ctx_dom_concat, ctx_obj_met_dom, ctx_dom_map; simpl.
        hnf. repeat(rewrite inter_union_r); repeat(rewrite inter_union_l).
        eapply fset_extens; eauto using subset_empty_l. hnf.
        intros v Hin. rewrite in_empty.
        repeat(rewrite in_union in Hin); repeat(rewrite in_inter in Hin).
        repeat(rewrite in_singleton in Hin).
        destruct_all; subst; eauto.
        all: try solve[ apply Frxz; repeat (rewrite in_union); tauto
                      | apply Frxc; repeat (rewrite in_union); tauto ].
        all: apply ctx_ok_obj_disjoint in H2; hnf in H2.
        all: rewrite <- in_empty with (x := v), <- H2.
        all: rewrite objctx_dom_ctx_dom in H1; simpl in H1;
          rewrite union_empty_l in H1.
        all: rewrite in_inter; split; eauto.
        all: rewrite ctx_dom_concat, ctx_dom_push.
        all: repeat (rewrite in_union); tauto.

    + intros Hhide Hvis. exfalso.
      rewrite <- (ctx_concat_empty_r G2), ctx_obj_met_assoc,
        ctx_concat_assoc in Hvis; simpl; eauto.
      apply ctx_obj_met_vis_hide_inv in Hvis as [? | ?]; subst; eauto.

    + rewrite ctx_obj_met_dom. repeat(rewrite notin_union, notin_singleton).
      repeat(split); eauto 2. apply ctx_ok_obj_disjoint in H2 as Hdisjoint.
      apply ctx_ok_obj_inv_l in H2. rewrite objctx_dom_ctx_dom.
      rewrite notin_union; split.
      * apply ctx_ok_middle_inv in H2 as [_ [_ HxG2fr]]. assumption.
      * hnf; hnf in Hdisjoint. intros Hin. rewrite <- in_empty with (x := x).
        rewrite <- Hdisjoint, in_inter; split; auto 2.
        repeat(rewrite ctx_dom_concat);
          repeat(rewrite ctx_dom_push); repeat(rewrite in_union).
        rewrite in_singleton; auto 3.

  - Case "ty_defs_cons".
    constructor*. rewrite <- subst_label_of_def. apply* subst_defs_hasnt.

Qed.


Global Instance subst_ty_trm: Substitution subst_trm subst_var (ty_trm _).
Proof.
  unfold Substitution. intros.
  eapply subst_rules_ind; eauto.
Qed.
Hint Resolve subst_ty_trm: core.

Global Instance subst_ty_lit: Substitution subst_lit subst_var (ty_lit _).
Proof.
  unfold Substitution. intros.
  eapply subst_rules_ind; eauto.
Qed.
Hint Resolve subst_ty_lit: core.

Global Instance subst_ty_defs: Substitution subst_defs subst_var (ty_defs _).
Proof.
  unfold Substitution. intros.
  eapply subst_rules_ind; eauto.
Qed.
Hint Resolve subst_ty_defs: core.

Global Instance subst_ty_def: Substitution subst_def subst_var (ty_def _).
Proof.
  unfold Substitution. intros.
  eapply subst_rules_ind; eauto.
Qed.
Hint Resolve subst_ty_def: core.

Lemma subst_env_vars_ok : forall xs ys G1 G2,
    ok (G1 & G2) ->
    ok (G1 & subst_env_vars xs ys G2).
Proof.
  intros. apply ok_concat_map; auto.
Qed.



Definition VKChangeInd x vk1 vk2 S G1 (A B: Set) (property: TypingProperty A B) :=
  forall G Go e (a: A) (b: B),
    property G Go e a b ->
    forall G2,
    G = G1 § x / vk1 ~ S §& G2 ->
    property (G1 § x / vk2 ~ S §& G2) Go e a b.
Definition VKChangeTInd x vk1 vk2 S G1 (A: Set) (AT: Typing A) := VKChangeInd x vk1 vk2 S G1 obj_typing.
Definition VKChangeIncapInd x vk1 vk2 S G1 (_: ReferenceIncapacitate) :=
  VKChangeInd x vk1 vk2 S G1 obj_ref_incap.
Definition VKChangeCapbndInd x vk1 vk2 S G1
           (C : typ_refcap_label)
           (_ : CapabilityBound C) :=
  VKChangeInd x vk1 vk2 S G1 obj_cap_bnd.
Definition VKChangeDInd x vk1 vk2 S G1 (A: Set) (AT: DecTyping A) := VKChangeInd x vk1 vk2 S G1 obj_dec_typing.


Lemma vkchange_ts_ind: forall x vk1 vk2 S G1,
  VKChangeTInd x vk1 vk2 S G1 (ty_var _) /\
  VKChangeInd x vk1 vk2 S G1 subtyping /\
  VKChangeIncapInd x vk1 vk2 S G1 (ty_incap _).
Proof.
  intros.
  apply ts_mutind; intros; subst; eauto.
  - pose proof (ctx_binds_middle_inv _ _ _ _ _ c).
    destruct_all; subst;
    apply ty_var_var;
    eapply ctx_binds_middle_vkchange; eauto.

  - eapply subtyp_met; eauto. introv Hx0fr Hneqyx0 Hy0fr.
    clear H s0 H0 s1.
    specialize (H1 _ Hx0fr _ Hneqyx0 Hy0fr).
    replace (ctx_met ((G1 § x / vk1 ~ S) §& G2) x0 TZ2 y TC2)
      with ((G1 § x / vk1 ~ S) §& ctx_met G2 x0 TZ2 y TC2)
      in H1 by auto.
    specialize (H1 _ eq_refl).
    replace (ctx_met ((G1 § x / vk2 ~ S) §& G2) x0 TZ2 y TC2)
      with ((G1 § x / vk2 ~ S) §& ctx_met G2 x0 TZ2 y TC2)
      by auto. assumption.
Qed.

Lemma vkchange_ty_capbnd: forall x vk1 vk2 S G1 C,
    VKChangeCapbndInd x vk1 vk2 S G1 (ty_capbnd _ C).
Proof.
  introv. hnf. induction 1; intros; subst; eauto.
  apply ty_capbnd_sel with (T := T) (S := S0); eauto.
  pose proof vkchange_ts_ind as [? _].
  eapply H1; eauto.
Qed.

Lemma vkchange_ctx_vis: forall Go x x0 vk1 vk2 S G1 G2 ,
    ctx_vis ((G1 § x / vk1 ~ S) §& G2 §§& Go) x0 ->
    ctx_vis ((G1 § x / vk2 ~ S) §& G2 §§& Go) x0.
Proof.
  introv Hvis. remember ((G1 § x / vk1 ~ S) §& G2 §§& Go) as G.
  generalize dependent G1. generalize dependent G2.
  generalize dependent Go.
  generalize dependent S.
  generalize dependent vk1. generalize dependent vk2.
  generalize dependent x.

  induction Hvis; intros; subst.
  - induction Go; simpl in *; eauto.
    + apply ctx_push_concat_inv in HeqG.
      destruct HeqG as [[? Heq] | [G2' Heq]]; subst; simpl; eauto.
      inversions Heq. eauto.
    + inversions HeqG; eauto.
    + apply ctx_push_concat_inv in HeqG.
      destruct HeqG as [[? Heq] | [G2' Heq]]; subst; simpl; eauto.
      inversions Heq. eauto.
  - induction Go; simpl in *; eauto.
    + apply ctx_push_concat_inv in HeqG as HeqG'.
      destruct HeqG' as [[? Heq] | [G2' Heq]]; subst; simpl; eauto.
      * inversions Heq. destruct (fv_decide x \{v}); eauto.
      * simpl in HeqG. inversions HeqG.
        destruct (fv_decide v2 \{v}); eauto.
        { rewrite in_singleton in m; subst; eauto. }
        { apply notin_singleton_r in n.
          eapply ctx_vis_n; eauto.
          specialize (IHHvis x vk2 vk1 S objctx_empty G2' G1).
          simpl in IHHvis. eauto. }
    + inversions HeqG. destruct (fv_decide v \{x0}); eauto.
      * rewrite in_singleton in m; subst; eauto.
      * apply notin_singleton_r in n.
        eapply ctx_vis_n; eauto.
        specialize (IHHvis x vk2 vk1 S objctx_empty G2 G1).
        simpl in IHHvis. eauto.
    + apply ctx_push_concat_inv in HeqG as HeqG'.
      destruct HeqG' as [[? Heq] | [G2' Heq]]; subst; simpl; eauto.
      * inversions Heq. destruct (fv_decide x \{v}); eauto.
      * simpl in HeqG. inversions HeqG.
        destruct (fv_decide v2 \{v}); eauto.
        { rewrite in_singleton in m; subst; eauto. }
        { apply notin_singleton_r in n.
          eapply ctx_vis_n; eauto.
          specialize (IHHvis x vk2 vk1 S objctx_empty G2' G1).
          simpl in IHHvis. eauto. }
Qed.

Corollary vkchange_ctx_vis_emptyctx: forall x x0 vk1 vk2 S G1 G2,
    ctx_vis ((G1 § x / vk1 ~ S) §& G2) x0 ->
    ctx_vis ((G1 § x / vk2 ~ S) §& G2) x0.
Proof.
  intros. pose proof (vkchange_ctx_vis objctx_empty). simpl in H0. eauto.
Qed.

Lemma vkchange_viewpoint_adapt: forall x x0 vk1 vk2 S T U G1 G2 e,
    (G1 § x / vk1 ~ S) §& G2 ⁏ e ⊢ x0 ▷ T ⟶ U ->
    (G1 § x / vk2 ~ S) §& G2 ⁏ e ⊢ x0 ▷ T ⟶ U.
Proof.
  introv Hadapt. hnf in Hadapt.
  pose proof vkchange_ty_capbnd as Hbnd. unfold VKChangeCapbndInd in Hbnd.
  pose proof vkchange_ts_ind as Hicp. unfold VKChangeIncapInd in Hicp.
  unfold viewpoint_adapt in *.
  destruct (mode_viewpoint_adaptation' typing_mode); eauto.
  unfold VKChangeInd in *. inversions Hadapt.
  specialize (Hicp x vk1 vk2 S G1). destruct Hicp as [_ [_ Hicp]].
  specialize (Hicp ((G1 § x / vk1 ~ S) §& G2) objctx_empty e T Tstrip H
                   _ eq_refl).
  specialize (Hbnd x vk1 vk2 S G1 typ_refcap_mutability
                   ((G1 § x / vk1 ~ S) §& G2) objctx_empty e T Tcap H0
                   _ eq_refl).
  constructor; eauto.
Qed.

Lemma vkchange_ty_trm: forall x vk1 vk2 S G1,
  VKChangeTInd x vk1 vk2 S G1 (ty_trm _) /\
  VKChangeTInd x vk1 vk2 S G1 (ty_lit _) /\
  VKChangeDInd x vk1 vk2 S G1 (ty_def _) /\
  VKChangeTInd x vk1 vk2 S G1 (ty_defs _).
Proof.
  intros.
  set (vkchange_ts_ind x vk1 vk2 S G1).
  destruct a as [? [? _]].
  apply rules_mutind; intros; subst; auto.

  all: econstructor;
    eauto 6 using vkchange_ctx_vis_emptyctx, vkchange_viewpoint_adapt,
    vkchange_ctx_vis.

  - Case "ty_def_fld".
    destruct Go; unfold ctx_obj in *. all: eauto.
  - Case "ty_def_met". introv Hxzfr Hneq Hxcfr.
    destruct Go; unfold ctx_obj, ctx_obj_met in *.
    + destruct (mode_visibility' typing_mode).
      * rewrite <- 2 ctx_concat_push; eauto.
      * specialize (H1 _ Hxzfr _ Hneq Hxcfr
                     (ctx_met (ctx_hide G2) xz TZ xc TC)).
        simpl in H1. eauto.
    + destruct (mode_visibility' typing_mode).
      * rewrite <- 3 ctx_concat_push; eauto.
      * specialize (H1 _ Hxzfr _ Hneq Hxcfr
                     (ctx_met_refined_heap
                        (ctx_hide (G2 § x0 / v_self ~ open x0 T))
                        xz TZ xc TC x0 T)).
      simpl in H1. eauto.
    + destruct (mode_visibility' typing_mode).
      * rewrite <- 2 ctx_concat_push; eauto.
      * specialize (H1 _ Hxzfr _ Hneq Hxcfr
                     (ctx_met_refined_heap
                        (ctx_hide G2) xz TZ xc TC y T)).
      simpl in H1. eauto.
Qed.

Lemma weaken_middle_trm : forall y' vky n T' G T T'' e (t : trm) x x' vkx,
    ctx_ok (G § x / vkx ~ open_rec n x' T) ->
    ok e ->
    renv_notin y' e ->
    y' \notin (fv T \u fv T' \u fv T'' \u fv t \u ctx_fv G \u ctx_dom G \u \{x}) ->
    G § y' / vky ~ T' § x / vkx ~ open_rec n y' T ⁏e ⊢ open_rec n y' t ∶ open_rec n y' T'' ->
    G⁏e ⊢ x' ∶ T' ->
    (mode_has_visibility typing_mode ->
     ctx_vis ((G § y' / vky ~ T') § x / vkx ~ open_rec n y' T) y' ->
     ctx_vis (G §& subst_ctx y' x' (x / vkx ~ open_rec n y' T)) x') ->
    G § x / vkx ~ open_rec n x' T ⁏e ⊢ open_rec n x' t ∶ open_rec n x' T''.
Proof.
  introv Hok Hoke Hne Hn2 Htyt Htyx Hvis.
  repeat rewrite (subst_intro (x:=y') x' n); auto.
  pose proof Hok.
  apply ctx_ok_push_inv_ok in Hok.
  apply ctx_ok_push_inv_unbound in H.
  rewrite ctx_concat_single.
  rewrite <- ctx_subst_single.
  eapply substitution with (property := (ty_trm _)); eauto.

  all: try rewrite  <- ctx_concat_single.
  all: try eassumption. all: auto 3.
  rewrite subst_fresh; auto.
  apply weaken_push; auto.
Qed.

(* Substituting reference to location makes equivalent types *)
Lemma tyq_deref_ref (e: renv) (v1 v2: var):
  (binds v1 v2 e) ->
  forall (v3: var),
    typing_equiv e v3 (subst_var v1 v2 v3).
Proof.
  intros.
  unfold subst_var, VarSubstVar.
  destruct (classicT (v3=v1));
  subst; auto.
Qed.

Lemma tyq_deref_var (e: renv) (v1 v2: var):
  typing_equiv e v1 v2 ->
  forall (v3: var),
  typing_equiv e v3 (v3 /[ v1 -> v2 ]).
Proof.
  introv H.
  induction H; intro.
  - rewrite subst_id. auto.
  - apply tyq_deref_ref; auto.
  -
    set (IHtyq_var v3).
    set (IHtyq_var (v3 /[ v2 -> v1])).
    (* using the fact that substituting back means ignore the first substitution *)
    rewrite subst_back in t0.
    symmetry in t0.
    apply (tyq_var_trans t t0).
  - (* cases on if the middle variable is the source variable *)
    set (classicT (v0=v2)).
    destruct s; subst; auto.
    + destruct (subst_cases v1 v3 v2);
      rewrite H1; auto.
    + set (IHtyq_var v0).
      set (IHtyq_var0 (v0 /[ v1 -> v2])).
      set (tyq_var_trans t t0).
      set (subst_trans v1 v3 n).
      rewrite e0 in t1.
      auto.
Qed.


End WithTypingMode.

Hint Resolve subst_sub: core.
Hint Resolve subst_ty_var: core.
Hint Resolve subst_ty_trm: core.
Hint Resolve subst_ty_lit: core.
Hint Resolve subst_ty_defs: core.

(** An alternative characterisation of a sealed type. Basically a type
    T is sealed at the de Bruijn level n if anything above level n
    isn't in part of its free variables. The 'base case' is given by
    is_sealed_sel. *)
Inductive is_sealed_typ : nat -> typ -> Prop :=
| is_sealed_top : forall n, is_sealed_typ n typ_top
| is_sealed_bot : forall n, is_sealed_typ n typ_bot
| is_sealed_rcd : forall n D, is_sealed_dec n D -> is_sealed_typ n (typ_rcd D)
| is_sealed_and : forall n T1 T2, is_sealed_typ n T1 ->
                             is_sealed_typ n T2 ->
                             is_sealed_typ n (typ_and T1 T2)
| is_sealed_or : forall n T1 T2, is_sealed_typ n T1 ->
                             is_sealed_typ n T2 ->
                             is_sealed_typ n (typ_or T1 T2)
| is_sealed_sel : forall n a L a', (forall n', n' >= n ->
                                     (avar_b n') <> a /\ (avar_b n') <> a') ->
                              is_sealed_typ n (typ_sel a L a')
| is_sealed_bnd : forall n T, is_sealed_typ (S n) T ->
                         is_sealed_typ n (typ_bnd T)
| is_sealed_fun : forall n T1 T2, is_sealed_typ n T1 ->
                             is_sealed_typ (S n) T2 ->
                             is_sealed_typ n (typ_fun T1 T2)
| is_sealed_nocap : forall n, is_sealed_typ n typ_nocap
with is_sealed_dec : nat -> dec -> Prop :=
| is_sealed_dec_typ : forall n L T U, is_sealed_typ (S n) T ->
                                 is_sealed_typ (S n) U ->
                                 is_sealed_dec n (dec_typ L T U)
| is_sealed_dec_fld : forall n L T U, is_sealed_typ n T ->
                                 is_sealed_typ n U ->
                                 is_sealed_dec n (dec_fld L T U)
| is_sealed_dec_met : forall n L TC TZ TR, is_sealed_typ (S n) TC ->
                                      is_sealed_typ n TZ ->
                                      is_sealed_typ (S (S n)) TR ->
                                      is_sealed_dec n (dec_met L TC TZ TR).
Hint Constructors is_sealed_typ is_sealed_dec : core.
Scheme is_sealed_typ_mut  := Induction for is_sealed_typ Sort Prop
with   is_sealed_dec_mut  := Induction for is_sealed_dec Sort Prop.
Combined Scheme is_sealed_mutind from is_sealed_typ_mut, is_sealed_dec_mut.

Class IsSealed (A : Set) := is_sealed : nat -> A -> Prop.
Instance TypIsSealed : IsSealed typ := is_sealed_typ.
Instance DecIsSealed : IsSealed dec := is_sealed_dec.
Existing Instances TypIsSealed DecIsSealed.


(* This is an equivalent characterisation of sealedness *)

Lemma sealed_is_sealed :
  (forall (T : typ) (n : nat),
      (forall (v : var), seal_rec n v T = T) -> is_sealed n T)
  /\ (forall (D : dec) (n : nat),
        (forall (v : var), seal_rec n v D = D) -> is_sealed n D).
Proof.
  apply typ_mutind; unfold is_sealed, TypIsSealed, DecIsSealed in *;
    try solve [ intros; eauto
              | introv IH Hs; constructor; eapply IH; intros v;
                specialize (Hs v); simpl in Hs; inversions Hs;
                rewrite H0 at 1; auto
              | introv IH1 IH2 Hs; constructor;
                try(eapply IH1); try(eapply IH2);
                intros v; specialize (Hs v); simpl in Hs; inversions Hs;
                try(rewrite H0 at 1); try(rewrite H1 at 1); auto ].

  - introv Hs. rename t into L, a0 into a'. constructor. introv Hge.
    assert (Hnea : forall m, a = avar_b m -> m < n).
    { introv Heq; subst. pick_fresh v. specialize (Hs v).
      destruct (classicT (n <= m)); try omega. inversions Hs. }
    assert (Hnea' : forall m', a' = avar_b m' -> m' < n).
    { introv Heq; subst. pick_fresh v. specialize (Hs v).
      destruct (classicT (n <= m')); try omega. inversions Hs. }
    destruct a; destruct a'; simpl in Hs.
    + specialize (Hnea _ eq_refl); specialize (Hnea' _ eq_refl).
      split; injection; try omega.
    + specialize (Hnea _ eq_refl). split; try congruence.
      injection; try omega.
    + specialize (Hnea' _ eq_refl). split; try congruence.
      injection; try omega.
    + split; try congruence.

  - introv IH1 IH2 IH3 Hs; constructor;
      try(eapply IH1); try(eapply IH2); try(eapply IH3);
      intros v; specialize (Hs v); simpl in Hs; inversions Hs;
      try(rewrite H0 at 1); try(rewrite H1 at 1); try(rewrite H2 at 1); auto.
Qed.

Lemma is_sealed_sealed :
  (forall (n : nat) (T : typ),
      is_sealed_typ n T -> (forall (v : var), seal_rec n v T = T))
  /\ (forall (n : nat) (D : dec),
        is_sealed_dec n D -> (forall (v : var), seal_rec n v D = D)).
Proof.
  apply is_sealed_mutind;
    try solve [ introv; eauto
              | unfold seal_rec, DecSealable, TypSealable; intros;
                simpl; congruence ].

  introv Hidx. intros v. simpl.
  assert (Hnea : forall m, a = avar_b m -> m < n).
  { intros m ?; subst. destruct (Nat.le_gt_cases n m); eauto.
    specialize (Hidx _ H). try tauto. }
  assert (Hnea' : forall m', a' = avar_b m' -> m' < n).
  { intros m' ?; subst. destruct (Nat.le_gt_cases n m'); eauto.
    specialize (Hidx m' H). try tauto. }

  destruct a; destruct a'; unfold seal_rec, AvarSealable; auto.
  - specialize (Hnea _ eq_refl). specialize (Hnea' _ eq_refl).
    destruct (classicT (n <= n0)); try omega.
    destruct (classicT (n <= n1)); try omega. auto.
  - specialize (Hnea _ eq_refl).
    destruct (classicT (n <= n0)); try omega. auto.
  - specialize (Hnea' _ eq_refl).
    destruct (classicT (n <= n0)); try omega. auto.
Qed.

Lemma is_sealed_iff_sealed_typ (T : typ) :
  is_sealed 0 T <-> sealed T.
Proof.
  split. pose proof (proj1 is_sealed_sealed 0 T). auto.
  pose proof (proj1 sealed_is_sealed T 0). auto.
Qed.

Lemma is_sealed_iff_sealed_dec (D : dec) :
  is_sealed 0 D <-> sealed D.
Proof.
  split. pose proof (proj2 is_sealed_sealed 0 D). auto.
  pose proof (proj2 sealed_is_sealed D 0). auto.
Qed.

Lemma sealed_subst_mutind :
  (forall (n : nat) (T : typ),
      is_sealed_typ n T -> (forall (x y : var), is_sealed_typ n (T /[ x -> y])))
  /\ (forall (n : nat) (D : dec),
        is_sealed_dec n D -> (forall (x y : var), is_sealed_dec n (D /[ x -> y]))).
Proof.
  apply is_sealed_mutind; simpl in *; eauto.
  introv Hidx. intros x y. constructor. introv Hge.
  specialize (Hidx _ Hge).
  destruct a; destruct a'; simpl; split; destruct_all; eauto; congruence.
Qed.

Lemma typ_sealed_subst (T : typ) x y :
  sealed T -> sealed (T /[ x -> y]).
Proof.
  repeat(rewrite <- is_sealed_iff_sealed_typ).
  pose proof (proj1 sealed_subst_mutind 0 T). introv Hs. eapply H. eauto.
Qed.

Lemma ctx_sealed_subst (G : ctx) x y :
  sealed G -> sealed G /[§ x -> y].
Proof.
  induction G; simpl in *; introv Hs;
    eauto using ctx_sealed_push, ctx_closed_hide_G, ctx_closed_stop_G.
  eapply ctx_sealed_push. eapply ctx_closed_push_G in Hs. eauto.
  eapply typ_sealed_subst. eapply ctx_closed_push_T. eauto.
Qed.
