Set Implicit Arguments.

Require Import Context Equivalence Indep GeneralTyping TFree.

Require Import LibExtra DotTactics AbstractSyntax TypingMode.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax TypingMode GeneralTyping Equivalence
        InertTightSubtyping  Indep.

(** T-Free variables (free variables used as values in terms) *)
(* In a typed term, all tfree variables are in the context.
   In rodot, they are also all bound after the last hide mark.
 *)

Section WithTypingMode.
Context (typing_mode: Mode).

Class TFreeDom (A B: Set) (AT: TFree A) (property: TypingProperty A B) :=
  tfree_dom G1 Go e (a: A) (b: B):
    property G1 Go e a b ->
    forall x,
    tfree x a ->
    x \in ctx_dom (G1 §§& Go).
Hint Unfold TFreeDom: core.

Definition TFreeDomInd (A B: Set) (AT: TFree A) (property: TypingProperty A B) :=
  forall G Go e (a: A) (b: B),
    property G Go e a b ->
    forall x,
    tfree x a ->
    x \in ctx_dom (G §§& Go).
Definition TFreeDomTInd (A: Set) (AT: TFree A) (t: Typing A) := TFreeDomInd AT obj_typing.
Definition TFreeDomDInd (A: Set) (AT: TFree A) (t: DecTyping A) := TFreeDomInd AT obj_dec_typing.

Lemma tfree_dom_rules_ind:
  TFreeDomTInd _ (ty_trm _) /\
  TFreeDomTInd _ (ty_lit _) /\
  TFreeDomDInd _ (ty_def _) /\
  TFreeDomTInd _ (ty_defs _).
Proof.
  apply rules_mutind.
  all: intros; unfold ctx_obj in *; unfold ctx_obj_met in *.
  all: match goal with
    H: tfree _ _ |- _ => inversions H;  eauto using typing_implies_dom
  end.
  -
    pick_fresh x0.
    assert ((x \in \{ x0} \u ctx_dom G) /\ x <> x0).
    split; eauto.
    destruct H1.
    apply (in_union_singleton_ne_l H1 H2).
  -
    pick_fresh x0.
    assert ((x \in \{ x0} \u ctx_dom G) /\ x <> x0).
    split; eauto.
    destruct H1.
    apply (in_union_singleton_ne_l H1 H2).
  -
    pick_fresh x0.
    assert ((x \in \{ x0} \u ctx_dom G) /\ x <> x0).
    split; eauto.
    destruct H0.
    apply (in_union_singleton_ne_l H0 H1).
  -
    pick_fresh x0.
    assert ((x \in \{ x0} \u ctx_dom G) /\ x <> x0).
    split; eauto.
    destruct H0.
    apply (in_union_singleton_ne_l H0 H1).
  - destruct Go.
    +
      pick_fresh x0.
      pick_fresh x1.
      assert ((x \in \{ x0} \u \{ x1} \u ctx_dom G) /\ x <> x0 /\ x <> x1).
      split; eauto.
      destruct typing_mode; eauto.
      destructc H0 as ? ? ?.
      eapply in_union_singleton_ne_l.
      eapply in_union_singleton_ne_l.
      all: eauto.
    +
      rename x0 into y.
      pick_fresh x0.
      pick_fresh x1.
      assert ((x \in \{ x0} \u \{ x1}  \u \{ y}  \u ctx_dom G) /\ x <> x0 /\ x <> x1) by (split; destruct typing_mode; eauto).
      destructc H0 as ? ? ?.
      eapply in_union_singleton_ne_l.
      eapply in_union_singleton_ne_l.
      all: eauto.
    +
      pick_fresh x0.
      pick_fresh x1.
      assert ((x \in \{ x0} \u \{ x1} \u ctx_dom G) /\ x <> x0 /\ x <> x1) by (split; destruct typing_mode; eauto).
      destructc H0 as ? ? ?.
      eapply in_union_singleton_ne_l.
      eapply in_union_singleton_ne_l.
      all: eauto.
  - inversions H3.
Qed.

Global Instance tfree_dom_trm: TFreeDom _ (@ty_trm _).
Proof.
  unfold TFreeDom.
  intros.
  (* Need to select the first conclusion because they do not depend on the set A *)
  eapply (proj41 tfree_dom_rules_ind); eauto.
Qed.
End WithTypingMode.

Section WithTypingMode.
Context (typing_mode: Mode).
Context (Hmv: mode_visibility' typing_mode = mode_visibility_hide_method).

Class TFreeVisible (A B: Set) (AT: TFree A) (property: TypingProperty A B) :=
  tfree_visible G1 G2 Go e (a: A) (b: B):
    property ((ctx_hide G1) §& G2) Go e a b ->
    forall x,
    tfree x a ->
    x \in ctx_dom (G2 §§& Go).
Hint Unfold TFreeVisible: core.

Definition TFreeVisibleInd (A B: Set) (AT: TFree A) (property: TypingProperty A B) :=
  forall G Go e (a: A) (b: B),
    property G Go e a b ->
    forall G1 G2,
      G = (ctx_hide G1) §& G2 ->
      forall x,
      tfree x a ->
      x \in ctx_dom (G2 §§& Go).
Definition TFreeVisibleTInd (A: Set) (AT: TFree A) (t: Typing A) := TFreeVisibleInd AT obj_typing.
Definition TFreeVisibleDInd (A: Set) (AT: TFree A) (t: DecTyping A) := TFreeVisibleInd AT obj_dec_typing.

Lemma tfree_visible_rules_ind:
  TFreeVisibleTInd _ (ty_trm _) /\
  TFreeVisibleTInd _ (ty_lit _) /\
  TFreeVisibleDInd _ (ty_def _) /\
  TFreeVisibleTInd _ (ty_defs _).
Proof.
  apply rules_mutind.
  all: intros; subst.
  all: unfold ctx_obj, ctx_obj_met in *.

  all: match goal with
    H: tfree _ _ |- _ => inversions H; eauto using ctx_vis_hide_dom
  end.
  - pick_fresh x0.
    assert (x \in ctx_dom (G2 § x0 / v_let ~ T)).
    eapply (H0 x0); auto.
    ctx_flatten. auto.
    ctx_simpl_dom.
    eapply in_union_singleton_ne_l; eauto.
  - pick_fresh x0.
    assert (x \in ctx_dom (G2 § x0 / v_let ~ T)).
    eapply (H0 x0); auto.
    ctx_flatten. auto.
    ctx_simpl_dom.
    eapply in_union_singleton_ne_l; eauto.
  - pick_fresh x0.
    assert (x \in ctx_dom (G2 § x0 / v_par ~ T)).
    eapply (H x0); auto.
    ctx_flatten. auto.
    ctx_simpl_dom.
    eapply in_union_singleton_ne_l; eauto.
  - pick_fresh x0.
    assert (x \in ctx_dom (G2 § x0 / v_par ~ T)).
    eapply (H x0); auto.
    ctx_flatten. auto.
    ctx_simpl_dom.
    eapply in_union_singleton_ne_l; eauto.
  - destruct Go;
    eauto using ctx_vis_hide_dom.
  - destruct Go.
    +
      pick_fresh x0.
      pick_fresh x1.
      assert (x \in ctx_dom (ctx_met (ctx_hide G2) x0 TZ x1 TC)).
      eapply (H x0 _ x1); auto.
       rewrite Hmv.
      ctx_flatten. auto.
      ctx_simpl_dom.
      eapply in_union_singleton_ne_l.
      eapply in_union_singleton_ne_l.
      Unshelve.
      all: eauto.
    +
      rename x0 into y.
      pick_fresh x0.
      pick_fresh x1.
      assert (x \in ctx_dom (ctx_met_refined_heap (ctx_hide (G2 § y / v_self ~ open y T)) x0 TZ x1 TC y T)).
      eapply (H x0 _ x1); auto.
 rewrite Hmv.
      ctx_flatten. auto.
      ctx_simpl_dom.
      eapply in_union_singleton_ne_l.
      eapply in_union_singleton_ne_l.
      Unshelve.
      all: eauto.
    +
      pick_fresh x0.
      pick_fresh x1.
      assert (x \in ctx_dom (ctx_met_refined_heap (ctx_hide G2) x0 TZ x1 TC y T)).
      eapply (H x0 _ x1); auto.
 rewrite Hmv.
      ctx_flatten. auto.
      ctx_simpl_dom.
      eapply in_union_singleton_ne_l.
      eapply in_union_singleton_ne_l.
      Unshelve.
      all: eauto.
  - inversions H3.
Qed.

Global Instance tfree_visible_trm: TFreeVisible _ (@ty_trm typing_mode).
Proof.
  unfold TFreeVisible.
  intros.
  (* Need to select the first conclusion because they do not depend on the set A *)
  eapply (proj41 tfree_visible_rules_ind); eauto.
Qed.

Lemma tfree_met_params G xz TZ xc TC w1 S e t TR:
  ctx_met_refined_heap (ctx_hide G) xz TZ xc TC w1 S ⁏ e ⊢ open_met xz xc t ∶ TR ->
  forall x,
  tfree x (open_met xc xz t) ->
  (x = xc) \/ (x = xz).
Proof.
  introv Hty Htf.

  eapply tfree_open_inv_trm in Htf.
  destruct Htf.
  eapply tfree_open_inv_trm in H.
  destruct H.

  ctx_split_from xz in Hty.

  assert (x \in ctx_dom (ctx_met_refined_heap ctx_empty xz TZ xc TC w1 S §§& objctx_empty)).
  eapply tfree_visible.
  apply Hty.
  eapply tfree_open.
  eapply tfree_open.
  auto.

  unfold ctx_obj in *.
  ctx_simpl_dom.
  repeat rewrite in_union in H0.
  destruct H0; auto.
  rewrite in_singleton in H0; auto.
  destruct H0; auto.
  rewrite in_singleton in H0; auto.
  eapply in_empty_inv in H0; contradiction.
  auto.
  auto.
Qed.

Lemma tfree_met_renaming (t: trm) xc xz w1 w2:
  xc \notin fv t ->
  xz \notin fv t ->
  (forall x, tfree x (open_met xc xz t) -> (x = xc) \/ (x = xz)) ->
  (forall x, tfree x (open_met w2 w1 t) -> (x = w1) \/ (x = w2)).
Proof.
  introv Hnc Hnz Hty Htf.
  eapply tfree_open_inv_trm in Htf.
  destruct Htf.
  eapply tfree_open_inv_trm in H.
  destruct H.
  - assert (tfree x (open_met xc xz t)).
    eapply tfree_open.
    eapply tfree_open.
    auto.
    destruct (Hty x H0); subst.
    all: eapply tfree_in_fv in H.
    all: contradiction.
  - auto.
  - auto.
Qed.

End WithTypingMode.

Hint Resolve tfree_dom_trm: core.
Hint Resolve tfree_visible_trm: core.