Set Implicit Arguments.

Require Import TypingMode.
Require Import LibExtra DotTactics AbstractSyntax.

(** * Definitions and lemmas about typing contexts. *)

(* In rodot, oustide variables are hidden to method body *)
Notation vis_hide typing_mode G := (
  match (mode_visibility' typing_mode) with
  | mode_visibility_all => G (* Everything is visible *)
  | mode_visibility_hide_method => (ctx_hide G) (* External variables hidden in a method *)
  end
  ) (only parsing).

(* Merges ctx and objctx into a context for typing the content of method bodies.
   This differ from `ctx_met (ctx_obj G Go)` in that this allows the receiver
   variable to get a more specific type. *)
Definition ctx_obj_met typing_mode (G: ctx)(Go: objctx)(xz: var)(TZ: typ)(xc: var)(TC: typ) :=
  match Go with
  | objctx_empty => ctx_met (vis_hide typing_mode G) xz TZ xc TC
  | objctx_lit x T => ctx_met_refined (vis_hide typing_mode (G § x / v_self ~ open x T)) xz TZ xc TC (typ_and (open x T) (open xc T))
  | objctx_heap y T => ctx_met_refined_heap (vis_hide typing_mode G) xz TZ xc TC y T
  (* Version without special treatment of xc:
  | objctx_lit x T => ctx_met (G § x / v_self ~ open x T) xz TZ xc TC
  | objctx_heap y T => ctx_met G xz TZ xc TC*)
  end.

Ltac notin_ctx_obj_solve := (unfold ctx_obj, ctx_obj_met in *; ctx_simpl_dom; notin_solve).

Lemma ctx_obj_met_assoc typing_mode G1 G2 Go xz TZ xc TC :
  ctx_obj_met typing_mode (G1 §& G2) Go xz TZ xc TC = G1 §& ctx_obj_met typing_mode (G2) Go xz TZ xc TC.
Proof.
  unfold ctx_obj, ctx_obj_met; destruct typing_mode; destruct Go; ctx_flatten; auto.
Qed.

Lemma ctx_obj_met_ok typing_mode G1 G2 Go xz TZ xc TC:
  ctx_ok (G1 §& G2 §§& Go) ->
  xz \notin (ctx_dom G1 \u ctx_dom G2 \u objctx_dom Go) ->
  xc <> xz ->
  xc \notin (ctx_dom G1 \u ctx_dom G2 \u objctx_dom Go) ->
  ctx_ok (G1 §& ctx_obj_met typing_mode G2 Go xz TZ xc TC).
Proof.
  intros.
  unfold ctx_obj_met. destruct Go; destruct typing_mode; simpls; constructor; ctx_simpl_dom; notin_solve; ctx_simpl_dom; notin_solve; ctx_simpl_dom; notin_solve.
  all: constructor; ctx_simpl_dom; notin_solve; ctx_simpl_dom; notin_solve; ctx_simpl_dom; notin_solve; eauto.
Qed.
Hint Resolve ctx_obj_met_ok: core.

Lemma ctx_obj_met_dom typing_mode G Go xz TZ xc TC:
  ctx_dom (ctx_obj_met typing_mode G Go xz TZ xc TC) = \{xz} \u \{xc} \u ctx_dom (G §§& Go) .
Proof.
  destruct Go;
  unfold ctx_obj_met, ctx_obj in *; destruct (mode_visibility' typing_mode); ctx_simpl_dom; ctx_simpl_dom; eauto.
Qed.

Lemma ctx_obj_met_ok_empty typing_mode Go xz TZ xc TC :
  xc <> xz ->
  xc \notin objctx_dom Go ->
  xz \notin objctx_dom Go ->
  ctx_ok (ctx_obj_met typing_mode ctx_empty Go xz TZ xc TC).
Proof.
  introv Hneq Hninc Hninz. destruct Go; simpl.
  all: destruct (mode_visibility' typing_mode).
  all: repeat(constructor); simpl; try(rewrite notin_union);
    try(split); eauto 3.
  all: apply notin_singleton; auto.
Qed.
