(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Weakening EWeakening Substitution.

(** * Typing helper lemmas. *)
(* This file contains lemmas about typing, which use weakening, substitution, etc *)

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma ty_trm_let_inv : forall G e t u U,
  ctx_ok G ->
  G⁏e ⊢ trm_let t u ∶ U ->
  exists T L,
  G⁏e ⊢ t ∶ T /\
  (forall x, x \notin L ->
    G § x / v_let ~ T ⁏e ⊢ open x u ∶ U).
Proof.
  introv Hok Hty.
  remember (trm_let t u) as Tlet.
  induction Hty; intros; inversions HeqTlet; eauto.
  forwards T0 L Htyt Htyu: IHHty; eauto.
  exists T0 (L \u ctx_dom G). splits. auto. intros.
  assert (x\notin L) as Hx by eauto. specialize (Htyu x Hx). eapply ty_trm_sub. eauto. eapply weaken_push; eauto.
Qed.

Lemma ty_trm_lit_inv : forall G e t u U,
  ctx_ok G ->
  G⁏e ⊢ trm_lit t u ∶ U ->
  exists T L,
  G⁏e ⊢ t ∶ T /\
  (forall x, x \notin L ->
    G § x / v_let ~ T ⁏e ⊢ open x u ∶ U).
Proof.
  introv Hok Hty.
  remember (trm_lit t u) as Tlet.
  induction Hty; intros; inversions HeqTlet; eauto.
  forwards T0 L Htyt Htyu: IHHty; eauto.
  exists T0 (L \u ctx_dom G). splits. auto. intros.
  assert (x\notin L) as Hx by eauto. specialize (Htyu x Hx). eapply ty_trm_sub. eauto. eapply weaken_push; eauto.
Qed.
Lemma ty_trm_read_inv G e x a T:
  G ⁏ e ⊢ trm_read (avar_f x) a ∶ T ->
  exists T1 T2,
  G ⁏ e ⊢ x ∶ (typ_rcd (dec_fld a T1 T2)).
Proof.
  introv Hty.
  remember (trm_read (avar_f x) a) as t.
  induction Hty; inversions Heqt; eauto.
Qed.

Lemma ty_trm_read_inv' G e x a T:
  G ⁏ e ⊢ trm_read (avar_f x) a ∶ T ->
  exists T1 T2 U,
  G ⁏ e ⊢ x ∶ (typ_rcd (dec_fld a T1 T2)) /\
  G ⁏ e ⊢ x ▷ T2 ⟶ U /\
  G ⁏ e ⊢ U <: T.
Proof.
  introv Hty.
  remember (trm_read (avar_f x) a) as t.
  induction Hty; inversions Heqt; eauto.
  - exists S T U; split; eauto.
  - forwards: IHHty. auto. destructecs H2.
    exists T1 T2 U0; split; eauto.
Qed.

Lemma ty_trm_call_inv G e x a y T:
  G ⁏ e ⊢ trm_call (avar_f x) a (avar_f y) ∶ T ->
  exists T1 T2 T3,
  G ⁏ e ⊢ x ∶ (typ_rcd (dec_met a T1 T2 T3)) /\
  G ⁏ e ⊢ y ∶ T2.
Proof.
  introv Hty.
  remember (trm_call (avar_f x) a (avar_f y)) as t.
  induction Hty; inversions Heqt; eauto.
Qed.

Lemma ty_trm_call_inv' G e x a y T:
  G ⁏ e ⊢ trm_call (avar_f x) a (avar_f y) ∶ T ->
  exists T1 T2 T3,
  G ⁏ e ⊢ x ∶ (typ_rcd (dec_met a T1 T2 T3)) /\
  G ⁏ e ⊢ y ∶ T2 /\ 
  G ⁏ e ⊢ T3  /[# 1 -> x]  /[-> y] <: T /\
  G ⁏ e ⊢ x ∶ T1 /[-> y].
Proof.
  introv Hty.
  remember (trm_call (avar_f x) a (avar_f y)) as t.
  induction Hty; inversions Heqt.
  - exists TC TZ TR. splits;   eauto.
  - forwards: IHHty; fweauto.
    destructecs H2.
    exists T1 T2 T3.
    splits;   eauto. 
Qed.

Lemma ty_trm_apply_inv G e x y T:
  G ⁏ e ⊢ trm_apply (avar_f x) (avar_f y) ∶ T ->
  exists T1 T2,
  G ⁏ e ⊢ x ∶ (typ_fun T1 T2) /\
  G ⁏ e ⊢ y ∶ T1 /\
  G ⁏ e ⊢ T2 /[-> y] <: T.
Proof.
  introv Hty.
  remember (trm_apply (avar_f x) (avar_f y)) as t.
  induction Hty; inversions Heqt; eauto. forwards: IHHty. auto. destructecs H2. exists T1 T2; splits; eauto.
Qed.
Lemma ty_trm_write_inv G e x a y T:
  G ⁏ e ⊢ trm_write (avar_f x) a (avar_f y) ∶ T ->
  exists S,
  G ⁏ e ⊢ y ∶ S /\
  G ⁏ e ⊢ x ∶ typ_rcd (dec_fld a S T).
Proof.
  introv Hty.
  remember (trm_write (avar_f x) a (avar_f y)) as t.
  induction Hty; inversions Heqt; eauto.
  forwards: IHHty; fweauto. destructecs H2. exists S. eauto.
Qed.
Lemma typing_swap G y1 y2 T1 T2 e t T:
  ok e ->
  ctx_ok G ->
  fresh (ctx_dom G \u ctx_fv G \u fv_renv e \u (fv T1) \u (fv T2)) 2 (y1::y2::nil)%list ->
  (G § y2 / v_let ~ T2) § y1 / v_let ~ T1 ⁏ e ⊢ t ∶ T ->
  (G § y1 / v_let ~ T1) § y2 / v_let ~ T2 ⁏ e ⊢ t ∶ T.
Proof.
  introv Hoke HokG Hfr Hty.
  sympls. rewrite ! notin_union in Hfr; destruct_all. rewrite notin_singleton in H2.

  pick_fresh y3.
  replace t with (t /[y2 -> y3] /[y3 -> y2]).
  replace T with (T /[y2 -> y3] /[y3 -> y2]).
  - eapply subst_end with (S:=T2)(vk:=v_let).
    eapply subst_ty_trm. eapply ewkn_ty_trm.
    ctx_split_between y2 y3.
    eapply weaken_ty.
    ctx_split_from y1.
    replace ((y1 / v_let ~ T1) § y3 / v_let ~ T2)  with (subst_ctx y2 y3 ((y1 / v_let ~ T1) § y3 / v_let ~ T2) ).
    eapply subst_ty with (S:=T2)(vk:=v_let).
    eapply subst_ty_trm. eapply ewkn_ty_trm.
    ctx_flatten.
    eapply weaken_push; eauto.
    ctx_flatten; eauto.
    eauto.
    eauto.
    eauto.
    ctx_flatten; eauto.
    ctx_flatten; eauto.
    sympls. rewrite ! subst_fresh; eauto.
    eauto.
    ctx_flatten; eauto.
    ctx_flatten; eauto.
    ctx_simpl_fv; eauto.
    eauto.
    eauto.
    eauto.
    rewrite ! subst_fresh; eauto.
    eauto.
  - rewrite subst_merge. rewrite subst_id; eauto. eauto. eauto.
  - rewrite subst_merge. rewrite subst_id; eauto. eauto. eauto.
Qed.

(* Do not have lemma like this because we do not have function to put list of vars onto context
Lemma rename_subst_vars_trm G e xs ys T1 T2 t T:
  ctx_ok G ->
  ok e ->
  length xs = length ys ->
  fresh (ctx_dom G) (length xs) xs ->
  fresh (ctx_dom G \u ctx_fv G \u fv_renv e \u fv T1 \u fv T2 \u fv T \u from_list xs) (length ys) ys ->
  (G § y2 / v_let ~ T2) § y1 / v_let ~ T1 ⁏ e ⊢ t ∶ T ->
  (G § x2 / v_let ~ T2) § x1 / v_let ~ T1 ⁏ e ⊢ ((t /[ y1 -> x1]) /[ y2 -> x2]) ∶ T.
Proof.
Qed.
*)


Lemma rename_subst_trm_2 G e x1 x2 y1 y2 T1 T2 t T:
  ctx_ok G ->
  ok e ->
  fresh (ctx_dom G) 2 (x1 :: x2 :: nil) ->
  fresh (ctx_dom G \u ctx_fv G \u fv_renv e \u fv T1 \u fv T2 \u fv T \u \{x1} \u \{x2}) 2 (y1 :: y2 :: nil) ->
  (G § y2 / v_let ~ T2) § y1 / v_let ~ T1 ⁏ e ⊢ t ∶ T ->
  (G § x2 / v_let ~ T2) § x1 / v_let ~ T1 ⁏ e ⊢ ((t /[ y1 -> x1]) /[ y2 -> x2]) ∶ T.
Proof.
  introv HokG Hoke Hfrx Hfry Hty.

  replace T with  ((T /[ y1 -> x1]) /[ y2 -> x2]).
  assert (ctx_ok ((G § x2 / v_let ~ T2) § x1 / v_let ~ T)) as HokG2 by eauto.
  assert (ctx_ok ((((G § x2 / v_let ~ T2) § x1 / v_let ~ T) § y2 / v_let ~ T2) § y1 / v_let ~ T)) as HokG4. {
    constructor. constructor. eauto. ctx_simpl_dom; notin_solve.  ctx_simpl_dom; notin_solve. 
  }

  eapply subst_end. eapply subst_ty_trm. eapply ewkn_ty_trm.
  eapply subst_end. eapply subst_ty_trm. eapply ewkn_ty_trm.
  - ctx_split_between x2 y2. eapply weaken_ty. ctx_flatten. eauto. eauto. ctx_flatten; eauto.
  - eauto.
  - eauto. ctx_simpl_fv. notin_solve.
  - eauto.
  - eauto.
  - rewrite subst_fresh; eauto.
  - eauto.
  - eauto.
  - ctx_simpl_fv. eauto.
  - eauto.
  - eauto.
  - rewrite subst_fresh; eauto.
  - eauto.
  - rewrite subst_fresh. rewrite subst_fresh; eauto. eapply notin_fv_subst_var; eauto.
Qed.


Lemma ty_trm_var_inv_vis G e x T:
  G ⁏ e ⊢ trm_var (avar_f x) ∶ T ->
  mode_visibility' typing_mode = mode_visibility_hide_method -> 
  ctx_vis G x.
Proof.
  introv Hty Hm.
  remember (trm_var (avar_f x)).
  reverts x.
  induction Hty; intros; inversions Heqt; auto.
Qed.
Lemma ty_trm_read_inv_vis G e x a T:
  G ⁏ e ⊢ trm_read (avar_f x) a ∶ T ->
  mode_visibility' typing_mode = mode_visibility_hide_method -> 
  ctx_vis G x.
Proof.
  introv Hty Hm.
  remember (trm_read (avar_f x) a).
  reverts x.
  induction Hty; intros; inversions Heqt; auto.
Qed.
Lemma ty_trm_call_inv_vis1 G e x1 a x2 T:
  G ⁏ e ⊢ trm_call (avar_f x1) a (avar_f x2) ∶ T ->
  mode_visibility' typing_mode = mode_visibility_hide_method -> 
  ctx_vis G x1.
Proof.
  introv Hty Hm.
  remember (trm_call (avar_f x1) a (avar_f x2)).
  reverts x1 x2.
  induction Hty; intros; inversions Heqt; auto.
  eapply IHHty. eauto.
Qed.
Lemma ty_trm_call_inv_vis2 G e x1 a x2 T:
  G ⁏ e ⊢ trm_call (avar_f x1) a (avar_f x2) ∶ T ->
  mode_visibility' typing_mode = mode_visibility_hide_method -> 
  ctx_vis G x2.
Proof.
  introv Hty Hm.
  remember (trm_call (avar_f x1) a (avar_f x2)).
  reverts x1 x2.
  induction Hty; intros; inversions Heqt; auto.
  eapply IHHty. eauto.
Qed.
Lemma ty_trm_apply_inv_vis1 G e x1 x2 T:
  G ⁏ e ⊢ trm_apply (avar_f x1)  (avar_f x2) ∶ T ->
  mode_visibility' typing_mode = mode_visibility_hide_method -> 
  ctx_vis G x1.
Proof.
  introv Hty Hm.
  remember (trm_apply (avar_f x1)  (avar_f x2)).
  reverts x1 x2.
  induction Hty; intros; inversions Heqt; auto.
  eapply IHHty. eauto.
Qed.
Lemma ty_trm_apply_inv_vis2 G e x1  x2 T:
  G ⁏ e ⊢ trm_apply (avar_f x1) (avar_f x2) ∶ T ->
  mode_visibility' typing_mode = mode_visibility_hide_method -> 
  ctx_vis G x2.
Proof.
  introv Hty Hm.
  remember (trm_apply (avar_f x1)  (avar_f x2)).
  reverts x1 x2.
  induction Hty; intros; inversions Heqt; auto.
  eapply IHHty. eauto.
Qed.

Lemma ty_def_label_eq G Go e d D:
  G §§ Go ⁏ e ⊢ d ∶d D ->
  label_of d = label_of D.
Proof.
  introv Hty.
  inversions Hty; simpls; auto.
Qed.


End WithTypingMode.