(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Program.Equality.

Require Import DotTactics LibExtra.
Require Import AbstractSyntax TypingMode Equivalence GeneralTyping
        Weakening EWeakening Visibility.
Require Import  Indep SubstCommon.

(** * Substitution of unbound variables *)

(* Class for substituting unbound named variables *)

Class UnboundSubstitution {A B: Set} {AS: SubstVar A} {AF: FreeVar A} {BS: SubstVar B} {BF: FreeVar B}
    (property: TypingProperty A B) :=
  unbound_substitution_obj (ys: list var) (S: typ) G1 (xs: list var) G2 Go e (a: A) (b: B):
      property (G1 §& G2) Go e a b ->
      ctx_ok (G1 §& G2 §§& Go) ->
      length ys = length xs ->
      fresh (ctx_fv G1 \u fv_renv e \u (ctx_dom G1) \u (ctx_dom G2) \u (objctx_dom Go)) (length xs) xs ->
      property (G1 §& (subst_ctx_pvars xs ys G2)) (subst_objctx_pvars xs ys Go) e (subst_pvars xs ys a) (subst_pvars xs ys b) .

(* Special cases of substituting unbound named variables *)
Section UnboundSubstitutionLemmata.
  Context {A B: Set} {AS: SubstVar A} {AF: FreeVar A} {BS: SubstVar B} {BF: FreeVar B}
    {property: TypingProperty A B}{AW: UnboundSubstitution property}.


Lemma usubstitution (ys: list var) G1 xs G2 e (a: A) (b: B):
  property (G1 §& G2) objctx_empty e a b ->
  ctx_ok (G1 §& G2) ->
  length ys = length xs ->
  fresh (ctx_fv G1 \u fv_renv e \u (ctx_dom G1) \u (ctx_dom G2)) (length xs) xs ->
  property (G1 §& (subst_ctx_pvars xs ys G2)) objctx_empty  e (subst_pvars xs ys a) (subst_pvars xs ys b).
Proof.
  intros.
  replace objctx_empty with (subst_objctx_pvars xs ys objctx_empty).
  eapply unbound_substitution_obj; eauto.
  auto.
Qed.

Lemma usubst_end G xs ys e (t : A) (T: B):
    property G objctx_empty e t T ->
    ctx_ok G ->
    length ys = length xs ->
    fresh (ctx_fv G \u fv_renv e \u (ctx_dom G)) (length xs) xs ->
    property G objctx_empty e (subst_pvars xs ys t) (subst_pvars xs ys  T).
Proof.
  intros.
  apply usubstitution with (G1:=G) (G2:=ctx_empty) (xs:=xs)  (ys:=ys) in H;
  try rewrite map_empty in *; try rewrite concat_empty_r in *; auto.
Qed.

Context {AO: Openable A} {AAS: AbstractSyntax AO AF AS} {ASI: SubstId AAS} {ASC: SubstCommut AAS}  {ASF: SubstFresh AAS} {AFSC: FvSubstCases AAS}.
Context {BO: Openable B} {BAS: AbstractSyntax BO BF BS} {BSI: SubstId BAS} {BSC: SubstCommut BAS}  {BSF: SubstFresh BAS} {BFSC: FvSubstCases BAS}.

Lemma usubst_end1:
  forall G x y e (t : A) (T: B),
    property (G ) objctx_empty e t T ->
    ctx_ok (G) ->
    x \notin ctx_fv G ->
    renv_notin x e ->
    x \notin ctx_dom G ->
    property G objctx_empty e (subst_var x y t) (subst_var x y T).
Proof.
  intros.
  rewrite <- subst_pvars_one.
  rewrite <- subst_pvars_one.
  eapply usubst_end; eauto. simpls. unfold renv_notin in H2. split; eauto; notin_solve.

Qed.



End UnboundSubstitutionLemmata.

(* Special cases of substituting unbound named variables in typing *)
Section UnboundSubstitutionTyLemmata.
  Context {A: Set}{AS: SubstVar A}{AF: FreeVar A}{AT: Typing A}
    (AW: UnboundSubstitution AT).

Lemma usubst_ty(ys: list var) G1 xs  G2 e (a: A) (b: typ):
  typing (G1 §& G2) e a b ->
  ctx_ok (G1 §& G2) ->
  length ys = length xs ->
  fresh (ctx_fv G1 \u fv_renv e \u (ctx_dom G1) \u (ctx_dom G2)) (length xs) xs ->
  typing (G1 §& (subst_ctx_pvars xs ys G2)) e (subst_pvars xs ys a) (subst_pvars xs ys b).
Proof.
  intros.
  eapply usubstitution; eauto.
Qed.
Hint Resolve usubst_ty: core.

End UnboundSubstitutionTyLemmata.

(* Re-stating UnboundSubstitution for mutual induction *)
Definition UnboundSubstitutionInd (A B: Set) (su: SubstVar A)  (fa: FreeVar A) (sub: SubstVar B) (fb: FreeVar B) (property: TypingProperty A B) :=
  (forall G Go e (v: A) T,
    property G Go e v T ->
    forall G1 G2 xs ys,
      G = G1 §& G2 ->
      ctx_ok (G1 §& G2 §§& Go) ->
      length ys = length xs ->
      fresh (ctx_fv G1 \u fv_renv e \u (ctx_dom G1) \u (ctx_dom G2) \u (objctx_dom Go)) (length xs) xs ->
      property (G1 §& (subst_ctx_pvars xs ys G2)) (subst_objctx_pvars xs ys Go) e (subst_pvars xs ys v) (subst_pvars xs ys T)).
Definition UnboundSubstitutionEInd (A: Set) (t: TypingEquiv A) (su: SubstVar A) (FA: FreeVar A)  :=
  UnboundSubstitutionInd (su) FA (su) FA (fun G Go e (a b: A) => typing_equiv e a b).
Definition UnboundSubstitutionTInd (A: Set) (t: Typing A) (su: SubstVar A) (FA: FreeVar A)  :=
  UnboundSubstitutionInd (su) FA (TypSubstVar) (TypFreeVar) (obj_typing).
Definition UnboundSubstitutionIncapInd (_: ReferenceIncapacitate) :=
  UnboundSubstitutionInd TypSubstVar TypFreeVar TypSubstVar TypFreeVar obj_ref_incap.
Definition UnboundSubstitutionCapbndInd (C : typ_refcap_label) (_ : CapabilityBound C) :=
  UnboundSubstitutionInd TypSubstVar TypFreeVar TypSubstVar TypFreeVar obj_cap_bnd.
Definition UnboundSubstitutionDInd (A: Set) (t: DecTyping A) (su: SubstVar A) (FA: FreeVar A)  :=
  UnboundSubstitutionInd (su) FA (subst_dec) DecFreeVar (obj_dec_typing).

Local Hint Extern 1 (_ §# _) => ctx_simpl_dom; notin_solve : core.

(*   Instances   *)

(* Instances for equivalence *)

(* Substituting one variable in equivalence *)
Lemma usubst_var_tyq_var x y e v (T: var):
  x \notin (fv_renv e) ->
  e ⊢ v ≈ T ->
  e ⊢ (v /[ x -> y]) ≈ (T /[ x -> y]).
Proof.
  introv Hn Heq.
  unfold typing_equiv in Heq. (* unfolding seems to be needed for dependent induction *)
  dependent induction Heq.
  - auto.
  -
    destruct (subst_cases2 x y v1) as [[Hv1x Hv1y] | [Hv1x Hv1v1]]; subst.
    + destruct (subst_cases2 x y v2) as [[Hv2x Hv2y] | [Hv2x Hv2v2]]; subst.
      * auto.
      * rewrite Hv1y, Hv2v2.
        false. eapply renv_binds_notin_false; eauto.
    +
      destruct (subst_cases2 x y v2) as [[Hv2x Hv2y] | [Hv2x Hv2v2]]; subst.
      * false. eapply renv_binds_notin_false; eauto.
      * rewrite Hv2v2, Hv1v1. auto.

  - symmetry. eapply IHHeq; eauto.
  - transitivity (v2 /[ x -> y]); eauto.
Qed.

(* Substituting parallel lists in equivalence *)
Lemma usubst_pvars_tyq_var xs ys e v (T: var):
  length ys = length xs ->
  fresh (fv_renv e) (length xs) xs ->
  e ⊢ v ≈ T ->
  e ⊢ (v //[ xs --> ys]) ≈ (T //[ xs --> ys]).
Proof.
  introv Hl Hfr Heq.
  reverts ys v T.
  induction xs; intros.
  - rewrite subst_pvars_nil_l. rewrite subst_pvars_nil_l. eauto.
  - destruct ys. simpls. discriminate.
    pick_fresh_with z (from_list xs \u from_list ys).
    inversions Hl.
    destructc Hfr as Hn Hfr.
    assert (a \notin from_list xs) as Hnaxs. {
      eapply from_list_fresh.
      eapply fresh_union_r2.
      eauto.
    }
    erewrite (subst_pvars_push_lr) with (v1:=z); unfold fv, VarFreeVar; eauto.
    erewrite (subst_pvars_push_lr) with (v1:=z); unfold fv, VarFreeVar; eauto.
    eapply usubst_var_tyq_var. notin_solve.
    eapply IHxs; eauto using usubst_var_tyq_var.
Qed.

Lemma usubst_tyq_var:
  UnboundSubstitutionEInd tyq_var VarSubstVar VarFreeVar.
Proof.
  unfold UnboundSubstitutionEInd, UnboundSubstitutionInd.
  introv Heq HG Hok Hl Hfr; subst.
  eapply usubst_pvars_tyq_var; eauto.
Qed.

Lemma usubst_pvars_tyq_typ y e x v (T: typ):
  length y = length x ->
  fresh (fv_renv e) (length x) x ->
  e ⊢ v ≈ T ->
  e ⊢ (v //[ x --> y]) ≈ (T //[ x --> y]).
Proof.
  introv Hl Hfr Heq.
  induction Heq.
  - apply tyq_typ_nulary.
    rewrite typ_subst_pvars_nulary.
    inversion H; auto. auto.
  - repeat rewrite typ_subst_pvars_andor; eauto.
  - autorewrite with simpl_subst_pvars; eauto.
    eapply (tyq_typ_fun); eauto. intros. pick_fresh_with z (from_list x \u from_list y). forwards: (H1 z); eauto.
    rewrite (subst_pvars_open_fresh_commut ) in H2 by eauto.
    rewrite (subst_pvars_open_fresh_commut ) in H2 by eauto.
    eapply tyq_typ_reopen; eauto. notin_solve.
    eapply notin_fv_subst_pvars; eauto.
    eapply notin_fv_subst_pvars; eauto.
  - (* fld *)
    autorewrite with simpl_subst_pvars.
    sympls; eauto.
  - (* typ*)
    autorewrite with simpl_subst_pvars.

    eapply (tyq_typ_typ); eauto.

    intros.
    pick_fresh_with z (from_list x \u from_list y). 

forwards: (H0 z); eauto. forwards: (H2 z); eauto.
    repeat rewrite (subst_pvars_open_fresh_commut ) in H3 by eauto.
    repeat rewrite (subst_pvars_open_fresh_commut ) in H4 by eauto.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_pvars; eauto.

    intros. pick_fresh_with z (from_list x \u from_list y). forwards: (H0 z); eauto. forwards: (H2 z); eauto.
    repeat rewrite (subst_pvars_open_fresh_commut ) in H3 by eauto.
    repeat rewrite (subst_pvars_open_fresh_commut ) in H4 by eauto.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_pvars; eauto.
  - rewrite typ_subst_pvars_sel by auto.
    rewrite typ_subst_pvars_sel by auto.
    repeat rewrite avar_subst_pvars_f.
    apply tyq_typ_sel_normal.
    eapply usubst_pvars_tyq_var; eauto.
  - rewrite typ_subst_pvars_sel by auto.
     apply tyq_typ_sel.
  - rewrite typ_subst_pvars_bnd .
    rewrite typ_subst_pvars_bnd .
    apply tyq_typ_bnd; eauto; intros. pick_fresh_with z (from_list x \u from_list y). forwards: (H0 z); eauto.
    repeat rewrite (subst_pvars_open_fresh_commut ) in H1 by eauto.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_pvars; eauto.
  - autorewrite with simpl_subst_pvars.
    apply tyq_typ_met; auto.
    intros. pick_fresh_with z (from_list x \u from_list y). forwards: (H0 z); eauto.
    repeat rewrite (subst_pvars_open_fresh_commut ) in H4 by eauto.
    eapply tyq_typ_reopen; eauto. notin_solve; eapply notin_fv_subst_pvars; eauto.
    intros. pick_fresh_with z1 (from_list x \u from_list y). pick_fresh_with z2 (from_list x \u from_list y). forwards: (H3 z1 z2); eauto.
    repeat rewrite (subst_pvars_open_fresh_commut ) in H4 by eauto.
    repeat rewrite <- (open_vars_single 1) in *.
    repeat rewrite open_vars_push in *.
    eapply tyq_typ_reopen_vars; eauto.
    fresh_solve; auto; eapply notin_fv_subst_pvars; eauto.
Qed.

Lemma usubst_tyq_typ:
  UnboundSubstitutionEInd tyq_typ TypSubstVar TypFreeVar.
Proof.
  hnf; intros; eapply usubst_pvars_tyq_typ; eauto.
Qed.

(* Lemmas about substitution applied to variables occurring in the typing context *)

(* A variable bound in context is not affected *)
Lemma subst_pvars_fresh_ctx_dom x G xs ys:
  x \in (ctx_dom G) ->
  fresh (ctx_dom G) (length xs) xs ->
  (x //[ xs --> ys]) = x.
Proof.
  introv Hbi Hfr.
  eapply subst_pvars_fresh.
  unfold fv, VarFreeVar.
  eapply fresh_subset; eauto.
  apply <- subset_singleton.
  auto.
Qed.

(* A variable bound to some type in a split context is not affected *)
Lemma subst_pvars_fresh_ctx_binds_concat x T G1 G2 xs ys L:
  ctx_binds x T (G1 §& G2) ->
  fresh (L \u (ctx_dom G1) \u (ctx_dom G2)) (length xs) xs ->
  (x //[ xs --> ys]) = x.
Proof.
  introv Hbi Hfr.
  eapply ctx_binds_dom in Hbi.
  eapply subst_pvars_fresh_ctx_dom; eauto.
  rewrite ctx_dom_concat.
  fresh_solve.
Qed.

Section WithTypingMode.
Context {typing_mode: Mode}.

(* A variable typed in a split context is not affected *)
Lemma subst_pvars_fresh_ty_concat (x: var) (T: typ) G1 G2 e xs ys L:
  G1 §& G2 ⁏ e ⊢ x ∶ T ->
  fresh (L \u (ctx_dom G1) \u (ctx_dom G2)) (length xs) xs ->
  (x //[ xs --> ys]) = x.
Proof.
  introv Hbi Hfr.
  eapply typing_implies_dom in Hbi.
  eapply subst_pvars_fresh_ctx_dom; eauto.
  rewrite ctx_dom_concat.
  fresh_solve.
Qed.

(* A variable typed in a split context and object context is not affected *)
Lemma subst_pvars_fresh_ty_concat_objctx (x: var) (T: typ) G1 G2 Go e xs ys L:
  G1 §& G2 §§& Go ⁏ e ⊢ x ∶ T ->
  fresh (L \u (ctx_dom G1) \u (ctx_dom (G2 §§& Go))) (length xs) xs ->
  (x //[ xs --> ys]) = x.
Proof.
  introv Hbi Hfr.
  rewrite ctx_obj_assoc in Hbi.
  eapply subst_pvars_fresh_ty_concat; eauto.
Qed.

(* A variable visible in a split context is not affected *)
Lemma subst_pvars_fresh_ctx_vis_concat (x: var) (T: typ) G1 G2 xs ys L:
  ctx_vis (G1 §& G2) x ->
  fresh (L \u (ctx_dom G1) \u (ctx_dom G2)) (length xs) xs ->
  (x //[ xs --> ys]) = x.
Proof.
  introv Hvis Hfr.
  eapply ctx_vis_in_dom in Hvis.
  eapply subst_pvars_fresh_ctx_dom; eauto.
  rewrite ctx_dom_concat.
  fresh_solve.
Qed.

(* A type of a variable in the context is not affected *)
Lemma subst_pvars_fresh_ctx_binds_fv x T G xs ys:
  ctx_binds x T (G) ->
  fresh (ctx_fv G) (length xs) xs ->
  (T //[ xs --> ys]) = T.
Proof.
  introv Hbi Hfr.
  eapply subst_pvars_fresh.
  unfold fv, VarFreeVar.
  eapply fresh_subset; eauto.
  eapply ctx_binds_fv_subset in Hbi.
  eauto.
Qed.


Lemma subst_pvars_push_fresh_typ (T: typ) xs ys z x:
  z \notin from_list xs ->
  z \notin from_list ys ->
  z \notin fv T ->
  subst_pvars xs ys T = subst_pvars (z::xs) (x::ys) T.
Proof.
  introv Hnxs Hnys HnG2.
  rewrite subst_pvars_push by auto.
  rewrite subst_fresh;  eauto.
  eapply notin_fv_subst_pvars; eauto.
Qed.

Lemma subst_pvars_push_fresh_ctx G2 xs ys z x:
  z \notin from_list xs ->
  z \notin from_list ys ->
  z \notin fv G2 ->
  subst_ctx_pvars xs ys G2 = subst_ctx_pvars (z::xs) (x::ys) G2.
Proof.
  introv Hnxs Hnys HnG2.
  induction G2; sympls; eauto; unfold CtxFreeVar in *.
  - rewrite <- IHG2.
    rewrite <- subst_pvars_push_fresh_typ; eauto.
    ctx_simpl_fv. notin_solve.
    ctx_simpl_fv. rewrite notin_union in HnG2. destruct HnG2. apply H0.
  - rewrite <- IHG2; eauto.
  - rewrite <- IHG2; eauto.
Qed.

Lemma subst_ctx_pvars_push G1 G2 xs ys x S2 vk:
  G1 §& subst_ctx_pvars xs ys G2 § x / vk ~ (S2 //[ xs --> ys])  = G1 §& subst_ctx_pvars xs ys (G2 § x / vk ~ S2 ) .
Proof.
  sympls. auto.
Qed.
Lemma in_subset A (x: A) (X Y: fset A):
  x \in X ->
  X \c Y ->
  x \in Y.
Proof.
  auto.
Qed.


(* Instances for var typing and subtyping *)
Lemma usubst_ts_ind:
  UnboundSubstitutionTInd (ty_var _) VarSubstVar VarFreeVar /\
  UnboundSubstitutionInd TypSubstVar TypFreeVar TypSubstVar TypFreeVar subtyping /\
  UnboundSubstitutionIncapInd (ty_incap _).
Proof.
  introv.
  apply ts_mutind.
  all: intros.
  all: autorewrite with simpl_subst_pvars; eauto 4.
  - Case "ty_var".
    subst.
    sympls.
    erewrite subst_pvars_fresh_ctx_binds_concat; eauto.

    destruct (ctx_binds_concat_inv _ _ c).
    + (* G1 - not substituted *)
      erewrite subst_pvars_fresh_ctx_binds_fv; eauto.
      eapply ctx_binds_concat_map_l in H; eauto.
    + (* G2 *)
      eapply ctx_binds_concat_map_r in H; eauto.
  - Case "ty_rec_intro".

    subst; sympls.
    assert (x \notin from_list xs) as Hnxxs. {
      eapply fresh_in_notin_list; eauto. eapply typing_implies_dom in o. ctx_simpl_dom.
      eapply (in_subset o).
      union_sub_solver.
    }

    erewrite subst_pvars_fresh_ty_concat; eauto.
    apply ty_rec_intro; auto.
    forwards: H; fweauto.
    erewrite subst_pvars_fresh_ty_concat in H1; eauto.
    rewrite (subst_pvars_open_fresh_commut ) in H1 by eauto.
    auto.

    intro.
    repeat rewrite <- (subst_pvars_open_fresh_commut ) by eauto. eauto.
  -
    subst; sympls.

    assert (fresh \{ x} (length xs) xs) as Hnxxs. {
      eapply fresh_from_list.
      eapply fresh_in_notin_list; eauto. eapply typing_implies_dom in o. ctx_simpl_dom.
      apply (in_subset o).
      union_sub_solver. eauto.
    }


    erewrite subst_pvars_fresh_ty_concat; eauto.
    rewrite (subst_pvars_open_fresh_commut ).
    apply ty_rec_elim; eauto.
    forwards: H; fweauto.
    rewrite typ_subst_pvars_bnd in H0.
    erewrite subst_pvars_fresh_ty_concat in H0; eauto.
    auto.
  - (* typ *)

    pick_fresh_with z (from_list xs \u from_list ys).

    assert (z \notin (from_list xs)) as Hnzxs by eauto.
    assert (z \notin (from_list ys)) as Hnzys by eauto.
    assert (z \notin (fv G2)) as HnzG2. { unfold fv, CtxFreeVar.
      notin_solve.
    }
    assert (fresh \{ z} (length xs) xs) as Hnzyxs by eauto.

    eapply subtyp_typ; intros.

    rewrite subst_intro with (x0:=z).
    rewrite subst_intro with (x0:=z) (X:=(S1 //[ xs --> ys])).
    repeat rewrite <- (subst_pvars_open_fresh_commut ) by eauto.
    rewrite <- subst_pvars_push by eauto.
    rewrite <- subst_pvars_push by eauto.
    rewrite (subst_pvars_push_fresh_ctx) with (z:=z) (x:=x) by eauto.
    eapply H; simpls; eauto.

    eapply notin_fv_subst_pvars; eauto. eapply notin_fv_subst_pvars; eauto.

    rewrite subst_intro with (x0:=z).
    rewrite subst_intro with (x0:=z) (X:=(T2 //[ xs --> ys])).
    repeat rewrite <- (subst_pvars_open_fresh_commut ) by eauto.
    rewrite <- subst_pvars_push by eauto.
    rewrite <- subst_pvars_push by eauto.
    rewrite (subst_pvars_push_fresh_ctx) with (z:=z) (x:=x) by eauto.
    eapply H0; simpls; eauto.

    eapply notin_fv_subst_pvars; eauto. eapply notin_fv_subst_pvars; eauto.

  - (* sel *)
    subst; sympls.
    rewrite subst_pvars_open_commut.
    eapply subtyp_sel2.

    forwards: H; fweauto.
    autorewrite with simpl_subst_pvars in H0. eauto.
  - (* sel *)
    subst; sympls.
    rewrite subst_pvars_open_commut.
    eapply subtyp_sel1.

    forwards: H; fweauto.
    autorewrite with simpl_subst_pvars in H0. eauto.

  - (* Function subtyping *)

    subst; sympls. fresh_constructor.
    repeat rewrite <- (subst_pvars_open_fresh_commut ) by eauto.
    rewrite  subst_ctx_pvars_push.
    eapply H0; eauto. ctx_flatten. constructor; eauto.

  - (* Method subtyping *)
    subst; sympls.
    fresh_constructor.

    repeat rewrite <- (subst_pvars_open_fresh_commut ) by eauto.
    rewrite  subst_ctx_pvars_push.
    eapply H0; eauto. ctx_flatten. constructor; eauto.

    intros Hsef Hindep. specialize (t Hsef).
    eapply subst_pvars_indep_b. auto. eapply t.
    eapply subst_pvars_indep_b_inv. symmetry. eauto. eauto.

    intros.

    repeat rewrite <- (subst_pvars_open_fresh_commut ) by eauto.
    rewrite  subst_ctx_pvars_push.
    rewrite  subst_ctx_pvars_push.
    eapply H1; eauto. ctx_flatten. constructor; eauto.

  - eapply usubst_tyq_typ in t. eauto.

  - subst; sympls.
    eapply ty_incap_sel.
    forwards: H; fweauto.  autorewrite with simpl_subst_pvars in H1. eauto.
    forwards: H0; fweauto.      rewrite subst_pvars_open_commut in H1. eauto.
  - (* Icp *)
    subst; sympls.

    eapply ty_incap_nocap. auto.
    forwards Hsu: H; fweauto. autorewrite with simpl_subst_pvars in Hsu. eauto.
Qed.

Global Instance usubst_ty_var: UnboundSubstitution (ty_var _).
Proof.
  instance_by_ind0 usubst_ts_ind.
Qed.
Hint Resolve usubst_ty_var: core.

Global Instance usubst_sub: UnboundSubstitution (@subtyping (subtyp _)).
Proof.
  instance_by_ind0 usubst_ts_ind.
Qed.
Hint Resolve usubst_sub: core.

Global Instance usubst_ty_incap: UnboundSubstitution  (ty_incap _).
Proof.
  instance_by_ind0 usubst_ts_ind.
Qed.
Hint Resolve usubst_ty_incap: core.

Lemma usubst_ty_var_u G1 xs ys G2 e (a: var) (b: typ):
  typing (G1 §& G2) (e) a b ->
  ctx_ok (G1 §& G2) ->
  length ys = length xs ->
  fresh (ctx_fv G1 \u fv_renv e \u (ctx_dom G1) \u (ctx_dom (G2))) (length xs) xs ->
  typing (G1 §& (subst_ctx_pvars xs ys G2)) e (a //[ xs --> ys ]) (subst_pvars xs ys b) .
Proof.
  intros.
  eapply usubst_ty; eauto.
Qed.
Hint Resolve usubst_ty_var_u: core.

(* Instances for viewpoint adaptation *)
Lemma usubst_ty_capbnd_ind:
  forall C, UnboundSubstitutionCapbndInd (ty_capbnd _ C).
Proof.
  introv. hnf. introv.
  induction 1; intros; subst;
    sympl in *;
    autorewrite with simpl_subst_pvars;
    try rewrite subst_open_commut;
      eauto 4.

  -  subst; sympls.
    eapply ty_capbnd_sel.
    assert (G1 §& subst_ctx_pvars xs ys G2 ⁏ e ⊢ (x //[ xs --> ys]) ∶ ((typ_rcd (dec_typ B S T)) //[ xs --> ys])).
    eapply usubst_ty; eauto.
      autorewrite with simpl_subst_pvars in H1.
    apply H1.
    forwards: IHty_capbnd; eauto.
    rewrite subst_pvars_open_commut in H1. eauto.
Qed.

Global Instance usubst_ty_capbnd:
  forall C, UnboundSubstitution (ty_capbnd _ C).
Proof. intros. hnf. intros. eapply usubst_ty_capbnd_ind; eauto. Qed.
Hint Resolve usubst_ty_capbnd: core.

Lemma usubst_ty_viewpoint_adapt xs ys G1 G2 Go e (T1 T2: typ) v:
  viewpoint_adapt _ v (G1 §& G2) Go e T1 T2 ->
  ctx_ok (G1 §& G2 §§& Go) ->
  length ys = length xs ->
  fresh (ctx_fv G1 \u fv_renv e \u (ctx_dom G1) \u (ctx_dom G2) \u (objctx_dom Go)) (length xs) xs ->
  viewpoint_adapt _
    (v //[ xs --> ys])
    (G1 §& subst_ctx_pvars xs ys G2) (subst_objctx_pvars xs ys Go)
    e (T1 //[ xs --> ys]) (T2 //[ xs --> ys]).
Proof.
  intros. hnf. hnf in H.
  pose proof usubst_ty_incap as Hicp. hnf in Hicp.
  pose proof usubst_ty_capbnd as Hcbp.
  unfold UnboundSubstitution in Hcbp.
  destruct (mode_viewpoint_adaptation' typing_mode); subst; eauto.
  inversions H.
  autorewrite with simpl_subst_pvars.
  constructor.
  - eapply Hicp; eauto.
  - eapply Hcbp; eauto.
Qed.

(* Instances for term typing *)

Lemma usubstitution_visible G1 G2 v x y:
  ctx_vis (G1 §& G2) x ->
  ctx_vis (G1 §& subst_ctx v y G2) x.
Proof.
  introv Hvis. remember (G1 §& G2) as G.
  generalize dependent G1. generalize dependent G2.
  induction Hvis; intros; subst.
  - apply ctx_push_concat_inv in HeqG.
    destruct HeqG as [[? ?] | [? ?]]; subst; simpl; eauto.
  - apply ctx_push_concat_inv in HeqG as ?.
    destruct H0 as [[? ?] | [? ?]]; subst; simpl in *; eauto.
    + inversions HeqG; simpl; eauto.
Qed.

Lemma usubstitution_visible_pvars G1 G2 v x y:
  ctx_vis ((G1) §& G2) x ->
  ctx_vis (G1 §& subst_ctx_pvars v y G2) x.
Proof.
  introv Hvis. remember ((G1 ) §& G2) as G.
  generalize dependent G1. generalize dependent G2.
  induction Hvis; intros; subst.
  - apply ctx_push_concat_inv in HeqG.
    destruct HeqG as [[? ?] | [? ?]]; subst; simpl; eauto.
  - apply ctx_push_concat_inv in HeqG as ?.
    destruct H0 as [[? ?] | [? ?]]; subst; simpl in *; eauto.
    + inversions HeqG; simpl; eauto.
Qed.

Lemma usubstitution_visible_pvars' L G1 G2 xs ys v:
  fresh (L \u (ctx_dom G1) \u (ctx_dom G2)) (length xs) xs ->
  ctx_vis (G1 §& G2) v ->
  ctx_vis (G1 §& subst_ctx_pvars xs ys G2) (subst_pvars xs ys v).
Proof.
  introv Hfr Hvis.
  erewrite subst_pvars_fresh_ctx_vis_concat; eauto.
  eapply usubstitution_visible_pvars; eauto.
Qed.



Lemma usubst_vis x0 G1 G2 e z S y:
  x0 §# G1 ->
  x0 §# G2 ->
  G1 §& G2 ⁏ e ⊢ z ∶ S ->
  (mode_has_visibility typing_mode -> ctx_vis (G1 §& G2) z) ->
  (mode_has_visibility typing_mode -> ctx_vis (G1 §& subst_ctx x0 y G2) (z \[ x0 -> y])).
Proof.
  intros.
  eapply usubstitution_visible.
  cases_if. assert (x0 §# G1 §& G2). eapply ctx_unbound_concat; eauto.
  eapply typing_implies_bound in H1. destruct H1.
  false. eapply ctx_binds_unbound_false. apply H1. auto.
  auto.
Qed.
Lemma usubst_vis_pvars xs ys G1 G2 e z S:
  length ys = length xs ->
  fresh (ctx_dom G1 \u ctx_dom G2) (length xs) xs ->
  G1 §& G2 ⁏ e ⊢ z ∶ S ->
  (mode_has_visibility typing_mode -> ctx_vis (G1 §& G2) z) ->
  (mode_has_visibility typing_mode -> ctx_vis (G1 §& subst_ctx_pvars xs ys G2) (z //[ xs --> ys])).
Proof.
  intros.
  eapply usubstitution_visible_pvars.
  erewrite subst_pvars_fresh_ty_concat; eauto.
Qed.


Lemma usubst_ty_var_intact  G1 xs ys G2 e (a: var) (b: typ):
  typing (G1 §& G2) (e) a b ->
  ctx_ok (G1 §& G2) ->
  length ys = length xs ->
  fresh (ctx_fv G1 \u fv_renv e \u (ctx_dom G1) \u (ctx_dom (G2) ) \u \{ a } \u (fv b)) (length xs) xs ->
  typing (G1 §& (subst_ctx_pvars xs ys G2)) e a b.
Proof.
  intros.
  replace a with (subst_pvars xs ys a).
  replace b with (subst_pvars xs ys b).
  eapply usubst_ty; eauto.
  rewrite subst_pvars_fresh; auto.
  rewrite subst_pvars_fresh; unfold fv, VarFreeVar; auto.
Qed.

Lemma fresh_ctx_obj_met G Go xz TZ xc TC xs:
  fresh (ctx_dom G \u objctx_dom Go \u \{xz} \u \{xc}) (length xs) xs ->
  fresh (ctx_dom (ctx_obj_met typing_mode G Go xz TZ xc TC)) (length xs) xs.
Proof.
  introv Hfr.
  unfold ctx_obj_met; destruct typing_mode; destruct Go; ctx_simpl_dom; simpls; fresh_solve.
Qed.

Lemma usubst_rules_ind:
  UnboundSubstitutionTInd (ty_trm _) TrmSubstVar TrmFreeVar /\
  UnboundSubstitutionTInd (ty_lit _) LitSubstVar LitFreeVar /\
  UnboundSubstitutionDInd (ty_def _) DefSubstVar DefFreeVar /\
  UnboundSubstitutionTInd (ty_defs _) DefsSubstVar DefsFreeVar.
Proof.

  (* Induction cases *)
  apply rules_mutind; intros.
  all: subst.
  (* Pushes substitution down *)
  all: sympls;
  autorewrite with simpl_subst_pvars;
  try rewrite subst_open_commut.

  - rewrite trm_subst_pvars_var.
    rewrite avar_subst_pvars_f.
    eapply ty_trm_var; eauto using usubstitution_visible_pvars'.
  -
    rewrite trm_subst_pvars_apply.
    rewrite avar_subst_pvars_f.
    rewrite avar_subst_pvars_f.
    rewrite subst_pvars_open_commut.
    pose proof o as Ho.
    pose proof o0 as Ho0.
    eapply usubst_ty in Ho0; eauto.
    eapply usubst_ty in Ho; eauto.
    rewrite typ_subst_pvars_fun in Ho.

    eapply ty_trm_apply; eauto using usubstitution_visible_pvars'.
  -
    rewrite trm_subst_pvars_read.
    rewrite avar_subst_pvars_f.
    eapply ty_trm_read.
    intro.
    eapply usubstitution_visible_pvars.
    erewrite subst_pvars_fresh_ty_concat; eauto.
    pose proof o as Ho.
    eapply usubst_ty in Ho; eauto.
    autorewrite with simpl_subst_pvars in Ho.
    eauto.
    eapply usubst_ty_viewpoint_adapt with (xs:=xs) (ys:=ys) in v. apply v. eauto.
    sympls; eauto.
    sympls; eauto.
  -
    rewrite trm_subst_pvars_write.
    rewrite avar_subst_pvars_f.
    rewrite avar_subst_pvars_f.

    pose proof o0 as Ho0.
    pose proof o1 as Ho1.
    eapply usubst_ty in Ho0; eauto.
    eapply usubst_ty in Ho1; eauto.
    autorewrite with simpl_subst_pvars in Ho0.

    eapply ty_trm_write; eauto using usubstitution_visible_pvars'.
    intro.

    pose proof (o H) as Ho.
    eapply usubst_ty in Ho; eauto.
    autorewrite with simpl_subst_pvars in Ho.

     eauto.
  -
    rewrite trm_subst_pvars_call.
    rewrite avar_subst_pvars_f.
    rewrite avar_subst_pvars_f.

    pose proof (o) as Ho.
    pose proof o0 as Ho0.
    pose proof o1 as Ho1.
    eapply usubst_ty in Ho; eauto.
    eapply usubst_ty in Ho0; eauto.
    eapply usubst_ty in Ho1; eauto.
    autorewrite with simpl_subst_pvars in Ho.
    rewrite (subst_pvars_open_commut ).
    rewrite (subst_pvars_open_commut ).

    eapply ty_trm_call; eauto using usubstitution_visible_pvars'.

    rewrite <- (subst_pvars_open_commut ); eauto.
  -
    rewrite trm_subst_pvars_let.
    eapply ty_trm_let. auto.
    intro_fresh_with x Hx (from_list xs \u from_list ys).
    rewrite subst_ctx_pvars_push.
    rewrite <- subst_pvars_open_fresh_commut; eauto.
    eapply H0; eauto.
    rewrite ctx_concat_push; eauto.
  -
    rewrite trm_subst_pvars_lit.
    eapply ty_trm_lit. auto.
    intro_fresh_with x Hx (from_list xs \u from_list ys).
    rewrite subst_ctx_pvars_push.
    rewrite <- subst_pvars_open_fresh_commut; eauto.
    eapply H0; eauto.
    rewrite ctx_concat_push; eauto.
  - eapply ty_trm_sub. eapply H; eauto. eapply usubstitution; eauto.
  -
    rewrite lit_subst_pvars_fun.
    eapply ty_lit_fun; auto.
    intro_fresh_with x Hx (from_list xs \u from_list ys).
    rewrite subst_ctx_pvars_push.
    rewrite <- subst_pvars_open_fresh_commut; eauto.
    rewrite <- subst_pvars_open_fresh_commut; eauto.
    eapply H; eauto.
    rewrite ctx_concat_push; eauto.
  - (*obj*)
    rewrite typ_subst_pvars_objlit.
    eapply ty_lit_obj with (L:=(from_list xs) \u L \u (ctx_dom (G1 §& G2))); eauto.
    intros.
    rewrite <- ! subst_pvars_open_fresh_commut; eauto.
  - (* def_typ *)
    eapply ty_def_typ.
    intro_fresh_with x Hx (from_list xs \u from_list ys).
    specialize (t x).
    rewrite <- subst_pvars_open_fresh_commut; eauto.
    rewrite <- subst_pvars_open_fresh_commut; eauto.
    eapply usubst_pvars_tyq_typ; eauto.
  - (* def_fld *)
    
    eapply ty_def_fld.
    +
      intros.
      rewrite ctx_obj_subst_pvars.
      rewrite ctx_obj_assoc in c.
      eapply usubstitution_visible_pvars'.
      rewrite ctx_obj_dom. eauto.
      auto.
      rewrite ctx_obj_dom. eauto.
    +
      rewrite ctx_obj_assoc in o.
      eapply usubst_ty in o; eauto.
      rewrite ctx_obj_subst_pvars. auto.
      rewrite ctx_obj_dom. eauto.
      rewrite ctx_obj_assoc in H0. auto.
      rewrite ctx_obj_dom. eauto.
  - (* def_met *)
    eapply ty_def_met.
    intro_fresh_with xz Hxz (from_list xs \u from_list ys).
    intros.
    rewrite <- ! subst_pvars_open_fresh_commut; eauto.
    rewrite ctx_obj_met_assoc.
    rewrite <- ctx_obj_met_subst_pvars.
    eapply H; eauto.
    rewrite ctx_obj_met_assoc; eauto.
    fresh_solve.

    eapply fresh_ctx_obj_met; fresh_solve.

    unfold ctx_obj_met; destruct typing_mode; destruct Go; ctx_simpl_dom; simpls; fresh_solve; eapply fresh_from_list; eauto.
    unfold ctx_obj_met; destruct typing_mode; destruct Go; ctx_simpl_dom; simpls; fresh_solve; eapply fresh_from_list; eauto.
    eapply fresh_ctx_obj_met.
    fresh_solve; eapply fresh_from_list; eauto.
  -
    rewrite defs_subst_pvars_cons ; eauto.
    rewrite defs_subst_pvars_nil ; eauto.
  -
    rewrite defs_subst_pvars_cons ; eauto.
    eapply ty_defs_cons; eauto.
    erewrite <- subst_pvars_label_of_def.
    eapply subst_pvars_defs_hasnt; eauto.
Qed.

Global Instance usubst_ty_trm: UnboundSubstitution (ty_trm _).
Proof.
  unfold UnboundSubstitution. intros.
  eapply usubst_rules_ind; eauto.
Qed.
Hint Resolve usubst_ty_trm: core.

Global Instance usubst_ty_lit: UnboundSubstitution (ty_lit _).
Proof.
  unfold UnboundSubstitution. intros.
  eapply usubst_rules_ind; eauto.
Qed.
Hint Resolve usubst_ty_lit: core.

Global Instance usubst_ty_defs: UnboundSubstitution (ty_defs _).
Proof.
  unfold UnboundSubstitution. intros.
  eapply usubst_rules_ind; eauto.
Qed.
Hint Resolve usubst_ty_defs: core.

Lemma subst_ctx_pvars_one : forall (a : ctx) (x y : var), (subst_ctx_pvars (x :: nil)( y :: nil) a ) = (subst_ctx x y a).
Proof.
  intros.
  unfold subst_ctx_pvars.
  induction a; eauto. rewrite -> subst_pvars_one. rewrite IHa. auto.  rewrite IHa. auto. rewrite IHa. auto.
Qed.


Lemma usubst_ty1_trm(y: var) G1 x  G2 e (a: trm) (b: typ):
  typing (G1 §& G2) e a b ->
  ctx_ok (G1 §& G2) ->
  x \notin ctx_fv G1 ->
  renv_notin x e ->
      x \notin ctx_dom G1 ->
      x \notin ctx_dom G2 ->
  typing (G1 §& (subst_ctx x y G2)) e (subst_var x y a) (subst_var x y b).
Proof.
  intros.
  rewrite <- subst_pvars_one.
  rewrite <- subst_pvars_one.
  rewrite <- subst_ctx_pvars_one.
  eapply usubst_ty; eauto. simpls. unfold renv_notin in H2. split; eauto; notin_solve.
Qed.

Lemma subtyp_reopen G e T1 T2 z x:
  ctx_ok G ->
  z \notin (ctx_dom G \u ctx_fv G \u fv_renv e \u fv T1 \u fv T2) ->
  G ⁏ e ⊢ open z T1 <: open z T2 ->
  G ⁏ e ⊢ open x T1 <: open x T2.
Proof.
  introv Hok Hn Hsub.

    rewrite subst_intro with (x0:=z).
    rewrite subst_intro with (x0:=z) (X:=T2).
    eapply usubst_end1; eauto. eauto. eauto.
Qed.

End WithTypingMode.
