(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Bool.Bool.
Require Export Coq.Program.Basics.
From TLC Require Import LibVar LibLN LibFset.
Require Import LibTactics LibSetExtra LibNatExtra.

(** * Lemmas about lists *)

(* Remove element at an index *)
Notation list_removeat n xs := (((LibList.take n xs) ++ (LibList.drop (S n) xs))%list) (only parsing).
Notation list_insertat n xs x := (((LibList.take n xs) ++ (x::(LibList.drop n xs)))%list) (only parsing).

Lemma length_eq_implies_rev_length_eq : forall A B (l1 : list A) (l2 : list B),
    length l1 = length l2 ->
    length (List.rev l1) = length (List.rev l2).
Proof.
  intros. rewrite ? List.rev_length. auto.
Qed.

Lemma rev_length_eq_implies_length_eq : forall A B (l1 : list A) (l2 : list B),
    length (List.rev l1) = length (List.rev l2) ->
    length l1 = length l2.
Proof.
  intros. rewrite ? List.rev_length in *. auto.
Qed.
Hint Resolve length_eq_implies_rev_length_eq : core.

Lemma liblist_length :
  LibList.length = length.
Proof.
  apply fun_ext_dep. intros A.
  apply fun_ext_dep. intros xs.
  induction xs as [ | x xs IHxs].
  - unfold LibList.length. simpl. auto.
  - simpl. rewrite <- IHxs. unfold LibList.length. auto.
Qed.

Lemma liblist_forall :
  LibList.Forall = List.Forall.
Proof.
  apply pred_ext_3; intros A P xs.
  induction xs; split; intro H; inversion H; subst; auto; destruct IHxs; auto.
Qed.

Lemma liblist_app : LibList.app = app.
Proof.
  apply fun_ext_dep. intros A.
  apply fun_ext_dep. intros xs.
  induction xs as [ | x xs IHxs].
  - reflexivity.
  - apply fun_ext_dep. intros ys.
    simpl. rewrite <- IHxs.
    reflexivity.
Qed.

Lemma liblist_rev :
  LibList.rev = List.rev.
Proof.
  apply fun_ext_dep. intros A.
  apply fun_ext_dep. intros xs.
  induction xs as [ | x xs IHxs].
  - reflexivity.
  - simpl. rewrite <- IHxs.
    rewrite LibList.rev_cons. rewrite liblist_app.
    reflexivity.
Qed.

Lemma liblist_fold_right: forall A B,
    @LibList.fold_right B A = @List.fold_right A B.
Proof.
  intros.
  apply fun_ext_dep. intros f.
  apply fun_ext_dep. intros acc.
  apply fun_ext_dep. intros xs.
  induction xs as [ | x xs IHxs].
  - reflexivity.
  - simpl. rewrite <- IHxs.
    reflexivity.
Qed.

Lemma liblist_map : LibList.map = List.map.
Proof.
  apply fun_ext_dep. intros A.
  apply fun_ext_dep. intros B.
  apply fun_ext_dep. intros f.
  apply fun_ext_dep. intros xs.
  induction xs as [ | x xs IHxs].
  - reflexivity.
  - simpl. rewrite <- IHxs.
    reflexivity.
Qed.

  Lemma list_length_cons (xs: list var) y:
    length (y::xs)%list = S (length xs).
  Proof. simpl. auto. Qed.

Lemma list_mem_take A (xs: list A) x m:
  LibList.mem x (LibList.take m xs) ->
  LibList.mem x xs.
Proof.
  introv Hmem.
  reverts xs.
  induction m; intros.
  - rewrite LibList.take_zero in Hmem. inversions Hmem.
  - destruct xs.
    + rewrite LibList.take_nil in Hmem. inversions Hmem.
    + rewrite LibList.take_succ in Hmem. eapply LibList.mem_cons_inv in Hmem. destruct Hmem.  
      * subst.  eapply LibList.mem_cons_l.
      *   eapply LibList.mem_cons_r; auto.
Qed.
Lemma list_mem_drop A (xs: list A) x m:
  LibList.mem x (LibList.drop m xs) ->
  LibList.mem x xs.
Proof.
  introv Hmem.
  reverts xs.
  induction m; intros.
  - rewrite LibList.drop_zero in Hmem. auto.
  - destruct xs.
    + rewrite LibList.drop_nil in Hmem. inversions Hmem.
    + rewrite LibList.drop_succ in Hmem.
      
      eapply LibList.mem_cons_r; auto.
Qed.

Lemma list_map_take A B (xs: list A) (f: A->B) m:
  LibList.map f (LibList.take m xs) = LibList.take m (LibList.map f xs).
Proof.
  reverts xs.
  induction m; intros.
  - rewrite ! LibList.take_zero. rewrite LibList.map_nil. auto.
  - destruct xs.
    + rewrite ! LibList.take_nil. rewrite LibList.map_nil. auto.
    + rewrite ! LibList.take_succ. rewrite ! LibList.map_cons. rewrite ! LibList.take_succ. rewrite IHm. auto.
Qed.

Lemma list_map_drop A B (xs: list A) (f: A->B) m:
  LibList.map f (LibList.drop m xs) = LibList.drop m (LibList.map f xs).
Proof.
  reverts xs.
  induction m; intros.
  - rewrite ! LibList.drop_zero. auto.
  - destruct xs.
    + rewrite ! LibList.drop_nil. rewrite LibList.map_nil. auto.
    + rewrite ! LibList.drop_succ. rewrite ! LibList.map_cons. rewrite ! LibList.drop_succ. rewrite IHm. auto.
Qed.

(* NO DUPLICATES *)

Lemma list_noduplicates_succ ns:
  LibList.noduplicates ns ->
  LibList.noduplicates (List.map S ns).
Proof.
  introv Hnd.
  induction Hnd. simpls. eapply LibList.noduplicates_nil.
    simpls. constructor; auto.
    intro. apply H. clear H Hnd IHHnd.
    induction l. simpls. false. eapply LibList.mem_nil_inv. eauto. simpls.
    destruct (LibList.mem_cons_inv H0). inversions H. auto.
    auto.
Qed.
Lemma list_mem_0_map_succ_false (ns: list nat):
 ~ LibList.mem 0 (List.map S ns).
Proof.
  - intro.
    induction ns.
    + simpls.
      eapply LibList.mem_nil_inv; eauto.
    + simpls. destruct (LibList.mem_cons_inv H). discriminate. auto.
Qed.

Lemma list_noduplicates_shift ns:
  LibList.noduplicates ns ->
  LibList.noduplicates (0::(List.map S ns)).
Proof.
  introv Hnd.
  eapply LibList.noduplicates_cons.
  - eapply list_mem_0_map_succ_false; eauto.
  - eapply list_noduplicates_succ; eauto.
Qed.

Lemma list_noduplicates_concat_middle_inv A (xs: list A) y ys:
  LibList.noduplicates (xs ++ y :: ys) ->
  LibList.noduplicates (xs ++ ys).
Proof.
  introv Hnd.
  induction xs.
  - rewrite List.app_nil_l in *. inversions Hnd; auto.
  - rewrite <- List.app_comm_cons in *.
    inversions Hnd.
    constructor;
    eauto.
    intro. apply H1.
    rewrite <- liblist_app in *.
    eapply LibList.mem_app_inv in H. destruct H.
    eapply LibList.mem_app_l; eauto.
    eapply LibList.mem_app_r; eauto.
Qed.

Lemma list_noduplicates_take A (xs: list A) m:
  LibList.noduplicates xs ->
  LibList.noduplicates (LibList.take m xs).
Proof.
  introv Hnd.
  reverts xs.
  induction m; intros.
  - rewrite LibList.take_zero. constructor.
  - destruct xs.
    + rewrite LibList.take_nil. constructor.
    + inversions Hnd.
      rewrite LibList.take_succ. constructor.
      intro. apply H1. eapply list_mem_take; eauto. eauto.
Qed.

Lemma list_noduplicates_drop A (xs: list A) m:
  LibList.noduplicates xs ->
  LibList.noduplicates (LibList.drop m xs).
Proof.
  introv Hnd.
  reverts xs.
  induction m; intros.
  - rewrite LibList.drop_zero. auto.
  - destruct xs.
    + rewrite LibList.drop_nil. constructor.
    + inversions Hnd.
      rewrite LibList.drop_succ. eauto.
Qed.

Lemma list_noduplicates_app_swap A (ns1 ns2: list A):
  LibList.noduplicates (ns1 ++ ns2) ->
  LibList.noduplicates (ns2 ++ ns1).
Proof.
  introv Hnd.
  rewrite <- liblist_app in *.
  forwards Ha: LibList.noduplicates_app_inv Hnd.
  destructecs Ha.
  eapply LibList.noduplicates_app; eauto.
Qed.

Lemma list_noduplicates_insert_middle ns1 ns2:
  LibList.noduplicates (ns1 ++ ns2) ->
  LibList.noduplicates (List.map S ns1 ++ 0 :: List.map S ns2).
Proof.
  intros Hnd.
  eapply list_noduplicates_app_swap.
  rewrite <- List.app_comm_cons.
  rewrite <- List.map_app.
  eapply list_noduplicates_shift.
  eapply list_noduplicates_app_swap. auto.
Qed.

Lemma from_list_append (X: Type) (xs ys: list X):
  from_list(xs++ys) = from_list(xs) \u from_list (ys).
Proof.
  induction xs; simpls.
  - rewrite from_list_nil. rewrite union_empty_l. auto.
  - rewrite from_list_cons. rewrite from_list_cons.
    rewrite IHxs. rewrite union_assoc. auto.
Qed.

Lemma list_mem_in_fromlist X x (l: list X):
  LibList.mem x l -> x \in (from_list l).
Proof. introv Hmem.
  induction Hmem.
  - rewrite from_list_cons. rewrite in_union. left. rewrite in_singleton. auto.
  - rewrite from_list_cons. rewrite in_union. right. auto.
Qed.

Lemma list_in_fromlist_mem X x (l: list X):
  x \in (from_list l) -> LibList.mem x l.
Proof. introv Hfl.
  induction l.
  - rewrite from_list_nil in Hfl. rewrite in_empty in Hfl. contradiction.
  - rewrite from_list_cons in Hfl. rewrite in_union in Hfl. destruct Hfl.
    rewrite in_singleton in H. subst. eauto.
    eauto.
Qed.

Lemma list_Nth_cut X n (xs: list X) (v: X):
  LibList.Nth n xs v ->
  exists (xs1 xs2: list X),
    xs = (xs1 ++ v::xs2)%list /\
    (length xs1 = n).
Proof.
  introv Hn.
  induction Hn.
  - exists (@nil X) l.
    eauto.
  - destructc IHHn as xs1 xs2 Hxs Hl.
    exists (y::xs1)%list xs2.
    subst.
    rewrite List.app_comm_cons.
    eauto.
Qed.
Lemma list_Nth_cut_take_drop X n (xs: list X) (v: X):
  LibList.Nth n xs v ->
   xs = ((LibList.take n xs) ++ v::(LibList.drop (S n) xs))%list.
Proof.
  introv Hn.
  induction Hn.
  - rewrite LibList.take_zero.
    rewrite LibList.drop_succ.
    rewrite LibList.drop_zero.
    rewrite List.app_nil_l. auto.
  - rewrite LibList.drop_succ.
    rewrite LibList.take_succ.
    rewrite <- List.app_comm_cons.
    rewrite <- IHHn. auto.
Qed.

Lemma list_make_drop X n m (y: X):
  n <= m ->
  LibList.drop n (LibList.make m y) = LibList.make (m - n) y.
Proof.
  introv Hnm.
  reverts m.
  induction n; intros.
  -    rewrite LibList.drop_zero. rewrite PeanoNat.Nat.sub_0_r. auto.
  - destruct m. inversions Hnm.
    eapply le_S_n in Hnm.
    rewrite LibList.make_succ.
    rewrite LibList.drop_succ.
    rewrite IHn by auto.
    rewrite PeanoNat.Nat.sub_succ.
    auto.
Qed.

Lemma list_make_take X n m (y: X):
  n <= m ->
  LibList.take n (LibList.make m y) = LibList.make n y.
Proof.
  introv Hnm.
  reverts m.
  induction n; intros.
  -    rewrite LibList.take_zero. rewrite LibList.make_zero. auto.
  - destruct m. inversions Hnm.
    eapply le_S_n in Hnm.
    rewrite LibList.make_succ.
    rewrite LibList.make_succ.
    rewrite LibList.take_succ.
    rewrite IHn by auto.
    auto.
Qed.

Lemma list_mem_Nth_first X xs (y: X): (* Also see LibList.mem_inv_middle_first *)
  LibList.mem y xs ->
  exists n,
  LibList.Nth n xs y /\ y \notin (from_list (LibList.take n xs)).
Proof.
  introv Hm.
  induction Hm.
  - exists 0. split. eauto. rewrite LibList.take_zero. rewrite (from_list_nil). notin_solve.
  - destruct (classicT (y= y0)).
    + subst.
    exists 0. split. eauto. rewrite LibList.take_zero. rewrite (from_list_nil). notin_solve.
    +
    destructecs IHHm.
    exists (S n0).
    split.
    eauto.
    rewrite LibList.take_succ. rewrite from_list_cons. rewrite notin_union, notin_singleton.
    auto.
Qed.
Lemma list_Nth_last_or_notin X xs (y: X):
  y \notin (from_list xs) \/
  exists n,
  LibList.Nth n xs y /\ y \notin (from_list (LibList.drop (S n) xs)).
Proof.
  induction xs.
  - left. rewrite from_list_nil. notin_solve.
  - destructecs IHxs.
    destruct (classicT (y = a)).
    + subst. right. exists 0.
      split. eauto. rewrite LibList.drop_succ. rewrite LibList.drop_zero. auto.
    + left. rewrite from_list_cons, notin_union, notin_singleton. auto.
    + (* From induction *)
      right.
     exists (S n).
    split.
    eauto.
    rewrite LibList.drop_succ. auto.
Qed.
Lemma list_mem_Nth_last X xs (y: X):
  LibList.mem y xs ->
  exists n,
  LibList.Nth n xs y /\ y \notin (from_list (LibList.drop (S n) xs)).
Proof.
  introv Hm.

  forwards Hc: list_Nth_last_or_notin xs y.
  destructecs Hc.
  - false. eapply list_mem_in_fromlist in Hm. contradiction.
  - eauto.
Qed.

Lemma list_Nth_app_invl X n (xs1 xs2: list X) v:
  LibList.Nth n (xs1 ++ xs2) v ->
  n < length xs1 ->
  LibList.Nth n xs1 v.
Proof.
  introv Hn Hl.
  rewrite <- liblist_app in Hn.
  eapply LibList.Nth_app_inv in Hn.
  destruct Hn.
  - auto.
  - false.
    destructc H as m Hm ?.
    subst.
    rewrite liblist_length in Hl.
    eapply lt_l_plus_false; eauto.
Qed.

Lemma list_decrease_increase ns m:
  ns = LibList.map (decrease_after m) (LibList.map (increase_after m) ns).
Proof.
 induction ns.
 - eauto.
 - rewrite LibList.map_cons. rewrite LibList.map_cons.
   rewrite <- IHns.
   rewrite decrease_increase_after. eapply f_equal. auto.
Qed.
Lemma list_increase_decrease ns m:
  ~ LibList.mem m ns ->
  ns = LibList.map (increase_after m) (LibList.map (decrease_after m) ns).
Proof.
  introv Hmem.
  induction ns.
  - eauto.
  - rewrite LibList.map_cons. rewrite LibList.map_cons.
    rewrite increase_decrease_after. eapply f_equal. eapply IHns; eauto. intro. eapply Hmem. inversions H; eauto.
Qed.

Lemma list_mem_decrease_after_inv a ns m:
  m <> a ->
  ~ LibList.mem m ns ->
  LibList.mem (decrease_after m a) (LibList.map (decrease_after m) ns)->
  LibList.mem a ns.
Proof.
  introv Hma Hnm Hmem.
  eapply (LibList.mem_map (increase_after m)) in Hmem.
  rewrite increase_decrease_after in Hmem by auto.
  replace ((LibList.map (increase_after m) (LibList.map (decrease_after m) ns))) with ns in Hmem.
  auto.
  eapply list_increase_decrease; eauto.
Qed.
Lemma list_mem_inrease_after_inv a ns m:
  LibList.mem (increase_after m a) (LibList.map (increase_after m) ns)->
  LibList.mem a ns.
Proof.
  introv Hmem.
  eapply (LibList.mem_map (decrease_after m)) in Hmem.
  rewrite decrease_increase_after in Hmem by auto.
  replace ((LibList.map (decrease_after m) (LibList.map (increase_after m) ns))) with ns in Hmem.
  auto.
  eapply list_decrease_increase; eauto.
Qed.



Lemma list_noduplicates_remove_middle ns m:
  LibList.noduplicates ns ->
  ~ LibList.mem (m) ns ->
  LibList.noduplicates (LibList.map (decrease_after (m)) ns).
Proof.
  introv Hnd Hmem.
  induction ns.
  - rewrite LibList.map_nil. eauto.
  - rewrite LibList.map_cons.
    inversions Hnd.
    constructor.
    * intro.
      apply H1.
      eapply list_mem_decrease_after_inv with (m:= m); eauto.
      intro; subst. apply Hmem. eapply LibList.mem_cons_l.
      * eapply IHns; eauto.
Qed.
Lemma list_noduplicates_select_middle ns m:
  LibList.noduplicates ns ->
  LibList.noduplicates (m :: LibList.map (increase_after m) ns).
Proof.
  introv Hnd.
  induction Hnd.
  - rewrite LibList.map_nil. constructor. intro. eapply LibList.mem_nil_inv; eauto. constructor.
  - inversions IHHnd.
    constructor.
    +
      intro. apply H2. rewrite LibList.map_cons in H0. eapply LibList.mem_cons_inv in H0. destruct H0.
      * unfold increase_after in H0. cases_if; subst.
        -- eapply PeanoNat.Nat.nle_succ_diag_l in C. contradiction.
        -- eapply PeanoNat.Nat.lt_irrefl in C. contradiction.
      * auto.
    + rewrite LibList.map_cons.
      constructor.
      * intro. apply H.
        eapply list_mem_inrease_after_inv; eauto.
      * auto.
Qed.

Lemma liblist_mem_map_inj(A B : Type) (f : A -> B) (l : list A) (x : A):
  (forall x y : A, f x = f y -> x = y) ->
  LibList.mem (f x) (LibList.map f l) ->
  LibList.mem x l.
Proof.
  introv Hinj Hmem.
  induction l.
  - rewrite LibList.map_nil in Hmem. inversions Hmem.
  - rewrite LibList.map_cons in Hmem. inversions Hmem.
    eapply Hinj in H0; subst. eauto.
    eauto.
Qed.

Lemma liblist_mem_map_inj_exists(A B : Type) (f : A -> B) (l : list A) (y : B):
  (forall x y : A, f x = f y -> x = y) ->
  LibList.mem y (LibList.map f l) ->
  exists x, y = (f x) /\
  LibList.mem x l.
Proof.
  introv Hinj Hmem.
  induction l.
  - rewrite LibList.map_nil in Hmem. inversions Hmem.
  - rewrite LibList.map_cons in Hmem. inversions Hmem.
    + exists a. splits; eauto.
    + forwards x Hx Hm: IHl; eauto.
Qed.

Lemma list_noduplicates_add ns m: 
  LibList.noduplicates ns ->
  LibList.noduplicates (LibList.map (plus m) ns).
Proof.
  introv Hnd.
  induction Hnd.
  - rewrite LibList.map_nil. constructor.
  - rewrite LibList.map_cons.
    constructor.
    intro. apply H.
    eapply liblist_mem_map_inj with (f:=(plus m)).
    intros. apply -> PeanoNat.Nat.add_cancel_l; eauto.
    eauto.    eauto.
Qed.

Lemma list_length_remove_middle A B (x1 x2: list A) y (x: list B) (ns: list nat):
  length (x1 ++ y :: x2) = length (length x :: ns) ->
  length x1 + length x2 = length ns.
Proof.
  introv Hlx.
  rewrite List.app_length in Hlx. simpl in Hlx.
  rewrite PeanoNat.Nat.add_succ_r in Hlx. inversions Hlx. auto.
Qed.


Lemma list_mem_succ_inv n ns:
  LibList.mem (S n) (List.map S ns) ->
  LibList.mem n ns.
Proof.
  introv Hmem.
    forwards: LibList.mem_map Nat.pred; eauto.
    rewrite ! liblist_map in H.
    rewrite List.map_map in H.
    rewrite PeanoNat.Nat.pred_succ in H.
    replace (fun x0 : nat => Nat.pred (S x0)) with (fun x0: nat => x0) in H.
    rewrite List.map_id in H; auto.
    eapply fun_ext_1. intro.
    rewrite (PeanoNat.Nat.pred_succ). auto.
Qed.

Lemma list_notmem_cons_inv (A: Type) (a b: A) xs:
  ~ LibList.mem a (b::xs) ->  
  a <> b /\ ~ LibList.mem a (xs).
Proof.
  introv Hm.
  split; intro.
  subst. apply Hm. auto.
  apply Hm. auto.
Qed.

Lemma list_map_S_pred_notmem_0 ns:
  ~ LibList.mem 0 ns ->
  (LibList.map S (LibList.map Nat.pred ns)) = ns.
Proof.
  introv Hnm.
  induction ns.
  - rewrite LibList.map_nil. auto.
  - eapply list_notmem_cons_inv in Hnm.
    destructecs Hnm.
    rewrite ! LibList.map_cons.
    rewrite PeanoNat.Nat.succ_pred by auto.
    rewrite IHns by auto.
    auto.
Qed.


Lemma list_map_S_pred_remove_0 ns:
  (LibList.map S (LibList.map Nat.pred (LibList.remove 0 ns))) = (LibList.remove 0 ns).
Proof.
  eapply list_map_S_pred_notmem_0.
  intro.
  eapply LibList.mem_remove_same_inv; eauto.
Qed.


Lemma list_remove_cons_eq (A: Type) l (a: A):
  LibList.remove a (a::l) = LibList.remove a l.
Proof.
  unfold LibList.remove.
  rewrite LibList.filter_cons.
  cases_if; auto.
Qed.
Lemma list_remove_cons_ne (A: Type) l (a b: A):
  a <> b ->
  LibList.remove a (b::l) = (b::(LibList.remove a l))%list.
Proof.
  introv Hne.
  unfold LibList.remove.
  rewrite LibList.filter_cons.
  cases_if; auto. false C. auto.
Qed.

Lemma notin_from_list_remove (A: Type) l (a: A):
  a \notin from_list (LibList.remove a l).
Proof.
  unfold LibList.remove.
  induction l.
  - rewrite LibList.filter_nil. rewrite from_list_nil. eauto.
  - rewrite LibList.filter_cons.
    cases_if.
    + rewrite from_list_cons. rewrite notin_union; split; notin_solve. rewrite notin_singleton. auto.
    + auto.
Qed.

Lemma notin_from_list_remove_from_list (A: Type) l (a b: A):
  a \notin from_list l ->
  a \notin from_list (LibList.remove b l).
Proof.
  unfold LibList.remove.
  induction l.
  - rewrite LibList.filter_nil. rewrite from_list_nil. eauto.
  - rewrite LibList.filter_cons.
    rewrite from_list_cons. rewrite notin_union.
    intros. destructecs H.
    cases_if.
    + rewrite from_list_cons. rewrite notin_union; split; notin_solve. auto.
    + auto.
Qed.


Lemma notin_from_list_drop_from_list (A: Type) l (a: A) n:
  a \notin from_list l ->
  a \notin from_list (LibList.drop n l).
Proof.
  unfold LibList.remove.
  reverts n.
  induction l; intros.
  - rewrite LibList.drop_nil. rewrite from_list_nil. eauto.
  - 
    rewrite from_list_cons in H. rewrite notin_union in H.
    destructecs H.
    destruct n.
    + rewrite LibList.drop_zero. rewrite from_list_cons. rewrite notin_union; auto.
    + rewrite LibList.drop_succ. auto.
Qed.
Lemma notin_from_list_take_from_list (A: Type) l (a: A) n:
  a \notin from_list l ->
  a \notin from_list (LibList.take n l).
Proof.
  unfold LibList.remove.
  reverts n.
  induction l; intros.
  - rewrite LibList.take_nil. rewrite from_list_nil. eauto.
  - 
    rewrite from_list_cons in H. rewrite notin_union in H.
    destructecs H.
    destruct n.
    + rewrite LibList.take_zero. rewrite from_list_nil. eauto.
    + rewrite LibList.take_succ.  rewrite from_list_cons. rewrite notin_union; auto.
Qed.
Lemma from_list_make (A: Type) l (a:A):
  from_list (LibList.make l a) = match l with | 0 => \{} | S n => \{a} end.
Proof.
  induction l.
  - rewrite LibList.make_zero. rewrite from_list_nil. auto.
  - rewrite LibList.make_succ. rewrite from_list_cons. rewrite IHl.
    destruct l. rewrite union_empty_r. auto. rewrite union_same. auto.
Qed.
Lemma notin_from_list_make (A: Type) l (a b:A):
  b <> a ->
  b \notin from_list (LibList.make l a).
Proof.
  introv Hn.
  forwards: (from_list_make l a). cases_if.
  rewrite H. notin_solve.
  rewrite H. rewrite notin_singleton. auto.
Qed.

Lemma from_list_remove (A: Type) l (a: A):
  from_list (LibList.remove a l) = (from_list l) \- \{a}.
Proof.
  unfold LibList.remove.
  induction l.
  - rewrite LibList.filter_nil. rewrite from_list_nil. rewrite remove_empty_r. auto.
  - rewrite LibList.filter_cons.
    cases_if.
    + rewrite ! from_list_cons. rewrite IHl.
      rewrite union_remove_swap0; eauto. rewrite notin_singleton; auto.
    + destruct(classicT(a0=a)). subst. rewrite IHl. rewrite from_list_cons. rewrite union_remove. rewrite remove_same. rewrite union_empty_l. auto.
      contradiction.
Qed.

Lemma from_list_remove_duplicates A (xs: list A):
  from_list (LibList.remove_duplicates xs) = (from_list xs).
Proof.
  induction xs; simpls; auto.
  rewrite ! from_list_cons.
  rewrite from_list_remove.
  rewrite IHxs. rewrite union_comm.
  rewrite remove_union_same. rewrite union_comm. auto.
Qed.

Lemma from_list_empty_inv A (xs: list A):
  from_list xs = \{} ->
  xs = nil.
Proof.
  introv He.
  destruct xs.
  - auto.
  - false.
    rewrite from_list_cons in He.
    eapply union_empty_inv in He.
    destructecs He.
    eapply singleton_ne_empty. symmetry; eauto.
Qed.

Lemma list_forall_app_invl A P (xs ys: list A):
  List.Forall P (xs ++ ys) -> List.Forall P xs.
Proof.
  introv Hf.
  induction xs.
  eauto.
  rewrite <- List.app_comm_cons in Hf. inversions Hf.
  eauto.
Qed.

Lemma list_forall_app_invr A P (xs ys: list A):
  List.Forall P (xs ++ ys) -> List.Forall P ys.
Proof.
  introv Hf.
  induction xs.
  eauto.
  rewrite <- List.app_comm_cons in Hf. inversions Hf.
  eauto.
Qed.
Lemma list_forall_app A P (xs ys: list A):
  List.Forall P xs ->
  List.Forall P ys ->
  List.Forall P (xs ++ ys).
Proof.
  introv Hfx Hfy.
  induction xs.
  eauto.
  rewrite <- List.app_comm_cons.
  inversions Hfx.
  constructor; eauto.
Qed.

Lemma list_forall_nth A P n (xs: list A) x:
  LibList.Nth n xs x ->
  List.Forall P xs ->
  P x.
Proof.
  introv Hn Hf.
  induction Hn; inversions Hf; eauto.
Qed.

Lemma list_forall_map A P (xs: list A) (f: A->A):
  List.Forall P xs ->
  (forall x, P x -> P (f x) )->
  List.Forall P (List.map f xs).
Proof.
  introv Hx Hf.
  induction xs.
  eauto.
  inversions Hx.
  constructor; eauto.
Qed.
Lemma list_forall_map_inv A P (xs: list A) (f: A->A):
  List.Forall P (List.map f xs) ->
  (forall x, P (f x) -> P x )->
  List.Forall P xs.
Proof.
  introv Hx Hf.
  induction xs.
  eauto.
  inversions Hx.
  constructor; eauto.
Qed.

Lemma list_removeat_length A (xs: list A) n:
  n < length xs ->
  length (list_removeat n xs) = Nat.pred (length xs).
Proof.
  introv Hlt.
  rewrite List.app_length.
  rewrite <- liblist_length.
  rewrite LibList.length_take, LibList.length_drop.
  +
    destruct xs.
    - simpls. eapply PeanoNat.Nat.nlt_0_r in Hlt. false Hlt.
    - rewrite liblist_length. simpls.
      eapply Minus.le_plus_minus_r. apply PeanoNat.Nat.lt_succ_r; auto.
  + rewrite LibNat.le_peano, liblist_length. apply Lt.lt_le_S; auto.
  + rewrite LibNat.le_peano, liblist_length. apply PeanoNat.Nat.lt_le_incl; auto.
Qed.

Lemma list_insertat_length A (xs: list A) n x:
  n <= length xs ->
  length (list_insertat n xs x) = S (length xs).
Proof.
  introv Hlt.
  rewrite List.app_length.
  rewrite <- liblist_length.
  rewrite LibList.length_take, LibList.length_cons, LibList.length_drop.
  + rewrite PeanoNat.Nat.add_assoc. rewrite PeanoNat.Nat.add_comm with (n:=n).
    rewrite <- PeanoNat.Nat.add_assoc.
    rewrite Minus.le_plus_minus_r.
    rewrite PeanoNat.Nat.add_1_l. auto.
    rewrite liblist_length. auto. 
  + rewrite LibNat.le_peano, liblist_length. auto.
  + rewrite LibNat.le_peano, liblist_length. auto.
Qed.

Lemma from_list_take_subset A (xs: list A) n:
  from_list (LibList.take n xs) \c from_list xs.
Proof.
  intro. intro.
  destruct (classicT (x \in (from_list xs))).
  - auto.
  - false notin_from_list_take_from_list; eauto.
Qed.

Lemma from_list_drop_subset A (xs: list A) n:
  from_list (LibList.drop n xs) \c from_list xs.
Proof.
  intro. intro.
  destruct (classicT (x \in (from_list xs))).
  - auto.
  - false notin_from_list_drop_from_list; eauto.
Qed.


Lemma from_list_removeat_subset A (xs: list A) n:
  from_list (list_removeat n xs) \c from_list xs.
Proof.
  rewrite from_list_append.
  eapply subset_union_l;
  eauto using from_list_take_subset, from_list_drop_subset.
Qed.
Lemma list_app_inv_length1 A (xs1 xs2 ys1 ys2: list A):
  (xs1 ++ xs2)%list = (ys1 ++ ys2)%list ->
  length xs1 = length ys1 ->
  xs1 = ys1 /\ xs2 = ys2.
Proof.
  introv Hl Heq.  
  reverts ys1.
  induction xs1; intros; destruct ys1; inversions Heq.
  - rewrite ! List.app_nil_l in Hl. auto.
  - rewrite <-! List.app_comm_cons in Hl.
    inversions Hl.
    forwards [He1 He2]: IHxs1; fweauto; subst.
    auto.
Qed.
Lemma list_app_inv_length2 A (xs1 xs2 ys1 ys2: list A):
  (xs1 ++ xs2)%list = (ys1 ++ ys2)%list ->
  length xs2 = length ys2 ->
  xs1 = ys1 /\ xs2 = ys2.
Proof.
  introv Hl Heq.
  eapply list_app_inv_length1.
  eauto.
  eapply PeanoNat.Nat.add_cancel_r with (p:=length xs2).
  rewrite Heq at 2.
  rewrite <- ! List.app_length.
  rewrite Hl; auto.
Qed.

Lemma list_drop_cons_Nth A (xs: list A) (x:A) (n: nat):
  LibList.Nth n xs x ->
  (LibList.drop n xs)%list = (x:: (LibList.drop (S n) xs))%list.
Proof.
  introv Hnth.
  reverts xs x.
  induction n; intros.
  - inversions Hnth. unfold LibList.drop. eauto.
  - inversions Hnth.
    rewrite LibList.drop_cons.
    rewrite LibList.drop_cons.
    eauto.
Qed.

Lemma list_split_at A (xs: list A) (x:A) (n: nat):
  LibList.Nth n xs x ->
  xs = ((LibList.take n xs) ++ x:: (LibList.drop (S n) xs))%list.
Proof.
  introv Hn.
  forwards Had: LibList.list_eq_take_app_drop n xs.
  eapply LibList.Nth_inbound; eauto.
  rewrite <- Had at 1.
  rewrite liblist_app. fequal.
  eapply list_drop_cons_Nth; eauto.
Qed.



Lemma list_Nth_take_length A ( xs: list A) n x:
  LibList.Nth n xs x ->
  length (LibList.take n xs) = n.
Proof.
  introv Hn.
  rewrite <- liblist_length.
  rewrite ! LibList.length_take. eauto. eapply LibList.Nth_inbound; eauto.
Qed.

Lemma list_Nth_drop_S_length A ( xs: list A) n x:
  LibList.Nth n xs x ->
  length (LibList.drop (S n) xs) = length xs - (S n).
Proof.
  introv Hn.
  rewrite <- liblist_length.
  rewrite ! LibList.length_drop. eauto.
  rewrite LibNat.le_peano. rewrite PeanoNat.Nat.le_succ_l. rewrite <- LibNat.lt_peano. eapply LibList.Nth_inbound; eauto.
Qed.

Lemma from_list_drop A (xs: list A) n:
  from_list (LibList.drop n xs) \c from_list xs.
Proof.
  reverts xs; induction n; intros.
  simpls. eapply subset_refl.
  destruct xs.
  simpls. eapply subset_refl.
  rewrite LibList.drop_succ. rewrite from_list_cons. eapply subset_trans. eapply IHn.  eapply subset_union_weak_r.
Qed.
Lemma from_list_take A (xs: list A) n:
  from_list (LibList.take n xs) \c from_list xs.
Proof.
  reverts xs; induction n; intros.
  simpls. rewrite from_list_nil. eapply subset_empty_l.
  destruct xs.
  simpls. eapply subset_refl.
  rewrite LibList.take_succ. rewrite ! from_list_cons. eapply subset_union_2. eapply subset_refl. eauto.
Qed.


Lemma list_removeat_map A B (l: list A) (f: A->B) n:
  LibList.map f (list_removeat n l) = list_removeat n (LibList.map f l).
  rewrite <- liblist_app.
  rewrite LibList.map_app.
  rewrite list_map_drop.
  rewrite list_map_take. auto.
Qed.


Lemma list_remove_insert_at A (S: list A) m x:
  LibList.Nth m S x ->
  list_insertat m (list_removeat m S) x = S.
Proof.
  introv Hn.
  forwards: LibList.Nth_inbound Hn.
  rewrite (list_split_at Hn) at 5.
  rewrite <- liblist_app.
  rewrite LibList.take_app_l.
  rewrite LibList.drop_app_r.
  rewrite LibList.length_take.
  rewrite PeanoNat.Nat.sub_diag.
  rewrite LibList.drop_zero.
  rewrite LibList.take_ge.
  auto.

  rewrite LibList.length_take. apply LibOrder.ge_refl.
  eapply LibOrder.lt_to_le; eauto.
  eapply LibOrder.lt_to_le; eauto.
  rewrite LibList.length_take. apply LibOrder.ge_refl.
  eapply LibOrder.lt_to_le; eauto.
  rewrite LibList.length_take. apply LibOrder.le_refl.
  eapply LibOrder.lt_to_le; eauto.

Qed.


