(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Bool.Bool.
Require Export Coq.Program.Basics.
From TLC Require Import LibVar LibLN LibFset.
Require Import LibTactics LibSetExtra LibListExtra LibNatExtra.


(** * List reordering *)

(* Reordering list: takes an input list and a list of indices, produces a list witch each element selected at an index *)
Inductive list_reorder A: (list A) -> (list nat) -> (list A) -> Prop :=
| list_reorder_nil xs:
    list_reorder xs nil nil
| list_reorder_cons xs y ys n ns:
    list_reorder xs ns ys ->
    LibList.Nth n xs y ->
    list_reorder xs (n::ns) (y::ys).

Lemma list_reorder_push0 A (xs0 xs1: list A) ns x:
  list_reorder xs0 ns xs1 ->
  list_reorder (x::xs0) (List.map S ns) xs1.
Proof.
  intros.
  induction H.
  - constructor.
  - constructor; auto.
Qed.

Lemma list_reorder_push A (xs0 xs1: list A) ns x:
  list_reorder xs0 ns xs1 ->
  list_reorder (x::xs0) (0::(List.map S ns)) (x :: xs1).
Proof.
  intros.
  constructor.
  eapply list_reorder_push0; auto.
  eapply LibList.Nth_zero.
Qed.

Lemma list_reorder_lengths A (xs0: list A) ns xs:
  list_reorder xs0 ns xs ->
  (length xs = length ns).
Proof.
  introv Hro.
  induction Hro.
  - auto.
  - simpls. rewrite IHHro. auto.
Qed.

Lemma list_reorder_one A (xs: list A) x:
  list_reorder (x::xs) (0::nil) (x::nil).
Proof.
  eapply list_reorder_cons.
  eapply list_reorder_nil.
  eapply LibList.Nth_zero.
Qed.
Lemma list_reorder_old_one A (xs: list A) x n:
  LibList.Nth n xs x ->
  list_reorder (xs) (n::nil) (x::nil).
Proof.
  eapply list_reorder_cons.
  eapply list_reorder_nil.
Qed.

Lemma list_reorder_same_length_exists A B (xs1 xs2: list A) (xs3: list B) ns:
  list_reorder xs1 ns xs2 ->
  length xs1 = length xs3 ->
  exists xs4,
    list_reorder xs3 ns xs4.
Proof.
  intros.
  induction H.
  - exists (@nil B). constructor.
  - destructc (IHlist_reorder H0) as ? ?.
    eapply LibList.Nth_inbound in H1.
    rewrite liblist_length in H1.
    rewrite H0 in H1.
    rewrite <- liblist_length in H1.
    eapply LibList.Nth_inbound_inv in H1.
    destructc H1 as x0 ?.
    exists (x0::x)%list.
    constructor; eauto.
Qed.

Lemma list_reorder_nth A (xs: list A) n x:
  LibList.Nth n xs x ->
  list_reorder xs (n::nil) (x::nil).
Proof.
  eapply list_reorder_cons.
  eapply list_reorder_nil.
Qed.



Lemma list_reorder_nth_nth A (xs0: list A) ns1 xs1 x n:
  list_reorder xs0 ns1 xs1 ->
  LibList.Nth n xs1 x ->
  LibList.Nth (LibList.nth n ns1) xs0 x.
Proof.
  introv Hrox Hnx.
  reverts n.
  induction Hrox; intros; simpls.
  - eapply LibList.Nth_nil_inv in Hnx. contradiction.
  - inversions Hnx.
    + rewrite LibList.nth_zero. eauto.
    + rewrite LibList.nth_succ. eapply IHHrox. eauto.
Qed.

Lemma list_reorder_nth_ys_nth_xs A (xs: list A) ns ys x n m:
  list_reorder xs ns ys ->
  LibList.Nth n ys x ->
  LibList.Nth n ns m ->
  LibList.Nth m xs x.
Proof.
  introv Hrox Hny Hnn.
  forwards: list_reorder_nth_nth; eauto.
  erewrite LibList.nth_of_Nth in H by eauto.
  eauto.
Qed.

Lemma list_reorder_nth_Nth A (xs0: list A) ns1 xs1 x n:
  list_reorder xs0 ns1 xs1 ->
  LibList.Nth n xs1 x ->
  exists (m: nat), LibList.Nth n ns1 m /\ LibList.Nth m xs0 x.
Proof.
  introv Hrox Hnx.
  reverts n.
  induction Hrox; intros; simpls.
  - eapply LibList.Nth_nil_inv in Hnx. contradiction.
  - inversions Hnx.
    + exists n. eauto.
    + destructc (IHHrox n1) as ? ? ?; eauto.
Qed.

Lemma list_reorder_nth_xs_nth_ys A (xs: list A) ns ys x n m:
  list_reorder xs ns ys ->
  LibList.Nth m xs x ->
  LibList.Nth n ns m ->
  LibList.Nth n ys x.
Proof.
  introv Hrox Hnx Hnn.
  reverts n.
  induction Hrox; intros; simpls.
  - eapply LibList.Nth_nil_inv in Hnn. contradiction.
  - inversions Hnn.
    +
    forwards: LibList.Nth_functional H Hnx. subst.
    auto.
    + forwards: IHHrox; fweauto. eauto.
Qed.


Lemma list_reorder2_nth_nth A (xs0 ys0: list A) ns1 xs1 ys1 x y n:
  list_reorder xs0 ns1 xs1 ->
  list_reorder ys0 ns1 ys1 ->
  LibList.Nth n xs1 x ->
  LibList.Nth n ys1 y ->
  exists m, LibList.Nth m xs0 x /\ LibList.Nth m ys0 y.
Proof.
  introv Hrox Hroy Hnx Hny.

  exists (LibList.nth n ns1).
  split.
  - eapply list_reorder_nth_nth; eauto.
  - eapply list_reorder_nth_nth; eauto.
Qed.

Lemma list_reorder_app A (xs: list A) ns1 ns2 ys1 ys2:
  list_reorder xs ns1 ys1 ->
  list_reorder xs ns2 ys2 ->
  list_reorder xs (ns1++ns2) (ys1++ys2).
Proof.
  introv Hre1 Hre2.
  reverts ys1.
  induction ns1; intros.
  - inversions Hre1.
    rewrite ! List.app_nil_l; auto.
  - inversions Hre1.
    rewrite <- ! List.app_comm_cons.
    constructor; auto.
Qed.

Lemma list_reorder_app_case_ys A (xs: list A) ns ys1 ys2:
  list_reorder xs ns (ys1++ys2) ->
  exists ns1 ns2,
    ns = (ns1 ++ ns2)%list /\
    list_reorder xs ns1 ys1 /\
    list_reorder xs ns2 ys2.
Proof.
  introv Hre.
  reverts ns ys2.
  induction ys1; intros; simpls.
  - exists (nil%list: list nat) ns; simpls; splits; eauto.
    eapply list_reorder_nil.
  - inversions Hre.
    forwards: IHys1; eauto.
    destructc H as ns1 ns2 Hns0 Hre1 Hre2.
    exists (n::ns1)%list ns2; simpls; splits.
    rewrite <- Hns0; auto.
    constructor; auto.
    auto.
Qed.

Lemma list_reorder_app_case A (xs: list A) ns1 ns2 ys1 ys2:
  length ns1 = length ys1 ->
  list_reorder xs (ns1 ++ ns2)%list (ys1++ys2) ->
  list_reorder xs ns1 ys1 /\ list_reorder xs ns2 ys2.
Proof.
  introv Hl Hre.
  forwards: list_reorder_app_case_ys Hre.
  destructecs H.
  forwards Hl01: list_reorder_lengths H1.
  forwards He: list_app_inv_length1; eauto. rewrite Hl, Hl01. auto.
  destructecs He; subst.
  auto.
Qed.

Lemma list_reorder_take A m (xs: list A) ns ys:
  list_reorder xs ns ys ->
  list_reorder xs (LibList.take m ns) (LibList.take m ys).
Proof.
  introv Hre.
  reverts ns ys.
  induction m; intros.
  - rewrite ! LibList.take_zero. constructor.
  - destruct ns; inversions Hre.
    + rewrite ! LibList.take_nil. constructor.
    + rewrite ! LibList.take_succ. constructor; eauto.
Qed.
Lemma list_reorder_drop A m (xs: list A) ns ys:
  list_reorder xs ns ys ->
  list_reorder xs (LibList.drop m ns) (LibList.drop m ys).
Proof.
  introv Hre.
  reverts ns ys.
  induction m; intros.
  - rewrite ! LibList.drop_zero. auto.
  - destruct ns; inversions Hre.
    + rewrite ! LibList.drop_nil. constructor.
    + rewrite ! LibList.drop_succ. eauto.
Qed.
Lemma list_reorder_app_lr A (xs1 xs2: list A) ns ys:
  list_reorder xs1 ns ys ->
  list_reorder (xs1++xs2)%list ns ys.
Proof.
  introv Hre1.
  induction Hre1.
  - constructor.
  - constructor; eauto.
    rewrite <- liblist_app.
    eapply LibList.Nth_app_l; eauto.
Qed.
Lemma list_reorder_app_ll A (xs1 xs2: list A) ns ys:
  list_reorder xs2 ns ys ->
  list_reorder (xs1++xs2)%list (LibList.map (plus (length xs1)) ns) ys.
Proof.
  introv Hre2.
  induction Hre2.
  - constructor.
  - constructor; eauto.
    rewrite <- liblist_app.
    rewrite <- liblist_length.
    rewrite PeanoNat.Nat.add_comm.
    eapply LibList.Nth_app_r; eauto.
Qed.

(* If we have an almost permutation of a list xs1, where one element is missing, we can remove the element from the source list and get a permutation *)
Lemma list_reorder_remove_middle (A: Type) xs1 (x: A) xs2 ns ys:
  list_reorder (xs1 ++ x::xs2) ns ys ->
  ~ LibList.mem (length xs1) ns ->
  list_reorder (xs1 ++ xs2) (LibList.map (decrease_after (length xs1)) ns) ys.   (* The result list is found by decreasing all indices pointing after x *)
Proof.
  intros Hre Hmem.
  reverts ys.
  (* Induction on the index list *)
  induction ns; intros; subst.
  - inversions Hre. constructor.
  - (* Some index a added to ns *)
    inversions Hre.
    rewrite LibList.map_cons.
    (* reorded *)
    constructor.
    * eapply IHns; eauto.
    * unfold decrease_after.
      rewrite <- liblist_app in H4.
      eapply LibList.Nth_app_inv in H4.
      cases_if; destruct H4.
      -- eapply LibList.Nth_inbound in H. rewrite LibNat.lt_peano in H. rewrite liblist_length in H. eapply Lt.lt_not_le in H. contradiction.
      -- destructc H as ? ? ?. subst.
         destruct (classicT(x0=0)).
         ++ subst. false.  apply Hmem. rewrite PeanoNat.Nat.add_0_r. rewrite liblist_length; eauto.
         ++ eapply PeanoNat.Nat.neq_0_r in n. destructc n as ? ?. subst. rewrite PeanoNat.Nat.add_succ_r. rewrite PeanoNat.Nat.pred_succ. rewrite PeanoNat.Nat.add_comm.

          rewrite <- liblist_app.
          inversions H0.
          eapply LibList.Nth_app_r'; eauto.
      -- rewrite <- liblist_app. eapply LibList.Nth_app_l; eauto.
      -- false. destructc H as ? ? ?. subst. rewrite liblist_length in C. rewrite PeanoNat.Nat.lt_add_lt_sub_l in C. rewrite PeanoNat.Nat.sub_diag in C. eapply PeanoNat.Nat.nlt_0_r in C. auto.
Qed.
(* Same reordering as list_reorder_remove_middle, state a bit differently *)
Lemma list_reorder_remove_middle_Nth (A: Type) xs (x: A) n ns ys:
  list_reorder xs ns ys ->
  LibList.Nth n xs x ->
  ~ LibList.mem n ns ->
  list_reorder (list_removeat n xs) (LibList.map (decrease_after n) ns) ys.   (* The result list is found by decreasing all indices pointing after x *)
Proof.
  introv Hre Hn Hmem.
  assert (length (LibList.take n xs) = n). {
    rewrite <- liblist_length.
    rewrite LibList.length_take. auto.
    eapply LibOrder.lt_to_le. eapply LibList.Nth_inbound; eauto.
  }
  forwards: list_reorder_remove_middle.
  erewrite <- list_Nth_cut_take_drop.
  eauto.
  eauto.
  rewrite H; auto.
  rewrite H in H0. auto.
Qed.

Lemma list_reorder_from_list (A:Type) (xs: list A) ns ys:
  list_reorder xs ns ys ->
  LibList.noduplicates ns ->
  length xs = length ns ->
  (from_list xs) = (from_list ys).
Proof.
  intros.
  reverts xs ns.
  induction ys; intros.
  - inversions H.  destruct xs. auto. simpls. discriminate.
  - inversions H. inversions H0.
    destructc (list_Nth_cut H7) as xs1 xs2 Hxs Hlxs1. subst.
    forwards: list_reorder_remove_middle; eauto.
    forwards: IHys; eauto.
    eapply list_noduplicates_remove_middle; eauto.
    eapply list_length_remove_middle in H1. rewrite List.app_length. rewrite liblist_map. rewrite List.map_length. auto.
    rewrite from_list_append in *.
    rewrite ! from_list_cons in *.
    rewrite <- H0.
    rewrite union_assoc.
    rewrite (union_comm (from_list xs1)).
    rewrite union_assoc. auto.
Qed.

Lemma list_reorder_from_list_subset (A:Type) (xs: list A) ns ys:
  list_reorder xs ns ys ->
  (from_list ys) \c (from_list xs).
Proof.
  intros.
  reverts xs ns.
  induction ys; intros.
  - rewrite from_list_nil. eapply subset_empty_l.
  - inversions H. rewrite from_list_cons.
    intro. intro.
    rewrite in_union in H; destruct H.
    rewrite in_singleton in H; subst. eapply list_mem_in_fromlist. eapply LibList.Nth_mem; eauto.
    eapply IHys; eauto.
Qed.


Lemma list_reorder_select_middle (A: Type) xs1 (x: A) xs2 ns ys:
  list_reorder (xs1 ++ xs2) ns ys ->
  list_reorder (xs1 ++ x :: xs2) ((length xs1)::(LibList.map (increase_after (length xs1)) ns)) (x::ys).
Proof.
  introv Hre.
  remember (xs1++xs2)%list as xs.
  induction Hre; intros; subst.
  - rewrite LibList.map_nil.
    eapply list_reorder_nth.
    rewrite <- liblist_app.
    eapply LibList.Nth_app_r'; eauto.
    rewrite liblist_length. auto.
  - specialize (IHHre (eq_refl _)).
    inversions IHHre.
    constructor; eauto.
    rewrite LibList.map_cons.
    constructor; eauto.
    rewrite <- liblist_app in *.
    eapply LibList.Nth_app_inv in H.
    destruct H; unfold increase_after; cases_if.
    + eapply LibList.Nth_inbound in H. rewrite LibNat.lt_peano in *. rewrite liblist_length in *. eapply Lt.lt_not_le in H. contradiction.
    + eapply LibList.Nth_app_l; auto.
    + destructc H as m Hml Hm; subst.
      rewrite <- PeanoNat.Nat.add_succ_r.
      rewrite PeanoNat.Nat.add_comm.
      eapply LibList.Nth_app_r; eauto.
    + destructc H as m Hml Hm; subst.
      rewrite liblist_length in *.
      rewrite <- (PeanoNat.Nat.add_0_r (length xs1)) in C at 2.

      eapply Plus.plus_lt_reg_l in C. 
      eapply PeanoNat.Nat.nlt_0_r in C.
      contradiction.
Qed.

Lemma list_reorder_insert_middle (A: Type) (x: A) xs ns1 ns2 (ys1 ys2: list A):
  list_reorder xs ns1 ys1 ->
  list_reorder xs ns2 ys2 ->
  list_reorder (x :: xs) ((List.map S ns1) ++ 0 :: (List.map S ns2)) (ys1 ++ x :: ys2).
Proof.
  introv Hre1 Hre2.
  eapply list_reorder_app.
  eapply list_reorder_push0; auto.
  eapply list_reorder_push; auto.
Qed.

Lemma list_reorder_insert_middle_drop_take (A: Type) (x: A) xs ns (ys: list A) m:
  list_reorder xs ns ys ->
  list_reorder (x :: xs) (list_insertat m (List.map S ns) 0) (list_insertat m ys x).
Proof.
  introv Hre.
  rewrite <- ! liblist_map.
  rewrite <- list_map_drop.
  rewrite <- list_map_take.
  rewrite ! liblist_map.
  eapply list_reorder_insert_middle.

  eapply list_reorder_take; auto.
  eapply list_reorder_drop; auto.
Qed.

(* Induction proving a property of two lists permuted *)
Lemma list_permutation_ind(P: forall (zs1 ws1 zs2 ws2: list var), Prop):
  (* Nil case *)
  (P nil nil nil nil) ->
  (* Cons case *)
  (
    forall zsl zsr z zs2 wsl wsr w ws2,
    length wsl = length zsl ->
    length wsr = length zsr ->
    (P (zsl++zsr)%list (wsl++wsr)%list zs2 ws2) ->
    (P (zsl++z::zsr)%list (wsl++w::wsr)%list (z::zs2)%list (w::ws2)%list)
  ) ->
  (* Conclusion *)
  forall xs1 ys1 xs2 ys2 ns,
  list_reorder xs1 ns xs2 ->
  list_reorder ys1 ns ys2 ->
  LibList.noduplicates ns ->
  length xs1 = length ns ->
  length ys1 = length ns ->
  (P xs1 ys1 xs2 ys2).
Proof.
  introv Hnil Hcons Hrex Hrey Hnd Hlx Hly.
  reverts ys1 ys2 xs1 ns.
  induction xs2; intros.
  - (* Nil case *)
    inversions Hrex.
    destruct xs1; simpls; try discriminate.
    inversions Hrey.
    destruct ys1; simpls; try discriminate.
    auto.
  - (* Cons case *)
    inversions Hrex. inversions Hrey.
    inversions Hnd.
    destructc (list_Nth_cut H4) as ? ? ? ?. subst. clear H4.
    destructc (list_Nth_cut H6) as ? ? ? Hlx1. subst.
    eapply list_length_remove_middle in Hlx.
    eapply list_length_remove_middle in Hly.

    eapply Hcons.
    auto.
    rewrite Hlx1 in Hly. rewrite <- Hly in Hlx. apply -> PeanoNat.Nat.add_cancel_l in Hlx. auto.

    forwards Hre0: list_reorder_remove_middle H3; eauto.
    forwards Hre1: list_reorder_remove_middle H2; eauto. rewrite Hlx1. auto.

    rewrite Hlx1 in *.
    eapply IHxs2; eauto.
    + eapply list_noduplicates_remove_middle; eauto.
    + rewrite List.app_length.
    rewrite liblist_map.
    rewrite List.map_length. auto.
    + rewrite List.app_length.
    rewrite liblist_map.
    rewrite List.map_length. rewrite Hlx1. auto.
Qed.


Lemma list_permuation_noduplicates A (ns1: list A) (ns2: list nat) ns3:
  list_reorder ns1 ns2 ns3 ->
  LibList.noduplicates ns1 ->
  LibList.noduplicates ns2 ->  
  LibList.noduplicates ns3.
Proof.
  intros.
  reverts ns1  ns3.
  induction H1; intros.
  - inversions H. constructor.
  - inversions H2. constructor.
    + intro.
      eapply LibList.mem_Nth in H2. destruct H2.
      destructc (list_reorder_nth_Nth H6 H2) as z Hz1 Hz2.
      pose proof (LibList.noduplicates_Nth_same_inv H0 Hz2 H8).
      subst.
      apply H. eapply LibList.Nth_mem; eauto.
    + eapply IHnoduplicates; eauto.
Qed.

Lemma list_reorder_trans (A: Type) ns1 ns2 ns3 (xs1 xs2 xs3: list A):
  list_reorder xs1 ns1 xs2 ->
  list_reorder xs2 ns2 xs3 ->
  list_reorder ns1 ns2 ns3 ->
  list_reorder xs1 ns3 xs3.
Proof.
  introv Hr1 Hr2 Hr3.
  reverts xs1 xs2 xs3;
  reverts ns1 ns3.
  induction ns2; intros.
  - inversions Hr2. inversions Hr3. constructor.
  - inversions Hr2. inversions Hr3.
    constructor.

    forwards: IHns2;  eauto.

    lets Hm: (list_reorder_nth_Nth) Hr1 H4.
    destructc Hm as m Hm1 Hm2.
    lets Hym: LibList.Nth_functional H6 Hm1.
    subst. auto.
Qed.

Lemma list_reorder_trans_exists (A: Type) ns1 ns2 (xs1 xs2 xs3: list A):
  list_reorder xs1 ns1 xs2 ->
  list_reorder xs2 ns2 xs3 ->
  exists ns3,
  (list_reorder xs1 ns3 xs3).
Proof.
  intros.
  forwards: list_reorder_lengths H.
  forwards: list_reorder_same_length_exists ns1 H1; eauto.
  destructc H2 as ws Hws.
  exists ws.
  eapply list_reorder_trans; eauto.
Qed.

Lemma list_permuation_trans (A: Type) ns1 ns2 (xs1 xs2 xs3: list A):
  list_reorder xs1 ns1 xs2 ->
  LibList.noduplicates ns1 ->
  length xs1 = length ns1 ->
  list_reorder xs2 ns2 xs3 ->
  LibList.noduplicates ns2 ->
  length xs2 = length ns2 ->
  exists ns3,
  (list_reorder xs1 ns3 xs3 /\ LibList.noduplicates ns3).
Proof.
  intros.
  forwards: list_reorder_same_length_exists ns1 H2. eapply list_reorder_lengths; eauto.
  destructc H5 as ws Hws.
  exists ws. splits.
  - eapply list_reorder_trans; eauto.
  - eapply list_permuation_noduplicates; eauto.
Qed.

Lemma list_reorder_gt (A: Type) (xs: list A) ns ys:
  list_reorder xs ns ys ->
  LibList.Forall (gt (length xs)) ns.
Proof.
  introv Hre.
  induction Hre.
  - constructor.
  - constructor.
    eapply LibList.Nth_inbound in H.
    rewrite liblist_length in H. rewrite LibNat.lt_peano in H. auto. auto.
Qed.

(* Existence of reordering index lists *)

(* The variants with _n are more general and allow using the same index list to reorder multiple lists.
   The additional conclusions about length and Forall are not strictly necessary,
   because they follow from reorder, but having them before the forall xs, allows using them before having a list of values
 *)

Ltac gt_solve := (unfold gt in *; eauto using le_0_n, le_n_S).

Lemma list_forall_gt_map_S n ns:
  LibList.Forall (gt n) ns ->
  LibList.Forall (gt (S n)) (List.map S ns).
Proof.
  induction ns.
  * simpls. constructor.
  * simpls. constructor; inversions H.  gt_solve.
    eapply IHns; eauto.
Qed.
Lemma list_forall_gt_map_plus n m ns:
  LibList.Forall (gt n) ns ->
  LibList.Forall (gt (m + n)) (List.map (plus m) ns).
Proof.
  induction ns.
  * simpls. constructor.
  * simpls. constructor; inversions H. 
    eapply Gt.plus_gt_compat_l; eauto.
    eapply IHns; eauto.
Qed.


Ltac gt_solve_map := (unfold gt,lt in *; eauto using le_0_n, le_n_S, LibList.Forall_cons, list_forall_gt_map_S).

(* Identity reordering *)
Lemma list_reorder_refl_n n:
  exists ns,
  LibList.noduplicates ns /\
  length ns = n /\
  LibList.Forall (gt n) ns /\
  forall A (xs: list A),
  (length xs) = n ->
  list_reorder xs ns xs.
Proof.
  induction n.
  - exists (@nil nat).
    splits.
    + constructor.
    + simpls. auto.
    + constructor.
    + intros. destruct xs; try discriminate.
      constructor.
  - destructc IHn as ns Hnd Hl Hf Hns.
    exists (0::(List.map S ns))%list.
    splits.
    +
      eapply list_noduplicates_shift. auto.
    +
      simpls.
      rewrite List.map_length. rewrite Hl. auto.
    + gt_solve_map.
    +
      intros. destruct xs; try discriminate.
      simpls.
      forwards: (Hns A xs). eauto.
      eapply list_reorder_push. auto.
Qed.

Lemma list_reorder_refl A (xs: list A):
  exists ns, list_reorder xs ns xs /\ LibList.noduplicates ns.
Proof.
  forwards ns Hnd Hl Hf Hre: (list_reorder_refl_n); eauto.
Qed.

(* Swapping first two items *)
Lemma list_reorder_swap_n n:
  exists ns,
  LibList.noduplicates ns /\
  length ns = S (S n) /\
  LibList.Forall (gt (S (S n))) ns /\
  forall A (a x: A) xs,
  length xs = n ->
  list_reorder (a :: x :: xs)%list ns (x :: a :: xs)%list.
Proof.
  destructc (list_reorder_refl_n n) as ns Hnd Hl Hf Hns.
  exists (1::0::(List.map S (List.map S ns)))%list.
  splits.
  - constructor.
    intro. eapply LibList.mem_cons_inv in H. destruct H; try discriminate.
    eapply list_mem_succ_inv in H.
    eapply list_mem_0_map_succ_false; eauto.
    eapply list_noduplicates_shift. eapply list_noduplicates_succ. auto.  
  - simpls. rewrite ! List.map_length. rewrite Hl. auto.
  - constructor; gt_solve_map.
  - 
    intros.
    forwards: (Hns A xs). eauto.
    constructor.
    eapply list_reorder_push; eauto.
    eapply list_reorder_push0; eauto.
    constructor.   constructor.
Qed.

Lemma list_reorder_swap A (a x: A) xs:
  exists ns, list_reorder (a :: x :: xs)%list ns (x :: a :: xs)%list /\ LibList.noduplicates ns.
Proof.
  forwards ns Hnd Hl Hf Hre: list_reorder_swap_n; eauto.
Qed.

(* Move an item from front to inside *)
Lemma list_reorder_shift_app (A: Type) x (xs ys: list A):
  exists ns,
  (list_reorder (x::(xs++ys)) ns (xs++x::ys) /\ LibList.noduplicates ns).
Proof.
  induction xs.
  - simpls. eapply list_reorder_refl.
  - simpls. destructc IHxs as ns Hr Hnd.
    eapply list_reorder_push with (x:=a) in Hr.
    forwards ns0 Hns0 Hnd0: (list_reorder_swap x a (xs++ys)).
    eapply list_permuation_trans; eauto.
    + eapply list_reorder_lengths in Hns0. rewrite <- Hns0. simpls. auto.
    + eapply list_noduplicates_shift; auto.
    + eapply list_reorder_lengths in Hr. rewrite <- Hr. simpls. rewrite ! List.app_length. fequal. simpls. rewrite plus_n_Sm. auto.
Qed.

Lemma list_reorder_shift_app_n m n:
  exists ns,
  LibList.noduplicates ns /\
  length ns = S (m + n) /\
  LibList.Forall (gt (S (m + n))) ns /\
  forall (A: Type) x (xs ys: list A),
  length xs = m ->
  length ys = n ->
  list_reorder (x::(xs++ys)) ns (xs++x::ys).
Proof.
  induction m.
  - forwards ns Hnd Hl Hf Hre: (list_reorder_refl_n (S n)).
    exists ns.
    splits.
    + auto.
    + simpls. auto.
    + simpls. auto.
    +
      intros.
      destruct xs; simpls; try discriminate. eapply Hre. simpls. auto.
  - (* Reorder the rest of the list *)
    destructc IHm as ns Hnd Hl Hf Hre.
    (* Separate the first element of the index list *)
    destruct ns; simpls; try discriminate.

    forwards ns0 Hnd0 Hl0 Hf0 Hre0: (list_reorder_swap_n (m+n)).
    destruct ns0; simpls; try discriminate.
    destruct ns0; simpls; try discriminate.
    inversions Hl0.

    assert (LibOrder.le m (LibList.length ns0)) as Hlen. {
      rewrite liblist_length.
      rewrite LibNat.le_peano.
      rewrite H0. eapply Plus.le_plus_l.
    }
    forwards Hns0app: LibList.list_eq_take_app_drop Hlen.
    rewrite liblist_app in Hns0app.

    assert(length (LibList.take m ns0) = m). {
      rewrite <- liblist_length.
      eapply LibList.length_take. eauto.
    }
    assert(length (LibList.drop m ns0) = n). {
      replace n with ((length ns0) - m).
      rewrite <- liblist_length.
      eapply LibList.length_drop. eauto.
      rewrite H0. eapply Minus.minus_plus.
    }

    forwards Hns0re: Hre n2 (LibList.take m ns0) (LibList.drop m ns0); eauto.

    rewrite Hns0app in Hns0re.

    exists (n1 :: LibList.take m ns0 ++ n2 :: LibList.drop m ns0)%list.

    splits.
    + eapply list_reorder_push with (x:=n1) in Hns0re.
      eapply list_permuation_noduplicates; eauto.
      eapply list_noduplicates_shift; eauto.
    + simpls. fequal. rewrite List.app_length. simpl. rewrite PeanoNat.Nat.add_succ_r. fequal.
      rewrite H0, H, H1; auto.
    + (* Forall *)
      rewrite <- liblist_app.
      eapply LibList.Forall_cons_inv in Hf0. destructc Hf0 as Hf0n1 Hf0.
      eapply LibList.Forall_cons_inv in Hf0. destructc Hf0 as Hf0n2 Hf0.
      forwards Hf0t: LibList.Forall_take m Hf0.
      forwards Hf0d: LibList.Forall_drop m Hf0.
      rewrite H0.
      eapply LibList.Forall_cons. auto.
      eapply LibList.Forall_app. auto.
      eapply LibList.Forall_cons; auto.
    + (* reorder *)
      intros.
      destruct xs; simpls; try discriminate.
      inversions H2.

      forwards Hre0xs: Hre0 x a (xs++ys)%list.
      rewrite List.app_length; auto.

      forwards Hrexs: Hre x xs ys; eauto.
      eapply list_reorder_push with (x:=a) in Hrexs.

      eapply list_reorder_trans; eauto using list_reorder_push.
Qed.

Lemma list_reorder_shift (A: Type) x (xs: list A):
  exists ns,
  (list_reorder (x::xs) ns (xs++x::nil) /\ LibList.noduplicates ns).
Proof.
  replace (x :: xs)%list with (x :: (xs++nil))%list.
  eapply list_reorder_shift_app. rewrite List.app_comm_cons. rewrite List.app_nil_r.
  auto.
Qed.

Lemma list_all_gt_decrease m n ns:
  LibList.Forall (gt (S m)) ns ->
  (S m) > n ->
  ~ LibList.mem n ns ->
  LibList.Forall (gt m) (LibList.map (decrease_after n) ns).
Proof.
  induction ns.
  - rewrite LibList.map_nil. constructor.
  - (* cons *)
    rewrite LibList.map_cons. constructor.
    +
      inversions H.
      eapply gt_decrease_after_from; eauto.
      intro; subst. eapply H1. eapply LibList.mem_cons_l.
    +
      inversions H.
      eauto.
Qed.

Lemma list_reorder_inverse (ns: list nat):
  LibList.noduplicates ns ->
  LibList.Forall (gt (length ns)) ns ->
  exists ns2,
  LibList.noduplicates ns2 /\
  length ns2 = length ns /\
  LibList.Forall (gt (length ns2)) ns2 /\
  forall (A: Type) (xs ys: list A),
  length xs = length ns ->
  list_reorder xs ns ys ->
  list_reorder ys ns2 xs.
Proof.
  introv Hnd Hf.
  remember (length ns) as n.
  reverts ns.
  induction n; intros.
  - destruct ns; simpls; try discriminate.
    exists (@nil nat).
    splits.
    + constructor.
    + simpls. auto.
    + constructor.
    + intros. destruct xs; simpls; try discriminate. constructor.
  - destruct ns; simpls; try discriminate.
    inversions Heqn.
    inversions Hnd.
    inversions Hf.
    forwards: (IHn (LibList.map (decrease_after n0) ns)).
    + eapply list_noduplicates_remove_middle; eauto.
    + rewrite liblist_map. rewrite List.map_length. auto.
    + eapply list_all_gt_decrease; eauto.
    +  clear IHn.
      destructc H as ns2 Hnd2 Hl2 Hf2 Hre2.

      assert (LibOrder.le n0 (LibList.length (LibList.map S ns2))) as Hlen. {
        rewrite LibList.length_map.
        rewrite liblist_length.
        rewrite LibNat.le_peano.
        unfold gt, lt in H3.
        rewrite Hl2. eapply le_S_n; eauto.
      }
      forwards Hns2app: LibList.list_eq_take_app_drop Hlen.
      set (ns3:=((LibList.take n0 (LibList.map S ns2)) ++ 0::(LibList.drop n0 (LibList.map S ns2)))%list).
      assert (length ns3 = S (length ns2)) as Hlen3. {
        unfold ns3.
        rewrite List.app_length. simpls.
        rewrite PeanoNat.Nat.add_succ_r.
        fequal.
        rewrite <- List.app_length.
        rewrite <- liblist_app.
        rewrite Hns2app.
        rewrite liblist_map.
        rewrite List.map_length. auto.
      }

      exists ns3.
      splits.
      * edestruct list_reorder_shift_app.
        destruct H.
        eapply list_permuation_noduplicates. eapply H.
        rewrite <- liblist_app. rewrite Hns2app.
        rewrite liblist_map.
        eapply list_noduplicates_shift; eauto.
        eauto.
      * rewrite <- Hl2. auto.
      * rewrite Hlen3. unfold ns3.
        rewrite <- liblist_app.
        eapply LibList.Forall_app.
        eapply LibList.Forall_take. rewrite liblist_map. gt_solve_map.
        constructor.
        eapply Gt.gt_Sn_O.
        eapply LibList.Forall_drop. rewrite liblist_map. gt_solve_map.
      * intros. unfold ns3 in *.

        destruct ys; simpls; inversions H0.

        eapply list_Nth_cut in H11. destructc H11 as xs1 xs2 Hxs Hlxs.
        subst.

        replace xs1 with (LibList.take (length xs1) (xs1++xs2)) at 3.
        replace xs2 with (LibList.drop (length xs1) (xs1++xs2)) at 2.
        rewrite liblist_map.
        eapply list_reorder_insert_middle_drop_take; eauto.

        rewrite <- liblist_app.
        eapply Hre2.
        rewrite liblist_app.
        rewrite List.app_length in H.
        rewrite List.app_length.
        simpls. rewrite PeanoNat.Nat.add_succ_r in H. inversion H. auto.

        rewrite liblist_app.

        eapply list_reorder_remove_middle; eauto.

        rewrite <- liblist_app.
        rewrite <- liblist_length.
        rewrite LibList.drop_app_length. auto.

        rewrite <- liblist_app.
        rewrite <- liblist_length.
        rewrite LibList.take_prefix_length. auto.
Qed.

Lemma list_reorder_inverse_for A (ns: list nat) (xs ys: list A):
  LibList.noduplicates ns ->
  length xs = length ns ->
  list_reorder xs ns ys ->
  exists ns2,
  list_reorder ys ns2 xs /\
  LibList.noduplicates ns2 /\
  length ns2 = length ns.
Proof.
  introv Hns Hl Hre.
  forwards: list_reorder_inverse; eauto.
  rewrite <- Hl.
  eapply list_reorder_gt; eauto.
  destructc H as ns2 Hnd2 Hl2 Hgt2 Hf; eauto.
Qed.

Lemma list_reorder_mem A (xs0: list A) ns xs x:
  list_reorder xs0 ns xs ->
  LibList.mem x xs ->
  LibList.mem x xs0.
Proof.
  introv Hro Hmem.
  reverts ns.
  induction Hmem; intros; simpls. inversions Hro. eapply LibList.Nth_mem in H4. auto.
  inversions Hro.
  eapply IHHmem. eauto. 
Qed.
Lemma list_reorder_notin A (xs0: list A) ns xs x:
  list_reorder xs0 ns xs ->
  x \notin (from_list xs0) ->
  x \notin (from_list xs).
Proof.
  introv Hro Hni.
  intro. apply Hni.
  apply list_mem_in_fromlist.
  eapply list_reorder_mem. eauto.
  eapply list_in_fromlist_mem. auto.
Qed.
Lemma list_reorder_disjoint A (xs1: list A) ns1 ys1 xs2 ns2 ys2:
  list_reorder xs1 ns1 ys1 ->
  list_reorder xs2 ns2 ys2 ->
  disjoint (from_list xs1) (from_list xs2) ->
  disjoint (from_list ys1) (from_list ys2).
Proof.
  introv Hre1 Hre2 Hdi.
  eapply disjoint_subset_r.
  eapply disjoint_subset_l.
  eauto.
  eapply list_reorder_from_list_subset. eauto.
  eapply list_reorder_from_list_subset. eauto.
Qed.



Lemma list_align_common (A: Type) (xs ys: list A):
  LibList.noduplicates xs ->
  LibList.noduplicates ys ->
  exists (zs xzs yzs: list A)(xns yns: list nat),
  list_reorder xs xns (zs++xzs) /\
  list_reorder ys yns (zs++yzs) /\
  LibList.noduplicates xns /\
  LibList.noduplicates yns /\
  LibList.noduplicates (xzs ++ yzs) /\
  length xns = length xs /\
  length yns = length ys.
Proof.
  introv Hndx Hndy.
  reverts ys.
  induction Hndx; intros.
  - destructc (list_reorder_refl ys) as yns Hrey Hndys.
    exists (@nil A) (@nil A) ys (@nil nat) yns.
    simpls; splits; eauto.
    eapply list_reorder_nil.
    eapply LibList.noduplicates_nil.
    eapply list_reorder_lengths in Hrey; eauto.
  - destruct (classicT(LibList.mem x ys)).
    + (* Common element *)
      eapply LibList.mem_Nth in m.
      destructc m as n Hn.
      eapply list_Nth_cut in Hn.
      destructc Hn as ys1 ys2 Hys Hlys1. subst.
      forwards: (IHHndx (ys1 ++ ys2)%list).
      eapply list_noduplicates_concat_middle_inv; eauto.
      destructc H0 as zs1 xzs1 yzs1 xns1 yns1 Hrex1 Hrey1 Hndx1 Hndy1 Hnd1 Hlxyns. destructc Hlxyns as Hlxns Hlyns.
      exists (x::zs1)%list xzs1 yzs1 (0::(List.map S xns1))%list ((length ys1)::(List.map (increase_after (length ys1)) yns1 ))%list.
      splits; eauto.
      * rewrite <- List.app_comm_cons.
        eapply list_reorder_push; auto.
      * rewrite <- List.app_comm_cons. rewrite <- liblist_map. eapply list_reorder_select_middle; eauto.
      * eapply list_noduplicates_shift; auto.
      * rewrite <- liblist_map. eapply list_noduplicates_select_middle; eauto.
      * simpls. fequal. rewrite List.map_length. auto.
      * simpls. rewrite List.app_length. simpl. rewrite PeanoNat.Nat.add_succ_r. fequal. rewrite List.map_length.  rewrite List.app_length in Hlyns. auto.
    + (* Element only in xs *)
      forwards: (IHHndx (ys)%list); auto.
      destructc H0 as zs1 xzs1 yzs1 xns1 yns1 Hrex1 Hrey1 Hndx1 Hndy1 Hnd1 Hlxyns. destructc Hlxyns as Hlxns Hlyns.
      exists zs1 (x::xzs1)%list yzs1 ((LibList.take (length zs1) (List.map S xns1)) ++ 0::(LibList.drop (length zs1) (List.map S xns1)))%list yns1.
      splits; eauto.
      *
        replace zs1 with (LibList.take (length zs1) (zs1++xzs1)) at 3.
        replace xzs1 with (LibList.drop (length zs1) (zs1++xzs1)) at 2.
        eapply list_reorder_insert_middle_drop_take; eauto.

        rewrite <- liblist_app.
        rewrite <- liblist_length.
        rewrite LibList.drop_app_length. auto.

        rewrite <- liblist_app.
        rewrite <- liblist_length.
        rewrite LibList.take_prefix_length. auto.
      *
        rewrite <- liblist_app.
        eapply LibList.noduplicates_app.
        eapply list_noduplicates_take. eapply list_noduplicates_succ; eauto.
        rewrite <- liblist_map.
        rewrite <- list_map_drop.
        rewrite liblist_map.
        eapply list_noduplicates_shift.
        eapply list_noduplicates_drop. auto.
        intros.
        rewrite <- (LibList.list_eq_take_app_drop) with (n:= (LibList.length zs1)) in Hndx1.
        eapply LibList.noduplicates_app_inv in  Hndx1.
        destructc Hndx1 as Hndx1t Hndx1d Hndx1m.
        apply Hndx1m.
        rewrite <- liblist_map in H1, H0.
        rewrite <- list_map_take in H0.
        rewrite <- list_map_drop in H1.
        eapply LibList.mem_cons_inv in H1.
        destruct x0.
        ** subst. false. eapply list_mem_0_map_succ_false; eauto. rewrite <- liblist_map. apply H0.
        ** destruct H1; try discriminate.
           exists x0. splits.
           eapply list_mem_succ_inv; rewrite <- liblist_map; rewrite liblist_length; eauto.
           eapply list_mem_succ_inv; rewrite <- liblist_map; rewrite liblist_length; eauto.
        ** eapply list_reorder_lengths in Hrex1. rewrite List.app_length in Hrex1. rewrite ! liblist_length. rewrite LibNat.le_peano. rewrite <- Hrex1.
           eapply Plus.le_plus_l.
      * rewrite <- List.app_comm_cons. constructor. rewrite <- liblist_app. intro. eapply LibList.mem_app_inv in H0. destruct H0.
        ** apply H. eapply list_reorder_mem; eauto. rewrite <- liblist_app. eapply LibList.mem_app_r; auto.
        ** apply n. eapply list_reorder_mem; eauto. rewrite <- liblist_app. eapply LibList.mem_app_r; auto.
        ** auto.
      * simpls. rewrite List.app_length. simpl. rewrite PeanoNat.Nat.add_succ_r. fequal. rewrite <- List.app_length. rewrite <- liblist_app. rewrite LibList.list_eq_take_app_drop.  rewrite List.map_length. auto.
        rewrite LibNat.le_peano. rewrite liblist_length. rewrite List.map_length. eapply list_reorder_lengths in Hrex1. rewrite <- Hrex1. rewrite List.app_length. eapply Plus.le_plus_l.
Qed.

(* Lemma useful after using list_align_common *)
Lemma list_align_common_subset X (A B: fset X) xs1 xs2:
  A \c B \u from_list xs1 ->
  A \c B \u from_list xs2 ->
  LibList.noduplicates (xs1 ++ xs2) ->
  A \c B.
Proof.
  introv Hs1 Hs2 Hnd.
  intro. intro.
  specialize (Hs1 x H).
  specialize (Hs2 x H).
  rewrite in_union in *.
  destruct Hs1; destruct Hs2; auto.
  false.
  eapply list_in_fromlist_mem in H0.
  eapply list_in_fromlist_mem in H1.
  rewrite <- liblist_app in Hnd.
  eapply LibList.noduplicates_app_inv in Hnd.
  destructc Hnd as ? ? ?.
  apply H4.
  exists x; auto.
Qed.

Lemma list_reorder_swap_app_n m n:
  exists ns,
  LibList.noduplicates ns /\
  length ns = m + n /\
  LibList.Forall (gt (m + n)) ns /\
  forall A (xs1 xs2: list A),
  length xs1 = m ->
  length xs2 = n ->
  list_reorder (xs1 ++ xs2)%list ns (xs2 ++ xs1)%list.
Proof.
  destructc (list_reorder_refl_n m) as ns1 Hnd1 Hl1 Hf1 Hns1.
  destructc (list_reorder_refl_n n) as ns2 Hnd2 Hl2 Hf2 Hns2.
  exists (((List.map (plus m)) ns2)++ns1)%list.
  splits.
  - rewrite <- liblist_app.
    eapply LibList.noduplicates_app.
    rewrite <- liblist_map.
    eapply list_noduplicates_add; eauto.
    eauto.
    intros.
    forwards Hgt: LibList.Forall_mem_inv Hf1 H0.
    rewrite <- liblist_map in H.
    forwards: liblist_mem_map_inj_exists H.
    intros. apply -> PeanoNat.Nat.add_cancel_l; eauto.
    forwards x0 Hx0 Hmemx0: H1. subst.
    eapply gt_plus_l_false; eauto.
  - rewrite List.app_length.
    rewrite List.map_length.
    rewrite Hl1, Hl2. eapply PeanoNat.Nat.add_comm.
  - rewrite <- liblist_app.
    rewrite <- liblist_map.
    eapply LibList.Forall_app.
    +
      rewrite liblist_map.
      eapply list_forall_gt_map_plus; eauto.
    +
    eapply LibList.Forall_pred_incl. eapply Hf1.
    unfold pred_incl; intros. eapply gt_plus_lr; eauto.
  - intros.
    forwards: Hns1; eauto.
    forwards: Hns2; eauto.
    eapply list_reorder_app.
    rewrite <- liblist_map.
    rewrite <- H.
    eapply list_reorder_app_ll; eauto.
    eapply list_reorder_app_lr; eauto.
Qed.

Lemma list_reorder_swap_app A (xs1 xs2: list A):
  exists ns,
  LibList.noduplicates ns /\
  length ns = (length (xs1 ++ xs2)) /\
  LibList.Forall (gt (length (xs1 ++ xs2))) ns /\
  list_reorder (xs1 ++ xs2)%list ns (xs2 ++ xs1)%list.
Proof.
  forwards ns Hnd Hl Hf Hre: (list_reorder_swap_app_n (length xs1) (length xs2)); eauto.
  rewrite List.app_length. exists ns; split; eauto.
Qed.

Lemma list_forall_reorder A P (xs ys: list A) (ns: list nat):
  list_reorder xs ns ys ->
  List.Forall P xs ->
  List.Forall P ys.
Proof.
  introv Hre Hf.
  induction Hre.
  eauto.
  constructor.
  eapply list_forall_nth; eauto.
  eauto.
Qed.
Lemma list_forall_reorder_inv A P (xs ys: list A) (ns: list nat):
  list_reorder xs ns ys ->
  length xs = length ns ->
  LibList.noduplicates ns ->
  List.Forall P ys ->
  List.Forall P xs.
Proof.
  introv Hre Hnd Hf.
  forwards: list_reorder_inverse_for; eauto.
  destructc H as ns2 Hre2 Hnd2 Hl2.
  eapply list_forall_reorder; eauto.
Qed.

Lemma list_reorder_exists A (xs: list A) ns:
  LibList.Forall (gt (length (xs))) ns ->
  exists ys,
    list_reorder xs ns ys.
Proof.
  introv Hf.
  induction Hf.
  exists ((nil):(list A)). constructor.
  destruct IHHf.

  forwards: LibList.Nth_inbound_inv.
  rewrite LibNat.lt_peano.
  rewrite liblist_length.
  unfold gt in H. eauto.
  destruct H1.
  exists (x1::x0)%list.
  constructor; eauto.
Qed.

Lemma list_reorder_move_n m n k:
  exists ns,
  LibList.noduplicates ns /\
  length ns = S (m + n + k) /\
  LibList.Forall (gt (S (m + n + k))) ns /\
  forall (A: Type) x (xs ys zs: list A),
  length xs = m ->
  length ys = n ->
  length zs = k ->
  list_reorder (xs++x::ys++zs) ns (xs++ys++x::zs).
Proof.
  forwards: list_reorder_shift_app_n (m+n) k.
  forwards: list_reorder_shift_app_n m (n+k).
  destructc H as ns1 Hnd1 Hl1 Hf1 Hr1.
  destructc H0 as ns2 Hnd2 Hl2 Hf2 Hr2.
  forwards: list_reorder_inverse ns2. eauto. rewrite Hl2. eauto.
  destructc H as ns3 Hnd3 Hl3 Hf3 Hr3.
  forwards: list_reorder_exists ns3 ns1. rewrite Hl3, Hl2. rewrite PeanoNat.Nat.add_assoc. auto.
  destructc H as ns4 Hns4.
  forwards Hl4: list_reorder_lengths Hns4.
  
  exists ns4.
  splits.
  -
  eapply list_permuation_noduplicates; eauto.
  -
  rewrite Hl4, Hl1.  auto.
  -
  rewrite liblist_forall.
  eapply list_forall_reorder; eauto.
  rewrite <- liblist_forall. rewrite Hl3, Hl2 in Hf3. rewrite PeanoNat.Nat.add_assoc in Hf3.  auto.
  -
  intros; subst.
  eapply list_reorder_trans. 3: apply Hns4.
  eapply Hr3 with (xs := (x:: xs ++ ys ++ zs)%list).
  rewrite Hl2. simpls. rewrite ! List.app_length. rewrite PeanoNat.Nat.add_assoc. auto.
  + eapply Hr2. auto. rewrite List.app_length; eauto.
  + rewrite ! List.app_assoc. eapply Hr1. rewrite List.app_length; eauto. auto.
Qed.
Lemma list_reorder_move_inv_n m n k:
  exists ns,
  LibList.noduplicates ns /\
  length ns = S (m + n + k) /\
  LibList.Forall (gt (S (m + n + k))) ns /\
  forall (A: Type) x (xs ys zs: list A),
  length xs = m ->
  length ys = n ->
  length zs = k ->
  list_reorder (xs++ys++x::zs) ns (xs++x::ys++zs).
Proof.
  forwards: list_reorder_move_n m n k.
  destructc H as ns Hnd Hl Hf Hre.
  forwards: list_reorder_inverse ns. eauto. rewrite Hl. eauto.
  destructc H as ns3 Hnd3 Hl3 Hf3 Hr3.
  exists ns3; splits; eauto.
  rewrite Hl3, Hl. auto.
  rewrite Hl3, Hl in Hf3. auto.
  intros; subst.
  eapply Hr3.
  rewrite Hl. rewrite ! List.app_length. simpl. rewrite PeanoNat.Nat.add_succ_r.  rewrite ! List.app_length. rewrite PeanoNat.Nat.add_assoc. auto.
  eapply Hre; auto.
Qed.

Lemma list_reorder_map A B (f: A->B) (xs: list A) ns ys:
  list_reorder xs ns ys ->
  list_reorder (List.map f xs) ns (List.map f ys).
Proof.
  introv Hre.
  induction Hre.
  - simpls. eapply list_reorder_nil.
  - simpls. eapply list_reorder_cons; eauto.
    rewrite <- liblist_map.
    eapply LibList.Nth_map; eauto.
Qed.

Lemma list_reorder_split_subset (A: Type) (xs: list A) (X Y: fset A):
  LibList.noduplicates xs ->
  (from_list xs \c (X \u Y)) ->
  exists ns xs1 xs2,
  LibList.noduplicates ns /\
  (from_list xs1 \c X) /\
  (from_list xs2 \c Y) /\
  list_reorder xs ns (xs1 ++ xs2).
Proof.
  introv Hndxs Hs.
  induction xs.
  - exists (@nil nat) (@nil A) (@nil A); splits; rewrite ? from_list_nil; try apply subset_empty_l.
    eapply LibList.noduplicates_nil.
    eapply list_reorder_nil.
  - inversions Hndxs.
    rewrite from_list_cons in Hs.
    forwards: subset_union_l1 Hs.
    forwards: subset_union_l2 Hs.
    rewrite subset_singleton, in_union in H.

    forwards Hr: IHxs H2 H0.
    destructecs Hr.

    destruct H.
    + (* a in X -> push to x1 *)
      eexists. exists (a::xs1)%list xs2.
      splits.
      4: rewrite <- List.app_comm_cons.
      (* Pushing to the front *)
      4: eapply list_reorder_push. 4: eapply Hr.
      eapply list_noduplicates_shift; auto.
      rewrite from_list_cons. apply subset_union_l. rewrite subset_singleton. auto.
      auto. auto.
    + (* a in Y -> push to x2 *)
      forwards Hra: list_reorder_app_case_ys Hr.
      destructecs Hra; subst.
    
      eexists. exists xs1 (a::xs2)%list.
      splits.
      (* Insert to the middle *)
      4: eapply list_reorder_insert_middle.
      4: eauto. 4: eauto.
      eapply list_noduplicates_insert_middle. auto.
      auto.
      rewrite from_list_cons. apply subset_union_l. rewrite subset_singleton. auto.
      auto.
Qed.

Lemma list_reorder_in_set (A: Type) (xs: list A) (X: fset A):
  LibList.noduplicates xs ->
  exists ns ys,
  list_reorder xs ns ys /\
  LibList.noduplicates ns /\
  (from_list ys = (X \n (from_list xs))).
Proof.
  introv Hndxs.
  induction xs.
  -
    exists (@nil nat) (@nil A); splits.
    eapply list_reorder_nil.
    eapply LibList.noduplicates_nil.
    rewrite inter_empty_r; auto.
  - inversions Hndxs.

    forwards Hr: IHxs H2.
    destructecs Hr.

    destruct (classicT (a \in X)).
    + (* a in X -> push to x1 *)
      exists (0::(List.map S ns))%list (a::ys)%list.
      splits.
      *
      (* Pushing to the front *)
      eapply list_reorder_push. eapply Hr0.
      *
      eapply list_noduplicates_shift; auto.
      *
      rewrite ! from_list_cons, Hr. rewrite inter_union_r. rewrite in_inter_singleton by auto. auto.
    + (* a in Y -> push to x2 *)
      exists (List.map S ns) ys.
      splits.
      *
      eapply list_reorder_push0. eapply Hr0.
      *
      eapply list_noduplicates_succ; auto.
      *
      rewrite ! from_list_cons, Hr. rewrite inter_union_r. rewrite inter_comm with (F:=\{a}). rewrite inter_singleton_notin by auto. rewrite union_empty_l. auto.
Qed.

Lemma list_permutation_mem A (S1 S2: list A) ns x:
  list_reorder S1 ns S2 ->
  LibList.noduplicates ns ->
  length S1 = length S2 ->  
  LibList.mem x S1 ->
  LibList.mem x S2.
Proof.
  introv Hre Hnd Hl Hbi.
  forwards Hlns: list_reorder_lengths Hre.

  forwards Hrei: list_reorder_inverse_for Hnd Hre. rewrite Hl. auto.
  destructecs Hrei.
  
  eapply list_reorder_mem; eauto.
Qed.

(* If the index-list contains all numbers, then the result list contains all the items, and equality of result lists means equality of source lists *)
Definition list_cover_interval (ns: list nat) (n: nat) :=
  forall m, m < n -> LibList.mem m ns.
Lemma list_cover_interval_0 ns:
  list_cover_interval ns 0.
Proof.
  unfold list_cover_interval.
  intros. apply PeanoNat.Nat.nlt_0_r in H. false H.
Qed.
Lemma list_cover_interval_nil_inv n:
  list_cover_interval nil n ->
  n = 0.
Proof.
  unfold list_cover_interval.
  intros.
  destruct n.
  - auto.
  - false. forwards: H n. auto. inversions H0.
Qed.
Lemma list_cover_interval_cons_mem_inv n ns m:
  list_cover_interval (n :: ns) m ->
  LibList.mem n ns ->
  list_cover_interval (ns) m.
Proof.
  introv Hci Hm.
  unfold list_cover_interval in *.
  intros.
  forwards: Hci; fweauto.
  inversions H0; auto.
Qed.

Lemma list_reorder_nil_ns_case (A: Type) (xs ys: list A):
  list_reorder xs nil ys ->
  ys = nil.
Proof.
  introv Hre. inversion Hre.
  auto.
Qed.

Lemma list_cover_interval_app_l (ns1 ns2: list nat) m:
  list_cover_interval ns1 m ->
  list_cover_interval (ns1++ns2) m.
Proof.
  introv Hci.
  unfold list_cover_interval in *.
  intros.
  forwards: Hci; fweauto.
  rewrite <- liblist_app.
  eapply LibList.mem_app_l; eauto.
Qed.
Lemma list_cover_interval_app_r (ns1 ns2: list nat) m:
  list_cover_interval ns2 m ->
  list_cover_interval (ns1++ns2) m.
Proof.
  introv Hci.
  unfold list_cover_interval in *.
  intros.
  forwards: Hci; fweauto.
  rewrite <- liblist_app.
  eapply LibList.mem_app_r; eauto.
Qed.

Lemma list_noduplicates_gt_cover_interval ns:
  LibList.noduplicates ns ->
  LibList.Forall (gt (length ns)) ns ->
  list_cover_interval ns (length ns).
Proof.
  introv Hnd Hgt.
  remember (length ns) as m.
  reverts ns.
  induction m; intros; subst.
  - eapply list_cover_interval_0; eauto.
  - unfold list_cover_interval in *.
    intros.
    destruct ns; inversions Heqm.
    destruct (classicT (m0 = n)).
    + subst. auto.
    + inversions Hnd.
      inversions Hgt.
      forwards: IHm (LibList.map (decrease_after n) ns) (decrease_after n m0).
      ++
      eapply list_noduplicates_remove_middle; eauto.
      ++
      rewrite <- liblist_length.
      rewrite LibList.length_map. auto.
      ++
      eapply list_all_gt_decrease; eauto.
      ++
      unfold decrease_after; cases_if.
      * rewrite <- (PeanoNat.Nat.pred_succ (length ns)). apply -> PeanoNat.Nat.pred_lt_mono. auto. intro. subst. eapply Le.le_n_0_eq in C. subst. contradiction.
      * unfold gt in H4. eapply Lt.lt_n_Sm_le in H4.
        eapply PeanoNat.Nat.lt_le_trans; eauto.
      ++ forwards: LibList.mem_map (increase_after n) H0.
         destruct (classicT(m0=n)).
         * subst. auto.
         * rewrite <- list_increase_decrease in H1 by auto.
           rewrite increase_decrease_after in H1 by auto. auto.
Qed.


Lemma list_reorder_push_inv (A: Type) (a: A) xs ns ys:
  list_reorder (a :: xs) ns ys ->
  LibList.mem 0 ns \/ list_reorder xs (List.map pred ns) ys.
Proof.
  introv Hre.
  reverts ys.
  induction ns; intros.
  - inversions Hre.
    right. eapply list_reorder_nil.
  - destruct a0.
    + (* found index 0 *)
      left. auto.
    + inversions Hre.
      forwards: IHns H2.
      destructecs H.
      * left. auto.
      * right.
        rewrite List.map_cons.
        rewrite PeanoNat.Nat.pred_succ.
        eapply list_reorder_cons. auto.
        inversions H4. auto.
Qed.

Lemma list_reorder_mem_zero (A: Type) (a: A) xs ns ys:
  list_reorder (a :: xs) ns ys ->
  LibList.mem a ys ->
  ~ LibList.mem a xs ->
  LibList.mem 0 ns.
Proof.
  introv Hre Hmys Hmxs.
  forwards Hi: list_reorder_push_inv Hre.
  destructecs Hi.
  - auto.
  - (* Index zero not found - contradiction with the membership *)
     false Hmxs.
     eapply list_reorder_mem; eauto.
Qed.

Lemma list_reorder_push_inv_remove (A: Type) (a: A) xs ns ys:
  list_reorder (a :: xs) ns ys ->
  ~ LibList.mem a xs ->
  list_reorder xs (List.map pred (LibList.remove 0 ns)) (LibList.remove a ys).
Proof.
  introv Hre Hm.
  reverts ys.
  induction ns; intros.
  - inversions Hre.
    unfold LibList.remove.
    rewrite ! LibList.filter_nil.
    simpl.
    eapply list_reorder_nil.
  - destruct a0.
    + (* found index 0 - remove *)
      inversions Hre.
      inversions H4.
      rewrite ! list_remove_cons_eq.
      eapply IHns. auto.
    + inversions Hre.
      inversions H4.
      assert (y <> a). {
        intro. subst.
        eapply LibList.Nth_mem in H5. false Hm; auto.
      }
      rewrite ! list_remove_cons_ne.
      rewrite List.map_cons.
      rewrite PeanoNat.Nat.pred_succ.
      eapply list_reorder_cons. auto. auto. auto. discriminate.
Qed.

Lemma list_reorder_functional A (xs: list A) ns ys1 ys2:
  list_reorder xs ns ys1 ->
  list_reorder xs ns ys2 ->
  ys1 = ys2.
Proof.
  introv Hre1 Hre2.
  reverts ys2.
  induction Hre1; intros.
  - inversions Hre2. auto.
  - inversions Hre2.  rewrite (IHHre1 ys0) by auto. forwards: LibList.Nth_functional H H5. subst; auto.
Qed.

Lemma list_cover_interval_reorder_subset (A: Type) (ns: list nat) (xs ys: list A):
  list_reorder xs ns ys ->
  (from_list xs) \c (from_list ys) ->
  LibList.noduplicates xs ->
  list_cover_interval ns (length xs).
Proof.
  introv Hs Hre.
  unfold list_cover_interval.
  remember (length xs) as m.
  reverts xs ns ys.
  induction m; intros.
  - eapply PeanoNat.Nat.nlt_0_r in H0. false H0.
  - destruct xs; inversions Heqm.
    inversions H.
    rewrite from_list_cons in *.
    forwards: subset_union_l1 Hre.
    forwards: subset_union_l2 Hre.
    rewrite subset_singleton in H.
    destruct m0.
    + (* 0 = first element a *)
      eapply list_reorder_mem_zero; eauto.
      eapply list_in_fromlist_mem; eauto.
    + (* Succ *)
      eapply Lt.lt_S_n in H0.

      forwards: list_reorder_push_inv_remove Hs H3.
      forwards: IHm H2; eauto.
      rewrite from_list_remove.
      intro. intro.
      rewrite in_remove. split. auto. rewrite notin_singleton. intro. subst.
      eapply list_in_fromlist_mem in H5. contradiction.
      
      forwards Hmemsp: LibList.mem_map S H5.
      rewrite <- liblist_map in Hmemsp.
      rewrite list_map_S_pred_remove_0 in Hmemsp.
      eapply LibList.mem_remove_inv; eauto.
Qed.

Lemma list_cover_interval_succ_inv ns m:
  list_cover_interval ns (S m) ->
  list_cover_interval ns m.
Proof.
  introv Hci.
  unfold list_cover_interval in *.
  intros; eapply Hci. eauto.
Qed.

Lemma list_cover_interval_reorder_eq_app (A: Type) (zs xs1 xs2: list A) ns ys:
  list_reorder (zs ++ xs1) ns ys ->
  list_reorder (zs ++ xs2) ns ys ->
  list_cover_interval ns (length zs + length xs1) ->
  length xs2 = length xs1 ->
  xs1 = xs2.
Proof.
  introv Hre1 Hre2 Hci Hl.
  remember (length xs1) as m.
  reverts xs1 xs2 ns ys.
  reverts zs.
  induction m; intros.
  - destruct xs1; inversions Heqm.
    destruct xs2; inversions Hl.
    auto.
  - destruct xs1; inversions Heqm.
    destruct xs2; inversions Hl.
    assert (a = a0). {
      (* head *) 
      forwards Hmemz: Hci (length zs).
      eapply PeanoNat.Nat.lt_add_pos_r.
      eapply PeanoNat.Nat.lt_0_succ.
      forwards Hnthz: LibList.mem_Nth Hmemz.
      destructecs Hnthz.

      forwards Hnth1: list_reorder_nth_xs_nth_ys Hre1 Hnthz.
      rewrite <- liblist_app.
      eapply LibList.Nth_app_r'. eapply LibList.Nth_zero. simpl. rewrite liblist_length. auto.

      forwards Hnth2: list_reorder_nth_xs_nth_ys Hre2 Hnthz.
      rewrite <- liblist_app.
      eapply LibList.Nth_app_r'. eapply LibList.Nth_zero. simpl. rewrite liblist_length. auto.
      eapply LibList.Nth_functional; eauto.
    }
    subst.
    fequal.
    {
      (* tail *)
      rewrite <- liblist_app in Hre1.
      rewrite LibList.app_cons_r in Hre1.
      rewrite liblist_app in Hre1.
      rewrite <- liblist_app in Hre2.
      rewrite LibList.app_cons_r in Hre2.
      rewrite liblist_app in Hre2.
      forwards: IHm Hre1 Hre2; auto.
      rewrite List.app_length.
      simpls.
      rewrite <- PeanoNat.Nat.add_assoc.
      rewrite PeanoNat.Nat.add_1_l.
      auto.
    }
Qed.


Lemma list_cover_interval_reorder_eq (A: Type) (xs1 xs2: list A) ns ys:
  list_reorder xs1 ns ys ->
  list_reorder xs2 ns ys ->
  list_cover_interval ns (length xs1) ->
  length xs2 = length xs1 ->
  xs1 = xs2.
Proof.
  introv Hre1 Hre2 Hci Hl.
  forwards: list_cover_interval_reorder_eq_app (@nil A) xs1 xs2; eauto.
Qed.

Lemma list_cover_interval_remove_n n ns m:
  list_cover_interval (n :: ns) m ->
  list_cover_interval (LibList.map (decrease_after n) ns) (Nat.pred m).
Proof.
  introv Hci.
  unfold list_cover_interval in *.
  intros.

  destruct ( Compare_dec.le_lt_dec n m0 ).
  - remember (decrease_after n (S m0)).
    pose proof Heqn0 as Heqn0'.
    unfold decrease_after in Heqn0'.
    cases_if.
    +
      forwards: Hci (S m0). apply PeanoNat.Nat.lt_succ_lt_pred; auto.
      inversions H0.
      * false PeanoNat.Nat.nle_succ_diag_l; eauto.
      * rewrite PeanoNat.Nat.pred_succ in Heqn0'.
        rewrite <- Heqn0'.
        eapply LibList.mem_map; auto.
    + false. forwards: PeanoNat.Nat.lt_le_trans C l. eapply PeanoNat.Nat.nlt_succ_diag_l; eauto.
  - remember (decrease_after n m0).
    pose proof Heqn0 as Heqn0'.
    unfold decrease_after in Heqn0'.
    cases_if.
    + false. forwards: PeanoNat.Nat.le_lt_trans C l. eapply PeanoNat.Nat.lt_irrefl; eauto.
    +
      subst. rewrite <- Heqn0'.
      eapply LibList.mem_map.
      forwards: Hci m0.
      eapply PeanoNat.Nat.lt_pred_lt; eauto.
      inversions H0.  eapply PeanoNat.Nat.lt_irrefl in C. false C.
      auto.
Qed.

Lemma list_cover_interval_2 xs A B nsA (xsA: list var) nsB xsB:
  list_reorder xs nsA xsA ->
  list_reorder xs nsB xsB ->
  LibList.noduplicates xs ->
  from_list xs \c A \u B ->
  from_list xsA = A \n from_list xs ->
  from_list xsB = B \n from_list xs ->
  list_cover_interval (nsA ++ nsB) (length xs).
Proof.
  introv HreA HreB Hnd Hs HeA HeB.
  forwards: list_cover_interval_reorder_subset (nsA ++ nsB)%list.
  eapply list_reorder_app; eauto. rewrite from_list_append. rewrite HeA, HeB. rewrite <- inter_union_l. eapply inter_subset; auto. eapply subset_refl.
  auto. 
  auto.
Qed.
Lemma list_cover_interval_3 xs A B C nsA (xsA: list var) nsB xsB nsC xsC:
  list_reorder xs nsA xsA ->
  list_reorder xs nsB xsB ->
  list_reorder xs nsC xsC ->
  LibList.noduplicates xs ->
  from_list xs \c A \u B \u C ->
  from_list xsA = A \n from_list xs ->
  from_list xsB = B \n from_list xs ->
  from_list xsC = C \n from_list xs ->
  list_cover_interval (nsA ++ nsB ++ nsC) (length xs).
Proof.
  introv HreA HreB HreC Hnd Hs HeA HeB HeC.
  forwards: list_cover_interval_reorder_subset (nsA ++ nsB ++ nsC)%list.
  eapply list_reorder_app; eauto.
  eapply list_reorder_app; eauto. rewrite 2 from_list_append. rewrite HeA, HeB, HeC. rewrite <- ! inter_union_l. eapply inter_subset; auto. eapply subset_refl.
  auto. 
  auto.
Qed.

(* Reordering two lists by two specs, producing four lists *)
Inductive list_reordering: (list var) -> (list var) -> (list var) -> (list var) -> (list var) -> (list var) -> Prop :=
  list_reordering_mk ns1 ns2 xs0 ys0 xs1 ys1 xs2 ys2:
    list_reorder xs0 ns1 xs1 ->
    list_reorder ys0 ns1 ys1 ->
    list_reorder xs0 ns2 xs2 ->
    list_reorder ys0 ns2 ys2 ->
    LibList.noduplicates ns1 ->
    LibList.noduplicates ns2 ->
    list_reordering xs0 ys0 xs1 ys1 xs2 ys2.


Lemma list_reordering_nil (xs ys: list var):
  (length xs = length ys) ->
  list_reordering  xs ys nil nil nil nil.
Proof.
  introv Hl.
  econstructor; try eapply LibList.noduplicates_nil; constructor.
Qed.

Lemma list_reordering_push  xs0 ys0 xs1 ys1 xs2 ys2 x y:
  list_reordering  xs0 ys0 xs1 ys1 xs2 ys2 ->
  list_reordering (x::xs0) (y::ys0) (x::xs1) (y::ys1) (x::xs2) (y::ys2).
Proof.
  intros.
  inversions H.
  econstructor.
  eapply list_reorder_push; eapply H0.
  eapply list_reorder_push; eapply H1.
  eapply list_reorder_push; eapply H2.
  eapply list_reorder_push; eapply H3.
  eapply list_noduplicates_shift; eauto.
  eapply list_noduplicates_shift; eauto.
Qed.

Lemma list_reordering_lengths xs0 ys0 xs1 ys1 xs2 ys2:
  list_reordering xs0 ys0 xs1 ys1 xs2 ys2 ->
  (length xs1 = length ys1) /\ (length xs2 = length ys2).
Proof.
  introv Hro.
  inversions Hro.
  eapply list_reorder_lengths in H.
  eapply list_reorder_lengths in H0.
  eapply list_reorder_lengths in H1.
  eapply list_reorder_lengths in H2.
  rewrite H.
  rewrite H0.
  rewrite H1.
  rewrite H2.
  auto.
Qed.

Lemma list_reordering_same (xs ys: list var):
  (length xs = length ys) ->
  list_reordering  xs ys xs ys xs ys.
Proof.
  introv Hl.
  reverts ys.
  induction xs; intros; simpls.
  - destruct ys; inversions Hl. eapply list_reordering_nil. auto.
  - destruct ys; inversions Hl. eauto using list_reordering_push.
Qed.



Lemma list_reordering_new_one1 xs0 ys0 xs1 ys1 xs2 ys2 x y:
  list_reordering xs0 ys0 xs1 ys1 xs2 ys2 ->
  list_reordering (x::xs0) (y::ys0) (x::nil) (y::nil) xs2 ys2.
Proof.
  introv Hro.
  inversions Hro.
  econstructor.
  eapply list_reorder_one.
  eapply list_reorder_one.

  eapply list_reorder_push0; eauto.
  eapply list_reorder_push0; eauto.

  constructor. intro. eapply LibList.mem_nil_inv; eauto. constructor.
  eapply list_noduplicates_succ; eauto.
Qed.


Lemma list_reordering_old_one1 xs0 ys0 xs1 ys1 xs2 ys2 x y n:
  list_reordering xs0 ys0 xs1 ys1 xs2 ys2 ->
  LibList.Nth n xs1 x ->
  LibList.Nth n ys1 y ->
  list_reordering (xs0) (ys0) (x::nil) (y::nil) xs2 ys2.
Proof.
  introv Hro Hnx Hny.
  inversions Hro.

  destructc (list_reorder2_nth_nth H H0 Hnx Hny) as m Hmx Hmy.

  econstructor.
  eapply list_reorder_old_one; eauto.
  eapply list_reorder_old_one; eauto.

  eauto.
  eauto.

  constructor. intro. eapply LibList.mem_nil_inv; eauto. constructor.
  eauto.
Qed.

Lemma list_reordering_r xs0 ys0 xs1 ys1 xs2 ys2:
  list_reordering xs0 ys0 xs1 ys1 xs2 ys2 ->
  list_reordering xs0 ys0 xs2 ys2 xs2 ys2.
Proof.
  introv Hro.
  inversions Hro.
  econstructor; eauto.
Qed.
Lemma list_reordering_swap xs0 ys0 xs1 ys1 xs2 ys2:
  list_reordering xs0 ys0 xs1 ys1 xs2 ys2 ->
  list_reordering xs0 ys0 xs2 ys2 xs1 ys1.
Proof.
  introv Hro.
  inversions Hro.
  econstructor; eauto.
Qed.
