(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Bool.Bool.
Require Export Coq.Program.Basics.

From TLC Require Import LibTactics.

(** * Lemmas about natural numbers *)

Lemma add_sub_move m a n b:
  a <= m ->
  n + b = m - a ->
  n + b + a = m.
Proof.
  intros.
  rewrite H0.
  eapply PeanoNat.Nat.sub_add; eauto.
Qed.

Lemma add_sub_eq_swap m a n b:
  a <= m ->
  b <= m ->
  n + b = m - a ->
  n + a = m - b.
Proof.
  intros.
  eapply add_sub_move in H; eauto.
  subst. rewrite PeanoNat.Nat.add_shuffle0. rewrite PeanoNat.Nat.add_sub. eauto.
Qed.

Lemma add_sub_le m a n b:
  a <= m ->
  n + b = m - a ->
  b <= m.
Proof.
  intros.
  eapply add_sub_move in H0; eauto.
  subst.
  eapply PeanoNat.Nat.le_trans.
  eapply PeanoNat.Nat.le_add_r.
  rewrite <- (PeanoNat.Nat.add_assoc).
  rewrite (PeanoNat.Nat.add_comm n (b+a)).
  rewrite <- (PeanoNat.Nat.add_assoc).
  eauto.
Qed.

Lemma ne_n_S m n:
  m <> n -> S m <> S n.
Proof.
  introv H0. intro. eapply H0. eapply PeanoNat.Nat.succ_inj. eauto.
Qed.

Lemma lt_l_plus_false m n:
  m + n < m ->
  False.
Proof.
  introv Hlt.
  rewrite <- (PeanoNat.Nat.add_0_r m) in Hlt at 2.
  eapply Plus.plus_lt_reg_l in Hlt.
  eapply PeanoNat.Nat.nlt_0_r; eauto.
Qed.
Lemma gt_plus_l_false m n:
  m > m + n ->
  False.
Proof.
  unfold gt.
  eapply lt_l_plus_false; eauto.
Qed.
Lemma gt_plus_lr m n x:
  m > x ->
  m + n > x.
Proof.
  unfold gt.
  introv Hgt.
  eapply PeanoNat.Nat.lt_lt_add_r.
  eauto.
Qed.

Definition decrease_after n m := if (Compare_dec.le_lt_dec n m ) then (Nat.pred m) else m.
Definition increase_after n m := if (Compare_dec.le_lt_dec n m ) then (S m) else m.

Lemma increase_decrease_after n m:
  n <> m ->
  increase_after n (decrease_after n m ) = m.
Proof.
  intros.
  unfold increase_after, decrease_after.
  cases_if; cases_if; eauto.
  -
  destruct (Compare_dec.le_lt_eq_dec n m C0).
  eapply PeanoNat.Nat.lt_succ_pred; eauto. contradiction.
  - eapply Lt.lt_not_le in C0. contradiction.
  - destruct (Compare_dec.le_lt_eq_dec n m C0).
    unfold lt in C.  erewrite  PeanoNat.Nat.lt_succ_pred in C; eauto. false. eapply H. eapply PeanoNat.Nat.le_antisymm; eauto.
    contradiction.
Qed.
Lemma decrease_increase_after n m:
  decrease_after n (increase_after n m ) = m.
Proof.
  intros.
  unfold increase_after, decrease_after.
  cases_if; cases_if; eauto.
  - eapply Lt.lt_not_le in C0. contradiction.
  - eapply PeanoNat.Nat.lt_succ_l in C. eapply Lt.lt_not_le in C. contradiction.
Qed.

(* Used in list_all_gt_decrease *)
Lemma gt_decrease_after_from m n a:
  S m > n ->
  S m > a ->
  n <> a ->
  m > decrease_after n a.
Proof.
  introv Hmn Hma Hna.
  unfold gt, lt in *.
  simpls. unfold decrease_after. cases_if.
  *
    destruct a.
    ** false. apply Hna. eapply Le.le_n_0_eq in C; subst. eauto.
    **
      rewrite PeanoNat.Nat.pred_succ.
      eapply le_S_n. eauto.
  *
    unfold lt in C. eapply le_S_n in Hmn. eapply PeanoNat.Nat.le_trans; eauto.
Qed.


(* Simpler version of Nat.strong_right_induction *)

Lemma nat_lt_induction (A: nat -> Prop):
  (forall n : nat, (forall m : nat, m < n -> A m) -> A n) ->
  forall n : nat, A n.
Proof.
  intros.
  forwards: PeanoNat.Nat.strong_right_induction A 0 n.  
  -
  hnf.
  split; subst; eauto.
  -
  intros. eapply H. intros. eapply H1. apply PeanoNat.Nat.le_0_l. auto.
  - apply PeanoNat.Nat.le_0_l.
  - auto.
Qed.

Lemma nat_lt_0S_induction (A: nat -> Prop):
  (A 0) ->
  (forall n : nat, 0 < n -> (forall m : nat, m < n -> A m) -> A n) ->
  forall n : nat, A n.
Proof.
  intros.
  eapply nat_lt_induction.
  intros. destruct n0; eauto using PeanoNat.Nat.lt_0_succ.
Qed.

Lemma nat_add_cases n1 n2:
  (n1 = 0 /\ n2 = 0) \/ 0 < n1 \/ 0 < n2.
Proof.
  destruct (PeanoNat.Nat.eq_0_gt_0_cases n1);
  destruct (PeanoNat.Nat.eq_0_gt_0_cases n2); eauto.
Qed.

Lemma nat_le_induction (A: nat -> Prop):
  (A 0) ->
  (forall n : nat, (forall m : nat, m <= n -> A m) -> A (S n)) ->
  forall n : nat, A n.
Proof.
  intros.
  eapply nat_lt_0S_induction.
  - auto.
  - intros. destruct n0; eauto using Lt.le_lt_n_Sm.
Qed.