Set Implicit Arguments.

From TLC Require Import LibVar LibLN LibFset LibTactics.
Require Import LibTactics.

(** * Lemmas about sets *)

Hint Extern 1 (?F \u ?G = ?G \u ?F) => rewrite union_comm; reflexivity : core.
Hint Extern 1 (?E \u ?F \u ?G = (?E \u ?F) \u ?G) => rewrite union_assoc; reflexivity : core.
Hint Extern 1 (?E \u ?F \u ?G = ?E \u ?G \u ?F) =>
rewrite (@union_comm _ F G); reflexivity : core.
Hint Extern 1 (?E \u ?F \u ?G = ?F \u ?E \u ?G) =>
rewrite (@union_comm_assoc _ E F G); reflexivity : core.
Hint Extern 1 (?E \u ?F \u ?E = ?F \u ?E) =>
rewrite union_comm_assoc; rewrite union_same; reflexivity : core.
Hint Extern 1 (?E \u \{} = ?E) => rewrite union_empty_r; reflexivity : core.
Hint Extern 1 (?E = ?E \u \{}) => rewrite union_empty_r; reflexivity : core.
Hint Extern 1 (\{} \u ?E = ?E) => rewrite union_empty_l; reflexivity : core.
Hint Extern 1 (?E = \{} \u ?E) => rewrite union_empty_l; reflexivity : core.

Lemma union_same_push : forall A (E F : fset A),
    E \u (E \u F) = E \u F.
Proof.
  intros. rewrite union_assoc, union_same. reflexivity.
Qed.

Lemma union_eq_cases : forall A (E F G H I : fset A),
    E = F \/ E = I \u F ->
    G = H \/ G = I \u H ->
    E \u G = F \u H \/ E \u G = I \u F \u H.
Proof.
  intros. destruct H0; destruct H1; subst; try solve [left; auto 2 | right; auto 2].
  - right. rewrite <- union_assoc. auto.
  - right.
    rewrite union_comm_assoc, <- union_assoc, union_same_push. auto 2.
Qed.


Lemma singleton_ne_empty A (l: A):
  \{} <> \{ l }.
Proof.
  intro.
  
  assert (l \in \{ l }).
  eapply in_singleton_self.
  rewrite <- H in H0.
  apply notin_empty in H0.
  contradiction.
Qed.

Lemma singleton_union_ne_empty A (l: A) ls:
  \{} <> ls \u \{ l }.
Proof.
  intro.
  
  assert (l \in ls \u \{ l }).
  rewrite in_union. right.
  eapply in_singleton_self.
  rewrite <- H in H0.
  apply notin_empty in H0.
  contradiction.
Qed.

Lemma singleton_eq_inv A (l1 l2: A):
  \{ l1 } = \{ l2 } ->
  l1 = l2.
Proof.
  intros.
  
  assert (l1 \in \{ l1 }).
  eapply in_singleton_self.
  rewrite H in H0.
  rewrite in_singleton in H0.
  assumption.
Qed.

Lemma singleton_eq_union_notin_inv A (l1 l2: A) ls:
  l1 \notin ls ->
  ls \u \{ l1 } = \{ l2 } ->
  ls = \{}.
Proof.
  intros.
  assert (l1 \in \{ l2 }).
  rewrite <- H0.
  rewrite in_union. right.
  eapply in_singleton_self.
  pose proof H1.
  rewrite in_singleton in H1.
  subst.
  destruct (fset_finite ls).
  induction x.
  + eauto.
  + false.
    assert( a \in ls).
    rewrite H1. compute.
    rewrite in_union.
    left.
    rewrite in_singleton.
    auto.
    assert( a \in \{ l2 }).
    rewrite <- H0.
    rewrite in_union.
    left.
    assumption.
    rewrite in_singleton in H4.
    subst.
    contradiction.
Qed.

Lemma in_union_singleton_ne_l A (l1 l2: A) ls:
  l1 \in \{ l2 } \u ls ->
  l1 <> l2 ->
  l1 \in ls.
Proof.
  intros.
  rewrite in_union in H.
  rewrite in_singleton in H.
  destruct H.
  - contradiction.
  - assumption.
Qed.

Lemma union_singleton_eq_notin_eq A (x: A) E F:
  x \notin E ->
  x \notin F ->
  \{ x } \u E = \{ x } \u F ->
  E = F.
Proof.
  introv HnE HnF Heq.
  eapply fset_extens.
  - intro. intro.
    assert ( x0 \in \{ x} \u E ).
    rewrite in_union; auto.
    rewrite Heq in H0.
    rewrite in_union in H0.
    destruct H0.
    + rewrite in_singleton in H0; subst.
      contradiction.
    + auto.
  - (* Symmetric case *)
    intro. intro.
    assert ( x0 \in \{ x} \u F ).
    rewrite in_union; auto.
    rewrite <- Heq in H0.
    rewrite in_union in H0.
    destruct H0.
    + rewrite in_singleton in H0; subst.
      contradiction.
    + auto.
Qed.


Hint Rewrite in_union: sets.
Hint Rewrite notin_union: sets.
Hint Rewrite in_singleton: sets.
Hint Rewrite notin_singleton: sets.
Hint Rewrite <- union_assoc: sets.
Hint Rewrite in_remove: sets.

Ltac rewrite_sets :=
  repeat match goal with
    H: _ = _ |- _ => rewrite H in *; clear H
  end.

Lemma union_remove_swap (X: Type) (A B: fset X) x y:
  x \notin B ->
  ((A \u B) \- \{ x} \u \{ y}) = (A \- \{ x} \u \{ y} \u B).
Proof.
  introv Hn.
  rewrite union_remove.
  assert (B \- \{ x} = B).
  apply fset_extens; hnf; intros.
  rewrite in_remove in H. destruct H; auto.
  rewrite in_remove. split. auto. rewrite notin_singleton. intro. subst. contradiction.
  rewrite H.
  auto.
  rewrite <- union_assoc.
  rewrite union_comm with (F:=B).
  auto.
Qed.

Lemma union_remove_swap2 (X: Type) (A B: fset X) x y:
  x \notin B ->
   (A \- \{ x} \u \{ y} \u B) = ((A \u B) \- \{ x} \u \{ y}).
Proof.
  introv Hn.
  symmetry.

  eapply union_remove_swap. auto.
Qed.
Lemma union_remove_swap2r (X: Type) (A B: fset X) x y:
  x \notin B ->
   B \u (A \- \{ x} \u \{ y}) = ((B \u A) \- \{ x} \u \{ y}).
Proof.
  introv Hn.
  rewrite union_comm.
  rewrite (union_comm B A).
  rewrite <- union_assoc.
  eapply union_remove_swap2. auto.
Qed.

Lemma union_remove_swap0 (X: Type): forall (A B : fset X) (x : X), x \notin A -> (A \u B) \- \{ x} = A \u (B \- \{ x}).
Proof.
  intros.
  eapply fset_extens.
  intro.  intro. autorewrite with sets in *. destruct H0; destruct H0; eauto.
  intro.  intro. autorewrite with sets in *. destruct H0; try destruct H0; splits; eauto. intro; subst; contradiction.
Qed.

Lemma union_zip A (X1 Y1 X2 Y2: fset A):
  (X1 \u X2) \u (Y1 \u Y2) = (X1 \u Y1) \u (X2 \u Y2).
Proof.
  rewrite union_assoc.
  rewrite <- (union_assoc X1).
  rewrite (union_assoc _ _ Y2).
  rewrite <- (union_assoc _ _ X2).
  rewrite (union_comm X2 Y1).
  eapply eq_refl.
Qed.

Lemma subset_union_zip A (X1 Y1 Z1 X2 Y2 Z2: fset A):
  X1 \c Y1 \u Z1 ->
  X2 \c Y2 \u Z2 ->
  (X1 \u X2) \c (Y1 \u Y2) \u (Z1 \u Z2).
Proof.
  intros.
  rewrite union_zip.
  eapply subset_union_2; auto.
Qed.

Lemma notin_remove A (X Y: fset A) x:
  x \notin X \/ x \in Y ->
  x \notin (X \- Y).
Proof.
  introv Hn.
  intro.
  rewrite in_remove in H.
  destruct H.
  destruct Hn;
  contradiction.
Qed.

Lemma union_remove_dist A (X Y: fset A) x:
  (X \- x) \u (Y \- x) = (X \u Y) \- x.
Proof.
  eapply fset_extens.
  - intro. intro.
    rewrite in_union in H.
    destruct H.
    rewrite in_remove in H. destruct H.
    rewrite in_remove. split; auto. rewrite in_union; auto.
    rewrite in_remove in H. destruct H.
    rewrite in_remove. split; auto. rewrite in_union; auto.
  - intro. intro.
    rewrite in_remove in H. destruct H.
    rewrite in_union in H. destruct H.
    rewrite in_union. left. rewrite in_remove; auto.
    rewrite in_union. right. rewrite in_remove; auto.
Qed.


Lemma remove_empty_r A (X: fset A):
  \{} \- X = \{}.
Proof.
  eapply fset_extens; intro; intro.
  - rewrite in_remove in H. destruct H. rewrite in_empty in H. contradiction.
  - rewrite in_empty in H. contradiction.
Qed.
Lemma remove_empty_l A (X: fset A):
  X \- \{} = X.
Proof.
  eapply fset_extens; intro; intro.
  - rewrite in_remove in H. destruct H. auto.
  - rewrite in_remove; eauto.
Qed.
Lemma remove_same A (X: fset A):
  X \- X = \{}.
Proof.
  eapply fset_extens; intro; intro.
  - rewrite in_remove in H. destruct H. contradiction.
  - rewrite in_empty in H. contradiction.
Qed.

Lemma subset_union_rr A (E F G: fset A):
  E \c F ->
  E \c (G \u F).
Proof.
  introv HEF.
  rewrite <- (union_empty_l E).
  eapply subset_union_2.
  eapply subset_empty_l.
  auto.
Qed.

Lemma subset_union_rl A (E F G: fset A):
  E \c F ->
  E \c (F \u G).
Proof.
  introv HEF.
  rewrite (union_comm).
  eapply subset_union_rr.
  auto.
Qed.

Lemma subset_singleton A x (X: fset A):
  \{ x } \c X <-> x \in X.
Proof.
  split; intros.
  apply H. rewrite in_singleton. auto.
  intro. rewrite in_singleton. intro. subst. auto.
Qed.

Lemma in_notin_neq (x y: var) A:
  x \in A ->
  y \notin A ->
  x <> y.
Proof.
  intros; intro; subst.
  contradiction.
Qed.

Lemma subset_union_l1 A (a b c: fset A): a \u b \c c -> a \c c.
Proof.
  introv Hab.
  intro. intro. apply Hab. rewrite in_union. auto.
Qed.
Lemma subset_union_l2 A (a b c: fset A): a \u b \c c -> b \c c.
Proof.
  introv Hab.
  intro. intro. apply Hab. rewrite in_union. auto.
Qed.
Lemma subset_trans A (a b c: fset A): a \c b -> b \c c -> a \c c.
Proof.
  introv Hab Hbc.
  intro. intro. apply Hbc. apply Hab. auto.
Qed.

Lemma subset_union_l (A: Type) (E F G: fset A):
  E \c G -> F \c G -> (E \u F) \c G.
Proof.
  intros.
  intro.
  intro.
  rewrite in_union in H1. destruct H1; eauto.
Qed.

Lemma subset_empty_inv (A: Type) (E: fset A):
  E \c \{} ->
  E = \{}.
Proof.
  introv Hs.
  eapply fset_extens; eauto using subset_empty_l.
Qed.

Lemma in_notin_disjoint (A : Type) (E F : fset A) :
  (forall x, x \in E -> x \notin F) ->
  disjoint E F.
Proof.
  intros.
  unfold disjoint.
  eapply fset_extens.
  intro. intro. rewrite in_inter in H0. destruct H0. eapply H in H0. contradiction.
  intro. intro. rewrite in_empty in H0. contradiction.
Qed.

Lemma union_subset_r (A: Type) (E F: fset A):
  E \c F -> E \u F = F.
Proof.
  introv Hs.
  eapply fset_extens.
  - eapply subset_union_l; auto using subset_refl.
  - eapply subset_union_rr. auto using subset_refl.
Qed.

Lemma remove_union_same (A:Type) (a b: fset A):
  (a \- b) \u b = a \u b.
Proof.
  eapply fset_extens; intro; rewrite ? in_union; rewrite ? in_remove; intros; repeat destruct H; auto.
  destruct (classicT( x \in b)); eauto.
Qed.

Lemma remove_union A (X Y Z: fset A):
  X \- (Y \u Z) =  (X \- Y) \- Z.
Proof.
  eapply fset_extens; intro; intro.
  rewrite ! in_remove, notin_union in *.
  destruct H as [? [? ?]]. eauto.
  rewrite ! in_remove, notin_union in *.
  destruct H as [[? ?] ?]. eauto.
Qed.

Lemma remove_remove A (X Y Z: fset A):
  (X \- Y) \- Z =  (X \- Z) \- Y.
Proof.
  rewrite <- remove_union.
  rewrite union_comm.
  rewrite remove_union.
  auto.
Qed.


Lemma inter_singleton_notin X (a:X) (A: fset X):
  a \notin A ->
  \{ a} \n A = \{}.
Proof.
  introv Hn.
  eapply fset_extens.
  - intro. rewrite in_inter, in_singleton. intro; destruct H; subst.
    contradiction.
  - intro. intro. rewrite in_empty in H. contradiction.
Qed.

Lemma notin_inter_singleton X (a:X) (A: fset X):
  \{ a} \n A = \{} -> 
  a \notin A.
Proof.
  introv Hn.
  intro.
  assert (a \in \{ a} \n A ).
  rewrite in_inter, in_singleton; auto.
  rewrite Hn in H0. rewrite in_empty in H0. auto.
Qed.

Lemma disjoint_singleton X (a:X) (A: fset X):
  a \notin A ->
  disjoint \{a} A.
Proof.
  introv Hn.
  unfold disjoint.
  eapply inter_singleton_notin; eauto.
Qed.
Lemma disjoint_singleton_r X (a:X) (A: fset X):
  a \notin A ->
  disjoint A \{a}.
Proof.
  eauto using disjoint_comm, disjoint_singleton.
Qed.

Lemma inter_union_l X (A B C: fset X):
  (A \u B) \n C = (A \n C) \u (B \n C).
Proof.
  eapply fset_extens; intro; intro;
  rewrite ! in_union, ! in_inter, ! in_union in *; repeat destruct H; eauto.
Qed.

Lemma inter_union_r X (A B C: fset X):
  A \n (C \u B) = (A \n C) \u (A \n B).
Proof.
  eapply fset_extens; intro; intro;
  rewrite ! in_union, ! in_inter, ! in_union in *; repeat destruct H; repeat destruct H0; eauto.
Qed.
Lemma inter_subset_l A (X Y: fset A):
  X \n Y \c X.
Proof.
  intro. intro. rewrite in_inter in H. destruct H. auto.
Qed.
Lemma inter_subset_r A (X Y: fset A):
  X \n Y \c Y.
Proof.
  intro. intro. rewrite in_inter in H. destruct H. auto.
Qed.
Lemma in_inter_singleton A a (X: fset A):
  a \in X ->
  X \n \{a} = \{a}.
Proof.
  introv Hi.
  eapply fset_extens.
  - eapply inter_subset_r; eauto.
  - intro; intro. rewrite in_inter. rewrite in_singleton in *. subst. auto.
Qed.
Lemma inter_subset_rl X (A B: fset X):
  A \c B ->
  A \c (A \n B).
Proof.
  introv Hin.
  intro. intro.
  rewrite in_inter. eauto.
Qed.
Lemma inter_subset_rr X (A B: fset X):
  A \c B ->
  A \c (B \n A).
Proof.
  introv Hin.
  intro. intro.
  rewrite in_inter. eauto.
Qed.

Lemma inter_subset X (A B C: fset X):
  A \c B ->
  A \c C ->
  A \c (B \n C).
Proof.
  introv Hin.
  intro. intro.
  rewrite in_inter. eauto.
Qed.

Lemma inter_subset_ll X (A B C: fset X):
  A \c B ->
  (A \n C) \c B.
Proof.
  introv Hin.
  intro. intro.
  rewrite in_inter in H; destruct H; eauto.
Qed.
Lemma inter_subset_lr X (A B C: fset X):
  A \c B ->
  (C \n A) \c B.
Proof.
  introv Hin.
  intro. intro.
  rewrite in_inter in H; destruct H; eauto.
Qed.

Lemma inter_subset2 X (A B C D: fset X):
  A \c B ->
  C \c D ->
  (A \n C) \c (B \n D).
Proof.
  introv HinA HinC.
  eapply  inter_subset.
  eapply inter_subset_ll; eauto.
  eapply inter_subset_lr; eauto.
Qed.


Lemma disjoint_union_l X (A B C: fset X):
  disjoint A B ->
  disjoint C B ->
  disjoint (A \u C) B.
Proof.
  introv Hdab Hdcb.
  unfold disjoint in *.
  rewrite inter_union_l.
  rewrite Hdab, Hdcb.
  rewrite union_empty_r; eauto.
Qed.

Lemma disjoint_union_r X (A B C: fset X):
  disjoint A B ->
  disjoint A C ->
  disjoint A (B \u C).
Proof.
  introv Hdab Hdcb.
  unfold disjoint in *.
  rewrite inter_union_r.
  rewrite Hdab, Hdcb.
  rewrite union_empty_r; eauto.
Qed.

Lemma union_empty_inv X (A B: fset X):
  A \u B = \{} ->
  A = \{} /\ B = \{}.
Proof.
  introv Heq.
  split; eapply fset_extens; intro; intro.
  - rewrite <- Heq. rewrite in_union. auto.
  - rewrite in_empty in H. contradiction.
  - rewrite <- Heq. rewrite in_union. auto.
  - rewrite in_empty in H. contradiction.
Qed.

Lemma disjoint_union_l_inv X (A B C: fset X):
  disjoint (A \u C) B ->
  disjoint A B /\
  disjoint C B.
Proof.
  introv Hd.
  unfold disjoint in *.
  rewrite inter_union_l in *.
  eapply union_empty_inv; eauto.
Qed.
Lemma disjoint_union_r_inv X (A B C: fset X):
  disjoint A (B \u C) ->
  disjoint A B /\
  disjoint A C.
Proof.
  introv Hd.
  unfold disjoint in *.
  rewrite inter_union_r in *.
  eapply union_empty_inv; eauto.
Qed.

Lemma disjoint_singleton_inv X (a:X) (A: fset X):
  disjoint \{a} A ->
  a \notin A.
Proof.
  introv Hn.
  unfold disjoint in Hn.
  eapply notin_inter_singleton; eauto.
Qed.
Lemma disjoint_singleton_inv_r X (a:X) (A: fset X):
  disjoint A \{a} ->
  a \notin A.
Proof.
  eauto using disjoint_comm, disjoint_singleton_inv.
Qed.

Lemma disjoint_subset_l X (A B C: fset X):
  disjoint A B ->
  C \c B ->
  disjoint A C.
Proof.
  introv Hdab HsCB.
  unfold disjoint in *.
  eapply fset_extens; intro; intro.
  - rewrite in_inter in H. destruct H. apply HsCB in H0. rewrite <- Hdab. rewrite in_inter. auto.
  - rewrite in_empty in H. false H.
Qed.
Lemma disjoint_subset_r X (A B C: fset X):
  disjoint A B ->
  C \c A ->
  disjoint C B.
Proof.
  introv Hdab HsCA.
  unfold disjoint in *.
  eapply fset_extens; intro; intro.
  - rewrite in_inter in H. destruct H. apply HsCA in H. rewrite <- Hdab. rewrite in_inter. auto.
  - rewrite in_empty in H. false H.
Qed.
Lemma disjoint_empty_l {A} (X : fset A) : disjoint \{} X.
Proof. hnf. rewrite inter_empty_l. eauto. Qed.

Lemma disjoint_empty_r {A} (X : fset A) : disjoint X \{}.
Proof. hnf. rewrite inter_empty_r. eauto. Qed.


Lemma union_disjoint_inv {A} (X Y Z : fset A) :
  disjoint X Y -> disjoint X Z ->
  X \u Y = X \u Z -> Y = Z.
Proof.
  introv HdisY HdisZ Heq. eapply fset_extens; hnf; introv Hin.
  - assert (Hin' : x \in X \u Y).
    { eapply subset_union_weak_r; eauto. }
    rewrite Heq, in_union in Hin'.
    repeat(match goal with
           | [ H : _ \/ _ |- _ ] => destruct H
           end); eauto. false.
    erewrite <- in_empty, <- HdisY, in_inter. eauto.
  - assert (Hin' : x \in X \u Z).
    { eapply subset_union_weak_r; eauto. }
    rewrite <- Heq, in_union in Hin'.
    repeat(match goal with
           | [ H : _ \/ _ |- _ ] => destruct H
           end); eauto. false.
    erewrite <- in_empty, <- HdisZ, in_inter. eauto.
Qed.

Lemma notin_remove_inv (a: var) x y:
  a \notin (x \- y) ->
  a \notin x \/ a \in y.
Proof.
  introv Hn.
  destruct (classicT (a\in x)).
  - right.   destruct (classicT (a\in y)). auto. assert (a \in (x\- y)). rewrite in_remove. auto. contradiction.
  - auto.
Qed.
Lemma notin_remove_notin (a: var) x y:
  a \notin (x \- y) ->
  a \notin y ->
  a \notin x.
Proof.
  introv Hnr Hny.
  destruct (notin_remove_inv Hnr); auto; contradiction.
Qed.

Lemma notin_subset A a (E F: fset A):
  F \c E ->
  a \notin E ->
  a \notin F.
Proof.
  introv Hsu Hne.
  intro. eapply Hsu in H. contradiction.
Qed.


Lemma remove_inter (A: Set) (X Y: fset A):
  X \- Y = X \- (X \n Y).
Proof.
  eapply fset_extens; intro; intro.
  rewrite in_remove in *. destructecs H. split; auto. intro. rewrite in_inter in H1. destructec H1. apply H; eauto.
  rewrite in_remove in *. destructecs H. split; auto. intro. apply H. rewrite in_inter. split; eauto.
Qed.

Lemma remove_disjoint (A: Set) (X Y: fset A):
  disjoint X Y ->
  X \- Y = X.
Proof.
  introv Hdi.
  hnf in Hdi.
  rewrite remove_inter. rewrite Hdi. rewrite remove_empty_l. auto.
Qed.

Lemma disjoint_in_false (A: Set) (X Y: fset A) (x: A):
  disjoint X Y ->
  x \in X ->
  x \in Y ->
  False.
Proof.
  introv Hd Hinx Hiny.
  hnf in Hd.
  erewrite <- in_empty.
  rewrite <- Hd.
  rewrite in_inter.
  eauto.
Qed.


Ltac union_in_solver_r := match goal with 
| |- ?a \in \{?a} => rewrite in_singleton; auto
| |- ?a \in ?b \u ?c => rewrite in_union; first[left;union_in_solver_r| right;union_in_solver_r]
| H: ?a \in ?b |- ?a \in ?b => exact H
end.

(* Solve goals of the form x \in A \u B \u ... *)
Ltac union_in_solver := union_in_solver_r.

Ltac union_sub_solver_destruct_l := match goal with 
| |- ?a \u ?b \c ?c => apply subset_union_l
end.

(* Destruct goals of the form A \u B \u ... \c X. *)
Ltac union_sub_solver_destruct_ls := repeat union_sub_solver_destruct_l.

Ltac union_sub_solver_r := match goal with 
| |- ?a \c ?a => eapply subset_refl
| |- \{} \c ?a => eapply subset_empty_l
| |- ?a \c ?b \u ?c => eapply subset_union_rl; union_sub_solver_r
| |- ?a \c ?b \u ?c => eapply subset_union_rr; union_sub_solver_r
| H: ?a \c ?b |- ?a \c ?b => exact H
end.

(* Solve goals of the form A \u B \u ... \c B \u A \u ... *)
Ltac union_sub_solver := union_sub_solver_destruct_ls; union_sub_solver_r.

