Set Implicit Arguments.

(** * Helper tactics *)

(* Works like set,
  if the definition already exists, does nothing,
  if the name is already used, generates a fresh name
*)
Ltac namegive t_var t := (
  match goal with
  | H := t |- _ => idtac
  | |- _ => let t_var0 := fresh t_var in set (t_var0:=t) in
  end
).
(* Finds a local definition for t *)
Ltac namefind t := (
  match goal with
  | H := t |- _ => H
  end
).

(* Converts an equality H: X = A into H: A:=X *)
Ltac eq_to_def A := (
  match goal with 
  | H: ?X = A |- _ => subst A; set (A:=X) in *
  end
).

(* A variant of eapply with different handling of implicits *)
Ltac iapply F := let H := fresh in pose proof F as H; eapply H; clear H.

(* Tactics for destruction of right-associative conjunctions and disjunctions, to avoid the square bracket syntax *)
Tactic Notation "destructc" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) :=
  destruct T as [H1 H2].
Tactic Notation "destructc" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) simple_intropattern(H3) :=
  destruct T as [H1 [H2 H3]].
Tactic Notation "destructc" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) simple_intropattern(H3) simple_intropattern(H4) :=
  destruct T as [H1 [H2 [H3 H4]]].
Tactic Notation "destructc" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) simple_intropattern(H3) simple_intropattern(H4)
  simple_intropattern(H5) :=
  destruct T as [H1 [H2 [H3 [H4 H5]]]].
Tactic Notation "destructc" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) simple_intropattern(H3) simple_intropattern(H4)
  simple_intropattern(H5) simple_intropattern(H6) :=
  destruct T as [H1 [H2 [H3 [H4 [H5 H6]]]]].
Tactic Notation "destructc" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) simple_intropattern(H3) simple_intropattern(H4)
  simple_intropattern(H5) simple_intropattern(H6) simple_intropattern(H7) :=
  destruct T as [H1 [H2 [H3 [H4 [H5 [H6 H7]]]]]].
Tactic Notation "destructc" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) simple_intropattern(H3) simple_intropattern(H4)
  simple_intropattern(H5) simple_intropattern(H6) simple_intropattern(H7) simple_intropattern(H8) :=
  destruct T as [H1 [H2 [H3 [H4 [H5 [H6 [H7 H8 ]]]]]]].
Tactic Notation "destructc" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) simple_intropattern(H3) simple_intropattern(H4)
  simple_intropattern(H5) simple_intropattern(H6) simple_intropattern(H7) simple_intropattern(H8) simple_intropattern(H9) :=
  destruct T as [H1 [H2 [H3 [H4 [H5 [H6 [H7 [H8 H9 ]]]]]]]].
Tactic Notation "destructc" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) simple_intropattern(H3) simple_intropattern(H4)
  simple_intropattern(H5) simple_intropattern(H6) simple_intropattern(H7) simple_intropattern(H8) simple_intropattern(H9) simple_intropattern(H10) :=
  destruct T as [H1 [H2 [H3 [H4 [H5 [H6 [H7 [H8 [H9 H10]]]]]]]]].
Tactic Notation "destructc" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) simple_intropattern(H3) simple_intropattern(H4)
  simple_intropattern(H5) simple_intropattern(H6) simple_intropattern(H7) simple_intropattern(H8) simple_intropattern(H9) simple_intropattern(H10) simple_intropattern(H11) :=
  destruct T as [H1 [H2 [H3 [H4 [H5 [H6 [H7 [H8 [H9 [H10 H11]]]]]]]]]].

Tactic Notation "destructd" constr(T) "as" simple_intropattern(H1) simple_intropattern(H2) simple_intropattern(H3) simple_intropattern(H4) :=
  destruct T as [H1 | [H2 | [H3 | H4]]].

(* Normal destruct tactic throws out the variable names in exists clauses, this one mostly preserves them *)
Ltac destructe H := match (type of H) with ex(fun a => _)  => destruct H as [?a H] end.
Ltac destructes H := destructe H; repeat destructe H.
(* This one also destructs conjunctions and disjunctions *)
Ltac destructec H := match (type of H) with | ex(fun a => _)  => destruct H as [?a H] | (_ /\ _) => destruct H as [?H H] | (_ \/ _) => destruct H as [H | H] end.
Ltac destructecs H := destructec H; repeat destructec H.

Ltac destructec_all := match goal with | H: ex(fun a => _) |- _  => destruct H as [?a H] | H: (_ /\ _) |- _ => destruct H as [?H H] | H: (_ \/ _) |- _ => destruct H as [H | H] end.
Ltac destructecs_all := repeat destructec_all.

Tactic Notation "reverts" hyp(H1) := (revert dependent H1).
Tactic Notation "reverts" hyp(H1) hyp(H2) := (revert dependent H2; revert dependent H1).
Tactic Notation "reverts" hyp(H1) hyp(H2) hyp(H3) := (revert dependent H3; revert dependent H2; revert dependent H1).
Tactic Notation "reverts" hyp(H1) hyp(H2) hyp(H3) hyp(H4) := (revert dependent H4; revert dependent H3; revert dependent H2; revert dependent H1).

(* Clear duplicate hypotheses and trivial equalities *)
Ltac cleanup := repeat match goal with | H1: ?T, H2: ?T |- _ => clear H1 | H: ?a=?a |- _ => clear H  end.
(* Eauto that does not apply to the last goal - useful after forwards tactic *)
Ltac fweauto := [> eauto.. |].
(* Eauto that does not apply to the first goal - useful after rewrite and apply-in tactic *)
Ltac rweauto := [> | eauto..].

(* autospecialize *)
Ltac autospecialize := repeat match goal with
| H0: ?T , H: ?T -> _ |- _ => specialize (H H0)
| H: ?x = ?x -> _ |- _ => specialize (H (eq_refl x))
end.
Ltac autospecialize_prop := repeat match goal with
| H0: ?T , H: ?T -> _ |- _ => pose proof (T: Prop) as PropCheck; clear PropCheck; specialize (H H0)
| H: ?x = ?x -> _ |- _ => specialize (H (eq_refl x))
end.
Ltac autodestruct := repeat match goal with
| H: ?P1 /\ ?P2 |- _ => destruct H
end.

Ltac hnde H := (hnf in H; destructecs H).
Ltac hndes H := (hnf in H; destructecs H; subst).
