Set Implicit Arguments.

From TLC Require Import LibVar LibLN.

(** * Lemmas about variables *)

Lemma var_decide : forall (x y : var), {x = y} + {x <> y}.
Proof.
  intros x y.
  pose proof (var_compare_eq x y).
  rewrite LibReflect.isTrue_eq_if in H.
  cases_if; auto.
Qed.

Lemma fv_decide : forall (x : var) s, {x \in s} + {x \notin s}.
Proof.
  intros. apply classicT.
Qed.

Fixpoint vars_zip_map {A} f (xs : list var) (ys : list var) (X : A) :=
  match xs, ys with
  | nil, _ => X
  | _, nil => X
  | cons x xs, cons y ys => f x y (vars_zip_map f xs ys X)
  end.

Notation "z '\[' x '->' y ']'" := (If z = x then y else z) (x at level 69, at level 7, left associativity).
Definition subst_fvar (x y z: var): var := If z = x then y else z.

(* Rewritings for if-then substitution *)
Lemma substi_same_var (v1 v2: var):
  v2 \[v1 -> v1] = v2.
Proof.
  cases_if; auto.
Qed.

Lemma substi_eq_var (v1 v2: var):
  v1 \[v1 -> v2] = v2.
Proof.
  cases_if. auto.
Qed.
Lemma substi_ne_var (v1 v2 v3: var):
  v1 <> v2 ->
  v1 \[v2 -> v3] = v1.
Proof.
  intros. cases_if. auto.
Qed.

Lemma substi_toeq_var (v1 v2: var):
  v2 \[v1 -> v2] = v2.
Proof.
  cases_if; auto.
Qed.

