(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra.
Require Import AbstractSyntax TypingMode GeneralTyping Substitution Indep.

Require Import Coq.Strings.String.

Section WithBasicMode.
Instance basic_mode: Mode := basic.

(** * A ⊥ typed term *)

Section OmegaIsBotTyped.
  Hint Resolve binds_single_eq : core.

  (** Using [null] as the field label *)
  Definition null : met_label := "null"%string.

  (** nPkg = ν (x : {null : ⊥}){null=x.null} *)
  Definition nPkg :=
    lit_obj
      (typ_rcd (dec_met null typ_top typ_top typ_bot))
      (defs_cons
         (defs_nil)
         (def_met null (trm_call (avar_b 2) null (avar_b 2)))).

  (** Ω = #<br />#
      let newPkg = ν (x : {null : ⊥}){null=x.null} in #<br />#
      nPkg.null *)
  Definition dot_omega :=
    trm_lit nPkg (trm_call (avar_b 0) null (avar_b 0)).

  (** ⊢ nPkg : μ(x:{null : ⊥}) *)
  Lemma nPkg_typed :
    ctx_empty⁏renv_empty ⊢ nPkg ∶
          typ_objlit basic_mode (typ_rcd (dec_met null typ_top typ_top typ_bot)).
  Proof.
    unfold nPkg.
    apply ty_lit_obj with
        (L:=\{}) (T:=(typ_rcd (dec_met null typ_top typ_top typ_bot))).
  intros x Hfr.
  simpl. cases_if; eauto.
  apply ty_defs_one.
  apply ty_def_met with (L:=\{x}). intros.
  assert ((open xz (open_rec 1 xc (trm_call (avar_f x) null (avar_f x)))) = (trm_call (avar_f x) null (avar_f x))).
  simpl; auto.
  rewrite H2.
  assert ((open xz (open_rec 1 xc typ_bot)) = (open x (open_rec 1 x typ_bot))).
  simpl; auto.
  rewrite H3.
  unfold ctx_obj_met.
  eapply ty_trm_call; intros; simpl; eauto.
  apply typ_indep_met_b; auto.
  Qed.

  (** ⊢ Ω : ⊥ *)
  Lemma omega_typed_bot :
    ctx_empty⁏renv_empty ⊢ dot_omega ∶ typ_bot.
  Proof.
    unfold dot_omega.
    apply (ty_trm_lit (L:=\{}) _ nPkg_typed).
    (* Prove type of the call with the type of nPkg in the context *)
    intros nPkg Hfr.
    (* Substitute the reference to nPkg *)
    simpl; cases_if.
    (* Rewrite the result type to the form that is used by method typing *)
    assert (typ_bot = (open nPkg (open_rec 1 nPkg typ_bot))).
    { simpl; auto. }
    rewrite H.
    eapply ty_trm_call with (TC:=typ_top) (TZ:=typ_top); eauto.
    (* Prove type of the method *)
    (* Note: on the line above, one would need eauto 8 to prove this automatically *)
    (* subtyping of method types *)
    simple eapply ty_var_sub.
    (* opening the object type *)
    simple eapply ty_rec_elim.
    (* subtyping to remove reference capabilities *)
    simple eapply ty_var_sub.
    2: simple eapply subtyp_and11.
    all: eauto.
  Qed.
End OmegaIsBotTyped.

End WithBasicMode.
