(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra.
Require Import AbstractSyntax TypingMode GeneralTyping Substitution Indep Incap.

Require Import Coq.Strings.String.
Require Import ReadableSyntax.

Section WithTypingMode.
Context (typing_mode: Mode).


Section ObjExamples.

  Definition fld: fld_label := "fld"%string.
  Definition A : typ_normal_label := "A"%string.
  Lemma example_fld_read_typed x y s fld A:
    x <> y ->
    x <> s ->
    s <> y ->
    ctx_empty⁏renv_empty ⊢ ❪
      Lit x = ν⟮s∶❴ⁿA∶⊤..⊤❵⟯∅⋏❴ⁿA=⊤❵ In
      Lit y = ν⟮s∶❴ᶠfld∶μ⟮s∶❴ⁿA∶⊤..⊤❵⟯..μ⟮s∶❴ⁿA∶⊤..⊤❵⟯❵⟯∅⋏❴ᶠfld=$x❵ In
      Read $y․fld
    ❫ ∶ ❪
      μ⟮s∶❴ⁿA∶⊤..⊤❵⟯
    ❫.
  Proof.
    introv Hxy Hxs Hsy.
    set (Lf := \{x} \u \{y} \u \{s}).
    eapply ty_trm_lit_obj_incap with (L:=Lf).
    eauto. simpl. apply subset_empty_l.
    intros.
    sympls.
    eapply ty_defs_one.
    eapply ty_def_typ with (L:=Lf); eauto.
    autorewrite with simpl_close. eauto.
    intros.
    autorewrite with simpl_close.
    eapply ty_trm_lit_obj_incap with (L:=Lf \u \{x0}).
    eauto. simpl. union_sub_solver.
    - (* Object *)
    intros.
    sympls.
    eapply ty_defs_one.
    repeat cases_if.
    autorewrite with simpl_close.
    repeat cases_if.
    sympls.
    repeat cases_if.

    eapply ty_def_fld.
    intros. 
    eapply SubstCommon.ctx_vis_objctx_concat; eauto.
    eapply ty_var_sub.
    eapply ty_var_var. unfold ctx_obj. eapply ctx_binds_n. notin_solve. eapply ctx_binds_e. eauto.
    - sympls. auto.
    - (* Read term *)
    intros.

    repeat cases_if.
    autorewrite with simpl_close.
    repeat cases_if.
    sympls.
    repeat cases_if.
    sympls.
    repeat cases_if.

    eapply ty_trm_read_incap.

    3: eauto.
    3: eauto.

    eauto.
    simple eapply ty_var_sub.
    eapply ty_rec_elim.
    eapply ty_var_var. eauto. auto.
    sympl.
    eapply subtyp_refl.
  Qed.

  (* Writing to a field of an object *)
  Lemma example_fld_write_typed x1 x2 y s fld A:
    x1 <> y ->
    x1 <> s ->
    x2 <> y ->
    x2 <> s ->
    x1 <> x2 ->
    s <> y ->
    ctx_empty⁏renv_empty ⊢ ❪
      Lit x1 = ν⟮s∶❴ⁿA∶⊤..⊤❵⟯∅⋏❴ⁿA=⊤❵ In
      Lit x2 = ν⟮s∶❴ⁿA∶⊤..⊤❵⟯∅⋏❴ⁿA=⊤❵ In
      Lit y = ν⟮s∶❴ᶠfld∶μ⟮s∶❴ⁿA∶⊤..⊤❵⟯..μ⟮s∶❴ⁿA∶⊤..⊤❵⟯❵⟯∅⋏❴ᶠfld=$x1❵ In
      Write $y․fld = $x2
    ❫ ∶ ❪
      μ⟮s∶❴ⁿA∶⊤..⊤❵⟯
    ❫.
  Proof.

    introv Hx1y Hx1s Hx2y Hx2s Hx1x2 Hsy.
    set (Lf := \{x1} \u \{x2} \u \{y} \u \{s}).
    eapply ty_trm_lit_obj_incap with (L:=Lf).
    eauto. simpl. apply subset_empty_l.
    intros.
    sympls.
    eapply ty_defs_one.
    eapply ty_def_typ with (L:=Lf); eauto.
    autorewrite with simpl_close. eauto.
    intros.
    autorewrite with simpl_close.
    eapply ty_trm_lit_obj_incap with (L:=Lf \u \{x}).
    eauto. simpl. union_sub_solver.
    intros.
    sympls.
    eapply ty_defs_one.
    eapply ty_def_typ with (L:=Lf); eauto.
    sympl.
    auto.
    intros.

    eapply ty_trm_lit with (L:=Lf \u \{x} \u \{x0}).
    cases_if.
    autorewrite with simpl_close.
    repeat cases_if.
    sympls.
    repeat cases_if.
    autorewrite with simpl_close.
    repeat cases_if.
    sympls.
    repeat cases_if.
    autorewrite with simpl_close.
    repeat cases_if.
    sympls.
    repeat cases_if.
    eapply ty_lit_obj with (L:=Lf \u \{x} \u \{x0}).

    - (* Object *)
    intros.
    sympls.
    eapply ty_defs_one.
    repeat cases_if.
    autorewrite with simpl_close.
    repeat cases_if.
    sympls.
    repeat cases_if.
    autorewrite with simpl_close.
    repeat cases_if.
    sympls.
    repeat cases_if.
    sympls.
    repeat cases_if.

    eapply ty_def_fld.
    intros. 
    eapply SubstCommon.ctx_vis_objctx_concat; eauto.
    eapply ty_var_sub.
    eapply ty_var_var. unfold ctx_obj. eapply ctx_binds_n. notin_solve. eapply ctx_binds_n.
    eauto.  eapply ctx_binds_e. eauto. 
    - sympls. auto.
    - (* Read term *)
    intros.
    repeat cases_if.
    * (* Normal *)
    autorewrite with simpl_close.
    repeat cases_if.
    sympls.
    repeat cases_if.
    sympls.
    repeat cases_if.
    sympls.
    repeat cases_if.

    eapply ty_trm_write.
    + (* Vis x3 *)
      eauto.
    + (* Vis x0 *)
      eauto.
    + (* Mut *)
      intros.
      typing_mode_contradiction typing_mode.
    +

    simple eapply ty_var_sub.
    eapply ty_rec_elim.
    eapply ty_var_sub.
    eapply ty_var_var. eauto. eapply subtyp_and11. auto.
    sympl.
    eapply subtyp_refl.
    +

    simple eapply ty_var_sub.
   
    eapply ty_var_var. eauto. auto.
    * (* Rodot *)

    repeat cases_if.
    autorewrite with simpl_close.
    repeat cases_if.
    sympls.
    repeat cases_if.
    sympls.
    repeat cases_if.
    sympls.
    repeat cases_if.

    eapply ty_trm_write.
    + (* Vis x3 *)
      eauto.
    + (* Vis x0 *)
      eauto.
    + (* Mut *)
      intros.
      simple eapply ty_var_sub.
      eapply ty_var_var. eauto. eapply subtyp_and12.
    +

      simple eapply ty_var_sub.
      eapply ty_rec_elim.
      eapply ty_var_sub.
      eapply ty_var_var. eauto. eapply subtyp_and11. auto.
      sympl.
      eapply subtyp_refl.
    +
      simple eapply ty_var_sub.
      eapply ty_var_var. eauto. auto.
  Qed.

  (* Value literals *)
  Definition ValueTop A s := ❪ν⟮s∶❴ⁿA∶⊤..⊤❵⟯∅⋏❴ⁿA=⊤❵❫.
  Definition ValueBot A s := ❪ν⟮s∶❴ⁿA∶⊥..⊥❵⟯∅⋏❴ⁿA=⊥❵❫.
  Definition ValueTyp A s := ❪μ⟮s∶❴ⁿA∶⊤..⊤❵⟯❫.

  Definition CommonInterface A m s r z := ❪μ⟮s∶ ❴ⁿA∶⊥..⊤❵ ∧ ❴ᵐ m ⟮r ∶ ⊤ , z ∶ $s ․ⁿ A ⟮ $s ⟯ ⟯ ∶ ⊤ ❵⟯ ❫.
  Definition Mut := ❪ ❴ᵗ 𝖬  ∶ ⊥ .. ⊥❵ ❫.

  Definition MutableObjectRecord A m s r z := ❪ ❴ⁿA∶ {Mut}..{Mut}❵ ∧ ❴ᵐ m ⟮r ∶ ⊤ , z ∶ $s ․ⁿ A ⟮ $s ⟯ ⟯ ∶ ⊤ ❵ ❫.
  Definition MutableObjectType A m s r z := ❪μ⟮s∶ {MutableObjectRecord  A m s r z} ⟯ ❫.
  Definition MutableObject A m s r z a vb := ❪ ν⟮s∶{MutableObjectRecord A m s r z }⟯∅⋏❴ⁿA={Mut}❵⋏❴ᵐ m⟮r,z⟯ =  Write $r․a = $vb ❵❫.

  Definition ImmutableObjectRecord A m s r z := ❪ ❴ⁿA∶ ⊤..⊤❵ ∧ ❴ᵐ m ⟮r ∶ ⊤ , z ∶ $s ․ⁿ A ⟮ $s ⟯ ⟯ ∶ ⊤ ❵ ❫.
  Definition ImmutableObjectType A m s r z := ❪μ⟮s∶ {ImmutableObjectRecord  A m s r z} ⟯ ❫.
  Definition ImmutableObject A m s r z a := ❪ ν⟮s∶{ImmutableObjectRecord A m s r z }⟯∅⋏❴ⁿA=⊤❵⋏❴ᵐ m⟮r,z⟯ = Read $r․a ❵❫.

  Definition ContainerRecord A m s r z o := ❪ ❴ⁿA∶ $o․ⁿA⟮$o⟯ .. $o․ⁿA⟮$o⟯ ❵ ∧ ❴ᵐ m ⟮r ∶ ⊤ , z ∶ $s ․ⁿ A ⟮ $s ⟯ ⟯ ∶ {ValueTyp A s} ❵ ❫.


End ObjExamples.

End WithTypingMode.