Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax TypingMode GeneralTyping Equivalence
        InertTightSubtyping  Indep
        GeneralToTight TightTyping InvertibleTyping
        RecordTypes RecordDefs InertContext PreciseTyping
        OperationalSemantics Substitution UnboundSubstitution Weakening
        HeapCorrespondence HeapUpdate HeapValid CanonicalForms
         Subenvironments Narrowing
        EnvCorrespondence EWeakening  TypeApproximation Shortening RestrictedTightSubtyping
        ConfigTyping ConfigVarkinds TFree Deref ReferenceTypes
        TypedOperationalSemantics MutableReachability Defs TFreeTyping
        TypedPreservation TypedSoundness HeapPreservation.

(** * Immutability guarantee *)

Section WithRodotMode.
Context (typing_mode: Mode).
Context (Hmm: mode_has_mutability typing_mode).

(* If a reference is mutable, then its mutability in the context is ⊥. *)
Lemma reference_writeable_inv G e x Tobj Tmut:
  inert G -> ok e -> renv_corr G e ->
  G⁏e ⊢# x ∶ (typ_mut_marker typ_bot) ->
  ctx_binds x (typ_objref Tobj (typ_mut_marker Tmut)) G ->
  (forall y, G⁏e ⊢# (open y Tmut) <: typ_bot).
Proof.
  introv Hin Hoke Hcor Hty Hbi.
  (* Precise typing from tight typing *)
  eapply tight_to_invertible in Hty.
  forwards Hsp: sel_premise Hin Hty.
  destructc Hsp as Tl Tu U Hpf Hl Hu.
  (* Precise typing from context *)
  eapply pf_bind in Hbi.
  eapply pf_and2 in Hbi.
  (* Equate *)
  destruct (pf_inert_unique_tight_bounds Hin Hbi Hpf).
  subst.
  all: assumption.
Qed.

Lemma binds_ro_not_mut G e x T :
  inert G -> sealed G -> renv_corr G e -> ok e ->
  ctx_binds x (typ_and (typ_bnd T) (typ_mut_marker typ_top)) G ->
  not (G ⁏ e ⊢ trm_var (avar_f x) ∶ (typ_mut_marker typ_bot)).
Proof.
  introv HinG HsG Hcor Hoke Hbi. hnf. introv Hbot.
  eapply ty_trm_var_inv in Hbot.
  eapply general_to_tight in Hbot; eauto.
  eapply reference_writeable_inv with (y:=x) in Hbot; eauto. simpl in Hbot.
  assert (Htyx : G ⁏ e ⊢# x ∶ typ_bot) by eauto.
  Unshelve. 2: exact typ_bot.
  eapply tight_to_invertible in Htyx; eauto.
  eapply invertible_bot_false in Htyx; eauto.
Qed.

Lemma binds_ro_not_mreach G t fs Sigma e Lv :
  sealed G -> ok e -> renv_corr G e ->
  (forall w, tfree w t \/ tfree w fs ->
        exists Tw, ctx_binds
                w (typ_and (typ_bnd Tw) (typ_mut_marker typ_top)) G) ->
  (forall y, not (G ⊢ (t ; fs ; Sigma ; e ; Lv) mreach y)).
Proof.
  intros HsG Hoke Hcor Hro y. remember (t ; fs ; Sigma ; e ; Lv) as C.
  induction 1; eauto. inversions HeqC. specialize (Hro w H0) as [Tw Hbiw].
  eapply binds_ro_not_mut with (e:=e); eauto.
Qed.

(* Lemma 5.? (MAdapt) *)
Lemma reference_writeable_adapt_inv G e x Tobj Tmut1 Tmut2:
  inert G -> ok e -> renv_corr G e ->
  G⁏e ⊢# x ∶ (typ_mut_marker typ_bot) ->
  ctx_binds x (typ_objref Tobj (typ_mut_marker (typ_or Tmut1 Tmut2))) G ->
  (forall y, G⁏e ⊢# (open y Tmut1) <: typ_bot) /\ (forall y, G⁏e ⊢# (open y Tmut2) <: typ_bot).
Proof.
  introv Hin Hoke Hcor Hty Hbi.
  pose proof (reference_writeable_inv Hin Hoke Hcor Hty Hbi).
  (* Follows by subtyping *)
  eauto.
Qed.

Lemma reference_dom_mut_top G e x:
  inert G ->
  x \in (ctx_dom G) ->
  G⁏e ⊢# x ∶ (typ_mut_marker typ_top).
Proof.
  introv Hin Hdom.

  eapply ctx_dom_to_binds in Hdom.
  destructc Hdom as t ?.

  eapply ty_var_mut_top_t.
  auto.
  eapply ty_var_t.

  all: eauto.
Qed.

Lemma reference_select_writeable_inv G e x:
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  x \in (ctx_dom G) ->
  (forall y, G⁏e ⊢ (typ_sel (avar_f x) (typ_label_refcap typ_refcap_mutability) (avar_f y)) <: typ_bot) ->
  G⁏e ⊢ x ∶ (typ_mut_marker typ_bot).
Proof.
  introv Hin Hsl Hoke Hcor Hdom Hsub.

  set (M := (typ_label_refcap typ_refcap_mutability)).

  assert (G⁏e ⊢# x ∶ (typ_mut_marker typ_top)) as Hty by eauto using reference_dom_mut_top.

  eapply tight_to_invertible in Hty; auto 1.
  eapply sel_premise in Hty; auto 1.
  destructc Hty as TL TU V Hpf HU HL. clear HU HL.

  assert  (forall y, G⁏e ⊢ (open y TU) <: typ_bot). {
  intro.

  pose proof (Hsub y) as Hsubo.
  eapply general_to_tight in Hsubo; eauto.
  eapply typ_approx_tight_to_s with (δ := ⊕) in Hsubo; eauto.
  destructc Hsubo as T3 T4 Haps Hapb Hsubo.

  inversions_typ Hapb.

  assert (G ⊢ typ_bot ↝ˢ⦇⊕⦈ typ_bot) as Hapbot by eauto.

  assert (forall y, G ⊢ typ_sel (avar_f x) M (avar_f y) ↝ˢ⦇⊕⦈ (open y TU)). {
    intro. eapply typ_approx_s_sel_u; eauto.
  }

  specialize (H y).

  destructc (typ_approx_s_com Hin Hsl Haps H) as T1 Hap31 HapU1.
  destructc (typ_approx_s_subcom Hin Hsl Hoke Hcor Hsubo Hap31 Hapbot) as T5 T6 Hap15 Hapbot6 Hsub56.

  inversions_typ Hapbot6.
  eapply typ_approx_s_sub with (e0:=e) in Haps; auto 1.
  eapply typ_approx_s_sub in HapU1; auto 1.
  eapply typ_approx_s_sub in Hap15; auto 1.


  eapply subtyp_trans.
  eapply subtyp_trans.
  eapply subtyp_restricted_to_general; eauto.
  eapply subtyp_restricted_to_general; eauto.
  eapply subtyp_restricted_to_general; eauto.
  }
  eapply precise_to_general in Hpf; eauto 1.
  eapply ty_var_sub.
  eapply Hpf. eauto.
Qed.


(* Lemma 19 (WMu) in "Reference mutability for DOT" *)
Lemma wmu G e y2 w2 T2 w1:
  inert (G § w2 / v_ref ~ T2) ->
  sealed (G § w2 / v_ref ~ T2) ->
  w2 \notin ctx_fv G ->
  w1 \in (ctx_dom G) ->
  ok e ->
  renv_notin w2 e ->
  renv_corr (G § w2 / v_ref ~ T2) (e & w2 ~ y2) ->
  w1 <> w2 ->
  (forall y, G § w2 / v_ref ~ T2⁏e & w2 ~ y2 ⊢ (typ_sel (avar_f w1) (typ_label_refcap typ_refcap_mutability) (avar_f y)) <: typ_bot) ->
  G⁏e ⊢ w1 ∶ (typ_mut_marker typ_bot).
Proof.
  introv Hin Hsl Hnw2G Hw1G Hoke Hnw2e Hcor Hnw Hsub.

  assert (Hfrw2 : w2 # e).
  { hnf in Hnw2e. rewrite in_union in Hnw2e. tauto. }
  destructc (renv_corr_push_inv Hfrw2 Hin Hcor) as ? ?.

  assert (G § w2 / v_ref ~ T2⁏e & w2 ~ y2 ⊢ w1 ∶ (typ_mut_marker typ_bot)). {
    eapply reference_select_writeable_inv; eauto.
    rewrite ctx_dom_push, in_union. tauto.
  }

  eapply shortening_typ_ref; eauto.
Qed.

Lemma write_typ_inv F e x a y T:
  F ⁏ e ⊢ trm_write (avar_f x) a (avar_f y) ∶ T ->
  F ⁏ e ⊢ x ∶ (typ_mut_marker typ_bot) /\ (mode_has_visibility typing_mode -> ((ctx_vis F x) /\ (ctx_vis F y))).
Proof.
  remember (trm_write (avar_f x) a (avar_f y)) as t.
  induction 1; inversions Heqt; eauto. repeat(split).
  - eapply H1; auto.
  - eapply H; auto.
  - eapply H0; auto.
Qed.

(* Lemma 10 in "Reference mutability for DOT" *)
Lemma mutated_mreach t1 s1 S1 e1 Lv1 F1 T t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ (t1 ; s1 ; S1 ; e1 ; Lv1) ∶ T ↦ F2 ⊢ (t2 ; s2 ; S2 ; e2 ; Lv2) -> (
      binds y (item_obj ds) S2 \/
      F1 ⊢ (t1 ; s1 ; S1 ; e1 ; Lv1) mreach y
    ).
Proof.
  introv Hbi Hred.
  (* In most cases, heap is not changed *)
  inversions Hred; auto.
  - (* write *)
    destruct (classicT (y = y1)).
    + (* changing the object - y is mreach*)
      subst.
      right.
      clear Hupd.
      eapply mreach_trm.
      3: eauto.
      eauto.
      eauto.
      inversions Htyc.
      invert_focus_typing H8.
      eauto.
    + (* not changing the object *)
      left.
      eapply heap_update_unchanged; eauto.
  - (* obj literal *)
    left.
    eapply binds_push_neq; eauto.
    (* prove that y is not the new variable by contradiction *)
    intro; subst.
    eapply binds_to_dom in Hbi.
    notin_solve.
    destruct H.
    rewrite in_union; right.
    rewrite in_union; right.
    rewrite in_union; left.
    rewrite in_union; left.
    auto.
  -
  (* function literals are not possible in rodot *)
typing_mode_contradiction typing_mode.
Qed.

(* Lemma 12 in "Reference mutability for DOT" *)
Lemma mreach_preserve_vars t1 s1 S e Lv1 F t2 s2 Lv2 y:
    F ⊢ (t1 ; s1 ; S ; e ; Lv1) mreach y ->
    (forall x,
      (tfree x t1 \/ tfree x s1) ->
      (tfree x t2 \/ tfree x s2)
    ) ->
    F ⊢ (t2 ; s2 ; S ; e ; Lv2) mreach y.
Proof.
  remember (t1 ; s1 ; S ; e ; Lv1) as C1.
  induction 1; subst; inversions HeqC1; introv Htf.
  - (* mreach_fld *) eapply mreach_fld; eauto.
  - (* mreach_trm *) eapply mreach_trm; eauto.
Qed.

(* If object defs pushed on the heap have a field with a location,
   then the defs in the literal have the same field with a corresponding reference
*)
Lemma deref_defs_has_epush_open_neq_inv2 e1 w1 y1 d a y:
  y <> y1 -> (* Assuming that the value of the field is not the new object itself *)
  ~(tfree y d) -> (* The value of the field is not in the literal defs - we can assume this because terms cannot contain location values *)
  w1 \notin (fv d) -> (* The location and reference to the new object are sufficiently fresh *)
  y1 # e1 ->
  ok (e1 & w1 ~ y1) ->
  renv_idempotent (e1 & w1 ~ y1) ->
  defs_has (deref_defs (e1 & w1 ~ y1) (open w1 d)) (def_fld a (avar_f y)) ->
  exists w,
    binds w y e1 /\ defs_has d (def_fld a (avar_f w)).
Proof.
  introv Hny Hnyd Hnwd Hnye Hoke Hide Hhas.
  (* Find the same field in the literal defs *)
  destructc (deref_defs_has_epush_open_neq_inv _ _ _ Hny Hhas) as a0 Hde Hhasd.
  clear Hhas.
  (* The reference in the field in literal defs is not the new reference *)
  assert (a0 <> avar_f w1). {
    intro; subst.
    eapply defs_has_fv in Hhasd. contradiction.
  }
  (* The new reference and location do not play a role *)
  rewrite deref_avar_open2 in Hde by auto.
  (* Is the variable in the literal a self-reference ? *)
  destruct (avar_open_cases a0 0 y1) as [[? ?] | [? ?]].
  * (* It is not a self-reference *)
    rewrite H0 in Hde.
    (* Induction on the environment *)
    induction e1 using env_ind.
    - (* Environment cannot be empty *)
      unfold deref in Hde. unfold deref_avar, deref_base in Hde. rewrite empty_def in Hde. subst.
      eapply defs_has_tfree in Hhasd. contradiction.
    - simpl_dom.
      rewrite notin_union in Hnye.
      destruct Hnye.
      specialize (IHe1 H3 (ok_remove Hoke)).
      rewrite notin_singleton in H2.
      (* Is the current variable the value of the fiels in the literal defs? *)
      destruct (avar_subst_cases a0 x v) as [[? ?] | [? ?]].
      + (* No. It is not affected, continue by induction. *)
        assert (deref e1 a0 = avar_f y).
        rewrite <- Hde.
        unfold deref, deref_avar. rewrite deref_base_push. rewrite H4. auto.
        unfold subst_var, AvarSubstVar in Hde.
        forwards Hex: IHe1.
        eapply renv_idempotent_concat_middle_inv; eauto. auto.
        destructc Hex as ? ? ?.
        exists x0.
        split; eauto.
        eapply binds_push_neq; auto. intro; subst.
        eapply binds_to_dom in H7.
        eapply ok_middle_inv_l in Hoke. auto.
      + (* Yes, this is the variable *)
        assert (v # e1) as Hvne. {
          intro.
          forwards Hine: (Hide v).
          simpl_dom. repeat rewrite in_union; auto.
          apply Hine. rewrite env_concat_fv. rewrite in_union. left. rewrite env_push_fv.
          rewrite in_union. right. compute. rewrite in_singleton. auto.
        }
        assert ((deref e1 (avar_f v)) = (avar_f v)) as Hdv. {
          eapply deref_notin_dom; auto.
        }
        subst. rewrite deref_avar_comm in Hde. inversion Hde.
        rewrite deref_avar_comm in Hdv. inversion Hdv.
        repeat rewrite H7.
        unfold deref, deref_var in H6.
        rewrite deref_base_push in H6.
        rewrite subst_eq_var in H6.
        unfold deref, deref_var in H7.
        rewrite H7 in H6; subst.
        inversion Hde; repeat rewrite H6.
        exists x. eauto.
  * (* It is a self-reference, but we know it cannot be the object itself  *)
    rewrite H0 in Hde.
    rewrite deref_notin_dom in Hde by auto.
    inversions Hde.
    contradiction.
Qed.

(* Lemma 18 (MUSub) in "Reference mutability for DOT" *)
Lemma musub G e T Tcap:
  G ⁏ e ⊢ T ∶cpb ⟦ typ_refcap_mutability ⟧ Tcap ->
  ((forall y, G ⁏ e ⊢ typ_top <: (open y Tcap)) \/ G ⁏ e ⊢ T <: (typ_mut_marker Tcap)).
Proof.
  introv Hcpb.
  induction Hcpb; eauto.
  - destruct IHHcpb; auto. right.
    eapply subtyp_trans.
    eapply subtyp_sel1. eauto. eauto. (*  eapply ty_var_sub. apply H. eapply subtyp_typ; eauto. eapply subtyp_refl.*)
  - destruct IHHcpb; destruct IHHcpb0; sympls; eauto.
    + right.
      eapply subtyp_trans. eapply subtyp_and12. eapply subtyp_trans. eapply H2. eapply subtyp_typ; eauto.
      intro. eapply subtyp_and2; eauto. eapply subtyp_trans with (T:= typ_top); eauto.
    + right.
      eapply subtyp_trans. eapply subtyp_and11. eapply subtyp_trans. eapply H1. eapply subtyp_typ; eauto.
      intro. eapply subtyp_and2; eauto. eapply subtyp_trans with (T:= typ_top); eauto.
    + right. eapply subtyp_trans.
      eapply (@subtyp_and _ (and_or_subtyp _)); eauto. eapply subtyp_trans. eapply subtyp_typ_and. eapply subtyp_typ; eauto.
  - destruct IHHcpb; destruct IHHcpb0; sympls; eauto.
    + left; intro; specialize (H1 y); specialize (H2 y); eauto.
    + left; intro; specialize (H1 y); eauto.
    + left; intro; specialize (H2 y); eauto.
    + right. (* Lemma 5.20 (OrTypSub) *)
      eapply subtyp_or2.
      eapply subtyp_trans; eauto. eapply subtyp_typ; intro; sympls; eauto.
      eapply subtyp_trans; eauto. eapply subtyp_typ; intro; sympls; eauto.
Qed.

Ltac invert_ty_config :=  inversion Htyc as [? ? ? ? ? ? ? ? HinG Hhc Htys' Htfs Hty Htft Hfv Hnull HclG Hclc HclT]; subst; clear HinG.
Ltac invert_ty_config2 :=  inversion Htyc2 as [? ? ? ? ? ? ? ? HinG2 Hhc2 Htys'2 Htfs2 Hty2 Htft2 Hfv2 Hnull2 HclG2 Hclc2 HclT'2]; subst.

Theorem mreach_preservation_read a f s1 S1 e1 Lv1 F1 T t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ ((trm_read a f) ; s1 ; S1 ; e1; Lv1) ∶ T ↦ F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) mreach y ->
    F1 ⊢ ((trm_read a f) ; s1 ; S1 ; e1; Lv1) mreach y.
Proof.
  introv Hbi Hred Hmr.

  assert (ty_config F2 (t2 ; s2 ; S2 ; e2; Lv2) T) as Htyc2. {
    eapply typed_preservation.
    eapply Hred.
  }

  inversions Hred; inversions H7; rename a0 into a.
  - (* read object *)
    destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hmva. typing_mode_contradiction typing_mode.
    hnf in Htcor.
    destructc Htcor as Tstrip Tcap Hicp Hcpb Htys HT2.
    remember (trm_read (avar_f w1) a) as t.
    remember ((trm_var (avar_f w2)) ; s2 ; S2 ; (e1 & w2 ~ y2);(fv (avar_f w1) \u fv s2 \u fv_heap S2 \u fv_renv e1 \u Lv1)) as C.
    remember (F1 § w2 / v_ref ~ T2) as F2.
    reverts Hmr. intro Hmr.
    clear Hbi.
    induction Hmr.
    + (* fld - by induction *)
      inversions HeqC.
      rename H into Hin2, H0 into Hmr2, H1 into Hbiy0S, H2 into Hhas0, H3 into Hmuty0.
      invert_ty_config.
      assert (ctx_vis F1 y0) as Hvisy by (eapply heap_correspond_heap_binds_vis; eauto).
      destructc Hhc as ? ? ? Hoke Hkdom ? Hrest.
      destructc Hrest as Hcor ?.
      destruct Hfv as [[HdomF1 HnfF1] HnfT].
      (* Show that it is reachable from the same field *)
      eapply mreach_fld with (y1 := y0); eauto.
      eapply ty_trm_var_inv in Hmuty0.
      eapply ty_trm_var.
      auto.
      (* Freshness of w2 *)
      replace (cons w2 nil )  with  (app nil (cons w2 nil)) in Hfrw2 by eauto.
      assert (y0 \in dom S2) as Hy0dS2 by (eapply binds_to_dom; eauto).
      assert (y2 \in fv yd) as Hy2fyd by (eapply defs_has_fv; eauto).
      assert (y2 \in fv_env S2) as Hy2fS2 by (eapply heap_bind_in_fv; eauto).
      assert (w2 \notin fv_env S2) as Hw2fS2 by (eapply fresh_in_config_notin_Sigma_im; eauto).
      assert (w2 \notin dom S2) as Hw2dS2 by (eapply fresh_in_config_notin_Sigma_dom; eauto).
      (* Show that the field had a mutable type *)
      eapply shortening_fld_typ_ref; eauto.
      eapply ctx_sealed_push; eauto.
      eapply closed_sealed0.  eauto.

      eapply fresh_in_config_notin_nothing_fresh; eauto.
      intro; subst. contradiction.
      intro; subst. contradiction.
      invert_ty_config2.
      destructc Hhc2 as ? ? ? ?Hoke ?Hkdom ? ?Hrest.
      destructc Hrest as ? ?. eauto.
    + (* term *)
      rename H0 into Htfree, H1 into Hbiwe, H2 into Htytrm.
      inversions HeqC.
      (* Is it the new reference in the term or an old one (on the stack)? *)
      destruct (classicT (w = w2)).
      * (* Reachable from the new referemce *)
        subst.
        set (F2 := F1 § w2 / v_ref ~ typ_and (typ_bnd R1) (typ_cap_adapt typ_refcap_mutability w1 Tcap)).
        invert_ty_config.
        assert (ctx_vis F1 y1) as Hvisy1 by (eapply heap_correspond_heap_binds_vis; eauto).

        destructc Hhc as ? ? ? Hoke ? ? Hrest. destructc Hrest as Hcor ?.
        invert_ty_config2.
        destructc Hhc2 as ? ? ? Hoke2 ? ? Hrest. destructc Hrest as Hcor2 ?.
        (* The reached location must be the new location *)
        eapply binds_push_eq_inv in Hbiwe; subst.

        eapply ty_trm_var_inv in Htytrm.
        eapply general_to_tight in Htytrm; eauto.

        assert( ctx_binds w2 (typ_and (typ_bnd R1) (typ_cap_adapt typ_refcap_mutability w1 Tcap) ) (F1 § w2 / v_ref ~ typ_and (typ_bnd R1) (typ_cap_adapt typ_refcap_mutability w1 Tcap) )) as Hbiw2 by eauto.

        destructc (reference_writeable_adapt_inv HinG2 Hoke2 Hcor2 Htytrm Hbiw2) as Hcapbot Hselbot.

        assert (w2 \notin ctx_fv F1) as Hnw2f.

        replace (cons w2 nil )  with  (app nil (cons w2 nil)) in Hfrw2 by eauto.
        destruct Hfv as [[HdomF1 HnfrF1] HnfrT].
        eapply fresh_in_config_notin_nothing_fresh; eauto.
        eapply mreach_fld. 4: apply Hhas. 3: apply Hby1S. auto.
        eapply mreach_trm with (w := w1); auto.


        -- eapply ty_trm_var. intro.  eapply inert_ty_vis in Htypw1; eauto.
          eapply wmu with (w2 := w2); auto.
          apply HinG2. apply HclG2.
          2: apply Hcor2.
          eapply typing_implies_dom in Htypw1. auto.
          intro y. specialize (Hcapbot y). specialize (Hselbot y). sympl in Hselbot. cases_if.
          eapply tight_to_general; eauto.
        --
          list_split_from_in w2 Hfrw2.
          assert (w2 \notin fv (typ_and (typ_bnd R1) (typ_cap_adapt typ_refcap_mutability w1 Tcap))).
          eapply fresh_in_config_notin_nothing_fresh; eauto.

          assert (forall y, F1 ⁏ e1 ⊢ (open y Tcap) <: typ_bot). {
            pose proof Hfrw2 as Hnw2S1.
            eapply fresh_in_config_notin_Sigma_im in Hnw2S1.
            pose proof Hfrw2 as Hnw2ed.
            eapply fresh_in_config_notin_e_dom in Hnw2ed.
            pose proof Hfrw2 as Hnw2ef.
            eapply fresh_in_config_notin_e_im in Hnw2ef.
            assert(y2 \in fv_env S2). { eapply heap_bind_in_fv. 2: apply Hby1S. eapply defs_has_fv; eauto. }

            pick_fresh z.
            assert (F1 ⁏ e1 ⊢ (open z Tcap) <: typ_bot). {
              eapply shortening_subtyp_bot_ref with (y:=y2); eauto.
              intro; subst. contradiction.
              eapply notin_fv_open. notin_solve.  notin_solve.
            }
            intro.
            replace (open y Tcap) with ((open z Tcap) /[z -> y]).
            replace (typ_bot) with (typ_bot /[z -> y]).
            eapply usubst_end1; eauto.
            simpls. auto.
            rewrite <- subst_intro; eauto.
          }

          eapply musub in Hcpb. destruct Hcpb.
          ++ (* Contradictory top<:bot *)
            eapply ty_trm_var.
            eauto.

            eapply binds_to_dom in Hby1S. rewrite <- H3 in Hby1S.

            eapply ctx_in_kdom_dom in Hby1S. eapply ctx_dom_to_binds in Hby1S. destruct Hby1S. eapply ty_var_sub. eauto.
            eapply subtyp_trans with (T:=(open y1 Tcap)).
            eapply subtyp_trans with (T:=typ_top); eauto.
            eapply subtyp_trans with (T:=typ_bot); eauto. eauto. eauto.

          ++
            eapply ty_trm_var.
            eauto.
            eapply ty_renv_binds_var; eauto.

      * (* From an old reference *)
        destruct Htfree.
        inversions H0; try contradiction.
        invert_ty_config2. inversions HinG2.
        eapply mreach_trm; eauto.
        eapply binds_push_neq_inv; eauto.
        eapply ty_trm_var_inv in Htytrm.
        invert_ty_config.
        assert (ctx_vis F1 w) as Hvisw by (eapply heap_correspond_renv_push_binds_vis; eauto).
        destructc Hhc as ? ? ? ? HkdomS Hkdom ?.

        destructc Hhc2 as ? ? ? ? ? ? Hrest.
        destructc Hrest as Hcor2 Hrest.
        destruct Hfv as [[HdomF1 HnfrF1] HnfrT].
        assert (y2 \in (dom S2)). {
          eapply ctx_binds_kdom in Hby2G.
          rewrite HkdomS in Hby2G. auto.
        }

        list_split_from_in w2 Hfrw2.
        forwards: fresh_in_config_notin_Sigma_dom; eauto 1.

        eapply ty_trm_var.
        eauto.
        eapply shortening_typ_ref with (y:=y2).
        6: apply n.
        all: eauto.

        hnf. intros. subst. auto.
  - (* read function*)
(* functions not supported in rodot. *)
    typing_mode_contradiction typing_mode.
Qed.

Theorem mreach_preservation_write a f a0 s1 S1 e1 Lv1 F1 T t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ ((trm_write a f a0) ; s1 ; S1 ; e1; Lv1) ∶ T ↦ F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) mreach y ->
    F1 ⊢ ((trm_write a f a0) ; s1 ; S1 ; e1; Lv1) mreach y.
Proof.
  introv Hbi Hred Hmr.

  assert (ty_config F2 (t2 ; s2 ; S2 ; e2; Lv2) T) as Htyc2. {
    eapply typed_preservation.
    eapply Hred.
  }

  inversions Hred; inversions H7; rename a1 into a.
  - (* write *)
    remember (trm_var (avar_f w3)) as t.
    remember ((fv (avar_f w1) \u fv (avar_f w3)) \u fv s2 \u fv_heap S1 \u fv_renv e2 \u Lv1) as Lv2.
    remember (t ; s2 ; S2 ; e2; Lv2) as C.
    clear Hbi.
    clear Hin.
    revert dependent Hmr. intro Hmr.

    induction Hmr.
    rename H into Hin.
    rename H0 into Hmr0.
    rename H1 into Hby0.
    rename H2 into Hhas.
    rename H3 into Htyp0.
    + inversions HeqC.
      destruct (classicT (y2 = y3)).
      * subst.
        destruct (classicT (y0 = y1)).
        ** (* From the updated object *)
          subst.
          destruct (classicT (a = a0)); subst.
          *** (* Reached from the updated field. Show that previously, it was reached from the focus *)
          eapply mreach_trm.
          auto.
          2: apply Hbw3y3e.
          left. eauto.
          (* Show that the reference must have been mutable (because the field is mutable) *)
          invert_ty_config.
          invert_focus_typing Hty.
          rename H2 into Htypw1.
          destructc Hhc as HokG HwkG HokS Hoke ? ? Hrest.
          destructc Hrest as Hcor ?.
          eapply ty_trm_var_inv in Htyp0.
          pose proof (var_typ_rcd_to_binds Hin HclG Hoke Hcor Htyp0) as HexLoc.
          pose proof (var_typ_rcd_to_binds Hin HclG Hoke Hcor Htypw1) as HexRef.
          destructc HexLoc as ? ? TfldLoc Hbiy1G Hhasy1a0 ? HTfldLoc.
          destructc HexRef as ? ? TfldRef ? Hhasw1a0 ? HTfldRef.
          assert (x1 = x). { rename Hcor into Hcor0.
              destruct_renv_corr_binds_to_ref_loc HokG Hcor0 Hbw1y1e TR0 TL0.
              ctx_binds_eq.
              ctx_binds_eq. inversions Hcor. auto.
        }
          subst.
          assert (typing_equiv e TfldRef TfldLoc). {
            eapply (record_has_tyq _ _ Hhasy1a0 Hhasw1a0). auto.
            Unshelve.
            eapply inert_binds_inert in Hbiy1G.
            inversions Hbiy1G.
            inversions H9.

            eauto. apply Hin.
            eapply inert_binds_inert in Hbiy1G.
            inversions Hbiy1G.
            inversions H9.
            auto. apply Hin.
          }
          eapply ty_trm_var; auto.
          eapply ty_var_sub; eauto.
          eapply subtyp_trans; eauto.
          eapply subtyp_trans. eapply subtyp_tyq; eauto. eauto.
          *** (* From different field of same object *)
          destructc (heap_update_nfld Hupd Hby0 Hhas n) as ? ? ?.
          eapply mreach_fld. auto.
          eapply IHHmr; eauto.
          apply H.
          apply H0.
          apply Htyp0.
        ** (* Reached from a field of a different object *)
          eapply mreach_fld.
          auto.
          eapply IHHmr; eauto.
          eapply heap_update_rev_inv; eauto.
          apply Hhas. apply Htyp0.
      * (* The field value differs *)
        pose proof (heap_update_inv_n Hupd Hby0 Hhas n) as Hui.
        destructc Hui as ds1 ? ?.
        eapply mreach_fld.
        auto.
        eapply IHHmr; eauto.
        apply H. apply H0.
        auto.
    + inversions HeqC.
      destruct H0.
      * inversions H0.
        eapply mreach_trm; eauto.
      * eapply mreach_trm; eauto.

Qed.

Theorem mreach_preservation_call a m a0 s1 S1 e1 Lv1 F1 T t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ ((trm_call a m a0) ; s1 ; S1 ; e1; Lv1) ∶ T ↦ F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) mreach y ->
    F1 ⊢ ((trm_call a m a0) ; s1 ; S1 ; e1; Lv1) mreach y.
Proof.
  introv Hbi Hred Hmr.

  assert (ty_config F2 (t2 ; s2 ; S2 ; e2; Lv2) T) as Htyc2. {
    eapply typed_preservation.
    eapply Hred.
  }

  inversions Hred; inversions H7; rename m0 into m.
  - (* call *)
    eapply mreach_preserve_vars; eauto.
    (* show that tfree variables are preserved *)
    intros.
    destruct H.
    + (* a variable in the body *)
      invert_ty_config.
      (* Code is the same as in preservation *)
      invert_focus_typing Hty.
      rename H2 into Htypw1M, H4 into Htypw1C.
      pose proof (canonical_forms_met_ref Hin HclG Hhc Hbw1ye Htypw1M) as Hcf.
      destruct Hcf as [L [yd' [t' [S [S' [S'' [HbniG [HblG [Hbie [Hhas' [Hsu [Htf' Hop]]]]]]]]]]]].
      binds_eq.
      apply defs_has_met_inv with (t := t') in Hhas; eauto.
      inversions Hhc; eauto. clear H4.
      pick_fresh xz. rename Fr into Hfrxz.
      pick_fresh xc. rename Fr into Hfrxc.
      assert (Hfr : xz \notin L) by eauto; specialize (Hop _ Hfr);
      clear Hfr.
      assert (Hfr : xc <> xz) by eauto; specialize (Hop _ Hfr); clear Hfr.
      assert (Hfr : xc \notin L) by eauto; specialize (Hop Hfr); clear Hfr.
      assert (Hfrzcop : xz \notin fv (open_rec 1 xc TR) \u
                           fv (open_rec 1 xc t)) by
          (eapply notin_union_l;
           match goal with
           | [ |- ?xc \notin fv (open_rec ?n ?w ?T) ] =>
             replace (open_rec n w T) with
                 (open_rec_vars n (w :: nil)%list T); eauto;
             eapply notin_open_rec_vars_list;
             try notin_solve;
             rewrite from_list_cons, from_list_nil, union_empty_r; eauto
           end).
      unfold ctx_obj in Hop.
      (* end same code *)
      assert ((x = w1) \/ (x = w2)).
      eapply tfree_met_renaming with (xc:=xc) (xz:=xz); eauto.
      notin_solve.
      notin_solve.
      destruct (mode_visibility' typing_mode) eqn: Hmv. typing_mode_contradiction typing_mode.
      eapply tfree_met_params; eauto.
      destruct H4; subst.
      eauto.
      eauto.
    + (* a variable on the stack *)
      eauto.
Qed.

Theorem mreach_preservation_apply a a0 s1 S1 e1 Lv1 F1 T t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ ((trm_apply a a0) ; s1 ; S1 ; e1; Lv1) ∶ T ↦ F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) mreach y ->
    F1 ⊢ ((trm_apply a a0) ; s1 ; S1 ; e1; Lv1) mreach y.
Proof.
  introv Hbi Hred Hmr.

  assert (ty_config F2 (t2 ; s2 ; S2 ; e2; Lv2) T) as Htyc2. {
    eapply typed_preservation.
    eapply Hred.
  }

  inversions Hred; inversions H7.

  - (* apply *)
    (* functions not supported in rodot. *)
    typing_mode_contradiction typing_mode.
Qed.

Lemma heap_loc_not_tfree_lit_obj F1 e1 S1 R d t y:
  heap_correspond F1 e1 S1 ->
  all_tfree_k_trm F1 (trm_lit (lit_obj R d) t) v_ref ->
  y \in dom S1 ->
  ~tfree y d.
Proof.
  introv Hhc Htft HyS.
  intro Htfyd.
  assert (all_tfree_k_defs F1 d v_ref) as Htfd. {
     intro. intro. apply Htft. eauto.
  }
  destructc Hhc as ? ? ? ? HdomS1 ? ?.
  rewrite <- HdomS1 in HyS.
  specialize (Htfd y Htfyd).
  eapply ctx_in_kdom_k in HyS; eauto.
  eapply (ctx_k_ne_false _ HyS Htfd).
  Unshelve. discriminate.
Qed.

Theorem mreach_preservation_lit l t1 s1 S1 e1 Lv1 F1 T t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ ((trm_lit l t1) ; s1 ; S1 ; e1; Lv1) ∶ T ↦ F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) mreach y ->
    F1 ⊢ ((trm_lit l t1) ; s1 ; S1 ; e1; Lv1) mreach y.
Proof.
  introv Hbi Hred Hmr.

  assert (ty_config F2 (t2 ; s2 ; S2 ; e2; Lv2) T) as Htyc2. {
    eapply typed_preservation.
    eapply Hred.
  }

  inversions Hred; inversions H7.

  - (* object literal *)
    (* assert that the location already is in the heap *)
    assert (y \in (dom S1)) as HyS.
    { eapply binds_to_dom; eauto. }
    clear Hbi.
    match goal with
      H: ?F20 ⊢ (?t20 ; ?s2 ; ?S20 ; ?e20 ; ?Lv20) mreach _ |- _ =>
        remember F20 as F2;
        remember t20 as t2;
        remember S20 as S2;
        remember e20 as e2;
        remember Lv20 as Lv2
    end.
    remember (t2 ; s2 ; S2 ; e2 ; Lv2) as C.
    destruct (mode_refcaps' typing_mode) eqn: Hmrc. typing_mode_contradiction typing_mode.

    induction Hmr.
    + (* field *)
      rename H3 into Hy1hasmuta.
      inversions HeqC.
      destruct (classicT (y1 = y0)).
      * (* new item *)
        subst.
        destruct (classicT (y0 = y2)).
        ** subst.  (* contradiction *)
           contradict HyS.
           list_split_from_in y2 Hfr.
           eapply fresh_in_config_notin_Sigma_dom; eauto.
        **
          rename y2 into y. rename y0 into y1.
          invert_ty_config.
          invert_ty_config2.
          eapply binds_push_eq_inv in H1.
          inversions H1.

          assert (~tfree y d) as Hntf by eauto using heap_loc_not_tfree_lit_obj.
          assert (w \notin fv d) as Hnwd. {
            list_split_from_in w Hfr.
            eapply fresh_in_config_notin_t in Hfr. intro. apply Hfr. compute. compute in H1. repeat rewrite in_union. auto.
          }
          assert (y <> y1) as Hnyy. auto.
          assert (y1 # e1) as Hnye. {
            list_split_from_in y1 Hfr.
            eapply fresh_in_config_notin_e_dom in Hfr. auto.
          }
          assert ( ok (e1 & w ~ y1)) as Hoke2. eauto.
          assert ( renv_idempotent (e1 & w ~ y1)) as Hide2. eapply renv_corr_idempotent; eauto.
          destructc Hhc2 as ? ? ? ? ? Hdomr Hrest. rewrite Hdomr. intro. auto.
          destructc (deref_defs_has_epush_open_neq_inv2 Hnyy Hntf Hnwd Hnye Hoke2 Hide2 H2) as x Hbixe1 Hhasda.
          unfold mode_refcaps' in *.

            eapply mreach_trm.  eauto. left. eapply defs_has_tfree_lit. apply Hhasda. auto.
          

          set (F2:=inert_G_for_obj F1 w y1 R (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))).
          assert(F2 ⁏ (e1 & w ~ y1) ⊢ x ∶ typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) as Hmutx. {

            destructc Hhc2 as ? ? ? ? ? ? Hrest. destructc Hrest as ? ?.

            eapply ty_trm_var_inv in Hy1hasmuta.
            eapply var_typ_rcd_to_binds in Hy1hasmuta.
            destructc Hy1hasmuta as ? ? ? Hbiy1 Hhasy1a Hbotx2 Hx2mut.
            eapply ctx_binds_push_inv_n in Hbiy1.
            eapply ctx_binds_push_inv_e in Hbiy1. inversions Hbiy1.

            eapply ty_trm_lit_obj_inv in Hty. destructc Hty as L0 HtydR.

            eapply (defs_has_record_has_ty_obj_fld) in HtydR.

            2: apply Hhasda.
            2: apply Hhasy1a.
            5: auto. 4: auto. 3: auto.
            eapply ty_var_sub.
            2: apply Hx2mut.
             eapply ewkn_push.
            unfold F2. eapply weaken_push. eapply HtydR. eauto. inversions HinG2. eauto.
              eauto. eauto.
              inversions HinG2. eauto.
              eauto. eauto. eauto.
         }
          eapply ty_trm_var; eauto.
          intro. eapply inert_dom_vis; eauto.

          destructc Hhc as ? ? ? ? HdomS1 Hdome1 Hrest. eapply binds_to_dom in Hbixe1.
          destructc Hrest as ? ?. rewrite <- Hdome1 in Hbixe1. eapply ctx_in_kdom_dom in Hbixe1. auto.

          eapply shortening_typ_loc;  eauto.
          6: eapply shortening_typ_ref. 13: eapply Hmutx. all: eauto.
          -- inversions HinG2. auto.
          -- eapply ctx_closed_push_G. eauto.
          -- assert (x \in dom e1) as Hxdome1 by eauto using binds_to_dom.
             hnf. intros. subst. auto.
          -- destructc Hhc as ? ? ? ? ? ? Hrest. destructc Hrest as ? ?.  eapply EnvCorrespondence.renv_corr_push_inv.
             eapply renv_corr_epush_inv; eauto. eapply ok_push_inv; eauto.
          -- destruct Hfv as [[HfvF1 HnfF1] ?].
            replace (ctx_fv (F1 § y1 / v_loc ~ typ_and (typ_bnd R) (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)))) with (ctx_fv (F1) \u fv (R)).
            rewrite notin_union; split.
            eapply fresh_in_config_notin_nothing_fresh. apply HnfF1.

            list_split_from_in w Hfr; eapply Hfr.
            list_split_from_in w Hfr; eapply fresh_in_config_notin_t in Hfr.
            intro HwR. apply Hfr. compute in HwR. compute. repeat rewrite in_union. eauto.
              compute. eauto. repeat rewrite union_empty_r. repeat rewrite union_empty_l. eauto.
          --
            eapply binds_to_dom in Hbixe1;  list_split_from_in w Hfr; eapply fresh_in_config_notin_e_dom in Hfr; eauto.
            intro; subst; contradiction.
      * (* unchanged item *)


        eapply binds_push_neq_inv in H1; auto.
        eapply mreach_fld; eauto.
        eapply IHHmr; eauto.
        eapply binds_to_dom; eauto.
        eapply ty_trm_var_inv in Hy1hasmuta.

        invert_ty_config2. destructc Hhc2 as HokG2 ? ? ? ? ? Hrest. destructc Hrest as ? ?.
        inversions H.
        invert_ty_config. destructc Hhc as ? ? ? ? Hkdom ? Hrest.
        destruct Hrest.
        clear Hfv2. clear IHHmr.

        assert (w # S1) as HnwS1. {
          list_split_from_in w Hfr.
          eapply fresh_in_config_notin_Sigma_dom. eauto.
        }
        assert (y1 \in dom S1) as HdomyS1. {
          eapply binds_to_dom in H1. auto.
        }
        assert (y0 §# F1) as Hny0dF1. {
          inversions HokG2. inversions H24. auto.
        }
        destruct Hfv as [[HdomF1 HnfrF1] HnfrT].

        assert (y0 \notin ctx_fv F1) as Hny0fF1. {
          list_split_from_in y0 Hfr.
          eapply fresh_in_config_notin_nothing_fresh; eauto.
        }

        assert (w \notin ctx_fv F1) as HnwF1. {
          list_split_from_in w Hfr.
          eapply fresh_in_config_notin_nothing_fresh; eauto.
        }


        eapply ty_trm_var.
        { rename H1 into Hbiy. eapply binds_to_dom in Hbiy. rewrite <- Hkdom in Hbiy.
          eapply ctx_in_kdom_dom in Hbiy. eapply ctx_dom_to_binds in Hbiy. destruct Hbiy.
          intro. eapply inert_binds_vis; eauto. eauto. eauto. }


        eapply shortening_fld_typ_loc.
        4: eauto.
        8: eapply shortening_fld_typ_ref.
        16: eapply Hy1hasmuta.
        all: eauto.
        *** (* sealed *)
          eapply ctx_closed_push_G; eauto.
        (* *** *)
        (* sympls. *)
        (* notin_solve. destructc Hfr as ? Hnw ?. repeat rewrite notin_union in Hnw. destruct_all. auto. *)
        ***
        intro; subst. contradiction.


    + (* term *)
      inversions HeqC.
      destruct (classicT (w0 = w)).
      * (* The new reference *)
        (* not possible, because we are assuming that the location already is in the heap *)
        subst.
        eapply binds_push_eq_inv in H1.
        subst.
        assert (y0 # S1).
        {
          eapply fresh_in_config_notin_Sigma_dom.
          list_split_from_in y0 Hfr.
          eapply Hfr.
        }
        contradiction.
      * (* tfree in the original term *)
        eapply binds_push_neq_inv in H1; auto.
        eapply mreach_trm with (w:=w0);
        eauto.
        destruct H0.
        eapply tfree_open_inv in H0.
        destruct H0; try contradiction.
        ** left. eauto.
        ** right. eauto.
        **
          invert_ty_config. pose proof Hhc as Hhc'. destructc Hhc' as ? ? ? ? Hkdom Hkdome Hrest. destructc Hrest as ? ?.
          inversions H.
          invert_ty_config2. destructc Hhc2 as ? ? ? ? ?Hkdom ?Hkdome Hrest. destructc Hrest as ? ?.
        assert (y0 \notin dom e1). {
          list_split_from_in y0 Hfr.
          eapply fresh_in_config_notin_e_dom. eauto.
        }
        assert (w0 \in dom e1). {
          eapply binds_to_dom in H1. auto.
        }
        assert (y0 §# F1). {
          inversions H. inversions H24. auto.
        }
     destruct Hfv as [[HdomF1 HnfrF1] HnfrT].
      assert (y0 \notin ctx_fv F1).
      list_split_from_in y0 Hfr.

      eapply fresh_in_config_notin_nothing_fresh; eauto.

       assert (w \notin ctx_fv F1).
      list_split_from_in w Hfr.

      eapply fresh_in_config_notin_nothing_fresh; eauto.

          eapply ty_trm_var_inv in H2.
          eapply ty_trm_var.
          { rename H1 into Hbiy. eapply binds_to_dom in Hbiy. rewrite <- Hkdome in Hbiy.
          eapply ctx_in_kdom_dom in Hbiy. eapply ctx_dom_to_binds in Hbiy. destruct Hbiy.
          intro. eapply inert_binds_vis; eauto. eauto. eauto. }
          eapply shortening_typ_loc. 4: eauto.
          8: eapply shortening_typ_ref.
          15: eapply H2.
          all: eauto.
     eapply ctx_closed_push_G; eauto.
          intro; subst. contradiction.
  - (* function literal *)
   (* functions not supported in rodot. *)
  typing_mode_contradiction typing_mode.
Qed.

Theorem mreach_preservation_let t_1 t_2 s1 S1 e1 Lv1 F1 T t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ ((trm_let t_1 t_2) ; s1 ; S1 ; e1; Lv1) ∶ T ↦ F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) mreach y ->
    F1 ⊢ ((trm_let t_1 t_2) ; s1 ; S1 ; e1; Lv1) mreach y.
Proof.
  introv Hbi Hred Hmr.

  assert (ty_config F2 (t2 ; s2 ; S2 ; e2; Lv2) T) as Htyc2. {
    eapply typed_preservation.
    eapply Hred.
  }

  inversions Hred; inversions H7.
  - (* let push *)
    eapply mreach_preserve_vars; eauto.
    intros.
    destruct H.
    + (* a variable in the focus *)
      eauto.
    + (* a variable on the stack *)
      inversions H; eauto.
Qed.

Theorem mreach_preservation_var a s1 S1 e1 Lv1 F1 T t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ ((trm_var a) ; s1 ; S1 ; e1; Lv1) ∶ T ↦ F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) mreach y ->
    F1 ⊢ ((trm_var a) ; s1 ; S1 ; e1; Lv1) mreach y.
Proof.
  introv Hbi Hred Hmr.

  assert (ty_config F2 (t2 ; s2 ; S2 ; e2; Lv2) T) as Htyc2. {
    eapply typed_preservation.
    eapply Hred.
  }

  inversions Hred; try inversions H7.
  - (* let loc *)
    eapply mreach_preserve_vars; eauto.
    intros.
    destruct H.
    + (* a variable on top *)
      eauto.
      eapply tfree_open_inv_trm in H.
      destruct H.
      * right. eapply tfree_stack_cons_l. eauto.
      * subst. eauto.
    + (* a variable on the stack *)
      right. eapply tfree_stack_cons_r. eauto.
Qed.

(* Theorem 11 (MP) in "Reference mutability for DOT" *)
Theorem mreach_preservation t1 s1 S1 e1 Lv1 F1 T t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ (t1 ; s1 ; S1 ; e1; Lv1) ∶ T ↦ F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) mreach y ->
    F1 ⊢ (t1 ; s1 ; S1 ; e1; Lv1) mreach y.
Proof.
  introv Hbi Hred Hmr.
  destruct t1.
  - eapply mreach_preservation_var; eauto.
  - eapply mreach_preservation_read; eauto.
  - eapply mreach_preservation_apply; eauto.
  - eapply mreach_preservation_let; eauto.
  - eapply mreach_preservation_lit; eauto.
  - eapply mreach_preservation_write; eauto.
  - eapply mreach_preservation_call; eauto.
Qed.

Theorem mreach_preservation_closure t1 s1 S1 e1 Lv1 F1 T n t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ (t1 ; s1 ; S1 ; e1; Lv1) ∶ T ⊣ n ↦* F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) mreach y ->
    F1 ⊢ (t1 ; s1 ; S1 ; e1; Lv1) mreach y.
Proof.
  introv Hbi Hred Hmr.
  remember (t1 ; s1 ; S1 ; e1; Lv1) as C1.
  induction Hred; eauto.
  intros; subst.
  eapply IHHred; eauto.
  destruct C2. destruct C3.
  destruct (heap_preservation_closure Hbi Hred).
  eapply mreach_preservation; eauto.
Qed.

Lemma mutated_mreach_closure t1 s1 S1 e1 Lv1 F1 T n t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ (t1 ; s1 ; S1 ; e1; Lv1) ∶ T ⊣ n ↦*  F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) -> (
      binds y (item_obj ds) S2 \/
      F1 ⊢ (t1 ; s1 ; S1 ; e1; Lv1) mreach y
    ).
Proof.
  introv Hbi Hred.
  
  remember_dependent_config t2 s2 S2 e2 Lv2 as C2.
  remember_dependent_config t1 s1 S1 e1 Lv1 as C1.

  induction Hred; intros; subst.
  - inversions HeqC2. eauto.
  -
    destruct C2 as [t3 s3 S3 e3 Lv3].
    specialize (  IHHred _ _ _ Hbi _ _ (eq_refl _) _ _ _ _ _  (eq_refl _)).
    destruct IHHred.
    assert (binds y (item_obj ds) S2 \/ F2 ⊢ t3 ; s3 ; S3 ; e3 ; Lv3 mreach y).
    eapply mutated_mreach; eauto.
    destruct H1. eauto.
    right.
    eapply mreach_preservation_closure; eauto.
    eauto.
Qed.

Theorem typed_immutability_guarantee t1 s1 S1 e1 Lv1 F1 T n t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ (t1 ; s1 ; S1 ; e1; Lv1) ∶ T ⊣ n ↦* F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) -> (
      binds y (item_obj ds) S2 \/
      F1 ⊢ (t1 ; s1 ; S1 ; e1; Lv1) mreach y
    ).
Proof.
  introv Hbi Hred.
  remember (t1 ; s1 ; S1 ; e1; Lv1) as C1.
  remember_dependent_config t2 s2 S2 e2 Lv2 as C2.

  induction Hred; intros; subst.
  - (* Base case *)
    inversions HeqC2.
    auto.
  - (* Step *)
    destruct C2 as [t3 s3 S3 e3 Lv3].
    specialize (IHHred (eq_refl _) _ _ _ _ _ (eq_refl _)).
    destruct IHHred.
    eapply mutated_mreach_closure; eauto. eauto.
Qed.

(* Theorem 9 (IG) in "Reference mutability for DOT" *)
(* Theorem 2 in "Pure methods for roDOT" *)
Theorem immutability_guarantee t1 s1 S1 e1 Lv1 F T n t2 s2 S2 e2 Lv2 y ds:
    binds y (item_obj ds) S1 ->
    ty_config F (t1 ; s1 ; S1 ; e1; Lv1) T ->
    (t1 ; s1 ; S1 ; e1; Lv1) ⊣ n ↦* (t2 ; s2 ; S2 ; e2; Lv2) -> (
      binds y (item_obj ds) S2 \/
      F ⊢ (t1 ; s1 ; S1 ; e1; Lv1) mreach y
    ).
Proof.
  introv Hbi Hty Hred.
  destruct (untyped_red_to_typed_closure Hty Hred).
  eapply typed_immutability_guarantee; eauto.
Qed.

(* If a configuration can be typed in a context where all the
   variables that do not appear in type position are assigned to
   read-only types then execution is side effects free as in it does
   not change the existing heap. *)
Lemma binds_ro_sef t1 s1 S1 e1 Lv1 G T n t2 s2 S2 e2 Lv2 :
  ty_config G (t1 ; s1 ; S1 ; e1; Lv1) T ->
  (forall w, tfree w t1 \/ tfree w s1 ->
        exists Tw, ctx_binds
                w (typ_and (typ_bnd Tw) (typ_mut_marker typ_top)) G) ->
  (t1 ; s1 ; S1 ; e1; Lv1) ⊣ n ↦* (t2 ; s2 ; S2 ; e2; Lv2) ->
  subheap_unchanged S1 S2.
Proof.
  introv Hty Hro Hred Hbiy.
  destruct itm.
  2: {
    false. inversions Hty. inversions H5.
    destruct H0 as [_ [_ [_ [_ [_ [_ Htys]]]]]].
    specialize (Htys _ (binds_to_dom Hbiy)). inversions Htys.
    - destruct typing_mode; simpl in H0, Hmm; discriminate.
    - eapply (binds_functional H1) in Hbiy. discriminate.
  }
  eapply immutability_guarantee in Hred; eauto.
  destruct Hred as [? | Hc]; eauto. false.
  inversions Hty. eapply binds_ro_not_mreach in Hc; eauto.
Qed.

End WithRodotMode.
