Set Implicit Arguments.

Require Import LibExtra.
Require Import AbstractSyntax TypingMode GeneralTyping.
Require Import RecordTypes InertContext TFree.

(** * Mutable Reachability *)

Class MutableReachability := mutable_reachability: ctx -> config -> var -> Prop.
Notation "F ⊢ C 'mreach' y" := (mutable_reachability F C y) (at level 40, C at level 59).

Section WithTypingMode.
Context (typing_mode: Mode).

(* Figure 11 in "Reference mutability for DOT" *)
(* Figure 3 in "Pure methods for roDOT" *)
(** We write F ⊢ C mreach y for a variable y that is mutable reachable from
    a configuration C of the abstract machine under the (mutability) typing given
    by an inert context F *)
Inductive mut_reach : MutableReachability :=
(** [      F ⊢ (t, fs, e, Σ) mreach y₁       ]  #<br># *)
(*  [        (y₁, ...{a = y₂}...) ∈ Σ        ]  #<br># *)
(*  [ F; e ⊢ y₁ : {a : ⊥..{M(r) : ⊥..⊥}}  ]  #<br># *)
(*  [―――――――――――――――――――――――――――――――――――――――――(Rea-Fld)]  #<br># *)
(*  [      F ⊢ (t, fs, ρ, Σ) mreach y₂       ]         *)
| mreach_fld : forall (F : ctx) (t : trm) (fs : stack) (e : renv) (Sigma : heap) Lv
                 (y1 y2 : var) (ds : defs) (a : fld_label),
    (* Mutable reachability is only defined for inert contexts *)
    inert F ->
    (* y₁ is mutable reachable *)
    F ⊢ (t; fs; Sigma; e; Lv) mreach y1 ->
    (* (y₁, ...{a = y₂}...) ∈ Σ *)
    binds y1 (item_obj ds) Sigma -> defs_has ds (def_fld a (avar_f y2)) ->
    (* y₁.a is mutable *)
    F ⁏ e ⊢ trm_var (avar_f y1) ∶ typ_rcd (dec_fld a typ_bot
        (typ_mut_marker typ_bot)) ->
    (* Then y₂ is mutable reachable *)
    F ⊢ (t; fs; Sigma; e; Lv) mreach y2

(** [   t tfree w ∨ fs tfree w    ]  #<br># *)
(*  [         (y, w) ∈ ρ          ]  #<br># *)
(*  [ F; e ⊢ w : {M(r) : ⊥..⊥}  ]  #<br># *)
(*  [――――――――――――――――――――――――――――――(Rea-Term)]  #<br># *)
(*  [ F ⊢ (t, fs, ρ, Σ) mreach y  ]         *)
| mreach_trm : forall (F : ctx) (t : trm) (fs : stack) (e : renv) (Sigma : heap) Lv
                 (y w : var),
    (* Mutable reachability is only defined for inert contexts *)
    inert F ->
    (* w is tfree in the hole t or the continuation fs *)
    (tfree w t) \/ (tfree w fs) ->
    (* e binds w to y (i.e. w points to y) *)
    binds w y e ->
    (* w is mutable *)
    F ⁏ e ⊢ trm_var (avar_f w) ∶ (typ_mut_marker typ_bot) ->
    (* Then y is mutable reachable *)
    F ⊢ (t; fs; Sigma; e; Lv) mreach y
.

End WithTypingMode.

Existing Instance mut_reach.
