Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import AbstractSyntax  TypingMode Equivalence GeneralTyping
        Weakening EWeakening Substitution Subenvironments SyntaxSimilarity
        ConfigSimilarity ConfigSefSimilarity Similarity.
Require Import InertContext GeneralToTight TightTyping EnvCorrespondence
        InvertibleTyping Indep InertTightSubtyping
        PreciseTyping RecordTypes Narrowing TFree Closedness.
Require Import HeapCorrespondence OperationalSemantics
        ConfigTyping ConfigVarkinds.
Require Import RoContext ReductionLemmas MethodCall TypedMethodCall ReferenceElimination
  ImmutabilityGuarantee ReferenceTypes HeapUpdate ConfigNotations.

(** * SEF guarantee *)

Inductive def_flds_sim : def -> def -> Prop :=
| def_flds_sim_typ : forall l T, def_flds_sim (def_typ l T) (def_typ l T)
| def_flds_sim_fld : forall l x y, def_flds_sim (def_fld l x) (def_fld l y)
| def_flds_sim_met : forall m t, def_flds_sim (def_met m t) (def_met m t).

Inductive defs_flds_sim : defs -> defs -> Prop :=
| defs_flds_sim_nil : defs_flds_sim defs_nil defs_nil
| defs_flds_sim_cons : forall d ds d' ds',
    (def_flds_sim d d') -> (defs_flds_sim ds ds') ->
    defs_flds_sim (defs_cons ds d) (defs_cons ds' d').

(* Two heap objects ds and ds' are similar by this relation when only
   their fields differ but not labels or type members or methods *)
Inductive heap_item_flds_sim : item -> item -> Prop :=
| heap_item_flds_sim_obj : forall ds ds',
    defs_flds_sim ds ds' ->
    heap_item_flds_sim (item_obj ds) (item_obj ds')
| heap_item_flds_sim_fun : forall t,
    heap_item_flds_sim (item_fun t) (item_fun t).

Instance DefFldsSimEquiv : RelationClasses.Equivalence def_flds_sim.
Proof.
  split; hnf.
  - intros. destruct x; constructor; eauto.
  - intros. inversions H; constructor; eauto.
  - intros. inversions H; inversions H0; constructor; eauto.
Qed.
Global Existing Instances DefFldsSimEquiv.

Instance DefsFldsSimEquiv : RelationClasses.Equivalence defs_flds_sim.
Proof.
  split; hnf.
  - intros. induction x; constructor; eauto; reflexivity.
  - intros. induction H; constructor; eauto. symmetry. easy.
  - intros. reverts z. induction H; intros.
    2: { inversions H1. eapply IHdefs_flds_sim in H6.
         constructor; eauto. rewrite H. easy. }
    inversions H0. constructor; eauto.
Qed.
Global Existing Instances DefsFldsSimEquiv.

Instance HeapItemFldsSimEquiv : RelationClasses.Equivalence
                                  heap_item_flds_sim.
Proof.
  split; hnf.
  - intros. destruct x; constructor; eauto. reflexivity.
  - intros. inversions H; constructor; eauto. symmetry. eauto.
  - intros. inversions H; inversions H0; constructor; eauto.
    rewrite H1; eauto.
Qed.
Global Existing Instances HeapItemFldsSimEquiv.

Lemma defs_update_flds_sim ds ds' a y :
  defs_update ds a y ds' ->
  defs_flds_sim ds ds'.
Proof. induction 1; constructor; eauto; try reflexivity. constructor. Qed.

Lemma red_heap_flds_sim_single c c' it it' y :
  c ↦ c' ->
  binds y it  (config_heap c ) ->
  binds y it' (config_heap c') ->
  heap_item_flds_sim it it'.
Proof.
  introv Hred Hbi Hbi'. inversions Hred; simpl in *; try binds_eq;
    try reflexivity.
  - hnf in H. destructc H as _ _ _ Hbin Hupd.
    destruct (classicT (y = yObj)) as [? | Hne]; subst.
    + destructc Hupd as ds ds' Hbi0 Hupd Hbi0'. binds_eq. binds_eq.
      constructor. eapply defs_update_flds_sim; eauto.
    + eapply Hbin in Hne; eauto. binds_eq. reflexivity.
  - eapply binds_push_inv in Hbi'. destruct_all; subst; try binds_eq;
      try reflexivity. false.
    eapply binds_to_dom in Hbi. eapply H. rewrite ! in_union. tauto.
  - eapply binds_push_inv in Hbi'. destruct_all; subst; try binds_eq;
      try reflexivity. false.
    eapply binds_to_dom in Hbi. eapply H. rewrite ! in_union. tauto.
Qed.

(* Lemma 22 in "Pure methods for roDOT" *)
Lemma red_heap_flds_sim c c' it it' y n :
  c ⊣n↦* c' ->
  binds y it  (config_heap c ) ->
  binds y it' (config_heap c') ->
  heap_item_flds_sim it it'.
Proof.
  introv Hred Hbi Hbi'. reverts c' it it'. induction n; intros;
    inversions Hred; try solve[binds_eq; reflexivity].
  eapply binds_to_dom in Hbi as Hbi2.
  eapply red_closure_mono_heap_dom, dom_to_binds in Hbi2; eauto.
  destructc Hbi2 as it2 Hbi2. forwards:IHn; fweauto. rewrite H.
  eapply red_heap_flds_sim_single; eauto.
Qed.

(* Definition 21 in "Pure methods for roDOT" *)
(* Two heaps are similar by this relation if for each variable in the set S
   that is in the common domain, the corresponding heap items in each heap
   are the same except they may have different fields *)
Definition heap_flds_sim (Sigma1 Sigma2 : heap) (S : vars) :=
  forall y itm1 itm2,
    (y \in S) ->
    binds y itm1 Sigma1 ->
    binds y itm2 Sigma2 ->
    heap_item_flds_sim itm1 itm2.

Section WithTypingMode.

Context (typing_mode : Mode).

End WithTypingMode.

Section WithSefMode.

Context (typing_mode : Mode).
Context (Hms : mode_has_sef typing_mode).
Context (Hmm : mode_has_mutability typing_mode).

Lemma ty_ro_loc G (e : env var) T w y :
  sealed G -> inert G -> binds w y e -> renv_corr G e -> ok e ->
  w \notin fv T ->
  G ⁏ e ⊢ typ_nocap <: T ->
  G ⁏ e ⊢ y ∶ T ->
  G ⁏ e ⊢ w ∶ T.
Proof.
  introv Hs Hi Hbi Hcor Hoke Hnw Hsu Hty.
  eapply reference_nocap_type; eauto; simpl.

  all: destruct (mode_has_incap' typing_mode) eqn: Hmi; try typing_mode_contradiction typing_mode. auto.
  all: destruct (mode_has_top_nm_subtype' typing_mode) eqn: Hmnm; try typing_mode_contradiction typing_mode. auto.
Qed.

Section SefG_I.

(** Side-effects-free (SEF) method type bound for a name m called with *)
(*     parameter whose type is TZ is *)
(*     { m:(z:TZ)(r:N) : ⊤ } *)
Notation sef_met_bnd m TZ :=
  (dec_met m typ_nocap TZ typ_top).
Notation sef_met_rcd m TZ := (typ_rcd (sef_met_bnd m TZ)).
(** A type T is read only when it's a super type of N *)
Notation is_ro_type G e T := (G ⁏ e ⊢ typ_nocap <: T).

Variable (w1 : var) (m : met_label) (w2 : var) (TZv : typ)
         (c1 : config) (G : ctx).
(** c₁ = ⟨w₁.m w₂; σ₁; ρ₁; Σ₁⟩ *)
Hypothesis (Htrmc1 : config_trm c1 = trm_call (avar_f w1) m (avar_f w2)).
(** The SEF condition says that w₂ has a read-only type S and *)
(*     w₁.m is a method type where the receiver expects something of type S: *)
(*     G ⊢ w₁.m : {m:(z:S)(r:N) : ⊤} and G ⊢ w₂ : S :> N *)
Hypothesis (Htysef : (G ⁏ (config_env c1) ⊢  w1 ∶ sef_met_rcd m TZv)
                     /\ (G ⁏ (config_env c1) ⊢ w2 ∶ TZv)
                     /\ (G ⁏ (config_env c1) ⊢ typ_nocap <: TZv)).
Hypothesis (Htycfg : exists T1, ty_config G c1 T1).

Lemma config_env_ok : ok (config_env c1).
Proof.
  destruct Htycfg as [T Htyc]. inversions Htyc.
  inversions H0. simpl. tauto.
Qed.
Hint Resolve config_env_ok : core.

Lemma config_has_null : has_null (config_fv c1).
Proof.
  destruct Htycfg as [T Htyc]. inversions Htyc.
  inversions H0. simpl. tauto.
Qed.
Hint Resolve config_has_null : core.

Lemma config_renv_corr : renv_corr G (config_env c1).
Proof. destruct Htycfg. destruct H. eauto. Qed.
Hint Resolve config_renv_corr : core.

Lemma config_heap_corr : heap_correspond G (config_env c1) (config_heap c1).
Proof. destruct Htycfg. destruct H. eauto. Qed.

Lemma config_ctx_inert : inert G.
Proof.  destruct Htycfg. destruct H. eauto. Qed.

Section W1W2Diff.

(** First assume that w₁ ≠ w₂ *)
Hypothesis (Hwdiff : w1 <> w2).

(** The plan is as follows:

      1. sef_neq_elim_ty_hole. Show that we can type Gw2 ⁏ ew2 ⊢
         ((config_trm c1) /[ w1 -> y1]) /[ w2 -> y2] ∶ Te where Gw2 is G
         but with references w eliminated for location y, and likewise
         for ew2 but for config_env c1.

      2. sef_neq_elim_heap_corr. Show heap correspondence
         (Gw2,ew2) ~ (subst_heap w2 y2 (subst_heap w1 y1 (config_heap c1))).

      3. sef_neq_ro_ty_hole. Use (1) sef_neq_ref_elim_ty_hole to show
         we can type the hole of c1 in a context where w have read
         only types.
         (Gw2 § w2 / v_ref ~ Twr2 § w1 / v_ref ~ Twr1 ⁏ (config_env c1)
             ⊢ (config_trm c1) ∶ Te

      4. sef_neq_ro_ty_heap. Weaken (2) to
           heap_correspond (Gw2 § w2 / v_ref ~ Twr2 § w1 / v_ref ~ Twr1)
                           (config_env c1)
                           (subst_heap w2 y2 (subst_heap w1 y1
                                               (config_heap c1)))

      5. sef_neq_ro_ty_config. Use (3) and (4) to show that if
         c₁ = ⟨w₁.m w₂; σ₁; ρ₁; Σ₁⟩ then one can type
         c₂ = ⟨w₁.m w₂; (); ρ₁; Σ₁[wᵢ=yᵢ]ᵢ₌₁,₂⟩ in an context where
         w₁, w₂ are given read-only types. The aim is to use this result
         to apply the immutability guarantee to the config c₂.

     6. Plans for next steps of proving the SEF guarantee is as given in
        the PDF in the documents branch.
*)

(* Lemma 18 in "Pure methods for roDOT" *)
Lemma sef_neq_elim_ty_hole Gw1 Gw2 ew1 ew2 y1 y2 :
  binds w1 y1 (config_env c1) ->
  binds w2 y2 (config_env c1) ->
  ref_elim_ctx G w1 y1 Gw1 ->
  ref_elim_ctx Gw1 w2 y2 Gw2 ->
  env_elim (config_env c1) w1 y1 ew1 ->
  env_elim ew1 w2 y2 ew2 ->
  exists Te, (Gw2 ⁏ ew2 ⊢ ((config_trm c1) /[ w1 -> y1]) /[ w2 -> y2] ∶ Te).
Proof.
  clear Htysef.
  destruct Htycfg as [T1 Htycfg']. clear Htycfg.
  destruct Htycfg'; simpl in *.
  rename H into HinG, H0 into Hhc, H1 into Htys, H2 into Htf,
    H5 into Hnfr, H4 into Htfk, H3 into Hty, H6 into Hhn, H7 into Hs,
    H8 into Hclc, H9 into HclU.
  clear Hnfr Htf Htys Htfk Hclc HclU Hhn Hs.
  introv Hbie1 Hbie2 HeG1 HeG2 Hee1 Hee2.
  destruct Hhc as
    [HokG [HwkG [Hoks [Hoke [Hsloc [Heref [Hcore Htyit]]]]]]].
  clear Hoks Hsloc Heref Htyit.

  pose proof (ref_elim_typ_trm (G:=G) (Go:=objctx_empty)
                               (G':=Gw1) (e:=e) (e':=ew1)
                               (w:=w1) (y:=y1)
                               (t:=t)
                               (T:=T)) as Hty1. simpl in *.
  specialize (Hty1 (inert_fresh_telescope HinG)
                HokG Hoke Hcore Hty HeG1 Hbie1 Hee1).

  pose proof (ref_elim_typ_trm (G:=Gw1) (Go:=objctx_empty)
                               (G':=Gw2) (e:=ew1) (e':=ew2)
                               (w:=w2) (y:=y2)
                               (t := t /[ w1 -> y1])
                               (T := T /[ w1 -> y1])) as Hty2. simpl in *.
  assert (HinGw1 : inert Gw1).
  { pose proof (renv_corr_binds_to_binds_after HokG Hcore Hbie1)
      as [Ty [Tw [Gw [Gy [Hbkw1 [Hbkw2 Htycor1]]]]]].
    eauto using ref_elim_ctx_inert. }
  assert (Hokew1 : ok ew1) by eauto using env_elim_ok.
  assert (Hcore1 : renv_corr Gw1 ew1).
  { eapply ref_elim_renv_corr
      with (e:=e) (G:=G) (w:=w1) (y:=y1); eauto using inert_fresh_telescope.
    introv Hinw1. apply fv_env_to_binds in Hinw1; eauto.
    destruct Hinw1 as [w Hbiw1].
    eapply renv_corr_binds_loc in Hbiw1; eauto.
    eapply renv_corr_binds_vk in Hbie1; eauto.
    destruct Hbie1 as [Hbie1 _]. assert (Hneq : v_loc <> v_ref) by easy.
    eauto using ctx_k_ne_false. }
  assert (Hbie21 : binds w2 y2 ew1) by eauto using env_elim_correct.

  eauto using inert_fresh_telescope.
Qed.

Lemma sef_neq_elim_heap_corr Gw1 Gw2 ew1 ew2 y1 y2 :
  binds w1 y1 (config_env c1) ->
  binds w2 y2 (config_env c1) ->
  ref_elim_ctx G w1 y1 Gw1 ->
  ref_elim_ctx Gw1 w2 y2 Gw2 ->
  env_elim (config_env c1) w1 y1 ew1 ->
  env_elim ew1 w2 y2 ew2 ->
  heap_correspond Gw2 ew2
                  (subst_heap w2 y2 (subst_heap w1 y1 (config_heap c1))).
Proof.
  clear Htysef.
  destruct Htycfg as [T1 Htycfg']. clear Htycfg.
  destruct Htycfg'; simpl in *.
  rename H into HinG, H0 into Hhc, H1 into Htys, H2 into Htf,
    H5 into Hnfr, H4 into Htfk, H3 into Hty, H6 into Hhn, H7 into Hs,
    H8 into Hclc, H9 into HclU.
  clear Hnfr Htf Htys Htfk Hclc HclU Hhn Hs.
  introv Hbie1 Hbie2 HeG1 HeG2 Hee1 Hee2.

  forwards: ref_elim_heap; try eassumption; eauto using inert_fresh_telescope.
  rename H into Hhc1.

  assert (HinGw1 : inert Gw1).
  { destruct Hhc as
      [HokG [HwkG [Hoks [Hoke [Hsloc [Heref [Hcore Htyit]]]]]]].
    pose proof (renv_corr_binds_to_binds_after HokG Hcore Hbie1)
      as [Ty [Tw [Gw [Gy [Hbkw1 [Hbkw2 Htycor1]]]]]].
    eauto using ref_elim_ctx_inert. }
  assert (Hokew1 : ok ew1) by eauto using env_elim_ok.
  assert (Hcore1 : renv_corr Gw1 ew1).
  { eapply ref_elim_renv_corr
      with (e:=e) (G:=G) (w:=w1) (y:=y1); eauto using inert_fresh_telescope.
    introv Hinw1. apply fv_env_to_binds in Hinw1; eauto.
    destruct Hinw1 as [w Hbiw1].
    destruct Hhc as
      [HokG [HwkG [Hoks [Hoke [Hsloc [Heref [Hcore Htyit]]]]]]].
    eapply renv_corr_binds_loc in Hbiw1; eauto using inert_ok.
    eapply renv_corr_binds_vk in Hbie1; eauto using inert_ok.
    destruct Hbie1 as [Hbie1 _]. assert (Hneq : v_loc <> v_ref) by easy.
    eauto using ctx_k_ne_false. }
  assert (Hbie21 : binds w2 y2 ew1) by eauto using env_elim_correct.

  forwards: ref_elim_heap; try eassumption; eauto using inert_fresh_telescope.
Qed.

Lemma sef_neq_ro_ty_heap Gw1 Gw2 y1 y2 Ty1 Twr1 Ty2 Twr2 :
  binds w1 y1 (config_env c1) ->
  binds w2 y2 (config_env c1) ->
  ctx_binds y1 Ty1 Gw2 ->
  ctx_binds y2 Ty2 Gw2 ->
  ref_elim_ctx G w1 y1 Gw1 ->
  ref_elim_ctx Gw1 w2 y2 Gw2 ->
  typ_ro Ty1 Twr1 ->
  typ_ro Ty2 Twr2 ->
  heap_correspond (Gw2 § w2 / v_ref ~ Twr2 § w1 / v_ref ~ Twr1)
    (config_env c1)
    (subst_heap w2 y2 (subst_heap w1 y1 (config_heap c1))).
Proof.
  introv Hbie1 Hbie2 HbiGw1 HbiGw2 Hre1 Hre2 Hro1 Hro2.

  assert (Hoke : ok (config_env c1)) by eauto.
  destruct (env_elim_binds_exists config_env_ok Hbie1) as
    [ew1 [Hee1 Hoke1]].
  eapply env_elim_correct with (y:=w2) in Hee1 as Hbie2'; eauto.
  destruct (env_elim_binds_exists Hoke1 Hbie2') as
    [ew2 [Hee2 Hoke2]].

  assert (HinG : inert G) by eauto using config_ctx_inert.
  assert (HcorGe : renv_corr G (config_env c1))
    by eauto using config_renv_corr.

  assert (Hlrne1 : w1 <> y1) by (eauto using renv_corr_diff).
  assert (Hlrne2 : w2 <> y2) by (eauto using renv_corr_diff).
  assert (Hr1l2ne : w1 <> y2) by (eauto using renv_corr_diff).
  assert (Hr2l1ne : w2 <> y1) by (eauto using renv_corr_diff).

  assert (HinG1: inert Gw1).
  { eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie1;
      eauto using inert_ok.
    destruct Hbie1 as [TL1 [TR1 [G2 [G3 [Hbkw1 [Hbky1 _]]]]]].
    eauto using ref_elim_ctx_inert. }

  assert (HcorGe1 : renv_corr Gw1 ew1).
  { eapply ref_elim_renv_corr with
      (G:=G) (e := config_env c1) (w:=w1) (y:=y1);
      eauto using inert_fresh_telescope.
    hnf. introv Hin. apply fv_env_to_binds in Hin; eauto using config_env_ok.
    destruct Hin as [w Hbie1'].
    eapply renv_corr_binds_loc with (G:=G) in Hbie1'; eauto using inert_ok.
    eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie1;
      eauto using inert_ok.
    destruct Hbie1 as [_ [TR1 [G2 [G3 [Hbk1 [_ _]]]]]].
    apply ctx_binds_k_after_k in Hbk1.
    pose proof (ctx_k_eq Hbie1' Hbk1). discriminate. }

  assert (Hbiew2 : binds w2 y2 ew1) by eauto using env_elim_correct.

  assert (HcorGe2 : renv_corr Gw2 ew2).
  { eapply ref_elim_renv_corr with
      (G:=Gw1) (e:=ew1) (w:=w2) (y:=y2); eauto using inert_fresh_telescope.
    hnf. introv Hin. apply fv_env_to_binds in Hin; eauto using config_env_ok.
    destruct Hin as [w Hbiew1].
    eapply renv_corr_binds_loc with (G:=Gw1) in Hbiew1; eauto using inert_ok.
    eapply renv_corr_binds_to_binds_after with (G1:=Gw1) in Hbiew2;
      eauto using inert_ok.
    destruct Hbiew2 as [_ [TR2 [G2 [G3 [Hbk2 [_ _]]]]]].
    apply ctx_binds_k_after_k in Hbk2.
    pose proof (ctx_k_eq Hbiew1 Hbk2). discriminate. }

  assert (HinG2 : inert Gw2).
  { eapply renv_corr_binds_to_binds_after with (G1:=Gw1) in Hbiew2;
      eauto using inert_ok.
    destruct Hbiew2 as [TL2 [TR2 [G2 [G3 [Hbkw2 [Hbky2 _]]]]]].
    eauto using ref_elim_ctx_inert. }

  assert (Hbky2 : exists Gy2, ctx_binds_k_after y2 v_loc Ty2 Gw2 Gy2).
  { eapply ctx_binds_to_ctx_binds_k_after in HbiGw2;
      eauto using inert_weakens. destruct HbiGw2 as [Gy2 [vy Hbk]]. exists Gy2.
    replace v_loc with vy; eauto.
    eapply renv_corr_binds_loc with (G:=Gw1) in Hbiew2; eauto.
    eapply ref_elim_ctx_vk_same with (G':=Gw2) in Hbiew2; eauto.
    eapply ctx_binds_k_after_k in Hbk. exact (ctx_k_eq Hbk Hbiew2). }

  assert (Hbky1 : exists Gy1, ctx_binds_k_after y1 v_loc Ty1 Gw2 Gy1).
  { eapply ctx_binds_to_ctx_binds_k_after in HbiGw1;
      eauto using inert_weakens. destruct HbiGw1 as [Gy1 [vy Hbk]]. exists Gy1.
    replace v_loc with vy; eauto. apply ctx_binds_k_after_k in Hbk.
    eapply renv_corr_binds_loc with (G:=G) in Hbie1; eauto.
    eapply ref_elim_ctx_vk_same with (G':=Gw1) in Hbie1; eauto.
    eapply ref_elim_ctx_vk_same with (G':=Gw2) in Hbie1; eauto.
    exact (ctx_k_eq Hbk Hbie1). }

  pose proof (sef_neq_elim_heap_corr Hbie1 Hbie2 Hre1 Hre2 Hee1 Hee2)
    as Hhc.
  destruct Hbky2 as [Gy2 Hbky2].

  eapply heap_correspond_epush_nowfr
    with (w:=w2) (T1:=Twr2) (T2:=Ty2) (y:=y2)
    in Hhc; eauto 3 using inert_typ_ro_corr, ref_elim_ctx_inert_typ_loc,
    ref_elim_ctx_ok_correct, env_elim_ok_notin.

  destruct Hbky1 as [Gy1 Hbky1].
  eapply heap_correspond_epush_nowfr
    with (w:=w1) (T1:=Twr1) (T2:=Ty1) (y:=y1)
    in Hhc; eauto 3.
  4: { eapply inert_typ_ro_corr; eauto.
       eapply renv_corr_binds_loc with (G:=G) in Hbie1 as HbiyG; eauto.
       eapply ctx_k_in_dom, ctx_dom_to_binds in HbiyG.
       destruct HbiyG as [TyG HbiyG].
       eapply ref_elim_ctx_neq_binds with (G':=Gw1) in HbiyG as HbiyG1;
         eauto.
       all: eauto using config_ctx_inert, inert_weakens,
           inert_ok, inert_fresh_telescope.
       eapply ref_elim_ctx_inert_typ_loc
         with (G':=Gw1) (G:=G) (e := config_env c1)
              (w0:=w1) (y0:=y1) (w:=w1) (y:=y1) in HbiyG1 as HinTyG;
         eauto.
       eapply ref_elim_ctx_neq_binds with (G':=Gw2) in HbiyG1;
         eauto using inert_fresh_telescope.
       eapply inert_type_subst with (x:=w2) (y:=y2) in HinTyG.
       rewrite <- (ctx_binds_functional HbiyG1 HbiGw1). easy. }
  3: { rewrite dom_push, notin_union, notin_singleton; split; eauto.
       eapply env_elim_ok_notin in Hee1; eauto using config_env_ok.
       introv Hin. apply dom_to_binds in Hin. destruct Hin as [v Hb].
       assert (Htmp : ok ew1) by eauto;
         pose proof (binds_to_dom (env_elim_binds_more Hee2 Htmp Hb));
         eauto. }
  2: { rewrite ctx_dom_push, notin_union, notin_singleton; split; eauto.
       eapply ref_elim_ctx_ok_correct in Hre1;
         eauto using config_ctx_inert, inert_ok.
       hnf in Hre2. destruct Hre2 as [vkw [Tw [G1 [G2 [HeqGw1 HeqGw2]]]]].
       rewrite HeqGw1, ctx_dom_concat in Hre1. simpl in Hre1.
       repeat(rewrite notin_union in Hre1); rewrite notin_singleton in Hre1.
       rewrite HeqGw2.
       rewrite <- ctx_dom_concat_map_r, ctx_dom_concat, notin_union. tauto. }

  eapply heap_correspond_epermute with (e := (ew2 & w2 ~ y2 & w1 ~ y1));
    eauto.

  - introv Hbi. apply binds_concat_inv in Hbi.
    rewrite dom_single, notin_singleton in Hbi.
    destruct Hbi as [Hbi | [_ Hbi]].
    1: apply binds_single_inv in Hbi; destruct Hbi as [? ?]; subst; eauto.
    apply binds_concat_inv in Hbi.
    rewrite dom_single, notin_singleton in Hbi.
    destruct Hbi as [Hbi | [_ Hbi]].
    1: apply binds_single_inv in Hbi; destruct Hbi as [? ?]; subst; eauto.
    eauto using env_elim_binds_more.

  - introv Hbi. destruct (classicT (w = w1)) as [? | Hne1]; subst.
    1: replace y with y1; eauto using binds_functional.
    destruct (classicT (w = w2)) as [? | Hne2]; subst.
    1: replace y with y2; eauto using binds_functional.
    repeat(eapply binds_push_neq); eauto using env_elim_correct.

Qed.

Lemma sef_neq_ro_ty_hole Gw1 Gw2 y1 y2 Ty1 Ty2 Twr1 Twr2 :
  ctx_binds y1 Ty1 Gw2 ->
  ctx_binds y2 Ty2 Gw2 ->
  binds w1 y1 (config_env c1) ->
  binds w2 y2 (config_env c1) ->
  ref_elim_ctx G w1 y1 Gw1 ->
  ref_elim_ctx Gw1 w2 y2 Gw2 ->
  typ_ro Ty1 Twr1 ->
  typ_ro Ty2 Twr2 ->
  (exists Te, (Gw2 § w2 / v_ref ~ Twr2 § w1 / v_ref ~ Twr1 ⁏ (config_env c1)
            ⊢ (config_trm c1) ∶ Te))
  /\ (inert (Gw2 § w2 / v_ref ~ Twr2 § w1 / v_ref ~ Twr1))
  /\ (sealed (Gw2 § w2 / v_ref ~ Twr2 § w1 / v_ref ~ Twr1)).
Proof.
  introv HbiG1 HbiG2 Hbie1 Hbie2 Hre1 Hre2 Hro1 Hro2.

  destruct (env_elim_binds_exists config_env_ok Hbie1) as
    [ew1 [Hee1 Hoke1]].

  eapply env_elim_correct with (y:=w2) in Hee1 as Hbie2'; eauto.
  destruct (env_elim_binds_exists Hoke1 Hbie2') as
    [ew2 [Hee2 Hoke2]].

  assert (Hokce : ok (config_env c1)) by eauto using config_env_ok.
  assert (HinG : inert G) by (destruct Htycfg as [T1 [? ?]]; eauto).
  assert (HsG : sealed G) by (destruct Htycfg as [T1 [? ?]]; eauto).
  assert (Hcor : renv_corr G (config_env c1))
    by (destruct Htycfg as [T1 [? ?]]; eauto).

  assert (Hlrne1 : w1 <> y1) by (eauto using renv_corr_diff).
  assert (Hlrne2 : w2 <> y2) by (eauto using renv_corr_diff).
  assert (Hr1l2ne : w1 <> y2) by (eauto using renv_corr_diff).
  assert (Hr2l1ne : w2 <> y1) by (eauto using renv_corr_diff).

  assert (HinGw1 : inert Gw1).
  { eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie1 as Htmp;
      eauto. destruct Htmp as [TL1 [TR1 [G12 [G11 [Hbkw1 [Hbky1 _]]]]]].
    eapply ref_elim_ctx_inert in Hre1 as HinGw1; eauto. }

  assert (HinGw2 : inert Gw2).
  { eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie1 as Htmp;
      eauto. destruct Htmp as [TL1 [TR1 [G12 [G11 [Hbkw1 [Hbky1 _]]]]]].
    eapply ref_elim_renv_corr in Hcor; eauto using inert_fresh_telescope.
    2: { hnf. intros Hin. apply fv_env_to_binds in Hin as [v Hbiw1']; eauto.
         eapply renv_corr_binds_vk with (G:=G) in Hbiw1' as [_ Hvkw1'];
           eauto.
         eapply env_elim_binds in Hee1; eauto.
         eapply renv_corr_binds_vk with (G:=G) in Hee1 as [Hvkw1 _]; eauto.
         eapply ctx_k_ne_false with (vk1:=v_loc) (vk2:=v_ref); eauto; easy. }

    eapply renv_corr_binds_to_binds_after with (G1:=Gw1) in Hbie2'
        as [TL2 [TR2 [G22 [G21 [Hbkw2 [Hbky2 _]]]]]]; eauto.
    eapply ref_elim_ctx_inert with (G':=Gw2) (w:=w2) (y:=y2)
      in HinGw1 as HinGw2; eauto. }

  assert (Hfrwr1 : w1 \notin fv Twr2).
  { introv Hin. eapply (typ_ro_fv Hro2) in Hin.
    eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie2 as Htmp;
      eauto. destruct Htmp as [TL2 [TR2 [G22 [G21 [Hbkw2 [Hbky2 _]]]]]].
    eapply ctx_binds_k_after_trans, ctx_binds_k_after_binds in Hbkw2; eauto.
    clear Hbky2.
    eapply ref_elim_ctx_neq_binds in Hbkw2;
      eauto using inert_fresh_telescope.
    eapply ref_elim_ctx_neq_binds in Hbkw2;
      eauto using inert_fresh_telescope.
    eapply (ctx_binds_functional HbiG2) in Hbkw2. subst.
    eapply fv_subst_var_subset in Hin. rewrite in_union, in_singleton in Hin.
    destruct Hin as [? | ?]; subst; eauto using renv_corr_diff.
    eapply subst_notfv in H; eauto. }

  assert (mode_refcaps' typing_mode = mode_refcaps_mutability) as Hmrc.
  {
    destruct (mode_refcaps' typing_mode) eqn: Hmrc. typing_mode_contradiction typing_mode. auto.
  }
  assert (mode_has_fun' typing_mode = false) as Hmf.
  {
    destruct (mode_has_fun' typing_mode) eqn: Hmf. typing_mode_contradiction typing_mode. auto.
  }
  assert (mode_refcap_corr' typing_mode = mode_refcap_corr_ref_mutability) as Hmrcr.
  {
    destruct (mode_refcap_corr' typing_mode) eqn: Hmrcr. typing_mode_contradiction typing_mode. auto.
  }
  assert (mode_visibility' typing_mode = mode_visibility_hide_method) as Hmv.
  {
    destruct (mode_visibility' typing_mode) eqn: Hmv. typing_mode_contradiction typing_mode. auto.
  }



  assert (HinGelim : inert ((Gw2 § w2 / v_ref ~ Twr2) § w1 / v_ref ~ Twr1)).
  {
    eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie1 as Htmp;
      eauto. destruct Htmp as [TL1 [TR1 [G12 [G11 [Hbkw1 [Hbky1 _]]]]]].
    eapply ref_elim_renv_corr in Hcor; eauto using inert_fresh_telescope.
    2: { hnf. intros Hin. apply fv_env_to_binds in Hin as [v Hbiw1']; eauto.
         eapply renv_corr_binds_vk with (G:=G) in Hbiw1' as [_ Hvkw1'];
           eauto.
         eapply env_elim_binds in Hee1; eauto.
         eapply renv_corr_binds_vk with (G:=G) in Hee1 as [Hvkw1 _]; eauto.
         eapply ctx_k_ne_false with (vk1:=v_loc) (vk2:=v_ref); eauto; easy. }

    eapply renv_corr_binds_to_binds_after with (G1:=Gw1) in Hbie2'
        as [TL2 [TR2 [G22 [G21 [Hbkw2 [Hbky2 _]]]]]]; eauto.

    constructor; eauto. constructor; eauto.

    1: eapply inert_binds_inert in HbiG2; eauto;
    destruct HbiG2 as [vkw Hity].
    5: eapply inert_binds_inert in HbiG1; eauto;
    destruct HbiG1 as [vkw Hity].
    1,5: inversions Hity; simpl in *; try easy.
    all: try match goal with | Hf: ?a = false, Ht: ?a=true |- _ => rewrite Hf in Ht; discriminate end.
    1,2: rename H into Hrcd, H0 into Hrfin, H1 into Hidp.
    1: inversions Hro2.
    2: inversions Hro1.
    1,2: rename H2 into Hrfro; constructor; eauto.
    1,2: hnf in *; rewrite Hmrc in *.
    1,2: inversions Hrfin.
    all: try solve[inversions Hrfro; eauto;
                   eapply refacp_inert_mutability_ref; false].

    4: simpl; rewrite notin_union, notin_singleton; split; eauto 2.
    4,1: clears G21 G22 Ty2 Ty1 G11 G12.
    2: destruct Hre1 as [vkw1 [Tw1 [G11 [G12 [? ?]]]]]; subst.
    1,2: destruct Hre2 as [vkw2 [Tw2 [G21 [G22 [? ?]]]]]; subst.
    2: apply inert_ok in HinG; apply ctx_ok_concat_inv in HinG.
    1,2: apply inert_ok in HinGw1; apply ctx_ok_concat_inv in HinGw1.
    2: destruct HinG as [HokG11 [_ Hdis11]].
    1,2: destruct HinGw1 as [HokG21 [_ Hdis22]].
    2: assert (Htmp : w1 §# G11 §& subst_ctx w1 y1 G12 ->
                      w1 §# G21 §& subst_ctx w2 y2 G22)
      by (rewrite H; repeat(rewrite ctx_dom_concat);
          rewrite ctx_dom_map; simpl;
          repeat(rewrite notin_union); rewrite notin_singleton; tauto);
    eapply Htmp; clear Htmp.
    1,2: hnf; rewrite <- ctx_dom_concat_map_r, ctx_dom_concat, in_union.
    2: inversions HokG11; clear H2; rename H5 into HfrG11.
    1: inversions HokG21; clear H1; rename H4 into HfrG21.
    1,2: intros [Hin1 | Hin2]; eauto.
    2: rename Hdis11 into Hdis. 1: rename Hdis22 into Hdis.
    1,2: simpl in Hdis.
    1,2: erewrite <- in_empty, <- Hdis, in_inter, in_union, in_singleton;
    eauto using in_singleton_self.

    3: simpl; rewrite notin_union; split.

    4,1: clears G11 G12 Ty2 Ty1.
    2:
      eapply ref_elim_ctx_renv_corr with (e := config_env c1) in Hre1;
    eauto using ref_elim_ctx_inert, inert_fresh_telescope.
    2: destruct Hre1 as
      [G01 [G02 [G2 [Ty1 [Tw1 [? [Hfry1 ?]]]]]]]; subst.
    2: eapply ref_elim_ctx_notin_fv with (e:=ew1); try eassumption; eauto;
    eauto using ref_elim_ctx_inert, inert_fresh_telescope,
      env_elim_correct; eauto.
    1: eapply ref_elim_ctx_renv_corr with (e:=ew1) in Hre2;
    eauto using ref_elim_ctx_inert, inert_fresh_telescope,
      env_elim_correct; eauto.
    1: destruct Hre2 as [G11 [G12 [G13 [Ty2 [Tw2 [? [Hfry2 ?]]]]]]]; subst.
    1,2: rewrite ctx_fv_concat, notin_union;
    split; eauto using notin_subst_ctx_fv_self.
    1: rename HinGw1 into HinGw. 2: rename HinG into HinGw.
    1,2: apply inert_concat_invert_l in HinGw; inversions HinGw; eauto.

    2: try easy.

    1: rename HbiG2 into HbiG. 2: rename HbiG1 into HbiG.
    1: rename Hro2 into Hro. 2 : rename Hro1 into Hro.
    1,2: eapply inert_binds_fv in HbiG as Hfr;
    eauto; apply typ_ro_fv in Hro as Hiny.
    1,2: hnf; hnf in Hiny; introv Hin; specialize (Hiny _ Hin).
    1: rename Hbie2 into Hbie. 2: rename Hbie1 into Hbie.
    2: {
      assert (HbiyTL : ctx_binds y1 TL1 G) by
      eauto using ctx_binds_k_after_trans, ctx_binds_k_after_binds.
      eapply renv_corr_ref_loc_fv with (G:=G) (Ty:=TL1) (Tw:=TR1) in Hbie;
      eauto using config_renv_corr, config_ctx_inert,
        inert_fresh_telescope, ctx_binds_k_after_binds.
      eapply ref_elim_ctx_neq_binds in HbiyTL as HbiyTL1;
        eauto using inert_fresh_telescope.
      rewrite subst_fresh in HbiyTL1 by eauto using renv_corr_ref_loc_fv.
      clear HbiyTL; rename HbiyTL1 into HbiyTL.
      eapply ref_elim_ctx_neq_binds in HbiyTL;
        eauto using inert_fresh_telescope.
      eapply ctx_binds_functional with (v1:=Ty1) in HbiyTL; eauto; subst.
      eapply fv_subst_var_subset in Hiny.
      rewrite in_union, in_singleton in Hiny.
      destruct Hiny as [Hiny | Hiny]; try tauto; eauto.
    }
    1: {
      assert (HbiyTL : exists TL2 TR2, ctx_binds y2 TL2 G /\ ctx_binds w2 TR2 G).
      { eapply renv_corr_binds_to_ref_loc with (G:=G) in Hbie;
          eauto using config_renv_corr.
        destruct Hbie as [? [? [? [? [? [? ?]]]]]]; eauto. }
      clear Hbky2 Hbkw2 TR2 TL2.
      destruct HbiyTL as [TL2 [TR2 [HbiyTL HbiyTR]]].
    eapply renv_corr_ref_loc_fv with (G:=G) (Ty:=TL2) (Tw:=TR2) in Hbie;
      eauto using config_renv_corr, config_ctx_inert, inert_fresh_telescope.
      eapply ref_elim_ctx_neq_binds in HbiyTL;
        eauto using inert_fresh_telescope.
      eapply ref_elim_ctx_neq_binds in HbiyTL as HbiyTL2;
        eauto using inert_fresh_telescope.
      rewrite subst_commut in HbiyTL2; eauto.
      destruct (classicT (y1 = w2)) as [? | _]; subst; try tauto.
      rewrite subst_fresh with (x:=w2) in HbiyTL2 by
          eauto using renv_corr_ref_loc_fv.
      clear HbiyTL; rename HbiyTL2 into HbiyTL.
      eapply ctx_binds_functional with (v1:=Ty2) in HbiyTL; eauto; subst.
      eapply fv_subst_var_subset in Hiny.
      rewrite in_union, in_singleton in Hiny.
      destruct Hiny as [Hiny | Hiny]; try tauto; eauto.
    }
  }

  pose proof (sef_neq_elim_ty_hole (Gw1:=Gw1) (Gw2:=Gw2)
                                       (ew1:=ew1) (ew2:=ew2)
                                       (y1:=y1) (y2:=y2)
                                       Hbie1 Hbie2 Hre1 Hre2 Hee1 Hee2)
    as [Te Htyelim].
  eapply weaken_push with (x:=w2) (vk:=v_ref) (t:=Twr2) in Htyelim;
    eauto.
  eapply weaken_push with (x:=w1) (vk:=v_ref) (t:=Twr1) in Htyelim;
    eauto.
  2,3,4 : inversions HinGelim; eauto 3.
  2: inversions H3; eauto.

  assert (HsGelim : sealed ((Gw2 § w2 / v_ref ~ Twr2) § w1 / v_ref ~ Twr1))
    by (
      repeat(eapply ctx_sealed_push);
      eauto using typ_ro_sealed, ctx_closed_binds, ref_elim_ctx_sealed
    ).

  repeat(split); eauto.

  assert (Hcorelimp0 : renv_corr ((Gw2 § w2 / v_ref ~ Twr2)
                                   § w1 / v_ref ~ Twr1)
                         (config_env c1)).
  { pose proof (sef_neq_ro_ty_heap Hbie1 Hbie2 HbiG1 HbiG2
                  Hre1 Hre2 Hro1 Hro2) as Hhc. clear Hcor.
    destruct Hhc as [_ [_ [_ [_ [_ [_ [Hcor _]]]]]]]. eauto. }

  assert (Hcorelimp : renv_corr ((Gw2 § w2 / v_ref ~ Twr2)
                                   § w1 / v_ref ~ Twr1)
                                (ew2 & w1 ~ y1)).
  { pose proof (sef_neq_ro_ty_heap Hbie1 Hbie2 HbiG1 HbiG2
                  Hre1 Hre2 Hro1 Hro2) as Hhc. clear Hcor.
    destruct Hhc as [_ [_ [_ [_ [_ [_ [Hcor _]]]]]]].
    eapply renv_corr_binds_more with (e' := ew2 & w2 ~ y2 & w1 ~ y1) in Hcor;
      eauto using renv_corr_eremove.

    - constructor. constructor. eauto.
      eapply env_elim_ok_notin with (e:=ew1); eauto.
      rewrite dom_concat, dom_single, notin_union, notin_singleton.
      split; eauto 1. introv Hbi. apply dom_to_binds in Hbi as [y1' Hbi].
      eapply env_elim_binds_more in Hee2; eauto 3.
      eapply binds_to_dom in Hee2. eapply env_elim_ok_notin in Hee2; eauto.

    - introv Hbi. apply binds_concat_inv in Hbi.
      rewrite dom_single, notin_singleton in Hbi.
      destruct Hbi as [Hbi | [_ Hbi]].
      1: apply binds_single_inv in Hbi; destruct Hbi as [? ?]; subst; eauto.
      apply binds_concat_inv in Hbi.
      rewrite dom_single, notin_singleton in Hbi.
      destruct Hbi as [Hbi | [_ Hbi]].
      1: apply binds_single_inv in Hbi; destruct Hbi as [? ?]; subst; eauto.
      eauto using env_elim_binds_more. }

  assert (Hcorelim : renv_corr ((Gw2 § w2 / v_ref ~ Twr2)
                                   § w1 / v_ref ~ Twr1)
                        ew2).
  { rewrite <- concat_empty_r in Hcorelimp.
    eapply renv_corr_eremove in Hcorelimp; eauto.
    rewrite concat_empty_r in Hcorelimp. easy. }

  rewrite Htrmc1 in *. simpl in *.
  destruct (classicT (w1 = w1)) as [_ | ?]; subst; try congruence.
  destruct (classicT (w2 = w1)) as [? | _]; subst; try congruence.
  destruct (classicT (w2 = w2)) as [_ | ?]; subst; try congruence.
  destruct (classicT (y1 = w2)) as [? | _]; subst; try congruence.

  {
    remember ((Gw2 § w2 / v_ref ~ Twr2) § w1 / v_ref ~ Twr1) as F.
    remember (trm_call (avar_f y1) m (avar_f y2)) as t.
    induction Htyelim; subst.
    all: try (inversions Heqt).

    (* Base case *)
    { rename H into Hvisy1, H0 into Hvisy2,
        H1 into Htyrcdy1, H4 into Htyopy1, H2 into Hty2, e into ew2,
        H3 into Hindep.
      simpl in Hindep. specialize (Hindep Hms).
      apply general_to_tight_typing, tight_to_invertible,
        invertible_to_precise_met_dec in Htyrcdy1; eauto.
      destruct Htyrcdy1 as
        [L [S0 [TC' [TZ' [TR' [Hty1 [HsubZ [HsubC [Hidp HsubR]]]]]]]]].
      clear Hvisy1 Hvisy2.
      eapply pf_inert_rcd_U in Hty1 as HeqS0; eauto.
      destruct HeqS0 as [U [S HeqS0]]; subst.
      apply pf_binds in Hty1 as Hbiy1.
      assert (ctx_binds w1 Twr1
                ((Gw2 § w2 / v_ref ~ Twr2) § w1 / v_ref ~ Twr1))
        as Hbiw1 by eauto.
      pose proof (renv_corr_binds_ref_loc_corr
                    (inert_ok HinGelim) Hcorelimp
                    (binds_push_eq _ _ _) Hbiw1 Hbiy1) as Htc.
      pose proof ty_ro_loc as Halpha.
      inversions Htc. simpl in *.
      rewrite Hmrcr in *.
      destruct H1 as [Tbnd [HeqSw HeqS]].
      apply pf_binds in Hty1 as HbipfGw. rename S into Tmbb, S0 into Sw.
      inversions HbipfGw; try congruence.
      clear H3. rename H6 into HbipfGw.
      inversions HbipfGw; try congruence.
      clear H3. rename H6 into HbipfGw.
      eapply (ctx_binds_functional HbipfGw) in HbiG1.
      rewrite <- HbiG1 in *. clear HbiG1.
      clear Hbiw1.
      destruct Htysef as [Htyw1 [Htyw2 Hir]].

      apply general_to_tight_typing, tight_to_invertible,
        invertible_to_precise_met_dec in Htyw1; eauto.
       destruct Htyw1 as
         [L0 [S0 [TC0' [TZ0' [TR0' [Htyw1 [HsubZ0 [HsubC0 HsubR0]]]]]]]].
       eapply pf_inert_rcd_U in Htyw1 as HeqS0; eauto.
       destruct HeqS0 as [U1 [S1 Heq]]. rewrite Heq in *. clear Heq.
       apply pf_binds in Htyw1 as HbipfG.
       eapply renv_corr_binds_ref_to_loc in HbipfG
           as [TL [Hbiy1G [_ Htycor]]]; eauto.
       inversions Htycor. rename H2 into Htycor. simpl in Htycor.
       rewrite Hmrcr in *.
       destruct Htycor as [Tbndw1 [HeqS1 HeqU0]].
       rewrite HeqS1 in *; rewrite HeqU0 in *; clear HeqS1 HeqU0.
       eapply ref_elim_ctx_neq_binds with (w:=w1) (y:=y1) in Hbiy1G;
         eauto using inert_fresh_telescope.
       eapply ref_elim_ctx_neq_binds with (w:=w2) (y:=y2) in Hbiy1G;
         eauto using inert_fresh_telescope.
       eapply ctx_binds_n with (v2:=w2) (vk:=v_ref) (t2:=Twr2) in Hbiy1G;
         eauto.
       eapply ctx_binds_n with (v2:=w1) (vk:=v_ref)
                               (t2:=typ_and (typ_bnd U)
                                      (typ_rcd
                                         (dec_typ (typ_label_refcap
                                                     typ_refcap_mutability)
                                            typ_bot Tbnd)))
         in Hbiy1G;
         eauto.
       apply (ctx_binds_functional Hbiy1) in Hbiy1G.
       simpl in Hbiy1G. inversions Hbiy1G. clear U0.

       eapply pf_record_sub in Htyw1 as Hhas; eauto.
       destruct Hhas as [[Hhas _] | [Hcontra _]];
         try(inversions Hcontra).
       eapply pf_record_sub in Hty1 as Hhas'; eauto.
       destruct Hhas' as [[Hhas' _] | [Hcontra _]];
         try(inversions Hcontra).
       eapply record_has_subst with (x:=w1) (y:=y1) in Hhas.
       rewrite subst_open_commut, substi_eq_var in Hhas.
       eapply record_has_subst with (x:=w2) (y:=y2) in Hhas.
       rewrite subst_open_commut, substi_ne_var in Hhas; eauto.
       eapply (record_has_functional _ Hhas) in Hhas'; eauto.
       simpl in Hhas'.

      exists (open_met w2 w1 TR').
      apply ty_trm_call with (TC:=TC') (TR:=TR') (TZ:=TZ'); eauto 3.
      all: inversions Hhas'.

      3 : {
        (* tyq_typ_open_indep_same says that if typ_indep_b 0 TC'[w₁=y₁]
            then open w₂ TC'[w₁=y₁] ≈ open y₂ TC'[w₁=y₁]. And so it suffices
            to show w₁ : open y₂ TC'[w₁=y₁]. By HsubC one can conclude
            open y₂ TC <: open y₂ TC'. By substitution and weakening,
            HsubC then gives open y₂ TC[w₁=y₁] <: open y₂ TC'[w₁=y₁].
            And so substitution on Htyopy1 gives y₁ : open y₂ TC[w₁=y₁] <:
            open y₂ TC'[w₁=y₁] ≈ open w₂ TC'[w₁=y₁].

            To get from y₁ to w₁, the idea is to know that open y₂ TC :> N.
            And then to show that if a location has a read-only type then
            one may derive the same type for the reference. *)

        pick_fresh_with z (fv_renv (config_env c1)).
        assert (Htmp : z \notin L) by eauto;
          specialize (HsubC _ Htmp); clear Htmp.

        epose proof (subst_end (x:=z) (y:=y2)
                       subst_sub _ _ HsubC)
          as HsubC'.
        assert (ctx_ok
                  (((Gw2 § w2 / v_ref ~ Twr2)
                      § w1 / v_ref ~ typ_and
                      (typ_bnd (subst_typ w2 y2
                                  (subst_typ w1 y1 U1)))
                      (typ_rcd
                         (dec_typ
                            (typ_label_refcap typ_refcap_mutability)
                            typ_bot Tbnd)))
                     § z / v_par ~ TZ)) as Htmp by eauto;
          specialize (HsubC' Htmp); clear Htmp.
        assert (z \notin
                  ctx_fv
                  ((Gw2 § w2 / v_ref ~ Twr2) § w1 / v_ref ~
                     typ_and (typ_bnd ((U1 /[ w1 -> y1]) /[ w2 -> y2]))
                     (typ_rcd
                        (dec_typ (typ_label_refcap typ_refcap_mutability)
                           typ_bot Tbnd)))) as Htmp.
        { simpl. repeat(rewrite notin_union). repeat(split); eauto.
          - introv Hin.
            apply (fv_subst_var_subset (U1 /[ w1 -> y1]) w2 y2) in Hin.
            repeat(rewrite notin_union in Fr).
            rewrite in_union, in_singleton in Hin;
              destruct Hin as [Hin | Hin]; subst; eauto.
            2: { destruct Fr as [[[[[[[[Fr _] _] _] _] _] _] _] _].
                 repeat(rewrite notin_singleton in Fr). tauto. }
            apply (fv_subst_var_subset U1 w1 y1) in Hin.
            rewrite in_union, in_singleton in Hin;
              destruct Hin as [Hin | Hin]; subst; eauto.
            2: { destruct Fr as [[[[[[[[Fr _] _] _] _] _] _] _] _].
                 repeat(rewrite notin_singleton in Fr). tauto. }
            destruct Fr as
              [[[[[[[[[[[[[Fr _] _] _] _] _] _] _] _] _] _] _] _] _].
            easy.
          - repeat(rewrite notin_union in Fr).
            destruct Fr as [[[[[[[[Fr _] _] _] _] _] _] _] _]. tauto. }
        specialize (HsubC' Htmp); clear Htmp.
        assert (Htmp : ok ew2) by eauto; specialize (HsubC' Htmp);
          clear Htmp.
        assert (Htmp : renv_notin z ew2) by eauto;
          specialize (HsubC' Htmp); clear Htmp.
        replace (TZ /[ z -> y2]) with TZ in HsubC'
            by (rewrite subst_fresh; eauto).
        specialize (HsubC' Hty2).
        assert (Htmp : mode_visibility' typing_mode = mode_visibility_hide_method ->
                       ctx_vis
                         ((Gw2 § w2 / v_ref ~ Twr2) § w1 / v_ref ~
                            typ_and (typ_bnd ((U1 /[ w1 -> y1]) /[ w2 -> y2]))
                            (typ_rcd
                               (dec_typ (typ_label_refcap
                                           typ_refcap_mutability)
                                  typ_bot Tbnd))) y2). {
          eauto using ctx_binds_weakens_vis.
        }
          specialize (HsubC' Htmp); clear Htmp.
        repeat(rewrite subst_open_commut_typ in HsubC').
        repeat(rewrite subst_eq_var in HsubC').
        replace (TC /[ z -> y2]) with TC in HsubC'
            by (rewrite subst_fresh; eauto).
        rewrite (subst_commut_ne (w:=w2) (x:=z)) in HsubC'; eauto.
        rewrite (subst_commut_ne (w:=w1) (x:=z)) in HsubC'; eauto.
        replace (TC0' /[ z -> y2]) with TC0' in HsubC'
            by (rewrite subst_fresh; eauto).
        eapply ebinm_subtyp with (e' := config_env c1) in HsubC';
          eauto using env_elim_binds_more.

        specialize (Hidp Hms Hindep).
        apply (tyq_typ_open_indep_same
                 (e := config_env c1) (y1:=w2) (y2:=y2)) in Hidp; eauto.
        eapply tyq_typ_sym, subtyp_tyq in Hidp.
        eapply ty_var_sub. 2: { exact Hidp. } clear Hidp.

        assert (Htmp : z \notin L0) by eauto;
          specialize (HsubC0 _ Htmp); clear Htmp.

        epose proof (subst_end (x:=z) (y:=w2)
                       subst_sub _ _ HsubC0)
          as HsubC0'.
        assert (ctx_ok (G § z / v_par ~ TZv)) as Htmp by eauto;
          specialize (HsubC0' Htmp); clear Htmp.
        assert (z \notin ctx_fv G) as Htmp by eauto;
          specialize (HsubC0' Htmp); clear Htmp.
        assert (ok (config_env c1)) as Htmp by eauto;
          specialize (HsubC0' Htmp); clear Htmp.
        assert (renv_notin z (config_env c1)) as Htmp by eauto;
          specialize (HsubC0' Htmp); clear Htmp.
        replace (TZv /[ z -> w2]) with TZv in HsubC0'
            by (rewrite subst_fresh; eauto).
        specialize (HsubC0' Htyw2).
        assert (mode_visibility' typing_mode = mode_visibility_hide_method ->
                ctx_vis G w2) as Htmp
          by (eauto using ctx_binds_weakens_vis);
          specialize (HsubC0' Htmp); clear Htmp.
        simpl in HsubC0'.
        rewrite subst_open_commut_typ, subst_eq_var in HsubC0'.
        replace (TC0' /[ z -> w2]) with TC0' in HsubC0'
            by (rewrite subst_fresh; eauto).

        eapply ref_elim_subtyp in HsubC0'; eauto using inert_fresh_telescope.
        simpl in HsubC'.
        rewrite subst_open_commut, substi_ne_var in HsubC0'; eauto.

        eapply ref_elim_subtyp in HsubC0'; eauto using inert_fresh_telescope.
        2: {
          eapply ref_elim_renv_corr with
          (e := config_env c1) (G := G); eauto using inert_fresh_telescope.
          hnf. introv Hin. apply fv_env_to_binds in Hin.
          destruct Hin as [w Hbi].
          pose proof (renv_corr_diff (inert_ok HinG) Hcor Hbie1 Hbi).
          easy. eauto. }
        simpl in HsubC0'.
        rewrite subst_open_commut, substi_eq_var in HsubC0'; eauto.
        eapply ebinm_subtyp with (e' := config_env c1) in HsubC0';
          eauto using env_elim_binds_more.

        eapply ebinm_ty_var with (e' := config_env c1) in Htyopy1;
          eauto using env_elim_binds_more.
        eapply ty_var_sub in Htyopy1; eauto.

        eapply Halpha; eauto.

        - eapply notin_fv_open; eauto.
          eapply notin_fv_subst_var; eauto.
          eapply subst_notfv; eauto.

        - eapply weaken_push. eapply weaken_push.
          all: eauto 3 using ref_elim_ctx_ok_correct.
          constructor.
          all: eauto 3 using ref_elim_ctx_ok_correct.
          apply inert_ok in HinGelim. inversions HinGelim. eauto.

        Unshelve.
        inversion HinGelim; subst. clear H3 H4 H6 H7 H8.
        rename H5 into Hinty. inversions Hinty.
        clear H2. rename H4 into HidpU1s. rename H1 into Hisrcd.
        eauto using open_record_type.

      }

      2: {
        (* First use reference elimination on Htyw₂ to type y₂ in Gw₂.
           Then use reference elimination on Hir and HsubZ₀.
           Finally use subgoal alpha. *)
        eapply ref_elim_typ_var in Htyw2; eauto using inert_fresh_telescope.
        eapply ref_elim_typ_var in Htyw2;
          eauto using inert_fresh_telescope.
        simpl in Htyw2.
        rewrite (subst_ne_var y1), subst_eq_var in Htyw2; eauto.

        eapply ref_elim_subtyp in Hir; eauto using inert_fresh_telescope.
        eapply ref_elim_subtyp in Hir;
          eauto using inert_fresh_telescope.
        simpl in Hir.

        eapply tight_to_general in HsubZ0; eauto.
        eapply ref_elim_subtyp in HsubZ0; eauto using inert_fresh_telescope.
        eapply ref_elim_subtyp in HsubZ0;
          eauto using inert_fresh_telescope.

        2,3,4 : eapply ref_elim_renv_corr with
          (e := config_env c1) (G := G); eauto using inert_fresh_telescope.
        2,3,4 : hnf; introv Hin; apply fv_env_to_binds in Hin; eauto.
        2,3,4 : destruct Hin as [w Hbi].
        2,3,4 : pose proof (renv_corr_diff (inert_ok HinG) Hcor Hbie1 Hbi);
        easy.

        eapply ebinm_ty_var with (e' := config_env c1) in Htyw2;
          eauto using env_elim_binds_more.
        eapply ebinm_subtyp with (e' := config_env c1) in Hir;
          eauto using env_elim_binds_more.
        eapply ebinm_subtyp with (e' := config_env c1) in HsubZ0;
          eauto using env_elim_binds_more.

        eapply weaken_push in Htyw2. eapply weaken_push in Htyw2.
        eapply weaken_push in HsubZ0. eapply weaken_push in HsubZ0.
        eapply weaken_push in Hir. eapply weaken_push in Hir.

        all : eauto 3 using ref_elim_ctx_ok_correct.
        2,4,6 : constructor.
        all : eauto 3 using ref_elim_ctx_ok_correct.
        2,3,4 : apply inert_ok in HinGelim; inversions HinGelim; eauto 3.

        eapply ty_var_sub. 2 : exact HsubZ0.
        eapply Halpha; eauto.
        Unshelve.

        eapply subst_notfv; eauto.
        all: try eapply weaken_subtyp.
        all: try unfold obj_typing.
        all: try eapply weaken_ty_var.

      }

      1: {
        (*
           w₁ is bound in the context to μ.U₁[wᵢ=yᵢ]ᵢ ∧ _ and the goal is
           to show it has type {m(TC₀',TZ₀):TR₀}[wᵢ=yᵢ]ᵢ.
           By assumption open y₁ U₁[wᵢ=yᵢ]ᵢ can project out
           {m(TC₀',TZ₀'):TR₀'}[wᵢ=yᵢ]ᵢ. Furthermore,
           by inertness, we know that U₁[wᵢ=yᵢ]ᵢ is independent of
           the self variable.

           Independence says that open w₁ U₁[wᵢ=yᵢ]ᵢ ≈ open y₁ U₁[wᵢ=yᵢ]ᵢ.
           And open w₁ U₁[wᵢ=yᵢ]ᵢ projects out
           {m:(TC₀',TZ₀'):TR₀'}[wᵢ=yᵢ]ᵢ[y₁=w₁].
           The lemma record_has_tyq_met then says that
           {m:(TC₀',TZ₀'):TR₀'}[wᵢ=yᵢ]ᵢ[y₁=w₁] ≈ {m:(TC₀',TZ₀'):TR₀'}[wᵢ=yᵢ]ᵢ.

         *)

        inversion HinGelim; subst. clear H3 H4 H6 H7 H8.
        rename H5 into Hinty. inversions Hinty.
        clear H2. rename H4 into HidpU1s. rename H1 into Hisrcd.

        eapply tyq_typ_open_indep_same with (y1:=y1) (y2:=w1) in HidpU1s;
          eauto.

        eapply pf_binds, inert_binds_fv in Htyw1; eauto.
        simpl in Htyw1. rewrite union_empty_l, notin_union in Htyw1.
        destruct Htyw1 as [Hfrw1 _].
        replace (subst_typ w1 y1 U1) with (U1 /[ w1 -> y1]) in * by eauto.
        rewrite (subst_fresh y1) in HidpU1s; eauto.
        rewrite (subst_fresh y1) in Hhas; eauto.
        rewrite (subst_fresh y1) in HbipfGw; eauto.
        rewrite (subst_fresh y1); eauto.

        eapply record_has_subst with (x:=y1) (y:=w1) in Hhas as Hhas'.
        rewrite subst_open_commut, substi_eq_var in Hhas'.
        assert (y1 \notin fv (U1 /[ w2 -> y2])) as HfryU1.
        { replace (subst_typ w2 y2 U1) with (U1 /[ w2 -> y2]) in HbipfGw
            by eauto.
          eapply inert_binds_fv in HbipfGw; eauto.
          simpl in HbipfGw. repeat(rewrite union_empty_r in HbipfGw).
          easy. }
        rewrite (subst_fresh w1) in Hhas'; eauto.

        eapply pf_record_has in Hhas' as Htyw1rcd.
        2: exact HinGelim.
        2: rewrite (subst_fresh y1); eauto.
        eapply precise_to_general in Htyw1rcd; eauto.
        rewrite (subst_fresh y1) in Htyw1rcd; eauto.
        simpl in Htyw1rcd.

        simpl in Hhas'.
        eapply record_has_tyq_met with (y:=y1) (e := config_env c1) in Hhas';
          eauto.

        3: { simpl in Hhas. exact Hhas. }
        2: { rewrite (subst_fresh y1) in Hisrcd; eauto. }

        eapply tyq_typ_sym, subtyp_tyq, (ty_var_sub Htyw1rcd) in Hhas';
          eauto.

      }
    (* End of base case *) }

    (* Inductive case *)
    { eauto. }

  }

Qed.

(* Lemma 19 in "Pure methods for roDOT" *)
(* c₁ = ⟨w₁.m w₂; σ₁; ρ₁; Σ₁⟩. Then we can type *)
(* c₂ = ⟨w₁.m w₂; (); ρ₁; Σ₁[wᵢ=yᵢ]ᵢ₌₁,₂⟩ in an environment where *)
(* w₁, w₂ are given read-only types. *)
Lemma sef_neq_ro_ty_config Gw1 Gw2 y1 y2 Ty1 Twr1 Ty2 Twr2 :
  binds w1 y1 (config_env c1) ->
  binds w2 y2 (config_env c1) ->
  ctx_binds y1 Ty1 Gw2 ->
  ctx_binds y2 Ty2 Gw2 ->
  ref_elim_ctx G w1 y1 Gw1 ->
  ref_elim_ctx Gw1 w2 y2 Gw2 ->
  typ_ro Ty1 Twr1 ->
  typ_ro Ty2 Twr2 ->
  exists U,
    ty_config (Gw2 § w2 / v_ref ~ Twr2 § w1 / v_ref ~ Twr1)
              ((config_trm c1) ; stack_nil
                               ; (subst_heap w2 y2
                                             (subst_heap w1 y1
                                                         (config_heap c1)))
                               ; (config_env c1)
                               ; (fv_config c1))
              U.
Proof.
  introv Hbie1 Hbie2 HbiG1 HbiG2 Hre1 Hre2 Htyr1 Htyr2.

  pose proof (sef_neq_ro_ty_hole
                (Gw1:=Gw1) (Gw2:=Gw2) (y1:=y1) (y2:=y2)
                (Ty1:=Ty1) (Ty2:=Ty2)
                (Twr1:=Twr1) (Twr2:=Twr2)) as Htytrm.
  specialize (Htytrm HbiG1 HbiG2 Hbie1 Hbie2 Hre1 Hre2 Htyr1 Htyr2).
  destruct Htytrm as [Htytrm [Hins Hseals]].

  pose proof (sef_neq_ro_ty_heap
                (Gw1:=Gw1) (Gw2:=Gw2) (y1:=y1) (y2:=y2)
                (Ty1:=Ty1) (Ty2:=Ty2)
                (Twr1:=Twr1) (Twr2:=Twr2)) as Hhcs.
  specialize (Hhcs Hbie1 Hbie2 HbiG1 HbiG2 Hre1 Hre2 Htyr1 Htyr2).

  assert (Hhn : has_null (fv_config c1)).
  { destruct Htycfg as [T1 Htcfg]. inversions Htcfg; eauto.
    destruct H6. eexists. repeat(rewrite in_union). eauto. }

  assert (Htfs : all_tfree_k_trm ((Gw2 § w2 / v_ref ~ Twr2)
                                   § w1 / v_ref ~ Twr1)
                                (config_trm c1) v_ref).
  { rewrite Htrmc1. hnf. introv Htf. inversions Htf; eexists; eauto. }

  assert (Hccs : config_closed
                   (config_trm c1 ; stack_nil
                                  ; subst_heap w2 y2
                                  (subst_heap w1 y1 (config_heap c1))
                                  ; config_env c1
                                  ; fv_config c1)).
  { destruct Htycfg as [T1 Htcfg]. inversions Htcfg. inversions H8.
    simpl in Htrmc1. rewrite Htrmc1 in *.
    constructor; simpl;
      eauto using stack_closed_nil, closed_subst_var, heap_closed_subst. }


  assert (Hlrne1 : w1 <> y1) by (eauto using renv_corr_diff).
  assert (Hlrne2 : w2 <> y2) by (eauto using renv_corr_diff).
  assert (Hr1l2ne : w1 <> y2) by (eauto using renv_corr_diff).
  assert (Hr2l1ne : w2 <> y1) by (eauto using renv_corr_diff).

  assert (HyS : y1 \in dom (config_heap c1) /\ y2 \in dom (config_heap c1)).
  { destruct Htycfg as [T1 Htcfg]. inversions Htcfg.
   clear H H0 H1 H2 H3 H4 H6 H7 H8 H9 T.
   destruct H5 as [[Hsd Hnf] _].
   destruct Hhcs as [_ [_ [_ [_ [HGL [HGR [Hrc _]]]]]]].
   simpl in Htrmc1. rewrite Htrmc1 in *. simpl in *.
   destruct (classicT (v_loc = v_ref)) as [? | _]; try discriminate.
   destruct (classicT (v_ref = v_ref)) as [_ | ?]; try tauto.
   split. 1: rename Hbie1 into Hbie. 2: rename Hbie2 into Hbie.
   all: eapply renv_corr_binds_loc in Hbie; try eassumption;
     eauto using inert_ok.
   all: eapply ctx_k_push_inv in Hbie;
     destruct Hbie as [[? _] | [_ Hbie]]; try tauto.
   all: eapply ctx_k_push_inv in Hbie;
     destruct Hbie as [[? _] | [_ Hbie]]; try tauto.
   all: eapply ctx_k_in_kdom in Hbie.
   all: rewrite HGL in Hbie; unfold subst_heap in Hbie.
   all: repeat(rewrite dom_map in Hbie); easy. }

  destruct Htytrm as [Te Htytrm]. exists typ_top.
  apply ty_config_c with (T:=Te); eauto using closed_typ_top.


  destruct Htycfg as [T1 Htcfg]. inversions Htcfg.
  clear H H0 H1 H2 H3 H4 H6 H7 H8 H9 T.
  destruct H5 as [[Hsd Hnf] _].
  destruct Hhcs as [_ [_ [_ [_ [HGL [HGR [Hrc _]]]]]]].
  simpl in Htrmc1. rewrite Htrmc1 in *. simpl.

  clear Htytrm Htfs Hhn Hccs Htysef Htycfg Hins Hseals.
  repeat(split).

  - introv Hin. repeat(rewrite in_union in *).
    destruct Hin as [? | [? | Hin]]; try tauto.
    eapply ref_elim_ctx_binds_not_more in Hin; eauto.
    eapply ref_elim_ctx_binds_not_more in Hin; eauto.
    apply Hsd in Hin. simpl in Hin.
    repeat(rewrite in_union in *). destruct_all; try tauto.

  - hnf. introv Hin. unfold fv, CtxFreeVar in Hin.
    repeat(rewrite ctx_fv_push in Hin). repeat(rewrite in_union in *).
    destruct Hin as [Hin | [Hin | Hin]].
    1: rename HbiG1 into HbiG, Htyr1 into Htyr.
    2: rename HbiG2 into HbiG, Htyr2 into Htyr.
    1,2: eapply ctx_binds_fv_subset in HbiG.
    1,2: eapply typ_ro_fv in Htyr.
    1,2: apply Htyr, HbiG in Hin.
    all: eapply ref_elim_ctx_fv_subset in Hin; eauto.
    all: eapply ref_elim_ctx_fv_subset in Hre1; eauto.
    all: eapply (subset_union_2 Hre1) in Hin; eauto using subset_refl.
    all: repeat(rewrite in_union in Hin).
    all: destruct Hin as [[Hin | Hin] | Hin].
    all: try solve[apply Hnf in Hin; simpl in Hin;
                   repeat(rewrite in_union in Hin); eauto].
    all: rewrite in_singleton in Hin; subst.
    all: simpl in *.
    all: destruct (classicT (v_loc = v_ref)) as [? | _]; try discriminate.
    all: destruct (classicT (v_ref = v_ref)) as [_ | ?]; try tauto.

  - hnf. simpl. eapply subset_empty_l.

Qed.

End W1W2Diff.

Section W1W2Same.

(** Now more pain but with w₁ = w₂ *)
Hypothesis (Hwsame : w1 = w2).

Let w := w1.

Lemma sef_eq_elim_ty_hole Gw ew y :
  binds w y (config_env c1) ->
  ref_elim_ctx G w y Gw ->
  env_elim (config_env c1) w y ew ->
  exists Te, (Gw ⁏ ew ⊢ ((config_trm c1) /[ w -> y]) ∶ Te).
Proof.
  clear Htysef.
  destruct Htycfg as [T1 Htycfg']. clear Htycfg.
  destruct Htycfg'; simpl in *.
  rename H into HinG, H0 into Hhc, H1 into Htys, H2 into Htf,
    H5 into Hnfr, H4 into Htfk, H3 into Hty, H6 into Hhn, H7 into Hs,
    H8 into Hclc, H9 into HclU.
  clear Hnfr Htf Htys Htfk Hclc HclU Hhn Hs.
  introv Hbie HeG Hee.
  destruct Hhc as
    [HokG [HwkG [Hoks [Hoke [Hsloc [Heref [Hcore Htyit]]]]]]].
  clear Hoks Hsloc Heref Htyit.

  pose proof (ref_elim_typ_trm (G:=G) (Go:=objctx_empty)
                               (G':=Gw) (e:=e) (e':=ew)
                               (w:=w) (y:=y)
                               (t:=t)
                               (T:=T)) as Htyelim. simpl in *.
  specialize (Htyelim (inert_fresh_telescope HinG)
                HokG Hoke Hcore Hty HeG Hbie Hee).
  eauto.
Qed.

Lemma sef_eq_elim_heap_corr Gw ew y :
  binds w y (config_env c1) ->
  ref_elim_ctx G w y Gw ->
  env_elim (config_env c1) w y ew ->
  heap_correspond Gw ew
    (subst_heap w y (config_heap c1)).
Proof.
  clear Htysef.
  destruct Htycfg as [T1 Htycfg']. clear Htycfg.
  destruct Htycfg'; simpl in *.
  rename H into HinG, H0 into Hhc, H1 into Htys, H2 into Htf,
    H5 into Hnfr, H4 into Htfk, H3 into Hty, H6 into Hhn, H7 into Hs,
    H8 into Hclc, H9 into HclU.
  clear Hnfr Htf Htys Htfk Hclc HclU Hhn Hs.
  introv Hbie HeG Hee.

  forwards: ref_elim_heap; try eassumption;
    eauto using inert_fresh_telescope.
Qed.

Lemma sef_eq_ro_ty_heap Gw y Ty Twr :
  binds w y (config_env c1) ->
  ctx_binds y Ty Gw ->
  ref_elim_ctx G w y Gw ->
  typ_ro Ty Twr ->
  heap_correspond (Gw § w / v_ref ~ Twr)
    (config_env c1)
    (subst_heap w y (config_heap c1)).
Proof.
  introv Hbie HbiGw Hre Hro.

  assert (Hoke : ok (config_env c1)) by eauto.
  destruct (env_elim_binds_exists config_env_ok Hbie) as
    [ew [Hee Hokew]].

  assert (HinG : inert G) by eauto using config_ctx_inert.
  assert (HcorGe : renv_corr G (config_env c1))
    by eauto using config_renv_corr.

  assert (Hlrne : w <> y) by (eauto using renv_corr_diff).

  assert (HinGw: inert Gw).
  { eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie;
      eauto using inert_ok.
    destruct Hbie as [TL1 [TR1 [G2 [G3 [Hbkw [Hbky _]]]]]].
    eauto using ref_elim_ctx_inert. }

  assert (HcorGew : renv_corr Gw ew).
  { eapply ref_elim_renv_corr with
      (G:=G) (e := config_env c1) (w:=w) (y:=y);
      eauto using inert_fresh_telescope.
    hnf. introv Hin. apply fv_env_to_binds in Hin; eauto using config_env_ok.
    destruct Hin as [v Hbie'].
    eapply renv_corr_binds_loc with (G:=G) in Hbie'; eauto using inert_ok.
    eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie;
      eauto using inert_ok.
    destruct Hbie as [_ [TR1 [G2 [G3 [Hbk [_ _]]]]]].
    apply ctx_binds_k_after_k in Hbk.
    pose proof (ctx_k_eq Hbie' Hbk). discriminate. }

  assert (Hbky : exists Gy, ctx_binds_k_after y v_loc Ty Gw Gy).
  { eapply ctx_binds_to_ctx_binds_k_after in HbiGw;
      eauto using inert_weakens. destruct HbiGw as [Gy [vy Hbk]]. exists Gy.
    replace v_loc with vy; eauto.
    eapply renv_corr_binds_loc with (G:=G) in Hbie; eauto.
    eapply ref_elim_ctx_vk_same with (G':=Gw) in Hbie; eauto.
    eapply ctx_binds_k_after_k in Hbk. exact (ctx_k_eq Hbk Hbie). }

  pose proof (sef_eq_elim_heap_corr Hbie Hre Hee) as Hhc.
  destruct Hbky as [Gy Hbky].

  eapply heap_correspond_epush_nowfr
    with (w:=w) (T1:=Twr) (T2:=Ty) (y:=y)
    in Hhc; eauto 3 using inert_typ_ro_corr, ref_elim_ctx_inert_typ_loc,
    ref_elim_ctx_ok_correct, env_elim_ok_notin.

  eapply heap_correspond_epermute with (e := (ew & w ~ y));
    eauto.

  - introv Hbi. apply binds_concat_inv in Hbi.
    rewrite dom_single, notin_singleton in Hbi.
    destruct Hbi as [Hbi | [_ Hbi]].
    1: apply binds_single_inv in Hbi; destruct Hbi as [? ?]; subst; eauto.
    eauto using env_elim_binds_more.

  - introv Hbi. destruct (classicT (w0 = w)) as [? | Hne]; subst.
    1: replace y0 with y; eauto using binds_functional.
    repeat(eapply binds_push_neq); eauto using env_elim_correct.

Qed.

Lemma sef_eq_ro_ty_hole Gw y Ty Twr :
  ctx_binds y Ty Gw ->
  binds w y (config_env c1) ->
  ref_elim_ctx G w y Gw ->
  typ_ro Ty Twr ->
  (exists Te, (Gw § w / v_ref ~ Twr ⁏ (config_env c1)
            ⊢ (config_trm c1) ∶ Te))
  /\ (inert (Gw § w / v_ref ~ Twr))
  /\ (sealed (Gw § w / v_ref ~ Twr)).
Proof.
  introv HbiG Hbie Hre Hro.

  destruct (env_elim_binds_exists config_env_ok Hbie) as
    [ew [Hee Hokew]].

  assert (Hokce : ok (config_env c1)) by eauto using config_env_ok.
  assert (HinG : inert G) by (destruct Htycfg as [T1 [? ?]]; eauto).
  assert (HsG : sealed G) by (destruct Htycfg as [T1 [? ?]]; eauto).
  assert (Hcor : renv_corr G (config_env c1))
    by (destruct Htycfg as [T1 [? ?]]; eauto).

  assert (Hlrne1 : w <> y) by (eauto using renv_corr_diff).



  assert (mode_refcaps' typing_mode = mode_refcaps_mutability) as Hmrc.
  {
    destruct (mode_refcaps' typing_mode) eqn: Hmrc. typing_mode_contradiction typing_mode. auto.
  }
  assert (mode_has_fun' typing_mode = false) as Hmf.
  {
    destruct (mode_has_fun' typing_mode) eqn: Hmf. typing_mode_contradiction typing_mode. auto.
  }
  assert (mode_refcap_corr' typing_mode = mode_refcap_corr_ref_mutability) as Hmrcr.
  {
    destruct (mode_refcap_corr' typing_mode) eqn: Hmrcr. typing_mode_contradiction typing_mode. auto.
  }
  assert (mode_visibility' typing_mode = mode_visibility_hide_method) as Hmv.
  {
    destruct (mode_visibility' typing_mode) eqn: Hmv. typing_mode_contradiction typing_mode. auto.
  }




  assert (HinGw : inert Gw).
  { eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie as Htmp;
      eauto. destruct Htmp as [TL1 [TR1 [G12 [G11 [Hbkw [Hbky _]]]]]].
    eapply ref_elim_ctx_inert in Hre as HinGw; eauto. }

  assert (HinGelim : inert ((Gw § w / v_ref ~ Twr))).
  {
    eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie as Htmp;
      eauto. destruct Htmp as [TL [TR [G12 [G11 [Hbkw [Hbky _]]]]]].
    eapply ref_elim_renv_corr in Hcor; eauto using inert_fresh_telescope.
    2: { hnf. intros Hin. apply fv_env_to_binds in Hin as [v Hbiw']; eauto.
         eapply renv_corr_binds_vk with (G:=G) in Hbiw' as [_ Hvkw'];
           eauto.
         eapply env_elim_binds in Hee; eauto.
         eapply renv_corr_binds_vk with (G:=G) in Hee as [Hvkw _]; eauto.
         eapply ctx_k_ne_false with (vk1:=v_loc) (vk2:=v_ref); eauto; easy. }

    constructor; eauto.

    1: eapply inert_binds_inert in HbiG; eauto;
    destruct HbiG as [vkw Hity].
    1: inversions Hity; simpl in *; try easy.
    all: try match goal with | Hf: ?a = false, Ht: ?a=true |- _ => rewrite Hf in Ht; discriminate end.
    1: rename H into Hrcd, H0 into Hrfin, H1 into Hidp.
    1: inversions Hro.
    1: rename H2 into Hrfro; constructor; eauto.
    1: hnf in *; info_typing_mode_contradiction typing_mode;
    inversions Hrfin.
    all: try solve[inversions Hrfro; eauto;
                   eapply refacp_inert_mutability_ref; false].

    1: clears Ty G11 G12.
    1: destruct Hre as [vkw [Tw [G21 [G22 [? ?]]]]]; subst.
    1: eapply inert_ok, ctx_ok_middle_inv in HinG.
    1: hnf; rewrite <- ctx_dom_concat_map_r, ctx_dom_concat, in_union.
    1: destruct HinG as [_ [HinG1 HinG2]].
    introv [? | ?]; try eauto.

    1: clears G11 G12 Ty.
    1: eapply ref_elim_ctx_renv_corr with (e := config_env c1) in Hre;
    eauto using ref_elim_ctx_inert, inert_fresh_telescope,
      env_elim_correct; eauto.
    1: destruct Hre as [G11 [G12 [G13 [Ty [Tw [? [Hfry ?]]]]]]]; subst.
    1: rewrite ctx_fv_concat, notin_union;
    split; eauto using notin_subst_ctx_fv_self.
    1: apply inert_concat_invert_l in HinG; inversions HinG; eauto.

    1: eapply inert_binds_fv in HbiG as Hfr;
    eauto; apply typ_ro_fv in Hro as Hiny.
    1: hnf; hnf in Hiny; introv Hin; specialize (Hiny _ Hin).
    assert (HbiyTL : ctx_binds y TL G) by
      eauto using ctx_binds_k_after_trans, ctx_binds_k_after_binds.
    eapply renv_corr_ref_loc_fv with (G:=G) (Ty:=TL) (Tw:=TR) in Hbie;
      eauto using config_renv_corr, config_ctx_inert,
        inert_fresh_telescope, ctx_binds_k_after_binds.
    eapply ref_elim_ctx_neq_binds in HbiyTL as HbiyTL;
      eauto using inert_fresh_telescope.
    eapply ctx_binds_functional with (v1:=Ty) in HbiyTL; eauto; subst.
    eapply fv_subst_var_subset in Hiny.
    rewrite in_union, in_singleton in Hiny.
    destruct Hiny as [Hiny | Hiny]; try tauto; eauto.
  }

  pose proof (sef_eq_elim_ty_hole (Gw:=Gw) (ew:=ew) (y:=y)
                                       Hbie Hre Hee)
    as [Te Htyelim].
  eapply weaken_push with (x:=w) (vk:=v_ref) (t:=Twr) in Htyelim;
    eauto.
  2 : inversions HinGelim; eauto 3.

  assert (HsGelim : sealed (Gw § w / v_ref ~ Twr))
    by (
      repeat(eapply ctx_sealed_push);
      eauto using typ_ro_sealed, ctx_closed_binds, ref_elim_ctx_sealed
    ).

  repeat(split); eauto.

  assert (Hcorelimp0 : renv_corr (Gw § w / v_ref ~ Twr)
                         (config_env c1)).
  { pose proof (sef_eq_ro_ty_heap Hbie HbiG Hre Hro) as Hhc. clear Hcor.
    destruct Hhc as [_ [_ [_ [_ [_ [_ [Hcor _]]]]]]]. eauto. }

  assert (Hcorelimp : renv_corr (Gw § w / v_ref ~ Twr)
                        (ew & w ~ y)).
  { pose proof (sef_eq_ro_ty_heap Hbie HbiG Hre Hro) as Hhc. clear Hcor.
    destruct Hhc as [_ [_ [_ [_ [_ [_ [Hcor _]]]]]]].
    eapply renv_corr_binds_more with (e' := ew & w ~ y) in Hcor; eauto.
    - constructor. eauto. eapply env_elim_ok_notin; eauto.
    - introv Hbi. apply binds_concat_inv in Hbi.
      rewrite dom_single, notin_singleton in Hbi.
      destruct Hbi as [Hbi | [_ Hbi]].
      1: apply binds_single_inv in Hbi; destruct Hbi as [? ?]; subst; eauto.
      eauto using env_elim_binds_more. }



  assert (Hcorelim : renv_corr (Gw § w / v_ref ~ Twr)
                       ew).
  { rewrite <- concat_empty_r in Hcorelimp.
    eapply renv_corr_eremove in Hcorelimp; eauto.
    rewrite concat_empty_r in Hcorelimp. easy. }

  rewrite Htrmc1 in *. simpl in *.
  destruct (classicT (w1 = w)) as [_ | ?]; subst;
    try solve [destruct n; eauto].
  replace w2 with w in *; eauto.
  destruct (classicT (w = w)) as [_ | ?]; subst; try congruence.

  {
    remember (Gw § w / v_ref ~ Twr) as F.
    remember (trm_call (avar_f y) m (avar_f y)) as t.
    induction Htyelim; subst.
    all: try (inversions Heqt).

    (* Base case *)
    { rename H into Hvisy; clear H0.
      rename H3 into Hindep, H4 into Htyopy, H2 into Hty,
        H1 into Htyrcdy, e into ew.
      simpl in Hindep. specialize (Hindep Hms).
      apply general_to_tight_typing, tight_to_invertible,
        invertible_to_precise_met_dec in Htyrcdy; eauto.
      destruct Htyrcdy as
        [L [S0 [TC' [TZ' [TR' [Hty1 [HsubZ [HsubC [Hidp HsubR]]]]]]]]].
      clear Hvisy.
      eapply pf_inert_rcd_U in Hty1 as HeqS0; eauto.
      destruct HeqS0 as [U [S HeqS0]]; subst.
      apply pf_binds in Hty1 as Hbiy.
      assert (ctx_binds w Twr (Gw § w / v_ref ~ Twr)) as Hbiw by eauto.
      pose proof (renv_corr_binds_ref_loc_corr
                    (inert_ok HinGelim) Hcorelimp
                    (binds_push_eq _ _ _) Hbiw Hbiy) as Htc.
      pose proof ty_ro_loc as Halpha.
      inversions Htc. simpl in *.
      rewrite Hmrcr in *.
      destruct H1 as [Tbnd [HeqSw HeqS]].
      apply pf_binds in Hty1 as HbipfGw. rename S into Tmbb, S0 into Sw.
      inversions HbipfGw; try congruence.
      clear H3. rename H6 into HbipfGw.
      eapply (ctx_binds_functional HbipfGw) in HbiG.
      rewrite <- HbiG in *. clear HbiG Hbiw.
      destruct Htysef as [Htyw1 [Htyw2 Hir]].

      apply general_to_tight_typing, tight_to_invertible,
        invertible_to_precise_met_dec in Htyw1; eauto.
      destruct Htyw1 as
        [L0 [S0 [TC0' [TZ0' [TR0' [Htyw1 [HsubZ0 [HsubC0 HsubR0]]]]]]]].
      eapply pf_inert_rcd_U in Htyw1 as HeqS0; eauto.
      destruct HeqS0 as [U1 [S1 Heq]]. rewrite Heq in *. clear Heq.
      apply pf_binds in Htyw1 as HbipfG.
      eapply renv_corr_binds_ref_to_loc in HbipfG
          as [TL [Hbiy1G [_ Htycor]]]; eauto.
      inversions Htycor. rename H2 into Htycor. simpl in Htycor.
      rewrite Hmrcr in *.
      destruct Htycor as [Tbndw1 [HeqS1 HeqU0]].
      rewrite HeqS1 in *; rewrite HeqU0 in *; clear HeqS1 HeqU0.
      eapply ref_elim_ctx_neq_binds with (w:=w) (y:=y) in Hbiy1G;
        eauto using inert_fresh_telescope.
      eapply ctx_binds_n with (v2:=w) (vk:=v_ref)
                              (t2:=typ_and (typ_bnd U)
                                     (typ_rcd
                                        (dec_typ (typ_label_refcap
                                                    typ_refcap_mutability)
                                           typ_bot Tbnd)))
        in Hbiy1G;
        eauto.
      apply (ctx_binds_functional Hbiy) in Hbiy1G.
      simpl in Hbiy1G. inversions Hbiy1G. clear U0.

      eapply pf_record_sub in Htyw1 as Hhas; eauto.
      destruct Hhas as [[Hhas _] | [Hcontra _]];
        try(inversions Hcontra).
      eapply pf_record_sub in Hty1 as Hhas'; eauto.
      destruct Hhas' as [[Hhas' _] | [Hcontra _]];
        try(inversions Hcontra).
      eapply record_has_subst with (x:=w) (y:=y) in Hhas.
      rewrite subst_open_commut, substi_eq_var in Hhas.
      eapply (record_has_functional _ Hhas) in Hhas'; eauto.
      simpl in Hhas'.

      exists (open_met w w TR').
      apply ty_trm_call with (TC:=TC') (TR:=TR') (TZ:=TZ'); eauto 3.
      all: inversions Hhas'.

      3 : {
        (* tyq_typ_open_indep_same says that if typ_indep_b 0 TC'[w₁=y₁]
           then open w₂ TC'[w₁=y₁] ≈ open y₂ TC'[w₁=y₁]. And so it suffices
           to show w₁ : open y₂ TC'[w₁=y₁]. By HsubC one can conclude
           open y₂ TC <: open y₂ TC'. By substitution and weakening,
           HsubC then gives open y₂ TC[w₁=y₁] <: open y₂ TC'[w₁=y₁].
           And so substitution on Htyopy1 gives y₁ : open y₂ TC[w₁=y₁] <:
           open y₂ TC'[w₁=y₁] ≈ open w₂ TC'[w₁=y₁].

           To get from y₁ to w₁, the idea is to know that open y₂ TC :> N.
           And then to show that if a location has a read-only type then
           one may derive the same type for the reference. *)

        pick_fresh_with z (fv_renv (config_env c1)).
        assert (Htmp : z \notin L) by eauto;
          specialize (HsubC _ Htmp); clear Htmp.

        epose proof (subst_end (x:=z) (y:=y)
                       subst_sub _ _ HsubC)
          as HsubC'.
        assert (ctx_ok
                  (((Gw § w / v_ref ~ typ_and
                       (typ_bnd (subst_typ w y U1))
                       (typ_rcd
                          (dec_typ
                             (typ_label_refcap typ_refcap_mutability)
                             typ_bot Tbnd)))
                      § z / v_par ~ TZ))) as Htmp by eauto;
          specialize (HsubC' Htmp); clear Htmp.
        assert (z \notin
                  ctx_fv
                  ((Gw § w / v_ref ~
                      typ_and (typ_bnd ((U1 /[ w -> y])))
                      (typ_rcd
                         (dec_typ (typ_label_refcap typ_refcap_mutability)
                            typ_bot Tbnd))))) as Htmp.
        { simpl. repeat(rewrite notin_union). repeat(split); eauto.
          - introv Hin.
            apply (fv_subst_var_subset U1 w y) in Hin.
            repeat(rewrite notin_union in Fr).
            rewrite in_union, in_singleton in Hin;
              destruct Hin as [Hin | Hin]; subst; eauto.
            2: { destruct Fr as [[[[[[[[Fr _] _] _] _] _] _] _] _].
                 repeat(rewrite notin_singleton in Fr). tauto. }
            destruct Fr as
              [[[[[[[[[[[[[Fr _] _] _] _] _] _] _] _] _] _] _] _] _].
            easy.
          - repeat(rewrite notin_union in Fr).
            destruct Fr as [[[[[[[[Fr _] _] _] _] _] _] _] _]. tauto. }
        specialize (HsubC' Htmp); clear Htmp.
        assert (Htmp : ok ew) by eauto; specialize (HsubC' Htmp);
          clear Htmp.
        assert (Htmp : renv_notin z ew) by eauto;
          specialize (HsubC' Htmp); clear Htmp.
        replace (TZ /[ z -> y]) with TZ in HsubC'
            by (rewrite subst_fresh; eauto).
        specialize (HsubC' Hty).
        assert (Htmp : mode_visibility' typing_mode =
                         mode_visibility_hide_method ->
                       ctx_vis
                         (Gw § w / v_ref ~
                            typ_and (typ_bnd (U1 /[ w -> y]))
                            (typ_rcd
                               (dec_typ (typ_label_refcap
                                           typ_refcap_mutability)
                                  typ_bot Tbnd))) y)
          by (eauto using ctx_binds_weakens_vis);
          specialize (HsubC' Htmp); clear Htmp.
        repeat(rewrite subst_open_commut_typ in HsubC').
        repeat(rewrite subst_eq_var in HsubC').
        replace (TC /[ z -> y]) with TC in HsubC'
            by (rewrite subst_fresh; eauto).
        rewrite (subst_commut_ne (w:=w) (x:=z)) in HsubC'; eauto.
        replace (TC0' /[ z -> y]) with TC0' in HsubC'
            by (rewrite subst_fresh; eauto).
        eapply ebinm_subtyp with (e' := config_env c1) in HsubC';
          eauto using env_elim_binds_more.
        specialize (Hidp Hms Hindep).
        apply (tyq_typ_open_indep_same
                 (e := config_env c1) (y1:=w) (y2:=y)) in Hidp; eauto.
        eapply tyq_typ_sym, subtyp_tyq in Hidp.
        eapply ty_var_sub. 2: { exact Hidp. } clear Hidp.

        assert (Htmp : z \notin L0) by eauto;
          specialize (HsubC0 _ Htmp); clear Htmp.

        epose proof (subst_end (x:=z) (y:=w)
                       subst_sub _ _ HsubC0)
          as HsubC0'.
        assert (ctx_ok (G § z / v_par ~ TZv)) as Htmp by eauto;
          specialize (HsubC0' Htmp); clear Htmp.
        assert (z \notin ctx_fv G) as Htmp by eauto;
          specialize (HsubC0' Htmp); clear Htmp.
        assert (ok (config_env c1)) as Htmp by eauto;
          specialize (HsubC0' Htmp); clear Htmp.
        assert (renv_notin z (config_env c1)) as Htmp by eauto;
          specialize (HsubC0' Htmp); clear Htmp.
        replace (TZv /[ z -> w2]) with TZv in HsubC0'
            by (rewrite subst_fresh; eauto).
        specialize (HsubC0' Htyw2).
        assert (mode_visibility' typing_mode = mode_visibility_hide_method ->
                ctx_vis G w) as Htmp
          by (eauto using ctx_binds_weakens_vis);
          specialize (HsubC0' Htmp); clear Htmp.
        simpl in HsubC0'.
        rewrite subst_open_commut_typ, subst_eq_var in HsubC0'.
        replace (TC0' /[ z -> w]) with TC0' in HsubC0'
            by (rewrite subst_fresh; eauto).

        eapply ref_elim_subtyp in HsubC0'; eauto using inert_fresh_telescope.
        simpl in HsubC'.
        rewrite subst_open_commut, substi_eq_var in HsubC0'; eauto.

        simpl in HsubC0'.
        eapply ebinm_subtyp with (e' := config_env c1) in HsubC0';
          eauto using env_elim_binds_more.

        eapply ebinm_ty_var with (e' := config_env c1) in Htyopy;
          eauto using env_elim_binds_more.
        eapply ty_var_sub in Htyopy; eauto.

        eapply Halpha; eauto.

        - eapply notin_fv_open; eauto.
          eapply subst_notfv; eauto.

        - eapply weaken_push.
          all: eauto 3 using ref_elim_ctx_ok_correct.

        Unshelve.
        inversion HinGelim; subst. clear H3 H4 H6 H7 H8.
        rename H5 into Hinty. inversions Hinty.
        clear H2. rename H4 into HidpU1s. rename H1 into Hisrcd.
        eauto using open_record_type.

      }

      2: {
        (* First use reference elimination on Htyw₂ to type y₂ in Gw₂.
           Then use reference elimination on Hir and HsubZ₀.
           Finally use subgoal alpha. *)
        eapply ref_elim_typ_var in Htyw2; eauto using inert_fresh_telescope.
        simpl in Htyw2. rewrite subst_eq_var in Htyw2; eauto.

        eapply ref_elim_subtyp in Hir; eauto using inert_fresh_telescope.
        simpl in Hir.

        eapply tight_to_general in HsubZ0; eauto.
        eapply ref_elim_subtyp in HsubZ0; eauto using inert_fresh_telescope.

        eapply ebinm_ty_var with (e' := config_env c1) in Htyw2;
          eauto using env_elim_binds_more.
        eapply ebinm_subtyp with (e' := config_env c1) in Hir;
          eauto using env_elim_binds_more.
        eapply ebinm_subtyp with (e' := config_env c1) in HsubZ0;
          eauto using env_elim_binds_more.

        eapply weaken_push in Htyw2.
        eapply weaken_push in HsubZ0.
        eapply weaken_push in Hir.

        all : eauto 3 using ref_elim_ctx_ok_correct.
        eapply ty_var_sub. 2 : exact HsubZ0.
        eapply Halpha; eauto.
        eapply subst_notfv; eauto.
        Unshelve.
        all: try eapply weaken_subtyp.
        all: try unfold obj_typing.
        all: try eapply weaken_ty_var.
      }

      1: {
        (*
           w₁ is bound in the context to μ.U₁[wᵢ=yᵢ]ᵢ ∧ _ and the goal is
           to show it has type {m(TC₀',TZ₀):TR₀}[wᵢ=yᵢ]ᵢ.
           By assumption open y₁ U₁[wᵢ=yᵢ]ᵢ can project out
           {m(TC₀',TZ₀'):TR₀'}[wᵢ=yᵢ]ᵢ. Furthermore,
           by inertness, we know that U₁[wᵢ=yᵢ]ᵢ is independent of
           the self variable.

           Independence says that open w₁ U₁[wᵢ=yᵢ]ᵢ ≈ open y₁ U₁[wᵢ=yᵢ]ᵢ.
           And open w₁ U₁[wᵢ=yᵢ]ᵢ projects out
           {m:(TC₀',TZ₀'):TR₀'}[wᵢ=yᵢ]ᵢ[y₁=w₁].
           The lemma record_has_tyq_met then says that
           {m:(TC₀',TZ₀'):TR₀'}[wᵢ=yᵢ]ᵢ[y₁=w₁] ≈ {m:(TC₀',TZ₀'):TR₀'}[wᵢ=yᵢ]ᵢ.

         *)

        inversion HinGelim; subst. clear H3 H4 H6 H7 H8.
        rename H5 into Hinty. inversions Hinty.
        clear H2. rename H4 into HidpU1s. rename H1 into Hisrcd.

        eapply tyq_typ_open_indep_same with (y1:=y) (y2:=w) in HidpU1s;
          eauto.

        eapply pf_binds, inert_binds_fv in Htyw1; eauto.
        simpl in Htyw1. rewrite union_empty_l, notin_union in Htyw1.
        destruct Htyw1 as [Hfrw1 _].
        replace (subst_typ w y U1) with (U1 /[ w -> y]) in * by eauto.
        rewrite (subst_fresh y) in HidpU1s; eauto.
        rewrite (subst_fresh y) in Hhas; eauto.
        rewrite (subst_fresh y) in HbipfGw; eauto.
        rewrite (subst_fresh y); eauto.

        eapply record_has_subst with (x:=y) (y:=w) in Hhas as Hhas'.
        rewrite subst_open_commut, substi_eq_var in Hhas'.
        assert (y \notin fv U1) as HfryU1.
        { replace (subst_typ w y U1) with (U1 /[ w -> y]) in HbipfGw
            by eauto.
          eapply inert_binds_fv in HbipfGw; eauto.
          simpl in HbipfGw. repeat(rewrite union_empty_r in HbipfGw).
          easy. }
        rewrite (subst_fresh w) in Hhas'; eauto.

        eapply pf_record_has in Hhas' as Htyw1rcd.
        2: exact HinGelim.
        2: rewrite (subst_fresh y); eauto.
        eapply precise_to_general in Htyw1rcd; eauto.
        rewrite (subst_fresh y) in Htyw1rcd; eauto.
        simpl in Htyw1rcd.

        simpl in Hhas'.
        eapply record_has_tyq_met with (y:=y) (e := config_env c1) in Hhas';
          eauto.

        2: { rewrite (subst_fresh y) in Hisrcd; eauto. }

        eapply tyq_typ_sym, subtyp_tyq, (ty_var_sub Htyw1rcd) in Hhas';
          eauto.

      }
    (* End of base case *) }

    (* Inductive case *)
    { eauto. }

  }

Qed.

Lemma sef_eq_ro_ty_config Gw y Ty Twr :
  binds w y (config_env c1) ->
  ctx_binds y Ty Gw ->
  ref_elim_ctx G w y Gw ->
  typ_ro Ty Twr ->
  exists U,
    ty_config (Gw § w / v_ref ~ Twr)
              ((config_trm c1) ; stack_nil
                               ; (subst_heap w y
                                    (config_heap c1))
                               ; (config_env c1)
                               ; (fv_config c1))
              U.
Proof.
  introv Hbie HbiG Hre Htyr.

  pose proof (sef_eq_ro_ty_hole
                (Gw:=Gw) (y:=y)
                (Ty:=Ty) (Twr:=Twr)) as Htytrm.
  specialize (Htytrm HbiG Hbie Hre Htyr).
  destruct Htytrm as [Htytrm [Hins Hseals]].

  pose proof (sef_eq_ro_ty_heap
                (Gw:=Gw) (y:=y)
                (Ty:=Ty) (Twr:=Twr)) as Hhcs.
  specialize (Hhcs Hbie HbiG Hre Htyr).

  assert (Hhn : has_null (fv_config c1)).
  { destruct Htycfg as [T1 Htcfg]. inversions Htcfg; eauto.
    destruct H6. eexists. repeat(rewrite in_union). eauto. }

  assert (Htfs : all_tfree_k_trm (Gw § w / v_ref ~ Twr)
                                (config_trm c1) v_ref).
  { rewrite Htrmc1. hnf. introv Htf. inversions Htf; eexists; eauto. }

  assert (Hccs : config_closed
                   (config_trm c1 ; stack_nil
                                  ; (subst_heap w y (config_heap c1))
                                  ; config_env c1
                                  ; fv_config c1)).
  { destruct Htycfg as [T1 Htcfg]. inversions Htcfg. inversions H8.
    simpl in Htrmc1. rewrite Htrmc1 in *.
    constructor; simpl;
      eauto using stack_closed_nil, closed_subst_var, heap_closed_subst. }


  assert (Hlrne : w <> y) by (eauto using renv_corr_diff).

  assert (HyS : y \in dom (config_heap c1)).
  { destruct Htycfg as [T1 Htcfg]. inversions Htcfg.
   clear H H0 H1 H2 H3 H4 H6 H7 H8 H9 T.
   destruct H5 as [[Hsd Hnf] _].
   destruct Hhcs as [_ [_ [_ [_ [HGL [HGR [Hrc _]]]]]]].
   simpl in Htrmc1. rewrite Htrmc1 in *. simpl in *.
   destruct (classicT (v_loc = v_ref)) as [? | _]; try discriminate.
   destruct (classicT (v_ref = v_ref)) as [_ | ?]; try tauto.
   eapply renv_corr_binds_loc in Hbie; try eassumption;
     eauto using inert_ok.
   eapply ctx_k_push_inv in Hbie;
     destruct Hbie as [[? _] | [_ Hbie]]; try tauto.
   eapply ctx_k_in_kdom in Hbie.
   rewrite HGL in Hbie; unfold subst_heap in Hbie.
   repeat(rewrite dom_map in Hbie); easy. }

  destruct Htytrm as [Te Htytrm]. exists typ_top.
  apply ty_config_c with (T:=Te); eauto using closed_typ_top.


  destruct Htycfg as [T1 Htcfg]. inversions Htcfg.
  clear H H0 H1 H2 H3 H4 H6 H7 H8 H9 T.
  destruct H5 as [[Hsd Hnf] _].
  destruct Hhcs as [_ [_ [_ [_ [HGL [HGR [Hrc _]]]]]]].
  simpl in Htrmc1. rewrite Htrmc1 in *. simpl.

  clear Htytrm Htfs Hhn Hccs Htysef Htycfg Hins Hseals.
  repeat(split).

  - introv Hin. repeat(rewrite in_union in *).
    destruct Hin as [? | Hin]; try tauto.
    eapply ref_elim_ctx_binds_not_more in Hin; eauto.
    apply Hsd in Hin. simpl in Hin.
    repeat(rewrite in_union in *). destruct_all; try tauto.

  - hnf. introv Hin. unfold fv, CtxFreeVar in Hin.
    repeat(rewrite ctx_fv_push in Hin). repeat(rewrite in_union in *).
    destruct Hin as [Hin | Hin].
    1: eapply ctx_binds_fv_subset in HbiG.
    1: eapply typ_ro_fv in Htyr.
    1: apply Htyr, HbiG in Hin.
    all: eapply ref_elim_ctx_fv_subset in Hin; eauto.
    all: eapply ref_elim_ctx_fv_subset in Hre; eauto.
    all: repeat(rewrite in_union in Hin).
    all: destruct Hin as [Hin | Hin].
    all: try solve[apply Hnf in Hin; simpl in Hin;
                   repeat(rewrite in_union in Hin); eauto].
    all: rewrite in_singleton in Hin; subst.
    all: simpl in *.
    all: destruct (classicT (v_loc = v_ref)) as [? | _]; try discriminate.
    all: destruct (classicT (v_ref = v_ref)) as [_ | ?]; try tauto.

  - hnf. simpl. eapply subset_empty_l.

Qed.

End W1W2Same.

(* Package both cases of w₁ = w₂ and w₁ ≠ w₂ of step 5 *)
Lemma sef_ro_ty_config y1 y2 :
  binds w1 y1 (config_env c1) ->
  binds w2 y2 (config_env c1) ->
  exists Gro U Tw1 Tw2,
    ((ty_config Gro
        ((config_trm c1) ; stack_nil
                         ; (subst_heap w2 y2
                              (subst_heap w1 y1
                                 (config_heap c1)))
                         ; (config_env c1)
                         ; (fv_config c1))
        U)
     /\ (ctx_binds w1 (typ_and (typ_bnd Tw1)
                        (typ_mut_marker typ_top))
          Gro)
     /\ (ctx_binds w2 (typ_and (typ_bnd Tw2)
                        (typ_mut_marker typ_top))
          Gro)).
Proof.
  introv Hbie1 Hbie2.
  epose proof sef_eq_ro_ty_config as Hrotyeq.
  epose proof sef_neq_ro_ty_config as Hrotyne.

  assert (HinG : inert G) by eauto using config_ctx_inert.
  assert (HcorGe : renv_corr G (config_env c1)) by
    eauto using config_ctx_inert.

  assert (Hnew1y1 : w1 <> y1) by (eauto using renv_corr_diff).
  assert (Hnew2y2 : w2 <> y2) by (eauto using renv_corr_diff).
  assert (Hnew1y2 : w1 <> y2) by (eauto using renv_corr_diff).
  assert (Hnew2y1 : w2 <> y1) by (eauto using renv_corr_diff).

  eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie1 as Hbiy1;
    eauto.
  destruct Hbiy1 as [TL1 [TR1 [G2 [G3 [Hbkw1 [Hbky1 Htycor1]]]]]].
  eapply ctx_binds_k_after_trans in Hbkw1 as Hbiy1; eauto.
  eapply inert_binds_k_inert in Hbiy1 as HinTL1; eauto.

  eapply renv_corr_ref_elim_ctx_exists with (G:=G) in Hbie1 as HeG1; eauto.
  destruct HeG1 as [Gw1 HeG1].
  eapply ref_elim_ctx_inert with (vkw:=v_ref) (vky:=v_loc) (G:=G) in HeG1
      as HinGw1; eauto.
  clears TR1.

  eapply ctx_binds_k_after_binds, ref_elim_ctx_neq_binds in Hbiy1;
    eauto using config_ctx_inert, inert_fresh_telescope.
  eapply inert_type_subst with (x:=w1) (y:=y1) in HinTL1.

  destruct (classicT (w1 = w2)) as [Hsame | Hdiff].

  - clear Hrotyne. subst. eapply (binds_functional Hbie1) in Hbie2; subst.
    rename w2 into w, y2 into y, Hbie1 into Hbie, Gw1 into Gw,
      HeG1 into HeG, Hbiy1 into Hbiy, TL1 into TL, HinTL1 into HinTL.
    eapply inert_typ_loc_ref, typ_ro_exists_bnd in HinTL
        as [Twr Hro].
    specialize (Hrotyeq eq_refl _ _ _ _ Hbie Hbiy HeG Hro)
      as [U Hrotyeq].
    exists. rewrite subst_heap_twice_same. repeat(split); eauto.
    eauto.

  - clear Hrotyeq.

    eapply renv_corr_binds_to_binds_after with (G1:=G) in Hbie2 as Hbiy2;
      eauto.
    destruct Hbiy2 as [TL2 [TR2 [G4 [G5 [Hbkw2 [Hbky2 Htycor2]]]]]].
    eapply ctx_binds_k_after_trans in Hbkw2 as Hbiy2; eauto.
    eapply inert_binds_k_inert in Hbiy2 as HinTL2; eauto.
    eapply ctx_binds_k_after_binds, ref_elim_ctx_neq_binds in Hbiy2;
      eauto using config_ctx_inert, inert_fresh_telescope.
    eapply inert_type_subst with (x:=w1) (y:=y1) in HinTL2.
    clears TR2.

    eapply env_elim_binds_exists in Hbie1 as Hee1;
      eauto using config_env_ok.
    destruct Hee1 as [ew1 [Hee1 Hokew1]].
    eapply env_elim_correct in Hee1 as Hbiy2ew1.
    3: exact Hbie2. 2: eauto.

    eapply ref_elim_renv_corr in HcorGe as HcorGe1;
      eauto using inert_fresh_telescope.
    2: { introv Hin. apply fv_env_to_binds in Hin as [w Hc].
         eapply renv_corr_diff in Hc.
         4: exact Hbie1. 3: eassumption.
         all: eauto using inert_ok. }
    eapply renv_corr_binds_to_binds_after with (G1:=Gw1) in Hbiy2ew1 as
        Hbiy1ew1; eauto.
    destruct Hbiy1ew1 as [TL3 [TR3 [G6 [G7 [Hbkw1e [Hbky1e Htycor1e]]]]]].

    eapply renv_corr_ref_elim_ctx_exists with (G:=Gw1) in
      Hbiy2ew1 as HeG2; eauto.
    clears TL3 TR3 G6 G7.
    destruct HeG2 as [Gw2 HeG2].

    eapply ref_elim_ctx_neq_binds in Hbiy1;
      eauto using inert_fresh_telescope.
    eapply ref_elim_ctx_neq_binds in Hbiy2;
      eauto using inert_fresh_telescope.

    eapply inert_type_subst with (x:=w2) (y:=y2) in HinTL2.
    eapply inert_type_subst with (x:=w2) (y:=y2) in HinTL1.

    eapply inert_typ_loc_ref, typ_ro_exists_bnd in HinTL1
        as [Twr1 Hro1].
    eapply inert_typ_loc_ref, typ_ro_exists_bnd in HinTL2
        as [Twr2 Hro2].

    specialize (Hrotyne Hdiff _ _ _ _ _ _ _ _
                  Hbie1 Hbie2 Hbiy1 Hbiy2 HeG1 HeG2 Hro1 Hro2)
      as [U Hrotyeq].

    exists. repeat(split); eauto. all: eauto.

Qed.

Section Step6.

Definition c2 : config :=
  ((config_trm c1) ; stack_nil
     ; (config_heap c1)
     ; (config_env c1)
     ; (fv_config c1)).
Lemma c2_def : c2 = ((config_trm c1) ; stack_nil
                       ; (config_heap c1)
                       ; (config_env c1)
                       ; (fv_config c1)).
Proof. easy. Qed.

Variable (y1 y2 : var).
Hypothesis
  (Hbiwy1 : binds w1 y1 (config_env c1))
  (Hbiwy2 : binds w2 y2 (config_env c1)).

Definition c1' : config :=
  ((config_trm c1) ; stack_nil
     ; (subst_heap w2 y2 (subst_heap w1 y1 (config_heap c1)))
     ; (config_env c1)
     ; (fv_config c1)).

Lemma heap_dom_c1' : dom (config_heap c1') = dom (config_heap c1).
Proof. simpl. unfold subst_heap. rewrite ! dom_map. easy. Qed.

Lemma c1'_ty_roctx :
  exists Gro U Tw1 Tw2,
    ((ty_config Gro c1' U)
     /\ (ctx_binds w1 (typ_and (typ_bnd Tw1)
                        (typ_mut_marker typ_top))
          Gro)
     /\ (ctx_binds w2 (typ_and (typ_bnd Tw2)
                        (typ_mut_marker typ_top))
          Gro)).
Proof. eauto using sef_ro_ty_config. Qed.

(* c₁' has the SEF property *)
Lemma c1'_sef n c1'' :
  c1' ⊣ n ↦* c1'' ->
  subheap_unchanged (config_heap c1') (config_heap c1'').
Proof.
  destruct c1''.
  destruct c1'_ty_roctx as [Gro [U [Tw1 [Tw2 [Htyro [Hbirow1 Hbirow2]]]]]].
  introv Hred. eapply binds_ro_sef; eauto.
  introv [Htf | Htf]. 2: inversions Htf.
  rewrite Htrmc1 in Htf. inversions Htf; eauto.
Qed.

(* c₂ is typed because it is just c₁ with the stack removed *)
Lemma ty_c2 : exists T2, ty_config G c2 T2.
Proof.
  rewrite c2_def. exists typ_top. destructc Htycfg as T1 Htc. inversions Htc.
  simpl. destruct H5 as [[? ?] ?].
  econstructor; eauto; try solve[constructor; eauto]. repeat(split).
  all: simpl; try rewrite union_empty_l.
  - eapply subset_trans. eassumption. introv.
    repeat(rewrite in_union). intros. tauto.
  - eapply subset_trans. eassumption. introv.
    repeat(rewrite in_union). intros. tauto.
  - eapply subset_empty_l.
  - destructc H6 as null Hhn. exists null. repeat(rewrite in_union); tauto.
  - inversions H8. constructor; eauto using stack_closed_nil.
Qed.

Lemma heap_ref_elim_sim Ltid :
  fv_config c2 = from_list Ltid ->
  (similar_conf_sef
     c2                      c1'
     (w2 :: w1 :: Ltid)%list  (y2 :: y1 :: Ltid)%list
     (keys (config_heap c1)) (keys (config_heap c1))).
Proof.
  eapply ref_elim_similar_conf_sef; eauto.
  destructc config_heap_corr as HokG HwkG HokS HokE Hloc Href Hrc _.
  constructor; eauto using config_env_ok.
  - rewrite <- Hloc, <- Href. eapply ctx_kdom_disjoint; eauto. discriminate.
  - rewrite <- Hloc. intros y Hin. eapply fv_env_to_binds in Hin; eauto.
    destructc Hin as w Hbi. eapply renv_corr_binds_loc in Hbi;
      eauto using ctx_k_in_kdom.
Qed.

Section RefElimRedHeap.

Variable (Ltid : list var).
Hypothesis (HLtid : fv_config c2 = from_list Ltid).

Lemma ref_elim_sim_red_append_heap c2' c1'' k :
  c2 ⊣k↦* c2' -> c1' ⊣k↦* c1'' ->
  exists LS1' LS2',
    (similar_with
       (LS1' ++ w2 :: w1 :: Ltid)%list   (LS2' ++ y2 :: y1 :: Ltid)%list
       (LS1' ++ keys (config_heap c1)) (LS2' ++ keys (config_heap c1))
       (config_heap c2')               (config_heap c1''))
    /\ (disjoint (from_list LS1') (fv_heap (config_heap c2)))
    /\ (disjoint (from_list LS2') (fv_heap (config_heap c1'))).
Proof.
  introv Hred2 Hred1.
  destructc c1'_ty_roctx as G1 U1 _ _ Hty1 _.
  destructc ty_c2 as U2 Hty2.
  forwards: conf_sim_red_append.
  3: eapply heap_ref_elim_sim; exact HLtid. all: eauto.
  destructc H as LS1' LS2' Hsim Hd1 Hd2. exists LS1' LS2'.
  eapply similar_conf_sef_heap_part in Hsim. splits; eauto; clear Hsim.
  1: destruct c2 as [t1 fs1 Sigma1 e1 Lv1] eqn:Heq1; rename Hd1 into Hd.
  2: destruct c1' as [t2 fs2 Sigma2 e2 Lv2] eqn:Heq2; rename Hd2 into Hd.
  all: repeat (eapply disjoint_union_r_inv in Hd as [? Hd]); tauto.
Qed.

(* If c₂ ↦* c₂' and c₁' ↦ c₁'' then the part of the heap we care about
   can only change up to identifying wᵢ with yᵢ. This is the first
   part of the imm_transport lemma. *)
Lemma ref_elim_sim_red_sim_heap c2' c1'' k :
  c2 ⊣k↦* c2' -> c1' ⊣k↦* c1'' ->
  similar_with
    (w2 :: w1 :: Ltid)%list   (y2 :: y1 :: Ltid)%list
    (keys (config_heap c1)) (keys (config_heap c1))
    (config_heap c2')       (config_heap c1'').
Proof.
  introv Hred2 Hred1.
  forwards: ref_elim_sim_red_append_heap; fweauto.
  destructc H as LS1' LS2' Hsim Hd1 Hd2.

  eapply similar_env_with_i_append_invr in Hsim.
  2: { eapply similar_env_with_equiv in Hsim.
       destructc Hsim as Hlen _. rewrite ! List.app_length in Hlen.
       simpl in Hlen. omega. }

  eapply similar_env_with_o_append_invr_disjoint_fv_r in Hsim; eauto.
  - eapply similar_env_with_equiv in Hsim.
    destructc Hsim as Hlen _. rewrite ! List.app_length in Hlen.
    simpl in Hlen. omega.
  - introv Hbi1' Hink1. rewrite env_from_list_keys_dom, <- heap_dom_c1'
      in Hink1.
    eapply dom_to_binds in Hink1. destructc Hink1 as it1 Hb1.
    eapply c1'_sef in Hred1; eauto.
    eapply (binds_functional Hred1) in Hbi1'; subst. clear Hred1.
    repeat (eapply disjoint_union_r_inv in Hd2 as [_ Hd2]).
    eapply fset_extens; eauto using subset_empty_l; introv Hin;
      rewrite in_inter in Hin; rewrite <- Hd2, in_inter; splits; try tauto.
    eapply env_binds_fv_subset in Hb1. firstorder.
Qed.

Lemma almost_sef_c2 c2' k y it it' :
  c2 ⊣k↦* c2' ->
  binds y it  (config_heap c2 )  ->
  binds y it' (config_heap c2') ->
  similar
    (w2 :: w1 :: Ltid)%list (y2 :: y1 :: Ltid)%list
    it'                   ((it /[ w1 -> y1]) /[ w2 -> y2]) /\
    disjoint (\{w1} \u \{w2}) (fields it') /\
    disjoint (\{w1} \u \{w2}) (fields it ).
Proof.
  introv Hred2 Hbi2 Hbi2'.
  destructc c1'_ty_roctx as Gro U1 _ _ Hty1 _.
  destructc ty_c2 as U2 Hty2.
  eapply conf_sim_red_create in Hred2 as Hred1.
  4: { eapply heap_ref_elim_sim; eauto. }
  all: eauto.
  destructc Hred1 as c1'' Hred1.
  forwards Hsim:ref_elim_sim_red_sim_heap; fweauto.
  eapply similar_env_with_equiv in Hsim. destructc Hsim as _ _ _ Hsim.

  assert (Hw1edc2' : w1 \in dom (config_env c2')).
  { destruct c2'. simpl. eapply red_closure_mono_e in Hred2.
    destructc Hred2 as e' Heq; subst.
    rewrite dom_concat. eapply binds_to_dom in Hbiwy1.
    rewrite in_union. tauto. }
  assert (Hw2edc2' : w2 \in dom (config_env c2')).
  { destruct c2'. simpl. eapply red_closure_mono_e in Hred2.
    destructc Hred2 as e' Heq; subst.
    rewrite dom_concat. eapply binds_to_dom in Hbiwy2.
    rewrite in_union. tauto. }

  splits.
  2: { eapply Safety.preservation_closure in Hred2 as Hty2'; eauto.
       inversions Hty2'. eapply ty_config_fields_not_refs; eauto.
       introv Hin. rewrite ! in_union, ! in_singleton in Hin.
       destruct Hin as [? | ?]; subst; eauto. }
  2: { eapply ty_config_fields_not_refs; eauto. simpl.
       eapply binds_to_dom in Hbiwy1. eapply binds_to_dom in Hbiwy2.
       introv Hin. rewrite ! in_union, ! in_singleton in Hin.
       destruct Hin as [? | ?]; subst; eauto. }

  assert (Hrely : simlist_relates
                    (keys (config_heap c1))
                    (keys (config_heap c1)) y y).
  { simpl in Hbi2. eapply binds_to_dom in Hbi2.
       rewrite <- env_from_list_keys_dom in Hbi2.
       eapply list_in_fromlist_mem, LibList.mem_Nth in Hbi2.
       destructc Hbi2 as n Hbi2. exists n. split; exact Hbi2. }

  assert (Hbi1'' : exists it1'', binds y it1'' (config_heap c1'')).
  { eapply red_closure_mono_heap_dom in Hred1.
    simpl in Hbi2. simpl in Hred1. unfold subst_heap in Hred1.
    rewrite ! dom_map in Hred1.
    eapply binds_to_dom, Hred1, dom_to_binds in Hbi2.
    destructc Hbi2 as ? ?. eauto. }
  destructc Hbi1'' as it1'' Hbi1''.

  forwards Hsimit:Hsim; try eassumption.

  simpl in Hbi2.
  eapply c1'_sef in Hred1.
  2: { simpl. unfold subst_heap. eapply binds_map, binds_map in Hbi2.
       exact Hbi2. }
  binds_eq. clear H.

  forwards:Hsim; fweauto. easy.
Qed.

Lemma def_sef_ref_elim_same d d' :
  def_flds_sim d d' ->
  similar
    (w2 :: w1 :: Ltid)%list (y2 :: y1 :: Ltid)%list
    d'                   ((d /[ w1 -> y1]) /[ w2 -> y2]) ->
  disjoint (\{w1} \u \{w2}) (fields d') ->
  disjoint (\{w1} \u \{w2}) (fields d ) ->
  d = d'.
Proof.
  introv Hdfs Hsim Hd' Hd. inversions Hdfs; eauto.
  simpl in Hsim; simpl in Hsim. eapply similar_def_fld_var in Hsim.
  unfold fields,FieldsDef,fields_def in Hd,Hd'.
  destruct y; destruct x.
  all: try solve[similar_syntax_invert; inversions Hsim; easy].

  simpl in Hsim.
  destruct (classicT (v0=w1)) as [Heq | Hne1].
  { subst. false. erewrite <- in_empty, <- Hd, in_inter, in_union,
      ! in_singleton. eauto. }
  destruct (classicT (v0=w2)) as [Heq | Hne2].
  { subst. false. erewrite <- in_empty, <- Hd, in_inter, in_union,
      ! in_singleton. eauto. }

  eapply similar_avar_f_var, similar_var_simlist_relates,
    simlist_relates_cons_cases in Hsim as [Hsim | [Hc1 Hc2]].
  2: { subst. false. erewrite <- in_empty, <- Hd', in_inter, in_union,
      ! in_singleton. eauto. }
  eapply simlist_relates_cons_cases in Hsim as [Hsim | [Hc1 Hc2]].
  2: { subst. false. erewrite <- in_empty, <- Hd', in_inter, in_union,
      ! in_singleton. eauto. }

  destructc Hsim as n Hn1 Hn2.
  eapply (LibList.Nth_functional Hn1) in Hn2; subst. easy.
Qed.

Lemma defs_sef_ref_elim_same ds ds' :
  defs_flds_sim ds ds' ->
  similar
    (w2 :: w1 :: Ltid)%list (y2 :: y1 :: Ltid)%list
    ds'                   ((ds /[ w1 -> y1]) /[ w2 -> y2]) ->
  disjoint (\{w1} \u \{w2}) (fields ds') ->
  disjoint (\{w1} \u \{w2}) (fields ds ) ->
  ds = ds'.
Proof.
  induction 1. intros; eauto.
  introv Hsim Hd' Hd. simpl in Hsim.
  eapply similar_defs_cons_defs in Hsim as Hsimdefs.
  eapply similar_defs_cons_def in Hsim as Hsimdef.

  eapply disjoint_union_r_inv in Hd as [Hds Hdd].
  eapply disjoint_union_r_inv in Hd' as [Hds' Hdd'].

  rewrite def_sef_ref_elim_same with (d:=d) (d':=d'); eauto.
  rewrite IHdefs_flds_sim; eauto.
Qed.

Lemma item_sef_ref_elim_same it it' :
  heap_item_flds_sim it it' ->
  similar
    (w2 :: w1 :: Ltid)%list (y2 :: y1 :: Ltid)%list
    it'                   ((it /[ w1 -> y1]) /[ w2 -> y2]) ->
  disjoint (\{w1} \u \{w2}) (fields it') ->
  disjoint (\{w1} \u \{w2}) (fields it ) ->
  it = it'.
Proof.
  introv Hfsim Hsim Hd' Hd. inversions Hfsim; eauto.
  rewrite defs_sef_ref_elim_same with (ds:=ds) (ds':=ds'); eauto.
  simpl in Hsim. eapply similar_item_obj_defs in Hsim; eauto.
Qed.

End RefElimRedHeap.

(* Lemma 17 in "Pure methods for roDOT" *)
Lemma imm_transport c3 k :
  c2 ⊣k↦* c3 -> subheap_unchanged (config_heap c1) (config_heap c3).
Proof.
  introv Hred Hbi1. destructc (fset_finite (fv_config c2)) as Ltid HLtid.

  eapply red_closure_mono_heap_dom in Hred as Hheapred. simpl in Hheapred.
  eapply binds_to_dom in Hbi1 as Hdom1.
  eapply Hheapred, dom_to_binds in Hdom1 as [itm3 Hbi3].

  forwards Hasef:almost_sef_c2; fweauto.
  destructc Hasef as Hsim Hd3 Hd1.
  forwards Hiteq:item_sef_ref_elim_same; fweauto.
  eapply red_heap_flds_sim; eauto. subst. eauto.
Qed.

End Step6.

(* Theorem 16 in "Pure methods for roDOT" *)
Theorem SefG_I c4 k :
  c1 ⊣k↦* c4 ->
  (exists j c3, frame_end_in c1 j c3 /\ j < k) \/
    (subheap_unchanged (config_heap c1) (config_heap c4)).
Proof.
  introv Hreds. forwards Hc:classic. destruct Hc as [Ht | Hf].
  exact Ht.

  right. eapply Classical_Prop.not_or_and in Hf.
  destructc Hf as Hnend _.
  pose proof imm_transport as Htrans.
  unfold c2 in Htrans. destruct c1. simpl in *. destruct c4.
  eapply frame_not_end_stays_within in Hnend.
  forwards Hsuf:Hnend; eauto. simpl in Hsuf.
  destructc Hsuf as fs ?; subst.
  eapply frame_within_truncate_stack in Hreds as Hredtrunc; eauto.
  destructc Hredtrunc as Lv1 Hredtrunc _.

  introv Hbiy.

  destructc Htycfg as T1 Htycfg0. inversions Htycfg0.
  clear H16 H15 H14 H13 H11 H8 H7 H6.
  rename H4 into HinG, H5 into Hhc, H9 into Htf.
  hnf in Htf.
  forwards Href1:Htf. 2: forwards Href2:Htf.
  2: eapply tfree_trm_call_r. 1: eapply tfree_trm_call_l.
  destructc Hhc as _ _ _ _ _ Henvref _ _.
  eapply ctx_k_in_kdom in Href1; eapply ctx_k_in_kdom in Href2.
  rewrite Henvref in *.
  eapply dom_to_binds in Href1; eapply dom_to_binds in Href2.
  destructc Href1 as y1 Hbiwy1; destructc Href2 as y2 Hbiwy2.
  eapply Htrans in Hredtrunc; eauto.
Qed.

End SefG_I.

End WithSefMode.
