(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Weakening EWeakening Substitution TFree.

Require Import OperationalSemantics TypedOperationalSemantics.
Require Import ConfigTyping TypedProgress.
Require Import TypedPreservation.

Require Import FreshChain FreshChainTactics.
Require Import TypedOperationalSemantics InertContext.
Require Import Similarity SyntaxSimilarity.
Require Import ExecutionTactics Renaming TypingHelpers ConfigTypingAccessories.

Require Import NothingFresh TypedPreservationLemmas LetLet Defs.
Require Import HeapCorrespondence ConfigVarkinds.
Require Import HeapUpdate ConfigValid.
Require Import EnvCorrespondence.

Section WithTypingMode.
Context (typing_mode: Mode).

(** * Lemmas about configuration typing premises  reduction steps *)

(* Partial configuration typing in a Read step *)
Lemma ty_config_accessories_tred_var G e t fs S w T Lv:
  ty_config_accessories G e (stack_cons (frame_let t) fs) S (trm_var (avar_f w)) T Lv ->
  ty_config_accessories G e fs S t /[-> w] T (fv (trm_var (avar_f w)) \u fv (stack_cons (frame_let t) fs) \u fv_heap S \u fv_renv e \u Lv).
Proof.
  introv Htca.
  unfold ty_config_accessories in *.
  destructecs Htca.
  eapply all_tfree_k_stack_push_inv in Htca1. destructecs Htca1.
  splits; eauto.

  - eapply ty_nothing_fresh_in_config_p; eauto.
  - eapply null_p; eauto.
  - eapply closed_let_loc; eauto.
Qed.

(* Partial configuration typing in a Let-push step *)
Lemma ty_config_accessories_tred_push G e t1 t2 fs S T Lv:
  ty_config_accessories G e fs S (trm_let t1 t2) T Lv ->
  ty_config_accessories G e (stack_cons (frame_let t2) fs) S t1 T (fv (trm_let t1 t2) \u fv fs \u fv_heap S \u fv_renv e \u Lv).
Proof.
  introv Htca.
  unfold ty_config_accessories in *.
  destructecs Htca.
  splits; eauto.
  - eapply all_tfree_k_let_push; eauto.
  - hnf in Htca2. hnf. eauto.
  - eapply ty_nothing_fresh_in_config_p; eauto.
  - eapply null_p; eauto.
  - eapply closed_let_push; eauto.
Qed.


(* Partial configuration typing in a Apply step *)
Lemma ty_config_accessories_tred_apply G e t fs S x z y T Lv:
  ty_item_s typing_mode G e S y ->
  binds y (item_fun t) S ->
  ty_config_accessories G e fs S (trm_apply (avar_f x) (avar_f z)) T Lv ->
  ty_config_accessories G e fs S t /[-> z] T (fv (trm_apply (avar_f x) (avar_f z)) \u fv fs \u fv_heap S \u fv_renv e \u Lv).
Proof.
  introv Htyit Hbi Htca.
  unfold ty_config_accessories in *.
  destructecs Htca.

  splits; eauto.
  - inversions Htyit. binds_eq. eauto. binds_eq.
  - eapply ty_nothing_fresh_in_config_p; eauto.
  - eapply null_p; eauto.
  - eapply closed_apply; eauto.
Qed.
Lemma ty_config_accessories_tred_apply' G e t fs S x z y T Lv:
  heap_correspond G e S ->
  binds y (item_fun t) S ->
  ty_config_accessories G e fs S (trm_apply (avar_f x) (avar_f z)) T Lv ->
  ty_config_accessories G e fs S t /[-> z] T (fv (trm_apply (avar_f x) (avar_f z)) \u fv fs \u fv_heap S \u fv_renv e \u Lv).
Proof.
  introv Hhc Hbi Htca.

  eapply ty_config_accessories_tred_apply; eauto.
  hnf in Hhc. destructecs Hhc. eapply Hhc. eapply binds_to_dom; eauto.
Qed.


(* Partial configuration typing in a Call step *)
Lemma ty_config_accessories_tred_call G e t ds fs S x z y T Lv m T0:
  ty_config_accessories G e fs S (trm_call (avar_f x) m (avar_f z)) T Lv ->
  binds y (item_obj ds) S ->
  defs_has ds (def_met m t) ->
  G ⁏ e ⊢ trm_call (avar_f x) m (avar_f z) ∶ T0 ->
  binds x y e ->
  heap_correspond G e S ->
  ok e ->
  ty_config_accessories G e fs S (open_met z x t) T (fv (trm_call (avar_f x) m (avar_f z)) \u fv fs \u fv_heap S \u fv_renv e \u Lv).
Proof.
  introv Htca Hbi Hhas Htyt Hbie Hhc Hoke.
  unfold ty_config_accessories in *.
  destructecs Htca.

  assert (  ty_item_s typing_mode G e S y ) as Htyit. hnf in Hhc. destructecs Hhc. eapply Hhc. eapply binds_to_dom; eauto.

  splits; eauto.
  - inversions Htyit. binds_eq. binds_eq.
    eapply all_tfree_k_call with (e:=e); eauto.
  - eapply ty_nothing_fresh_in_config_p; eauto.
  - eapply null_p; eauto.
  - eapply closed_call; eauto.
Qed.

(* Partial configuration typing in a Write step *)
Lemma ty_config_accessories_tred_write G e w2 w0 fs Sa Sigma3 a T Lva y1 y3:
  ty_config_accessories G e fs Sa (trm_write (avar_f w0) a (avar_f w2)) T Lva ->
  heap_update Sa y1 a y3 Sigma3 ->
  ty_config_accessories G e fs Sigma3 (trm_var (avar_f w2)) T (fv (trm_write (avar_f w0) a (avar_f w2)) \u fv fs \u fv_heap Sa \u fv_renv e \u Lva).
Proof.
  introv Htca Hu. (* Htyit Hbi Hhas Htyt Hbie Hhc Hoke. *)
  unfold ty_config_accessories in *.
  destructecs Htca.

  splits; eauto.
  - eapply ty_nothing_fresh_in_config_p; eauto.
  - eapply null_p; eauto.
  - inversions Htca6. constructor; eauto. intro. intro. simpl. auto.
    eapply Closedness.heap_closed_update; eauto.
Qed.


Lemma ty_config_accessories_tred_let_obj G e t R ds fs S y w T Lv (Td: typ) :
  ty_config_accessories G e fs S (trm_lit (lit_obj R ds) t) T Lv ->
  G ⁏ e ⊢ lit_obj R ds ∶ Td ->
  fresh (ctx_dom G \u ctx_fv G \u fv R) 2 (y :: w :: nil) ->
  ty_config_accessories (inert_G_for_obj_loc typing_mode G w y R) (e & w ~ y) fs (S & y ~ item_obj (Deref.deref_defs (e & w ~ y) ds /[-> w])) t /[-> w] T (fv (trm_lit (lit_obj R ds) t) \u fv fs \u fv_heap S \u fv_renv e \u Lv).
Proof.
  introv Htca Htyd Hfr.
  unfold ty_config_accessories in *.
  destructecs Htca.
  splits.
  *
    eapply inert_G_for_obj_loc_inert; eauto.
    eapply typ_objref_loc_inert; eauto.
    eapply typ_objref_ref_inert; eauto.
  *
    eauto.
  *
    eapply all_tfree_k_let_obj; eauto.
  * 
    eapply tred_ctx_dom_lit_obj; eauto.
  *
    eapply null_p; eauto.
  * eapply tred_sealed_lit_obj; eauto.
  * eapply closed_let_lit_obj; eauto.
  *
    eauto.
Qed.


Lemma ty_config_accessories_tred_let_fun G e t R ds fs S y w T Lv (Td: typ) UN Lva2:
  mode_has_fun' typing_mode = true ->
  ty_config G (trm_lit (lit_fun R ds) t ; fs ; S ; e ; Lv) T ->
  ty_config_accessories G e fs S (trm_lit (lit_fun R ds) t) T Lv ->
  G ⁏ e ⊢ lit_fun R ds ∶ Td ->
  fresh (ctx_dom G \u ctx_fv G \u fv R \u fv UN) 2 (y :: w :: nil) ->
  ((fv R \u fv ds) \u fv t) \u fv fs \u fv_heap S \u fv_renv e \u Lv \u fv UN  \c Lva2 ->
  closed  (typ_fun R UN) ->
  ty_config_accessories (inert_G_for_item G w y (typ_fun R UN)) (e & w ~ y) fs (S & y ~ item_fun ds) t /[-> w] T Lva2.
Proof.
  introv Hm.
  introv Htyc Htca Htyd Hfr Hsu HclU.
  unfold ty_config_accessories in *.
  destructecs Htca.
  splits.
    + eapply inert_G_for_item_fun_inert; eauto. sympl. fresh_solve. auto.
    + apply all_tfree_k_Gpush. apply all_tfree_k_Gpush; eauto. eauto.
    + eapply tred_lit_fun_tfree_k_trm; eauto.
    + 

      eapply ty_nothing_fresh_in_config_push2_Lv_T. eapply Htca3. sympl.
      eapply subset_trans. 2: eapply Hsu. union_sub_solver.
      sympl. 
      eapply subset_trans. 2: eapply Hsu. union_sub_solver.
    +   destructc Htca4 as null ?. exists null. eapply Hsu. union_in_solver.
    + eapply tred_lit_fun_sealed; eauto.
    + eapply config_closed_Lv. eapply tred_lit_fun_closed; eauto.
    + eauto.

Qed.


Lemma heap_correspond_accessories_tred_let_obj G S e w1 y1 d R0:
  heap_correspond_accessories G e S ->
  fresh (ctx_dom G \u ctx_fv G \u fv R0 \u dom S \u dom e) 2 (y1 :: w1 :: nil) ->
  heap_correspond_accessories
  (inert_G_for_item G w1 y1
     match mode_refcaps' typing_mode with
     | mode_refcaps_none => typ_and (typ_bnd R0) typ_top
     | mode_refcaps_mutability => typ_and (typ_bnd R0) (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))
     end) (e & w1 ~ y1) (S & y1 ~ item_obj (Deref.deref_defs (e & w1 ~ y1) d /[-> w1])).
Proof.
  introv Hhc Hfr.
  hnf in Hhc; destructecs Hhc.

  hnf; splits.
  - eauto.
  - eauto.
  -  eauto.
  -  eauto.
  - sympl. cases_if. cases_if. simpl_dom. rewrite Hhc4. auto.
  - sympl. cases_if. cases_if. simpl_dom. rewrite Hhc5. auto.
  -
  destruct (mode_refcaps' typing_mode) eqn: Hm;
  eapply renv_corr_push_obj; eauto.
  forwards: typ_objref_loc_corr. rewrite Hm in H. eauto.
  forwards: typ_objref_loc_corr. rewrite Hm in H. eauto.
Qed.


Lemma heap_correspond_accessories_tred_let_fun G S e w1 y1 t2 T0 UN:
  mode_has_fun' typing_mode = true ->
  heap_correspond_accessories G e S ->
  fresh (ctx_dom G \u ctx_fv G \u fv T0 \u fv UN \u dom S \u dom e) 2 (y1 :: w1 :: nil) ->
  heap_correspond_accessories (inert_G_for_item G w1 y1 (typ_fun T0 UN)) (e & w1 ~ y1) (S & y1 ~ item_fun t2).
Proof.
  introv Hm Hhc Hfr.
  hnf in Hhc; destructecs Hhc.

  hnf; splits.
  - eauto.
  - eauto.
  -  eauto.
  -  eauto.
  - sympl. cases_if. cases_if. simpl_dom. rewrite Hhc4. auto.
  - sympl. cases_if. cases_if. simpl_dom. rewrite Hhc5. auto.
  -
  eapply renv_corr_push_fun; eauto. econstructor; auto.

Qed.

Lemma heap_correspond_accessories_update G S e y1 a y2 S2:
  heap_correspond_accessories G e S ->
  heap_update S y1 a y2 S2 ->
  heap_correspond_accessories G e S2.
Proof.
  introv Hhc Hu.
  hnf in *.
  destructecs_all.
  splits; auto 1.
  rewrite <- Hu2. auto.
Qed.

End WithTypingMode.
