
Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode.
Require Import GeneralTyping Equivalence Deref EWeakening
EnvCorrespondence ReferenceTypes  Substitution Indep InertContext.

Section WithTypingMode.
Context (typing_mode: Mode).

(** * Lemmas about instantiating objects on the heap *)

(* Partially dereferencing a variable preserves its type.
   This is a generalized version of ty_deref_var, which allows
   the renv for typing to be longer than the renv for dereferencing,
   for the purpose of induction on the latter. *)
Lemma ty_deref_var2 G e1 e2 (v: var) T:
   inert G ->
   ok (e1 & e2) ->
   renv_corr G (e1 & e2) ->
   renv_corr G (e1) ->
   typing G (e1 & e2) v T ->
   typing G (e1 & e2) (deref e1 v) T
.
Proof.
  introv Hok.
  revert v e2.
  induction e1.
  - rewrite <- empty_def.
    intros. rewrite concat_empty_l in *.
    rewrite empty_def. auto.
  - introv Hoke Hcor2 Hcor1 Hty.
    destruct a.
    rewrite cons_to_push in *.
    unfold deref, deref_var.
    rewrite deref_base_push.
    rewrite <- concat_assoc in *.
    assert (renv_corr G e1) by (eapply renv_corr_epush_inv; eauto).
    apply IHe1; auto.
    rewrite concat_assoc in *.
    (* eapply renv_corr_epush_inv; eauto.*)
    compute. destruct (classicT (v = v0)); subst.
    + inversion Hcor1; subst.
      * false. eapply empty_push_inv;eauto.
      * destruct (eq_push_inv H0) as [?[? ?]]; subst.
        eapply ty_renv_binds_var; eauto.
        apply binds_middle_eq. destruct (ok_middle_inv Hoke). auto.
  + auto.
Qed.

(* Dereferencing a variable preserves its type *)
Lemma ty_deref_var G e1 (v: var) T:
   inert G ->
   ok e1 ->
   renv_corr G e1 ->
   typing G e1 v T ->
   typing G e1 (deref e1 v) T
.
Proof.
  intros.
  replace e1 with (e1 & empty) in * at 0 1.
  apply ty_deref_var2; auto.
  rewrite concat_empty_r in *. auto.
  auto using concat_empty_r.
Qed.

(* Dereferencing a definition preserves its type *)
Lemma ty_deref_def G Go e1 d T:
   inert (G §§& Go) ->
   ok e1 ->
   renv_corr (G §§& Go) e1 ->
   obj_dec_typing G Go e1 d T ->
   obj_dec_typing G Go e1 (deref_def e1 d) T
.
Proof.
  introv Hin Hoke1 Hcor1 Hty.
  destruct d.
  - (* type - unchanged *)
    unfold deref_def. auto.
  - (* field - changed *)
    unfold deref_def.
    inversion Hty; subst.
    rewrite deref_avar_comm.
    apply ty_def_fld.
    intros. eapply inert_ty_vis; eauto.
    eapply ty_deref_var; eauto.
    apply ty_deref_var; auto.
  - (* type - unchanged *)
    unfold deref_def. auto.
Qed.

Lemma label_of_deref e d:
  label_of (deref_def e d) = label_of d.
Proof.
  unfold label_of. unfold DefLabelOf.
  unfold deref_def.
  destruct d.
  all: auto.
Qed.

Lemma defs_hasnt_deref e ds l:
  defs_hasnt ds l ->
  defs_hasnt (deref_defs e ds) l.
Proof.
  intros Hhasnt.
  unfold defs_hasnt in *.
  induction ds.
  - unfold deref_defs; auto.
  - rewrite deref_defs_cons_comm.
    unfold get_def in *.
    case_if.
    + subst.
      rewrite label_of_deref in *.
      case_if in Hhasnt.
    + rewrite label_of_deref in *.
      case_if in Hhasnt.
      auto.
Qed.

(* Lemma 4 in "Reference mutability for DOT" *)
Lemma ty_deref_defs G Go e1 ds T:
  inert (G §§& Go) ->
  ok e1 ->
  renv_corr (G §§& Go) e1 ->
  obj_typing G Go e1 ds T ->
  obj_typing G Go e1 (deref_defs e1 ds) T.
Proof.
  introv Hin Hoke1 Hcor1 Hty.
  induction Hty.
  - (* one *)
    rewrite deref_defs_cons_comm.
    apply ty_defs_one.
    apply ty_deref_def; auto.
  - (* cons *)
    rewrite deref_defs_cons_comm.
    apply ty_defs_cons.
    apply IHHty; auto.
    apply ty_deref_def; auto.
    apply defs_hasnt_deref.
    rewrite label_of_deref.
    auto.
Qed.

End WithTypingMode.
