Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax TypingMode GeneralTyping Equivalence
        InertTightSubtyping  Indep
        GeneralToTight TightTyping InvertibleTyping
        RecordTypes RecordDefs InertContext PreciseTyping
        OperationalSemantics Substitution UnboundSubstitution Weakening
        HeapCorrespondence HeapUpdate HeapValid CanonicalForms
         Subenvironments Narrowing
        EnvCorrespondence EWeakening  TypeApproximation Shortening RestrictedTightSubtyping
        ConfigTyping ConfigVarkinds TFree Deref
        TypedOperationalSemantics MutableReachability Defs TFreeTyping
        TypedPreservation TypedSoundness.

(** * Preservation of heap objects *)

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma heap_preservation t1 s1 S1 e1 Lv1 F1 T  t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ (t1 ; s1 ; S1 ; e1; Lv1) ∶ T ↦ F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    exists ds2, binds y (item_obj ds2) S2.
Proof.
  introv Hbi Hred.
  inversions Hred; eauto.
  - (* write *)
    inversions Hupd. destructc H0 as ? Hdoms ? ? ? ? Hrest. destructc Hrest as ? HbiS2.
    specialize (H1 y (item_obj ds)).
    destruct (classicT (y = y1)).
    + subst. exists x0. auto.
    + exists ds. auto.
  - (* lit_obj *)
    exists ds.
    list_split_from_in y0 Hfr.
    eapply fresh_in_config_notin_Sigma_dom in Hfr.
    eapply binds_push_neq. auto. eapply binds_to_dom in Hbi.
    intro; subst. contradiction.
  - (* lit_fun *)
    exists ds.
    list_split_from_in y0 Hfr.
    eapply fresh_in_config_notin_Sigma_dom in Hfr.
    eapply binds_push_neq. auto. eapply binds_to_dom in Hbi.
    intro; subst. contradiction.
Qed.

Lemma heap_preservation_closure t1 s1 S1 e1 Lv1 F1 T n t2 s2 S2 e2 Lv2 F2 y ds:
    binds y (item_obj ds) S1 ->
    F1 ⊢ (t1 ; s1 ; S1 ; e1; Lv1) ∶ T ⊣ n ↦* F2 ⊢ (t2 ; s2 ; S2 ; e2; Lv2) ->
    exists ds2, binds y (item_obj ds2) S2.
Proof.
  introv Hbi Hred.
  remember (t1 ; s1 ; S1 ; e1; Lv1) as C1.
  remember_dependent_config t2 s2 S2 e2 Lv2 as C2.

  induction Hred; intros; subst.
  - inversions HeqC2. exists ds. auto.
  - destruct C2.
    assert (exists ds2, binds y (item_obj ds2) S).
    eapply IHHred; eauto.
    destruct H0.
     eapply heap_preservation. apply H0. apply H.
Qed.

End WithTypingMode.