(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping EnvCorrespondence TFree ConfigVarkinds Defs RecordDefs ReferenceTypes.

(** * Updating the Heap *)
(* defs_update ds1 a x ds2 means that
   updating the value of field a in ds1 to x results in ds2 *)
Inductive defs_update : defs -> fld_label -> var -> defs -> Prop :=
| defs_update_var : forall ds a t x,
    defs_update (defs_cons ds (def_fld a t)) a x (defs_cons ds (def_fld a (avar_f x)))
| defs_update_push : forall ds d a x ds',
    defs_update ds a x ds' ->
    label_of d <> (label_fld a) ->
    defs_update (defs_cons ds d) a x (defs_cons ds' d).

(* heap_update s1 x a y s2 means that
   updating the value of field a in an object at location x in heap s1 to y
   results in heap s2 *)
Definition heap_update (s : heap) (x : var) (a : fld_label) (y : var) (s' : heap) :=
  ok s /\
  ok s' /\
  (dom s) = (dom s') /\
  (forall z v, z <> x ->
          binds z v s ->
          binds z v s') /\
  exists ds ds', binds x (item_obj ds) s /\
              defs_update ds a y ds' /\
              binds x (item_obj ds') s'.

(* Updating a field does not add any new labels *)
Lemma defs_update_hasnt: forall ds a x ds' l,
    defs_update ds a x ds' ->
    defs_hasnt ds l ->
    defs_hasnt ds' l.
Proof.
  introv H; gen l; induction H; intros.
  - unfold defs_hasnt; simpl.
    inversion H; cases_if.
    auto.
  - unfold defs_hasnt; simpl.
    rename H1 into Hhasnt.
    assert (label_of d <> l).
    { inversion Hhasnt. cases_if. auto. }
    cases_if.
    assert (defs_hasnt ds l).
    { inversion Hhasnt. cases_if. auto. }
    apply IHdefs_update; auto.
Qed.

Lemma defs_update_hasnt_inv: forall ds a x ds',
    defs_update ds a x ds' ->
    defs_hasnt ds (label_fld a) -> False.
Proof.
  introv H. induction H; intros.
  - inversion H; cases_if; congruence.
  - assert (defs_hasnt ds (label_fld a)).
    { inversion H1; cases_if. }
    auto.
Qed.

Lemma defs_update_functional ds f x ds1 ds2:
  defs_update ds f x ds1 ->
  defs_update ds f x ds2 ->
  ds1 = ds2.
Proof.
  introv Hup1 Hup2.
  reverts ds2.
  induction Hup1; intros; inversions Hup2.
  - auto.
  - simpls. contradiction.
  - simpls. contradiction. 
  - fequal. eapply IHHup1; eauto.
Qed.

Require Import RecordTypes InertContext PreciseTyping HeapCorrespondence CanonicalForms.

Section WithTypingMode.
Context (typing_mode: Mode).

(* Updating a field in defs to a variable of a correct type
   preserves the type of the defs. *)
Lemma defs_update_typing: forall G Go e (ds : defs) U (ds' : defs) a x T,
    defs_update ds a x ds' ->
    record_type U ->
    G §§ Go⁏e ⊢ ds ∶ U ->
    record_has U (dec_fld a T T) ->
    (mode_has_visibility typing_mode -> ctx_vis (G §§& Go) x) ->
    G §§& Go⁏e ⊢ x ∶ T ->
    G §§ Go⁏e ⊢ ds' ∶ U.
Proof.
  introv H; gen U T. induction H; intros.
  - (* var case - update happens here *)
    inversions H0.
    + (* ty_defs_one - last member *)
      inversions H10.
      inversions H1. eauto.
    + (* ty_defs_cons - some members after this one *)
      inversions H10.
      assert (T1 = T) by (eapply (unique_rcd_trm H); eauto); subst.
      eauto.
  - (* push case - update happens later *)
    inversions H2. inversions H.
    assert (record_type T0).
    { destruct H1 as [ls ?H]. inversions H1.
      exists ls0; auto. }
    inversion H3; subst T0 U D0.
    + constructor; eauto using defs_update_hasnt.
    + inversions H10. inversions H12; simpl in *.
      exfalso. eapply defs_update_hasnt_inv; eauto.
Qed.

(* Updating a field to a location variable preserves correct variable kinds
   in the definitions *)
Lemma defs_update_varkinds: forall G (ds : defs) (ds' : defs) a x,
    defs_update ds a x ds' ->
    ctx_k G x v_loc ->
    defs_varkinds G ds ->
    defs_varkinds G ds'.
Proof.
  introv Hu Hvkx Hvkds.
  induction Hu.
  - inversion Hvkx. inversion Hvkds; subst.
    auto.
  - inversion Hvkds. auto.
Qed.

Lemma heap_update_ty_item_s G e S1 S2 x vObjL ds0 a vValL vArgL ds' vObjR vArgR S T:
  inert G ->
  sealed G ->
  ok e ->
  renv_corr G e ->
  (forall x0 : var, x0 \in dom S1 -> ty_item_s typing_mode G e S1 x0) ->
  x \in dom S2 ->
  dom S1 = dom S2 ->
  ctx_kdom v_loc G = dom S1 ->
  binds vObjL (item_obj ds0) S1 ->
  defs_has ds0 (def_fld a (avar_f vValL)) ->
  defs_update ds0 a vArgL ds' ->
  binds vObjR vObjL e ->
  binds vArgR vArgL e ->
  binds vObjL (item_obj ds') S2 ->
  G ⁏ e ⊢ vArgR ∶ S ->
  G ⁏ e ⊢ vObjR ∶ typ_rcd (dec_fld a S T) ->
  ( forall (z : var) (v : item), z <> vObjL -> binds z v S1 -> binds z v S2) ->
  ty_item_s typing_mode G e S2 x.
Proof.
  introv Hin HclG Hoke Hcor   Htyit Hindom Hdom Hdoms.
  introv HbiObjL Hhas Hup HbiObjR HbiArgR HbiS2 HtyArgR HtyObjR HbiSS.

  pose proof (inert_ok Hin) as HokG.
  pose proof (inert_weakens Hin) as HwkG.
  rewrite <- Hdom in Hindom.
  set (HindomG := Hindom).
  rewrite <- Hdoms in HindomG.
  pose proof (ctx_kdom_to_binds _ HokG HwkG HindomG) as [?T HbixG].
  pose proof (Htyit _ Hindom) as Htyv.
  pose proof (dom_to_binds Hindom) as [?v HbixS1].

  destruct (var_decide x vObjL); subst.
  - (* x0 = x *) inversions Htyv.
    + (* λ-abstraction case is impossible because x is bound to an object *)
      exfalso. pose proof (binds_functional HbiObjL H1). congruence.
    + pose proof (binds_functional H0 HbiObjL) as Hiteq. inversions Hiteq.
      eapply ty_item_obj_s; eauto.
      assert ( G ⁏ e ⊢ vObjL ∶ typ_rcd (dec_fld a S T)) as HtyObjL.
      eapply ty_renv_binds_var; eauto.
      assert ( G ⁏ e ⊢ vArgL ∶ S) as HtyArgL.
      eapply ty_renv_binds_var; eauto.
      pose proof (var_typ_rcd_to_binds Hin HclG Hoke Hcor HtyObjL)
        as [?S [?S [?T [?Hbi [Hrchas [?Hty _]]]]]].
      pose proof (ctx_binds_functional Hbi H) as Hbif. inversions Hbif.
      eapply defs_update_typing; eauto using ty_defs_record_type.
      eapply defs_update_varkinds; eauto.

      destruct (renv_corr_binds_vk HokG Hcor HbiArgR). auto.
  - (* x0 ≠ x *)
    apply (HbiSS _ _ n) in HbixS1. rename n into Hneq.
    inversions Htyv; eauto.
Qed.

Lemma heap_update_ty_item_s_n G e S1 S2 x vObjL a vArgL :
  inert G ->
  ok e ->
  ty_item_s typing_mode G e S1 x -> 
  heap_update S1 vObjL a vArgL S2 -> 
  (forall (z : var) (v : item), z <> vObjL -> binds z v S1 -> binds z v S2) ->
  x <> vObjL ->
  ty_item_s typing_mode G e S2 x.
Proof.
  introv Hin Hoke Htyit Hup.
  introv HbiSS Hnx.

  assert (  x \in dom S1 ) as Hindom. {
    inversions Htyit; eapply binds_to_dom; eauto.
  }

  pose proof (dom_to_binds Hindom) as [?v HbixS1].
  apply (HbiSS _ _ Hnx) in HbixS1. 
  inversions Htyit; eauto.
Qed.

(* Updating a field in a heap corresponding to an inert context with variables
   coming from a correctly typed write term preserves the heap correspondence *)
Lemma heap_update_inert: forall G e s s' vObjR vObjL vArgR vArgL a T,
    inert G ->
    sealed G ->
    heap_correspond G e s ->
    ctx_k G vObjR v_ref ->
    ctx_k G vArgR v_ref ->
    G⁏e ⊢ trm_write (avar_f vObjR) a (avar_f vArgR) ∶ T ->
    binds vObjR vObjL e ->
    binds vArgR vArgL e ->
    heap_update s vObjL a vArgL s' ->
    heap_correspond G e s'.
Proof.
  introv Hin Hcl Hst HvkObj HvkArg Hty HbiObj HbiArg Hsto.
  remember (trm_write (avar_f vObjR) a (avar_f vArgR)) as t.
  induction Hty; inversions Heqt; eauto.
  rename H2 into HtyObjR.
  pose proof (canonical_forms_fld_ref Hin Hcl Hst HbiObj HtyObjR) as [?ds [?vValL [HbiObjL [?H ?H]]]].
  unfold heap_update in Hsto.
  destruct Hsto as
      [?Hoks [?Hoks' [?Hdom [?Hbi [?ds [?ds' [HbiObjL1 [?Hdefup ?Hbis']]]]]]]].
  pose proof (binds_functional HbiObjL HbiObjL1) as Hveq; inversions Hveq; clear H1.
  destruct Hst as [?H [?H [?H [Hoke [Hdoms [?H [Hcor Htyit]]]]]]].

  unfold heap_correspond; splits; try congruence; auto.
  intros x Hindom. clear H5 H1.

  eapply heap_update_ty_item_s; eauto.

Qed.

Lemma defs_update_exists: forall ds a t x,
  defs_has ds (def_fld a t) ->
  exists ds', defs_update ds a x ds'.
Proof.
  introv Hhas.
  eapply (defs_has_ind Hhas); intros.
  - exists (defs_cons d (def_fld a (avar_f x))); constructor.
  - destructc H as ds' Hds'.
    exists (defs_cons ds' d0).
    constructor; auto.
Qed.

Lemma heap_updateable: forall s x ds a t y,
    ok s ->
    binds x (item_obj ds) s ->
    defs_has ds (def_fld a t) ->
    exists s' ds',
      ok s' /\
      dom s = dom s' /\
      (forall z v, z <> x ->
              binds z v s ->
              binds z v s') /\
      defs_update ds a y ds' /\
      binds x (item_obj ds') s'.
Proof.
  intros s; induction s using env_ind; intros.
  - exfalso. apply (binds_empty_inv H0).
  - pose proof (defs_update_exists y H1) as [?ds' ?H].
    pose proof (binds_push_inv H0) as [[?H ?H] | [?H ?H]]; subst.
    + exists (s & x ~ item_obj (ds')) ds'.
      splits; auto.
      * simpl_dom; auto.
      * intros.
        pose proof (binds_push_inv H4) as [[?H ?H] | [?H ?H]]; try congruence.
        auto.
    + assert (Hoks: ok s) by auto.
      pose proof (IHs x0 _ _ _ y Hoks H4 H1) as
          [?s' [?ds' [?Hoks' [?Hdeq [?H [?H ?H]]]]]].
      exists (s' & x ~ v) (ds'0). splits; auto.
      * apply ok_push; auto 1.
        pose proof (ok_push_inv H) as [_ ?H].
        rewrite <- Hdeq; auto.
      * simpl_dom. rewrite Hdeq; auto.
      * intros.
        pose proof (binds_push_inv H9) as [[?H ?H] | [?H ?H]]; subst; auto.
Qed.

Lemma heap_update_exists_ref: forall G e s vObjRef vObjLoc vArgRef vArgLoc a S T,
    inert G ->
    sealed G ->
    heap_correspond G e s ->
    G⁏e ⊢ vObjRef ∶ typ_rcd (dec_fld a S T) ->
    G⁏e ⊢ vArgRef ∶ S ->
    binds vObjRef vObjLoc e ->
    binds vArgRef vArgLoc e ->
    exists s', heap_update s vObjLoc a vArgLoc s'.
Proof.
  introv Hin Hcl Hst Hty HTy HbiObj HbiArg.
  pose proof (canonical_forms_fld_ref Hin Hcl Hst HbiObj Hty) as [?ds [?t [?Hbi [?Hdef _]]]].
  unfold heap_correspond in Hst.
  destruct Hst as [?H [Hwk [Hok ?H]]].
  pose proof (heap_updateable vArgLoc Hok Hbi Hdef) as [?s' [?ds' [?H ?H]]]; destruct_all.
  exists s'; unfold heap_update; splits; auto.
  exists ds ds'; splits; auto.
Qed.

Lemma heap_update_exists: forall G e s vObjRef vArgRef a S T,
    inert G ->
    sealed G ->
    heap_correspond G e s ->
    G⁏e ⊢ vObjRef ∶ typ_rcd (dec_fld a S T) ->
    G⁏e ⊢ vArgRef ∶ S ->
    ctx_k G vObjRef v_ref ->
    ctx_k G vArgRef v_ref ->
    exists vObjLoc vArgLoc s' ,
    binds vObjRef vObjLoc e /\
    binds vArgRef vArgLoc e /\
    heap_update s vObjLoc a vArgLoc s'.
Proof.
  introv Hin Hcl Hst Hty HTy HvkObj HvkArg.
  destruct (heap_correspond_exists_loc Hst HvkObj) as [vObjLoc HbiObj].
  destruct (heap_correspond_exists_loc Hst HvkArg) as [vArgLoc HbiArg].
  exists vObjLoc vArgLoc.
  destruct (heap_update_exists_ref Hin Hcl Hst Hty HTy HbiObj HbiArg) as [s' Hrest].
  exists s'.
  auto.
Qed.

(* If an updated heap has an item, then the old heap has the same item or it was the changed item *)
Lemma heap_update_binds_rev S1 y1 a y3 S2 x v:
  heap_update S1 y1 a y3 S2 ->
  binds x v S2 ->
  (binds x v S1 \/ x = y1).
Proof.
  introv Hu Hbi.
  destructc Hu as ? ? Hdom HbiS ds ds' Hrest. destructc Hrest as Hbiy1 Hdsu Hbiy2.
  destruct (classicT(x = y1)).
  - auto.
  - left.
    pose proof (binds_to_dom Hbi) as Hdomx.
    rewrite <- Hdom in Hdomx.
    eapply dom_to_binds in Hdomx.
    destruct Hdomx.
    pose proof H1.
    eapply HbiS in H1; eauto.
    pose proof (binds_functional Hbi H1).
    subst. auto.
Qed.

(* If a heap has an item, then the updated heap has the same item or it was the changed item *)
Lemma heap_update_unchanged S1 y1 a y3 S2 y ds:
    heap_update S1 y1 a y3 S2 ->
    binds y (item_obj ds) S1 ->
    y <> y1 ->
    binds y (item_obj ds) S2.
Proof.
  introv Hu Hbi1 Hn.
  destructc Hu as ? ? ? ? ?.
  eauto.
Qed.

End WithTypingMode.

Lemma heap_update_binds_cases S1 yObj1 m yArg1 ds S2 y:
  heap_update S1 yObj1 m yArg1 S2 ->
  binds y ds S1  ->
  (y <> yObj1 /\ binds y ds S2) \/ (y = yObj1 /\ exists ds1 ds2, (ds=(item_obj ds1)) /\ binds y (item_obj ds2) S2 /\ defs_update ds1 m yArg1 ds2).
Proof.
  intros.
  destructc H as ? ? ? ? ? ? Hrest. destructc Hrest as ? ? ?.
  destruct (classicT(y=yObj1)).
  - subst. right. split. auto. exists x x0; splits; eauto. binds_eq. auto.
  - left; eauto.
Qed.
Lemma heap_update_dom  S1 yObj1 m yArg1 S2:
  heap_update S1 yObj1 m yArg1 S2 ->
  dom S1 = dom S2.
Proof.
  intros.
  destructc H as ? ? ? ? ? ? Hrest. destructc Hrest as ? ? ?.
  auto.
Qed.
Lemma heap_update_notin_dom  S1 yObj1 m yArg1 S2 y:
  heap_update S1 yObj1 m yArg1 S2 ->
  y # S1 ->
  y # S2.
Proof.
  intros.
  destructc H as ? ? ? ? ? ? Hrest. destructc Hrest as ? ? ?.
  rewrite <- H2. auto.
Qed.

Lemma defs_update_fields ds1 a y2 ds2:
  defs_update ds1 a y2 ds2 ->
  fields ds2 \c fields ds1 \u \{y2}.
Proof.
  introv Hu.
  induction Hu; unfold fields, FieldsDefs, fields_defs; fold fields_defs.
  - eapply subset_union_l. eapply subset_union_rl. eapply subset_union_weak_l.
    eapply subset_union_rr. unfold fields, FieldsDef, fields_def. eapply subset_refl.
  - eapply subset_union_l.
    rewrite <- (union_comm (fields d)).
    rewrite <- union_assoc.
    eapply subset_union_rr.
    eapply IHHu.
    eapply subset_union_rl.
    eapply subset_union_rr.
    eapply subset_refl.
Qed.

Lemma heap_fields_binds_defs_has x S:
  ok S ->
  x \in fields S ->
  exists y ds,
  binds y ds S /\
  x \in fields ds.
Proof.
  introv Hok Hf.
  induction S;
  unfold fields, FieldsHeap, fields_heap in Hf.
  - rewrite in_empty in Hf; contradiction.
  - destruct a. rewrite cons_to_push in *.
    fold fields_heap in Hf.
    rewrite in_union in Hf.
    destruct Hf.
    forwards y ds Hbi Hin: IHS; eauto.
    exists y ds; split. eapply binds_concat_left_ok; eauto. eauto.
    exists v i; split; eauto.
Qed.

Lemma defs_has_fields y ds a:
  defs_has ds (def_fld a (avar_f y)) ->
  y \in fields ds.
Proof.
  introv Hds.
  eapply (defs_has_ind Hds); intros.
  unfold fields, FieldsDefs, fields_defs; fold fields_defs.
  unfold fields, FieldsDef, fields_def; fold fields_defs.
  rewrite in_union, in_singleton; eauto.
  unfold fields, FieldsDefs, fields_defs; fold fields_defs.
  rewrite in_union; eauto.
Qed.

Lemma heap_binds_fields x y ds S:
  binds y ds S ->
  x \in fields ds ->
  x \in fields S.
Proof.

  introv Hbi Hf.
  induction S using env_ind.
  - eapply binds_empty_inv in Hbi. contradiction.
  - eapply binds_push_inv in Hbi; destruct_all; subst.
    rewrite <- cons_to_push.
    unfold fields, FieldsHeap, fields_heap; fold fields_heap.
    rewrite in_union; eauto.
    forwards: IHS H0.
    rewrite <- cons_to_push.
    unfold fields, FieldsHeap, fields_heap; fold fields_heap.
    rewrite in_union; eauto.
Qed.

Lemma heap_update_fields S1 y1 a y2 S2:
  heap_update S1 y1 a y2 S2 ->
  fields S2 \c fields S1 \u \{y2}.
Proof.
  introv Hu. pose proof Hu as Hu1.
  hnf in Hu; destruct_all.
  intro. intro Hin.
  forwards y ds Hbi Hinds: (heap_fields_binds_defs_has H0 Hin).
  forwards: heap_update_binds_rev; eauto.
  destruct H6; subst.
  - forwards: heap_binds_fields H6 Hinds. rewrite in_union; eauto.
  - binds_eq.
    forwards: defs_update_fields H4.
    eapply H7 in Hinds.
    rewrite in_union in Hinds.
    destruct Hinds.
    rewrite in_union. left. eapply heap_binds_fields; eauto.
    rewrite in_union. eauto.
Qed.


(* If two heaps with the same order of keys are updated,
  then either the last object is updated,
  or the rest of the heap is updated *)
Lemma heap_update_keys_eq_cases S1 S2 x f y z ita itb:
  keys S1 = keys S2 ->
  heap_update (S1 & z ~ ita) x f y (S2 & z ~ itb) ->
  (
    heap_update (S1) x f y (S2) /\
    x <> z /\
   ita = itb
  ) \/ (
    S1 = S2 /\
    x = z /\
    exists dsa dsb,
      ita = item_obj dsa /\
      itb = item_obj dsb /\
      defs_update dsa f y dsb
  ).
Proof.
  introv Heq Hup.
  unfold heap_update in *.
  destructecs Hup.
  destruct (classicT(x=z)).
  - subst.
    right.
    eapply binds_push_eq_inv in Hup4.
    eapply binds_push_eq_inv in Hup. subst.

      eapply ok_push_inv in Hup0. destructecs Hup0.
    splits.
    +

    eapply env_extens. eauto. auto.
    intros. forwards: Hup3 z0 v.

      intro; subst. apply Hup0. eapply binds_to_dom; eauto.
      eapply binds_push_neq. auto.
      intro; subst. apply Hup0. eapply binds_to_dom; eauto.
      eapply binds_push_neq_inv. eapply H0.
      intro; subst. apply Hup0. eapply binds_to_dom; eauto.
    + auto.
    + exists ds ds'. auto.
  - left.
    splits.
    splits.
    + auto.
    + auto.
    + rewrite <- ! keys_from_list_dom. rewrite Heq. auto.
    + intros.
      eapply ok_push_inv in Hup0. destructecs Hup0.
      forwards: Hup3 z0 v. auto.
      eapply binds_push_neq. eapply H0.
      intro; subst. apply Hup0. eapply binds_to_dom; eauto.
      eapply binds_push_neq_inv. eapply H1.
      intro; subst. apply Hup0. eapply binds_to_dom; eauto.
    + exists ds ds'; splits.
      eapply binds_push_neq_inv; eauto. auto.
      eapply binds_push_neq_inv; eauto. 
    + auto.
    + forwards: Hup3 z ita. auto. eapply binds_push_eq. eapply binds_push_eq_inv in H. auto.
Qed.
(* An empty heap can not be updated *)
Lemma heap_update_empty_inv S x f y:
  heap_update empty x f y S ->
  False.
Proof.
  introv Hu. hnf in Hu. destructecs Hu.
  eapply binds_empty_inv; eauto.
Qed.

(* Updating the heap while preserving the order of keys is deterministic *)
Lemma heap_update_keys_eq S1 S2 x f y  S:
  keys S = keys S1 ->
  keys S = keys S2 ->
  heap_update S x f y S1 ->
  heap_update S x f y S2 ->
  S1 = S2.
Proof.
  introv Heq1 Heq2 Hup1 Hup2.

  reverts S1 S2.
  induction S using env_ind; intros.
  - false. eapply heap_update_empty_inv; eauto.
  -

    lets Hup1': Hup1.
    lets Hup2': Hup2.

    hnf in Hup1, Hup2.
    destructecs Hup1.
    destructecs Hup2.

    destruct (env_case S1). subst. eapply binds_empty_inv in Hup1. false.
    destruct (env_case S2). subst. eapply binds_empty_inv in Hup2. false.
      
    destructecs H.
    destructecs H0.
    subst.

    rewrite concat_def, single_def, keys_def, ! LibList.app_cons_one_r, ! LibList.map_cons in Heq1. inversions Heq1.
    rewrite concat_def, single_def, keys_def, ! LibList.app_cons_one_r, ! LibList.map_cons in Heq2. inversions Heq2.

    rewrite <- keys_def in H1, H2.

    forwards: heap_update_keys_eq_cases Hup1'. auto.
    forwards: heap_update_keys_eq_cases Hup2'. auto.

    destructecs H; destructecs H0; try contradiction.
    +
      forwards: IHS H3 H5; fweauto. subst. binds_eq.
      forwards: Hup11 x2. auto. eapply binds_push_eq.
      forwards: Hup5 x2. auto. eapply binds_push_eq. auto.
    + subst.
      binds_eq.
      forwards: defs_update_functional Hup7 Hup13. subst.
      eapply binds_push_eq_inv in Hup6. eapply binds_push_eq_inv in Hup1. eapply binds_push_eq_inv in Hup2. inversions Hup1. inversions Hup2. auto.

Qed.