Set Implicit Arguments.

Require Import LibExtra.
Require Import AbstractSyntax Deref HeapUpdate OperationalSemantics ReductionLemmas.

(** * Lemmas about free variables and reduction *)

Lemma red_fv_config_more_single c c' :
  c ↦ c' -> fv_config c \c fv_config c'.
Proof.
  introv Hred. destruct Hred; simpl.
  all: try solve[introv Hin; repeat(rewrite in_union in *); tauto].
Qed.

Lemma red_fv_config_more c n c' :
  c ⊣n↦* c' -> fv_config c \c fv_config c'.
Proof.
  introv Hred. reverts c'. induction n; intros; inversions Hred;
    eauto using subset_refl.
  autospecialize. eapply red_fv_config_more_single in H2.
  eauto using subset_trans.
Qed.

(* Set of free variables added between C1 and C2 *)
Definition config_fv_extension (C1 C2: config): fset var :=
  ((dom (config_heap C2)) \- (dom (config_heap C1))) \u  ((dom (config_env C2)) \- (dom (config_env C1))).


Lemma config_fv_extension_diag C:
  config_fv_extension C C = \{}.
Proof.
  unfold config_fv_extension. rewrite ! remove_same, union_empty_r. auto.
Qed.

Lemma red_confing_fv_extension_disjoint C1 C2:
  C1 ↦ C2 ->
  disjoint (fv_config C1) (config_fv_extension C1 C2).
Proof.
  introv Hred.
  unfold config_fv_extension.
  inversions Hred; simpl; simpl_dom; rewrite ?union_remove, ? remove_same, ? union_empty_r, ? union_empty_l; try eapply disjoint_empty_r.
  - rewrite remove_disjoint.
    eapply disjoint_singleton_r. notin_solve.
    eapply disjoint_comm.
    eapply disjoint_singleton_r. notin_solve.
  - erewrite HeapUpdate.heap_update_dom at 2 by eauto.
    rewrite ?union_remove, ? remove_same, ? union_empty_r, ? union_empty_l; try eapply disjoint_empty_r.
  - rewrite ! remove_disjoint.
    eapply disjoint_union_r.
    eapply disjoint_singleton_r. notin_solve.
    eapply disjoint_singleton_r. notin_solve.
    eapply disjoint_comm.
    eapply disjoint_singleton_r. notin_solve.
    eapply disjoint_comm.
    eapply disjoint_singleton_r. notin_solve.
  - rewrite ! remove_disjoint.
    eapply disjoint_union_r.
    eapply disjoint_singleton_r. notin_solve.
    eapply disjoint_singleton_r. notin_solve.
    eapply disjoint_comm.
    eapply disjoint_singleton_r. notin_solve.
    eapply disjoint_comm.
    eapply disjoint_singleton_r. notin_solve.
Qed.

Lemma config_fv_extension_subset_trans C1 C2 C3:
  config_fv_extension C1 C3 \c config_fv_extension C1 C2 \u config_fv_extension C2 C3.
Proof.

  unfold config_fv_extension.
  rewrite union_zip.
  eapply subset_union_2.
    intro. intros. rewrite in_remove in H. rewrite in_union, ? in_remove. destruct H; destruct (fv_decide x (dom (config_heap C2))); eauto.
    intro. intros. rewrite in_remove in H. rewrite in_union, ? in_remove. destruct H; destruct (fv_decide x (dom (config_env C2))); eauto.
Qed.


Lemma red_closure_confing_fv_extension_disjoint C1 C2 n:
  C1  ⊣ n ↦*  C2 ->
  disjoint (fv_config C1) (config_fv_extension C1 C2).
Proof.
  introv Hred.
  induction Hred.
  - rewrite config_fv_extension_diag. eapply disjoint_empty_r.
  - forwards: red_confing_fv_extension_disjoint. eauto.
    eapply disjoint_subset_l.
    2: eapply config_fv_extension_subset_trans.
    eapply disjoint_union_r. eauto. eapply disjoint_subset_r. eauto. eapply red_fv_config_more; eauto.
Qed.

Lemma config_fv_extension_trans C1 C2 C3 n1 n2:
  C1 ⊣n1↦* C2 ->
  C2 ⊣n2↦* C3 ->
  config_fv_extension C1 C3 = config_fv_extension C1 C2 \u config_fv_extension C2 C3.
Proof.
  introv Hred1 Hred2.
  eapply fset_extens.
  - eapply config_fv_extension_subset_trans; eauto.
  -
  unfold config_fv_extension.
  rewrite union_zip.
  eapply subset_union_2.
  rewrite union_comm.
  +
  intro. intros. rewrite in_remove. rewrite in_union, ? in_remove in H. destructecs H; splits; auto 1.
  * eapply notin_subset.
  eapply red_closure_mono_heap_dom; eauto. eauto.
  * 
  eapply red_closure_mono_heap_dom; eauto.
  +
  intro. intros. rewrite in_remove. rewrite in_union, ? in_remove in H. destructecs H; splits; auto 1.
  * 
  eapply red_closure_mono_env_dom; eauto.
  * eapply notin_subset.
  eapply red_closure_mono_env_dom; eauto. eauto.
Qed.
