Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax OperationalSemantics.
Require Import Lia.

(** * Lemmas about untyped reduction *)

Lemma red_closure_mono_e t fs S e Lv t2 fs2 S2 e2 Lv2 n:
  (t; fs; S; e; Lv) ⊣ n ↦* (t2; fs2; S2; e2; Lv2) ->
  exists e3,
    e2 = e & e3.
Proof.
  introv Hred.
  remember_dependent_config t fs S e Lv as C1.
  remember_dependent_config t2 fs2 S2 e2 Lv2 as C2.
  induction Hred; intros; subst.
  - inversions HeqC1.
    eexists. rewrite concat_empty_r. auto.
  - destruct C2.
    forwards: IHHred; eauto.
    destructc H0 as e3 He3; subst.
    inversions H; try solve [exists e3; eauto].
    all: eexists; rewrite <- concat_assoc; eauto.
Qed.

Lemma red_closure_mono_Lv t fs S e Lv t2 fs2 S2 e2 Lv2 n:
  (t; fs; S; e; Lv) ⊣ n ↦* (t2; fs2; S2; e2; Lv2) ->
  Lv \c Lv2.
Proof.
  introv Hred.
  remember_dependent_config t fs S e Lv as C1.
  remember_dependent_config t2 fs2 S2 e2 Lv2 as C2.
  induction Hred; intros; subst.
  - inversions HeqC1.
    eapply subset_refl.
  - destruct C2.
    forwards: IHHred; eauto.
    eapply subset_trans. eauto.
    inversions H;
    repeat rewrite union_assoc;
    eapply subset_union_weak_r.
Qed.


Lemma red_mono_heap_dom c1 c2 :
  c1 ↦ c2 ->
  dom (config_heap c1) \c dom (config_heap c2).
Proof.
  destruct c1. destruct c2. simpl.
  introv Hred. inversions Hred; eauto using subset_refl.
  - destructc H2 as _ _ Heq _. rewrite Heq. eauto using subset_refl.
  - rewrite dom_push. eauto using subset_union_weak_r.
  - rewrite dom_push. eauto using subset_union_weak_r.
Qed.
Lemma red_mono_env_dom c1 c2 :
  c1 ↦ c2 ->
  dom (config_env c1) \c dom (config_env c2).
Proof.
  destruct c1. destruct c2. simpl.
  introv Hred. inversions Hred; simpl_dom; eauto using subset_refl, subset_union_weak_r.
Qed.

Lemma red_closure_mono_heap_dom c1 c2 n :
  c1 ⊣n↦* c2 ->
  dom (config_heap c1) \c dom (config_heap c2).
Proof.
  introv Hred. reverts c2. induction n; intros; inversions Hred;
    eauto using subset_refl.
  forwards:IHn; fweauto. eapply red_mono_heap_dom in H2. firstorder.
Qed.

Lemma red_closure_mono_env_dom c1 c2 n :
  c1 ⊣n↦* c2 ->
  dom (config_env c1) \c dom (config_env c2).
Proof.
  introv Hred. reverts c2. induction n; intros; inversions Hred;
    eauto using subset_refl.
  forwards:IHn; fweauto. eapply red_mono_env_dom in H2. firstorder.
Qed.

Lemma red_stack_cases t1 fs1 S1 e1 Lv1 t2 fs2 S2 e2 Lv2 :
  t1 ; fs1 ; S1 ; e1 ; Lv1 ↦ t2 ; fs2 ; S2 ; e2 ; Lv2 ->
  stack_suffix fs1 fs2 \/
  exists w1 f1,
    t1 = trm_var (avar_f w1) /\
    fs1 = (stack_cons f1 fs2).
Proof.
  introv Hred.
  inversions Hred;
  try solve [left; eapply stack_suffix_refl].
  - right. exists x (frame_let t); eauto.
  - left. eapply stack_suffix_cons. eapply stack_suffix_refl.
Qed.

Lemma red_closure_invl C1 C2 n:
  C1  ⊣ (S n) ↦* C2 ->
  exists C3,
  C1  ↦  C3 /\
  C3 ⊣ n ↦*  C2.
Proof.
  introv Hred.
  reverts C1 C2.
  induction n; intros; inversions Hred.
  - inversions H0; eauto.
  - 
    forwards Hex: IHn H0.
    destructecs Hex.
    exists C0; split; auto.
    eapply red_trans; eauto.
Qed.

Lemma red_closure_short C1 C2 m n:
  C1 ⊣ n ↦* C2 ->
  m <= n ->
  exists C3 mn, 
  C1 ⊣ m ↦* C3 /\
  C3 ⊣ mn ↦* C2 /\
  n = m + mn.
Proof.
  introv Hred Hmn.
  reverts C1 C2 n.
  induction m; intros.
  - exists C1 n.
    splits; eauto.
  - forwards: IHm Hred. lia.
    destructecs H.
    destruct n.
    lia.
    apply le_S_n in Hmn.
    destruct mn. lia.
    eapply red_closure_invl in H1.
    destructecs H1.
    exists C0 mn; splits.
    eauto using red_trans.
    auto.  
    rewrite Nat.add_succ_comm.
    auto.
Qed.

Lemma prop_ex_4 (A B C D: Type) (f: A -> B -> C -> D -> Prop) (a: A) (b: B) (c: C) (d: D):
  (exists ax bx cx dx,
  f ax bx cx dx /\
  ax = a /\ bx = b /\ cx = c /\ dx = d) ->
  f a b c d.
Proof.
  introv Hex.
  destructecs Hex.
  subst. auto.
Qed.
Lemma prop_ex_2 (A B: Type) (f: A -> B -> Prop) (a: A) (b: B):
  (exists ax bx ,
  f ax bx /\
  ax = a /\ bx = b ) ->
  f a b.
Proof.
  introv Hex.
  destructecs Hex.
  subst. auto.
Qed.
