Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping
        InertTightSubtyping
        GeneralToTight TightTyping InvertibleTyping
        RecordTypes RecordDefs InertContext PreciseTyping
        OperationalSemantics Substitution Renaming Weakening
        HeapCorrespondence HeapUpdate CanonicalForms
         Narrowing
        EnvCorrespondence EWeakening Visibility
        ConfigTyping ConfigVarkinds TFree Deref ReferenceTypes
        TypedOperationalSemantics Defree Defs Closedness.
Require Import TypedReductionEquivalence NothingFresh TypedPreservationLemmas.

(** * Preservation lemma for typed reduction *)

Section WithTypingMode.
Context (typing_mode: Mode).

(* Solves goals  x \notin fv (open y T) which come up
   after notin_solve in inert_obj_typ and preservation_helper.
   Perhaps should be stated as a lemma instead.
*)
Local Hint Extern 1 (?x \notin fv (open_rec ?n ?y ?T ) ) =>
  destruct (fv_open_cases y T n) as [HFoc | HFoc];
  rewrite HFoc; auto : core.


Ltac invert_ty_config :=  inversion Htyc as [? ? ? ? ? ? ? ? HinG Hhc Htys Htfs Hty Htft Hfv Hnull HclG Hclc HclT]; subst; clear HinG.

Lemma typing_call_met_recvtyp (TC S0: typ) (xc xz w1 w2: var):
  xc \notin fv S0 \u fv TC \u \{w1} \u \{w2}->
  xz \notin fv S0 \u fv TC \u \{w1} \u \{w2} ->
  xc <> xz ->
 (((typ_and (open xz TC) (typ_and (open w1 S0) (open xc S0)))/[ xz -> w2]) = (typ_and (open w2 TC) (typ_and (open w1 S0) (open xc S0)))).
Proof.
  introv Hnxc Hnxz Hnxcz.
      {
        sympl.
        assert ((S0 /[->w1] /[ xz -> w2]) = (open w1 S0)). rewrite subst_fresh; eauto.
        assert ((TC /[ xz -> w2]) = TC). rewrite subst_fresh; eauto.
        assert ((S0/[->xc]  /[ xz -> w2]) = (open xc S0)). rewrite subst_fresh; eauto.
        rewrite H.
        rewrite subst_open_commut.
        rewrite  H0. 
        rewrite substi_eq_var. rewrite H1. auto.
      }
Qed.

Lemma typing_call_met_recvty G e TC S0 S1 (w1 w2: var):
  inert G ->
  G ⁏ e ⊢ w1 ∶ TC /[-> w2] ->
  ctx_binds w1 (typ_and (typ_bnd S0) S1) G ->
  G ⁏ e ⊢ w1 ∶ typ_and TC /[-> w2] (typ_and S0 /[-> w1] S0 /[-> w1]).
Proof.
  introv HinG Hty Hbiw1G.
        apply ty_var_and. auto.
        apply ty_var_and.
        all: apply ty_rec_elim; [eapply ty_var_sub; [ | eapply subtyp_and11 ]; eauto | eapply inert_bnd_indep; eauto].
Qed.

Lemma typing_call_met G e S w1 y1 d m t w2 T:
  inert G ->
  sealed G ->
  ok e ->
  heap_correspond G e S ->
  binds w1 y1 e ->
  binds y1 (item_obj d) S ->
  defs_has d (def_met m t) ->
  G ⁏ e ⊢ trm_call (avar_f w1) m (avar_f w2) ∶ T ->
  G ⁏ e ⊢ open_met w2 w1 t∶ T.
Proof.
  introv Hin HclG Hoke Hhc Hbw1ye Hbiw1S Hhas Hty.
  assert (Hok : ctx_ok G) by eauto.

  invert_focus_typing Hty.

  pose proof (canonical_forms_met_ref Hin HclG Hhc Hbw1ye H1) as Hcf.
  destruct Hcf as [L [yd' [t' [S0 [S1 [S2 [HbniG [HblG [Hbie [Hhas' [Hsu [Htf' Hop]]]]]]]]]]]].
  binds_eq.
  clear Hsu.

  apply defs_has_met_inv with (t := t') in Hhas; auto.
  inversions Hhas; eauto; clear H.
  rename H2 into Htw2, H1 into Htw1m, H3 into Htw1c.

  pick_fresh xz. rename Fr into Hfrxz.
  pick_fresh xc. rename Fr into Hfrxc.
  assert (Hfr : xz \notin L) by eauto; specialize (Hop _ Hfr);
  clear Hfr.
  assert (Hfr : xc <> xz) by eauto; specialize (Hop _ Hfr); clear Hfr.
  assert (Hfr : xc \notin L) by eauto; specialize (Hop Hfr); clear Hfr.
  assert (Hfrzcop : xz \notin fv (open_rec 1 xc TR) \u
                       fv (open_rec 1 xc t)) by
      (eapply notin_union_l;
       match goal with
       | [ |- ?xc \notin fv (open_rec ?n ?w ?T) ] =>
         replace (open_rec n w T) with
             (open_rec_vars n (w :: nil)%list T); eauto;
         eapply notin_open_rec_vars_list;
         try notin_solve;
         rewrite from_list_cons, from_list_nil, union_empty_r; eauto
       end).
  unfold ctx_obj in Hop.

      assert (ctx_met_refined_heap G xz TZ xc TC w1 S0 ⁏ e ⊢ open_met xz xc t
      ∶ open_met xz xc TR) as Hop2.
      {

        (* Deal with ctx_hide in Hop *)
        ctx_split_from xz in Hop.
        destruct (mode_visibility' typing_mode) eqn: Hmv; auto.
        eapply unhiding in Hop.
        ctx_flatten in Hop.
        auto.
      }
        clear Hop. rename Hop2 into Hop.


      assert (G § xc / v_recv ~ ((typ_and (open xz TC) (typ_and (open w1 S0) (open xc S0)))/[ xz -> w2]) ⁏ e ⊢ ((open_met xz xc t) /[ xz -> w2])
      ∶ ((open_met xz xc TR) /[ xz -> w2])).
      {
          ctx_split_from xc.
          rewrite <- (ctx_subst_empty xz w2).
          rewrite <- ctx_subst_push.
          eapply subst_ty; eauto using ewkn_ty_trm.
          3: notin_solve.
          3: eapply weaken_end; auto. 3: rewrite subst_fresh; auto. 3: apply Htw2.
          ctx_flatten. apply Hop.
          ctx_flatten; eauto. destruct typing_mode; eauto. notin_solve.
          apply ctx_weakens_map; auto.
          apply ctx_ok_concat_map.
          ctx_flatten. destruct typing_mode; auto.
          intros. apply ctx_vis_n. auto.
          eapply inert_ty_vis; eauto.
      }
      assert( (((typ_and (open xz TC) (typ_and (open w1 S0) (open xc S0)))/[ xz -> w2]) = (typ_and (open w2 TC) (typ_and (open w1 S0) (open xc S0))))).
      eapply typing_call_met_recvtyp; eauto.

      assert (G § xc / v_recv ~ (typ_and (open w2 TC) (typ_and (open w1 S0) (open xc S0))) ⁏ e ⊢ open_met w2 xc t ∶ open_met w2 xc TR) as TyC.
      {
        rewrite <- H1.
        rewrite subst_intro with (x:=xz) (y:=w2); auto.
        rewrite subst_intro with (x:=xz) (y:=w2); auto.
      }
      assert (G ⁏ e ⊢ w1 ∶ (typ_and (open w2 TC) (typ_and (open w1 S0) (open w1 S0)))) as HtyX.
      {
        eapply typing_call_met_recvty; eauto.
      }
      rewrite open_commut with (t0 := t) ; eauto.
      rewrite open_commut with (t0 := TR) ; eauto.
      rewrite subst_intro with (x:=xc) (y:=w1).
      rewrite subst_intro with (x:=xc) (y:=w1).

      eapply subst_end with (property:=ty_trm _); auto.
      rewrite <- open_commut with (t0 := TR); auto.
      rewrite <- open_commut with (t0 := t); auto.
      apply TyC.
      assert (((typ_and (open w2 TC) (typ_and (open w1 S0) (open xc S0))) /[ xc -> w1]) = typ_and (open w2 TC) (typ_and (open w1 S0) (open w1 S0))) as Heqa.
      sympl.
      assert (( S0/[->w1]  /[ xc -> w1]) = (open w1 S0)) as How1S. rewrite subst_fresh; eauto.
      assert (( TC/[->w2]  /[ xc -> w1]) = open w2 TC) as How2C. rewrite subst_fresh; eauto.
      rewrite How1S.
      rewrite How2C. rewrite subst_open_commut. rewrite substi_eq_var.  rewrite subst_fresh; eauto.
      rewrite Heqa.
      apply HtyX.
      eauto using inert_binds_vis.
      auto.
      auto.
Qed.


Lemma typing_apply_fun G e t' z T0 L' S:
  inert G ->
  ok e  ->
  G ⁏ e ⊢ z ∶ S ->
  (forall x0 : var, x0 \notin L' -> G §ᶻ x0 ~ S ⁏ e ⊢ t' /[-> x0] ∶ T0 /[-> x0]) ->
  G ⁏ e ⊢ t' /[-> z] ∶ T0 /[-> z].
Proof.
  introv HinG Hoke Htyz Hop'.

    pick_fresh v.
    
    pose proof (@subst_ty typing_mode trm (subst_trm) (ty_trm typing_mode) _ _ z S G v v_par ctx_empty e
                           (open v t') (open v T0))
      as Hsubst.
    simpl in Hsubst.
    assert (Hvfr' : v \notin L') by eauto;
      specialize (Hop' v Hvfr'); clear Hvfr'.
    specialize (Hsubst Hop').
    assert (Hok' : ctx_ok (G § v / v_par ~ S)) by eauto;
      specialize (Hsubst Hok'); clear Hok'.
    assert (Hvfr' : v \notin ctx_fv G) by eauto;
      specialize (Hsubst Hvfr'); clear Hvfr'.
    assert (Hvfr' : renv_notin v e) by eauto;
      specialize (Hsubst Hoke Hvfr'); clear Hvfr'.
    rewrite subst_fresh in Hsubst; eauto.
    specialize (Hsubst Htyz).
    lapply Hsubst; intro Hsubst2.
*
      
      assert (SubstOpenCommut TrmAbstractSyntax) by eapply TransformSubstOpenCommut.
      assert (SubstFresh TrmAbstractSyntax) by eapply TransformSubstFresh.

      repeat(rewrite subst_open_commut in Hsubst2).
      assert (Heq : v \[ v -> z] = z) by (cases_if; eauto).
      rewrite Heq in Hsubst2; clear Heq.
      rewrite subst_fresh in Hsubst2; eauto.
      repeat(rewrite subst_fresh in Hsubst2; eauto). 
    * (* visibility *)
      intro.
      eapply inert_ty_vis; eauto.
Qed.

(* Lemma 2 in "Reference mutability for DOT"
   Part of theorem 6 in "Reference mutability for DOT" *)
Lemma typed_preservation : forall (F1 F2 : ctx) (C1 C2 : config) (T : typ),
    F1 ⊢ C1 ∶ T ↦ F2 ⊢ C2 ->
    ty_config F2 C2 T.
Proof.
  introv Hstep. inversion Hstep; subst.

  - (* Case tred_read_obj *)
    rename F1 into G, F' into G'.
    invert_ty_config.
    assert (Hok : ctx_ok G) by eauto.
    assert (Hoke : ok e) by eauto.
    assert (HGecor : renv_corr G e) by eauto.
    assert (Hfrw2G : w2 §# G) by eauto using heap_correspond_notin_dom.


    (* Invert the focus typing ty_trm_read *)
    invert_focus_typing Hty.
    rename H0 into Htyw1.
    (* New reference type correspondence *)
    assert (inert_typ (typ_objref R1 R2) v_loc) as Hint by eauto using inert_binds_k_inert.
    assert (typ_corr T2 (typ_objref R1 R2)). {
      destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hmva; subst.
      * constructor.
        destruct (mode_refcap_corr' typing_mode) eqn: Hmcor;
        try (typing_mode_contradiction typing_mode); eauto.
      * (* rodot *)
        destructc Htcor as Tstrip Tcap ? ? ? ?.
        subst.
        constructor.
        destruct (mode_refcap_corr' typing_mode) eqn: Hmcor;
        try (typing_mode_contradiction typing_mode).
        eexists; split.
        + apply eq_refl.
        + inversions Hint.
          hnf in H7.
          destruct (mode_refcaps' typing_mode) eqn: Hmrc;
          try (typing_mode_contradiction typing_mode).
          inversions H7. auto.
    }
    (* Inertness: amounts to showing T2 inert *)
    assert (inert_typ T2 v_ref) by eauto using inert_typ_corr_inert.
    assert (inert (G § w2 / v_ref ~ T2)).
    { constructor; auto. all: destruct Hfrfvw2; auto. }
    assert (Hhc2 : heap_correspond (G § w2 / v_ref ~ T2) (e & w2 ~ y2) Sigma).
    { eapply heap_correspond_epush; eauto 3. }

    destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hmva.
      + (* basic modes *)
        (* Construct the typing of the target configuration. *)
        eapply ty_config_c with (T := U); eauto.
        * (* w₂ has type T₀ *)
          eapply ty_trm_var; eauto.

          pose proof (canonical_forms_fld_ref Hin HclG Hhc Hbwye Htyw1) as Hcf.
          destructc Hcf as ds' y2' Hby1S' Hhas' Htypy2 Hst Hy2loc.
          binds_eq. rename ds' into yd.
          apply defs_has_inv with (t := avar_f y2) in Hhas'; eauto;
          inversions Hhas'; rename y2' into y2.

          apply ty_renv_binds_rev_var with (y := y2); eauto.
          destruct (mode_refcap_corr' typing_mode) eqn: Hmcor; try typing_mode_contradiction typing_mode; auto.
          apply weaken_push; auto.
          apply ewkn_push; eauto.
          eapply typing_viewpoint_adapt_same; eauto.
       * (* Everything not in type-position are references *)
          hnf. introv Htf. inversions Htf. hnf. exists (typ_and (typ_bnd R1) R2) G. eauto.
       * eapply ty_nothing_fresh_in_config_push; eauto.
       * eapply null_p; eauto.
       * eapply tred_sealed_read; eauto.
       * eapply closed_read; eauto.
     + (* rodot *)
        subst.
        destructc Htcor as Tstrip Tcap ? ? ? ?.
        subst.
        clear Htys.
        eapply ty_config_c
          with (T := (typ_and Tstrip (typ_cap_adapt typ_refcap_mutability
                                                    w1 Tcap))); eauto.
        eapply ty_trm_var; eauto.

        pose proof (canonical_forms_fld_ref Hin HclG Hhc Hbwye Htypw1) as Hcf.
        destructc Hcf as ds' y2' Hby1S' Hhas' Htypy2 Hst Hy2loc.
        binds_eq. rename ds' into yd.
        apply defs_has_inv with (t := avar_f y2) in Hhas'; eauto;
        inversions Hhas'; rename y2' into y2.
        *
          eapply ty_var_and.
          ++ (* record *)
            eapply reference_incap_type.
            7: eapply binds_tail.
            destruct (mode_has_incap' typing_mode) eqn: Hmm; try typing_mode_contradiction typing_mode; auto.
            2: {
              eapply ctx_sealed_push; eauto.  eapply closed_sealed0. eauto.
            }
            1, 2, 3: eauto.
            
            eapply @weaken_push with (P:= (@obj_typing _ (ty_var _))); auto. eapply weaken_ty_var.
            eapply @ewkn_push with (P:= (@obj_typing _ (ty_var _))); auto. eapply ewkn_ty_var.
            eapply Htypy2; eauto.
            eapply @weaken_push with (P:= (@obj_ref_incap (ty_incap _))); auto. eapply weaken_ty_incap.
            eapply @ewkn_push with (P:= (@obj_ref_incap (ty_incap _))); auto. eapply ewkn_ty_incap.
          ++ (* capabilities *)
            eapply ty_var_sub.
            eapply ty_var_var; eauto.
            eapply subtyp_trans. eapply subtyp_and12.
            eapply subtyp_typ. eauto.
            intros.
            eapply (@subtyp_or _ (and_or_subtyp _));
            eauto.
        * (* Everything not in type-position are references *)
          hnf. introv Htf. inversions Htf. hnf. eexists. exists G. eauto.
        * eapply ty_nothing_fresh_in_config_push; eauto.
        * eapply null_p; eauto.
        * eapply tred_sealed_read; eauto.
        * eapply closed_read; eauto.
  - (* Case tred_read_fun *)
    rename F1 into G, F' into G'.
    invert_ty_config.
    assert (Hok : ctx_ok G) by eauto.
    assert (Hoke : ok e) by eauto.
    assert (HGecor : renv_corr G e) by eauto.
    assert (Hfrw2G : w2 §# G) by eauto using heap_correspond_notin_dom.
    assert (Hfrw2e : w2 # e) by eauto using heap_correspond_notin_dom.
    assert (Hhc2 : heap_correspond (G § w2 / v_ref ~ typ_fun S T1)
                                   (e & w2 ~ y2)
                                   Sigma).
    { eapply heap_correspond_epush; eauto 3. constructor. auto. }

    (* Invert the focus typing ty_trm_read *)
    invert_focus_typing Hty.

    assert (HinGw2 : inert (G § w2 / v_ref ~ typ_fun S T1)).
    { constructor; auto; hnf; intros Hw2in.
      - destruct Hfv as [[_ Hvf] _]. apply Hfrw2. auto.
      - destruct Hfv as [[_ Hvf] _]. apply Hfrw2, Hvf.
        apply ctx_binds_k_after_binds, ctx_binds_fv_subset in Hbk. auto. }
    econstructor; eauto.
    * (* Current hole has the right type for the awaiting continuation *)
      eapply tred_lit_fun_ty_trm; eauto.
    * (* Everything not in type-position are references *)
      eapply all_tfree_k_let_fun; eauto.
    * eapply tred_ctx_dom_lit_fun; eauto.
    * eapply null_p; eauto.
    * eapply ctx_sealed_push; eauto.
      eapply ctx_binds_k_after_binds in Hbk.
      eapply ctx_closed_binds; eauto.
    * eapply closed_read; eauto.
  - (* Case tred_write *)
    invert_ty_config.
    remember (trm_write (avar_f w1) a (avar_f w3)) as tz.
    induction Hty; try congruence.
    { (* Base case *)
      inversions Heqtz.
      pose proof (canonical_forms_fld_ref Hin HclG Hhc Hbw1y1e H2) as Hcf.
      destructc Hcf as _ _ _ _ _ Hst _.
      econstructor; eauto 7 using heap_update_inert.
      * eapply ty_nothing_fresh_in_config_p; eauto.
      * eapply null_p; eauto.
      * eapply closed_write; eauto.
    }
    { (* Inductive case *)
      apply IHHty; eauto using ty_stack_sub. }

  - (* Case tred_call *)
    rename F2 into G.
    invert_ty_config.

(*    invert_focus_typing Hty.*)

    assert (Hok : ctx_ok G) by eauto.
    assert (ok e) as Hoke by eauto.
    assert (HGecor : renv_corr G e) by eauto.

    econstructor; eauto.
    * (* Opening method gives the right type. *)
      eapply typing_call_met; eauto.
    * (* Everything not in type-position are references *)
      eapply all_tfree_k_call; eauto.
    * eapply ty_nothing_fresh_in_config_p; eauto.
    * eapply null_p; eauto.
    * eapply closed_call; eauto.
  - (* Case tred_apply *)
    rename F2 into G.
    invert_ty_config.
    invert_focus_typing Hty.

    assert (Htyst : ty_item_s _ G e Sigma y).
    { unfold heap_correspond in Hhc. destruct_all.
      eapply H15. apply binds_to_dom in HbyS. eauto. }
    inversions Htyst; try solve [binds_eq].
    rename H2 into Hctxby, H4 into Hop, H5 into Htf.

    assert (Hdomeq : ctx_kdom v_ref G = dom e)
      by (hnf in Hhc; destruct_all; eauto).
    assert (Hxine : x \in dom e) by (eapply binds_to_dom in Hbxye; eauto).
    assert (Hcor : renv_corr G e) by eauto.
    assert (Hok : ctx_ok G) by eauto.
    assert (ok e) as Hoke by eauto.
    pose proof (renv_corr_binds_to_binds_after Hok Hcor Hbxye) as Hbk.
    destructc Hbk as Tx Ty Gx Gy Hbkx Hbky Hcorxy.
    assert (Hctxkx : ctx_k G x v_ref) by eauto.
    rewrite <- Hdomeq in Hxine.
    pose proof (canonical_forms_fun Hin HclG Hhc Hctxkx Hctxby) as Hcan.
    destructc Hcan as y' L' T'' t' Hbxye' HbyS' Hcan'.
    destructecs Hcan'.
    binds_eq. binds_eq. binds_eq. clear H2.
    econstructor; eauto.
    * eapply typing_apply_fun; eauto.
    * eapply ty_nothing_fresh_in_config_p; eauto.
    * eapply null_p; eauto.
    * eapply closed_apply; eauto.
  - (* Case tred_let_obj *)
    rename F1 into G.
    invert_ty_config.
    invert_focus_typing Hty.

    rename H into Htyd.
    assert (HokG : ctx_ok G) by eauto.
    assert (Hoke : ok e) by eauto.
    assert (HGecor : renv_corr G e) by eauto.
    assert (HfryG : y §# G) by (eauto 3 using heap_correspond_notin_dom).
    assert (Hfrwe : w # e) by (eauto 3 using heap_correspond_notin_dom).
    assert (HfrwG : w §# G) by (eauto 3 using heap_correspond_notin_dom).
    assert(HfrwGy : w §# G § y / v_loc ~ typ_objref_loc typing_mode R).
    { apply ctx_unbound_push; auto. }
    assert (HinL : inert_typ (typ_objref_loc typing_mode R) v_loc).
    { eapply typ_objref_loc_inert; eauto. }
    assert (HinR : inert_typ (typ_objref_loc typing_mode R) v_ref) by eauto using inert_typ_loc_ref.
    assert (typ_corr (typ_objref_loc typing_mode R) (typ_objref_loc typing_mode R)) by eauto using typ_objref_loc_corr.
    pose proof (ctx_binds_e y v_loc T G) as HbyGy.
    pose proof (ctx_binds_e w v_ref T G) as HbwGw.
    inversions Htyd.
    econstructor; eauto.
    * eapply inert_G_for_obj_loc_inert; eauto.
    * 
      destruct (mode_refcaps' typing_mode).
      all: eapply heap_correspond_push_obj; eauto;
        replace (fv d) with (fv_defs d) by eauto; simpl; eauto.
      all: hnf; hnf in Htft; auto.
    * eapply tred_lit_obj_ty_trm; eauto.
    * eapply all_tfree_k_let_obj; eauto.
    * (* Freshness *)
      eapply tred_ctx_dom_lit_obj; eauto.
    * eapply null_p; eauto.
    * eapply tred_sealed_lit_obj; eauto.
    * eapply closed_let_lit_obj; eauto.
  - (* Case tred_let_fun *)
    rename F1 into G, H1 into Htyt2,
    H2 into Hhc, H0 into Htyp.
    inversion Htyp; subst.
    rename H4 into Htybod.
    assert (HfryG : y \notin ctx_dom G).
    { eapply heap_correspond_notin_dom; eauto 2. }
    assert (HfrwG : w \notin ctx_dom G).
    { eapply heap_correspond_notin_dom; eauto 2. }
    assert (HfrwGy : w §# G § y / v_loc ~ typ_fun T0 U).
    { simpl; eauto 2. }
    assert (Hcokyw : ctx_ok (G §& ((y / v_loc ~ typ_fun T0 U)
                                     §& (w / v_ref ~ typ_fun T0 U)))).
    { constructor; eauto. }
    assert (Hcoky : ctx_ok (G §& (y / v_loc ~ typ_fun T0 U))).
    { constructor; eauto. }
    assert (Heok : ok e) by eauto.
    econstructor; eauto.
    + eapply inert_G_for_item_fun_inert; eauto. simpls; eauto.
    + eapply tred_lit_fun_hc; eauto.
      fresh_solve; sympls; eauto.
    + eapply tred_lit_fun_trm; eauto.
    + eapply tred_lit_fun_tfree_k_trm; eauto.
    + inversions Htyc.
      eapply ty_nothing_fresh_in_config_push2; auto.
    + inversions Htyc. eapply null_p; eauto.
    + eapply tred_lit_fun_sealed; eauto.
    + eapply tred_lit_fun_closed; eauto.
    + inversions Htyc. auto.
  - (* Case tred_let_push *)
    rename F2 into G.
    invert_ty_config.
    invert_focus_typing Hty.
    assert (Hok : ctx_ok G) by eauto.
    assert (HGecor : renv_corr G e) by eauto.
    econstructor; eauto.
    * eapply all_tfree_k_let_push; eauto.
    * hnf. eauto.
    * eapply ty_nothing_fresh_in_config_p; eauto.
    * eapply null_p; eauto.
    * eapply closed_let_push; eauto.
  - (* Case tred_let_loc *)
    rename F2 into G.
    invert_ty_config.
    invert_focus_typing Hty.
    inversions Htys.
    rename H6 into Hyts, H8 into Hop.
    hnf in Htfs. econstructor; eauto.
    * hnf. eauto using tfree_stack_cons_r.
    * eapply @renaming_fresh; eauto.
      apply CloseUnboundSubstIntro. apply ewkn_ty_trm.
    * hnf. eapply all_tfree_k_open_trm; eauto.
      hnf. eauto using tfree_stack_cons_l.
    * eapply ty_nothing_fresh_in_config_p; eauto.
    * eapply null_p; eauto.
    * eapply closed_let_loc; eauto.
Qed.

Corollary typed_preservation_closure
  (F F' : ctx) (C C' : config) (T : typ) (n : nat):
    ty_config F C T ->
    F ⊢ C ∶ T ⊣n↦* F' ⊢ C' ->
    ty_config F' C' T.
Proof.
  introv Htyc Hstep. induction Hstep; eauto using typed_preservation.
Qed.

End WithTypingMode.
