Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping
        InertTightSubtyping
        GeneralToTight TightTyping InvertibleTyping
        RecordTypes RecordDefs InertContext PreciseTyping
        OperationalSemantics Substitution Renaming Weakening
        HeapCorrespondence HeapUpdate CanonicalForms
         Narrowing
        EnvCorrespondence EWeakening Visibility
        ConfigTyping ConfigVarkinds TFree Deref ReferenceTypes
        TypedOperationalSemantics Defree Defs Closedness FreshChain FreshChainTactics.
Require Import TypedReductionEquivalence NothingFresh.

(** * Helper lemmas for preservation for typed reduction *)

Section WithTypingMode.
Context (typing_mode: Mode).

Local Hint Unfold record_type: core.

(* If a location type is inert, then the corresponding reference type is inert *)
Lemma inert_typ_corr_inert Tloc Tref:
  inert_typ Tloc v_loc ->
  typ_corr Tref Tloc ->
  inert_typ Tref v_ref.
Proof.
  introv Hinl Hcor.
  inversions Hinl.
  - inversions Hcor. auto.
  - inversions Hcor.
      destruct (mode_refcap_corr' typing_mode) eqn: Hm.
    + subst.
      destruct typing_mode; unfold mode_refcap_corr' in *; try discriminate; eauto.
    + (* rodot *)
      subst.
      destructc H4 as ? ? ?.
      subst.

      constructor; eauto. unfold refcap_inert_type.
      destruct (mode_refcaps' typing_mode) eqn: Hmr;
      try typing_mode_contradiction typing_mode.
      constructor.
Qed.

Lemma ctx_vis_hide_fv G:
  ctx_fv (vis_hide typing_mode G) = ctx_fv G.
Proof.
  destruct typing_mode; auto.
Qed.


Lemma closed_typ_objref_loc x R:
  closed (open x R) ->
  closed (typ_objref_loc typing_mode R).
Proof.
  introv HclR.
  destruct typing_mode; intro; intro; simpls;   eapply closed_open_down in HclR; rewrite HclR; auto.
Qed.

Lemma typing_viewpoint_adapt_same G e (y: var) w T U:
  mode_viewpoint_adaptation' typing_mode = mode_viewpoint_adaptation_same ->
  G ⁏ e ⊢ w ▷ T ⟶ U ->
  G ⁏ e ⊢ y ∶ T ->
  G ⁏ e ⊢ y ∶ U.
Proof.
  introv Hm Hva Hty.
  hnf in Hva.
  rewrite Hm in Hva.
  subst. auto.
Qed.

Lemma ty_capbnd_top G e (y: var) w T:
mode_has_mutability typing_mode ->
  G ⁏ e ⊢ y ∶ T ->
  G ⁏ e ⊢ y ∶ typ_cap_adapt typ_refcap_mutability w typ_top.
Proof.
  introv Hm Hty.
  eapply ty_var_sub.
  eapply ty_var_mut_top; eauto.
  eapply subtyp_typ. eauto.
  intros.
  sympl.
  eapply subtyp_or11.
Qed.

Lemma typ_objref_loc_inert G e R d T0:
  G ⁏ e ⊢ lit_obj R d ∶ T0 ->
  ctx_ok G ->
  inert_typ (typ_objref_loc typing_mode R) v_loc.
Proof.
  introv Htyd HokG.
  destruct (mode_refcaps' typing_mode) eqn: Hm; eapply inert_obj_typ; eauto using refcap_type_top;
  hnf; destruct (mode_refcaps' typing_mode) eqn: Hmrc; try typing_mode_contradiction typing_mode; constructor.
Qed.
Lemma typ_objref_ref_inert G e R d T0:
  G ⁏ e ⊢ lit_obj R d ∶ T0 ->
  ctx_ok G ->
  inert_typ (typ_objref_loc typing_mode R) v_ref.
Proof.
  introv Htyd HokG.
  destruct (mode_refcaps' typing_mode) eqn: Hm; eapply inert_obj_typ; eauto using refcap_type_top;
  hnf; destruct (mode_refcaps' typing_mode) eqn: Hmrc; try typing_mode_contradiction typing_mode; constructor.
Qed.
Lemma typ_objref_loc_corr R:
  typ_corr (typ_objref_loc typing_mode R) (typ_objref_loc typing_mode R).
Proof.
  destruct ( mode_refcaps' typing_mode) eqn: Hmrc;
  constructor; destruct ( mode_refcap_corr' typing_mode) eqn: Hmcor;
  try typing_mode_contradiction typing_mode; eauto.
Qed.


Lemma closed_let_loc t fs w Sigma e Lva Lvb:
  config_closed (trm_var (avar_f w) ; stack_cons (frame_let t) fs ; Sigma ; e ; Lva) ->
  config_closed (t /[-> w] ; fs ; Sigma ; e ; Lvb).
Proof.
  introv Hclc.
  inversions Hclc. pose proof stack_closed_t H5. eapply stack_closed_fs in H5.
  constructor; eauto.
Qed.

Lemma closed_let_lit_obj t fs w y R d Sigma e Lva Lvb:
  config_closed (trm_lit (lit_obj R d) t ; fs ; Sigma ; e ; Lva) ->
  config_closed (t /[-> w] ; fs ; Sigma & y ~ item_obj (deref_defs (e & w ~ y) d /[-> w]) ; e & w ~ y ; Lvb).
Proof.
  introv Hclc.
  inversions Hclc. constructor; eauto.
      eapply closed_trm_lit_obj_r; eauto.
      constructor; eauto. eapply closed_item_obj'.
      eapply closed_deref_defs. eapply closed_trm_lit_obj_l; eauto.
Qed.
Lemma closed_apply t' fs x z Sigma e Lva Lvb y:
  binds y (item_fun t') Sigma ->
  config_closed (trm_apply (avar_f x) (avar_f z) ; fs ; Sigma ; e ;  Lva) ->
  config_closed (t' /[-> z] ; fs ; Sigma ; e ; Lvb).
Proof.
  introv Hbi Hclc.
  inversions Hclc. constructor; eauto.
  eapply heap_binds_fun_closed; eauto.
Qed.

Lemma closed_apply' t' fs x z Sigma Sigma' e Lva Lvb Lvc y:
  binds y (item_fun t') Sigma' ->
  config_closed (trm_apply (avar_f x) (avar_f z) ; fs ; Sigma' ; e ;  Lva) ->
  config_closed (trm_apply (avar_f x) (avar_f z) ; fs ; Sigma ; e ;  Lvc) ->
  config_closed (t' /[-> z] ; fs ; Sigma ; e ; Lvb).
Proof.
  introv Hbi Hclc' Hclc.
  inversions Hclc.
  inversions Hclc'. constructor; eauto.
  eapply heap_binds_fun_closed; eauto.
Qed.


Lemma closed_read w1 w2 y2 a fs Sigma e Lva Lvb:
  config_closed (trm_read (avar_f w1) a ; fs ; Sigma ; e ; Lva) ->
  config_closed (trm_var (avar_f w2) ; fs ; Sigma ; e & w2 ~ y2 ; Lvb).
Proof.
  introv Hclc.
  inversions Hclc. constructor; eauto. eapply closed_trm_avar_f; eauto.
Qed.

Lemma closed_let_push t1 t2 fs Sigma e Lva Lvb:
  config_closed (trm_let t1 t2 ; fs ; Sigma ; e ; Lva) ->
  config_closed (t1 ; stack_cons (frame_let t2) fs ; Sigma ; e ; Lvb).
Proof.
  introv Hclc. 
  inversions Hclc.
  constructor; eauto. eapply closed_trm_let_l; eauto. eapply stack_closed_let; eauto. intros. eapply closed_trm_let_r; eauto.
Qed.

Lemma closed_write w1 w3 a y1 y3 fs Sigma Sigma2 e Lva Lvb:
  heap_update Sigma y1 a y3 Sigma2 ->
  config_closed (trm_write (avar_f w1) a (avar_f w3) ; fs ; Sigma ; e ; Lva) ->
  config_closed (trm_var (avar_f w3) ; fs ; Sigma2 ; e ; Lvb).
Proof.
  introv Hu Hclc.
  inversions Hclc. constructor; eauto. eapply closed_trm_avar_f; eauto.
 eapply heap_closed_update; eauto.
Qed.

Lemma closed_call t w1 w2 a yd y fs Sigma e Lva Lvb:
  binds y (item_obj yd) Sigma ->
  defs_has yd (def_met a t) ->
  config_closed (trm_call (avar_f w1) a (avar_f w2) ; fs ; Sigma ; e ; Lva) ->
  config_closed (t /[# 1 -> w1] /[-> w2] ; fs ; Sigma ; e ; Lvb).
Proof.
  introv Hbi Hhas Hclc.
  inversions Hclc. constructor; eauto. eapply heap_binds_met_closed; eauto.
Qed.
Lemma closed_call' t w1 w2 a yd y fs Sigma Sigma' e Lva Lvb Lvc:
  binds y (item_obj yd) Sigma' ->
  defs_has yd (def_met a t) ->
  config_closed (trm_call (avar_f w1) a (avar_f w2) ; fs ; Sigma' ; e ; Lva) ->
  config_closed (trm_call (avar_f w1) a (avar_f w2) ; fs ; Sigma ; e ; Lvc) ->
  config_closed (t /[# 1 -> w1] /[-> w2] ; fs ; Sigma ; e ; Lvb).
Proof.
  introv Hbi Hhas Hclc' Hclc.
  inversions Hclc. 
  inversions Hclc'. constructor; eauto. eapply heap_binds_met_closed; eauto.
Qed.


Lemma tred_sealed_read G w2 T2:
  sealed G ->
  closed T2 ->
  sealed (G §ʷ w2 ~ T2).
Proof.
  introv HslG HclT2.
  eapply ctx_sealed_push; eauto. eapply closed_sealed0; eauto.
Qed.

Lemma tred_sealed_lit_obj G w y R d t fs Sigma e Lv:
  sealed G ->
  config_closed (trm_lit (lit_obj R d) t ; fs ; Sigma ; e ; Lv) ->
  sealed (inert_G_for_item G w y (typ_objref_loc typing_mode R)).
Proof.
  introv HslG Hclc.

      assert (closed (typ_objref_loc typing_mode R)). {
        inversions Hclc.
        pick_fresh z.
        eapply closed_trm_lit_obj_t with (w:=z) in H2.
        eapply closed_typ_objref_loc. eapply H2.
      }
      eapply ctx_sealed_push.
      eapply ctx_sealed_push. eauto.
      eapply closed_sealed0; eauto.
      eapply closed_sealed0; eauto.
Qed.


Lemma all_tfree_k_let_push G fs t1 t2:
  all_tfree_k_stack G fs v_ref ->
  all_tfree_k_trm G (trm_let t1 t2) v_ref ->
  all_tfree_k_stack G (stack_cons (frame_let t2) fs) v_ref.
Proof.
  introv Htfs Htft.
  hnf. introv Htf. hnf in Htf. inversions Htf; eauto.
Qed.

Lemma all_tfree_k_call G e w1 m w2 T0 t yd Si y:
  G ⁏ e ⊢ trm_call (avar_f w1) m (avar_f w2) ∶ T0 ->
  all_tfree_k_trm G (trm_call (avar_f w1) m (avar_f w2)) v_ref ->
  binds w1 y e ->
  binds y (item_obj yd) Si ->
  defs_has yd (def_met m t) ->
  heap_correspond G e Si ->
  sealed G ->
  inert G ->
  ok e ->
  all_tfree_k_trm G t /[# 1 -> w1] /[-> w2] v_ref.
Proof.
  introv Hty Htff.
  introv Hbw1ye HbyS Hhas Hhc HclG Hin Hoke.
    (* Everything not in type-position are references *)
      invert_focus_typing Hty.

      pose proof (canonical_forms_met_ref Hin HclG Hhc Hbw1ye H1) as Hcf.
      destruct Hcf as [L [yd' [t' [S0 [S1 [S2 [HbniG [HblG [Hbie [Hhas' [Hsu [Htf' Hop]]]]]]]]]]]].
      binds_eq.

      apply defs_has_met_inv with (t := t') in Hhas; auto.
      inversions Hhas; eauto; clear H.
      rename H2 into Htw2, H1 into Htw1m, H3 into Htw1c.
      eapply all_tfree_k_open_trm; eauto.
Qed.

Lemma all_tfree_k_let_obj G w y R t d:
  all_tfree_k_trm G (trm_lit (lit_obj R d) t) v_ref ->
  fresh (ctx_dom G) 2 (w::y::nil) ->
  all_tfree_k_trm (inert_G_for_obj_loc typing_mode G w y R) t /[-> w] v_ref.
Proof. 
  introv Htft Hfr.
  apply all_tfree_k_open_trm; try solve [ hnf; simpl; eauto ].
  apply all_tfree_k_Gpush; auto. apply all_tfree_k_Gpush; auto.
  hnf. eauto.
Qed.
Lemma all_tfree_k_let_obj_defs G w y R t d:
  all_tfree_k_trm G (trm_lit (lit_obj R d) t) v_ref ->
  fresh (ctx_dom G) 2 (w::y::nil) ->
  all_tfree_k_defs (inert_G_for_obj_loc typing_mode G w y R) d  v_ref.
Proof. 
  introv Htft Hfr.
  apply all_tfree_k_Gpush; auto. apply all_tfree_k_Gpush; auto.
  hnf. eauto.
Qed.




Lemma all_tfree_k_let_fun G w  S T:
  all_tfree_k_trm (G §ʷ w ~ typ_fun S T) (trm_var (avar_f w)) v_ref.
Proof. 

(* Everything not in type-position are references *)
      hnf. introv Htf. inversions Htf. hnf. exists (typ_fun S T) G. eauto.
Qed.



Lemma inert_G_for_obj_loc_inert G w y R:
  inert G ->
  fresh (ctx_dom G \u ctx_fv G \u fv R ) 2 (y :: w:: nil) ->
  inert_typ (typ_objref_loc typing_mode R) v_loc ->
  inert_typ (typ_objref_loc typing_mode R) v_ref ->
  inert (inert_G_for_obj_loc typing_mode G w y R).
Proof.
  introv Hin Hfr HinL HinR.
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hfr.
  fresh_chain_cleanup Fr.

  destruct (mode_refcaps' typing_mode).
  all: constructor; auto.
  constructor; auto.
  all:  sympl; try  fresh_chain_solve Fr; try  fresh_chain_solve2 Fr. 
  constructor; auto.
  all:  sympl; try  fresh_chain_solve Fr; try  fresh_chain_solve2 Fr. 
Qed.

Lemma inert_G_for_item_fun_inert G w y T U:
  mode_has_fun typing_mode ->
  inert G ->
  fresh ( ctx_fv G \u ctx_dom G \u fv (typ_fun T U)) 2 (w::y::nil) ->
  inert (inert_G_for_item G w y (typ_fun T U)).
Proof.
  introv Hm HinG Hfr.
  assert (Hfrw: w \notin ctx_fv G \u fv (typ_fun T U)) by ( simpl; notin_solve).
  assert (Hfry: y \notin ctx_fv G \u fv (typ_fun T U)) by ( simpl; notin_solve).
  constructor; auto. simpl. simpl in Hfrw. auto.
Qed.


Lemma tred_ctx_dom_lit_obj G w y R t fs Sigma e d Lv (T: typ):
  (ctx_dom G \c fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv /\
       nothing_fresh_in_config G (trm_lit (lit_obj R d) t ; fs ; Sigma ; e ; Lv)) /\
      nothing_fresh_in_config T (trm_lit (lit_obj R d) t ; fs ; Sigma ; e ; Lv) ->
(ctx_dom (inert_G_for_obj_loc typing_mode G w y R) \c
 fv t /[-> w] \u
 fv fs \u
 fv_heap (Sigma & y ~ item_obj (deref_defs (e & w ~ y) d /[-> w])) \u
 fv_renv (e & w ~ y) \u fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv /\
 nothing_fresh_in_config (inert_G_for_obj_loc typing_mode G w y R)
   (t /[-> w] ; fs ; Sigma & y ~ item_obj (deref_defs (e & w ~ y) d /[-> w]) ; e & w ~ y
    ; fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv)) /\
nothing_fresh_in_config T
  (t /[-> w] ; fs ; Sigma & y ~ item_obj (deref_defs (e & w ~ y) d /[-> w]) ; e & w ~ y
   ; fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv).
Proof.
introv Hfv.
    destruct (mode_refcaps' typing_mode) eqn: Hmrc;
      destruct (mode_objlit_refcaps' typing_mode) eqn: Hmol;
      try typing_mode_contradiction typing_mode.

        --
       eapply ty_nothing_fresh_in_config_push2; auto.
       unfold nothing_fresh_in_config.
       sympls. union_sub_solver.
        --
       eapply ty_nothing_fresh_in_config_push2; auto.
       unfold nothing_fresh_in_config.
       sympls. union_sub_solver.
Qed.

Lemma tred_lit_obj_ty_trm G w y R e t U L:
  ctx_ok G ->
  ok e ->
(  forall x : var,
     x \notin L ->
     G §ˣ x ~ typ_objlit typing_mode R
 ⁏ e ⊢ t /[-> x] ∶ U) ->
  fresh (ctx_dom G \u dom e) 2 (w::y::nil) ->
  inert_G_for_obj_loc typing_mode G w y R ⁏ e & w ~ y ⊢ t /[-> w] ∶ U.
Proof.
  introv HokG Hoke Htyt Hfr.
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hfr.
  fresh_chain_cleanup Fr.
  eapply @renaming_fresh; eauto.
      apply CloseUnboundSubstIntro.
      apply ewkn_ty_trm. fresh_chain_solve Fr. fresh_chain_solve2 Fr.
      destruct (mode_refcaps' typing_mode) eqn: Hmrc;
      destruct (mode_objlit_refcaps' typing_mode) eqn: Hmol;
      try typing_mode_contradiction typing_mode.
      all: intros.
      all: ctx_split_between y z.
      all: assert (Weakening (ty_trm basic)); eauto.
      all: apply weaken_ty; simpl; try fresh_chain_solve2 Fr; eauto.
      (* Needs to change after adding mutability *)
      all: eapply ewkn_push;  try fresh_chain_solve2 Fr; eauto.
Qed.

Lemma tred_ctx_dom_lit_fun G w2 S T1 e y2 w1 a fs Sigma Lv (T: typ) G2:
  ctx_binds_k_after y2 v_loc (typ_fun S T1) G G2 ->
  (ctx_dom G \c fv (trm_read (avar_f w1) a) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv /\
  nothing_fresh_in_config G (trm_read (avar_f w1) a ; fs ; Sigma ; e ; Lv)) /\ nothing_fresh_in_config T (trm_read (avar_f w1) a ; fs ; Sigma ; e ; Lv) ->
  (ctx_dom (G §ʷ w2 ~ typ_fun S T1) \c
  fv (trm_var (avar_f w2)) \u fv fs \u fv_heap Sigma \u fv_renv (e & w2 ~ y2) \u fv (trm_read (avar_f w1) a) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv /\
  nothing_fresh_in_config (G §ʷ w2 ~ typ_fun S T1)
  (trm_var (avar_f w2) ; fs ; Sigma ; e & w2 ~ y2 ; fv (trm_read (avar_f w1) a) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv)) /\
  nothing_fresh_in_config T (trm_var (avar_f w2) ; fs ; Sigma ; e & w2 ~ y2 ; fv (trm_read (avar_f w1) a) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv).
Proof.
  introv Hbk Hfv.
  destructc Hfv as [? ?] ?. eapply ty_nothing_fresh_in_config_push; eauto.  eapply nothing_fresh_in_config_ctx_binds. eauto. eapply ctx_binds_k_after_binds; eauto.
Qed.

Lemma tred_lit_fun_ty_trm G w y e S T Sf Tf wf a U yf Sigma yd:
  mode_has_fun' typing_mode = true ->
  G ⁏ e ⊢ wf ▷ Tf ⟶ U ->
  G ⁏ e ⊢ wf ∶ typ_rcd (dec_fld a Sf Tf) ->
  binds wf yf e ->
  heap_correspond G e Sigma ->
  inert G ->
  sealed G ->
  binds yf (item_obj yd) Sigma ->
  inert (G §ʷ w ~ typ_fun S T) ->
  defs_has yd (def_fld a (avar_f y)) ->
  w # e ->
  w §# G ->
  renv_corr G e->
  heap_correspond (G §ʷ w ~ typ_fun S T) (e & w ~ y) Sigma ->
  G §ʷ w ~ typ_fun S T ⁏ e & w ~ y ⊢ trm_var (avar_f w) ∶ U.
Proof.
  introv Hm.
  introv Hva Htyf Hbwye Hhc Hin HclG Hby1S.
  introv HinGw2 Hhas Hfrw2e Hfrw2G Hcor Hhc2.

 eapply ty_trm_var.
      + intro. eauto.
      +

      apply ty_renv_binds_rev_var with (y := y); eauto.
      destruct (mode_refcap_corr' typing_mode) eqn: Hmcor;
        try typing_mode_contradiction typing_mode; auto.
      hnf in Hva.
      destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hvar; try typing_mode_contradiction typing_mode.
      subst.

      pose proof (canonical_forms_fld_ref Hin HclG Hhc Hbwye Htyf) as Hcf.
      destructc Hcf as ds' y2' Hby1S' Hhas' Htypy2 Hst Hy2loc.
      binds_eq. rename ds' into ds.
      apply defs_has_inv with (t := avar_f y) in Hhas'; eauto;
        inversions Hhas'; rename y2' into y.
        
        eapply (@weaken_push) with (P:=(ty_var _)); eauto.
        eapply (@ewkn_push) with (P:= (ty_var _)); eauto.
Qed.

Lemma tred_lit_fun_hc G w y T0 U e Sigma (t1 t2: trm) L0:
  mode_has_fun typing_mode ->
  heap_correspond G e Sigma ->
  (forall x : var, x \notin L0 -> G §ᶻ x ~ T0 ⁏ e ⊢ t1 /[-> x] ∶ U /[-> x]) ->
  fresh (ctx_dom G \u ctx_fv G  \u fv (typ_fun T0 U) \u fv_env Sigma \u dom e \u fv t1 ) 2 (w::y::nil) ->
  all_tfree_k_trm G (trm_lit (lit_fun T0 t1) t2) v_ref ->
  heap_correspond (inert_G_for_item G w y (typ_fun T0 U)) (e & w ~ y) (Sigma & y ~ item_fun t1).
Proof.
  introv Hm.
  introv Hhc Htybod Hfr.
  introv Htft.
  
  assert (ctx_ok (G §& (ctx_empty §ʸ y ~ typ_fun T0 U))) as Hcoky.
  ctx_flatten. eauto.

  eapply heap_correspond_epush; eauto 2.
  * eapply heap_correspond_push; eauto 2.
    eapply ty_item_fun_s with (L := L0 \u \{y}); simpl; eauto 2.
    { introv Hfrx.
      assert (HfrxL0 : x \notin L0) by eauto;
        specialize (Htybod _ HfrxL0); clear HfrxL0.
      epose proof (weaken_ty_trm G (x / v_par ~ T0)) as Hwk.
      assert (Hcwk : ctx_weakens (y / v_loc ~ typ_fun T0 U))
        by (simpl; eauto).
      specialize (Hwk objctx_empty e (open x t1) (open x U)).
      specialize (Hwk Htybod Hcwk Hcoky).
      simpl in Hwk. auto. }
    { apply all_tfree_k_Gpush; auto. hnf in Htft. hnf. eauto. }
  * rewrite fv_env_push, notin_union.
    split; eauto. sympl. eauto.
  * constructor. eauto.
Qed.

Lemma tred_lit_fun_trm G w y T0 U e t2 A L:
( forall x : var, x \notin L -> G §ˣ x ~ typ_fun T0 U ⁏ e ⊢ t2 /[-> x] ∶ A) ->
  ctx_ok G ->
   w §# G §ʸ y ~ typ_fun T0 U ->
  y §# G ->
  ok e ->
  w # e ->
  y # e ->
  inert_G_for_item G w y (typ_fun T0 U) ⁏ e & w ~ y ⊢ t2 /[-> w] ∶ A.
Proof.
  introv Htyt2.
  introv HokG HfrwG HfryG Hoke Hfrwe Hfrye.
      eapply @renaming_fresh; auto.
      apply CloseUnboundSubstIntro. apply ewkn_ty_trm. 
      intros.
      ctx_split_between y z.
      apply weaken_ty; auto.
      autorewrite with ctx_flatten.
      eapply ewkn_push; eauto.
      ctx_flatten; eauto.
Qed.

Lemma tred_lit_fun_tfree_k_trm G w y T0 U t2 t1:
  y §# G ->
  w §# G §ʸ y ~ typ_fun T0 U ->
  all_tfree_k_trm G (trm_lit (lit_fun T0 t1) t2) v_ref ->
  all_tfree_k_trm (inert_G_for_item G w y (typ_fun T0 U)) t2 /[-> w] v_ref.
Proof.
  introv HfryG HfrwGy Htft.
  apply all_tfree_k_open_trm; try solve [ hnf; simpl; eauto ].
        apply all_tfree_k_Gpush; auto. apply all_tfree_k_Gpush; auto.
        hnf. hnf in Htft. eauto.
Qed.

Lemma tred_lit_fun_sealed G w y T0 U T t2 t1 fs Sigma e Lv:
  ty_config G (trm_lit (lit_fun T0 t1) t2 ; fs ; Sigma ; e ; Lv) T ->
  closed (typ_fun T0 U) ->
  sealed (inert_G_for_item G w y (typ_fun T0 U)).
Proof.
  introv Htyc HclT2.
inversions Htyc.
      eapply ctx_sealed_push.
      eapply ctx_sealed_push.
      eauto.       eapply closed_sealed0; eauto.      eapply closed_sealed0; eauto.
Qed.

Lemma tred_lit_fun_closed G w y T0 T t2 t1 fs Sigma e Lv:
  ty_config G (trm_lit (lit_fun T0 t1) t2 ; fs ; Sigma ; e ; Lv) T ->
  config_closed (t2 /[-> w] ; fs ; Sigma & y ~ item_fun t1 ; e & w ~ y ; fv (trm_lit (lit_fun T0 t1) t2) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv).
Proof.
  introv Htyc.
  inversions Htyc. inversions H15. constructor; eauto.
      eapply closed_trm_lit_fun_t2. eauto.
      constructor; eauto.
      pick_fresh z. assert (closed (open z t1)).
      eapply closed_trm_lit_fun_t1. eauto.
      eapply closed_item_fun'; eauto.
Qed.

Lemma config_closed_Lv   t  fs  Si  e Lv1 Lv2:
  config_closed (t ; fs ; Si ; e ; Lv1) ->
  config_closed (t ; fs ; Si ; e ; Lv2).
Proof.
  introv Hcl.
  inversions Hcl.
  econstructor; eauto.
Qed.

End WithTypingMode.
