Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping
        InvertibleTyping InertContext
        OperationalSemantics Weakening
        HeapCorrespondence HeapUpdate CanonicalForms
        
        EnvCorrespondence Visibility
        ConfigTyping TFree Deref ReferenceTypes
        TypedOperationalSemantics Defree Closedness.
Require Import TypedReductionEquivalence ConfigNotations.

(** * Progress lemma for typed reduction *)

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma nothing_fresh_in_config_read U1 x TcapN a fs Sigma e Lv (S0: typ) (G: ctx) x2 (U: typ) (null: var):
  ctx_dom G \c \{ x} \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv ->
  nothing_fresh_in_config G (trm_read (avar_f x) a ; fs ; Sigma ; e ; Lv) ->
  nothing_fresh_in_config U (trm_read (avar_f x) a ; fs ; Sigma ; e ; Lv) ->
  null \in Lv ->
  ctx_binds x2 (typ_and (typ_bnd U1) S0) G ->
  fv TcapN \c ctx_dom G \u ctx_fv G \u dom e \u fv_env e \u fv U \u fv fs \u \{ null} ->
  nothing_fresh_in_config (typ_and (typ_bnd U1) (typ_cap_adapt typ_refcap_mutability x TcapN)) (trm_read (avar_f x) a ; fs ; Sigma ; e ; Lv).
Proof.
  introv HnfrGd HnfrGf HnfrU Hnull HbGx2' HfvN.
        unfold nothing_fresh_in_config.

        assert (fv U1 \c fv (typ_bnd U1) ) as HfvU1. {
          intro. intro Hx1. compute. compute in Hx1. auto.
        }
        assert (fv (typ_bnd U1)  \c fv (typ_and (typ_bnd U1) S0) ) as Hfvbnd. {
          intro. intro Hx1. compute. compute in Hx1. rewrite in_union. auto.
        }
        assert (fv (typ_and (typ_bnd U1) S0) \c fv G) as Hfvand. {
          eapply ctx_binds_fv_subset; eauto.
        }
        assert (fv (typ_bnd U1)  \c fv G) as HfvbndG. {
          intro. intro. auto.
        }

        assert (fv (typ_and (typ_bnd U1) (typ_cap_adapt typ_refcap_mutability x TcapN)) \c fv (typ_bnd U1) \u fv (typ_cap_adapt typ_refcap_mutability x TcapN)) as Hfvandadapt.
         intro. intro. compute. compute in H. auto.

        assert (fv (typ_cap_adapt typ_refcap_mutability x TcapN) \c fv TcapN \u \{ x } ) as H12b.

          intro. intro Hx1. compute. compute in Hx1. rewrite union_empty_l in Hx1. rewrite union_empty_r in Hx1. auto.

        intro. intro Hfvx1. apply Hfvandadapt in Hfvx1. rewrite in_union in Hfvx1. destruct Hfvx1 as [Hfvx1 | Hfvx1]. apply HfvbndG in Hfvx1. apply HnfrGf in Hfvx1. auto.
        apply H12b in Hfvx1. rewrite in_union in Hfvx1. destruct Hfvx1 as [Hfvx1 | Hfvx1]. apply HfvN in Hfvx1.
        rewrite in_union in Hfvx1.
        destruct Hfvx1 as [Hfvx1 | Hfvx1].
        apply HnfrGd in Hfvx1. auto.
        rewrite in_union in Hfvx1.
        destruct Hfvx1 as [Hfvx1 | Hfvx1].
        apply HnfrGf in Hfvx1. auto.
        rewrite in_union in Hfvx1.
        destruct Hfvx1 as [Hfvx1 | Hfvx1].
        repeat rewrite in_union; right; auto.
        rewrite in_union in Hfvx1.
        destruct Hfvx1 as [Hfvx1 | Hfvx1].
        repeat rewrite in_union; right; auto.
        rewrite in_union in Hfvx1.
        destruct Hfvx1 as [Hfvx1 | Hfvx1].
        apply HnfrU in Hfvx1. auto.
        repeat rewrite in_union; right; auto.
        rewrite in_union in Hfvx1.
        destruct Hfvx1 as [Hfvx1 | Hfvx1]. auto.
        rewrite in_singleton in Hfvx1. subst. right. right. auto.
        rewrite in_singleton in Hfvx1. subst.
   rewrite in_union. left. compute. rewrite in_singleton. auto.
Qed.

(* Proof of Lemma 1 in "Reference mutability for DOT" *)
Lemma typed_progress_read G x a fs Sigma e Lv U U0 S T :
  ty_config G (trm_read (avar_f x) a ; fs ; Sigma ; e ; Lv) U ->
  inert G ->
  heap_correspond G e Sigma ->
  ty_stack G e U0 fs U ->
  all_tfree_k_stack G fs v_ref ->
  (mode_visibility' typing_mode = mode_visibility_hide_method -> ctx_vis G x) ->
  G ⁏ e ⊢ x ∶ typ_rcd (dec_fld a S T) ->
  G ⁏ e ⊢ x ▷ T ⟶ U0 ->
  all_tfree_k_trm G (trm_read (avar_f x) a) v_ref ->
  (ctx_dom G \c fv (trm_read (avar_f x) a) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv /\
       nothing_fresh_in_config G (trm_read (avar_f x) a ; fs ; Sigma ; e ; Lv)) /\ nothing_fresh_in_config U (trm_read (avar_f x) a ; fs ; Sigma ; e ; Lv) ->
  has_null Lv ->
  sealed G ->
  config_closed (trm_read (avar_f x) a ; fs ; Sigma ; e ; Lv) ->
  closed U ->
  (exists F' c', G ⊢ trm_read (avar_f x) a ; fs ; Sigma ; e ; Lv ∶ U ↦ F' ⊢ c').
Proof.
  introv Htc Hin Hcor Hts Hvks Hvis Htyx Hva Hvkt Hnf.
  introv Hnfr HclG Hclc HclU.

(* T is field type from term typing *)
    (* U0 is viewpoint-adapted field type from term typing *)
    (* Assertions *)
    assert (ctx_k G x v_ref) as Hvk by auto.
    pick_fresh_with w (fv fs).
    assert (w # Sigma) by auto.
    assert (w # e) by auto.
    pose proof (canonical_forms_fld Hin HclG Hcor Hvk Htyx).
    (* Proving case *)
     destruct_all.
    rename x3 into null, H2 into Hnull.
    rename x1 into ds, H1 into Hbxe, H3 into Hbx0S, H4 into Hhas,
    H5 into Htypx2, H6 into Hst, H7 into Hctxk.
    clear Htypx2 Hst.
    (* x2 is the value of the field on the heap *)
    pose proof Hctxk as Hctx2k.
    hnf in Hctx2k. destruct Hctx2k as [T2 [G2 Hbkaf]].
    apply ctx_binds_k_after_binds in Hbkaf as HbGx2'.
    apply ctx_k_in_kdom in Hctxk.
    assert (Hdomeq : ctx_kdom v_loc G = dom Sigma)
      by (unfold heap_correspond in Hcor; destruct_all; eauto);
      rewrite Hdomeq in Hctxk; clear Hdomeq.
    apply dom_to_binds in Hctxk as HbSx2.
    assert (Htysx2 : ty_item_s _ G e Sigma x2)
      by (unfold heap_correspond in Hcor; destruct_all; eauto).
    clear HbSx2.
    (* Proving existence of step *)
    inversions Htysx2.
    + rename H2 into HbGx2,
      H3 into HbSx2, H4 into Hoptyp, H5 into Htf.
      pose proof (inert_binds_inert Hin HbGx2) as Hint.
      inversions Hint. inversions H1. 
      find_Lv LvC.
      exists (G § w / v_ref ~ typ_fun S0 T0)
        (trm_var (avar_f w) ; fs ; Sigma ; e & w ~ x2 ; LvC).
      eapply tred_read_fun; eauto.
      ctx_binds_eq. eauto.
    + rename H1 into HbGx2, H2 into HbSx2, H4 into Htyds0, H5 into Hdvk.
      ctx_binds_eq; clear H1.
      (* U1 is record type of x2, S0 is refcap type of x2 *)
      destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hm.
      * (* basic mode *)
        find_Lv LvC.
        exists (G § w / v_ref ~ (typ_objref U1 S0))
          (trm_var (avar_f w) ; fs ; Sigma ; e & w ~ x2 ; LvC).

        eapply tred_read_obj with (R1 := U1); rewrite ? Hm;
        eauto using typ_corr_obj.
        { apply ctx_binds_fv_subset in HbGx2'.
          rewrite <- notin_union, union_comm, union_subset_r; auto. }
        eapply nothing_fresh_in_config_ctx_binds; eauto.
        eapply ctx_closed_binds_closed; eauto.
      * (* rodot mode *)
        subst. hnf in Hva. rewrite Hm in *.
        find_Lv LvC.

        rename Htyx into Htyfld, Hts into Hs.
        assert (ok e) as Hoke by eauto.
        assert (ctx_ok G) as HokG by eauto.
        assert (ctx_weakens G) as HwkG by eauto. rename H10 into HnfrGf, H8 into HnfrGd, H9 into HnfrU.
        rename H0 into Hnwe. rename w into wVal.
        hnf in Hva. (* The type of the new reference is derived from the viewpoint adaptation *)
        inversions Hva.

        pose proof (typing_implies_dom Htyfld) as Hdomx.
        inversions Hclc. clear H5 H9. rename H8 into Hcls.
        destructc (adapt_ref_mutability_nothing_fresh null HokG HwkG HclG Hoke Hdomx Hcls HclU Hs Htyfld H0 H1 ) as TN0 TN TstripN TcapN ? ? Hrest.
        destructc Hrest as HicpN HcpbN HfvN HclN.
        subst.
        exists (G § wVal / v_ref ~ (rodot_read_ref_type U1 typ_sel x TcapN)).
        exists  (trm_var (avar_f wVal) ; fs ; Sigma ; e & wVal ~ x2 ; LvC).
        eapply tred_read_obj with (R1 := U1).
        11: {
          rewrite Hm.
          exists TstripN, TcapN; eauto.
        }
        all: eauto.
        { simpl. repeat(rewrite union_empty_r, union_empty_l).
          apply ctx_binds_fv_subset in HbGx2'.
          rewrite <- notin_union. rewrite union_assoc.
          rewrite union_comm with (E := ctx_fv G).
          erewrite union_subset_r with (F := ctx_fv G);
            eauto using subset_union_l1.
          repeat(rewrite notin_union). repeat(split); auto.
          hnf. introv Hwin. apply HfvN in Hwin.
          repeat(rewrite in_union in Hwin).
          destruct Hwin as [ HwinGd
                           | [ HwinGfv
                             | [ Hwined
                               | [ Hwinefv
                                 | [ HwinU
                                   | [ Hqins | Hwnull ]]]]]].
          all: try solve [ apply Fr; auto
                         | apply Fr; simpl; repeat(rewrite in_union);
                           tauto ].
          rewrite in_singleton in Hwnull; subst.
          apply Fr; auto. repeat(rewrite in_union); tauto. }
       eapply nothing_fresh_in_config_read; eauto.

    ***
  assert(closed (typ_and (typ_bnd U1) S0)). {
    eapply ctx_closed_binds_closed; eauto.
  }
  eapply closed_typ_and. eapply closed_typ_and_l. eauto.
  eapply closed_typ_cap_adapt. eauto.
Qed.

(* Lemma 1 in "Reference mutability for DOT"
   Part of Theorem 6 in "Reference mutability for DOT" *)
Theorem typed_progress : forall F c T,
    ty_config F c T ->
    answer c \/
    exists F' c', F ⊢ c ∶ T ↦ F' ⊢ c'.
Proof.
  introv Htc. rename T into U.
  inversion Htc as [Gamma e fs Sigma Lv t T
                          H Hin Hcor Hts Hvks HT Hvkt].
  subst.
  rename H2 into HclG, H3 into Hclc, H4 into HclU.
  induction HT.

  - Case "ty_trm_var".
    destruct fs; eauto 3. right. exists G.
    destruct f.
    find_Lv LvC.
    exists (open x t ; fs ; Sigma ; e ; LvC).
    eapply tred_let_var; eauto.

  - Case "ty_trm_apply".
    rename H0 into Hnfr.
    rename H4 into Htyx.
    assert (ctx_k G x v_ref) as Hvk by auto.
    pose proof (canonical_forms_fun Hin HclG Hcor Hvk Htyx) as Hcf.
    destructecs Hcf.
    right. destruct_all.
    find_Lv LvC.
    rename Hcf0 into Hbxye, Hcf1 into HbyS, Hcf4 into Hst, Hcf5 into Htf, Hcf6 into Hoptyp.

    exists G (open z t ; fs ; Sigma ; e ; LvC).
    eapply tred_apply; eauto.

  - Case "ty_trm_read".
    rename H into Hvis, H2 into Htyx, H1 into Hnfr.
    rename H3 into Hva, H0 into Hnf.
    right.
    eapply typed_progress_read; eauto.

  - Case "ty_trm_write".
    rename H3 into Htyxm, H4 into Htyx, H5 into Htyy.
    assert (ctx_k G x v_ref) as Hvkx by auto.
    assert (ctx_k G y v_ref) as Hvky by auto.
    pose proof (heap_update_exists Hin HclG Hcor Htyx Htyy Hvkx Hvky) as [?s' ?H].
    destruct_all. right. rename s' into x', x1 into Sigma'.
      find_Lv LvC.
    exists G (trm_var (avar_f y) ; fs ; Sigma' ; e ; LvC).
    eapply tred_write; eauto.

  - Case "ty_trm_call".
    rename H3 into Htyx, H4 into Htyy.
    assert (ctx_k G x v_ref) as Hvkx by auto.
    assert (ctx_k G y v_ref) as Hvky by auto.
    pose proof (canonical_forms_met Hin HclG Hcor Hvkx Htyx).
    destruct_all. right.
    rename x1 into L, x2 into d, x3 into t, H2 into Hbxe, H3 into Hbx0S,
    H4 into Hhas, H5 into Htf, H6 into Hoptyp.
      find_Lv LvC.
    exists G (open_met y x t ; fs ; Sigma ; e ; LvC).
    eapply tred_call; eauto.

  - Case "ty_trm_let".
    right.
      find_Lv LvC.
    exists G (t ; (stack_cons (frame_let u) fs) ; Sigma ; e ; LvC).
    eapply tred_let_push; eauto.

  - Case "ty_llit".
      pose proof H as Htylit.
    destruct l; pick_fresh_with x (fv fs);
      assert (x # Sigma) by auto;
      assert (x # e) by auto;
      inversions H; right; eauto 6.
    + rename t into T.
      pick_fresh_with y ((fv fs) \u \{x}).
      find_Lv LvC.

      destructc H1 as null Hnull.
      eapply fun_nothing_fresh with (null:=null) in Htylit; eauto.
      destructc Htylit as LNN TfnN TcontN Htystack Htylit Htycont Hrest.
      destructc Hrest as Hfv HclN.
      exists (G §  y / v_loc ~ TfnN § x / v_ref ~ TfnN) ((open x u) ; fs ; (Sigma & y ~ item_fun t0) ; (e & x ~ y); LvC); simpl.
      eapply tred_let_fun.
      5:{ repeat(rewrite notin_union). repeat(split); auto.
          hnf. introv Hyin. apply Hfv in Hyin.
          repeat(rewrite in_union in Hyin).
          destruct Hyin as [ HyinGd
                           | [ HyinGfv
                             | [ Hyined
                               | [ Hyinefv
                                 | [ HyinU
                                   | [ Hyinfs
                                     | [ HyinT
                                       | [ Hyinu
                                         | [ Hyint0
                                           | Hynull ]]]]]]]]].
          all: try solve [ apply Fr0; auto
                         | apply Fr0; simpl; repeat(rewrite in_union);
                           tauto ].
          rewrite in_singleton in Hynull; subst.
          apply Fr0; auto. repeat(rewrite in_union); tauto. }
      5: { repeat(rewrite notin_union). repeat(split); auto.
           hnf. introv Hxin. apply Hfv in Hxin.
          repeat(rewrite in_union in Hxin).
           destruct Hxin as [ HxinGd
                            | [ HxinGfv
                              | [ Hxined
                                | [ Hxinefv
                                  | [ HxinU
                                    | [ Hxinfs
                                      | [ HxinT
                                        | [ Hxinu
                                          | [ Hxint0
                                            | Hxnull ]]]]]]]]].
           all: try solve [ apply Fr; auto
                          | apply Fr; simpl; repeat(rewrite in_union);
                            tauto ].
           rewrite in_singleton in Hxnull; subst.
           apply Fr; auto. repeat(rewrite in_union); tauto. }
      9: apply Htystack. 7: apply Htycont. 6: apply Htylit.
      all: eauto.
      simpl.
      repeat(rewrite notin_union).
      repeat(rewrite notin_union in Fr).
      repeat(rewrite notin_union in Fr0).
      destruct_all; repeat(split); eauto 6.

      destruct H0 as [[? ?] ?].
      unfold nothing_fresh_in_config in *. intro. intro. eapply Hfv in H6.
      repeat rewrite in_union in *. destruct_all; auto.
      apply H in H6. repeat rewrite in_union in *. destruct_all; auto.
      all: try solve [repeat (right; auto)].

      left. unfold fv, TrmFreeVar, fv_trm. repeat rewrite in_union; auto.
      left. unfold fv, TrmFreeVar, fv_trm. repeat rewrite in_union; auto.
      left. unfold fv, TrmFreeVar, fv_trm. repeat rewrite in_union; auto.

      apply H0 in H6. repeat rewrite in_union in *. destruct_all; auto.
      all: try solve [repeat (right; auto)].
      apply H1 in H6. repeat rewrite in_union in *. destruct_all; auto.
      all: try solve [repeat (right; auto)].

      left. unfold fv, TrmFreeVar, fv_trm. repeat rewrite in_union; auto.
      left. unfold fv, TrmFreeVar, fv_trm. repeat rewrite in_union; auto.
      left. unfold fv, TrmFreeVar, fv_trm. repeat rewrite in_union; auto.
      rewrite in_singleton in H6. subst. repeat right. auto.
      inversions Hclc; auto.
      intro. inversions Hclc. eapply closed_trm_lit_fun_t1; eauto.
      intro. inversions Hclc. eapply closed_trm_lit_fun_t2; eauto.
      intro. inversions Hclc. eapply closed_trm_lit_fun_T; eauto.

    + rename t into T. pick_fresh_with w (fv fs).
      find_Lv LvC.
      exists (inert_G_for_item G w x (typ_objref_loc typing_mode T))
        ((open w u)
           ; fs
           ; (Sigma & x ~ (item_obj (deref_defs (e & w ~ x) (open w d))))
           ; (e & w ~ x) ; LvC).
      eapply tred_let_obj; eauto 6.
      simpl; repeat(rewrite notin_union);
        repeat(rewrite notin_union in Fr);
        repeat(rewrite notin_union in Fr0).
      destruct_all; repeat(split); eauto 6.

  - Case "ty_sub".
    apply IHHT; eauto using ty_stack_sub.

Qed.

Lemma typed_step_not_answer : forall (F1 F2 : ctx) (C1 C2 : config) (T : typ),
    F1 ⊢ C1 ∶ T ↦ F2 ⊢ C2 ->
    not (answer C1).
Proof.
  introv Hstep. hnf. introv Hans.
  inversions Hans. inversions Hstep; try congruence.
Qed.

End WithTypingMode.
