Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping
        InertTightSubtyping 
        GeneralToTight TightTyping InvertibleTyping
        RecordTypes InertContext PreciseTyping
        OperationalSemantics Substitution Weakening
        HeapCorrespondence HeapUpdate CanonicalForms
         Narrowing
        EnvCorrespondence EWeakening Closedness
        ConfigTyping ConfigVarkinds TFree Deref
        TypedOperationalSemantics Defree Defs.

Section WithTypingMode.
Context (typing_mode: Mode).

(* Basic sanity check *)
Lemma typed_red_ty_config : forall (F1 F2 : ctx) (C1 C2 : config) (T : typ),
    F1 ⊢ C1 ∶ T ↦ F2 ⊢ C2 ->
    ty_config F1 C1 T.
Proof. introv Hstep. inversions Hstep; eauto 3. Qed.

(** This is Lemma 5.169 (REq) in the technical report. *)
Lemma typed_red_to_untyped : forall (F1 F2 : ctx) (C1 C2 : config) (T : typ),
    F1 ⊢ C1 ∶ T ↦ F2 ⊢ C2 ->
    C1 ↦ C2.
Proof. introv Htypred. inversions Htypred; eauto 3. Qed.

Lemma nothing_fresh_in_config_ctx_binds G C x T:
  nothing_fresh_in_config G C ->
  ctx_binds x T G ->
  nothing_fresh_in_config T C.
Proof.
  introv Hnfr Hbi.
  unfold nothing_fresh_in_config in *.
  assert (fv T \c fv G).
  eapply ctx_binds_fv_subset; eauto.
  hnf in *.
  eauto.
Qed.


End WithTypingMode.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma untyped_red_to_typed_helper : forall Gamma c c' U,
    ty_config Gamma c U ->
    c ↦ c' ->
    exists Gamma', Gamma ⊢ c ∶ U ↦ Gamma §& Gamma' ⊢ c'.
Proof.
  introv Htc Hred. gen c'.
  pose proof Htc as Htc'.
  inversion Htc as [Gamma' e s Sigma Lv t T U' Hin Hhc Hs Hvks Ht Hvkt]; subst;
    clear Htc.
  rename H1 into HclG, H2 into Hclc, H3 into HclU.
  induction Ht; intros.

  - (* Variable case *)
    assert (ctx_k G x v_ref) as Hvk by auto.
    exists ctx_empty; simpl. inversions Hred. inversions Hs; eauto.
    destruct (all_tfree_k_stack_push_inv Hvks).
    econstructor; try eassumption; eauto.

  - (* Function application case *)
    inversions Hred. exists ctx_empty. simpl.
    eapply tred_apply; eauto.

  - (* Field read case *)
    rename H into Hnfr, H0 into Hnull, H1 into Hvis,
      H2 into Htyfld, H3 into Hva.
    inversions Hred.
    rename H6 into Hfrw, H7 into Hbixe, H8 into HbiyS, H9 into Hhasdfs.
    simpl in Hfrw. destruct Hfrw as [Hfrw _].
    assert (HwfrG : wVal §# G) by (eapply heap_correspond_notin_dom; eauto).
    assert (Hoke : ok e) by (destruct Hhc; destruct_all; auto).
    assert (HokG : ctx_ok G) by (destruct Hhc; destruct_all; auto).
    assert (HwkG : ctx_weakens G) by (destruct Hhc; destruct_all; auto).
    assert (Hvk : ctx_k G x v_ref) by auto.
    pose proof (canonical_forms_fld Hin HclG Hhc Hvk Htyfld)
      as [?y [ds' [t' [Hbie [HbBis [Has [Hty [Hsub [HvkLoc Hrest']]]]]]]]].
    binds_eq. binds_eq.
    pose proof (defs_has_inv Has Hhasdfs) as HasInv.
    inversions HasInv. clear Has.
    destruct HvkLoc as [Tbi [?G Hbk]].
    assert (Hyts: ty_item_s _ G e Sigma yVal).
    { destruct Hhc as [_ [_ [_ [_ [HGls [HGrs [_ Hhc]]]]]]].
      apply ctx_binds_kdom in Hbk. apply Hhc. rewrite <- HGls. auto. }
    inversions Hyts.
    + (* Reading a function (lambda) *)
      apply ctx_binds_k_after_binds in Hbk as HbiyG.
      ctx_binds_eq. clear H2 H3 H4 t L.
      exists (ctx_empty § wVal / v_ref ~ (typ_fun S0 T0)).
      eapply tred_read_fun; eauto.
    + (* Reading an object *)
      apply ctx_binds_k_after_binds in Hbk as HbiyG.
      ctx_binds_eq. clear H1 H3 H4 ds.
      destruct (mode_viewpoint_adaptation' typing_mode) eqn: Hm.
      * (* basic modes *)
        exists (ctx_empty § wVal / v_ref ~ (typ_objref U1 S0)).
        rewrite <- ctx_concat_single.
        pose proof (inert_binds_inert Hin (ctx_binds_k_after_binds Hbk)) as HinU0.
        eapply tred_read_obj; rewrite ? Hm; eauto.
        { simpl. rewrite notin_union. repeat(split); auto.
          all: destruct Hnfr as [[_ Hnfr] _].
          all: apply ctx_binds_fv_subset in HbiyG; simpl in HbiyG.
          all: hnf; intros Hwin.
          2,3: specialize (HbiyG wVal); rewrite in_union in HbiyG;
          assert (wVal \in fv G) as HwinG by auto.
          all: hnf in Hfrw; apply Hfrw.
          all: apply Hnfr; auto. }
        destructc Hnfr as [? ?] ?.
        eapply nothing_fresh_in_config_ctx_binds; eauto.
        ** eapply ctx_closed_binds_closed; eauto.
      * (* rodot *)
        hnf in Hva. (* The type of the new reference is derived from the viewpoint adaptation *)
        rewrite Hm in *.
        inversions Hva.
        pose proof (typing_implies_dom Htyfld) as Hdomx.
        destructc Hnull as null Hnull.
        inversions Hclc. clear H6 H10. rename H9 into Hcls.
        destructc (adapt_ref_mutability_nothing_fresh null HokG HwkG HclG Hoke Hdomx Hcls HclU Hs Htyfld H0 H1 ) as TN0 TN TstripN TcapN ? ? Hrest.
        destructc Hrest as HicpN HcpbN HfvN HclN.
        subst.
        exists (ctx_empty § wVal / v_ref ~ (rodot_read_ref_type U1 typ_sel x TcapN)).
        eapply tred_read_obj.
        11: {
          rewrite Hm.
          exists TstripN, TcapN; eauto.
        }
        all: eauto.
        { simpl. repeat(rewrite notin_union). repeat(split); auto.
          all: destruct Hnfr as [[Hnfrdom Hnfrfv] HnfrfvU].
          all: apply ctx_binds_fv_subset in HbiyG; simpl in HbiyG.
          all: hnf; intros Hwin.
          2: specialize (HbiyG wVal); rewrite in_union in HbiyG.
          all: hnf in Hfrw; apply Hfrw.
          1,2: apply Hnfrfv; auto.

          specialize (HfvN _ Hwin). repeat(rewrite in_union in HfvN).
          destruct HfvN as
            [HfvNGd
            | [HfvNGf | [HfvNed | [HfvNef | [HfvNeU | [HfvNs | HfvNn]]]]]];
            [ apply Hnfrdom; auto
            | apply Hnfrfv; auto
            | repeat(rewrite in_union); right; auto
            | repeat(rewrite in_union); right; auto
            | apply HnfrfvU; auto
            | repeat(rewrite in_union); right; auto
            | rewrite in_singleton in HfvNn; subst;
                repeat(rewrite in_union); right; auto ]. }

        unfold nothing_fresh_in_config.
        unfold nothing_fresh_in_config in Hnfr.
        destructecs Hnfr.
        destructecs Hnfr0.

        sympls.
        union_sub_solver_destruct_ls; try union_sub_solver.
        ++
        assert (fv U1 \c fv (typ_bnd U1) ) as HfvU1.
          sympls. eapply subset_refl.
        assert (fv (typ_bnd U1)  \c fv (typ_and (typ_bnd U1) S0) ) as Hfvbnd.
          sympls. union_sub_solver.
        assert (fv (typ_and (typ_bnd U1) S0) \c fv G) as Hfvand.
          eapply ctx_binds_fv_subset; eauto.
        assert (fv (typ_bnd U1)  \c fv G) as HfvbndG.
          intro. intro. auto.
          
        eapply subset_trans.
        eapply subset_trans.
        eapply HfvU1.
        eapply HfvbndG.
        eapply subset_trans.
        eapply Hnfr0. union_sub_solver.
        ++
        eapply subset_trans. eapply HfvN.
        union_sub_solver_destruct_ls; try union_sub_solver.
        eapply subset_singleton. union_in_solver.
       
      ++   eapply closed_typ_and. eapply closed_typ_and_l.  eapply ctx_closed_binds_closed; eauto. eapply closed_typ_rcd. eapply closed_dec_typ with (L:=\{}).  intros; simpls. eapply closed_typ_bot.
          intros;simpls. cases_if. eapply closed_typ_or. eapply HclN. intro. intro. simpls. auto.
  - (* Assignment case *)
    rename H3 into Hmut. rename H4 into Htypx. rename H5 into Htypy.
    exists ctx_empty. simpl.
    inversions Hred.
    rename H11 into Hupd, H12 into Hbixe, H13 into Hbiye.
    destruct (canonical_forms_fld_ref Hin HclG Hhc Hbixe Htypx). destruct_all.
    rename x0 into ds, H into HbiyS, H0 into Hhas, H1 into Htypx1,
    H2 into Hst, H3 into Hctxkx1.
    assert (ok e) as Hoke by (destruct Hhc; destruct_all; auto).
    eapply tred_write; eauto 3.

  - (* Method call case *)
    assert (ctx_k G x v_ref) as Hvkx by auto.
    assert (ctx_k G y v_ref) as Hvky by auto.
    inversions Hred. exists ctx_empty.
    rename H15 into HbiyS, H16 into Hhas, H17 into Hbixe.
    eapply tred_call; eauto.

  - (* Let-term case *)
    inversions Hred. exists ctx_empty.
    eapply tred_let_push; eauto.

  - (* Let-literal case *)
    inversions Hred.
    + (* Function case *)
      destructc H0 as null Hnull.
      eapply fun_nothing_fresh with (null:=null) in H1; eauto.
      rename H11 into Hfryw.
      destructc H1 as LNN TfnN TcontN Htystack Htylit Htycont [Hfv HclN].
      exists (y / v_loc ~ TfnN § w / v_ref ~ TfnN); simpl.
      eapply tred_let_fun.
      5: {
        rewrite notin_union. split; hnf; introv Hyin.
        - destruct H as [[_ HnfG] _]. apply HnfG in Hyin.
          hnf in Hfryw. destruct Hfryw as [Hfry _]. auto.
        - specialize (Hfv _ Hyin); clear Hyin.
          repeat(rewrite in_union in Hfv).
          destruct H as [[HnfGd HnfGfv] HnfU].
          hnf in Hfryw. destruct Hfryw as [Hfry _].
          destruct Hfv as [ HyinGd
                          | [ HyinGfv
                            | [ Hyined
                              | [ Hyinefv
                                | [ HyinU
                                  | [ Hyins
                                    | [ HyinT0
                                      | [ Hyinu | [ Hyint1 | Hynull ]]]]]]]]].
          all: try solve [ apply Hfry; auto
                         | apply Hfry; repeat(rewrite in_union);
                           right; auto
                         | apply Hfry; simpl; repeat(rewrite in_union);
                           auto].
          rewrite in_singleton in Hynull; subst.
          apply Hfry; auto. repeat(rewrite in_union); auto.
      }
      5: {
        rewrite notin_union. split; hnf; introv Hwin.
        - destruct H as [[_ HnfG] _]. apply HnfG in Hwin.
          hnf in Hfryw. destruct Hfryw as [_ Hfrw]. simpl in Hfrw.
          destruct Hfrw as [Hfrw _]. apply Hfrw. rewrite in_union.
          left; auto.
        - specialize (Hfv _ Hwin); clear Hwin.
          repeat(rewrite in_union in Hfv).
          destruct H as [[HnfGd HnfGfv] HnfU].
          hnf in Hfryw. destruct Hfryw as [_ Hfrw].  simpl in Hfrw.
          destruct Hfrw as [Hfrw _].
          rewrite notin_union in Hfrw. destruct Hfrw as [Hfrw _].
          destruct Hfv as [ HwinGd
                          | [ HwinGfv
                            | [ Hwined
                              | [ Hwinefv
                                | [ HwinU
                                  | [ Hwins
                                    | [ HwinT0
                                      | [ Hwinu | [ Hwint1 | Hwnull ]]]]]]]]].
          all: try solve [ apply Hfrw; auto
                         | apply Hfrw; repeat(rewrite in_union);
                           right; auto
                         | apply Hfrw; simpl; repeat(rewrite in_union);
                           auto].
          rewrite in_singleton in Hwnull; subst.
          apply Hfrw; auto. repeat(rewrite in_union); auto.
      }
      9: apply Htystack. 7: apply Htycont. 6: apply Htylit.
      all: eauto.
      inversions Htylit; auto.
      destruct H as [[? ?] ?].
      unfold nothing_fresh_in_config in *. intro. intro. eapply Hfv in H4. repeat rewrite in_union in *. destruct_all; auto.
      apply H in H4. repeat rewrite in_union in *. destruct_all; auto.
      all: try solve [repeat (right; auto)].
      apply H0 in H4. repeat rewrite in_union in *. destruct_all; auto.
      all: try solve [repeat (right; auto)].
      apply H1 in H4. repeat rewrite in_union in *. destruct_all; auto.
      all: try solve [repeat (right; auto)].
      left. unfold fv, TrmFreeVar, fv_trm. repeat rewrite in_union; auto.
      left. unfold fv, TrmFreeVar, fv_trm. repeat rewrite in_union; auto.
      left. unfold fv, TrmFreeVar, fv_trm. repeat rewrite in_union; auto.
      rewrite in_singleton in H4. subst. auto.
      inversions Hclc; auto.
      intro. inversions Hclc. eapply closed_trm_lit_fun_t1; eauto.
      intro. inversions Hclc. eapply closed_trm_lit_fun_t2; eauto.
      intro. inversions Hclc. eapply closed_trm_lit_fun_T; eauto.
    + (* Object case *)
      pose proof (@tred_let_obj typing_mode).
      rename H11 into Hfryw.
      destruct (mode_refcaps' typing_mode).
      * (* basic mode *)
        exists (inert_G_for_obj ctx_empty w y T0 typ_top).
        simpl.
        eapply H4; eauto.
        { destruct H as [[HnfGd HnfGfv] HnfU].
          hnf in Hfryw. simpl in Hfryw.
          destruct Hfryw as [Hfry [Hfrw _]].
          rewrite notin_union in Hfrw. destruct Hfrw as [Hfrw _].
          rewrite notin_union. split; hnf; introv Hyin; auto.
          apply Hfry. repeat(rewrite in_union). auto. }
        { destruct H as [[HnfGd HnfGfv] HnfU].
          hnf in Hfryw. simpl in Hfryw.
          destruct Hfryw as [Hfry [Hfrw _]].
          rewrite notin_union in Hfrw. destruct Hfrw as [Hfrw _].
          rewrite notin_union. split; hnf; introv Hwin; auto.
          apply Hfrw. repeat(rewrite in_union). auto. }
      * (* rodot *)
        exists (inert_G_for_obj ctx_empty w y T0 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))).
        eapply H4; eauto.
        { destruct H as [[HnfGd HnfGfv] HnfU].
          hnf in Hfryw. simpl in Hfryw.
          destruct Hfryw as [Hfry [Hfrw _]].
          rewrite notin_union in Hfrw. destruct Hfrw as [Hfrw _].
          rewrite notin_union. split; hnf; introv Hyin; auto.
          apply Hfry. repeat(rewrite in_union). auto. }
        { destruct H as [[HnfGd HnfGfv] HnfU].
          hnf in Hfryw. simpl in Hfryw.
          destruct Hfryw as [Hfry [Hfrw _]].
          rewrite notin_union in Hfrw. destruct Hfrw as [Hfrw _].
          rewrite notin_union. split; hnf; introv Hwin; auto.
          apply Hfrw. repeat(rewrite in_union). auto. }
  - (* Subtyping case *)
    apply IHHt; eauto using ty_stack_sub.
Qed.

Corollary untyped_red_to_typed : forall Gamma c c' U,
    ty_config Gamma c U ->
    c ↦ c' ->
    exists Gamma', Gamma ⊢ c ∶ U ↦ Gamma' ⊢ c'.
Proof.
  intros. eapply untyped_red_to_typed_helper in H; eauto.
  destruct_all. eauto.
Qed.

Corollary typed_to_untyped_red_closure : forall (F1 F2 : ctx) (C1 C2 : config)
                                       (T : typ) (n : nat),
    F1 ⊢ C1 ∶ T ⊣ n ↦* F2 ⊢ C2 ->
    C1 ⊣ n ↦* C2.
Proof. induction 1; eauto using typed_red_to_untyped. Qed.


End WithTypingMode.
