Set Implicit Arguments.

Require Import LibExtra DotTactics TypingMode.
Require Import AbstractSyntax OperationalSemantics TypedOperationalSemantics.

(** * Lemmas about typed reduction *)

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma typed_red_closure_answer_case G1 C1 T n G2 C2:
  G1 ⊢ C1 ∶ T ⊣ n ↦* G2 ⊢ C2 ->
  answer C1 ->
  n = 0 /\ G1 = G2 /\ C1 = C2.
Proof.
  introv Hred Hans.
  reverts G2 C2.
  induction n; intros.
  - inversions Hred. auto.
  - inversions Hred.
    forwards: IHn; fweauto.
    destructecs H; subst.
    inversions Hans.
    inversions H4; discriminate.
Qed.

Lemma typed_red_closure_invl G1 C1 T G2 C2 n:
  G1 ⊢ C1 ∶ T ⊣ (S n) ↦* G2 ⊢ C2 ->
  exists G3 C3,
  G1 ⊢ C1 ∶ T ↦ G3 ⊢ C3 /\
  G3 ⊢ C3 ∶ T ⊣ n ↦* G2 ⊢ C2.
Proof.
  introv Hred.
  reverts G1 C1 T.
  reverts G2 C2.
  induction n; intros; inversions Hred.
  - inversions H0; eauto.
  - 
    forwards Hex: IHn H0.
    destructecs Hex.
    exists G3 C0; split; auto.
    eapply typed_red_step; eauto.
Qed.

Lemma typed_red_closure_concat G1 C1 T G2 C2 n1 G3 C3 n2:
  G1 ⊢ C1 ∶ T ⊣ n1 ↦* G2 ⊢ C2 ->
  G2 ⊢ C2 ∶ T ⊣ n2 ↦* G3 ⊢ C3 ->
  G1 ⊢ C1 ∶ T ⊣ (n1 + n2) ↦* G3 ⊢ C3.
Proof.
  introv Hred1 Hred2.
  reverts G2 C2 G3 C3.
  reverts n1.
  induction n2; intros.
  - inversions Hred2. rewrite Nat.add_0_r. auto.
  - forwards Hstep: typed_red_closure_invl Hred2.
    destructecs Hstep.
    rewrite <- Nat.add_succ_comm. eapply IHn2.
    eapply typed_red_step; eauto.
    auto.
Qed.

Lemma typed_red_stepl G1 C1 T G2 C2 G3 C3 n2:
  G1 ⊢ C1 ∶ T ↦ G2 ⊢ C2 ->
  G2 ⊢ C2 ∶ T ⊣ n2 ↦* G3 ⊢ C3 ->
  G1 ⊢ C1 ∶ T ⊣ (S n2) ↦* G3 ⊢ C3.
Proof.
  introv Hred1 Hred2.
  forwards: typed_red_step Hred1; fweauto.
  forwards: typed_red_closure_concat; eauto.
Qed.
 
 
Lemma typed_red_mono_G G1 C1 T G2 C2:
  G1 ⊢ C1 ∶ T ↦ G2 ⊢ C2 ->
  exists G3,
    G2 = G1 §& G3 /\
    ctx_weakens G3.
Proof.
  introv Hred.
  inversions Hred.
  - exists (ctx_empty §ʷ w2 ~ T2). rewrite ? ctx_concat_push. simpl. eauto.
  - exists (ctx_empty §ʷ w2 ~ (typ_fun S T1)). rewrite ? ctx_concat_push. simpl. eauto.
  - exists (ctx_empty). rewrite ? ctx_concat_push. simpl. eauto.
  - exists (ctx_empty). rewrite ? ctx_concat_push. simpl. eauto.
  - exists (ctx_empty). rewrite ? ctx_concat_push. simpl. eauto.
  - exists (inert_G_for_obj_loc typing_mode ctx_empty w y R ). rewrite ? ctx_concat_push. simpl. eauto.
  - exists ((ctx_empty §ʸ y ~ A1) §ʷ w ~ A1). rewrite ? ctx_concat_push. simpl. eauto.
  - exists (ctx_empty). rewrite ? ctx_concat_push. simpl. eauto.
  - exists (ctx_empty). rewrite ? ctx_concat_push. simpl. eauto.
Qed.

Lemma typed_red_closure_mono_G G1 C1 T n G2 C2:
  G1 ⊢ C1 ∶ T ⊣ n ↦* G2 ⊢ C2 ->
  exists G3,
    G2 = G1 §& G3 /\
    ctx_weakens G3.
Proof.
  introv Hred.
  induction Hred; intros; subst.
  - exists ctx_empty.  simpls. eauto.
  - destructecs IHHred; subst.
    forwards: typed_red_mono_G H.
    destructecs H0; subst.
  exists (G3 §& G0). split. rewrite ctx_concat_assoc. auto. eapply ctx_weakens_concat; eauto. 

Qed.


End WithTypingMode.
