Set Implicit Arguments.

Require Import LibExtra DotTactics TypingMode.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax OperationalSemantics TypedOperationalSemantics SimilarityPreservation TypedReductionEquivalence TypedPreservation ConfigSimilarity ConfigValid ConfigTyping Similarity.
Require Import ReductionFreeVars.
Require Import Lia.
Require Import TypedProgress Safety TypedSoundness.

(** * Lemmas about preservation of similarity *)

Section WithTypingMode.
Context (typing_mode: Mode).


Lemma config_hastyp_preservation C1 C2:
  config_hastyp typing_mode C1 ->
  C1 ↦ C2 ->
  config_hastyp typing_mode C2.
Proof.
  introv Hht Hred.
  inversions Hht.
  forwards G' Htred: untyped_red_to_typed; eauto.
  eapply typed_preservation in Htred.
  econstructor; eauto.
Qed.
Lemma config_hastyp_preservation_closure C1 C2 n:
  config_hastyp typing_mode C1 ->
  C1 ⊣ n ↦* C2 ->
  config_hastyp typing_mode C2.
Proof.
  introv Hht Hredc. reverts C2.
  induction n; intros; inversions Hredc.
  - eauto.
  - eapply config_hastyp_preservation. eapply IHn. eauto. eauto.
Qed.

Theorem similar_conf_1to1_preservation_closure Ca1 Ca2 Cb1 Cb2 n:
  similar_conf_1to1 Ca1 Cb1 ->
  config_hastyp typing_mode Ca1 ->
  config_hastyp typing_mode Cb1 ->
  Ca1 ⊣ n ↦* Ca2 ->
  Cb1 ⊣ n ↦* Cb2 ->
  similar_conf_1to1 Ca2 Cb2.
Proof.
  reverts Ca2 Cb2.
  induction n; intros;
  inversions H2; inversions H3.
  - eauto.
  - forwards: IHn; eauto.
    forwards: similar_conf_1to1_preservation H2.
    eapply config_hastyp_valid. eapply config_hastyp_preservation_closure. eapply H0. eauto.
    eapply config_hastyp_valid. eapply config_hastyp_preservation_closure. eapply H1. eauto.
    eauto.
    eauto.
    eauto.
Qed.


Lemma similar_preservation_closure Ca1 Ca2 Cb1 Cb2 (xa xb: list var) n:
  similar xa xb Ca1 Cb1 ->
  config_hastyp typing_mode Ca1 ->
  config_hastyp typing_mode Cb1 ->
  vars_conf_unique xa Ca1 ->
  vars_conf_unique xb Cb1 ->
  Ca1 ⊣ n ↦* Ca2 ->
  Cb1 ⊣ n ↦* Cb2 ->
  exists (ya yb: list var),
  similar (ya ++ xa) (yb ++ xb) Ca2 Cb2 /\
  vars_conf_unique (ya ++ xa) Ca2 /\
  vars_conf_unique (yb ++ xb) Cb2 /\
  (config_fv_extension Ca1 Ca2) = (from_list ya) /\
  (config_fv_extension Cb1 Cb2) = (from_list yb).
Proof.
  introv Hsi Htya Htyb Hvua Hvub Hra Hrb.
  reverts Ca2 Cb2.
  induction n; intros;
  inversions Hra; inversions Hrb.
  - exists (@nil var) (@nil var).
    rewrite ! List.app_nil_l, from_list_nil, ! config_fv_extension_diag.
    splits; auto 1.
  - forwards Hn: IHn H0 H1.
    destructecs Hn.
    forwards Hs: similar_preservation Hn0 H2 H4.
    eapply config_hastyp_valid. eapply config_hastyp_preservation_closure. eapply Htya. eauto.
    eapply config_hastyp_valid. eapply config_hastyp_preservation_closure. eapply Htyb. eauto.
    eauto.
    eauto.
    destructecs Hs.
    exists (ya0 ++ ya)%list (yb0 ++ yb)%list; splits.
    rewrite <- ! List.app_assoc. auto.
    rewrite <- ! List.app_assoc. auto.
    rewrite <- ! List.app_assoc. auto.
    erewrite config_fv_extension_trans. 2: eauto. 2: eapply red_trans. 2: eapply red_refl. 2: auto. rewrite Hn3, Hs3, from_list_append. auto.
    erewrite config_fv_extension_trans. 2: eauto. 2: eapply red_trans. 2: eapply red_refl. 2: auto. rewrite Hn, Hs, from_list_append. auto.
Qed.

Lemma step_not_answer : forall (C1 C2 : config),
     C1 ↦  C2 ->
    not (answer C1).
Proof.
  introv Hstep. hnf. introv Hans.
  inversions Hans. inversions Hstep.
Qed.


Lemma similar_preservation_closure_answer_overshoot Ca1 Ca2 Cb1 Cb2 (xa xb: list var) n1 n2:
  similar xa xb Ca1 Cb1 ->
  config_hastyp typing_mode Ca1 ->
  config_hastyp typing_mode Cb1 ->
  vars_conf_unique xa Ca1 ->
  vars_conf_unique xb Cb1 ->
  Ca1 ⊣ n1 ↦* Ca2 ->
  Cb1 ⊣ n2 ↦* Cb2 ->
  answer Ca2 ->
  n2 <= n1.
Proof.
  introv Hsi Htya Htyb Hvua Hvub Hra Hrb Ha1.
  destruct (Nat.lt_ge_cases n1 n2) as [? | ? ].
  -
    forwards Hred: ReductionLemmas.red_closure_short n1 Hrb.
    lia.
    destructecs Hred. subst.
    forwards Hsic: similar_preservation_closure Hsi Hra Hred0; fweauto.  
    destructecs Hsic. 
    assert (mn > 0). lia. destruct mn. lia.
    eapply ReductionLemmas.red_closure_invl in Hred1.
    destructecs Hred1.
    forwards: similar_conf_answer_case. exists (ya ++ xa)%list (yb ++ xb)%list; splits; eauto.  hnf in Hsic0. destruct Ca2, C3. destructecs Hsic0. eapply similar_length with (a:=t). eauto. auto.
    false. eapply step_not_answer; eauto.
  - auto.
Qed.
Lemma similar_preservation_closure_answer_overshoot_r Ca1 Ca2 Cb1 Cb2 (xa xb: list var) n1 n2:
  similar xa xb Ca1 Cb1 ->
  config_hastyp typing_mode Ca1 ->
  config_hastyp typing_mode Cb1 ->
  vars_conf_unique xa Ca1 ->
  vars_conf_unique xb Cb1 ->
  Ca1 ⊣ n1 ↦* Ca2 ->
  Cb1 ⊣ n2 ↦* Cb2 ->
  answer Cb2 ->
  n1 <= n2.
Proof.
  introv Hsi Htya Htyb Hvua Hvub Hra Hrb Ha1.
  eapply similar_symm in Hsi.
  eapply similar_preservation_closure_answer_overshoot; eauto.
Qed.

Lemma similar_preservation_closure_answer_steps Ca1 Ca2 Cb1 Cb2 (xa xb: list var) n1 n2:
  similar xa xb Ca1 Cb1 ->
  config_hastyp typing_mode Ca1 ->
  config_hastyp typing_mode Cb1 ->
  vars_conf_unique xa Ca1 ->
  vars_conf_unique xb Cb1 ->
  Ca1 ⊣ n1 ↦* Ca2 ->
  Cb1 ⊣ n2 ↦* Cb2 ->
  answer Ca2 ->
  answer Cb2 ->
  n2 = n1.
Proof.
  introv Hsi Htya Htyb Hvua Hvub Hra Hrb Ha1 Ha2.

  forwards: similar_preservation_closure_answer_overshoot; eauto.
  forwards: similar_preservation_closure_answer_overshoot_r; eauto.

  eapply Nat.le_antisymm; eauto.
Qed.

Lemma similar_preservation_closure_answer Ca1 Ca2 Cb1 Cb2 (xa xb: list var) n1 n2:
  similar xa xb Ca1 Cb1 ->
  config_hastyp typing_mode Ca1 ->
  config_hastyp typing_mode Cb1 ->
  vars_conf_unique xa Ca1 ->
  vars_conf_unique xb Cb1 ->
  Ca1 ⊣ n1 ↦* Ca2 ->
  Cb1 ⊣ n2 ↦* Cb2 ->
  answer Ca2 ->
  answer Cb2 ->
  n1 = n2 /\
  exists (ya yb: list var),
  similar (ya ++ xa) (yb ++ xb) Ca2 Cb2 /\
  vars_conf_unique (ya ++ xa) Ca2 /\
  vars_conf_unique (yb ++ xb) Cb2 /\
  (config_fv_extension Ca1 Ca2) = (from_list ya) /\
  (config_fv_extension Cb1 Cb2) = (from_list yb).
Proof.
  introv Hsi Htya Htyb Hvua Hvub Hra Hrb Ha1 Ha2.
  assert (n2 = n1).
  eapply similar_preservation_closure_answer_steps; eauto. subst.
  
  split. auto.
  eapply similar_preservation_closure; eauto.
Qed.


(*
Theorem similar_conf_1to1_typed_preservation,
but providing the fourth configuration.
*)
Theorem similar_conf_1to1_closure_case Ca1 Ca2 Cb1 n:
  similar_conf_1to1 Ca1 Cb1 ->
  config_hastyp typing_mode Ca1 ->
  config_hastyp typing_mode Cb1 ->
  Ca1 ⊣ n ↦* Ca2 ->
  exists Cb2,
  Cb1 ⊣ n ↦* Cb2 /\
  similar_conf_1to1 Ca2 Cb2.
Proof.
  introv Hsi Hty1 Hty2 Hct.
  reverts Ca2.
  (* Induction by execution length *)
  induction n; intros.
  - inversions Hct.
    exists Cb1; auto.
  - (* Add one step *)
    inversions Hct.
    forwards HCb: IHn; fweauto.
    destructecs HCb.
    inversions Hty1.
    inversions Hty2.
    (* Use preservation to show that the configurations obtained by induction have types *)
    forwards HtyC2: preservation_closure H0; fweauto.
    forwards HtyCb2: preservation_closure HCb0; fweauto.
    inversions HtyCb2.
    inversions HtyC2.
    (* Use progress to show that next step exists *)
    forwards Hcases: terminates_or_not Cb2 H3; fweauto.
    destructecs Hcases.
    + exists c'. splits. econstructor; eauto.
      eapply similar_conf_1to1_preservation; eauto using config_typed_valid.
    + (* answers cannot progress *) 
      forwards: similar_conf_answer_case Hcases.
      eapply similar_conf_1to1_symm; eauto.
      forwards [? ?]: untyped_red_to_typed; eauto.
      false typed_step_not_answer; eauto.
Qed.

Lemma similar_conf_1to1_terminates_in C1 C2 n:
  similar_conf_1to1 C1 C2 ->
  config_hastyp typing_mode C1 ->
  config_hastyp typing_mode C2 ->
  config_terminates_in C1 n ->
  config_terminates_in C2 n.
Proof.
  introv Hsi Hty1 Hty2 Hct.

  unfold config_terminates_in in *.
  destructecs Hct.

  forwards HCb2: similar_conf_1to1_closure_case; fweauto.
  destructecs HCb2.
  forwards: similar_conf_answer_case; eauto.
Qed.


End WithTypingMode.