Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import AbstractSyntax TypingMode GeneralTyping
        InertTightSubtyping
        GeneralToTight TightTyping InvertibleTyping
        RecordTypes InertContext PreciseTyping
        OperationalSemantics Substitution Weakening
        HeapCorrespondence HeapUpdate CanonicalForms
         Narrowing
        EnvCorrespondence EWeakening Visibility
        ConfigTyping ConfigVarkinds TFree Deref ReferenceTypes
        TypedOperationalSemantics Defree Defs Closedness.
Require Import TypedReductionEquivalence TypedProgress TypedPreservation.

(** * Soundness theorem for typed reduction *)

Require Import Logic.Classical_Prop Logic.Classical_Pred_Type.

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma untyped_red_to_typed_closure : forall Gamma c c' U n,
    ty_config Gamma c U ->
    c  ⊣ n ↦*  c' ->
    exists Gamma', Gamma ⊢ c ∶ U  ⊣ n ↦*  Gamma' ⊢ c'.
Proof.
  intros. induction H0. exists Gamma. auto.
  specialize (IHred_closure H). destructc IHred_closure as Gamma2 Hred2.
  pose proof Hred2 as Hty2.
  eapply typed_preservation_closure in Hty2; eauto.
  eapply untyped_red_to_typed in H1; eauto. destructc H1 as G2 HG2. exists G2. econstructor; eauto.
Qed.


(* Theorem 7 in "Reference mutability for DOT" *)
(* WARNING: the proof is not constructive. *)
Theorem typed_red_soundness : forall (F : ctx) (C : config) (T : typ),
    (* If a configuration is well-typed *)
    ty_config F C T ->
    (* It either converges to an answer using typed reduction *)
    (exists (n : nat) (Fa : ctx) (Ca : config),
        answer Ca /\ (F ⊢ C ∶ T ⊣n↦* Fa ⊢ Ca)) \/
    (* Or diverges using typed reduction *)
    (forall (n : nat), exists (Fn : ctx) (Cn : config),
          F ⊢ C ∶ T ⊣n↦* Fn ⊢ Cn).
Proof.
  introv Htc.
  (* Uses classical logic to avoid having to construct a decision
     procedure to decide whether a well-typed config. terminates *)
  destruct (classic  (exists (n : nat) (Fa : ctx) (Ca : config),
                         answer Ca /\ (F ⊢ C ∶ T ⊣n↦* Fa ⊢ Ca)))
    as [? | Hdiv]; eauto.
  (* Configuration C diverges *)
  right. intros. induction n; eauto.
  destruct IHn as [Fn [Cn Hstepn]].
  apply typed_preservation_closure in Hstepn as Htcn; eauto.
  apply typed_progress in Htcn as Hprog; eauto.
  destruct Hprog as [Hnans | [FSn [CSn Hstep]]].
  - exfalso. eapply Hdiv. exists n Fn Cn. split; eauto.
  - exists FSn CSn. eauto.
Qed.

End WithTypingMode.
