Set Implicit Arguments.

Require Import Coq.Init.Datatypes.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax Generalization
        Similarity SyntaxSimilarity ConfigSimilarity OperationalSemantics HeapUpdate Deref Defs ConfigValid.
Require Import ReductionFreeVars.

(** * Preservation of similarity by reduction *)

(* Lemmas for heap similarity *)

Lemma similar_renv_binds xs1 xs2 ys1 ys2 (S1 S2: renv) x1 x2 it1 it2:
  similar_env_with similar_var xs1 xs2 ys1 ys2 S1 S2 ->
  similar ys1 ys2 x1 x2 ->
  binds x1 it1 S1 ->
  binds x2 it2 S2 ->
  similar xs1 xs2 it1 it2.
Proof.
  intros.
  unfold similar, similar_var, similar_env in H.
  eapply similar_var_simlist_relates in H0.
  eapply similar_env_with_binds; eauto.
Qed.
Lemma similar_renv_binds_to_binds xs1 xs2 ys1 ys2 (S1 S2: renv) x1 x2 it1 it2:
  ok S1 ->
  fresh \{} (length xs1) xs1 ->
  similar_env_with similar_var xs1 xs2 ys1 ys2 S1 S2 ->
  similar ys1 ys2 x1 x2 ->
  binds x1 it1 S1 ->
  similar xs1 xs2 it1 it2 ->
  binds x2 it2 S2.
Proof.
  introv Hoke Hfrx1.
  intros.
  unfold similar, similar_var, similar_env in H.
  eapply similar_var_simlist_relates in H0.
  forwards: (similar_env_with_cases H0 H); eauto.
  destruct_all.
  - eapply binds_to_dom in H1. contradiction.
  - binds_eq.
    forwards: similar_noduplicates_var_eq H2 H5; subst; eauto.
Qed.
Lemma similar_renv_binds_to xs1 xs2 ys1 ys2 (S1 S2: renv) x1 x2 it1:
  similar_env_with similar_var xs1 xs2 ys1 ys2 S1 S2 ->
  similar ys1 ys2 x1 x2 ->
  binds x1 it1 S1 ->
  exists it2,
  binds x2 it2 S2 /\
  similar xs1 xs2 it1 it2.
Proof.
  intros.
  unfold similar, similar_var, similar_env in H.
  eapply similar_var_simlist_relates in H0.
  forwards: (similar_env_with_cases H0 H); eauto.
  destruct_all.
  - eapply binds_to_dom in H1. contradiction.
  - binds_eq. eauto.
Qed.

Lemma similar_heap_binds xs1 xs2 ys1 ys2 (S1 S2: heap) x1 x2 it1 it2:
  similar_env_with similar_item xs1 xs2 ys1 ys2 S1 S2 ->
  similar ys1 ys2 x1 x2 ->
  binds x1 it1 S1 ->
  binds x2 it2 S2 ->
  similar xs1 xs2 it1 it2.
Proof.
  intros.
  unfold similar, similar_heap, similar_env in H.
  eapply similar_var_simlist_relates in H0.
  eapply similar_env_with_binds; eauto.

Qed.

Lemma similar_defs_update xs1 xs2 ds11 ds12 ds21 ds22 yArg1 yArg2 m (*ys1 ys2*):
(*  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  similar_env_with similar_item xs1 xs2 ys1 ys2 S1 S2 ->*)
  similar xs1 xs2 ds11 ds21 ->
  similar xs1 xs2 yArg1 yArg2 ->
  defs_update ds11 m yArg1 ds12 ->
  defs_update ds21 m yArg2 ds22 ->
  similar xs1 xs2 ds12 ds22.
Proof.
  introv Hsds HsArg Hu1 Hu2.
  reverts ds21 ds22.
  induction Hu1; intros.
  -
    destruct_similar_by_subst (fv x \u fv yArg2) Hsds ds. sympls.
    destruct_similar_by_subst_by xds3 (fv (defs_cons ds (def_fld a t)) \u fv ds21) HsArg yArg. sympls. fresh_solve.
    invert_subst_vars_in Hads.
    invert_subst_vars.
    invert_subst_vars.
    autorewrite with simpl_subst_vars in *.
    sympls.
    inversions Hu2.
    +
      eapply similar_defs_cons.
      eapply similar_defs_cons_defs; eauto.
      eapply similar_def_fld; eauto.
      eapply similar_avar_f; eauto.
    + simpls. contradiction.
  -
    inversions Hu2.
    + false.
      eapply similar_defs_cons_def in Hsds.
      unfold similar, similar_defs, similar_def, similar_by_subst in Hsds.
      destruct_all.
      invert_subst_vars.
      rewrite <- subst_vars_label_of_def in H.
      sympls. contradiction.
    +
      eapply similar_defs_cons.
      eapply IHHu1; eauto.
      eapply similar_defs_cons_defs; eauto.
      eapply similar_defs_cons_def; eauto.
Qed.

Lemma similar_deref_avar xs1 xs2 (ds1 ds2: avar) e1 e2:
  ok e1 ->
  ok e2 ->
  renv_idempotent e1 ->
  renv_idempotent e2 ->
  similar xs1 xs2 ds1 ds2 ->
  similar xs1 xs2 e1 e2 ->
  similar xs1 xs2 (deref e1 ds1) (deref e2 ds2).
Proof.
  introv Hoke1 Hoke2 Hidm1 Hidm2 Hsds Hse.
  destruct (deref_avar_cases ds1 Hidm1 Hoke1); eauto;
  destruct (deref_avar_cases ds2 Hidm2 Hoke2); eauto.
  - destruct H; destruct H0; subst.
    rewrite H. rewrite H0. auto.
  - false.
    destruct_all; subst.
    eapply similar_avar_b_case in Hsds. discriminate.
    eapply similar_avar_f_var in Hsds.
    eapply similar_var_simlist_relates in Hsds.
    forwards: (similar_env_with_cases Hsds Hse).
    destruct_all.
    + eapply binds_to_dom in H2. contradiction.
    + eapply binds_to_dom in H0. contradiction.
  - false.
    destruct_all; subst.
    eapply similar_symm in Hsds.
    eapply similar_avar_b_case in Hsds. discriminate.
    eapply similar_avar_f_var in Hsds.
    eapply similar_var_simlist_relates in Hsds.
    forwards: (similar_env_with_cases Hsds Hse).
    destruct_all.
    + eapply binds_to_dom in H3. contradiction.
    + eapply binds_to_dom in H1. contradiction.
  - destruct_all; subst.
    eapply similar_avar_f_var in Hsds.
    rewrite ! deref_avar_comm.
    eapply similar_avar_f.
    forwards: similar_renv_binds; eauto.
Qed.

Lemma similar_deref_def xs1 xs2 (ds1 ds2: def) e1 e2:
  ok e1 ->
  ok e2 ->
  renv_idempotent e1 ->
  renv_idempotent e2 ->
  similar xs1 xs2 ds1 ds2 ->
  similar xs1 xs2 e1 e2 ->
  similar xs1 xs2 (deref_def e1 ds1) (deref_def e2 ds2).
Proof.
  introv Hoke1 Hoke2 Hidm1 Hidm2 Hsds Hse.
  destruct ds1.
  - (* typ *)
    unfold similar, similar_defs, similar_def, similar_by_subst in Hsds.
    destruct_all. subst.
    destruct x; autorewrite with simpl_subst_vars in *; try discriminate.
    inversions H3.
    sympls.
    unfold similar, similar_defs, similar_def, similar_by_subst.
    exists (def_typ t t2) x0.
    splits;  autorewrite with simpl_subst_vars in *; sympls; eauto.
  - (* fld *)
    unfold similar, similar_defs, similar_def, similar_by_subst in Hsds.
    destruct_all. subst.
    destruct x; autorewrite with simpl_subst_vars in *; try discriminate.
    inversions H3.
    sympls.
    eapply similar_def_fld.
    eapply similar_deref_avar; eauto.
    unfold similar, similar_avar, similar_by_subst.
    exists a0 x0; splits; eauto.
  - (* met *)
    unfold similar, similar_defs, similar_def, similar_by_subst in Hsds.
    destruct_all. subst.
    destruct x; autorewrite with simpl_subst_vars in *; try discriminate.
    inversions H3.
    sympls.
    unfold similar, similar_defs, similar_def, similar_by_subst.
    exists (def_met m t0) x0.
    splits;  autorewrite with simpl_subst_vars in *; sympls; eauto.

Qed.

Lemma similar_deref_defs xs1 xs2 (ds1 ds2: defs) e1 e2:
  ok e1 ->
  ok e2 ->
  renv_idempotent e1 ->
  renv_idempotent e2 ->
  similar xs1 xs2 ds1 ds2 ->
  similar xs1 xs2 e1 e2 ->
  similar xs1 xs2 (deref_defs e1 ds1) (deref_defs e2 ds2).
Proof.
  introv Hoke1 Hoke2 Hidm1 Hidm2 Hsds Hse.
  reverts ds2.
  induction ds1; intros.
  - sympls.
    unfold similar, similar_defs, similar_def, similar_by_subst in Hsds.
    destruct_all. subst. destruct x; autorewrite with simpl_subst_vars in *; try discriminate. sympls.
    eapply similar_defs_nil. eapply similar_length in Hse. auto.
  - sympls.
    unfold similar, similar_defs, similar_def, similar_by_subst in Hsds.
    destruct_all. subst. destruct x; autorewrite with simpl_subst_vars in *; try discriminate. sympls.
    inversions H3.
    eapply similar_defs_cons.
    eapply IHds1. exists x x0; splits; eauto using subset_union_l1.
    eapply similar_deref_def; eauto.
    exists d0 x0; splits; eauto using subset_union_l2.
Qed.

Lemma similar_heap_update xs1 xs2 ys1 ys2 S1 S2 yObj1 yObj2 yArg1 yArg2 S1u S2u it1 it2 x1 x2 m:
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  similar_with xs1 xs2 ys1 ys2 S1 S2 ->
  similar xs1 xs2 x1 x2 ->
  similar xs1 xs2 yObj1 yObj2 ->
  similar xs1 xs2 yArg1 yArg2 ->
  similar xs1 xs2 it1 it2 ->
  heap_update S1 yObj1 m yArg1 S1u ->
  heap_update S2 yObj2 m yArg2 S2u ->
  binds x1 it1 S1 ->
  binds x2 it2 S2 ->
  exists it1u it2u,
  binds x1 it1u S1u /\
  binds x2 it2u S2u /\
  similar xs1 xs2 it1u it2u.
Proof.
  introv Hfr1 Hfr2.
  introv HsS Hsx HsObj HsArg Hsit Hu1 Hu2 Hbi1 Hbi2.
  forwards Heq: similar_noduplicates_var_cases Hsx HsObj; eauto.
  destructc Heq as Heq1 Heq2.

  destruct (heap_update_binds_cases Hu1 Hbi1);
  destruct (heap_update_binds_cases Hu2 Hbi2); destruct_all.
  - (* Not changed *)
    exists it1 it2. splits; eauto.
  - apply Heq2 in H0. contradiction.
  - apply Heq1 in H. contradiction.
  - (* Changed *)
    subst. exists (item_obj x4) (item_obj x0). splits; eauto.
    eapply similar_item_obj.
    applys similar_defs_update HsArg; eauto.
    eapply similar_item_obj_defs; eauto.
Qed.




(* This lemma needs 1-to-1 similarity. Otherwise, update could happen on different objects *)
Lemma similar_reduce_write_heap_update xs1 xs2 S1 S2 S1u S2u yObj1 yObj2 yArg1 yArg2 m:
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  similar xs1 xs2 S1 S2 ->
  similar xs1 xs2 yObj1 yObj2 ->
  similar xs1 xs2 yArg1 yArg2 ->
  heap_update S1 yObj1 m yArg1 S1u ->
  heap_update S2 yObj2 m yArg2 S2u ->
  similar xs1 xs2 S1u S2u.
Proof.
  introv Hfr1 Hfr2 HS HObj HArg2.
  introv Hup1 Hup2.
  unfold similar, similar_heap, similar_env.
  unfold similar, similar_heap, similar_env in HS.
  eapply similar_env_with_ind_list_split with (S3 := S1) (S4:=S2); eauto; intros.
  clear HS. reverts ys1 ys2.
  induction H2; intros; subst.
  - constructor. auto.
  - forwards Hit: similar_heap_update H2 Hup1 H0; eauto.
    eapply similar_var_middle; eauto.
    eapply similar_length in HObj. rewrite ! List.app_length in HObj. simpl in HObj. rewrite H4 in HObj. eapply Plus.plus_reg_l in HObj. inversion HObj. auto.
    destructc Hit as it1u it2u Hbi1u Hbi2u Hsit.
    eapply similar_env_dom; eauto.
    eapply IHsimilar_env_with with (ys1 := (ys0 ++ x1 :: nil)%list) (ys2 := (ys3 ++ x2 :: nil)%list); eauto.
    rewrite <- List.app_assoc. rewrite <- List.app_comm_cons. rewrite List.app_nil_l. auto.
    rewrite ! List.app_length. rewrite H4. simpls. auto.
    rewrite <- List.app_assoc. rewrite <- List.app_comm_cons. rewrite List.app_nil_l. auto.
  - eapply similar_env_notin.
    eapply IHsimilar_env_with with (ys1 := (ys0 ++ x1 :: nil)%list) (ys2 := (ys3 ++ x2 :: nil)%list); eauto.
    rewrite <- List.app_assoc. rewrite <- List.app_comm_cons. rewrite List.app_nil_l. auto.
    rewrite ! List.app_length. rewrite H3. simpls. auto.
    rewrite <- List.app_assoc. rewrite <- List.app_comm_cons. rewrite List.app_nil_l. auto.
    eapply heap_update_notin_dom; eauto.
    eapply heap_update_notin_dom; eauto.
Qed.


(* Lemmas for similarity in reduction steps *)

Lemma similar_reduce_read_trm xs1 xs2 m wObj1 wObj2 wVal1 wVal2:
  similar xs1 xs2 (trm_read (avar_f wObj1) m) (trm_read (avar_f wObj2) m) ->
  similar (wVal1::xs1) (wVal2::xs2) (trm_var (avar_f wVal1)) (trm_var (avar_f wVal2)).
Proof.
  intros.
  pose proof (similar_length xs1 xs2 _ _ H) as Hl.
  eapply similar_trm_var.
  eapply similar_avar_f.
  eapply simlist_relates_similar_var. simpls. rewrite Hl. auto.
  eapply simlist_relates_push_l.
Qed.

Lemma similar_reduce_read_renv xs1 xs2 w1 w2 y1 y2 e1 e2 wo1 wo2 yo1 yo2 S1 S2 ds1 ds2 a:
  similar xs1 xs2 (trm_read (avar_f wo1) a) (trm_read (avar_f wo2) a) ->
  binds wo1 yo1 e1 ->
  binds wo2 yo2 e2 ->
  binds yo1 (item_obj ds1) S1 ->
  binds yo2 (item_obj ds2) S2 ->
  defs_has ds1 (def_fld a (avar_f y1)) ->
  defs_has ds2 (def_fld a (avar_f y2)) ->
  similar xs1 xs2 S1 S2 ->
  similar xs1 xs2 e1 e2 ->
  w1 \notin from_list xs1 ->
  w2 \notin from_list xs2 ->
  similar (w1 :: xs1) (w2 :: xs2) (e1 & w1 ~ y1) (e2 & w2 ~ y2).
Proof.
  introv Hs Hbiwo1 Hbiwo2 Hbiyo1 Hbiyo2 Hhas1 Hhas2 HsS Hse.
  introv Hnw1 Hnw2.

  assert (similar xs1 xs2 wo1 wo2). {
    eapply similar_avar_f_var.
    eapply similar_trm_read_var; eauto.
  }
  assert (similar xs1 xs2 yo1 yo2). {
    eapply similar_renv_binds; eauto.
  }
  assert (similar xs1 xs2 ds1 ds2). {
    eapply similar_item_obj_defs.
    eapply similar_heap_binds; eauto.
  }
  assert (similar xs1 xs2 y1 y2). {
    eapply similar_avar_f_var.
    eapply similar_def_fld_var.
    eapply similar_defs_has; eauto.
    sympls. auto.
  }
  eapply similar_env_push; eauto.
  eapply similar_extend; eauto.
Qed.

Lemma similar_reduce_write_trm xs1 xs2 m wObj1 wObj2 wArg1 wArg2:
  similar xs1 xs2 (trm_write (avar_f wObj1) m (avar_f wArg1)) (trm_write (avar_f wObj2) m (avar_f wArg2)) ->
  similar xs1 xs2 (trm_var (avar_f wArg1)) (trm_var (avar_f wArg2)).
Proof.
  intros.
  eapply similar_trm_var.
  eapply similar_trm_write_var2; eauto.
Qed.

Lemma similar_reduce_write_heap xs1 xs2 e1 e2 S1 S2 S1u S2u wObj1 wObj2 yObj1 yObj2 wArg1 wArg2 yArg1 yArg2 m:
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  similar xs1 xs2 (trm_write (avar_f wObj1) m (avar_f wArg1)) (trm_write (avar_f wObj2) m (avar_f wArg2)) ->
  similar xs1 xs2 S1 S2 ->
  similar xs1 xs2 e1 e2 ->
  binds wObj1 yObj1 e1 ->
  binds wArg1 yArg1 e1 ->
  binds wObj2 yObj2 e2 ->
  binds wArg2 yArg2 e2 ->
  heap_update S1 yObj1 m yArg1 S1u ->
  heap_update S2 yObj2 m yArg2 S2u ->
  similar xs1 xs2 S1u S2u.
Proof.
  introv Hfr1 Hfr2.
  introv Hsimt HsimS Hsime.
  introv HbiObj1 HbiArg1 HbiObj2 HbiArg2.
  introv Hup1 Hup2.

  assert (similar xs1 xs2 wArg1 wArg2). {
    eapply similar_avar_f_var.
    eapply similar_trm_write_var2; eauto.
  }

  assert (similar xs1 xs2 wObj1 wObj2). {
    eapply similar_avar_f_var.
    eapply similar_trm_write_var1; eauto.
  }

  assert (similar xs1 xs2 yObj1 yObj2). {
    eapply similar_renv_binds; eauto.
  }
  assert (similar xs1 xs2 yArg1 yArg2). {
    applys similar_renv_binds HbiArg2; eauto.
  }

  applys similar_reduce_write_heap_update Hup1 Hup2; eauto.
Qed.

Lemma similar_reduce_apply_trm xs1 xs2 e1 e2 S1 S2 wObj1 wObj2 yObj1 yObj2 wArg1 wArg2 t1 t2:
  similar xs1 xs2 (trm_apply (avar_f wObj1) (avar_f wArg1)) (trm_apply (avar_f wObj2) (avar_f wArg2)) ->
  similar xs1 xs2 S1 S2 ->
  similar xs1 xs2 e1 e2 ->
  binds yObj1 (item_fun t1) S1 ->
  binds yObj2 (item_fun t2) S2 ->
  binds wObj1 yObj1 e1 ->
  binds wObj2 yObj2 e2 ->
  similar xs1 xs2 (open wArg1 t1) (open wArg2 t2).
Proof.
  introv Hsimt HsimS Hsime.
  introv HbiObj1S HbiObj2S HbiObj1 HbiObj2.

  eapply similar_open_rec.
  eapply similar_item_fun_trm.
  eapply similar_heap_binds; eauto.
  eapply similar_renv_binds; eauto.
  eapply similar_avar_f_var; eauto.
  eapply similar_trm_apply_var1; eauto.
  eapply similar_avar_f_var; eauto.
  eapply similar_trm_apply_var2; eauto.
Qed.

Lemma similar_reduce_call_trm xs1 xs2 e1 e2 S1 S2 wObj1 wObj2 yObj1 yObj2 wArg1 wArg2 ds1 ds2 t1 t2 m:
  similar xs1 xs2 (trm_call (avar_f wObj1) m (avar_f wArg1)) (trm_call (avar_f wObj2) m (avar_f wArg2)) ->
  similar xs1 xs2 S1 S2 ->
  similar xs1 xs2 e1 e2 ->
  binds yObj1 (item_obj ds1) S1 ->
  binds yObj2 (item_obj ds2) S2 ->
  binds wObj1 yObj1 e1 ->
  binds wObj2 yObj2 e2 ->
  defs_has ds1 (def_met m t1) ->
  defs_has ds2 (def_met m t2) ->
  similar xs1 xs2 (open_met wArg1 wObj1 t1) (open_met wArg2 wObj2 t2).
Proof.
  introv Hsimt HsimS Hsime.
  introv HbiObj1S HbiObj2S HbiObj1 HbiObj2.
  introv Hhas1 Hhas2.

  assert (similar xs1 xs2 wArg1 wArg2). {
    eapply similar_avar_f_var.
    eapply similar_trm_call_var2; eauto.
  }

  assert (similar xs1 xs2 wObj1 wObj2). {
    eapply similar_avar_f_var.
    eapply similar_trm_call_var1; eauto.
  }

  assert (similar xs1 xs2 yObj1 yObj2). {
    eapply similar_renv_binds; eauto.
  }

  assert (similar xs1 xs2 ds1 ds2). {
    eapply similar_item_obj_defs; eauto.
    eapply similar_heap_binds; eauto.
  }

  assert (similar xs1 xs2 t1 t2). {
    eapply similar_def_met_trm; eauto.
    eapply similar_defs_has; eauto.
    sympls. auto.
  }


  eapply similar_open_rec; eauto.
  eapply similar_open_rec; eauto.
Qed.


Lemma similar_reduce_var_trm xs1 xs2 t1 t2 s1 s2 x1 x2:
  similar xs1 xs2 (trm_var (avar_f x1)) (trm_var (avar_f x2)) ->
  similar xs1 xs2 (stack_cons (frame_let t1) s1) (stack_cons (frame_let t2) s2) ->
  similar xs1 xs2 (open x1 t1) (open x2 t2).
Proof.
  introv Hst Hss.
  eapply similar_open_rec; eauto.
  eapply similar_stack_let_trm; eauto.
  eapply similar_avar_f_var.
  eapply similar_trm_var_var; eauto.
Qed.

Lemma similar_reduce_lit_trm xs1 xs2 t12 t22 w1 w2 y1 y2 l1 l2:
  similar xs1 xs2 (trm_lit l1 t12) (trm_lit l2 t22) ->
  similar (w1 :: y1 :: xs1) (w2 :: y2 :: xs2) (open w1 t12) (open w2 t22).
Proof.
  introv Hsimt.
  eapply similar_open_rec.
  - eapply similar_trm_lit_trm; eauto. eapply similar_extend. eapply similar_extend; eauto.
  - eapply similar_var_push. simpls. eapply similar_length in Hsimt. rewrite Hsimt; auto.
Qed.

Lemma similar_reduce_fun_heap xs1 xs2 T1 T2 t11 t21 t12 t22 w1 w2 y1 y2 S1 S2:
  similar xs1 xs2 (trm_lit (lit_fun T1 t11) t12) (trm_lit (lit_fun T2 t21) t22) ->
  similar xs1 xs2 S1 S2 ->
  w1 \notin ((dom S1) \u \{y1}) ->
  w2 \notin ((dom S2) \u \{y2}) ->
  y1 \notin ((dom S1) \u from_list xs1) ->
  y2 \notin ((dom S2) \u from_list xs2) ->
  similar (w1 :: y1 :: xs1) (w2 :: y2 :: xs2) (S1 & y1 ~ item_fun t11) (S2 & y2 ~ item_fun t21).
Proof.
  intros.
  eapply similar_extend_notin.
  - eapply similar_env_push; eauto.
    eapply similar_extend.
    eapply similar_item_fun.
    eapply similar_lit_fun_trm.
    eapply similar_trm_lit_lit; eauto.
  - simpl_dom. notin_solve.
  - simpl_dom. notin_solve.
Qed.

Lemma similar_reduce_obj_heap xs1 xs2 T1 T2 ds1 t12 ds2 t22 w1 w2 y1 y2 S1 S2 e1 e2:
  similar xs1 xs2 (trm_lit (lit_obj T1 ds1) t12) (trm_lit (lit_obj T2 ds2) t22) ->
  similar xs1 xs2 e1 e2 ->
  similar xs1 xs2 S1 S2 ->
  ok e1 ->
  ok e2 ->
  renv_idempotent e1 ->
  renv_idempotent e2 ->
  w1 \notin (dom S1 \u \{y1} \u dom e1 \u fv_env e1 \u from_list xs1) ->
  w2 \notin (dom S2 \u \{y2} \u dom e2 \u fv_env e2 \u from_list xs2) ->
  y1 \notin ((dom S1) \u from_list xs1 \u dom e1) ->
  y2 \notin ((dom S2) \u from_list xs2 \u dom e2) ->
  similar (w1 :: y1 :: xs1) (w2 :: y2 :: xs2) (S1 & y1 ~ item_obj (deref_defs (e1 & w1 ~ y1) (open w1 ds1))) (S2 & y2 ~ item_obj (deref_defs (e2 & w2 ~ y2) (open w2 ds2))).
Proof.
  introv Hst Hse HsS.
  introv Hoke1 Hoke2 Hidm1 Hidm2.
  introv Hnw1 Hnw2 Hny1 Hny2.

  assert (similar xs1 xs2 ds1 ds2). {
    eapply similar_lit_obj_defs; eauto.
    eapply similar_trm_lit_lit; eauto.
  }
  eapply similar_env_notin.

  - eapply similar_env_with_ie_push_notin; eauto.
    eapply similar_env_with_o_push.
    eapply similar_env_with_o_push; eauto.
    eapply similar_item_obj.
    eapply similar_deref_defs; eauto.
    eapply renv_idempotent_push; eauto.
    eapply renv_idempotent_push; eauto.

    eapply similar_open_rec.
    eapply similar_extend; eauto.
    eapply similar_extend; eauto.
    eapply similar_var_push. simpls. eapply similar_length in Hse. eauto.

    eapply similar_env_push; eauto.
    eapply similar_extend_notin. eauto. eauto. eauto.
    eapply similar_extend.
    eapply similar_var_push. simpls. eapply similar_length in Hse. eauto.
    rewrite from_list_cons; notin_solve.
    rewrite from_list_cons; notin_solve.
  - simpl_dom. notin_solve.
  - simpl_dom. notin_solve.
Qed.

Lemma similar_reduce_lit_renv xs1 xs2 w1 w2 y1 y2 e1 e2:
  similar xs1 xs2 e1 e2 ->
  y1 # e1 ->
  y2 # e2 ->
  w1 \notin (\{y1} \u from_list xs1) ->
  w2 \notin (\{y2} \u from_list xs2) ->
  similar (w1 :: y1 :: xs1) (w2 :: y2 :: xs2)  (e1 & w1 ~ y1) (e2 & w2 ~ y2).
Proof.
  introv Hse Hny1 Hny2 Hnw1 Hnw2.
  eapply similar_env_push; eauto.
  eapply similar_env_notin; eauto.
  eapply similar_env_with_o_push; eauto.
  eapply similar_extend.
  eapply similar_var_push. eapply similar_length in Hse. auto.
Qed.

Lemma similar_reduce_let_trm xs1 xs2 t11 t12 t21 t22:
  similar xs1 xs2 (trm_let t11 t12) (trm_let t21 t22) ->
  similar xs1 xs2 t11 t21.
Proof.
  introv.
  eapply similar_trm_let_1.
Qed.

Lemma similar_reduce_let_stack xs1 xs2 t11 t12 t21 t22 fs1 fs2:
  similar xs1 xs2 fs1 fs2 ->
  similar xs1 xs2 (trm_let t11 t12) (trm_let t21 t22) ->
  similar xs1 xs2 (stack_cons (frame_let t12) fs1) (stack_cons (frame_let t22) fs2).
Proof.
  introv Hsimfs Hsimt.
  eapply similar_stack_let; eauto.
  eapply similar_trm_let_2; eauto.
Qed.

(* Similar configurations reduce to similar configurations *)
(* A single object on the heap cannot relate to two object on the heap,
  because then on one side, only one of them would be modified *)
Lemma similar_preservation Ca1 Ca2 Cb1 Cb2 (xa xb: simlist):
  similar xa xb Ca1 Cb1 ->
  config_valid Ca1 ->
  config_valid Cb1 ->
  vars_conf_unique xa Ca1 ->
  vars_conf_unique xb Cb1 ->
  Ca1 ↦ Ca2 ->
  Cb1 ↦ Cb2 ->
  exists (ya yb: simlist),
  similar (ya ++ xa) (yb ++ xb) Ca2 Cb2 /\
  vars_conf_unique (ya ++ xa) Ca2 /\
  vars_conf_unique (yb ++ xb) Cb2 /\
  (config_fv_extension Ca1 Ca2) = (from_list ya) /\
  (config_fv_extension Cb1 Cb2) = (from_list yb).
Proof.
  introv Hs Hcva Hcvb Hsfa Hsfb Hstepa Hstepb.
  (* Unfold conf similarity and get conf parts *)
  unfold similar, similar_conf in *.
  destruct Cb1.
  destruct Cb2.
  (* Step cases *)
  inversions Hstepa.
  (* Get similarity of conf parts *)
  all: destruct_all.
  (* Unfold similarity of focus *)
  all:
    match goal with
    | H: (@similar trm similar_trm _ _ _ _) |- _ =>
        pose proof H as ?Hsimt;
        unfold similar, similar_trm, similar_by_subst in H;
        destruct_all
    end.
  (* Find common ancestor of focuses *)
  all: invert_subst_vars.

  (* Invert step of Cb1 *)
  all: autorewrite with simpl_subst_vars in Hstepb;
  inversions Hstepb.
  - (* trm_read *)
    exists (wVal :: nil)%list (wVal0::nil)%list.
    splits.
    * (* similarity *)
      simpls.
      invert_subst_vars.
      invert_subst_vars.
      autorewrite with simpl_subst_vars in Hsimt.
      splits.
      + (* term *)
        eapply similar_reduce_read_trm; eauto.
      + eapply similar_extend; eauto.
      + eapply similar_extend_notin.
        auto.
        destruct_all.
        notin_solve.
        destruct_all.
        notin_solve.
      + (* Destruct and invert to show that wObj is similar to wObj0 *)
        unfold vars_conf_unique in *.
        destruct_all.
        autorewrite with simpl_subst_vars in H14; eauto.
        sympls.
        eapply similar_reduce_read_renv; eauto.
        intro. eapply H16 in H17. eapply H; eauto.
        intro. eapply H14 in H17. eapply H10; eauto.
    * eapply vars_conf_unique_prepend1.
      rewrite H12; eauto. simpls. eauto.
      rewrite H12; eauto. repeat rewrite notin_union in H. destruct_all; simpls.  notin_solve.
       rewrite in_union. right. simpl_dom.  rewrite in_union. rewrite in_singleton. auto.
    * replace ( \{ wObj0}) with (fv (trm_read (avar_f wObj0) m)).
      eapply vars_conf_unique_prepend1.
      rewrite H10; eauto. autorewrite with simpl_subst_vars in Hsfb; eauto.
      simpls. repeat rewrite notin_union in H15. destruct_all; simpls.  notin_solve.
      rewrite in_union. right. simpl_dom. rewrite in_union.  rewrite in_singleton. auto.
      simpls. auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom. rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l. 
      rewrite remove_disjoint. auto.
      eapply disjoint_comm. eapply disjoint_singleton_r. notin_solve.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom. rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l. 
      rewrite remove_disjoint. auto.
      eapply disjoint_comm. eapply disjoint_singleton_r. notin_solve.
  - (* trm_write *)
    exists (@nil var) (@nil var).
    splits.
    *
      simpls.
      invert_subst_vars.
      invert_subst_vars.
      invert_subst_vars_in H9.
      invert_subst_vars_in H12.
      autorewrite with simpl_subst_vars in Hsimt.
      splits; eauto.
      + (* term *)
        autorewrite with simpl_subst_vars.
        eapply similar_reduce_write_trm; eauto.
      + (* heap *)
        unfold vars_conf_unique in *.
        destruct_all.
        eapply similar_reduce_write_heap; eauto.
    * eapply vars_conf_unique_prepend0. rewrite H11. rewrite H13. eauto.
    * replace ( \{ wObj0} \u \{ wArg0}) with (fv (trm_write (avar_f wObj0) m (avar_f wArg0))).
      eapply vars_conf_unique_prepend0.
      rewrite H9. rewrite H12. autorewrite with simpl_subst_vars in Hsfb; eauto.
      simpls. auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom.
      rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l.
      erewrite <- heap_update_dom by eauto.
      rewrite remove_same. apply eq_refl.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom.
      rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l.
      erewrite <- heap_update_dom by eauto.
      rewrite remove_same. apply eq_refl.
  - (* trm_call *)
    exists (@nil var) (@nil var).
    splits.
    *
      simpls.
      invert_subst_vars.
      invert_subst_vars.
      invert_subst_vars_in H9.
      invert_subst_vars_in H12.
      autorewrite with simpl_subst_vars in Hsimt.
      splits; eauto.
      + (* term *)
        eapply similar_reduce_call_trm; eauto.
    * eapply vars_conf_unique_prepend0. rewrite H11. rewrite H13. eauto.
    * replace ( \{ wObj0} \u \{ wArg0}) with (fv (trm_call a m a0) /[ x0 --> xb] ).
      eapply vars_conf_unique_prepend0. eauto.
      autorewrite with simpl_subst_vars. rewrite <- H9. rewrite <- H12. simpls. auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom.
      rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l.
      auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom.
      rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l.
      auto.
  - (* trm_apply *)
    exists (@nil var) (@nil var).
    splits.
    *
      simpls.
      invert_subst_vars.
      invert_subst_vars.
      invert_subst_vars_in H8.
      invert_subst_vars_in H9.
      autorewrite with simpl_subst_vars in Hsimt.
      splits; eauto.
      + (* term *)
        eapply similar_reduce_apply_trm; eauto.
    * eapply vars_conf_unique_prepend0. rewrite H11. rewrite H10. eauto.
    * replace ( \{ wFun0} \u \{ wArg0}) with (fv (trm_apply a a0) /[ x0 --> xb]).
      eapply vars_conf_unique_prepend0. eauto.
      autorewrite with simpl_subst_vars. rewrite <- H9. rewrite <- H8. simpls. auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom.
      rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l.
      auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom.
      rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l.
      auto.
  - (* lit_fun *)
    exists (w::y::(@nil var))%list (w0::y0::(@nil var))%list.
    splits.
    *
      simpls.
      invert_subst_vars.
      invert_subst_vars.
      autorewrite with simpl_subst_vars in Hsimt.
      splits; eauto.
      + (* term *)
        eapply similar_reduce_lit_trm; eauto.
      + (* stack *)
        eapply similar_extend; eauto.
        eapply similar_extend; eauto.
      + (* heap *)
        unfold vars_conf_unique in *.
        destruct_all.
        autorewrite with simpl_subst_vars in H14; eauto.
        sympls.
        eapply similar_reduce_fun_heap; eauto; notin_solve.
        intro. eapply H16 in H17. eapply H; eauto.
        intro. eapply H14 in H17. eapply H8; eauto.
      + (* renv *)
        unfold vars_conf_unique in *.
        destruct_all.
        autorewrite with simpl_subst_vars in H14; eauto.
        sympls.
        eapply similar_reduce_lit_renv; eauto.
        notin_solve.
        intro. eapply H16 in H17. eapply H11. rewrite in_union. left. eauto.
        notin_solve.
        intro. eapply H14 in H17. eapply H9. rewrite in_union. left. eauto.
    *
      inversions H.
      inversions H12.
      eapply vars_conf_unique_prepend2; eauto. rewrite H10. eauto.
      rewrite H10. notin_solve.
      simpl_dom. repeat rewrite in_union; rewrite ! in_singleton; eauto.
      rewrite H10. notin_solve.
      simpl_dom. repeat rewrite in_union; rewrite ! in_singleton; eauto.
    * replace ((fv T0 \u fv_trm t2) \u fv_trm (x /[ x0 --> xb])) with (fv (trm_lit l x) /[ x0 --> xb]).
      inversions H11.
      inversions H12.
      eapply vars_conf_unique_prepend2; eauto. sympls.
      autorewrite with simpl_subst_vars.
      rewrite <- H8. sympls. notin_solve.
      simpl_dom. repeat rewrite in_union; rewrite ! in_singleton; eauto.
      autorewrite with simpl_subst_vars.
      rewrite <- H8. sympls. notin_solve.
      simpl_dom. repeat rewrite in_union; rewrite ! in_singleton; eauto.
      autorewrite with simpl_subst_vars. rewrite <- H8. sympls. auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom. rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l. 
      rewrite ? remove_disjoint. apply union_comm.
      eapply disjoint_comm. eapply disjoint_singleton_r. notin_solve.
      eapply disjoint_comm. eapply disjoint_singleton_r. notin_solve.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom. rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l. 
      rewrite ? remove_disjoint. apply union_comm.
      eapply disjoint_comm. eapply disjoint_singleton_r. notin_solve.
      eapply disjoint_comm. eapply disjoint_singleton_r. notin_solve.
  - (* lit_fun vs lit_obj *)
    invert_subst_vars.
    autorewrite with simpl_subst_vars in H8. discriminate.
  - (* lit_fun vs lit_obj *)
    invert_subst_vars.
    autorewrite with simpl_subst_vars in H8. discriminate.
  - (* lit_obj *)
    exists (w::y::(@nil var))%list (w0::y0::(@nil var))%list.
    splits.
    *
      simpls.
      invert_subst_vars.
      invert_subst_vars.
      autorewrite with simpl_subst_vars in Hsimt.
      splits; eauto.
      + (* term *)
        eapply similar_reduce_lit_trm; eauto.
      + (* stack *)
        eapply similar_extend; eauto.
        eapply similar_extend; eauto.
      + (* heap *)
        assert(renv_idempotent e) as Him1.
        eapply config_valid_renv_idempotent; eauto.
        assert(renv_idempotent e1) as Him2.
        eapply config_valid_renv_idempotent; eauto.
        inversions Hcva.
        inversions Hcvb.
        unfold vars_conf_unique in *.
        destruct_all.
        autorewrite with simpl_subst_vars in H23; eauto.
        sympls.
        eapply similar_reduce_obj_heap; eauto 1; notin_solve.
        intro. eapply H25 in H26. eapply H11; rewrite in_union; eauto.
        intro. eapply H23 in H26. eapply H9; rewrite in_union; eauto.
        intro. eapply H25 in H26. eapply H; eauto.
        intro. eapply H23 in H26. eapply H8; eauto.
      + (* renv *)
        unfold vars_conf_unique in *.
        destruct_all.
        autorewrite with simpl_subst_vars in H14; eauto.
        sympls.
        eapply similar_reduce_lit_renv; eauto.
        notin_solve.
        intro. eapply H16 in H17. eapply H11. rewrite in_union. left. eauto.
        notin_solve.
        intro. eapply H14 in H17. eapply H9. rewrite in_union. left. eauto.
    *
      inversions H.
      inversions H12.
      eapply vars_conf_unique_prepend2; eauto. rewrite H10. eauto.
      rewrite H10. notin_solve.
      simpl_dom. repeat rewrite in_union; rewrite ! in_singleton; eauto.
      rewrite H10. notin_solve.
      simpl_dom. repeat rewrite in_union; rewrite ! in_singleton; eauto.
    * replace ((fv T0 \u fv_defs ds0) \u fv_trm (x /[ x0 --> xb])) with (fv (trm_lit l x) /[ x0 --> xb]).
      inversions H11.
      inversions H12.
      eapply vars_conf_unique_prepend2; eauto. sympls.
      autorewrite with simpl_subst_vars.
      rewrite <- H8. sympls. notin_solve.
      simpl_dom. repeat rewrite in_union; rewrite ! in_singleton; eauto.
      autorewrite with simpl_subst_vars.
      rewrite <- H8. sympls. notin_solve.
      simpl_dom. repeat rewrite in_union; rewrite ! in_singleton; eauto.
      autorewrite with simpl_subst_vars. rewrite <- H8. sympls. auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom. rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l. 
      rewrite ? remove_disjoint. apply union_comm.
      eapply disjoint_comm. eapply disjoint_singleton_r. notin_solve.
      eapply disjoint_comm. eapply disjoint_singleton_r. notin_solve.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom. rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l. 
      rewrite ? remove_disjoint. apply union_comm.
      eapply disjoint_comm. eapply disjoint_singleton_r. notin_solve.
      eapply disjoint_comm. eapply disjoint_singleton_r. notin_solve.
  - (* trm_var *)
    exists (@nil var) (@nil var).
    splits.
    *
      simpls.
      invert_subst_vars.
      invert_subst_vars.
      autorewrite with simpl_subst_vars in Hsimt.
      splits; eauto.
      + (* term *)
        eapply similar_reduce_var_trm; eauto.
      + (* stack *)
        eapply similar_stack_let_stack; eauto.
    * eapply vars_conf_unique_prepend0; eauto. rewrite H8. auto.
    * replace (\{x0} ) with (fv (trm_var (avar_f x0 ))).
      replace (fv t \u fv s0) with (fv (stack_cons (frame_let t) s0)).
      eapply vars_conf_unique_prepend0; eauto. rewrite H7. autorewrite with simpl_subst_vars in Hsfb. eapply Hsfb.
      sympls. auto.
      sympls. auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom.
      rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l.
      auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom.
      rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l.
      auto.
  - (* trm_let *)
    exists (@nil var) (@nil var).
    splits.
    *
      simpls.
      autorewrite with simpl_subst_vars in Hsimt.
      splits; eauto.
      + (* term *) eapply similar_reduce_let_trm; eauto.
      + (* stack *) eapply similar_reduce_let_stack; eauto.
    * eapply vars_conf_unique_prepend0; eauto.
    * replace ((fv_trm (x1 /[ x0 --> xb]) \u fv_trm (x2 /[ x0 --> xb]))) with (fv  (trm_let (x1 /[ x0 --> xb]) (x2 /[ x0 --> xb]))).
      eapply vars_conf_unique_prepend0; eauto. autorewrite with simpl_subst_vars in Hsfb. eapply Hsfb.
      sympls. auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom.
      rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l.
      auto.
    * unfold config_fv_extension. rewrite ? from_list_cons, from_list_nil. simpl. simpl_dom.
      rewrite ? union_remove, ? remove_same, ? union_empty_r, ? union_empty_l.
      auto.
Qed.


Theorem similar_conf_1to1_preservation Ca1 Ca2 Cb1 Cb2:
  similar_conf_1to1 Ca1 Cb1 ->
  config_valid Ca1 ->
  config_valid Cb1 ->
  Ca1 ↦ Ca2 ->
  Cb1 ↦ Cb2 ->
  similar_conf_1to1 Ca2 Cb2.
Proof.
  introv HsiC Hva1 Hvb1 Hra Hrb.
  unfold similar_conf_1to1 in *.
  destruct_all.
  forwards: similar_preservation; eauto.
  destructc H3 as ya yb Hsi2 Hua2 Hub2 ?H.
  exists (ya++x)%list (yb ++ x0)%list.
  splits; eauto.
  eapply similar_conf_length in Hsi2.
  eauto.
Qed.
