Set Implicit Arguments.

Require Import Coq.Init.Datatypes.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import
        AbstractSyntax GeneralTyping Generalization Equivalence
        Similarity ContextSimilarity SyntaxSimilarity OperationalSemantics HeapUpdate Deref Defs ConfigValid TypingSimilarity.
Require Import UnboundSubstitution.
Require Import ConfigTyping TFree TFreeSubst.

(** * Similarity of typed configurations *)

Existing Instance similar_ctx.

Lemma similarity_all_tfree_k xs1 xs2 G1 G2 (t1 t2: trm) vk:
  fresh \{} (length xs1) xs1 ->
  ctx_ok G2 ->
  similar xs1 xs2 G1 G2->
  similar xs1 xs2 t1 t2->
  all_tfree_k G1 t1 vk ->
  all_tfree_k G2 t2 vk.
Proof.
  introv Hfr1 HokG2 HsiG HsiT Htf.
  unfold all_tfree_k in *.
  intros.
  unfold similar, similar_trm, similar_by_subst in HsiT.
  destruct_all; subst.
  eapply tfree_subst_case in H.
  destruct_all; subst.
  forwards: Htf (x1 /[ x0 --> xs1]).
  eapply tfree_subst; eauto.

  pose proof (tfree_in_fv_trm H4).

  assert (similar xs1 xs2 (x1 /[ x0 --> xs1]) (x1 /[ x0 --> xs2])).
    unfold similar, similar_var, similar_by_subst. exists x1 x0; splits; eauto. fresh_solve.
    eapply fresh_fv_subst_vars_in; eauto. eapply subset_refl.
    eapply fresh_fv_subst_vars_in; eauto. eapply subset_refl.
  unfold fv, VarFreeVar. eapply subset_singleton. eauto.

  eapply similarity_ctx_k; eauto.
Qed.


Section WithTypingMode.
Context {typing_mode: Mode}.

Lemma typing_similarity_stack xs1 xs2 G1 G2 e1 e2 (x1 x2: stack) (U1 U2: typ) (T1 T2: typ):
  ok e1 ->
  ctx_ok G2 ->
  disjoint (dom e1) (fv_env e1) ->
  dom e1 \c ctx_dom G1 ->
  dom e2 \c ctx_dom G2 ->
  ctx_dom G1 \c (from_list xs1) ->
  fresh \{} (length xs1) xs1 ->
  fresh \{} (length xs2) xs2 ->
  similar xs1 xs2 e1 e2 ->
  similar xs1 xs2 x1 x2 ->
  similar xs1 xs2 T1 T2 ->
  similar xs1 xs2 U1 U2 ->
  similar xs1 xs2 G1 G2->
  ty_stack G1 e1 U1 x1 T1 ->
  ty_stack G2 e2 U2 x2 T2.
Proof.
  introv Hoke HokG2 Hne1 Hse1G1 Hse2G2 HsG1xs1. introv Hfrxs1 Hfrxs2 Hsie Hsix HsiT HsiU HsiG.
  intros.  reverts U2 G2. reverts xs1 xs2 T2 x2.
  induction H; intros.
  all: similar_syntax_inverts.
  - eapply ty_stack_nil.
    eapply typing_similarity; eauto.
  -

    pose proof (similar_exists_premises T Hfrxs1 Hfrxs2 HsG1xs1 Hse1G1 Hse2G2 Hsie HsiG) as HT.
    destructc HT as ys1 ys2 Ty Hfrys1 Hfrys2 HsG1ys1 Hsie2 HsiG2 HsiT2 Hlys.

    eapply ty_stack_let with (T:=Ty) (L:=L \u ctx_dom G \u ctx_dom G2 \u fv_renv e \u fv_renv e2 \u from_list (ys1++xs1) \u from_list (ys2++xs2)).
    eapply (IHty_stack Hoke Hne1 Hse1G1 (ys1 ++ xs1)%list HsG1ys1 Hfrys1 (ys2 ++ xs2)%list); eauto.
    eapply similar_out_append; eauto.
    eapply similar_out_append; eauto.
    intros.
    assert (similar (ys1 ++ xs1) (ys2 ++ xs2) S U2) as HsiS2.
    eapply similar_out_append; eauto.
    lets HT: (similar_push_cons_premises (z:=x) v_let Hfrys1 Hfrys2 HsG1ys1 Hse1G1 Hse2G2 HokG2 HsiS2 Hsie2 HsiG2).
    destructc HT as Hfrzs1 Hfrzs2 HsG1zs1 Hse1Gy Hse2Gy HokGy Hsie3 HsiG3.
    eauto.
    assert (x \notin L) as HnL by eauto.
    eapply typing_similarity with (xs3:= (x::ys1++xs1)%list) (xs4:= (x::ys2++xs2)%list) (x2:=open x t )(T1:=T); eauto.
    eapply similar_open_rec_push; eauto.
    eapply similar_out_append; eauto.
    eapply similar_extend; eauto.
Qed.

End WithTypingMode.
