(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import Types TypeInduction.
Require Import Logctx.

(** * Syntactic [/\]-context *)

Inductive andctx: Set :=
| andctx_nil: andctx (* The hole *)
| andctx_l (oc: andctx) (T:typ): andctx (* Hole on the left of a type *)
| andctx_r (T:typ) (oc: andctx): andctx (* Hole on the right of a type *)
.

(* Filling the hole in a union context *)
Fixpoint andctx_typ (oc: andctx) (T:typ): typ := 
  match oc with
  | andctx_nil => T
  | andctx_l ocl Tr => (typ_and (andctx_typ ocl T) Tr)
  | andctx_r Tl ocr => (typ_and Tl (andctx_typ ocr T))
  end.

Fixpoint andctx_logctx (oc: andctx): logctx := 
  match oc with
  | andctx_nil => logctx_nil
  | andctx_l ocl Tr => logctx_andl (andctx_logctx ocl) Tr
  | andctx_r Tl ocr => logctx_andr Tl (andctx_logctx ocr)
  end.

Definition logctx_isand oc := (exists oco, oc = andctx_logctx oco).

Lemma andctx_logctx_typ oc T:
  (andctx_typ oc T) = (logctx_typ (andctx_logctx oc) T).
Proof.
  induction oc; simpls; try rewrite IHoc; eauto.
Qed.
Lemma andctx_logctx_isand oc:
  logctx_isand (andctx_logctx oc).
Proof.
  induction oc; hnf; simpls.
  - exists andctx_nil; simpls; eauto.
  - hnf in IHoc; destruct IHoc. exists (andctx_l x T); simpls; rewrite H; eauto.
  - hnf in IHoc; destruct IHoc. exists (andctx_r T x); simpls; rewrite H; eauto.
Qed.