Set Implicit Arguments.

Require Import Coq.Bool.Bool.
Require Import LibExtra DotTactics SyntaxClasses.
Require Import TransformCollect Vars.

(** * Lemmata about free variables in renv *)

Section EnvLemmas.
Context {A: Set}.
Lemma env_push_values (e: env A) v1 v2:
  values (e & (v1 ~ v2)) = cons v2 (values e).
Proof.
  rewrite values_def.
  rewrite concat_def.
  rewrite single_def.
  compute.
  auto.
Qed.

Context {AF: FreeVar A}.

Lemma env_push_fv (e: env A) v1 v2:
  fv_env (e & (v1 ~ v2)) = fv_env e \u fv v2.
Proof.
  unfold fv_env.
  rewrite env_push_values.
  compute.
  auto.
Qed.

Lemma env_concat_fv(e1 e2: env A):
  fv_env (e1 & e2) = fv_env e1 \u fv_env e2.
Proof.
  induction e2 using env_ind.
  -  rewrite concat_empty_r. rewrite empty_def. compute. rewrite values_def. compute. eauto.
  - rewrite concat_assoc.
    rewrite env_push_fv.
    rewrite env_push_fv.
    rewrite IHe2.
    rewrite union_assoc. auto.
Qed.

Lemma env_push_in_fv_l v (e: env A) v1 v2:
  v \in fv_env e ->
  v \in fv_env (e & (v1 ~ v2)).
Proof.
  rewrite env_push_fv.
  rewrite in_union.
  left. auto.
Qed.

Lemma env_push_obj_in_fv_r v1 v2 x e:
  v2 \in fv x ->
  v2 \in fv_env (e & (v1 ~ x)).
Proof.
  rewrite env_push_fv.
  rewrite in_union.
  right.
  unfold fv.
  unfold fv_in_values.
  unfold fv in H.
  auto.
Qed.


Lemma env_binds_fv_subset x (a: A) S:
  binds x a S ->
  fv a \c fv_env S.
Proof.
  intros.
  induction S using env_ind.
  - false.
    eapply binds_empty_inv; eauto.
  -
    destruct (binds_push_inv H) as [[? ?] | [? ?]]; subst.
    + intro. intro. rewrite env_push_fv. rewrite in_union. auto.
    + eapply subset_trans; eauto. intro. intro.  rewrite env_push_fv. rewrite in_union. auto.
Qed.

End EnvLemmas.