(** remove printing ~ *)

Set Implicit Arguments.

Require Import Coq.Strings.String.

(** ** Labels *)

(** Labels are strings. *)
Definition typ_normal_label: Set := string.
Definition fld_label: Set := string.
Definition met_label: Set := string.

Inductive typ_refcap_label: Set :=
(** Reference capability type name for mutability ([M]) *)
| typ_refcap_mutability
.

(* Part of Figure 3 in "Reference mutability for DOT" *)
(* Part of Figure 1 in "Pure methods for roDOT" *)
(** Type member lables:
- [typ_label_normal] represents a Normal type member ([A])
- [typ_label_refcap] represents a Reference capability ([M]) *)
Inductive typ_label: Set :=
| typ_label_normal: typ_normal_label -> typ_label
| typ_label_refcap: typ_refcap_label -> typ_label
.
Notation typ_label_mutability := (typ_label_refcap typ_refcap_mutability) (only parsing).

(** *** Term and type members
 Type member labels ([A], [B], [M]),
 field member labels ([a]),
 method member labels ([m]).  *)
Inductive label: Set :=
| label_typ: typ_label -> label
| label_fld: fld_label -> label
| label_met: met_label -> label
.

(** Some syntax will contain labels *)
Class LabelOf (A : Set) := label_of : A -> label.
Arguments label_of _ _ !_ /.
