(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import Types.
Require Import TypeInduction.

(** * Syntactic context forming a type with unions and intersections *)

Inductive logctx: Set :=
| logctx_nil: logctx (* The hole *)
| logctx_orl (oc: logctx) (T:typ): logctx (* Hole on the left of a type *)
| logctx_orr (T:typ) (oc: logctx): logctx (* Hole on the right of a type *)
| logctx_andl (oc: logctx) (T:typ): logctx (* Hole on the left of a type *)
| logctx_andr (T:typ) (oc: logctx): logctx (* Hole on the right of a type *)
.

Fixpoint logctx_typ (oc: logctx) (T:typ): typ := 
  match oc with
  | logctx_nil => T
  | logctx_orl ocl Tr => (typ_or (logctx_typ ocl T) Tr)
  | logctx_orr Tl ocr => (typ_or Tl (logctx_typ ocr T))
  | logctx_andl ocl Tr => (typ_and (logctx_typ ocl T) Tr)
  | logctx_andr Tl ocr => (typ_and Tl (logctx_typ ocr T))
  end.

Fixpoint logctx_extend (oc oci: logctx): logctx := 
  match oc with
  | logctx_nil => oci
  | logctx_orl ocl Tr => logctx_orl (logctx_extend ocl oci) Tr
  | logctx_orr Tl ocr => logctx_orr Tl (logctx_extend ocr oci)
  | logctx_andl ocl Tr => logctx_andl (logctx_extend ocl oci) Tr
  | logctx_andr Tl ocr => logctx_andr Tl (logctx_extend ocr oci)
  end.

Lemma logctx_extend_assoc oc1 oc2 oc3:
  logctx_extend (logctx_extend oc1 oc2 ) oc3 = logctx_extend oc1 (logctx_extend oc2 oc3).
Proof.
  reverts oc2 oc3.
  induction oc1; intros; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
Qed.

Lemma logctx_extend_orl oc Tl Tr:
  logctx_typ oc (typ_or Tl Tr) = logctx_typ (logctx_extend oc (logctx_orl logctx_nil Tr) ) Tl.
Proof.
  induction oc; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
Qed.
Lemma logctx_extend_orr oc Tl Tr:
  logctx_typ oc (typ_or Tl Tr) = logctx_typ (logctx_extend oc (logctx_orr Tl logctx_nil) ) Tr.
Proof.
  induction oc; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
Qed.
Lemma logctx_extend_andl oc Tl Tr:
  logctx_typ oc (typ_and Tl Tr) = logctx_typ (logctx_extend oc (logctx_andl logctx_nil Tr) ) Tl.
Proof.
  induction oc; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
Qed.
Lemma logctx_extend_andr oc Tl Tr:
  logctx_typ oc (typ_and Tl Tr) = logctx_typ (logctx_extend oc (logctx_andr Tl logctx_nil) ) Tr.
Proof.
  induction oc; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
Qed.
Lemma logctx_extend_nil_l oc:
  logctx_extend oc logctx_nil = oc.
Proof.
  induction oc; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
Qed.
Lemma logctx_fold_andr oc Tl Tr:
  typ_and Tl (logctx_typ oc Tr)  = logctx_typ (logctx_andr Tl oc) Tr.
Proof.
  simpls; eauto.
Qed.
Lemma logctx_fold_andl oc Tl Tr:
  typ_and (logctx_typ oc Tl) Tr = logctx_typ (logctx_andl oc Tr) Tl.
Proof.
  simpls; eauto.
Qed.
Lemma logctx_fold_orr oc Tl Tr:
  typ_or Tl (logctx_typ oc Tr)  = logctx_typ (logctx_orr Tl oc) Tr.
Proof.
  simpls; eauto.
Qed.
Lemma logctx_fold_orl oc Tl Tr:
  typ_or (logctx_typ oc Tl) Tr = logctx_typ (logctx_orl oc Tr) Tl.
Proof.
  simpls; eauto.
Qed.
Lemma logctx_extend_typ oc1 oc2 T:
  logctx_typ (logctx_extend oc1 oc2 ) T = logctx_typ oc1 (logctx_typ oc2 T).
Proof.
  induction oc1; simpls.
  - auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
  - fequal. auto.
Qed.

Fixpoint logctx_height (oc: logctx): nat := 
  match oc with
  | logctx_nil => 0
  | logctx_orl ocl Tr => S (logctx_height ocl)
  | logctx_orr Tl ocr => S (logctx_height ocr)
  | logctx_andl ocl Tr => S (logctx_height ocl)
  | logctx_andr Tl ocr => S (logctx_height ocr)
  end.

Lemma logctx_extend_height oc1 oc2:
  logctx_height(logctx_extend oc1 oc2) =   logctx_height oc1 + logctx_height oc2.
Proof.
  intros.
  induction oc1;
  simpls; eauto.
Qed.
Lemma logctx_reverse_destruct oc:
  oc = logctx_nil \/ (
    exists ocl T,
      oc = logctx_extend ocl (logctx_orl logctx_nil T)
  ) \/ (
    exists ocr T,
      oc = logctx_extend ocr (logctx_orr T logctx_nil)
  ) \/ (
    exists ocl T,
      oc = logctx_extend ocl (logctx_andl logctx_nil T)
  ) \/ (
    exists ocr T,
      oc = logctx_extend ocr (logctx_andr T logctx_nil)
  ).
Proof.
  induction oc; destruct_all; subst; eauto.
  -  right. left. exists logctx_nil T. simpls. eauto.
  -  right. left. exists (logctx_orl x T). simpls. eauto.
  -  right. right. left. exists (logctx_orl x T). simpls. eauto.
  -  right. right. right. left. exists (logctx_orl x T). simpls. eauto.
  -  right. right. right. right. exists (logctx_orl x T). simpls. eauto.

  -  right. right. left. exists logctx_nil T. simpls. eauto.
  -  right. left. exists (logctx_orr T x). simpls. eauto.
  -  right. right. left. exists (logctx_orr T x). simpls. eauto.
  -  right. right. right. left. exists (logctx_orr T x). simpls. eauto.
  -  right. right. right. right. exists (logctx_orr T x). simpls. eauto.

  -  right. right. right. left. exists logctx_nil T. simpls. eauto.
  -  right. left. exists (logctx_andl x T). simpls. eauto.
  -  right. right. left. exists (logctx_andl x T). simpls. eauto.
  -  right. right. right. left. exists (logctx_andl x T). simpls. eauto.
  -  right. right. right. right. exists (logctx_andl x T). simpls. eauto.

  -  right. right. right. right. exists logctx_nil T. simpls. eauto.
  -  right. left. exists (logctx_andr T x). simpls. eauto.
  -  right. right. left. exists (logctx_andr T x). simpls. eauto.
  -  right. right. right. left. exists (logctx_andr T x). simpls. eauto.
  -  right. right. right. right. exists (logctx_andr T x). simpls. eauto.
Qed.
(* Induction from below *)
Lemma logctx_reverse_ind (P : logctx -> Prop):
  P logctx_nil ->
  (forall oc : logctx, P oc -> forall T : typ, P (logctx_extend oc (logctx_orl logctx_nil T ))) ->
  (forall (T : typ) (oc : logctx), P oc -> P (logctx_extend oc (logctx_orr T logctx_nil))) ->
  (forall oc : logctx, P oc -> forall T : typ, P (logctx_extend oc (logctx_andl logctx_nil T ))) ->
  (forall (T : typ) (oc : logctx), P oc -> P (logctx_extend oc (logctx_andr T logctx_nil))) ->
  forall o : logctx, P o.
Proof.
  intros.
  remember (logctx_height o) as i. 
  reverts o.
  induction i; intros; subst.
  -
  destruct o; inversions Heqi. eauto.
  -
  forwards: logctx_reverse_destruct o.
  destruct_all; subst; simpls.
  + inversions Heqi.
  + eapply H0. eapply IHi. rewrite logctx_extend_height in Heqi. simpls. rewrite Nat.add_1_r in Heqi. inversions Heqi. auto.
  + eapply H1. eapply IHi. rewrite logctx_extend_height in Heqi. simpls. rewrite Nat.add_1_r in Heqi. inversions Heqi. auto.
  + eapply H2. eapply IHi. rewrite logctx_extend_height in Heqi. simpls. rewrite Nat.add_1_r in Heqi. inversions Heqi. auto.
  + eapply H3. eapply IHi. rewrite logctx_extend_height in Heqi. simpls. rewrite Nat.add_1_r in Heqi. inversions Heqi. auto.
Qed.

Definition logctx_typ_separate T1 T2 := ((forall oc2, T1 <> logctx_typ oc2 T2) /\ (forall oc1, T2 <> logctx_typ oc1 T1)).
Lemma logctx_typ_separate_symm T1 T2:
  logctx_typ_separate T1 T2 ->
  logctx_typ_separate T2 T1.
Proof.
  intros.
  unfold logctx_typ_separate in *.
  intros.
  forwards H1 H2: H; eauto.
Qed.
Lemma logctx_typ_separate_bnd_bot T:
  logctx_typ_separate (typ_bnd T) typ_bot.
Proof.
  unfold logctx_typ_separate in *.
  intros; split.
  - destruct oc2; simpls; discriminate.
  - destruct oc1; simpls; discriminate.
Qed.
Lemma logctx_typ_separate_sel_bot x A y:
  logctx_typ_separate (typ_sel x A y) typ_bot.
Proof.
  unfold logctx_typ_separate in *.
  intros; split.
  - destruct oc2; simpls; discriminate.
  - destruct oc1; simpls; discriminate.
Qed.
Lemma logctx_typ_separate_bnd_nocap T:
  logctx_typ_separate (typ_bnd T) typ_nocap.
Proof.
  unfold logctx_typ_separate in *.
  intros; split.
  - destruct oc2; simpls; discriminate.
  - destruct oc1; simpls; discriminate.
Qed.
Lemma logctx_typ_separate_sel_nocap x A y:
  logctx_typ_separate (typ_sel x A y) typ_nocap.
Proof.
  unfold logctx_typ_separate in *.
  intros; split.
  - destruct oc2; simpls; discriminate.
  - destruct oc1; simpls; discriminate.
Qed.
(*
Lemma logctx_typ_separate_self T oc1 oc2:
  logctx_typ_separate (logctx_typ oc1 T) (logctx_typ oc2 T) ->
  False.
Proof.
  intros.
  hnf in *. destruct H.
Qed.*)

(* Follow the same path, with possibly different types *)
Inductive logctx_follow: logctx->logctx->Prop :=
| logctx_follow_nil :
  logctx_follow logctx_nil logctx_nil
| logctx_follow_orl oc1 oc2 Tr1 Tr2:
  logctx_follow oc1 oc2 -> 
  logctx_follow (logctx_orl oc1 Tr1) (logctx_orl oc2 Tr2)
| logctx_follow_orr oc1 oc2 Tl1 Tl2: (* Common prefix right *)
  logctx_follow oc1 oc2 -> 
  logctx_follow (logctx_orr Tl1 oc1) (logctx_orr Tl2 oc2)
| logctx_follow_andl oc1 oc2 Tr1 Tr2: (* Common prefix left *)
  logctx_follow oc1 oc2 -> 
  logctx_follow (logctx_andl oc1 Tr1) (logctx_andl oc2 Tr2)
| logctx_follow_andr oc1 oc2 Tl1 Tl2: (* Common prefix right *)
  logctx_follow oc1 oc2 -> 
  logctx_follow (logctx_andr Tl1 oc1) (logctx_andr Tl2 oc2)
.
Hint Constructors logctx_follow: core.

Lemma logctx_follow_refl oc:
  logctx_follow oc oc.
Proof.
  induction oc; eauto.
Qed.
Hint Resolve logctx_follow_refl: core.

Lemma logctx_follow_symm oc1 oc2:
  logctx_follow oc1 oc2 ->
  logctx_follow oc2 oc1.
Proof.
  introv Hfol. reverts oc2.
  induction oc1; intros; inversions Hfol; eauto.
Qed.

Lemma logctx_follow_trans oc1 oc2 oc3:
  logctx_follow oc1 oc2 ->
  logctx_follow oc2 oc3 ->
  logctx_follow oc1 oc3.
Proof.
  introv Hf1 Hf2. reverts oc3.
  induction Hf1; intros; inversions Hf2; eauto.
Qed.

(* Not prefixes *)
Inductive logctx_separate: logctx->logctx->Prop :=
| logctx_separate_orl_e oc1 oc2 Tr: (* Common prefix left *)
  logctx_separate oc1 oc2 -> 
  logctx_separate (logctx_orl oc1 Tr) (logctx_orl oc2 Tr)
| logctx_separate_orr_e oc1 oc2 Tl: (* Common prefix right *)
  logctx_separate oc1 oc2 -> 
  logctx_separate (logctx_orr Tl oc1) (logctx_orr Tl oc2)
| logctx_separate_andl_e oc1 oc2 Tr: (* Common prefix left *)
  logctx_separate oc1 oc2 -> 
  logctx_separate (logctx_andl oc1 Tr) (logctx_andl oc2 Tr)
| logctx_separate_andr_e oc1 oc2 Tl: (* Common prefix right *)
  logctx_separate oc1 oc2 -> 
  logctx_separate (logctx_andr Tl oc1) (logctx_andr Tl oc2)
| logctx_separate_orl_n oc1 oc2 Tr:
  (forall oc3 Tr3, oc2 <> (logctx_orl oc3 Tr3) /\ oc2 <> logctx_nil) ->
  logctx_separate (logctx_orl oc1 Tr) oc2
| logctx_separate_orr_n oc1 oc2 Tl:
  (forall oc3 Tl3, oc2 <> (logctx_orr Tl3 oc3) /\ oc2 <> logctx_nil) ->
  logctx_separate (logctx_orr Tl oc1) oc2
| logctx_separate_andl_n oc1 oc2 Tr:
  (forall oc3 Tr3, oc2 <> (logctx_andl oc3 Tr3) /\ oc2 <> logctx_nil) ->
  logctx_separate (logctx_andl oc1 Tr) oc2
| logctx_separate_andr_n oc1 oc2 Tl:
  (forall oc3 Tl3, oc2 <> (logctx_andr Tl3 oc3) /\ oc2 <> logctx_nil) ->
  logctx_separate (logctx_andr Tl oc1) oc2
.

Lemma logctx_separate_symm oc1 oc2:
  logctx_separate oc1 oc2 ->
  logctx_separate oc2 oc1.
Proof.
  introv Hse.
  induction Hse.
  eapply logctx_separate_orl_e; eauto.
  eapply logctx_separate_orr_e; eauto.
  eapply logctx_separate_andl_e; eauto.
  eapply logctx_separate_andr_e; eauto.
  all: destruct oc2.

  forwards H1 H2: (H oc1 Tr); eauto. contradiction.
  forwards H1 H2: (H oc2 T); eauto. contradiction.
  eapply logctx_separate_orr_n; intros; split; discriminate.
  eapply logctx_separate_andl_n; intros; split; discriminate.
  eapply logctx_separate_andr_n; intros; split; discriminate.

  forwards H1 H2: (H oc1 Tl); eauto. contradiction.
  eapply logctx_separate_orl_n; intros; split; discriminate.
  forwards H1 H2: (H oc2 T); eauto. contradiction.
  eapply logctx_separate_andl_n; intros; split; discriminate.
  eapply logctx_separate_andr_n; intros; split; discriminate.

  forwards H1 H2: (H oc1 Tr); eauto. contradiction.
  eapply logctx_separate_orl_n; intros; split; discriminate.
  eapply logctx_separate_orr_n; intros; split; discriminate.
  forwards H1 H2: (H oc2 T); eauto. contradiction.
  eapply logctx_separate_andr_n; intros; split; discriminate.

  forwards H1 H2: (H oc1 Tl); eauto. contradiction.
  eapply logctx_separate_orl_n; intros; split; discriminate.
  eapply logctx_separate_orr_n; intros; split; discriminate.
  eapply logctx_separate_andl_n; intros; split; discriminate.
  forwards H1 H2: (H oc2 T); eauto. contradiction.
Qed.

Lemma logctx_intersect_cases ocS S ocT T:
  logctx_typ ocS S = logctx_typ ocT T ->
  (ocS = ocT /\ S = T) \/
  (exists ocST, ocT = logctx_extend ocS ocST /\ S = logctx_typ ocST T) \/
  (exists ocTS, ocS = logctx_extend ocT ocTS /\ T = logctx_typ ocTS S) \/
  logctx_separate ocS ocT.
Proof.
  introv Heq.
  reverts ocS S.
  induction ocT; intros; simpls.
  - right. right. left. exists ocS; eauto.
  - destruct ocS; simpls; subst; try discriminate.
    + right. left.
      forwards HocST: IHocT; eauto.
    + (* Agreeing *) 
      inversions Heq.
      forwards HocST: IHocT; eauto.
      destruct_all; subst; eauto 6 using logctx_separate_orl_e.
    + (* Disagreeing *)
      right. right. right.
      eapply logctx_separate_orr_n; intros; split; discriminate.
  - destruct ocS; simpls; subst; try discriminate.
    + right. left.
      forwards HocST: IHocT; eauto.
    + (* Disagreeing *)
      right. right. right.
      eapply logctx_separate_orl_n; intros; split; discriminate.
    + (* Agreeing *)
      inversions Heq.
      forwards HocST: IHocT; eauto.
      destruct_all; subst; eauto 6 using logctx_separate_orr_e.
  - destruct ocS; simpls; subst; try discriminate.
    + right. left.
      forwards HocST: IHocT; eauto.
    + (* Agreeing *) 
      inversions Heq.
      forwards HocST: IHocT; eauto.
      destruct_all; subst; eauto 6 using logctx_separate_andl_e.
    + (* Disagreeing *)
      right. right. right.
      eapply logctx_separate_andr_n; intros; split; discriminate.
  - destruct ocS; simpls; subst; try discriminate.
    + right. left.
      forwards HocST: IHocT; eauto.
    + (* Disagreeing *)
      right. right. right.
      eapply logctx_separate_andl_n; intros; split; discriminate.
    + (* Agreeing *)
      inversions Heq.
      forwards HocST: IHocT; eauto.
      destruct_all; subst; eauto 6 using logctx_separate_andr_e.
Qed.

Lemma logctx_intersect_cases_nonnil ocS S ocT T:
  logctx_typ ocS S = logctx_typ ocT T ->
  (ocS = ocT /\ S = T) \/
  (exists ocST, ocT = logctx_extend ocS ocST /\ S = logctx_typ ocST T /\ ocST <> logctx_nil) \/
  (exists ocTS, ocS = logctx_extend ocT ocTS /\ T = logctx_typ ocTS S /\ ocTS <> logctx_nil) \/
  logctx_separate ocS ocT.
Proof.
  introv Heq.
  forwards: logctx_intersect_cases Heq.
  destruct_all; eauto.
  -
  assert  (x = logctx_nil \/ x <> logctx_nil) as Hx. {
    destruct x; auto; right; discriminate.
  }
  destruct Hx; subst; simpls; rewrite ? logctx_extend_nil_l; auto ; right; left; eauto.
  -
  assert  (x = logctx_nil \/ x <> logctx_nil) as Hx. {
    destruct x; auto; right; discriminate.
  }
  destruct Hx; subst; simpls; rewrite ? logctx_extend_nil_l; auto ; right;  right ;left; eauto.
Qed.


Lemma logctx_intersect_separate_case ocS S ocT T:
  logctx_typ_separate S T ->
  logctx_typ ocS S = logctx_typ ocT T ->
  logctx_separate ocS ocT.
Proof.
  intros.
  forwards: logctx_intersect_cases ocS S ocT T; eauto.
  destruct_all; subst; hnf in H; destruct_all.
  - forwards: H logctx_nil; simpls; contradiction.
  - forwards: H x; contradiction.
  - forwards: H1 x; contradiction.
  - eauto.
Qed.

Lemma logctx_separate_cross_r ocS S ocT T:
  logctx_typ ocS S = logctx_typ ocT T ->
  logctx_separate ocS ocT ->
  (forall S1, exists ocS1T,
    logctx_typ ocS S1 = logctx_typ ocS1T T  /\
    (forall T1, exists ocST1,
    logctx_typ ocST1 S = logctx_typ ocT T1 /\
    logctx_typ ocST1 S1 = logctx_typ ocS1T T1 /\
    logctx_follow ocST1 ocS) /\
    logctx_follow ocT ocS1T
  ).
Proof.
  introv Heq Hse.
  reverts S T.
  induction Hse; intros; simpls; inversions Heq.
  + (* Agreeing *)
    forwards ocS1T HocS1T HexST1 HfolT: (IHHse S T H0 S1); eauto.
    exists (logctx_orl ocS1T Tr).
    splits. rewrite HocS1T. auto.
    intros.
    forwards ocST1 HocST1 HocS1T1 HfolS: (HexST1 T1).
    exists (logctx_orl ocST1 Tr).
    simpls.
    rewrite HocST1.
    rewrite HocS1T1.
    auto.
    auto.
  + (* Agreeing *)
    forwards ocS1T HocS1T  HexST1 HfolT: (IHHse S T H0 S1); eauto.
    exists (logctx_orr Tl ocS1T).
    splits. rewrite HocS1T. auto.
    intros.
    forwards ocST1 HocST1 HocS1T1 HfolS: (HexST1 T1).
    exists (logctx_orr Tl ocST1).
    simpls.
    rewrite HocST1.
    rewrite HocS1T1.
    auto.
    auto.
  + (* Agreeing *)
    forwards ocS1T HocS1T  HexST1 HfolT: (IHHse S T H0 S1); eauto.
    exists (logctx_andl ocS1T Tr).
    splits. rewrite HocS1T. auto.
    intros.
    forwards ocST1 HocST1 HocS1T1 HfolS: (HexST1 T1).
    exists (logctx_andl ocST1 Tr).
    simpls.
    rewrite HocST1.
    rewrite HocS1T1.
    auto.
    auto.
  + (* Agreeing *)
    forwards ocS1T HocS1T  HexST1 HfolT: (IHHse S T H0 S1); eauto.
    exists (logctx_andr Tl ocS1T).
    splits. rewrite HocS1T. auto.
    intros.
    forwards ocST1 HocST1 HocS1T1 HfolS: (HexST1 T1).
    exists (logctx_andr Tl ocST1).
    simpls.
    rewrite HocST1.
    rewrite HocS1T1.
    auto.
    auto.
  + destruct oc2; simpls.
    * forwards: H oc1 S; eauto. destruct_all. contradiction.
    * forwards: H oc2 T0; eauto. destruct_all. contradiction.
    * inversions H1.
      exists (logctx_orr (logctx_typ oc1 S1) oc2). simpls. splits. auto.
      intros.
      exists (logctx_orl oc1 (logctx_typ oc2 T1)). simpls. auto.
      auto.
    * inversions H1.
    * inversions H1.
  + destruct oc2; simpls.
    * forwards: H oc1 S; eauto. destruct_all. contradiction.
    * inversions H1.
      exists (logctx_orl oc2 (logctx_typ oc1 S1) ). simpls. splits. auto.
      intros.
      exists (logctx_orr  (logctx_typ oc2 T1) oc1). simpls. auto.
      auto.
    * forwards: H oc2 T0; eauto. destruct_all. contradiction.
    * inversions H1.
    * inversions H1.
  + destruct oc2; simpls.
    * forwards: H oc1 S; eauto. destruct_all. contradiction.
    * inversions H1.
    * inversions H1.
    * forwards: H oc2 T0; eauto. destruct_all. contradiction.
    * inversions H1.
      exists (logctx_andr (logctx_typ oc1 S1) oc2). simpls. splits. auto.
      intros.
      exists (logctx_andl oc1 (logctx_typ oc2 T1)). simpls. auto.
      auto.
  + destruct oc2; simpls.
    * forwards: H oc1 S; eauto. destruct_all. contradiction.
    * inversions H1.
    * inversions H1.
    * inversions H1.
      exists (logctx_andl oc2 (logctx_typ oc1 S1) ). simpls. splits. auto.
      intros.
      exists (logctx_andr  (logctx_typ oc2 T1) oc1). simpls. auto.
      auto.
    * forwards: H oc2 T0; eauto. destruct_all. contradiction.
Qed.

Lemma logctx_same_typ_eq_inv oc T1 T2:
  (logctx_typ oc T1) = (logctx_typ oc T2) ->
  T1 = T2.
Proof.
  introv Heq.
  induction oc; simpls; eauto; inversions Heq; eauto.
Qed.
Lemma logctx_follow_eq_inv oc1 oc2 T1 T2:
  logctx_follow oc1 oc2 ->
  (logctx_typ oc1 T1) = (logctx_typ oc2 T2) ->
  T1 = T2 /\ oc1 = oc2.
Proof.
  introv Hf Heq.
  induction Hf; simpls; eauto; inversions Heq; forwards : IHHf; destruct_all; subst; eauto.
Qed.

Lemma logctx_recombine T U S1 S2 ocT1 ocT2 ocS ocSU:
  logctx_separate ocT1 ocS ->
  logctx_follow ocT1 ocT2 ->
  logctx_follow ocS ocSU ->
  logctx_typ ocT1 T = logctx_typ ocS S1 ->
  logctx_typ ocT2 T = logctx_typ ocS S2 ->
  logctx_typ ocT1 U = logctx_typ ocSU S1 ->
  logctx_typ ocT2 U = logctx_typ ocSU S2.
Proof.
  introv Hsep HfT HfS HT1 HT2 HU.
  reverts ocT2 ocSU T U. reverts S1 S2.
  induction Hsep; intros; simpls; inversions HfT; simpls.
  (* Common path - by induction *)
  - inversions HfS; simpls; inversions HU; inversions HT2; inversions HT1.
    forwards: IHHsep H1; eauto.
    fequal. eauto.
  - inversions HfS; simpls; inversions HU; inversions HT2; inversions HT1.
    forwards: IHHsep H1; eauto.
    fequal. eauto.
  - inversions HfS; simpls; inversions HU; inversions HT2; inversions HT1.
    forwards: IHHsep H1; eauto.
    fequal. eauto.
  - inversions HfS; simpls; inversions HU; inversions HT2; inversions HT1.
    forwards: IHHsep H1; eauto.
    fequal. eauto.
  (* Divergence *)
  - destruct oc2; simpls; forwards: H oc1 T; try forwards: H oc2 T0; destruct_all; try contradiction; inversions HT1; inversions HT2; inversions HfS; simpls; inversions HU.
    symmetry in H6.
    forwards: logctx_follow_eq_inv H6; eauto; destruct_all; subst.
    forwards: logctx_follow_eq_inv H8; eauto; destruct_all; subst.
    fequal.
  - destruct oc2; simpls; forwards: H oc1 T; try forwards: H oc2 T0; destruct_all; try contradiction; inversions HT1; inversions HT2; inversions HfS; simpls; inversions HU.
    symmetry in H7.
    forwards: logctx_follow_eq_inv H6; eauto; destruct_all; subst.
    forwards: logctx_follow_eq_inv H7; eauto; destruct_all; subst.
    fequal.
  - destruct oc2; simpls; forwards: H oc1 T; try forwards: H oc2 T0; destruct_all; try contradiction; inversions HT1; inversions HT2; inversions HfS; simpls; inversions HU.
    symmetry in H6.
    forwards: logctx_follow_eq_inv H6; eauto; destruct_all; subst.
    forwards: logctx_follow_eq_inv H8; eauto; destruct_all; subst.
    fequal.
  - destruct oc2; simpls; forwards: H oc1 T; try forwards: H oc2 T0; destruct_all; try contradiction; inversions HT1; inversions HT2; inversions HfS; simpls; inversions HU.
    symmetry in H7.
    forwards: logctx_follow_eq_inv H6; eauto; destruct_all; subst.
    forwards: logctx_follow_eq_inv H7; eauto; destruct_all; subst.
    fequal.
Qed.

Lemma logctx_follow_extend oc ocp ocl:
  logctx_follow (ocl) (ocp) ->
  logctx_follow (logctx_extend oc ocl) (logctx_extend oc ocp).
Proof.
  induction oc; simpls; auto.
Qed.

Inductive typ_log: typ -> Prop :=
| typ_log_notlog T:
  typ_notlog T ->
  typ_log T
| typ_log_log T1 T2 TB: 
  typ_log T1 ->
  typ_log T2 ->
  typ_andor TB ->
  typ_log (TB T1 T2).

Lemma typ_typ_log T:
  typ_log T.
Proof.
  induction T.
  all: try solve [eapply typ_log_notlog; splits; hnf; intros; discriminate].
  all: eapply typ_log_log; eauto.
Qed.
