(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics.
Require Import Coq.Program.Equality.
Require Import Types.
Require Import TypeInduction.

Require Import Orctx Logctx Andctx.

(** * Conversion between [\/]-context and [/\],[\/]-context *)

Fixpoint orctx_logctx (oc: orctx): logctx := 
  match oc with
  | orctx_nil => logctx_nil
  | orctx_l ocl Tr => logctx_orl (orctx_logctx ocl) Tr
  | orctx_r Tl ocr => logctx_orr Tl (orctx_logctx ocr)
  end.

Lemma orctx_logctx_typ oc T:
  (orctx_typ oc T) = (logctx_typ (orctx_logctx oc) T).
Proof.
  induction oc; simpls; try rewrite IHoc; eauto.
Qed.

(* Logctx that is equal to some orctx *)
Definition logctx_isor oc := (exists oco, oc = orctx_logctx oco).

Lemma logctx_isor_nil:
  logctx_isor logctx_nil.
Proof.
  hnf.
  exists orctx_nil; simpls; auto. 
Qed.
Hint Resolve logctx_isor_nil: core.

Lemma logctx_isor_orctx oc:
  logctx_isor (orctx_logctx oc).
Proof.
  induction oc; hnf; simpls.
  - exists orctx_nil; simpls; eauto.
  - hnf in IHoc; destruct IHoc. exists (orctx_l x T); simpls; rewrite H; eauto.
  - hnf in IHoc; destruct IHoc. exists (orctx_r T x); simpls; rewrite H; eauto.
Qed.
Hint Resolve logctx_isor_orctx: core.

Lemma logctx_follows_isor oc1 oc2:
  logctx_isor oc1 ->
  logctx_follow oc1 oc2 ->
  logctx_isor oc2.
Proof.
  introv His Hfol.
  induction Hfol; hnf.
  - exists orctx_nil; simpls; eauto.
  - hnf in His; destruct_all. destruct x; simpls; inversions H.
    forwards: IHHfol logctx_isor_orctx. hnf in H. destruct_all; subst.
     exists (orctx_l x0 Tr2); simpls; auto.
  - hnf in His; destruct_all. destruct x; simpls; inversions H.
    forwards: IHHfol logctx_isor_orctx. hnf in H. destruct_all; subst.
     exists (orctx_r Tl2 x0); simpls; auto.
  - hnf in His; destruct_all. destruct x; simpls; inversions H.
  - hnf in His; destruct_all. destruct x; simpls; inversions H.
Qed.

Lemma orctx_logctx_extend ocl ocr1 ocr2:
  orctx_logctx ocl = logctx_extend ocr1 ocr2 ->
  exists ocl1 ocl2,
     orctx_logctx ocl1 = ocr1 /\
     orctx_logctx ocl2 = ocr2 /\
     ocl = orctx_extend ocl1 ocl2.
Proof.
  introv Hex.
  reverts ocr2 ocl.
  induction ocr1; intros; simpls.
  - exists orctx_nil ocl; simpls; auto.
  - destruct ocl; inversions Hex.
    forwards: IHocr1 H0. destruct_all; subst.
    exists (orctx_l x T) x0; simpls; auto.
  - destruct ocl; inversions Hex.
    forwards: IHocr1 H1. destruct_all; subst.
    exists (orctx_r T x) x0; simpls; auto.
  - destruct ocl; inversions Hex.
  - destruct ocl; inversions Hex.
Qed.





(* A logctx that starts with an and-sequence and ends with an or-sequence (there is no and after an or) *)

Definition logctx_andor oc := exists oc1 oc2, oc = (logctx_extend (andctx_logctx oc1) (orctx_logctx oc2)).

Lemma logctx_andor_andr T oc:
  logctx_andor oc ->
  logctx_andor (logctx_andr T oc).
Proof.
  introv Ha.
  hnf in *.
  destruct_all; subst.
  exists (andctx_r T x) x0.
  simpls. auto.
Qed.
Hint Resolve logctx_andor_andr: core.
Lemma logctx_andor_andl T oc:
  logctx_andor oc ->
  logctx_andor (logctx_andl oc T).
Proof.
  introv Ha.
  hnf in *.
  destruct_all; subst.
  exists (andctx_l x T) x0.
  simpls. auto.
Qed.
Hint Resolve logctx_andor_andl: core.
Lemma logctx_andor_orctx oc:
  logctx_andor (orctx_logctx oc).
Proof.
  hnf in *.
  exists (andctx_nil) oc; simpls; auto.
Qed.
Hint Resolve logctx_andor_orctx: core.
Lemma logctx_andor_andctx oc:
  logctx_andor (andctx_logctx oc).
Proof.
  hnf in *.
  exists oc (orctx_nil); rewrite logctx_extend_nil_l; auto.
Qed.
Hint Resolve logctx_andor_andctx: core.

Lemma logctx_andor_orl_inv oc T:
  logctx_andor (logctx_orl oc T) ->
  logctx_andor oc /\ logctx_isor oc.
Proof.
  introv Ha.
  hnf in *.
  destruct_all.
  destruct x; simpls.
  destruct x0; simpls; inversions H.
  splits; hnf.
  exists (andctx_nil) x0; simpls; auto.
  eauto.
  inversions H.
  inversions H.
Qed.
Lemma logctx_andor_orr_inv oc T:
  logctx_andor (logctx_orr T oc) ->
  logctx_andor oc /\ logctx_isor oc.
Proof.
  introv Ha.
  hnf in *.
  destruct_all.
  destruct x; simpls.
  destruct x0; simpls; inversions H.
  splits; hnf.
  exists (andctx_nil) x0; simpls; auto.
  eauto.
  inversions H.
  inversions H.
Qed.

Lemma logctx_andor_andl_inv oc T:
  logctx_andor (logctx_andl oc T) ->
  logctx_andor oc.
Proof.
  introv Ha.
  hnf in *.
  destruct_all.
  destruct x; simpls.
  destruct x0; simpls; inversions H.
  inversions H. eauto.
  inversions H.
Qed.
Lemma logctx_andor_andr_inv oc T:
  logctx_andor (logctx_andr T oc) ->
  logctx_andor oc.
Proof.
  introv Ha.
  hnf in *.
  destruct_all.
  destruct x; simpls.
  destruct x0; simpls; inversions H.
  inversions H.
  inversions H.  eauto.
Qed.


Lemma logctx_follow_andor oc1 oc2:
  logctx_follow oc1 oc2 ->
  logctx_andor oc1 ->
  logctx_andor oc2.
Proof.
  introv Hf Ha.
  reverts oc2.
  induction oc1; intros; inversions Hf; eauto.
  - eapply logctx_andor_orl_inv in Ha; destruct_all; subst.
    forwards: logctx_follows_isor; eauto.
    hnf. hnf in H1.
    destruct_all; subst. 
    eauto.
    exists andctx_nil (orctx_l x Tr2); simpls; eauto.
  - eapply logctx_andor_orr_inv in Ha; destruct_all; subst.
    forwards: logctx_follows_isor; eauto.
    hnf. hnf in H1.
    destruct_all; subst. 
    eauto.
    exists andctx_nil (orctx_r Tl2 x); simpls; eauto.
  - eapply logctx_andor_andl_inv in Ha. eauto.
  - eapply logctx_andor_andr_inv in Ha. eauto.
Qed.

Lemma orctx_logctx_extend_and_l_false oc1 oc2 oc3 T:
  orctx_logctx oc1 = logctx_extend oc2 (logctx_andl oc3 T) -> False.
Proof.
  introv Heq.
  reverts oc1.
  induction oc2; intros; simpls.
  - destruct oc1; simpls; inversions Heq.
  - destruct oc1; simpls; inversions Heq. eauto.
  - destruct oc1; simpls; inversions Heq. eauto.
  - destruct oc1; simpls; inversions Heq.
  - destruct oc1; simpls; inversions Heq.
Qed.
Lemma orctx_logctx_extend_and_r_false oc1 oc2 oc3 T:
  orctx_logctx oc1 = logctx_extend oc2 (logctx_andr T oc3) -> False.
Proof.
  introv Heq.
  reverts oc1.
  induction oc2; intros; simpls.
  - destruct oc1; simpls; inversions Heq.
  - destruct oc1; simpls; inversions Heq. eauto.
  - destruct oc1; simpls; inversions Heq. eauto.
  - destruct oc1; simpls; inversions Heq.
  - destruct oc1; simpls; inversions Heq.
Qed.
Lemma logctx_follow_orctx oc1 oc2:
  logctx_follow (orctx_logctx oc1)  (orctx_logctx oc2) ->
  orctx_follow oc1 oc2.
Proof.
  introv Hf.
  dependent induction Hf; destruct oc1; inversions x0; destruct oc2; inversions x; eauto.
Qed.

