Require Import LibExtra AbstractSyntax.

Declare Custom Entry term_custom.

(** * Readable syntax for terms *)
(*
  Use bold parentheses ❪❫ to enter readable syntax.
  
  Associativity applies to the level, not the notation itself, and determines what to do when there are muliple ones at the same level.
  Higher level is less binding.

*)

Notation "'❪' e '❫'" := e (e custom term_custom at level 2, only parsing).
Notation "'$' v" := (avar_f v) (in custom term_custom at level 1, only parsing).
Notation "'#' n" := (avar_b n) (in custom term_custom at level 1, only parsing).
Notation "'Let' x '=' t1 'In' t2 " := (trm_let t1 (close x t2))  (in custom term_custom at level 0, x ident, only parsing, right associativity).
Notation "'Lit' x '=' l 'In' t " := (trm_lit l (close x t))  (in custom term_custom at level 0, x ident, only parsing, right associativity).
Notation "'Read' x '․' a " := (trm_read x a)  (in custom term_custom at level 0, x at level 1, only parsing).
Notation "'Write' x1 '․' a '=' x2 " := (trm_write x1 a x2)  (in custom term_custom at level 0, x1 at level 1, x2 at level 1, only parsing).
Notation "'Call' x1 '․' m x2" := (trm_call x1 m x2)  (in custom term_custom at level 0, x1 at level 1, x2 at level 1, only parsing).

Notation "'λ' '⟮' z '∶' T '⟯' t " := (lit_fun T (close z t)) (in custom term_custom at level 2, only parsing).
Notation "'ν' '⟮' s '∶' T '⟯' ds " := (lit_obj (close s T) (close s ds)) (in custom term_custom at level 2, only parsing).

Notation "'μ' T " := (typ_bnd T) (in custom term_custom at level 2, only parsing).
Notation "'μ' '⟮' r '∶' T '⟯' " := (typ_bnd (close r T)) (in custom term_custom at level 2, only parsing).
Notation "x1 '․' A '⟮' x2 '⟯' " := (typ_sel x1 A x2) (in custom term_custom at level 2, only parsing).
Notation "x1 '․ⁿ' A '⟮' x2 '⟯' " := (typ_sel x1 (typ_label_normal A) x2) (in custom term_custom at level 2, only parsing).

Notation "'⊤'" := (typ_top) (in custom term_custom at level 2, only parsing).
Notation "'⊥'" := (typ_bot) (in custom term_custom at level 2, only parsing).
Notation "'𝖭'" := (typ_nocap) (in custom term_custom at level 1, only parsing).
Notation "'𝖬'" := (typ_label_refcap typ_refcap_mutability) (in custom term_custom at level 2, only parsing).

Notation "T1 '∧' T2" := (typ_and T1 T2) (in custom term_custom at level 3, only parsing, left associativity).
Notation "T1 '∨' T2" := (typ_or T1 T2) (in custom term_custom at level 3, only parsing, left associativity).

Notation "'❴ᵗ' A '∶' TL '..' TU  '❵'" := (typ_rcd (dec_typ A TL TU)) (in custom term_custom at level 2, only parsing).
Notation "'❴ⁿ' A '∶' TL '..' TU  '❵'" := (typ_rcd (dec_typ (typ_label_normal A) TL TU)) (in custom term_custom at level 2, only parsing).
Notation "'❴ᶜ' A '∶' TL '..' TU  '❵'" := (typ_rcd (dec_typ (typ_label_refcap A) TL TU)) (in custom term_custom at level 2, only parsing).
Notation "'❴ᶠ' a '∶' TL '..' TU  '❵'" := (typ_rcd (dec_fld a TL TU)) (in custom term_custom at level 2, only parsing).
Notation "'❴ᵐ' m '∶' TC ',' TZ '→' TR  '❵'" := (typ_rcd (dec_met m TC TZ TR)) (in custom term_custom at level 2, only parsing).

Notation "'❴ᵗ' A '⟮' r '⟯' '∶' TL '..' TU '❵'" := (typ_rcd (dec_typ A (close r TL) (close r TU))) (in custom term_custom at level 2, only parsing).
Notation "'❴ᵐ' m '⟮' r '∶' TC ',' z '∶' TZ '⟯' '∶' TR '❵'" := (typ_rcd (dec_met m (close z TC) TZ (close_met z r TR))) (in custom term_custom at level 2, only parsing).

Notation "'❴ᵗ' A '=' T '❵'" := (def_typ A T) (in custom term_custom at level 2, only parsing).
Notation "'❴ⁿ' A '=' T '❵'" := (def_typ (typ_label_normal A) T) (in custom term_custom at level 2, only parsing).
Notation "'❴ᶜ' A '=' T '❵'" := (def_typ (typ_label_refcap A) T) (in custom term_custom at level 2, only parsing).
Notation "'❴ᶠ' A '=' x '❵'" := (def_fld A x) (in custom term_custom at level 2, only parsing).
Notation "'❴ᵐ' m '=' t  '❵'" := (def_met m t) (in custom term_custom at level 2, only parsing).
Notation "'❴ᵗ' A '⟮' r '⟯' '=' T '❵'" := (def_typ A (close r T)) (in custom term_custom at level 2, only parsing).
Notation "'❴ᵐ' m '⟮' r ',' z '⟯' '=' t '❵'" := (def_met m (close_met z r t)) (in custom term_custom at level 2, only parsing).
Notation "T1 '⋏' T2" := (defs_cons T1 T2) (in custom term_custom at level 3, only parsing, left associativity).
Notation "∅" := (defs_nil) (in custom term_custom at level 3, only parsing, left associativity).

Notation "{ x }" := x (in custom term_custom at level 0, x constr, only parsing).
Notation "x" := x (in custom term_custom at level 0, x ident, only parsing).

Local Lemma TEST_syntax_read (x:var) (a: fld_label):
  ❪Read $x․a❫ = (trm_read (avar_f x) a).
Proof. simple apply eq_refl. Qed.

Local Lemma TEST_syntax_call (x y:var) (a: met_label):
  ❪Call $x․a $y❫ = (trm_call (avar_f x) a (avar_f y)).
Proof. simple apply eq_refl. Qed.


Local Lemma TEST_syntax_let_let (x1 x2 y1 y2:var) (a1 a2: fld_label) (t: trm):
  ❪Let y1 = Read $x1․a1 In Let y2 = Read $x2․a2 In t❫ = (trm_let (trm_read (avar_f x1) a1) (close y1 (trm_let (trm_read (avar_f x2) a2) (close y2 t)))).
Proof. simple apply eq_refl. Qed.

Local Lemma TEST_syntax_object (A: typ_label) (a: fld_label) (m: met_label) (r1 r2 z: var):
  ❪μ⟮r1∶❴ᵐ m⟮r2∶𝖭,z∶$r1․A⟮$z⟯ ∨ ⊥⟯∶⊤❵∧❴ᵗ A∶⊤..⊤❵∧❴ᶠ a∶⊤..⊤❵⟯∧❴ᵗ𝖬∶⊥..⊤❵❫ =
  typ_and (typ_bnd (close r1 (typ_and (typ_and
              (typ_rcd (dec_met m (close z typ_nocap) (typ_or (typ_sel (avar_f r1) A (avar_f z)) typ_bot) typ_top))
              (typ_rcd (dec_typ A typ_top typ_top))) (typ_rcd (dec_fld a typ_top typ_top)))))
  (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_top)).
Proof. simple apply eq_refl. Qed.