Set Implicit Arguments.

Require Import LibExtra DotTactics SyntaxClasses.

Require Import AbstractSyntax  TFree.

(** * T-free variables and substitution *)

Class TFreeSubst {A: Set} {AS: SubstVar A} (AT: TFree A) := {
  tfree_subst ys xs vy (ty:A):
    tfree vy ty ->
    tfree (vy /[ ys --> xs]) (ty /[ ys --> xs]);
  tfree_subst_case ys xs vx (ty:A):
    tfree vx (ty /[ ys --> xs]) ->
    exists vy,
      vx = (vy /[ ys --> xs]) /\
      tfree vy ty
}.
Definition TFreeSubstInd (A: Set) (AS: SubstVar A) (AT: TFree A) :=
  forall vy (ty: A),
  tfree vy ty ->
  forall ys xs,
    tfree (vy /[ ys --> xs]) (ty /[ ys --> xs]).

Lemma tfree_subst_vars:
  TFreeSubstInd subst_trm tfree_trm /\
  TFreeSubstInd subst_lit tfree_lit /\
  TFreeSubstInd subst_def tfree_def /\
  TFreeSubstInd subst_defs tfree_defs.
Proof.
  eapply tfree_mutind.
  all: intros; subst; autorewrite with simpl_subst_vars; eauto.
  all: forwards: (H ys xs); eauto.
Qed.


Definition TFreeSubstCaseInd (A: Set) (AS: SubstVar A) (AT: TFree A) :=
  forall vx (tx: A),
  tfree vx tx ->
  forall ys xs ty,
  tx = (ty /[ ys --> xs]) ->
  exists vy,
    vx = (vy /[ ys --> xs]) /\
    tfree vy ty.

Lemma tfree_subst_vars_case:
  TFreeSubstCaseInd subst_trm tfree_trm /\
  TFreeSubstCaseInd subst_lit tfree_lit /\
  TFreeSubstCaseInd subst_def tfree_def /\
  TFreeSubstCaseInd subst_defs tfree_defs.
Proof.
  eapply tfree_mutind.
  all: intros; subst; repeat invert_subst_vars; eauto.
  all: forwards: H; eauto.
  all: destruct_all; eauto.
Qed.

Instance tfree_trm_subst_vars_case: TFreeSubst tfree_trm.
Proof.
  split; intros.
  eapply tfree_subst_vars; eauto.
  eapply tfree_subst_vars_case; eauto.
Qed.
Instance tfree_defs_subst_vars_case: TFreeSubst tfree_defs.
Proof.
  split; intros.
  eapply tfree_subst_vars; eauto.
  eapply tfree_subst_vars_case; eauto.
Qed.