(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra. 
Require Import DotTactics.
Require Import SyntaxClassDefs SyntaxClassProperties SyntaxClassLemmas SyntaxClassSubstLemmas.

Generalizable Variables A B.

(** * Equalities between normal and depth operations *)
Section Open_depth.
  Context `{AS: AbstractSyntax A}.
  Lemma open_eq_open_depth : forall n x a,
      open_rec n x a = open_depth_rec n x a 0.
  Proof.
    intros.
    unfold open_depth_rec.
    rewrite Nat.add_0_r.
    auto.
  Qed.
End Open_depth.

Section Subst_depth.
  Context `{AS: AbstractSyntax A}.
  Lemma subst_depth: forall {x y} (n : nat) (a : A),
    subst_var x y a = (ignore_second (subst_var x y)) a n.
  Proof. auto. Qed.
End Subst_depth.


Section Close_depth.
  Context `{AS: AbstractSyntax A} {Cl: Closing AS}.
  Lemma close_eq_close_depth : forall n x a,
      close_rec n x a = close_depth_rec n x a 0.
  Proof.
    intros.
    unfold close_depth_rec.
    rewrite Nat.add_0_r.
    auto.
  Qed.
End Close_depth.


Section About_OpenTwice.
  Context `{OpenTwice}.

  Lemma open_depth_twice_comp : forall n x y,
      (open_depth_rec n y) ∘∘ (open_depth_rec n x) = (open_depth_rec n x).
  Proof.
    intros.
    apply fun_ext_dep. intros a.
    apply fun_ext_dep. intros m.
    unfold compose12, open_depth_rec.
    rewrite (fold_compose (open_rec _ _)), open_twice_comp.
    auto.
  Qed.

End About_OpenTwice.

Section About_OpenCommut.
  Context `{OpenCommut A}.

  Lemma open_depth_commut_comp : forall m n x y,
      m <> n ->
      (open_depth_rec m x) ∘∘ (open_depth_rec n y)
      = (open_depth_rec n y) ∘∘ (open_depth_rec m x).
  Proof.
    intros.
    unfold open_depth. rewrite lift_depth_commute12; auto.
    intros. apply open_commut_comp.
    rewrite Nat.add_cancel_r; auto.
  Qed.
End About_OpenCommut.


Section About_SubstOpenCommut.
  Context `{SubstOpenCommut}.

  Lemma subst_open_commut_depth : forall n x y z,
      (ignore_second (subst_var x y)) ∘∘ (open_depth_rec n z) =
      open_depth_rec n (z \[x -> y]) ∘∘ (ignore_second (subst_var x y)).
  Proof.
    intros.
    apply fun_ext_dep. intros a.
    apply fun_ext_dep. intros m.
    unfold compose12, open_depth, ignore_second.
    rewrite subst_open_commut; auto.
  Qed.

End About_SubstOpenCommut.

Section Seal_depth.
  Context `{AS: AbstractSyntax A} `{Sealable A}.
  Lemma seal_eq_seal_depth : forall n x a,
      seal_rec n x a = seal_depth_rec n x a 0.
  Proof.
    intros.
    unfold seal_depth_rec.
    rewrite Nat.add_0_r.
    auto.
  Qed.
End Seal_depth.

Section About_ClosingSanity.
  Section OpenClose.
    Context `{OpenClose}.
    Lemma open_depth_close_depth : forall n x,
        (open_depth_rec n x ∘∘ close_depth_rec n x) = open_depth_rec n x.
    Proof.
      intros.
      apply fun_ext_dep. intros a.
      apply fun_ext_dep. intros m.
      unfold open_depth, close_depth, compose12.
      rewrite (fold_compose (open_rec _ _)).
      rewrite open_close_comp.
      auto.
    Qed.
  End OpenClose.
  Section CloseOpen.
    Context `{CloseOpen}.
    Lemma close_depth_open_depth : forall n x,
        (close_depth_rec n x ∘∘ open_depth_rec n x) = close_depth_rec n x.
    Proof.
      intros.
      apply fun_ext_dep. intros a.
      apply fun_ext_dep. intros m.
      unfold open_depth, close_depth, compose12.
      rewrite (fold_compose (close_rec _ _)).
      rewrite close_open_comp.
      auto.
    Qed.
  End CloseOpen.
  Section CloseCommut.
    Context `{CloseCommut A}.
    Lemma close_commut_depth : forall m n x y,
      x <> y ->
      (close_depth_rec m x) ∘∘ (close_depth_rec n y)
      = (close_depth_rec n y) ∘∘ (close_depth_rec m x).
    Proof.
      intros.
      apply fun_ext_dep. intros a.
      apply fun_ext_dep. intros m0.
      unfold open_depth, close_depth, ignore_second, compose12.

      rewrite (close_commut (m + m0) (n + m0) a).
      auto. auto.
    Qed.
  End CloseCommut.
  Section OpenCloseCommut.
    Context `{OpenCloseCommut A}.
    Lemma open_close_commut_depth : forall m n x y,
      x <> y ->
      m <> n ->
      (close_depth_rec m x) ∘∘ (open_depth_rec n y)
      = (open_depth_rec n y) ∘∘ (close_depth_rec m x).
    Proof.
      intros.
      apply fun_ext_dep. intros a.
      apply fun_ext_dep. intros m0.
      unfold open_depth, close_depth, ignore_second, compose12.
      assert ((m + m0) <> (n + m0)). intro. apply H4. eapply PeanoNat.Nat.add_cancel_r; eauto.
      rewrite (open_close_commut a H3 H5 ).
      auto.
    Qed.
  End OpenCloseCommut.
  Section SubstOpenClose.
    Context `{SubstOpenClose}.
    Lemma subst_open_close_depth : forall n x y,
      (ignore_second (subst_var x y)) ∘∘ (open_depth_rec n x)
      = (open_depth_rec n y) ∘∘ (close_depth_rec n x).
    Proof.
      intros.
      apply fun_ext_dep. intros a.
      apply fun_ext_dep. intros m.
      unfold open_depth, close_depth, ignore_second, compose12.
      rewrite (fold_compose (subst_var _ _)).
      rewrite (subst_open_close_comp (n + m) x y).
      auto.
    Qed.
  End SubstOpenClose.
  Section CloseSubst.
    Context `{CloseSubst}.
    Lemma close_subst_depth : forall n x y,
      (close_depth_rec n y) ∘∘ (ignore_second (subst_var x y))
      = (close_depth_rec n y) ∘∘ (close_depth_rec n x).
    Proof.
      intros.
      apply fun_ext_dep. intros a.
      apply fun_ext_dep. intros m.
      unfold open_depth, close_depth, ignore_second, compose12.

      rewrite (close_subst (n + m) x y).
      auto.
    Qed.
  End CloseSubst.
  Section SubstCloseCommut.
    Context `{SubstCloseCommut}.
    Lemma subst_close_commut_depth : forall n x y z,
      z <> x -> z <> y ->
      (ignore_second (subst_var x y)) ∘∘ (close_depth_rec n z)
      = (close_depth_rec n z) ∘∘ (ignore_second (subst_var x y)).
    Proof.
      intros.
      apply fun_ext_dep. intros a.
      apply fun_ext_dep. intros m.
      unfold open_depth, close_depth, ignore_second, compose12.
      rewrite (subst_close_commut (n + m) a); eauto.
    Qed.
  End SubstCloseCommut.
End About_ClosingSanity.