Set Implicit Arguments.

Require Import LibExtra.
Require Import DotTactics.
Require Import SyntaxClassDefs SyntaxClassProperties SyntaxClassLemmas.

Generalizable Variables A B.

Section Subst_env.
  Context `{AS: AbstractSyntax A}.

  Lemma subst_single: forall x y z a,
    z ~ (a /[x -> y]) = subst_env x y (z ~ a).
  Proof.
    intros. rewrite map_single; auto.
  Qed.
End Subst_env.



Section About_SubstFresh.
  Context `{SubstFresh A}.

  Lemma subst_fresh_env: forall x y G,
    x \notin fv_env G ->
    subst_env x y G = G.
  Proof.
    intros x y G. induction G using env_ind; introv Hnfv.
    - rewrite map_empty; auto.
    - rewrite fv_env_push, notin_union in Hnfv.
      rewrite map_concat. destruct Hnfv. rewrite IHG by auto.
      rewrite map_single. rewrite subst_fresh by auto.
      auto.
  Qed.

End About_SubstFresh.

Section About_SubstOpenCommut.
  Context `{SubstOpenCommut}.

  Lemma subst_open_rec_vars_commut : forall x y ys t n,
    t /[#n --> ys] /[x -> y] = t /[x -> y] /[#n --> (List.map (subst_fvar x y) ys)].
  Proof.
    intros x y ys. induction ys.
    - auto.
    - intros. simpl. rewrite IHys. rewrite subst_open_commut.
      auto.
  Qed.

  (* stronger than subst_open_vars_commut_fresh_length *)
  Lemma subst_open_vars_commut_notin: forall x y ys a n,
      x \notin from_list ys ->
      a /[#n --> ys] /[x -> y] = a /[x -> y] /[#n --> ys].
  Proof.
    intros. rewrite subst_open_rec_vars_commut.
    rewrite notin_vars_map; auto.
  Qed.
  (* weaker than subst_open_vars_commut_notin_length *)
  Lemma subst_open_vars_commut_fresh: forall x y ys a n,
      (fresh \{ x } (length ys) ys) ->
      a /[#n --> ys] /[x -> y] = a /[x -> y] /[#n --> ys].
  Proof.
    intros. rewrite subst_open_rec_vars_commut.
    rewrite fresh_vars_map; auto.
  Qed.

  Lemma subst_open_commut_ne : forall x y z X n,
    z <> x ->
    X /[#n -> z] /[x -> y] = X /[x -> y] /[#n -> z].
  Proof.
    intros. rewrite subst_open_commut. cases_if. auto.
  Qed.

  Lemma subst_vars_open_fresh_commut n x T xs ys:
     fresh \{x} (length xs) xs ->
     T /[#n -> x] /[ xs --> ys] = T /[ xs --> ys] /[#n -> x].
  Proof.
    introv Hfr.
    revert dependent ys.
    induction xs; intros; simpls.
    - auto.
    - destruct ys. auto.
      destruct Hfr.
      rewrite notin_singleton in H3.
      rewrite IHxs.
      rewrite subst_open_commut.
      rewrite substi_ne_var.
      auto.
      auto.
      fresh_solve.
  Qed.

End About_SubstOpenCommut.




  Section SubstOpenClose.
    Context `{SubstOpenClose}.

    Lemma subst_open_close n x y (a: A):
      a /[# n -> x] /[x -> y] = a /[#n <- x] /[# n -> y].
    Proof.
      intros.
      pose proof (subst_open_close_comp n x y).
      eapply fun_eq_1 with (x1:=a) in H3.
      eapply H3.
    Qed.

  End SubstOpenClose.
  Section CloseSubst.
    Context `{CloseSubst}.
    Lemma close_subst n x y (a: A):
      a /[x -> y] /[#n <- y] = a /[#n <- x] /[#n <- y].
    Proof.
      intros.
      pose proof (close_subst_comp n x y).
      eapply fun_eq_1 with (x1:=a) in H3.
      eapply H3.
    Qed.
    Lemma close_subst_depth : forall n x y,
      (close_depth_rec n y) ∘∘ (ignore_second (subst_var x y))
      = (close_depth_rec n y) ∘∘ (close_depth_rec n x).
    Proof.
      intros.
      apply fun_ext_dep. intros a.
      apply fun_ext_dep. intros m.
      unfold open_depth, close_depth, ignore_second, compose12.

      rewrite (close_subst (n + m) x y).
      auto.
    Qed.
  End CloseSubst.



  Section SubstCloseCommut.
    Context `{SubstCloseCommut}.
    Lemma subst_close_commut : forall n x y z (a: A),
      z <> x -> z <> y ->
      a /[#n <- z] /[x -> y] = a /[x -> y] /[#n <- z].
    Proof.
      intros.
      forwards Hc: subst_close_commut_comp n x y z; eauto.
      eapply fun_eq_1 with (x1:=a) in Hc.
      eapply Hc.
    Qed.
    Lemma subst_close_commut_depth : forall n x y z,
      z <> x -> z <> y ->
      (ignore_second (subst_var x y)) ∘∘ (close_depth_rec n z)
      = (close_depth_rec n z) ∘∘ (ignore_second (subst_var x y)).
    Proof.
      intros.
      apply fun_ext_dep. intros a.
      apply fun_ext_dep. intros m.
      unfold open_depth, close_depth, ignore_second, compose12.
      rewrite (subst_close_commut (n + m) a); eauto.
    Qed.
  End SubstCloseCommut.





Section About_FvSubstCases.
  Context `{FvSubstCases}.
  Lemma subst_notfv x y:
    x <> y ->
    forall X,
      x \notin (fv X/[x -> y]).
  Proof.
    intros Hn ?.
    destruct (fv_subst_cases x y X) as [[Hfv Htrivial] | [? [? Hfveq]]].
    - destruct Htrivial.
      + contradiction.
      + rewrite Hfv. auto.
    - rewrite Hfveq.
      rewrite notin_union.
      split.
      + intro Hxins.
        rewrite in_remove in Hxins.
        destruct Hxins as [HxinX Hxnx].
        rewrite notin_singleton in Hxnx. auto.
      + rewrite notin_singleton. auto.
  Qed.
  Lemma notin_fv_subst_var x xs ys (X: A):
    x \notin (fv X) ->
    x <> ys ->
    x \notin (fv X/[xs -> ys]).
  Proof.
    introv HnX Hnys.
    destruct (fv_subst_cases xs ys X) as [[Hfv Htrivial] | [? [? Hfveq]]].
    - rewrite Hfv. auto.
    - rewrite Hfveq.
      rewrite notin_union.
      split.
      eapply notin_remove. auto.
      rewrite notin_singleton; auto.
  Qed.
 


  Lemma fresh_fv_subst_var x y ws a:
    fresh ((fv a) \u \{ y }) (length ws) ws ->
    fresh (fv (a /[ x -> y])) (length ws) ws.
  Proof.
    introv Hfr.
    destruct (fv_subst_cases x y a).
    - destruct H3.
      rewrite H3. fresh_solve.
    - destruct H3 as [? [? ?]]. rewrite H5.
      fresh_solve.
      eapply fresh_remove. fresh_solve.
  Qed.


  Lemma fv_subst_var_subset a x y:
    (fv (a /[x -> y])) \c ((fv a) \u \{y}).
  Proof.
    introv.
    rewrite in_union.
    pose proof (fv_subst_cases x y a).
    destruct_all; intros.
    * subst. rewrite H3 in H5. auto.
    * rewrite H3 in H5. auto.
    * rewrite H5 in H6.
      rewrite in_union in H6.
      rewrite in_remove in H6.
      destruct_all; auto.
  Qed.


  Lemma in_fv_subst_var x xs ys (X: A):
    x \in (fv X) ->
    x <> xs ->
    x \in (fv X /[xs -> ys]).
  Proof.
    introv HnX Hnys.
    destruct (fv_subst_cases xs ys X) as [[Hfv Htrivial] | [? [? Hfveq]]].
    - rewrite Hfv. auto.
    - rewrite Hfveq.
      rewrite in_union. left. rewrite in_remove. split; auto.
  Qed.


  Lemma in_fv_env_subst E x x1 x2 :
    x <> x1 -> (x \in fv_env E) ->
    x \in fv_env (subst_env x1 x2 E).
  Proof.
    introv Hne Hin. reverts E.
    eapply (env_ind
              (fun E => x \in fv_env E -> x \in fv_env (subst_env x1 x2 E))).
    - rewrite fv_env_empty, in_empty. tauto.
    - introv IH Hin. rewrite fv_env_push, in_union in Hin.
      rewrite map_push, fv_env_push, in_union.
      destruct Hin as [Hin | Hin]; try tauto. right.
      eapply in_fv_subst_var; eauto.
  Qed.

End About_FvSubstCases.

Section Subst_Commut.
  Context `{SubstCommut}.
  Lemma subst_commut x y w z (X:A):
    x <> w ->
    (y <> w \/ x = z) ->
    X /[w -> z] /[x -> y] = X /[x -> y] /[w -> z \[x -> y]].
  Proof.
    intros.
    forwards: (subst_commut_comp H3 H4).
    unfold compose in H5.
    apply fun_eq_1 with (x1:=X) in H5.
    auto.
  Qed.
  Lemma subst_commut_ne x y w z (X:A):
    x <> w ->
    x <> z ->
    y <> w ->
    X /[w -> z] /[x -> y] = X /[x -> y] /[w -> z].
  Proof.
    intros.
    forwards: (@subst_commut x y w z X).
    auto. auto.
    cases_if. auto.
  Qed.


End Subst_Commut.

Section SubstId.
  Context `{AS: AbstractSyntax}.
  Context {SI: SubstId AS}.
  Lemma subst_id (v: var) (a: A):
    a /[v -> v] = a.
  Proof.
    pose proof (fun_eq_1 a (subst_id_comp v)). rewrite H2. compute. auto.
  Qed.
End SubstId.

Section Subst.
  Context `{AS: AbstractSyntax}.
  Context {SI: SubstId AS}.
  Context {SC: SubstCommut AS}.
  Context {SF: SubstFresh AS}.
  Context {FSC: FvSubstCases AS}.

  Lemma subst_merge x y z (X:A):
    x <> y ->
    y \notin fv X ->
    X /[x -> y] /[y -> z] = X /[x -> z].
  Proof.
    intros.
    lets subst_commut': subst_commut.
    forwards Hr: (subst_commut' y z x y X); auto.
    cases_if. rewrite Hr.
    rewrite subst_fresh with (x0:=y); auto.
  Qed.
  Lemma subst_merge_same x y (X:A):
    x <> y ->
    y \notin fv X ->
    X  /[x -> y] /[y -> x] = X.
  Proof.
    intros.
    forwards: (subst_merge (x:=x) (y:=y) x); eauto.
    rewrite H4.
    rewrite subst_id_comp.
    auto.
   Qed.
  (* Lemmas about substituion that replace multiple variables by the same variable *)
  Lemma subst_commut_to_same (X: A) x y z:
    X /[x -> y] /[z -> y] =  X /[z -> y] /[x -> y].
  Proof.
    destruct (classicT(x=y)).
    subst.
    rewrite ! subst_id. auto.
    destruct (classicT(z=y)).
    subst.
    rewrite ! subst_id. auto.
    destruct (classicT(x=z)).
    subst. auto.
    eapply subst_commut_ne; auto.
  Qed.
  Lemma subst_vars_commut_to_same (X: A) xs y z:
    let ys := (LibList.make (length xs) y) in
    X /[xs --> ys] /[z -> y] = X /[z -> y] /[xs --> ys].
  Proof.
    intros. unfold ys; clear ys.
    induction xs; simpls.
    - auto.
    -
      rewrite subst_commut_to_same.
      rewrite IHxs.
      auto.
  Qed.

  Lemma subst_twice (X: A) x y:
    X /[x -> y] /[x -> y] = X /[x -> y].
  Proof.
    destruct (classicT(x=y)).
    - subst.
      rewrite subst_id. auto.
    -
    rewrite subst_fresh. auto.
    eapply subst_notfv. auto.
  Qed.
End Subst.
Section SubstClose.
  Context `{AS: AbstractSyntax}.
  Context {SI: SubstId AS}.
  Context {SC: SubstCommut AS}.
  Context {SF: SubstFresh AS}.
  Context {FSC: FvSubstCases AS}.
  Context {CA: Closing AS}.
  Context {SCC: SubstCloseCommut CA}.

  Lemma close_subst_vars (t: A) n v (xs ys: list var):
    length xs = length ys ->
    v \notin (from_list xs \u from_list ys) ->
    t /[ xs --> ys] /[#n <- v] = t /[#n <- v] /[ xs --> ys].
  Proof.
    introv Hl.
    reverts ys t.
    induction xs; introv Hl; destruct ys; try discriminate.
    - simpls. auto.
    - introv Hn.
      simpls.  rewrite ! from_list_cons in Hn.
      rewrite <- IHxs.
      rewrite (subst_close_commut); eauto.
      eauto.
      eauto.
  Qed.
End SubstClose.

Section OpenTwiceClosed.
  Context `{AS: AbstractSyntax A}.
  Context {OT: OpenTwice H}.
  Context {SOC: SubstOpenCommut AS}.
  Context {CA: Closing AS}.

  Lemma closed_subst_var x y (T: A):
    closed T ->
    closed (T /[ x -> y]).
  Proof.
    unfold closed.
    introv Hcl.
    intros.
    rewrite <- (Hcl n x).
    rewrite (subst_open_commut x y x).
    cases_if.
    rewrite open_twice.
    auto.
  Qed.
  (* Variant of closing just for index 0 *)
  Lemma closed0_subst_var (T: A) x y:
    (forall z : var, T /[->z] = T) ->
    (forall z : var, T /[x -> y]/[->z] = T/[x -> y]).
  Proof.
    introv Ho.
    intro.
    remember (var_freshes \{x} 1 ) as ws.
    destruct ws.
    destruct x0; inversions f.
    rewrite <- (Ho v).
    rewrite subst_open_commut_ne.
    rewrite open_twice. auto. eauto.
  Qed.
End OpenTwiceClosed.


Section OpenClose.
  Context `{AS: AbstractSyntax A}.
  Context {OT: OpenTwice H}.
  Context {OC: OpenCommut H}.
  Context {SOC: SubstOpenCommut AS}.
  Context {CA: Closing AS}.
  Context {SOCA: SubstOpenClose CA}.

  Lemma closed_open_close x y (T: A):
    closed T ->
    closed T /[<- x] /[-> y].
  Proof.
    introv Hcl.
    rewrite <- subst_open_close.
    eapply closed_subst_var.
    rewrite Hcl. auto.
  Qed.
End OpenClose.


