(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation TransformationTerm TransformationLifting Weakening EWeakening Substitution.
Require Import ConfigTyping TFree TFreeTyping HeapCorrespondence InertContext ConfigTypingAccessories.
Require Import TypedOperationalSemantics OperationalSemantics ConfigVarkinds EnvCorrespondence SealClosed.
Require Import Sealing.

(** * Transformations of machine configurations and their elements *)
Definition TransformHeap := Transform renv ctx heap.
Definition TransformItem := Transform TrfContextTy var item.
Definition TransformStack := Transform TrfContextTy (typ*typ) stack.
Definition TransformConfig := Transform ctx typ config.

Section WithTypingMode.
Context (typing_mode: Mode).

(** Type-safe configuration transformation *)
Definition TransformCondConfTyped (TEa: TypedElement ctx typ config) :=
  match TEa with
    〈G ⊢ C ∶ T〉 => ty_config G C T
  end.

(* *Type-safe heap transformation *)
Definition TransformCondHeapCorr (TEa: TypedElement renv ctx heap) :=
  match TEa with
    〈e ⊢ Si ∶ G〉 => 
      heap_correspond_accessories G e Si ->
      heap_correspond G e Si
  end.

(** Type-safe stack transformation *)
Definition TransformCondStackTyped (TEa: TypedElement TrfContextTy (typ*typ) stack) :=
  match TEa with
    〈G⁏e ⊢ fs ∶ (Tt, T)〉 => ty_stack G e Tt fs T
  end.

(** Lift term typing to configurations, but only in the focus *)
Inductive trf_focus_conf (Cond: LiftCond) (TL: TermTransform): TransformConfig:=
  trf_focus_conf_focus Ga Gb ea eb fs S ta tb Tta Ttb Tsa Tsb Ta Tb Lva Lvb:
    ty_config_accessories Ga ea fs S ta Ta Lva ->
    ty_config_accessories Gb eb fs S tb Tb Lvb ->
    heap_correspond Ga ea S ->
    heap_correspond Gb eb S ->
    ty_stack Ga ea Tsa fs Ta ->
    ty_stack Gb eb Tsb fs Tb ->
    〈Ga⁏ea ⊢ ta ∶ Tta〉 ∼TL∼> 〈Gb⁏eb ⊢ tb ∶ Ttb〉 ->
    Ga ⁏ ea ⊢ Tta <: Tsa ->
    Gb ⁏ eb ⊢ Ttb <: Tsb ->
    〈Ga ⊢ ta; fs; S; ea; Lva ∶ Ta〉 ∼(Cond _ _ _)∼∼(trf_focus_conf Cond TL)∼> 〈Gb ⊢ tb; fs; S; eb; Lvb ∶ Tb〉.

(* Part of Figure 24 in "Pure methods for roDOT" *)
(** Lift term typing to configurations, but only in the focus, type identical variant *)
Inductive trf_focus_idtyp_conf (TL: TermTransform): TransformConfig:=
  (** (TRFF-Focus) *)
  trf_focus_idtyp_conf_focus G e fs S ta tb Tt Ts T Lva Lvb:
    ty_config_accessories G e fs S ta T Lva ->
    ty_config_accessories G e fs S tb T Lvb ->
    heap_correspond G e S ->
    ty_stack G e Ts fs T ->
    〈G⁏e ⊢ ta ∶ Tt〉 ∼TL∼> 〈G⁏e ⊢ tb ∶ Tt〉 ->
    G ⁏ e ⊢ Tt <: Ts ->
    〈G ⊢ ta; fs; S; e; Lva ∶ T〉 ∼(trf_focus_idtyp_conf TL)∼> 〈G ⊢ tb; fs; S; e; Lvb ∶ T〉.

Hint Constructors trf_focus_conf trf_focus_idtyp_conf: core.

(** Lift term typing to stacks *)
Inductive trf_lift_stack (Cond: LiftCond) (TL: TermTransform): TransformStack :=
| trf_lift_stack_e Ga ea Gb eb fs ta tb T1a T1b T2a T2b T3a T3b L:
    (forall x, x \notin L ->
      〈Ga § x / v_let ~ T1a⁏ea ⊢ (open x ta) ∶ T2a〉 ∼(trf_local_trm typing_mode Cond TL)∼> 〈Gb § x / v_let ~ T1b⁏eb ⊢ (open x tb) ∶ T2b〉
    ) ->
    ty_stack Ga ea T2a fs T3a ->
    ty_stack Gb eb T2b fs T3b ->
    〈Ga⁏ea ⊢ (stack_cons (frame_let ta) fs) ∶ (T1a, T3a)〉 ∼(Cond TrfContextTy _ _)∼∼(trf_lift_stack Cond TL)∼> 〈Gb⁏eb ⊢ (stack_cons (frame_let tb) fs) ∶ (T1b, T3b)〉
| trf_lift_stack_n Ga ea Gb eb fsa fsb t T1a T1b T2a T2b T3a T3b L:
    〈Ga⁏ea ⊢ fsa ∶ (T2a,T3a)〉 ∼(trf_lift_stack Cond TL)∼> 〈Gb⁏eb ⊢ fsb ∶ (T2b,T3b)〉 ->
    (forall x, x \notin L ->
      Ga § x / v_let ~ T1a ⁏ea ⊢ open x t ∶ T2a
    ) ->
    (forall x, x \notin L ->
      Gb § x / v_let ~ T1b ⁏eb ⊢ open x t ∶ T2b
    ) ->
    〈Ga⁏ea ⊢ (stack_cons (frame_let t) fsa) ∶ (T1a,T3a)〉 ∼(Cond TrfContextTy _ _)∼∼(trf_lift_stack Cond TL)∼> 〈Gb⁏eb ⊢ (stack_cons (frame_let t) fsb) ∶ (T1b,T3b)〉.
(* Lift term typing to stacks, type identical variant *)
Inductive trf_lift_idtyp_stack (TL: TermTransform): TransformStack :=
| trf_lift_idtyp_stack_e Ga ea fs ta tb T1a T2a T3a L:
    (forall x, x \notin L ->
      〈Ga § x / v_let ~ T1a⁏ea ⊢ (open x ta) ∶ T2a〉 ∼(trf_local_idtyp_trm typing_mode TL)∼> 〈Ga § x / v_let ~ T1a⁏ea ⊢ (open x tb) ∶ T2a〉
    ) ->
    ty_stack Ga ea T2a fs T3a ->
    〈Ga⁏ea ⊢ (stack_cons (frame_let ta) fs) ∶ (T1a, T3a)〉 ∼(trf_lift_idtyp_stack TL)∼> 〈Ga⁏ea ⊢ (stack_cons (frame_let tb) fs) ∶ (T1a, T3a)〉
| trf_lift_idtyp_stack_n Ga ea fsa fsb t T1a T2a T3a L:
    〈Ga⁏ea ⊢ fsa ∶ (T2a,T3a)〉 ∼(trf_lift_idtyp_stack TL)∼> 〈Ga⁏ea ⊢ fsb ∶ (T2a,T3a)〉 ->
    (forall x, x \notin L ->
      Ga § x / v_let ~ T1a ⁏ea ⊢ open x t ∶ T2a
    ) ->
    〈Ga⁏ea ⊢ (stack_cons (frame_let t) fsa) ∶ (T1a,T3a)〉 ∼(trf_lift_idtyp_stack TL)∼> 〈Ga⁏ea ⊢ (stack_cons (frame_let t) fsb) ∶ (T1a,T3a)〉.
(* Lift term typing to items *)
Inductive trf_lift_item (Cond: LiftCond) (TL: TermTransform): TransformItem :=
| trf_lift_item_fun Ga Gb ea eb y ta tb Sa Sb Ta Tb L:
    mode_has_fun' typing_mode = true ->
    ctx_binds y (typ_fun Sa Ta) Ga ->
    ctx_binds y (typ_fun Sb Tb) Gb ->
    (forall x, x \notin L ->
       〈(Ga § x / v_par ~ Sa) ⁏ ea ⊢  (open x ta) ∶ (open x Ta)〉 ∼(trf_local_trm typing_mode Cond TL)∼> 〈(Gb § x / v_par ~ Sb) ⁏ eb ⊢ (open x tb) ∶ (open x Tb)〉
    ) ->
    all_tfree_k Ga ta v_ref ->
    all_tfree_k Gb tb v_ref ->
    〈Ga⁏ea ⊢ (item_fun ta) ∶ y〉 ∼(Cond TrfContextTy _ _)∼∼(trf_lift_item Cond TL)∼> 〈Gb⁏eb ⊢ (item_fun tb) ∶ y〉
| trf_lift_item_obj Ga Gb ea eb y dsa dsb Ua Ub Sa Sb:
    ctx_binds y (typ_objref Ua Sa) Ga ->
    ctx_binds y (typ_objref Ub Sb) Gb ->
    〈Ga§§(objctx_heap y Ua)⁏ea ⊢ dsa ∶ (open y Ua)〉 ∼(trf_local_defs typing_mode Cond TL )∼> 〈Gb§§(objctx_heap y Ub)⁏eb ⊢ dsb ∶ (open y Ub)〉 ->
    defs_varkinds Ga dsa ->
    defs_varkinds Gb dsb ->
    〈Ga⁏ea ⊢ (item_obj dsa) ∶ y〉 ∼(Cond TrfContextTy _ _)∼∼(trf_lift_item Cond TL)∼> 〈Gb⁏eb ⊢ (item_obj dsb) ∶ y〉.
(* Lift term typing to items, type identical variant *)
Inductive trf_lift_idtyp_item (TL: TermTransform): TransformItem :=
| trf_lift_idtyp_item_fun G e y ta tb S T L:
    mode_has_fun' typing_mode = true ->
    ctx_binds y (typ_fun S T) G ->
    (forall x, x \notin L ->
       〈(G § x / v_par ~ S) ⁏ e ⊢  (open x ta) ∶ (open x T)〉 ∼(trf_local_idtyp_trm typing_mode TL)∼> 〈(G § x / v_par ~ S) ⁏ e ⊢ (open x tb) ∶ (open x T)〉
    ) ->
    all_tfree_k G ta v_ref ->
    all_tfree_k G tb v_ref ->
    〈G⁏e ⊢ (item_fun ta) ∶ y〉 ∼(trf_lift_idtyp_item TL)∼> 〈G⁏e ⊢ (item_fun tb) ∶ y〉
| trf_lift_idtyp_item_obj G e y dsa dsb U S:
    ctx_binds y (typ_objref U S) G ->
    〈G§§(objctx_heap y U)⁏e ⊢ dsa ∶ (open y U)〉 ∼(trf_local_idtyp_defs typing_mode TL )∼> 〈G§§(objctx_heap y U)⁏e ⊢ dsb ∶ (open y U)〉 ->
    defs_varkinds G dsa ->
    defs_varkinds G dsb ->
    〈G⁏e ⊢ (item_obj dsa) ∶ y〉 ∼(trf_lift_idtyp_item TL)∼> 〈G⁏e ⊢ (item_obj dsb) ∶ y〉.

(* Apply the local transformation at exactly one place on the heap (= one place in the body of one method in one item on the heap) *)
Inductive trf_lift_heap (Cond: LiftCond) (TL: TermTransform): TransformHeap :=
| trf_lift_heap_e Ga Gb ea eb ita itb S x:
    (* Preceding items *)
    (forall y, y \in dom S ->
      ty_item_s typing_mode Ga ea S y
    ) ->
    (forall y, y \in dom S ->
      ty_item_s typing_mode Gb eb S y
    ) ->
    (* Changed item *)
    〈Ga⁏ea ⊢ ita ∶ x〉 ∼(trf_lift_item Cond TL)∼> 〈Gb⁏eb ⊢ itb ∶ x〉 ->
    〈ea ⊢ (S & x ~ ita) ∶ Ga〉 ∼(Cond _ _ _)∼∼(trf_lift_heap Cond TL)∼> 〈eb ⊢ (S & x ~ itb) ∶ Gb〉
| trf_lift_heap_n Ga Gb ea eb Sa Sb x it:
    〈ea ⊢ Sa ∶ Ga〉 ∼(trf_lift_heap Cond TL)∼> 〈eb ⊢ Sb ∶ Gb〉 ->
    ty_item_s typing_mode Ga ea (x ~ it) x ->
    ty_item_s typing_mode Gb eb (x ~ it) x ->
    〈ea ⊢ (Sa & x ~ it) ∶ Ga〉 ∼(Cond _ _ _)∼∼(trf_lift_heap Cond TL)∼> 〈eb ⊢ (Sb & x ~ it) ∶ Gb〉.
Inductive trf_lift_idtyp_heap (TL: TermTransform): TransformHeap :=
| trf_lift_idtyp_heap_e G e ita itb S x:
    (* Preceding items *)
    (forall y, y \in dom S ->
      ty_item_s typing_mode G e S y
    ) ->
    (* Changed item *)
    〈G⁏e ⊢ ita ∶ x〉 ∼(trf_lift_idtyp_item TL)∼> 〈G⁏e ⊢ itb ∶ x〉 ->
    〈e ⊢ (S & x ~ ita) ∶ G〉 ∼(trf_lift_idtyp_heap TL)∼> 〈e ⊢ (S & x ~ itb) ∶ G〉
| trf_lift_idtyp_heap_n G e Sa Sb x it:
    〈e ⊢ Sa ∶ G〉 ∼(trf_lift_idtyp_heap TL)∼> 〈e ⊢ Sb ∶ G〉 ->
    ty_item_s typing_mode G e (x ~ it) x ->
    〈e ⊢ (Sa & x ~ it) ∶ G〉 ∼(trf_lift_idtyp_heap TL)∼> 〈e ⊢ (Sb & x ~ it) ∶ G〉.

(* Apply the local transformation at exactly one place in the configuration (= one place in the focus, or one place on the stack, or one place on the heap *)
Inductive trf_lift_conf (Cond: LiftCond) (TL: TermTransform): TransformConfig :=
(* In focus term *)
| trf_lift_conf_focus Ga Gb Ta Tb Ca Cb:
    〈Ga ⊢ Ca ∶ Ta〉 ∼(trf_focus_conf Cond (trf_local_trm typing_mode Cond TL))∼> 〈Gb ⊢ Cb ∶ Tb〉 ->
    〈Ga ⊢ Ca ∶ Ta〉 ∼(Cond _ _ _)∼∼(trf_lift_conf Cond TL)∼> 〈Gb ⊢ Cb ∶ Tb〉
(* On the heap *)
| trf_lift_conf_heap Ga Gb e fs Sa Sb t Tta Ttb Ta Tb Lva Lvb:
    ty_config_accessories Ga e fs Sa t Ta Lva ->
    ty_config_accessories Gb e fs Sb t Tb Lvb ->
    heap_correspond_accessories Ga e Sa ->
    heap_correspond_accessories Gb e Sb ->
    〈e ⊢ Sa ∶ Ga〉 ∼(trf_lift_heap Cond TL)∼> 〈e ⊢ Sb ∶ Gb〉 ->
    ty_stack Ga e Tta fs Ta ->
    ty_stack Gb e Ttb fs Tb ->
    Ga⁏e ⊢ t ∶ Tta ->
    Gb⁏e ⊢ t ∶ Ttb ->
    〈Ga ⊢ (t; fs; Sa; e; Lva) ∶ Ta〉 ∼(Cond _ _ _)∼∼(trf_lift_conf Cond TL)∼> 〈Gb ⊢ (t; fs; Sb; e; Lvb) ∶ Tb〉
(* On the stack *)
| trf_lift_conf_stack Ga Gb e fsa fsb S t Tta Ttb Ta Tb Lva Lvb:
    ty_config_accessories Ga e fsa S t Ta Lva ->
    ty_config_accessories Gb e fsb S t Tb Lvb ->
    heap_correspond Ga e S ->
    heap_correspond Gb e S ->
    〈Ga⁏e ⊢ fsa ∶ (Tta,Ta)〉 ∼(trf_lift_stack Cond TL)∼> 〈Gb⁏e ⊢ fsb ∶ (Ttb,Tb)〉 ->
    Ga⁏e ⊢ t ∶ Tta ->
    Gb⁏e ⊢ t ∶ Ttb ->
    〈Ga ⊢ (t; fsa; S; e; Lva) ∶ Ta〉 ∼(Cond _ _ _)∼∼(trf_lift_conf Cond TL)∼> 〈Gb ⊢ (t; fsb; S; e; Lvb) ∶ Tb〉.

(* Part of Figure 24 in "Pure methods for roDOT" *)
Inductive trf_lift_idtyp_conf (TL: TermTransform): TransformConfig :=
(** (TRFC-Focus) In focus term *)
| trf_lift_idtyp_conf_focus G T Ca Cb:
    〈G ⊢ Ca ∶ T〉 ∼(trf_focus_idtyp_conf (trf_local_idtyp_trm typing_mode TL))∼> 〈G ⊢ Cb ∶ T〉 ->
    〈G ⊢ Ca ∶ T〉 ∼(trf_lift_idtyp_conf TL)∼> 〈G ⊢ Cb ∶ T〉
(** (TRFC-Heap) On the heap *)
| trf_lift_idtyp_conf_heap G e fs Sa Sb t Tt T Lva Lvb:
    ty_config_accessories G e fs Sa t T Lva ->
    ty_config_accessories G e fs Sb t T Lvb ->
    heap_correspond_accessories G e Sa ->
    heap_correspond_accessories G e Sb ->
    〈e ⊢ Sa ∶ G〉 ∼(trf_lift_idtyp_heap TL)∼> 〈e ⊢ Sb ∶ G〉 ->
    ty_stack G e Tt fs T ->
    G⁏e ⊢ t ∶ Tt ->
    〈G ⊢ (t; fs; Sa; e; Lva) ∶ T〉 ∼(trf_lift_idtyp_conf TL)∼> 〈G ⊢ (t; fs; Sb; e; Lvb) ∶ T〉
(** (TRFC-Stack) On the stack *)
| trf_lift_idtyp_conf_stack G e fsa fsb S t Tt T Lva Lvb:
    ty_config_accessories G e fsa S t T Lva ->
    ty_config_accessories G e fsb S t T Lvb ->
    heap_correspond G e S ->
    〈G⁏e ⊢ fsa ∶ (Tt,T)〉 ∼(trf_lift_idtyp_stack TL)∼> 〈G⁏e ⊢ fsb ∶ (Tt,T)〉 ->
    G⁏e ⊢ t ∶ Tt ->
    〈G ⊢ (t; fsa; S; e; Lva) ∶ T〉 ∼(trf_lift_idtyp_conf TL)∼> 〈G ⊢ (t; fsb; S; e; Lvb) ∶ T〉.

Global Instance TransformFocusConfigTyped Cond trf
  (TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires TransformCondConfTyped (trf_focus_conf Cond trf).
Proof.
  hnf. intros.
  induction H.
  inversions H; inversions H0; destruct_all. pose proof H1 as Hhc. inversions H1; destruct_all.
  eapply TI in H5. hnf in H5.
  econstructor; eauto.
Qed.

(** Transformations lifted to heaps are type safe *)
Lemma trf_lift_heap_correspond Cond
  (trf : TermTransform)
  (TI : TransformRequires (TransformCondTermTyped typing_mode) trf):
    TransformRequires TransformCondHeapCorr (trf_lift_heap Cond trf).
Proof.
  hnf.
  introv Htrf.
  hnf. destruct TEa.
  rename T into G, a into Si, G into e.
  introv Hhca.
  inversions Hhca; destruct_all.
  rename H into HokG, H0 into HwkG, H1 into HokSi, H2 into Hoke.
  rename H3 into HdSi, H4 into Hde, H5 into Hcor.
  econstructor; try splits; eauto.
  intros x Hx.
  clear HdSi Hde Hcor.
  remember (〈 e ⊢ Si ∶ G 〉) as TEa.
  reverts x. reverts e Si G.
  hnf in Htrf.
  induction Htrf; intros.
  + (* transformation_lift_heap_e *)
    inversions HeqTEa.
    rename H into Htya, H0 into Htyb, H1 into Htrfit, H2 into Hcond.
    simpl_dom. rewrite in_union, in_singleton in Hx. destruct Hx; subst.
    * (* x *)
      forwards Htrfit': Htrfit.
      inversions Htrfit'.
      - (* fun *)
        rename H6 into Hm, H7 into HbiGa, H8 into HbiGb, H9 into Htrft, H10 into Htfa, H11 into Htfb.
        eapply ty_item_fun_s with (L:=L \u ctx_dom G); eauto.
        intros x0 Hx0. forwards Htrftx0: Htrft x0; eauto.
        eapply trf_requires in Htrftx0. destruct_all; eauto.
      - (* obj *)
        rename H6 into HbiGa, H7 into HbiGb, H8 into Htrfds, H9 into Hvka, H10 into Hvkb, H11 into Hcondds.
        eapply ty_item_obj_s. eapply HbiGa. 2: eauto. eapply binds_tail. 2: eauto.
        eapply trf_requires in Htrfds; eauto.
    * (* S1 *)
      rename H into Hx0.
      forwards: Htya; eauto.
      assert (x # x0 ~ ita) as Hxx0. {
        simpl_dom. rewrite notin_singleton.
        eapply ok_push_inv in HokSi. destruct HokSi.
        intro; subst.
        contradiction.
      }
      inversions H.
      - eapply ty_item_fun_s with (L:=L \u \{x0}); eauto.
      - eapply ty_item_obj_s; eauto.
  + (* transformation_lift_heap_n *)
    inversions HeqTEa.
    rename H into HtrfS, H0 into Htya, H1 into Htyb.
    simpl_dom. rewrite in_union, in_singleton in Hx. destruct Hx; subst.
    * 
      inversions Htya.
      - eapply ty_item_fun_s with (L:=L \u \{x}); eauto.
      - eapply ty_item_obj_s; eauto.
    *
      assert (x0 # x ~ it) as Hxx0. {
        simpl_dom. rewrite notin_singleton.
        eapply ok_push_inv in HokSi. destruct HokSi.
        intro; subst.
        contradiction.
      }
      forwards Htyx0: IHHtrf Sa x0; fweauto.
      inversions Htyx0.
      - eapply ty_item_fun_s with (L:=L \u \{x0}); eauto.
      - eapply ty_item_obj_s; eauto.
Qed.

(** Transformations lifted to stacks are type safe *)
Instance TransformLiftStackTyped Cond
  (trf : TermTransform)
  (TI : TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires TransformCondStackTyped (trf_lift_stack Cond trf).
Proof.
  hnf.
  introv Htrf.
  induction Htrf.
  - eapply ty_stack_let with (L:=L \u ctx_dom Ga); eauto.
    intros x Hx.
    forwards Htrf: (H x); eauto.
    eapply trf_requires in Htrf; eauto.
  - eapply ty_stack_let with (L:=L \u ctx_dom Ga); eauto.
Qed.

Instance TransformLiftIdtypStackTyped
  (trf : TermTransform)
  (TI : TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires TransformCondStackTyped (trf_lift_idtyp_stack trf).
Proof.
  hnf.
  introv Htrf.
  induction Htrf.
  - eapply ty_stack_let with (L:=L \u ctx_dom Ga); eauto.
    intros x Hx.
    forwards Htrf: (H x); eauto.
    eapply trf_requires in Htrf; eauto.
  - eapply ty_stack_let with (L:=L \u ctx_dom Ga); eauto.
Qed.

(* Transformations lifted to configurations are type safe *)
Instance TransformLiftConfigTyped Cond
  (trf: TermTransform)
  (TI: TransformRequires (TransformCondTermTyped typing_mode)  trf):
  TransformRequires TransformCondConfTyped (trf_lift_conf Cond trf).
Proof.
  hnf. intros.
  induction H.
  - (* Focus *)
    eapply trf_requires in H; eauto.
  - (* Heap *)
    assert (heap_correspond Ga e Sa) as Hcor. {
      eapply trf_lift_heap_correspond in H3; eauto.
    }
    inversions H0; inversions H; destruct_all.
    econstructor; eauto.
  -
    inversions H0; inversions H; destruct_all.
    econstructor; eauto.
    eapply trf_requires in H3; eauto.
Qed.


(** ** Reduction *)

(** trf1 reduces to trf2 in one step *)
Class TransformConfigReducesTo1(trf1 trf2: TransformConfig) :=
  trf_config_reduces_to_1 Ga1 Gb1 Ta Tb Ca1 Cb1 Ga2 Gb2 Ca2 Cb2 :
    〈Ga1 ⊢ Ca1 ∶ Ta〉 ∼trf1∼> 〈Gb1 ⊢ Cb1 ∶ Tb〉 ->
    Ga1 ⊢ Ca1 ∶ Ta ↦ Ga2 ⊢ Ca2 ->
    Gb1 ⊢ Cb1 ∶ Tb ↦ Gb2 ⊢ Cb2 ->
    〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼trf2∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉.

(** trf1 reduces to trf2 in some number of steps *)
Class TransformConfigReducesTo(trf1 trf2: TransformConfig) :=
  trf_config_reduces_to Ga1 Gb1 Ta Tb Ca1 Cb1 :
    〈Ga1 ⊢ Ca1 ∶ Ta〉 ∼trf1∼> 〈Gb1 ⊢ Cb1 ∶ Tb〉 ->
    exists n1 n2 Ga2 Gb2 Ca2 Cb2, 
    Ga1 ⊢ Ca1 ∶ Ta ⊣ n1 ↦* Ga2 ⊢ Ca2 /\
    Gb1 ⊢ Cb1 ∶ Tb ⊣ n2 ↦* Gb2 ⊢ Cb2 /\
    〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼trf2∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉.

(** trf1 reduces to trf2, the left side in less or equal number of steps as the right *)
Class TransformConfigReducesToShort(trf1 trf2: TransformConfig) :=
  trf_config_reduces_to_short Ga1 Gb1 Ta Tb Ca1 Cb1 :
    〈Ga1 ⊢ Ca1 ∶ Ta〉 ∼trf1∼> 〈Gb1 ⊢ Cb1 ∶ Tb〉 ->
    exists n1 n2 Ga2 Gb2 Ca2 Cb2, 
    n2 <= n1 /\
    Ga1 ⊢ Ca1 ∶ Ta ⊣ n1 ↦* Ga2 ⊢ Ca2 /\
    Gb1 ⊢ Cb1 ∶ Tb ⊣ n2 ↦* Gb2 ⊢ Cb2 /\
    〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼trf2∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉.

(** trf1 reduces to trf2, the left side in less or equal number of steps as the right, given that the left side terminates *)
(* Definition 56 in "Pure methods for roDOT" *)
Class TransformConfigReducesToShortIfTerminates(trf1 trf2: TransformConfig) :=
  trf_config_reduces_to_short_if_terminates Ga1 Gb1 Ta Tb Ca1 Cb1 :
    〈Ga1 ⊢ Ca1 ∶ Ta〉 ∼trf1∼> 〈Gb1 ⊢ Cb1 ∶ Tb〉 ->
    config_terminates Ca1 ->
    exists n1 n2 Ga2 Gb2 Ca2 Cb2, 
    n2 <= n1 /\
    Ga1 ⊢ Ca1 ∶ Ta ⊣ n1 ↦* Ga2 ⊢ Ca2 /\
    Gb1 ⊢ Cb1 ∶ Tb ⊣ n2 ↦* Gb2 ⊢ Cb2 /\
    〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼trf2∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉.

(** Preserving termination *)
Definition TransformCondTerminates (TEa: TypedElement ctx typ config) :=
  match TEa with
    〈_ ⊢ C ∶ _〉 => config_terminates C
  end.

(* Part of Definition 54 in "Pure methods for roDOT" *)
(** Preserving answers *)
Definition TransformIsAnswer (TEa: TypedElement ctx typ config) :=
  match TEa with
    〈_ ⊢ C ∶ _〉 => answer C
  end.

Global Instance TransformPreservesIsAnswerFocus Cond (trf: TermTransform) (TI: TransformPreserves TransformIsAnswerTerm trf):
  TransformPreserves TransformIsAnswer  (trf_focus_conf Cond trf).
Proof.
  hnf; introv Htrf Hans.
  hnf in *.
  inversions Htrf.
  inversions Hans.
  eapply TI in H5. inversions H5. econstructor.
  hnf. eauto.
Qed.

Global Instance TransformPreservesIsAnswerLift Cond (trf: TermTransform) (TI: TransformPreserves TransformIsAnswerTerm  trf):
  TransformPreserves TransformIsAnswer (trf_lift_conf Cond trf).
Proof.
  hnf; introv Htrf Hans.
  inversions Htrf.
  - eapply TransformPreservesIsAnswerFocus in H; eauto. eapply IsAnswerTransformationTrm; eauto.
  - hnf; inversions Hans; eauto.
  - hnf; inversions Hans; eauto.
    inversions H3.
Qed.

Global Instance TransformPreservesIsAnswerFocusIdtyp (trf: TermTransform) (TI: TransformPreserves TransformIsAnswerTerm trf):
  TransformPreserves TransformIsAnswer  (trf_focus_idtyp_conf trf).
Proof.
  hnf; introv Htrf Hans.
  hnf in *.
  inversions Htrf.
  inversions Hans.
  eapply TI in H3. inversions H3. econstructor.
  hnf. eauto.
Qed.

Global Instance TransformPreservesIsAnswerLiftIdtyp (trf: TermTransform) (TI: TransformPreserves TransformIsAnswerTerm  trf):
  TransformPreserves TransformIsAnswer (trf_lift_idtyp_conf trf).
Proof.
  hnf; introv Htrf Hans.
  inversions Htrf.
  - eapply TransformPreservesIsAnswerFocusIdtyp in H; eauto. eapply IsAnswerTransformationTrmIdtyp; eauto.
  - hnf; inversions Hans; eauto.
  - hnf; inversions Hans; eauto.
    inversions H2.
Qed.


Lemma transformation_inverse_focus_conf Cond (trf: TermTransform):
  trf_inverse (trf_focus_conf Cond trf) = trf_focus_conf (trf_liftcond_inverse Cond) (trf_inverse trf).
Proof.
  eapply pred_ext_2.
  intros.
  unfold trf_inverse.
  splits; intros.
  - inversions H. econstructor; eauto.
  - inversions H. econstructor; eauto.
Qed.

Lemma trf_inverse_lift_heap Cond (trf: TermTransform):
  trf_inverse (trf_lift_heap Cond trf) = trf_lift_heap (trf_liftcond_inverse Cond) (trf_inverse trf).
Proof.
  eapply pred_ext_2.
  intros. splits; intros.
  -
   induction H.
      * (* = *)
        eapply trf_lift_heap_e; eauto.
        inversions H1.
        ** (* fun *)
          eapply trf_lift_item_fun with (L:=L); eauto.
          intros. 
          forwards: H13 x0 H1.
          rewrite <- (trf_local_inverse_trm) with (Cond:=Cond).
          auto.
        ** (* obj *)
          eapply trf_lift_item_obj; eauto.
          intros. 
          rewrite <- trf_local_inverse_defs.
          auto.
      * (* <> *)
        eapply trf_lift_heap_n; eauto.
  - induction H.
      * (* = *)
        eapply trf_lift_heap_e; eauto.
        inversions H1.
        ** (* fun *)
          eapply trf_lift_item_fun with (L:=L); eauto.
          intros. 
          forwards: H13 x0 H1.
          rewrite <- trf_local_inverse_trm in H3.
          auto.
        ** (* obj *)
          eapply trf_lift_item_obj; eauto.
          intros. 
          rewrite <- trf_local_inverse_defs in H12.
          auto.
      * (* <> *)
        eapply trf_lift_heap_n; eauto.
Qed.

Lemma trf_inverse_lift_stack Cond (trf: TermTransform):
  trf_inverse (trf_lift_stack Cond trf) = trf_lift_stack (trf_liftcond_inverse Cond) (trf_inverse trf).
Proof.
  eapply pred_ext_2.
  intros. splits; intros.
  -
   induction H; intros.
      * eapply trf_lift_stack_e; eauto.
          intros x Hx. 
          forwards: H x Hx.
          rewrite <- trf_local_inverse_trm. 
          auto. 
      * eapply trf_lift_stack_n; eauto.
  - induction H; intros.
      * eapply trf_lift_stack_e; eauto.
          intros x Hx. 
          forwards: H x Hx.
          rewrite <- trf_local_inverse_trm in H3.
          auto.
      * eapply trf_lift_stack_n; eauto.
Qed.

Lemma trf_inverse_lift_conf Cond (trf: TermTransform):
  trf_inverse (trf_lift_conf Cond trf) = trf_lift_conf (trf_liftcond_inverse Cond) (trf_inverse trf).
Proof.
  eapply pred_ext_2.
  intros.
  splits; intros.
  - inversions H.
    inversions H0.
    + (* term *)
    eapply trf_lift_conf_focus; eauto.
    econstructor; eauto. 
    rewrite <- trf_local_inverse_trm.
    auto.
    + (* heap *)
      eapply trf_lift_conf_heap; eauto.
      rewrite <- trf_inverse_lift_heap; auto.
    + (* stack *)
      eapply trf_lift_conf_stack; eauto.
      rewrite <- trf_inverse_lift_stack; auto.
  - (* The other direction - copy of the above, except rewriting happens in hypotheses *)
    inversions H.
    inversions H0.
    + (* term *)
      eapply trf_lift_conf_focus; eauto.
      econstructor; eauto.
      rewrite <- trf_local_inverse_trm in H13.
      auto.
    + (* heap *)
      eapply trf_lift_conf_heap; eauto.
      rewrite <- trf_inverse_lift_heap in H4; auto.
    + (* stack *)
      eapply trf_lift_conf_stack; eauto.
      rewrite <- trf_inverse_lift_stack in H4; auto.
Qed.

Lemma transformation_inverse_focus_idtyp_conf (trf: TermTransform):
  trf_inverse (trf_focus_idtyp_conf trf) = trf_focus_idtyp_conf (trf_inverse trf).
Proof.
  eapply pred_ext_2.
  intros.
  unfold trf_inverse.
  splits; intros.
  - inversions H. econstructor; eauto.
  - inversions H. econstructor; eauto.
Qed.

Lemma trf_inverse_lift_idtyp_heap (trf: TermTransform) (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf):
  trf_inverse (trf_lift_idtyp_heap trf) = trf_lift_idtyp_heap (trf_inverse trf).
Proof.
  eapply pred_ext_2.
  intros. splits; intros.
  -
   induction H.
      * (* = *)
        eapply trf_lift_idtyp_heap_e; eauto.
        inversions H0.
        ** (* fun *)
          eapply trf_lift_idtyp_item_fun with (L:=L); eauto.
          intros. 
          forwards: H8 x0 H0.
          rewrite <- (trf_local_idtyp_inverse_trm).
          auto. auto.
        ** (* obj *)
          eapply trf_lift_idtyp_item_obj; eauto.
          intros. 
          rewrite <- trf_local_idtyp_inverse_defs.
          auto. auto.
      * (* <> *)
        eapply trf_lift_idtyp_heap_n; eauto.
  - induction H.
      * (* = *)
        eapply trf_lift_idtyp_heap_e; eauto.
        inversions H0.
        ** (* fun *)
          eapply trf_lift_idtyp_item_fun with (L:=L); eauto.
          intros. 
          forwards: H8 x0 H0.
          rewrite <- trf_local_idtyp_inverse_trm in H1.
          auto. auto.
        ** (* obj *)
          eapply trf_lift_idtyp_item_obj; eauto.
          intros. 
          rewrite <- trf_local_idtyp_inverse_defs in H7.
          auto. auto.
      * (* <> *)
        eapply trf_lift_idtyp_heap_n; eauto.
Qed.

Lemma trf_inverse_lift_idtyp_stack (trf: TermTransform) (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf):
  trf_inverse (trf_lift_idtyp_stack trf) = trf_lift_idtyp_stack (trf_inverse trf).
Proof.
  eapply pred_ext_2.
  intros. splits; intros.
  -
   induction H; intros.
      * eapply trf_lift_idtyp_stack_e; eauto.
          intros x Hx. 
          forwards: H x Hx.
          rewrite <- trf_local_idtyp_inverse_trm. 
          auto. auto. 
      * eapply trf_lift_idtyp_stack_n; eauto.
  - induction H; intros.
      * eapply trf_lift_idtyp_stack_e; eauto.
          intros x Hx. 
          forwards: H x Hx.
          rewrite <- trf_local_idtyp_inverse_trm in H1.
          auto. auto.
      * eapply trf_lift_idtyp_stack_n; eauto.
Qed.

Lemma trf_inverse_lift_idtyp_conf (trf: TermTransform) (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf):
  trf_inverse (trf_lift_idtyp_conf trf) = trf_lift_idtyp_conf (trf_inverse trf).
Proof.
  eapply pred_ext_2.
  intros.
  splits; intros.
  - inversions H.
    inversions H0.
    + (* term *)
    eapply trf_lift_idtyp_conf_focus; eauto.
    econstructor; eauto. 
    rewrite <- trf_local_idtyp_inverse_trm.
    auto. auto.
    + (* heap *)
      eapply trf_lift_idtyp_conf_heap; eauto.
      rewrite <- trf_inverse_lift_idtyp_heap; auto.
    + (* stack *)
      eapply trf_lift_idtyp_conf_stack; eauto.
      rewrite <- trf_inverse_lift_idtyp_stack; auto.
  - (* The other direction - copy of the above, except rewriting happens in hypotheses *)
    inversions H.
    inversions H0.
    + (* term *)
      eapply trf_lift_idtyp_conf_focus; eauto.
      econstructor; eauto.
      rewrite <- trf_local_idtyp_inverse_trm in H8.
      auto. auto.
    + (* heap *)
      eapply trf_lift_idtyp_conf_heap; eauto.
      rewrite <- trf_inverse_lift_idtyp_heap in H4; auto.
    + (* stack *)
      eapply trf_lift_idtyp_conf_stack; eauto.
      rewrite <- trf_inverse_lift_idtyp_stack in H3; auto.
Qed.

Instance TrfFocusIdtypEnsuresCond trf:
  TransformEnsures (@TransformCondTypeIdentical _ _ _) (trf_focus_idtyp_conf trf).
Proof.
  hnf; intros.
  hnf in *; intros.
  inversions H.
  auto.
Qed.
Instance TrfLiftIdtypEnsuresCond trf:
  TransformEnsures (@TransformCondTypeIdentical _ _ _) (trf_lift_idtyp_conf trf).
Proof.
  hnf; intros.
  hnf in *; intros.
  inversions H; auto.
Qed.

(* Conversions between idtyp and general definitions *)
Lemma trf_focus_idtyp_conf_imp trf C1 C2:
  C1 ∼(trf_focus_idtyp_conf trf)∼> C2 -> 
  C1 ∼(trf_focus_conf TransformCondTypeIdentical trf)∼> C2.
Proof.
  introv Htrf.
  inversions Htrf.
  econstructor; eauto.
  constructor; auto.
Qed.
Lemma trf_lift_idtyp_stack_imp trf C1 C2 (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf):
  C1 ∼(trf_lift_idtyp_stack trf)∼> C2 -> 
  C1 ∼(trf_lift_stack TransformCondTypeIdentical trf)∼> C2.
Proof.
  introv Htrf.
  induction Htrf.
  -
  econstructor; eauto.
  intros. rewrite  trf_local_trm_idtyp_eq. eauto. auto.
  econstructor; eauto.
  -
  econstructor; eauto.
  econstructor; eauto.
Qed.
Lemma trf_lift_idtyp_item_imp trf C1 C2 (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf):
  C1 ∼(trf_lift_idtyp_item trf)∼> C2 -> 
  C1 ∼(trf_lift_item TransformCondTypeIdentical trf)∼> C2.
Proof.
  introv Htrf.
  induction Htrf.
  -
  econstructor; eauto.
  intros. rewrite  trf_local_trm_idtyp_eq. eauto. auto.
  econstructor; eauto.
  -
  econstructor; eauto.
  intros. rewrite  trf_local_defs_idtyp_eq. eauto. auto.
  econstructor; eauto.
Qed.

Lemma trf_lift_idtyp_heap_imp trf C1 C2 (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf):
  C1 ∼(trf_lift_idtyp_heap trf)∼> C2 -> 
  C1 ∼(trf_lift_heap TransformCondTypeIdentical trf)∼> C2.
Proof.
  introv Htrf.
  induction Htrf.
  -
  econstructor; eauto.
  eapply trf_lift_idtyp_item_imp; eauto.
  econstructor; eauto.
  -
  econstructor; eauto.
  econstructor; eauto.
Qed.


Lemma trf_lift_idtyp_conf_imp trf C1 C2 (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf):
  C1 ∼(trf_lift_idtyp_conf trf)∼> C2 -> 
  C1 ∼(trf_lift_conf TransformCondTypeIdentical trf)∼> C2.
Proof.
  introv Htrf.
  induction Htrf.
  -
  eapply trf_lift_conf_focus; eauto.
  eapply trf_focus_idtyp_conf_imp; eauto.
  rewrite  trf_local_trm_idtyp_eq. eauto. auto.
  econstructor; eauto.
  -
  eapply trf_lift_conf_heap; eauto.
  eapply trf_lift_idtyp_heap_imp; eauto.
  econstructor; eauto.
  -
  eapply trf_lift_conf_stack; eauto.
  eapply trf_lift_idtyp_stack_imp; eauto.
  econstructor; eauto.
Qed.

(*
Cannot have an equality like this, because the non-type-identical version can have different e-parameter
Lemma trf_focus_conf_idtyp_eq trf:
  (trf_focus_conf TransformCondTypeIdentical trf) = (trf_focus_idtyp_conf trf).
Proof.
  eapply pred_ext_2; intros.
  split; intros; inversions H.
  - inversions H7. econstructor. eauto.

  - eapply (proj41 (trf_local_local_idtyp_ind trf)); eauto. 
  - eapply (proj41 (trf_local_idtyp_local_ind trf)); eauto.
Qed.
*)


Lemma trf_lift_idtyp_heap_correspond
  (trf : TermTransform) (TE: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf)
  (TI : TransformRequires (TransformCondTermTyped typing_mode) trf):
    TransformRequires TransformCondHeapCorr (trf_lift_idtyp_heap trf).
Proof.
  hnf; intros.
  eapply trf_lift_idtyp_heap_imp in H.
  eapply trf_lift_heap_correspond. eauto. eauto. auto.
Qed.

(* Typing is required *)
Global Instance TransformFocusIdtypConfigTyped trf
  (TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires TransformCondConfTyped (trf_focus_idtyp_conf trf).
Proof.
  hnf; intros.
  eapply trf_focus_idtyp_conf_imp in H.
  eapply TransformFocusConfigTyped; eauto.
Qed.

Global Instance TransformLiftIdtypConfigTyped trf
  (TI: TransformRequires (TransformCondTermTyped typing_mode) trf)
  (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf):
  TransformRequires TransformCondConfTyped (trf_lift_idtyp_conf trf).
Proof.
  hnf; intros.
  eapply trf_lift_idtyp_conf_imp in H.
  eapply TransformLiftConfigTyped; eauto. auto.
Qed.

Global Instance WknCtxPushCompatTransformationStack trf
  (TI: TransformWeakeningCompat _ _ trf):
  TransformWknCtxPushCompat _ _ (trf_lift_idtyp_stack trf).
Proof.
  hnf. introv Htrf Hwk1 Hwk2 Hid.
  reverts TEa2 TEb2.
  induction Htrf; intros.
  - hnf in Hid. destruct TEa2, TEb2. inversions Hid.
    hnf in Hwk1. inversions Hwk1.
    hnf in Hwk2. inversions Hwk2.
    eapply trf_lift_idtyp_stack_e with (L:=L \u ctx_dom (Ga § x / vk ~ Tx)).
    intros.
    eapply WeakeningCompatTransformationTrm; eauto.
    + eapply (H x0).  notin_solve.
    +
      hnf.
      ctx_split_between x x0.
      rewrite ctx_concat_single.
      rewrite ctx_concat_single with (G:= (x / vk ~ Tx)).
      rewrite ! ctx_concat_assoc.
      econstructor;  eauto. ctx_flatten. eauto.
    +
      hnf.
      ctx_split_between x x0.
      rewrite ctx_concat_single.
      rewrite ctx_concat_single with (G:= (x / vk ~ Tx)).
      rewrite ! ctx_concat_assoc.
      econstructor;  eauto. ctx_flatten. eauto.
    +
      hnf. eauto.
    +
      eapply ty_stack_push; eauto.
  -
    hnf in Hid. destruct TEa2, TEb2. inversions Hid.
    hnf in Hwk1. inversions Hwk1.
    hnf in Hwk2. inversions Hwk2.

    eapply trf_lift_idtyp_stack_n.
    eapply IHHtrf.
    +
      hnf. econstructor; eauto.
    +
      hnf. econstructor; eauto.
    +
      hnf. eauto.
    + intros.
      ctx_split_between x x0.
      eapply weaken_ty; ctx_flatten; eauto.
Qed.
Global Instance WknEnvPushCompatTransformationStack trf (TI: TransformEWeakeningCompat _ _ trf):
  TransformWknEnvPushCompat _ _ (trf_lift_idtyp_stack trf).
Proof.
  hnf. introv Htrf Hwk1 Hwk2 Hid.
  reverts TEa2 TEb2.
  induction Htrf; intros.
  - hnf in Hid. destruct TEa2, TEb2. inversions Hid.
    hnf in Hwk1. inversions Hwk1.
    hnf in Hwk2. inversions Hwk2.
    eapply trf_lift_idtyp_stack_e with (L:=L \u ctx_dom Ga).
    intros.
    eapply EWeakeningCompatTransformationTrm; eauto.
    + eapply (H x1).  notin_solve.
    +
      hnf.
      rewrite <- concat_empty_r with (E := ea & x0 ~ xy0).
      rewrite <- concat_empty_r with (E := ea) at 1.
      econstructor; rewrite ? concat_empty_r; eauto.
    +
      hnf.
      rewrite <- concat_empty_r with (E := ea & x0 ~ xy0).
      rewrite <- concat_empty_r with (E := ea) at 1.
      econstructor; rewrite ? concat_empty_r; eauto.
    +
      hnf. eauto.
    +
      eapply ty_stack_epush; eauto.
  -
    hnf in Hid. destruct TEa2, TEb2. inversions Hid.
    hnf in Hwk1. inversions Hwk1.
    hnf in Hwk2. inversions Hwk2.

    eapply trf_lift_idtyp_stack_n.
    eapply IHHtrf.
    +
      hnf. econstructor; eauto.
    +
      hnf. econstructor; eauto.
    +
      hnf. eauto.
    + intros.
      rewrite <- concat_empty_r with (E := ea & x0 ~ xy0).
      eapply ewkn; rewrite ? concat_empty_r; eauto.
Qed.

Lemma trf_lift_idtyp_stack_seal (TL: TermTransform) (TI: TransformSealingCompat _ _ TL) G e fsa fsb T2 T3 x:
    closed fsa ->
    closed fsb ->
      〈G⁏e ⊢ fsa ∶ (T2, T3)〉 ∼(trf_lift_idtyp_stack TL)∼> 〈G⁏e ⊢ fsb ∶ ( T2, T3)〉 ->
      〈seal x G⁏e ⊢ fsa ∶ (seal x T2,seal x T3)〉 ∼(trf_lift_idtyp_stack TL)∼> 〈seal x G⁏e ⊢ fsb ∶ (seal x T2,seal x T3)〉.
Proof.
  introv Hcla Hclb Htrf.
  remember (〈  G ⁏ e ⊢ fsa ∶ (T2, T3) 〉 ).
  remember (〈  G ⁏ e ⊢ fsb ∶ (T2 , T3) 〉 ).
  reverts fsa fsb T2 T3.
  induction Htrf; intros; inversions Heqt; inversions Heqt0.
  - eapply trf_lift_idtyp_stack_e with (L:=L) (T2a:=T2a /[# 0 ->> x]).
    intros.
    eapply SealingCompatTransformationTrm. eauto.
    eapply (H x0). eauto.
    rewrite <- ctx_seal_push.
    eapply stack_closed_t in Hcla.
    rewrite <- closed_sealed0 with (X:=(ta /[-> x0])) at 2 by eauto. 
    econstructor.
    eapply stack_closed_t in Hclb.
    rewrite <- closed_sealed0 with (X:=(tb /[-> x0])) at 2 by eauto. 
    econstructor.
    hnf; eauto.
    eapply seal_ty_stack; eauto using stack_closed_fs.
  - eapply trf_lift_idtyp_stack_n with (L:=L) (T2a:=T2a /[# 0 ->> x]).
    eapply IHHtrf; eauto using stack_closed_fs.
    intros.
    eapply stack_closed_t in Hcla.
    rewrite <- closed_sealed0 with (X:=(t /[-> x0])) by eauto. 
    rewrite <- ctx_seal_push.
    eapply sealing_obj with (Go:=objctx_empty). eauto.

Qed.

Lemma trf_lift_idtyp_stack_seal_closed (TL: TermTransform) (TI: TransformSealingCompat _ _ TL) G e fsa fsb T2 T3 x:
    closed fsa ->
    closed fsb ->
  sealed G ->
  closed T3 ->
      〈G⁏e ⊢ fsa ∶ (T2, T3)〉 ∼(trf_lift_idtyp_stack TL)∼> 〈G⁏e ⊢ fsb ∶ ( T2, T3)〉 ->
      〈G⁏e ⊢ fsa ∶ (seal x T2,T3)〉 ∼(trf_lift_idtyp_stack TL)∼> 〈G⁏e ⊢ fsb ∶ (seal x T2, T3)〉.
Proof.
  introv Hcla Hclb HclG HclT Htrf.
  forwards: trf_lift_idtyp_stack_seal x Htrf; eauto.
  rewrite HclG in H.
  rewrite (closed_sealed0 HclT) in H.
  auto.
Qed.

Global Instance SymmetricTransformFocusIdtypConfig trf (TI: SymmetricTransform trf):
  SymmetricTransform (trf_focus_idtyp_conf trf).
Proof.
  hnf; intros.
  hnf.
  inversions H.
  econstructor; eauto.
Qed.

Global Instance SymmetricTransformLiftIdtypConfig trf (TI: SymmetricTransform trf) (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf):
  SymmetricTransform (trf_lift_idtyp_conf trf).
Proof.
  hnf; intros.
  eapply trf_inverse_intro in H. 
  rewrite trf_inverse_lift_idtyp_conf in H.
  rewrite trf_symmetric_inverse in H. eauto. eauto.
  auto.
Qed.

End WithTypingMode.
