(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation Weakening EWeakening Substitution TFree.
Require Import TransformationSimilarity TransformationPreservation TransformationTerm.
Require Import TransformationConfig OperationalSemantics TypedOperationalSemantics.
Require Import ContextSimilarity ConfigSimilarity.
Require Import ConfigTyping ConfigTypingAccessories TypedProgress.
Require Import TypedPreservation.
Require Import TransformationIterateSim.
Require Import Similarity TypingSimilarity.

Require Import TransformationLifting.
Require Import FreshChain FreshChainTactics.
Require Import TypedOperationalSemantics InertContext.
Require Import Similarity SyntaxSimilarity.
Require Import ExecutionTactics Renaming TypingHelpers.

Require Import NothingFresh TypedPreservation LetLet Defs.
Require Import HeapCorrespondence EnvCorrespondence ConfigVarkinds.
Require Import HeapUpdate ConfigValid Deref TransformationTyping.

Require Import HeapUpdate.
Require Import HeapInstantiate.
Require Import RecordTypes Subrecord.

Section WithTypingMode.
Context (typing_mode: Mode).

(** * Transformations and steps of reduction *)

(* In this file, we assume that the fresh variables are the same on both sides - this is solved at a higher level by similarity *)

(* Transformation of heap does not affect the keys and their order *)
Lemma trf_lift_idtyp_heap_keys TL G e Sa Sb:
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  keys Sa = keys Sb.
Proof.
  introv Htrf.
  remember (〈 e ⊢ Sa ∶ G 〉).
  remember (〈 e ⊢ Sb ∶ G 〉).
  reverts Sa Sb G e.
  induction Htrf; intros; inversions Heqt; inversions Heqt0.
  - rewrite !env_push_keys. auto.
  - rewrite !env_push_keys. erewrite IHHtrf; eauto.
Qed.

Lemma trf_lift_idtyp_defs_update TL G e dsa1 dsb1 dsa2 dsb2 x f y U T Ty S:
  〈 G §§ objctx_heap x U ⁏ e ⊢ dsa1 ∶ T /[-> x] 〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈 G §§ objctx_heap x U ⁏ e ⊢ dsb1 ∶ T /[-> x] 〉 ->
  defs_update dsa1 f y dsa2 ->
  defs_update dsb1 f y dsb2 ->
  (mode_visibility' typing_mode = mode_visibility_hide_method -> ctx_vis (G §§& objctx_heap x U) y) ->
  ( G §§& objctx_heap x U ⁏ e ⊢ y ∶ Ty) ->
  ctx_binds x (typ_and (typ_bnd U ) S) G ->
  record_has (T/[->x]) (dec_fld f Ty Ty) ->
  subrecord T/[->x] U/[->x] ->
  record_type U /[-> x] ->
  〈 G §§ objctx_heap x U ⁏ e ⊢ dsa2 ∶ T /[-> x] 〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈 G §§ objctx_heap x U ⁏ e ⊢ dsb2 ∶ T /[-> x] 〉.
Proof.
  introv Htrf Hua Hub Hvisy Htyy HbiGx HhasUf Hsur HrcU.
  remember ( 〈 G §§ objctx_heap x U ⁏ e ⊢ dsa1 ∶ T /[-> x] 〉 ).
  remember ( 〈 G §§ objctx_heap x U ⁏ e ⊢ dsb1 ∶ T /[-> x] 〉 ).
  reverts G x U e.
  reverts T.
  reverts dsa1 dsb1 dsa2 dsb2.
  induction Htrf; intros; inversions Heqt; inversions Heqt0.
  - inversions Hua; inversions Hub; inversions H.
    inversions H3.
  - destruct T0; try discriminate.
    destruct T0_2; try discriminate.
    clear H7.
    sympl in H4. inversions H4.
    inversions Hua; inversions Hub.
    + eapply trf_local_idtyp_defs_consl. 
      auto.

      inversions H0.
      destruct d; sympls; inversions H9.
      (* Need to unify Ty and t *)


      forwards: record_has_functional HhasUf.
      eapply subrecord_record. eauto. eauto. eapply rh_andr. eapply rh_one. sympl; eauto.
      inversions H0.


      eapply ty_def_fld. auto. unfold ctx_obj. auto.
      simpls; auto.
      simpls; auto.
    + simpl in H9. contradiction.
    + simpl in H9. contradiction.
    + eapply trf_local_idtyp_defs_consl.
      forwards: IHHtrf dsla dslb.
      8: eauto.
      8: eauto.

      eauto.
      eauto.

      2: eauto.
      2: eauto.
      3: eauto.
      2: { sympls. eapply subrecord_and_inv. eauto. }

      inversions HhasUf.
      eauto.
      inversions H8.
      destruct d; inversions H7.

      eapply ty_def_label_eq in H0. sympl in H0. contradiction.

      sympls.

      eauto.
      auto.
      eauto.

      eapply defs_update_hasnt; eauto.
      eapply defs_update_hasnt; eauto.
  - inversions Hua; inversions Hub; lets H': H; inversions H.
    forwards: defs_update_functional H6 H8. subst.
    eapply trf_local_idtyp_defs_consr.
    auto.
    destruct T0; inversions H4.
    destruct T0_2; inversions H9.
    destruct d; inversions H3.
    sympls.

    eapply defs_update_typing. eauto. 
    eapply subrecord_and_inv in Hsur.
    eapply subrecord_record; eauto.
    eauto.
    3: eauto.
    inversions HhasUf. eauto.
    inversions H9.

    eauto.
    eapply defs_update_hasnt; eauto.
    eapply defs_update_hasnt; eauto.
Qed.

Lemma trf_lift_idtyp_item_obj_update TL G e dsa1 dsb1 dsa2 dsb2 x f y U S Ty:
  〈 G ⁏ e ⊢ item_obj dsa1 ∶ x 〉 ∼ trf_lift_idtyp_item typing_mode TL ∼> 〈 G ⁏ e ⊢ item_obj dsb1 ∶ x 〉 ->
  defs_update dsa1 f y dsa2 ->
  defs_update dsb1 f y dsb2 ->
  (mode_visibility' typing_mode = mode_visibility_hide_method -> ctx_vis (G §§& objctx_heap x U) y) ->
  ( G §§& objctx_heap x U ⁏ e ⊢ y ∶ Ty) ->
  ctx_binds x (typ_and (typ_bnd U ) S) G ->
  record_has (U/[->x]) (dec_fld f Ty Ty) ->
  record_type U /[-> x] ->
  ctx_k G y v_loc ->
  〈 G ⁏ e ⊢ item_obj dsa2 ∶ x 〉 ∼ trf_lift_idtyp_item typing_mode TL ∼> 〈 G ⁏ e ⊢ item_obj dsb2 ∶ x 〉.
Proof.
  introv Htrf Hua Hub Hvisy Htyy HbiGx HhasUf HrcU Hk.
  inversions Htrf.

  ctx_binds_eq.
  econstructor.
  eauto.
  eauto.
  eapply trf_lift_idtyp_defs_update; eauto.
  eapply subrecord_refl.
  eapply defs_update_varkinds; eauto.
  eapply defs_update_varkinds; eauto.
Qed.

Lemma ty_item_s_push_n_inv G e S x it y:
  ty_item_s typing_mode G e (S & x ~ it) y ->
  x <> y ->
  ty_item_s typing_mode G e S y.
Proof.
  introv Hty Hn.
  inversions Hty.
  eapply binds_push_neq_inv in H1.
  eauto.
  auto.
  eapply binds_push_neq_inv in H0.
  eauto.
  auto.
Qed.

Lemma ty_item_s_push_n G e S x it y:
  ty_item_s typing_mode G e S y ->
  x <> y ->
  ty_item_s typing_mode G e (S & x ~ it) y.
Proof.
  introv Hty Hn.
  inversions Hty.
  eauto.
  eauto.
Qed.

Lemma ty_item_s_concat_invl G e S1 S2 y:
  ok (S1 & S2) ->
  ty_item_s typing_mode G e (S1 & S2) y ->
  y \in dom S1 ->
  ty_item_s typing_mode G e (S1) y.
Proof.
  introv Hok Hty Hn.
  inversions Hty.

  eapply binds_concat_ok_inv in H1.  
  destructecs H1; try contradiction; eauto. auto.

  eapply binds_concat_ok_inv in H0.  
  destructecs H0; try contradiction; eauto. auto.
Qed.

Lemma ty_item_s_concat_invr G e S1 S2 y:
  ok (S1 & S2) ->
  ty_item_s typing_mode G e (S1 & S2) y ->
  y \in dom S2 ->
  ty_item_s typing_mode G e (S2) y.
Proof.
  introv Hok Hty Hn.
  inversions Hty.

  eapply binds_concat_ok_inv in H1.  
  destructecs H1; try contradiction; eauto. auto.

  eapply binds_concat_ok_inv in H0.  
  destructecs H0; try contradiction; eauto. auto.
Qed.


Lemma trf_lift_idtyp_heap_update TL G e Sa1 Sb1 Sa2 Sb2 x f y U S Ty Sc1 Sc2:
  〈 e ⊢ Sa1 ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb1 ∶ G 〉 ->
  heap_update Sa1 x f y Sa2 ->
  heap_update Sb1 x f y Sb2 ->
  keys Sa1 = keys Sa2 ->
  keys Sb1 = keys Sb2 ->

  (mode_visibility' typing_mode = mode_visibility_hide_method -> ctx_vis (G §§& objctx_heap x U) y) ->
  ( G §§& objctx_heap x U ⁏ e ⊢ y ∶ Ty) ->
  ctx_binds x (typ_and (typ_bnd U ) S) G ->
  record_has (U/[->x]) (dec_fld f Ty Ty) ->
  record_type U /[-> x] ->
  ctx_k G y v_loc ->
  heap_correspond G e (Sa1 & Sc1) ->
  heap_correspond G e (Sa2 & Sc2) ->
  〈 e ⊢ Sa2 ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb2 ∶ G 〉.
Proof.
  introv Htrf Hua Hub Hka Hkb Hvisy Htyy HbiGx.
  introv HhasUf HrcU Hk.
  introv HcorSa1 HcorSa2.
  remember (〈 e ⊢ Sa1 ∶ G 〉).
  remember (〈 e ⊢ Sb1 ∶ G 〉).
  reverts Sa1 Sb1 G e.
  reverts Sa2 Sb2.
  reverts Sc1 Sc2.
  induction Htrf; intros; inversions Heqt; inversions Heqt0.
  - (* Transformed here *)
    eapply keys_push_eq_inv in Hkb.
    eapply keys_push_eq_inv in Hka.
    destructecs Hkb.
    destructecs Hka.
    subst.

    forwards: heap_update_keys_eq_cases Hua. auto.
    forwards: heap_update_keys_eq_cases Hub. auto.
    destructecs H1;     destructecs H2; try contradiction.
    + (* Updated elsewhere *)
      forwards: heap_update_keys_eq H3 H5; fweauto. subst.

      hnf in HcorSa1.
      hnf in HcorSa2.
      destructecs HcorSa1.
      destructecs HcorSa2.

      eapply trf_lift_idtyp_heap_e.
      intros.
      forwards: HcorSa2 y1.  simpl_dom. union_in_solver.

      eapply  ty_item_s_concat_invl in H2.
      eapply  ty_item_s_concat_invl in H2.
      auto.
      eauto.
      eauto.
      eauto.
      simpl_dom. union_in_solver.
        

      auto.
    + (* updated here *) subst.
      eapply trf_lift_idtyp_heap_e. auto.
      eapply trf_lift_idtyp_item_obj_update; eauto.
  - (* Transformed elsewhere*)

    eapply keys_push_eq_inv in Hkb.
    eapply keys_push_eq_inv in Hka.
    destructecs Hkb.
    destructecs Hka.
    subst.
    forwards: heap_update_keys_eq_cases Hua. auto.
    forwards: heap_update_keys_eq_cases Hub. auto.

    destructecs H1;     destructecs H2; try contradiction.
    + subst.
      eapply trf_lift_idtyp_heap_n. 
      rewrite <- concat_assoc in HcorSa2, HcorSa1.
      applys IHHtrf H3 H5; eauto.
      auto.
    + subst. inversions H9.
      forwards: defs_update_functional H1 H2. subst.
      eapply trf_lift_idtyp_heap_n.
      auto.
      hnf in HcorSa2. destructecs HcorSa2.

      forwards: HcorSa2 x0.  simpl_dom. union_in_solver.

      eapply  ty_item_s_concat_invl in H3.
      eapply  ty_item_s_concat_invr in H3.
      auto.
      eauto.
      simpl_dom. union_in_solver.
      eauto.
      simpl_dom. union_in_solver.
Qed.



Lemma trf_lift_idtyp_item_epush TL e G ita itb x w y T
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL):
  ctx_ok G ->
  w §# G ->
  ok e ->
  w # e ->
  〈 G ⁏ e ⊢ ita ∶ x 〉 ∼ trf_lift_idtyp_item typing_mode TL ∼> 〈 G ⁏ e ⊢ itb ∶ x 〉->
  〈 G §ʷ w ~ T ⁏ e & w ~ y ⊢ ita ∶ x 〉 ∼ trf_lift_idtyp_item typing_mode TL ∼> 〈 G §ʷ w ~ T ⁏ e & w ~ y ⊢ itb ∶ x 〉.
Proof.
  introv HonG HnG Hoke Hne Htrf.
  inversions Htrf.
  - eapply trf_lift_idtyp_item_fun with (L:=L \u ctx_dom G \u dom e \u \{w}); eauto.
    eapply ctx_binds_n. intro; subst. eapply HnG. eapply ctx_binds_dom in H5. eauto. eauto.
    intros.

    eapply trf_local_idtyp_trm_ewkn_push. auto.
    ctx_split_between w x0.

    eapply WeakeningCompatTransformationTrm. eauto.
    eapply (H6 x0).
    eauto.
    rewrite ctx_concat_single. econstructor; ctx_flatten. eauto. econstructor. eauto. eauto. eauto.
    rewrite ctx_concat_single. econstructor; ctx_flatten. eauto. econstructor. eauto. eauto. eauto.
    hnf; eauto.
    eauto.
    eauto.
    eauto.
  - eapply trf_lift_idtyp_item_obj; eauto.
    eapply ctx_binds_n. intro; subst. eapply HnG. eapply ctx_binds_dom in H4. eauto. eauto.
    intros.

    eapply trf_local_idtyp_defs_ewkn_push. auto.
    eapply WeakeningCompatTransformationDefs. eauto.
    eapply (H5).
    rewrite ctx_concat_single. rewrite <- ctx_concat_empty_r with (G1:=G) at 1. rewrite <- ctx_concat_empty_r with (G1:=G §& (ctx_empty §ʷ w ~ T)) at 1. econstructor; ctx_flatten. eauto. econstructor. eauto. eauto. eauto.
    rewrite ctx_concat_single. rewrite <- ctx_concat_empty_r with (G1:=G) at 1. rewrite <- ctx_concat_empty_r with (G1:=G §& (ctx_empty §ʷ w ~ T)) at 1. econstructor; ctx_flatten. eauto. econstructor. eauto. eauto. eauto.

    hnf; eauto.
    eauto.
    eauto.
    eauto.
    eapply weaken_defs_varkinds; eauto.
    eapply weaken_defs_varkinds; eauto.
Qed.

Lemma trf_lift_idtyp_heap_epush TL e G Sa Sb w y T
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL):
  ctx_ok G ->
  w §# G ->
  ok e ->
  w # e ->
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  〈 e & w ~ y ⊢ Sa ∶ G § w / v_ref ~ T〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e & w ~ y ⊢ Sb ∶ G § w / v_ref ~ T〉.
Proof.
  introv HonG HnG Hoke Hne Htrf.
  remember (〈 e ⊢ Sa ∶ G 〉).
  remember (〈 e ⊢ Sb ∶ G 〉).
  reverts Sa Sb G e.
  induction Htrf; intros; subst; inversions Heqt; inversions Heqt0.
  - eapply trf_lift_idtyp_heap_e.
    intros.
    eapply HeapCorrespondence.ty_item_s_epush; eauto. 
    eapply trf_lift_idtyp_item_epush; eauto.
  - eapply trf_lift_idtyp_heap_n.
    eapply IHHtrf; eauto.
    eapply HeapCorrespondence.ty_item_s_epush; eauto.
Qed.

Lemma trf_lift_idtyp_item_Gpush TL e G ita itb x w T vk
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL):
  ctx_ok G ->
  w §# G ->
  ok e ->
  w # e ->
  〈 G ⁏ e ⊢ ita ∶ x 〉 ∼ trf_lift_idtyp_item typing_mode TL ∼> 〈 G ⁏ e ⊢ itb ∶ x 〉->
  〈 G § w / vk ~ T ⁏ e ⊢ ita ∶ x 〉 ∼ trf_lift_idtyp_item typing_mode TL ∼> 〈 G § w / vk ~ T ⁏ e  ⊢ itb ∶ x 〉.
Proof.
  introv HonG HnG Hoke Hne Htrf.
  inversions Htrf.
  - eapply trf_lift_idtyp_item_fun with (L:=L \u ctx_dom G \u dom e \u \{w}); eauto.
    eapply ctx_binds_n. intro; subst. eapply HnG. eapply ctx_binds_dom in H5. eauto. eauto.
    intros.

    ctx_split_between w x0.

    eapply WeakeningCompatTransformationTrm. eauto.
    eapply (H6 x0).
    eauto.
    rewrite ctx_concat_single. econstructor; ctx_flatten. eauto. econstructor. eauto. eauto. eauto.
    rewrite ctx_concat_single. econstructor; ctx_flatten. eauto. econstructor. eauto. eauto. eauto.
    hnf; eauto.

  - eapply trf_lift_idtyp_item_obj; eauto.
    eapply ctx_binds_n. intro; subst. eapply HnG. eapply ctx_binds_dom in H4. eauto. eauto.
    intros.

    eapply WeakeningCompatTransformationDefs. eauto.
    eapply (H5).
    rewrite ctx_concat_single. rewrite <- ctx_concat_empty_r with (G1:=G) at 1. rewrite <- ctx_concat_empty_r with (G1:=G §& (ctx_empty § w / vk ~ T)) at 1. econstructor; ctx_flatten. eauto. econstructor. eauto. eauto. eauto.
    rewrite ctx_concat_single. rewrite <- ctx_concat_empty_r with (G1:=G) at 1. rewrite <- ctx_concat_empty_r with (G1:=G §& (ctx_empty § w  / vk ~ T)) at 1. econstructor; ctx_flatten. eauto. econstructor. eauto. eauto. eauto.

    hnf; eauto.
    eauto.
    eapply weaken_defs_varkinds; eauto.
    eapply weaken_defs_varkinds; eauto.
Qed.

Lemma trf_lift_idtyp_heap_Gpush TL e G Sa Sb w T vk
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL):
  ctx_ok G ->
  w §# G ->
  ok e ->
  w # e ->
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  〈 e  ⊢ Sa ∶ G § w / vk ~ T〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G § w / vk ~ T〉.
Proof.
  introv HonG HnG Hoke Hne Htrf.
  remember (〈 e ⊢ Sa ∶ G 〉).
  remember (〈 e ⊢ Sb ∶ G 〉).
  reverts Sa Sb G e.
  induction Htrf; intros; subst; inversions Heqt; inversions Heqt0.
  - eapply trf_lift_idtyp_heap_e.
    intros.
    eapply HeapCorrespondence.ty_item_s_Gpush; eauto. 
    eapply trf_lift_idtyp_item_Gpush; eauto.
  - eapply trf_lift_idtyp_heap_n.
    eapply IHHtrf; eauto.
    eapply HeapCorrespondence.ty_item_s_Gpush; eauto.
Qed.


Lemma trf_lift_idtyp_heap_binds_met_cases TL e G Sa Sb y ta tb:
  binds y (item_fun ta) Sa ->
  binds y (item_fun tb) Sb ->
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  (
    ta = tb \/ (* Not the transformed item *)
    exists T U L, (* The transformed item *)
      ctx_binds y (typ_fun U T) G /\
      forall x : var,
      x \notin L ->
        〈 G § x /v_par ~ U⁏e ⊢ ta/[-> x]  ∶ T/[-> x]  〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G § x /v_par ~ U⁏e ⊢ tb/[-> x]  ∶ T/[-> x] 〉
  ).
Proof.
  introv Hbia Hbib Htrf.
  remember (〈 e ⊢ Sa ∶ G 〉).
  remember (〈 e ⊢ Sb ∶ G 〉).
  reverts Sa Sb G e.
  induction Htrf; intros; subst; inversions Heqt; inversions Heqt0.
  - destruct (classicT (x=y)).
    + (* Found the transformed item *)
      subst.
      eapply binds_push_eq_inv in Hbib.
      eapply binds_push_eq_inv in Hbia.
      subst.
      inversions H0.
      right.
      exists T S0 L. splits.
      auto.
      intros. eapply H8. auto.
    + eapply binds_push_neq_inv in Hbib.
      eapply binds_push_neq_inv in Hbia.
      binds_eq.
      left. auto.
      auto.
      auto.
  - (* Transformation is later *)
    destruct (classicT (x=y)).
    + (* Found the non-transformed item *)
      subst.
      eapply binds_push_eq_inv in Hbib.
      eapply binds_push_eq_inv in Hbia.
      subst.
      inversions Hbia.
      left. auto.
    + eapply binds_push_neq_inv in Hbib.
      eapply binds_push_neq_inv in Hbia.
      (* Recurse into the smaller heap *)
      eapply IHHtrf; eauto.
      auto.
      auto.
Qed.
Lemma defs_has_one_inv d1 d2:
  (defs_has (defs_cons defs_nil d1) d2) -> d1 = d2.
Proof.
  introv Hh.
  hnf in Hh.
  simpls.
  cases_if. 
  inversions Hh. auto.
Qed.
Lemma trf_lift_idtyp_heap_binds_defs_met_cases TL e G y dsa dsb m ta tb U U1:
  defs_has dsa (def_met m ta) ->
  defs_has dsb (def_met m tb) ->
  〈 G §§ objctx_heap y U1 ⁏ e ⊢ dsa ∶ U /[-> y] 〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈 G §§ objctx_heap y U1 ⁏ e ⊢ dsb ∶ U /[-> y] 〉 ->
  (
    ta = tb \/ (* Not the transformed item *)
    exists TZ TC TR  L , (* The transformed item *)
     (forall xz : var,
     xz \notin L ->
     forall xc : var,
     xc <> xz ->
     xc \notin L ->
     〈 ctx_met_refined_heap (vis_hide typing_mode G) xz TZ /[-> y] xc TC /[# 1 -> y] y U1 ⁏ e ⊢ ta /[# 1 -> xc] /[-> xz] ∶ TR /[# 2 -> y] /[# 1 -> xc] /[-> xz] 〉
     ∼ trf_local_idtyp_trm typing_mode TL
     ∼> 〈 ctx_met_refined_heap (vis_hide typing_mode G) xz TZ /[-> y] xc TC /[# 1 -> y] y U1 ⁏ e ⊢ tb /[# 1 -> xc] /[-> xz] ∶ TR /[# 2 -> y] /[# 1 -> xc] /[-> xz] 〉
    ) /\
    RecordTypes.record_has U /[-> y] (dec_met m TC/[# 1 -> y] TZ/[-> y] TR/[# 2 -> y])
  ). 
Proof.

  introv Hbia Hbib Htrf.
  remember (〈 G §§ objctx_heap y U1 ⁏ e ⊢ dsa ∶ U /[-> y] 〉).
  remember (〈 G §§ objctx_heap y U1 ⁏ e ⊢ dsb ∶ U /[-> y] 〉).
  reverts dsa dsb G e.
  reverts y U U1.
  induction Htrf; intros; subst; inversions Heqt; inversions Heqt0.
  - eapply defs_has_one_inv in Hbia.
    eapply defs_has_one_inv in Hbib.
    subst.
    inversions H.
    destruct U; inversions H1.
    destruct d; inversions H4. sympls. right. exists t0 t t1 L. splits. eapply H2. eapply rh_one.
  - (* transformation later *)
    eapply defs_has_cons_inv in Hbia.
    eapply defs_has_cons_inv in Hbib.
    destruct Hbia; destruct Hbib.
    + subst. inversions H5. auto.
    + subst. forwards: defs_has_hasnt_neq H5; eauto. contradiction.
    + subst. forwards: defs_has_hasnt_neq H3; eauto. contradiction.
    +
      destruct U; inversions H4. sympls.
      forwards HHtrf: IHHtrf; fweauto.
      destructecs HHtrf. auto.
      right. exists TZ TC TR L; splits. eauto.
      eapply rh_andl. auto.
  - (* transformation here *)
    eapply defs_has_cons_inv in Hbia.
    eapply defs_has_cons_inv in Hbib.
    destruct Hbia; destruct Hbib.
    + subst. inversions H.
          destruct U; inversions H4.
    destruct U3; inversions H6.
    destruct d; inversions H3. sympls. right. exists t0 t t1 L; splits. eapply H5. eapply rh_andr. eapply rh_one.
    + subst. inversions H. forwards: defs_has_hasnt_neq H5; eauto. contradiction.
    + subst. inversions H. forwards: defs_has_hasnt_neq H3; eauto. contradiction.
    + forwards: defs_has_met_inv H3 H5. eauto.
Qed.

Lemma trf_lift_idtyp_heap_binds_obj_met_cases TL e G Sa Sb y dsa dsb m ta tb:
  binds y (item_obj dsa) Sa ->
  binds y (item_obj dsb) Sb ->
  defs_has dsa (def_met m ta) ->
  defs_has dsb (def_met m tb) ->
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  (
    ta = tb \/ (* Not the transformed item *)
    exists U1 TZ TC TR S L  , (* The transformed item *)
     (forall xz : var,
     xz \notin L ->
     forall xc : var,
     xc <> xz ->
     xc \notin L ->
     〈 ctx_met_refined_heap (vis_hide typing_mode G) xz TZ /[-> y] xc TC /[# 1 -> y] y U1 ⁏ e ⊢ ta /[# 1 -> xc] /[-> xz] ∶ TR /[# 2 -> y] /[# 1 -> xc] /[-> xz] 〉
     ∼ trf_local_idtyp_trm typing_mode TL
     ∼> 〈 ctx_met_refined_heap (vis_hide typing_mode G) xz TZ /[-> y] xc TC /[# 1 -> y] y U1 ⁏ e ⊢ tb /[# 1 -> xc] /[-> xz] ∶ TR /[# 2 -> y] /[# 1 -> xc] /[-> xz] 〉
    )/\
    ctx_binds y (typ_and (typ_bnd U1) S) G /\
    RecordTypes.record_has U1 /[-> y] (dec_met m TC/[# 1 -> y] TZ/[-> y] TR/[# 2 -> y])
  ).
Proof.
  introv Hbia Hbib Hha Hhb Htrf.
  remember (〈 e ⊢ Sa ∶ G 〉).
  remember (〈 e ⊢ Sb ∶ G 〉).
  reverts Sa Sb G e.
  induction Htrf; intros; subst; inversions Heqt; inversions Heqt0.
  - destruct (classicT (x=y)).
    + (* Found the transformed item *)
      subst.
      eapply binds_push_eq_inv in Hbib.
      eapply binds_push_eq_inv in Hbia.
      subst.
      inversions H0.
      forwards: trf_lift_idtyp_heap_binds_defs_met_cases H7; fweauto.
      destruct H0. auto.
      right. destructecs H0. exists U TZ TC TR S0. eauto.
    + eapply binds_push_neq_inv in Hbib.
      eapply binds_push_neq_inv in Hbia.
      binds_eq.
      forwards: defs_has_met_inv Hha Hhb. eauto.
  
      auto.
      auto.
  - (* Transformation is later *)
    destruct (classicT (x=y)).
    + (* Found the non-transformed item *)
      subst.
      eapply binds_push_eq_inv in Hbib.
      eapply binds_push_eq_inv in Hbia.
      subst.
      inversions Hbia.
      forwards: defs_has_met_inv Hha Hhb. eauto.
    + eapply binds_push_neq_inv in Hbib.
      eapply binds_push_neq_inv in Hbia.
      (* Recurse into the smaller heap *)
      eapply IHHtrf; eauto.
      auto.
      auto.
Qed.


Lemma trf_red_trm_let1 (TL: TermTransform) G (e: renv) t1a t1b t2 fs S Lva Lvb (T: typ) Lva2 Lvb2 Tl:
  〈 G ⁏ e ⊢ t1a ∶ Tl 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G ⁏ e ⊢ t1b ∶ Tl 〉 ->
  G ⊢ trm_let t1a t2 ; fs ; S ; e ; Lva ∶ T ↦ G ⊢ t1a ; stack_cons (frame_let t2) fs ; S ; e ; Lva2 ->
  G ⊢ trm_let t1b t2 ; fs ; S ; e ; Lvb ∶ T ↦ G ⊢ t1b ; stack_cons (frame_let t2) fs ; S ; e ; Lvb2 ->
  ty_config_accessories G e (stack_cons (frame_let t2) fs) S t1a T Lva2 ->
  ty_config_accessories G e (stack_cons (frame_let t2) fs) S t1b T Lvb2 ->
  HeapCorrespondence.heap_correspond G e S ->
  ty_stack G e Tl (stack_cons (frame_let t2) fs) T ->
  〈 G ⊢ t1a ; stack_cons (frame_let t2) fs ; S ; e ; Lva2 ∶ T 〉 ∼ trf_lift_idtyp_conf typing_mode TL ∼> 〈 G ⊢ t1b ; stack_cons (frame_let t2) fs ; S ; e ; Lvb2 ∶ T 〉 .
Proof.
  introv Htrf Htya Htyb Htyca Htycb Hhc Htys.
  eapply trf_lift_idtyp_conf_focus.
  eapply trf_focus_idtyp_conf_focus; eauto.
Qed.



Lemma trf_local_deref_def TL G Go e da db T:
   〈 G §§ Go ⁏ e ⊢ da ∶ T 〉 ∼ trf_local_idtyp_def typing_mode TL ∼> 〈 G §§ Go ⁏ e ⊢ db ∶ T 〉 ->
   〈 G §§ Go ⁏ e ⊢ (deref_def e da) ∶ T 〉 ∼ trf_local_idtyp_def typing_mode TL ∼> 〈 G §§ Go ⁏ e ⊢ (deref_def e db) ∶ T 〉.
Proof.
  introv Hty.
  inversions Hty.
  unfold deref_def.
  econstructor; eauto.
Qed.


Lemma trf_local_deref_defs TL G Go e da db T:
  inert (G §§& Go) ->
  ok e ->
  renv_corr (G §§& Go) e ->
   〈 G §§ Go ⁏ e ⊢ da ∶ T 〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈 G §§ Go ⁏ e ⊢ db ∶ T 〉 ->
   〈 G §§ Go ⁏ e ⊢ (deref_defs e da) ∶ T 〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈 G §§ Go ⁏ e ⊢ (deref_defs e db) ∶ T 〉.
Proof.
  introv Hin Hoke1 Hcor1 Hty.
  remember (〈 G §§ Go ⁏ e ⊢ da ∶ T 〉).
  remember (〈 G §§ Go ⁏ e ⊢ db ∶ T 〉).
  reverts da db T.
  induction Hty; intros; subst; inversions Heqt;  inversions Heqt0.

  - (* one *) econstructor.
    eapply trf_local_deref_def; eauto.
  - (* cons *)
    rewrite ! deref_defs_cons_comm.
    eapply trf_local_idtyp_defs_consl.
    apply IHHty; auto.
    apply ty_deref_def; auto.
    +
    apply defs_hasnt_deref.
    rewrite label_of_deref.
    auto.
    +
    apply defs_hasnt_deref.
    rewrite label_of_deref.
    auto.
  -
    rewrite ! deref_defs_cons_comm.
    eapply trf_local_idtyp_defs_consr.
    apply trf_local_deref_def; auto.
    apply ty_deref_defs; auto.
    +
    apply defs_hasnt_deref.
    rewrite label_of_deref.
    auto.
    +
    apply defs_hasnt_deref.
    rewrite label_of_deref.
    auto.
Qed.

End WithTypingMode.