(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation TransformationLifting Weakening EWeakening Substitution.
Require Import TypingHelpers TransformationConfig ConfigTyping.
Require Import Similarity ConfigSimilarity ContextSimilarity SyntaxSimilarity SimilarityPreservation TypingSimilarity
 TypedReductionEquivalence OperationalSemantics TypedOperationalSemantics TypedPreservation ConfigValid.
Require Import TransformationPreservation TransformationTerm TransformationConfig TransformationSimilarity.

Section WithTypingMode.
Context (typing_mode: Mode).

(* Iterate transformation in union with similarity. *)
Definition trf_lift_sim_conf (TL: TermTransform): TransformConfig :=
  (trf_iterate (TransformCondConfTyped typing_mode) (trf_union (trf_similarity typing_mode) (trf_lift_idtyp_conf typing_mode TL))).

Global Instance TransformIterateSimPreservesIsAnswer
  (trf: TermTransform) (TI: TransformPreserves TransformIsAnswerTerm  trf):
  TransformPreserves TransformIsAnswer (trf_lift_sim_conf trf).
Proof.
  unfold trf_lift_sim_conf.
  eapply TransformPreservesIterate.
  eapply TransformPreservesUnion.
  eapply TransformSimilarityPreservesIsAnswer.
  eapply TransformPreservesIsAnswerLiftIdtyp.
  auto.
Qed.

Global Instance TransformIterateSimTransitive (trf: TermTransform): TransitiveTransform (trf_lift_sim_conf trf).
Proof.
  eapply TransformIterateTransitive.
Qed.

Global Instance TransformIterateSimTyped trf (TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires (TransformCondConfTyped typing_mode) (trf_lift_sim_conf trf).
Proof.
  eapply TransformRequiresIterate.
Qed.

Global Instance TransformIterateSimSymmetric (trf: TermTransform)
    (TS: SymmetricTransform trf)
    (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf)
    (TT: TransformRequires (TransformCondTermTyped typing_mode) trf):
  SymmetricTransform (trf_lift_sim_conf trf).
Proof.
  eapply TransformIterateSymmetric.
  eapply SymmetricTransformUnion.
  eapply TransformSimilaritySymmetric.
  eapply SymmetricTransformLiftIdtypConfig.
  auto.
  auto.
  eapply TransformRequiresUnion.
  eapply TransformSimilarityTyped.
  eapply TransformLiftIdtypConfigTyped. auto.
  auto.
Qed.


End WithTypingMode.

Section WithTypingMode.
Context (typing_mode: Mode).

(* Constructing trf_lift_sim_conf with a fixed number of steps *)

Lemma trf_lift_sim_conf_lift (TL: TermTransform) Ca Cb Ga Gb Ta Tb:
  ty_config Ga Ca Ta ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_lift_idtyp_conf typing_mode TL ∼> 〈 Gb ⊢ Cb ∶ Tb 〉 ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 Gb ⊢ Cb ∶ Tb 〉.
Proof.
  introv Hty Htrf.
  eapply trf_iterate_1.
  hnf. auto.
  hnf. auto.
Qed.

Lemma trf_lift_sim_conf_sim (TL: TermTransform) Ca Cb Ga Gb Ta Tb:
  ty_config Ga Ca Ta ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_similarity typing_mode ∼> 〈 Gb ⊢ Cb ∶ Tb 〉 ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 Gb ⊢ Cb ∶ Tb 〉.
Proof.
  introv Hty Htrf.
  eapply trf_iterate_1.
  hnf. auto.
  hnf. auto.
Qed.


Lemma trf_lift_sim_conf_lift_sim (TL: TermTransform) Ca Cb Cc Ga Gb Gc Ta Tb Tc:
  ty_config Ga Ca Ta ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_lift_idtyp_conf typing_mode TL ∼> 〈 Gb ⊢ Cb ∶ Tb 〉 ->
  〈 Gb ⊢ Cb ∶ Tb 〉 ∼ trf_similarity typing_mode ∼> 〈 Gc ⊢ Cc ∶ Tc 〉 ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 Gc ⊢ Cc ∶ Tc 〉.
Proof.
  introv Hty Htrf Hsi.
  eapply trf_iterate_S.
  eapply trf_iterate_1.
  hnf. auto.
  right. eauto.
  left. eauto.
Qed.
Lemma trf_lift_sim_conf_sim_lift (TL: TermTransform) Ca Cb Cc Ga Gb Gc Ta Tb Tc:
  ty_config Ga Ca Ta ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_similarity typing_mode ∼> 〈 Gb ⊢ Cb ∶ Tb 〉  ->
  〈 Gb ⊢ Cb ∶ Tb 〉 ∼ trf_lift_idtyp_conf typing_mode TL ∼> 〈 Gc ⊢ Cc ∶ Tc 〉 ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 Gc ⊢ Cc ∶ Tc 〉.
Proof.
  introv Hty Htrf Hsi.
  eapply trf_iterate_S.
  eapply trf_iterate_1.
  hnf. auto.
  left. eauto.
  right. eauto.
Qed.

Lemma trf_lift_sim_conf_sim_lift_sim (TL: TermTransform) Ca Cb Cc Cd Ga Gb Gc Gd Ta Tb Tc Td:
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_similarity typing_mode ∼> 〈 Gb ⊢ Cb ∶ Tb 〉  ->
  〈 Gb ⊢ Cb ∶ Tb 〉 ∼ trf_lift_idtyp_conf typing_mode TL ∼> 〈 Gc ⊢ Cc ∶ Tc 〉 ->
  〈 Gc ⊢ Cc ∶ Tc 〉 ∼ trf_similarity typing_mode ∼> 〈 Gd ⊢ Cd ∶ Td 〉  ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 Gd ⊢ Cd ∶ Td 〉.
Proof.
  introv Hsi1 Htrf Hsi2.
  eapply trf_iterate_S.
  eapply trf_iterate_S.
  eapply trf_iterate_1.
  eapply TransformSimilarityTyped; eauto.
  left. eauto.
  right. eauto.
  left. eauto.
Qed.

Lemma trf_lift_sim_conf_sim_lift_lift_sim (TL: TermTransform) Ca Cb Cc Cd Ga Gb Gc Gd Ta Tb Tc Td Ge Ce Te:
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_similarity typing_mode ∼> 〈 Gb ⊢ Cb ∶ Tb 〉  ->
  〈 Gb ⊢ Cb ∶ Tb 〉 ∼ trf_lift_idtyp_conf typing_mode TL ∼> 〈 Ge ⊢ Ce ∶ Te 〉 ->
  〈 Ge ⊢ Ce ∶ Te 〉 ∼ trf_lift_idtyp_conf typing_mode TL ∼> 〈 Gc ⊢ Cc ∶ Tc 〉 ->
  〈 Gc ⊢ Cc ∶ Tc 〉 ∼ trf_similarity typing_mode ∼> 〈 Gd ⊢ Cd ∶ Td 〉  ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 Gd ⊢ Cd ∶ Td 〉.
Proof.
  introv Hsi1 Htrf1 Htrf2 Hsi2.
  eapply trf_iterate_S.
  eapply trf_iterate_S.
  eapply trf_iterate_S.
  eapply trf_iterate_1.
  eapply TransformSimilarityTyped; eauto.
  left. eauto.
  right. eauto.
  right. eauto.
  left. eauto.
Qed.
Lemma trf_lift_sim_conf_lift_lift (TL: TermTransform) Ca Cb Cc Ga Gb Gc Ta Tb Tc:
  ty_config Ga Ca Ta ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_lift_idtyp_conf typing_mode TL ∼>  〈 Gc ⊢ Cc ∶ Tc 〉  ->
  〈 Gc ⊢ Cc ∶ Tc 〉  ∼ trf_lift_idtyp_conf typing_mode TL ∼> 〈 Gb ⊢ Cb ∶ Tb 〉 ->
  〈 Ga ⊢ Ca ∶ Ta 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 Gb ⊢ Cb ∶ Tb 〉.
Proof.
  introv Hty Htrf1 Htrf2.
  eapply trf_iterate_S.
  eapply trf_iterate_1.
  hnf. auto.
  right. eauto.
  right. eauto.
Qed.

End WithTypingMode.