(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation Weakening EWeakening Substitution TransformationTerm.
Require Import Narrowing Substitution Subenvironments  FreshChain FreshChainTactics.
Require Import Sealing.
Require Import Visibility.

Section WithTypingMode.
Context (typing_mode: Mode).

(* Lifts a local transformation to a term transformation *)
(* This definition sort-of re-types the whole program (it does not use the original typing),
  so it does not exactly correspond to the idea of a program transformation working at an already typed tree and looking locally at just one place.  *)
Inductive trf_local_trm (Cond: LiftCond) (TL: TermTransform): TermTransform :=
(* Apply the local trasformation at this point *)
| trf_local_trm_local Ga Gb ea eb Ta Tb ta tb Sa Sb:
    〈Ga⁏ea ⊢ ta ∶ Ta〉 ∼TL∼> 〈Gb⁏eb ⊢ tb ∶ Tb〉 ->
    Ga ⁏ ea ⊢ Ta <: Sa ->
    Gb ⁏ eb ⊢ Tb <: Sb ->
    〈Ga⁏ea ⊢ ta ∶ Sa〉 ∼(Cond TrfContextTy typ trm)∼∼(trf_local_trm Cond TL)∼> 〈Gb⁏eb ⊢ tb ∶ Sb〉
(* Let argument *)
| trf_local_trm_letl Ga Gb ea eb tla tlb Ta Tb Tla Tlb tr L Sa Sb:
    〈Ga⁏ea ⊢ tla ∶ Tla〉 ∼(trf_local_trm Cond TL)∼> 〈Gb⁏eb ⊢ tlb ∶ Tlb〉 ->
    (forall x, x \notin L -> (Ga § x / v_let ~ Tla) ⁏ ea ⊢ (open x tr) ∶ Ta) ->
    (forall x, x \notin L -> (Gb § x / v_let ~ Tlb) ⁏ eb ⊢ (open x tr) ∶ Tb) ->
    Ga ⁏ ea ⊢ Ta <: Sa ->
    Gb ⁏ eb ⊢ Tb <: Sb ->
    〈Ga⁏ea ⊢ (trm_let tla tr) ∶ Sa〉 ∼(Cond TrfContextTy typ trm)∼∼(trf_local_trm Cond TL)∼> 〈Gb⁏eb ⊢ (trm_let tlb tr) ∶ Sb〉
(* Let continuation *)
| trf_local_trm_letr Ga Gb ea eb tl Ta Tb Tla Tlb tra trb L Sa Sb:
    (forall x, x \notin L ->
      〈Ga § x / v_let ~ Tla⁏ea ⊢ (open x tra) ∶ Ta〉 ∼(trf_local_trm Cond TL)∼> 〈Gb § x / v_let ~ Tlb⁏eb ⊢ (open x trb) ∶ Tb〉
    ) ->
    Ga ⁏ ea ⊢ tl ∶ Tla ->
    Gb ⁏ eb ⊢ tl ∶ Tlb ->
    Ga ⁏ ea ⊢ Ta <: Sa ->
    Gb ⁏ eb ⊢ Tb <: Sb ->
    〈Ga⁏ea ⊢ (trm_let tl tra) ∶ Sa〉 ∼(Cond TrfContextTy typ trm)∼∼(trf_local_trm Cond TL)∼> 〈Gb⁏eb ⊢ (trm_let tl trb) ∶ Sb〉
(* Literal argument *)
| trf_local_trm_litl Ga Gb ea eb tla tlb Ta Tb Tla Tlb tr L Sa Sb:
    〈Ga⁏ea ⊢ tla ∶ Tla〉 ∼(trf_local_lit Cond TL)∼> 〈Gb⁏eb ⊢ tlb ∶ Tlb〉 ->
    (forall x, x \notin L ->
      (Ga § x / v_let ~ Tla) ⁏ ea ⊢ (open x tr) ∶ Ta
    ) ->
    (forall x, x \notin L ->
      (Gb § x / v_let ~ Tlb) ⁏ eb ⊢ (open x tr) ∶ Tb
    ) ->
    Ga ⁏ ea ⊢ Ta <: Sa ->
    Gb ⁏ eb ⊢ Tb <: Sb ->
    〈Ga⁏ea ⊢ (trm_lit tla tr) ∶ Sa〉 ∼(Cond TrfContextTy typ trm)∼∼(trf_local_trm Cond TL)∼> 〈Gb⁏eb ⊢ (trm_lit tlb tr) ∶ Sb〉
(* Literal continuation *)
| trf_local_trm_litr Ga Gb ea eb tl Ta Tb Tla Tlb tra trb L Sa Sb:
    (forall x, x \notin L ->
      〈Ga § x / v_let ~ Tla⁏ea ⊢ (open x tra) ∶ Ta〉 ∼(trf_local_trm Cond TL)∼> 〈Gb § x / v_let ~ Tlb⁏eb ⊢ (open x trb) ∶ Tb〉
    ) ->
    Ga ⁏ ea ⊢ tl ∶ Tla ->
    Gb ⁏ eb ⊢ tl ∶ Tlb ->
    Ga ⁏ ea ⊢ Ta <: Sa ->
    Gb ⁏ eb ⊢ Tb <: Sb ->
    〈Ga⁏ea ⊢ (trm_lit tl tra) ∶ Sa〉 ∼(Cond TrfContextTy typ trm)∼∼(trf_local_trm Cond TL)∼> 〈Gb⁏eb ⊢ (trm_lit tl trb) ∶ Sb〉
with trf_local_lit (Cond: LiftCond) (TL: TermTransform): LitTransform :=
(* Function literal *)
| trf_local_lit_fun Ga Gb ea eb Ta Tb Ua Ub ta tb L:
    mode_has_fun typing_mode ->
    (forall x, x \notin L ->
      〈Ga § x / v_par ~ Ta⁏ea ⊢ (open x ta) ∶ (open x Ua)〉 ∼(trf_local_trm Cond TL)∼> 〈Gb § x / v_par ~ Tb⁏eb ⊢ (open x tb) ∶ (open x Ub)〉
    ) ->
    〈Ga⁏ea ⊢ (lit_fun Ta ta) ∶ (typ_fun Ta Ua)〉 ∼(Cond TrfContextTy typ lit)∼∼(trf_local_lit Cond TL)∼> 〈Gb⁏eb ⊢ (lit_fun Tb tb) ∶ (typ_fun Tb Ub)〉
(* Object literal *)
| trf_local_lit_obj Ga Gb ea eb Ta Tb dsa dsb L:
    (forall x, x \notin L ->
      〈Ga§§(objctx_lit x Ta)⁏ea ⊢ (open x dsa) ∶ (open x Ta)〉 ∼(trf_local_defs Cond TL)∼> 〈Gb§§(objctx_lit x Tb)⁏eb ⊢ (open x dsb) ∶ (open x Tb)〉
    ) ->
    typ_indep_b 0 Ta ->
    typ_indep_b 0 Tb ->
    〈Ga⁏ea ⊢ (lit_obj Ta dsa) ∶ (typ_objlit typing_mode Ta)〉 ∼(Cond TrfContextTy typ lit)∼∼(trf_local_lit Cond TL)∼> 〈Gb⁏eb ⊢ (lit_obj Tb dsb) ∶ (typ_objlit typing_mode Tb)〉
with trf_local_def (Cond: LiftCond) (TL: TermTransform): DefTransform :=
(* Method definition *)
| trf_local_def_met Ga Gb Goa Gob ea eb Ta Tb TCa TCb TZa TZb m ta tb L:
    ( forall xz, xz \notin L ->
      forall xc, xc <> xz -> xc \notin L ->
      〈(ctx_obj_met typing_mode Ga Goa xz TZa xc TCa)⁏ea ⊢ (open_met xz xc ta) ∶ (open_met xz xc Ta)〉 ∼(trf_local_trm Cond TL)∼> 〈(ctx_obj_met typing_mode Gb Gob xz TZb xc TCb)⁏eb ⊢ (open_met xz xc tb) ∶ (open_met xz xc Tb)〉
    ) ->
    〈Ga §§ Goa⁏ea ⊢ (def_met m ta) ∶ (dec_met m TCa TZa Ta)〉 ∼(Cond TrfContextObjTy dec def)∼∼(trf_local_def Cond TL)∼> 〈Gb §§ Gob⁏eb ⊢ (def_met m tb) ∶ (dec_met m TCb TZb Tb)〉
with trf_local_defs (Cond: LiftCond) (TL: TermTransform): DefsTransform :=
(* Last definition *)
| trf_local_defs_one Ga Gb Goa Gob ea eb Da Db dsla dslb:
    〈Ga §§ Goa⁏ea ⊢ dsla ∶ Da〉 ∼(trf_local_def Cond TL)∼> 〈Gb §§ Gob⁏eb ⊢ dslb ∶ Db〉 ->
    〈Ga §§ Goa⁏ea ⊢ (defs_cons defs_nil dsla) ∶ typ_rcd Da〉 ∼(Cond TrfContextObjTy typ defs)∼∼(trf_local_defs Cond TL)∼> 〈Gb §§ Gob⁏eb ⊢ (defs_cons defs_nil dslb) ∶ typ_rcd Db〉
(* Definition  *)
| trf_local_defs_consl Ga Gb Goa Gob ea eb Ta Tb Da Db dsla dslb dsr:
    〈Ga §§ Goa⁏ea ⊢ dsla ∶ Ta〉 ∼(trf_local_defs Cond TL)∼> 〈Gb §§ Gob⁏eb ⊢ dslb ∶ Tb〉 ->
    Ga §§ Goa⁏ea ⊢ dsr ∶d Da ->
    Gb §§ Gob⁏eb ⊢ dsr ∶d Db ->
    defs_hasnt dsla (label_of dsr) ->
    defs_hasnt dslb (label_of dsr) ->
    〈Ga §§ Goa⁏ea ⊢ (defs_cons dsla dsr) ∶ (typ_and Ta (typ_rcd Da))〉 ∼(Cond TrfContextObjTy typ defs)∼∼(trf_local_defs Cond TL)∼> 〈Gb §§ Gob⁏eb ⊢ (defs_cons dslb dsr) ∶ (typ_and Tb (typ_rcd Db))〉
(* Definition  *)
| trf_local_defs_consr Ga Gb Goa Gob ea eb Ta Tb Da Db dsl dsra dsrb:
    〈Ga§§Goa⁏ea ⊢ dsra ∶ Da〉 ∼(trf_local_def Cond TL)∼> 〈Gb§§Gob⁏eb ⊢ dsrb ∶ Db〉 ->
    Ga §§ Goa⁏ea ⊢ dsl ∶ Ta ->
    Gb §§ Gob⁏eb ⊢ dsl ∶ Tb ->
    defs_hasnt dsl (label_of dsra) ->
    defs_hasnt dsl (label_of dsrb) ->
    〈Ga §§ Goa⁏ea ⊢ (defs_cons dsl dsra) ∶ (typ_and Ta (typ_rcd Da))〉 ∼(Cond TrfContextObjTy typ defs)∼∼(trf_local_defs Cond TL)∼> 〈Gb §§ Gob⁏eb ⊢ (defs_cons dsl dsrb) ∶ (typ_and Tb (typ_rcd Db))〉
.

Scheme trf_local_trm_mut    := Induction for trf_local_trm Sort Prop
with   trf_local_lit_mut    := Induction for trf_local_lit Sort Prop
with   trf_local_def_mut    := Induction for trf_local_def Sort Prop
with   trf_local_defs_mut   := Induction for trf_local_defs Sort Prop.
Combined Scheme trf_local_mutind from trf_local_trm_mut, trf_local_lit_mut,
trf_local_def_mut, trf_local_defs_mut.

(* Part of Figure 23 in "Pure methods for roDOT" *)
(* Version of trf_local_trm, where contexts and types on both sides are the same *)
Inductive trf_local_idtyp_trm (TL: TermTransform): TermTransform :=
(** (TRFL-Local) Apply the local trasformation at this point *)
| trf_local_idtyp_trm_local G e T ta tb S:
    〈G⁏e ⊢ ta ∶ T〉 ∼TL∼> 〈G⁏e ⊢ tb ∶ T〉 ->
    G ⁏ e ⊢ T <: S ->
    〈G⁏e ⊢ ta ∶ S〉 ∼(trf_local_idtyp_trm TL)∼> 〈G⁏e ⊢ tb ∶ S〉
(** (TRFL-Let1) Let argument *)
| trf_local_idtyp_trm_letl G e tla tlb T Tl tr L S:
    〈G⁏e ⊢ tla ∶ Tl〉 ∼(trf_local_idtyp_trm TL)∼> 〈G⁏e ⊢ tlb ∶ Tl〉 ->
    (forall x, x \notin L -> (G § x / v_let ~ Tl) ⁏ e ⊢ (open x tr) ∶ T) ->
    G ⁏ e ⊢ T <: S ->
    〈G⁏e ⊢ (trm_let tla tr) ∶ S〉 ∼(trf_local_idtyp_trm TL)∼> 〈G⁏e ⊢ (trm_let tlb tr) ∶ S〉
(* (TRFL-Let2) Let continuation *)
| trf_local_idtyp_trm_letr G e tl T Tl tra trb L S:
    (forall x, x \notin L ->
      〈G § x / v_let ~ Tl⁏e ⊢ (open x tra) ∶ T〉 ∼(trf_local_idtyp_trm TL)∼> 〈G § x / v_let ~ Tl⁏e ⊢ (open x trb) ∶ T〉
    ) ->
    G ⁏ e ⊢ tl ∶ Tl ->
    G ⁏ e ⊢ T <: S ->
    〈G⁏e ⊢ (trm_let tl tra) ∶ S〉 ∼(trf_local_idtyp_trm TL)∼> 〈G⁏e ⊢ (trm_let tl trb) ∶ S〉
(* (TRFL-Lit1) Literal argument *)
| trf_local_idtyp_trm_litl G e tla tlb T Tl tr L S:
    〈G⁏e ⊢ tla ∶ Tl〉 ∼(trf_local_idtyp_lit TL)∼> 〈G⁏e ⊢ tlb ∶ Tl〉 ->
    (forall x, x \notin L ->
      (G § x / v_let ~ Tl) ⁏ e ⊢ (open x tr) ∶ T
    ) ->
    G ⁏ e ⊢ T <: S ->
    〈G⁏e ⊢ (trm_lit tla tr) ∶ S〉 ∼(trf_local_idtyp_trm TL)∼> 〈G⁏e ⊢ (trm_lit tlb tr) ∶ S〉
(* (TRFL-Lit2) Literal continuation *)
| trf_local_idtyp_trm_litr G e tl T Tl tra trb L S:
    (forall x, x \notin L ->
      〈G § x / v_let ~ Tl⁏e ⊢ (open x tra) ∶ T〉 ∼(trf_local_idtyp_trm TL)∼> 〈G § x / v_let ~ Tl⁏e ⊢ (open x trb) ∶ T〉
    ) ->
    G ⁏ e ⊢ tl ∶ Tl ->
    G ⁏ e ⊢ T <: S ->
    〈G⁏e ⊢ (trm_lit tl tra) ∶ S〉 ∼(trf_local_idtyp_trm TL)∼> 〈G⁏e ⊢ (trm_lit tl trb) ∶ S〉
(* Part of Figure 23 in "Pure methods for roDOT" *)
with trf_local_idtyp_lit (TL: TermTransform): LitTransform :=
(* Function literal *)
| trf_local_idtyp_lit_fun G e T U ta tb L:
    mode_has_fun typing_mode ->
    (forall x, x \notin L ->
      〈G § x / v_par ~ T⁏e ⊢ (open x ta) ∶ (open x U)〉 ∼(trf_local_idtyp_trm TL)∼> 〈G § x / v_par ~ T⁏e ⊢ (open x tb) ∶ (open x U)〉
    ) ->
    〈G⁏e ⊢ (lit_fun T ta) ∶ (typ_fun T U)〉 ∼(trf_local_idtyp_lit TL)∼> 〈G⁏e ⊢ (lit_fun T tb) ∶ (typ_fun T U)〉
(*  (TRFL-Lit1) Object literal *)
| trf_local_idtyp_lit_obj G e T dsa dsb L:
    (forall x, x \notin L ->
      〈G§§(objctx_lit x T)⁏e ⊢ (open x dsa) ∶ (open x T)〉 ∼(trf_local_idtyp_defs TL)∼> 〈G§§(objctx_lit x T)⁏e ⊢ (open x dsb) ∶ (open x T)〉
    ) ->
    typ_indep_b 0 T ->
    〈G⁏e ⊢ (lit_obj T dsa) ∶ (typ_objlit typing_mode T)〉 ∼(trf_local_idtyp_lit TL)∼> 〈G⁏e ⊢ (lit_obj T dsb) ∶ (typ_objlit typing_mode T)〉
(* Part of Figure 23 in "Pure methods for roDOT" *)
with trf_local_idtyp_def (TL: TermTransform): DefTransform :=
(* (TRFL-Met) Method definition *)
| trf_local_idtyp_def_met G Go e T TC TZ m ta tb L:
    ( forall xz, xz \notin L ->
      forall xc, xc <> xz -> xc \notin L ->
      〈(ctx_obj_met typing_mode G Go xz TZ xc TC)⁏e ⊢ (open_met xz xc ta) ∶ (open_met xz xc T)〉 ∼(trf_local_idtyp_trm TL)∼> 〈(ctx_obj_met typing_mode G Go xz TZ xc TC)⁏e ⊢ (open_met xz xc tb) ∶ (open_met xz xc T)〉
    ) ->
    〈G §§ Go⁏e ⊢ (def_met m ta) ∶ (dec_met m TC TZ T)〉 ∼(trf_local_idtyp_def TL)∼> 〈G §§ Go⁏e ⊢ (def_met m tb) ∶ (dec_met m TC TZ T)〉
(* Part of Figure 23 in "Pure methods for roDOT" *)
with trf_local_idtyp_defs (TL: TermTransform): DefsTransform :=
(* (TRFL-Met) Last definition *)
| trf_local_idtyp_defs_one G Go e D dsla dslb:
    〈G §§ Go⁏e ⊢ dsla ∶ D〉 ∼(trf_local_idtyp_def TL)∼> 〈G §§ Go⁏e ⊢ dslb ∶ D〉 ->
    〈G §§ Go⁏e ⊢ (defs_cons defs_nil dsla) ∶ typ_rcd D〉 ∼(trf_local_idtyp_defs TL)∼> 〈G §§ Go⁏e ⊢ (defs_cons defs_nil dslb) ∶ typ_rcd D〉
(* (TRFL-And1) Definition *)
| trf_local_idtyp_defs_consl G Go e T D dsla dslb dsr:
    〈G §§ Go⁏e ⊢ dsla ∶ T〉 ∼(trf_local_idtyp_defs TL)∼> 〈G §§ Go⁏e ⊢ dslb ∶ T〉 ->
    G §§ Go⁏e ⊢ dsr ∶d D ->
    defs_hasnt dsla (label_of dsr) ->
    defs_hasnt dslb (label_of dsr) ->
    〈G §§ Go⁏e ⊢ (defs_cons dsla dsr) ∶ (typ_and T (typ_rcd D))〉 ∼(trf_local_idtyp_defs TL)∼> 〈G §§ Go⁏e ⊢ (defs_cons dslb dsr) ∶ (typ_and T (typ_rcd D))〉
(* (TRFL-And2) Definition *)
| trf_local_idtyp_defs_consr G Go e T D dsl dsra dsrb:
    〈G§§Go⁏e ⊢ dsra ∶ D〉 ∼(trf_local_idtyp_def TL)∼> 〈G§§Go⁏e ⊢ dsrb ∶ D〉 ->
    G §§ Go⁏e ⊢ dsl ∶ T ->
    defs_hasnt dsl (label_of dsra) ->
    defs_hasnt dsl (label_of dsrb) ->
    〈G §§ Go⁏e ⊢ (defs_cons dsl dsra) ∶ (typ_and T (typ_rcd D))〉 ∼(trf_local_idtyp_defs TL)∼> 〈G §§ Go⁏e ⊢ (defs_cons dsl dsrb) ∶ (typ_and T (typ_rcd D))〉
.

Scheme trf_local_idtyp_trm_mut    := Induction for trf_local_idtyp_trm Sort Prop
with   trf_local_idtyp_lit_mut    := Induction for trf_local_idtyp_lit Sort Prop
with   trf_local_idtyp_def_mut    := Induction for trf_local_idtyp_def Sort Prop
with   trf_local_idtyp_defs_mut   := Induction for trf_local_idtyp_defs Sort Prop.
Combined Scheme trf_local_idtyp_mutind from trf_local_idtyp_trm_mut, trf_local_idtyp_lit_mut,
trf_local_idtyp_def_mut, trf_local_idtyp_defs_mut.

Lemma trf_local_idtyp_type_identical_ind (trf: TermTransform):
  ImplyTransformInd (trf_local_idtyp_trm trf) (@TransformCondTypeIdentical TrfContextTy typ trm) /\
  ImplyTransformInd (trf_local_idtyp_lit trf) (@TransformCondTypeIdentical TrfContextTy typ lit) /\
  ImplyTransformInd (trf_local_idtyp_def trf) (@TransformCondTypeIdentical TrfContextObjTy dec def) /\
  ImplyTransformInd (trf_local_idtyp_defs trf) (@TransformCondTypeIdentical TrfContextObjTy typ defs).
Proof.
  unfold ImplyTransformInd.
  eapply trf_local_idtyp_mutind; intros; unfold TransformCondTypeIdentical; eauto.
Qed.
Instance TrfLocalIdtypTrmEnsuresCond trf: TransformEnsures (@TransformCondTypeIdentical TrfContextTy typ trm) (trf_local_idtyp_trm trf).
Proof.
  hnf; intros.
  eapply (proj41 (trf_local_idtyp_type_identical_ind  trf)); eauto.
Qed.

Lemma trf_local_cond_ind Cond (trf: TermTransform):
  ImplyTransformInd (trf_local_trm Cond trf) (Cond TrfContextTy typ trm) /\
  ImplyTransformInd (trf_local_lit Cond trf) (Cond TrfContextTy typ lit) /\
  ImplyTransformInd (trf_local_def Cond trf) (Cond TrfContextObjTy dec def) /\
  ImplyTransformInd (trf_local_defs Cond trf) (Cond TrfContextObjTy typ defs).
Proof.
  unfold ImplyTransformInd.
  eapply trf_local_mutind; intros; eauto.
Qed.

Instance TrfLocalTrmEnsuresCond (Cond: LiftCond) trf: TransformEnsures (Cond TrfContextTy typ trm) (trf_local_trm Cond trf).
Proof.
  hnf; intros.
  eapply (proj41 (trf_local_cond_ind Cond trf)); eauto.
Qed.
Instance TrfLocalLitEnsuresCond (Cond: LiftCond) trf: TransformEnsures (Cond TrfContextTy typ lit) (trf_local_lit Cond trf).
Proof.
  hnf; intros.
  eapply (proj42 (trf_local_cond_ind Cond trf)); eauto.
Qed.

Lemma trf_local_idtyp_local_ind (trf: TermTransform):
  ImplyTransformInd (trf_local_idtyp_trm trf) (trf_local_trm TransformCondTypeIdentical trf) /\
  ImplyTransformInd (trf_local_idtyp_lit trf) (trf_local_lit TransformCondTypeIdentical trf) /\
  ImplyTransformInd (trf_local_idtyp_def trf) (trf_local_def TransformCondTypeIdentical trf) /\
  ImplyTransformInd (trf_local_idtyp_defs trf) (trf_local_defs TransformCondTypeIdentical trf).
Proof.
  unfold ImplyTransformInd.
  eapply trf_local_idtyp_mutind; intros.
  - econstructor; eauto. hnf; auto.
  - eapply trf_local_trm_letl; hnf; eauto.
  - eapply trf_local_trm_letr; hnf; eauto.
  - eapply trf_local_trm_litl; hnf; eauto.
  - eapply trf_local_trm_litr; hnf; eauto.
  - eapply trf_local_lit_fun; hnf; eauto.
  - eapply trf_local_lit_obj; hnf; eauto.
  - eapply trf_local_def_met; hnf; eauto.
  - eapply trf_local_defs_one; hnf; eauto.
  - eapply trf_local_defs_consl; hnf; eauto.
  - eapply trf_local_defs_consr; hnf; eauto.
Qed.

Lemma trf_local_local_idtyp_ind (trf: TermTransform)
  (TTI: TransformEnsures (@TransformCondTypeIdentical TrfContextTy typ trm) trf):
  ImplyTransformInd (trf_local_trm TransformCondTypeIdentical trf) (trf_local_idtyp_trm trf) /\
  ImplyTransformInd (trf_local_lit TransformCondTypeIdentical trf) (trf_local_idtyp_lit trf) /\
  ImplyTransformInd (trf_local_def TransformCondTypeIdentical trf) (trf_local_idtyp_def trf) /\
  ImplyTransformInd (trf_local_defs TransformCondTypeIdentical trf) (trf_local_idtyp_defs trf).
Proof.
  unfold ImplyTransformInd.
  eapply trf_local_mutind; intros; hnf in c; destructecs c; subst.
  - inversions c0. forwards: TTI t. inversions H. econstructor; eauto.
  -
    inversions c0; subst. 
    eapply TrfLocalTrmEnsuresCond in t. hnf in t. inversions t.
    eapply trf_local_idtyp_trm_letl; hnf; eauto.
  -
    inversions c0; subst.
    eapply trf_local_idtyp_trm_letr with (L:=L); hnf; eauto.
    intros.
    eapply TrfLocalTrmEnsuresCond in t. 2: eauto.
    hnf in t. inversions t. inversions H1.
    eapply H; eauto. 
  -
    inversions c0; subst. 
    eapply TrfLocalLitEnsuresCond in t. hnf in t. inversions t.
    eapply trf_local_idtyp_trm_litl; hnf; eauto.
  -
    inversions c0; subst. 
    eapply trf_local_idtyp_trm_litr; hnf; eauto.
    intros.
    eapply TrfLocalTrmEnsuresCond in t. 2: eauto.
    hnf in t. inversions t. inversions H1.
    eapply H; eauto. 
  -
    inversions c0; subst.
    inversions c;
    eapply trf_local_idtyp_lit_fun; hnf; eauto.
  -
    inversions c0; subst.
    destruct (mode_objlit_refcaps' typing_mode) in c; inversions c;
    eapply trf_local_idtyp_lit_obj; hnf; eauto.
  -
    inversions c0; subst.
    inversions c; subst.
    eapply trf_local_idtyp_def_met; hnf; eauto.
  -
    inversions c0; subst.
    inversions c; subst. eapply trf_local_idtyp_defs_one; hnf; eauto.
  -
    inversions c0; subst.
    inversions c; subst.
    eapply trf_local_idtyp_defs_consl; hnf; eauto.
  -
    inversions c0; subst.
    inversions c; subst.
    eapply trf_local_idtyp_defs_consr; hnf; eauto.
Qed.

Lemma trf_local_trm_idtyp_eq trf
  (TTI: TransformEnsures (@TransformCondTypeIdentical TrfContextTy typ trm) trf):
  (trf_local_trm TransformCondTypeIdentical trf) = (trf_local_idtyp_trm trf).
Proof.
  eapply pred_ext_2; intros.
  split.
  - eapply (proj41 (trf_local_local_idtyp_ind TTI)); eauto. 
  - eapply (proj41 (trf_local_idtyp_local_ind trf)); eauto.
Qed.
Lemma trf_local_lit_idtyp_eq trf
  (TTI: TransformEnsures (@TransformCondTypeIdentical TrfContextTy typ trm) trf):
  (trf_local_lit TransformCondTypeIdentical trf) = (trf_local_idtyp_lit trf).
Proof.
  eapply pred_ext_2; intros.
  split.
  - eapply (proj42 (trf_local_local_idtyp_ind TTI)); eauto. 
  - eapply (proj42 (trf_local_idtyp_local_ind trf)); eauto.
Qed.
Lemma trf_local_def_idtyp_eq trf
  (TTI: TransformEnsures (@TransformCondTypeIdentical TrfContextTy typ trm) trf):
  (trf_local_def TransformCondTypeIdentical trf) = (trf_local_idtyp_def trf).
Proof.
  eapply pred_ext_2; intros.
  split.
  - eapply (proj43 (trf_local_local_idtyp_ind TTI)); eauto. 
  - eapply (proj43 (trf_local_idtyp_local_ind trf)); eauto.
Qed.

Lemma trf_local_defs_idtyp_eq trf
  (TTI: TransformEnsures (@TransformCondTypeIdentical TrfContextTy typ trm) trf):
  (trf_local_defs TransformCondTypeIdentical trf) = (trf_local_idtyp_defs trf).
Proof.
  eapply pred_ext_2; intros.
  split.
  - eapply (proj44 (trf_local_local_idtyp_ind TTI)); eauto. 
  - eapply (proj44 (trf_local_idtyp_local_ind trf)); eauto.
Qed.



Lemma trf_local_inverse_ind Cond (trf: TermTransform):
  InverseTransformIndCond Cond trf trf_local_trm /\
  InverseTransformIndCond Cond trf trf_local_lit /\
  InverseTransformIndCond Cond trf trf_local_def /\
  InverseTransformIndCond Cond trf trf_local_defs.
Proof.
  unfold InverseTransformInd, ImplyTransformInd.
  eapply trf_local_mutind; intros.
  - econstructor. unfold trf_inverse; eauto. auto. auto.
    unfold trf_liftcond_inverse; eauto.
  - unfold trf_inverse at 1 in H. unfold trf_inverse at 1.
    eapply trf_local_trm_letl; eauto.
  - unfold trf_inverse at 1 in H. unfold trf_inverse at 1.
    eapply trf_local_trm_letr; eauto.
  - unfold trf_inverse at 1 in H. unfold trf_inverse at 1.
    eapply trf_local_trm_litl; eauto.
  - unfold trf_inverse at 1 in H. unfold trf_inverse at 1.
    eapply trf_local_trm_litr; eauto.
  - unfold trf_inverse at 1 in H. unfold trf_inverse at 1.
    eapply trf_local_lit_fun; eauto.
  - unfold trf_inverse at 1 in H. unfold trf_inverse at 1.
    eapply trf_local_lit_obj; eauto.
  - unfold trf_inverse at 1 in H. unfold trf_inverse at 1.
    eapply trf_local_def_met; eauto.
  - unfold trf_inverse at 1 in H. unfold trf_inverse at 1.
    eapply trf_local_defs_one; eauto.
  - unfold trf_inverse at 1 in H. unfold trf_inverse at 1.
    eapply trf_local_defs_consl; eauto.
  - unfold trf_inverse at 1 in H. unfold trf_inverse at 1.
    eapply trf_local_defs_consr; eauto.
Qed.

Lemma trf_local_inverse_trm Cond (trf1: TermTransform):
  trf_inverse (trf_local_trm Cond trf1) = trf_local_trm (trf_liftcond_inverse Cond) (trf_inverse trf1).
Proof.
  destructc (trf_local_inverse_ind Cond trf1) as Htrm ?.
  destructc (trf_local_inverse_ind (trf_liftcond_inverse Cond) (trf_inverse trf1)) as Htrmi ?.
  hnf in Htrm, Htrmi.
  eapply pred_ext_2.
  intros.
  unfold trf_inverse.
  split; intros; destruct_all.
  - eapply Htrm in H1.
    unfold trf_inverse in *; eauto.
  - eapply Htrmi in H1.
    unfold trf_inverse in *; eauto.
Qed.


Lemma trf_local_inverse_lit Cond (trf1: TermTransform):
  trf_inverse (trf_local_lit Cond trf1) = trf_local_lit (trf_liftcond_inverse Cond) (trf_inverse trf1).
Proof.
  destructc (trf_local_inverse_ind Cond trf1) as Htrm Hlit ?.
  destructc (trf_local_inverse_ind (trf_liftcond_inverse Cond) (trf_inverse trf1)) as Htrmi Hliti ?.
  hnf in Htrm, Htrmi.
  eapply pred_ext_2.
  intros.
  unfold trf_inverse.

  split; intros; destruct_all.
  - eapply Hlit in H1.
    unfold trf_inverse in *; eauto.
  - eapply Hliti in H1.
    unfold trf_inverse in *; eauto.
Qed.

Lemma trf_local_inverse_def Cond (trf1: TermTransform):
  trf_inverse (trf_local_def Cond trf1) = trf_local_def (trf_liftcond_inverse Cond) (trf_inverse trf1).
Proof.
  destructc (trf_local_inverse_ind Cond trf1) as Htrm ?  Hdef ?.
  destructc (trf_local_inverse_ind (trf_liftcond_inverse Cond) (trf_inverse trf1)) as Htrmi ? Hdefi ?.
  hnf in Htrm, Htrmi.
  eapply pred_ext_2.
  intros.
  unfold trf_inverse.
  split; intros; destruct_all.
  - eapply Hdef in H3.
    unfold trf_inverse in *; eauto.
  - eapply Hdefi in H3.
    unfold trf_inverse in *; eauto.
Qed.

Lemma trf_local_inverse_defs Cond (trf1: TermTransform):
  trf_inverse (trf_local_defs Cond trf1) = trf_local_defs (trf_liftcond_inverse Cond) (trf_inverse trf1).
Proof.
  destructc (trf_local_inverse_ind Cond trf1) as Htrm ? ? Hdefs.
  destructc (trf_local_inverse_ind (trf_liftcond_inverse Cond) (trf_inverse trf1)) as Htrmi ? ? Hdefsi.
  hnf in Htrm, Htrmi.
  eapply pred_ext_2.
  intros.
  unfold trf_inverse.
  split; intros; destruct_all.
  - eapply Hdefs in H3.
    unfold trf_inverse in *; eauto.
  - eapply Hdefsi in H3.
    unfold trf_inverse in *; eauto.
Qed.


Lemma trf_liftcond_inverse_idtyp:
  trf_liftcond_inverse TransformCondTypeIdentical = TransformCondTypeIdentical.
Proof.
 unfold trf_liftcond_inverse, TransformCondTypeIdentical.
  intros.
  eapply fun_ext_5; intros. destruct x4, x5.
  eapply prop_ext; split; intros; destructecs_all; split; auto.
Qed.

Lemma trf_ensures_typeid_inverse (trf1: TermTransform):
  TransformEnsures (TransformCondTypeIdentical (A:=trm)) (trf1) ->
  TransformEnsures (TransformCondTypeIdentical (A:=trm)) (trf_inverse trf1).
Proof.
  introv TTI.
  hnf. hnf in TTI. intros. hnf in H. forwards: TTI H. hnf in H0. hnf. destruct TEa, TEb. destructecs H0. auto.
Qed.

Lemma trf_local_idtyp_inverse_trm (trf1: TermTransform)
  (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf1):
  trf_inverse (trf_local_idtyp_trm trf1) = trf_local_idtyp_trm (trf_inverse trf1).
Proof.
  rewrite <- trf_local_trm_idtyp_eq.
  rewrite trf_local_inverse_trm.
  rewrite <- trf_local_trm_idtyp_eq.
  fequal.
  apply trf_liftcond_inverse_idtyp.
  apply trf_ensures_typeid_inverse; auto.
  auto.
Qed.
Lemma trf_local_idtyp_inverse_lit (trf1: TermTransform)
  (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf1):
  trf_inverse (trf_local_idtyp_lit trf1) = trf_local_idtyp_lit (trf_inverse trf1).
Proof.
  rewrite <- trf_local_lit_idtyp_eq.
  rewrite trf_local_inverse_lit.
  rewrite <- trf_local_lit_idtyp_eq.
  fequal.
  apply trf_liftcond_inverse_idtyp.
  apply trf_ensures_typeid_inverse; auto.
  auto.
Qed.

Lemma trf_local_idtyp_inverse_def (trf1: TermTransform)
  (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf1):
  trf_inverse (trf_local_idtyp_def trf1) = trf_local_idtyp_def (trf_inverse trf1).
Proof.
  rewrite <- trf_local_def_idtyp_eq.
  rewrite trf_local_inverse_def.
  rewrite <- trf_local_def_idtyp_eq.
  fequal.
  eapply trf_liftcond_inverse_idtyp.
  apply trf_ensures_typeid_inverse; auto.
  auto.
Qed.

Lemma trf_local_idtyp_inverse_defs (trf1: TermTransform)
  (TTI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) trf1):
  trf_inverse (trf_local_idtyp_defs trf1) = trf_local_idtyp_defs (trf_inverse trf1).
Proof.
  rewrite <- trf_local_defs_idtyp_eq.
  rewrite trf_local_inverse_defs.
  rewrite <- trf_local_defs_idtyp_eq.
  fequal.
  eapply trf_liftcond_inverse_idtyp.
  apply trf_ensures_typeid_inverse; auto.
  auto.
Qed.

Lemma trf_local_idtyp_trm_subsumption (TL: TermTransform) G e ta tb T S:
  〈G⁏e ⊢ ta ∶ T〉 ∼(trf_local_idtyp_trm TL)∼> 〈G⁏e ⊢ tb ∶ T〉 ->
  G ⁏ e ⊢ T <: S ->
  〈G⁏e ⊢ ta ∶ S〉 ∼(trf_local_idtyp_trm TL)∼> 〈G⁏e ⊢ tb ∶ S〉.
Proof.
  introv Htrf Hsu.
  inversions Htrf.
  - econstructor; eauto.
  - eapply trf_local_idtyp_trm_letl with (S:=S); eauto.
  - eapply trf_local_idtyp_trm_letr with (S:=S); eauto.
  - eapply trf_local_idtyp_trm_litl with (S:=S); eauto.
  - eapply trf_local_idtyp_trm_litr with (S:=S); eauto.
Qed.

Lemma trf_cond_trm_typed_subsumption G e t S T:
  TransformCondTermTyped typing_mode (〈 G ⁏ e ⊢ t ∶ T 〉) ->
  G ⁏ e ⊢ T <: S ->
  TransformCondTermTyped typing_mode (〈 G ⁏ e ⊢ t ∶ S 〉).
Proof.
  introv Hcnd Hsu.
  hnf in *.
  eapply ty_trm_sub; eauto.
Qed.

Lemma trf_local_typed_ind Cond (trf: TermTransform)(TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  CondTransformInd (TransformCondTermTyped typing_mode) (trf_local_trm Cond trf) /\
  CondTransformInd (TransformCondLitTyped typing_mode) (trf_local_lit Cond trf) /\
  CondTransformInd (TransformCondDefTyped typing_mode) (trf_local_def Cond trf) /\
  CondTransformInd (TransformCondDefsTyped typing_mode) (trf_local_defs Cond trf).
Proof.
  eapply trf_local_mutind; intros.
  - eapply trf_cond_trm_typed_subsumption. eapply trf_requires; eauto. eauto.
  - forwards Hty1: H; eauto.
  - eapply trf_cond_trm_typed_subsumption; rweauto.  simpls. eapply ty_trm_let with (L:=L \u ctx_dom Ga); eauto. 
  - eapply trf_cond_trm_typed_subsumption; rweauto.  simpls. forwards Hty2: H; eauto.
  - eapply trf_cond_trm_typed_subsumption; rweauto.  simpls. eapply ty_trm_lit with (L:=L \u ctx_dom Ga) ; eauto.
  - simpls. eapply ty_lit_fun with (L:=L \u ctx_dom Ga); eauto.
  - simpls. eapply ty_lit_obj with (L:=L \u ctx_dom Ga); eauto.
  - simpls.
    eapply ty_def_met with (L:=L \u ctx_dom Ga \u objctx_dom Goa); eauto.
  - simpls; eauto.
  - simpls; eauto.
  - simpls; eauto.
Qed.

Global Instance TransformLocalTrmTyped Cond trf (TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires (TransformCondTermTyped typing_mode) (trf_local_trm Cond trf).
Proof.
  hnf. intros.
  eapply trf_local_typed_ind; eauto.
Qed.

Global Instance TransformTrmDefsTyped Cond trf (TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires (TransformCondDefsTyped typing_mode) (trf_local_defs Cond trf).
Proof.
  hnf. intros.
  eapply trf_local_typed_ind; eauto.
Qed.

Lemma trf_local_idtyp_typed_ind (trf: TermTransform)(TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  CondTransformInd (TransformCondTermTyped typing_mode) (trf_local_idtyp_trm trf) /\
  CondTransformInd (TransformCondLitTyped typing_mode) (trf_local_idtyp_lit trf) /\
  CondTransformInd (TransformCondDefTyped typing_mode) (trf_local_idtyp_def trf) /\
  CondTransformInd (TransformCondDefsTyped typing_mode) (trf_local_idtyp_defs trf).
Proof.
  eapply trf_local_idtyp_mutind; intros.
  - eapply trf_cond_trm_typed_subsumption. eapply trf_requires; eauto. eauto.
  - forwards Hty1: H; eauto.
  - eapply trf_cond_trm_typed_subsumption; rweauto.  simpls. eapply ty_trm_let with (L:=L \u ctx_dom G); eauto. 
  - eapply trf_cond_trm_typed_subsumption; rweauto.  simpls. forwards Hty2: H; eauto.
  - eapply trf_cond_trm_typed_subsumption; rweauto.  simpls. eapply ty_trm_lit with (L:=L \u ctx_dom G) ; eauto.
  - simpls. eapply ty_lit_fun with (L:=L \u ctx_dom G); eauto.
  - simpls. eapply ty_lit_obj with (L:=L \u ctx_dom G); eauto.
  - simpls.
    eapply ty_def_met with (L:=L \u ctx_dom G \u objctx_dom Go); eauto.
  - simpls; eauto.
  - simpls; eauto.
  - simpls; eauto.
Qed.

Global Instance TransformLocalIdtypTrmTyped trf (TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires (TransformCondTermTyped typing_mode) (trf_local_idtyp_trm trf).
Proof.
  hnf. intros.
  eapply trf_local_idtyp_typed_ind; eauto.
Qed.

Global Instance TransformLocalIdtypDefsTyped trf (TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires (TransformCondDefsTyped typing_mode) (trf_local_idtyp_defs trf).
Proof.
  hnf. intros.
  eapply trf_local_idtyp_typed_ind; eauto.
Qed.
Global Instance TransformLocalIdtypLitTyped trf (TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires (TransformCondLitTyped typing_mode) (trf_local_idtyp_lit trf).
Proof.
  hnf. intros.
  eapply trf_local_idtyp_typed_ind; eauto.
Qed.

Global Instance TransformLocalIdtypDefTyped trf (TI: TransformRequires (TransformCondTermTyped typing_mode) trf):
  TransformRequires (TransformCondDefTyped typing_mode) (trf_local_idtyp_def trf).
Proof.
  hnf. intros.
  eapply trf_local_idtyp_typed_ind; eauto.
Qed.



Lemma trf_cond_trm_ok_subsumption G e (t: trm) (S T: typ):
  TransformCondOk (〈 G ⁏ e ⊢ t ∶ T 〉) ->
  TransformCondOk  (〈 G ⁏ e ⊢ t ∶ S 〉).
Proof.
  introv Hcnd.
  hnf in *.
  auto.
Qed.

Lemma trf_local_ok_ind Cond (trf: TermTransform)(TI: TransformRequires (@TransformCondOk _ _) trf):
  CondTransformInd (@TransformCondOk _ _) (trf_local_trm Cond trf) /\
  CondTransformInd (@TransformCondOk _ _) (trf_local_lit Cond trf) /\
  CondTransformInd (@TransformCondObjOk _ _) (trf_local_def Cond trf) /\
  CondTransformInd (@TransformCondObjOk _ _) (trf_local_defs Cond trf).
Proof.
  eapply trf_local_mutind; intros; try assumption.
  - eapply trf_cond_trm_ok_subsumption. eapply trf_requires; eauto.
  - simpls. pick_fresh z. forwards HpkG Hoke: (H z); eauto.
  - simpls. pick_fresh z. forwards HpkG Hoke: (H z); eauto.
  - simpls. pick_fresh z. forwards HpkG Hoke: (H z); eauto.
  - simpls. pick_fresh z. forwards HpkG Hoke: (H z); eauto.
  - simpls. pick_fresh zz. pick_fresh zc. forwards H1: H zz zc; eauto. unfold ctx_obj_met in H1. destruct H1. destruct Goa; destruct (mode_visibility' typing_mode); inversions H0; eauto; inversions H4; eauto.
Qed.

Global Instance OkTransformTrm Cond trf (TI: TransformRequires (@TransformCondOk _ _)  trf):
  TransformRequires (@TransformCondOk _ _) (trf_local_trm Cond trf).
Proof.
  hnf. intros.
  eapply (proj41 (trf_local_ok_ind Cond TI)); eauto.
Qed.

Global Instance OkTransformTrmDefs Cond trf (TI: TransformRequires (@TransformCondOk _ _)  trf):
  TransformRequires (@TransformCondObjOk _ _) (trf_local_defs Cond trf).
Proof.
  hnf. intros.
  eapply trf_local_ok_ind; eauto.
Qed.

Lemma trf_local_weakening_compat_ind (trf: TermTransform)(TI: TransformWeakeningCompat _ _ trf):
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_trm trf) (@TransformWeakening _ _) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_lit trf) (@TransformWeakening _ _) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_def trf) (@TransformObjWeakening _ _) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_defs trf) (@TransformObjWeakening _ _).
Proof.

  Local Ltac invert_weakenings := (
    repeat match goal with 
    | H: TransformWeakening ?TEa ?Teb |- _ => inversions H
    | H: TransformObjWeakening ?TEa ?Teb |- _ => inversions H
    end;
    match goal with
    | H: ?G1 §& ?G2 §& ?G3 = ?G4 §& ?G5 §& ?G6 |- _ => rewrite H; clears G1 G2 G3
    end;
    try match goal with
    | H: ?Ga1 §& ?Ga3 ⁏ ?e0 ⊢ ?T <: ?T1 |- _ => eapply weaken in H; [| eauto| eauto using ctx_ok_concat_l]
    end
  ).


  eapply trf_local_idtyp_mutind; intros; destruct TEc, TEd; subst;
  match goal with  (* H1 must be inverted so that trf_local_idtyp_trm_local can be applied, but also kept so that TI can be applied *)
    H: TransformCondTypeIdentical ?TEc ?TEd |- _ =>
    let H':=fresh H in pose proof H as H'; inversions H'
  end.
  - destruct G1.
    invert_weakenings.
    eapply trf_local_idtyp_trm_local; eauto. eapply TI. eauto. econstructor; eauto. econstructor; eauto. econstructor; eauto.
  - 
    invert_weakenings.
    forwards: H; eauto. hnf; eauto.
    match goal with 
    | H: forall x : var, x \notin ?L -> ?P |- _ => assert (forall x : var, x \notin (L \u ctx_dom Ga2) -> P) by (intros; eapply H; eauto)
    end.
    match type of o with 
    |  context A [Ga1 §& Ga3]  => let x:= context A [Ga1 §& Ga2 §& Ga3] in assert (x)
    end. intros. rewrite <- ! ctx_concat_push with (G1:=Ga3). eapply weaken_ty; eauto using ctx_ok_concat_l. 
    eapply trf_local_idtyp_trm_letl; eauto.
  - invert_weakenings. 

    eapply trf_local_idtyp_trm_letr with (L:= L \u ctx_dom (Ga1 §& Ga2 §& Ga3)); eauto.
    +
    intros.
    eapply (H x); eauto.
    rewrite <- ! ctx_concat_push with (G1:=Ga3). econstructor; eauto.
    ctx_flatten. eauto.
    rewrite <- ! ctx_concat_push with (G1:=Ga3). econstructor; eauto.
    ctx_flatten. eauto.
    hnf; eauto.
    +
    eapply weaken_ty; eauto using ctx_ok_concat_l.
  - invert_weakenings.
    eapply trf_local_idtyp_trm_litl; eauto.
    +
    apply H; eauto.
    hnf; eauto.
    +
    intros.
    rewrite <- ! ctx_concat_push with (G1:=Ga3).
    eapply weaken_ty; eauto using ctx_ok_concat_l.
  - invert_weakenings.
    eapply trf_local_idtyp_trm_litr with (L:= L \u ctx_dom (Ga1 §& Ga2 §& Ga3)); eauto.
    intros.
    +

    eapply (H x); eauto.
    rewrite <- ! ctx_concat_push with (G1:=Ga3).
    econstructor;
    ctx_flatten; eauto.
    rewrite <- ! ctx_concat_push with (G1:=Ga3).
    econstructor;
    ctx_flatten; eauto.
    hnf; eauto.
    +
    eapply weaken_ty; eauto using ctx_ok_concat_l.
  - invert_weakenings.
    eapply trf_local_idtyp_lit_fun with (L:= L \u ctx_dom (Ga1 §& Ga2 §& Ga3)); eauto. intros.  eapply (H x);  rewrite <- ? ctx_concat_push with (G1:=Ga3); eauto.
    econstructor; rewrite ? ctx_concat_push; eauto.
    econstructor; rewrite ? ctx_concat_push; eauto.   hnf; eauto.
  - invert_weakenings.
    eapply trf_local_idtyp_lit_obj with (L:= L \u ctx_dom (Ga1 §& Ga2 §& Ga3)); eauto. intros; eapply H; eauto. eauto. hnf; eauto.
  - invert_weakenings.
    eapply trf_local_idtyp_def_met with (L:= L \u ctx_dom (Ga1 §& Ga2 §& Ga3 §§& Go)); eauto.
    intros.
    rewrite ctx_obj_met_assoc.
    eapply H; eauto. eauto.
    rewrite ctx_obj_met_assoc; eauto.
    rewrite ! ctx_obj_assoc, !ctx_dom_concat, ctx_obj_dom in H0, H3.
    econstructor; eauto.
    rewrite ctx_obj_met_assoc; eauto.
    rewrite ! ctx_obj_assoc, !ctx_dom_concat, ctx_obj_dom in H0, H3.
    econstructor; eauto.
    hnf; eauto.
  - invert_weakenings.
    eapply trf_local_idtyp_defs_one; eauto.
    eapply H; eauto.
    hnf; eauto.
  - invert_weakenings.
    eapply trf_local_idtyp_defs_consl; eauto.
    eapply H; eauto.
    hnf; eauto.
    eapply weaken; eauto.
    rewrite ctx_obj_assoc in H12. eauto using ctx_ok_concat_l.
  - invert_weakenings.
    eapply trf_local_idtyp_defs_consr; eauto.
    eapply H; eauto.
    hnf; eauto.
    eapply weaken; eauto.
    rewrite ctx_obj_assoc in H12. eauto using ctx_ok_concat_l.
Qed.

(* Lemma 53 in "Pure methods for roDOT" *)
Global Instance WeakeningCompatTransformationTrm trf (TI: TransformWeakeningCompat _ _ trf):
  TransformWeakeningCompat _ _ (trf_local_idtyp_trm trf).
Proof.
  hnf. intros.
  eapply trf_local_weakening_compat_ind; eauto.
Qed.
Global Instance WeakeningCompatTransformationDefs trf (TI: TransformWeakeningCompat _ _ trf):
  TransformObjWeakeningCompat _ _ (trf_local_idtyp_defs trf).
Proof.
  hnf. intros.
  eapply trf_local_weakening_compat_ind; eauto.
Qed.


Lemma trf_local_eweakening_compat_ind (trf: TermTransform)(TI: TransformEWeakeningCompat _ _ trf):
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_trm trf) (@TransformEWeakening _ _) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_lit trf) (@TransformEWeakening _ _) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_def trf) (@TransformObjEWeakening _ _) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_defs trf) (@TransformObjEWeakening _ _).
Proof.

  Local Ltac invert_eweakenings := (
    repeat match goal with 
    | H: TransformEWeakening ?TEa ?Teb |- _ => inversions H
    | H: TransformObjEWeakening ?TEa ?Teb |- _ => inversions H
    end;
    match goal with
    | H: ?G1 & ?G2 & ?G3 = ?G4 & ?G5 & ?G6 |- _ => rewrite H; clears G1 G2 G3
    end;
    try  match goal with
    | H: ?G0 ⁏ ?e1 & ?e3 ⊢ ?T <: ?T1, Hok: ok (?e1 & ?e2 & ?e3) |- _ =>  eapply ewkn with (e5:=e2) in H; [| eauto]
    end
  ).


  eapply trf_local_idtyp_mutind; intros; destruct TEc, TEd; subst;
  match goal with  (* H1 must be inverted so that trf_local_idtyp_trm_local can be applied, but also kept so that TI can be applied *)
    H: TransformCondTypeIdentical ?TEc ?TEd |- _ =>
    let H':=fresh H in pose proof H as H'; inversions H'
  end.
  - destruct G1.

    invert_eweakenings.
    eapply trf_local_idtyp_trm_local; eauto. eapply TI. eauto. econstructor; eauto. econstructor; eauto. econstructor; eauto.
  - 
    invert_eweakenings.
    forwards: H; eauto. hnf; eauto.
    match goal with 
    | H: forall x : var, x \notin ?L -> ?P |- _ => assert (forall x : var, x \notin (L \u dom e2) -> P) by (intros; eapply H; eauto)
    end.
    match type of o with 
    |  context A [e1 & e3]  => let x:= context A [e1 & e2 & e3] in assert (x)
    end. intros. eapply ewkn; eauto using ctx_ok_concat_l. 
    eapply trf_local_idtyp_trm_letl; eauto.
  - invert_eweakenings. 

    eapply trf_local_idtyp_trm_letr with (L:= L \u ctx_dom G \u dom (e1 & e2 & e3)); eauto.
    +
    intros.
    eapply (H x). eauto.
    econstructor; eauto.
    econstructor; eauto.
    hnf; eauto.
    +
    eapply ewkn; eauto.
  - invert_eweakenings.
    eapply trf_local_idtyp_trm_litl; eauto.
    +
    apply H; eauto.
    hnf; eauto.
    +
    intros.
    eapply ewkn; eauto using ctx_ok_concat_l.
  - invert_eweakenings.
    eapply trf_local_idtyp_trm_litr with (L:= L \u ctx_dom G \u dom (e1 & e2 & e3)); eauto.
    intros.
    +

    eapply (H x); eauto.
    econstructor;
    ctx_flatten; eauto.
  
    +
    eapply ewkn; eauto using ctx_ok_concat_l.
  - invert_eweakenings.
    eapply trf_local_idtyp_lit_fun with (L:= L \u ctx_dom G \u dom (e1 & e2 & e3)); eauto. intros; eapply H; eauto. eauto. hnf; eauto.
  - invert_eweakenings.
    eapply trf_local_idtyp_lit_obj with (L:= L \u ctx_dom G \u dom (e1 & e2 & e3)); eauto. intros; eapply H; eauto. eauto. hnf; eauto.
  - invert_eweakenings.
    eapply trf_local_idtyp_def_met with (L:= L \u ctx_dom (G §§& Go) \u dom (e1 & e2 & e3)); eauto.
    intros.
    rewrite ctx_obj_dom in H0, H3.
    forwards HokG: ctx_obj_met_ok ctx_empty G Go xz xc; fweauto.
    rewrite ctx_obj_assoc. rewrite ctx_concat_empty_l. auto.
    rewrite ctx_concat_empty_l in HokG.
    eauto.
    eapply (H xz); eauto. eauto.
    econstructor; eauto.
  - invert_eweakenings.
    eapply trf_local_idtyp_defs_one; eauto.
    eapply H; eauto.
    hnf; eauto.
  - invert_eweakenings.
    eapply trf_local_idtyp_defs_consl; eauto.
    eapply H; eauto.
    hnf; eauto.
    eapply ewkn; eauto.
  - invert_eweakenings.
    eapply trf_local_idtyp_defs_consr; eauto.
    eapply H; eauto.
    hnf; eauto.
    eapply ewkn; eauto.
Qed.


Global Instance EWeakeningCompatTransformationTrm trf (TI: TransformEWeakeningCompat _ _ trf):
  TransformEWeakeningCompat _ _ (trf_local_idtyp_trm trf).
Proof.
  hnf. intros.
  eapply trf_local_eweakening_compat_ind; eauto.
Qed.
Global Instance EWeakeningCompatTransformationDefs trf (TI: TransformEWeakeningCompat _ _ trf):
  TransformObjEWeakeningCompat _ _ (trf_local_idtyp_defs trf).
Proof.
  hnf. intros.
  eapply trf_local_eweakening_compat_ind; eauto.
Qed.


Global Instance IsAnswerTransformationTrm Cond trf (TI: TransformPreserves TransformIsAnswerTerm  trf):
   TransformPreserves TransformIsAnswerTerm (trf_local_trm Cond trf).
Proof.
  hnf. intros.
  unfold TransformIsAnswerTerm in *.
  (* Discriminate solves all cases except trf_local_trm_local *)
  inversions H; destructecs H0; try discriminate.
  forwards: TI (〈 Ga ⁏ ea  ⊢ ta ∶ Ta 〉); eauto.
Qed.

Global Instance IsAnswerTransformationTrmIdtyp trf (TI: TransformPreserves TransformIsAnswerTerm  trf):
   TransformPreserves TransformIsAnswerTerm (trf_local_idtyp_trm trf).
Proof.
  hnf. intros.
  unfold TransformIsAnswerTerm in *.
  (* Discriminate solves all cases except trf_local_trm_local *)
  inversions H; destructecs H0; try discriminate.
  forwards: TI (〈 G ⁏ e  ⊢ ta ∶ T 〉); eauto.
Qed.

Lemma is_answer_transformation_trm_inv trf (TI: TransformPreserves TransformIsAnswerTerm (trf_local_trm trf_liftcond_true trf)):
  TransformPreserves TransformIsAnswerTerm trf.
Proof.
  hnf. intros.
  destruct TEa, TEb.
  destruct G0, G.
  eapply trf_preserves.

  eapply trf_local_trm_local; eauto.
  2: eauto. eauto.
Qed.


Lemma subctx_middle Ga1 x vk Txa Ga2 e Txb:
  ctx_ok ((Ga1 § x / vk ~ Txa) §& Ga2) ->
  ok e ->
  Ga1 ⁏ e ⊢ Txb <: Txa ->
  e ⊢ (Ga1 § x / vk ~ Txb) §& Ga2 ⪯ (Ga1 § x / vk ~ Txa) §& Ga2.
Proof.
  introv HokG Hoke Hsu.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr HokG.

  eapply subctx_concat.
  eapply subctx_grow; eauto.
  eapply subctx_refl.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

Qed.

Lemma trf_local_narrowing_compat_ind (trf: TermTransform)(TI: TransformNarrowingCompat _ _ _ trf):
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_trm trf) (@TransformNarrowing _ _ _) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_lit trf) (@TransformNarrowing _ _ _) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_def trf) (@TransformObjNarrowing _ _ _) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_defs trf) (@TransformObjNarrowing _ _ _).
Proof.

  Local Ltac invert_narrowings := (
    repeat match goal with 
    | H: TransformNarrowing _ ?TEa ?Teb |- _ => inversions H
    | H: TransformObjNarrowing _ ?TEa ?Teb |- _ => inversions H
    end ;
    match goal with
    | H: (?Ga0 § ?x0 / ?vk0 ~ ?Txb0) §& ?Ga3 = (?Ga1 § ?x / ?vk ~ ?Txb) §& ?Ga2  |- _ => rewrite H; clears Ga0 x0 vk0 Txb0 Ga3
    end;
    let Hsubctx := fresh "Hsubctx" in (
    forwards Hsubctx: subctx_middle; fweauto );
    try match goal with
    | H: (?Ga1 § ?x / ?vk ~ ?Txa) §& ?Ga2 ⁏ ?e0 ⊢ ?T <: ?T1, Hs: ?e ⊢ ?G1 ⪯ ?G2 |- _ => eapply narrow_subtyping in H; [| eapply narrow_subtyp| eapply Hs ]
    end
  ).


  eapply trf_local_idtyp_mutind; intros; destruct TEc, TEd; subst;
  match goal with  (* H1 must be inverted so that trf_local_idtyp_trm_local can be applied, but also kept so that TI can be applied *)
    H: TransformCondTypeIdentical ?TEc ?TEd |- _ =>
    let H':=fresh H in pose proof H as H'; inversions H'
  end.
  - destruct G1.
    invert_narrowings.
    eapply trf_local_idtyp_trm_local; eauto. eapply TI. eauto. constructor; eauto. constructor; eauto. constructor; eauto.
  - 
    invert_narrowings.

    forwards: H; eauto. hnf; eauto.

    match goal with 
    | H: forall x : var, x \notin ?L -> ?P |- _ => assert (forall x : var, x \notin (L \u ctx_dom Ga2) -> P) by (intros; eapply H; eauto)
    end.
  
    eapply trf_local_idtyp_trm_letl with (L:=L\u ctx_dom ((Ga1 § x / vk ~ Txb) §&  Ga2) ); eauto.
    intros.
    forwards: H1 x0. eauto.

    eapply narrow_ty; eauto using ctx_ok_concat_l.

  - invert_narrowings. 

    eapply trf_local_idtyp_trm_letr with (L:= L \u ctx_dom ((Ga1 § x / vk ~ Txb) §& Ga2)); eauto.
    +
    intros.
    eapply (H x0); eauto.
    rewrite <- ! ctx_concat_push with (G1:=Ga2). econstructor; eauto.
    ctx_flatten. eauto.
    rewrite <- ! ctx_concat_push with (G1:=Ga2). econstructor; eauto.
    ctx_flatten. eauto.
    hnf; eauto.
    +
    eapply narrow_ty; eauto using ctx_ok_concat_l.
  - invert_narrowings.
    eapply trf_local_idtyp_trm_litl with (L:=L\u ctx_dom ((Ga1 § x / vk ~ Txb) §&  Ga2) ); eauto.
    +
    apply H; eauto.
    hnf; eauto.
    +
    intros.
    rewrite <- ! ctx_concat_push with (G1:=Ga2).
    eapply narrow_ty. eauto. eauto using ctx_ok_concat_l.
    ctx_flatten. eapply subctx_push; eauto.
  - invert_narrowings.
    eapply trf_local_idtyp_trm_litr with (L:=L\u ctx_dom ((Ga1 § x / vk ~ Txb) §&  Ga2) ); eauto.
    intros.
    +
    eapply (H x0); eauto.
    rewrite <- ! ctx_concat_push with (G1:=Ga2).
    econstructor;
    ctx_flatten; eauto.
    rewrite <- ! ctx_concat_push with (G1:=Ga2).
    econstructor;
    ctx_flatten; eauto.
    hnf; eauto.
    +
    eapply narrow_ty; eauto using ctx_ok_concat_l.
    eapply narrow_ty_lit.
  - invert_narrowings.
    eapply trf_local_idtyp_lit_fun with (L:= L \u ctx_dom ((Ga1 § x / vk ~ Txb) §&  Ga2)); eauto. intros; eapply (H x0); eauto.
    rewrite <- ! ctx_concat_push with (G1:=Ga2).
    econstructor; rewrite ? ctx_concat_push; eauto.
    rewrite <- ! ctx_concat_push with (G1:=Ga2).
    econstructor; rewrite ? ctx_concat_push; eauto.
    hnf; eauto.
  - invert_narrowings.
    eapply trf_local_idtyp_lit_obj with (L:= L \u ctx_dom ((Ga1 § x / vk ~ Txb) §&  Ga2 )); eauto. intros.
    fresh_chain_init Fr. fresh_chain_add_fresh Fr H0. fresh_chain_cleanup Fr.
    rewrite fresh_chain_eq_swap_set_set in Fr by apply true.
    rewrite <- fresh_chain_eq_last_var_set in Fr by apply true.
    rewrite fresh_chain_eq_merge_var_ctxdom in Fr by apply true.
    rewrite fresh_chain_eq_merge_var_ctxdompush in Fr by apply true.
    eapply (H x0); eauto.
    fresh_chain_solve Fr.
    econstructor; rewrite ? ctx_concat_push; eauto. unfold ctx_obj. 
    fresh_chain_apply Fr.
    rewrite fresh_chain_eq_merge_var_ctxdom by apply true.
    rewrite fresh_chain_eq_merge_var_ctxdompush by apply true.
    fresh_chain_solve Fr.
    econstructor; rewrite ? ctx_concat_push; eauto. unfold ctx_obj. 
    fresh_chain_apply Fr.
    rewrite fresh_chain_eq_merge_var_ctxdom by apply true.
    rewrite fresh_chain_eq_merge_var_ctxdompush by apply true.
    fresh_chain_solve Fr.
    econstructor; rewrite ? ctx_concat_push; eauto.
  - invert_narrowings. 

    rewrite ctx_obj_assoc in H6; eapply H6. 
    rewrite <- ! ctx_obj_assoc in Hsubctx. 
  
    eapply trf_local_idtyp_def_met with (L:= L \u ctx_dom ((Ga1 § x / vk ~ Txb) §&  Ga2 §§& Go)); eauto.
    intros.
    rewrite ctx_obj_met_assoc.
    eapply H; eauto. eauto.

    rewrite ctx_obj_met_assoc; eauto.
    rewrite ! ctx_obj_assoc, !ctx_dom_concat, ctx_obj_dom in H0, H3.
    econstructor; eauto.

    rewrite ctx_obj_met_assoc; eauto.
    rewrite ! ctx_obj_assoc, !ctx_dom_concat, ctx_obj_dom in H0, H3.
    econstructor; eauto.

    hnf; eauto.
  - invert_narrowings.
    rewrite ctx_obj_assoc in H6; eapply H6.
    eapply trf_local_idtyp_defs_one; eauto.
    eapply H; eauto.
    hnf; eauto.
  - invert_narrowings.
    rewrite ctx_obj_assoc in H6; eapply H6.
    eapply trf_local_idtyp_defs_consl; eauto.
    eapply H; eauto.
    hnf; eauto.
    eapply narrow; eauto.
    rewrite ctx_obj_assoc in H6.
    rewrite ctx_obj_assoc. eauto using ctx_ok_concat_l.
    eapply subctx_middle; eauto. eapply ctx_ok_obj_inv_l; eauto.
  - invert_narrowings.
    rewrite ctx_obj_assoc in H6; eapply H6.
    eapply trf_local_idtyp_defs_consr; eauto.
    eapply H; eauto.
    hnf; eauto.
    eapply narrow; eauto.
    rewrite ctx_obj_assoc in H6.
    rewrite ctx_obj_assoc. eauto using ctx_ok_concat_l.
    eapply subctx_middle; eauto. eapply ctx_ok_obj_inv_l; eauto.
Qed.

Global Instance NarrowingCompatTransformationTrm trf (TI: TransformNarrowingCompat _ _ _ trf):
  TransformNarrowingCompat _ _ _ (trf_local_idtyp_trm trf).
Proof.
  hnf. intros.
  eapply trf_local_narrowing_compat_ind; eauto.
Qed.
Global Instance NarrowingCompatTransformationDefs trf (TI: TransformNarrowingCompat _ _ _ trf):
  TransformObjNarrowingCompat _ _ _ (trf_local_idtyp_defs trf).
Proof.
  hnf. intros.
  eapply trf_local_narrowing_compat_ind; eauto.
Qed.

Lemma trf_local_seal_compat_ind (trf: TermTransform)(TI: TransformSealingCompat _ _  trf) x:
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_trm trf) (@TransformSealing _ _ _  _ x) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_lit trf) (@TransformSealing _ _ _  _ x) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_def trf) (@TransformObjSealing _ _  _ _ x) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_defs trf) (@TransformObjSealing _ _ _ _ x).
Proof.
  eapply trf_local_idtyp_mutind; intros; destruct TEc, TEd; subst;
  match goal with  (* H1 must be inverted so that trf_local_idtyp_trm_local can be applied, but also kept so that TI can be applied *)
    H: TransformCondTypeIdentical ?TEc ?TEd |- _ =>
    let H':=fresh H in pose proof H as H'; inversions H'
  end.

  Local Ltac invert_sealings := (
    repeat match goal with 
    | H: TransformSealing _ ?TEa ?Teb |- _ => inversions H
    | H: TransformObjSealing _ ?TEa ?Teb |- _ => inversions H
    end
  ).
  Local Ltac sympl_seal := change seal_rec_trm with (@seal_rec trm TrmSealable);
   change seal_rec_def with (@seal_rec def DefSealable);
   change seal_rec_dec with (@seal_rec dec DecSealable);
   change seal_rec_defs with (@seal_rec defs DefsSealable);
   change seal_rec_typ with (@seal_rec typ TypSealable).

  - destruct G1.
    invert_sealings.
    eapply trf_local_idtyp_trm_local; eauto. eapply (TI x). eauto.
    constructor; eauto. constructor; eauto. constructor; eauto.
    eapply sealing_obj with (Go:=objctx_empty); eauto.
  - 
    invert_sealings.

    forwards: H; eauto. hnf; eauto.
  
    eapply trf_local_idtyp_trm_letl with (L:=L\u ctx_dom G \u \{x} ); eauto.
    intros. 
    forwards: o x0. eauto.
    sympl_seal.   
    rewrite ! open_seal_down.
    rewrite <- ctx_seal_push.
    eapply sealing_obj with (Go:=objctx_empty); eauto.
    eapply sealing_obj with (Go:=objctx_empty); eauto.

  - 
    invert_sealings.

    eapply trf_local_idtyp_trm_letr with (L:=L\u ctx_dom G \u \{x} ) (Tl:=Tl /[# 0 ->> x] ); eauto.
    intros.

    forwards: (H x0); eauto. hnf; eauto.
    sympl_seal. 
    rewrite ! open_seal_down.
    rewrite <- ctx_seal_push. eauto.
    eapply sealing_obj with (Go:=objctx_empty); eauto.
    eapply sealing_obj with (Go:=objctx_empty); eauto.

  - 
    invert_sealings.

    eapply trf_local_idtyp_trm_litl with (L:=L\u ctx_dom G \u \{x} ) (Tl:=Tl /[# 0 ->> x] ); eauto.


    forwards: (H); eauto. hnf; eauto.
    sympl_seal.
    intros.     
    rewrite ! open_seal_down.
    rewrite <- ctx_seal_push. eauto.
    eapply sealing_obj with (Go:=objctx_empty); eauto.
    eapply sealing_obj with (Go:=objctx_empty); eauto.

  - 
    invert_sealings.

    eapply trf_local_idtyp_trm_litr with (L:=L\u ctx_dom G \u \{x} ) (Tl:=Tl /[# 0 ->> x] ); eauto.
    intros.

    forwards: (H x0); eauto. hnf; eauto.
    sympl_seal. 
    rewrite ! open_seal_down.
    rewrite <- ctx_seal_push. eauto.
    eapply sealing_obj with (Go:=objctx_empty); eauto.
    eapply sealing_obj with (Go:=objctx_empty); eauto.
  - 
    invert_sealings.

    eapply trf_local_idtyp_lit_fun with (L:=L\u ctx_dom G \u \{x} ) ; eauto.
    intros.     

    forwards: (H x0); eauto. hnf; eauto.
    sympl_seal.

    rewrite ! open_seal_down.
    rewrite <- ctx_seal_push. eauto.
  - 
    invert_sealings.
    rewrite seal_typ_objlit.
    eapply trf_local_idtyp_lit_obj with (L:=L\u ctx_dom G \u \{x} ); eauto.
    intros.

    forwards: (H x0); eauto. hnf; eauto.
    sympl_seal. 
    rewrite ! open_seal_down.
    eauto.
    eapply typ_indep_seal; eauto.
  - 
    invert_sealings.
    eapply trf_local_idtyp_def_met with (L:=L\u ctx_dom G \u \{x} ); eauto.
    intros.

    forwards: H xz xc; eauto. hnf; eauto.
    sympl_seal.
    rewrite ! (open_seal_down 1).
    rewrite ! (open_seal_down).

    rewrite ! (@open_seal_down trm TrmOpenable TrmFreeVar TrmSubstVar TrmAbstractSyntax _ _ _ 0) .
    rewrite ! (@open_seal_down trm TrmOpenable TrmFreeVar TrmSubstVar TrmAbstractSyntax _ _ _ 0 x xz (tb /[# 1 -> xc]) ) .
    rewrite <- seal_ctx_obj_met. eauto.
  - 
    invert_sealings.

    eapply trf_local_idtyp_defs_one; eauto.
    forwards: (H); eauto. hnf; eauto.

  - 
    invert_sealings.

    eapply trf_local_idtyp_defs_consl; eauto using seal_defs_hasnt.
    forwards: (H); eauto. hnf; eauto.

    sympl_seal.
    eapply sealing_obj with (Go0:=Go); eauto.


 - 
    invert_sealings.

    eapply trf_local_idtyp_defs_consr; eauto using seal_defs_hasnt.
    forwards: (H); eauto. hnf; eauto.

    sympl_seal.
    eapply sealing_obj with (Go0:=Go); eauto.

Unshelve.
all: apply sealing_subtyp.
Qed.


Global Instance SealingCompatTransformationTrm (trf: TermTransform) (TI: TransformSealingCompat _ _ trf):
  TransformSealingCompat _ _ (trf_local_idtyp_trm trf).
Proof.
  intros.
  hnf. intros.
  eapply (trf_local_seal_compat_ind TI x); eauto.
Qed.
Global Instance SealingCompatTransformationDefs (trf: TermTransform) (TI: TransformSealingCompat _ _ trf):
  TransformObjSealingCompat _ _ (trf_local_idtyp_defs trf).
Proof.
  intros.
  hnf. intros.
  eapply trf_local_seal_compat_ind; eauto.
Qed.

Lemma trf_unhiding_met G0 G2 Go xz TZ xc TC (ta: trm) (T: typ) e:
TransformUnhiding (〈 ctx_obj_met typing_mode (ctx_hide G0 §& G2) Go xz TZ xc TC ⁏ e ⊢ ta ∶ T 〉)
  (〈 ctx_obj_met typing_mode (G0 §& G2) Go xz TZ xc TC ⁏ e ⊢  ta ∶ T  〉).
Proof.
  
    unfold ctx_obj_met in *;
    destruct (mode_visibility' typing_mode);
    destruct Go;

      ctx_split_from xz; rewrite <- ? ctx_concat_hide, <- ctx_concat_assoc; try constructor.
Qed.

Lemma trf_local_unhide_compat_ind (trf: TermTransform)(TI: TransformUnhidingCompat _ _ trf):
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_trm trf) (@TransformUnhiding _  _ ) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_lit trf) (@TransformUnhiding _   _ ) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_def trf) (@TransformObjUnhiding  _ _ ) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_defs trf) (@TransformObjUnhiding  _ _ ).
Proof.
  eapply trf_local_idtyp_mutind; intros; destruct TEc, TEd; subst;
  match goal with  (* H1 must be inverted so that trf_local_idtyp_trm_local can be applied, but also kept so that TI can be applied *)
    H: TransformCondTypeIdentical ?TEc ?TEd |- _ =>
    let H':=fresh H in pose proof H as H'; inversions H'
  end.

  Local Ltac invert_unhidings := (
    repeat match goal with 
    | H: TransformUnhiding ?TEa ?Teb |- _ => inversions H
    | H: TransformObjUnhiding  ?TEa ?Teb |- _ => inversions H
    end ;
    match goal with
    | Hh: ctx_hide ?G0 §& ?G3 = ctx_hide ?G1 §& ?G2,
      Hv: ?G0 §& ?G3 = ?G1 §& ?G2 |- _ =>  rewrite Hv; clears G0 G3
    end
  ).

  - destruct G1.
    invert_unhidings.
    eapply trf_local_idtyp_trm_local; eauto.
    eapply (TI ). eauto.
    constructor; eauto. constructor; eauto. constructor; eauto.
    eapply unhiding with (Go:=objctx_empty); eauto.
  - 
    invert_unhidings.

    forwards: H; eauto. hnf; eauto.
  
    eapply trf_local_idtyp_trm_letl with (L:=L\u ctx_dom G0 \u ctx_dom G2  ); eauto.
    intros. 
    forwards: o x. eauto.
   
    rewrite <- ctx_concat_push.
    eapply unhiding with (Go:=objctx_empty); eauto.
    eapply unhiding with (Go:=objctx_empty); eauto.

  - 
    invert_unhidings.

    eapply trf_local_idtyp_trm_letr with (L:=L\u ctx_dom G0 \u ctx_dom G2 ) (Tl:=Tl); eauto.
    intros.

    forwards: (H x). eauto.
    rewrite <- ctx_concat_push. eauto.
    rewrite <- ctx_concat_push. eauto.
     hnf; eauto.

    rewrite <- ctx_concat_push. eauto.
    eapply unhiding with (Go:=objctx_empty); eauto.
    eapply unhiding with (Go:=objctx_empty); eauto.

  - 
    invert_unhidings.

    eapply trf_local_idtyp_trm_litl with (L:=L\u ctx_dom G0 \u ctx_dom G2  ) (Tl:=Tl ); eauto.


    forwards: (H); eauto. hnf; eauto.

    intros.
    rewrite <- ctx_concat_push.
     
    eapply unhiding with (Go:=objctx_empty); eauto.
    eapply unhiding with (Go:=objctx_empty); eauto.

  - 
    invert_unhidings.

    eapply trf_local_idtyp_trm_litr with (L:=L\u ctx_dom G0 \u ctx_dom G2  ) (Tl:=Tl ); eauto.
    intros.

    forwards: (H x).
 eauto.
    rewrite <- ctx_concat_push. eauto.
    rewrite <- ctx_concat_push. eauto.
    hnf; eauto.

   rewrite <- ctx_concat_push. eauto.
    eapply unhiding with (Go:=objctx_empty); eauto.
    eapply unhiding with (Go:=objctx_empty); eauto.
  - 
    invert_unhidings.

    eapply trf_local_idtyp_lit_fun with (L:=L\u ctx_dom G0 \u ctx_dom G2  ) ; eauto.
    intros.     

    forwards: (H x); eauto. 
    rewrite <- ctx_concat_push. econstructor; eauto.
    rewrite <- ctx_concat_push. econstructor; eauto.
    hnf; eauto.

  - 
    invert_unhidings.

    eapply trf_local_idtyp_lit_obj with (L:=L\u ctx_dom G0 \u ctx_dom G2 ); eauto.
    intros.

    forwards: (H x); eauto. hnf; eauto.

  - 
    invert_unhidings.
    eapply trf_local_idtyp_def_met with (L:=L\u ctx_dom G0 \u ctx_dom G2 ); eauto.
    intros.

    forwards: H xz xc; eauto.
    eapply trf_unhiding_met.
    eapply trf_unhiding_met.

    hnf; eauto.
  - 
    invert_unhidings.

    eapply trf_local_idtyp_defs_one; eauto.
    forwards: (H); eauto. hnf; eauto.

  - 
    invert_unhidings.

    eapply trf_local_idtyp_defs_consl; eauto using seal_defs_hasnt.
    forwards: (H); eauto. hnf; eauto.
    eapply unhiding with (Go0:=Go); eauto.
 - 
    invert_unhidings.

    eapply trf_local_idtyp_defs_consr; eauto using seal_defs_hasnt.
    forwards: (H); eauto. hnf; eauto.
    eapply unhiding with (Go0:=Go); eauto.
Qed.


Global Instance UnhidingCompatTransformationTrm (trf: TermTransform) (TI: TransformUnhidingCompat _ _ trf):
  TransformUnhidingCompat _ _ (trf_local_idtyp_trm trf).
Proof.
  intros.
  hnf. intros.
  eapply (trf_local_unhide_compat_ind TI); eauto.
Qed.
Global Instance UnhidingCompatTransformationDefs (trf: TermTransform) (TI: TransformUnhidingCompat _ _ trf):
  TransformObjUnhidingCompat _ _ (trf_local_idtyp_defs trf).
Proof.
  intros.
  hnf. intros.
  eapply trf_local_unhide_compat_ind; eauto.
Qed.



Lemma trf_subst_ty_trm G e (a: trm) (T: typ) G0 e0 a0 T0 y S:
  〈 G ⁏ e ⊢ a ∶ T 〉 ∼ TransformSubst typing_mode y S ∼> 〈 G0 ⁏ e0 ⊢ a0 ∶ T0 〉 ->
  G ⁏ e ⊢ a ∶ T ->
  G0 ⁏ e0 ⊢ a0 ∶ T0.
Proof.
  introv Htrf Hty.
  inversions Htrf.
  eapply subst_ty_epush; eauto.
Qed.
Lemma trf_subst_ty_lit G e (a: lit) (T: typ) G0 e0 a0 T0 y S:
  〈 G ⁏ e ⊢ a ∶ T 〉 ∼ TransformSubst typing_mode y S ∼> 〈 G0 ⁏ e0 ⊢ a0 ∶ T0 〉 ->
  G ⁏ e ⊢ a ∶ T ->
  G0 ⁏ e0 ⊢ a0 ∶ T0.
Proof.
  introv Htrf Hty.
  inversions Htrf.
  eapply subst_ty_epush; eauto.
Qed.

Lemma trf_subst_ty_defs G Go e (a: defs) (T: typ) G0 Go0 e0 a0 T0 y S:
  〈 G §§ Go ⁏ e ⊢ a ∶ T 〉 ∼ TransformObjSubst typing_mode y S ∼> 〈 G0 §§ Go0 ⁏ e0 ⊢ a0 ∶ T0 〉 ->
  G §§ Go ⁏ e ⊢ a ∶ T ->
  G0 §§ Go0 ⁏ e0 ⊢ a0 ∶ T0.
Proof.
  introv Htrf Hty.
  inversions Htrf.
  eapply substitution_obj; eauto.
Qed.

(*
Lemma trf_subst_ty_def G Go e (a: def) (T: dec) G0 Go0 e0 a0 T0 y S:
  〈 G §§ Go ⁏ e ⊢ a ∶ T 〉 ∼ TransformObjSubst typing_mode y S ∼> 〈 G0 §§ Go0 ⁏ e0 ⊢ a0 ∶ T0 〉 ->
  G §§ Go ⁏ e ⊢ a ∶ T ->
  G0 §§ Go0 ⁏ e0 ⊢ a0 ∶ T0.
Proof.
  introv Htrf Hty.
  inversions Htrf.
  eapply substitution_obj; eauto.
Qed. *)


Global Instance TransformSubstPreservesTermTyped y S: TransformPreserves (TransformCondTermTyped typing_mode) (TransformSubst typing_mode y S).
Proof.
  hnf; intros.
  hnf; intros. hnf in H0.
  destruct TEa, TEb, G, G0.
  eapply trf_subst_ty_trm; eauto.
Qed.

Lemma trf_subst_trm_push (y: var) (S: typ) G0 x vk G2 e0 (tr: trm) (T: typ) x0 Tl:
  x0 \notin (ctx_dom G0 \u \{x} \u ctx_dom G2 \u \{y} ) ->
  TransformSubst typing_mode y S (〈 (G0 § x / vk ~ S) §& G2 ⁏ e0 & x ~ y ⊢ tr ∶ T 〉) (〈 G0 §& subst_ctx x y G2 ⁏ e0 ⊢ (subst_trm x y tr) ∶ T /[ x -> y] 〉) ->
  TransformSubst typing_mode y S (〈 (G0 § x / vk ~ S) §& G2 §ˣ x0 ~ Tl ⁏ e0 & x ~ y ⊢ tr  /[-> x0] ∶ T 〉) (〈 G0 §& subst_ctx x y G2 §ˣ x0 ~ Tl /[ x -> y] ⁏ e0 ⊢ (subst_trm x y tr) /[-> x0] ∶ T /[ x -> y] 〉).
Proof.
  introv Hnx0 Htrf.
  inversions Htrf.
  eapply eq_push_inv in H0. destructecs H0; subst.
  rewrite ? H, ? H3 in *.
  eapply ctx_ok_middle_eq_inv in H. destructecs H; subst.

  rewrite <- subst_open_commut_ne by notin_solve.
  rewrite <- ctx_concat_push.
  rewrite <- ctx_concat_push.
  rewrite <- ctx_subst_push.
  econstructor; eauto.
  -
  ctx_flatten. eauto.
  -
  ctx_flatten. eapply weaken_push. eauto. eapply ctx_ok_concat_map. eapply ctx_ok_push_concat; eauto.
  eauto.
  -
  intros.
  ctx_flatten in H1.
  inversions H1.
  + false. eapply Hnx0; rewrite ! in_union, in_singleton; auto.
  + forwards: H11; fweauto.
    ctx_flatten. eapply ctx_vis_n; eauto.
  -
    eauto.
Qed.

Lemma trf_subst_trm_push_o (y: var) (S: typ) G0 x vk G2 e0 (tr: trm) (T: typ) x0 Tl:
  x0 \notin (ctx_dom G0 \u \{x} \u ctx_dom G2 \u \{y} ) ->
  TransformSubst typing_mode y S (〈 (G0 § x / vk ~ S) §& G2 ⁏ e0 & x ~ y ⊢ tr ∶ T 〉) (〈 G0 §& subst_ctx x y G2 ⁏ e0 ⊢ (subst_trm x y tr) ∶ T /[ x -> y] 〉) ->
  TransformSubst typing_mode y S (〈 (G0 § x / vk ~ S) §& G2 §ᶻ x0 ~ Tl ⁏ e0 & x ~ y ⊢ tr  /[-> x0] ∶ T /[->x0] 〉) (〈 G0 §& subst_ctx x y G2 §ᶻ x0 ~ Tl /[ x -> y] ⁏ e0 ⊢ (subst_trm x y tr) /[-> x0] ∶ T /[ x -> y] /[->x0]〉).
Proof.
  introv Hnx0 Htrf.
  inversions Htrf.
  eapply eq_push_inv in H0. destructecs H0; subst.
  rewrite ? H, ? H3 in *.
  eapply ctx_ok_middle_eq_inv in H. destructecs H; subst.

  rewrite <- subst_open_commut_ne by notin_solve.
  rewrite <- subst_open_commut_ne by notin_solve. (* This is the only difference from trf_subst_trm_push *)
  rewrite <- ctx_concat_push.
  rewrite <- ctx_concat_push.
  rewrite <- ctx_subst_push.
  econstructor; eauto.
  -
  ctx_flatten. eauto.
  -
  ctx_flatten. eapply weaken_push. eauto. eapply ctx_ok_concat_map. eapply ctx_ok_push_concat; eauto.
  eauto.
  -
  intros.
  ctx_flatten in H1.
  inversions H1.
  + false. eapply Hnx0; rewrite ! in_union, in_singleton; auto.
  + forwards: H11; fweauto.
    ctx_flatten. eapply ctx_vis_n; eauto.
  -
    eauto.
Qed.


Lemma ctx_vis_obj_met_inv G0 x G2 Go xz xc TZ TC:
xz <> x ->
xc <> x ->
ctx_ok ((G0) §& G2 §§& Go) ->
ctx_vis ((G0) §& ctx_obj_met typing_mode G2 Go xz TZ xc TC) x ->  ( ctx_vis ((G0 ) §& G2 §§& Go) x).
Proof.
  introv Hnc Hnz Hok Hv.
destruct Go; unfold ctx_obj, ctx_obj_met in *; sympl; destruct ( mode_visibility' typing_mode );
ctx_flatten in *.
-
eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
all: auto.
-
eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
inversions Hv.
all: auto.
-
eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
all: auto.
-
eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
inversions Hv.
all: auto.
-
eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
all: auto.
-
eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
inversions Hv.
all: auto.
Qed.

Lemma ctx_vis_push_n_ok G v v2 vk t:
  ctx_vis G v ->
  ctx_ok (G § v2 / vk ~ t) ->
  ctx_vis (G § v2 / vk ~ t) v.
Proof.
introv Hvis Hok.
eapply ctx_vis_n.
intro; subst.
inversions Hok.
eapply H4. eapply ctx_vis_in_dom; eauto.
auto.
Qed.


Lemma ctx_vis_obj_met_compat G0 x vk S G2 Go y xz xc TZ TC:
xz <> x ->
xc <> x ->
ctx_ok ((G0 § x / vk ~ S) §& ctx_obj_met typing_mode G2 Go xz TZ xc TC) ->
( ctx_vis ((G0 § x / vk ~ S) §& G2 §§& Go) x -> ctx_vis (G0 §& subst_ctx x y G2 §§& Go /[§§ x -> y]) y ) ->
( ctx_vis ((G0 § x / vk ~ S) §& ctx_obj_met typing_mode G2 Go xz TZ xc TC) x -> ctx_vis (G0 §& subst_ctx x y (ctx_obj_met typing_mode G2 Go xz TZ xc TC)) y ).
Proof.
  introv Hnc Hnz Hok Hvis.
  introv Hv.
destruct Go; unfold ctx_obj, ctx_obj_met in *; sympl; destruct ( mode_visibility' typing_mode );
ctx_flatten in *.
-
fresh_chain_init Fr.

fresh_chain_add_fresh Fr Hok. fresh_chain_cleanup Fr.
rewrite ? fresh_chain_eq_merge_var_objctxdom, ? fresh_chain_eq_merge_var_ctxdom, ? fresh_chain_eq_merge_var_ctxdompush in Fr by (apply true).

eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
forwards: Hvis Hv.
eapply ctx_vis_push_n_ok.
eapply ctx_vis_push_n_ok.
auto.
fresh_chain_solve2 Fr.
fresh_chain_solve2 Fr.
all: eauto.
-

eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
inversions Hv.
all: eauto.

-
fresh_chain_init Fr.

fresh_chain_add_fresh Fr Hok. fresh_chain_cleanup Fr.
rewrite ? fresh_chain_eq_merge_var_objctxdom, ? fresh_chain_eq_merge_var_ctxdom, ? fresh_chain_eq_merge_var_ctxdompush in Fr by (apply true).

eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
forwards: Hvis Hv.
eapply ctx_vis_push_n_ok.
eapply ctx_vis_push_n_ok.
inversions H; eauto.
fresh_chain_solve2 Fr.
fresh_chain_solve2 Fr.
all: eauto.
-

eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
inversions Hv.
all: eauto.
-
fresh_chain_init Fr.

fresh_chain_add_fresh Fr Hok. fresh_chain_cleanup Fr.
rewrite ? fresh_chain_eq_merge_var_objctxdom, ? fresh_chain_eq_merge_var_ctxdom, ? fresh_chain_eq_merge_var_ctxdompush in Fr by (apply true).

eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
forwards: Hvis Hv.
eapply ctx_vis_push_n_ok.
eapply ctx_vis_push_n_ok.
auto.
fresh_chain_solve2 Fr.
fresh_chain_solve2 Fr.
all: eauto.
-

eapply ctx_vis_push_n_inv in Hv.
eapply ctx_vis_push_n_inv in Hv.
inversions Hv.
all: eauto.

Qed.

Lemma trf_subst_trm_push_met (y: var) (S: typ) G0 x vk G2 e0 (ta: trm) (T: typ) xc xz TC TZ Go:
  xc \notin (ctx_dom G0 \u \{x} \u ctx_dom G2 \u objctx_dom Go \u \{y} ) ->
  xz \notin (ctx_dom G0 \u \{x} \u ctx_dom G2 \u objctx_dom Go \u \{y} ) ->
  xc <> xz ->
  TransformSubst typing_mode y S (〈 (G0 § x / vk ~ S) §& G2 §§& Go ⁏ e0 & x ~ y ⊢ ta ∶ T 〉) (〈 G0 §& subst_ctx x y G2  §§& Go/[§§ x -> y] ⁏ e0 ⊢ (subst_trm x y ta) ∶ T /[ x -> y] 〉) ->
  TransformSubst typing_mode y S (〈 ctx_obj_met typing_mode ((G0 § x / vk ~ S) §& G2) Go xz TZ xc TC ⁏ e0 & x ~ y ⊢ ta /[# 1 -> xc] /[-> xz]
   ∶ T /[# 1 -> xc] /[-> xz] 〉)
  (〈 ctx_obj_met typing_mode (G0 §& subst_ctx x y G2) Go /[§§ x -> y] xz (subst_typ x y TZ) xc (subst_typ x y TC) ⁏ e0
   ⊢ (subst_trm x y ta) /[# 1 -> xc] /[-> xz] ∶ (subst_typ x y T) /[# 1 -> xc] /[-> xz] 〉).
Proof.
  introv Hnxc Hnxz Hncz Htrf. 
  inversions Htrf.
  eapply eq_push_inv in H0. destructecs H0; subst.
  rewrite ? H, ? H3 in *.
  rewrite ctx_obj_assoc in H.

  lets Hok: H5.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H5.

  eapply fresh_chain_add_fresh_var in Fr. 3: eapply Hnxz. 2: sympl. 2: union_sub_solver.
  eapply fresh_chain_add_fresh_var in Fr. 3: { eapply notin_union. split. eapply Hnxc. eapply notin_singleton. eauto. } 2: sympl. 2: union_sub_solver.


  eapply ctx_ok_middle_eq_inv in H.
  2: fresh_chain_solve2 Fr.
  destructecs H; subst.
  sympls.
  rewrite <- subst_open_commut_ne by notin_solve.
  rewrite <- subst_open_commut_ne by notin_solve.
  rewrite <- subst_open_commut_ne by notin_solve.
  rewrite <- subst_open_commut_ne by notin_solve.
  rewrite ctx_obj_met_assoc.
  rewrite ctx_obj_met_assoc.
  rewrite <- ctx_obj_met_subst_var.

  2: rewrite ctx_obj_met_dom. 2: rewrite ! notin_union, ! notin_singleton; splits; try fresh_chain_solve2 Fr.

  econstructor; eauto.
  -
    rewrite ctx_obj_met_subst_var.

    rewrite <- ctx_concat_empty_r with (G1:=(subst_ctx x y G2)).
    rewrite ctx_obj_met_assoc with (G2:=ctx_empty).
    rewrite ctx_concat_assoc.
    eapply ctx_obj_met_weakens_var'. eauto.

    eapply ctx_obj_met_ok; ctx_simpl_dom; eauto.
    fresh_chain_solve2 Fr.
    fresh_chain_solve2 Fr.
    fresh_chain_solve2 Fr.
    rewrite ctx_obj_met_dom.
    fresh_chain_solve2 Fr.
  -
    intros. 
    eapply ctx_vis_obj_met_compat with (x:=x) (vk:=vk)(S:=S).
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    eapply ctx_obj_met_ok; ctx_simpl_dom; eauto.
    eauto.
    eauto.
Qed.


Lemma ctx_obj_weaken_ty_var G Go e (x: var) T:
  ctx_ok (G §§& Go) ->
  G ⁏ e ⊢ x ∶ T ->
  G §§& Go ⁏ e ⊢ x ∶ T.
Proof.
  introv Hok Hty.
  destruct Go; unfold ctx_obj in *; auto 1.
  eapply weaken_push; try inversions Hok; eauto.
Qed.

Lemma ctx_obj_weaken_ty_trm G Go e (x: trm) T:
  ctx_ok (G §§& Go) ->
  G ⁏ e ⊢ x ∶ T ->
  G §§& Go ⁏ e ⊢ x ∶ T.
Proof.
  introv Hok Hty.
  destruct Go; unfold ctx_obj in *; auto 1.
  eapply weaken_push; try inversions Hok; eauto.
Qed.

Lemma ctx_vis_objlit_inv G x0 T x:
  x <> x0 ->
  ctx_vis (G §§& objctx_lit x0 T) x ->
  ctx_vis G x.
Proof.
  introv Hnx Hvis.
  eapply ctx_vis_objctx_inv in Hvis.
  destruct Hvis.
  auto.
  unfold objctx_dom in H. rewrite in_singleton in H. contradiction.
Qed.


Lemma trf_local_subst_compat_ind (trf: TermTransform)(TI: TransformSubstCompat _ _ _ trf) y S:
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_trm trf) (@TransformSubst _ _ _ _ _ y S) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_lit trf) (@TransformSubst _ _ _ _ _ y S) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_def trf) (@TransformObjSubst _ _ _ _ _ y S) /\
  CondCompatTransformInd (@TransformCondTypeIdentical _ _ _) (trf_local_idtyp_defs trf) (@TransformObjSubst _ _ _ _ _ y S).
Proof.
  eapply trf_local_idtyp_mutind; intros; destruct TEc, TEd; subst;
  match goal with  (* H1 must be inverted so that trf_local_idtyp_trm_local can be applied, but also kept so that TI can be applied *)
    H: TransformCondTypeIdentical ?TEc ?TEd |- _ =>
    let H':=fresh H in pose proof H as H'; inversions H'
  end.

  Local Ltac invert_substs := (
    repeat match goal with 
    | H: TransformSubst _ _ _ ?TEa ?Teb |- _ => inversions H
    | H: TransformObjSubst _ _ _ ?TEa ?Teb |- _ => inversions H
    end;
    match goal with
    | H: ?e & ?xa ~ ?y = ?e & ?xb ~ ?y |- _ =>   eapply eq_push_inv in H; destructecs H; subst
    end;
    try match goal with 
    | H: (?G0 § ?x / ?vk0 ~ ?S) §& ?G3 = (?G1 § ?x / ?vk ~ ?S) §& ?G2 |- _ => eapply ctx_ok_middle_eq_inv in H; [|auto]; destructecs H; subst
    end;
    try match goal with
    | H: (?G1 § ?x / ?vk ~ ?S) §& ?G2 ⁏ ?e0 & ?x ~ ?y ⊢ ?T <: ?S0 |- _ => eapply substitution_epush in H; eauto using subst_sub
    end
  ).

  - 
    destruct G1.
    invert_substs.

    eapply trf_local_idtyp_trm_local; eauto.
    eapply TI; eauto. constructor; eauto. constructor; eauto. constructor; eauto.
  - generalize H0.
 
    invert_substs.
    intros.

    rewrite ? H1, ? H6 in *.
    eapply trf_local_idtyp_trm_letl with (L:=L \u ctx_dom G0 \u \{x} \u ctx_dom G2 \u \{y} ).
    eapply H. econstructor; eauto. econstructor; eauto.
    econstructor; eauto.
    intros.
    eapply trf_subst_ty_trm. 2: eapply (o x0).
    eapply trf_subst_trm_push. notin_solve. econstructor; eauto. auto. eauto.
  - invert_substs.
    rewrite ? H1, ? H6 in *.
    eapply trf_local_idtyp_trm_letr with (L:=L \u ctx_dom G0 \u \{ x} \u ctx_dom G2\u \{y}).
    intros.
    eapply (H x0); eauto.
    eapply trf_subst_trm_push. notin_solve.
    econstructor; eauto.
    eapply trf_subst_trm_push. notin_solve.
    econstructor; eauto.
    econstructor; eauto.
    eapply trf_subst_ty_trm. 2: eapply (o ). econstructor; eauto.  eauto.
  - invert_substs.
    rewrite ? H1, ? H6 in *.
    eapply trf_local_idtyp_trm_litl with (L:=L \u ctx_dom G0 \u \{x} \u ctx_dom G2 \u \{y} ).
    eapply H. econstructor; eauto.
    econstructor; eauto.
    econstructor; eauto.
    intros.
    eapply trf_subst_ty_trm. 2: eapply (o x0).
    eapply trf_subst_trm_push. notin_solve. econstructor; eauto. auto.  eauto.
  - invert_substs.
    rewrite ? H1, ? H6 in *.
    eapply trf_local_idtyp_trm_litr with (L:=L \u ctx_dom G0 \u \{x} \u ctx_dom G2 \u \{y} ) (Tl := Tl /[ x -> y]).
    intros.
    eapply (H x0).
    eauto.
    eapply trf_subst_trm_push. notin_solve.
    econstructor; eauto.
    eapply trf_subst_trm_push. notin_solve.
    econstructor; eauto.
    econstructor; eauto.
    eapply trf_subst_ty_lit. 2: eapply (o ). econstructor; eauto.  eauto.
  - invert_substs.
    rewrite ? H1, ? H6 in *.
    eapply trf_local_idtyp_lit_fun with (L:=L \u ctx_dom G0 \u \{x} \u ctx_dom G2 \u \{y} \u fv U ).
    auto.
    intros.
    eapply (H x0). eauto.
    eapply trf_subst_trm_push_o. eauto.     econstructor; eauto.
    eapply trf_subst_trm_push_o. eauto.     econstructor; eauto.
    econstructor; eauto.
  - invert_substs.
    rewrite ? H1, ? H6 in *.
    rewrite SubstCommon.typ_subst_objlit.
    eapply trf_local_idtyp_lit_obj with (L:=L  \u ctx_dom G0 \u \{x} \u ctx_dom G2 \u \{y} ).
    intros.
    eapply (H x0). eauto.
    +
      sympls.
      rewrite <- subst_open_commut_ne.
      rewrite <- subst_open_commut_ne.
      replace ( objctx_lit x0 T /[ x -> y]) with (( objctx_lit x0 T) /[§§ x -> y]).
      econstructor.
      eauto.
      eauto.
      eauto.
      eauto.
      intros.
      eapply ctx_vis_objlit_inv in H5.
      eapply H14 in H5.
      eapply SubstCommon.ctx_vis_objctx_concat. eauto.
      eauto.
      eauto.
      eauto.    
      eauto.
      eauto.
    +
      sympls.
      rewrite <- subst_open_commut_ne.
      rewrite <- subst_open_commut_ne.
      replace ( objctx_lit x0 T /[ x -> y]) with (( objctx_lit x0 T) /[§§ x -> y]).
      econstructor.
      eauto.
      eauto.
      eauto.
      eauto.
      intros.
      eapply ctx_vis_objlit_inv in H5.
      eapply H14 in H5.
      eapply SubstCommon.ctx_vis_objctx_concat. eauto.
      eauto.
      eauto.
      eauto.    
      eauto.
      eauto.
    + constructor; eauto.
    + eauto.

  - invert_substs.
      rewrite ? H1, ? H7 in *.
      eapply trf_local_idtyp_def_met with (L:=L  \u ctx_dom G0 \u \{x} \u ctx_dom G2 \u objctx_dom Go \u \{y}).
      intros.
      fresh_chain_init Fr.
      fresh_chain_add_fresh Fr H8.
      fresh_chain_cleanup Fr.
      
      eapply (H xz _ xc); eauto.
      +

        eapply trf_subst_trm_push_met; eauto. sympl. rewrite ! ctx_obj_assoc. rewrite ctx_obj_subst. econstructor; eauto.
        rewrite <- ctx_obj_assoc. auto.
        rewrite <- ctx_obj_subst. rewrite <- ctx_obj_assoc. eapply ctx_obj_weaken_ty_var.
        fresh_chain_solve2 Fr.
        eauto.
        fresh_chain_solve2 Fr.
        rewrite <- ctx_obj_assoc.
        rewrite <- ctx_obj_subst.
        rewrite <- ctx_obj_assoc.
        eapply H15.
        fresh_chain_solve2 Fr.
        fresh_chain_solve2 Fr.

    +

    eapply trf_subst_trm_push_met; eauto. sympl. rewrite ! ctx_obj_assoc. rewrite ctx_obj_subst. econstructor; eauto.
    rewrite <- ctx_obj_assoc. auto.
    rewrite <- ctx_obj_subst. rewrite <- ctx_obj_assoc. eapply ctx_obj_weaken_ty_var.
    fresh_chain_solve2 Fr.
     eauto.
    fresh_chain_solve2 Fr.
    rewrite <- ctx_obj_assoc.
    rewrite <- ctx_obj_subst.
    rewrite <- ctx_obj_assoc.
    eapply H15.
    fresh_chain_solve2 Fr.
    fresh_chain_solve2 Fr.
    +
    hnf; eauto.

  - invert_substs.
    rewrite ? H1, ? H7 in *.
    eapply trf_local_idtyp_defs_one.
    eapply (H).
    +
      econstructor. eauto. eauto. eauto. eauto. eauto.
    +
      econstructor. eauto. eauto. eauto. eauto. eauto.
    +
      constructor; eauto.
  - invert_substs.
    rewrite ? H1, ? H7 in *.
    eapply trf_local_idtyp_defs_consl.
    eapply H.
    econstructor. eauto. eauto. eauto. eauto. eauto.
    econstructor. eauto. eauto. eauto. eauto. eauto.

    constructor; eauto.
    sympls.
    eapply substitution_obj. eauto.
    eauto.
    eauto.
    eauto.
    eauto.
    eauto.
    eapply subst_defs_hasnt. rewrite <- subst_label_of_def. auto.
    eapply subst_defs_hasnt. rewrite <- subst_label_of_def. auto.
  - invert_substs.
    rewrite ? H1, ? H7 in *.
    eapply trf_local_idtyp_defs_consr.
    eapply H.
    econstructor. eauto. eauto. eauto. eauto. eauto.
    econstructor. eauto. eauto. eauto. eauto. eauto.
    constructor; eauto.
    sympls.
    eapply substitution_obj. eauto.
    eauto.
    eauto.
    eauto.
    eauto.
    eauto.
    eapply subst_defs_hasnt. rewrite <- subst_label_of_def. auto.
    eapply subst_defs_hasnt. rewrite <- subst_label_of_def. auto.
Unshelve.
  eauto.
Qed.


Global Instance SubstCompatTransformationTrm (trf: TermTransform) (TI: TransformSubstCompat _ _ _ trf):
  TransformSubstCompat _ _ _ (trf_local_idtyp_trm trf).
Proof.
  hnf. intros. hnf. intros.
  eapply trf_local_subst_compat_ind; eauto.
Qed.
Global Instance SubstCompatTransformationDefs (trf: TermTransform) (TI: TransformSubstCompat _ _ _ trf):
  TransformObjSubstCompat _ _ _ (trf_local_idtyp_defs trf).
Proof.
  hnf. intros.  hnf. intros.
  eapply trf_local_subst_compat_ind; eauto.
Qed.

End WithTypingMode.