(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation Weakening EWeakening Substitution TFree.
Require Import TransformationSimilarity TransformationPreservation TransformationTerm.
Require Import TransformationConfig OperationalSemantics TypedOperationalSemantics.
Require Import ContextSimilarity ConfigSimilarity.
Require Import ConfigTyping ConfigTypingAccessories TypedProgress.
Require Import TypedPreservation.
Require Import TransformationIterateSim.
Require Import Similarity TypingSimilarity.

Require Import TransformationLifting.
Require Import FreshChain FreshChainTactics.
Require Import TypedOperationalSemantics InertContext.
Require Import Similarity SyntaxSimilarity.
Require Import ExecutionTactics Renaming TypingHelpers.

Require Import NothingFresh TypedPreservation LetLet Defs.
Require Import HeapCorrespondence ConfigVarkinds.
Require Import HeapUpdate ConfigValid.

Require Import TransformationSimilaritySteps ConfigTypingSteps TransformationConfigSteps.
Require Import Defs LetLet EnvCorrespondence.
Require Import Subenvironments Narrowing.
Require Import Defree.
Require Import Visibility ConfigNotations.
Require Import TypedPreservationLemmas TypedReductionLemmas.
Require Import TransformationTyping TransformationLiftingPreservationInstantiate TransformationLiftingPreservationFunction.
Require Import TransformationSimilarityPreservation.
Require Import TypedReductionEquivalence.
Require Import Lia.

(** * Preservation of lifted transformations during reduction. *)

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma transformation_lifted_focus_or_preserved_focus_let_1 (TL : TermTransform)
  G tla tr fs S e Lva F' C3 tlb Lvb F'0 C4 T Tt Tl L:
  〈 G ⁏ e ⊢ tla ∶ Tl 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G ⁏ e ⊢ tlb ∶ Tl 〉 ->
  G ⊢ trm_let tla tr ; fs ; S ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_let tlb tr ; fs ; S ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  heap_correspond G e S ->
  ty_stack G e Tt fs T ->
  (forall x : var, x \notin L -> G §ˣ x ~ Tl ⁏ e ⊢ tr /[-> x] ∶ Tt) ->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htrf Hpr1 Hpr2 Hhc Htys Htyt.
  (* Determine config and context after step *)
  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  inversions Hpr1; inversions H8.
  inversions Hpr2; inversions H8.

  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  forwards Hac3: ty_config_to_accessories HtyC3.
  forwards Hac4: ty_config_to_accessories HtyC4.

  eapply trf_lift_sim_conf_lift.
  eauto.

  eapply trf_red_trm_let1; eauto.
Qed.

Lemma transformation_lifted_focus_or_preserved_focus_let_2 (TL : TermTransform)
  G tl tra fs S e Lva F' C3 trb Lvb F'0 C4 T Tt Tl L:
  G ⁏ e ⊢ tl ∶ Tl ->
  G ⊢ trm_let tl tra ; fs ; S ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_let tl trb ; fs ; S ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  heap_correspond G e S ->
  ty_stack G e Tt fs T ->
  (forall x : var, x \notin L -> 〈 G §ˣ x ~ Tl ⁏ e ⊢ tra /[-> x] ∶ Tt 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G §ˣ x ~ Tl ⁏ e ⊢ trb /[-> x] ∶ Tt 〉) ->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt  Hpr1 Hpr2 Hhc Htys Htrf.
  (* Determine config and context after step *)
  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2. 
  inversions Hpr1; inversions H8.
  inversions Hpr2; inversions H8.

  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.


  forwards Hac3: ty_config_to_accessories HtyC3.
  forwards Hac4: ty_config_to_accessories HtyC4.


  eapply trf_lift_sim_conf_lift.
  eauto.

  eapply trf_lift_idtyp_conf_stack; eauto.
  eapply trf_lift_idtyp_stack_e; eauto.
Qed.

Lemma transformation_lifted_focus_or_preserved_focus_lit_1 (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
  G tla tr fs S e Lva F' C3 tlb Lvb F'0 C4 T Tt Tl L:
  〈 G ⁏ e ⊢ tla ∶ Tl 〉 ∼ trf_local_idtyp_lit typing_mode TL ∼> 〈 G ⁏ e ⊢ tlb ∶ Tl 〉->
  G ⊢ trm_lit tla tr ; fs ; S ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_lit tlb tr ; fs ; S ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  heap_correspond G e S ->
  ty_stack G e Tt fs T ->
  (forall x : var, x \notin L -> G §ˣ x ~ Tl ⁏ e ⊢ tr /[-> x] ∶ Tt) ->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htrf Hpr1 Hpr2 Hhc Htys Htyt.
  (* Determine config and context after step *)
  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  inversions Hpr1; inversions H8;
  inversions Hpr2; inversions H8.

  * (* Obj *)
    (* The transformed literal is an object.
       The transformation is in one of the defs and it will be the same on the heap *)
    eapply transformation_lifted_focus_or_preserved_focus_lit_1obj; eauto.
  
  * (* mismatch *) inversions Htrf.
  * (* mismatch *) inversions Htrf.
  * (* Fun *)
    eapply transformation_lifted_focus_or_preserved_focus_lit_1fun. 12: eapply Htyt. all: eauto.
Qed.



Lemma transformation_lifted_focus_or_preserved_focus_lit_2 (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
  G tl tra fs S e Lva F' C3 trb Lvb F'0 C4 T Tt Tl L:
  G ⁏ e ⊢ tl ∶ Tl ->
  G ⊢ trm_lit tl tra ; fs ; S ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_lit tl trb ; fs ; S ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  heap_correspond G e S ->
  ty_stack G e Tt fs T ->
  (forall x : var, x \notin L -> 〈 G §ˣ x ~ Tl ⁏ e ⊢ tra /[-> x] ∶ Tt 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G §ˣ x ~ Tl ⁏ e ⊢ trb /[-> x] ∶ Tt 〉) ->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt  Hpr1 Hpr2 Hhc Htys Htrf.
 (* Determine config and context after step *)
  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  inversions Hpr1; inversions H8;
  inversions Hpr2; inversions H8.
  * (* Obj *)
  eapply transformation_lifted_focus_or_preserved_focus_lit_2obj; eauto.
  * (* Fun *)
  eapply transformation_lifted_focus_or_preserved_focus_lit_2fun; eauto.
Qed.

Lemma trf_lift_obj_binds_fld_eq (TL : TermTransform) G e Go  yda ydb f ya yb U:
  defs_has yda (def_fld f (avar_f ya)) ->
  defs_has ydb (def_fld f (avar_f yb)) ->
  〈 G §§ Go ⁏ e ⊢ yda ∶ U  〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈 G §§ Go⁏ e ⊢ ydb ∶ U  〉 ->
  ya = yb.
Proof.
  introv Hhasa Hhasb Htrf.
  remember (〈 G §§ Go ⁏ e ⊢ yda ∶ U  〉 ).
  remember (〈  G §§ Go ⁏ e ⊢ ydb ∶ U 〉 ).
  reverts yda ydb U.
  induction Htrf; intros; subst; inversions Heqt; inversions Heqt0.
  - inversions H. inversions Hhasa. cases_if.
  - eapply defs_has_cons_inv in Hhasa.
    eapply defs_has_cons_inv in Hhasb.
    destructecs_all.
    + subst. inversions Hhasb. auto.
    + subst. sympls. 
    forwards: defs_has_hasnt_neq (def_fld f (avar_f ya)) (def_fld f (avar_f ya)). eauto. eapply H1. contradiction.
    +
    subst. sympls. 
    forwards: defs_has_hasnt_neq (def_fld f (avar_f yb)) (def_fld f (avar_f yb)). eauto. eapply H2. contradiction.
    + eauto.
  -
    eapply defs_has_cons_inv in Hhasa.
    eapply defs_has_cons_inv in Hhasb.
    destructecs_all.
    + subst. inversions H.
    + subst. sympls. 
    forwards: defs_has_hasnt_neq (def_fld f (avar_f ya)) (def_fld f (avar_f ya)). eauto. eapply H2. contradiction.
    +
    subst. sympls. 
    forwards: defs_has_hasnt_neq (def_fld f (avar_f yb)) (def_fld f (avar_f yb)). eauto. eapply H1. contradiction.
    +
    forwards: defs_has_inv Hhasa Hhasb. inversions H3. auto. 
Qed.

Lemma trf_lift_heap_binds_fld_eq (TL : TermTransform) G e Sa Sb y yda ydb f ya yb:
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  binds y (item_obj yda) Sa ->
  defs_has yda (def_fld f (avar_f ya)) ->
  binds y (item_obj ydb) Sb ->
  defs_has ydb (def_fld f (avar_f yb)) ->
  ya = yb.
Proof.
  introv Htrf Hbia Hhasa Hbib Hhasb.
  remember (〈 e ⊢ Sa ∶ G 〉 ).
  remember (〈 e ⊢ Sb ∶ G 〉 ).
  reverts Sa Sb.
  induction Htrf; intros; subst; inversions Heqt; inversions Heqt0.
  -

    eapply binds_push_inv in Hbia. destructecs Hbia.
    + subst. eapply binds_push_eq_inv in Hbib.
      inversions H0.
      inversions H5.


      eapply trf_lift_obj_binds_fld_eq; eauto.
   +
     eapply binds_push_neq_inv in Hbib. binds_eq.
      forwards: defs_has_inv Hhasa Hhasb. inversions H1. auto. auto.
  -

    eapply  binds_push_inv in Hbia. destructecs Hbia.
    + subst. 
      eapply binds_push_eq_inv in Hbib. 
      inversions Hbib.
      forwards: defs_has_inv Hhasa Hhasb. inversions H1. auto.
    +
      eapply binds_push_neq_inv in Hbib.
      eauto.
      auto.
Qed.


Lemma tred_read_rename G f Tw w0 fs S e Lva T wa wb y:
  wb \notin (fv (trm_read (avar_f w0) f) \u fv fs \u fv_heap S \u fv_renv e \u Lva)  ->
  wb \notin ctx_fv G \u fv Tw -> 
  (
    G ⊢  trm_read (avar_f w0) f ; fs ; S ; e ; Lva ∶ T
    ↦ 
    G §ʷ wa ~ Tw ⊢ trm_var (avar_f wa) ; fs ; S ; e & wa ~ y ; fv (trm_read (avar_f w0) f) \u fv fs \u fv_heap S \u fv_renv e \u Lva
  ) ->  (
    G ⊢ trm_read (avar_f w0) f ; fs ; S ; e ; Lva ∶ T
    ↦ 
    (G §ʷ wb ~ Tw) ⊢ trm_var (avar_f wb) ; fs ; S ; e & wb ~ y ; fv (trm_read (avar_f w0) f) \u fv fs \u fv_heap S \u fv_renv e \u Lva 
  ) .
Proof.
  introv Hnwb HnwbG  Htr.
  inversions Htr; try discriminate.
  inversions H6.

  forwards HeqS: eq_push_inv H9. destructecs HeqS.
  subst.
  eapply tred_read_obj; eauto.

  forwards HeqS: eq_push_inv H9. destructecs HeqS.
  subst.
  eapply tred_read_fun; eauto.

Qed.



Lemma trf_similarity_config_read G e (w0: var) f fs S wa wb y Lva T Tw:
  ok e ->
  config_valid ( trm_read (avar_f w0) f ; fs ; S ; e ; Lva)  ->
  wa \notin (fv (trm_read (avar_f w0) f) \u fv fs \u fv_heap S \u fv_renv e \u Lva) ->
  wb \notin (fv (trm_read (avar_f w0) f) \u fv fs \u fv_heap S \u fv_renv e \u Lva)  ->
  wb \notin ctx_fv G \u fv Tw ->
  y \in dom S ->
  (
    G ⊢  trm_read (avar_f w0) f ; fs ; S ; e ; Lva ∶ T
    ↦ 
    G §ʷ wa ~ Tw ⊢ trm_var (avar_f wa) ; fs ; S ; e & wa ~ y ; fv (trm_read (avar_f w0) f) \u fv fs \u fv_heap S \u fv_renv e \u Lva
  ) ->
  〈 (G §ʷ wa ~ Tw) ⊢ trm_var (avar_f wa) ; fs ; S ; e & wa ~ y ; fv (trm_read (avar_f w0) f) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T〉
  ∼ trf_similarity typing_mode ∼>
  〈 (G §ʷ wb ~ Tw) ⊢ trm_var (avar_f wb) ; fs ; S ; e & wb ~ y ; fv (trm_read (avar_f w0) f) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T〉.
Proof.
  introv Hoke Hva Hnwa Hnwb HnwbG HyS Htr.
  forwards: tred_read_rename Htr; fweauto.
  constructor.
  (* Solve typing conditions *)
  eapply (typed_preservation Htr); eauto.
  eapply (typed_preservation H); eauto.
  (* Base the similarity on the previous configuration *)
  forwards Hsim: similar_conf_1to1_refl_vars Hva.
  destructecs Hsim.
  hnf in Hsim1.
  destructecs Hsim1.
  forwards Hsimlit: similar_trm_read_case Hsim2.
  destructecs Hsimlit. inversions Hsimlit0.

  hnf.

  exists (wa::xs)%list (wb::xs)%list; splits.
  -
    eapply vars_conf_unique_t_Lv_subset.
    eapply vars_conf_unique_Lv_subset.
    eapply vars_conf_unique_prepend1_concat_e. eauto.
    sympl. eauto.
    simpl_dom. union_in_solver.
    union_sub_solver.
    union_sub_solver.
  -
    eapply vars_conf_unique_t_Lv_subset.
    eapply vars_conf_unique_Lv_subset.
    eapply vars_conf_unique_prepend1_concat_e. eauto.
    sympl. eauto.
    simpl_dom. union_in_solver.
    union_sub_solver.
    union_sub_solver.
  - simpl. auto.
  -
    hnf; splits.
    +
      eapply similar_trm_var.
      eapply similar_avar_f.
      eapply similar_var_push. auto.
    +
      eapply similar_extend.
      auto.
    + (* heap *)

      eapply similar_extend_notin. eauto. eauto. eauto.
    + 
      eapply similar_env_push.
        eauto.
      eapply similar_extend.
      eapply similar_refl_by_fv. unfold fv, VarFreeVar. eapply subset_singleton. rewrite Hsim. union_in_solver.
      rewrite Hsim. eauto.
      rewrite Hsim. eauto.
Qed.


Lemma typing_viewpoint_adapt_loc G e (y: var) w T U S:
  G ⁏ e ⊢ w ▷ T ⟶ U ->
  G ⁏ e ⊢ y ∶ T ->
  G ⁏ e ⊢ y ∶ S ->
  (inert_typ S v_loc) ->
  G ⁏ e ⊢ y ∶ U.
Proof.
  introv Hva HtyT HtyS HinS.
  hnf in Hva.
  destruct (mode_viewpoint_adaptation' typing_mode ) eqn: Hm.
  subst. auto.
  inversions Hva.
  eapply ty_var_and.
  eapply ty_var_sub.
  eapply HtyT.
  eapply Incap.ty_incap_subtyp; eauto.
  inversions HinS.
  typing_mode_contradiction typing_mode.
  hnf in H2.
  destruct (mode_refcaps' typing_mode) eqn: Hrc.   typing_mode_contradiction typing_mode.
  inversions H2.
  eapply ty_var_sub.
  eapply HtyS.
  eapply subtyp_trans.
  eapply subtyp_and12.
  eapply subtyp_typ. eauto. eauto.
Qed.


Lemma ty_nothing_fresh_in_config_push1_Lv_T (G:ctx) w T (U:typ) (t: trm) fs Sigma1 e Lv t2 fs2 Lv2 y vk:
   ty_nothing_fresh_in_config G U (t ; fs ; Sigma1 ; e ; Lv) ->
   fv_config (t ; fs ; Sigma1 ; e ; Lv) \c Lv2  ->
   fv T \c Lv2 ->
   ty_nothing_fresh_in_config (G § w /vk ~ T ) U (t2; fs2 ; Sigma1 ; e & (w ~ y); Lv2).
Proof.
  introv Hfr Hsuc HsuT.

  unfold nothing_fresh_in_config in *.
  destructc Hfr as [? ?] ?.
  repeat split;  ctx_simpl_dom; simpl_dom; change (@fv ctx CtxFreeVar) with (@ctx_fv); ctx_simpl_fv; rewrite ? env_push_fv.
  change (@fv var _ ?a) with \{a}.
  union_sub_solver_destruct_ls.
  union_sub_solver.

  eapply subset_trans. eapply H.
  eapply subset_trans. eapply Hsuc.
  union_sub_solver.
  union_sub_solver_destruct_ls.
  union_sub_solver.

  eapply subset_trans. eapply H0.
  eapply subset_trans. eapply Hsuc.
  union_sub_solver.
  union_sub_solver_destruct_ls.
  eapply subset_trans. eapply H1.
  eapply subset_trans. eapply Hsuc.
  union_sub_solver.
Qed.

Lemma transformation_lifted_focus_or_preserved_heap_read (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
  G a f fs Sa Sb e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢ trm_read a f ∶ Tt->
  G ⊢ trm_read a f  ; fs ; Sa ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_read a f  ; fs ; Sb ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_read a f ; fs ; Sa ; e ; Lva) T ->
  ty_config G (trm_read a f ; fs ; Sb ; e ; Lvb) T ->
  ty_stack G e Tt fs T ->
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Htys Htrf.

  (* Basic assertions about typing *)
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.

  forwards Hhca: ty_config_to_heap_accessories Htyc1.
  forwards Hhcb: ty_config_to_heap_accessories Htyc2.

  step_invert_simple Hpr1 HtyC3;
  step_invert_simple Hpr2 HtyC4.


  (* Pick fresh variable for the new reference *)
  pick_fresh_with  z (fv fs).

  (* Information about the field *)
  forwards Htyti: ty_trm_read_inv' Htyt.
  destructecs Htyti.
  inversions Htyc1.
  forwards Hcf: CanonicalForms.canonical_forms_fld Htyti0; fweauto.
  destructecs Hcf.
  binds_eq.
  binds_eq.
  binds_eq.
  forwards Hheq: defs_has_inv Hhas Hcf2. inversions Hheq.
  rename vVal into y.

  (* Unify the location *)

  forwards: trf_lift_heap_binds_fld_eq Htrf; fweauto. subst.
  clear H0.
  rename w0 into w, y0 into y.

  (* Use similarity to use the chosen reference *)

  destructecs Hfrw2.
  destructecs Hfrw0.

  assert (y \in dom Sa) as HySa. {
    hnf in Hhca; destructecs Hhca.  
    eapply ctx_k_in_kdom in Hcf5.
    rewrite Hhca4 in Hcf5. auto. 
  }
  assert (y \in dom Sb) as HySb. {
    hnf in Hhcb; destructecs Hhcb.  
    eapply ctx_k_in_kdom in Hcf5.
    rewrite Hhcb4 in Hcf5. auto. 
  }

  rename w2 into wa, T2 into Twa.
  rename w3 into wb, T0 into Twb.

  assert (
    〈 G §ʷ wa ~ Twa ⊢ trm_var (avar_f wa) ; fs ; Sa ; e & wa ~ y ; \{ w} \u fv fs \u fv_heap Sa \u fv_renv e \u Lva ∶ T 〉
    ∼ trf_similarity typing_mode ∼>
    〈 G §ʷ z ~ Twa ⊢ trm_var (avar_f z) ; fs ; Sa ; e & z ~ y ; \{ w} \u fv fs \u fv_heap Sa \u fv_renv e \u Lva ∶ T 〉
  ) as Hsia. {

    forwards Hsi: trf_similarity_config_read w wa z Hpr1.

    hnf in Hhca; destructecs Hhca.  eauto.
    eapply config_typed_valid; eauto. sympl; eauto. sympl; eauto. sympl; eauto.
    auto. sympl in Hsi.  auto.
  }

  assert (
    〈 G §ʷ z ~ Twb ⊢ trm_var (avar_f z) ; fs ; Sb ; e & z ~ y ; \{ w} \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb ∶ T 〉
    ∼ trf_similarity typing_mode ∼>
    〈 G §ʷ wb ~ Twb ⊢ trm_var (avar_f wb) ; fs ; Sb ; e & wb ~ y ; \{ w} \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb ∶ T 〉
  ) as Hsib. {

    forwards Hsi: trf_similarity_config_read w wb z Hpr2.

    hnf in Hhcb; destructecs Hhcb.  eauto.
    eapply config_typed_valid; eauto. sympl; eauto. sympl; eauto. sympl; eauto.
    auto. sympl in Hsi.
    eapply TransformSimilaritySymmetric in Hsi.
    auto.
  }
  (* Find common type for the variable *)

  forwards Hyeq: ctx_binds_k_after_eq Hby2G Hby2G0.
  destructecs Hyeq; subst.
  rename T7 into Ty.
  (*

  forwards Hky: ctx_k_in_dom Hcf5.
  forwards HbiyG: ctx_dom_to_binds Hky.
  eauto.
  eauto.
  destructecs HbiyG. *)

  assert (inert_typ Ty v_loc) as HinlTy. {
      forwards: inert_binds_k_inert Hby2G0. eauto. auto.
  }

  assert (inert_typ Ty v_ref) as HinTy. {
    eapply inert_typ_loc_ref; eauto.
  }
  assert (typ_corr Ty Ty) as HcorTy. {
    eapply inert_typ_corr_loc_refl; eauto.
  }
  assert (inert (G §ʷ z ~ Ty)) as HinGz. {
    econstructor; eauto.
  }
  assert (heap_correspond (G §ʷ z ~ Ty) (e & z ~ y) Sa) as Hhcz. {
    eapply heap_correspond_epush; eauto.  
  }
  assert (heap_correspond (G §ʷ z ~ Ty) (e & z ~ y) Sb) as Hhczb. {
    inversions Htyc2.
    eapply heap_correspond_epush; eauto.  
  }


  assert (renv_corr (G §ʷ z ~ Ty) (e & z ~ y)) as Hcorrz. {
    hnf in Hhcz. destructecs Hhcz. auto.
  }

  assert ((G §ʷ z ~ Ty)  ⁏ (e & z ~ y) ⊢ z ∶ T9) as Htyz. {
    eapply  ReferenceTypes.ty_renv_binds_rev_sametype_var with (v2:=y) (U:=Ty).
    eauto.
    eapply ctx_binds_e.
    eapply ctx_binds_n with (v:=y). eauto.
    eapply ctx_binds_k_after_binds; eauto.
    eapply binds_push_eq.
    eauto.
    auto.
      eapply weaken_push. 
      eapply ewkn_push.
    eauto.
    eauto.
    eauto.
    eauto.
    eauto.
  } 
  assert ((G §ʷ z ~ Ty)  ⁏ (e & z ~ y) ⊢ z ∶ Tt) as HtyzTt. {
    eapply ty_var_sub with (T:=U).
    eapply typing_viewpoint_adapt_loc with (w:=w) (T:=T9) (S:=Ty).
    eapply ewkn_push; eauto.
    eapply weaken_push; eauto.
    eauto.
    eauto.
    auto.
    eapply ewkn_push; eauto.
    eapply weaken_push; eauto.
  }



  (* Re-type the configurations with this type *)

  set (Lv2:=( \{ w} \u fv fs \u fv_heap Sa \u fv_renv e \u Lva ) \u ( \{ w} \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb)).

  (* Similarity with the common type in the context - left side *)
  assert ( 
    〈 G §ʷ z ~ Twa ⊢ trm_var (avar_f z) ; fs ; Sa ; e & z ~ y ; \{ w} \u fv fs \u fv_heap Sa \u fv_renv e \u Lva ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
    〈 G §ʷ z ~ Ty ⊢ trm_var (avar_f z) ; fs ; Sa ; e & z ~ y ; Lv2 ∶ T 〉
  ) as Hsiat. {
    eapply trf_similarity_retype_Lv.

    *
      inversions Hsia; eauto.
    *
      unfold Lv2. sympl.

      inversions Hsia.
      inversions H18.


      econstructor.
      + auto.
      +  auto.
      + eapply ty_stack_epush. eauto. eauto. eapply ty_stack_push. eauto. eauto. eapply Htys. 
      + apply all_tfree_k_Gpush; eauto.
      + eapply ty_trm_var; eauto.
      + eapply all_tfree_k_trm_avar. eapply ctx_k_push_e.
      + eapply ty_nothing_fresh_in_config_push1_Lv_T; eauto. sympl; union_sub_solver. 

        eapply subset_trans with (b:=ctx_fv G).
        eapply ctx_binds_fv_subset. eapply ctx_binds_k_after_binds; eauto.
        destruct H12.
        destruct H0. eapply subset_trans. eapply H3.  sympl; union_sub_solver. 

      + hnf. exists w. union_in_solver. 
      + eapply ctx_sealed_push. auto. eapply ctx_closed_binds. eapply H15. eapply ctx_binds_k_after_binds; eauto.
      +   eapply closed_read; eauto.
      + auto.
    * unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsiac: TransformSimilarityTransitive Hsia Hsiat. clear Hsia Hsiat.

  assert ( 
    〈 G §ʷ z ~ Ty ⊢ trm_var (avar_f z) ; fs ; Sb ; e & z ~ y ; Lv2 ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
    〈 G §ʷ z ~ Twb ⊢ trm_var (avar_f z) ; fs ; Sb ; e & z ~ y ; \{ w} \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb ∶ T 〉
  ) as Hsibt. {
    eapply TransformSimilaritySymmetric.
    eapply trf_similarity_retype_Lv.
    *
      inversions Hsib; eauto.
    *
      unfold Lv2. sympl.

      inversions Hsib.
      inversions H18.


      econstructor.
      + auto.
      +  auto.
      + eapply ty_stack_epush. eauto. eauto. eapply ty_stack_push. eauto. eauto. eapply Htys.
      + apply all_tfree_k_Gpush; eauto.
      + eapply ty_trm_var; eauto.
      + eapply all_tfree_k_trm_avar. eapply ctx_k_push_e.
      + inversions Htyc2. eapply ty_nothing_fresh_in_config_push1_Lv_T; eauto. sympl; union_sub_solver. 

        eapply subset_trans with (b:=ctx_fv G).
        eapply ctx_binds_fv_subset. eapply ctx_binds_k_after_binds; eauto.
        destruct H37.
        destruct H0. eapply subset_trans. eapply H3.  sympl; union_sub_solver. 

      + hnf. exists w. union_in_solver.
      + eapply ctx_sealed_push. auto. eapply ctx_closed_binds. eapply H15. eapply ctx_binds_k_after_binds; eauto.
      +  inversions Htyc2. eapply closed_read; eauto.
      + auto.
    * unfold Lv2. sympl. union_sub_solver.
  }
  forwards Hsibc: TransformSimilarityTransitive Hsibt Hsib. clear Hsib Hsibt.


  eapply trf_lift_sim_conf_sim_lift_sim.
  eapply Hsiac. 2: eapply Hsibc.


  eapply trf_lift_idtyp_conf_heap.

  5: {
    eapply trf_lift_idtyp_heap_epush; eauto.
  }

  - inversions Hsiac. eapply ty_config_to_accessories; eauto.
  - inversions Hsibc. eapply ty_config_to_accessories; eauto.

  - inversions Hsiac. inversions H18. eapply heap_correspond_to_accessories; eauto.

  - inversions Hsibc. inversions H18. eapply heap_correspond_to_accessories; eauto.
  -  eapply ty_stack_epush. eauto. eauto. eapply ty_stack_push. eauto. eauto. eapply Htys.
  - eauto.

Qed.


Lemma transformation_lifted_focus_or_preserved_heap_apply (TL : TermTransform)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
    G a a0 fs Sa Sb e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢ trm_apply a a0 ∶ Tt ->
  G ⊢ trm_apply a a0 ; fs ; Sa ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_apply a a0 ; fs ; Sb ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_apply a a0 ; fs ; Sa ; e ; Lva) T ->
  ty_config G (trm_apply a a0 ; fs ; Sb ; e ; Lvb) T->
  ty_stack G e Tt fs T ->
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Htys Htrf.
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.
  forwards Hhca: ty_config_to_heap_accessories Htyc1.
  forwards Hhcb: ty_config_to_heap_accessories Htyc2.

  forwards Htyca1: ty_config_to_accessories Htyc1.
  forwards Htyca2: ty_config_to_accessories Htyc2.


  step_invert_simple Hpr1 HtyC3.
  step_invert_simple Hpr2 HtyC4.

  forwards Hbi: binds_functional Hbxye Hbxye0. subst.

  (* Cases on whether the selected item is transformed *)
  forwards: trf_lift_idtyp_heap_binds_met_cases Htrf.
  eapply HbyS.
  eapply HbyS0.

  destruct H.
  * (* Not the transformed function - keep the same *)
    subst.
    eapply trf_lift_sim_conf_lift. eauto.

    eapply trf_lift_idtyp_conf_heap; eauto.
    eapply ty_config_accessories_tred_apply'; eauto. inversions Htyc1; auto.
    eapply ty_config_accessories_tred_apply'; eauto. inversions Htyc2; auto.

    forwards Htyf: ty_trm_apply_inv Htyt. destructecs Htyf.

    forwards Hcf: CanonicalForms.canonical_forms_fun Htyf0; fweauto. inversions Htyc; auto. inversions Htyc; eauto.  inversions Htyc; eauto. destructecs Hcf.
    forwards: typing_apply_fun Htyf1; fweauto.
    inversions Htyc1. eauto.
    binds_eq.
    binds_eq.
    eauto.

  * (* The transformed item - copy to focus *)
    destructecs H.

    (* Prepare in advance, so parts of it can be used for the intermediate config *)
    assert (
      ty_config_accessories F'0 e fs Sa t /[-> z0] T (fv (trm_apply (avar_f x0) (avar_f z0)) \u fv fs \u fv_heap Sa \u fv_renv e \u Lva)
    ) as Hhcaa. {   eapply ty_config_accessories_tred_apply'; eauto. inversions Htyc1; auto. }
    assert (
      ty_config_accessories F'0 e fs Sb t /[-> z0] T (fv (trm_apply (avar_f x0) (avar_f z0)) \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb)
    ) as Hhcab. {
      hnf in Hhcaa; destructecs Hhcaa.
      hnf in Htyca1; destructecs Htyca1.
      hnf in Htyca2; destructecs Htyca2.
      hnf; splits; eauto.
      - eapply ty_nothing_fresh_in_config_p; eauto.
      - eapply null_p; eauto.
      - eapply TypedPreservationLemmas.closed_apply'; eauto.
    }

    eapply trf_lift_sim_conf_lift_lift with (Gc:=F'0) (Tc:=T) (Cc:=t /[-> z0] ; fs ; Sb ; e ; fv (trm_apply (avar_f x0) (avar_f z0)) \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb).
    + eauto.
    + (* Heap transform *)
      eapply trf_lift_idtyp_conf_heap; eauto. 
      forwards Htyf: ty_trm_apply_inv Htyt. destructecs Htyf.
      forwards Hcf: CanonicalForms.canonical_forms_fun Htyf0; fweauto. inversions Htyc; auto. inversions Htyc; eauto.  inversions Htyc; eauto. destructecs Hcf.
      forwards: typing_apply_fun Htyf1; fweauto.
      inversions Htyc1. eauto.
      binds_eq.
      binds_eq.
      eauto.
    + (* Focus transform *)

      forwards Htyf: ty_trm_apply_inv Htyt. destructecs Htyf.
      forwards Hcf: CanonicalForms.canonical_forms_fun Htyf0; fweauto. inversions Htyc; auto. inversions Htyc; eauto.  inversions Htyc; eauto. destructecs Hcf.


      inversions Htyc2.

      binds_eq.
      binds_eq.
      binds_eq.
      ctx_binds_eq.

      eapply ty_trm_apply_inv in Htyt.
      destructecs Htyt.

      assert ( F'0 ⁏ e ⊢ U' /[-> z0] <: T2 /[-> z0]). {
        pick_fresh zz.
        forwards: Hcf zz. eauto.
        rewrite subst_intro with (x:=zz).
        rewrite subst_intro with (x:=zz) (y0:=z0).
        eapply subst_end. eapply subst_sub. eapply ewkn_sub. eapply H3. eauto. eauto.  eauto.  eauto.  rewrite subst_fresh. eauto.
        eauto.
        eauto.
        eauto.
        eauto.
      }

      eapply trf_lift_idtyp_conf_focus.
      econstructor.
      eauto.

      eapply ty_config_accessories_tred_apply'; eauto.
      eauto.
      eapply Htys.
      eapply trf_local_idtyp_trm_subst_wz with (vk:=v_par) (L:=L) (Tt := T'). eauto. eauto.
      eauto.
      eauto.
      eapply ty_trm_var. eapply ty_trm_apply_inv_vis2. eauto. eapply ty_var_sub. eauto. eapply Hcf4. eapply H.
      eapply subtyp_trans.
      eapply H3. eapply Htyf.
Qed.

Lemma transformation_lifted_focus_or_preserved_heap_let (TL : TermTransform)
    G t1 t2 fs Sa Sb e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢ trm_let t1 t2 ∶ Tt ->
  G ⊢ trm_let t1 t2 ; fs ; Sa ; e ; Lva ∶ T ↦ F' ⊢ C3->
  G ⊢ trm_let t1 t2 ; fs ; Sb ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_let t1 t2 ; fs ; Sa ; e ; Lva) T->
  ty_config G (trm_let t1 t2 ; fs ; Sb ; e ; Lvb) T->
  ty_stack G e Tt fs T ->
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Htys Htrf.
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.
  forwards Hhca: ty_config_to_heap_accessories Htyc1.
  forwards Hhcb: ty_config_to_heap_accessories Htyc2.

  forwards Htyca1: ty_config_to_accessories Htyc1.
  forwards Htyca2: ty_config_to_accessories Htyc2.

  step_invert_simple Hpr1 HtyC3.
  step_invert_simple Hpr2 HtyC4.
  eapply trf_lift_sim_conf_lift. eauto.

  inversions HtyC3.

  eapply trf_lift_idtyp_conf_heap; eauto.

  eapply ty_config_accessories_tred_push; eauto.
  eapply ty_config_accessories_tred_push; eauto.
Qed.


Lemma transformation_lifted_focus_or_preserved_heap_lit (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
   G l t fs Sa Sb e Lva F' C3 Lvb F'0 C4 T Tt:
   G ⁏ e ⊢ trm_lit l t ∶ Tt ->
   G ⊢ trm_lit l t ; fs ; Sa ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
   G ⊢ trm_lit l t ; fs ; Sb ; e ; Lvb ∶ T ↦ F'0 ⊢ C4->
   ty_config G (trm_lit l t ; fs ; Sa ; e ; Lva) T ->
   ty_config G (trm_lit l t ; fs ; Sb ; e ; Lvb) T->
   ty_stack G e Tt fs T ->
   〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
   〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Htys Htrf.
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.
  forwards Hhca: ty_config_to_heap_accessories Htyc1.
  forwards Hhcb: ty_config_to_heap_accessories Htyc2.

  forwards Htyca1: ty_config_to_accessories Htyc1.
  forwards Htyca2: ty_config_to_accessories Htyc2.

  step_invert_simple Hpr1 HtyC3.
  step_invert_simple Hpr2 HtyC4.

  - eapply transformation_lifted_focus_or_preserved_heap_lit_2obj; eauto.
  - eapply transformation_lifted_focus_or_preserved_heap_lit_2fun; eauto.
Qed.


Lemma transformation_lifted_focus_or_preserved_heap_write (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
    G a f a0 fs Sa Sb e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢ trm_write a f a0 ∶ Tt ->
  G ⊢ trm_write a f a0 ; fs ; Sa ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢trm_write a f a0; fs ; Sb ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_write a f a0 ; fs ; Sa ; e ; Lva) T ->
  ty_config G (trm_write a f a0 ; fs ; Sb ; e ; Lvb) T->
  ty_stack G e Tt fs T ->
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  (* write - needs similarity on heap *)

  introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Htys Htrf.
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.
  forwards Hhca: ty_config_to_heap_accessories Htyc1.
  forwards Hhcb: ty_config_to_heap_accessories Htyc2.

  forwards Htyca1: ty_config_to_accessories Htyc1.
  forwards Htyca2: ty_config_to_accessories Htyc2.

  step_invert_simple Hpr1 HtyC3.
  step_invert_simple Hpr2 HtyC4.

  binds_eq.
  binds_eq.

  forwards Hsia: tred_write_similar Hupd; fweauto.
  destructecs Hsia.

  forwards Hsib: tred_write_similar Hupd0; fweauto.
  destructecs Hsib.

  inversions HtyC3.

  eapply trf_lift_sim_conf_sim_lift_sim.
  eapply Hsia.
  2: eapply trf_symmetric. 2: eapply Hsib.

  eapply trf_lift_idtyp_conf_heap.
  -

    eapply ty_config_accessories_tred_write; eauto.
  -
    eapply ty_config_accessories_tred_write; eauto.
  -

    forwards: heap_correspond_accessories_update Hsia0; fweauto. eauto.
  -
    forwards: heap_correspond_accessories_update Hsib0; fweauto. eauto.
  -
    forwards Hcor: trf_lift_idtyp_heap_correspond Htrf Hhca. eauto. eauto.
  
    eapply ty_trm_write_inv in Htyt.
    destructecs Htyt.
    inversions Htyc.
    forwards Hcf: CanonicalForms.canonical_forms_fld H12 H14 Htyt. eauto. eauto.
    destructecs Hcf. 
    binds_eq.

    hnf in Hcor. destructecs Hcor.

    forwards: renv_corr_binds_loc Hcor6 Hbw3y3e. eauto.


    eapply trf_lift_idtyp_heap_update with (Sc1 := empty) (Sc2:=empty); eauto.
    +
      intros. unfold ctx_obj. eapply inert_dom_vis. eauto. eapply ctx_k_in_dom; eauto.
    + sympl. eapply ty_var_sub.
      eapply ReferenceTypes.ty_renv_binds_var; eauto. eauto.
    +
      eapply RecordTypes.open_record_type.
      eapply PreciseTyping.pf_bind in Hcf7.
      eapply PreciseTyping.pf_rcd_T. eauto. eapply Hcf7.
    +
      rewrite concat_empty_r. auto.
    +
      inversions Hsia. inversions H31.
      rewrite concat_empty_r. auto.
  -
    eauto.
  -
    eauto.
Qed.

Lemma transformation_lifted_focus_or_preserved_heap_call (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
  (TLU: TransformUnhidingCompat typ trm TL)
    G a m a0 fs Sa Sb e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢ trm_call a m a0 ∶ Tt ->
  G ⊢trm_call a m a0; fs ; Sa ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_call a m a0 ; fs ; Sb ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_call a m a0 ; fs ; Sa ; e ; Lva) T ->
  ty_config G (trm_call a m a0; fs ; Sb ; e ; Lvb) T->
  ty_stack G e Tt fs T ->
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
 introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Htys Htrf.
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.
  forwards Hhca: ty_config_to_heap_accessories Htyc1.
  forwards Hhcb: ty_config_to_heap_accessories Htyc2.

  forwards Htyca1: ty_config_to_accessories Htyc1.
  forwards Htyca2: ty_config_to_accessories Htyc2.


  step_invert_simple Hpr1 HtyC3.
  step_invert_simple Hpr2 HtyC4.

  binds_eq.

  (* Cases on whether the selected item is transformed *)
  forwards Hbim: trf_lift_idtyp_heap_binds_obj_met_cases Htrf.
  eapply HbyS.
  eapply HbyS0.
  eauto.
  eauto.

  assert (ok e) as Hoke. {
   hnf in Hhca. destructecs Hhca. auto.
  }

  destructecs Hbim.
  * (* Not the transformed function - keep the same *)
    subst.
    eapply trf_lift_sim_conf_lift. eauto.

    eapply trf_lift_idtyp_conf_heap; eauto.
    eapply ty_config_accessories_tred_call; eauto. inversions Htyc1; auto.
    eapply ty_config_accessories_tred_call; eauto. inversions Htyc2; auto.

    lets Htyt': Htyt.
    forwards Htyf: ty_trm_call_inv Htyt'. destructecs Htyf.

    forwards Hcf: CanonicalForms.canonical_forms_met Htyf0; fweauto. inversions Htyc; auto. inversions Htyc; eauto.  inversions Htyc; eauto. destructecs Hcf.

    forwards: typing_call_met Htyt; fweauto.
    inversions Htyc1. eauto.     inversions Htyc1. eauto. 
    binds_eq.
    binds_eq.

    forwards: defs_has_met_inv Hhas Hcf4. subst.
    eauto.

  * (* The transformed item - copy to focus *)
   
    (* Prepare in advance, so parts of it can be used for the intermediate config *)
    assert (
      ty_config_accessories F'0 e fs Sa t /[# 1 -> w0] /[-> w3] T  (fv (trm_call (avar_f w0) m (avar_f w3)) \u fv fs \u fv_heap Sa \u fv_renv e \u Lva)
    ) as Hhcaa. {   eapply ty_config_accessories_tred_call; eauto. inversions Htyc1; auto. }
    assert (
      ty_config_accessories F'0 e fs Sb t /[# 1 -> w0] /[-> w3] T  (fv (trm_call (avar_f w0) m (avar_f w3)) \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb)
    ) as Hhcab. {
      hnf in Hhcaa; destructecs Hhcaa.
      hnf in Htyca1; destructecs Htyca1.
      hnf in Htyca2; destructecs Htyca2.
      hnf; splits; eauto.
      - eapply ty_nothing_fresh_in_config_p; eauto.
      - eapply null_p; eauto.
      - eapply TypedPreservationLemmas.closed_call'. eapply HbyS. eapply Hhas. eauto. eauto.
    }

    eapply trf_lift_sim_conf_lift_lift with (Gc:=F'0) (Tc:=T) (Cc:=t /[# 1 -> w0] /[-> w3] ; fs ; Sb ; e ; fv (trm_call (avar_f w0) m (avar_f w3)) \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb).
    + eauto.
    + (* Heap transform *)
    eapply trf_lift_idtyp_conf_heap; eauto. 
    forwards Htyf: ty_trm_call_inv Htyt. destructecs Htyf.
    forwards Hcf: CanonicalForms.canonical_forms_met Htyf0; fweauto. inversions Htyc; auto. inversions Htyc; eauto.  inversions Htyc; eauto. destructecs Hcf.
    forwards: typing_call_met Htyt; fweauto.
    inversions Htyc1. eauto.    inversions Htyc1. eauto. 
    binds_eq.
    binds_eq.
    forwards: defs_has_met_inv Hhas Hcf4. subst.
    eauto.
+ (* Focus transform *)

    forwards Htyf: ty_trm_call_inv' Htyt. destructecs Htyf.
    forwards Hcf: CanonicalForms.canonical_forms_met Htyf0; fweauto. inversions Htyc; auto. inversions Htyc; eauto.  inversions Htyc; eauto. destructecs Hcf.
    destructecs Hcf5.

    inversions Htyc2.

    binds_eq.
    binds_eq.
    forwards: defs_has_met_inv Hhas Hcf4. subst.
    ctx_binds_eq.
    forwards: PreciseTyping.pf_bind Hcf0.
    forwards: PreciseTyping.pf_rcd_T H5 H1. 
    forwards: RecordTypes.record_has_functional Hbim Hcf7.
    eapply RecordTypes.open_record_type; eauto.
    simpl. auto.
    inversions H3.

    eapply trf_lift_idtyp_conf_focus.
    econstructor.
    eauto.

    eapply ty_config_accessories_tred_call; eauto.
    eauto.
    eapply Htys.

    forwards: trf_local_idtyp_trm_subst_met Hbim0 Hbw1ye.
    eauto.
    eauto.
    eauto.
    eauto.
    eauto.
    eauto.
    eauto.
    eauto.

    {
      eapply ty_trm_sub. eauto.

      pick_fresh yw3.

      forwards: Hcf9 yw3. eauto.
      forwards: subst_end w3 H3. eapply subst_sub. eapply ewkn_sub. eauto. eauto. eauto. eauto. rewrite subst_fresh. eauto. eauto.
      eapply ty_trm_call_inv_vis2; eauto.
      rewrite <- subst_intro in H4.
      rewrite <- subst_intro in H4. eapply H4.

      eapply notin_fv_open; eauto.
      eauto.
    }

    eauto.
    eauto.
    eapply H3.

    {
      eapply subtyp_trans.
      2: eapply Htyf2.

      pick_fresh yw3.
      pick_fresh yw0.

      forwards: Hcf5 yw3 yw0; fweauto.

      ctx_split_between yw3 yw0 in H3.

      forwards: subst_ty w3 H3. eapply subst_sub. eapply ewkn_sub.
      ctx_flatten.
      eauto. eauto.
      eauto. eauto. eauto. 

      rewrite subst_fresh.
      eapply weaken_end. eauto. sympls. eauto. ctx_flatten. eauto. eauto.

      intros. ctx_flatten. eapply ctx_vis_n. eauto. eapply ty_trm_call_inv_vis2; eauto.
      ctx_flatten in H4.
      rewrite <- ! subst_intro in H4.

      rewrite ! open_commut with (n:=1) in H4.

      forwards: subst_end w0 H4. eapply subst_sub. eapply ewkn_sub. ctx_flatten. eauto.
      ctx_simpl_fv. eauto. eauto. eauto.
      ctx_flatten. 
      rewrite subst_fresh. auto.

      eapply notin_fv_open; eauto.
      eauto.
      rewrite <- ! subst_intro in H11.
      rewrite ! open_commut with (n:=1). eauto.
      eauto.
      eauto.
      eapply notin_fv_open; eauto.
      eapply notin_fv_open. eauto. eapply notin_fv_open; eauto.

      eauto.
      eauto.
      eapply notin_fv_open; eauto.
      eapply notin_fv_open. eauto. eapply notin_fv_open; eauto.
      eauto.
    }

Qed.

Lemma transformation_lifted_focus_or_preserved_stack_read (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
    G a f fsa fsb S e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢ trm_read a f ∶ Tt->
  G ⊢ trm_read a f  ; fsa ; S ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_read a f  ; fsb ; S ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_read a f ; fsa ; S ; e ; Lva) T ->
  ty_config G (trm_read a f ; fsb ; S ; e ; Lvb) T ->
  heap_correspond G e S ->
  〈 G ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ∼ trf_lift_idtyp_stack typing_mode TL ∼> 〈 G ⁏ e ⊢ fsb ∶ (Tt, T) 〉->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt Hpr1 Hpr2 Htyca Htycb Hhc Htrf.



  (* Basic assertions about typing *)
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.

  forwards Hhca: ty_config_to_heap_accessories Htyca.
  forwards Hhcb: ty_config_to_heap_accessories Htycb.

  step_invert_simple Hpr1 HtyC3.
  step_invert_simple Hpr2 HtyC4.
  
  (* Unify the location *)
  binds_eq.
  binds_eq.
  forwards:defs_has_inv Hhas Hhas0.
  inversions H1. 
  clear H0 H.
  rename w0 into w, y0 into y.

  (* Pick fresh variable for the new reference *)
  pick_fresh_with  z (fv fsa \u fv fsb).

  (* Information about the field *)
  forwards Htyti: ty_trm_read_inv' Htyt.
  destructecs Htyti.
  inversions Htyca.
  forwards Hcf: CanonicalForms.canonical_forms_fld Htyti0; fweauto.
  destructecs Hcf.
  binds_eq.
  binds_eq.
  forwards Hheq: defs_has_inv Hhas Hcf2. inversions Hheq.
  rename vVal into y.

  (* Use similarity to use the chosen reference *)

  destructecs Hfrw2.
  destructecs Hfrw0.

  assert (y \in dom S) as HyS. {
    hnf in Hhc; destructecs Hhc.  
    eapply ctx_k_in_kdom in Hcf5.
    rewrite Hhc4 in Hcf5. auto. 
  }

  assert (
    〈 G §ʷ w2 ~ T2 ⊢ trm_var (avar_f w2) ; fsa ; S ; e & w2 ~ y ; \{ w} \u fv fsa \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
    ∼ trf_similarity typing_mode ∼>
    〈 G §ʷ z ~ T2 ⊢ trm_var (avar_f z) ; fsa ; S ; e & z ~ y ; \{ w} \u fv fsa \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
  ) as Hsia. {

    forwards Hsi: trf_similarity_config_read w w2 z Hpr1.

    hnf in Hhca; destructecs Hhca.  eauto.
    eapply config_typed_valid; eauto. sympl; eauto. sympl; eauto. sympl; eauto.
    auto. sympl in Hsi.  auto.
  }

  assert (
    〈 G §ʷ z ~ T0 ⊢ trm_var (avar_f z) ; fsb ; S ; e & z ~ y ; \{ w} \u fv fsb \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉
    ∼ trf_similarity typing_mode ∼>
    〈 G §ʷ w3 ~ T0 ⊢ trm_var (avar_f w3) ; fsb ; S ; e & w3 ~ y ; \{ w} \u fv fsb \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉
  ) as Hsib. {

    forwards Hsi: trf_similarity_config_read w w3 z Hpr2.

    hnf in Hhcb; destructecs Hhcb.  eauto.
    eapply config_typed_valid; eauto. sympl; eauto. sympl; eauto. sympl; eauto.
    auto. sympl in Hsi.
    eapply TransformSimilaritySymmetric in Hsi.
    auto.
  }


  (* Find common type for the variable *)

  forwards Hyeq: ctx_binds_k_after_eq Hby2G Hby2G0.
  destructecs Hyeq; subst.
  rename T7 into Ty.
  rename T2 into Twa, T0 into Twb.
  rename w2 into wa, w3 into wb.

  assert (inert_typ Ty v_loc) as HinlTy. {
    forwards: inert_binds_k_inert Hby2G0. eauto. auto.
  }

  assert (inert_typ Ty v_ref) as HinTy. {
  eapply inert_typ_loc_ref; eauto.
  }
  assert (typ_corr Ty Ty) as HcorTy. {
  eapply inert_typ_corr_loc_refl; eauto.
  }
  assert (inert (G §ʷ z ~ Ty)) as HinGz. {
  econstructor; eauto.
  }
  assert (heap_correspond (G §ʷ z ~ Ty) (e & z ~ y) S) as Hhcz. {
  eapply heap_correspond_epush; eauto.  
  }


  assert (renv_corr (G §ʷ z ~ Ty) (e & z ~ y)) as Hcorrz. {
    hnf in Hhcz. destructecs Hhcz. auto.
  }

  assert ((G §ʷ z ~ Ty)  ⁏ (e & z ~ y) ⊢ z ∶ T9) as Htyz. {
    eapply  ReferenceTypes.ty_renv_binds_rev_sametype_var with (v2:=y) (U:=Ty).
    eauto.
    eapply ctx_binds_e.
    eapply ctx_binds_n with (v:=y). eauto.
    eapply ctx_binds_k_after_binds; eauto.
    eapply binds_push_eq.
    eauto.
    auto.
      eapply weaken_push. 
      eapply ewkn_push.
    eauto.
    eauto.
    eauto.
    eauto.
    eauto.
  } 
  assert ((G §ʷ z ~ Ty)  ⁏ (e & z ~ y) ⊢ z ∶ Tt) as HtyzTt. {
    eapply ty_var_sub with (T:=U).
    eapply typing_viewpoint_adapt_loc with (w:=w) (T:=T9) (S:=Ty).
    eapply ewkn_push; eauto.
    eapply weaken_push; eauto.
    eauto.
    eauto.
    auto.
    eapply ewkn_push; eauto.
    eapply weaken_push; eauto.
  }



  (* Re-type the configurations with this type *)

  set (Lv2:=( \{ w} \u fv fsa \u fv_heap S \u fv_renv e \u Lva ) \u ( \{ w} \u fv fsb \u fv_heap S \u fv_renv e \u Lvb)).

  (* Similarity with the common type in the context - left side *)
  assert ( 
    〈 G §ʷ z ~ Twa ⊢ trm_var (avar_f z) ; fsa ; S ; e & z ~ y ; \{ w} \u fv fsa \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
    〈 G §ʷ z ~ Ty ⊢ trm_var (avar_f z) ; fsa ; S ; e & z ~ y ; Lv2 ∶ T 〉
  ) as Hsiat. {
    eapply trf_similarity_retype_Lv.

  *
    inversions Hsia; eauto.
  *
    unfold Lv2. sympl.

    inversions Hsia.
    inversions H18.

    forwards Htrfty: TransformLiftIdtypStackTyped TLT.
    eapply Htrfty in Htrf. hnf in Htrf. sympl in Htrf.

    econstructor.
    + auto.
    + auto.
    + eapply ty_stack_epush. eauto. eauto. eapply ty_stack_push. eauto. eauto.
      eapply Htrf.
    + apply all_tfree_k_Gpush; eauto.
    + eapply ty_trm_var; eauto.
    + eapply all_tfree_k_trm_avar. eapply ctx_k_push_e.
    + eapply ty_nothing_fresh_in_config_push1_Lv_T; eauto. sympl; union_sub_solver. 
      eapply subset_trans with (b:=ctx_fv G).
      eapply ctx_binds_fv_subset. eapply ctx_binds_k_after_binds; eauto.
      destruct H11.
      destruct H1. eapply subset_trans. eapply H3.  sympl; union_sub_solver. 

    + hnf. exists w. union_in_solver. 
    + eapply ctx_sealed_push. auto. eapply ctx_closed_binds. eapply H14. eapply ctx_binds_k_after_binds; eauto.
    +   eapply closed_read; eauto.
    + auto.
  * unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsiac: TransformSimilarityTransitive Hsia Hsiat. clear Hsia Hsiat.

  assert ( 
    〈 G §ʷ z ~ Ty ⊢ trm_var (avar_f z) ; fsb ; S ; e & z ~ y ; Lv2 ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
    〈 G §ʷ z ~ Twb ⊢ trm_var (avar_f z) ; fsb ; S ; e & z ~ y ; \{ w} \u fv fsb \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉
  ) as Hsibt. {
    eapply TransformSimilaritySymmetric.
    eapply trf_similarity_retype_Lv.
    *
      inversions Hsib; eauto.
    *
      unfold Lv2. sympl.
      inversions Hsib.
      inversions H18.
      forwards Htrfty: TransformLiftIdtypStackTyped TLTI.
      rewrite <- trf_inverse_lift_idtyp_stack in Htrfty.
      eapply Htrfty in Htrf. hnf in Htrf. sympl in Htrf.
      econstructor.
      + auto.
      + auto.
      + eapply ty_stack_epush. eauto. eauto. eapply ty_stack_push. eauto. eauto. 
        eapply Htrf.
      + inversions Htycb. apply all_tfree_k_Gpush; eauto.
      + eapply ty_trm_var; eauto.
      + eapply all_tfree_k_trm_avar. eapply ctx_k_push_e.
      + inversions Htycb. eapply ty_nothing_fresh_in_config_push1_Lv_T; eauto. sympl; union_sub_solver. 
        eapply subset_trans with (b:=ctx_fv G).
        eapply ctx_binds_fv_subset. eapply ctx_binds_k_after_binds; eauto.
        destruct H26.
        destruct H1. eapply subset_trans. eapply H3.  sympl; union_sub_solver. 
      + hnf. exists w. union_in_solver.
      + eapply ctx_sealed_push. auto. eapply ctx_closed_binds. eapply H14. eapply ctx_binds_k_after_binds; eauto.
      + inversions Htycb. eapply closed_read; eauto.
      + auto.
      + eauto.
      * unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsibc: TransformSimilarityTransitive Hsibt Hsib. clear Hsib Hsibt.

  eapply trf_lift_sim_conf_sim_lift_sim.
  eapply Hsiac. 2: eapply Hsibc.

  eapply trf_lift_idtyp_conf_stack with (Tt:=Tt).

  4: {
    eapply WknEnvPushCompatTransformationStack with (TEa1:=〈 G §ʷ z ~ Ty ⁏ e ⊢ fsa ∶ (Tt, T) 〉) (TEb1:=〈 G §ʷ z ~ Ty ⁏ e ⊢ fsb ∶ (Tt, T) 〉).
    eauto.
    eapply WknCtxPushCompatTransformationStack. eauto. eapply Htrf.
    econstructor; eauto.
    econstructor; eauto.
    econstructor; eauto.
    econstructor; eauto.
    econstructor; eauto.
    econstructor; eauto.
  }

- inversions Hsiac. eapply ty_config_to_accessories; eauto.
- inversions Hsibc. eapply ty_config_to_accessories; eauto.
- auto. 
- eauto.
Qed.


Lemma transformation_lifted_focus_or_preserved_stack_apply (TL : TermTransform)
    G a a0 fsa fsb S e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢ trm_apply a a0  ∶ Tt->
  G ⊢ trm_apply a a0 ; fsa ; S ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_apply a a0 ; fsb ; S ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_apply a a0  ; fsa ; S ; e ; Lva) T ->
  ty_config G (trm_apply a a0  ; fsb ; S ; e ; Lvb) T ->
  heap_correspond G e S ->
  〈 G ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ∼ trf_lift_idtyp_stack typing_mode TL ∼> 〈 G ⁏ e ⊢ fsb ∶ (Tt, T) 〉->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Hhc Htrf.
  (* Config typing *)
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.
  forwards Htyca1: ty_config_to_accessories Htyc1.
  forwards Htyca2: ty_config_to_accessories Htyc2.
  (* Step *)
  step_invert_simple Hpr1 HtyC3.
  step_invert_simple Hpr2 HtyC4.
  binds_eq.
  binds_eq.
  (* Only one transformation, on the stack *)
  eapply trf_lift_sim_conf_lift. eauto.

  eapply trf_lift_idtyp_conf_stack.
  eapply ty_config_accessories_tred_apply'; eauto.
  eapply ty_config_accessories_tred_apply'; eauto.
  auto.
  eauto.
  forwards: ty_trm_apply_inv Htyt. destructecs H0.

  forwards: CanonicalForms.canonical_forms_fun H1; fweauto. inversions Htyc; auto. inversions Htyc; auto. destructecs H3.
  forwards: typing_apply_fun H2; fweauto.
  binds_eq.
  binds_eq.
   eauto.
Qed.

Lemma transformation_lifted_focus_or_preserved_stack_let (TL : TermTransform)
    G t1 t2 fsa fsb S e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢ trm_let t1 t2 ∶ Tt->
  G ⊢ trm_let t1 t2 ; fsa ; S ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_let t1 t2 ; fsb ; S ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_let t1 t2 ; fsa ; S ; e ; Lva) T ->
  ty_config G (trm_let t1 t2 ; fsb ; S ; e ; Lvb) T ->
  heap_correspond G e S ->
  〈 G ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ∼ trf_lift_idtyp_stack typing_mode TL ∼> 〈 G ⁏ e ⊢ fsb ∶ (Tt, T) 〉->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Hhc Htrf.
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.
  forwards Htyca1: ty_config_to_accessories Htyc1.
  forwards Htyca2: ty_config_to_accessories Htyc2.
  step_invert_simple Hpr1 HtyC3.
  step_invert_simple Hpr2 HtyC4.


  forwards Htylet: ty_let_inv Htyt. eauto. destructecs Htylet.

  eapply trf_lift_sim_conf_lift. eauto.

  eapply trf_lift_idtyp_conf_stack.

  eapply ty_config_accessories_tred_push; eauto.
  eapply ty_config_accessories_tred_push; eauto.
  auto.
  eapply trf_lift_idtyp_stack_n.
  eauto.
  eauto.
  eapply Htylet0.
Qed.


Lemma transformation_lifted_focus_or_preserved_stack_lit (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
    G l t  fsa fsb S e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢trm_lit l t  ∶ Tt->
  G ⊢ trm_lit l t ; fsa ; S ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_lit l t ; fsb ; S ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_lit l t ; fsa ; S ; e ; Lva) T ->
  ty_config G (trm_lit l t ; fsb ; S ; e ; Lvb) T ->
  heap_correspond G e S ->
  〈 G ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ∼ trf_lift_idtyp_stack typing_mode TL ∼> 〈 G ⁏ e ⊢ fsb ∶ (Tt, T) 〉->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Hhc Htrf.
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.


  step_invert_simple Hpr1 HtyC3;
  step_invert_simple Hpr2 HtyC4.
  * (* obj *)
    eapply transformation_lifted_focus_or_preserved_stack_lit_2obj; eauto.
  * (* fun *)
    eapply transformation_lifted_focus_or_preserved_stack_lit_2fun; eauto.
Qed.


Lemma transformation_lifted_focus_or_preserved_stack_write (TL : TermTransform)
    G a f a0 fsa fsb S e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢ trm_write a f a0 ∶ Tt->
  G ⊢ trm_write a f a0  ; fsa ; S ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_write a f a0  ; fsb ; S ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_write a f a0; fsa ; S ; e ; Lva) T ->
  ty_config G (trm_write a f a0 ; fsb ; S ; e ; Lvb) T ->
  heap_correspond G e S ->
  〈 G ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ∼ trf_lift_idtyp_stack typing_mode TL ∼> 〈 G ⁏ e ⊢ fsb ∶ (Tt, T) 〉->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  (* write - needs similarity on heap *)
  introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Hhc Htrf.
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.
  forwards Hhca: ty_config_to_heap_accessories Htyc1.
  forwards Hhcb: ty_config_to_heap_accessories Htyc2.

  forwards Htyca1: ty_config_to_accessories Htyc1.
  forwards Htyca2: ty_config_to_accessories Htyc2.

  step_invert_simple Hpr1 HtyC3.
  step_invert_simple Hpr2 HtyC4.

  binds_eq.
  binds_eq.

  cleanup.

  forwards Hsia: tred_write_similar Hupd; fweauto.
  destructecs Hsia.

  forwards Hsib: tred_write_similar Hupd0; fweauto.
  destructecs Hsib.
  (* In order to type the new term, we need more typing information *)
  assert (sealed F'0) as HclG. inversions Htyc; auto.
  forwards Htyw: ty_trm_write_inv; fweauto. destructecs Htyw.
  forwards Hcf: CanonicalForms.canonical_forms_fld_ref Hbw1y1e; fweauto. destructecs Hcf.

  eapply trf_lift_sim_conf_sim_lift_sim.
  eapply Hsia.
  2: eapply trf_symmetric. 2: eapply Hsib.

  forwards: heap_update_keys_eq Hsia0 Hsib0. auto. auto.
  subst.

  eapply trf_lift_idtyp_conf_stack.

  eapply ty_config_accessories_tred_write; eauto.
  eapply ty_config_accessories_tred_write; eauto.
  assert (all_tfree_k_trm F'0 (trm_write (avar_f w0) a (avar_f w2)) v_ref). inversions Htyc. auto.
  forwards: heap_update_inert  Htyt; fweauto. auto.
  eauto.
  eauto.
Qed.


Lemma transformation_lifted_focus_or_preserved_stack_call (TL : TermTransform)
    G a m a0 fsa fsb S e Lva F' C3 Lvb F'0 C4 T Tt:
  G ⁏ e ⊢ trm_call a m a0 ∶ Tt->
  G ⊢ trm_call a m a0 ; fsa ; S ; e ; Lva ∶ T ↦ F' ⊢ C3 ->
  G ⊢ trm_call a m a0 ; fsb ; S ; e ; Lvb ∶ T ↦ F'0 ⊢ C4 ->
  ty_config G (trm_call a m a0 ; fsa ; S ; e ; Lva) T ->
  ty_config G (trm_call a m a0 ; fsb ; S ; e ; Lvb) T ->
  heap_correspond G e S ->
  〈 G ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ∼ trf_lift_idtyp_stack typing_mode TL ∼> 〈 G ⁏ e ⊢ fsb ∶ (Tt, T) 〉 ->
  〈 F' ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 F'0 ⊢ C4 ∶ T 〉.
Proof.
  introv Htyt Hpr1 Hpr2 Htyc1 Htyc2 Hhc Htrf.
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.

  forwards Htyca1: ty_config_to_accessories Htyc1.
  forwards Htyca2: ty_config_to_accessories Htyc2.

  step_invert_simple Hpr1 HtyC3.
  step_invert_simple Hpr2 HtyC4.
  eapply trf_lift_sim_conf_lift. eauto.
  binds_eq.
  binds_eq.
 
  forwards: defs_has_met_inv Hhas Hhas0. subst.
  eapply trf_lift_idtyp_conf_stack.

  eapply ty_config_accessories_tred_call; eauto.
  eapply ty_config_accessories_tred_call; eauto.
  auto.
  eauto.
  eapply typing_call_met; eauto. inversions Htyc; eauto.
Qed.

Lemma transformation_lifted_focus_or_preserved
  (TL: TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TRT: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TTI: TransformEnsures (@TransformCondTypeIdentical _ _ _)  TL)
  (TLA: TransformPreserves TransformIsAnswerTerm TL)
  (TRA: TransformPreserves TransformIsAnswerTerm (trf_inverse TL))
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLN: TransformNarrowingCompat _ typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
  (TLU: TransformUnhidingCompat typ trm TL)

  C1 C2 G T:
  (
    〈G ⊢ C1 ∶ T〉 ∼(trf_lift_idtyp_conf typing_mode TL)∼> 〈G ⊢ C2 ∶ T〉 ->
    ( (* 3 possibilities: finished, in focus, or transferred *)
      (answer C1 /\ answer C2) \/
      〈G ⊢ C1 ∶ T〉 ∼(trf_focus_idtyp_conf typing_mode TL)∼> 〈G ⊢ C2 ∶ T〉 \/
      exists C3 C4 G3 G4, 
        G ⊢ C1 ∶ T ↦ G3 ⊢ C3 /\
        G ⊢ C2 ∶ T ↦ G4 ⊢ C4 /\
        〈G3 ⊢ C3 ∶ T〉 ∼(trf_lift_sim_conf typing_mode TL)∼> 〈G4 ⊢ C4 ∶ T〉
    )
  ).
  introv Hlift.
  (* Typing for C1 *)
  forwards HtypedL: TransformLiftIdtypConfigTyped TLT TTI.
  forwards Htyc1: HtypedL Hlift.
  (* Typing for C2 *)
  forwards HtypedR: TransformLiftIdtypConfigTyped TRT. eapply trf_ensures_typeid_inverse; eauto.
  forwards Hliftinv: trf_inverse_intro Hlift.
  rewrite trf_inverse_lift_idtyp_conf in Hliftinv.
  forwards Htyc2: HtypedR Hliftinv.
  hnf in Htyc1, Htyc2.

  (* Progress *)
  forwards Hpr1: typed_progress Htyc1.
  forwards Hpr2: typed_progress Htyc2.

  (* Solve answer cases *)
  destructecs Hpr1; [>| destructecs Hpr2].
  -- (* C1 is answer *)
    forwards HansL: TransformPreservesIsAnswerLiftIdtyp TLA.
    forwards Hans2: HansL Hlift Hpr1.
    inversions Hans2.
    eauto.

  -- (* C2 is answer *)
    forwards HansR: TransformPreservesIsAnswerLiftIdtyp TRA.
    forwards Hans2: HansR Hliftinv Hpr2.
    inversions Hans2.
    eauto.
  -- (* Neither is answer *)
  right.
  (* Type the resulting configs *)
  rename c' into C3, c'0 into C4.
  forwards HtyC3: typed_preservation Hpr1.
  forwards HtyC4: typed_preservation Hpr2.

  (* Cases on where the transormation applies *)
  pose proof Hlift as Hlift'.
  inversions Hlift.
  - (* Focus *)
    (* Continue inverting the place of the transformation. *)
    (* The change can be anywhere, but we are only interesed in the top level,
      so it is inversion instead of induction *)
    inversions H0.
    inversions H8.
    + (* Focus root - directly one of the possible cases *)
      left. econstructor; eauto.
    + (* Let trm 1 - change is moved within focus *)
      right.

      (* Fix the configs and the steps *)
      eexists. eexists. eexists. eexists.
      splits. eapply Hpr1. eapply Hpr2.

      eapply transformation_lifted_focus_or_preserved_focus_let_1 with (L:=L \u ctx_dom G); eauto.
      intros. eapply ty_trm_sub. eapply H11; eauto. eapply weaken_push; eauto.

    + (* Let trm 2 - change is moved onto stack *)
      right.

      (* Fix the configs and the steps *)
      eexists. eexists. eexists. eexists.
      splits. eapply Hpr1. eapply Hpr2.

      eapply transformation_lifted_focus_or_preserved_focus_let_2 with (Tt:=T0) ; eauto.
      eapply ty_stack_sub; eauto.

    + (* Lit trm 1 *)
      (* The literal is transformed -> will transfer to the heap *)
      right.

      eexists. eexists. eexists. eexists.
      splits. eapply Hpr1. eapply Hpr2.

      eapply transformation_lifted_focus_or_preserved_focus_lit_1  with (Tt:=T0); eauto.
      eapply ty_stack_sub; eauto.
    
    + (* Lit 2 *)
      right.

      eexists. eexists. eexists. eexists.
      splits. eapply Hpr1. eapply Hpr2.

      eapply transformation_lifted_focus_or_preserved_focus_lit_2  with (Tt:=T0); eauto.
      eapply ty_stack_sub; eauto.

  - (* Heap *)

    (* For steps that do not read the heap, the transformation stays the same.
       For a "call" step, which reads from the heap, the transformation may copy to the focus.
       During the copying, the argument is substituted.
    *)
      right. exists C3 C4 F' F'0 ; splits.
      auto.
      auto.

      (* Cases on the term determine the step *)

      destruct t.
      + (* Var - popping from the stack *)
      destruct fs.
      * (* answer - not possible *) inversions Hpr1; inversions H15.
      * (* popping *)

        step_invert_simple Hpr1 HtyC3.
        step_invert_simple Hpr2 HtyC4.

        inversions H8. forwards: ty_trm_var_inv H9.
        eapply trf_lift_sim_conf_lift. eauto.
        eapply trf_lift_idtyp_conf_heap; eauto.
        eapply ty_config_accessories_tred_var; eauto.
        eapply ty_config_accessories_tred_var; eauto.

      eapply @renaming_fresh; eauto.
      apply CloseUnboundSubstIntro. apply ewkn_ty_trm. inversions HtyC4. eapply HeapCorrespondence.heap_correspond_to_ok_e; eauto.
    + (* Read *)
      (* Weakening with a new variable *)
      eapply transformation_lifted_focus_or_preserved_heap_read; eauto.
    + (* Apply *)
      eapply transformation_lifted_focus_or_preserved_heap_apply; eauto.
    + (* Let *)
      eapply transformation_lifted_focus_or_preserved_heap_let; eauto.
    + (* Lit *)
      eapply transformation_lifted_focus_or_preserved_heap_lit; eauto.
    + (* Write *)
      eapply transformation_lifted_focus_or_preserved_heap_write; eauto.
    + (* call *)
      eapply transformation_lifted_focus_or_preserved_heap_call; eauto.
  - (* Stack *)

   (* For steps that do not pop the stack, the transformation stays the same.
      For the step which pops the stack, if the transformation is in the first frame,
      it moves to the focus *)

      right. exists C3 C4 F' F'0 ; splits.
      auto.
      auto.

      (* Cases on the term determine the step *)

      destruct t.
      + (* Var - popping from the stack *)
        (* Where is the transformation  *)
      inversions H6.
      * (* On top -> moving to focus *)
        step_invert_simple Hpr1 HtyC3.
        step_invert_simple Hpr2 HtyC4.
        (* Apply to focus *)
        eapply trf_lift_sim_conf_lift. eauto.
        eapply trf_lift_idtyp_conf_focus. 
        eapply trf_focus_idtyp_conf_focus.

        eapply ty_config_accessories_tred_var; eauto.
        eapply ty_config_accessories_tred_var; eauto.
        auto.
        eauto.
        eapply trf_local_idtyp_trm_subst_w; eauto. eauto.
      * (* Deeper -> still applies in stack *)

        step_invert_simple Hpr1 HtyC3.
        step_invert_simple Hpr2 HtyC4.

        eapply trf_lift_sim_conf_lift. eauto.
        eapply trf_lift_idtyp_conf_stack.

        eapply ty_config_accessories_tred_var; eauto.
        eapply ty_config_accessories_tred_var; eauto.
        auto.
        eauto. eapply @renaming_fresh; eauto.
        apply CloseUnboundSubstIntro. apply ewkn_ty_trm. eapply ty_trm_var_inv; eauto. 
        eapply ty_trm_var_inv_vis; eauto.

     + (* read - needs weakening *)
        eapply transformation_lifted_focus_or_preserved_stack_read; eauto.

     + (* apply - no change *)
        eapply transformation_lifted_focus_or_preserved_stack_apply; eauto.
     + (* let - push - move down *)
        eapply transformation_lifted_focus_or_preserved_stack_let; eauto.
     + (* lit needs weakening *)
        eapply transformation_lifted_focus_or_preserved_stack_lit; eauto.
     + (* write - needs similarity on heap *)
        eapply transformation_lifted_focus_or_preserved_stack_write; eauto.
     + (* call - no change *)
        eapply transformation_lifted_focus_or_preserved_stack_call; eauto.
  -- eauto.
Qed.


(* Lemma 55 in "Pure methods for roDOT" *)
Theorem transformation_preserved
  (TL: TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TRT: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TRA: TransformPreserves TransformIsAnswerTerm (trf_inverse TL))
  (TTI: TransformEnsures (@TransformCondTypeIdentical _ _ _)  TL)
  (TLA: TransformPreserves TransformIsAnswerTerm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLU: TransformUnhidingCompat typ trm TL)
  (TFS: TransformConfigReducesToShort typing_mode (trf_focus_idtyp_conf typing_mode TL) (trf_similarity typing_mode))
  Ca1 Ca2 Cb1 Ga1 Ga2 Gb1 Ta Tb na:
  (* The original program terminates in na steps *)
  Ga1 ⊢ Ca1 ∶ Ta ⊣na↦* Ga2 ⊢ Ca2 ->
  answer Ca2 ->
  (* Transformation of the program *)
  〈Ga1 ⊢ Ca1 ∶ Ta〉 ∼(trf_lift_sim_conf typing_mode TL)∼> 〈Gb1 ⊢ Cb1 ∶ Tb〉 ->
  (* Exists transformed execution *)
  exists nb Gb2 Cb2,
    (* Which is not longer *)
    nb <= na /\
    (* Reaches an answer *)
    Gb1 ⊢ Cb1 ∶ Tb ⊣nb↦* Gb2 ⊢ Cb2 /\
    answer Cb2 /\
    (* Which is transformed *)
    〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼(trf_lift_sim_conf typing_mode TL)∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉.
Proof.
  introv Hred Hans Htrf.
  (* First induction is by number of steps.
    The length of the transformation may increase, by reduction,
    but the length of execution cannot increase by transformation
  *)
  reverts Ga1 Ca1 Ta.
  reverts Ga2 Ca2.
  reverts Gb1 Cb1 Tb.
  induction na using nat_le_induction; intros.
  - (* 0 steps *)
    inversions Hred.
    exists 0 Gb1 Cb1; splits.
    + auto.
    + eauto.
    + 
      forwards HansL: TransformIterateSimPreservesIsAnswer TLA.
      forwards: HansL Htrf.
      hnf; eauto. hnf in H. auto.
    + auto.
  - (* na+1 steps *)
    (* Induction by the transformation *)
    unfold trf_lift_sim_conf in Htrf.
    remember (〈 Ga1 ⊢ Ca1 ∶ Ta 〉).
    remember (〈 Gb1 ⊢ Cb1 ∶ Tb 〉).
    reverts Ga1 Ca1 Ta. reverts Gb1 Cb1 Tb.
    induction Htrf; intros; subst.
    + (* No transformation *)
      inversions Heqt.
      exists (S na) Ga2 Ca2; splits; eauto.
      eapply trf_iterate_0. 
      hnf. eapply typed_preservation_closure; eauto.
    + (* Some transformation *)
      (* previous transformations *)

      destruct TEb as [Gc1 Tc1 Cc1].

      forwards Hleft: IHHtrf; eauto.
      destructecs Hleft.
    
      destruct H1.
      ++ (* Similarity *)

      exists nb.


      forwards Hrsi: transform_config_reduces_to_n_exists_inv H1 Hleft1;
      rewrite ? trf_symmetric_inverse.
      apply TransformConfigReducesTo1Similarity.
      apply TransformSimilaritySymmetric.
      apply TransformSimilarityPreservesIsAnswer.
      apply TransformSimilaritySymmetric.
      apply TransformSimilarityTyped.
      apply TransformSimilaritySymmetric.
      
      destructecs Hrsi.
      exists Gb0 Cb0; splits; auto.
      forwards Hsipa: TransformSimilarityPreservesIsAnswer Hrsi Hleft2. apply Hsipa.
      eapply trf_iterate_S. eapply Hleft. left. auto.
      ++  (* Transformation *)
      (* The type does not change (Tc1 = Tb) *)
      forwards Htyeq: TrfLiftIdtypEnsuresCond H1.
      hnf in Htyeq. inversions Htyeq.
      (* The configuration before transformation is typed *)
      assert (ty_config Gb1 Cc1 Tb) as HtyCc1. { 
        forwards: TransformLiftIdtypConfigTyped H1. auto. auto.
        hnf in H2. auto.
      }
      assert (ty_config Gb2 Cb2 Tb) as HtyCb2. { 
        eapply typed_preservation_closure; eauto.
      }

      forwards Hrfc: transformation_lifted_focus_or_preserved H1; fweauto.
      destructecs Hrfc.
      * (* Answer *)
        forwards: typed_red_closure_answer_case Hleft1; fweauto.
        destructecs H2; subst.
        exists 0 Gb2 Cb1; splits; auto.
        eapply trf_iterate_S. eapply Hleft. right. auto.
      * (* Focus *)
        forwards Hfc: TFS Hrfc.
        destructecs Hfc.
        forwards Hshort: typed_red_closure_answer_split Hleft1 Hleft2 Hfc1. auto.
        destructecs Hshort. subst nb.

        forwards Hrsi: transform_config_reduces_to_n_exists_inv Hfc Hshort1; 
        rewrite ? trf_symmetric_inverse.
        apply TransformConfigReducesTo1Similarity.
        apply TransformSimilaritySymmetric.
        apply TransformSimilarityPreservesIsAnswer.
        apply TransformSimilaritySymmetric.
        apply TransformSimilarityTyped.
        apply TransformSimilaritySymmetric.
        destructecs Hrsi.

        assert (answer Cb3). {
          forwards: TransformSimilarityPreservesIsAnswer Hrsi. hnf; auto. auto.
        }

        exists (n2 + n32) Gb3 Cb3; splits.
        lia. eapply typed_red_closure_concat; eauto. auto.
        eapply trf_transitive. eauto.
        eapply trf_lift_sim_conf_sim.  auto. eapply trf_transitive; eauto.
        

      * (* Transferred - apply induction with one less step *)
        forwards Hstep: typed_red_closure_answer_split1 Hleft1 Hrfc0. auto. auto.
        destructecs Hstep. subst nb.
        forwards Hstepl: typed_red_closure_invl Hleft1.
        destructecs Hstepl.

        forwards Hsim0: trf_similarity_step Hstepl0 Hrfc0. auto.
      
        forwards Hnext: H Hstepl. lia. auto.
        eapply trf_transitive.
        eapply trf_lift_sim_conf_sim. eapply typed_preservation; eauto. eauto. eauto.
  
        destructecs Hnext.

        exists (S nb) Gb0 Cb0; splits.
        lia.
        eapply typed_red_stepl; eauto. auto.
        eapply trf_transitive. eauto. eauto.
Qed.

Lemma tred_answer_terminates G1 C1 T n G2 C2:
  G1 ⊢ C1 ∶ T ⊣ n ↦* G2 ⊢ C2 ->
  answer C2 ->
  config_terminates C1.
Proof.
  introv Hred Hans.
  hnf.
  exists n.
  hnf.
  exists C2; split.
  auto.
  eapply typed_to_untyped_red_closure; eauto.
Qed.

(* Theorem 25 in "Pure methods for roDOT" *)
Theorem transformation_preserved_if_terminates
  (TL: TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TRT: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TRA: TransformPreserves TransformIsAnswerTerm (trf_inverse TL))
  (TTI: TransformEnsures (@TransformCondTypeIdentical _ _ _)  TL)
  (TLA: TransformPreserves TransformIsAnswerTerm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLU: TransformUnhidingCompat typ trm TL)
  (TFS: TransformConfigReducesToShortIfTerminates typing_mode (trf_focus_idtyp_conf typing_mode TL) (trf_similarity typing_mode))
  Ca1 Ca2 Cb1 Ga1 Ga2 Gb1 Ta Tb na:
  (* The original program terminates in na steps *)
  Ga1 ⊢ Ca1 ∶ Ta ⊣na↦* Ga2 ⊢ Ca2 ->
  answer Ca2 ->
  (* Transformation of the program *)
  〈Ga1 ⊢ Ca1 ∶ Ta〉 ∼(trf_lift_sim_conf typing_mode TL)∼> 〈Gb1 ⊢ Cb1 ∶ Tb〉 ->
  (* Exists transformed execution *)
  exists nb Gb2 Cb2,
    (* Which is not longer *)
    nb <= na /\
    (* Reaches an answer *)
    Gb1 ⊢ Cb1 ∶ Tb ⊣nb↦* Gb2 ⊢ Cb2 /\
    answer Cb2 /\
    (* Which is transformed *)
    〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼(trf_lift_sim_conf typing_mode TL)∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉.
Proof.
  introv Hred Hans Htrf.
  (* First induction is by number of steps.
    The length of the transformation may increase, by reduction,
    but the length of execution cannot increase by transformation
  *)
  reverts Ga1 Ca1 Ta.
  reverts Ga2 Ca2.
  reverts Gb1 Cb1 Tb.
  induction na using nat_le_induction; intros.
  - (* 0 steps *)
    inversions Hred.
    exists 0 Gb1 Cb1; splits.
    + auto.
    + eauto.
    + 
      forwards HansL: TransformIterateSimPreservesIsAnswer TLA.
      forwards: HansL Htrf.
      hnf; eauto. hnf in H. auto.
    + auto.
  - (* na+1 steps *)
    (* Induction by the transformation *)
    unfold trf_lift_sim_conf in Htrf.
    remember (〈 Ga1 ⊢ Ca1 ∶ Ta 〉).
    remember (〈 Gb1 ⊢ Cb1 ∶ Tb 〉).
    reverts Ga1 Ca1 Ta. reverts Gb1 Cb1 Tb.
    induction Htrf; intros; subst.
    + (* No transformation *)
      inversions Heqt.
      exists (S na) Ga2 Ca2; splits; eauto.
      eapply trf_iterate_0. 
      hnf. eapply typed_preservation_closure; eauto.
    + (* Some transformation *)
      (* previous transformations *)

      destruct TEb as [Gc1 Tc1 Cc1].

      forwards Hleft: IHHtrf; eauto.
      destructecs Hleft.
    
      destruct H1.
      ++ (* Similarity *)

      exists nb.


      forwards Hrsi: transform_config_reduces_to_n_exists_inv H1 Hleft1;
      rewrite ? trf_symmetric_inverse.
      apply TransformConfigReducesTo1Similarity.
      apply TransformSimilaritySymmetric.
      apply TransformSimilarityPreservesIsAnswer.
      apply TransformSimilaritySymmetric.
      apply TransformSimilarityTyped.
      apply TransformSimilaritySymmetric.
      
      destructecs Hrsi.
      exists Gb0 Cb0; splits; auto.
      forwards Hsipa: TransformSimilarityPreservesIsAnswer Hrsi Hleft2. apply Hsipa.
      eapply trf_iterate_S. eapply Hleft. left. auto.
      ++  (* Transformation *)
      (* The type does not change (Tc1 = Tb) *)
      forwards Htyeq: TrfLiftIdtypEnsuresCond H1.
      hnf in Htyeq. inversions Htyeq.
      (* The configuration before transformation is typed *)
      assert (ty_config Gb1 Cc1 Tb) as HtyCc1. { 
        forwards: TransformLiftIdtypConfigTyped H1. auto. auto.
        hnf in H2. auto.
      }
      assert (ty_config Gb2 Cb2 Tb) as HtyCb2. { 
        eapply typed_preservation_closure; eauto.
      }

      forwards Hrfc: transformation_lifted_focus_or_preserved H1; fweauto.
      destructecs Hrfc.
      * (* Answer *)
        forwards: typed_red_closure_answer_case Hleft1; fweauto.
        destructecs H2; subst.
        exists 0 Gb2 Cb1; splits; auto.
        eapply trf_iterate_S. eapply Hleft. right. auto.
      * (* Focus *)
        forwards Hfc: TFS Hrfc.
        eapply tred_answer_terminates; eauto.
        destructecs Hfc.
        forwards Hshort: typed_red_closure_answer_split Hleft1 Hleft2 Hfc1. auto.
        destructecs Hshort. subst nb.

        forwards Hrsi: transform_config_reduces_to_n_exists_inv Hfc Hshort1; 
        rewrite ? trf_symmetric_inverse.
        apply TransformConfigReducesTo1Similarity.
        apply TransformSimilaritySymmetric.
        apply TransformSimilarityPreservesIsAnswer.
        apply TransformSimilaritySymmetric.
        apply TransformSimilarityTyped.
        apply TransformSimilaritySymmetric.
        destructecs Hrsi.

        assert (answer Cb3). {
          forwards: TransformSimilarityPreservesIsAnswer Hrsi. hnf; auto. auto.
        }

        exists (n2 + n32) Gb3 Cb3; splits.
        lia. eapply typed_red_closure_concat; eauto. auto.
        eapply trf_transitive. eauto.
        eapply trf_lift_sim_conf_sim.  auto. eapply trf_transitive; eauto.
        

      * (* Transferred - apply induction with one less step *)
        forwards Hstep: typed_red_closure_answer_split1 Hleft1 Hrfc0. auto. auto.
        destructecs Hstep. subst nb.
        forwards Hstepl: typed_red_closure_invl Hleft1.
        destructecs Hstepl.

        forwards Hsim0: trf_similarity_step Hstepl0 Hrfc0. auto.
      
        forwards Hnext: H Hstepl. lia. auto.
        eapply trf_transitive.
        eapply trf_lift_sim_conf_sim. eapply typed_preservation; eauto. eauto. eauto.
  
        destructecs Hnext.

        exists (S nb) Gb0 Cb0; splits.
        lia.
        eapply typed_red_stepl; eauto. auto.
        eapply trf_transitive. eauto. eauto.
Qed.


End WithTypingMode.