(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation Weakening EWeakening Substitution TFree.
Require Import TransformationSimilarity TransformationPreservation TransformationTerm.
Require Import TransformationConfig OperationalSemantics TypedOperationalSemantics.
Require Import ContextSimilarity ConfigSimilarity.
Require Import ConfigTyping ConfigTypingAccessories TypedProgress.
Require Import TypedPreservation.
Require Import TransformationIterateSim.
Require Import Similarity TypingSimilarity.

Require Import TransformationLifting.
Require Import FreshChain FreshChainTactics.
Require Import TypedOperationalSemantics InertContext.
Require Import Similarity SyntaxSimilarity.
Require Import ExecutionTactics Renaming TypingHelpers.

Require Import NothingFresh TypedPreservation LetLet Defs.
Require Import HeapCorrespondence ConfigVarkinds.
Require Import HeapUpdate ConfigValid ConfigNotations.

Require Import TransformationSimilaritySteps ConfigTypingSteps TransformationConfigSteps.
Require Import Defs LetLet EnvCorrespondence.
Require Import Subenvironments Narrowing.
Require Import Defree.
Require Import Visibility.
Require Import TypedPreservationLemmas.
Require Import TransformationTyping.
Require Import Sealing SealClosed.

(** * Preservation of lifted transformations during reduction involving function items. *)

Section WithTypingMode.
Context (typing_mode: Mode).

Lemma trf_retype_fun G T t1a t1b t2a t2b fsa fsb Sa Sb e Lva Lvb w y U Tt L Tc:
  ty_config G (trm_lit (lit_fun T t1a) t2a ; fsa ; Sa ; e ; Lva) Tc ->
  ty_config G (trm_lit (lit_fun T t1b) t2b ; fsb ; Sb ; e ; Lvb) Tc ->

  G ⁏ e ⊢ lit_fun T t1a ∶ typ_fun T U ->
  ( forall x : var, x \notin L -> G §ˣ x ~ typ_fun T U ⁏ e ⊢ t2a /[-> x] ∶ Tt) ->
  ty_stack G e Tt fsa Tc ->

  G ⁏ e ⊢ lit_fun T t1b ∶ typ_fun T U ->
  ( forall x : var, x \notin L -> G §ˣ x ~ typ_fun T U ⁏ e ⊢ t2b /[-> x] ∶ Tt) ->
  ty_stack G e Tt fsb Tc ->

  fresh (ctx_dom G \u ctx_fv G \u fv T \u fv_env Sa \u fv_env Sb \u dom e \u fv_env e \u fv t1a \u fv t2a \u fv t1b \u fv t2b \u fv Tc \u Lva \u Lvb \u fv fsa \u fv fsb) 2 (w :: y :: nil) ->

  exists UN Lva2 Lvb2,
  ty_config (inert_G_for_item G w y (typ_fun T UN)) (t2a /[-> w] ; fsa ; Sa & y ~ item_fun t1a ; e & w ~ y ; Lva2) Tc /\
  ty_config (inert_G_for_item G w y (typ_fun T UN)) (t2b /[-> w] ; fsb ; Sb & y ~ item_fun t1b ; e & w ~ y ; Lvb2) Tc /\ (
    exists TcontN LNN,
      ty_stack G e TcontN fsa Tc /\
      ty_stack G e TcontN fsb Tc /\ 
      (forall x : var, x \notin LNN -> G §ˣ x ~ typ_fun T UN ⁏ e ⊢ t2a /[-> x] ∶ TcontN)  /\
      (forall x : var, x \notin LNN -> G §ˣ x ~ typ_fun T UN ⁏ e ⊢ t2b /[-> x] ∶ TcontN)
  )  /\
  (exists L0, forall x : var, x \notin L0 -> G §ᶻ x ~ T ⁏ e ⊢ t1a /[-> x] ∶ UN /[-> x]) /\
  (exists L1, forall x : var, x \notin L1 -> G §ᶻ x ~ T ⁏ e ⊢ t1b /[-> x] ∶ UN /[-> x]) /\
  (fv_config (trm_lit (lit_fun T t1a) t2a ; fsa ; Sa ; e ; Lva)) \c Lva2 /\
  (fv_config (trm_lit (lit_fun T t1b) t2b ; fsb ; Sb ; e ; Lvb)) \c Lvb2 /\
  closed (typ_fun T UN) /\
  fv UN \c Lva2 /\
  fv UN \c Lvb2 /\
  w \notin fv UN /\
  y \notin fv UN.
Proof.
  introv Htyca Htycb. (* Htyc2a Htyc2b. clear Htyc2a Htyc2b.*)
  introv Htyfa Htyta Htyfsa.
  introv Htyfb Htytb Htyfsb.
  introv Hfr.

  lets Htyca': Htyca.
  lets Htycb': Htycb.

  inversions Htyca.
  destruct H13 as [null Hnull].
  inversions H15.
  inversions Htycb.
  inversions H27.

  (* Find a common type *)
  pose proof fun_nothing_fresh' as HnfU.
  specialize (HnfU typing_mode G e Tc L null T (typ_fun T U) Tt (fv t1a \u fv t2a \u fv t1b \u fv t2b \u fv fsa \u fv fsb)).
  forwards HnfU1: HnfU; fweauto.
  clear HnfU. eapply Closedness.closed_trm_lit_fun_T; eauto.
  destructecs HnfU1.

  forwards Hnfta: HnfU1 fsa t1a t2a; fweauto.
  union_sub_solver.
  intros. eapply Closedness.closed_trm_lit_fun_t1; eauto.
  intros. eapply Closedness.closed_trm_lit_fun_t2; eauto.

  forwards Hnftb: HnfU1 fsb t1b t2b; fweauto.
  union_sub_solver.
  intros. eapply Closedness.closed_trm_lit_fun_t1; eauto.
  intros. eapply Closedness.closed_trm_lit_fun_t2; eauto.

  destructecs Hnfta.
  destructecs Hnftb.
  inversions Hnfta1.
  inversions Hnftb1.

  assert (w \notin \{ null}) as Hnwn. {
    rewrite notin_singleton. 
    intro. subst.
    assert ( null \notin Lva). notin_solve.
    contradiction.
  }
  assert (w \notin fv U0) as HnwU. {
  eapply notin_subset.
  eapply subset_trans with (b:=fv (typ_fun T U0)).
  sympls. union_sub_solver.
  eapply HnfU0.
  notin_solve.
  }
  assert (y \notin \{ null}) as Hnyn. {
    rewrite notin_singleton. 
    intro. subst.
    assert ( null \notin Lva). notin_solve.
    contradiction.
  }
  assert (y \notin fv U0) as HnyU. {
  eapply notin_subset.
  eapply subset_trans with (b:=fv (typ_fun T U0)).
  sympls. union_sub_solver.
  eapply HnfU0.
  notin_solve.
  }
  assert (ok e) as Hoke. {
    hnf in H5; destructecs H5; eauto.
  }
  assert (ctx_ok G) as HokG. {
    eauto.
  }


  exists U0.
  exists (fv_config (trm_lit (lit_fun T t1a) t2a ; fsa ; Sa ; e ; Lva) \u fv U0).
  exists (fv_config (trm_lit (lit_fun T t1b) t2b ; fsb ; Sb ; e ; Lvb) \u fv U0).
  splits.
  -
    econstructor.
    + eapply inert_G_for_item_fun_inert; eauto. sympl. fresh_solve. auto.
    + eapply tred_lit_fun_hc; eauto. sympl. fresh_solve. auto.
    + eapply ty_stack_epush. eauto. eauto. eapply ty_stack_push. eauto. eauto. eapply ty_stack_push. eauto. eauto. eauto.
    + apply all_tfree_k_Gpush. apply all_tfree_k_Gpush; eauto. eauto.
    + eapply tred_lit_fun_trm; eauto.
    + eapply tred_lit_fun_tfree_k_trm; eauto.
    + 

      eapply ty_nothing_fresh_in_config_push2_Lv_T. eapply H11. union_sub_solver.
      sympl. union_sub_solver.
    + exists null. union_in_solver.
    + eapply tred_lit_fun_sealed; eauto.
    + eapply config_closed_Lv. eapply tred_lit_fun_closed; eauto.
    + eauto.
  -
    econstructor.
    + eapply inert_G_for_item_fun_inert; eauto. sympl. fresh_solve. auto.
    + eapply tred_lit_fun_hc; eauto. sympl. fresh_solve. auto.
    + eapply ty_stack_epush. eauto. eauto. eapply ty_stack_push. eauto. eauto. eapply ty_stack_push. eauto. eauto. eauto.
    + apply all_tfree_k_Gpush. apply all_tfree_k_Gpush; eauto. eauto.
    + eapply tred_lit_fun_trm; eauto.
    + eapply tred_lit_fun_tfree_k_trm; eauto.
    + 

      eapply ty_nothing_fresh_in_config_push2_Lv_T. eapply H23. union_sub_solver.
      sympl. union_sub_solver.
    + destructecs H25. exists null0. union_in_solver.
    + eapply tred_lit_fun_sealed; eauto.
    + eapply config_closed_Lv. eapply tred_lit_fun_closed; eauto.
    + eauto.
  - exists TcontN LNN. splits; eauto.
  - exists L0. eauto.
  - exists L1. eauto.
  - eapply subset_union_weak_l.
  - eapply subset_union_weak_l.
  - splits.
    -- auto. 
  -- eapply subset_union_weak_r.
  -- eapply subset_union_weak_r.
   -- auto. 
   -- auto. 
Qed.



Lemma trf_retype_fun' G T t1a t1b t2a t2b fsa fsb Sa Sb e Lva Lvb w y U Tt L Tc null:
  let Ca := (trm_lit (lit_fun T t1a) t2a ; fsa ; Sa ; e ; Lva) in
  let Cb := (trm_lit (lit_fun T t1b) t2b ; fsb ; Sb ; e ; Lvb) in
  ty_config G Ca Tc ->
  ty_config G Cb Tc ->

  G ⁏ e ⊢ lit_fun T t1a ∶ typ_fun T U ->
  ( forall x : var, x \notin L -> G §ˣ x ~ typ_fun T U ⁏ e ⊢ t2a /[-> x] ∶ Tt) ->
  ty_stack G e Tt fsa Tc ->

  G ⁏ e ⊢ lit_fun T t1b ∶ typ_fun T U ->
  ( forall x : var, x \notin L -> G §ˣ x ~ typ_fun T U ⁏ e ⊢ t2b /[-> x] ∶ Tt) ->
  ty_stack G e Tt fsb Tc ->

  null \in Lva ->
  let Lv2 := (fv_config Ca \u fv_config Cb \u fv U \u ctx_dom G \u ctx_fv G) in
  fresh (Lv2) 2 (w :: y :: nil) ->

  let UN :=  (seal_rec 1 null U) in
  let TtN :=  (seal null Tt) in

  ty_config (inert_G_for_item G w y (typ_fun T  UN)) (t2a /[-> w] ; fsa ; Sa & y ~ item_fun t1a ; e & w ~ y ; Lv2) Tc /\
  ty_config (inert_G_for_item G w y (typ_fun T  UN)) (t2b /[-> w] ; fsb ; Sb & y ~ item_fun t1b ; e & w ~ y ; Lv2) Tc /\ (
    exists LNN,
      ty_stack G e TtN fsa Tc /\
      ty_stack G e TtN fsb Tc /\ 
      (forall x : var, x \notin LNN -> G §ˣ x ~ typ_fun T UN ⁏ e ⊢ t2a /[-> x] ∶ TtN)  /\
      (forall x : var, x \notin LNN -> G §ˣ x ~ typ_fun T UN ⁏ e ⊢ t2b /[-> x] ∶ TtN)
  )  /\
  (exists L0, forall x : var, x \notin L0 -> G §ᶻ x ~ T ⁏ e ⊢ t1a /[-> x] ∶ UN /[-> x]) /\
  (exists L1, forall x : var, x \notin L1 -> G §ᶻ x ~ T ⁏ e ⊢ t1b /[-> x] ∶ UN/[-> x]) /\
  closed (typ_fun T UN).
Proof.
  introv Htyca Htycb. (* Htyc2a Htyc2b. clear Htyc2a Htyc2b.*)
  introv Htyfa Htyta Htyfsa.
  introv Htyfb Htytb Htyfsb.
  introv Hnull.
  intro Lv2.
  introv Hfr.
  intros. unfold UN, TtN.

  lets Htyca': Htyca.
  lets Htycb': Htycb.

  inversions Htyca.
  inversions H15.
  inversions Htycb.
  inversions H28.

  clear dependent T1.
  clear dependent T0.

  unfold Lv2 in *.
  assert (w <> null) as Hnwn. {
    intro. subst.
    assert ( null \notin Lva). notin_solve.
    contradiction.
  }
  assert (y <> null) as Hnyn. {
    intro. subst.
    assert ( null \notin Lva). notin_solve.
    contradiction.
  }
  assert (w \notin fv U /[# 1 ->> null]) as HnwUn. {
    eapply  fv_seal_notin; eauto.
  }
  assert (y \notin fv U /[# 1 ->> null]) as HnyUn. {
    eapply  fv_seal_notin; eauto.
  }


  assert (ok e) as Hoke. {
    hnf in H5; destructecs H5; eauto.
  }
  assert (ctx_ok G) as HokG. {
    eauto.
  }
  assert (mode_has_fun' typing_mode = true) as Hm. {
    inversions Htyfa. auto.
  }

  eapply seal_ty_lit_fun with (x:=null) in Htyfa.
  eapply seal_ty_lit_fun with (x:=null) in Htyfb.



    inversions Htyfa.
    inversions Htyfb.

  assert (forall x : var, x \notin Lv2 \u L0 -> G §ᶻ x ~ T ⁏ e ⊢ t1a /[-> x] ∶ U /[# 1 ->> null] /[-> x]) as Htyt1a. {
    unfold Lv2.
    intros.
    eapply H19. eauto.
  }
  assert (forall x : var, x \notin Lv2 \u L -> G §ˣ x ~ typ_fun T U /[# 1 ->> null] ⁏ e ⊢ t2a /[-> x] ∶ TtN) as Htyt2a. {
    unfold Lv2.
    intros.
    eapply seal_ty_lit_fun_cont with (L2:= Lv2 \u L); eauto. 
    intros. eapply Closedness.closed_trm_lit_fun_t2; eauto.
    intros. eapply Closedness.closed_trm_lit_fun_T; eauto.
  }

  assert (forall x : var, x \notin Lv2 \u L1 -> G §ᶻ x ~ T ⁏ e ⊢ t1b /[-> x] ∶ U /[# 1 ->> null] /[-> x]) as Htyt1b. {
    unfold Lv2.
    intros.
    eapply H23. eauto.
  }
  assert (forall x : var, x \notin Lv2 \u L -> G §ˣ x ~ typ_fun T U /[# 1 ->> null] ⁏ e ⊢ t2b /[-> x] ∶ TtN) as Htyt2b. {
    unfold Lv2.
    intros.
    eapply seal_ty_lit_fun_cont with (L2:= Lv2 \u L); eauto. 
    intros. eapply Closedness.closed_trm_lit_fun_t2; eauto.
    intros. eapply Closedness.closed_trm_lit_fun_T; eauto.
  }
  assert (closed (typ_fun T U /[# 1 ->> null])) as Hclf. {

   intro.
      intros.
      sympls.
      erewrite Closedness.closed_trm_lit_fun_T by eauto. 
      fequal.
      erewrite closed_open_down with (x0:=y). eauto.
      rewrite open_seal_down.
      eapply sealed_closed.
  }
  eapply seal_ty_stack_closed with (y0:=null)  in Htyfsa.
  eapply seal_ty_stack_closed with (y0:=null)  in Htyfsb.

  splits.
  -
    econstructor.
    + eapply inert_G_for_item_fun_inert; eauto. sympl. fresh_solve. auto.
    + eapply tred_lit_fun_hc; eauto. sympl. fresh_solve. all: sympls; eauto.
    + eapply ty_stack_epush. eauto. eauto. eapply ty_stack_push. eauto. eauto. eapply ty_stack_push. eauto. eauto. eauto.
    + apply all_tfree_k_Gpush. apply all_tfree_k_Gpush; eauto. eauto.
    + eapply tred_lit_fun_trm with (L:=Lv2 \u L); eauto.
    + eapply tred_lit_fun_tfree_k_trm; eauto.
    + 

      eapply ty_nothing_fresh_in_config_push2_Lv_T. eapply H11. union_sub_solver.
      sympl. union_sub_solver_destruct_ls. union_sub_solver.  eapply subset_trans. eapply fv_seal_subset. union_sub_solver_destruct_ls.  union_sub_solver.  rewrite subset_singleton. union_in_solver.
    + exists null. union_in_solver.
    + eapply tred_lit_fun_sealed; eauto. 
   
    + eapply config_closed_Lv. eapply tred_lit_fun_closed; eauto.
    + eauto.
  -
    econstructor.
    + eapply inert_G_for_item_fun_inert; eauto. sympl. fresh_solve. auto.
    + eapply tred_lit_fun_hc; eauto. sympl. fresh_solve. all: sympls; eauto.
    + eapply ty_stack_epush. eauto. eauto. eapply ty_stack_push. eauto. eauto. eapply ty_stack_push. eauto. eauto. eauto.
    + apply all_tfree_k_Gpush. apply all_tfree_k_Gpush; eauto. eauto.
    + eapply tred_lit_fun_trm with (L:=Lv2 \u L); eauto.
    + eapply tred_lit_fun_tfree_k_trm; eauto.
    + 

      eapply ty_nothing_fresh_in_config_push2_Lv_T. eapply H11. union_sub_solver.
      sympl. union_sub_solver_destruct_ls. union_sub_solver.  eapply subset_trans. eapply fv_seal_subset. union_sub_solver_destruct_ls.  union_sub_solver.  rewrite subset_singleton. union_in_solver.
    + exists null. union_in_solver.
    + eapply tred_lit_fun_sealed; eauto. 
    
    + eapply config_closed_Lv. eapply tred_lit_fun_closed; eauto.
    + eauto.
  - exists (Lv2 \u L). splits; eauto.
  - exists L0. eauto.
  - exists L1. eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - eauto.
  - intro. intros. forwards: H3 n x. injection H. sympl. intros. fequal; eauto.
  - eauto.
  - intro. intros. forwards: H2 n x. injection H. sympl. intros. fequal; eauto.
  - eauto.
Qed.

Lemma transformation_lifted_focus_or_preserved_focus_lit_1fun (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLSl: TransformSealingCompat typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
    G  T0 t1 T1 t3 tr fs S e Lva C3 Lvb C4 T Tt Tl L G3 G4:
  〈 G ⁏ e ⊢ lit_fun T0 t1 ∶ Tl  〉 ∼ trf_local_idtyp_lit typing_mode TL ∼> 〈 G ⁏ e ⊢  lit_fun T1 t3 ∶ Tl 〉->
  G ⊢ trm_lit (lit_fun T0 t1) tr ; fs ; S ; e ; Lva ∶ T ↦ G3 ⊢ C3 ->
  G ⊢ trm_lit (lit_fun T1 t3) tr ; fs ; S ; e ; Lvb ∶ T ↦ G4 ⊢ C4 ->
  heap_correspond G e S ->
  ty_stack G e Tt fs T ->
  (forall x : var, x \notin L -> G §ˣ x ~ Tl ⁏ e ⊢ tr /[-> x] ∶ Tt) ->
  〈 G3 ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 G4 ⊢ C4 ∶ T 〉.
Proof.
  introv Htrf Hpr1 Hpr2 Hhc Htys Htyt.


  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  step_invert_simple Hpr1' HtyC3.
  step_invert_simple Hpr2' HtyC4.



  lets Htrf': Htrf.
  inversions Htrf'.

  assert (has_null Lva) as Hnull. inversions Htyc. auto.
  destructc Hnull as null Hnull.

  (* Pick fresh variables to use in the transformation *)
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr. eauto. eapply HeapCorrespondence.heap_correspond_to_ok_e; eauto. eapply HeapCorrespondence.heap_correspond_to_ok_s; eauto.
  fresh_chain_pick Fr w1.
  fresh_chain_pick Fr y1.

  (* Assert similarity from both sides, renaming the chosen variables to the fresh variables *)

  assert ( exists U0,
    〈 inert_G_for_item G w y (typ_fun T0 U0)
       ⊢ t5 /[-> w] ; fs ; S & y ~ item_fun t2 ; e & w ~ y ; fv (trm_lit (lit_fun T0 t2) t5) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
       ∼ trf_similarity typing_mode
       ∼> 〈 inert_G_for_item G w1 y1 (typ_fun T0 U0)
          ⊢ t5 /[-> w1] ; fs ; S & y1 ~ item_fun t2 ; e & w1 ~ y1 ; fv (trm_lit (lit_fun T0 t2) t5) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
  /\  A1 = (typ_fun T0 U0)
  ) as Hsi1. {
  (* Left side *) 
  inversions H9. exists U0. split.
  forwards Hsi1: trf_similarity_config_lit_fun w1 y1 Hpr1'.
  eauto.
  eapply config_typed_valid; eauto. apply Hfr.
  sympl. fresh_chain_solve Fr.
  fresh_chain_solve Fr. fresh_chain_solve Fr. auto. auto.
  }

  assert (exists U0,
〈 inert_G_for_item G w1 y1 (typ_fun T0 U0)
       ⊢ t5 /[-> w1] ; fs ; S & y1 ~ item_fun t1 ; e & w1 ~ y1 ; fv (trm_lit (lit_fun T0 t1) t5) \u fv fs \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉
       ∼ trf_similarity typing_mode
       ∼> 〈 inert_G_for_item G w0 y0 (typ_fun T0 U0)
          ⊢ t5 /[-> w0] ; fs ; S & y0 ~ item_fun t1 ; e & w0 ~ y0 ; fv (trm_lit (lit_fun T0 t1) t5) \u fv fs \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉 
  /\  A2 = (typ_fun T0 U0)
) as Hsi2.
  {
  (* Right side needs symmetry *)
    inversions H12. exists U0. split.

    forwards Hsi2: trf_similarity_config_lit_fun w1 y1 Hpr2'.
    eauto.
    eapply config_typed_valid; eauto. auto.
    sympl. fresh_chain_solve Fr.
    fresh_chain_solve Fr. fresh_chain_solve Fr.
    eapply TransformSimilaritySymmetric in Hsi2.
    auto. auto.
  }
  destructecs Hsi1.
  destructecs Hsi2.
  subst.
  (* Typing of literal on both sides *)
  assert ( G ⁏ e ⊢lit_fun T0 t1 ∶ typ_fun T0 U) as Htylita. {
             eapply trf_inverse_intro in Htrf.
          forwards: TransformLocalIdtypLitTyped TLTI.
          rewrite <- trf_local_idtyp_inverse_lit in H.
          eapply H in Htrf. hnf in Htrf. eapply Htrf.
          eauto.
  }
  assert ( G ⁏ e ⊢lit_fun T0 t2 ∶ typ_fun T0 U) as Htylitb. {
          forwards: TransformLocalIdtypLitTyped TLT.
          eapply H in Htrf. hnf in Htrf. eapply Htrf.
  }


  set (Lv2 :=  (((fv T0 \u fv t2) \u fv t5) \u fv fs \u fv_heap S \u fv_renv e \u Lva) \u
              (((fv T0 \u fv t1) \u fv t5) \u fv fs \u fv_heap S \u fv_renv e \u Lvb) \u fv U \u ctx_dom G \u ctx_fv G (*fv U \u fv (trm_lit (lit_fun T0 t2) t5) \u fv fs \u fv_heap S \u fv_renv e \u Lva \u  fv (trm_lit (lit_fun T0 t1) t5) \u fv fs \u fv_heap S \u fv_renv e \u Lvb *) ).

  forwards Hrety': trf_retype_fun' w1 y1 null Htyc Htyc0. eapply Htylitb.  eauto. eauto. eapply Htylita. eauto. eauto. eauto.
  forwards Hrety: Hrety'. splits; sympl; fresh_chain_solve Fr.
  clear Hrety'.
  destructecs Hrety.
  destructecs Hrety2.
  destructecs Hrety3.
  destructecs Hrety4.


  (* Similarity with the common type in the context - left side *)
  assert ( 
   〈 inert_G_for_item G w1 y1 (typ_fun T0 U0)
            ⊢ t5 /[-> w1] ; fs ; S & y1 ~ item_fun t2 ; e & w1 ~ y1 ; fv (trm_lit (lit_fun T0 t2) t5) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
  〈 inert_G_for_item G w1 y1 (typ_fun T0 (seal_rec 1 null U))
            ⊢ t5 /[-> w1] ; fs ; S & y1 ~ item_fun t2 ; e & w1 ~ y1 ; Lv2 ∶ T 〉
  ) as Hsi1a. {
 eapply trf_similarity_retype_Lv.
 inversions Hsi0; eauto. unfold Lv2. sympl. sympl in Hrety0. eapply Hrety0. unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsi1b: TransformSimilarityTransitive Hsi0 Hsi1a. clear Hsi0 Hsi1a.

  (* Similarity with the common type in the context - right side *)

  assert ( 
    〈 inert_G_for_item G w1 y1 (typ_fun T0  (seal_rec 1 null U))
            ⊢ t5 /[-> w1] ; fs ; S & y1 ~ item_fun t1 ; e & w1 ~ y1 ; Lv2 ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
 〈 inert_G_for_item G w1 y1 (typ_fun T0 U1)
       ⊢ t5 /[-> w1] ; fs ; S & y1 ~ item_fun t1 ; e & w1 ~ y1 ; fv (trm_lit (lit_fun T0 t1) t5) \u fv fs \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉
  ) as Hsi2a. {

  eapply TransformSimilaritySymmetric.
 eapply trf_similarity_retype_Lv. 
 inversions Hsi3; eauto. eapply Hrety1. unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsi2b: TransformSimilarityTransitive Hsi2a Hsi3. clear Hsi3 Hsi2a.

  (* Solve the transformation with similarities on both sides *)
  eapply trf_lift_sim_conf_sim_lift_sim. eapply Hsi1b. 2: eapply Hsi2b.


  (* Transformaption happens on the heap *)
  eapply trf_lift_idtyp_conf_heap.
  - (* Left config accessories *)
    eapply ty_config_accessories_tred_let_fun; eauto 1.
    eapply ty_config_to_accessories; eauto.
    fresh_solve; try eapply fv_seal_notin;
    fresh_chain_solve Fr.

    unfold Lv2; union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_trans. eapply fv_seal_subset.
    union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_singleton. union_in_solver.
  -  (* Right config accessories *)
    eapply ty_config_accessories_tred_let_fun; eauto 1.
    eapply ty_config_to_accessories; eauto.
    fresh_solve; try eapply fv_seal_notin;
    fresh_chain_solve Fr.

    unfold Lv2; union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_trans. eapply fv_seal_subset.
    union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_singleton. union_in_solver.
  - (* Left heap accessories *)
    eapply heap_correspond_accessories_tred_let_fun. auto. eapply heap_correspond_to_accessories; eauto. fresh_solve; try eapply fv_seal_notin; fresh_chain_solve Fr.
  - (* Right heap accessories *)
    eapply heap_correspond_accessories_tred_let_fun. auto. eapply heap_correspond_to_accessories; eauto. fresh_solve; try eapply fv_seal_notin; fresh_chain_solve Fr.
  - (* Heap transformation *)
    hnf in Hhc. destructecs Hhc.
    eapply trf_lift_idtyp_heap_e.
    + (* Typing of old items *)
    intros y2 Hy2.
    eapply ty_item_s_epush; try fresh_chain_solve Fr. eapply  ty_item_s_Gpush; try fresh_chain_solve Fr; eauto.
    ctx_simpl_dom. fresh_chain_solve Fr.
  
    + (* Transformation of the new item *)
      eapply trf_lift_idtyp_item_fun with (L:= (L \u \{w0} \u \{y0} \u dom e \u ctx_dom G \u LNN \u L2 \u \{ w1} \u \{ y1} )).
      eauto.
      eapply ctx_binds_n. fresh_chain_solve Fr.
      eapply ctx_binds_e.
      intros.
  
      forwards: SealingCompatTransformationTrm TLSl  null.
      forwards: H0  (H7 x).
      eauto.
    econstructor.
    econstructor. hnf; eauto.
    rewrite ! open_seal_down.
assert (sealed G). inversions Htyc0. auto.
assert (closed T0). intro. intros. forwards: HclT0 n x0. sympl in H4. injection H4. auto.
assert (closed (t2 /[-> x])).
inversions  Htyc. inversions H28. eapply Closedness.closed_trm_lit_fun_t1; eauto.
assert (closed (t1 /[-> x])).
inversions  Htyc0. inversions H29. eapply Closedness.closed_trm_lit_fun_t1; eauto.
sympl in H2. rewrite H3 in H2.
eapply closed_sealed0 in H4.
eapply closed_sealed0 in H5.
eapply closed_sealed0 in H6.
rewrite H4 in H2.
rewrite H5 in H2.
rewrite H6 in H2.


assert (ctx_ok (inert_G_for_item G w1 y1 (typ_fun T0 U /[# 1 ->> null]) §ᶻ x ~ T0)) as HokGx. {
 econstructor. fresh_chain_solve2 Fr.  eauto.
}

eapply trf_local_idtyp_trm_ewkn_push. auto.
ctx_split_between y1 x.

eapply WeakeningCompatTransformationTrm. eauto.

eapply H2.
rewrite ctx_concat_single. econstructor. eauto. ctx_flatten. auto.  auto.
rewrite ctx_concat_single. econstructor. eauto. ctx_flatten. auto.  auto.
hnf; eauto. auto. auto. fresh_chain_solve Fr. 

      apply all_tfree_k_Gpush. apply all_tfree_k_Gpush. hnf in Htft. hnf. eauto. fresh_chain_solve Fr.  fresh_chain_solve2 Fr.
      apply all_tfree_k_Gpush. apply all_tfree_k_Gpush. hnf in Htft. hnf. eauto. fresh_chain_solve Fr.  fresh_chain_solve2 Fr.
 
  - 
    eapply ty_stack_epush; try fresh_chain_solve Fr.
    eapply ty_stack_push; try (ctx_simpl_dom; fresh_chain_solve Fr).
    eapply ty_stack_push; try fresh_chain_solve Fr. 
    eauto.
  - eapply @renaming_fresh with (L := (L \u \{w0} \u \{y0} \u dom e \u ctx_dom G \u LNN)) (vk:=v_let) (T:=typ_fun T0 U /[# 1 ->> null]).
      *
      apply CloseUnboundSubstIntro.
      * eauto.
      *
      apply ewkn_ty_trm.
      * fresh_chain_solve2 Fr.
      * fresh_chain_solve2 Fr.
      * 
      intros.
      ctx_split_between y1 z.
      assert (Weakening (ty_trm typing_mode)); eauto.
       apply @weaken_ty. eauto. sympl.    eapply ewkn_push. eapply Hrety2.  simpl; eauto.
    
      eauto. fresh_chain_solve Fr. eauto. fresh_chain_solve2 Fr.
      *
      eauto.
      * eauto.

Qed.



Lemma transformation_lifted_focus_or_preserved_focus_lit_2fun (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLSl: TransformSealingCompat typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
    G  ta tb tf Tf fs S e Lva C3 Lvb C4 T Tt Tl L G3 G4:
  (forall x : var,
       x \notin L ->
       〈 G §ˣ x ~ Tl ⁏ e ⊢ ta /[-> x] ∶ Tt 〉 ∼ trf_local_idtyp_trm typing_mode TL
       ∼> 〈 G §ˣ x ~ Tl ⁏ e ⊢ tb /[-> x] ∶ Tt 〉) ->
  G ⊢ trm_lit (lit_fun Tf tf) ta ; fs ; S ; e ; Lva ∶ T ↦ G3 ⊢ C3 ->
  G ⊢ trm_lit (lit_fun Tf tf) tb ; fs ; S ; e ; Lvb ∶ T ↦ G4 ⊢ C4 ->
  heap_correspond G e S ->
  ty_stack G e Tt fs T ->
  G ⁏ e ⊢ lit_fun Tf tf ∶ Tl ->
  〈 G3 ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 G4 ⊢ C4 ∶ T 〉.
Proof.
  introv Htrf Hpr1 Hpr2 Hhc Htys Htyt.


  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  step_invert_simple Hpr1' HtyC3.
  step_invert_simple Hpr2' HtyC4.

  lets Htyt': Htyt. 
  inversions Htyt'.

  assert (has_null Lva) as Hnull. inversions Htyc. auto.
  destructc Hnull as null Hnull.

  rename t3 into tf, t2 into tca, t4 into tcb, T1 into Tf.
  rename w into wa, y into ya, w0 into wb, y0 into yb.

  (* Pick fresh variables to use in the transformation *)
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr. eauto. eapply HeapCorrespondence.heap_correspond_to_ok_e; eauto. eapply HeapCorrespondence.heap_correspond_to_ok_s; eauto.
  fresh_chain_pick Fr wc.
  fresh_chain_pick Fr yc.

  (* Assert similarity from both sides, renaming the chosen variables to the fresh variables *)

  assert ( exists U0,
    〈 inert_G_for_item G wa ya (typ_fun Tf U0)
       ⊢ tca /[-> wa] ; fs ; S & ya ~ item_fun tf ; e & wa ~ ya ; fv (trm_lit (lit_fun Tf tf) tca) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
       ∼ trf_similarity typing_mode
       ∼> 〈 inert_G_for_item G wc yc (typ_fun Tf U0)
          ⊢ tca /[-> wc] ; fs ; S & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tca) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
  /\  A1 = (typ_fun Tf U0)
  ) as Hsi1. {
  (* Left side *) 
  inversions H9. exists U0. split.
  forwards Hsi1: trf_similarity_config_lit_fun wc yc Hpr1'.
  eauto.
  eapply config_typed_valid; eauto. apply Hfr.
  sympl. fresh_chain_solve Fr.
  fresh_chain_solve Fr. fresh_chain_solve Fr. auto. auto.
  }

  assert (exists U0,
〈 inert_G_for_item G wc yc (typ_fun Tf U0)
       ⊢ tcb /[-> wc] ; fs ; S & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tcb) \u fv fs \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉
       ∼ trf_similarity typing_mode
       ∼> 〈 inert_G_for_item G wb yb (typ_fun Tf U0)
          ⊢ tcb /[-> wb] ; fs ; S & yb ~ item_fun tf ; e & wb ~ yb ; fv (trm_lit (lit_fun Tf tf) tcb) \u fv fs \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉 
  /\  A2 = (typ_fun Tf U0)
) as Hsi2.
  {
  (* Right side needs symmetry *)
    inversions H12. exists U0. split.

    forwards Hsi2: trf_similarity_config_lit_fun wc yc Hpr2'.
    eauto.
    eapply config_typed_valid; eauto. auto.
    sympl. fresh_chain_solve Fr.
    fresh_chain_solve Fr. fresh_chain_solve Fr.
    eapply TransformSimilaritySymmetric in Hsi2.
    auto. auto.
  }
  destructecs Hsi1.
  destructecs Hsi2.
  subst.
  (* Typing of literal on both sides *)
  assert ( forall x : var,
       x \notin L ->
       G §ˣ x ~ typ_fun Tf U ⁏ e ⊢ tcb /[-> x] ∶ Tt) as Htycnta. {
      intros.
      specialize (Htrf x H).
             eapply trf_inverse_intro in Htrf.
          forwards Htrfty: TransformLocalIdtypTrmTyped TLTI.
          rewrite <- trf_local_idtyp_inverse_trm in Htrfty.
          eapply Htrfty in Htrf. hnf in Htrf. sympl in Htrf. eapply Htrf.
          eauto. 
  }
  assert ( forall x : var,
       x \notin L ->
       G §ˣ x ~ typ_fun Tf U ⁏ e ⊢ tca /[-> x] ∶ Tt) as Htycntb. {
      intros.
      specialize (Htrf x H).
          forwards Htrfty: TransformLocalIdtypTrmTyped TLT.
          eapply Htrfty in Htrf. hnf in Htrf. eapply Htrf.
  }


  set (Lv2 :=  (((fv Tf \u fv tf) \u fv tca) \u fv fs \u fv_heap S \u fv_renv e \u Lva) \u
              (((fv Tf \u fv tf) \u fv tcb) \u fv fs \u fv_heap S \u fv_renv e \u Lvb) \u fv U \u ctx_dom G \u ctx_fv G ).

  forwards Hrety': trf_retype_fun' wc yc null Htyc Htyc0.
  eapply Htyt. eapply Htycntb. eauto.
  eapply Htyt. eapply Htycnta. eauto.
  eauto.

  forwards Hrety: Hrety'. splits; sympl; fresh_chain_solve Fr.
  clear Hrety'.
  destructecs Hrety.
  destructecs Hrety2.
  destructecs Hrety3.
  destructecs Hrety4.


  (* Similarity with the common type in the context - left side *)
  assert ( 
    〈 inert_G_for_item G wc yc (typ_fun Tf U0)
            ⊢ tca /[-> wc] ; fs ; S & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tca) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
    〈 inert_G_for_item G wc yc (typ_fun Tf (seal_rec 1 null U))
            ⊢ tca /[-> wc] ; fs ; S & yc ~ item_fun tf ; e & wc ~ yc ; Lv2 ∶ T 〉
  ) as Hsi1a. {
    eapply trf_similarity_retype_Lv.
    inversions Hsi0; eauto.
    unfold Lv2.
    sympl. sympl in Hrety0. eapply Hrety0. unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsi1b: TransformSimilarityTransitive Hsi0 Hsi1a. clear Hsi0 Hsi1a.

  (* Similarity with the common type in the context - right side *)

  assert ( 
    〈 inert_G_for_item G wc yc (typ_fun Tf (seal_rec 1 null U))
            ⊢ tcb /[-> wc] ; fs ; S & yc ~ item_fun tf ; e & wc ~ yc ; Lv2 ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
 〈 inert_G_for_item G wc yc (typ_fun Tf U1)
       ⊢ tcb /[-> wc] ; fs ; S & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tcb) \u fv fs \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉
  ) as Hsi2a. {

  eapply TransformSimilaritySymmetric.
 eapply trf_similarity_retype_Lv. 
 inversions Hsi3; eauto. eapply Hrety1. unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsi2b: TransformSimilarityTransitive Hsi2a Hsi3. clear Hsi3 Hsi2a.

  (* Solve the transformation with similarities on both sides *)
  eapply trf_lift_sim_conf_sim_lift_sim. eapply Hsi1b. 2: eapply Hsi2b.
  assert (wc \notin Lva). fresh_chain_solve Fr.
  assert (yc \notin Lva). fresh_chain_solve Fr.
  assert (wc \notin fv U /[# 1 ->> null]). {
    eapply fv_seal_notin. intro. subst. contradiction. fresh_chain_solve Fr.
  }
  assert (yc \notin fv U /[# 1 ->> null]). {
    eapply fv_seal_notin. intro. subst. contradiction. fresh_chain_solve Fr.
  }

  (* Transformaption happens on the focus *)
  eapply trf_lift_idtyp_conf_focus.
  eapply trf_focus_idtyp_conf_focus with (Tt:=Tt  /[# 0 ->> null] ).
  - (* Left config accessories *)
    eapply ty_config_accessories_tred_let_fun; eauto 1.
    eapply ty_config_to_accessories; eauto.
    fresh_solve; try eapply fv_seal_notin;
    fresh_chain_solve Fr.

    unfold Lv2; union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_trans. eapply fv_seal_subset.
    union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_singleton. union_in_solver.
  -  (* Right config accessories *)
    eapply ty_config_accessories_tred_let_fun; eauto 1.
    eapply ty_config_to_accessories; eauto.
    fresh_solve; try eapply fv_seal_notin;
    fresh_chain_solve Fr.

    unfold Lv2; union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_trans. eapply fv_seal_subset.
    union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_singleton. union_in_solver.
  - (* Heap *)
    eapply tred_lit_fun_hc; eauto.
    sympl.
    fresh_solve; fresh_chain_solve Fr.
  - 
    eapply ty_stack_epush; try fresh_chain_solve Fr.
    eapply ty_stack_push; try (ctx_simpl_dom; fresh_chain_solve Fr).
    eapply ty_stack_push; try fresh_chain_solve Fr. 
    eauto.
  -
    eapply trf_local_idtyp_trm_subst_w with (L:=L \u ctx_dom G \u \{ wc} \u \{ yc}) (vk:=v_let).
    auto.
    auto.
    fresh_chain_solve2 Fr.
    fresh_chain_solve2 Fr.
    eauto.
    intros x Hx.
    forwards Htrfx: Htrf x . auto.
          forwards HSl: SealingCompatTransformationTrm TLSl  null.
          forwards Htrfxs: HSl (Htrfx).
    econstructor.
    econstructor.
    hnf; eauto.

    assert (sealed G). inversions Htyc0. auto.
    assert (closed Tf). intro. intros. forwards: HclT0 n x0. sympl in H6. injection H6. auto.
    assert (closed (tca /[-> x])).
    inversions  Htyc. inversions H29. eapply Closedness.closed_trm_lit_fun_t2; eauto.
    assert (closed (tcb /[-> x])).
    inversions  Htyc0. inversions H30. eapply Closedness.closed_trm_lit_fun_t2; eauto.
    sympl in Htrfxs.
    rewrite H5 in Htrfxs.


    eapply closed_sealed0 in H6.
    eapply closed_sealed0 in H7.
    eapply closed_sealed0 in H8.
    rewrite H6 in Htrfxs.
    rewrite H7 in Htrfxs.
    rewrite H8 in Htrfxs.


    eapply trf_local_idtyp_trm_ewkn_push. auto.
    ctx_split_between yc x.

    eapply WeakeningCompatTransformationTrm. eauto.
    eapply Htrfxs.
    rewrite ctx_concat_single. econstructor. eauto. ctx_flatten. econstructor. fresh_chain_solve2 Fr. ctx_simpl_dom. notin_solve.  fresh_chain_solve Fr.
    rewrite ctx_concat_single. econstructor. eauto. ctx_flatten. econstructor. fresh_chain_solve2 Fr. ctx_simpl_dom. notin_solve.  fresh_chain_solve Fr.
    hnf; eauto. 
    econstructor. fresh_chain_solve2 Fr. ctx_simpl_dom. notin_solve.  fresh_chain_solve Fr.  fresh_chain_solve Fr.
  -
    eapply subtyp_refl.
Qed.

Lemma transformation_lifted_focus_or_preserved_heap_lit_2fun (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
    G Tf tf tr fs Sa Sb e Lva C3 Lvb C4 T Tt  G3 G4:
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->  
  G ⊢ trm_lit (lit_fun Tf tf) tr ; fs ; Sa ; e ; Lva ∶ T ↦ G3 ⊢ C3 ->
  G ⊢ trm_lit (lit_fun Tf tf) tr ; fs ; Sb ; e ; Lvb ∶ T ↦ G4 ⊢ C4 ->
  G ⁏ e ⊢ trm_lit (lit_fun Tf tf) tr ∶ Tt ->
  ty_stack G e Tt fs T ->
(*
  G ⁏ e ⊢ lit_obj R d ∶ Tl ->
  (forall x : var, x \notin L -> G §ˣ x ~ Tl ⁏ e ⊢ tr /[-> x] ∶ Tt) -> *)
  〈 G3 ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 G4 ⊢ C4 ∶ T 〉.
Proof.
  introv Htrf Hpr1 Hpr2 Htyt Htys.

  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  step_invert_simple Hpr1' HtyC3.
  step_invert_simple Hpr2' HtyC4.

  forwards Htl: ty_trm_lit_inv Htyt. eauto.
  destructecs Htl.
  
  lets Htylit: Htl0.
  inversions Htylit.

  assert (has_null Lva) as Hnull. inversions Htyc. auto.
  destructc Hnull as null Hnull.

  rename t3 into tf, t4 into tc, T1 into Tf.
  rename w into wa, y into ya, w0 into wb, y0 into yb.

  (* Pick fresh variables to use in the transformation *)
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr. eauto. eapply HeapCorrespondence.heap_correspond_to_ok_e; eauto. eapply HeapCorrespondence.heap_correspond_to_ok_s; eauto. eapply HeapCorrespondence.heap_correspond_to_ok_s; eauto.
  fresh_chain_pick Fr wc.
  fresh_chain_pick Fr yc.

  (* Assert similarity from both sides, renaming the chosen variables to the fresh variables *)

  assert ( exists U0,
    〈 inert_G_for_item G wa ya (typ_fun Tf U0)
       ⊢ tc /[-> wa] ; fs ; Sa & ya ~ item_fun tf ; e & wa ~ ya ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fs \u fv_heap Sa \u fv_renv e \u Lva ∶ T 〉
       ∼ trf_similarity typing_mode
       ∼> 〈 inert_G_for_item G wc yc (typ_fun Tf U0)
          ⊢ tc /[-> wc] ; fs ; Sa & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fs \u fv_heap Sa \u fv_renv e \u Lva ∶ T 〉
  /\  A1 = (typ_fun Tf U0)
  ) as Hsi1. {
  (* Left side *) 
  inversions H9. exists U0. split.
  forwards Hsi1: trf_similarity_config_lit_fun wc yc Hpr1'.
  eauto.
  eapply config_typed_valid; eauto. apply Hfr.
  sympl. fresh_chain_solve Fr.
  fresh_chain_solve Fr. fresh_chain_solve Fr. auto. auto.
  }

  assert (exists U0,
〈 inert_G_for_item G wc yc (typ_fun Tf U0)
       ⊢ tc /[-> wc] ; fs ; Sb & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb ∶ T 〉
       ∼ trf_similarity typing_mode
       ∼> 〈 inert_G_for_item G wb yb (typ_fun Tf U0)
          ⊢ tc /[-> wb] ; fs ; Sb & yb ~ item_fun tf ; e & wb ~ yb ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb ∶ T 〉 
  /\  A2 = (typ_fun Tf U0)
) as Hsi2.
  {
  (* Right side needs symmetry *)
    inversions H12. exists U0. split.

    forwards Hsi2: trf_similarity_config_lit_fun wc yc Hpr2'.
    eauto.
    eapply config_typed_valid; eauto. auto.
    sympl. fresh_chain_solve Fr.
    fresh_chain_solve Fr. fresh_chain_solve Fr.
    eapply TransformSimilaritySymmetric in Hsi2.
    auto. auto.
  }
  destructecs Hsi1.
  destructecs Hsi2.
  subst.

  set (Lv2 := 
    (((fv Tf \u fv tf) \u fv tc) \u fv fs \u fv_heap Sa \u fv_renv e \u Lva) \u
    (((fv Tf \u fv tf) \u fv tc) \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb) \u
    fv U \u ctx_dom G \u ctx_fv G
  ).

  forwards Hrety': trf_retype_fun' wc yc null Htyc Htyc0.
  eapply Htl0. eapply Htl. eauto.
  eapply Htl0. eapply Htl. eauto.
  eauto.

  forwards Hrety: Hrety'. splits; sympl; fresh_chain_solve Fr.
  clear Hrety'.
  destructecs Hrety.
  destructecs Hrety2.
  destructecs Hrety3.
  destructecs Hrety4.


  (* Similarity with the common type in the context - left side *)
  assert ( 
    〈 inert_G_for_item G wc yc (typ_fun Tf U0)
            ⊢ tc /[-> wc] ; fs ; Sa & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fs \u fv_heap Sa \u fv_renv e \u Lva ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
    〈 inert_G_for_item G wc yc (typ_fun Tf (seal_rec 1 null U))
            ⊢ tc /[-> wc] ; fs ; Sa & yc ~ item_fun tf ; e & wc ~ yc ; Lv2 ∶ T 〉
  ) as Hsi1a. {
    eapply trf_similarity_retype_Lv.
    inversions Hsi0; eauto.
    unfold Lv2.
    sympl. sympl in Hrety0. eapply Hrety0. unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsi1b: TransformSimilarityTransitive Hsi0 Hsi1a. clear Hsi0 Hsi1a.

  (* Similarity with the common type in the context - right side *)

  assert ( 
    〈 inert_G_for_item G wc yc (typ_fun Tf (seal_rec 1 null U))
            ⊢ tc /[-> wc] ; fs ; Sb & yc ~ item_fun tf ; e & wc ~ yc ; Lv2 ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
 〈 inert_G_for_item G wc yc (typ_fun Tf U1)
       ⊢ tc /[-> wc] ; fs ; Sb & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fs \u fv_heap Sb \u fv_renv e \u Lvb ∶ T 〉
  ) as Hsi2a. {

  eapply TransformSimilaritySymmetric.
 eapply trf_similarity_retype_Lv. 
 inversions Hsi3; eauto. eapply Hrety1. unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsi2b: TransformSimilarityTransitive Hsi2a Hsi3. clear Hsi3 Hsi2a.

  (* Solve the transformation with similarities on both sides *)
  eapply trf_lift_sim_conf_sim_lift_sim. eapply Hsi1b. 2: eapply Hsi2b.
  assert (wc \notin Lva). fresh_chain_solve Fr.
  assert (yc \notin Lva). fresh_chain_solve Fr.
  assert (wc \notin fv U /[# 1 ->> null]). {
    eapply fv_seal_notin. intro. subst. contradiction. fresh_chain_solve Fr.
  }
  assert (yc \notin fv U /[# 1 ->> null]). {
    eapply fv_seal_notin. intro. subst. contradiction. fresh_chain_solve Fr.
  }

  (* Transformaption happens on the focus *)
  eapply trf_lift_idtyp_conf_heap.
 - (* Left config accessories *)
    eapply ty_config_accessories_tred_let_fun; eauto 1.
    eapply ty_config_to_accessories; eauto.
    fresh_solve; try eapply fv_seal_notin;
    fresh_chain_solve Fr.

    unfold Lv2; union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_trans. eapply fv_seal_subset.
    union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_singleton. union_in_solver.
  -  (* Right config accessories *)
    eapply ty_config_accessories_tred_let_fun; eauto 1.
    eapply ty_config_to_accessories; eauto.
    fresh_solve; try eapply fv_seal_notin;
    fresh_chain_solve Fr.

    unfold Lv2; union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_trans. eapply fv_seal_subset.
    union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_singleton. union_in_solver.
  - (* Left heap accessories *)
    eapply heap_correspond_accessories_tred_let_fun. auto. eapply heap_correspond_to_accessories; eauto. fresh_solve; try eapply fv_seal_notin; fresh_chain_solve Fr.
  - (* Right heap accessories *)
    eapply heap_correspond_accessories_tred_let_fun. auto. eapply heap_correspond_to_accessories; eauto. fresh_solve; try eapply fv_seal_notin; fresh_chain_solve Fr.
  - (* Heap transformation *)
    eapply trf_lift_idtyp_heap_n.
    + (* Transformation of old items *)
    eapply trf_lift_idtyp_heap_epush.
    eauto.
    eauto.
    eauto.
    fresh_chain_solve Fr.
    ctx_simpl_dom.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    eapply trf_lift_idtyp_heap_Gpush.
    eauto.
    eauto.
    eauto.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    eauto.
  
    + (* Typing of the new item *)

      econstructor. eauto.
      eapply ctx_binds_n. fresh_chain_solve Fr.
      eapply ctx_binds_e.
      eapply binds_single_eq.
      intros.
      eapply ewkn_push.
      ctx_split_between yc x.
      eapply weaken_ty.
      ctx_flatten. eapply Hrety3. eauto.
      eauto.
      ctx_flatten. fresh_chain_solve2 Fr.
      fresh_chain_solve Fr.
      fresh_chain_solve Fr.


      apply all_tfree_k_Gpush. apply all_tfree_k_Gpush. hnf in Htft. hnf. eauto. fresh_chain_solve Fr.  fresh_chain_solve2 Fr.

 
  - (* Stack transformation*)
    eapply ty_stack_epush; try fresh_chain_solve Fr.
    eapply ty_stack_push; try (ctx_simpl_dom; fresh_chain_solve Fr).
    eapply ty_stack_push; try fresh_chain_solve Fr. 
    eauto.
  - (* Focus tranformation *)
    eapply @renaming_fresh with (L := (L \u \{wc} \u \{yc} \u dom e \u ctx_dom G \u LNN)) (vk:=v_let) (T:=typ_fun Tf U /[# 1 ->> null]).
      *
        apply CloseUnboundSubstIntro.
      * eauto.
      *
        apply ewkn_ty_trm.
      * fresh_chain_solve2 Fr.
      * fresh_chain_solve2 Fr.
      * 
        intros.
        ctx_split_between yc z.
        assert (Weakening (ty_trm typing_mode)); eauto.
        apply @weaken_ty. eauto. sympl.    eapply ewkn_push. eapply Hrety2.  simpl; eauto.
    
        eauto. fresh_chain_solve Fr. eauto. fresh_chain_solve2 Fr.
      *
        eauto.
      * eauto.
Qed.

Lemma transformation_lifted_focus_or_preserved_stack_lit_2fun (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
    G Tf tf tr fsa fsb S e Lva C3 Lvb C4 T Tt  G3 G4:
  〈 G ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ∼ trf_lift_idtyp_stack typing_mode TL ∼> 〈 G ⁏ e ⊢ fsb ∶ (Tt, T) 〉->
  G ⊢ trm_lit (lit_fun Tf tf) tr ; fsa ; S ; e ; Lva ∶ T ↦ G3 ⊢ C3 ->
  G ⊢ trm_lit (lit_fun Tf tf) tr ; fsb ; S ; e ; Lvb ∶ T ↦ G4 ⊢ C4 ->
  heap_correspond G e S ->  
  G ⁏ e ⊢ trm_lit (lit_fun Tf tf) tr ∶ Tt ->
  〈 G3 ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 G4 ⊢ C4 ∶ T 〉.
Proof.

  introv Htrf Hpr1 Hpr2 Htyt Htys .


  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  step_invert_simple Hpr1' HtyC3.
  step_invert_simple Hpr2' HtyC4.

  forwards Htl: ty_trm_lit_inv Htyt. eauto.
  destructecs Htl.
  
  lets Htylit: Htl0.
  inversions Htylit.

  assert (has_null Lva) as Hnull. inversions Htyc. auto.
  destructc Hnull as null Hnull.

  rename t3 into tf, t4 into tc, T1 into Tf.
  rename w into wa, y into ya, w0 into wb, y0 into yb.

  (* Pick fresh variables to use in the transformation *)
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr. eauto. eapply HeapCorrespondence.heap_correspond_to_ok_e; eauto. eapply HeapCorrespondence.heap_correspond_to_ok_s; eauto.
  fresh_chain_pick Fr wc.
  fresh_chain_pick Fr yc.

  (* Assert similarity from both sides, renaming the chosen variables to the fresh variables *)

  assert ( exists U0,
    〈 inert_G_for_item G wa ya (typ_fun Tf U0)
       ⊢ tc /[-> wa] ; fsa ; S & ya ~ item_fun tf ; e & wa ~ ya ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fsa \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
       ∼ trf_similarity typing_mode
       ∼> 〈 inert_G_for_item G wc yc (typ_fun Tf U0)
          ⊢ tc /[-> wc] ; fsa ; S & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fsa \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
  /\  A1 = (typ_fun Tf U0)
  ) as Hsi1. {
  (* Left side *) 
  inversions H9. exists U0. split.
  forwards Hsi1: trf_similarity_config_lit_fun wc yc Hpr1'.
  eauto.
  eapply config_typed_valid; eauto. apply Hfr.
  sympl. fresh_chain_solve Fr.
  fresh_chain_solve Fr. fresh_chain_solve Fr. auto. auto.
  }

  assert (exists U0,
〈 inert_G_for_item G wc yc (typ_fun Tf U0)
       ⊢ tc /[-> wc] ; fsb ; S & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fsb \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉
       ∼ trf_similarity typing_mode
       ∼> 〈 inert_G_for_item G wb yb (typ_fun Tf U0)
          ⊢ tc /[-> wb] ; fsb ; S & yb ~ item_fun tf ; e & wb ~ yb ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fsb \u fv_heap S \u fv_renv e \u Lvb ∶ T 〉 
  /\  A2 = (typ_fun Tf U0)
) as Hsi2.
  {
  (* Right side needs symmetry *)
    inversions H12. exists U0. split.

    forwards Hsi2: trf_similarity_config_lit_fun wc yc Hpr2'.
    eauto.
    eapply config_typed_valid; eauto. auto.
    sympl. fresh_chain_solve Fr.
    fresh_chain_solve Fr. fresh_chain_solve Fr.
    eapply TransformSimilaritySymmetric in Hsi2.
    auto. auto.
  }
  destructecs Hsi1.
  destructecs Hsi2.
  subst.

  (* Typing of stack on both sides *)
  assert (ty_stack G e Tt fsb T) as Htyfsb. {
      intros.
             eapply trf_inverse_intro in Htrf.
          forwards Htrfty: TransformLiftIdtypStackTyped TLTI.
          rewrite <- trf_inverse_lift_idtyp_stack in Htrfty.
          eapply Htrfty in Htrf. hnf in Htrf. sympl in Htrf. eapply Htrf.
          eauto. 
  }
  assert (ty_stack G e Tt fsa T) as Htyfsa.
  {
      intros.
          forwards Htrfty: TransformLiftIdtypStackTyped TLT.
          eapply Htrfty in Htrf. hnf in Htrf. eapply Htrf.
  }

  set (Lv2 :=
    (((fv Tf \u fv tf) \u fv tc) \u fv fsa \u fv_heap S \u fv_renv e \u Lva) \u
    (((fv Tf \u fv tf) \u fv tc) \u fv fsb \u fv_heap S \u fv_renv e \u Lvb) \u
    fv U \u ctx_dom G \u ctx_fv G
  ).

  forwards Hrety': trf_retype_fun' wc yc null Htyc Htyc0.
  eapply Htl0. eapply Htl. eapply Htyfsa.
  eapply Htl0. eapply Htl. eapply Htyfsb.
  eauto.

  forwards Hrety: Hrety'. splits; sympl; fresh_chain_solve Fr.
  clear Hrety'.
  destructecs Hrety.
  destructecs Hrety2.
  destructecs Hrety3.
  destructecs Hrety4.


  (* Similarity with the common type in the context - left side *)
  assert ( 
    〈 inert_G_for_item G wc yc (typ_fun Tf U0)
            ⊢ tc /[-> wc] ; fsa ; S & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fsa \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
    〈 inert_G_for_item G wc yc (typ_fun Tf (seal_rec 1 null U))
            ⊢ tc /[-> wc] ; fsa ; S & yc ~ item_fun tf ; e & wc ~ yc ; Lv2 ∶ T 〉
  ) as Hsi1a. {
    eapply trf_similarity_retype_Lv.
    inversions Hsi0; eauto.
    unfold Lv2.
    sympl. sympl in Hrety0. eapply Hrety0. unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsi1b: TransformSimilarityTransitive Hsi0 Hsi1a. clear Hsi0 Hsi1a.

  (* Similarity with the common type in the context - right side *)

  assert ( 
    〈 inert_G_for_item G wc yc (typ_fun Tf (seal_rec 1 null U))
            ⊢ tc /[-> wc] ; fsb ; S & yc ~ item_fun tf ; e & wc ~ yc ; Lv2 ∶ T 〉
    ∼ trf_similarity typing_mode ∼> 
 〈 inert_G_for_item G wc yc (typ_fun Tf U1)
       ⊢ tc /[-> wc] ; fsb ; S & yc ~ item_fun tf ; e & wc ~ yc ; fv (trm_lit (lit_fun Tf tf) tc) \u fv fsb \u fv_heap S\u fv_renv e \u Lvb ∶ T 〉
  ) as Hsi2a. {
    eapply TransformSimilaritySymmetric.
    eapply trf_similarity_retype_Lv. 
    inversions Hsi3; eauto. eapply Hrety1. unfold Lv2. sympl. union_sub_solver.
  }

  forwards Hsi2b: TransformSimilarityTransitive Hsi2a Hsi3. clear Hsi3 Hsi2a.

  (* Solve the transformation with similarities on both sides *)
  eapply trf_lift_sim_conf_sim_lift_sim. eapply Hsi1b. 2: eapply Hsi2b.
  assert (wc \notin Lva). fresh_chain_solve Fr.
  assert (yc \notin Lva). fresh_chain_solve Fr.
  assert (wc \notin fv U /[# 1 ->> null]). {
    eapply fv_seal_notin. intro. subst. contradiction. fresh_chain_solve Fr.
  }
  assert (yc \notin fv U /[# 1 ->> null]). {
    eapply fv_seal_notin. intro. subst. contradiction. fresh_chain_solve Fr.
  }

  (* Transformaption happens on the focus *)
  eapply trf_lift_idtyp_conf_stack with (Tt:=Tt /[# 0 ->> null]).
 - (* Left config accessories *)
    eapply ty_config_accessories_tred_let_fun; eauto 1.
    eapply ty_config_to_accessories; eauto.
    fresh_solve; try eapply fv_seal_notin;
    fresh_chain_solve Fr.

    unfold Lv2; union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_trans. eapply fv_seal_subset.
    union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_singleton. union_in_solver.
  -  (* Right config accessories *)
    eapply ty_config_accessories_tred_let_fun; eauto 1.
    eapply ty_config_to_accessories; eauto.
    fresh_solve; try eapply fv_seal_notin;
    fresh_chain_solve Fr.

    unfold Lv2; union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_trans. eapply fv_seal_subset.
    union_sub_solver_destruct_ls; try union_sub_solver . eapply subset_singleton. union_in_solver.
  - (* Heap *)
    eapply tred_lit_fun_hc; eauto.
    sympl.
    fresh_solve; fresh_chain_solve Fr.
  -

    assert(〈 inert_G_for_item G wc yc (typ_fun Tf U /[# 1 ->> null]) ⁏ e ⊢ fsa ∶ (Tt /[# 0 ->> null], T) 〉 ∼ TransformWknEnvPush (A:=stack) ∼> 〈 inert_G_for_item G wc yc (typ_fun Tf U /[# 1 ->> null]) ⁏ e & wc ~ yc ⊢ fsa ∶ (Tt /[# 0 ->> null], T) 〉).
    econstructor; fresh_chain_solve2 Fr.

    assert(〈 inert_G_for_item G wc yc (typ_fun Tf U /[# 1 ->> null]) ⁏ e ⊢ fsb ∶ (Tt /[# 0 ->> null], T) 〉 ∼ TransformWknEnvPush (A:=stack) ∼> 〈 inert_G_for_item G wc yc (typ_fun Tf U /[# 1 ->> null]) ⁏ e & wc ~ yc ⊢ fsb ∶ (Tt /[# 0 ->> null], T) 〉).
    econstructor; fresh_chain_solve2 Fr.

    eapply WknEnvPushCompatTransformationStack. eauto. 2: eauto. 2: eauto.

    assert(〈 G § yc / v_loc ~ (typ_fun Tf U /[# 1 ->> null]) ⁏ e ⊢ fsa ∶ (Tt /[# 0 ->> null], T) 〉  ∼ TransformWknCtxPush (A:=stack) ∼> 〈 inert_G_for_item G wc yc (typ_fun Tf U /[# 1 ->> null]) ⁏ e ⊢ fsa ∶ (Tt /[# 0 ->> null], T) 〉 ).
    econstructor; fresh_chain_solve2 Fr.

    assert(〈 G § yc / v_loc ~ (typ_fun Tf U /[# 1 ->> null])  ⁏ e ⊢ fsb ∶ (Tt /[# 0 ->> null], T) 〉  ∼ TransformWknCtxPush (A:=stack) ∼> 〈 inert_G_for_item G wc yc (typ_fun Tf U /[# 1 ->> null]) ⁏ e ⊢ fsb ∶ (Tt /[# 0 ->> null], T) 〉 ).
    econstructor; fresh_chain_solve2 Fr.

    eapply WknCtxPushCompatTransformationStack. eauto. 2: eauto. 2: eauto.


    assert(〈 G ⁏ e ⊢ fsa ∶ (Tt /[# 0 ->> null], T) 〉  ∼ TransformWknCtxPush (A:=stack) ∼> 〈 G § yc / v_loc ~ (typ_fun Tf U /[# 1 ->> null])  ⁏ e ⊢ fsa ∶ (Tt /[# 0 ->> null], T) 〉 ).
    econstructor; fresh_chain_solve2 Fr.

    assert(〈 G  ⁏ e ⊢ fsb ∶ (Tt /[# 0 ->> null], T) 〉  ∼ TransformWknCtxPush (A:=stack) ∼> 〈G § yc / v_loc ~ (typ_fun Tf U /[# 1 ->> null])  ⁏ e ⊢ fsb ∶ (Tt /[# 0 ->> null], T) 〉 ).
    econstructor; fresh_chain_solve2 Fr.

    eapply WknCtxPushCompatTransformationStack. eauto. 2: eauto. 2: eauto.

    eapply trf_lift_idtyp_stack_seal_closed.

    eauto.
    inversions Htyc. inversions H33. auto.
    inversions Htyc0. inversions H33. auto.
    inversions Htyc0. auto.
    inversions Htyc0. auto.

    eapply Htrf.
    hnf; eauto.
    hnf; eauto.
    hnf; eauto.

  - eapply @renaming_fresh with (L := (L \u \{wc} \u \{yc} \u dom e \u ctx_dom G \u LNN)) (vk:=v_let) (T:=typ_fun Tf U /[# 1 ->> null]).
      *
        apply CloseUnboundSubstIntro.
      * eauto.
      *
        apply ewkn_ty_trm.
      * fresh_chain_solve2 Fr.
      * fresh_chain_solve2 Fr.
      * 
        intros.
        ctx_split_between yc z.
        assert (Weakening (ty_trm typing_mode)); eauto.
        apply @weaken_ty. eauto. sympl.    eapply ewkn_push.
        eapply Hrety2.  simpl; eauto.
        eauto. fresh_chain_solve Fr. eauto. fresh_chain_solve2 Fr.
      *
        eauto.
      * eauto.
Qed.


End WithTypingMode.