(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation Weakening EWeakening Substitution TFree.
Require Import TransformationSimilarity TransformationPreservation TransformationTerm.
Require Import TransformationConfig OperationalSemantics TypedOperationalSemantics.
Require Import ContextSimilarity ConfigSimilarity.
Require Import ConfigTyping ConfigTypingAccessories TypedProgress.
Require Import TypedPreservation.
Require Import TransformationIterateSim.
Require Import Similarity TypingSimilarity.

Require Import TransformationLifting.
Require Import FreshChain FreshChainTactics.
Require Import TypedOperationalSemantics InertContext.
Require Import Similarity SyntaxSimilarity.
Require Import ExecutionTactics Renaming TypingHelpers.

Require Import NothingFresh TypedPreservation LetLet Defs.
Require Import HeapCorrespondence ConfigVarkinds.
Require Import HeapUpdate ConfigValid.

Require Import TransformationSimilaritySteps ConfigTypingSteps TransformationConfigSteps.
Require Import Defs LetLet EnvCorrespondence.
Require Import Subenvironments Narrowing.
Require Import Defree.
Require Import Visibility.
Require Import TypedPreservationLemmas TransformationTyping.

Section WithTypingMode.
Context (typing_mode: Mode).

(** * Preservation of lifted transformations during reduction involving objects. *)

Lemma transformation_open_def_tyq (TL : TermTransform) (TLN: TransformNarrowingCompat typing_mode typ trm TL):
    forall G Go e (da db : def) (D : dec) w y n,
    inert G -> renv_corr G e -> ctx_ok (G §§& Go) -> ok e ->
   〈 G §§ Go ⁏ e ⊢ da ∶ open_rec n w D 〉 ∼ trf_local_idtyp_def typing_mode TL ∼> 〈G §§ Go ⁏ e ⊢ db ∶ open_rec n w D 〉->
    e ⊢ w ≈ y ->
    w \notin fv D ->
    typ_indep_b n (typ_rcd D) ->
   〈 G §§ Go ⁏ e ⊢ da ∶ open_rec n y D 〉 ∼ trf_local_idtyp_def typing_mode TL ∼> 〈G §§ Go ⁏ e ⊢ db ∶ open_rec n y D 〉.
Proof.
  introv Hin Hcor Hok Hoke Htyd Htyv Hfr Hindep.
  remember (open_rec n w D) as Z.
  inversions Htyd.
  rename H3 into HeqZ.
    apply open_dec_met_inv in HeqZ as HeqD.
    destruct HeqD as [TC' [TZ' [TR' Heq]]]. subst.
    simpl in HeqZ. inversions HeqZ. simpl. rename H0 into Hty.
     simpl in Hfr. eapply trf_local_idtyp_def_met with (L := L \u ctx_dom (G §§& Go)).
    introv Hfrxz Hneq Hfrxc.
    assert (HfrxzL : xz \notin L) by auto.
    assert (HfrxcL : xc \notin L) by auto.
    specialize (Hty _ HfrxzL _ Hneq HfrxcL); clear HfrxzL HfrxcL.
    assert (Hsubctx :
              subctx e
                (ctx_obj_met typing_mode G Go xz (open_rec_typ n y TZ')
                         xc (open_rec_typ (S n) y TC'))
                (ctx_obj_met typing_mode G Go xz (open_rec_typ n w TZ')
                         xc (open_rec_typ (S n) w TC'))).
    { unfold ctx_obj_met.
      assert (Hindep' : typ_indep_b n TZ').
        { inversions_tyq Hindep. auto. }
        destruct Go; unfold ctx_obj in *;
        eapply subctx_grow;
        try eapply subctx_grow; try solve [destruct (mode_visibility' typing_mode); auto].
      - inversions_tyq Hindep.
        rename H3 into HindepTC, H5 into HindepTZ, H6 into HindepTR.
        apply subtyp_tyq. eapply tyq_typ_open.
        eapply tyq_typ_open_rec_indep; eauto 2.
      - iapply subtyp_and; auto.
        inversions_tyq Hindep.
        rename H3 into HindepTC, H5 into HindepTZ, H6 into HindepTR.
        apply subtyp_tyq. eapply tyq_typ_open.
        eapply tyq_typ_open_rec_indep; eauto 2.
      - iapply subtyp_and; auto.
        inversions_tyq Hindep.
        rename H3 into HindepTC, H5 into HindepTZ, H6 into HindepTR.
        apply subtyp_tyq. eapply tyq_typ_open.
        eapply tyq_typ_open_rec_indep; eauto 2.
      }

    forwards: TransformNarrowingCompatSubctx.
    2: eapply NarrowingCompatTransformationTrm.
    2: eapply TLN.
    eapply TrfLocalIdtypTrmEnsuresCond.

    forwards: H Hty.
    econstructor. auto. eapply Hsubctx.
    econstructor. auto. eapply Hsubctx.
    econstructor; eauto.

    eapply trf_local_idtyp_trm_subsumption. eauto.
    eapply subtyp_tyq; eauto 2. 
    assert (Hindep' : typ_indep_b (S (S n)) TR').
    { inversions_tyq Hindep; auto. }
    assert (Htyq : e ⊢ (open_rec (S (S n)) w TR') ≈ (open_rec (S (S n)) y TR')).
    { eapply tyq_typ_open_rec_indep; eauto 2. }
    apply tyq_typ_open_rec with (n := 1) (y := xc) in Htyq.
    apply tyq_typ_open_rec with (n := 0) (y := xz) in Htyq. auto.
Qed.


Lemma transformation_open_defs_tyq (TL : TermTransform) (TLN: TransformNarrowingCompat typing_mode typ trm TL):
forall G Go e (dsa dsb : defs) (Ds : typ) w y n,
    inert G -> renv_corr G e -> ctx_ok (G §§& Go) -> ok e->
   〈 G §§ Go ⁏ e ⊢ dsa ∶ open_rec n w Ds 〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈G §§ Go ⁏ e ⊢ dsb ∶ open_rec n w Ds 〉->
    e ⊢ w ≈ y ->
    w \notin fv Ds ->
    typ_indep_b n Ds ->
   〈 G §§ Go ⁏ e ⊢ dsa ∶ open_rec n y Ds 〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈G §§ Go ⁏ e ⊢ dsb ∶ open_rec n y Ds 〉.
Proof.
  introv Hin Hcor Hok Hoke Htyd Htyv Hfr Hindep.
  remember (〈 G §§ Go ⁏ e ⊢ dsa ∶ Ds /[# n -> w]〉 ).
  remember (〈 G §§ Go ⁏ e ⊢ dsb ∶ Ds /[# n -> w]〉 ).
  reverts dsa dsb Ds.
  induction Htyd; intros; inversions Heqt; inversions Heqt0.

  - rename H1 into Hrcdeq.
    apply open_typ_rcd_inv in Hrcdeq as ?. destruct H0. subst. simpl in Hrcdeq.
    inversions Hrcdeq.  simpl. econstructor. eapply transformation_open_def_tyq; eauto.
  - apply open_and_inv in H4 as HeqD.
    destruct HeqD as [T' [U' Heq]]. subst. simpl in *.
    inversions H4. rename H6 into HeqZ.
    apply open_typ_rcd_inv in HeqZ as HeqD.
    destruct HeqD as [D' Heq]. subst. simpl in HeqZ. inversions HeqZ.
    simpl in *.
    eapply trf_local_idtyp_defs_consl; eauto 2.
    + eapply IHHtyd; eauto 3 using notin_union_r2, notin_union_r1.
      inversions_tyq Hindep; auto.
    + eapply open_def_tyq; eauto 3 using notin_union_r2, notin_union_r1.
      inversions_tyq Hindep; auto.
  - apply open_and_inv in H4 as HeqD.
    destruct HeqD as [T' [U' Heq]]. subst. simpl in *.
    inversions H4. rename H6 into HeqZ.
    apply open_typ_rcd_inv in HeqZ as HeqD.
    destruct HeqD as [D' Heq]. subst. simpl in HeqZ. inversions HeqZ.
    simpl in *.
    eapply trf_local_idtyp_defs_consr; eauto 2.
      
    + eapply transformation_open_def_tyq; eauto 3 using notin_union_r2, notin_union_r1.
      inversions_tyq Hindep; auto.
    + eapply open_defs_tyq; eauto 3 using notin_union_r2, notin_union_r1.
      inversions_tyq Hindep; auto.
Qed.

Lemma trf_instantiation_met0z (TL : TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
L G e w y z T S (t t0: trm) TZ TC TR L0:
  z \notin (L \u fv T \u fv S \u ctx_fv G \u dom e \u fv_env e \u \{ w } \u \{ y }) ->
  ok (e & w ~ y) ->
  ctx_ok ((inert_G_for_obj G w y T S § z / v_self ~ T /[-> z])) ->
  typ_indep_b 0 T ->
  (forall xz : var,
     xz \notin L0 ->
     forall xc : var,
     xc <> xz ->
     xc \notin L0 ->
  〈 ctx_obj_met typing_mode (G § z / v_self ~ T /[-> z]) (objctx_heap z T) xz TZ xc TC ⁏ e ⊢ open_met xz xc t ∶ open_met xz xc TR 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 ctx_obj_met typing_mode (G § z / v_self ~ open z T) (objctx_heap z T) xz TZ xc TC ⁏ e ⊢ open_met xz xc t0 ∶ open_met xz xc TR 〉
  ) ->
  forall xz, xz \notin (L0 \u \{ z }\u \{ w}\u \{ y} \u (ctx_dom G)) ->
  forall xc, xc <> xz -> xc \notin (L0 \u \{ z }\u \{ w}\u \{ y} \u (ctx_dom G)) ->
  〈 (inert_G_for_obj G w y T S § z / v_self ~ T /[-> z]) §& ctx_met_refined_heap (vis_hide typing_mode ctx_empty) xz TZ xc TC z T ⁏ e & w ~ y ⊢ t /[# 1 -> xc] /[-> xz] ∶ TR /[# 1 -> xc] /[-> xz] 〉 
  ∼ trf_local_idtyp_trm typing_mode TL ∼>
  〈 (inert_G_for_obj G w y T S § z / v_self ~ T /[-> z]) §& ctx_met_refined_heap (vis_hide typing_mode ctx_empty) xz TZ xc TC z T ⁏ e & w ~ y ⊢ t0 /[# 1 -> xc] /[-> xz] ∶ TR /[# 1 -> xc] /[-> xz] 〉.
Proof.
  introv Hn Hokw HokG Hind Hty.
  intros.
  forwards Htyx: Hty xz xc; fweauto.

  assert (
    ctx_ok (ctx_met_refined_heap
      match mode_visibility' typing_mode with
      | mode_visibility_all => G §ˢ z ~ T /[-> z]
      | mode_visibility_hide_method => ctx_hide (G §ˢ z ~ T /[-> z])
      end xz TZ xc TC z T)
  ).
  {
    inversions HokG.
    inversions H4.
    inversions H5.
    destruct (mode_visibility' typing_mode) eqn: Hm; ctx_flatten.
    constructor; eauto.
    constructor; eauto.
  
  }

  assert (
    〈  ctx_obj_met typing_mode (G § z / v_self ~ T /[-> z]) (objctx_heap z T) xz TZ xc TC ⁏ e & w ~ y ⊢ open_met xz xc t ∶ open_met xz xc TR〉 
    ∼ trf_local_idtyp_trm typing_mode TL ∼>
    〈  ctx_obj_met typing_mode (G § z / v_self ~ T /[-> z]) (objctx_heap z T) xz TZ xc TC ⁏ e & w ~ y ⊢ open_met xz xc t0 ∶ open_met xz xc TR  〉
  ) as Htye. {
    eapply trf_local_idtyp_trm_ewkn_push. eauto.
    eapply Htyx; eauto.
    sympls.
    auto.
    auto.
    eapply ok_push_inv in Hokw. destructecs Hokw.
    auto.
  }
  sympls.
  destruct (mode_visibility' typing_mode) eqn: Hm.
  * (* basic *)
    ctx_split_between y z.
    eapply WeakeningCompatTransformationTrm.
    ++
      eauto.
    ++
      apply Htye.
    ++
      ctx_split_between z y.
      replace ( G §& ctx_met_refined_heap (inert_G_for_obj ctx_empty w y T S §ˢ z ~ T /[-> z]) xz TZ xc TC z T ) with 
      ( G §& inert_G_for_obj ctx_empty w y T S §& ctx_met_refined_heap (ctx_empty §ˢ z ~ T /[-> z]) xz TZ xc TC z T ).
      2: {
        ctx_flatten. auto.
      }
      econstructor; eauto.
      ctx_flatten. eauto.
    ++
      ctx_split_between z y.
      replace ( G §& ctx_met_refined_heap (inert_G_for_obj ctx_empty w y T S §ˢ z ~ T /[-> z]) xz TZ xc TC z T ) with 
      ( G §& inert_G_for_obj ctx_empty w y T S §& ctx_met_refined_heap (ctx_empty §ˢ z ~ T /[-> z]) xz TZ xc TC z T ).
      2: {
        ctx_flatten. auto.
      }
      econstructor; eauto.
      ctx_flatten. eauto.
    ++ hnf; eauto.
      * (* basic *)
        ctx_split_between y z.
        eapply WeakeningCompatTransformationTrm.
    ++
      eauto.
    ++
      apply Htye.
    ++
      ctx_split_between z y.
      rewrite <- ctx_concat_hide.
      rewrite <- ctx_concat_hide.

      replace (  ctx_met_refined_heap (G §& ctx_hide (inert_G_for_obj ctx_empty w y T S §ˢ z ~ T /[-> z])) xz TZ xc TC z T ) with 
      ( G §& ( inert_G_for_obj ctx_empty w y T S) §&(ctx_met_refined_heap ( ctx_hide( ctx_empty §ˢ z ~ T /[-> z])) xz TZ xc TC z T) ).
      2: {
        ctx_flatten. auto.
      }

      replace ( ctx_met_refined_heap (G §& ctx_hide (ctx_empty §ˢ z ~ T /[-> z])) xz TZ xc TC z T ) with 
      (G §&  ctx_met_refined_heap (ctx_hide (ctx_empty §ˢ z ~ T /[-> z])) xz TZ xc TC z T).
      2: {
        ctx_flatten. auto.
      }
 
      econstructor; eauto.
      ctx_flatten.
      inversions HokG.
      inversions H5.
      inversions H6.
      destruct (mode_visibility' typing_mode) eqn: Hmv; ctx_flatten.
      constructor; eauto.
      constructor; eauto.
      constructor; eauto.
      constructor; eauto.
 
    ++
      ctx_split_between z y.
      rewrite <- ctx_concat_hide.
      rewrite <- ctx_concat_hide.

      replace (  ctx_met_refined_heap (G §& ctx_hide (inert_G_for_obj ctx_empty w y T S §ˢ z ~ T /[-> z])) xz TZ xc TC z T ) with 
      ( G §& ( inert_G_for_obj ctx_empty w y T S) §&(ctx_met_refined_heap ( ctx_hide( ctx_empty §ˢ z ~ T /[-> z])) xz TZ xc TC z T) ).
      2: {
        ctx_flatten. auto.
      }

      replace ( ctx_met_refined_heap (G §& ctx_hide (ctx_empty §ˢ z ~ T /[-> z])) xz TZ xc TC z T ) with 
      (G §&  ctx_met_refined_heap (ctx_hide (ctx_empty §ˢ z ~ T /[-> z])) xz TZ xc TC z T).
      2: {
        ctx_flatten. auto.
      }
      
      econstructor; eauto.
      ctx_flatten.
      inversions HokG.
      inversions H5.
      inversions H6.
      destruct (mode_visibility' typing_mode) eqn: Hmv; ctx_flatten.
      constructor; eauto.
      constructor; eauto.
      constructor; eauto.
      constructor; eauto.
    ++
      hnf; eauto.
Qed.


Lemma trf_instantiation_met0 (TL : TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
L G e w y z T S (t t0: trm) TZ TC TR L0:
  z \notin (L \u fv T \u fv S \u ctx_fv G \u dom e \u fv_env e \u \{ w } \u \{ y }) ->
  ok (e & w ~ y) ->
  ctx_ok ((inert_G_for_obj G w y T S § z / v_self ~ T /[-> z])) ->
  typ_indep_b 0 T ->
  (forall xz : var,
     xz \notin L0 ->
     forall xc : var,
     xc <> xz ->
     xc \notin L0 ->
  〈 ctx_obj_met typing_mode (G § z / v_self ~ T /[-> z]) (objctx_heap z T) xz TZ xc TC ⁏ e ⊢ open_met xz xc t ∶ open_met xz xc TR 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 ctx_obj_met typing_mode (G § z / v_self ~ open z T) (objctx_heap z T) xz TZ xc TC ⁏ e ⊢ open_met xz xc t0 ∶ open_met xz xc TR 〉
  ) ->
  forall xz, xz \notin (L0 \u \{ z }\u \{ w}\u \{ y} \u (ctx_dom G)) ->
  forall xc, xc <> xz -> xc \notin (L0 \u \{ z }\u \{ w}\u \{ y} \u (ctx_dom G)) ->
  〈 inert_G_for_obj G w y T S §& (subst_ctx z w (ctx_met_refined_heap (vis_hide typing_mode ctx_empty) xz TZ xc TC z T)) ⁏ e & w ~ y ⊢ ((open_met xz xc t) /[ z -> w]) ∶ ((open_met xz xc TR) /[ z -> w]) 〉 
  ∼ trf_local_idtyp_trm typing_mode TL ∼>
  〈 inert_G_for_obj G w y T S §& (subst_ctx z w (ctx_met_refined_heap (vis_hide typing_mode ctx_empty) xz TZ xc TC z T)) ⁏ e & w ~ y ⊢ ((open_met xz xc t0) /[ z -> w]) ∶ ((open_met xz xc TR) /[ z -> w]) 〉.
Proof.
  introv Hn Hokw HokG Hind Hty.
  intros.

  assert (
    inert_G_for_obj G w y T S
    §& subst_ctx z w
      (ctx_met_refined_heap
          match mode_visibility' typing_mode with
          | mode_visibility_all => ctx_empty
          | mode_visibility_hide_method => ctx_hide ctx_empty
          end xz TZ xc TC z T) ⁏ e  & w ~ y  ⊢ w ∶ T /[-> z] /[ z -> w]
  ) as Htyw.
  {
    eapply instantiation_met0_sub; eauto.

  }
  assert (
  ctx_ok
    ((inert_G_for_obj G w y T S §ˢ z ~ T /[-> z])
    §& ctx_met_refined_heap
          match mode_visibility' typing_mode with
          | mode_visibility_all => ctx_empty
          | mode_visibility_hide_method => ctx_hide ctx_empty
          end xz TZ xc TC z T)
  ) as HokG3.
  {
    intros. destruct (mode_visibility' typing_mode). simpl. eauto.
    simpl. eauto. ctx_flatten in HokG.
    inversions HokG.
    inversions H4.
    inversions H5.
    constructor; eauto.
    constructor; eauto.
  }

  assert (
    TransformEWeakening
      (〈 (inert_G_for_obj G w y T S §ˢ z ~ T /[-> z])
            §& ctx_met_refined_heap
                match mode_visibility' typing_mode with
                | mode_visibility_all => ctx_empty
                | mode_visibility_hide_method => ctx_hide ctx_empty
                end xz TZ xc TC z T ⁏ e & w ~ y ⊢  t /[# 1 -> xc] /[-> xz] ∶ TR /[# 1 -> xc] /[-> xz]  〉)
        (〈 (inert_G_for_obj G w y T S §ˢ z ~ T /[-> z])
            §& ctx_met_refined_heap
                match mode_visibility' typing_mode with
                | mode_visibility_all => ctx_empty
                | mode_visibility_hide_method => ctx_hide ctx_empty
                end xz TZ xc TC z T ⁏ e & w ~ y & z ~ w ⊢  t /[# 1 -> xc] /[-> xz] ∶ TR /[# 1 -> xc] /[-> xz]  〉)
  ) as Hwkz.
  {
    eapply trf_ewkn_push.
    eauto.
    eauto.
  }
  assert (
  TransformEWeakening
     (〈 (inert_G_for_obj G w y T S §ˢ z ~ T /[-> z])
          §& ctx_met_refined_heap
               match mode_visibility' typing_mode with
               | mode_visibility_all => ctx_empty
               | mode_visibility_hide_method => ctx_hide ctx_empty
               end xz TZ xc TC z T ⁏ e & w ~ y ⊢  t0 /[# 1 -> xc] /[-> xz] ∶ TR /[# 1 -> xc] /[-> xz]  〉)
       (〈 (inert_G_for_obj G w y T S §ˢ z ~ T /[-> z])
          §& ctx_met_refined_heap
               match mode_visibility' typing_mode with
               | mode_visibility_all => ctx_empty
               | mode_visibility_hide_method => ctx_hide ctx_empty
               end xz TZ xc TC z T ⁏ e & w ~ y & z ~ w ⊢  t0 /[# 1 -> xc] /[-> xz] ∶ TR /[# 1 -> xc] /[-> xz]  〉)
  ) as Hwkz0.
  {
    eapply trf_ewkn_push.
    eauto.
    eauto.
  }

  assert ( forall (ax: trm) (Tx: typ), 
  TransformSubst typing_mode w T /[-> z]
          (〈 (inert_G_for_obj G w y T S §ˢ z ~ T /[-> z])
              §& ctx_met_refined_heap match mode_visibility' typing_mode with
                                      | mode_visibility_all => ctx_empty
                                      | mode_visibility_hide_method => ctx_hide ctx_empty
                                      end xz TZ xc TC z T ⁏ e & w ~ y & z ~ w ⊢ ax ∶ Tx 〉)
          (〈 inert_G_for_obj G w y T S
              §& subst_ctx z w
                  (ctx_met_refined_heap match mode_visibility' typing_mode with
                                        | mode_visibility_all => ctx_empty
                                        | mode_visibility_hide_method => ctx_hide ctx_empty
                                        end xz TZ xc TC z T) ⁏ e & w ~ y ⊢ ax /[ z -> w] ∶ Tx /[ z -> w] 〉)
  ) as Hsuz.
  {
    intros. 
    forwards Hsuz: (@trf_subst typing_mode typ trm) (inert_G_for_obj G w y T S ) z ( ctx_met_refined_heap
        match mode_visibility' typing_mode with
        | mode_visibility_all => ctx_empty
        | mode_visibility_hide_method => ctx_hide ctx_empty
        end xz TZ xc TC z T ) Htyw.

    * ctx_flatten.
      
      constructor. constructor.
      eapply ctx_ok_vis_hide; eauto.
      destruct (mode_visibility' typing_mode); ctx_flatten; notin_solve.
      destruct (mode_visibility' typing_mode); ctx_flatten; notin_solve.

    * sympls; eauto using ctx_dom.
    * unfold renv_notin. notin_solve. auto.
      rewrite env_push_fv. notin_solve.
      unfold fv, VarFreeVar. auto.

    * intros. destruct (mode_visibility' typing_mode). simpl. eauto.
      simpl. eauto. ctx_flatten in H3.
      inversions H3. false. eapply notin_same with (x:=z). notin_solve.
      inversions H10. false. eapply notin_same with (x:=z). notin_solve.
      inversions H11.

    * apply Hsuz.
  }
    eapply SubstCompatTransformationTrm with (y:=w) (S:=T /[-> z]).
  + auto.
  + 
    eapply EWeakeningCompatTransformationTrm. eauto.
    eapply trf_instantiation_met0z with (xc:=xc) (xz:=xz); eauto.
    eapply Hwkz.
    eapply Hwkz0.
    hnf; eauto.
  +
    eapply Hsuz.
  +
    eapply Hsuz.
  + hnf; eauto.
Qed.


Lemma trf_instantiation_met (TL : TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) ( TL))
  L G e w y z T S m t t0 TZ TC TR L0:
  z \notin (L \u fv T \u fv S \u ctx_fv G \u dom e \u fv_env e \u \{ w } \u \{ y }) ->
  ok (e & w ~ y) ->
  ctx_ok ((inert_G_for_obj G w y T S § z / v_self ~ T /[-> z])) ->
  typ_indep_b 0 T ->
  (forall xz : var,
     xz \notin L0 ->
     forall xc : var,
     xc <> xz ->
     xc \notin L0 -> 
  〈 ctx_obj_met typing_mode (G § z / v_self ~ open z T) (objctx_heap z T) xz TZ xc TC ⁏ e ⊢ open_met xz xc t ∶ open_met xz xc TR 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 ctx_obj_met typing_mode (G § z / v_self ~ open z T) (objctx_heap z T) xz TZ xc TC ⁏ e ⊢ open_met xz xc t0 ∶ open_met xz xc TR 〉
  ) ->
  〈 inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y ⊢ def_met m (t /[ z -> w]) ∶ dec_met m (TC /[ z -> w]) (TZ /[ z -> w]) (TR /[ z -> w]) 〉 ∼ trf_local_idtyp_def typing_mode TL ∼> 〈 inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y ⊢ def_met m (t0 /[ z -> w]) ∶ dec_met m (TC /[ z -> w]) (TZ /[ z -> w]) (TR /[ z -> w]) 〉.
Proof.
  introv Hn Hokw HokG Hind Hty.
  eapply trf_local_idtyp_def_met with (L:=(L \u L0 \u fv T \u fv S \u ctx_fv G \u dom e \u fv_env e \u \{ w} \u \{ y} \u ctx_dom G\u \{ z})).
  sympls.
  intros.
  ctx_split_from xz.

  forwards Hty1: trf_instantiation_met0 w y z xz xc.
  8: eapply Hty.
   eauto.
   eauto.
   eauto.
   eauto.
   eauto.
   eauto.
   eauto.
   eauto.
   eauto.
   eauto.

  sympls.

  repeat (
    rewrite subst_open_commut in Hty1;
    try rewrite substi_eq_var in Hty1;
    try rewrite substi_ne_var in Hty1 by auto
  ).

  eapply TransformNarrowingCompatSubctx.
  ++
    eapply TrfLocalIdtypTrmEnsuresCond.
  ++
    eapply NarrowingCompatTransformationTrm. eauto.
  ++
    eapply Hty1.
  ++
    econstructor. eauto.
    constructor; auto.
    +
      assert (ctx_ok G).
      inversions HokG. inversions H4. inversions H5. auto.

      destruct (mode_visibility' typing_mode).
      * constructor; eauto. constructor; auto 3. constructor; auto 3. inversions HokG. eauto.
      * eauto. constructor; eauto. constructor; auto 3.
    +
      destruct (mode_visibility' typing_mode); eauto 3. constructor; eauto.
    + sympls.
      iapply subtyp_and; auto.
      iapply subtyp_and; auto.

      replace (T /[ z -> w]) with T.
      apply subtyp_tyq.
      auto.
      rewrite subst_fresh; auto.

      replace (T /[ z -> w]) with T. auto.
      rewrite subst_fresh; auto.
  ++
    econstructor. eauto.
    constructor; auto.
    +
      assert (ctx_ok G).
      inversions HokG. inversions H4. inversions H5. auto.

      destruct (mode_visibility' typing_mode).
      * constructor; eauto. constructor; auto 3. constructor; auto 3. inversions HokG. eauto.
      * eauto. constructor; eauto. constructor; auto 3.
    +
      destruct (mode_visibility' typing_mode); eauto 3. constructor; eauto.
    + sympls.
      iapply subtyp_and; auto.
      iapply subtyp_and; auto.

      replace (T /[ z -> w]) with T.
      apply subtyp_tyq.
      auto.
      rewrite subst_fresh; auto.

      replace (T /[ z -> w]) with T. auto.
      rewrite subst_fresh; auto.
  ++
    hnf; eauto.
Qed.


(* A definition typed in an object literal will be typed on the heap *)
Lemma trf_instantiation_def (TL : TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
(TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) (TL))
 L G e w y z T S D d d0:
  z \notin (L \u fv d \u fv T \u fv S \u ctx_fv G \u dom e \u fv_env e \u \{ w } \u \{ y }) ->
  ok (e & w ~ y) ->
  ctx_ok ((inert_G_for_obj G w y T S § z / v_self ~ T /[-> z])) ->
  typ_indep_b 0 T ->
  〈 G §§ objctx_lit z T ⁏ e ⊢ d /[-> z] ∶ D 〉 ∼ trf_local_idtyp_def typing_mode TL ∼> 〈 G §§ objctx_lit z T ⁏ e ⊢ d0 /[-> z] ∶ D 〉 ->
  〈 inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y ⊢ d /[-> z]  /[ z -> w] ∶ D /[ z -> w] 〉 ∼ trf_local_idtyp_def typing_mode TL ∼> 〈 inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y ⊢ d0  /[-> z] /[ z -> w] ∶ D /[ z -> w] 〉.
Proof.
  introv Hn Hokw HokG Hindep Htyd.
  inversions Htyd.
  eapply trf_instantiation_met; eauto.
Qed.

(* A definition list typed in an object literal will be typed on the heap *)
Lemma trf_instantiation_defs (TL : TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
(TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) ( TL))
 L G e w y T S T0 (ds ds0: defs):
  ok (e & w ~ y) ->
  ctx_ok (inert_G_for_obj G w y T S) ->
  typ_indep_b 0 T ->
  (forall x : var,
    x \notin L ->
    〈 G §§ objctx_lit x T ⁏ e ⊢ ds /[-> x] ∶ T0 /[-> x] 〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈 G §§ objctx_lit x T ⁏ e ⊢ ds0 /[-> x] ∶ T0 /[-> x] 〉
  ) ->
  〈 inert_G_for_obj G w y T S §§ objctx_heap y T⁏ e & w ~ y ⊢ ds /[-> w] ∶ T0 /[-> w] 〉
    ∼ trf_local_idtyp_defs typing_mode TL ∼>
  〈 inert_G_for_obj G w y T S §§ objctx_heap y T ⁏ e & w ~ y ⊢ ds0 /[-> w] ∶ T0 /[-> w] 〉.
Proof.
  introv Hoke HokG Hin Hdefs.
  pick_fresh_gen
    (L \u \{ w} \u \{ y} \u ctx_dom G \u ctx_fv G \u fv ds \u fv ds0  \u fv T \u fv S \u dom e
       \u fv_env e \u dom (e & w ~ y) \u fv_env (e & w ~ y)
       \u ctx_fv (inert_G_for_obj G w y T S) \u fv T0) z.

  assert (Hzds : z \notin (fv ds)). notin_solve.
  assert (Hzds0 : z \notin (fv ds0)). notin_solve.
  rewrite subst_intro with (x:=z) (y0:=w) by auto.
  rewrite subst_intro with (x:=z) (y0:=w) by auto.
  rewrite subst_intro with (x:=z) (y0:=w) by auto.
  assert (z \notin L) as HzL by notin_solve.
  specialize (Hdefs z HzL).
  remember ( 〈G §§ objctx_lit z T ⁏ e ⊢ ds /[-> z] ∶ T0 /[-> z]〉  ).
  remember ( 〈G §§ objctx_lit z T ⁏ e ⊢ ds0 /[-> z] ∶ T0 /[-> z]〉  ).
  reverts ds ds0 T0.

  induction Hdefs; intros; inversion Heqt; inversions Heqt0; sympls.

  + destruct T0; inversions H9. sympls.

    destruct ds; inversions H5.
    destruct ds0; inversions H10.
    sympls.
    apply trf_local_idtyp_defs_one.
    eapply trf_instantiation_def with (L:=L); eauto.
  + 
    destruct T1; inversions H7. sympls.
    destruct T1_2; inversions H5. sympls.
    destruct ds; inversions H8.
    destruct ds0; inversions H13.
    apply trf_local_idtyp_defs_consl.
    --
      eapply IHHdefs; eauto. sympls. notin_solve. sympls. notin_solve. sympls. eauto.
    --
      eapply instantiation_def with (L:=L); auto. sympls. eauto.
    --
      apply subst_defs_hasnt.
      unfold label_of, DefLabelOf in *.
      destruct d0; sympls; auto.
    --
      apply subst_defs_hasnt.
      unfold label_of, DefLabelOf in *.
      destruct d0; sympls; auto.
  +
    destruct T1; inversions H7. sympls.
    destruct T1_2; inversions H5. sympls.
    destruct ds; inversions H8.
    destruct ds0; inversions H13.
    apply trf_local_idtyp_defs_consr.
    --
      eapply trf_instantiation_def with (L:=L); auto. sympls. eauto.
    --
      sympls.
      rewrite <- subst_intro.
      rewrite <- subst_intro.
      eapply instantiation_defs' with (z:=z) (L:=L); eauto. sympls; eauto. eauto. eauto.
    --
      apply subst_defs_hasnt.
      unfold label_of, DefLabelOf in *.
      destruct d0; sympls; auto.
    --
      apply subst_defs_hasnt.
      unfold label_of, DefLabelOf in *.
      destruct d1; sympls; auto.
Qed.

Lemma transformation_lifted_focus_or_preserved_focus_lit_1obj (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
    G da db Ra Rb tr fs S e Lva C3 Lvb C4 T Tt Tl L G3 G4:
  〈 G ⁏ e ⊢ (lit_obj Ra da) ∶ Tl 〉 ∼ trf_local_idtyp_lit typing_mode TL ∼> 〈 G ⁏ e ⊢ (lit_obj Rb db) ∶ Tl 〉->
  G ⊢ trm_lit (lit_obj Ra da) tr ; fs ; S ; e ; Lva ∶ T ↦ G3 ⊢ C3 ->
  G ⊢ trm_lit (lit_obj Rb db) tr ; fs ; S ; e ; Lvb ∶ T ↦ G4 ⊢ C4 ->
  heap_correspond G e S ->
  ty_stack G e Tt fs T ->
  (forall x : var, x \notin L -> G §ˣ x ~ Tl ⁏ e ⊢ tr /[-> x] ∶ Tt) ->
  〈 G3 ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 G4 ⊢ C4 ∶ T 〉.
Proof.
  introv Htrf Hpr1 Hpr2 Hhc Htys Htyt.

  (* The transformed literal is an object.
       The transformation is in one of the defs and it will be the same on the heap *)

  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  step_invert_simple Hpr1' HtyC3.
  step_invert_simple Hpr2' HtyC4.

  lets Htrf': Htrf.
  inversions Htrf'.

  (* Pick fresh variables to use in the transformation *)
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr. eauto. eapply HeapCorrespondence.heap_correspond_to_ok_e; eauto. eapply HeapCorrespondence.heap_correspond_to_ok_s; eauto.
  fresh_chain_pick Fr w1.
  fresh_chain_pick Fr y1.
  (* Assert similarity from both sides, renaming the chosen variables to the fresh variables *)
  forwards Hsi1: trf_similarity_config_lit_obj w1 y1 Hpr1'.
  eauto.
  eapply config_typed_valid; eauto.
  sympl. fresh_chain_solve Fr.
  sympl. eauto.
  fresh_chain_solve Fr. fresh_chain_solve Fr.
  forwards Hsi2: trf_similarity_config_lit_obj w1 y1 Hpr2'.
  eauto.
  eapply config_typed_valid; eauto.
  sympl. fresh_chain_solve Fr.
  sympl. eauto.
  fresh_chain_solve Fr. fresh_chain_solve Fr.
  eapply TransformSimilaritySymmetric in Hsi2.

  (* Solve the transformation with similarities on both sides *)
  eapply trf_lift_sim_conf_sim_lift_sim. eapply Hsi1. 2: eapply Hsi2.


  (* Solve the transformation in the middle, using the fresh variables *)
  clear Hsi1 Hsi2.

  (* Transformaption happens on the heap *)
  eapply trf_lift_idtyp_conf_heap; eauto.
  - (* Left config accessories *)
    inversion Htyc. subst. 
    eapply ty_trm_lit_inv in H10. destructecs H10.
    eapply ty_config_accessories_tred_let_obj.
    eapply ty_config_to_accessories; eauto.
    eapply H0. fresh_chain_solve Fr. eauto.
  -  (* Right config accessories *)
    inversion Htyc0. subst.
    eapply ty_trm_lit_inv in H10. destructecs H10.
    eapply ty_config_accessories_tred_let_obj.
    eapply ty_config_to_accessories; eauto.
    eapply H0. fresh_chain_solve Fr. eauto.
  - (* Left heap accessories *)
    eapply heap_correspond_accessories_tred_let_obj. eapply heap_correspond_to_accessories; eauto. fresh_chain_solve Fr.
  - (* Right heap accessories *)
    eapply heap_correspond_accessories_tred_let_obj. eapply heap_correspond_to_accessories; eauto. fresh_chain_solve Fr.
  - (* Heap transformation *)
    hnf in Hhc. destructecs Hhc.
    (* The transformation applies here on the last item *)
    eapply trf_lift_idtyp_heap_e; eauto.
    + (* Typing of old items *)
    intros y2 Hy2.
    eapply ty_item_s_epush; try fresh_chain_solve Fr. eapply  ty_item_s_Gpush; try fresh_chain_solve Fr; eauto.
    ctx_simpl_dom. fresh_chain_solve Fr.
    + (* Transformation of the new item *)
      rename H7 into HindepR0.
      rename H1 into Htrfd.

      (* Cases on the type of the item *)
      destruct (mode_refcaps' typing_mode) eqn: Hm;
      destruct (mode_objlit_refcaps' typing_mode) eqn: Hml; try typing_mode_contradiction typing_mode;
      destruct (mode_refcap_corr' typing_mode) eqn: Hmc; try typing_mode_contradiction typing_mode.

      * (* No capabilities *)

        assert (ok (e & w1 ~ y1)) as Hoke1. {
          constructor. auto. fresh_chain_solve Fr.
        }
        assert (e & w1 ~ y1 ⊢ w1 ≈ y1) as Htyq1. {
          eauto.
        }
        assert (ctx_ok (inert_G_for_obj G w1 y1 R0 typ_top)) as HokG1. {
          fresh_chain_solve2 Fr.
        }
        assert (ctx_kdom v_ref (inert_G_for_obj G w1 y1 R0 typ_top) = dom (e & w1 ~ y1)) as Hkdom1. {
          simpl_dom.
          simpl; cases_if; cases_if. rewrite Hhc5. auto.
        }

        assert (all_tfree_k_defs (inert_G_for_obj G w1 y1 R0 typ_top) d /[-> w1] v_ref) as Htfd. {
          apply all_tfree_k_open; eauto.
          intro. intro. inversions Htyc. hnf in H10. eapply ctx_k_push_n. eapply ctx_k_push_n.  eapply H10; eauto. fresh_chain_solve Fr.  fresh_chain_solve2 Fr.
          apply ctx_k_push_e. 
        }
        assert (forall x : var, x \notin L0 -> G §§ objctx_lit x R0 ⁏ e ⊢ d /[-> x] ∶ R0 /[-> x]) as Htyl. {  
          intros. forwards: Htrfd H.
          forwards: TransformLocalIdtypDefsTyped H0. eauto. hnf in H1. eapply H1.
        }
        assert (inert_G_for_obj G w1 y1 R0 typ_top §§ objctx_heap y1 R0 ⁏ e & w1 ~ y1 ⊢ d /[-> w1] ∶ R0 /[-> w1]) as Htyd. {
          eapply instantiation_defs; eauto.
        }
        assert (all_tfree_k_defs (inert_G_for_obj G w1 y1 R0 typ_top) d0 /[-> w1] v_ref) as Htfd0. {
          apply all_tfree_k_open; eauto.
          intro. intro. inversions Htyc0. hnf in H10. eapply ctx_k_push_n. eapply ctx_k_push_n.  eapply H10; eauto. fresh_chain_solve Fr.  fresh_chain_solve2 Fr.
          apply ctx_k_push_e. 
        }
        assert (forall x : var, x \notin L0 -> G §§ objctx_lit x R0 ⁏ e ⊢ d0 /[-> x] ∶ R0 /[-> x]) as Htyl0. {  
          intros. forwards: Htrfd H.
          eapply trf_inverse_intro in H0.
          forwards: TransformLocalIdtypDefsTyped TLTI.
          rewrite <- trf_local_idtyp_inverse_defs in H1.
          eapply H1 in H0. hnf in H0. eapply H0.
          eauto.
        }
        assert (inert_G_for_obj G w1 y1 R0 typ_top §§ objctx_heap y1 R0 ⁏ e & w1 ~ y1 ⊢ d0 /[-> w1] ∶ R0 /[-> w1]) as Htyd0. {
          eapply instantiation_defs; eauto.
        }


        assert (renv_corr (inert_G_for_obj G w1 y1 R0 typ_top) (e & w1 ~ y1)) as Hcor1. {
          apply renv_corr_push_obj; try fresh_chain_solve Fr; auto. constructor. rewrite Hmc. auto.
        }

        assert (inert (inert_G_for_obj G w1 y1 R0 typ_top §§& objctx_heap y1 R0)) as HinGwy. {
          eapply ctx_inert_for_obj_heap; sympl; try fresh_chain_solve Fr; eauto.
          eapply RecordDefs.inert_obj_typ; eauto.
          hnf. rewrite Hm. constructor.
        }
        assert (inert (inert_G_for_obj G w1 y1 R0 typ_top)) as HinGwy'. auto.
        eapply trf_lift_idtyp_item_obj.
        -- (* item type *)
          eapply ctx_binds_n. fresh_chain_solve Fr.
          eapply ctx_binds_e.
        -- (* defs transformation *)

          eapply trf_local_deref_defs.
          auto.
          fresh_chain_solve2 Fr.
          auto.
          eapply transformation_open_defs_tyq; eauto.
          eapply trf_instantiation_defs; eauto.
          fresh_chain_solve2 Fr.
        --
          eapply all_tfree_defs_deref; eauto.
        --
          eapply all_tfree_defs_deref; eauto.
      * (* Rodot capabilities *)

        assert (ok (e & w1 ~ y1)) as Hoke1. {
          constructor. auto. fresh_chain_solve Fr.
        }
        assert (e & w1 ~ y1 ⊢ w1 ≈ y1) as Htyq1. {
          eauto.
        }
        assert (ctx_ok (inert_G_for_obj G w1 y1 R0 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)))) as HokG1. {
          fresh_chain_solve2 Fr.
        }
        assert (ctx_kdom v_ref (inert_G_for_obj G w1 y1 R0 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) = dom (e & w1 ~ y1)) as Hkdom1. {
          simpl_dom.
          simpl; cases_if; cases_if. rewrite Hhc5. auto.
        }

        assert (all_tfree_k_defs (inert_G_for_obj G w1 y1 R0 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) d /[-> w1] v_ref) as Htfd. {
          apply all_tfree_k_open; eauto.
          intro. intro. inversions Htyc. hnf in H10. eapply ctx_k_push_n. eapply ctx_k_push_n.  eapply H10; eauto. fresh_chain_solve Fr.  fresh_chain_solve2 Fr.
          apply ctx_k_push_e. 
        }
        assert (forall x : var, x \notin L0 -> G §§ objctx_lit x R0 ⁏ e ⊢ d /[-> x] ∶ R0 /[-> x]) as Htyl. {  
          intros. forwards: Htrfd H.
          forwards: TransformLocalIdtypDefsTyped H0. eauto. hnf in H1. eapply H1.
        }
        assert (inert_G_for_obj G w1 y1 R0 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) §§ objctx_heap y1 R0 ⁏ e & w1 ~ y1 ⊢ d /[-> w1] ∶ R0 /[-> w1]) as Htyd. {
          eapply instantiation_defs; eauto.
        }
        assert (all_tfree_k_defs (inert_G_for_obj G w1 y1 R0 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) d0 /[-> w1] v_ref) as Htfd0. {
          apply all_tfree_k_open; eauto.
          intro. intro. inversions Htyc0. hnf in H10. eapply ctx_k_push_n. eapply ctx_k_push_n.  eapply H10; eauto. fresh_chain_solve Fr.  fresh_chain_solve2 Fr.
          apply ctx_k_push_e. 
        }
        assert (forall x : var, x \notin L0 -> G §§ objctx_lit x R0 ⁏ e ⊢ d0 /[-> x] ∶ R0 /[-> x]) as Htyl0. {  
          intros. forwards: Htrfd H.
          eapply trf_inverse_intro in H0.
          forwards: TransformLocalIdtypDefsTyped TLTI.
          rewrite <- trf_local_idtyp_inverse_defs in H1.
          eapply H1 in H0. hnf in H0. eapply H0.
          eauto.
        }
        assert (inert_G_for_obj G w1 y1 R0 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) §§ objctx_heap y1 R0 ⁏ e & w1 ~ y1 ⊢ d0 /[-> w1] ∶ R0 /[-> w1]) as Htyd0. {
          eapply instantiation_defs; eauto.
        }
        assert (renv_corr (inert_G_for_obj G w1 y1 R0 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))) (e & w1 ~ y1)) as Hcor1. {
          apply renv_corr_push_obj; try fresh_chain_solve Fr; auto. constructor. rewrite Hmc. eauto.
        }

        assert (inert (inert_G_for_obj G w1 y1 R0 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)) §§& objctx_heap y1 R0)) as HinGwy. {
          eapply ctx_inert_for_obj_heap; sympl; try fresh_chain_solve Fr; eauto.
          eapply RecordDefs.inert_obj_typ; eauto.
          hnf. rewrite Hm. constructor.
        }
        assert (inert (inert_G_for_obj G w1 y1 R0 (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot)))) as HinGwy'. auto.

        eapply trf_lift_idtyp_item_obj.
        -- (* item type *)
          eapply ctx_binds_n. fresh_chain_solve Fr.
          eapply ctx_binds_e.
        -- (* defs transformation *)

          eapply trf_local_deref_defs.
          auto.
          fresh_chain_solve2 Fr.
          auto.
          eapply transformation_open_defs_tyq; eauto.
          eapply trf_instantiation_defs; eauto.
          fresh_chain_solve2 Fr.
        --
          eapply all_tfree_defs_deref; eauto.
        --
          eapply all_tfree_defs_deref; eauto.

  - (* Stack typing *)
    eapply ty_stack_epush; try fresh_chain_solve Fr.
    eapply ty_stack_push; try (ctx_simpl_dom; fresh_chain_solve Fr).
    eapply ty_stack_push; try fresh_chain_solve Fr. 
    eauto.
  - (* Focus typing *)
    
      destruct (mode_refcaps' typing_mode) eqn: Hmrc;
      destruct (mode_objlit_refcaps' typing_mode) eqn: Hmol;
      try typing_mode_contradiction typing_mode.
      + (* No capabilities *)
      eapply @renaming_fresh with (L := (L \u \{w0} \u \{y0} \u dom e \u ctx_dom G)).
      *
      apply CloseUnboundSubstIntro.
      * eauto.
      *
      apply ewkn_ty_trm.
      * fresh_chain_solve2 Fr.
      * fresh_chain_solve2 Fr.
      * 
      intros.
      ctx_split_between y1 z.
      assert (Weakening (ty_trm basic)); eauto.
       apply weaken_ty; simpl; eauto.
      eapply ewkn_push. eapply Htyt; eauto. eauto. fresh_chain_solve Fr. fresh_chain_solve2 Fr.
      *
      eauto.
      * eauto.
    + (* Rodot capabilities *)
      eapply @renaming_fresh with (L := (L \u \{w0} \u \{y0} \u dom e \u ctx_dom G)).
      *
        apply CloseUnboundSubstIntro.
      * eauto.
      *
        apply ewkn_ty_trm.
      * fresh_chain_solve2 Fr.
      * fresh_chain_solve2 Fr.
      * 
        intros.
        ctx_split_between y1 z.
        assert (Weakening (ty_trm basic)); eauto.
        apply weaken_ty; simpl; eauto.
        eapply ewkn_push. eapply Htyt; eauto. eauto. fresh_chain_solve Fr. fresh_chain_solve2 Fr.
      * eauto.
      * eauto.
Qed.


Lemma transformation_lifted_focus_or_preserved_focus_lit_2obj (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
    G d R ta tb fs S e Lva C3 Lvb C4 T Tt Tl L G3 G4:
  ( forall x : var,
       x \notin L ->
       〈 G §ˣ x ~ Tl ⁏ e ⊢ ta /[-> x] ∶ Tt 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G §ˣ x ~ Tl ⁏ e ⊢ tb /[-> x] ∶ Tt 〉) ->
  G ⊢ trm_lit (lit_obj R d) ta ; fs ; S ; e ; Lva ∶ T ↦ G3 ⊢ C3 ->
  G ⊢ trm_lit (lit_obj R d) tb ; fs ; S ; e ; Lvb ∶ T ↦ G4 ⊢ C4 ->
  heap_correspond G e S ->
  ty_stack G e Tt fs T ->
  G ⁏ e ⊢ lit_obj R d ∶ Tl ->
  〈 G3 ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 G4 ⊢ C4 ∶ T 〉.
Proof.
  introv Htrf Hpr1 Hpr2 Hhc Htys Htyt.

  (* The transformed literal is an object.
       The transformation is in one of the defs and it will be the same on the heap *)

  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  step_invert_simple Hpr1' HtyC3.
  step_invert_simple Hpr2' HtyC4.

  lets Htyt': Htyt.
  inversions Htyt'.

  (* Pick fresh variables to use in the transformation *)
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr. eauto. eapply HeapCorrespondence.heap_correspond_to_ok_e; eauto. eapply HeapCorrespondence.heap_correspond_to_ok_s; eauto.
  fresh_chain_pick Fr w1.
  fresh_chain_pick Fr y1.
  (* Assert similarity from both sides, renaming the chosen variables to the fresh variables *)
  forwards Hsi1: trf_similarity_config_lit_obj w1 y1 Hpr1'.
  eauto.
  eapply config_typed_valid; eauto.
  sympl. fresh_chain_solve Fr.
  sympl. eauto.
  fresh_chain_solve Fr. fresh_chain_solve Fr.
  forwards Hsi2: trf_similarity_config_lit_obj w1 y1 Hpr2'.
  eauto.
  eapply config_typed_valid; eauto.
  sympl. fresh_chain_solve Fr.
  sympl. eauto.
  fresh_chain_solve Fr. fresh_chain_solve Fr.
  eapply TransformSimilaritySymmetric in Hsi2.

  (* Solve the transformation with similarities on both sides *)
  eapply trf_lift_sim_conf_sim_lift_sim. eapply Hsi1. 2: eapply Hsi2.

    assert (HinL : inert_typ (typ_objref_loc typing_mode R) v_loc).
    { eapply typ_objref_loc_inert; eauto. }
    assert (HinR : inert_typ (typ_objref_loc typing_mode R) v_ref) by eauto using inert_typ_loc_ref.
    assert (Hcor: typ_corr (typ_objref_loc typing_mode R) (typ_objref_loc typing_mode R)) by eauto using typ_objref_loc_corr.

  (* Solve the transformation in the middle, using the fresh variables *)

  clear Hsi1 Hsi2.

  eapply trf_lift_idtyp_conf_focus.
  eapply trf_focus_idtyp_conf_focus with (Tt:=Tt).
  - (* Left config accessories *)
    inversion Htyc. subst. 
    eapply ty_trm_lit_inv in H10. destructecs H10.
    eapply ty_config_accessories_tred_let_obj.
    eapply ty_config_to_accessories; eauto.
    eapply H0. fresh_chain_solve Fr. eauto.
  -  (* Right config accessories *)
    inversion Htyc0. subst.
    eapply ty_trm_lit_inv in H10. destructecs H10.
    eapply ty_config_accessories_tred_let_obj.
    eapply ty_config_to_accessories; eauto.
    eapply H0. fresh_chain_solve Fr. eauto.
  - 

      destruct (mode_refcaps' typing_mode).
      eapply heap_correspond_push_obj; try fresh_chain_solve Fr; eauto.
      inversions Htyc.
      hnf; hnf in H11; auto.

      eapply heap_correspond_push_obj; try fresh_chain_solve Fr; sympl; eauto.
      inversions Htyc.
      hnf; hnf in H11; auto.
  - 
    eapply ty_stack_epush; try fresh_chain_solve Fr.
    eapply ty_stack_push; try (ctx_simpl_dom; fresh_chain_solve Fr).
    eapply ty_stack_push; try fresh_chain_solve Fr. 
    eauto.

  -
    eapply trf_local_idtyp_trm_subst_w with (L:=L \u ctx_dom G \u \{ w1} \u \{ y1}) (vk:=v_let).
    auto.
    auto.
    fresh_chain_solve2 Fr.
    fresh_chain_solve2 Fr.
    eauto.
    intros x Hx.
    forwards Htrfx: Htrf x . auto.

    assert (sealed G). inversions Htyc0. auto.
    (*assert (closed R). intro. intros. forwards: HclT0 n x0. sympl in H6. injection H6. auto.*)
    assert (closed (t /[-> x])).
    inversions  Htyc. inversions H18. eapply Closedness.closed_trm_lit_obj_r; eauto.
    assert (closed (t0 /[-> x])).
    inversions  Htyc0. inversions H19. eapply Closedness.closed_trm_lit_obj_r; eauto.

    sympl in Htrfx.

    eapply trf_local_idtyp_trm_ewkn_push. auto.
    ctx_split_between y1 x.

    eapply WeakeningCompatTransformationTrm. eauto.
    eapply Htrfx.
    rewrite ctx_concat_single.

    destruct (mode_objlit_refcaps' typing_mode) eqn: Hmo;
    destruct (mode_refcaps' typing_mode) eqn: Hmr; try typing_mode_contradiction typing_mode.

    econstructor. eauto. ctx_flatten. econstructor. fresh_chain_solve2 Fr. ctx_simpl_dom. notin_solve.  fresh_chain_solve Fr.
    econstructor. eauto. ctx_flatten. econstructor. fresh_chain_solve2 Fr. ctx_simpl_dom. notin_solve.  fresh_chain_solve Fr.

    rewrite ctx_concat_single.
    destruct (mode_objlit_refcaps' typing_mode) eqn: Hmo;
    destruct (mode_refcaps' typing_mode) eqn: Hmr; try typing_mode_contradiction typing_mode.

    econstructor. eauto. ctx_flatten. econstructor. fresh_chain_solve2 Fr. ctx_simpl_dom. notin_solve.  fresh_chain_solve Fr.
    econstructor. eauto. ctx_flatten. econstructor. fresh_chain_solve2 Fr. ctx_simpl_dom. notin_solve.  fresh_chain_solve Fr.

    hnf; eauto. 
    econstructor. fresh_chain_solve2 Fr. ctx_simpl_dom. notin_solve.  fresh_chain_solve Fr.  fresh_chain_solve Fr.
  - eapply subtyp_refl.
Qed.


Lemma transformation_lifted_focus_or_preserved_heap_lit_2obj (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLSl: TransformSealingCompat typ trm TL)
    G d R tr fs Sa Sb e Lva C3 Lvb C4 T Tt  G3 G4:
  〈 e ⊢ Sa ∶ G 〉 ∼ trf_lift_idtyp_heap typing_mode TL ∼> 〈 e ⊢ Sb ∶ G 〉 ->  
  G ⊢ trm_lit (lit_obj R d) tr ; fs ; Sa ; e ; Lva ∶ T ↦ G3 ⊢ C3 ->
  G ⊢ trm_lit (lit_obj R d) tr ; fs ; Sb ; e ; Lvb ∶ T ↦ G4 ⊢ C4 ->
  G ⁏ e ⊢ trm_lit (lit_obj R d) tr ∶ Tt ->
  ty_stack G e Tt fs T ->
  〈 G3 ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 G4 ⊢ C4 ∶ T 〉.
Proof.
  introv Htrf Hpr1 Hpr2 Htyt Htys.

  (* The transformed literal is an object.
       The transformation is in one of the defs and it will be the same on the heap *)

  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  step_invert_simple Hpr1' HtyC3.
  step_invert_simple Hpr2' HtyC4.

  inversions Htyc.
  inversions Htyc0.

  forwards Htl: ty_trm_lit_inv Htyt. eauto.
  destructecs Htl.
  
  lets Htylit: Htl0.
  inversions Htylit.

  (* Pick fresh variables to use in the transformation *)
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr. eauto.
    eapply HeapCorrespondence.heap_correspond_to_ok_e; eauto.
    eapply HeapCorrespondence.heap_correspond_to_ok_s; eauto.
    eapply HeapCorrespondence.heap_correspond_to_ok_s; eauto.
  fresh_chain_pick Fr w1.
  fresh_chain_pick Fr y1.
  (* Assert similarity from both sides, renaming the chosen variables to the fresh variables *)
  forwards Hsi1: trf_similarity_config_lit_obj w1 y1 Hpr1'.
  eauto.
  eapply config_typed_valid; eauto.
  sympl. fresh_chain_solve Fr.
  sympl. eauto.
  fresh_chain_solve Fr. fresh_chain_solve Fr.
  forwards Hsi2: trf_similarity_config_lit_obj w1 y1 Hpr2'.
  eauto.
  eapply config_typed_valid; eauto.
  sympl. fresh_chain_solve Fr.
  sympl. eauto.
  fresh_chain_solve Fr. fresh_chain_solve Fr.
  eapply TransformSimilaritySymmetric in Hsi2.

  (* Solve the transformation with similarities on both sides *)
  eapply trf_lift_sim_conf_sim_lift_sim. eapply Hsi1. 2: eapply Hsi2.

    assert (HinL : inert_typ (typ_objref_loc typing_mode R) v_loc).
    { eapply typ_objref_loc_inert; eauto. }
    assert (HinR : inert_typ (typ_objref_loc typing_mode R) v_ref) by eauto using inert_typ_loc_ref.
    assert (Hcor: typ_corr (typ_objref_loc typing_mode R) (typ_objref_loc typing_mode R)) by eauto using typ_objref_loc_corr.

  (* Solve the transformation in the middle, using the fresh variables *)
  clear Hsi1 Hsi2.

  assert (
    ty_config_accessories
    (inert_G_for_item G w1 y1
     match mode_refcaps' typing_mode with
     | mode_refcaps_none => typ_and (typ_bnd R) typ_top
     | mode_refcaps_mutability =>
         typ_and (typ_bnd R)
           (typ_rcd
              (dec_typ
                 (typ_label_refcap
                    typ_refcap_mutability) typ_bot
                 typ_bot))
     end) (e & w1 ~ y1) fs
  (Sa &
   y1 ~
   item_obj (Deref.deref_defs (e & w1 ~ y1) d /[-> w1]))
  t0 /[-> w1] T
  (fv (trm_lit (lit_obj R d) t0) \u
   fv fs \u fv_heap Sa \u fv_renv e \u Lva)
) as Hac2. {

    eapply ty_config_accessories_tred_let_obj.
    eapply ty_config_to_accessories; eauto.
    eapply Htl0. fresh_chain_solve Fr.
}

assert (
  heap_correspond_accessories
  (inert_G_for_item G w1 y1
     match mode_refcaps' typing_mode with
     | mode_refcaps_none => typ_and (typ_bnd R) typ_top
     | mode_refcaps_mutability =>
         typ_and (typ_bnd R)
           (typ_rcd
              (dec_typ
                 (typ_label_refcap
                    typ_refcap_mutability) typ_bot
                 typ_bot))
     end) (e & w1 ~ y1)
  (Sa &
   y1 ~
   item_obj (Deref.deref_defs (e & w1 ~ y1) d /[-> w1]))
) as Hhac2. {
  eapply heap_correspond_accessories_tred_let_obj. auto.
  eapply heap_correspond_to_accessories; eauto.
  fresh_solve; try eapply fv_seal_notin; fresh_chain_solve Fr.
}

  eapply trf_lift_idtyp_conf_heap.
  - (* Left config accessories *)
    apply Hac2.
  -  (* Right config accessories *)

    eapply ty_config_accessories_tred_let_obj.
    eapply ty_config_to_accessories; eauto.
    eapply Htl0. fresh_chain_solve Fr.
  - (* Left heap accessories *)
    apply Hhac2.
  - (* Right heap accessories *)
    eapply heap_correspond_accessories_tred_let_obj. auto. eapply heap_correspond_to_accessories; eauto. fresh_solve; try eapply fv_seal_notin; fresh_chain_solve Fr.
  - (* Heap transformation *)
    eapply trf_lift_idtyp_heap_n.
    + (* Transformation of old items *)
    eapply trf_lift_idtyp_heap_epush.
    eauto.
    eauto.
    eauto.
    fresh_chain_solve Fr.
    ctx_simpl_dom.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    eapply trf_lift_idtyp_heap_Gpush.
    eauto.
    eauto.
    eauto.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    eauto.
  
    + (* Typing of the new item *)

    assert (inert (inert_G_for_item G w1 y1 (typ_objref_loc typing_mode R) )) as HinGwy. {
    hnf in Hac2. destructecs Hac2. eapply Hac0.
    }

    assert (renv_corr (inert_G_for_item G w1 y1 (typ_objref_loc typing_mode R) ) (e & w1 ~ y1)) as HcorGwy. {
    hnf in Hhac2. destructecs Hhac2. eauto.
    }

    assert (ok (e & w1 ~ y1)) as Hokewy. {
    hnf in Hhac2. destructecs Hhac2. eauto.
    }
      assert (ctx_kdom v_ref (inert_G_for_item G w1 y1 (typ_objref_loc typing_mode R)) = dom (e & w1 ~ y1)) as Hkdom.
      {
    hnf in Hhac2. destructecs Hhac2. eapply Hhac6.
    }

    assert (all_tfree_k_defs (inert_G_for_item G w1 y1 (typ_objref_loc typing_mode R)) d v_ref) as Hk. 
    {
    eapply all_tfree_k_let_obj_defs.
    eauto.
    fresh_chain_solve Fr.
    }

    destruct (mode_refcaps' typing_mode) eqn: Hm.
  *
    assert(
      inert_G_for_obj G w1 y1 R typ_top
      §§ objctx_heap y1 R ⁏ e & w1 ~ y1
      ⊢ d /[-> w1] ∶ R /[-> w1]
    ) as Htyd. {
      eapply instantiation_defs; auto 1.
      eapply inert_ok; eauto.
      eauto.
    }
    eapply ty_item_obj_s.
    eapply ctx_binds_n. fresh_chain_solve Fr.
    eapply ctx_binds_e.
    eapply binds_single_eq.
    eauto.

    eapply HeapInstantiate.ty_deref_defs; auto 1.
    eapply open_defs_tyq with (y:=y1) (w:=w1); auto 1.

    eapply inert_ok; eauto.

    eauto.
    fresh_chain_solve Fr.
    eapply all_tfree_defs_deref; auto 1.
    eapply inert_ok; eauto. eapply Htyd.
    hnf in Hhac2. destructecs Hhac2.
    apply all_tfree_k_open; eauto.
    apply ctx_k_push_e.

    *

      assert(
        inert_G_for_obj G w1 y1 R (typ_rcd (dec_typ (typ_label_refcap typ_refcap_mutability) typ_bot typ_bot))
        §§ objctx_heap y1 R ⁏ e & w1 ~ y1
        ⊢ d /[-> w1] ∶ R /[-> w1]
      ) as Htyd. {
        eapply instantiation_defs; auto 1.
        eapply inert_ok; eauto.
        eauto.
      }
      eapply ty_item_obj_s.
      eapply ctx_binds_n. fresh_chain_solve Fr.
      eapply ctx_binds_e.
      eapply binds_single_eq.
      eauto.

      eapply HeapInstantiate.ty_deref_defs; auto 1.
      eapply open_defs_tyq with (y:=y1) (w:=w1); auto 1.

      eapply inert_ok; eauto.

      eauto.
      fresh_chain_solve Fr.
      eapply all_tfree_defs_deref; auto 1.
      eapply inert_ok; eauto. eapply Htyd.
      hnf in Hhac2. destructecs Hhac2.
      apply all_tfree_k_open; eauto.
      apply ctx_k_push_e.
  - 
    eapply ty_stack_epush; try fresh_chain_solve Fr.
    eapply ty_stack_push; try (ctx_simpl_dom; fresh_chain_solve Fr).
    eapply ty_stack_push; try fresh_chain_solve Fr. 
    eapply Htys.
  - (* Focus typing *)
    
      destruct (mode_refcaps' typing_mode) eqn: Hmrc;
      destruct (mode_objlit_refcaps' typing_mode) eqn: Hmol;
      try typing_mode_contradiction typing_mode.
      + (* No capabilities *)
      eapply @renaming_fresh with (L := (L \u \{w0} \u \{y0} \u dom e \u ctx_dom G)).
      *
      apply CloseUnboundSubstIntro.
      * eauto.
      *
      apply ewkn_ty_trm.
      * fresh_chain_solve2 Fr.
      * fresh_chain_solve2 Fr.
      * 
      intros.
      ctx_split_between y1 z.
      assert (Weakening (ty_trm basic)); eauto.
       apply weaken_ty; simpl; eauto.
      eapply ewkn_push. eapply Htl; eauto. eauto. fresh_chain_solve Fr. fresh_chain_solve2 Fr.
      *
      eauto.
      * eauto.
    + (* Rodot capabilities *)
      eapply @renaming_fresh with (L := (L \u \{w0} \u \{y0} \u dom e \u ctx_dom G)).
      *
      apply CloseUnboundSubstIntro.
      * eauto.
      *
      apply ewkn_ty_trm.
      * fresh_chain_solve2 Fr.
      * fresh_chain_solve2 Fr.
      * 
      intros.
      ctx_split_between y1 z.
      assert (Weakening (ty_trm basic)); eauto.
       apply weaken_ty; simpl; eauto.
      eapply ewkn_push. eapply Htl; eauto. eauto. fresh_chain_solve Fr. fresh_chain_solve2 Fr.
      * eauto.
      * eauto.
Qed.


Lemma transformation_lifted_focus_or_preserved_stack_lit_2obj (TL : TermTransform)
  (TLT: TransformRequires (TransformCondTermTyped typing_mode) TL)
  (TLTI: TransformRequires (TransformCondTermTyped typing_mode) (trf_inverse TL))
  (TLI: TransformEnsures (TransformCondTypeIdentical (A:=trm)) TL)
  (TLN: TransformNarrowingCompat typing_mode typ trm TL)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
    G d R tr fsa fsb S e Lva C3 Lvb C4 T Tt  G3 G4:
  〈 G ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ∼ trf_lift_idtyp_stack typing_mode TL ∼> 〈 G ⁏ e ⊢ fsb ∶ (Tt, T) 〉->
  G ⊢ trm_lit (lit_obj R d) tr ; fsa ; S ; e ; Lva ∶ T ↦ G3 ⊢ C3 ->
  G ⊢ trm_lit (lit_obj R d) tr ; fsb ; S ; e ; Lvb ∶ T ↦ G4 ⊢ C4 ->
  heap_correspond G e S ->  
  G ⁏ e ⊢ trm_lit (lit_obj R d) tr ∶ Tt ->
  〈 G3 ⊢ C3 ∶ T 〉 ∼ trf_lift_sim_conf typing_mode TL ∼> 〈 G4 ⊢ C4 ∶ T 〉.
Proof.
  introv Htrf Hpr1 Hpr2 Hhc Htyt.

  (* The transformed literal is an object.
       The transformation is in one of the defs and it will be the same on the heap *)

  lets Hpr1': Hpr1.
  lets Hpr2': Hpr2.
  forwards HtyC3: typed_preservation Hpr1'.
  forwards HtyC4: typed_preservation Hpr2'.

  step_invert_simple Hpr1' HtyC3.
  step_invert_simple Hpr2' HtyC4.

  inversions Htyc.
  inversions Htyc0.

  forwards Htl: ty_trm_lit_inv Htyt. eauto.
  destructecs Htl.
  
  lets Htylit: Htl0.
  inversions Htylit.

  (* Pick fresh variables to use in the transformation *)
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr. eauto.
  eapply HeapCorrespondence.heap_correspond_to_ok_e; eauto.
  eapply HeapCorrespondence.heap_correspond_to_ok_s; eauto.
  fresh_chain_pick Fr w1.
  fresh_chain_pick Fr y1.
  (* Assert similarity from both sides, renaming the chosen variables to the fresh variables *)
  forwards Hsi1: trf_similarity_config_lit_obj w1 y1 Hpr1'.
  eauto.
  eapply config_typed_valid; eauto.
  sympl. fresh_chain_solve Fr.
  sympl. eauto.
  fresh_chain_solve Fr. fresh_chain_solve Fr.
  forwards Hsi2: trf_similarity_config_lit_obj w1 y1 Hpr2'.
  eauto.
  eapply config_typed_valid; eauto.
  sympl. fresh_chain_solve Fr.
  sympl. eauto.
  fresh_chain_solve Fr. fresh_chain_solve Fr.
  eapply TransformSimilaritySymmetric in Hsi2.

  (* Solve the transformation with similarities on both sides *)
  eapply trf_lift_sim_conf_sim_lift_sim. eapply Hsi1. 2: eapply Hsi2.

    assert (HinL : inert_typ (typ_objref_loc typing_mode R) v_loc).
    { eapply typ_objref_loc_inert; eauto. }
    assert (HinR : inert_typ (typ_objref_loc typing_mode R) v_ref) by eauto using inert_typ_loc_ref.
    assert (Hcor: typ_corr (typ_objref_loc typing_mode R) (typ_objref_loc typing_mode R)) by eauto using typ_objref_loc_corr.

  (* Solve the transformation in the middle, using the fresh variables *)

  clear Hsi1 Hsi2.

  eapply trf_lift_idtyp_conf_stack.
  - (* Left config accessories *)
    eapply ty_config_accessories_tred_let_obj.
    eapply ty_config_to_accessories; eauto.
    eapply Htl0. fresh_chain_solve Fr.
  -  (* Right config accessories *)

    eapply ty_config_accessories_tred_let_obj.
    eapply ty_config_to_accessories; eauto.
    eapply Htl0. fresh_chain_solve Fr.
  -
      destruct (mode_refcaps' typing_mode).
      eapply heap_correspond_push_obj; try fresh_chain_solve Fr; eauto.

      hnf; hnf in H11; auto.

      eapply heap_correspond_push_obj; try fresh_chain_solve Fr; sympl; eauto.
      hnf; hnf in H11; auto.
  -
    assert(〈 inert_G_for_item G w1 y1 (typ_objref_loc typing_mode  R) ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ∼ TransformWknEnvPush (A:=stack) ∼> 〈 inert_G_for_item G w1 y1 (typ_objref_loc typing_mode R) ⁏ e & w1 ~ y1 ⊢ fsa ∶ (Tt, T) 〉).
    econstructor; fresh_chain_solve2 Fr.

    assert(〈 inert_G_for_item G w1 y1 (typ_objref_loc typing_mode R) ⁏ e ⊢ fsb ∶ (Tt, T) 〉 ∼ TransformWknEnvPush (A:=stack) ∼> 〈 inert_G_for_item G w1 y1 (typ_objref_loc typing_mode R) ⁏ e & w1 ~ y1 ⊢ fsb ∶ (Tt, T) 〉).
    econstructor; fresh_chain_solve2 Fr.

    eapply WknEnvPushCompatTransformationStack. eauto. 2: eauto. 2: eauto.

    assert(〈 G § y1 / v_loc ~ (typ_objref_loc typing_mode R) ⁏ e ⊢ fsa ∶ (Tt, T) 〉  ∼ TransformWknCtxPush (A:=stack) ∼> 〈 inert_G_for_item G w1 y1 (typ_objref_loc typing_mode R) ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ).
    econstructor; fresh_chain_solve2 Fr.

    assert(〈 G § y1 / v_loc ~ (typ_objref_loc typing_mode R)  ⁏ e ⊢ fsb ∶ (Tt, T) 〉  ∼ TransformWknCtxPush (A:=stack) ∼> 〈 inert_G_for_item G w1 y1 (typ_objref_loc typing_mode R) ⁏ e ⊢ fsb ∶ (Tt, T) 〉 ).
    econstructor; fresh_chain_solve2 Fr.

    eapply WknCtxPushCompatTransformationStack. eauto. 2: eauto. 2: eauto.


    assert(〈 G ⁏ e ⊢ fsa ∶ (Tt, T) 〉  ∼ TransformWknCtxPush (A:=stack) ∼> 〈 G § y1 / v_loc ~ (typ_objref_loc typing_mode R)  ⁏ e ⊢ fsa ∶ (Tt, T) 〉 ).
    econstructor; fresh_chain_solve2 Fr.

    assert(〈 G  ⁏ e ⊢ fsb ∶ (Tt, T) 〉  ∼ TransformWknCtxPush (A:=stack) ∼> 〈G § y1 / v_loc ~ (typ_objref_loc typing_mode R)  ⁏ e ⊢ fsb ∶ (Tt, T) 〉 ).
    econstructor; fresh_chain_solve2 Fr.

    eapply WknCtxPushCompatTransformationStack. eauto. 2: eauto. 2: eauto.

    eapply Htrf.
    hnf; eauto.
    hnf; eauto.
    hnf; eauto.

  - (* Focus typing *)
    
      destruct (mode_refcaps' typing_mode) eqn: Hmrc;
      destruct (mode_objlit_refcaps' typing_mode) eqn: Hmol;
      try typing_mode_contradiction typing_mode.
      + (* No capabilities *)
      eapply @renaming_fresh with (L := (L \u \{w0} \u \{y0} \u dom e \u ctx_dom G)).
      *
        apply CloseUnboundSubstIntro.
      * eauto.
      *
        apply ewkn_ty_trm.
      * fresh_chain_solve2 Fr.
      * fresh_chain_solve2 Fr.
      * 
        intros.
        ctx_split_between y1 z.
        assert (Weakening (ty_trm basic)); eauto.
        apply weaken_ty; simpl; eauto.
        eapply ewkn_push. eapply Htl; eauto. eauto. fresh_chain_solve Fr. fresh_chain_solve2 Fr.
      *
        eauto.
      * eauto.
    + (* Rodot capabilities *)
      eapply @renaming_fresh with (L := (L \u \{w0} \u \{y0} \u dom e \u ctx_dom G)).
      *
        apply CloseUnboundSubstIntro.
      * eauto.
      *
        apply ewkn_ty_trm.
      * fresh_chain_solve2 Fr.
      * fresh_chain_solve2 Fr.
      * 
        intros.
        ctx_split_between y1 z.
        assert (Weakening (ty_trm basic)); eauto.
        apply weaken_ty; simpl; eauto.
        eapply ewkn_push. eapply Htl; eauto. eauto. fresh_chain_solve Fr. fresh_chain_solve2 Fr.
      * eauto.
      * eauto.
Qed.

End WithTypingMode.