(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation TransformationLifting Weakening EWeakening Substitution.
Require Import ConfigTyping TFree TFreeTyping HeapCorrespondence InertContext.
Require Import TypedOperationalSemantics OperationalSemantics ConfigVarkinds EnvCorrespondence.

Require Import TransformationConfig TypedProgress TypedReductionEquivalence.
Require Import TypedPreservation TypedSoundness.
Require Import Safety.

(** * Preservation of transformations during reduction. *)

Section WithTypingMode.
Context (typing_mode: Mode).


Section TransformConfigReducesTo1Self.
  Context (trf: TransformConfig) {TR1: TransformConfigReducesTo1 typing_mode trf trf}.
  Lemma transform_config_reduces_to_n n Ga1 Gb1 Ta Tb Ca1 Cb1 Ga2 Gb2 Ca2 Cb2:
    〈Ga1 ⊢ Ca1 ∶ Ta〉 ∼trf∼> 〈Gb1 ⊢ Cb1 ∶ Tb〉 ->
    Ga1 ⊢ Ca1 ∶ Ta ⊣n↦* Ga2 ⊢ Ca2 ->
    Gb1 ⊢ Cb1 ∶ Tb ⊣n↦* Gb2 ⊢ Cb2 ->
    〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼trf∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉.
  Proof.
    reverts Ga2 Gb2 Ca2 Cb2.
    induction n; intros.
    - inversions H0.
      inversions H1. eauto.
    - inversions H0.
      inversions H1.
      forwards: IHn; eauto.
  Qed.
  Lemma transform_config_reduces_to_n_exists
      (TA: TransformPreserves TransformIsAnswer trf) (TL: TransformRequires (TransformCondConfTyped typing_mode) trf)
      n Ga1 Gb1 Ta Tb Ca1 Cb1 Gb2 Cb2:
    〈Ga1 ⊢ Ca1 ∶ Ta〉 ∼trf∼> 〈Gb1 ⊢ Cb1 ∶ Tb〉 ->
    Gb1 ⊢ Cb1 ∶ Tb ⊣n↦* Gb2 ⊢ Cb2 -> (* Note that this is the right side *)
    exists Ga2 Ca2,
    Ga1 ⊢ Ca1 ∶ Ta ⊣n↦* Ga2 ⊢ Ca2 /\
    〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼trf∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉.
  Proof.
    reverts Gb2 Cb2.
    induction n; intros.
    - inversions H0.
      exists Ga1 Ca1. split; eauto.
    - inversions H0.
      forwards G5 C5 Hred25 Htrf45: IHn; eauto.
      pose proof (typed_step_not_answer H6) as Hna.

      assert (~ answer C5) as Hna5.
      intro. eapply Hna.
      forwards Ha5: trf_preserves Htrf45; eauto.
      forwards Hty4: trf_requires Htrf45; eauto.
      hnf in Hty4.
      (* progress *)
      lets Hprg: (typed_progress Hty4).
      destruct Hprg.
      + (* answer *) contradiction.
      + destructc H0 as G4 C0 Hr5.
        exists G4 C0.
        splits; eauto.
  Qed.
End TransformConfigReducesTo1Self.

Section TransformConfigReducesTo1SelfInv.
  Context (trf: TransformConfig) {TR1: TransformConfigReducesTo1 typing_mode (trf_inverse trf) (trf_inverse trf)}.

  Lemma transform_config_reduces_to_n_exists_inv (TA: TransformPreserves TransformIsAnswer (trf_inverse trf)) (TL: TransformRequires (TransformCondConfTyped typing_mode) (trf_inverse trf))
      n Ga1 Gb1 Ta Tb Ca1 Cb1 Ga2 Ca2:
    〈Ga1 ⊢ Ca1 ∶ Ta〉 ∼trf∼> 〈Gb1 ⊢ Cb1 ∶ Tb〉 ->
    Ga1 ⊢ Ca1 ∶ Ta ⊣n↦* Ga2 ⊢ Ca2 -> (* Note that this is the left side *)
    exists Gb2 Cb2,
    Gb1 ⊢ Cb1 ∶ Tb ⊣n↦* Gb2 ⊢ Cb2 /\
    〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼trf∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉.
  Proof.
    intros.
    assert (〈Gb1 ⊢ Cb1 ∶ Tb〉 ∼(trf_inverse trf)∼> 〈Ga1 ⊢ Ca1 ∶ Ta〉).
     hnf. auto. clear H.
    eapply transform_config_reduces_to_n_exists in H1; eauto.
  Qed.

End TransformConfigReducesTo1SelfInv.

Section TransformConfigReducesTo1Self.
  Context {trf: TransformConfig} {TR1: TransformConfigReducesTo1 typing_mode trf trf}
    {TR1I: TransformConfigReducesTo1 typing_mode (trf_inverse trf) (trf_inverse trf)}.

  Lemma trf_reduces_to_self_terminates
    (TA: TransformPreserves TransformIsAnswer trf)
    (TAI: TransformPreserves TransformIsAnswer (trf_inverse trf))
    (TL: TransformRequires (TransformCondConfTyped typing_mode) trf)
    (TLI: TransformRequires (TransformCondConfTyped typing_mode) (trf_inverse trf)) :
      (TransformPreserves TransformCondTerminates trf).
  Proof.
    hnf. intros. hnf in H0. hnf.
    destruct TEa, TEb.
    destructc H0 as n C3 Ha3 Hred1.
    exists n.
    pose proof H as Hty1.
    eapply trf_requires in Hty1; eauto.
    forwards G3 Htyred1: (untyped_red_to_typed_closure Hty1 Hred1).
    forwards He: transform_config_reduces_to_n_exists_inv Htyred1; eauto.
    destructc He as G4 C4 Htyred2 Htrf34.
    exists C4.
    split.
    -
      clear TAI.
      eapply TA in Htrf34. unfold TransformIsAnswer in Htrf34; auto.
    -
      eapply typed_to_untyped_red_closure; eauto.
  Qed.

End TransformConfigReducesTo1Self.

End WithTypingMode.