(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation TFree.
Require Import TransformationSimilarity TransformationPreservation TransformationTerm.
Require Import TransformationConfig OperationalSemantics TypedOperationalSemantics.
Require Import ContextSimilarity ConfigSimilarity.
Require Import ConfigTyping ConfigTypingAccessories TypedProgress.
Require Import TypedPreservation.
Require Import TransformationIterateSim.
Require Import Similarity TypingSimilarity.

Require Import TransformationLifting.
Require Import TypedOperationalSemantics InertContext.
Require Import Similarity SyntaxSimilarity.
Require Import ExecutionTactics TypingHelpers.

Require Import TypedPreservation.

Require Import TypedSoundness TypedReductionEquivalence TypedReductionLemmas.

(** * Preservation of similarity transformation during reduction. *)

Section WithTypingMode.
Context (typing_mode: Mode).

(* If there is an execution to an answer, then any other execution cannot be longer *)
Lemma typed_red_closure_answer_split G1 G2 G3 C1 C2 C3 T n2 n3:
  ty_config G1 C1 T ->
  G1 ⊢ C1 ∶ T ⊣ n2 ↦* G2 ⊢ C2 ->
  answer C2 ->
  G1 ⊢ C1 ∶ T ⊣ n3 ↦* G3 ⊢ C3 ->
  exists n32,
    n2 = n3 + n32 /\
    exists G32 C32,
      G3 ⊢ C3 ∶ T ⊣ n32 ↦* G32 ⊢ C32 /\
      〈G2 ⊢ C2 ∶ T〉 ∼(trf_similarity typing_mode)∼> 〈G32 ⊢ C32 ∶ T〉 /\
      answer C32.
Proof.
  introv Htyc.
  reverts G1 C1 G2 C2.
  reverts G3 C3.

  induction n3; intros.
  + inversions H1.
    exists n2; split. rewrite Nat.add_0_l; auto.
    exists G2 C2; splits. auto.
    eapply TransformSimilarityReflexive. hnf. eapply typed_preservation_closure; eauto.
    auto.
  + inversions H1.
    forwards Hex: IHn3; fweauto.
    destructecs Hex.
    destruct n32.
    - (* n32=0 - Contradiction *)
      false.
      inversions Hex1.
      forwards: typed_step_not_answer H7.
      auto.
    - exists n32; splits.
      rewrite Nat.add_succ_comm; auto.
      forwards Hstepl: typed_red_closure_invl Hex1.
      destructecs Hstepl.
      forwards: trf_similarity_step Hstepl0 H7. eapply typed_preservation_closure; eauto.
      forwards Htred3: transform_config_reduces_to_n_exists_inv H1;
      rewrite ? trf_symmetric_inverse.
      apply TransformConfigReducesTo1Similarity.
      apply TransformSimilaritySymmetric.
      apply TransformSimilarityPreservesIsAnswer.
      apply TransformSimilaritySymmetric.
      apply TransformSimilarityTyped.
      apply TransformSimilaritySymmetric.
      eauto.
      destructecs Htred3.
      exists Gb2 Cb2; splits; eauto.
      eapply trf_transitive; eauto.
      forwards: TransformSimilarityPreservesIsAnswer typing_mode Htred3; eauto.
Qed.

Lemma typed_red_closure_answer_split1 G1 G2 G3 C1 C2 C3 T n2:
  ty_config G1 C1 T ->
  G1 ⊢ C1 ∶ T ⊣ n2 ↦* G2 ⊢ C2 ->
  answer C2 ->
  G1 ⊢ C1 ∶ T ↦ G3 ⊢ C3 ->
  exists n32,
  n2 = S n32 /\
  exists G32 C32,
  G3 ⊢ C3 ∶ T ⊣ n32 ↦* G32 ⊢ C32 /\
  〈G2 ⊢ C2 ∶ T〉 ∼(trf_similarity typing_mode)∼> 〈G32 ⊢ C32 ∶ T〉 /\
  answer C32.
Proof.
  introv Htyc Hredc Hans Hred.
  forwards: typed_red_step Hred. eauto.
  forwards: typed_red_closure_answer_split Hredc H; fweauto.
  destructecs H0.
  exists n32; splits. rewrite Nat.add_1_l in H1. auto.
  exists G32 C32; eauto.
Qed.

Theorem typed_red_config_terminates Ga1 Ca1 T Ga2 Ca2 n:
  ty_config Ga1 Ca1 T ->
  Ga1 ⊢ Ca1 ∶ T ⊣ n ↦* Ga2 ⊢  Ca2 ->
  config_terminates Ca1 ->
  config_terminates Ca2.
Proof.
  introv Hty Hred Hter.
  hnde Hter.
  hnde Hter.

  forwards Htr: untyped_red_to_typed_closure Hty Hter.
  destructecs Htr.
  forwards Hans: typed_red_closure_answer_split Htr Hred; fweauto.
  destructecs Hans. subst.
  exists n32.
  exists C32. split.
  auto.
  eapply typed_to_untyped_red_closure. eauto.
Qed.

Theorem typed_red_config_terminates1 Ga1 Ca1 T Ga2 Ca2:
  ty_config Ga1 Ca1 T ->
  Ga1 ⊢ Ca1 ∶ T ↦ Ga2 ⊢ Ca2 ->
  config_terminates Ca1 ->
  config_terminates Ca2.
Proof.
  introv Hty Hred Hter.
  eapply typed_red_config_terminates; eauto 1.
  eapply typed_red_step. eapply typed_red_refl. eauto.
Qed.



End WithTypingMode.