(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation Weakening EWeakening Substitution TFree.
Require Import TransformationSimilarity TransformationPreservation TransformationTerm.
Require Import TransformationConfig OperationalSemantics TypedOperationalSemantics.
Require Import ContextSimilarity ConfigSimilarity.
Require Import ConfigTyping TypedProgress.
Require Import TypedPreservation.
Require Import TransformationIterateSim.
Require Import Similarity TypingSimilarity.

Require Import TransformationLifting.
Require Import FreshChain FreshChainTactics.
Require Import TypedOperationalSemantics InertContext.
Require Import Similarity SyntaxSimilarity.
Require Import ExecutionTactics Renaming TypingHelpers.

Require Import NothingFresh TypedPreservation LetLet Defs.
Require Import HeapCorrespondence ConfigVarkinds.
Require Import HeapUpdate ConfigValid.

Section WithTypingMode.
Context (typing_mode: Mode).

(** * Lemmas about similarity transformation and reduction steps. *)

Lemma tred_read_obj_rename G x a fs Sigma e Lv U w U1 S0 x2 LvC w0:
  w0 \notin (fv (trm_read (avar_f x) a) \u fv fs \u fv_heap Sigma \u fv_renv e \u Lv) ->
  w0 \notin ctx_fv G ->
  w0 \notin fv (typ_and (typ_bnd U1) S0) ->
  G ⊢ trm_read (avar_f x) a ; fs ; Sigma ; e ; Lv ∶ U ↦ G §ʷ w ~ typ_and (typ_bnd U1) S0 ⊢ trm_var (avar_f w) ; fs ; Sigma ; e & w ~ x2 ; LvC ->
  G ⊢ trm_read (avar_f x) a ; fs ; Sigma ; e ; Lv ∶ U ↦ G §ʷ w0 ~ typ_and (typ_bnd U1) S0 ⊢ trm_var (avar_f w0) ; fs ; Sigma ; e & w0 ~ x2 ; LvC.
Proof.
  introv Hw0C Hw0G Hw0T Htr.
  inversions Htr; try discriminate.
  inversions H6.
  forwards Heqe: eq_push_inv H9. destructecs Heqe. subst.
  eapply tred_read_obj; eauto.
Qed.

Lemma tred_lit_obj_rename G R w t fs S y e d Lva T w0 y0:
  fresh (fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap S \u fv_renv e \u Lva) (length (y0 :: w0 :: nil)) (y0 :: w0 :: nil) ->
  y0 \notin ctx_fv G \u fv R ->
  w0 \notin ctx_fv G \u fv R ->
    G ⊢ trm_lit (lit_obj R d) t ; fs ; S ; e ; Lva ∶ T
  ↦ 
    ((G §ʸ y ~ 
     (typ_objref_loc typing_mode R)) §ʷ w ~
     (typ_objref_loc typing_mode R)) ⊢
     (t /[-> w] ; fs ; S & y ~ item_obj (Deref.deref_defs (e & w ~ y) d /[-> w]) ; e & w ~ y
     ; fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap S \u fv_renv e \u Lva)
  ->
  G ⊢ trm_lit (lit_obj R d) t ; fs ; S ; e ; Lva ∶ T
        ↦ 
  ((G §ʸ y0 ~
    (typ_objref_loc typing_mode R)) §ʷ w0 ~
   (typ_objref_loc typing_mode R)) ⊢
  (t /[-> w0] ; fs ; S & y0 ~ item_obj (Deref.deref_defs (e & w0 ~ y0) d /[-> w0]) ; e & w0 ~ y0
   ; fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap S \u fv_renv e \u Lva).
Proof.
  introv Hw0C Hw0G Hw0T Htr.
  inversions Htr; try discriminate.
  inversions H6.
  
  assert (w \notin fv t0). {
    sympl in Hfr. destructecs Hfr. 
    fresh_chain_init Fr. fresh_chain_add_fresh Fr Hfr1. fresh_chain_cleanup Fr. fresh_chain_solve Fr.
  }
  eapply open_fresh_injective in H11.
  forwards HeqS: eq_push_inv H12. destructecs HeqS.
  
  eapply tred_let_obj; eauto.
  auto.
  auto.
Qed.

Lemma vars_conf_unique_t_Lv_subset t1 t2 fs S e Lv xs:
  vars_conf_unique xs (t1 ; fs ; S ; e ; Lv) ->
  fv t1 \c Lv ->
  vars_conf_unique xs (t2 ; fs ; S ; e ; Lv).
Proof.
  introv Hsf Hsu.
  hnf in Hsf. hnf.
  destructecs Hsf.
  splits.
  auto.
  eapply subset_trans. eauto.
  union_sub_solver.
Qed.

(* Computes freshness chain for similar_conf_1to1_lit_obj. Doing it in the lemma for some 
reason made Qed take really long *)

Lemma similar_conf_1to1_lit_obj_fresh_chain t wa fs S ya e d Lv R xs1:
  ok e ->
  fresh_in_config (wa::ya::nil) (trm_lit (lit_obj R d) t; fs; S; e; Lv) ->
  vars_conf_unique xs1 (config_mk (trm_lit (lit_obj R d) t) fs S e Lv) ->
  fresh_chain_is
        (fresh_chain_var (fresh_element_var wa)
           (fresh_chain_var (fresh_element_var ya)
              (fresh_chain_subset (fresh_element_set (from_list xs1))
                 (fresh_chain_set (fresh_element_set Lv)
                    (fresh_chain_set (fresh_element_renv e)
                       (fresh_chain_set (fresh_element_set (fv_in_values fv S))
                          (fresh_chain_set (fresh_element_set (dom S))
                             (fresh_chain_set (fresh_element_element fs)
                                (fresh_chain_set (fresh_element_element R)
                                   (fresh_chain_set (fresh_element_element d) (fresh_chain_set (fresh_element_element t) fresh_chain_nil))))))))))).
Proof.
  introv Hoke Hfr Hva.

  fresh_chain_init Fra.
  fresh_chain_add_fresh Fra Hfr.
  sympl in Fra.
  fresh_chain_cleanup Fra.

  rewrite <- (fresh_chain_eq_prepend_susbet_subset) with (X:=from_list xs1) (fc:=(fresh_chain_set (fresh_element_set Lv) _)) in Fra.
  3, 4, 5: apply true.
  2: { simpls. hnf in Hva. destructecs Hva. sympls. eapply subset_trans. eauto. union_sub_solver. } 
  apply Fra.
Qed.


Lemma similar_conf_1to1_lit_obj t wa wb fs S ya yb e d Lv R:
  ok e ->
  fresh_in_config (wa::ya::nil) (trm_lit (lit_obj R d) t; fs; S; e; Lv) ->
  fresh_in_config (wb::yb::nil) (trm_lit (lit_obj R d) t; fs; S; e; Lv) ->
  ConfigValid.config_valid (trm_lit (lit_obj R d) t; fs; S; e; Lv) -> 
  similar_conf_1to1
  (t /[-> wa] ; fs ; S & ya ~ item_obj (Deref.deref_defs (e & wa ~ ya) d /[-> wa]) ; e & wa ~ ya ; (fv_config (trm_lit (lit_obj R d) t; fs; S; e; Lv)) )
  (t /[-> wb] ; fs ; S & yb ~ item_obj (Deref.deref_defs (e & wb ~ yb) d /[-> wb]) ; e & wb ~ yb ; (fv_config (trm_lit (lit_obj R d) t; fs; S; e; Lv)) ).
Proof.
  introv Hoke Hfra Hfrb Hva.

  (* Forward reasoning *)
  forwards Hrefl: similar_conf_1to1_refl Hva.
  hnf in Hrefl.
  destructecs Hrefl.

  forwards Fra: similar_conf_1to1_lit_obj_fresh_chain Hoke Hfra Hrefl0.
  forwards Frb: similar_conf_1to1_lit_obj_fresh_chain Hoke Hfrb Hrefl1.

  (* Back reasoning *)
  hnf. 
  (* Use the lists and prepend the new variables *)
  exists (wa::ya::xs1)%list (wb::yb::xs2)%list; splits.
  -
    eapply vars_conf_unique_t_Lv_subset.
    eapply vars_conf_unique_Lv_subset.


    eapply vars_conf_unique_prepend2_concat. eauto.
    sympls. fresh_chain_solve Fra.
    rewrite ! dom_single. 
    union_in_solver.
    sympls. fresh_chain_solve Fra.
    notin_solve.
    rewrite ! dom_single.
    union_in_solver.
    fresh_chain_solve Fra.
    union_sub_solver.
    union_sub_solver.
  -
    eapply vars_conf_unique_t_Lv_subset.
    eapply vars_conf_unique_Lv_subset.

    eapply vars_conf_unique_prepend2_concat. eauto.
    sympls. fresh_chain_solve Frb.
    rewrite ! dom_single. 
    union_in_solver.
    sympls. fresh_chain_solve Frb.
    notin_solve.
    rewrite ! dom_single.
    union_in_solver.
    fresh_chain_solve Frb.
    union_sub_solver.
    union_sub_solver.
  - simpl; auto.
  - hnf in Hrefl; destructecs Hrefl.
    hnf; splits.

  +
  forwards Hsimlit: similar_trm_lit_case Hrefl3.
  destructecs Hsimlit. inversions Hsimlit0.
  forwards Hsimobj: similar_lit_obj_case Hsimlit1.
  destructecs Hsimobj. inversions Hsimobj0.


  eapply similar_open_rec_push.
  eapply similar_extend.
  auto.


  +
  eapply similar_extend.
  eapply similar_extend.
  auto.
    + (* heap *)
      eapply SimilarityPreservation.similar_reduce_obj_heap; eauto 2 using config_valid_renv_idempotent.
       
      sympls. fresh_chain_solve Fra.
      sympls. fresh_chain_solve Frb.
      sympls. fresh_chain_solve Fra.
      sympls. fresh_chain_solve Frb.

   + 
      eapply similar_env_push.
      eapply similar_extend_notin.
      eauto.
      fresh_chain_solve Fra.
      fresh_chain_solve Frb.
      eapply similar_extend.
      eapply similar_var_push; auto.
      *
      rewrite from_list_cons, notin_union; split.
      fresh_chain_solve Fra.
      hnf in Hrefl0. destructecs Hrefl0.
      eapply notin_subset. eauto.
      sympl. fresh_chain_solve Fra.
      *
      rewrite from_list_cons, notin_union; split.
      fresh_chain_solve Frb.
      hnf in Hrefl1. destructecs Hrefl1.
      eapply notin_subset. eauto.
      sympl. fresh_chain_solve Frb.
Qed.

Lemma trf_similarity_config_lit_obj G R w t fs S y e d Lva T w0 y0:
  ok e ->
  config_valid (trm_lit (lit_obj R d) t ; fs ; S ; e ; Lva)  ->
  fresh (fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap S \u fv_renv e \u Lva) (length (y0 :: w0 :: nil)) (y0 :: w0 :: nil) ->
  fresh (fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap S \u fv_renv e \u Lva) (length (y :: w :: nil)) (y :: w :: nil) ->
  y0 \notin ctx_fv G \u fv R ->
  w0 \notin ctx_fv G \u fv R ->
  (
    G ⊢ trm_lit (lit_obj R d) t ; fs ; S ; e ; Lva ∶ T
    ↦ 
    ((G §ʸ y ~ (typ_objref_loc typing_mode R)) §ʷ w ~ (typ_objref_loc typing_mode R)) ⊢
    (t /[-> w] ; fs ; S & y ~ item_obj (Deref.deref_defs (e & w ~ y) d /[-> w]) ; e & w ~ y
   ; fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap S \u fv_renv e \u Lva)
  ) ->
  〈 (G §ʸ y ~
    (typ_objref_loc typing_mode R)) §ʷ w ~
   (typ_objref_loc typing_mode R)
  ⊢ t /[-> w] ; fs ; S & y ~ item_obj (Deref.deref_defs (e & w ~ y) d /[-> w]) ; e & w ~ y
    ; fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T 〉 ∼ trf_similarity typing_mode ∼>
  〈 (G §ʸ y0 ~
      (typ_objref_loc typing_mode R)) §ʷ w0 ~
     (typ_objref_loc typing_mode R)
  ⊢ t /[-> w0] ; fs ; S & y0 ~ item_obj (Deref.deref_defs (e & w0 ~ y0) d /[-> w0]) ; e & w0 ~ y0
  ; fv (trm_lit (lit_obj R d) t) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T 〉.
Proof.
  introv Hoke Hva Hw0C HwC Hw0G Hw0T Htr.
  forwards: tred_lit_obj_rename Hw0C; fweauto.
  constructor.
  (* Solve typing conditions *)
  eapply (typed_preservation Htr); eauto.
  eapply (typed_preservation H); eauto.
  (* Base the similarity on the previous configuration *)
  forwards Hsim: similar_conf_1to1_refl_vars Hva.
  destructecs Hsim.
  hnf in Hsim1.
  destructecs Hsim1.
  forwards Hsimlit: similar_trm_lit_case Hsim2.
  destructecs Hsimlit. inversions Hsimlit0.
  forwards Hsimobj: similar_lit_obj_case Hsimlit1.
  destructecs Hsimobj. inversions Hsimobj0.

  eapply similar_conf_1to1_lit_obj; eauto.

Qed.

Lemma tred_lit_fun_rename G T1 T2 wb t1 t2 fs S yb e Lva T wa ya:
  fresh (fv (trm_lit (lit_fun T1 t1) t2) \u fv fs \u fv_heap S \u fv_renv e \u Lva) (length (yb :: wb :: nil)) (yb :: wb :: nil) ->
  yb \notin ctx_fv G \u fv (typ_fun T1 T2) ->
  wb \notin ctx_fv G \u fv (typ_fun T1 T2) ->
  (
      G ⊢  trm_lit (lit_fun T1 t1) t2  ; fs ; S ; e ; Lva ∶ T
      ↦ 
      inert_G_for_item G wa ya (typ_fun T1 T2)
        ⊢ t2 /[-> wa] ; fs ; S & ya ~ item_fun t1 ; e & wa ~ ya ; fv (trm_lit (lit_fun T1 t1) t2) \u fv fs \u fv_heap S \u fv_renv e \u Lva
    )->
  (
    G ⊢  trm_lit (lit_fun T1 t1) t2  ; fs ; S ; e ; Lva ∶ T
    ↦ 
    inert_G_for_item G wb yb (typ_fun T1 T2)
       ⊢ t2 /[-> wb] ; fs ; S & yb ~ item_fun t1 ; e & wb ~ yb ; fv (trm_lit (lit_fun T1 t1) t2) \u fv fs \u fv_heap S \u fv_renv e \u Lva
  ).
Proof.
  introv Hw0C Hw0G Hw0T Htr.
  inversions Htr; try discriminate.
  inversions H6.

  assert (wa \notin fv t4). {
    sympl in Hfr. destructecs Hfr. 
    fresh_chain_init Fr. fresh_chain_add_fresh Fr Hfr1. fresh_chain_cleanup Fr. fresh_chain_solve Fr.
  }
  eapply open_fresh_injective in H11.
  forwards HeqS: eq_push_inv H12. destructecs HeqS.
  
  eapply tred_let_fun; eauto.
  auto.
  auto.
Qed.


Lemma similar_conf_1to1_lit_fun_fresh_chain t1 wa fs S ya e t2 Lv T1 xs1:
  ok e ->
  fresh_in_config (wa::ya::nil) (trm_lit (lit_fun T1 t1) t2; fs; S; e; Lv) ->
  vars_conf_unique xs1 (config_mk (trm_lit (lit_fun T1 t1) t2) fs S e Lv) ->
  fresh_chain_is
        (fresh_chain_var (fresh_element_var wa)
           (fresh_chain_var (fresh_element_var ya)
              (fresh_chain_subset (fresh_element_set (from_list xs1))
                 (fresh_chain_set (fresh_element_set Lv)
                    (fresh_chain_set (fresh_element_renv e)
                       (fresh_chain_set (fresh_element_set (fv_in_values fv S))
                          (fresh_chain_set (fresh_element_set (dom S))
                             (fresh_chain_set (fresh_element_element fs)
                                (fresh_chain_set (fresh_element_element T1)
                                   (fresh_chain_set (fresh_element_element t1) (fresh_chain_set (fresh_element_element t2) fresh_chain_nil))))))))))).
Proof.
  introv Hoke Hfr Hva.

  fresh_chain_init Fra.
  fresh_chain_add_fresh Fra Hfr.
  sympl in Fra.
  fresh_chain_cleanup Fra.

  rewrite <- (fresh_chain_eq_prepend_susbet_subset) with (X:=from_list xs1) (fc:=(fresh_chain_set (fresh_element_set Lv) _)) in Fra.
  3, 4, 5: apply true.
  2: { simpls. hnf in Hva. destructecs Hva. sympls. eapply subset_trans. eauto. union_sub_solver. } 
  apply Fra.
Qed.



Lemma similar_conf_1to1_lit_fun  wa wb fs S ya yb e t1 t2 Lv T1:
  ok e ->
  fresh_in_config (wa::ya::nil) ( trm_lit (lit_fun T1 t1) t2; fs; S; e; Lv) ->
  fresh_in_config (wb::yb::nil) ( trm_lit (lit_fun T1 t1) t2; fs; S; e; Lv) ->
  ConfigValid.config_valid ( trm_lit (lit_fun T1 t1) t2; fs; S; e; Lv) -> 
  similar_conf_1to1
  (t2 /[-> wa] ; fs ;  S & ya ~ item_fun t1 ; e & wa ~ ya ; (fv_config ( trm_lit (lit_fun T1 t1) t2; fs; S; e; Lv)) )
  (t2 /[-> wb] ; fs ;  S & yb ~ item_fun t1 ; e & wb ~ yb ; (fv_config ( trm_lit (lit_fun T1 t1) t2; fs; S; e; Lv)) ).
Proof.
  introv Hoke Hfra Hfrb Hva.

  (* Forward reasoning *)
  forwards Hrefl: similar_conf_1to1_refl Hva.
  hnf in Hrefl.
  destructecs Hrefl.

  forwards Fra: similar_conf_1to1_lit_fun_fresh_chain Hoke Hfra Hrefl0.
  forwards Frb: similar_conf_1to1_lit_fun_fresh_chain Hoke Hfrb Hrefl1.

  (* Back reasoning *)
  hnf. 
  (* Use the lists and prepend the new variables *)
  exists (wa::ya::xs1)%list (wb::yb::xs2)%list; splits.
  -
    eapply vars_conf_unique_t_Lv_subset.
    eapply vars_conf_unique_Lv_subset.


    eapply vars_conf_unique_prepend2_concat. eauto.
    sympls. fresh_chain_solve Fra.
    rewrite ! dom_single. 
    union_in_solver.
    sympls. fresh_chain_solve Fra.
    notin_solve.
    rewrite ! dom_single.
    union_in_solver.
    fresh_chain_solve Fra.
    union_sub_solver.
    union_sub_solver.
  -
    eapply vars_conf_unique_t_Lv_subset.
    eapply vars_conf_unique_Lv_subset.

    eapply vars_conf_unique_prepend2_concat. eauto.
    sympls. fresh_chain_solve Frb.
    rewrite ! dom_single. 
    union_in_solver.
    sympls. fresh_chain_solve Frb.
    notin_solve.
    rewrite ! dom_single.
    union_in_solver.
    fresh_chain_solve Frb.
    union_sub_solver.
    union_sub_solver.
  - simpl; auto.
  - hnf in Hrefl; destructecs Hrefl.
    hnf; splits.

  +
    forwards Hsimlit: similar_trm_lit_case Hrefl3.
    destructecs Hsimlit. inversions Hsimlit0.
    forwards Hsimobj: similar_lit_fun_case Hsimlit1.
    destructecs Hsimobj. inversions Hsimobj0.

    eapply similar_open_rec_push.
    eapply similar_extend.
    auto.

  +
    eapply similar_extend.
    eapply similar_extend.
    auto.
  + (* heap *)
      eapply SimilarityPreservation.similar_reduce_fun_heap; eauto 2 using config_valid_renv_idempotent.
       
      sympls. fresh_chain_solve Fra.
      sympls. fresh_chain_solve Frb.

  + 
      eapply similar_env_push.
      eapply similar_extend_notin.
      eauto.
      fresh_chain_solve Fra.
      fresh_chain_solve Frb.
      eapply similar_extend.
      eapply similar_var_push; auto.
      *
      rewrite from_list_cons, notin_union; split.
      fresh_chain_solve Fra.
      hnf in Hrefl0. destructecs Hrefl0.
      eapply notin_subset. eauto.
      sympl. fresh_chain_solve Fra.
      *
      rewrite from_list_cons, notin_union; split.
      fresh_chain_solve Frb.
      hnf in Hrefl1. destructecs Hrefl1.
      eapply notin_subset. eauto.
      sympl. fresh_chain_solve Frb.
Qed.

Lemma trf_similarity_config_lit_fun G e T1 T2 (t1 t2: trm) fs S wa wb ya yb Lva T:
  ok e ->
  config_valid ( trm_lit (lit_fun T1 t1) t2 ; fs ; S ; e ; Lva)  ->
  fresh (fv (trm_lit (lit_fun T1 t1) t2) \u fv fs \u fv_heap S \u fv_renv e \u Lva) (length (ya :: wa :: nil)) (ya :: wa :: nil) ->
  fresh (fv (trm_lit (lit_fun T1 t1) t2) \u fv fs \u fv_heap S \u fv_renv e \u Lva) (length (yb :: wb :: nil)) (yb :: wb :: nil) ->
  yb \notin ctx_fv G \u fv (typ_fun T1 T2)->
  wb \notin ctx_fv G\u fv (typ_fun T1 T2) -> 
  (
    G ⊢  trm_lit (lit_fun T1 t1) t2  ; fs ; S ; e ; Lva ∶ T
    ↦ 
    inert_G_for_item G wa ya (typ_fun T1 T2)
       ⊢ t2 /[-> wa] ; fs ; S & ya ~ item_fun t1 ; e & wa ~ ya ; fv (trm_lit (lit_fun T1 t1) t2) \u fv fs \u fv_heap S \u fv_renv e \u Lva
  ) ->
  〈 (G §ʸ ya ~ (typ_fun T1 T2)) §ʷ wa ~ (typ_fun T1 T2) ⊢ t2 /[-> wa] ; fs ; S & ya ~ item_fun t1 ; e & wa ~ ya ; fv (trm_lit (lit_fun T1 t1) t2) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T 〉
  ∼ trf_similarity typing_mode ∼>
  〈 (G §ʸ yb ~ (typ_fun T1 T2)) §ʷ wb ~ (typ_fun T1 T2) ⊢ t2 /[-> wb] ; fs ; S & yb ~ item_fun t1 ; e & wb ~ yb ; fv (trm_lit (lit_fun T1 t1) t2) \u fv fs \u fv_heap S \u fv_renv e \u Lva ∶ T〉.

Proof.
  introv Hoke Hva Hw0C HwC Hw0G Hw0T Htr.
  forwards: tred_lit_fun_rename HwC; fweauto.
  constructor.
  (* Solve typing conditions *)
  eapply (typed_preservation Htr); eauto.
  eapply (typed_preservation H); eauto.
  (* Base the similarity on the previous configuration *)
  forwards Hsim: similar_conf_1to1_refl_vars Hva.
  destructecs Hsim.
  hnf in Hsim1.
  destructecs Hsim1.
  forwards Hsimlit: similar_trm_lit_case Hsim2.
  destructecs Hsimlit. inversions Hsimlit0.
  forwards Hsimobj: similar_lit_fun_case Hsimlit1.
  destructecs Hsimobj. inversions Hsimobj0.

  eapply similar_conf_1to1_lit_fun; eauto.
Qed.



(* Lemmas for reordering environments *)

(* Environment keys can be reordered *)
Lemma env_reorder_keys_exists A (S1 S2: env A):
  ok S1 ->
  ok S2 ->
  dom S1 = dom S2 ->
  exists ns,
  list_reorder (keys S1) ns (keys S2) /\
  List.length ns = List.length S1 /\
  LibList.noduplicates ns.
Proof.
  introv Hok1 Hok2 Hdom.
  reverts S2.
  induction S1; intros.
  - destruct S2; rewrite dom_def in Hdom; simpls.
    + exists (@nil nat).
      rewrite keys_def, LibList.map_nil.
      splits.
      *
      eapply list_reorder_nil.
      * simpl. auto.
      *
      eapply LibList.noduplicates_nil.
    + false. erewrite <- in_empty. rewrite Hdom. rewrite in_union. rewrite in_singleton. auto.
  -
    (* Find the item in S2 *)
    assert (fst a \in dom S2) as Hin. {
     rewrite dom_def in Hdom. simpl in Hdom. rewrite <- dom_def in Hdom.
     rewrite <- Hdom. union_in_solver.
    }
    forwards Hmem: env_dom_mem Hin.
    destructecs Hmem.
    eapply LibList.mem_Nth in Hmem.
    destructecs Hmem.
    (* Apply induction with that item removed *)
    forwards HS1: IHS1 (list_removeat n S2).
    + destruct a. rewrite cons_to_push in Hok1. eapply ok_push_inv. eauto.
    + eapply env_removeat_ok; eauto.
    + erewrite env_removeat_dom_remove. 3: eapply  Hmem. simpls.
      destruct a. rewrite env_dom_cons in Hdom. simpl. rewrite <- Hdom.
      rewrite union_comm.
      rewrite union_remove_swap0. rewrite remove_same. auto.
      rewrite cons_to_push in Hok1. eapply ok_push_inv; eauto.
      auto.
    + destructecs HS1.

      forwards Hl: LibList.Nth_inbound Hmem. rewrite liblist_length in Hl. rewrite LibNat.lt_peano in Hl.

      assert (n <= length ns). {
        forwards: list_reorder_lengths HS0. rewrite <- H. rewrite keys_def. rewrite liblist_map. rewrite List.map_length. rewrite list_removeat_length.
        eapply PeanoNat.Nat.lt_le_pred. auto. auto.
      }

      exists (list_insertat n (List.map S ns) 0). splits.
      *
      rewrite keys_def in HS0.
      rewrite list_removeat_map in HS0.
      rewrite keys_def.
      rewrite LibList.map_cons.
      rewrite liblist_map.
      
      rewrite <- list_remove_insert_at with (S0 := List.map fst S2) (m:=n) (x:= fst a).
      eapply list_reorder_insert_middle_drop_take.
      rewrite <- liblist_map. auto.
      rewrite <- liblist_map. 

      eapply LibList.Nth_map with (f:=fst) in Hmem. simpl in Hmem. auto.
      * rewrite list_insertat_length. simpl. rewrite List.map_length. rewrite HS2. auto. rewrite List.map_length. auto. 

      * rewrite <- liblist_map. rewrite <- list_map_take, <- list_map_drop.
        rewrite liblist_map. eapply list_noduplicates_insert_middle.
        rewrite <- liblist_app. rewrite LibList.list_eq_take_app_drop. auto.
        rewrite LibNat.le_peano, liblist_length. auto.
Qed.


(* Environment can be reordered by its keys *)
Lemma env_reorder_keys_env A (S1: env A) ks2 ns:
  list_reorder (keys S1) ns ks2 ->
  exists S2,
    keys S2 = ks2 /\
    list_reorder S1 ns S2.
Proof.
  introv Hre.

  forwards Hrek: list_reorder_same_length_exists Hre.
  rewrite keys_def. rewrite liblist_map. eapply List.map_length.
  destructecs Hrek.
  exists xs4.
  split.

  forwards: list_reorder_map (@fst var A) Hrek .
  rewrite keys_def in Hre. rewrite liblist_map in Hre.
  forwards: list_reorder_functional Hre H.
  rewrite keys_def, liblist_map. auto.
  auto.
Qed.



Lemma env_reorder_binds A (S1 S2: env A) ns x v:
  list_reorder S1 ns S2 ->
  LibList.noduplicates ns ->
  length S1 = length S2 ->  
  ok S1 ->
  ok S2 -> (* Not necessary, but easy to show *)
  binds x v S1 ->
  binds x v S2.
Proof.
  introv Hre Hnd Hl HokS1 HokS2 Hbi.

  eapply env_mem_binds.
  eapply list_permutation_mem; eauto.
  eapply env_binds_mem; eauto.
  auto.
Qed.


Lemma env_reorder_notin A (S1 S2: env A) ns x:
  list_reorder S1 ns S2 ->
  LibList.noduplicates ns ->
  length S1 = length S2 ->  
  ok S1 ->
  ok S2 -> (* Not necessary, but easy to show *)
  x # S1 ->
  x # S2.
Proof.
  introv Hre Hnd Hl HokS1 HokS2 Hbi.
  intro.
  eapply env_dom_mem in H. destructecs H.
  forwards: list_reorder_mem H; fweauto.
  eapply Hbi.
  eapply env_mem_dom. eauto.
Qed.

Lemma env_reorder_dom A (S1 S2: env A) ns:
  list_reorder S1 ns S2 ->
  LibList.noduplicates ns ->
  length S1 = length S2 ->  
  ok S1 ->
  ok S2 -> (* Not necessary, but easy to show *)
  dom S1 = dom S2.
Proof.
  introv Hre Hnd Hl HokS1 HokS2.
  eapply fset_extens; intro; intro.
  -
    eapply env_dom_mem in H. destructecs H.
    eapply env_mem_dom.
    eapply list_permutation_mem; eauto.
  -
    eapply env_dom_mem in H. destructecs H.
    eapply env_mem_dom.
    forwards: list_reorder_mem H; eauto. 
Qed.

Lemma heap_fv_mem (S: heap) a:
  a \in fv_env S ->
  exists x p,
    LibList.mem (x, p) S /\
    a \in fv p.
Proof.
  introv Hdom.
  induction S.
  - rewrite <- empty_def in Hdom; rewrite fv_env_empty in Hdom. rewrite in_empty in Hdom. contradiction.
  - destruct a0. rewrite cons_to_push in Hdom. rewrite fv_env_push in Hdom.
    rewrite in_union in Hdom. destructecs Hdom; subst.
    + forwards Hp: IHS; fweauto.
      destructecs Hp.
      eauto.
    + (* here *)
      exists v i. splits.
      eauto. auto.
Qed.

Lemma heap_mem_fv (S: heap) a p x:
  LibList.mem (a, p) S ->
  x \in fv p ->
  x \in fv_env S.
Proof.
  introv Hmem Hin.
  induction S.
  - inversions Hmem.
  - inversions Hmem.
    + (* here *)
       rewrite cons_to_push.
       rewrite fv_env_push. rewrite in_union. auto.
    + 
      destruct a0.
      rewrite cons_to_push.
      rewrite fv_env_push.
      rewrite in_union. auto.
Qed.



Lemma heap_reorder_fv (S1 S2: heap) ns:
  list_reorder S1 ns S2 ->
  LibList.noduplicates ns ->
  length S1 = length S2 ->  
  ok S1 ->
  ok S2 -> (* Not necessary, but easy to show *)
  fv_heap S1 = fv_heap S2.
Proof.
  introv Hre Hnd Hl HokS1 HokS2.
  eapply fset_extens; intro; intro.
  -
    rewrite in_union in *.
    destruct H.
    +
      eapply env_dom_mem in H. destructecs H.
      left.
      eapply env_mem_dom.
      eapply list_permutation_mem; eauto.
    +
      right.
      eapply heap_fv_mem in H. destructecs H.
      eapply heap_mem_fv; eauto.
      eapply list_permutation_mem; eauto.
  -
    rewrite in_union in *.
    destruct H.
    + left.
      eapply env_dom_mem in H. destructecs H.
      eapply env_mem_dom.
      forwards: list_reorder_mem H; eauto. 
    + right.
      eapply heap_fv_mem in H. destructecs H.
      eapply heap_mem_fv.
      forwards: list_reorder_mem H0; eauto. eauto.
Qed.


Lemma heap_update_reorder S1 y1 a y2 S2 ns S3:
  heap_update S1 y1 a y2 S2 ->
  list_reorder S2 ns S3 ->
  LibList.noduplicates ns ->
  keys S3 = keys S1 ->
  length S2 = length S3 ->
  heap_update S1 y1 a y2 S3.
Proof.
  introv Hu Hre Hnd Heq Hl.
  hnf in Hu. destructecs Hu.
  hnf; splits.
  - auto. 
  - eapply env_keys_eq_ok. symmetry. eauto. eauto.
  - rewrite <- ! keys_from_list_dom. rewrite Heq. auto.
  - intros. eapply env_reorder_binds; eauto.
    
    eapply env_keys_eq_ok. symmetry. eauto. auto.
  - exists ds ds'; splits; auto.
    eapply env_reorder_binds; eauto.
    eapply env_keys_eq_ok. symmetry. eauto. auto.
Qed.

(* For an updated heap, there is a version that has the same order as the original heap *)
Lemma heap_update_keys_reorder S1 y1 a y2 S2 :
  heap_update S1 y1 a y2 S2 ->
  exists S3 ns,
  heap_update S1 y1 a y2 S3 /\
  list_reorder S2 ns S3 /\
  LibList.noduplicates ns /\
  length S2 = length S3 /\
  keys S3 = keys S1.
Proof.
  introv Hu.
  lets Hu': Hu.
  hnf in Hu. destructecs Hu.
  symmetry in Hu2.
  forwards Hre: env_reorder_keys_exists Hu2; fweauto.
  destructecs Hre.
  forwards HreS: env_reorder_keys_env Hre0.
  destructecs HreS.
  exists S0 ns; splits; auto.

  eapply heap_update_reorder; eauto.
  eapply list_reorder_lengths in HreS. rewrite HreS. auto.
  eapply list_reorder_lengths in HreS. rewrite HreS. auto.
Qed.

Lemma similar_env_with_reorder_env {A: Set} {AS: Similar A} xsa xsb ysa ysb (S1: env A) ns S2 S3:
  similar_env_with AS xsa xsb ysa ysb S1 S3 ->
  list_reorder S3 ns S2 ->
  LibList.noduplicates ns ->
  length S3 = length S2 ->
  ok S3 ->
  ok S2 ->
  similar_env_with AS xsa xsb ysa ysb S1 S2.
Proof.
  introv Hsi Hre Hnd Hl Hok3 Hok2.
  induction Hsi.
  - eapply similar_env_empty. auto.
  - eapply similar_env_dom.
    hnf.
    auto.
    eauto.
    eapply env_reorder_binds; eauto.
    eauto.
  - eapply similar_env_notin.
    hnf. eauto.
    auto.
    eapply env_reorder_notin; eauto.
Qed.


Lemma similar_heap_diag_reorder_env xsa xsb (S1: heap) ns S2 S3:
  similar xsa xsb S1 S3 ->
  list_reorder S3 ns S2 ->
  LibList.noduplicates ns ->
  length S3 = length S2 ->
  ok S3 ->
  ok S2 ->
  similar xsa xsb S1 S2.
Proof.
  eapply similar_env_with_reorder_env.
Qed.


Lemma heap_update_keys_similar (S1: heap) y1 a y2 S2 xs:
  heap_update S1 y1 a y2 S2 ->
  similar xs xs S2 S2 ->
  exists S3 ns,
  heap_update S1 y1 a y2 S3 /\
  similar xs xs S2 S3 /\
  list_reorder S2 ns S3 /\
  LibList.noduplicates ns /\
  length S2 = length S3 /\
  keys S3 = keys S1.
Proof.
  introv Hu Hsi.

  forwards Hre: heap_update_keys_reorder Hu; fweauto.
  destructecs Hre.

  exists S3 ns; splits; auto 1.
  
  eapply similar_heap_diag_reorder_env; eauto.

  hnf in Hu. destructecs Hu. auto.
  hnf in Hre0. destructecs Hre0. auto.
Qed.

Lemma heap_reorder_ty_item_s G e S1 S2 ns x:
  ok S1 ->
  ok S2 ->
  ty_item_s typing_mode G e S1 x ->
  list_reorder S1 ns S2 ->
  LibList.noduplicates ns ->
  length S1 = length S2 ->
  ty_item_s typing_mode G e S2 x.
Proof.
  introv HokS1 HokS2 Htyit Hre Hnd Hl.
  inversions Htyit.
  eapply (env_reorder_binds Hre) in H1; eauto.
  eapply (env_reorder_binds Hre) in H0; eauto.
Qed.



Lemma heap_reorder_correspond G e S1 S2 ns:
  ok S2 ->
  heap_correspond G e S1 ->
  list_reorder S1 ns S2 ->
  LibList.noduplicates ns ->
  length S1 = length S2 ->
  heap_correspond G e S2.
Proof.
  introv HokS2 Hhc Hre Hnd Hl.
  hnf in Hhc. destructecs Hhc.
  forwards: env_reorder_dom Hre; fweauto.

  hnf. splits; auto 1.

  rewrite <- H. auto.
  intros.
  rewrite <- H in H0.
  eapply  heap_reorder_ty_item_s; eauto.
Qed.

Lemma heap_reorder_closed S1 S2 ns:
  ok S1 ->
  heap_closed S1 ->
  list_reorder S1 ns S2 ->
  heap_closed S2.
Proof.
  introv Hok Hc Hre.
  induction Hre.
  -  rewrite <- empty_def. constructor.
  - destruct y. rewrite cons_to_push. constructor. auto.
    eapply Closedness.heap_binds_closed. eauto.
    eapply env_mem_binds. eapply LibList.Nth_mem. eauto. auto.
Qed.
  
Lemma tred_write_similar F'0 w2 fs Sigma2 e w0 Sa Lva T a y1 y3:
  binds w0 y1 e ->
  binds w2 y3 e ->
  heap_update Sa y1 a y3 Sigma2 ->
  ty_config F'0 ( trm_var (avar_f w2) ; fs ; Sigma2 ; e ; fv (trm_write (avar_f w0) a (avar_f w2)) \u fv fs \u fv_heap Sa \u fv_renv e \u Lva )  T ->
  exists Sigma3,
    heap_update Sa y1 a y3 Sigma3 /\
    keys Sigma3 = keys Sa /\
    〈 F'0 ⊢ trm_var (avar_f w2) ; fs ; Sigma2 ; e ; fv (trm_write (avar_f w0) a (avar_f w2)) \u fv fs \u fv_heap Sa \u fv_renv e \u Lva ∶ T 〉
    ∼ trf_similarity typing_mode ∼>
    〈 F'0 ⊢ trm_var (avar_f w2) ; fs ; Sigma3 ; e ; fv (trm_write (avar_f w0) a (avar_f w2)) \u fv fs \u fv_heap Sa \u fv_renv e \u Lva ∶ T 〉.
Proof.
  introv Hbiw0e Hbiw2e Hu Hty.
  forwards Hva: config_typed_valid Hty.
  forwards Hsi: similar_conf_1to1_refl_vars Hva.
  destructecs Hsi.
  hnf in Hsi1. destructecs Hsi1.
  forwards Huk: heap_update_keys_similar Hu Hsi4.
  destructecs Huk.
  assert (ok S3) as HokS3. { hnf in Huk0. destructecs Huk0. auto. }
  assert (ok Sigma2) as HokS2. { hnf in Hu. destructecs Hu. auto. }
  forwards: heap_reorder_fv Huk2; fweauto.

  exists S3. splits.
  +
    auto.
  +
    auto.
  +
    hnf. econstructor.
    -
      eauto.
    -
      inversions Hty.
      econstructor; eauto.
      eapply heap_reorder_correspond; eauto.
      destructecs H12. destructecs H0. splits. split.
      eapply subset_trans. eapply H1.
      rewrite H. sympls. union_sub_solver.
      hnf.
      rewrite <- H. eapply H0.
      hnf.
      rewrite <- H. eapply H12.
      inversions H16.
      constructor; auto 1.
      eapply heap_reorder_closed; eauto.
    -
      econstructor.
      exists xs; splits.
      eauto.
      eapply vars_conf_unique_S_subset. eauto. rewrite H. eapply subset_refl. auto.
      hnf; splits; auto 1.
Qed.

End WithTypingMode.
