(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation TransformationTerm TransformationLifting Weakening EWeakening Substitution.
Require Import TypingHelpers Narrowing.
Require Import ReadableSyntax FreshChain FreshChainTactics.
Require Import Visibility.
Require Import Lia.
Require Import Sealing. 
Require Import Subenvironments.
Require Import Closedness.

(** * A transformation that swaps to consecutive method calls. *)

Section WithTypingMode.
Context (typing_mode: Mode).

(* Definition 26 in "Pure methods for roDOT" *)
(* A local transformation which swaps two calls to read-only methods in a let-chain *)
Inductive trf_swap_calls_local: TermTransform :=
  trf_swap_calls_local_at (x1 x2: var) (m1 m2: met_label) (xp1 xp2 y1 y2: var) t G e (T1 T2 T TZ1 TZ2: typ):
    ctx_ok G ->
    ok e ->
    closed t ->
    fresh (ctx_dom G \u ctx_fv G \u fv_renv e \u fv T \u fv T1 \u fv T2) 2 (y1::y2::nil)%list -> (* Fresh variable names to use for the locals *)
    (* Read-only arguments *)
    G ⁏ e ⊢ x1 ∶ ❪❴ᵐ m1 ∶ 𝖭 , TZ1 → ⊤ ❵❫ ->
    G ⁏ e ⊢ x2 ∶ ❪❴ᵐ m2 ∶ 𝖭 , TZ2 → ⊤ ❵❫ ->
    G ⁏ e ⊢ ❪𝖭❫ <: TZ1 ->
    G ⁏ e ⊢ ❪𝖭❫ <: TZ2 ->
    G ⁏ e ⊢ ❪Call $x1․m1 $xp1❫ ∶ T1 ->
    G ⁏ e ⊢ ❪Call $x2․m2 $xp2❫ ∶ T2 ->
    G ⁏ e ⊢ xp1 ∶ TZ1 ->
    G ⁏ e ⊢ xp2 ∶ TZ2 ->
    G §ˣ y2 ~ T2 §ˣ y1 ~ T1 ⁏ e ⊢ t ∶ T ->
    〈G ⁏ e ⊢ ❪Let y1=Call $x1․m1 $xp1 In Let y2=Call $x2․m2 $xp2 In t❫ ∶ T〉 ∼trf_swap_calls_local∼> 〈G ⁏ e ⊢ ❪Let y2=Call $x2․m2 $xp2 In Let y1=Call $x1․m1 $xp1 In t❫ ∶ T〉.

Global Instance TransformSwapCallsIdtyp:
  TransformEnsures (@TransformCondTypeIdentical _ _ _) (trf_swap_calls_local).
Proof.
  hnf; intros.
  hnf in *; intros.
  inversions H.
  auto.
Qed.


Global Instance TransformSwapReadsRequiresOk: TransformRequires (@TransformCondOk _ _) trf_swap_calls_local.
Proof.
  hnf; intros.
  inversions H.
  hnf.
  eauto.
Qed.

Global Instance TransformSwapCallsSymmetric: SymmetricTransform trf_swap_calls_local.
Proof.
  hnf; intros.
  hnf; intros.
  inversions H.
  eapply trf_swap_calls_local_at; eauto.
  fresh_solve; eauto.
  eapply typing_swap; eauto.
Qed.

Global Instance TransformSwapCallsTyped: TransformRequires (TransformCondTermTyped typing_mode) (trf_swap_calls_local).
Proof.
  hnf in *.
  introv Htr.
  inversions Htr.
  rename H into HokG, H0 into Hoke.
  rename H3 into Htymet1, H4 into Htymet2.
  rename H7 into Htycall1, H8 into Htycall2.
  rename H2 into Hfr, H1 into Hcl.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hfr.
  fresh_chain_cleanup Fr.
  fresh_chain_include_typing Fr Htycall1.
  fresh_chain_include_typing Fr Htycall2.

  let LFr := (fresh_chain_get_set Fr) in
  eapply ty_trm_let with (L:=LFr).
  eauto.
  intros. simpl in H.
  fresh_chain_add_fresh Fr H.

  (* Rewrite the closing into list *)
  rewrite trm_close_let_close_fresh_vars by (simpls; fresh_chain_solve Fr).
  sympls. 

  let LFr := (fresh_chain_get_set Fr) in
  eapply ty_trm_let with (L:=LFr).

  eapply weaken_push.
  eapply Htycall2. auto. fresh_chain_solve Fr.
  intros. simpl in H.
  fresh_chain_add_fresh Fr H.

  rewrite trm_open_open_vars.
  rewrite <- subst_open_close_vars_closed.

  unfold subst_vars.
  rewrite subst_commut_ne by (fresh_chain_solve Fr).
  eapply rename_subst_trm_2; [>(fresh_chain_solve Fr)..|].
  eapply typing_swap; [>(fresh_chain_solve Fr)..|].
  apply H11.
  simpl; auto.
  fresh_chain_solve Fr.
  auto.
Unshelve.
  eapply weaken_ty_trm.
Qed.

(* Lemma 59 in "Pure methods for roDOT" *)
Global Instance TransformSwapCallsWeakening: TransformWeakeningCompat _ _ trf_swap_calls_local.
Proof.
  hnf in *.
  introv Htrf Hwk1 Hwk2 Hti.
  inversions Hwk1.
  inversions Hwk2.
  inversions Hti. inversions H5.
  rename ea0 into ea.
  rename H into Hwk, H0 into HokG, H1 into Hoke.

  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG HokG.

  inversions Htrf.
  clears Ga0 Ga4 Ga5.
  (* Add freshness conditions into a chain *)
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H12.
  fresh_chain_cleanup Fr.

  clear H9 H10 H4 Hoke.

  rename H11 into Hclt, H13 into Htymet1, H14 into Htymet2, H15 into Htyz1, H16 into Htyz2, H17 into Htycall1, H18 into Htycall2, H19 into Htyt.
  fresh_chain_include_typing Fr Htycall1.
  fresh_chain_include_typing Fr Htycall2.
 
  fresh_chain_add_gather Fr; [> fresh_chain_solve FrG .. |].

  (* The variables y1,y2 are not fresh in G2, so replace them by new fresh variables *)
  fresh_chain_pick Fr z1.
  fresh_chain_pick Fr z2.

  rewrite ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  (* Prepare replacement of closed terms *)
  forwards Ht12: (close_vars_subst_vars t (y1 :: y2 :: nil)%list (z1 :: z2 :: nil)%list 0).
  auto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

  forwards Ht21: (close_vars_subst_vars t (y2 :: y1 :: nil)%list (z2 :: z1 :: nil)%list 0).
  eauto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  rewrite subst_commut_ne in Ht12 by (fresh_chain_solve Fr).

  (* Replace closed terms *)
  rewrite Ht12, Ht21.
  rewrite <- ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  eapply trf_swap_calls_local_at with (T1:=T1) (T2:=T2) (TZ1:=TZ1) (TZ2:=TZ2).
  - fresh_chain_solve FrG.
  - fresh_chain_solve Fr.
  - eapply closed_subst_var. eapply closed_subst_var. eauto.
  - fresh_chain_solve Fr.
  - eapply weaken_ty; eauto. fresh_chain_solve FrG.
  - eapply weaken_ty; eauto. fresh_chain_solve FrG.
  - eapply weaken_ty; eauto. fresh_chain_solve FrG.
  - eapply weaken_ty; eauto. fresh_chain_solve FrG.
  - eapply weaken_ty; eauto. fresh_chain_solve FrG.
  - eapply weaken_ty; eauto. fresh_chain_solve FrG.
  - eapply weaken_ty; eauto. fresh_chain_solve FrG.
  - eapply weaken_ty; eauto. fresh_chain_solve FrG.
  -
    rewrite <- ctx_concat_push.
    rewrite <- ctx_concat_push.
    eapply weaken_ty. ctx_flatten.
    eapply rename_subst_trm_2.
    fresh_chain_solve FrG.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    auto 1.
    auto 1.
    fresh_chain_solve FrG.

Unshelve.
  eapply weaken_ty_var.
  eapply weaken_ty_var.
  eapply weaken_subtyp.
  eapply weaken_subtyp.
  eapply weaken_ty_trm.
  eapply weaken_ty_trm.
  eapply weaken_ty_var.
  eapply weaken_ty_var.
  eapply weaken_ty_trm.
Qed.

Global Instance TransformSwapCallsEWeakening: TransformEWeakeningCompat _ _ trf_swap_calls_local.
Proof.
  hnf in *.
  introv Htrf Hwk1 Hwk2 Hti.
  inversions Hwk1.
  inversions Hwk2.
  inversions Hti. inversions H3.
  rename Ga0 into Ga.
  rename H1 into HokG, H0 into Hoke.

  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG Hoke.

  inversions Htrf.
  clears e0 e4 e5.
  (* Add freshness conditions into a chain *)
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H11.
  fresh_chain_cleanup Fr.

  clear H8 H9 HokG.

  rename H10 into Hclt, H12 into Htymet1, H13 into Htymet2, H14 into Htyz1, H15 into Htyz2, H16 into Htycall1, H17 into Htycall2, H18 into Htyt.
  fresh_chain_include_typing Fr Htycall1.
  fresh_chain_include_typing Fr Htycall2.
 
  fresh_chain_add_gather Fr; [> fresh_chain_solve FrG .. |].

  (* The variables y1,y2 are not fresh in G2, so replace them by new fresh variables *)
  fresh_chain_pick Fr z1.
  fresh_chain_pick Fr z2.

  rewrite ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  (* Prepare replacement of closed terms *)
  forwards Ht12: (close_vars_subst_vars t (y1 :: y2 :: nil)%list (z1 :: z2 :: nil)%list 0).
  auto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

  forwards Ht21: (close_vars_subst_vars t (y2 :: y1 :: nil)%list (z2 :: z1 :: nil)%list 0).
  eauto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  rewrite subst_commut_ne in Ht12 by (fresh_chain_solve Fr).

  (* Replace closed terms *)
  rewrite Ht12, Ht21.
  rewrite <- ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  eapply trf_swap_calls_local_at with (T1:=T1) (T2:=T2) (TZ1:=TZ1) (TZ2:=TZ2).
  - fresh_chain_solve Fr.
  - fresh_chain_solve FrG.
  - eapply closed_subst_var. eapply closed_subst_var. eauto.
  - fresh_chain_solve Fr.
  - eapply ewkn; eauto. fresh_chain_solve FrG.
  - eapply ewkn; eauto. fresh_chain_solve FrG.
  - eapply ewkn; eauto. fresh_chain_solve FrG.
  - eapply ewkn; eauto. fresh_chain_solve FrG.
  - eapply ewkn; eauto. fresh_chain_solve FrG.
  - eapply ewkn; eauto. fresh_chain_solve FrG.
  - eapply ewkn; eauto. fresh_chain_solve FrG.
  - eapply ewkn; eauto. fresh_chain_solve FrG.
  -

    eapply ewkn.
    eapply rename_subst_trm_2.
    fresh_chain_solve Fr.
    fresh_chain_solve FrG.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    auto 1.
    auto 1.
    fresh_chain_solve FrG.

Unshelve.
  eapply ewkn_ty_var.
  eapply ewkn_ty_var.
  eapply ewkn_sub.
  eapply ewkn_sub.
  eapply ewkn_ty_trm.
  eapply ewkn_ty_trm.
  eapply ewkn_ty_var.
  eapply ewkn_ty_var.
  eapply ewkn_ty_trm.
Qed.


Global Instance TransformSwapCallsNarrowing: TransformNarrowingCompat _ _ _ trf_swap_calls_local.
Proof.
  hnf in *.
  introv Htrf Hna1 Hna2 Hti.
  inversions Hna1.
  inversions Hna2.
  inversions Hti. inversions H5.
  rename e0 into e.
  rename H into HokG, H0 into Hoke.

  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG HokG.

  inversions Htrf.
  clears Ga0 x0 Txa0 Ga3.
  (* Add freshness conditions into a chain *)
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H13.
  fresh_chain_cleanup Fr.

  clear H10 H3 H11 Hoke.

  rename H12 into Hclt, H14 into Htyx1, H15 into Htyx2, H16 into Hsuz1, H17 into Hsuz2, H18 into Hty1, H19 into Hty2, H20 into Htyp1, H21 into Htyp2.
  fresh_chain_include_typing Fr Hty1.
  fresh_chain_include_typing Fr Hty2.

  fresh_chain_add_gather Fr; [> fresh_chain_solve FrG .. |].

  (* The variables y1,y2 are not fresh in G2, so replace them by new fresh variables *)
  fresh_chain_pick Fr z1.
  fresh_chain_pick Fr z2.

  fresh_chain_add_var FrG z1. 2: fresh_chain_solve Fr.
  fresh_chain_add_var FrG z2. 2: fresh_chain_solve Fr.

  rewrite ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  (* Prepare replacement of closed terms *)
  forwards Ht12: (close_vars_subst_vars t (y1 :: y2 :: nil)%list (z1 :: z2 :: nil)%list 0).
  auto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

  forwards Ht21: (close_vars_subst_vars t (y2 :: y1 :: nil)%list (z2 :: z1 :: nil)%list 0).
  eauto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  rewrite subst_commut_ne in Ht12 by (fresh_chain_solve Fr).

  (* Replace closed terms *)
  rewrite Ht12, Ht21.
  rewrite <- ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).


assert (ctx_ok ((Ga1 § x / vk ~ Txb) §& Ga2 §§& objctx_empty)) as HokG. {
  unfold ctx_obj.
  fresh_chain_solve FrG.
}

assert ( @subctx (subctx0 typing_mode) e ((Ga1 § x / vk ~ Txb) §& Ga2) ((Ga1 § x / vk ~ Txa) §& Ga2) ) as Hsubctx. {
      eapply subctx_concat. eapply subctx_grow. eapply subctx_refl.
    fresh_chain_solve FrG.
    fresh_chain_solve FrG.
    eauto.
    fresh_chain_solve FrG.
    fresh_chain_solve FrG.
}


  eapply trf_swap_calls_local_at with (T1:=T1) (T2:=T2).
  - fresh_chain_solve FrG.
  - fresh_chain_solve Fr.
  - eapply closed_subst_var. eapply closed_subst_var. eauto.
  - fresh_chain_solve Fr.
  - eapply narrow; eauto. eapply Htyx1. eauto.
  - eapply narrow; eauto. eapply Htyx2. eauto.
  - eapply narrow; eauto. eapply Hsuz1. eauto.
  - eapply narrow; eauto. eapply Hsuz2. eauto.
  - eapply narrow; eauto. eapply Hty1. eauto.
  - eapply narrow; eauto. eapply Hty2. eauto.
  - eapply narrow; eauto. eapply Htyp1. eauto.
  - eapply narrow; eauto. eapply Htyp2. eauto.
  -
    rewrite <- ctx_concat_push.
    rewrite <- ctx_concat_push.
    eapply narrow.
    3: {
      ctx_flatten.
      eapply subctx_push.
      eapply subctx_push.
      eapply subctx_concat.
      eapply subctx_grow.
      eapply subctx_refl.

      fresh_chain_solve FrG.
      fresh_chain_solve FrG.
      eauto.
    fresh_chain_solve FrG.
    fresh_chain_solve FrG.
    fresh_chain_apply FrG. rewrite fresh_chain_eq_merge_var_ctxdom, fresh_chain_eq_merge_var_ctxdompush; try apply true. fresh_chain_solve FrG.
    fresh_chain_apply FrG. rewrite fresh_chain_eq_merge_var_ctxdom, fresh_chain_eq_merge_var_ctxdompush; try apply true. fresh_chain_solve FrG.
    fresh_chain_apply FrG. rewrite fresh_chain_eq_merge_var_ctxdom, fresh_chain_eq_merge_var_ctxdompush; try apply true. fresh_chain_solve FrG.
    fresh_chain_apply FrG. rewrite fresh_chain_eq_merge_var_ctxdom, fresh_chain_eq_merge_var_ctxdompush; try apply true. fresh_chain_solve FrG.
    }
    eapply rename_subst_trm_2.
    fresh_chain_solve FrG.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    auto 1.
    unfold ctx_obj.
    fresh_chain_solve FrG.

Unshelve.
  eapply narrow_ty_var.
  eapply narrow_ty_var.
  eapply narrow_subtyp.
  eapply narrow_subtyp.
  eapply narrow_ty_trm.
  eapply narrow_ty_trm.
  eapply narrow_ty_var.
  eapply narrow_ty_var.
  eapply narrow_ty_trm.
Qed.

Global Instance TransformSwapCallsPreservesAnswer:
  TransformPreserves TransformIsAnswerTerm trf_swap_calls_local.
Proof. (* easy, this transformation only works on non-answers *)
  hnf. intros.
  hnf in *.
  destruct TEa, TEb.
  inversions H.
  inversions H0.
  inversions H.
Qed.

Lemma trf_swap_calls_local_subst_rewrite x1 xp1 f1 x2 xp2 f2 (t: trm) (y1 y2 x0 y: var):
  fresh (\{x1} \u \{x2} \u \{xp1} \u \{xp2} \u \{x0} \u \{y}) 2 (y1 :: y2 :: nil) ->
  (
    (trm_let (trm_call (avar_f x1) f1 (avar_f xp1)) (trm_let (trm_call (avar_f x2) f2 (avar_f xp2)) t /[<- y2]) /[<- y1]) /[ x0 -> y]
  ) = (
    (trm_let (trm_call (avar_f x1 /[ x0 -> y]) f1 (avar_f xp1 /[ x0 -> y])) (trm_let (trm_call (avar_f x2 /[ x0 -> y]) f2 (avar_f xp2 /[ x0 -> y])) t /[ x0 -> y] /[<- y2]) /[<- y1])
  ).
Proof.
  introv Hfr.
  
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hfr.
  fresh_chain_cleanup Fr.

    autorewrite with simpl_close. (* Push closing down *)
    sympl. (* Push substitution down *)
    rewrite ! subst_close_commut. (* Swap closing and subst *)
    autorewrite with simpl_close.
    fequal. (* Solve first trm_let *)
    fequal. (* Solve second trm_let *)

    assert (x2 \[ x0 -> y] <> y1). {
      cases_if.
      fresh_chain_solve Fr.
      fresh_chain_solve Fr.
    }
    assert (xp2 \[ x0 -> y] <> y1). {
      cases_if.
      fresh_chain_solve Fr.
      fresh_chain_solve Fr.
    }

    fequal. (* Solve second trm_call *)

    rewrite If_r with (P:= x2 = y1).
    rewrite If_r with (P:= x2 /[ x0 -> y] = y1). auto. auto.

    fresh_chain_solve Fr.

    rewrite If_r with (P:= xp2 = y1).
    rewrite If_r with (P:= xp2 /[ x0 -> y] = y1). auto. auto.

    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.

Qed.


Local Instance typing_mode':Mode := typing_mode.

Global Instance TransformSwapCallsSubstCompat: TransformSubstCompat typing_mode _ _ trf_swap_calls_local.
Proof.

intros.
  introv Htrf Hsu1 Hsu2 Hti.
  hnf in *.
  destruct TEa2, TEb2.
  inversions Hti.
  inversions Hsu1.
  inversions Hsu2.

  rewrite <- H in H6.

  inversions Htrf.

  rewrite H15 in *.
  eapply eq_push_inv in H16. destructecs H16; subst.

  fresh_chain_init FrG.
  fresh_chain_add_fresh FrG H8.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H22.
  fresh_chain_cleanup Fr.

  forwards Htysu1: (substitution_epush ) (@subst_ty_trm typing_mode)    H28; fweauto.
  forwards Htysu2: (substitution_epush ) (@subst_ty_trm typing_mode)    H27; fweauto.

  forwards Htysux1: (substitution_epush ) (@subst_ty_var typing_mode)    H23; fweauto.
  forwards Htysux2: (substitution_epush ) (@subst_ty_var typing_mode)    H24; fweauto.

  forwards HtysuZ1: (substitution_epush ) (@subst_sub typing_mode)    H25; fweauto.
  forwards HtysuZ2: (substitution_epush ) (@subst_sub typing_mode)    H26; fweauto.

  forwards Htysuxp1: (substitution_epush ) (@subst_ty_var typing_mode)    H29; fweauto.
  forwards Htysuxp2: (substitution_epush ) (@subst_ty_var typing_mode)    H30; fweauto.

  sympl in Htysux1.
  sympl in Htysux2.
  sympl in HtysuZ1.
  sympl in HtysuZ2.
  sympl in Htysuxp1.
  sympl in Htysuxp2.


  fresh_chain_include_typing Fr H28.
  fresh_chain_include_typing Fr H27.

  rewrite fresh_element_eq_containing_renv_push_dom in Fr by apply true.
  rewrite fresh_element_eq_containing_renv_push_fv in Fr by apply true.
  rewrite fresh_element_eq_containing_renv_push_env in Fr by apply true.

  rewrite fresh_element_eq_containing_ctx_concat_r in Fr by apply true.
  rewrite fresh_element_eq_containing_ctx_concat_l in Fr by apply true.
  rewrite fresh_element_eq_containing_ctx_push_ctx in Fr by apply true.

  fresh_chain_add_var FrG y2. 2: fresh_chain_solve Fr.
  fresh_chain_add_var FrG y1. 2: fresh_chain_solve Fr.

  (* Make it easier to solve ctx_ok with substitution *)
  rewrite fresh_element_eq_containing_refl with (fe:=(fresh_element_ctxdom G3)) in FrG by apply true .
  rewrite fresh_element_eq_ctxdom_map with (f:=subst_var x0 y) in FrG at 2 by apply true.

  assert (ctx_ok (G0 §& subst_ctx x0 y G3)) as HokGs. {
    eapply ctx_ok_concat_map. fresh_chain_apply FrG. fresh_chain_solve FrG.
  }

  (* Rewrite one side into the form needed by trf_swap_reads_local_at *)

  rewrite trf_swap_calls_local_subst_rewrite. 2: fresh_chain_solve Fr.
  rewrite trf_swap_calls_local_subst_rewrite.

  eapply trf_swap_calls_local_at.
  9: eauto. (* Solve first call typing, this resolves ?T1 *)
  9: eauto. (* Solve second call typing, this resolves ?T2 *)
  5: eauto.
  5: eauto.
  5: eauto.
  5: eauto.
  5: eauto.
  5: eauto.
  - auto.
  - auto.
  - eapply closed_subst_var; eauto.
  - fresh_solve;
    rewrite ? ctx_fv_concat; try apply notin_fv_subst_var;
    try fresh_chain_solve Fr; try fresh_chain_solve FrG;  try fresh_chain_solve2 FrG.
    rewrite notin_union; split;  try fresh_chain_solve Fr.
    eapply notin_subst_ctx_fv;  try fresh_chain_solve Fr.
    rewrite notin_union; split;  try fresh_chain_solve Fr.
    eapply notin_subst_ctx_fv;  try fresh_chain_solve Fr.
  -

    rewrite <- ! ctx_concat_push.
    rewrite <- ! ctx_subst_push.

    eapply substitution_epush. eapply (@subst_ty_trm typing_mode).
    +
      ctx_flatten; eauto.
    +
      fresh_chain_solve FrG.
    + auto.
    +
      eauto.
    +
      ctx_flatten; eauto.
      eapply weaken_push.
      eapply weaken_push.
      eauto.

      fresh_chain_solve FrG.
      fresh_chain_apply FrG.
      fresh_chain_simpl; fresh_chain_solve FrG.

      fresh_chain_apply FrG. 
      fresh_chain_simpl; fresh_chain_solve FrG.

      fresh_chain_apply FrG. rewrite <- fresh_chain_eq_last_var_set by apply true.
      rewrite fresh_chain_eq_merge_var_ctxdompush by apply true.
      rewrite fresh_chain_eq_merge_var_ctxdom by apply true.
      fresh_chain_solve FrG.
    + intros.
      ctx_flatten in H8.
      eapply ctx_vis_push_n_inv in H8.
      eapply ctx_vis_push_n_inv in H8.
      ctx_flatten.
      eapply ctx_vis_n.   fresh_chain_solve Fr.
      eapply ctx_vis_n.   fresh_chain_solve Fr.
      eauto.
      fresh_chain_solve Fr.
      fresh_chain_solve Fr.
  -  fresh_chain_solve Fr.
Qed.

Lemma seal_trf_swap_calls_local x1 m1 xp1 x2 m2 xp2 t x:
  (trm_let (trm_call (avar_f x1) m1 (avar_f xp1)) (trm_let (trm_call (avar_f x2) m2 (avar_f xp2)) t ) ) /[# 0 ->> x] =
  (trm_let (trm_call (avar_f x1) m1 (avar_f xp1)) (trm_let (trm_call (avar_f x2) m2 (avar_f xp2)) t /[# 2 ->> x] ) )  .
Proof.  
  sympl. auto.
Qed.




Global Instance TransformSwapCallsSealing: TransformSealingCompat _ _ trf_swap_calls_local.
Proof.
  hnf in *.
  introv Htrf Hwk1 Hwk2 Hti.
  inversions Hwk1.
  inversions Hwk2.
  inversions Hti.
   inversions H.
  rename e0 into e.

  fresh_chain_init FrG.


  inversions Htrf.


  (* Add freshness conditions into a chain *)
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H11.
  fresh_chain_cleanup Fr.

  rename H10 into Hclt, H12 into Htymet1, H13 into Htymet2, H14 into Htyz1, H15 into Htyz2, H16 into Htycall1, H17 into Htycall2, H18 into Htyt.
  fresh_chain_include_typing Fr Htycall1.
  fresh_chain_include_typing Fr Htycall2.
 
  fresh_chain_add_gather Fr; [> fresh_chain_solve FrG .. |].

  (* The variables y1,y2 are not fresh in G2, so replace them by new fresh variables *)
  fresh_chain_pick Fr z1.
  fresh_chain_pick Fr z2.

  rewrite ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  (* Prepare replacement of closed terms *)
  forwards Ht12: (close_vars_subst_vars t (y1 :: y2 :: nil)%list (z1 :: z2 :: nil)%list 0).
  auto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

  forwards Ht21: (close_vars_subst_vars t (y2 :: y1 :: nil)%list (z2 :: z1 :: nil)%list 0).
  eauto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  rewrite subst_commut_ne in Ht12 by (fresh_chain_solve Fr).


  rewrite seal_trf_swap_calls_local.
  rewrite seal_trf_swap_calls_local.


  (* Replace closed terms *)
  rewrite Ht12, Ht21.

  erewrite <- ! close_vars_seal_commut.


  2: simpl; lia.
  2: fresh_chain_solve Fr.
  2: simpl; lia.
  2: fresh_chain_solve Fr.

  rewrite <- ! trm_close_let_close_fresh_vars by (sympl; fresh_chain_solve Fr).
 
  assert (closed t /[ y1 -> z1] /[ y2 -> z2]) as Hclts. {
    eapply closed_subst_var; eapply closed_subst_var; eauto.
  }

  eapply trf_swap_calls_local_at with (T1:=T1/[# 0 ->> x] ) (T2:=T2/[# 0 ->> x] ) (TZ1:=TZ1/[# 0 ->> x] ) (TZ2:=TZ2/[# 0 ->> x] ).
  - apply ctx_seal_ok. auto.
  - fresh_chain_solve Fr.
  - erewrite closed_sealed; eapply closed_subst_var; eapply closed_subst_var; eauto.
  -
    rewrite <- ctx_seal_dom. fresh_solve; try eapply fv_seal_notin; try eapply ctx_fv_seal_notin; try fresh_chain_solve Fr.

  -
    replace (x1) with (x1/[# 0 ->> x]).
    replace (typ_rcd (dec_met m1 typ_nocap TZ1 /[# 0 ->> x] typ_top)) with ((typ_rcd (dec_met m1 typ_nocap TZ1 typ_top))/[# 0 ->> x]).
    eapply sealing_obj with (Go:=objctx_empty). auto.
    sympl; auto.
    sympl; auto.

  -
    replace (x2) with (x2/[# 0 ->> x]).
    replace (typ_rcd (dec_met m2 typ_nocap TZ2 /[# 0 ->> x] typ_top)) with ((typ_rcd (dec_met m2 typ_nocap TZ2 typ_top))/[# 0 ->> x]).
    eapply sealing_obj with (Go:=objctx_empty). auto.
    sympl; auto.
    sympl; auto.
  
  -
    replace (typ_nocap) with (typ_nocap/[# 0 ->> x]).

    eapply sealing_obj with (Go:=objctx_empty). auto.
    sympl; auto.

  -
    replace (typ_nocap) with (typ_nocap/[# 0 ->> x]).

    eapply sealing_obj with (Go:=objctx_empty). auto.
    sympl; auto.

 - replace (trm_call (avar_f x1) m1 (avar_f xp1)) with ((trm_call (avar_f x1) m1 (avar_f xp1))/[# 0 ->> x]).
    eapply sealing_obj with (Go:=objctx_empty). auto.
    sympl; auto.
  - replace (trm_call (avar_f x2) m2 (avar_f xp2)) with ((trm_call (avar_f x2) m2 (avar_f xp2))/[# 0 ->> x]).
    eapply sealing_obj with (Go:=objctx_empty). auto.
    sympl; auto.


  -
    replace (xp1) with (xp1/[# 0 ->> x]).

    eapply sealing_obj with (Go:=objctx_empty). auto.
    sympl; auto.

  -
    replace (xp2) with (xp2/[# 0 ->> x]).

    eapply sealing_obj with (Go:=objctx_empty). auto.
    sympl; auto.

  - erewrite closed_sealed with (X:=(t /[ y1 -> z1] /[ y2 -> z2])) by eapply Hclts.
    erewrite <- closed_sealed with (X:=(t /[ y1 -> z1] /[ y2 -> z2])) (n:=0) (y:=x) by eapply Hclts.
    rewrite <- ctx_seal_push.
    rewrite <- ctx_seal_push.

    eapply sealing_obj with (Go:=objctx_empty).
    eapply rename_subst_trm_2.
    auto.
    auto.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    auto 1.

Unshelve.

  5: eapply TransformSealClose.
  5: eapply TransformSealClose.

  all: try eapply sealing_ty_trm.
  all: try eapply sealing_ty_var.
  all: try eapply sealing_subtyp.
Qed.


Global Instance TransformSwapCallsUnhiding: TransformUnhidingCompat _ _ trf_swap_calls_local.
Proof.
  hnf in *.
  introv Htrf Hwk1 Hwk2 Hti.
  inversions Hwk1.
  inversions Hwk2.
  inversions Hti. inversions H.
  rename e0 into e.

  inversions Htrf.
  clears G0 G3.
  (* Add freshness conditions into a chain *)
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr H9.
  fresh_chain_cleanup Fr.

  rename H8 into Hclt, H10 into Htymet1, H11 into Htymet2, H12 into Htyz1, H13 into Htyz2, H14 into Htycall1, H15 into Htycall2, H16 into Htyt.
  fresh_chain_include_typing Fr Htycall1.
  fresh_chain_include_typing Fr Htycall2.
 
  fresh_chain_add_gather Fr. eapply ctx_ok_concat_inv in H6. destructecs H6. auto. eapply ctx_ok_concat_inv in H6. destructecs H6. inversion H0. auto.

  (* The variables y1,y2 are not fresh in G2, so replace them by new fresh variables *)
  fresh_chain_pick Fr z1.
  fresh_chain_pick Fr z2.

  rewrite ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  (* Prepare replacement of closed terms *)
  forwards Ht12: (close_vars_subst_vars t (y1 :: y2 :: nil)%list (z1 :: z2 :: nil)%list 0).
  auto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.

  forwards Ht21: (close_vars_subst_vars t (y2 :: y1 :: nil)%list (z2 :: z1 :: nil)%list 0).
  eauto.
  fresh_chain_solve Fr.
  fresh_chain_solve Fr.
  rewrite subst_commut_ne in Ht12 by (fresh_chain_solve Fr).

  (* Replace closed terms *)
  rewrite Ht12, Ht21.
  rewrite <- ! trm_close_let_close_fresh_vars by (simpl; fresh_chain_solve Fr).

  eapply trf_swap_calls_local_at with (T1:=T1) (T2:=T2) (TZ1:=TZ1) (TZ2:=TZ2).
  - eapply ctx_ok_hide_l_concat_inv; eauto.
  - fresh_chain_solve Fr.
  - eapply closed_subst_var. eapply closed_subst_var. eauto.
  - fresh_chain_solve Fr.

  - eapply unhiding with (Go:=objctx_empty); eauto.
  - eapply unhiding with (Go:=objctx_empty); eauto.
  - eapply unhiding with (Go:=objctx_empty); eauto.
  - eapply unhiding with (Go:=objctx_empty); eauto.
  - eapply unhiding with (Go:=objctx_empty); eauto.
  - eapply unhiding with (Go:=objctx_empty); eauto.
  - eapply unhiding with (Go:=objctx_empty); eauto.
  - eapply unhiding with (Go:=objctx_empty); eauto.
  -
    rewrite <- ctx_concat_push.
    rewrite <- ctx_concat_push.
    eapply unhiding with (Go:=objctx_empty); eauto. ctx_flatten.
    eapply rename_subst_trm_2.
    eauto.
    eauto.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    eauto.

Unshelve.
  eapply unhiding_ty_var.
  eapply unhiding_ty_var.
  eapply unhiding_subtyp.
  eapply unhiding_subtyp.
  eapply unhiding_ty_trm.
  eapply unhiding_ty_trm.
  eapply unhiding_ty_var.
  eapply unhiding_ty_var.
  eapply unhiding_ty_trm.
Qed.


End WithTypingMode.
