Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation TransformationLifting Weakening EWeakening Substitution.
Require Import TypingHelpers TransformationConfig ConfigTyping.
Require Import Similarity ConfigSimilarity ContextSimilarity SyntaxSimilarity SimilarityPreservation TypingSimilarity
 TypedReductionEquivalence OperationalSemantics TypedOperationalSemantics TypedPreservation ConfigValid TransformationSwapReads TransformationSimilarity TransformationPreservation TypedProgress ExecutionTactics.
Require Import Defs.
Require Import HeapCorrespondence TransformationTerm TFree.
Require Import TypedSimilarityPreservation.

Require Import SefGuarantee TransformationSwapCalls FreshChain FreshChainTactics.
Require Import TransformationSimilarityPreservation.
Require Import MethodCall TypedMethodCall ReductionLemmas TypedReductionLemmas.
Require Import TypedSoundness.
Require Import Safety.
Require Import Lia.
Require Import ReductionFreeVars.
Require Import ConfigSimilarity.

Require Import TransformationLiftingPreservation TransformationIterateSim TransformationSwapCallsPreservation.
Require Import TypedSoundness.

(** * Transformation guarantee for call-swapping *)

Section WithTypingMode.
Context (typing_mode: Mode).
(* Only works in mode with mutability and SEF guarantee *)
Context (Hms : mode_has_sef typing_mode).
Context (Hmm : mode_has_mutability typing_mode).

(* Lemma 58 in "Pure methods for roDOT" *)
Theorem swap_calls_transformation_preservation Ca1 Ca2 Cb1 Ga1 Ga2 Gb1 Ta Tb na:
  (* The original program terminates in na steps *)
  Ga1 ⊢ Ca1 ∶ Ta ⊣na↦* Ga2 ⊢ Ca2 ->
  answer Ca2 ->
  (* Transformation of the program *)
  〈Ga1 ⊢ Ca1 ∶ Ta〉 ∼(trf_lift_sim_conf typing_mode (trf_swap_calls_local typing_mode))∼> 〈Gb1 ⊢ Cb1 ∶ Tb〉 ->
  (* Exists transformed execution *)
  exists nb Gb2 Cb2,
    (* Which is not longer *)
    nb <= na /\
    (* Reaches an answer *)
    Gb1 ⊢ Cb1 ∶ Tb ⊣nb↦* Gb2 ⊢ Cb2 /\
    answer Cb2 /\
    (* Which is transformed *)
    〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼(trf_lift_sim_conf typing_mode (trf_swap_calls_local typing_mode))∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉.
Proof.
  intros.
  eapply transformation_preserved_if_terminates; eauto.
  eapply TransformSwapCallsTyped.
  rewrite trf_symmetric_inverse. eapply TransformSwapCallsTyped. eapply TransformSwapCallsSymmetric.
  rewrite trf_symmetric_inverse. eapply TransformSwapCallsPreservesAnswer. eapply TransformSwapCallsSymmetric.
  eapply TransformSwapCallsIdtyp.
  eapply TransformSwapCallsPreservesAnswer.
  eapply TransformSwapCallsSubstCompat.
  eapply TransformSwapCallsEWeakening.
  eapply TransformSwapCallsWeakening.
  eapply TransformSwapCallsSealing.
  eapply TransformSwapCallsNarrowing.
  eapply TransformSwapCallsUnhiding.
  eapply transformation_swap_calls_reduces_to_similarity.

  all: auto.
Qed. 

Definition term_terminates_in t n :=
  exists C,   config_initial t C /\   config_terminates_in C n.

(* Theorem 27 in "Pure methods for roDOT" *)
Theorem swap_calls_transformation_guarantee ta tb Ta Tb:
  〈ctx_empty⁏empty ⊢ ta ∶ Ta〉 ∼(trf_local_idtyp_trm typing_mode (trf_swap_calls_local typing_mode))∼> 〈ctx_empty⁏empty ⊢ tb ∶ Tb〉 ->
  closed ta ->
  closed tb ->
  closed Ta ->
  closed Tb ->
  (* Exists transformed execution *)
  exists Ca Cb,
    config_initial ta Ca /\
    config_initial tb Cb /\
    (
      ( exists  na nb Ga2 Gb2 Ca2 Cb2,
        (* Which is not longer *)
        nb <= na /\
        (* Reaches an answer *)
        ctx_empty ⊢ Ca ∶ Ta ⊣na↦* Ga2 ⊢ Ca2 /\
        ctx_empty ⊢ Cb ∶ Tb ⊣nb↦* Gb2 ⊢ Cb2 /\
        answer Ca2 /\
        answer Cb2 /\
        (* Which is transformed *)
        〈Ga2 ⊢ Ca2 ∶ Ta〉 ∼(trf_lift_sim_conf typing_mode (trf_swap_calls_local typing_mode))∼> 〈Gb2 ⊢ Cb2 ∶ Tb〉
      ) \/ (
         forall (n : nat),
         exists (Ca2 Cb2 : config),
           Ca ⊣n↦* Ca2 /\
           Cb ⊣n↦* Cb2
      )
    ).
Proof.
  introv Htrf Hcla Hclb HclTa HclTb.

  forwards Htyb: ty_config_initial Hclb HclTb.
  { 
    eapply trf_inverse_intro in Htrf.
    rewrite trf_local_idtyp_inverse_trm in Htrf.
    rewrite trf_symmetric_inverse in Htrf.
    forwards: TransformLocalIdtypTrmTyped.
    eapply TransformSwapCallsTyped.
    forwards: H. eauto. hnf in H0. eapply H0.
    eapply TransformSwapCallsSymmetric.
    eapply TransformSwapCallsIdtyp.
  }
  destructecs Htyb. destruct C. inversions Htyb0.

  forwards Htya: ty_config_initial Hcla HclTa.
  {
    forwards: TransformLocalIdtypTrmTyped.
    eapply TransformSwapCallsTyped.
    forwards: H. eauto.
    hnf in H0.
    eapply H0.
  }
  destructecs Htya. destruct C. inversions Htya0.

  (* Unify types *)
  forwards: TrfLocalIdtypTrmEnsuresCond Htrf.
  inversions H. inversions H0.

  assert (〈 ctx_empty ⊢ (t0 ; stack_nil ; empty ; empty ; Lv0) ∶ Tb  〉 ∼ trf_lift_sim_conf typing_mode (trf_swap_calls_local typing_mode) ∼> 〈 ctx_empty  ⊢  (t ; stack_nil ; empty ; empty ; Lv) ∶ Tb 〉). {
    eapply trf_lift_sim_conf_lift.
    auto.

    eapply trf_lift_idtyp_conf_focus.
    eapply trf_focus_idtyp_conf_focus with (Ts:=Tb).
    5: eauto. 4: eapply ty_stack_nil.
    4: auto.
    4: auto.
    eapply ConfigTypingAccessories.ty_config_to_accessories; eauto.
    eapply ConfigTypingAccessories.ty_config_to_accessories; eauto.
    eauto.
  }

  forwards Hsa:typed_red_soundness Htya.
  forwards Hsb:typed_red_soundness Htyb.

  destructecs Hsa.
  - forwards Hpr: swap_calls_transformation_preservation H. eauto. auto.
    destructecs Hpr.
    exists (t0 ; stack_nil ; empty ; empty ; Lv0) (t ; stack_nil ; empty ; empty ; Lv); splits.
    econstructor.    econstructor.
    left.
    exists n nb.
    exists Fa Gb2 Ca Cb2; splits; auto.
  - destructecs Hsb.
    + (* contradictory case *)
      false.
      eapply trf_symmetric in H.

      forwards Hpr: swap_calls_transformation_preservation H. eauto. auto.
      destructecs Hpr. specialize (Hsa (S nb)). destructecs Hsa.
      forwards: typed_red_closure_answer_split Hpr1 Hsa. eauto. auto.
      destructecs H0.
      assert ( S nb + n32 <= nb). rewrite H1 at 2. auto.
      lia.
    + 
      exists (t0 ; stack_nil ; empty ; empty ; Lv0) (t ; stack_nil ; empty ; empty ; Lv); splits.
      econstructor.    econstructor.

      right. intros. forwards: Hsa n. forwards: Hsb n. destructecs H0. destructecs H1. exists Cn Cn0; splits.
      eapply typed_to_untyped_red_closure; eauto.
      eapply typed_to_untyped_red_closure; eauto.
Qed. 

End WithTypingMode.