(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation TransformationLifting Weakening EWeakening Substitution.
Require Import TypingHelpers TransformationConfig ConfigTyping.
Require Import Similarity ConfigSimilarity ContextSimilarity SyntaxSimilarity SimilarityPreservation TypingSimilarity
 TypedReductionEquivalence OperationalSemantics TypedOperationalSemantics TypedPreservation ConfigValid TransformationSwapReads TransformationSimilarity TransformationPreservation TypedProgress ExecutionTactics.
Require Import Defs.
Require Import HeapCorrespondence TransformationTerm TFree.
Require Import TypedSimilarityPreservation.

Require Import SefGuarantee TransformationSwapCalls FreshChain FreshChainTactics.
Require Import TransformationSimilarityPreservation.
Require Import MethodCall TypedMethodCall ReductionLemmas TypedReductionLemmas.
Require Import TypedSoundness.
Require Import Safety.
Require Import Lia.
Require Import ReductionFreeVars.
Require Import ConfigSimilarity.
Require Import Defree.

(** * Preservation of call-swapping transformation *)

Section WithTypingMode.
Context (typing_mode: Mode).
(* Only works in mode with mutability and SEF guarantee *)
Context (Hms : mode_has_sef typing_mode).
Context (Hmm : mode_has_mutability typing_mode).


Definition trf_swap_calls_focus := trf_focus_idtyp_conf typing_mode (trf_swap_calls_local typing_mode).
Definition trf_swap_calls_lift := trf_lift_idtyp_conf typing_mode (trf_swap_calls_local typing_mode).

Lemma similar_stack_suffix_cons_false f fsb fsa xs1 xs2:
  stack_suffix (stack_cons f fsb) fsa ->
  similar xs2 xs1 fsa fsb ->
  False.
Proof.
  introv Hsu Hsi.
  reverts fsb f; induction fsa; intros.
  -
  eapply similar_stack_nil_case in Hsi. subst.
  hnf in Hsu. 
  destructecs Hsu.
  destruct fs3; simpls; discriminate.
  -
  eapply similar_stack_cons_case in Hsi.    
  destructecs Hsi; subst.

  hnf in Hsu. 
  destructecs Hsu.

  eapply IHfsa with (f:=w2); eauto.

  destruct fs3; simpls; inversions Hsu.
  hnf. exists stack_nil. simpl. eauto.
  exists (stack_append fs3 (stack_cons f0 stack_nil)).
  rewrite stack_append_assoc. simpl. auto.
Qed.

Lemma frame_end_in_within_le C1 C2 C3 n m:
  config_hastyp typing_mode C1 ->
  C1 ⊣ n ↦* C3 ->
  frame_within C1 n ->
  frame_end_in C1 m C2 ->
  n <= m.
Proof.
  introv Hty Hred Hfw Hfe.
  hnf in *.
  destructecs Hfe.
  destruct C1, C2; simpls; subst.
  destruct (Nat.lt_ge_cases m n).
  - false.
    forwards Hrs: red_closure_short Hred H.
    destructecs Hrs.

    forwards Hfws: Hfw (Datatypes.S m).
    lia.
    eauto.
    inversions Hrs0.

    forwards Hsi: TypedSimilarityPreservation.similar_conf_1to1_preservation_closure.
    4: eapply Hfe0.
    4: eapply H1.
    eapply similar_conf_1to1_refl. eapply config_hastyp_valid; eauto. eauto. eauto.
    hnf in Hsi. destructecs Hsi.
    destruct C2.
    hnf in Hsi. destructecs Hsi.
    eapply similar_trm_var_case in Hsi3. destructecs Hsi3. subst.
    inversions H3.
    simpls.
    eapply similar_symm in Hsi4.
    eapply similar_stack_cons_case in Hsi4.
    destructecs Hsi4; subst.
    eapply similar_stack_suffix_cons_false; eauto.
  - auto.
Qed.


Lemma wkn_ewkn_end (A B : Set) (property: TypingProperty A B) {EW: EWeakening property} {W: Weakening property} G1 G2 e1 e2 (a: A) (T: B):
  property G1 objctx_empty e1 a T ->
  ok (e1 & e2) ->
  ctx_ok (G1 §& G2) ->
  ctx_weakens G2 ->
  property (G1 §& G2) objctx_empty (e1 & e2) a T.
Proof.
  introv Hty Hoke HokG HwkG.
  rewrite <- ctx_concat_empty_r with (G1:= (G1 §& G2)).
  rewrite <- concat_empty_r with (E:= (e1 & e2)).
  eapply ewkn.
  eapply weaken.
  rewrite ctx_concat_empty_r, concat_empty_r.
  all: auto.
Qed.

Lemma typed_sef_guarantee G x m y fs S e Lv1 T n G2 w S2 e2 Lv2 Ty:
  ty_config G (trm_call (avar_f x) m (avar_f y) ; fs ; S ; e ; Lv1) T ->
  G ⊢ (trm_call (avar_f x) m (avar_f y) ; fs ; S ; e ; Lv1) ∶ T ⊣ n ↦* G2 ⊢ ((trm_var (avar_f w)) ; fs ; S2 ; e & e2 ; Lv2) ->
  frame_within (trm_call (avar_f x) m (avar_f y) ; fs ; S ; e ; Lv1) n ->
  G ⁏ e ⊢ x ∶ typ_rcd (dec_met m typ_nocap Ty typ_top) ->
  G ⁏ e ⊢ y ∶ Ty ->
  G ⁏ e ⊢ typ_nocap <: Ty ->
  (forall (y : var) (itm : item), binds y itm S -> binds y itm S2).
Proof.
  introv Htyc Htred Hfw Htx Hty Hsu.

  forwards Hsef: SefG_I. 5: eexists; eapply Htyc.
  auto.
  auto.
  sympl; eauto.  
  sympl; splits; eauto.
  eapply typed_to_untyped_red_closure; eauto.

  destructecs Hsef.
  - (* Not possible *)
    forwards: frame_end_in_within_le Hfw Hsef0.
    eapply config_hastyp_c; eauto.
    eapply typed_to_untyped_red_closure; eauto.
    lia.
  - simpl in Hsef. auto.
Qed.


Lemma subheap_unchanged_empty_l S:
  subheap_unchanged (empty: heap) S.
Proof.
  introv Hbi.
  eapply binds_empty_inv in Hbi.
  false.
Qed.
Lemma subheap_unchanged_empty_r_inv S:
  subheap_unchanged S (empty: heap) ->
  S = (empty: heap).
Proof.
  introv Hbi.
  induction S using env_ind.
  auto.
  forwards: Hbi x v. eauto.
  eapply binds_empty_inv in H.
  false.
Qed.



Lemma subheap_unchanged_split (S1 S2: heap):
  subheap_unchanged S2 S1 ->
  ok S1 ->
  ok S2 ->
  exists S3,  
    ok (S2 & S3) /\
    dom(S2 & S3) = dom S1 /\
    subheap_unchanged (S2 & S3) S1.
Proof.
  introv Hsu Hok1 Hok2.
  reverts S2.
  induction S1 using env_ind; intros.

  - exists (empty: heap).
    eapply subheap_unchanged_empty_r_inv in Hsu.
    subst. rewrite !concat_empty_r.
    split; auto.
  - destruct (fv_decide x (dom S2)).
    + (* x in S2 *)
      eapply dom_to_binds in m.
      destructecs m.
      eapply binds_middle_exists in m.
      destructecs m.
      subst.
      assert (ok (S0 & S3)) as Hok03 by eauto.
      assert (v=v0). {
        forwards: Hsu x v0. eapply binds_middle_eq.
        eapply ok_permute_push in Hok2. eapply ok_push_inv in Hok2. destruct Hok2. eauto. eapply binds_push_eq_inv in H. auto.
      } subst.

      forwards HS1: IHS1 (S0 & S3) Hok03.

      eauto.
      intros.

      assert (x<>y) as Hnxy. {
        intro. subst.
        eapply ok_permute_push in Hok2.
        eapply ok_push_inv in Hok2. destruct Hok2. eapply binds_to_dom in H. contradiction.
      }

      eapply binds_push_neq_inv.
      eapply Hsu.

      eapply binds_weaken2; eauto. auto.
      destructecs HS1.

      assert (ok (S0 & x ~ v0 & (S3 & S2)) ) as Hok032. {
        eapply ok_swap_after.
        rewrite concat_assoc.

        constructor. auto. rewrite HS2. eapply ok_push_inv in Hok1. destruct Hok1. auto.
      }

      exists S2; splits.
      rewrite <- concat_assoc. auto.

      simpl_dom. rewrite <- HS2. eapply fset_extens; union_sub_solver.
      intros.
      rewrite <- concat_assoc in H.
      eapply binds_permute_push in H.
      eapply binds_push_inv in H.
      destructecs H; subst.
      eapply binds_push_eq.
      rewrite concat_assoc in H.
      eapply HS1 in H.
      eapply binds_push_neq; auto. auto.
    +
      forwards: IHS1 S2. eauto.
      intros.

      assert (x<>y) as Hnxy. {
        intro. subst.
        eapply binds_to_dom in H. contradiction.
      }
      forwards: Hsu H. eapply binds_push_neq_inv; eauto. auto.
      destructecs H.
      exists (S3 & x ~v); splits.
      rewrite concat_assoc .
      constructor. auto. rewrite H1. eapply ok_push_inv in Hok1. destruct Hok1. auto.
      simpl_dom. rewrite <- H1. eapply fset_extens; union_sub_solver.
      intros.
      rewrite concat_assoc in H2. eapply binds_push_inv in H2. destructecs H2; subst.
      eapply binds_push_eq. eapply binds_push_neq; eauto.
Qed.

Lemma subheap_unchanged_symm (S2 S3: heap):
  dom S2 = dom S3 ->
  subheap_unchanged S2 S3 ->
  subheap_unchanged S3 S2.
Proof.
  introv Hdom Hsu.
  intros.
  forwards: binds_to_dom H.
  rewrite <- Hdom in H0.
  eapply dom_to_binds in H0.
  destructecs H0.
  forwards: Hsu H0.
  binds_eq.
  auto.
Qed.

Lemma subheap_unchanged_concat_r (S2 S3 S1: heap):
  dom S2 = dom S3 ->
  subheap_unchanged S2 S3 ->
  subheap_unchanged (S1 & S2) (S1 & S3).
Proof.
  introv Hdom Hsu.
  intros.
  eapply binds_concat_inv in H.
  destructecs H.
  eapply binds_concat_right. eauto.
  eapply binds_concat_left. eauto.

  intro Hbi3. eapply dom_to_binds in Hbi3.
  destructecs Hbi3.
  forwards: subheap_unchanged_symm Hsu Hbi3. auto.
  eapply binds_to_dom in H1.
  contradiction.
Qed.

Lemma subheap_unchanged_concat_l (S2 S3 S1: heap):
  subheap_unchanged S2 S3 ->
  subheap_unchanged (S2 & S1) (S3 & S1).
Proof.
  introv Hsu.
  intros.
  eapply binds_concat_inv in H.
  destructecs H.
  eapply binds_concat_right. eauto.
  eapply binds_concat_left. eauto. auto.
Qed.


Lemma similar_with_heap_elements_rf (S1 S2 S3: heap) xs ys  ws zs:
  similar_with xs ys ws zs S1 S2 ->
  ok S3 ->
  dom S2 = dom S3 ->
  subheap_unchanged S2 S3 ->
  similar_with xs ys ws zs S1 S3.
Proof.
  introv Hsi Hok3 Hdom Hsu.
  hnf. hnf in Hsi.

  induction Hsi.
  - eapply similar_env_empty. auto.
  - eapply similar_env_dom. hnf. eapply IHHsi; eauto.
  eauto. eauto. eauto.
  - eapply similar_env_notin. hnf. eapply IHHsi; eauto.
  eauto. eauto.  
Qed.

Lemma similar_with_heap_elements_lf (S1 S2 S3: heap) xs ys  ws zs:
  similar_with xs ys ws zs S2 S1 ->
  ok S3 ->
  dom S2 = dom S3 ->
  subheap_unchanged S2 S3 ->
  similar_with xs ys ws zs S3 S1.
Proof.
  introv Hsi Hok3 Hdom Hsu.
  eapply similar_env_with_comm.
  eapply similar_env_with_comm in Hsi.
  eapply similar_with_heap_elements_rf; eauto.
Qed.

Lemma similar_with_heap_elements_rb (S1 S2 S3: heap) xs ys  ws zs:
  similar_with xs ys ws zs S1 S2 ->
  ok S3 ->
  dom S2 = dom S3 ->
  subheap_unchanged S3 S2 ->
  similar_with xs ys ws zs S1 S3.
Proof.
  introv Hsi Hok3 Hdom Hsu.
  eapply similar_with_heap_elements_rf; eauto.
  eapply subheap_unchanged_symm; eauto.
Qed.
Lemma similar_with_heap_elements_lb (S1 S2 S3: heap) xs ys  ws zs:
  similar_with xs ys ws zs S2 S1 ->
  ok S3 ->
  dom S2 = dom S3 ->
  subheap_unchanged S3 S2 ->
  similar_with xs ys ws zs S3 S1.
Proof.
  introv Hsi Hok3 Hdom Hsu.
  eapply similar_with_heap_elements_lf; eauto.
  eapply subheap_unchanged_symm; eauto.
Qed.
Lemma similar_heap_elements_rf (S1 S2 S3: heap) xs ys:
  similar xs ys S1 S2 ->
  ok S3 ->
  dom S2 = dom S3 ->
  subheap_unchanged S2 S3 ->
  similar xs ys S1 S3.
Proof.
  introv Hsi Hok3 Hdom Hsu.
  eapply similar_with_heap_elements_rf; eauto.
Qed.


Lemma similar_conf_1to1_vars_heap_bi
  ta fsa Sa ea Lva tb fsb Sb eb Lvb xs ys Sc:
  similar xs ys (ta ; fsa ; Sa ; ea ; Lva) (tb ; fsb ; Sb ; eb ; Lvb) ->
  ok Sc ->
  dom (Sc) = dom Sb -> 
  (forall (y : var) (itm : item), binds y itm Sb -> binds y itm Sc) ->
  similar xs ys (ta ; fsa ; Sa ; ea ; Lva) (tb ; fsb ; Sc ; eb ; Lvb).
Proof.
  introv Hsim Hok Hdom Hbi.
  hnde Hsim.
  hnf; splits; auto 1.
  eapply similar_heap_elements_rf; eauto.
Qed.



Lemma config_hastyp_Lv_subset t fs S e Lva Lvb:
  config_hastyp typing_mode (t ; fs ; S ; e ; Lva) ->
  Lva \c Lvb ->
  config_hastyp typing_mode (t ; fs ; S ; e ; Lvb).
Proof.
  introv Hhas Hsu.
  inversions Hhas.
  econstructor.
  eapply ty_config_Lv_subset; eauto.
Qed.


Lemma config_hastyp_stack_nil_Lv_subset t fs S e Lva Lvb:
  config_hastyp typing_mode (t ; fs ; S ; e ; Lva) ->
  fv fs \u Lva \c Lvb ->
  config_hastyp typing_mode (t ; stack_nil ; S ; e ; Lvb).
Proof.
  introv Hhas Hsu.
  inversions Hhas.

  inversions H.
  rename Gamma into G.

  destructecs H14.
  set (dfc := defree_ctx_with_null G e (Lva \u fv t) null).
  assert (defree_ctx_ok dfc) as Hok. constructor; auto. eauto. eauto. eauto.
  assert (defree_ctx_get dfc G e) as Hget. constructor; auto.
  assert (defree_ctx_closed dfc) as HclG. constructor; auto.
  assert (defree_ctx_null dfc null) as Hnull. constructor; auto.


  destructc (defree_typ_closed T Hok HclG Hnull) as dfs Hva Hfv HclTfn.

  assert (G ⁏ e ⊢ t ∶ (defree_by_rec 0 dfs T)). {
  eapply defree_ty_trm; eauto.
  econstructor. auto. 
  inversions H16. auto.
  unfold defree_ctx_fv. unfold dfc. union_sub_solver.
  }

  eapply config_hastyp_c with (Gamma := G) (U:=(defree_by_rec 0 dfs T)).
  eapply ty_config_c; eauto.
  - econstructor. eauto.
  - 
    forwards: subset_union_l2 Hsu. forwards: subset_union_l1 Hsu. 
    destruct H12. destruct H2.

  assert (
    ctx_dom G \c fv t \u fv stack_nil \u fv_heap S \u fv_renv e \u Lvb) as HsubGd. {
    eapply subset_trans. eauto. union_sub_solver.
  }
  assert (
    ctx_fv G \c fv t \u fv stack_nil \u fv_heap S \u fv_renv e \u Lvb) as HsubGf. {
    eapply subset_trans. eapply H4. union_sub_solver.
  }
  assert (
    \{null} \c fv t \u fv stack_nil \u fv_heap S \u fv_renv e \u Lvb) as Hsubn. {
    eapply subset_trans. eapply subset_singleton. eauto. union_sub_solver.
  }
  assert (
    defree_ctx_fv dfc \c (fv t \u fv stack_nil \u fv_heap S \u fv_renv e \u Lvb)) as Hdc. {
    unfold defree_ctx_fv, dfc.
    union_sub_solver.
  }
    split; try split; auto 1. 
    intro.
    intro.
    forwards: defree_fv2 H11. eauto.
    rewrite in_union in H12. destruct H12.
    forwards: Hfv H12. destruct H13. contradiction.
    eapply Hdc in H13. auto.
    eapply Hdc in H12. auto.
  - exists null; eauto. eapply Hsu. rewrite in_union; auto.
  - inversions H16. econstructor; eauto.  intro. intro. simpl. auto.
Qed.


Lemma vars_conf_unique_fs_nil_Lv_subset t fs S e Lva Lvb xs:
  vars_conf_unique xs (t ; fs ; S ; e ; Lva) ->
  fv fs \u Lva \c Lvb ->
  vars_conf_unique xs (t ; stack_nil ; S ; e ; Lvb).
Proof.
  introv Hsf Hsu.
  hnf in Hsf. hnf.
  destructecs Hsf.
  splits.
  auto.
  eapply subset_trans. eauto.
  forwards: subset_union_l2 Hsu.
  forwards: subset_union_l1 Hsu.
  union_sub_solver.
Qed.


Lemma ty_config_disjoint_Se t fs S e Lv G T:
  ty_config G (t ; fs ; S ; e ; Lv) T ->
  disjoint (dom e) (dom S).
Proof.
  introv Hty.
  eapply config_typed_valid in Hty.
  inversions Hty.
  auto.
Qed.

Lemma close_subst_fresh n x y (a: trm):
  y \notin fv a ->
  a /[x -> y] /[#n <- y] = a /[#n <- x].
Proof.
  introv Hn.
  rewrite close_subst.
  rewrite close_unbound with (x0:=y). auto.
  eapply notin_fv_close_notin; eauto.
Qed.

Lemma trm_closed_close_open_swap (t: trm) y1 y2 w1 w2:
  closed t ->
  y1 <> y2 ->
  t /[<- y2] /[# 1 <- y1] /[# 1 -> w1] /[-> w2] =   t /[<- y1] /[# 1 <- y2] /[# 1 -> w2] /[-> w1].
Proof.
  introv Hclt Hny.
  pick_fresh z1.
  pick_fresh z2.

  replace (t /[<- y2] /[# 1 <- y1]) with (t /[y2 -> z2] /[y1 -> z1] /[<- z2] /[# 1 <- z1]).
  replace (t /[<- y1] /[# 1 <- y2]) with (t /[y2 -> z2] /[y1 -> z1] /[<- z1] /[# 1 <- z2]).
  2: {
    rewrite close_subst_fresh.
    rewrite close_commut.
    rewrite close_subst_fresh.
    rewrite close_commut.
    auto. eauto. eauto. eauto.
    eapply notin_fv_subst_var; eauto.
  }
  2: {
    rewrite subst_commut_ne.
    rewrite close_subst_fresh.
    rewrite close_commut.
    rewrite close_subst_fresh.
    rewrite close_commut.
    auto. eauto. eauto. eauto.
    eapply notin_fv_subst_var; eauto.
    all: eauto.
  }
  assert (closed (t /[ y2 -> z2] /[ y1 -> z1] )).
    eapply closed_subst_var.
    eapply closed_subst_var. auto.

  rewrite <- subst_open_close.
  rewrite <- subst_open_close.
  rewrite <- subst_open_commut_ne by eauto.
  rewrite <- subst_open_commut_ne by eauto.

  rewrite open_close_commut by eauto.
  rewrite H.
  rewrite open_close_commut with (x:=z1) by eauto.
  rewrite H.
  
  rewrite <- subst_open_close.
  rewrite <- subst_open_close.
  rewrite H.
  rewrite H.
  
  rewrite subst_commut_ne by eauto. auto.
Qed.

(* Lemma 60 in "Pure methods for roDOT" *)
Lemma transformation_swap_calls_reduces_to_similarity:
  TransformConfigReducesToShortIfTerminates typing_mode (trf_swap_calls_focus) (trf_similarity typing_mode).
Proof.
  hnf.
  introv Htrf Htrm.
  unfold trf_swap_calls_focus.

  forwards Htyi: @trf_ensures Htrf. eapply TrfFocusIdtypEnsuresCond.
  inversions Htyi.

  assert (ty_config Gb1 Ca1 Tb) as HtyCa1. {
    forwards: @trf_requires Htrf. unfold trf_swap_calls_focus.
    eapply TransformFocusIdtypConfigTyped.
    eapply TransformSwapCallsTyped.
    eauto.
  }
  assert (ty_config Gb1 Cb1 Tb) as HtyCb1. {
    eapply trf_symmetric in Htrf. 
    forwards: @trf_requires Htrf. unfold trf_swap_calls_focus.
    eapply TransformFocusIdtypConfigTyped.
    eapply TransformSwapCallsTyped.
    eauto.
  }
  assert (config_valid Ca1) as Hva1. {
    eapply config_typed_valid; eauto.
  }

  (* Invert the transformation *)
  inversions Htrf.
  rename H3 into Htyca, H4 into Htycb, H5 into Hhc, H6 into Htys, H7 into Htrfl, H8 into Hsu.

  hnf in Hhc; destructecs Hhc.
  inversion Htrfl.
  (* Store the whole terms *)
  eq_to_def ta.
  eq_to_def tb.
  subst.

  rename H7 into Hfr.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hfr.
  fresh_chain_cleanup Fr.

  rename H5 into Hoke, H6 into Hclt, H8 into Htyx1, H9 into Htyx2, H10 into HsuTZ1, H11 into HsuTZ2, H14 into Htyxp1, H15 into Htyxp2, H12 into Htym1, H13 into Htym2.
  
  fresh_chain_include_typing Fr Htyx1.
  fresh_chain_include_typing Fr Htyx2.
  fresh_chain_include_typing Fr Htyxp1.
  fresh_chain_include_typing Fr Htyxp2.

  set (ta' := trm_let (trm_call (avar_f x1) m1 (avar_f xp1)) (trm_let (trm_call (avar_f x2) m2 (avar_f xp2)) t /[<- y2]/[#1 <- y1])).

  assert ( ta = ta' ) as Heqta. {
    assert (x2 <> y1). fresh_chain_solve Fr.
    assert (xp2 <> y1). fresh_chain_solve Fr.
    unfold ta, ta'.
    autorewrite with simpl_close.
    cases_if.
    cases_if.
    sympl. auto.
  }
  rewrite Heqta in *. clear Heqta ta.

  set (tb' := trm_let (trm_call (avar_f x2) m2 (avar_f xp2)) (trm_let (trm_call (avar_f x1) m1 (avar_f xp1)) t /[<- y1]/[#1 <- y2])).

  assert ( tb = tb' ) as Heqtb. {
    assert (x1 <> y2). fresh_chain_solve Fr.
    assert (xp1 <> y2). fresh_chain_solve Fr.
    unfold tb, tb'.
    autorewrite with simpl_close.
    cases_if.
    cases_if.
    sympl. auto.
  }
  rewrite Heqtb in *. clear Heqtb tb.

  (* Execute left side *)
  (* Make first step into the call. *)
  machine_step HtyCa1 Ca2 step_invert_simple.

  (* Execute the call *)
  forwards HredCa3: tred_call_end_exists HtyCa2.
  { (* Termination *)
    applys typed_red_config_terminates1 HredCa2; eauto.
  }

  destructecs HredCa3. rename HredCa3 into HfwCa3, HredCa0 into HredCa3.
  forwards HtyCa3: typed_preservation_closure HtyCa2 HredCa3.

  machine_step HtyCa3 Ca4 step_invert_simple.
  machine_step HtyCa4 Ca5 step_invert_simple.

  (* Execute the call *)
  forwards HredCa6: tred_call_end_exists HtyCa5.
 { (* Termination *)
    applys typed_red_config_terminates1 HredCa5; eauto.
    applys typed_red_config_terminates1 HredCa4; eauto.
    applys typed_red_config_terminates HredCa3; eauto.
    applys typed_red_config_terminates1 HredCa2; eauto.
  }
  destructecs HredCa6. rename HredCa6 into HfwCa6, HredCa0 into HredCa6.
  forwards HtyCa6: typed_preservation_closure HtyCa5 HredCa6.
  machine_step HtyCa6 Ca7 step_invert_simple.

  rename S2 into Sa1, S0 into Sa2, e2 into ea1, e0 into ea2.
  rename Lv3 into Lva6, Lv1 into Lva5, Lv2 into Lva3, Lv0 into Lva4, Lv4 into Lva7, Lv into Lva2.

  (* Use SEF gurarantee to relate the heaps *)
  assert (subheap_unchanged S Sa1) as HsuSa3. {
    intros.
    forwards: typed_sef_guarantee HtyCa2 HredCa3 HfwCa3; eauto.
  }

  rename GCa2 into G.

  forwards HGCa5: typed_red_closure_mono_G G GCa5 HredCa3.
  destructecs HGCa5. rename G3 into GGCa5.

  assert (ok (e & ea1)) as Hokea1. {
    eapply ty_config_env_ok; eauto.
  }
  assert (ctx_ok (G §& GGCa5)) as HokGGCa5. {
    inversions HtyCa5. eauto.
  }

  assert (subheap_unchanged Sa1 Sa2) as HsuSa6. {
    introv Hbi.
    forwards: typed_sef_guarantee TZ2 HtyCa5 HredCa6 HfwCa6 Hbi.
    rewrite HGCa0.
    eapply wkn_ewkn_end.  eapply ewkn_ty_var. eapply weaken_ty_var. eapply Htyx2. auto. auto. auto.
    rewrite HGCa0.
    eapply wkn_ewkn_end.  eapply ewkn_ty_var. eapply weaken_ty_var. eapply Htyxp2. auto. auto. auto.
    rewrite HGCa0.
    eapply wkn_ewkn_end. eapply ewkn_sub. eapply weaken_subtyp. eapply HsuTZ2. auto. auto. auto.
    auto.
  }
  
  (* Stackless execution *)

  forwards HtsCa3: MethodCall.frame_within_truncate_stack stack_nil HfwCa3.
  2: eapply typed_to_untyped_red_closure.  2: eapply HredCa3. lia.
  destructecs HtsCa3.
  
  forwards HtsCa6: MethodCall.frame_within_truncate_stack stack_nil HfwCa6.
  2: eapply typed_to_untyped_red_closure.  2: eapply HredCa6. lia.
  destructecs HtsCa6.

  rename Lv0 into Lvsa3, Lv1 into Lvsa6, HtsCa0 into Hredsa3, HtsCa3 into HLvsa3, HtsCa1 into Hredsa6, HtsCa6 into HLvsa6.

  (* Split heaps *)
  assert (ok S) as HokS by (eapply ty_config_heap_ok; eauto).
  assert (ok Sa1) as HokSa1 by (eapply ty_config_heap_ok; eauto).
  assert (ok Sa2) as HokSa2 by (eapply ty_config_heap_ok; eauto).

  
  forwards HsuSa3e: subheap_unchanged_split HsuSa3; fweauto.
  destructecs HsuSa3e.

  forwards HsuSa6e: subheap_unchanged_split HsuSa6; fweauto.
  destructecs HsuSa6e.

  rename S3 into Sam1, S0 into Sam2.

  (* Assertions about variables *)

  assert (x1 \in Lva2) as Hx1Lv. {
      unfold Lva2, ta'. sympl. union_in_solver.
  }
  assert (xp1 \in Lva2) as Hxp1Lv. {
      unfold Lva2, ta'. sympl. union_in_solver.
  }

  assert (ok (e & ea1 & ea2)) as Hokea10. {
    eapply ty_config_env_ok; eauto.
  }

  set (Lvz := (\{x1} \u \{x2} \u \{xp1} \u \{xp2} \u (fv t /[<- y2] /[# 1 <- y1]) \u fv fs0 \u fv_heap S \u fv_renv e)). 
  assert (Lvz \c Lva2) as HLva2. {
     unfold Lva2, Lvz, ta'. sympl.
     union_sub_solver.
  }

  assert (fv (trm_call (avar_f x1) m1 (avar_f xp1)) \u fv stack_nil \u fv_heap S \u fv_renv e \u Lvz = Lvz) as HLvzm1. {
    eapply fset_extens; unfold Lvz; sympl;      union_sub_solver.
  }
  assert (fv (trm_call (avar_f x2) m2 (avar_f xp2)) \u fv stack_nil \u fv_heap S \u fv_renv e \u Lvz = Lvz) as HLvzm2. {
    eapply fset_extens; unfold Lvz; sympl;      union_sub_solver.
  }

  assert (Lvz \c Lva5) as HLva5. {
    assert( Lva4 \c Lva5). unfold Lva5.     union_sub_solver.
    assert( Lva3 \c Lva4). unfold Lva4.     union_sub_solver.

    assert( Lva2 \c Lva3). eapply red_closure_mono_Lv. eapply typed_to_untyped_red_closure. eauto.
    eapply subset_trans. eapply HLva2.
    eapply subset_trans. eauto.
    eapply subset_trans. eauto.
    eauto.
  }

  assert (Lvz \c Lva7) as HLva7. {
    assert( Lva6 \c Lva7). unfold Lva7.     union_sub_solver.
    assert( Lva5 \c Lva6). eapply red_closure_mono_Lv. eapply typed_to_untyped_red_closure. eauto.
    eapply subset_trans. eapply HLva5.
    eapply subset_trans. eauto.
    eauto.
  }

  (* Reflective similarity of common start *)

  forwards HsiCa3: similar_conf_1to1_refl_vars (trm_call (avar_f x1) m1 (avar_f xp1) ; stack_nil ; S ; e ; Lvz).
  eapply config_valid_change. eapply config_typed_valid. eauto.
  destructecs HsiCa3.

  rewrite HLvzm1 in HsiCa3.

  assert (fresh \{} (length xs) xs) as Hfrxs. {
    hnde HsiCa0. auto.
  }

  forwards: similar_conf_1to1_refl_by ( trm_call (avar_f x2) m2 (avar_f xp2) ; stack_nil ; S ; e ; Lvz) xs.
  eapply config_valid_change. eapply config_typed_valid. eauto.
  auto.
  rewrite HsiCa3. sympl. unfold Lvz.
  union_sub_solver_destruct_ls; try union_sub_solver.

  (* Execute right side *)
  lets HtyCc1: HtyCb1.

  machine_step HtyCc1 Cc2 step_invert_simple.

  rename Lv into Lvb2.

  (* Before executing the first call on the right, relate to second call on the left to show termination *)
  assert (similar xs xs (trm_call (avar_f x2) m2 (avar_f xp2) ; stack_nil ; S ; e ; Lvz) (trm_call (avar_f x2) m2 (avar_f xp2) ; stack_nil ; Sa1; e & ea1 ; Lvz)) as HsiCa. {
    eapply similar_conf_1to1_vars_heap_bi. 4: eapply HsuSa3e. 3: auto. 2: auto.
    eapply similar_conf_1to1_vars_heap_renv_extend_r.
    eauto.
    auto.
    eapply disjoint_fresh. eauto. rewrite HsiCa3. sympl.
    forwards Hdi: red_closure_confing_fv_extension_disjoint Hredsa3. sympl in Hdi.
    unfold config_fv_extension in Hdi.
    eapply disjoint_comm.
    eapply disjoint_subset_l.
    eapply disjoint_subset_r.
    eapply Hdi.


    unfold Lvz. replace (fv t /[<- y2] /[# 1 <- y1] ) with (fv t /[<- y1] /[# 1 <- y2]).  union_sub_solver. rewrite ! fv_close. rewrite remove_remove. auto.
    sympl.
    eapply subset_union_2. 
    rewrite <- HsuSa3e1. rewrite dom_concat. rewrite union_remove, remove_same. rewrite remove_disjoint. union_sub_solver. eapply disjoint_comm. eapply ok_concat_disjoint. auto.
    rewrite dom_concat. rewrite union_remove, remove_same. rewrite remove_disjoint. union_sub_solver. eapply disjoint_comm. eapply ok_concat_disjoint. auto.
  }

  assert (Lvz \c Lvb2) as HLvb2. {
    assert (fv t /[<- y2] /[# 1 <- y1] = fv t /[<- y1] /[# 1 <- y2]) as Hfvt. {
      rewrite ! fv_close.
      rewrite remove_remove. auto.
    }
     unfold Lvb2, Lvz, tb'. sympl. rewrite Hfvt.
     union_sub_solver.
  }

  (* Execute the call *)
  forwards HredCc3: tred_call_end_exists_or_within (Datatypes.S n0) HtyCc2. 
  destructecs HredCc3.
  
  2: { (* Non-terminating case contradicts termination of second call on the left *)
    false.
    rename HredCc3 into HfwCc3, HredCc0 into HredCc3.
    forwards HtyCc3: typed_preservation_closure HtyCc2 HredCc3.
    destruct C2.

    (* Truncate stack in the overshooting execution *)
    forwards HtsCc3: frame_within_truncate_stack' HfwCc3.
    2: eapply typed_to_untyped_red_closure.  2: eapply HredCc3. lia.
    destructecs HtsCc3. subst s.
    
    (* Similarity of overshooting execution on the right with terminating exection on the left *)
    forwards Hsipra:  similar_preservation_closure_answer_overshoot xs xs Hredsa6 HtsCc1. 
    eapply similar_conf_Lv. eapply similar_symm. eapply HsiCa.
    eapply config_hastyp_stack_nil_Lv_subset. eapply config_hastyp_c. eauto. union_sub_solver.
    eapply config_hastyp_stack_nil_Lv_subset. eapply config_hastyp_c. eauto. union_sub_solver.
    hnf. split. auto.  rewrite HsiCa3. union_sub_solver.
    hnf. split. auto.  rewrite HsiCa3. union_sub_solver.
    constructor.
    eapply Nat.nle_succ_diag_l; eauto.
  }

  rename HredCc3 into HfwCc3, HredCc0 into HredCc3.
  forwards HtyCc3: typed_preservation_closure HtyCc2 HredCc3.

  machine_step HtyCc3 Cc4 step_invert_simple.
  machine_step HtyCc4 Cc5 step_invert_simple.

  rename S2 into Sb1, e2 into eb1.
  rename Lv0 into Lvb5, Lv2 into Lvb3, Lv into Lvb4.

  (* Split heaps *)

  assert (subheap_unchanged S Sb1) as HsuSc3. {
    intros.
    forwards: typed_sef_guarantee HtyCc2 HredCc3 HfwCc3 H0; eauto.
  }

  assert (ok (e & eb1)) as Hokeb1. {
    eapply ty_config_env_ok; eauto.
  }

  assert (ok Sb1) as HokSb1 by (eapply ty_config_heap_ok; eauto).

  forwards HsuSc3e: subheap_unchanged_split HsuSc3; fweauto.
  destructecs HsuSc3e.

  (* Assertions about variables *)
  assert (Lvz \c Lvb5) as HLvb5. {
    assert( Lvb4 \c Lvb5). unfold Lvb5.     union_sub_solver.
    assert( Lvb3 \c Lvb4). unfold Lvb4.     union_sub_solver.
    assert( Lvb2 \c Lvb3). eapply red_closure_mono_Lv. eapply typed_to_untyped_red_closure. eauto.
    eapply subset_trans. eapply HLvb2.
    eapply subset_trans. eauto.
    eapply subset_trans. eauto.
    eauto.
  }

  (* Stackless execution *)
  
  forwards HtsCc3: MethodCall.frame_within_truncate_stack stack_nil HfwCc3.
  2: eapply typed_to_untyped_red_closure.  2: eapply HredCc3. lia.
  destructecs HtsCc3.

  (* Before executing the first call on the right, relate to second call on the left to show termination *)
  assert (similar xs xs (trm_call (avar_f x1) m1 (avar_f xp1) ; stack_nil ; S ; e ; Lvz) (trm_call (avar_f x1) m1 (avar_f xp1) ; stack_nil ; Sb1; e & eb1 ; Lvz)) as HsiCc. {
    eapply similar_conf_1to1_vars_heap_bi. 4: eapply HsuSc3e. 3: auto. 2: auto.
    eapply similar_conf_1to1_vars_heap_renv_extend_r.
    eauto.
    auto.
    eapply disjoint_fresh. eauto. rewrite HsiCa3.
    sympl.


    forwards Hdi: red_closure_confing_fv_extension_disjoint HtsCc0. sympl in Hdi.
    unfold config_fv_extension in Hdi.
    eapply disjoint_comm.
    eapply disjoint_subset_l.
    eapply disjoint_subset_r.
    eapply Hdi.

    unfold Lvz. replace (fv t /[<- y2] /[# 1 <- y1] ) with (fv t /[<- y1] /[# 1 <- y2]).  union_sub_solver. rewrite ! fv_close. rewrite remove_remove. auto.
    sympl.
    eapply subset_union_2. 
    rewrite <- HsuSc3e1. rewrite dom_concat. rewrite union_remove, remove_same. rewrite remove_disjoint. union_sub_solver. eapply disjoint_comm. eapply ok_concat_disjoint. auto.
    rewrite dom_concat. rewrite union_remove, remove_same. rewrite remove_disjoint. union_sub_solver. eapply disjoint_comm. eapply ok_concat_disjoint. auto.
  }


  (* Execute the call *)
  forwards HredCc6: tred_call_end_exists_or_within (Datatypes.S n) HtyCc5.
  destructecs HredCc6.

  2: { (* Non-terminating case contradicts termination of first call on the left *)
    false.
    rename HredCc6 into HfwCc6, HredCc0 into HredCc6.
    forwards HtyCc6: typed_preservation_closure HtyCc5 HredCc6.
    destruct C2.

    (* Truncate stack in the overshooting execution *)
  
    forwards HtsCc6: frame_within_truncate_stack' HfwCc6.
    2: eapply typed_to_untyped_red_closure.  2: eapply HredCc6. lia.
    destructecs HtsCc6. subst s.
    
    (* Similarity of overshooting execution on the right with terminating exection on the left *)
    forwards Hsiprb:  similar_preservation_closure_answer_overshoot xs xs Hredsa3 HtsCc2.
    eapply similar_conf_Lv. eapply HsiCc.
    eapply config_hastyp_stack_nil_Lv_subset. eapply config_hastyp_c. eauto. union_sub_solver.
    eapply config_hastyp_stack_nil_Lv_subset. eapply config_hastyp_c. eauto. union_sub_solver.
    hnf. split. auto.  rewrite HsiCa3. union_sub_solver.
    hnf. split. auto.  rewrite HsiCa3. union_sub_solver.
    constructor.
    eapply Nat.nle_succ_diag_l; eauto.
  }

  rename HredCc6 into HfwCc6, HredCc0 into HredCc6.
  forwards HtyCc6: typed_preservation_closure HtyCc5 HredCc6.
  machine_step HtyCc6 Cc7 step_invert_simple.

  rename S2 into Sb2, e2 into eb2.
  rename Lv into Lvb7, Lv2 into Lvb6.

  (* Relate stacks *)


  forwards HGCc5: typed_red_closure_mono_G GCc2 GCc5 HredCc3.
  destructecs HGCc5. rename G3 into GGCc5.

  assert (ctx_ok ( GCc2 §& GGCc5)) as HokGGCc5. {
    inversions HtyCc5. eauto.
  }

  assert (subheap_unchanged Sb1 Sb2) as HsuSc6. {
    introv Hbi.
    forwards: typed_sef_guarantee TZ1 HtyCc5 HredCc6 HfwCc6 Hbi.
    rewrite HGCc0.
    eapply wkn_ewkn_end.  eapply ewkn_ty_var. eapply weaken_ty_var. eapply Htyx1. auto. auto. auto.
    rewrite HGCc0.
    eapply wkn_ewkn_end.  eapply ewkn_ty_var. eapply weaken_ty_var. eapply Htyxp1. auto. auto. auto.
    rewrite HGCc0.
    eapply wkn_ewkn_end. eapply ewkn_sub. eapply weaken_subtyp. eapply HsuTZ1. auto. auto. auto.
    auto.
  }
  
  (* Stackless execution *)
  
  forwards HtsCc6: MethodCall.frame_within_truncate_stack stack_nil HfwCc6.
  2: eapply typed_to_untyped_red_closure.  2: eapply HredCc6. lia.
  destructecs HtsCc6.

  (* Split heaps *)

  assert (ok Sb2) as HokSb2 by (eapply ty_config_heap_ok; eauto).

  forwards HsuSc6e: subheap_unchanged_split HsuSc6; fweauto.
  destructecs HsuSc6e.

  rename S3 into Sbm1, S0 into Sbm2.

  (* Assertions about variables *)

  assert (x2 \in Lvb2) as Hx2Lv. {
      unfold Lvb2, tb'. sympl. union_in_solver.
  }
  assert (xp2 \in Lvb2) as Hxp2Lv. {
      unfold Lvb2, tb'. sympl. union_in_solver.
  }

  assert (ok (e & eb1 & eb2)) as Hokeb12. {
    eapply ty_config_env_ok; eauto.
  }

  assert (Lvz \c Lvb7) as HLvb7. {
    assert( Lvb6 \c Lvb7). unfold Lvb7.     union_sub_solver.
    assert( Lvb5 \c Lvb6). eapply red_closure_mono_Lv. eapply typed_to_untyped_red_closure. eauto.
    eapply subset_trans. eapply HLvb5.
    eapply subset_trans. eauto.
    eauto.
  }

  (* Similarity on calls *)
  forwards Hsipra:  similar_preservation_closure_answer xs xs Hredsa3 HtsCc1. 
  eapply similar_conf_Lv. eapply HsiCc.
  eapply config_hastyp_stack_nil_Lv_subset. eapply config_hastyp_c. eauto. union_sub_solver.
  eapply config_hastyp_stack_nil_Lv_subset. eapply config_hastyp_c. eauto. union_sub_solver.
  eapply vars_conf_unique_fs_nil_Lv_subset. eauto. sympl.  union_sub_solver.
  hnf. split. auto.  rewrite HsiCa3. union_sub_solver.
  constructor.
  constructor.

  forwards Hsiprb:  similar_preservation_closure_answer xs xs Hredsa6 HtsCc0.
  eapply similar_conf_Lv. eapply similar_symm. eapply HsiCa.
  eapply config_hastyp_stack_nil_Lv_subset. eapply config_hastyp_c. eauto. union_sub_solver.
  eapply config_hastyp_stack_nil_Lv_subset. eapply config_hastyp_c. eauto. union_sub_solver.
  hnf. split. auto.  rewrite HsiCa3. union_sub_solver.
  hnf. split. auto.  rewrite HsiCa3. union_sub_solver.
  constructor.
  constructor.

  destructecs Hsipra.
  destructecs Hsiprb.
  subst.

  rename Hsipra1 into Hsipra, Hsipra2 into Hvcuya, Hsipra3 into Hvcuyb, Hsipra4 into Hdomya, Hsipra into Hdomyb.
  rename Hsiprb1 into Hsiprb, Hsiprb2 into Hvcuya0, Hsiprb3 into Hvcuyb0, Hsiprb4 into Hdomya0, Hsiprb into Hdomyb0.

  (* Domains *)

  unfold ReductionFreeVars.config_fv_extension in Hdomya; sympl in Hdomya;
  unfold ReductionFreeVars.config_fv_extension in Hdomyb; sympl in Hdomyb;
  unfold ReductionFreeVars.config_fv_extension in Hdomya0; sympl in Hdomya0;
  unfold ReductionFreeVars.config_fv_extension in Hdomyb0; sympl in Hdomyb0.

  rewrite ok_dom_remove in Hdomya by eauto;
  rewrite <- HsuSa3e1 in Hdomya;
  rewrite ok_dom_remove in Hdomya by auto.


  rewrite ok_dom_remove in Hdomyb0 by eauto;
  rewrite <- HsuSc3e1 in Hdomyb0;
  rewrite ok_dom_remove in Hdomyb0 by auto.

  rewrite ok_dom_remove in Hdomya0 by auto;
  rewrite <- HsuSa6e1 in Hdomya0;
  rewrite ok_dom_remove in Hdomya0 by auto.


  rewrite ok_dom_remove in Hdomyb by auto;
  rewrite <- HsuSc6e1 in Hdomyb;
  rewrite ok_dom_remove in Hdomyb by auto.

  assert (length ya = length yb) as Hlyab. {
    forwards Hl: similar_length Hsipra.
    rewrite ! List.app_length in Hl.
    apply -> PeanoNat.Nat.add_cancel_r in Hl.
    auto.
  }

  assert (length ya0 = length yb0) as Hlyab0. {
    forwards Hl: similar_length Hsiprb.
    rewrite ! List.app_length in Hl.
    apply -> PeanoNat.Nat.add_cancel_r in Hl.
    auto.
  }

  assert (fresh (from_list xs) (length ya0) ya0) as Hfrya0. {
    hnde Hvcuya0. eapply fresh_append_fresh_l in Hvcuya1. auto.
  }
  assert (fresh (from_list xs) (length yb0) yb0) as Hfryb0. {
    hnde Hvcuyb0. eapply fresh_append_fresh_l in Hvcuyb1. auto.
  }
  assert (fresh (from_list xs) (length ya) ya) as Hfrya. {
    hnde Hvcuya. eapply fresh_append_fresh_l in Hvcuya1. auto.
  }
  assert (fresh (from_list xs) (length yb) yb) as Hfryb. {
    hnde Hvcuyb. eapply fresh_append_fresh_l in Hvcuyb1. auto.
  }

  assert (disjoint ( dom( e & eb1 & eb2 )) (dom Sb2)) as HdieS3. {
    eapply ty_config_disjoint_Se; eauto.
  }
  assert (disjoint ( dom( e & ea1 & ea2 )) (dom Sa2)) as HdieSa2. {
    eapply ty_config_disjoint_Se; eauto.
  }

  assert (HdieS3u : disjoint ((dom e \u dom eb1) \u dom eb2) ((dom S \u dom Sbm1) \u dom Sbm2)). {
    simpl_dom.
    rewrite <- HsuSc6e1, <- HsuSc3e1 in HdieS3. auto.
  }
  assert (HdieS0u : disjoint ((dom e \u dom ea1) \u dom ea2) ((dom S \u dom Sam1) \u dom Sam2)). {
    simpl_dom.
    rewrite <- HsuSa6e1, <- HsuSa3e1 in HdieSa2. auto.
  }

  assert (disjoint (from_list ya0) (from_list ya)) as Hidsya. {
    simpl_dom.
    rewrite <- Hdomya, <- Hdomya0.

    eapply ok_concat_disjoint in HsuSa6e0. rewrite <- HsuSa3e1 in HsuSa6e0.
    eapply ok_concat_disjoint in HsuSc6e0. rewrite <- HsuSc3e1 in HsuSc6e0.
    eapply ok_concat_disjoint in Hokeb12. rewrite dom_concat in Hokeb12. 
    eapply ok_concat_disjoint in Hokea10. rewrite dom_concat in Hokea10.

    Ltac destruct_disjoint := match goal with 
    | H: disjoint (?A \u ?B) ?C |- _ => apply disjoint_union_l_inv in H; destructec H
    | H: disjoint ?C (?A \u ?B)  |- _ => apply disjoint_union_r_inv in H; destructec H
    end.

    repeat destruct_disjoint.

    eapply disjoint_union_r; eapply disjoint_union_l; auto 2 using disjoint_comm.
  }
  assert (disjoint (from_list yb0) (from_list yb)) as Hidsyb. {
    simpl_dom.
    rewrite <- Hdomyb, <- Hdomyb0.

    eapply ok_concat_disjoint in HsuSa6e0. rewrite <- HsuSa3e1 in HsuSa6e0.
    eapply ok_concat_disjoint in HsuSc6e0. rewrite <- HsuSc3e1 in HsuSc6e0.
    eapply ok_concat_disjoint in Hokeb12. rewrite dom_concat in Hokeb12. 
    eapply ok_concat_disjoint in Hokea10. rewrite dom_concat in Hokea10.

    repeat destruct_disjoint.

    eapply disjoint_union_r; eapply disjoint_union_l; auto 2 using disjoint_comm.
  }
 
  (* Compose typed execution *)

  set (noa := (Datatypes.S( Datatypes.S(Datatypes.S(Datatypes.S(0)+ n2))+ n1))).
  set (toa := t /[<- y2] /[# 1 <- y1] /[# 1 -> w] /[-> w0] ).

  assert (GCc2 ⊢ ta' ; fs0 ; S ; e ; Lva ∶ Tb ⊣ noa ↦* GCa7 ⊢ (toa ; fs0 ; Sa2 ; e & ea1 & ea2 ; Lva7) ) as Hexeca. {
    assert (((open_rec_trm 1 w t /[<- y2] /[# 1 <- y1]) /[-> w0]) = toa). {
      sympl. unfold toa. auto. 
    }
    eapply typed_red_step. 2: eapply HredCa7.
    eapply typed_red_closure_concat. 2: eapply HredCa6.
    eapply typed_red_step. 2: eapply HredCa5.
    eapply typed_red_step. 2: eapply HredCa4.
    eapply typed_red_closure_concat. 2: eapply HredCa3.
    eapply typed_red_step. 2: eapply HredCa2. eauto.
  }
  set (nob := (Datatypes.S( Datatypes.S(Datatypes.S(Datatypes.S(0)+ n1))+ n2))).
  set (tob := t /[<- y1] /[# 1 <- y2] /[# 1 -> w1] /[-> w2] ).
  assert (GCc2 ⊢ tb' ; fs0 ; S ; e ; Lvb ∶ Tb ⊣ nob ↦* GCc7 ⊢ (tob ; fs0 ; Sb2 ; e & eb1 & eb2 ; Lvb7) )as Hexevcb. {
    assert (((open_rec_trm 1 w1 t /[<- y1] /[# 1 <- y2]) /[-> w2]) = tob). {
      sympl. unfold tob. auto. 
    }
    eapply typed_red_step. 2: eapply HredCc7.
    eapply typed_red_closure_concat. 2: eapply HredCc6.
    eapply typed_red_step. 2: eapply HredCc5.
    eapply typed_red_step. 2: eapply HredCc4.
    eapply typed_red_closure_concat. 2: eapply HredCc3.
    eapply typed_red_step. 2: eapply HredCc2. eauto.
  }

  assert(noa = nob) as Heqn. {
    unfold noa, nob; lia.
  }
  rewrite Heqn in Hexeca.

  (* Fill the goal *)

  exists nob nob.
  exists GCa7 GCc7.
  exists (toa ; fs0 ; Sa2 ; e & ea1 & ea2 ; Lva7) (tob; fs0 ; Sb2 ; e & eb1 & eb2 ; Lvb7).
  splits.
  -
    auto.
  - auto.
  - auto.
  -
    econstructor.
  + eapply typed_preservation_closure. 2: eapply Hexeca. eauto.
  + eapply typed_preservation_closure. 2: eapply Hexevcb. eauto.
  + hnf.
    exists (ya ++ ya0 ++ xs)%list (yb ++ yb0 ++ xs)%list; splits.
    * hnf; split. 
      ** eapply fresh_append.
          eauto.
          rewrite union_empty_l.
        eapply fresh_append.
        eapply disjoint_fresh. eauto. auto 2 using disjoint_comm.
        eapply fresh_union_l; eapply fresh_swap; eauto.
      ** rewrite !from_list_append. rewrite <- Hdomya, <- Hdomya0, HsiCa3. rewrite <- HsuSa6e1, dom_concat, <- HsuSa3e1, dom_concat. rewrite ! dom_concat. union_sub_solver.
    * hnf; split. 
      ** eapply fresh_append.
          eauto.
          rewrite union_empty_l.
        eapply fresh_append.
        eapply disjoint_fresh. eauto. auto 2 using disjoint_comm.
        eapply fresh_union_l; eapply fresh_swap; eauto.
      ** rewrite !from_list_append. rewrite <- Hdomyb, <- Hdomyb0, HsiCa3. rewrite <- HsuSc6e1, dom_concat, <- HsuSc3e1, dom_concat. rewrite ! dom_concat. union_sub_solver.
    * rewrite! List.app_length. rewrite Hlyab, Hlyab0. auto.
    * hnf; splits.
      ** 
        assert (similar (ya ++ xs) (yb ++ xs) w w2). {
          hnde Hsipra.
          eapply similar_trm_var_var in Hsipra0.
          eapply similar_avar_f_var in Hsipra0. auto.
        }
        assert (similar (ya0 ++ xs) (yb0 ++ xs) w0 w1). {
          hnde Hsiprb.
          eapply similar_trm_var_var in Hsiprb0.
          eapply similar_avar_f_var in Hsiprb0. auto.
        }

        unfold toa, tob.
        rewrite trm_closed_close_open_swap with (y2:=y1). 
        eapply similar_open_rec.
        eapply similar_open_rec.
        eapply similar_out_append.
        eapply similar_out_append.
        2: auto.
        2: auto.
        eapply similar_refl_by_fv.
        rewrite HsiCa3.
        unfold Lvz.
        union_sub_solver.
      eapply similar_out_append_middle. auto. auto. auto.
      eapply similar_out_append. auto. auto. auto. fresh_chain_solve Fr.
      **
        eapply similar_out_append.
        eapply similar_out_append.
        eapply similar_refl_by_fv.
        rewrite HsiCa3.
        unfold Lvz.
        union_sub_solver.
        auto.
        auto.
      ** (* Heaps *)
        assert (fresh (dom S) (length ya0) ya0) as Hfrya0e. {
          eapply fresh_subset. 2: eapply Hfrya0. rewrite HsiCa3. unfold Lvz. union_sub_solver.
        }
        assert (fresh (dom S) (length yb0) yb0) as Hfryb0e. {
          hnde Hvcuyb0. eapply fresh_append_fresh_l in Hvcuyb1.
          eapply fresh_subset. 2: eapply Hfryb0. rewrite HsiCa3. unfold Lvz. union_sub_solver.
        }
        assert (fresh (dom S) (length ya) ya) as Hfryae. {
          hnde Hvcuya. eapply fresh_append_fresh_l in Hvcuya1.
          eapply fresh_subset. 2: eapply Hfrya. rewrite HsiCa3. unfold Lvz. union_sub_solver.
        }
        assert (fresh (dom S) (length yb) yb) as Hfrybe. {
          hnde Hvcuyb. eapply fresh_append_fresh_l in Hvcuyb1.
          eapply fresh_subset. 2: eapply Hfryb. rewrite HsiCa3. unfold Lvz. union_sub_solver.
        }
        assert (fresh (dom Sbm1) (length yb) yb) as Hfrybe1. {
          eapply disjoint_fresh. eauto.
          eapply disjoint_subset_r. eapply Hidsyb.
          rewrite <- Hdomyb0. union_sub_solver.
        } 
        assert (fresh (dom Sam1) (length ya0) ya0) as Hfryae2. {
          eapply disjoint_fresh. eauto.
          eapply disjoint_subset_r. eapply disjoint_comm. eapply Hidsya.
          rewrite <- Hdomya. union_sub_solver.
        }
        assert (fresh (dom Sbm2) (length yb0) yb0) as Hfrybe3. {
          eapply disjoint_fresh. eauto.
          eapply disjoint_subset_r. eapply disjoint_comm. eapply Hidsyb.
          rewrite <- Hdomyb. union_sub_solver.
        }
        assert (fresh (dom Sam2) (length ya) ya) as Hfryae0. {
          eapply disjoint_fresh. eauto.
          eapply disjoint_subset_r. eapply Hidsya.
          rewrite <- Hdomya0. union_sub_solver.
        }
        assert (fresh (dom Sam1) (length xs) xs) as Hfrxse2. {
          eapply fresh_subset with (F:=(from_list ya)). 
          rewrite <- Hdomya. union_sub_solver.
          eapply fresh_swap; eauto.
        }
        assert (fresh (dom Sbm2) (length xs) xs) as Hfrxse3. {
          eapply fresh_subset with (F:=(from_list yb)). 
          rewrite <- Hdomyb. union_sub_solver.
          eapply fresh_swap; eauto.
        }
        assert (fresh (dom Sam2) (length xs) xs) as Hfrxse0. {
          eapply fresh_subset with (F:=(from_list ya0)). 
          rewrite <- Hdomya0. union_sub_solver.
          eapply fresh_swap; eauto.
        }
        assert (fresh (dom Sbm1) (length xs) xs) as Hfrxse1. {
          eapply fresh_subset with (F:=(from_list yb0)). 
          rewrite <- Hdomyb0. union_sub_solver.
          eapply fresh_swap; eauto.
        }


        assert (ok (S & Sbm1 & Sbm2)) as HokSbm12. { 

          eapply ok_concat; auto.
          rewrite HsuSc3e1.
          eapply ok_concat_disjoint. auto.
        }
        assert (ok (S & Sam1 & Sam2)) as Hok45. { 
          eapply ok_concat; auto.
          rewrite HsuSa3e1.
          eapply ok_concat_disjoint. auto.
        }

        assert (dom (Sa1 & Sam2) = dom (S & Sam1 & Sam2)) as Hdom25. {
          rewrite !dom_concat, <-HsuSa3e1, dom_concat. auto.
        }
        assert (dom (Sb1 & Sbm2) = dom (S & Sbm1 & Sbm2)) as Hdom17. {
          rewrite !dom_concat, <-HsuSc3e1, dom_concat. auto.
        }

        (* ya yb side *)
        assert (similar (ya ++ xs) (yb ++ xs) Sa1 Sb2) as HsiSa13. {
          hnde Hsipra. auto.
        }

        forwards HsiSa117: similar_with_heap_elements_rb HsiSa13 HsuSc6e; fweauto.
        forwards HsuSam117: similar_with_heap_elements_lb HsiSa117 HsuSa3e; fweauto.

        assert (subheap_unchanged  (S & Sbm1 & Sbm2) (Sb1 & Sbm2) ) as Hsu67. {
          eapply subheap_unchanged_concat_l. eauto.
        }

        forwards HsuSam167: similar_with_heap_elements_rb HsuSam117 Hsu67; fweauto. auto. auto.
        clear HsiSa117 HsuSam117.

        (* ya0 yb0 side *)
        assert (similar (ya0 ++ xs) (yb0 ++ xs) Sa2 Sb1) as HsiSa21. {
          hnde Hsiprb. auto.
        }

        forwards HsiSa26: similar_with_heap_elements_rb HsiSa21 HsuSc3e; fweauto.
        forwards HsuSa156: similar_with_heap_elements_lb HsiSa26 HsuSa6e; fweauto.

        assert (subheap_unchanged  (S & Sam1 & Sam2) (Sa1 & Sam2) ) as Hsu45. {
          eapply subheap_unchanged_concat_l. eauto.
        }

        forwards HsuSam156: similar_with_heap_elements_lb HsuSa156 Hsu45; fweauto. auto. auto.

        applys similar_with_heap_elements_lf HsuSa6e. 2: auto. 2: auto.
        applys similar_with_heap_elements_lf Hsu45. 2: auto. 2: symmetry; auto.
        applys similar_with_heap_elements_rf HsuSc6e.  2: auto. 2: auto.
        applys similar_with_heap_elements_rf Hsu67. 2: auto. 2: symmetry; auto.

        (* Align the envs *)
        eapply similar_env_with_er_swap_after. eapply ok_swap_after; eauto.
        eapply similar_env_with_e_concat.
        eapply similar_env_with_e_concat.
        eapply similar_env_append_fresh_r.
        eapply similar_env_append_fresh_r.
        hnde HsiCa1. auto.
        1-6: auto.
        ***
        assert (similar_with (ya ++ xs) (yb ++ xs) ya yb Sam1 Sbm2). {
          eapply similar_env_with_i_append_swap in HsuSam167.
          eapply similar_env_with_ie_invr in HsuSam167.
          auto.
          auto.
          auto.
          auto.
          eapply fresh_disjoint; eauto.
          rewrite dom_concat.
          eapply disjoint_union_l.
          eapply fresh_disjoint; eauto.
          eapply fresh_disjoint; eauto.
          auto.
        }
        

        eapply similar_env_with_o_append_m. 2: auto. 2: auto.
        eapply similar_env_with_i_append_fresh_l. auto.

        eapply fresh_append. auto.
        eapply fresh_union_l. auto.
                  eapply fresh_swap; eauto.


        eapply fresh_append. auto.
        eapply fresh_union_l. auto.
                  eapply fresh_swap; eauto.
        rewrite ! List.app_length. rewrite Hlyab0. auto.

        ***
        assert (similar_with (ya0 ++ xs) (yb0 ++ xs) ya0 yb0 Sam2 Sbm1). {
          eapply similar_env_with_i_append_swap in HsuSa156.
          eapply similar_env_with_ie_invr in HsuSa156.
          auto.
          auto.
          auto.
          auto.
    
          rewrite <-HsuSa3e1,  dom_concat.
          eapply disjoint_union_l.
          eapply fresh_disjoint; eauto.
          eapply fresh_disjoint; eauto.
          eapply fresh_disjoint; eauto.
          auto.
        }
        

        eapply similar_env_with_o_append_r. 2: auto.
        eapply similar_env_with_i_append_fresh_r.
        eapply similar_env_with_i_append_fresh_l. auto.
        all: auto.

      ** (* Envs *)
        assert (fresh (dom e) (length ya0) ya0) as Hfrya0e. {
          eapply fresh_subset. 2: eapply Hfrya0. rewrite HsiCa3. unfold Lvz. union_sub_solver.
        }
        assert (fresh (dom e) (length yb0) yb0) as Hfryb0e. {
          hnde Hvcuyb0. eapply fresh_append_fresh_l in Hvcuyb1.
          eapply fresh_subset. 2: eapply Hfryb0. rewrite HsiCa3. unfold Lvz. union_sub_solver.
        }
        assert (fresh (dom e) (length ya) ya) as Hfryae. {
          hnde Hvcuya. eapply fresh_append_fresh_l in Hvcuya1.
          eapply fresh_subset. 2: eapply Hfrya. rewrite HsiCa3. unfold Lvz. union_sub_solver.
        }
        assert (fresh (dom e) (length yb) yb) as Hfrybe. {
          hnde Hvcuyb. eapply fresh_append_fresh_l in Hvcuyb1.
          eapply fresh_subset. 2: eapply Hfryb. rewrite HsiCa3. unfold Lvz. union_sub_solver.
        }
        assert (fresh (dom eb1) (length yb) yb) as Hfrybe1. {
          eapply disjoint_fresh. eauto.
          eapply disjoint_subset_r. eapply Hidsyb.
          rewrite <- Hdomyb0. union_sub_solver.
        } 
        assert (fresh (dom ea1) (length ya0) ya0) as Hfryaea1. {
          eapply disjoint_fresh. eauto.
          eapply disjoint_subset_r. eapply disjoint_comm. eapply Hidsya.
          rewrite <- Hdomya. union_sub_solver.
        }
        assert (fresh (dom eb2) (length yb0) yb0) as Hfrybe3. {
          eapply disjoint_fresh. eauto.
          eapply disjoint_subset_r. eapply disjoint_comm. eapply Hidsyb.
          rewrite <- Hdomyb. union_sub_solver.
        }
        assert (fresh (dom ea2) (length ya) ya) as Hfryaea2. {
          eapply disjoint_fresh. eauto.
          eapply disjoint_subset_r. eapply Hidsya.
          rewrite <- Hdomya0. union_sub_solver.
        }
        assert (fresh (dom ea1) (length xs) xs) as Hfrxsea1. {
          eapply fresh_subset with (F:=(from_list ya)). 
          rewrite <- Hdomya. union_sub_solver.
          eapply fresh_swap; eauto.
        }
        assert (fresh (dom eb2) (length xs) xs) as Hfrxse3. {
          eapply fresh_subset with (F:=(from_list yb)). 
          rewrite <- Hdomyb. union_sub_solver.
          eapply fresh_swap; eauto.
        }
        assert (fresh (dom ea2) (length xs) xs) as Hfrxsea2. {
          eapply fresh_subset with (F:=(from_list ya0)). 
          rewrite <- Hdomya0. union_sub_solver.
          eapply fresh_swap; eauto.
        }
        assert (fresh (dom eb1) (length xs) xs) as Hfrxse1. {
          eapply fresh_subset with (F:=(from_list yb0)). 
          rewrite <- Hdomyb0. union_sub_solver.
          eapply fresh_swap; eauto.
        }

        (* Align the envs *)
        eapply similar_env_with_er_swap_after. eapply ok_swap_after; eauto.
        eapply similar_env_with_e_concat.
        eapply similar_env_with_e_concat.
        eapply similar_env_append_fresh_r.
        eapply similar_env_append_fresh_r.
        hnde HsiCa1. auto.
        1-6: auto.
        ***
        assert (similar (ya ++ xs) (yb ++ xs) (e &  ea1) ( e & eb1 & eb2)) as Hsiea113. {
          hnde Hsipra. auto.
        }

        assert (similar_with (ya ++ xs) (yb ++ xs) ya yb ea1 eb2). {
          eapply similar_env_with_i_append_swap in Hsiea113.
          eapply similar_env_with_ie_invr in Hsiea113.
          auto.
          auto.
          auto.
          auto.
          eapply fresh_disjoint; eauto.
          rewrite dom_concat.
          eapply disjoint_union_l.
          eapply fresh_disjoint; eauto.
          eapply fresh_disjoint; eauto.
          auto.
        }
        

        eapply similar_env_with_o_append_m. 2: auto. 2: auto.
        eapply similar_env_with_i_append_fresh_l. auto.

        eapply fresh_append. auto.
        eapply fresh_union_l. auto.
                  eapply fresh_swap; eauto.


        eapply fresh_append. auto.
        eapply fresh_union_l. auto.
                  eapply fresh_swap; eauto.
        rewrite ! List.app_length. rewrite Hlyab0. auto.

        ***
        assert (similar (ya0 ++ xs) (yb0 ++ xs) ( e & ea1 & ea2)  (e & eb1)) as Hsie120. {
          hnde Hsiprb. auto.
        }

        assert (similar_with (ya0 ++ xs) (yb0 ++ xs) ya0 yb0 ea2 eb1). {
          eapply similar_env_with_i_append_swap in Hsie120.
          eapply similar_env_with_ie_invr in Hsie120.
          auto.
          auto.
          auto.
          auto.
          rewrite dom_concat.
          eapply disjoint_union_l.
          eapply fresh_disjoint; eauto.
          eapply fresh_disjoint; eauto.
          eapply fresh_disjoint; eauto.
          auto.
        }
        

        eapply similar_env_with_o_append_r. 2: auto.
        eapply similar_env_with_i_append_fresh_r.
        eapply similar_env_with_i_append_fresh_l. auto.
        all: auto.
Qed.


End WithTypingMode.