(** remove printing ~ *)

Set Implicit Arguments.

Require Import LibExtra DotTactics AbstractSyntax.
Require Import Context Equivalence Indep GeneralTyping.
Require Import Transformation Weakening EWeakening Substitution TFree.
Require Import TransformationSimilarity TransformationPreservation TransformationTerm.
Require Import TransformationConfig OperationalSemantics TypedOperationalSemantics.
Require Import ContextSimilarity ConfigSimilarity.
Require Import ConfigTyping ConfigTypingAccessories TypedProgress.
Require Import TypedPreservation.
Require Import TransformationIterateSim.
Require Import Similarity TypingSimilarity.

Require Import TransformationLifting.
Require Import FreshChain FreshChainTactics.
Require Import TypedOperationalSemantics InertContext.
Require Import Similarity SyntaxSimilarity.
Require Import ExecutionTactics Renaming TypingHelpers.

Require Import NothingFresh TypedPreservation LetLet Defs.
Require Import HeapCorrespondence ConfigVarkinds.
Require Import HeapUpdate ConfigValid.

Require Import TransformationSimilaritySteps ConfigTypingSteps.
Require Import Defs LetLet EnvCorrespondence.
Require Import Subenvironments Narrowing.
Require Import Defree.
Require Import Visibility.
Require Import TypedPreservationLemmas.

Section WithTypingMode.
Context (typing_mode: Mode).

(** * Lemmas about transformation of typed configurations *)

Lemma trf_subst_w G e (ta: trm) w (T: typ) Tt x vk:
  ctx_ok G ->
  x \notin ( ctx_dom G \u  ctx_fv G \u fv_renv e \u fv Tt \u fv ta \u fv T) ->
  G ⁏ e ⊢ trm_var (avar_f w) ∶ Tt ->
  〈 G § x / vk ~ Tt ⁏ e & x ~ w ⊢ ta /[-> x] ∶ T 〉 ∼ TransformSubst typing_mode w Tt ∼> 〈 G ⁏ e ⊢ ta /[-> w] ∶ T 〉.
Proof.
  introv HokG Hnx Htyw.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx.
  fresh_chain_cleanup Fr.

  pose proof trf_subst (A:=trm) (T:=typ) Tt (y:=w) G (e:=e)  (x:=x) vk ctx_empty as Htrf.
  specialize (Htrf (ta /[-> x]) T).
  forwards: Htrf.
  - fresh_chain_solve Fr.
  - fresh_chain_solve Fr.
  - unfold renv_notin. fresh_chain_solve Fr.
  - rewrite subst_fresh. eapply ty_trm_var_inv;  eauto. fresh_chain_solve Fr.
  - intros. eapply ty_trm_var_inv_vis;  eauto.
  - 
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_fresh in H by ( fresh_chain_solve Fr).
    rewrite subst_fresh in H by ( fresh_chain_solve Fr).
    hnf. apply H.
Qed.

Lemma trf_subst_wz G e (ta: trm) w (T: typ) Tt x vk:
  ctx_ok G ->
  x \notin ( ctx_dom G \u  ctx_fv G \u fv_renv e \u fv Tt \u fv ta \u fv T) ->
  G ⁏ e ⊢ trm_var (avar_f w) ∶ Tt ->
  〈 G § x / vk ~ Tt ⁏ e & x ~ w ⊢ ta /[-> x] ∶ T  /[-> x]〉 ∼ TransformSubst typing_mode w Tt ∼> 〈 G ⁏ e ⊢ ta /[-> w] ∶ T  /[-> w]〉.
Proof.
  introv HokG Hnx Htyw.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx.
  fresh_chain_cleanup Fr.

  pose proof trf_subst (A:=trm) (T:=typ) Tt (y:=w) G (e:=e)  (x:=x) vk ctx_empty as Htrf.
  specialize (Htrf (ta /[-> x]) T /[-> x]).
  forwards: Htrf.
  - fresh_chain_solve Fr.
  - fresh_chain_solve Fr.
  - unfold renv_notin. fresh_chain_solve Fr.
  - rewrite subst_fresh. eapply ty_trm_var_inv;  eauto. fresh_chain_solve Fr.
  - intros. eapply ty_trm_var_inv_vis;  eauto.
  - 
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_fresh in H by ( fresh_chain_solve Fr).
    rewrite subst_fresh in H by ( fresh_chain_solve Fr).
    hnf. apply H.
Qed.
Lemma trf_subst_wz0 G e (ta: trm) w (T: typ) Tt x vk:
  ctx_ok G ->
  x \notin ( ctx_dom G \u  ctx_fv G \u fv_renv e \u fv Tt \u fv ta \u fv T) ->
  G ⁏ e ⊢ trm_var (avar_f w) ∶ Tt  ->
  〈 G § x / vk ~ Tt ⁏ e & x ~ w ⊢ ta /[-> x] ∶ T  /[-> x]〉 ∼ TransformSubst typing_mode w Tt ∼> 〈 G ⁏ e ⊢ ta /[-> w] ∶ T  /[-> w]〉.
Proof.
  introv HokG Hnx Htyw.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx.
  fresh_chain_cleanup Fr.

  pose proof trf_subst (A:=trm) (T:=typ) Tt (y:=w) G (e:=e)  (x:=x) vk ctx_empty as Htrf.
  specialize (Htrf (ta /[-> x]) T /[-> x]).
  forwards: Htrf.
  - fresh_chain_solve Fr.
  - fresh_chain_solve Fr.
  - unfold renv_notin. fresh_chain_solve Fr.
  - rewrite subst_fresh. eapply ty_trm_var_inv;  eauto. fresh_chain_solve Fr.
  - intros. eapply ty_trm_var_inv_vis;  eauto.
  - 
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_fresh in H by ( fresh_chain_solve Fr).
    rewrite subst_fresh in H by ( fresh_chain_solve Fr).
    hnf. apply H.
Qed.

Lemma trf_subst_wz1 G e (ta: trm) w (T: typ) Tt x vk:
  ctx_ok G ->
  x \notin ( ctx_dom G \u  ctx_fv G \u fv_renv e \u fv ta \u fv T) ->
  G ⁏ e ⊢ trm_var (avar_f w) ∶ Tt /[ x -> w] ->
  〈 G § x / vk ~ Tt ⁏ e & x ~ w ⊢ ta /[#1-> x] ∶ T  /[#1-> x]〉 ∼ TransformSubst typing_mode w Tt ∼> 〈 G ⁏ e ⊢ ta /[#1-> w] ∶ T  /[#1-> w]〉.
Proof.
  introv HokG Hnx Htyw.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx.
  fresh_chain_cleanup Fr.

  pose proof trf_subst (A:=trm) (T:=typ) Tt (y:=w) G (e:=e)  (x:=x) vk ctx_empty as Htrf.
  specialize (Htrf (ta /[#1-> x]) T /[#1-> x]).
  forwards: Htrf.
  - fresh_chain_solve Fr.
  - fresh_chain_solve Fr.
  - unfold renv_notin. fresh_chain_solve Fr.
  -  eapply ty_trm_var_inv;  eauto.
  - intros. eapply ty_trm_var_inv_vis;  eauto.
  - 
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_fresh in H by ( fresh_chain_solve Fr).
    rewrite subst_fresh in H by ( fresh_chain_solve Fr).
    hnf. apply H.
Qed.

Lemma trf_subst_wz1' G e (ta: trm) w (T: typ) Tt x vk x' vk' T':
  ctx_ok G ->
  x' \notin ctx_dom G ->
  x \notin ( ctx_dom G \u  ctx_fv G \u fv_renv e \u fv ta \u fv T) ->
  x <> x' ->
  G § x' / vk' ~ T' /[ x -> w] ⁏ e ⊢ trm_var (avar_f w) ∶ Tt /[ x -> w] ->
  〈 G § x / vk ~ Tt § x' / vk' ~ T' ⁏ e & x ~ w ⊢ ta /[#1-> x] ∶ T  /[#1-> x]〉 ∼ TransformSubst typing_mode w Tt ∼> 〈 G § x' / vk' ~ T' /[x-> w] ⁏ e ⊢ ta /[#1-> w] ∶ T  /[#1-> w]〉.
Proof.
  introv HokG Hnx' Hnx Hnxx Htyw.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx.
  fresh_chain_cleanup Fr.

  assert (ctx_ok ((G § x / vk ~ Tt) § x' / vk' ~ T')) as HokGx.
  econstructor.
  fresh_chain_solve Fr.
  eauto.

  pose proof trf_subst (A:=trm) (T:=typ) Tt (y:=w) G (e:=e)  (x:=x) vk (ctx_empty § x' / vk' ~ T') as Htrf.
  specialize (Htrf (ta /[#1-> x]) T /[#1-> x]).
  forwards: Htrf.
  - auto.
  - fresh_chain_solve Fr.
  - unfold renv_notin. fresh_chain_solve Fr.
  -  eapply ty_trm_var_inv;  eauto.
  - intros. eapply ty_trm_var_inv_vis;  eauto.
  - 
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_fresh with (X:=ta) in H by ( fresh_chain_solve Fr).
    rewrite subst_fresh with (X:=T) in H by ( fresh_chain_solve Fr).
    hnf. apply H.
Qed.

Lemma trf_subst_wz0' G e (ta: trm) w (T: typ) Tt x vk x' vk' T':
  ctx_ok G ->
  x' \notin ctx_dom G ->
  x \notin ( ctx_dom G \u  ctx_fv G \u fv_renv e \u fv ta \u fv T) ->
  x <> x' ->
  G § x' / vk' ~ T' /[ x -> w] ⁏ e ⊢ trm_var (avar_f w) ∶ Tt /[ x -> w] ->
  〈 G § x / vk ~ Tt § x' / vk' ~ T' ⁏ e & x ~ w ⊢ ta /[-> x] ∶ T  /[-> x]〉 ∼ TransformSubst typing_mode w Tt ∼> 〈 G § x' / vk' ~ T' /[x-> w] ⁏ e ⊢ ta /[-> w] ∶ T  /[-> w]〉.
Proof.
  introv HokG Hnx' Hnx Hnxx Htyw.

  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx.
  fresh_chain_cleanup Fr.

  assert (ctx_ok ((G § x / vk ~ Tt) § x' / vk' ~ T')) as HokGx.
  econstructor.
  fresh_chain_solve Fr.
  eauto.

  pose proof trf_subst (A:=trm) (T:=typ) Tt (y:=w) G (e:=e)  (x:=x) vk (ctx_empty § x' / vk' ~ T') as Htrf.
  specialize (Htrf (ta /[-> x]) T /[-> x]).
  forwards: Htrf.
  - auto.
  - fresh_chain_solve Fr.
  - unfold renv_notin. fresh_chain_solve Fr.
  -  eapply ty_trm_var_inv;  eauto.
  - intros. eapply ty_trm_var_inv_vis;  eauto.
  - 
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_open_commut in H. cases_if.
    rewrite subst_fresh with (X:=ta) in H by ( fresh_chain_solve Fr).
    rewrite subst_fresh with (X:=T) in H by ( fresh_chain_solve Fr).
    hnf. apply H.
Qed.

Lemma  trf_ewkn_push' (A AT: Type) G e (t: A) (T: AT) w y:
  ctx_ok G ->
  ok e ->
  w # e ->
  〈 G⁏e ⊢ t  ∶ T 〉 ∼ @TransformEWeakening AT A ∼> 〈 G ⁏ e & w ~ y ⊢ t ∶ T 〉.
Proof.
  introv HokG Hoke Hne.
  rewrite <- concat_empty_r with (E:=e) at 1.
  rewrite <- concat_empty_r with (E:=e & w ~ y ) at 1.
  econstructor; eauto. rewrite concat_empty_r. eauto.
Qed.
Lemma  trf_ewkn_obj_push' (A AT: Type) G Go e (t: A) (T: AT) w y:
  ctx_ok (G§§& Go )->
  ok e ->
  w # e ->
  〈 G §§ Go ⁏ e ⊢ t  ∶ T 〉 ∼ @TransformObjEWeakening AT A ∼> 〈 G  §§ Go ⁏ e & w ~ y ⊢ t ∶ T 〉.
Proof.
  introv HokG Hoke Hne.
  rewrite <- concat_empty_r with (E:=e) at 1.
  rewrite <- concat_empty_r with (E:=e & w ~ y ) at 1.
  econstructor; eauto. rewrite concat_empty_r. eauto.
Qed.

Lemma trf_local_idtyp_trm_ewkn_push
  (TL: TermTransform)
  (TLE: TransformEWeakeningCompat typ trm TL)
  G e ta tb T w y:
  〈 G⁏ e ⊢ ta ∶ T 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G ⁏ e ⊢ tb ∶ T 〉 ->
  ctx_ok G ->
  ok e ->
  y # e ->
  〈 G⁏ e & y ~ w  ⊢ ta ∶ T 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G ⁏ e & y ~ w  ⊢ tb ∶ T 〉.
Proof.
  introv Htrf HokG Hoke Hnye.
  eapply EWeakeningCompatTransformationTrm. auto.

  eapply Htrf.
  eapply trf_ewkn_push' with (w:=y) (y:=w); eauto.
  eapply trf_ewkn_push' with (w:=y) (y:=w); eauto.
  econstructor; auto.
Qed.
Lemma trf_local_idtyp_defs_ewkn_push
  (TL: TermTransform)
  (TLE: TransformEWeakeningCompat typ trm TL)
  G Go e ta tb T w y:
  〈 G §§ Go ⁏ e ⊢ ta ∶ T 〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈 G §§ Go  ⁏ e ⊢ tb ∶ T 〉 ->
  ctx_ok (G §§& Go) ->
  ok e ->
  y # e ->
  〈 G §§ Go ⁏ e & y ~ w  ⊢ ta ∶ T 〉 ∼ trf_local_idtyp_defs typing_mode TL ∼> 〈 G §§ Go  ⁏ e & y ~ w  ⊢ tb ∶ T 〉.
Proof.
  introv Htrf HokG Hoke Hnye.
  eapply EWeakeningCompatTransformationDefs. auto.

  eapply Htrf.
  eapply trf_ewkn_obj_push' with (w:=y) (y:=w); eauto.
  eapply trf_ewkn_obj_push' with (w:=y) (y:=w); eauto.
  econstructor; auto.
Qed.

Lemma trf_local_idtyp_trm_subst_w
  (TL: TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  G e ta tb w T Tt L vk:
  ctx_ok G ->
  ok e ->
  G ⁏ e ⊢ trm_var (avar_f w) ∶ Tt ->
  (
    forall x : var,
    x \notin L ->
    〈 G § x / vk ~ Tt ⁏ e ⊢ ta /[-> x] ∶ T 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G § x / vk  ~ Tt ⁏ e ⊢ tb /[-> x] ∶ T 〉
  ) ->
  〈 G ⁏ e ⊢ ta /[-> w] ∶ T 〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G ⁏ e ⊢ tb /[-> w] ∶ T 〉.
Proof.
  introv HokG Hoke Htyw Htyt.
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr.
  fresh_chain_pick Fr y.

  eapply SubstCompatTransformationTrm.
  auto.

  eapply trf_local_idtyp_trm_ewkn_push with (w:=w) (y:=y). auto.

  eapply (Htyt y). fresh_chain_solve Fr.
  fresh_chain_solve Fr. fresh_chain_solve Fr. fresh_chain_solve Fr.
  
  eapply trf_subst_w; eauto. fresh_chain_solve Fr.
  eapply trf_subst_w; eauto. fresh_chain_solve Fr.
  econstructor; auto.
Qed.

Lemma trf_local_idtyp_trm_subst_wz
  (TL: TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  G e ta tb w T Tt L vk:
  ctx_ok G ->
  ok e ->
  G ⁏ e ⊢ trm_var (avar_f w) ∶ Tt ->
  (
    forall x : var,
    x \notin L ->
    〈 G § x / vk ~ Tt ⁏ e ⊢ ta /[-> x] ∶ T  /[-> x]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G § x / vk  ~ Tt ⁏ e ⊢ tb /[-> x] ∶ T  /[-> x]〉
  ) ->
  〈 G ⁏ e ⊢ ta /[-> w] ∶ T /[-> w]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G ⁏ e ⊢ tb /[-> w] ∶ T /[-> w] 〉.
Proof.
  introv HokG Hoke Htyw Htyt.
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr.
  fresh_chain_pick Fr y.

  eapply SubstCompatTransformationTrm.
  auto.

  eapply trf_local_idtyp_trm_ewkn_push with (w:=w) (y:=y). auto.

  eapply (Htyt y). fresh_chain_solve Fr.
  fresh_chain_solve Fr. fresh_chain_solve Fr. fresh_chain_solve Fr.
 
  eapply trf_subst_wz; eauto. fresh_chain_solve Fr.
  eapply trf_subst_wz; eauto. fresh_chain_solve Fr.
  econstructor; auto.
Qed.
Lemma trf_local_idtyp_trm_subst_wz0
  (TL: TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  G e ta tb w T Tt vk x:
  ctx_ok G ->
  ok e ->
  x \notin (ctx_dom G \u ctx_fv G \u fv_renv e \u fv Tt \u fv ta \u fv tb \u fv T) ->
  G ⁏ e ⊢ trm_var (avar_f w) ∶ Tt ->
  (
    〈 G § x / vk ~ Tt ⁏ e ⊢ ta /[-> x] ∶ T  /[-> x]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G § x / vk  ~ Tt ⁏ e ⊢ tb /[-> x] ∶ T  /[-> x]〉
  ) ->
  〈 G ⁏ e ⊢ ta /[-> w] ∶ T /[-> w]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G ⁏ e ⊢ tb /[-> w] ∶ T /[-> w] 〉.
Proof.
  introv HokG Hoke Hnx Htyw Htyt.
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx.
  fresh_chain_cleanup Fr.

  eapply SubstCompatTransformationTrm.
  auto.

  eapply trf_local_idtyp_trm_ewkn_push with (w:=w) (y:=x). auto.

  eapply (Htyt ). fresh_chain_solve Fr.
  fresh_chain_solve Fr. fresh_chain_solve Fr.

  eapply trf_subst_wz0; eauto. fresh_chain_solve Fr.
  eapply trf_subst_wz0; eauto. fresh_chain_solve Fr.
  econstructor; auto.
Qed.
Lemma trf_local_idtyp_trm_subst_wz1
  (TL: TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  G e ta tb w T Tt vk x:
  ctx_ok G ->
  ok e ->
  x \notin (ctx_dom G \u ctx_fv G \u fv_renv e  \u fv ta \u fv tb \u fv T) ->
  G ⁏ e ⊢ trm_var (avar_f w) ∶ Tt/[ x -> w]  ->
  (
    〈 G § x / vk ~ Tt ⁏ e ⊢ ta /[#1-> x] ∶ T  /[#1-> x]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G § x / vk  ~ Tt ⁏ e ⊢ tb /[#1-> x] ∶ T  /[#1-> x]〉
  ) ->
  〈 G ⁏ e ⊢ ta /[#1-> w] ∶ T /[#1-> w]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G ⁏ e ⊢ tb /[#1-> w] ∶ T /[#1-> w] 〉.
Proof.
  introv HokG Hoke Hnx Htyw Htyt.
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx.
  fresh_chain_cleanup Fr.

  eapply SubstCompatTransformationTrm.
  auto.

  eapply trf_local_idtyp_trm_ewkn_push with (w:=w) (y:=x). auto.

  eapply (Htyt ). fresh_chain_solve Fr.
  fresh_chain_solve Fr. fresh_chain_solve Fr.

  
  eapply trf_subst_wz1; eauto. fresh_chain_solve Fr.
  eapply trf_subst_wz1; eauto. fresh_chain_solve Fr.
  econstructor; auto.
Qed.

Lemma trf_local_idtyp_trm_subst_wz1'
  (TL: TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  G e ta tb w T Tt vk x x' vk' T':
  ctx_ok G ->
  x' \notin ctx_dom G ->
  x' <> x ->
  ok e ->
  x \notin (ctx_dom G \u ctx_fv G \u fv_renv e  \u fv ta \u fv tb \u fv T) ->
  G § x' / vk' ~ T' /[ x -> w] ⁏ e ⊢ trm_var (avar_f w) ∶ Tt/[ x -> w]  ->
  (
    〈 G § x / vk ~ Tt § x' / vk' ~ T'⁏ e ⊢ ta /[#1-> x] ∶ T  /[#1-> x]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G § x / vk  ~ Tt § x' / vk' ~ T' ⁏ e ⊢ tb /[#1-> x] ∶ T  /[#1-> x]〉
  ) ->
  〈 G§ x' / vk' ~ T' /[ x -> w] ⁏ e ⊢ ta /[#1-> w] ∶ T /[#1-> w]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G § x' / vk' ~ T' /[ x -> w]⁏ e ⊢ tb /[#1-> w] ∶ T /[#1-> w] 〉.
Proof.
  introv HokG HnxG Hnxx Hoke Hnx Htyw Htyt.
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx.
  fresh_chain_cleanup Fr.

  eapply SubstCompatTransformationTrm.
  auto.

  eapply trf_local_idtyp_trm_ewkn_push with (w:=w) (y:=x). auto.

  eapply (Htyt ). econstructor. fresh_chain_solve Fr. eauto.
  fresh_chain_solve Fr. fresh_chain_solve Fr.
  
  eapply trf_subst_wz1'; eauto. fresh_chain_solve Fr.
  eapply trf_subst_wz1'; eauto. fresh_chain_solve Fr.
  econstructor; auto.
Qed.

Lemma trf_local_idtyp_trm_subst_wz0'
  (TL: TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  G e ta tb w T Tt vk x x' vk' T':
  ctx_ok G ->
  x' \notin ctx_dom G ->
  x' <> x ->
  ok e ->
  x \notin (ctx_dom G \u ctx_fv G \u fv_renv e  \u fv ta \u fv tb \u fv T) ->
  G § x' / vk' ~ T' /[ x -> w] ⁏ e ⊢ trm_var (avar_f w) ∶ Tt/[ x -> w]  ->
  (
    〈 G § x / vk ~ Tt § x' / vk' ~ T'⁏ e ⊢ ta /[-> x] ∶ T  /[-> x]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G § x / vk  ~ Tt § x' / vk' ~ T' ⁏ e ⊢ tb /[-> x] ∶ T  /[-> x]〉
  ) ->
  〈 G§ x' / vk' ~ T' /[ x -> w] ⁏ e ⊢ ta /[-> w] ∶ T /[-> w]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G § x' / vk' ~ T' /[ x -> w]⁏ e ⊢ tb /[-> w] ∶ T /[-> w] 〉.
Proof.
  introv HokG HnxG Hnxx Hoke Hnx Htyw Htyt.
  fresh_chain_init Fr.
  fresh_chain_add_fresh Fr Hnx.
  fresh_chain_cleanup Fr.

  eapply SubstCompatTransformationTrm.
  auto.

  eapply trf_local_idtyp_trm_ewkn_push with (w:=w) (y:=x). auto.

  eapply (Htyt ). econstructor. fresh_chain_solve Fr. eauto.
  fresh_chain_solve Fr. fresh_chain_solve Fr.

    eapply trf_subst_wz0'; eauto. fresh_chain_solve Fr.
  eapply trf_subst_wz0'; eauto. fresh_chain_solve Fr.
  econstructor; auto.
Qed.

Lemma trf_local_idtyp_trm_subst_met
  (TL: TermTransform)
  (TLS: TransformSubstCompat typing_mode typ trm TL)
  (TLE: TransformEWeakeningCompat typ trm TL)
  (TLW: TransformWeakeningCompat typ trm TL)
  (TLU: TransformUnhidingCompat typ trm TL)
  G e ta tb wc wz T TC TZ L yc Ty Sw Sy:
  inert G ->
  ok e ->
  G ⁏ e ⊢ trm_var (avar_f wz) ∶ TZ ->
  G ⁏ e ⊢ trm_var (avar_f wc) ∶ TC /[-> wz]  ->
  (
        forall xz : var,
       xz \notin L ->
       forall xc : var,
       xc <> xz ->
       xc \notin L ->
       〈 ctx_met_refined_heap (vis_hide typing_mode G) xz TZ xc TC yc Ty ⁏ e ⊢ ta  /[# 1 -> xc]/[-> xz]∶ T /[# 1 -> xc]/[-> xz]〉
        ∼ trf_local_idtyp_trm typing_mode TL ∼>
       〈 ctx_met_refined_heap (vis_hide typing_mode G) xz TZ xc TC yc Ty ⁏ e ⊢ tb  /[# 1 -> xc]/[-> xz] ∶ T /[# 1 -> xc]/[-> xz]〉
  ) ->
   ctx_binds wc (typ_and (typ_bnd Ty) Sw) G ->
   ctx_binds yc (typ_and (typ_bnd Ty) Sy) G ->
   binds wc yc e ->
  〈 G ⁏ e ⊢ ta /[# 1 -> wc]/[-> wz] ∶ T  /[# 1 -> wc]/[-> wz]〉 ∼ trf_local_idtyp_trm typing_mode TL ∼> 〈 G ⁏ e ⊢ tb  /[# 1 -> wc]/[-> wz] ∶ T /[# 1 -> wc]/[-> wz]〉.
Proof.
  introv HinG Hoke Htywz Htywc Htyt HbiwcG HbiycG Hbie.
  forwards HokG: inert_ok HinG.
  fresh_chain_init Fr.
  fresh_chain_add_gather Fr.
  fresh_chain_pick Fr xz.
  fresh_chain_pick Fr xc.

  forwards Htyty: Htyt xz xc. fresh_chain_solve Fr. fresh_chain_solve Fr. fresh_chain_solve Fr.
  clear Htyt.

  assert (G §& (ctx_empty §ʳ xc ~ typ_and TC /[-> xz] (typ_and Ty /[-> wc] Ty /[-> xc])) /[§ xz -> wz] ⁏ e ⊢ wz ∶ TZ /[ xz -> wz]) as Htywzz. {
    eapply weaken_end; auto.
    rewrite subst_fresh; auto.
    eapply ty_trm_var_inv in Htywz. auto.
    fresh_chain_solve Fr.
    ctx_flatten. eauto.
    ctx_flatten. fresh_chain_solve Fr.
  }
      assert( (((typ_and (open xz TC) (typ_and (open yc Ty) (open xc Ty)))/[ xz -> wz]) = (typ_and (open wz TC) (typ_and (open yc Ty) (open xc Ty))))) as HeqTC.
      eapply typing_call_met_recvtyp; try fresh_chain_solve Fr; eauto.

      assert (G ⁏ e ⊢ wc ∶ (typ_and (open wz TC) (typ_and (open wc Ty) (open wc Ty)))) as Htywcc.
      {
        eapply ty_trm_var_inv in Htywc.
        eapply typing_call_met_recvty; eauto.
      }

    assert (
      〈 ctx_met_refined_heap G xz TZ xc TC  yc Ty ⁏ e ⊢ ta /[# 1 -> xc] /[-> xz] ∶ T /[# 1 -> xc] /[-> xz] 〉 ∼ trf_local_idtyp_trm typing_mode TL
      ∼> 〈 ctx_met_refined_heap G xz TZ  xc TC yc Ty ⁏ e ⊢ tb /[# 1 -> xc] /[-> xz] ∶ T /[# 1 -> xc] /[-> xz] 〉
    ) as Htrf. {
      ctx_split_from xz in Htyty.
      ctx_split_from xz.
      destruct (mode_visibility' typing_mode) eqn: Hmv; auto.
      eapply UnhidingCompatTransformationTrm. eauto. eapply Htyty. constructor; eauto. constructor; eauto. hnf; eauto.
    }
    clear Htyty.
    rewrite ! open_commut with (m:=0)(n:=1).

    eapply trf_local_idtyp_trm_subst_wz1 with (x:=xc) (Tt := typ_and TC/[->wz] (typ_and Ty/[->yc] Ty/[->xc])) (vk:=v_recv).
    auto. auto.  fresh_chain_solve Fr. fresh_chain_solve Fr. sympl; notin_solve; try eapply notin_fv_open; try fresh_chain_solve Fr.
    eapply ty_trm_var. eapply ty_trm_var_inv_vis; eauto.
    eauto.
    sympls.

    rewrite subst_fresh.

    rewrite subst_fresh.
    rewrite <- subst_intro with (x:=xc) (y:=wc).

    eapply ty_var_sub.
    eapply Htywcc.
    eapply subtyp_tyq.
    constructor. auto. eauto.
    constructor. auto.

    eapply tyq_typ_open_indep_same.
    eapply inert_bnd_indep. eauto. eauto.
    eauto.
    eauto.
    try eapply notin_fv_open; try fresh_chain_solve Fr.
    try eapply notin_fv_open; try fresh_chain_solve Fr.
    try eapply notin_fv_open; try fresh_chain_solve Fr.


    rewrite ! open_commut with (m:=1)(n:=0).

    rewrite <- HeqTC.

    eapply trf_local_idtyp_trm_subst_wz0' with (Tt := TZ) (vk:=v_par) (x:=xz).
    auto. auto.  fresh_chain_solve Fr. fresh_chain_solve Fr.
    fresh_chain_solve Fr. fresh_chain_solve Fr.
    sympl; notin_solve; try eapply notin_fv_open; try fresh_chain_solve Fr.


    eapply weaken_push.
    rewrite subst_fresh.
    eauto.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.
    fresh_chain_solve Fr.

    eapply Htrf.
    eauto.

    all: eauto.


Qed.


Lemma trf_ewkn_push G e (t: trm) (T: typ) w y:
  ctx_ok G ->
  ok ( e & w ~ y ) ->
  〈 G ⁏ e ⊢ t ∶ T〉 ∼ TransformEWeakening (A:=trm) ∼>  〈 G ⁏ e & w ~ y ⊢ t ∶ T〉.
Proof.
  introv HokG Hoke.
  rewrite <- concat_empty_r with (E:=e) at 1.
  rewrite <- concat_empty_r with (E:=(e & w ~ y )) at 1.
  econstructor; try rewrite concat_empty_r; eauto.
Qed.

End WithTypingMode.